(function () {
  'use strict';

  // ::- Persistent data structure representing an ordered mapping from
  // strings to values, with some convenient update methods.
  function OrderedMap(content) {
    this.content = content;
  }

  OrderedMap.prototype = {
    constructor: OrderedMap,

    find: function(key) {
      for (var i = 0; i < this.content.length; i += 2)
        { if (this.content[i] === key) { return i } }
      return -1
    },

    // :: (string) → ?any
    // Retrieve the value stored under `key`, or return undefined when
    // no such key exists.
    get: function(key) {
      var found = this.find(key);
      return found == -1 ? undefined : this.content[found + 1]
    },

    // :: (string, any, ?string) → OrderedMap
    // Create a new map by replacing the value of `key` with a new
    // value, or adding a binding to the end of the map. If `newKey` is
    // given, the key of the binding will be replaced with that key.
    update: function(key, value, newKey) {
      var self = newKey && newKey != key ? this.remove(newKey) : this;
      var found = self.find(key), content = self.content.slice();
      if (found == -1) {
        content.push(newKey || key, value);
      } else {
        content[found + 1] = value;
        if (newKey) { content[found] = newKey; }
      }
      return new OrderedMap(content)
    },

    // :: (string) → OrderedMap
    // Return a map with the given key removed, if it existed.
    remove: function(key) {
      var found = this.find(key);
      if (found == -1) { return this }
      var content = this.content.slice();
      content.splice(found, 2);
      return new OrderedMap(content)
    },

    // :: (string, any) → OrderedMap
    // Add a new key to the start of the map.
    addToStart: function(key, value) {
      return new OrderedMap([key, value].concat(this.remove(key).content))
    },

    // :: (string, any) → OrderedMap
    // Add a new key to the end of the map.
    addToEnd: function(key, value) {
      var content = this.remove(key).content.slice();
      content.push(key, value);
      return new OrderedMap(content)
    },

    // :: (string, string, any) → OrderedMap
    // Add a key after the given key. If `place` is not found, the new
    // key is added to the end.
    addBefore: function(place, key, value) {
      var without = this.remove(key), content = without.content.slice();
      var found = without.find(place);
      content.splice(found == -1 ? content.length : found, 0, key, value);
      return new OrderedMap(content)
    },

    // :: ((key: string, value: any))
    // Call the given function for each key/value pair in the map, in
    // order.
    forEach: function(f) {
      for (var i = 0; i < this.content.length; i += 2)
        { f(this.content[i], this.content[i + 1]); }
    },

    // :: (union<Object, OrderedMap>) → OrderedMap
    // Create a new map by prepending the keys in this map that don't
    // appear in `map` before the keys in `map`.
    prepend: function(map) {
      map = OrderedMap.from(map);
      if (!map.size) { return this }
      return new OrderedMap(map.content.concat(this.subtract(map).content))
    },

    // :: (union<Object, OrderedMap>) → OrderedMap
    // Create a new map by appending the keys in this map that don't
    // appear in `map` after the keys in `map`.
    append: function(map) {
      map = OrderedMap.from(map);
      if (!map.size) { return this }
      return new OrderedMap(this.subtract(map).content.concat(map.content))
    },

    // :: (union<Object, OrderedMap>) → OrderedMap
    // Create a map containing all the keys in this map that don't
    // appear in `map`.
    subtract: function(map) {
      var result = this;
      map = OrderedMap.from(map);
      for (var i = 0; i < map.content.length; i += 2)
        { result = result.remove(map.content[i]); }
      return result
    },

    // :: number
    // The amount of keys in this map.
    get size() {
      return this.content.length >> 1
    }
  };

  // :: (?union<Object, OrderedMap>) → OrderedMap
  // Return a map with the given content. If null, create an empty
  // map. If given an ordered map, return that map itself. If given an
  // object, create a map from the object's properties.
  OrderedMap.from = function(value) {
    if (value instanceof OrderedMap) { return value }
    var content = [];
    if (value) { for (var prop in value) { content.push(prop, value[prop]); } }
    return new OrderedMap(content)
  };

  var orderedmap = OrderedMap;

  function findDiffStart$1(a, b, pos) {
    for (var i = 0;; i++) {
      if (i == a.childCount || i == b.childCount)
        { return a.childCount == b.childCount ? null : pos }

      var childA = a.child(i), childB = b.child(i);
      if (childA == childB) { pos += childA.nodeSize; continue }

      if (!childA.sameMarkup(childB)) { return pos }

      if (childA.isText && childA.text != childB.text) {
        for (var j = 0; childA.text[j] == childB.text[j]; j++)
          { pos++; }
        return pos
      }
      if (childA.content.size || childB.content.size) {
        var inner = findDiffStart$1(childA.content, childB.content, pos + 1);
        if (inner != null) { return inner }
      }
      pos += childA.nodeSize;
    }
  }

  function findDiffEnd$1(a, b, posA, posB) {
    for (var iA = a.childCount, iB = b.childCount;;) {
      if (iA == 0 || iB == 0)
        { return iA == iB ? null : {a: posA, b: posB} }

      var childA = a.child(--iA), childB = b.child(--iB), size = childA.nodeSize;
      if (childA == childB) {
        posA -= size; posB -= size;
        continue
      }

      if (!childA.sameMarkup(childB)) { return {a: posA, b: posB} }

      if (childA.isText && childA.text != childB.text) {
        var same = 0, minSize = Math.min(childA.text.length, childB.text.length);
        while (same < minSize && childA.text[childA.text.length - same - 1] == childB.text[childB.text.length - same - 1]) {
          same++; posA--; posB--;
        }
        return {a: posA, b: posB}
      }
      if (childA.content.size || childB.content.size) {
        var inner = findDiffEnd$1(childA.content, childB.content, posA - 1, posB - 1);
        if (inner) { return inner }
      }
      posA -= size; posB -= size;
    }
  }

  // ::- A fragment represents a node's collection of child nodes.
  //
  // Like nodes, fragments are persistent data structures, and you
  // should not mutate them or their content. Rather, you create new
  // instances whenever needed. The API tries to make this easy.
  var Fragment$1 = function Fragment(content, size) {
    this.content = content;
    // :: number
    // The size of the fragment, which is the total of the size of its
    // content nodes.
    this.size = size || 0;
    if (size == null) { for (var i = 0; i < content.length; i++)
      { this.size += content[i].nodeSize; } }
  };

  var prototypeAccessors$7 = { firstChild: { configurable: true },lastChild: { configurable: true },childCount: { configurable: true } };

  // :: (number, number, (node: Node, start: number, parent: Node, index: number) → ?bool, ?number)
  // Invoke a callback for all descendant nodes between the given two
  // positions (relative to start of this fragment). Doesn't descend
  // into a node when the callback returns `false`.
  Fragment$1.prototype.nodesBetween = function nodesBetween (from, to, f, nodeStart, parent) {
      if ( nodeStart === void 0 ) { nodeStart = 0; }

    for (var i = 0, pos = 0; pos < to; i++) {
      var child = this.content[i], end = pos + child.nodeSize;
      if (end > from && f(child, nodeStart + pos, parent, i) !== false && child.content.size) {
        var start = pos + 1;
        child.nodesBetween(Math.max(0, from - start),
                           Math.min(child.content.size, to - start),
                           f, nodeStart + start);
      }
      pos = end;
    }
  };

  // :: ((node: Node, pos: number, parent: Node) → ?bool)
  // Call the given callback for every descendant node. The callback
  // may return `false` to prevent traversal of a given node's children.
  Fragment$1.prototype.descendants = function descendants (f) {
    this.nodesBetween(0, this.size, f);
  };

  // :: (number, number, ?string, ?string) → string
  // Extract the text between `from` and `to`. See the same method on
  // [`Node`](#model.Node.textBetween).
  Fragment$1.prototype.textBetween = function textBetween (from, to, blockSeparator, leafText) {
    var text = "", separated = true;
    this.nodesBetween(from, to, function (node, pos) {
      if (node.isText) {
        text += node.text.slice(Math.max(from, pos) - pos, to - pos);
        separated = !blockSeparator;
      } else if (node.isLeaf && leafText) {
        text += leafText;
        separated = !blockSeparator;
      } else if (!separated && node.isBlock) {
        text += blockSeparator;
        separated = true;
      }
    }, 0);
    return text
  };

  // :: (Fragment) → Fragment
  // Create a new fragment containing the combined content of this
  // fragment and the other.
  Fragment$1.prototype.append = function append (other) {
    if (!other.size) { return this }
    if (!this.size) { return other }
    var last = this.lastChild, first = other.firstChild, content = this.content.slice(), i = 0;
    if (last.isText && last.sameMarkup(first)) {
      content[content.length - 1] = last.withText(last.text + first.text);
      i = 1;
    }
    for (; i < other.content.length; i++) { content.push(other.content[i]); }
    return new Fragment$1(content, this.size + other.size)
  };

  // :: (number, ?number) → Fragment
  // Cut out the sub-fragment between the two given positions.
  Fragment$1.prototype.cut = function cut (from, to) {
    if (to == null) { to = this.size; }
    if (from == 0 && to == this.size) { return this }
    var result = [], size = 0;
    if (to > from) { for (var i = 0, pos = 0; pos < to; i++) {
      var child = this.content[i], end = pos + child.nodeSize;
      if (end > from) {
        if (pos < from || end > to) {
          if (child.isText)
            { child = child.cut(Math.max(0, from - pos), Math.min(child.text.length, to - pos)); }
          else
            { child = child.cut(Math.max(0, from - pos - 1), Math.min(child.content.size, to - pos - 1)); }
        }
        result.push(child);
        size += child.nodeSize;
      }
      pos = end;
    } }
    return new Fragment$1(result, size)
  };

  Fragment$1.prototype.cutByIndex = function cutByIndex (from, to) {
    if (from == to) { return Fragment$1.empty }
    if (from == 0 && to == this.content.length) { return this }
    return new Fragment$1(this.content.slice(from, to))
  };

  // :: (number, Node) → Fragment
  // Create a new fragment in which the node at the given index is
  // replaced by the given node.
  Fragment$1.prototype.replaceChild = function replaceChild (index, node) {
    var current = this.content[index];
    if (current == node) { return this }
    var copy = this.content.slice();
    var size = this.size + node.nodeSize - current.nodeSize;
    copy[index] = node;
    return new Fragment$1(copy, size)
  };

  // : (Node) → Fragment
  // Create a new fragment by prepending the given node to this
  // fragment.
  Fragment$1.prototype.addToStart = function addToStart (node) {
    return new Fragment$1([node].concat(this.content), this.size + node.nodeSize)
  };

  // : (Node) → Fragment
  // Create a new fragment by appending the given node to this
  // fragment.
  Fragment$1.prototype.addToEnd = function addToEnd (node) {
    return new Fragment$1(this.content.concat(node), this.size + node.nodeSize)
  };

  // :: (Fragment) → bool
  // Compare this fragment to another one.
  Fragment$1.prototype.eq = function eq (other) {
    if (this.content.length != other.content.length) { return false }
    for (var i = 0; i < this.content.length; i++)
      { if (!this.content[i].eq(other.content[i])) { return false } }
    return true
  };

  // :: ?Node
  // The first child of the fragment, or `null` if it is empty.
  prototypeAccessors$7.firstChild.get = function () { return this.content.length ? this.content[0] : null };

  // :: ?Node
  // The last child of the fragment, or `null` if it is empty.
  prototypeAccessors$7.lastChild.get = function () { return this.content.length ? this.content[this.content.length - 1] : null };

  // :: number
  // The number of child nodes in this fragment.
  prototypeAccessors$7.childCount.get = function () { return this.content.length };

  // :: (number) → Node
  // Get the child node at the given index. Raise an error when the
  // index is out of range.
  Fragment$1.prototype.child = function child (index) {
    var found = this.content[index];
    if (!found) { throw new RangeError("Index " + index + " out of range for " + this) }
    return found
  };

  // :: (number) → ?Node
  // Get the child node at the given index, if it exists.
  Fragment$1.prototype.maybeChild = function maybeChild (index) {
    return this.content[index]
  };

  // :: ((node: Node, offset: number, index: number))
  // Call `f` for every child node, passing the node, its offset
  // into this parent node, and its index.
  Fragment$1.prototype.forEach = function forEach (f) {
    for (var i = 0, p = 0; i < this.content.length; i++) {
      var child = this.content[i];
      f(child, p, i);
      p += child.nodeSize;
    }
  };

  // :: (Fragment) → ?number
  // Find the first position at which this fragment and another
  // fragment differ, or `null` if they are the same.
  Fragment$1.prototype.findDiffStart = function findDiffStart$1$1 (other, pos) {
      if ( pos === void 0 ) { pos = 0; }

    return findDiffStart$1(this, other, pos)
  };

  // :: (Fragment) → ?{a: number, b: number}
  // Find the first position, searching from the end, at which this
  // fragment and the given fragment differ, or `null` if they are the
  // same. Since this position will not be the same in both nodes, an
  // object with two separate positions is returned.
  Fragment$1.prototype.findDiffEnd = function findDiffEnd$1$1 (other, pos, otherPos) {
      if ( pos === void 0 ) { pos = this.size; }
      if ( otherPos === void 0 ) { otherPos = other.size; }

    return findDiffEnd$1(this, other, pos, otherPos)
  };

  // : (number, ?number) → {index: number, offset: number}
  // Find the index and inner offset corresponding to a given relative
  // position in this fragment. The result object will be reused
  // (overwritten) the next time the function is called. (Not public.)
  Fragment$1.prototype.findIndex = function findIndex (pos, round) {
      if ( round === void 0 ) { round = -1; }

    if (pos == 0) { return retIndex$1(0, pos) }
    if (pos == this.size) { return retIndex$1(this.content.length, pos) }
    if (pos > this.size || pos < 0) { throw new RangeError(("Position " + pos + " outside of fragment (" + (this) + ")")) }
    for (var i = 0, curPos = 0;; i++) {
      var cur = this.child(i), end = curPos + cur.nodeSize;
      if (end >= pos) {
        if (end == pos || round > 0) { return retIndex$1(i + 1, end) }
        return retIndex$1(i, curPos)
      }
      curPos = end;
    }
  };

  // :: () → string
  // Return a debugging string that describes this fragment.
  Fragment$1.prototype.toString = function toString () { return "<" + this.toStringInner() + ">" };

  Fragment$1.prototype.toStringInner = function toStringInner () { return this.content.join(", ") };

  // :: () → ?Object
  // Create a JSON-serializeable representation of this fragment.
  Fragment$1.prototype.toJSON = function toJSON () {
    return this.content.length ? this.content.map(function (n) { return n.toJSON(); }) : null
  };

  // :: (Schema, ?Object) → Fragment
  // Deserialize a fragment from its JSON representation.
  Fragment$1.fromJSON = function fromJSON (schema, value) {
    if (!value) { return Fragment$1.empty }
    if (!Array.isArray(value)) { throw new RangeError("Invalid input for Fragment.fromJSON") }
    return new Fragment$1(value.map(schema.nodeFromJSON))
  };

  // :: ([Node]) → Fragment
  // Build a fragment from an array of nodes. Ensures that adjacent
  // text nodes with the same marks are joined together.
  Fragment$1.fromArray = function fromArray (array) {
    if (!array.length) { return Fragment$1.empty }
    var joined, size = 0;
    for (var i = 0; i < array.length; i++) {
      var node = array[i];
      size += node.nodeSize;
      if (i && node.isText && array[i - 1].sameMarkup(node)) {
        if (!joined) { joined = array.slice(0, i); }
        joined[joined.length - 1] = node.withText(joined[joined.length - 1].text + node.text);
      } else if (joined) {
        joined.push(node);
      }
    }
    return new Fragment$1(joined || array, size)
  };

  // :: (?union<Fragment, Node, [Node]>) → Fragment
  // Create a fragment from something that can be interpreted as a set
  // of nodes. For `null`, it returns the empty fragment. For a
  // fragment, the fragment itself. For a node or array of nodes, a
  // fragment containing those nodes.
  Fragment$1.from = function from (nodes) {
    if (!nodes) { return Fragment$1.empty }
    if (nodes instanceof Fragment$1) { return nodes }
    if (Array.isArray(nodes)) { return this.fromArray(nodes) }
    if (nodes.attrs) { return new Fragment$1([nodes], nodes.nodeSize) }
    throw new RangeError("Can not convert " + nodes + " to a Fragment" +
                         (nodes.nodesBetween ? " (looks like multiple versions of prosemirror-model were loaded)" : ""))
  };

  Object.defineProperties( Fragment$1.prototype, prototypeAccessors$7 );

  var found$1 = {index: 0, offset: 0};
  function retIndex$1(index, offset) {
    found$1.index = index;
    found$1.offset = offset;
    return found$1
  }

  // :: Fragment
  // An empty fragment. Intended to be reused whenever a node doesn't
  // contain anything (rather than allocating a new empty fragment for
  // each leaf node).
  Fragment$1.empty = new Fragment$1([], 0);

  function compareDeep$1(a, b) {
    if (a === b) { return true }
    if (!(a && typeof a == "object") ||
        !(b && typeof b == "object")) { return false }
    var array = Array.isArray(a);
    if (Array.isArray(b) != array) { return false }
    if (array) {
      if (a.length != b.length) { return false }
      for (var i = 0; i < a.length; i++) { if (!compareDeep$1(a[i], b[i])) { return false } }
    } else {
      for (var p in a) { if (!(p in b) || !compareDeep$1(a[p], b[p])) { return false } }
      for (var p$1 in b) { if (!(p$1 in a)) { return false } }
    }
    return true
  }

  // ::- A mark is a piece of information that can be attached to a node,
  // such as it being emphasized, in code font, or a link. It has a type
  // and optionally a set of attributes that provide further information
  // (such as the target of the link). Marks are created through a
  // `Schema`, which controls which types exist and which
  // attributes they have.
  var Mark$1 = function Mark(type, attrs) {
    // :: MarkType
    // The type of this mark.
    this.type = type;
    // :: Object
    // The attributes associated with this mark.
    this.attrs = attrs;
  };

  // :: ([Mark]) → [Mark]
  // Given a set of marks, create a new set which contains this one as
  // well, in the right position. If this mark is already in the set,
  // the set itself is returned. If any marks that are set to be
  // [exclusive](#model.MarkSpec.excludes) with this mark are present,
  // those are replaced by this one.
  Mark$1.prototype.addToSet = function addToSet (set) {
    var copy, placed = false;
    for (var i = 0; i < set.length; i++) {
      var other = set[i];
      if (this.eq(other)) { return set }
      if (this.type.excludes(other.type)) {
        if (!copy) { copy = set.slice(0, i); }
      } else if (other.type.excludes(this.type)) {
        return set
      } else {
        if (!placed && other.type.rank > this.type.rank) {
          if (!copy) { copy = set.slice(0, i); }
          copy.push(this);
          placed = true;
        }
        if (copy) { copy.push(other); }
      }
    }
    if (!copy) { copy = set.slice(); }
    if (!placed) { copy.push(this); }
    return copy
  };

  // :: ([Mark]) → [Mark]
  // Remove this mark from the given set, returning a new set. If this
  // mark is not in the set, the set itself is returned.
  Mark$1.prototype.removeFromSet = function removeFromSet (set) {
    for (var i = 0; i < set.length; i++)
      { if (this.eq(set[i]))
        { return set.slice(0, i).concat(set.slice(i + 1)) } }
    return set
  };

  // :: ([Mark]) → bool
  // Test whether this mark is in the given set of marks.
  Mark$1.prototype.isInSet = function isInSet (set) {
    for (var i = 0; i < set.length; i++)
      { if (this.eq(set[i])) { return true } }
    return false
  };

  // :: (Mark) → bool
  // Test whether this mark has the same type and attributes as
  // another mark.
  Mark$1.prototype.eq = function eq (other) {
    return this == other ||
      (this.type == other.type && compareDeep$1(this.attrs, other.attrs))
  };

  // :: () → Object
  // Convert this mark to a JSON-serializeable representation.
  Mark$1.prototype.toJSON = function toJSON () {
    var obj = {type: this.type.name};
    for (var _ in this.attrs) {
      obj.attrs = this.attrs;
      break
    }
    return obj
  };

  // :: (Schema, Object) → Mark
  Mark$1.fromJSON = function fromJSON (schema, json) {
    if (!json) { throw new RangeError("Invalid input for Mark.fromJSON") }
    var type = schema.marks[json.type];
    if (!type) { throw new RangeError(("There is no mark type " + (json.type) + " in this schema")) }
    return type.create(json.attrs)
  };

  // :: ([Mark], [Mark]) → bool
  // Test whether two sets of marks are identical.
  Mark$1.sameSet = function sameSet (a, b) {
    if (a == b) { return true }
    if (a.length != b.length) { return false }
    for (var i = 0; i < a.length; i++)
      { if (!a[i].eq(b[i])) { return false } }
    return true
  };

  // :: (?union<Mark, [Mark]>) → [Mark]
  // Create a properly sorted mark set from null, a single mark, or an
  // unsorted array of marks.
  Mark$1.setFrom = function setFrom (marks) {
    if (!marks || marks.length == 0) { return Mark$1.none }
    if (marks instanceof Mark$1) { return [marks] }
    var copy = marks.slice();
    copy.sort(function (a, b) { return a.type.rank - b.type.rank; });
    return copy
  };

  // :: [Mark] The empty set of marks.
  Mark$1.none = [];

  // ReplaceError:: class extends Error
  // Error type raised by [`Node.replace`](#model.Node.replace) when
  // given an invalid replacement.

  function ReplaceError$1(message) {
    var err = Error.call(this, message);
    err.__proto__ = ReplaceError$1.prototype;
    return err
  }

  ReplaceError$1.prototype = Object.create(Error.prototype);
  ReplaceError$1.prototype.constructor = ReplaceError$1;
  ReplaceError$1.prototype.name = "ReplaceError";

  // ::- A slice represents a piece cut out of a larger document. It
  // stores not only a fragment, but also the depth up to which nodes on
  // both side are ‘open’ (cut through).
  var Slice$1 = function Slice(content, openStart, openEnd) {
    // :: Fragment The slice's content.
    this.content = content;
    // :: number The open depth at the start.
    this.openStart = openStart;
    // :: number The open depth at the end.
    this.openEnd = openEnd;
  };

  var prototypeAccessors$1$5 = { size: { configurable: true } };

  // :: number
  // The size this slice would add when inserted into a document.
  prototypeAccessors$1$5.size.get = function () {
    return this.content.size - this.openStart - this.openEnd
  };

  Slice$1.prototype.insertAt = function insertAt (pos, fragment) {
    var content = insertInto$1(this.content, pos + this.openStart, fragment, null);
    return content && new Slice$1(content, this.openStart, this.openEnd)
  };

  Slice$1.prototype.removeBetween = function removeBetween (from, to) {
    return new Slice$1(removeRange$1(this.content, from + this.openStart, to + this.openStart), this.openStart, this.openEnd)
  };

  // :: (Slice) → bool
  // Tests whether this slice is equal to another slice.
  Slice$1.prototype.eq = function eq (other) {
    return this.content.eq(other.content) && this.openStart == other.openStart && this.openEnd == other.openEnd
  };

  Slice$1.prototype.toString = function toString () {
    return this.content + "(" + this.openStart + "," + this.openEnd + ")"
  };

  // :: () → ?Object
  // Convert a slice to a JSON-serializable representation.
  Slice$1.prototype.toJSON = function toJSON () {
    if (!this.content.size) { return null }
    var json = {content: this.content.toJSON()};
    if (this.openStart > 0) { json.openStart = this.openStart; }
    if (this.openEnd > 0) { json.openEnd = this.openEnd; }
    return json
  };

  // :: (Schema, ?Object) → Slice
  // Deserialize a slice from its JSON representation.
  Slice$1.fromJSON = function fromJSON (schema, json) {
    if (!json) { return Slice$1.empty }
    var openStart = json.openStart || 0, openEnd = json.openEnd || 0;
    if (typeof openStart != "number" || typeof openEnd != "number")
      { throw new RangeError("Invalid input for Slice.fromJSON") }
    return new Slice$1(Fragment$1.fromJSON(schema, json.content), openStart, openEnd)
  };

  // :: (Fragment, ?bool) → Slice
  // Create a slice from a fragment by taking the maximum possible
  // open value on both side of the fragment.
  Slice$1.maxOpen = function maxOpen (fragment, openIsolating) {
      if ( openIsolating === void 0 ) { openIsolating=true; }

    var openStart = 0, openEnd = 0;
    for (var n = fragment.firstChild; n && !n.isLeaf && (openIsolating || !n.type.spec.isolating); n = n.firstChild) { openStart++; }
    for (var n$1 = fragment.lastChild; n$1 && !n$1.isLeaf && (openIsolating || !n$1.type.spec.isolating); n$1 = n$1.lastChild) { openEnd++; }
    return new Slice$1(fragment, openStart, openEnd)
  };

  Object.defineProperties( Slice$1.prototype, prototypeAccessors$1$5 );

  function removeRange$1(content, from, to) {
    var ref = content.findIndex(from);
    var index = ref.index;
    var offset = ref.offset;
    var child = content.maybeChild(index);
    var ref$1 = content.findIndex(to);
    var indexTo = ref$1.index;
    var offsetTo = ref$1.offset;
    if (offset == from || child.isText) {
      if (offsetTo != to && !content.child(indexTo).isText) { throw new RangeError("Removing non-flat range") }
      return content.cut(0, from).append(content.cut(to))
    }
    if (index != indexTo) { throw new RangeError("Removing non-flat range") }
    return content.replaceChild(index, child.copy(removeRange$1(child.content, from - offset - 1, to - offset - 1)))
  }

  function insertInto$1(content, dist, insert, parent) {
    var ref = content.findIndex(dist);
    var index = ref.index;
    var offset = ref.offset;
    var child = content.maybeChild(index);
    if (offset == dist || child.isText) {
      if (parent && !parent.canReplace(index, index, insert)) { return null }
      return content.cut(0, dist).append(insert).append(content.cut(dist))
    }
    var inner = insertInto$1(child.content, dist - offset - 1, insert);
    return inner && content.replaceChild(index, child.copy(inner))
  }

  // :: Slice
  // The empty slice.
  Slice$1.empty = new Slice$1(Fragment$1.empty, 0, 0);

  function replace$2($from, $to, slice) {
    if (slice.openStart > $from.depth)
      { throw new ReplaceError$1("Inserted content deeper than insertion position") }
    if ($from.depth - slice.openStart != $to.depth - slice.openEnd)
      { throw new ReplaceError$1("Inconsistent open depths") }
    return replaceOuter$1($from, $to, slice, 0)
  }

  function replaceOuter$1($from, $to, slice, depth) {
    var index = $from.index(depth), node = $from.node(depth);
    if (index == $to.index(depth) && depth < $from.depth - slice.openStart) {
      var inner = replaceOuter$1($from, $to, slice, depth + 1);
      return node.copy(node.content.replaceChild(index, inner))
    } else if (!slice.content.size) {
      return close$1(node, replaceTwoWay$1($from, $to, depth))
    } else if (!slice.openStart && !slice.openEnd && $from.depth == depth && $to.depth == depth) { // Simple, flat case
      var parent = $from.parent, content = parent.content;
      return close$1(parent, content.cut(0, $from.parentOffset).append(slice.content).append(content.cut($to.parentOffset)))
    } else {
      var ref = prepareSliceForReplace$1(slice, $from);
      var start = ref.start;
      var end = ref.end;
      return close$1(node, replaceThreeWay$1($from, start, end, $to, depth))
    }
  }

  function checkJoin$1(main, sub) {
    if (!sub.type.compatibleContent(main.type))
      { throw new ReplaceError$1("Cannot join " + sub.type.name + " onto " + main.type.name) }
  }

  function joinable$2($before, $after, depth) {
    var node = $before.node(depth);
    checkJoin$1(node, $after.node(depth));
    return node
  }

  function addNode$1(child, target) {
    var last = target.length - 1;
    if (last >= 0 && child.isText && child.sameMarkup(target[last]))
      { target[last] = child.withText(target[last].text + child.text); }
    else
      { target.push(child); }
  }

  function addRange$1($start, $end, depth, target) {
    var node = ($end || $start).node(depth);
    var startIndex = 0, endIndex = $end ? $end.index(depth) : node.childCount;
    if ($start) {
      startIndex = $start.index(depth);
      if ($start.depth > depth) {
        startIndex++;
      } else if ($start.textOffset) {
        addNode$1($start.nodeAfter, target);
        startIndex++;
      }
    }
    for (var i = startIndex; i < endIndex; i++) { addNode$1(node.child(i), target); }
    if ($end && $end.depth == depth && $end.textOffset)
      { addNode$1($end.nodeBefore, target); }
  }

  function close$1(node, content) {
    if (!node.type.validContent(content))
      { throw new ReplaceError$1("Invalid content for node " + node.type.name) }
    return node.copy(content)
  }

  function replaceThreeWay$1($from, $start, $end, $to, depth) {
    var openStart = $from.depth > depth && joinable$2($from, $start, depth + 1);
    var openEnd = $to.depth > depth && joinable$2($end, $to, depth + 1);

    var content = [];
    addRange$1(null, $from, depth, content);
    if (openStart && openEnd && $start.index(depth) == $end.index(depth)) {
      checkJoin$1(openStart, openEnd);
      addNode$1(close$1(openStart, replaceThreeWay$1($from, $start, $end, $to, depth + 1)), content);
    } else {
      if (openStart)
        { addNode$1(close$1(openStart, replaceTwoWay$1($from, $start, depth + 1)), content); }
      addRange$1($start, $end, depth, content);
      if (openEnd)
        { addNode$1(close$1(openEnd, replaceTwoWay$1($end, $to, depth + 1)), content); }
    }
    addRange$1($to, null, depth, content);
    return new Fragment$1(content)
  }

  function replaceTwoWay$1($from, $to, depth) {
    var content = [];
    addRange$1(null, $from, depth, content);
    if ($from.depth > depth) {
      var type = joinable$2($from, $to, depth + 1);
      addNode$1(close$1(type, replaceTwoWay$1($from, $to, depth + 1)), content);
    }
    addRange$1($to, null, depth, content);
    return new Fragment$1(content)
  }

  function prepareSliceForReplace$1(slice, $along) {
    var extra = $along.depth - slice.openStart, parent = $along.node(extra);
    var node = parent.copy(slice.content);
    for (var i = extra - 1; i >= 0; i--)
      { node = $along.node(i).copy(Fragment$1.from(node)); }
    return {start: node.resolveNoCache(slice.openStart + extra),
            end: node.resolveNoCache(node.content.size - slice.openEnd - extra)}
  }

  // ::- You can [_resolve_](#model.Node.resolve) a position to get more
  // information about it. Objects of this class represent such a
  // resolved position, providing various pieces of context information,
  // and some helper methods.
  //
  // Throughout this interface, methods that take an optional `depth`
  // parameter will interpret undefined as `this.depth` and negative
  // numbers as `this.depth + value`.
  var ResolvedPos$1 = function ResolvedPos(pos, path, parentOffset) {
    // :: number The position that was resolved.
    this.pos = pos;
    this.path = path;
    // :: number
    // The number of levels the parent node is from the root. If this
    // position points directly into the root node, it is 0. If it
    // points into a top-level paragraph, 1, and so on.
    this.depth = path.length / 3 - 1;
    // :: number The offset this position has into its parent node.
    this.parentOffset = parentOffset;
  };

  var prototypeAccessors$2$2 = { parent: { configurable: true },doc: { configurable: true },textOffset: { configurable: true },nodeAfter: { configurable: true },nodeBefore: { configurable: true } };

  ResolvedPos$1.prototype.resolveDepth = function resolveDepth (val) {
    if (val == null) { return this.depth }
    if (val < 0) { return this.depth + val }
    return val
  };

  // :: Node
  // The parent node that the position points into. Note that even if
  // a position points into a text node, that node is not considered
  // the parent—text nodes are ‘flat’ in this model, and have no content.
  prototypeAccessors$2$2.parent.get = function () { return this.node(this.depth) };

  // :: Node
  // The root node in which the position was resolved.
  prototypeAccessors$2$2.doc.get = function () { return this.node(0) };

  // :: (?number) → Node
  // The ancestor node at the given level. `p.node(p.depth)` is the
  // same as `p.parent`.
  ResolvedPos$1.prototype.node = function node (depth) { return this.path[this.resolveDepth(depth) * 3] };

  // :: (?number) → number
  // The index into the ancestor at the given level. If this points at
  // the 3rd node in the 2nd paragraph on the top level, for example,
  // `p.index(0)` is 1 and `p.index(1)` is 2.
  ResolvedPos$1.prototype.index = function index (depth) { return this.path[this.resolveDepth(depth) * 3 + 1] };

  // :: (?number) → number
  // The index pointing after this position into the ancestor at the
  // given level.
  ResolvedPos$1.prototype.indexAfter = function indexAfter (depth) {
    depth = this.resolveDepth(depth);
    return this.index(depth) + (depth == this.depth && !this.textOffset ? 0 : 1)
  };

  // :: (?number) → number
  // The (absolute) position at the start of the node at the given
  // level.
  ResolvedPos$1.prototype.start = function start (depth) {
    depth = this.resolveDepth(depth);
    return depth == 0 ? 0 : this.path[depth * 3 - 1] + 1
  };

  // :: (?number) → number
  // The (absolute) position at the end of the node at the given
  // level.
  ResolvedPos$1.prototype.end = function end (depth) {
    depth = this.resolveDepth(depth);
    return this.start(depth) + this.node(depth).content.size
  };

  // :: (?number) → number
  // The (absolute) position directly before the wrapping node at the
  // given level, or, when `depth` is `this.depth + 1`, the original
  // position.
  ResolvedPos$1.prototype.before = function before (depth) {
    depth = this.resolveDepth(depth);
    if (!depth) { throw new RangeError("There is no position before the top-level node") }
    return depth == this.depth + 1 ? this.pos : this.path[depth * 3 - 1]
  };

  // :: (?number) → number
  // The (absolute) position directly after the wrapping node at the
  // given level, or the original position when `depth` is `this.depth + 1`.
  ResolvedPos$1.prototype.after = function after (depth) {
    depth = this.resolveDepth(depth);
    if (!depth) { throw new RangeError("There is no position after the top-level node") }
    return depth == this.depth + 1 ? this.pos : this.path[depth * 3 - 1] + this.path[depth * 3].nodeSize
  };

  // :: number
  // When this position points into a text node, this returns the
  // distance between the position and the start of the text node.
  // Will be zero for positions that point between nodes.
  prototypeAccessors$2$2.textOffset.get = function () { return this.pos - this.path[this.path.length - 1] };

  // :: ?Node
  // Get the node directly after the position, if any. If the position
  // points into a text node, only the part of that node after the
  // position is returned.
  prototypeAccessors$2$2.nodeAfter.get = function () {
    var parent = this.parent, index = this.index(this.depth);
    if (index == parent.childCount) { return null }
    var dOff = this.pos - this.path[this.path.length - 1], child = parent.child(index);
    return dOff ? parent.child(index).cut(dOff) : child
  };

  // :: ?Node
  // Get the node directly before the position, if any. If the
  // position points into a text node, only the part of that node
  // before the position is returned.
  prototypeAccessors$2$2.nodeBefore.get = function () {
    var index = this.index(this.depth);
    var dOff = this.pos - this.path[this.path.length - 1];
    if (dOff) { return this.parent.child(index).cut(0, dOff) }
    return index == 0 ? null : this.parent.child(index - 1)
  };

  // :: (number, ?number) → number
  // Get the position at the given index in the parent node at the
  // given depth (which defaults to `this.depth`).
  ResolvedPos$1.prototype.posAtIndex = function posAtIndex (index, depth) {
    depth = this.resolveDepth(depth);
    var node = this.path[depth * 3], pos = depth == 0 ? 0 : this.path[depth * 3 - 1] + 1;
    for (var i = 0; i < index; i++) { pos += node.child(i).nodeSize; }
    return pos
  };

  // :: () → [Mark]
  // Get the marks at this position, factoring in the surrounding
  // marks' [`inclusive`](#model.MarkSpec.inclusive) property. If the
  // position is at the start of a non-empty node, the marks of the
  // node after it (if any) are returned.
  ResolvedPos$1.prototype.marks = function marks () {
    var parent = this.parent, index = this.index();

    // In an empty parent, return the empty array
    if (parent.content.size == 0) { return Mark$1.none }

    // When inside a text node, just return the text node's marks
    if (this.textOffset) { return parent.child(index).marks }

    var main = parent.maybeChild(index - 1), other = parent.maybeChild(index);
    // If the `after` flag is true of there is no node before, make
    // the node after this position the main reference.
    if (!main) { var tmp = main; main = other; other = tmp; }

    // Use all marks in the main node, except those that have
    // `inclusive` set to false and are not present in the other node.
    var marks = main.marks;
    for (var i = 0; i < marks.length; i++)
      { if (marks[i].type.spec.inclusive === false && (!other || !marks[i].isInSet(other.marks)))
        { marks = marks[i--].removeFromSet(marks); } }

    return marks
  };

  // :: (ResolvedPos) → ?[Mark]
  // Get the marks after the current position, if any, except those
  // that are non-inclusive and not present at position `$end`. This
  // is mostly useful for getting the set of marks to preserve after a
  // deletion. Will return `null` if this position is at the end of
  // its parent node or its parent node isn't a textblock (in which
  // case no marks should be preserved).
  ResolvedPos$1.prototype.marksAcross = function marksAcross ($end) {
    var after = this.parent.maybeChild(this.index());
    if (!after || !after.isInline) { return null }

    var marks = after.marks, next = $end.parent.maybeChild($end.index());
    for (var i = 0; i < marks.length; i++)
      { if (marks[i].type.spec.inclusive === false && (!next || !marks[i].isInSet(next.marks)))
        { marks = marks[i--].removeFromSet(marks); } }
    return marks
  };

  // :: (number) → number
  // The depth up to which this position and the given (non-resolved)
  // position share the same parent nodes.
  ResolvedPos$1.prototype.sharedDepth = function sharedDepth (pos) {
    for (var depth = this.depth; depth > 0; depth--)
      { if (this.start(depth) <= pos && this.end(depth) >= pos) { return depth } }
    return 0
  };

  // :: (?ResolvedPos, ?(Node) → bool) → ?NodeRange
  // Returns a range based on the place where this position and the
  // given position diverge around block content. If both point into
  // the same textblock, for example, a range around that textblock
  // will be returned. If they point into different blocks, the range
  // around those blocks in their shared ancestor is returned. You can
  // pass in an optional predicate that will be called with a parent
  // node to see if a range into that parent is acceptable.
  ResolvedPos$1.prototype.blockRange = function blockRange (other, pred) {
      if ( other === void 0 ) { other = this; }

    if (other.pos < this.pos) { return other.blockRange(this) }
    for (var d = this.depth - (this.parent.inlineContent || this.pos == other.pos ? 1 : 0); d >= 0; d--)
      { if (other.pos <= this.end(d) && (!pred || pred(this.node(d))))
        { return new NodeRange$1(this, other, d) } }
  };

  // :: (ResolvedPos) → bool
  // Query whether the given position shares the same parent node.
  ResolvedPos$1.prototype.sameParent = function sameParent (other) {
    return this.pos - this.parentOffset == other.pos - other.parentOffset
  };

  // :: (ResolvedPos) → ResolvedPos
  // Return the greater of this and the given position.
  ResolvedPos$1.prototype.max = function max (other) {
    return other.pos > this.pos ? other : this
  };

  // :: (ResolvedPos) → ResolvedPos
  // Return the smaller of this and the given position.
  ResolvedPos$1.prototype.min = function min (other) {
    return other.pos < this.pos ? other : this
  };

  ResolvedPos$1.prototype.toString = function toString () {
    var str = "";
    for (var i = 1; i <= this.depth; i++)
      { str += (str ? "/" : "") + this.node(i).type.name + "_" + this.index(i - 1); }
    return str + ":" + this.parentOffset
  };

  ResolvedPos$1.resolve = function resolve (doc, pos) {
    if (!(pos >= 0 && pos <= doc.content.size)) { throw new RangeError("Position " + pos + " out of range") }
    var path = [];
    var start = 0, parentOffset = pos;
    for (var node = doc;;) {
      var ref = node.content.findIndex(parentOffset);
        var index = ref.index;
        var offset = ref.offset;
      var rem = parentOffset - offset;
      path.push(node, index, start + offset);
      if (!rem) { break }
      node = node.child(index);
      if (node.isText) { break }
      parentOffset = rem - 1;
      start += offset + 1;
    }
    return new ResolvedPos$1(pos, path, parentOffset)
  };

  ResolvedPos$1.resolveCached = function resolveCached (doc, pos) {
    for (var i = 0; i < resolveCache$1.length; i++) {
      var cached = resolveCache$1[i];
      if (cached.pos == pos && cached.doc == doc) { return cached }
    }
    var result = resolveCache$1[resolveCachePos$1] = ResolvedPos$1.resolve(doc, pos);
    resolveCachePos$1 = (resolveCachePos$1 + 1) % resolveCacheSize$1;
    return result
  };

  Object.defineProperties( ResolvedPos$1.prototype, prototypeAccessors$2$2 );

  var resolveCache$1 = [], resolveCachePos$1 = 0, resolveCacheSize$1 = 12;

  // ::- Represents a flat range of content, i.e. one that starts and
  // ends in the same node.
  var NodeRange$1 = function NodeRange($from, $to, depth) {
    // :: ResolvedPos A resolved position along the start of the
    // content. May have a `depth` greater than this object's `depth`
    // property, since these are the positions that were used to
    // compute the range, not re-resolved positions directly at its
    // boundaries.
    this.$from = $from;
    // :: ResolvedPos A position along the end of the content. See
    // caveat for [`$from`](#model.NodeRange.$from).
    this.$to = $to;
    // :: number The depth of the node that this range points into.
    this.depth = depth;
  };

  var prototypeAccessors$1$1$1 = { start: { configurable: true },end: { configurable: true },parent: { configurable: true },startIndex: { configurable: true },endIndex: { configurable: true } };

  // :: number The position at the start of the range.
  prototypeAccessors$1$1$1.start.get = function () { return this.$from.before(this.depth + 1) };
  // :: number The position at the end of the range.
  prototypeAccessors$1$1$1.end.get = function () { return this.$to.after(this.depth + 1) };

  // :: Node The parent node that the range points into.
  prototypeAccessors$1$1$1.parent.get = function () { return this.$from.node(this.depth) };
  // :: number The start index of the range in the parent node.
  prototypeAccessors$1$1$1.startIndex.get = function () { return this.$from.index(this.depth) };
  // :: number The end index of the range in the parent node.
  prototypeAccessors$1$1$1.endIndex.get = function () { return this.$to.indexAfter(this.depth) };

  Object.defineProperties( NodeRange$1.prototype, prototypeAccessors$1$1$1 );

  var emptyAttrs$1 = Object.create(null);

  // ::- This class represents a node in the tree that makes up a
  // ProseMirror document. So a document is an instance of `Node`, with
  // children that are also instances of `Node`.
  //
  // Nodes are persistent data structures. Instead of changing them, you
  // create new ones with the content you want. Old ones keep pointing
  // at the old document shape. This is made cheaper by sharing
  // structure between the old and new data as much as possible, which a
  // tree shape like this (without back pointers) makes easy.
  //
  // **Do not** directly mutate the properties of a `Node` object. See
  // [the guide](/docs/guide/#doc) for more information.
  var Node$2 = function Node(type, attrs, content, marks) {
    // :: NodeType
    // The type of node that this is.
    this.type = type;

    // :: Object
    // An object mapping attribute names to values. The kind of
    // attributes allowed and required are
    // [determined](#model.NodeSpec.attrs) by the node type.
    this.attrs = attrs;

    // :: Fragment
    // A container holding the node's children.
    this.content = content || Fragment$1.empty;

    // :: [Mark]
    // The marks (things like whether it is emphasized or part of a
    // link) applied to this node.
    this.marks = marks || Mark$1.none;
  };

  var prototypeAccessors$3$1 = { nodeSize: { configurable: true },childCount: { configurable: true },textContent: { configurable: true },firstChild: { configurable: true },lastChild: { configurable: true },isBlock: { configurable: true },isTextblock: { configurable: true },inlineContent: { configurable: true },isInline: { configurable: true },isText: { configurable: true },isLeaf: { configurable: true },isAtom: { configurable: true } };

  // text:: ?string
  // For text nodes, this contains the node's text content.

  // :: number
  // The size of this node, as defined by the integer-based [indexing
  // scheme](/docs/guide/#doc.indexing). For text nodes, this is the
  // amount of characters. For other leaf nodes, it is one. For
  // non-leaf nodes, it is the size of the content plus two (the start
  // and end token).
  prototypeAccessors$3$1.nodeSize.get = function () { return this.isLeaf ? 1 : 2 + this.content.size };

  // :: number
  // The number of children that the node has.
  prototypeAccessors$3$1.childCount.get = function () { return this.content.childCount };

  // :: (number) → Node
  // Get the child node at the given index. Raises an error when the
  // index is out of range.
  Node$2.prototype.child = function child (index) { return this.content.child(index) };

  // :: (number) → ?Node
  // Get the child node at the given index, if it exists.
  Node$2.prototype.maybeChild = function maybeChild (index) { return this.content.maybeChild(index) };

  // :: ((node: Node, offset: number, index: number))
  // Call `f` for every child node, passing the node, its offset
  // into this parent node, and its index.
  Node$2.prototype.forEach = function forEach (f) { this.content.forEach(f); };

  // :: (number, number, (node: Node, pos: number, parent: Node, index: number) → ?bool, ?number)
  // Invoke a callback for all descendant nodes recursively between
  // the given two positions that are relative to start of this node's
  // content. The callback is invoked with the node, its
  // parent-relative position, its parent node, and its child index.
  // When the callback returns false for a given node, that node's
  // children will not be recursed over. The last parameter can be
  // used to specify a starting position to count from.
  Node$2.prototype.nodesBetween = function nodesBetween (from, to, f, startPos) {
      if ( startPos === void 0 ) { startPos = 0; }

    this.content.nodesBetween(from, to, f, startPos, this);
  };

  // :: ((node: Node, pos: number, parent: Node) → ?bool)
  // Call the given callback for every descendant node. Doesn't
  // descend into a node when the callback returns `false`.
  Node$2.prototype.descendants = function descendants (f) {
    this.nodesBetween(0, this.content.size, f);
  };

  // :: string
  // Concatenates all the text nodes found in this fragment and its
  // children.
  prototypeAccessors$3$1.textContent.get = function () { return this.textBetween(0, this.content.size, "") };

  // :: (number, number, ?string, ?string) → string
  // Get all text between positions `from` and `to`. When
  // `blockSeparator` is given, it will be inserted whenever a new
  // block node is started. When `leafText` is given, it'll be
  // inserted for every non-text leaf node encountered.
  Node$2.prototype.textBetween = function textBetween (from, to, blockSeparator, leafText) {
    return this.content.textBetween(from, to, blockSeparator, leafText)
  };

  // :: ?Node
  // Returns this node's first child, or `null` if there are no
  // children.
  prototypeAccessors$3$1.firstChild.get = function () { return this.content.firstChild };

  // :: ?Node
  // Returns this node's last child, or `null` if there are no
  // children.
  prototypeAccessors$3$1.lastChild.get = function () { return this.content.lastChild };

  // :: (Node) → bool
  // Test whether two nodes represent the same piece of document.
  Node$2.prototype.eq = function eq (other) {
    return this == other || (this.sameMarkup(other) && this.content.eq(other.content))
  };

  // :: (Node) → bool
  // Compare the markup (type, attributes, and marks) of this node to
  // those of another. Returns `true` if both have the same markup.
  Node$2.prototype.sameMarkup = function sameMarkup (other) {
    return this.hasMarkup(other.type, other.attrs, other.marks)
  };

  // :: (NodeType, ?Object, ?[Mark]) → bool
  // Check whether this node's markup correspond to the given type,
  // attributes, and marks.
  Node$2.prototype.hasMarkup = function hasMarkup (type, attrs, marks) {
    return this.type == type &&
      compareDeep$1(this.attrs, attrs || type.defaultAttrs || emptyAttrs$1) &&
      Mark$1.sameSet(this.marks, marks || Mark$1.none)
  };

  // :: (?Fragment) → Node
  // Create a new node with the same markup as this node, containing
  // the given content (or empty, if no content is given).
  Node$2.prototype.copy = function copy (content) {
      if ( content === void 0 ) { content = null; }

    if (content == this.content) { return this }
    return new this.constructor(this.type, this.attrs, content, this.marks)
  };

  // :: ([Mark]) → Node
  // Create a copy of this node, with the given set of marks instead
  // of the node's own marks.
  Node$2.prototype.mark = function mark (marks) {
    return marks == this.marks ? this : new this.constructor(this.type, this.attrs, this.content, marks)
  };

  // :: (number, ?number) → Node
  // Create a copy of this node with only the content between the
  // given positions. If `to` is not given, it defaults to the end of
  // the node.
  Node$2.prototype.cut = function cut (from, to) {
    if (from == 0 && to == this.content.size) { return this }
    return this.copy(this.content.cut(from, to))
  };

  // :: (number, ?number) → Slice
  // Cut out the part of the document between the given positions, and
  // return it as a `Slice` object.
  Node$2.prototype.slice = function slice (from, to, includeParents) {
      if ( to === void 0 ) { to = this.content.size; }
      if ( includeParents === void 0 ) { includeParents = false; }

    if (from == to) { return Slice$1.empty }

    var $from = this.resolve(from), $to = this.resolve(to);
    var depth = includeParents ? 0 : $from.sharedDepth(to);
    var start = $from.start(depth), node = $from.node(depth);
    var content = node.content.cut($from.pos - start, $to.pos - start);
    return new Slice$1(content, $from.depth - depth, $to.depth - depth)
  };

  // :: (number, number, Slice) → Node
  // Replace the part of the document between the given positions with
  // the given slice. The slice must 'fit', meaning its open sides
  // must be able to connect to the surrounding content, and its
  // content nodes must be valid children for the node they are placed
  // into. If any of this is violated, an error of type
  // [`ReplaceError`](#model.ReplaceError) is thrown.
  Node$2.prototype.replace = function replace$1 (from, to, slice) {
    return replace$2(this.resolve(from), this.resolve(to), slice)
  };

  // :: (number) → ?Node
  // Find the node directly after the given position.
  Node$2.prototype.nodeAt = function nodeAt (pos) {
    for (var node = this;;) {
      var ref = node.content.findIndex(pos);
        var index = ref.index;
        var offset = ref.offset;
      node = node.maybeChild(index);
      if (!node) { return null }
      if (offset == pos || node.isText) { return node }
      pos -= offset + 1;
    }
  };

  // :: (number) → {node: ?Node, index: number, offset: number}
  // Find the (direct) child node after the given offset, if any,
  // and return it along with its index and offset relative to this
  // node.
  Node$2.prototype.childAfter = function childAfter (pos) {
    var ref = this.content.findIndex(pos);
      var index = ref.index;
      var offset = ref.offset;
    return {node: this.content.maybeChild(index), index: index, offset: offset}
  };

  // :: (number) → {node: ?Node, index: number, offset: number}
  // Find the (direct) child node before the given offset, if any,
  // and return it along with its index and offset relative to this
  // node.
  Node$2.prototype.childBefore = function childBefore (pos) {
    if (pos == 0) { return {node: null, index: 0, offset: 0} }
    var ref = this.content.findIndex(pos);
      var index = ref.index;
      var offset = ref.offset;
    if (offset < pos) { return {node: this.content.child(index), index: index, offset: offset} }
    var node = this.content.child(index - 1);
    return {node: node, index: index - 1, offset: offset - node.nodeSize}
  };

  // :: (number) → ResolvedPos
  // Resolve the given position in the document, returning an
  // [object](#model.ResolvedPos) with information about its context.
  Node$2.prototype.resolve = function resolve (pos) { return ResolvedPos$1.resolveCached(this, pos) };

  Node$2.prototype.resolveNoCache = function resolveNoCache (pos) { return ResolvedPos$1.resolve(this, pos) };

  // :: (number, number, union<Mark, MarkType>) → bool
  // Test whether a given mark or mark type occurs in this document
  // between the two given positions.
  Node$2.prototype.rangeHasMark = function rangeHasMark (from, to, type) {
    var found = false;
    if (to > from) { this.nodesBetween(from, to, function (node) {
      if (type.isInSet(node.marks)) { found = true; }
      return !found
    }); }
    return found
  };

  // :: bool
  // True when this is a block (non-inline node)
  prototypeAccessors$3$1.isBlock.get = function () { return this.type.isBlock };

  // :: bool
  // True when this is a textblock node, a block node with inline
  // content.
  prototypeAccessors$3$1.isTextblock.get = function () { return this.type.isTextblock };

  // :: bool
  // True when this node allows inline content.
  prototypeAccessors$3$1.inlineContent.get = function () { return this.type.inlineContent };

  // :: bool
  // True when this is an inline node (a text node or a node that can
  // appear among text).
  prototypeAccessors$3$1.isInline.get = function () { return this.type.isInline };

  // :: bool
  // True when this is a text node.
  prototypeAccessors$3$1.isText.get = function () { return this.type.isText };

  // :: bool
  // True when this is a leaf node.
  prototypeAccessors$3$1.isLeaf.get = function () { return this.type.isLeaf };

  // :: bool
  // True when this is an atom, i.e. when it does not have directly
  // editable content. This is usually the same as `isLeaf`, but can
  // be configured with the [`atom` property](#model.NodeSpec.atom) on
  // a node's spec (typically used when the node is displayed as an
  // uneditable [node view](#view.NodeView)).
  prototypeAccessors$3$1.isAtom.get = function () { return this.type.isAtom };

  // :: () → string
  // Return a string representation of this node for debugging
  // purposes.
  Node$2.prototype.toString = function toString () {
    if (this.type.spec.toDebugString) { return this.type.spec.toDebugString(this) }
    var name = this.type.name;
    if (this.content.size)
      { name += "(" + this.content.toStringInner() + ")"; }
    return wrapMarks$1(this.marks, name)
  };

  // :: (number) → ContentMatch
  // Get the content match in this node at the given index.
  Node$2.prototype.contentMatchAt = function contentMatchAt (index) {
    var match = this.type.contentMatch.matchFragment(this.content, 0, index);
    if (!match) { throw new Error("Called contentMatchAt on a node with invalid content") }
    return match
  };

  // :: (number, number, ?Fragment, ?number, ?number) → bool
  // Test whether replacing the range between `from` and `to` (by
  // child index) with the given replacement fragment (which defaults
  // to the empty fragment) would leave the node's content valid. You
  // can optionally pass `start` and `end` indices into the
  // replacement fragment.
  Node$2.prototype.canReplace = function canReplace (from, to, replacement, start, end) {
      if ( replacement === void 0 ) { replacement = Fragment$1.empty; }
      if ( start === void 0 ) { start = 0; }
      if ( end === void 0 ) { end = replacement.childCount; }

    var one = this.contentMatchAt(from).matchFragment(replacement, start, end);
    var two = one && one.matchFragment(this.content, to);
    if (!two || !two.validEnd) { return false }
    for (var i = start; i < end; i++) { if (!this.type.allowsMarks(replacement.child(i).marks)) { return false } }
    return true
  };

  // :: (number, number, NodeType, ?[Mark]) → bool
  // Test whether replacing the range `from` to `to` (by index) with a
  // node of the given type would leave the node's content valid.
  Node$2.prototype.canReplaceWith = function canReplaceWith (from, to, type, marks) {
    if (marks && !this.type.allowsMarks(marks)) { return false }
    var start = this.contentMatchAt(from).matchType(type);
    var end = start && start.matchFragment(this.content, to);
    return end ? end.validEnd : false
  };

  // :: (Node) → bool
  // Test whether the given node's content could be appended to this
  // node. If that node is empty, this will only return true if there
  // is at least one node type that can appear in both nodes (to avoid
  // merging completely incompatible nodes).
  Node$2.prototype.canAppend = function canAppend (other) {
    if (other.content.size) { return this.canReplace(this.childCount, this.childCount, other.content) }
    else { return this.type.compatibleContent(other.type) }
  };

  // :: ()
  // Check whether this node and its descendants conform to the
  // schema, and raise error when they do not.
  Node$2.prototype.check = function check () {
    if (!this.type.validContent(this.content))
      { throw new RangeError(("Invalid content for node " + (this.type.name) + ": " + (this.content.toString().slice(0, 50)))) }
    var copy = Mark$1.none;
    for (var i = 0; i < this.marks.length; i++) { copy = this.marks[i].addToSet(copy); }
    if (!Mark$1.sameSet(copy, this.marks))
      { throw new RangeError(("Invalid collection of marks for node " + (this.type.name) + ": " + (this.marks.map(function (m) { return m.type.name; })))) }
    this.content.forEach(function (node) { return node.check(); });
  };

  // :: () → Object
  // Return a JSON-serializeable representation of this node.
  Node$2.prototype.toJSON = function toJSON () {
    var obj = {type: this.type.name};
    for (var _ in this.attrs) {
      obj.attrs = this.attrs;
      break
    }
    if (this.content.size)
      { obj.content = this.content.toJSON(); }
    if (this.marks.length)
      { obj.marks = this.marks.map(function (n) { return n.toJSON(); }); }
    return obj
  };

  // :: (Schema, Object) → Node
  // Deserialize a node from its JSON representation.
  Node$2.fromJSON = function fromJSON (schema, json) {
    if (!json) { throw new RangeError("Invalid input for Node.fromJSON") }
    var marks = null;
    if (json.marks) {
      if (!Array.isArray(json.marks)) { throw new RangeError("Invalid mark data for Node.fromJSON") }
      marks = json.marks.map(schema.markFromJSON);
    }
    if (json.type == "text") {
      if (typeof json.text != "string") { throw new RangeError("Invalid text node in JSON") }
      return schema.text(json.text, marks)
    }
    var content = Fragment$1.fromJSON(schema, json.content);
    return schema.nodeType(json.type).create(json.attrs, content, marks)
  };

  Object.defineProperties( Node$2.prototype, prototypeAccessors$3$1 );

  var TextNode = /*@__PURE__*/(function (Node) {
    function TextNode(type, attrs, content, marks) {
      Node.call(this, type, attrs, null, marks);

      if (!content) { throw new RangeError("Empty text nodes are not allowed") }

      this.text = content;
    }

    if ( Node ) { TextNode.__proto__ = Node; }
    TextNode.prototype = Object.create( Node && Node.prototype );
    TextNode.prototype.constructor = TextNode;

    var prototypeAccessors$1 = { textContent: { configurable: true },nodeSize: { configurable: true } };

    TextNode.prototype.toString = function toString () {
      if (this.type.spec.toDebugString) { return this.type.spec.toDebugString(this) }
      return wrapMarks$1(this.marks, JSON.stringify(this.text))
    };

    prototypeAccessors$1.textContent.get = function () { return this.text };

    TextNode.prototype.textBetween = function textBetween (from, to) { return this.text.slice(from, to) };

    prototypeAccessors$1.nodeSize.get = function () { return this.text.length };

    TextNode.prototype.mark = function mark (marks) {
      return marks == this.marks ? this : new TextNode(this.type, this.attrs, this.text, marks)
    };

    TextNode.prototype.withText = function withText (text) {
      if (text == this.text) { return this }
      return new TextNode(this.type, this.attrs, text, this.marks)
    };

    TextNode.prototype.cut = function cut (from, to) {
      if ( from === void 0 ) { from = 0; }
      if ( to === void 0 ) { to = this.text.length; }

      if (from == 0 && to == this.text.length) { return this }
      return this.withText(this.text.slice(from, to))
    };

    TextNode.prototype.eq = function eq (other) {
      return this.sameMarkup(other) && this.text == other.text
    };

    TextNode.prototype.toJSON = function toJSON () {
      var base = Node.prototype.toJSON.call(this);
      base.text = this.text;
      return base
    };

    Object.defineProperties( TextNode.prototype, prototypeAccessors$1 );

    return TextNode;
  }(Node$2));

  function wrapMarks$1(marks, str) {
    for (var i = marks.length - 1; i >= 0; i--)
      { str = marks[i].type.name + "(" + str + ")"; }
    return str
  }

  // ::- Instances of this class represent a match state of a node
  // type's [content expression](#model.NodeSpec.content), and can be
  // used to find out whether further content matches here, and whether
  // a given position is a valid end of the node.
  var ContentMatch = function ContentMatch(validEnd) {
    // :: bool
    // True when this match state represents a valid end of the node.
    this.validEnd = validEnd;
    this.next = [];
    this.wrapCache = [];
  };

  var prototypeAccessors$4$1 = { inlineContent: { configurable: true },defaultType: { configurable: true },edgeCount: { configurable: true } };

  ContentMatch.parse = function parse (string, nodeTypes) {
    var stream = new TokenStream(string, nodeTypes);
    if (stream.next == null) { return ContentMatch.empty }
    var expr = parseExpr(stream);
    if (stream.next) { stream.err("Unexpected trailing text"); }
    var match = dfa(nfa(expr));
    checkForDeadEnds(match, stream);
    return match
  };

  // :: (NodeType) → ?ContentMatch
  // Match a node type, returning a match after that node if
  // successful.
  ContentMatch.prototype.matchType = function matchType (type) {
    for (var i = 0; i < this.next.length; i += 2)
      { if (this.next[i] == type) { return this.next[i + 1] } }
    return null
  };

  // :: (Fragment, ?number, ?number) → ?ContentMatch
  // Try to match a fragment. Returns the resulting match when
  // successful.
  ContentMatch.prototype.matchFragment = function matchFragment (frag, start, end) {
      if ( start === void 0 ) { start = 0; }
      if ( end === void 0 ) { end = frag.childCount; }

    var cur = this;
    for (var i = start; cur && i < end; i++)
      { cur = cur.matchType(frag.child(i).type); }
    return cur
  };

  prototypeAccessors$4$1.inlineContent.get = function () {
    var first = this.next[0];
    return first ? first.isInline : false
  };

  // :: ?NodeType
  // Get the first matching node type at this match position that can
  // be generated.
  prototypeAccessors$4$1.defaultType.get = function () {
    for (var i = 0; i < this.next.length; i += 2) {
      var type = this.next[i];
      if (!(type.isText || type.hasRequiredAttrs())) { return type }
    }
  };

  ContentMatch.prototype.compatible = function compatible (other) {
    for (var i = 0; i < this.next.length; i += 2)
      { for (var j = 0; j < other.next.length; j += 2)
        { if (this.next[i] == other.next[j]) { return true } } }
    return false
  };

  // :: (Fragment, bool, ?number) → ?Fragment
  // Try to match the given fragment, and if that fails, see if it can
  // be made to match by inserting nodes in front of it. When
  // successful, return a fragment of inserted nodes (which may be
  // empty if nothing had to be inserted). When `toEnd` is true, only
  // return a fragment if the resulting match goes to the end of the
  // content expression.
  ContentMatch.prototype.fillBefore = function fillBefore (after, toEnd, startIndex) {
      if ( toEnd === void 0 ) { toEnd = false; }
      if ( startIndex === void 0 ) { startIndex = 0; }

    var seen = [this];
    function search(match, types) {
      var finished = match.matchFragment(after, startIndex);
      if (finished && (!toEnd || finished.validEnd))
        { return Fragment$1.from(types.map(function (tp) { return tp.createAndFill(); })) }

      for (var i = 0; i < match.next.length; i += 2) {
        var type = match.next[i], next = match.next[i + 1];
        if (!(type.isText || type.hasRequiredAttrs()) && seen.indexOf(next) == -1) {
          seen.push(next);
          var found = search(next, types.concat(type));
          if (found) { return found }
        }
      }
    }

    return search(this, [])
  };

  // :: (NodeType) → ?[NodeType]
  // Find a set of wrapping node types that would allow a node of the
  // given type to appear at this position. The result may be empty
  // (when it fits directly) and will be null when no such wrapping
  // exists.
  ContentMatch.prototype.findWrapping = function findWrapping (target) {
    for (var i = 0; i < this.wrapCache.length; i += 2)
      { if (this.wrapCache[i] == target) { return this.wrapCache[i + 1] } }
    var computed = this.computeWrapping(target);
    this.wrapCache.push(target, computed);
    return computed
  };

  ContentMatch.prototype.computeWrapping = function computeWrapping (target) {
    var seen = Object.create(null), active = [{match: this, type: null, via: null}];
    while (active.length) {
      var current = active.shift(), match = current.match;
      if (match.matchType(target)) {
        var result = [];
        for (var obj = current; obj.type; obj = obj.via)
          { result.push(obj.type); }
        return result.reverse()
      }
      for (var i = 0; i < match.next.length; i += 2) {
        var type = match.next[i];
        if (!type.isLeaf && !type.hasRequiredAttrs() && !(type.name in seen) && (!current.type || match.next[i + 1].validEnd)) {
          active.push({match: type.contentMatch, type: type, via: current});
          seen[type.name] = true;
        }
      }
    }
  };

  // :: number
  // The number of outgoing edges this node has in the finite
  // automaton that describes the content expression.
  prototypeAccessors$4$1.edgeCount.get = function () {
    return this.next.length >> 1
  };

  // :: (number) → {type: NodeType, next: ContentMatch}
  // Get the _n_​th outgoing edge from this node in the finite
  // automaton that describes the content expression.
  ContentMatch.prototype.edge = function edge (n) {
    var i = n << 1;
    if (i >= this.next.length) { throw new RangeError(("There's no " + n + "th edge in this content match")) }
    return {type: this.next[i], next: this.next[i + 1]}
  };

  ContentMatch.prototype.toString = function toString () {
    var seen = [];
    function scan(m) {
      seen.push(m);
      for (var i = 1; i < m.next.length; i += 2)
        { if (seen.indexOf(m.next[i]) == -1) { scan(m.next[i]); } }
    }
    scan(this);
    return seen.map(function (m, i) {
      var out = i + (m.validEnd ? "*" : " ") + " ";
      for (var i$1 = 0; i$1 < m.next.length; i$1 += 2)
        { out += (i$1 ? ", " : "") + m.next[i$1].name + "->" + seen.indexOf(m.next[i$1 + 1]); }
      return out
    }).join("\n")
  };

  Object.defineProperties( ContentMatch.prototype, prototypeAccessors$4$1 );

  ContentMatch.empty = new ContentMatch(true);

  var TokenStream = function TokenStream(string, nodeTypes) {
    this.string = string;
    this.nodeTypes = nodeTypes;
    this.inline = null;
    this.pos = 0;
    this.tokens = string.split(/\s*(?=\b|\W|$)/);
    if (this.tokens[this.tokens.length - 1] == "") { this.tokens.pop(); }
    if (this.tokens[0] == "") { this.tokens.shift(); }
  };

  var prototypeAccessors$1$2$1 = { next: { configurable: true } };

  prototypeAccessors$1$2$1.next.get = function () { return this.tokens[this.pos] };

  TokenStream.prototype.eat = function eat (tok) { return this.next == tok && (this.pos++ || true) };

  TokenStream.prototype.err = function err (str) { throw new SyntaxError(str + " (in content expression '" + this.string + "')") };

  Object.defineProperties( TokenStream.prototype, prototypeAccessors$1$2$1 );

  function parseExpr(stream) {
    var exprs = [];
    do { exprs.push(parseExprSeq(stream)); }
    while (stream.eat("|"))
    return exprs.length == 1 ? exprs[0] : {type: "choice", exprs: exprs}
  }

  function parseExprSeq(stream) {
    var exprs = [];
    do { exprs.push(parseExprSubscript(stream)); }
    while (stream.next && stream.next != ")" && stream.next != "|")
    return exprs.length == 1 ? exprs[0] : {type: "seq", exprs: exprs}
  }

  function parseExprSubscript(stream) {
    var expr = parseExprAtom(stream);
    for (;;) {
      if (stream.eat("+"))
        { expr = {type: "plus", expr: expr}; }
      else if (stream.eat("*"))
        { expr = {type: "star", expr: expr}; }
      else if (stream.eat("?"))
        { expr = {type: "opt", expr: expr}; }
      else if (stream.eat("{"))
        { expr = parseExprRange(stream, expr); }
      else { break }
    }
    return expr
  }

  function parseNum(stream) {
    if (/\D/.test(stream.next)) { stream.err("Expected number, got '" + stream.next + "'"); }
    var result = Number(stream.next);
    stream.pos++;
    return result
  }

  function parseExprRange(stream, expr) {
    var min = parseNum(stream), max = min;
    if (stream.eat(",")) {
      if (stream.next != "}") { max = parseNum(stream); }
      else { max = -1; }
    }
    if (!stream.eat("}")) { stream.err("Unclosed braced range"); }
    return {type: "range", min: min, max: max, expr: expr}
  }

  function resolveName(stream, name) {
    var types = stream.nodeTypes, type = types[name];
    if (type) { return [type] }
    var result = [];
    for (var typeName in types) {
      var type$1 = types[typeName];
      if (type$1.groups.indexOf(name) > -1) { result.push(type$1); }
    }
    if (result.length == 0) { stream.err("No node type or group '" + name + "' found"); }
    return result
  }

  function parseExprAtom(stream) {
    if (stream.eat("(")) {
      var expr = parseExpr(stream);
      if (!stream.eat(")")) { stream.err("Missing closing paren"); }
      return expr
    } else if (!/\W/.test(stream.next)) {
      var exprs = resolveName(stream, stream.next).map(function (type) {
        if (stream.inline == null) { stream.inline = type.isInline; }
        else if (stream.inline != type.isInline) { stream.err("Mixing inline and block content"); }
        return {type: "name", value: type}
      });
      stream.pos++;
      return exprs.length == 1 ? exprs[0] : {type: "choice", exprs: exprs}
    } else {
      stream.err("Unexpected token '" + stream.next + "'");
    }
  }

  // The code below helps compile a regular-expression-like language
  // into a deterministic finite automaton. For a good introduction to
  // these concepts, see https://swtch.com/~rsc/regexp/regexp1.html

  // : (Object) → [[{term: ?any, to: number}]]
  // Construct an NFA from an expression as returned by the parser. The
  // NFA is represented as an array of states, which are themselves
  // arrays of edges, which are `{term, to}` objects. The first state is
  // the entry state and the last node is the success state.
  //
  // Note that unlike typical NFAs, the edge ordering in this one is
  // significant, in that it is used to contruct filler content when
  // necessary.
  function nfa(expr) {
    var nfa = [[]];
    connect(compile(expr, 0), node());
    return nfa

    function node() { return nfa.push([]) - 1 }
    function edge(from, to, term) {
      var edge = {term: term, to: to};
      nfa[from].push(edge);
      return edge
    }
    function connect(edges, to) { edges.forEach(function (edge) { return edge.to = to; }); }

    function compile(expr, from) {
      if (expr.type == "choice") {
        return expr.exprs.reduce(function (out, expr) { return out.concat(compile(expr, from)); }, [])
      } else if (expr.type == "seq") {
        for (var i = 0;; i++) {
          var next = compile(expr.exprs[i], from);
          if (i == expr.exprs.length - 1) { return next }
          connect(next, from = node());
        }
      } else if (expr.type == "star") {
        var loop = node();
        edge(from, loop);
        connect(compile(expr.expr, loop), loop);
        return [edge(loop)]
      } else if (expr.type == "plus") {
        var loop$1 = node();
        connect(compile(expr.expr, from), loop$1);
        connect(compile(expr.expr, loop$1), loop$1);
        return [edge(loop$1)]
      } else if (expr.type == "opt") {
        return [edge(from)].concat(compile(expr.expr, from))
      } else if (expr.type == "range") {
        var cur = from;
        for (var i$1 = 0; i$1 < expr.min; i$1++) {
          var next$1 = node();
          connect(compile(expr.expr, cur), next$1);
          cur = next$1;
        }
        if (expr.max == -1) {
          connect(compile(expr.expr, cur), cur);
        } else {
          for (var i$2 = expr.min; i$2 < expr.max; i$2++) {
            var next$2 = node();
            edge(cur, next$2);
            connect(compile(expr.expr, cur), next$2);
            cur = next$2;
          }
        }
        return [edge(cur)]
      } else if (expr.type == "name") {
        return [edge(from, null, expr.value)]
      }
    }
  }

  function cmp(a, b) { return b - a }

  // Get the set of nodes reachable by null edges from `node`. Omit
  // nodes with only a single null-out-edge, since they may lead to
  // needless duplicated nodes.
  function nullFrom(nfa, node) {
    var result = [];
    scan(node);
    return result.sort(cmp)

    function scan(node) {
      var edges = nfa[node];
      if (edges.length == 1 && !edges[0].term) { return scan(edges[0].to) }
      result.push(node);
      for (var i = 0; i < edges.length; i++) {
        var ref = edges[i];
        var term = ref.term;
        var to = ref.to;
        if (!term && result.indexOf(to) == -1) { scan(to); }
      }
    }
  }

  // : ([[{term: ?any, to: number}]]) → ContentMatch
  // Compiles an NFA as produced by `nfa` into a DFA, modeled as a set
  // of state objects (`ContentMatch` instances) with transitions
  // between them.
  function dfa(nfa) {
    var labeled = Object.create(null);
    return explore(nullFrom(nfa, 0))

    function explore(states) {
      var out = [];
      states.forEach(function (node) {
        nfa[node].forEach(function (ref) {
          var term = ref.term;
          var to = ref.to;

          if (!term) { return }
          var known = out.indexOf(term), set = known > -1 && out[known + 1];
          nullFrom(nfa, to).forEach(function (node) {
            if (!set) { out.push(term, set = []); }
            if (set.indexOf(node) == -1) { set.push(node); }
          });
        });
      });
      var state = labeled[states.join(",")] = new ContentMatch(states.indexOf(nfa.length - 1) > -1);
      for (var i = 0; i < out.length; i += 2) {
        var states$1 = out[i + 1].sort(cmp);
        state.next.push(out[i], labeled[states$1.join(",")] || explore(states$1));
      }
      return state
    }
  }

  function checkForDeadEnds(match, stream) {
    for (var i = 0, work = [match]; i < work.length; i++) {
      var state = work[i], dead = !state.validEnd, nodes = [];
      for (var j = 0; j < state.next.length; j += 2) {
        var node = state.next[j], next = state.next[j + 1];
        nodes.push(node.name);
        if (dead && !(node.isText || node.hasRequiredAttrs())) { dead = false; }
        if (work.indexOf(next) == -1) { work.push(next); }
      }
      if (dead) { stream.err("Only non-generatable nodes (" + nodes.join(", ") + ") in a required position (see https://prosemirror.net/docs/guide/#generatable)"); }
    }
  }

  // For node types where all attrs have a default value (or which don't
  // have any attributes), build up a single reusable default attribute
  // object, and use it for all nodes that don't specify specific
  // attributes.
  function defaultAttrs(attrs) {
    var defaults = Object.create(null);
    for (var attrName in attrs) {
      var attr = attrs[attrName];
      if (!attr.hasDefault) { return null }
      defaults[attrName] = attr.default;
    }
    return defaults
  }

  function computeAttrs(attrs, value) {
    var built = Object.create(null);
    for (var name in attrs) {
      var given = value && value[name];
      if (given === undefined) {
        var attr = attrs[name];
        if (attr.hasDefault) { given = attr.default; }
        else { throw new RangeError("No value supplied for attribute " + name) }
      }
      built[name] = given;
    }
    return built
  }

  function initAttrs(attrs) {
    var result = Object.create(null);
    if (attrs) { for (var name in attrs) { result[name] = new Attribute(attrs[name]); } }
    return result
  }

  // ::- Node types are objects allocated once per `Schema` and used to
  // [tag](#model.Node.type) `Node` instances. They contain information
  // about the node type, such as its name and what kind of node it
  // represents.
  var NodeType$1 = function NodeType(name, schema, spec) {
    // :: string
    // The name the node type has in this schema.
    this.name = name;

    // :: Schema
    // A link back to the `Schema` the node type belongs to.
    this.schema = schema;

    // :: NodeSpec
    // The spec that this type is based on
    this.spec = spec;

    this.groups = spec.group ? spec.group.split(" ") : [];
    this.attrs = initAttrs(spec.attrs);

    this.defaultAttrs = defaultAttrs(this.attrs);

    // :: ContentMatch
    // The starting match of the node type's content expression.
    this.contentMatch = null;

    // : ?[MarkType]
    // The set of marks allowed in this node. `null` means all marks
    // are allowed.
    this.markSet = null;

    // :: bool
    // True if this node type has inline content.
    this.inlineContent = null;

    // :: bool
    // True if this is a block type
    this.isBlock = !(spec.inline || name == "text");

    // :: bool
    // True if this is the text node type.
    this.isText = name == "text";
  };

  var prototypeAccessors$5$1 = { isInline: { configurable: true },isTextblock: { configurable: true },isLeaf: { configurable: true },isAtom: { configurable: true } };

  // :: bool
  // True if this is an inline type.
  prototypeAccessors$5$1.isInline.get = function () { return !this.isBlock };

  // :: bool
  // True if this is a textblock type, a block that contains inline
  // content.
  prototypeAccessors$5$1.isTextblock.get = function () { return this.isBlock && this.inlineContent };

  // :: bool
  // True for node types that allow no content.
  prototypeAccessors$5$1.isLeaf.get = function () { return this.contentMatch == ContentMatch.empty };

  // :: bool
  // True when this node is an atom, i.e. when it does not have
  // directly editable content.
  prototypeAccessors$5$1.isAtom.get = function () { return this.isLeaf || this.spec.atom };

  // :: () → bool
  // Tells you whether this node type has any required attributes.
  NodeType$1.prototype.hasRequiredAttrs = function hasRequiredAttrs () {
    for (var n in this.attrs) { if (this.attrs[n].isRequired) { return true } }
    return false
  };

  NodeType$1.prototype.compatibleContent = function compatibleContent (other) {
    return this == other || this.contentMatch.compatible(other.contentMatch)
  };

  NodeType$1.prototype.computeAttrs = function computeAttrs$1 (attrs) {
    if (!attrs && this.defaultAttrs) { return this.defaultAttrs }
    else { return computeAttrs(this.attrs, attrs) }
  };

  // :: (?Object, ?union<Fragment, Node, [Node]>, ?[Mark]) → Node
  // Create a `Node` of this type. The given attributes are
  // checked and defaulted (you can pass `null` to use the type's
  // defaults entirely, if no required attributes exist). `content`
  // may be a `Fragment`, a node, an array of nodes, or
  // `null`. Similarly `marks` may be `null` to default to the empty
  // set of marks.
  NodeType$1.prototype.create = function create (attrs, content, marks) {
    if (this.isText) { throw new Error("NodeType.create can't construct text nodes") }
    return new Node$2(this, this.computeAttrs(attrs), Fragment$1.from(content), Mark$1.setFrom(marks))
  };

  // :: (?Object, ?union<Fragment, Node, [Node]>, ?[Mark]) → Node
  // Like [`create`](#model.NodeType.create), but check the given content
  // against the node type's content restrictions, and throw an error
  // if it doesn't match.
  NodeType$1.prototype.createChecked = function createChecked (attrs, content, marks) {
    content = Fragment$1.from(content);
    if (!this.validContent(content))
      { throw new RangeError("Invalid content for node " + this.name) }
    return new Node$2(this, this.computeAttrs(attrs), content, Mark$1.setFrom(marks))
  };

  // :: (?Object, ?union<Fragment, Node, [Node]>, ?[Mark]) → ?Node
  // Like [`create`](#model.NodeType.create), but see if it is necessary to
  // add nodes to the start or end of the given fragment to make it
  // fit the node. If no fitting wrapping can be found, return null.
  // Note that, due to the fact that required nodes can always be
  // created, this will always succeed if you pass null or
  // `Fragment.empty` as content.
  NodeType$1.prototype.createAndFill = function createAndFill (attrs, content, marks) {
    attrs = this.computeAttrs(attrs);
    content = Fragment$1.from(content);
    if (content.size) {
      var before = this.contentMatch.fillBefore(content);
      if (!before) { return null }
      content = before.append(content);
    }
    var after = this.contentMatch.matchFragment(content).fillBefore(Fragment$1.empty, true);
    if (!after) { return null }
    return new Node$2(this, attrs, content.append(after), Mark$1.setFrom(marks))
  };

  // :: (Fragment) → bool
  // Returns true if the given fragment is valid content for this node
  // type with the given attributes.
  NodeType$1.prototype.validContent = function validContent (content) {
    var result = this.contentMatch.matchFragment(content);
    if (!result || !result.validEnd) { return false }
    for (var i = 0; i < content.childCount; i++)
      { if (!this.allowsMarks(content.child(i).marks)) { return false } }
    return true
  };

  // :: (MarkType) → bool
  // Check whether the given mark type is allowed in this node.
  NodeType$1.prototype.allowsMarkType = function allowsMarkType (markType) {
    return this.markSet == null || this.markSet.indexOf(markType) > -1
  };

  // :: ([Mark]) → bool
  // Test whether the given set of marks are allowed in this node.
  NodeType$1.prototype.allowsMarks = function allowsMarks (marks) {
    if (this.markSet == null) { return true }
    for (var i = 0; i < marks.length; i++) { if (!this.allowsMarkType(marks[i].type)) { return false } }
    return true
  };

  // :: ([Mark]) → [Mark]
  // Removes the marks that are not allowed in this node from the given set.
  NodeType$1.prototype.allowedMarks = function allowedMarks (marks) {
    if (this.markSet == null) { return marks }
    var copy;
    for (var i = 0; i < marks.length; i++) {
      if (!this.allowsMarkType(marks[i].type)) {
        if (!copy) { copy = marks.slice(0, i); }
      } else if (copy) {
        copy.push(marks[i]);
      }
    }
    return !copy ? marks : copy.length ? copy : Mark$1.empty
  };

  NodeType$1.compile = function compile (nodes, schema) {
    var result = Object.create(null);
    nodes.forEach(function (name, spec) { return result[name] = new NodeType$1(name, schema, spec); });

    var topType = schema.spec.topNode || "doc";
    if (!result[topType]) { throw new RangeError("Schema is missing its top node type ('" + topType + "')") }
    if (!result.text) { throw new RangeError("Every schema needs a 'text' type") }
    for (var _ in result.text.attrs) { throw new RangeError("The text node type should not have attributes") }

    return result
  };

  Object.defineProperties( NodeType$1.prototype, prototypeAccessors$5$1 );

  // Attribute descriptors

  var Attribute = function Attribute(options) {
    this.hasDefault = Object.prototype.hasOwnProperty.call(options, "default");
    this.default = options.default;
  };

  var prototypeAccessors$1$3$1 = { isRequired: { configurable: true } };

  prototypeAccessors$1$3$1.isRequired.get = function () {
    return !this.hasDefault
  };

  Object.defineProperties( Attribute.prototype, prototypeAccessors$1$3$1 );

  // Marks

  // ::- Like nodes, marks (which are associated with nodes to signify
  // things like emphasis or being part of a link) are
  // [tagged](#model.Mark.type) with type objects, which are
  // instantiated once per `Schema`.
  var MarkType = function MarkType(name, rank, schema, spec) {
    // :: string
    // The name of the mark type.
    this.name = name;

    // :: Schema
    // The schema that this mark type instance is part of.
    this.schema = schema;

    // :: MarkSpec
    // The spec on which the type is based.
    this.spec = spec;

    this.attrs = initAttrs(spec.attrs);

    this.rank = rank;
    this.excluded = null;
    var defaults = defaultAttrs(this.attrs);
    this.instance = defaults && new Mark$1(this, defaults);
  };

  // :: (?Object) → Mark
  // Create a mark of this type. `attrs` may be `null` or an object
  // containing only some of the mark's attributes. The others, if
  // they have defaults, will be added.
  MarkType.prototype.create = function create (attrs) {
    if (!attrs && this.instance) { return this.instance }
    return new Mark$1(this, computeAttrs(this.attrs, attrs))
  };

  MarkType.compile = function compile (marks, schema) {
    var result = Object.create(null), rank = 0;
    marks.forEach(function (name, spec) { return result[name] = new MarkType(name, rank++, schema, spec); });
    return result
  };

  // :: ([Mark]) → [Mark]
  // When there is a mark of this type in the given set, a new set
  // without it is returned. Otherwise, the input set is returned.
  MarkType.prototype.removeFromSet = function removeFromSet (set) {
    for (var i = 0; i < set.length; i++) { if (set[i].type == this) {
      set = set.slice(0, i).concat(set.slice(i + 1));
      i--;
    } }
    return set
  };

  // :: ([Mark]) → ?Mark
  // Tests whether there is a mark of this type in the given set.
  MarkType.prototype.isInSet = function isInSet (set) {
    for (var i = 0; i < set.length; i++)
      { if (set[i].type == this) { return set[i] } }
  };

  // :: (MarkType) → bool
  // Queries whether a given mark type is
  // [excluded](#model.MarkSpec.excludes) by this one.
  MarkType.prototype.excludes = function excludes (other) {
    return this.excluded.indexOf(other) > -1
  };

  // SchemaSpec:: interface
  // An object describing a schema, as passed to the [`Schema`](#model.Schema)
  // constructor.
  //
  //   nodes:: union<Object<NodeSpec>, OrderedMap<NodeSpec>>
  //   The node types in this schema. Maps names to
  //   [`NodeSpec`](#model.NodeSpec) objects that describe the node type
  //   associated with that name. Their order is significant—it
  //   determines which [parse rules](#model.NodeSpec.parseDOM) take
  //   precedence by default, and which nodes come first in a given
  //   [group](#model.NodeSpec.group).
  //
  //   marks:: ?union<Object<MarkSpec>, OrderedMap<MarkSpec>>
  //   The mark types that exist in this schema. The order in which they
  //   are provided determines the order in which [mark
  //   sets](#model.Mark.addToSet) are sorted and in which [parse
  //   rules](#model.MarkSpec.parseDOM) are tried.
  //
  //   topNode:: ?string
  //   The name of the default top-level node for the schema. Defaults
  //   to `"doc"`.

  // NodeSpec:: interface
  //
  //   content:: ?string
  //   The content expression for this node, as described in the [schema
  //   guide](/docs/guide/#schema.content_expressions). When not given,
  //   the node does not allow any content.
  //
  //   marks:: ?string
  //   The marks that are allowed inside of this node. May be a
  //   space-separated string referring to mark names or groups, `"_"`
  //   to explicitly allow all marks, or `""` to disallow marks. When
  //   not given, nodes with inline content default to allowing all
  //   marks, other nodes default to not allowing marks.
  //
  //   group:: ?string
  //   The group or space-separated groups to which this node belongs,
  //   which can be referred to in the content expressions for the
  //   schema.
  //
  //   inline:: ?bool
  //   Should be set to true for inline nodes. (Implied for text nodes.)
  //
  //   atom:: ?bool
  //   Can be set to true to indicate that, though this isn't a [leaf
  //   node](#model.NodeType.isLeaf), it doesn't have directly editable
  //   content and should be treated as a single unit in the view.
  //
  //   attrs:: ?Object<AttributeSpec>
  //   The attributes that nodes of this type get.
  //
  //   selectable:: ?bool
  //   Controls whether nodes of this type can be selected as a [node
  //   selection](#state.NodeSelection). Defaults to true for non-text
  //   nodes.
  //
  //   draggable:: ?bool
  //   Determines whether nodes of this type can be dragged without
  //   being selected. Defaults to false.
  //
  //   code:: ?bool
  //   Can be used to indicate that this node contains code, which
  //   causes some commands to behave differently.
  //
  //   defining:: ?bool
  //   Determines whether this node is considered an important parent
  //   node during replace operations (such as paste). Non-defining (the
  //   default) nodes get dropped when their entire content is replaced,
  //   whereas defining nodes persist and wrap the inserted content.
  //   Likewise, in _inserted_ content the defining parents of the
  //   content are preserved when possible. Typically,
  //   non-default-paragraph textblock types, and possibly list items,
  //   are marked as defining.
  //
  //   isolating:: ?bool
  //   When enabled (default is false), the sides of nodes of this type
  //   count as boundaries that regular editing operations, like
  //   backspacing or lifting, won't cross. An example of a node that
  //   should probably have this enabled is a table cell.
  //
  //   toDOM:: ?(node: Node) → DOMOutputSpec
  //   Defines the default way a node of this type should be serialized
  //   to DOM/HTML (as used by
  //   [`DOMSerializer.fromSchema`](#model.DOMSerializer^fromSchema)).
  //   Should return a DOM node or an [array
  //   structure](#model.DOMOutputSpec) that describes one, with an
  //   optional number zero (“hole”) in it to indicate where the node's
  //   content should be inserted.
  //
  //   For text nodes, the default is to create a text DOM node. Though
  //   it is possible to create a serializer where text is rendered
  //   differently, this is not supported inside the editor, so you
  //   shouldn't override that in your text node spec.
  //
  //   parseDOM:: ?[ParseRule]
  //   Associates DOM parser information with this node, which can be
  //   used by [`DOMParser.fromSchema`](#model.DOMParser^fromSchema) to
  //   automatically derive a parser. The `node` field in the rules is
  //   implied (the name of this node will be filled in automatically).
  //   If you supply your own parser, you do not need to also specify
  //   parsing rules in your schema.
  //
  //   toDebugString:: ?(node: Node) -> string
  //   Defines the default way a node of this type should be serialized
  //   to a string representation for debugging (e.g. in error messages).

  // MarkSpec:: interface
  //
  //   attrs:: ?Object<AttributeSpec>
  //   The attributes that marks of this type get.
  //
  //   inclusive:: ?bool
  //   Whether this mark should be active when the cursor is positioned
  //   at its end (or at its start when that is also the start of the
  //   parent node). Defaults to true.
  //
  //   excludes:: ?string
  //   Determines which other marks this mark can coexist with. Should
  //   be a space-separated strings naming other marks or groups of marks.
  //   When a mark is [added](#model.Mark.addToSet) to a set, all marks
  //   that it excludes are removed in the process. If the set contains
  //   any mark that excludes the new mark but is not, itself, excluded
  //   by the new mark, the mark can not be added an the set. You can
  //   use the value `"_"` to indicate that the mark excludes all
  //   marks in the schema.
  //
  //   Defaults to only being exclusive with marks of the same type. You
  //   can set it to an empty string (or any string not containing the
  //   mark's own name) to allow multiple marks of a given type to
  //   coexist (as long as they have different attributes).
  //
  //   group:: ?string
  //   The group or space-separated groups to which this mark belongs.
  //
  //   spanning:: ?bool
  //   Determines whether marks of this type can span multiple adjacent
  //   nodes when serialized to DOM/HTML. Defaults to true.
  //
  //   toDOM:: ?(mark: Mark, inline: bool) → DOMOutputSpec
  //   Defines the default way marks of this type should be serialized
  //   to DOM/HTML. When the resulting spec contains a hole, that is
  //   where the marked content is placed. Otherwise, it is appended to
  //   the top node.
  //
  //   parseDOM:: ?[ParseRule]
  //   Associates DOM parser information with this mark (see the
  //   corresponding [node spec field](#model.NodeSpec.parseDOM)). The
  //   `mark` field in the rules is implied.

  // AttributeSpec:: interface
  //
  // Used to [define](#model.NodeSpec.attrs) attributes on nodes or
  // marks.
  //
  //   default:: ?any
  //   The default value for this attribute, to use when no explicit
  //   value is provided. Attributes that have no default must be
  //   provided whenever a node or mark of a type that has them is
  //   created.

  // ::- A document schema. Holds [node](#model.NodeType) and [mark
  // type](#model.MarkType) objects for the nodes and marks that may
  // occur in conforming documents, and provides functionality for
  // creating and deserializing such documents.
  var Schema = function Schema(spec) {
    // :: SchemaSpec
    // The [spec](#model.SchemaSpec) on which the schema is based,
    // with the added guarantee that its `nodes` and `marks`
    // properties are
    // [`OrderedMap`](https://github.com/marijnh/orderedmap) instances
    // (not raw objects).
    this.spec = {};
    for (var prop in spec) { this.spec[prop] = spec[prop]; }
    this.spec.nodes = orderedmap.from(spec.nodes);
    this.spec.marks = orderedmap.from(spec.marks);

    // :: Object<NodeType>
    // An object mapping the schema's node names to node type objects.
    this.nodes = NodeType$1.compile(this.spec.nodes, this);

    // :: Object<MarkType>
    // A map from mark names to mark type objects.
    this.marks = MarkType.compile(this.spec.marks, this);

    var contentExprCache = Object.create(null);
    for (var prop$1 in this.nodes) {
      if (prop$1 in this.marks)
        { throw new RangeError(prop$1 + " can not be both a node and a mark") }
      var type = this.nodes[prop$1], contentExpr = type.spec.content || "", markExpr = type.spec.marks;
      type.contentMatch = contentExprCache[contentExpr] ||
        (contentExprCache[contentExpr] = ContentMatch.parse(contentExpr, this.nodes));
      type.inlineContent = type.contentMatch.inlineContent;
      type.markSet = markExpr == "_" ? null :
        markExpr ? gatherMarks(this, markExpr.split(" ")) :
        markExpr == "" || !type.inlineContent ? [] : null;
    }
    for (var prop$2 in this.marks) {
      var type$1 = this.marks[prop$2], excl = type$1.spec.excludes;
      type$1.excluded = excl == null ? [type$1] : excl == "" ? [] : gatherMarks(this, excl.split(" "));
    }

    this.nodeFromJSON = this.nodeFromJSON.bind(this);
    this.markFromJSON = this.markFromJSON.bind(this);

    // :: NodeType
    // The type of the [default top node](#model.SchemaSpec.topNode)
    // for this schema.
    this.topNodeType = this.nodes[this.spec.topNode || "doc"];

    // :: Object
    // An object for storing whatever values modules may want to
    // compute and cache per schema. (If you want to store something
    // in it, try to use property names unlikely to clash.)
    this.cached = Object.create(null);
    this.cached.wrappings = Object.create(null);
  };

  // :: (union<string, NodeType>, ?Object, ?union<Fragment, Node, [Node]>, ?[Mark]) → Node
  // Create a node in this schema. The `type` may be a string or a
  // `NodeType` instance. Attributes will be extended
  // with defaults, `content` may be a `Fragment`,
  // `null`, a `Node`, or an array of nodes.
  Schema.prototype.node = function node (type, attrs, content, marks) {
    if (typeof type == "string")
      { type = this.nodeType(type); }
    else if (!(type instanceof NodeType$1))
      { throw new RangeError("Invalid node type: " + type) }
    else if (type.schema != this)
      { throw new RangeError("Node type from different schema used (" + type.name + ")") }

    return type.createChecked(attrs, content, marks)
  };

  // :: (string, ?[Mark]) → Node
  // Create a text node in the schema. Empty text nodes are not
  // allowed.
  Schema.prototype.text = function text (text$1, marks) {
    var type = this.nodes.text;
    return new TextNode(type, type.defaultAttrs, text$1, Mark$1.setFrom(marks))
  };

  // :: (union<string, MarkType>, ?Object) → Mark
  // Create a mark with the given type and attributes.
  Schema.prototype.mark = function mark (type, attrs) {
    if (typeof type == "string") { type = this.marks[type]; }
    return type.create(attrs)
  };

  // :: (Object) → Node
  // Deserialize a node from its JSON representation. This method is
  // bound.
  Schema.prototype.nodeFromJSON = function nodeFromJSON (json) {
    return Node$2.fromJSON(this, json)
  };

  // :: (Object) → Mark
  // Deserialize a mark from its JSON representation. This method is
  // bound.
  Schema.prototype.markFromJSON = function markFromJSON (json) {
    return Mark$1.fromJSON(this, json)
  };

  Schema.prototype.nodeType = function nodeType (name) {
    var found = this.nodes[name];
    if (!found) { throw new RangeError("Unknown node type: " + name) }
    return found
  };

  function gatherMarks(schema, marks) {
    var found = [];
    for (var i = 0; i < marks.length; i++) {
      var name = marks[i], mark = schema.marks[name], ok = mark;
      if (mark) {
        found.push(mark);
      } else {
        for (var prop in schema.marks) {
          var mark$1 = schema.marks[prop];
          if (name == "_" || (mark$1.spec.group && mark$1.spec.group.split(" ").indexOf(name) > -1))
            { found.push(ok = mark$1); }
        }
      }
      if (!ok) { throw new SyntaxError("Unknown mark type: '" + marks[i] + "'") }
    }
    return found
  }

  // ParseOptions:: interface
  // These are the options recognized by the
  // [`parse`](#model.DOMParser.parse) and
  // [`parseSlice`](#model.DOMParser.parseSlice) methods.
  //
  //   preserveWhitespace:: ?union<bool, "full">
  //   By default, whitespace is collapsed as per HTML's rules. Pass
  //   `true` to preserve whitespace, but normalize newlines to
  //   spaces, and `"full"` to preserve whitespace entirely.
  //
  //   findPositions:: ?[{node: dom.Node, offset: number}]
  //   When given, the parser will, beside parsing the content,
  //   record the document positions of the given DOM positions. It
  //   will do so by writing to the objects, adding a `pos` property
  //   that holds the document position. DOM positions that are not
  //   in the parsed content will not be written to.
  //
  //   from:: ?number
  //   The child node index to start parsing from.
  //
  //   to:: ?number
  //   The child node index to stop parsing at.
  //
  //   topNode:: ?Node
  //   By default, the content is parsed into the schema's default
  //   [top node type](#model.Schema.topNodeType). You can pass this
  //   option to use the type and attributes from a different node
  //   as the top container.
  //
  //   topMatch:: ?ContentMatch
  //   Provide the starting content match that content parsed into the
  //   top node is matched against.
  //
  //   context:: ?ResolvedPos
  //   A set of additional nodes to count as
  //   [context](#model.ParseRule.context) when parsing, above the
  //   given [top node](#model.ParseOptions.topNode).

  // ParseRule:: interface
  // A value that describes how to parse a given DOM node or inline
  // style as a ProseMirror node or mark.
  //
  //   tag:: ?string
  //   A CSS selector describing the kind of DOM elements to match. A
  //   single rule should have _either_ a `tag` or a `style` property.
  //
  //   namespace:: ?string
  //   The namespace to match. This should be used with `tag`.
  //   Nodes are only matched when the namespace matches or this property
  //   is null.
  //
  //   style:: ?string
  //   A CSS property name to match. When given, this rule matches
  //   inline styles that list that property. May also have the form
  //   `"property=value"`, in which case the rule only matches if the
  //   property's value exactly matches the given value. (For more
  //   complicated filters, use [`getAttrs`](#model.ParseRule.getAttrs)
  //   and return false to indicate that the match failed.) Rules
  //   matching styles may only produce [marks](#model.ParseRule.mark),
  //   not nodes.
  //
  //   priority:: ?number
  //   Can be used to change the order in which the parse rules in a
  //   schema are tried. Those with higher priority come first. Rules
  //   without a priority are counted as having priority 50. This
  //   property is only meaningful in a schema—when directly
  //   constructing a parser, the order of the rule array is used.
  //
  //   consuming:: ?boolean
  //   By default, when a rule matches an element or style, no further
  //   rules get a chance to match it. By setting this to `false`, you
  //   indicate that even when this rule matches, other rules that come
  //   after it should also run.
  //
  //   context:: ?string
  //   When given, restricts this rule to only match when the current
  //   context—the parent nodes into which the content is being
  //   parsed—matches this expression. Should contain one or more node
  //   names or node group names followed by single or double slashes.
  //   For example `"paragraph/"` means the rule only matches when the
  //   parent node is a paragraph, `"blockquote/paragraph/"` restricts
  //   it to be in a paragraph that is inside a blockquote, and
  //   `"section//"` matches any position inside a section—a double
  //   slash matches any sequence of ancestor nodes. To allow multiple
  //   different contexts, they can be separated by a pipe (`|`)
  //   character, as in `"blockquote/|list_item/"`.
  //
  //   node:: ?string
  //   The name of the node type to create when this rule matches. Only
  //   valid for rules with a `tag` property, not for style rules. Each
  //   rule should have one of a `node`, `mark`, or `ignore` property
  //   (except when it appears in a [node](#model.NodeSpec.parseDOM) or
  //   [mark spec](#model.MarkSpec.parseDOM), in which case the `node`
  //   or `mark` property will be derived from its position).
  //
  //   mark:: ?string
  //   The name of the mark type to wrap the matched content in.
  //
  //   ignore:: ?bool
  //   When true, ignore content that matches this rule.
  //
  //   closeParent:: ?bool
  //   When true, finding an element that matches this rule will close
  //   the current node.
  //
  //   skip:: ?bool
  //   When true, ignore the node that matches this rule, but do parse
  //   its content.
  //
  //   attrs:: ?Object
  //   Attributes for the node or mark created by this rule. When
  //   `getAttrs` is provided, it takes precedence.
  //
  //   getAttrs:: ?(union<dom.Node, string>) → ?union<Object, false>
  //   A function used to compute the attributes for the node or mark
  //   created by this rule. Can also be used to describe further
  //   conditions the DOM element or style must match. When it returns
  //   `false`, the rule won't match. When it returns null or undefined,
  //   that is interpreted as an empty/default set of attributes.
  //
  //   Called with a DOM Element for `tag` rules, and with a string (the
  //   style's value) for `style` rules.
  //
  //   contentElement:: ?union<string, (dom.Node) → dom.Node>
  //   For `tag` rules that produce non-leaf nodes or marks, by default
  //   the content of the DOM element is parsed as content of the mark
  //   or node. If the child nodes are in a descendent node, this may be
  //   a CSS selector string that the parser must use to find the actual
  //   content element, or a function that returns the actual content
  //   element to the parser.
  //
  //   getContent:: ?(dom.Node, schema: Schema) → Fragment
  //   Can be used to override the content of a matched node. When
  //   present, instead of parsing the node's child nodes, the result of
  //   this function is used.
  //
  //   preserveWhitespace:: ?union<bool, "full">
  //   Controls whether whitespace should be preserved when parsing the
  //   content inside the matched element. `false` means whitespace may
  //   be collapsed, `true` means that whitespace should be preserved
  //   but newlines normalized to spaces, and `"full"` means that
  //   newlines should also be preserved.

  // ::- A DOM parser represents a strategy for parsing DOM content into
  // a ProseMirror document conforming to a given schema. Its behavior
  // is defined by an array of [rules](#model.ParseRule).
  var DOMParser = function DOMParser(schema, rules) {
    var this$1$1 = this;

    // :: Schema
    // The schema into which the parser parses.
    this.schema = schema;
    // :: [ParseRule]
    // The set of [parse rules](#model.ParseRule) that the parser
    // uses, in order of precedence.
    this.rules = rules;
    this.tags = [];
    this.styles = [];

    rules.forEach(function (rule) {
      if (rule.tag) { this$1$1.tags.push(rule); }
      else if (rule.style) { this$1$1.styles.push(rule); }
    });

    // Only normalize list elements when lists in the schema can't directly contain themselves
    this.normalizeLists = !this.tags.some(function (r) {
      if (!/^(ul|ol)\b/.test(r.tag) || !r.node) { return false }
      var node = schema.nodes[r.node];
      return node.contentMatch.matchType(node)
    });
  };

  // :: (dom.Node, ?ParseOptions) → Node
  // Parse a document from the content of a DOM node.
  DOMParser.prototype.parse = function parse (dom, options) {
      if ( options === void 0 ) { options = {}; }

    var context = new ParseContext(this, options, false);
    context.addAll(dom, null, options.from, options.to);
    return context.finish()
  };

  // :: (dom.Node, ?ParseOptions) → Slice
  // Parses the content of the given DOM node, like
  // [`parse`](#model.DOMParser.parse), and takes the same set of
  // options. But unlike that method, which produces a whole node,
  // this one returns a slice that is open at the sides, meaning that
  // the schema constraints aren't applied to the start of nodes to
  // the left of the input and the end of nodes at the end.
  DOMParser.prototype.parseSlice = function parseSlice (dom, options) {
      if ( options === void 0 ) { options = {}; }

    var context = new ParseContext(this, options, true);
    context.addAll(dom, null, options.from, options.to);
    return Slice$1.maxOpen(context.finish())
  };

  DOMParser.prototype.matchTag = function matchTag (dom, context, after) {
    for (var i = after ? this.tags.indexOf(after) + 1 : 0; i < this.tags.length; i++) {
      var rule = this.tags[i];
      if (matches(dom, rule.tag) &&
          (rule.namespace === undefined || dom.namespaceURI == rule.namespace) &&
          (!rule.context || context.matchesContext(rule.context))) {
        if (rule.getAttrs) {
          var result = rule.getAttrs(dom);
          if (result === false) { continue }
          rule.attrs = result;
        }
        return rule
      }
    }
  };

  DOMParser.prototype.matchStyle = function matchStyle (prop, value, context, after) {
    for (var i = after ? this.styles.indexOf(after) + 1 : 0; i < this.styles.length; i++) {
      var rule = this.styles[i];
      if (rule.style.indexOf(prop) != 0 ||
          rule.context && !context.matchesContext(rule.context) ||
          // Test that the style string either precisely matches the prop,
          // or has an '=' sign after the prop, followed by the given
          // value.
          rule.style.length > prop.length &&
          (rule.style.charCodeAt(prop.length) != 61 || rule.style.slice(prop.length + 1) != value))
        { continue }
      if (rule.getAttrs) {
        var result = rule.getAttrs(value);
        if (result === false) { continue }
        rule.attrs = result;
      }
      return rule
    }
  };

  // : (Schema) → [ParseRule]
  DOMParser.schemaRules = function schemaRules (schema) {
    var result = [];
    function insert(rule) {
      var priority = rule.priority == null ? 50 : rule.priority, i = 0;
      for (; i < result.length; i++) {
        var next = result[i], nextPriority = next.priority == null ? 50 : next.priority;
        if (nextPriority < priority) { break }
      }
      result.splice(i, 0, rule);
    }

    var loop = function ( name ) {
      var rules = schema.marks[name].spec.parseDOM;
      if (rules) { rules.forEach(function (rule) {
        insert(rule = copy$2(rule));
        rule.mark = name;
      }); }
    };

      for (var name in schema.marks) { loop( name ); }
    var loop$1 = function ( name ) {
      var rules$1 = schema.nodes[name$1].spec.parseDOM;
      if (rules$1) { rules$1.forEach(function (rule) {
        insert(rule = copy$2(rule));
        rule.node = name$1;
      }); }
    };

      for (var name$1 in schema.nodes) { loop$1(); }
    return result
  };

  // :: (Schema) → DOMParser
  // Construct a DOM parser using the parsing rules listed in a
  // schema's [node specs](#model.NodeSpec.parseDOM), reordered by
  // [priority](#model.ParseRule.priority).
  DOMParser.fromSchema = function fromSchema (schema) {
    return schema.cached.domParser ||
      (schema.cached.domParser = new DOMParser(schema, DOMParser.schemaRules(schema)))
  };

  // : Object<bool> The block-level tags in HTML5
  var blockTags = {
    address: true, article: true, aside: true, blockquote: true, canvas: true,
    dd: true, div: true, dl: true, fieldset: true, figcaption: true, figure: true,
    footer: true, form: true, h1: true, h2: true, h3: true, h4: true, h5: true,
    h6: true, header: true, hgroup: true, hr: true, li: true, noscript: true, ol: true,
    output: true, p: true, pre: true, section: true, table: true, tfoot: true, ul: true
  };

  // : Object<bool> The tags that we normally ignore.
  var ignoreTags = {
    head: true, noscript: true, object: true, script: true, style: true, title: true
  };

  // : Object<bool> List tags.
  var listTags = {ol: true, ul: true};

  // Using a bitfield for node context options
  var OPT_PRESERVE_WS = 1, OPT_PRESERVE_WS_FULL = 2, OPT_OPEN_LEFT = 4;

  function wsOptionsFor(preserveWhitespace) {
    return (preserveWhitespace ? OPT_PRESERVE_WS : 0) | (preserveWhitespace === "full" ? OPT_PRESERVE_WS_FULL : 0)
  }

  var NodeContext = function NodeContext(type, attrs, marks, pendingMarks, solid, match, options) {
    this.type = type;
    this.attrs = attrs;
    this.solid = solid;
    this.match = match || (options & OPT_OPEN_LEFT ? null : type.contentMatch);
    this.options = options;
    this.content = [];
    // Marks applied to this node itself
    this.marks = marks;
    // Marks applied to its children
    this.activeMarks = Mark$1.none;
    // Marks that can't apply here, but will be used in children if possible
    this.pendingMarks = pendingMarks;
    // Nested Marks with same type
    this.stashMarks = [];
  };

  NodeContext.prototype.findWrapping = function findWrapping (node) {
    if (!this.match) {
      if (!this.type) { return [] }
      var fill = this.type.contentMatch.fillBefore(Fragment$1.from(node));
      if (fill) {
        this.match = this.type.contentMatch.matchFragment(fill);
      } else {
        var start = this.type.contentMatch, wrap;
        if (wrap = start.findWrapping(node.type)) {
          this.match = start;
          return wrap
        } else {
          return null
        }
      }
    }
    return this.match.findWrapping(node.type)
  };

  NodeContext.prototype.finish = function finish (openEnd) {
    if (!(this.options & OPT_PRESERVE_WS)) { // Strip trailing whitespace
      var last = this.content[this.content.length - 1], m;
      if (last && last.isText && (m = /[ \t\r\n\u000c]+$/.exec(last.text))) {
        if (last.text.length == m[0].length) { this.content.pop(); }
        else { this.content[this.content.length - 1] = last.withText(last.text.slice(0, last.text.length - m[0].length)); }
      }
    }
    var content = Fragment$1.from(this.content);
    if (!openEnd && this.match)
      { content = content.append(this.match.fillBefore(Fragment$1.empty, true)); }
    return this.type ? this.type.create(this.attrs, content, this.marks) : content
  };

  NodeContext.prototype.popFromStashMark = function popFromStashMark (mark) {
    for (var i = this.stashMarks.length - 1; i >= 0; i--)
      { if (mark.eq(this.stashMarks[i])) { return this.stashMarks.splice(i, 1)[0] } }
  };

  NodeContext.prototype.applyPending = function applyPending (nextType) {
    for (var i = 0, pending = this.pendingMarks; i < pending.length; i++) {
      var mark = pending[i];
      if ((this.type ? this.type.allowsMarkType(mark.type) : markMayApply(mark.type, nextType)) &&
          !mark.isInSet(this.activeMarks)) {
        this.activeMarks = mark.addToSet(this.activeMarks);
        this.pendingMarks = mark.removeFromSet(this.pendingMarks);
      }
    }
  };

  NodeContext.prototype.inlineContext = function inlineContext (node) {
    if (this.type) { return this.type.inlineContent }
    if (this.content.length) { return this.content[0].isInline }
    return node.parentNode && !blockTags.hasOwnProperty(node.parentNode.nodeName.toLowerCase())
  };

  var ParseContext = function ParseContext(parser, options, open) {
    // : DOMParser The parser we are using.
    this.parser = parser;
    // : Object The options passed to this parse.
    this.options = options;
    this.isOpen = open;
    var topNode = options.topNode, topContext;
    var topOptions = wsOptionsFor(options.preserveWhitespace) | (open ? OPT_OPEN_LEFT : 0);
    if (topNode)
      { topContext = new NodeContext(topNode.type, topNode.attrs, Mark$1.none, Mark$1.none, true,
                                   options.topMatch || topNode.type.contentMatch, topOptions); }
    else if (open)
      { topContext = new NodeContext(null, null, Mark$1.none, Mark$1.none, true, null, topOptions); }
    else
      { topContext = new NodeContext(parser.schema.topNodeType, null, Mark$1.none, Mark$1.none, true, null, topOptions); }
    this.nodes = [topContext];
    // : [Mark] The current set of marks
    this.open = 0;
    this.find = options.findPositions;
    this.needsBlock = false;
  };

  var prototypeAccessors$6$1 = { top: { configurable: true },currentPos: { configurable: true } };

  prototypeAccessors$6$1.top.get = function () {
    return this.nodes[this.open]
  };

  // : (dom.Node)
  // Add a DOM node to the content. Text is inserted as text node,
  // otherwise, the node is passed to `addElement` or, if it has a
  // `style` attribute, `addElementWithStyles`.
  ParseContext.prototype.addDOM = function addDOM (dom) {
    if (dom.nodeType == 3) {
      this.addTextNode(dom);
    } else if (dom.nodeType == 1) {
      var style = dom.getAttribute("style");
      var marks = style ? this.readStyles(parseStyles(style)) : null, top = this.top;
      if (marks != null) { for (var i = 0; i < marks.length; i++) { this.addPendingMark(marks[i]); } }
      this.addElement(dom);
      if (marks != null) { for (var i$1 = 0; i$1 < marks.length; i$1++) { this.removePendingMark(marks[i$1], top); } }
    }
  };

  ParseContext.prototype.addTextNode = function addTextNode (dom) {
    var value = dom.nodeValue;
    var top = this.top;
    if (top.options & OPT_PRESERVE_WS_FULL ||
        top.inlineContext(dom) ||
        /[^ \t\r\n\u000c]/.test(value)) {
      if (!(top.options & OPT_PRESERVE_WS)) {
        value = value.replace(/[ \t\r\n\u000c]+/g, " ");
        // If this starts with whitespace, and there is no node before it, or
        // a hard break, or a text node that ends with whitespace, strip the
        // leading space.
        if (/^[ \t\r\n\u000c]/.test(value) && this.open == this.nodes.length - 1) {
          var nodeBefore = top.content[top.content.length - 1];
          var domNodeBefore = dom.previousSibling;
          if (!nodeBefore ||
              (domNodeBefore && domNodeBefore.nodeName == 'BR') ||
              (nodeBefore.isText && /[ \t\r\n\u000c]$/.test(nodeBefore.text)))
            { value = value.slice(1); }
        }
      } else if (!(top.options & OPT_PRESERVE_WS_FULL)) {
        value = value.replace(/\r?\n|\r/g, " ");
      } else {
        value = value.replace(/\r\n?/g, "\n");
      }
      if (value) { this.insertNode(this.parser.schema.text(value)); }
      this.findInText(dom);
    } else {
      this.findInside(dom);
    }
  };

  // : (dom.Element, ?ParseRule)
  // Try to find a handler for the given tag and use that to parse. If
  // none is found, the element's content nodes are added directly.
  ParseContext.prototype.addElement = function addElement (dom, matchAfter) {
    var name = dom.nodeName.toLowerCase(), ruleID;
    if (listTags.hasOwnProperty(name) && this.parser.normalizeLists) { normalizeList(dom); }
    var rule = (this.options.ruleFromNode && this.options.ruleFromNode(dom)) ||
        (ruleID = this.parser.matchTag(dom, this, matchAfter));
    if (rule ? rule.ignore : ignoreTags.hasOwnProperty(name)) {
      this.findInside(dom);
      this.ignoreFallback(dom);
    } else if (!rule || rule.skip || rule.closeParent) {
      if (rule && rule.closeParent) { this.open = Math.max(0, this.open - 1); }
      else if (rule && rule.skip.nodeType) { dom = rule.skip; }
      var sync, top = this.top, oldNeedsBlock = this.needsBlock;
      if (blockTags.hasOwnProperty(name)) {
        sync = true;
        if (!top.type) { this.needsBlock = true; }
      } else if (!dom.firstChild) {
        this.leafFallback(dom);
        return
      }
      this.addAll(dom);
      if (sync) { this.sync(top); }
      this.needsBlock = oldNeedsBlock;
    } else {
      this.addElementByRule(dom, rule, rule.consuming === false ? ruleID : null);
    }
  };

  // Called for leaf DOM nodes that would otherwise be ignored
  ParseContext.prototype.leafFallback = function leafFallback (dom) {
    if (dom.nodeName == "BR" && this.top.type && this.top.type.inlineContent)
      { this.addTextNode(dom.ownerDocument.createTextNode("\n")); }
  };

  // Called for ignored nodes
  ParseContext.prototype.ignoreFallback = function ignoreFallback (dom) {
    // Ignored BR nodes should at least create an inline context
    if (dom.nodeName == "BR" && (!this.top.type || !this.top.type.inlineContent))
      { this.findPlace(this.parser.schema.text("-")); }
  };

  // Run any style parser associated with the node's styles. Either
  // return an array of marks, or null to indicate some of the styles
  // had a rule with `ignore` set.
  ParseContext.prototype.readStyles = function readStyles (styles) {
    var marks = Mark$1.none;
    style: for (var i = 0; i < styles.length; i += 2) {
      for (var after = null;;) {
        var rule = this.parser.matchStyle(styles[i], styles[i + 1], this, after);
        if (!rule) { continue style }
        if (rule.ignore) { return null }
        marks = this.parser.schema.marks[rule.mark].create(rule.attrs).addToSet(marks);
        if (rule.consuming === false) { after = rule; }
        else { break }
      }
    }
    return marks
  };

  // : (dom.Element, ParseRule) → bool
  // Look up a handler for the given node. If none are found, return
  // false. Otherwise, apply it, use its return value to drive the way
  // the node's content is wrapped, and return true.
  ParseContext.prototype.addElementByRule = function addElementByRule (dom, rule, continueAfter) {
      var this$1$1 = this;

    var sync, nodeType, markType, mark;
    if (rule.node) {
      nodeType = this.parser.schema.nodes[rule.node];
      if (!nodeType.isLeaf) {
        sync = this.enter(nodeType, rule.attrs, rule.preserveWhitespace);
      } else if (!this.insertNode(nodeType.create(rule.attrs))) {
        this.leafFallback(dom);
      }
    } else {
      markType = this.parser.schema.marks[rule.mark];
      mark = markType.create(rule.attrs);
      this.addPendingMark(mark);
    }
    var startIn = this.top;

    if (nodeType && nodeType.isLeaf) {
      this.findInside(dom);
    } else if (continueAfter) {
      this.addElement(dom, continueAfter);
    } else if (rule.getContent) {
      this.findInside(dom);
      rule.getContent(dom, this.parser.schema).forEach(function (node) { return this$1$1.insertNode(node); });
    } else {
      var contentDOM = rule.contentElement;
      if (typeof contentDOM == "string") { contentDOM = dom.querySelector(contentDOM); }
      else if (typeof contentDOM == "function") { contentDOM = contentDOM(dom); }
      if (!contentDOM) { contentDOM = dom; }
      this.findAround(dom, contentDOM, true);
      this.addAll(contentDOM, sync);
    }
    if (sync) { this.sync(startIn); this.open--; }
    if (mark) { this.removePendingMark(mark, startIn); }
  };

  // : (dom.Node, ?NodeBuilder, ?number, ?number)
  // Add all child nodes between `startIndex` and `endIndex` (or the
  // whole node, if not given). If `sync` is passed, use it to
  // synchronize after every block element.
  ParseContext.prototype.addAll = function addAll (parent, sync, startIndex, endIndex) {
    var index = startIndex || 0;
    for (var dom = startIndex ? parent.childNodes[startIndex] : parent.firstChild,
             end = endIndex == null ? null : parent.childNodes[endIndex];
         dom != end; dom = dom.nextSibling, ++index) {
      this.findAtPoint(parent, index);
      this.addDOM(dom);
      if (sync && blockTags.hasOwnProperty(dom.nodeName.toLowerCase()))
        { this.sync(sync); }
    }
    this.findAtPoint(parent, index);
  };

  // Try to find a way to fit the given node type into the current
  // context. May add intermediate wrappers and/or leave non-solid
  // nodes that we're in.
  ParseContext.prototype.findPlace = function findPlace (node) {
    var route, sync;
    for (var depth = this.open; depth >= 0; depth--) {
      var cx = this.nodes[depth];
      var found = cx.findWrapping(node);
      if (found && (!route || route.length > found.length)) {
        route = found;
        sync = cx;
        if (!found.length) { break }
      }
      if (cx.solid) { break }
    }
    if (!route) { return false }
    this.sync(sync);
    for (var i = 0; i < route.length; i++)
      { this.enterInner(route[i], null, false); }
    return true
  };

  // : (Node) → ?Node
  // Try to insert the given node, adjusting the context when needed.
  ParseContext.prototype.insertNode = function insertNode (node) {
    if (node.isInline && this.needsBlock && !this.top.type) {
      var block = this.textblockFromContext();
      if (block) { this.enterInner(block); }
    }
    if (this.findPlace(node)) {
      this.closeExtra();
      var top = this.top;
      top.applyPending(node.type);
      if (top.match) { top.match = top.match.matchType(node.type); }
      var marks = top.activeMarks;
      for (var i = 0; i < node.marks.length; i++)
        { if (!top.type || top.type.allowsMarkType(node.marks[i].type))
          { marks = node.marks[i].addToSet(marks); } }
      top.content.push(node.mark(marks));
      return true
    }
    return false
  };

  // : (NodeType, ?Object) → bool
  // Try to start a node of the given type, adjusting the context when
  // necessary.
  ParseContext.prototype.enter = function enter (type, attrs, preserveWS) {
    var ok = this.findPlace(type.create(attrs));
    if (ok) { this.enterInner(type, attrs, true, preserveWS); }
    return ok
  };

  // Open a node of the given type
  ParseContext.prototype.enterInner = function enterInner (type, attrs, solid, preserveWS) {
    this.closeExtra();
    var top = this.top;
    top.applyPending(type);
    top.match = top.match && top.match.matchType(type, attrs);
    var options = preserveWS == null ? top.options & ~OPT_OPEN_LEFT : wsOptionsFor(preserveWS);
    if ((top.options & OPT_OPEN_LEFT) && top.content.length == 0) { options |= OPT_OPEN_LEFT; }
    this.nodes.push(new NodeContext(type, attrs, top.activeMarks, top.pendingMarks, solid, null, options));
    this.open++;
  };

  // Make sure all nodes above this.open are finished and added to
  // their parents
  ParseContext.prototype.closeExtra = function closeExtra (openEnd) {
    var i = this.nodes.length - 1;
    if (i > this.open) {
      for (; i > this.open; i--) { this.nodes[i - 1].content.push(this.nodes[i].finish(openEnd)); }
      this.nodes.length = this.open + 1;
    }
  };

  ParseContext.prototype.finish = function finish () {
    this.open = 0;
    this.closeExtra(this.isOpen);
    return this.nodes[0].finish(this.isOpen || this.options.topOpen)
  };

  ParseContext.prototype.sync = function sync (to) {
    for (var i = this.open; i >= 0; i--) { if (this.nodes[i] == to) {
      this.open = i;
      return
    } }
  };

  prototypeAccessors$6$1.currentPos.get = function () {
    this.closeExtra();
    var pos = 0;
    for (var i = this.open; i >= 0; i--) {
      var content = this.nodes[i].content;
      for (var j = content.length - 1; j >= 0; j--)
        { pos += content[j].nodeSize; }
      if (i) { pos++; }
    }
    return pos
  };

  ParseContext.prototype.findAtPoint = function findAtPoint (parent, offset) {
    if (this.find) { for (var i = 0; i < this.find.length; i++) {
      if (this.find[i].node == parent && this.find[i].offset == offset)
        { this.find[i].pos = this.currentPos; }
    } }
  };

  ParseContext.prototype.findInside = function findInside (parent) {
    if (this.find) { for (var i = 0; i < this.find.length; i++) {
      if (this.find[i].pos == null && parent.nodeType == 1 && parent.contains(this.find[i].node))
        { this.find[i].pos = this.currentPos; }
    } }
  };

  ParseContext.prototype.findAround = function findAround (parent, content, before) {
    if (parent != content && this.find) { for (var i = 0; i < this.find.length; i++) {
      if (this.find[i].pos == null && parent.nodeType == 1 && parent.contains(this.find[i].node)) {
        var pos = content.compareDocumentPosition(this.find[i].node);
        if (pos & (before ? 2 : 4))
          { this.find[i].pos = this.currentPos; }
      }
    } }
  };

  ParseContext.prototype.findInText = function findInText (textNode) {
    if (this.find) { for (var i = 0; i < this.find.length; i++) {
      if (this.find[i].node == textNode)
        { this.find[i].pos = this.currentPos - (textNode.nodeValue.length - this.find[i].offset); }
    } }
  };

  // : (string) → bool
  // Determines whether the given [context
  // string](#ParseRule.context) matches this context.
  ParseContext.prototype.matchesContext = function matchesContext (context) {
      var this$1$1 = this;

    if (context.indexOf("|") > -1)
      { return context.split(/\s*\|\s*/).some(this.matchesContext, this) }

    var parts = context.split("/");
    var option = this.options.context;
    var useRoot = !this.isOpen && (!option || option.parent.type == this.nodes[0].type);
    var minDepth = -(option ? option.depth + 1 : 0) + (useRoot ? 0 : 1);
    var match = function (i, depth) {
      for (; i >= 0; i--) {
        var part = parts[i];
        if (part == "") {
          if (i == parts.length - 1 || i == 0) { continue }
          for (; depth >= minDepth; depth--)
            { if (match(i - 1, depth)) { return true } }
          return false
        } else {
          var next = depth > 0 || (depth == 0 && useRoot) ? this$1$1.nodes[depth].type
              : option && depth >= minDepth ? option.node(depth - minDepth).type
              : null;
          if (!next || (next.name != part && next.groups.indexOf(part) == -1))
            { return false }
          depth--;
        }
      }
      return true
    };
    return match(parts.length - 1, this.open)
  };

  ParseContext.prototype.textblockFromContext = function textblockFromContext () {
    var $context = this.options.context;
    if ($context) { for (var d = $context.depth; d >= 0; d--) {
      var deflt = $context.node(d).contentMatchAt($context.indexAfter(d)).defaultType;
      if (deflt && deflt.isTextblock && deflt.defaultAttrs) { return deflt }
    } }
    for (var name in this.parser.schema.nodes) {
      var type = this.parser.schema.nodes[name];
      if (type.isTextblock && type.defaultAttrs) { return type }
    }
  };

  ParseContext.prototype.addPendingMark = function addPendingMark (mark) {
    var found = findSameMarkInSet(mark, this.top.pendingMarks);
    if (found) { this.top.stashMarks.push(found); }
    this.top.pendingMarks = mark.addToSet(this.top.pendingMarks);
  };

  ParseContext.prototype.removePendingMark = function removePendingMark (mark, upto) {
    for (var depth = this.open; depth >= 0; depth--) {
      var level = this.nodes[depth];
      var found = level.pendingMarks.lastIndexOf(mark);
      if (found > -1) {
        level.pendingMarks = mark.removeFromSet(level.pendingMarks);
      } else {
        level.activeMarks = mark.removeFromSet(level.activeMarks);
        var stashMark = level.popFromStashMark(mark);
        if (stashMark && level.type && level.type.allowsMarkType(stashMark.type))
          { level.activeMarks = stashMark.addToSet(level.activeMarks); }
      }
      if (level == upto) { break }
    }
  };

  Object.defineProperties( ParseContext.prototype, prototypeAccessors$6$1 );

  // Kludge to work around directly nested list nodes produced by some
  // tools and allowed by browsers to mean that the nested list is
  // actually part of the list item above it.
  function normalizeList(dom) {
    for (var child = dom.firstChild, prevItem = null; child; child = child.nextSibling) {
      var name = child.nodeType == 1 ? child.nodeName.toLowerCase() : null;
      if (name && listTags.hasOwnProperty(name) && prevItem) {
        prevItem.appendChild(child);
        child = prevItem;
      } else if (name == "li") {
        prevItem = child;
      } else if (name) {
        prevItem = null;
      }
    }
  }

  // Apply a CSS selector.
  function matches(dom, selector) {
    return (dom.matches || dom.msMatchesSelector || dom.webkitMatchesSelector || dom.mozMatchesSelector).call(dom, selector)
  }

  // : (string) → [string]
  // Tokenize a style attribute into property/value pairs.
  function parseStyles(style) {
    var re = /\s*([\w-]+)\s*:\s*([^;]+)/g, m, result = [];
    while (m = re.exec(style)) { result.push(m[1], m[2].trim()); }
    return result
  }

  function copy$2(obj) {
    var copy = {};
    for (var prop in obj) { copy[prop] = obj[prop]; }
    return copy
  }

  // Used when finding a mark at the top level of a fragment parse.
  // Checks whether it would be reasonable to apply a given mark type to
  // a given node, by looking at the way the mark occurs in the schema.
  function markMayApply(markType, nodeType) {
    var nodes = nodeType.schema.nodes;
    var loop = function ( name ) {
      var parent = nodes[name];
      if (!parent.allowsMarkType(markType)) { return }
      var seen = [], scan = function (match) {
        seen.push(match);
        for (var i = 0; i < match.edgeCount; i++) {
          var ref = match.edge(i);
          var type = ref.type;
          var next = ref.next;
          if (type == nodeType) { return true }
          if (seen.indexOf(next) < 0 && scan(next)) { return true }
        }
      };
      if (scan(parent.contentMatch)) { return { v: true } }
    };

    for (var name in nodes) {
      var returned = loop( name );

      if ( returned ) { return returned.v; }
    }
  }

  function findSameMarkInSet(mark, set) {
    for (var i = 0; i < set.length; i++) {
      if (mark.eq(set[i])) { return set[i] }
    }
  }

  // DOMOutputSpec:: interface
  // A description of a DOM structure. Can be either a string, which is
  // interpreted as a text node, a DOM node, which is interpreted as
  // itself, a `{dom: Node, contentDOM: ?Node}` object, or an array.
  //
  // An array describes a DOM element. The first value in the array
  // should be a string—the name of the DOM element, optionally prefixed
  // by a namespace URL and a space. If the second element is plain
  // object, it is interpreted as a set of attributes for the element.
  // Any elements after that (including the 2nd if it's not an attribute
  // object) are interpreted as children of the DOM elements, and must
  // either be valid `DOMOutputSpec` values, or the number zero.
  //
  // The number zero (pronounced “hole”) is used to indicate the place
  // where a node's child nodes should be inserted. If it occurs in an
  // output spec, it should be the only child element in its parent
  // node.

  // ::- A DOM serializer knows how to convert ProseMirror nodes and
  // marks of various types to DOM nodes.
  var DOMSerializer = function DOMSerializer(nodes, marks) {
    // :: Object<(node: Node) → DOMOutputSpec>
    // The node serialization functions.
    this.nodes = nodes || {};
    // :: Object<?(mark: Mark, inline: bool) → DOMOutputSpec>
    // The mark serialization functions.
    this.marks = marks || {};
  };

  // :: (Fragment, ?Object) → dom.DocumentFragment
  // Serialize the content of this fragment to a DOM fragment. When
  // not in the browser, the `document` option, containing a DOM
  // document, should be passed so that the serializer can create
  // nodes.
  DOMSerializer.prototype.serializeFragment = function serializeFragment (fragment, options, target) {
      var this$1$1 = this;
      if ( options === void 0 ) { options = {}; }

    if (!target) { target = doc$1(options).createDocumentFragment(); }

    var top = target, active = null;
    fragment.forEach(function (node) {
      if (active || node.marks.length) {
        if (!active) { active = []; }
        var keep = 0, rendered = 0;
        while (keep < active.length && rendered < node.marks.length) {
          var next = node.marks[rendered];
          if (!this$1$1.marks[next.type.name]) { rendered++; continue }
          if (!next.eq(active[keep]) || next.type.spec.spanning === false) { break }
          keep += 2; rendered++;
        }
        while (keep < active.length) {
          top = active.pop();
          active.pop();
        }
        while (rendered < node.marks.length) {
          var add = node.marks[rendered++];
          var markDOM = this$1$1.serializeMark(add, node.isInline, options);
          if (markDOM) {
            active.push(add, top);
            top.appendChild(markDOM.dom);
            top = markDOM.contentDOM || markDOM.dom;
          }
        }
      }
      top.appendChild(this$1$1.serializeNodeInner(node, options));
    });

    return target
  };

  DOMSerializer.prototype.serializeNodeInner = function serializeNodeInner (node, options) {
      if ( options === void 0 ) { options = {}; }

    var ref =
        DOMSerializer.renderSpec(doc$1(options), this.nodes[node.type.name](node));
      var dom = ref.dom;
      var contentDOM = ref.contentDOM;
    if (contentDOM) {
      if (node.isLeaf)
        { throw new RangeError("Content hole not allowed in a leaf node spec") }
      if (options.onContent)
        { options.onContent(node, contentDOM, options); }
      else
        { this.serializeFragment(node.content, options, contentDOM); }
    }
    return dom
  };

  // :: (Node, ?Object) → dom.Node
  // Serialize this node to a DOM node. This can be useful when you
  // need to serialize a part of a document, as opposed to the whole
  // document. To serialize a whole document, use
  // [`serializeFragment`](#model.DOMSerializer.serializeFragment) on
  // its [content](#model.Node.content).
  DOMSerializer.prototype.serializeNode = function serializeNode (node, options) {
      if ( options === void 0 ) { options = {}; }

    var dom = this.serializeNodeInner(node, options);
    for (var i = node.marks.length - 1; i >= 0; i--) {
      var wrap = this.serializeMark(node.marks[i], node.isInline, options);
      if (wrap) {
  (wrap.contentDOM || wrap.dom).appendChild(dom);
        dom = wrap.dom;
      }
    }
    return dom
  };

  DOMSerializer.prototype.serializeMark = function serializeMark (mark, inline, options) {
      if ( options === void 0 ) { options = {}; }

    var toDOM = this.marks[mark.type.name];
    return toDOM && DOMSerializer.renderSpec(doc$1(options), toDOM(mark, inline))
  };

  // :: (dom.Document, DOMOutputSpec) → {dom: dom.Node, contentDOM: ?dom.Node}
  // Render an [output spec](#model.DOMOutputSpec) to a DOM node. If
  // the spec has a hole (zero) in it, `contentDOM` will point at the
  // node with the hole.
  DOMSerializer.renderSpec = function renderSpec (doc, structure, xmlNS) {
      if ( xmlNS === void 0 ) { xmlNS = null; }

    if (typeof structure == "string")
      { return {dom: doc.createTextNode(structure)} }
    if (structure.nodeType != null)
      { return {dom: structure} }
    if (structure.dom && structure.dom.nodeType != null)
      { return structure }
    var tagName = structure[0], space = tagName.indexOf(" ");
    if (space > 0) {
      xmlNS = tagName.slice(0, space);
      tagName = tagName.slice(space + 1);
    }
    var contentDOM = null, dom = xmlNS ? doc.createElementNS(xmlNS, tagName) : doc.createElement(tagName);
    var attrs = structure[1], start = 1;
    if (attrs && typeof attrs == "object" && attrs.nodeType == null && !Array.isArray(attrs)) {
      start = 2;
      for (var name in attrs) { if (attrs[name] != null) {
        var space$1 = name.indexOf(" ");
        if (space$1 > 0) { dom.setAttributeNS(name.slice(0, space$1), name.slice(space$1 + 1), attrs[name]); }
        else { dom.setAttribute(name, attrs[name]); }
      } }
    }
    for (var i = start; i < structure.length; i++) {
      var child = structure[i];
      if (child === 0) {
        if (i < structure.length - 1 || i > start)
          { throw new RangeError("Content hole must be the only child of its parent node") }
        return {dom: dom, contentDOM: dom}
      } else {
        var ref = DOMSerializer.renderSpec(doc, child, xmlNS);
          var inner = ref.dom;
          var innerContent = ref.contentDOM;
        dom.appendChild(inner);
        if (innerContent) {
          if (contentDOM) { throw new RangeError("Multiple content holes") }
          contentDOM = innerContent;
        }
      }
    }
    return {dom: dom, contentDOM: contentDOM}
  };

  // :: (Schema) → DOMSerializer
  // Build a serializer using the [`toDOM`](#model.NodeSpec.toDOM)
  // properties in a schema's node and mark specs.
  DOMSerializer.fromSchema = function fromSchema (schema) {
    return schema.cached.domSerializer ||
      (schema.cached.domSerializer = new DOMSerializer(this.nodesFromSchema(schema), this.marksFromSchema(schema)))
  };

  // : (Schema) → Object<(node: Node) → DOMOutputSpec>
  // Gather the serializers in a schema's node specs into an object.
  // This can be useful as a base to build a custom serializer from.
  DOMSerializer.nodesFromSchema = function nodesFromSchema (schema) {
    var result = gatherToDOM(schema.nodes);
    if (!result.text) { result.text = function (node) { return node.text; }; }
    return result
  };

  // : (Schema) → Object<(mark: Mark) → DOMOutputSpec>
  // Gather the serializers in a schema's mark specs into an object.
  DOMSerializer.marksFromSchema = function marksFromSchema (schema) {
    return gatherToDOM(schema.marks)
  };

  function gatherToDOM(obj) {
    var result = {};
    for (var name in obj) {
      var toDOM = obj[name].spec.toDOM;
      if (toDOM) { result[name] = toDOM; }
    }
    return result
  }

  function doc$1(options) {
    // declare global: window
    return options.document || window.document
  }

  var model = /*#__PURE__*/Object.freeze({
    __proto__: null,
    ContentMatch: ContentMatch,
    DOMParser: DOMParser,
    DOMSerializer: DOMSerializer,
    Fragment: Fragment$1,
    Mark: Mark$1,
    MarkType: MarkType,
    Node: Node$2,
    NodeRange: NodeRange$1,
    NodeType: NodeType$1,
    ReplaceError: ReplaceError$1,
    ResolvedPos: ResolvedPos$1,
    Schema: Schema,
    Slice: Slice$1
  });

  // Mappable:: interface
  // There are several things that positions can be mapped through.
  // Such objects conform to this interface.
  //
  //   map:: (pos: number, assoc: ?number) → number
  //   Map a position through this object. When given, `assoc` (should
  //   be -1 or 1, defaults to 1) determines with which side the
  //   position is associated, which determines in which direction to
  //   move when a chunk of content is inserted at the mapped position.
  //
  //   mapResult:: (pos: number, assoc: ?number) → MapResult
  //   Map a position, and return an object containing additional
  //   information about the mapping. The result's `deleted` field tells
  //   you whether the position was deleted (completely enclosed in a
  //   replaced range) during the mapping. When content on only one side
  //   is deleted, the position itself is only considered deleted when
  //   `assoc` points in the direction of the deleted content.

  // Recovery values encode a range index and an offset. They are
  // represented as numbers, because tons of them will be created when
  // mapping, for example, a large number of decorations. The number's
  // lower 16 bits provide the index, the remaining bits the offset.
  //
  // Note: We intentionally don't use bit shift operators to en- and
  // decode these, since those clip to 32 bits, which we might in rare
  // cases want to overflow. A 64-bit float can represent 48-bit
  // integers precisely.

  var lower16 = 0xffff;
  var factor16 = Math.pow(2, 16);

  function makeRecover(index, offset) { return index + offset * factor16 }
  function recoverIndex(value) { return value & lower16 }
  function recoverOffset(value) { return (value - (value & lower16)) / factor16 }

  // ::- An object representing a mapped position with extra
  // information.
  var MapResult = function MapResult(pos, deleted, recover) {
    if ( deleted === void 0 ) { deleted = false; }
    if ( recover === void 0 ) { recover = null; }

    // :: number The mapped version of the position.
    this.pos = pos;
    // :: bool Tells you whether the position was deleted, that is,
    // whether the step removed its surroundings from the document.
    this.deleted = deleted;
    this.recover = recover;
  };

  // :: class extends Mappable
  // A map describing the deletions and insertions made by a step, which
  // can be used to find the correspondence between positions in the
  // pre-step version of a document and the same position in the
  // post-step version.
  var StepMap = function StepMap(ranges, inverted) {
    if ( inverted === void 0 ) { inverted = false; }

    this.ranges = ranges;
    this.inverted = inverted;
  };

  StepMap.prototype.recover = function recover (value) {
    var diff = 0, index = recoverIndex(value);
    if (!this.inverted) { for (var i = 0; i < index; i++)
      { diff += this.ranges[i * 3 + 2] - this.ranges[i * 3 + 1]; } }
    return this.ranges[index * 3] + diff + recoverOffset(value)
  };

  // : (number, ?number) → MapResult
  StepMap.prototype.mapResult = function mapResult (pos, assoc) {
    if ( assoc === void 0 ) { assoc = 1; }
   return this._map(pos, assoc, false) };

  // : (number, ?number) → number
  StepMap.prototype.map = function map (pos, assoc) {
    if ( assoc === void 0 ) { assoc = 1; }
   return this._map(pos, assoc, true) };

  StepMap.prototype._map = function _map (pos, assoc, simple) {
    var diff = 0, oldIndex = this.inverted ? 2 : 1, newIndex = this.inverted ? 1 : 2;
    for (var i = 0; i < this.ranges.length; i += 3) {
      var start = this.ranges[i] - (this.inverted ? diff : 0);
      if (start > pos) { break }
      var oldSize = this.ranges[i + oldIndex], newSize = this.ranges[i + newIndex], end = start + oldSize;
      if (pos <= end) {
        var side = !oldSize ? assoc : pos == start ? -1 : pos == end ? 1 : assoc;
        var result = start + diff + (side < 0 ? 0 : newSize);
        if (simple) { return result }
        var recover = pos == (assoc < 0 ? start : end) ? null : makeRecover(i / 3, pos - start);
        return new MapResult(result, assoc < 0 ? pos != start : pos != end, recover)
      }
      diff += newSize - oldSize;
    }
    return simple ? pos + diff : new MapResult(pos + diff)
  };

  StepMap.prototype.touches = function touches (pos, recover) {
    var diff = 0, index = recoverIndex(recover);
    var oldIndex = this.inverted ? 2 : 1, newIndex = this.inverted ? 1 : 2;
    for (var i = 0; i < this.ranges.length; i += 3) {
      var start = this.ranges[i] - (this.inverted ? diff : 0);
      if (start > pos) { break }
      var oldSize = this.ranges[i + oldIndex], end = start + oldSize;
      if (pos <= end && i == index * 3) { return true }
      diff += this.ranges[i + newIndex] - oldSize;
    }
    return false
  };

  // :: ((oldStart: number, oldEnd: number, newStart: number, newEnd: number))
  // Calls the given function on each of the changed ranges included in
  // this map.
  StepMap.prototype.forEach = function forEach (f) {
    var oldIndex = this.inverted ? 2 : 1, newIndex = this.inverted ? 1 : 2;
    for (var i = 0, diff = 0; i < this.ranges.length; i += 3) {
      var start = this.ranges[i], oldStart = start - (this.inverted ? diff : 0), newStart = start + (this.inverted ? 0 : diff);
      var oldSize = this.ranges[i + oldIndex], newSize = this.ranges[i + newIndex];
      f(oldStart, oldStart + oldSize, newStart, newStart + newSize);
      diff += newSize - oldSize;
    }
  };

  // :: () → StepMap
  // Create an inverted version of this map. The result can be used to
  // map positions in the post-step document to the pre-step document.
  StepMap.prototype.invert = function invert () {
    return new StepMap(this.ranges, !this.inverted)
  };

  StepMap.prototype.toString = function toString () {
    return (this.inverted ? "-" : "") + JSON.stringify(this.ranges)
  };

  // :: (n: number) → StepMap
  // Create a map that moves all positions by offset `n` (which may be
  // negative). This can be useful when applying steps meant for a
  // sub-document to a larger document, or vice-versa.
  StepMap.offset = function offset (n) {
    return n == 0 ? StepMap.empty : new StepMap(n < 0 ? [0, -n, 0] : [0, 0, n])
  };

  StepMap.empty = new StepMap([]);

  // :: class extends Mappable
  // A mapping represents a pipeline of zero or more [step
  // maps](#transform.StepMap). It has special provisions for losslessly
  // handling mapping positions through a series of steps in which some
  // steps are inverted versions of earlier steps. (This comes up when
  // ‘[rebasing](/docs/guide/#transform.rebasing)’ steps for
  // collaboration or history management.)
  var Mapping = function Mapping(maps, mirror, from, to) {
    // :: [StepMap]
    // The step maps in this mapping.
    this.maps = maps || [];
    // :: number
    // The starting position in the `maps` array, used when `map` or
    // `mapResult` is called.
    this.from = from || 0;
    // :: number
    // The end position in the `maps` array.
    this.to = to == null ? this.maps.length : to;
    this.mirror = mirror;
  };

  // :: (?number, ?number) → Mapping
  // Create a mapping that maps only through a part of this one.
  Mapping.prototype.slice = function slice (from, to) {
      if ( from === void 0 ) { from = 0; }
      if ( to === void 0 ) { to = this.maps.length; }

    return new Mapping(this.maps, this.mirror, from, to)
  };

  Mapping.prototype.copy = function copy () {
    return new Mapping(this.maps.slice(), this.mirror && this.mirror.slice(), this.from, this.to)
  };

  // :: (StepMap, ?number)
  // Add a step map to the end of this mapping. If `mirrors` is
  // given, it should be the index of the step map that is the mirror
  // image of this one.
  Mapping.prototype.appendMap = function appendMap (map, mirrors) {
    this.to = this.maps.push(map);
    if (mirrors != null) { this.setMirror(this.maps.length - 1, mirrors); }
  };

  // :: (Mapping)
  // Add all the step maps in a given mapping to this one (preserving
  // mirroring information).
  Mapping.prototype.appendMapping = function appendMapping (mapping) {
    for (var i = 0, startSize = this.maps.length; i < mapping.maps.length; i++) {
      var mirr = mapping.getMirror(i);
      this.appendMap(mapping.maps[i], mirr != null && mirr < i ? startSize + mirr : null);
    }
  };

  // :: (number) → ?number
  // Finds the offset of the step map that mirrors the map at the
  // given offset, in this mapping (as per the second argument to
  // `appendMap`).
  Mapping.prototype.getMirror = function getMirror (n) {
    if (this.mirror) { for (var i = 0; i < this.mirror.length; i++)
      { if (this.mirror[i] == n) { return this.mirror[i + (i % 2 ? -1 : 1)] } } }
  };

  Mapping.prototype.setMirror = function setMirror (n, m) {
    if (!this.mirror) { this.mirror = []; }
    this.mirror.push(n, m);
  };

  // :: (Mapping)
  // Append the inverse of the given mapping to this one.
  Mapping.prototype.appendMappingInverted = function appendMappingInverted (mapping) {
    for (var i = mapping.maps.length - 1, totalSize = this.maps.length + mapping.maps.length; i >= 0; i--) {
      var mirr = mapping.getMirror(i);
      this.appendMap(mapping.maps[i].invert(), mirr != null && mirr > i ? totalSize - mirr - 1 : null);
    }
  };

  // :: () → Mapping
  // Create an inverted version of this mapping.
  Mapping.prototype.invert = function invert () {
    var inverse = new Mapping;
    inverse.appendMappingInverted(this);
    return inverse
  };

  // : (number, ?number) → number
  // Map a position through this mapping.
  Mapping.prototype.map = function map (pos, assoc) {
      if ( assoc === void 0 ) { assoc = 1; }

    if (this.mirror) { return this._map(pos, assoc, true) }
    for (var i = this.from; i < this.to; i++)
      { pos = this.maps[i].map(pos, assoc); }
    return pos
  };

  // : (number, ?number) → MapResult
  // Map a position through this mapping, returning a mapping
  // result.
  Mapping.prototype.mapResult = function mapResult (pos, assoc) {
    if ( assoc === void 0 ) { assoc = 1; }
   return this._map(pos, assoc, false) };

  Mapping.prototype._map = function _map (pos, assoc, simple) {
    var deleted = false;

    for (var i = this.from; i < this.to; i++) {
      var map = this.maps[i], result = map.mapResult(pos, assoc);
      if (result.recover != null) {
        var corr = this.getMirror(i);
        if (corr != null && corr > i && corr < this.to) {
          i = corr;
          pos = this.maps[corr].recover(result.recover);
          continue
        }
      }

      if (result.deleted) { deleted = true; }
      pos = result.pos;
    }

    return simple ? pos : new MapResult(pos, deleted)
  };

  function TransformError(message) {
    var err = Error.call(this, message);
    err.__proto__ = TransformError.prototype;
    return err
  }

  TransformError.prototype = Object.create(Error.prototype);
  TransformError.prototype.constructor = TransformError;
  TransformError.prototype.name = "TransformError";

  // ::- Abstraction to build up and track an array of
  // [steps](#transform.Step) representing a document transformation.
  //
  // Most transforming methods return the `Transform` object itself, so
  // that they can be chained.
  var Transform = function Transform(doc) {
    // :: Node
    // The current document (the result of applying the steps in the
    // transform).
    this.doc = doc;
    // :: [Step]
    // The steps in this transform.
    this.steps = [];
    // :: [Node]
    // The documents before each of the steps.
    this.docs = [];
    // :: Mapping
    // A mapping with the maps for each of the steps in this transform.
    this.mapping = new Mapping;
  };

  var prototypeAccessors$6 = { before: { configurable: true },docChanged: { configurable: true } };

  // :: Node The starting document.
  prototypeAccessors$6.before.get = function () { return this.docs.length ? this.docs[0] : this.doc };

  // :: (step: Step) → this
  // Apply a new step in this transform, saving the result. Throws an
  // error when the step fails.
  Transform.prototype.step = function step (object) {
    var result = this.maybeStep(object);
    if (result.failed) { throw new TransformError(result.failed) }
    return this
  };

  // :: (Step) → StepResult
  // Try to apply a step in this transformation, ignoring it if it
  // fails. Returns the step result.
  Transform.prototype.maybeStep = function maybeStep (step) {
    var result = step.apply(this.doc);
    if (!result.failed) { this.addStep(step, result.doc); }
    return result
  };

  // :: bool
  // True when the document has been changed (when there are any
  // steps).
  prototypeAccessors$6.docChanged.get = function () {
    return this.steps.length > 0
  };

  Transform.prototype.addStep = function addStep (step, doc) {
    this.docs.push(this.doc);
    this.steps.push(step);
    this.mapping.appendMap(step.getMap());
    this.doc = doc;
  };

  Object.defineProperties( Transform.prototype, prototypeAccessors$6 );

  function mustOverride() { throw new Error("Override me") }

  var stepsByID = Object.create(null);

  // ::- A step object represents an atomic change. It generally applies
  // only to the document it was created for, since the positions
  // stored in it will only make sense for that document.
  //
  // New steps are defined by creating classes that extend `Step`,
  // overriding the `apply`, `invert`, `map`, `getMap` and `fromJSON`
  // methods, and registering your class with a unique
  // JSON-serialization identifier using
  // [`Step.jsonID`](#transform.Step^jsonID).
  var Step = function Step () {};

  Step.prototype.apply = function apply (_doc) { return mustOverride() };

  // :: () → StepMap
  // Get the step map that represents the changes made by this step,
  // and which can be used to transform between positions in the old
  // and the new document.
  Step.prototype.getMap = function getMap () { return StepMap.empty };

  // :: (doc: Node) → Step
  // Create an inverted version of this step. Needs the document as it
  // was before the step as argument.
  Step.prototype.invert = function invert (_doc) { return mustOverride() };

  // :: (mapping: Mappable) → ?Step
  // Map this step through a mappable thing, returning either a
  // version of that step with its positions adjusted, or `null` if
  // the step was entirely deleted by the mapping.
  Step.prototype.map = function map (_mapping) { return mustOverride() };

  // :: (other: Step) → ?Step
  // Try to merge this step with another one, to be applied directly
  // after it. Returns the merged step when possible, null if the
  // steps can't be merged.
  Step.prototype.merge = function merge (_other) { return null };

  // :: () → Object
  // Create a JSON-serializeable representation of this step. When
  // defining this for a custom subclass, make sure the result object
  // includes the step type's [JSON id](#transform.Step^jsonID) under
  // the `stepType` property.
  Step.prototype.toJSON = function toJSON () { return mustOverride() };

  // :: (Schema, Object) → Step
  // Deserialize a step from its JSON representation. Will call
  // through to the step class' own implementation of this method.
  Step.fromJSON = function fromJSON (schema, json) {
    if (!json || !json.stepType) { throw new RangeError("Invalid input for Step.fromJSON") }
    var type = stepsByID[json.stepType];
    if (!type) { throw new RangeError(("No step type " + (json.stepType) + " defined")) }
    return type.fromJSON(schema, json)
  };

  // :: (string, constructor<Step>)
  // To be able to serialize steps to JSON, each step needs a string
  // ID to attach to its JSON representation. Use this method to
  // register an ID for your step classes. Try to pick something
  // that's unlikely to clash with steps from other modules.
  Step.jsonID = function jsonID (id, stepClass) {
    if (id in stepsByID) { throw new RangeError("Duplicate use of step JSON ID " + id) }
    stepsByID[id] = stepClass;
    stepClass.prototype.jsonID = id;
    return stepClass
  };

  // ::- The result of [applying](#transform.Step.apply) a step. Contains either a
  // new document or a failure value.
  var StepResult = function StepResult(doc, failed) {
    // :: ?Node The transformed document.
    this.doc = doc;
    // :: ?string Text providing information about a failed step.
    this.failed = failed;
  };

  // :: (Node) → StepResult
  // Create a successful step result.
  StepResult.ok = function ok (doc) { return new StepResult(doc, null) };

  // :: (string) → StepResult
  // Create a failed step result.
  StepResult.fail = function fail (message) { return new StepResult(null, message) };

  // :: (Node, number, number, Slice) → StepResult
  // Call [`Node.replace`](#model.Node.replace) with the given
  // arguments. Create a successful result if it succeeds, and a
  // failed one if it throws a `ReplaceError`.
  StepResult.fromReplace = function fromReplace (doc, from, to, slice) {
    try {
      return StepResult.ok(doc.replace(from, to, slice))
    } catch (e) {
      if (e instanceof ReplaceError$1) { return StepResult.fail(e.message) }
      throw e
    }
  };

  // ::- Replace a part of the document with a slice of new content.
  var ReplaceStep = /*@__PURE__*/(function (Step) {
    function ReplaceStep(from, to, slice, structure) {
      Step.call(this);
      // :: number
      // The start position of the replaced range.
      this.from = from;
      // :: number
      // The end position of the replaced range.
      this.to = to;
      // :: Slice
      // The slice to insert.
      this.slice = slice;
      this.structure = !!structure;
    }

    if ( Step ) { ReplaceStep.__proto__ = Step; }
    ReplaceStep.prototype = Object.create( Step && Step.prototype );
    ReplaceStep.prototype.constructor = ReplaceStep;

    ReplaceStep.prototype.apply = function apply (doc) {
      if (this.structure && contentBetween(doc, this.from, this.to))
        { return StepResult.fail("Structure replace would overwrite content") }
      return StepResult.fromReplace(doc, this.from, this.to, this.slice)
    };

    ReplaceStep.prototype.getMap = function getMap () {
      return new StepMap([this.from, this.to - this.from, this.slice.size])
    };

    ReplaceStep.prototype.invert = function invert (doc) {
      return new ReplaceStep(this.from, this.from + this.slice.size, doc.slice(this.from, this.to))
    };

    ReplaceStep.prototype.map = function map (mapping) {
      var from = mapping.mapResult(this.from, 1), to = mapping.mapResult(this.to, -1);
      if (from.deleted && to.deleted) { return null }
      return new ReplaceStep(from.pos, Math.max(from.pos, to.pos), this.slice)
    };

    ReplaceStep.prototype.merge = function merge (other) {
      if (!(other instanceof ReplaceStep) || other.structure || this.structure) { return null }

      if (this.from + this.slice.size == other.from && !this.slice.openEnd && !other.slice.openStart) {
        var slice = this.slice.size + other.slice.size == 0 ? Slice$1.empty
            : new Slice$1(this.slice.content.append(other.slice.content), this.slice.openStart, other.slice.openEnd);
        return new ReplaceStep(this.from, this.to + (other.to - other.from), slice, this.structure)
      } else if (other.to == this.from && !this.slice.openStart && !other.slice.openEnd) {
        var slice$1 = this.slice.size + other.slice.size == 0 ? Slice$1.empty
            : new Slice$1(other.slice.content.append(this.slice.content), other.slice.openStart, this.slice.openEnd);
        return new ReplaceStep(other.from, this.to, slice$1, this.structure)
      } else {
        return null
      }
    };

    ReplaceStep.prototype.toJSON = function toJSON () {
      var json = {stepType: "replace", from: this.from, to: this.to};
      if (this.slice.size) { json.slice = this.slice.toJSON(); }
      if (this.structure) { json.structure = true; }
      return json
    };

    ReplaceStep.fromJSON = function fromJSON (schema, json) {
      if (typeof json.from != "number" || typeof json.to != "number")
        { throw new RangeError("Invalid input for ReplaceStep.fromJSON") }
      return new ReplaceStep(json.from, json.to, Slice$1.fromJSON(schema, json.slice), !!json.structure)
    };

    return ReplaceStep;
  }(Step));

  Step.jsonID("replace", ReplaceStep);

  // ::- Replace a part of the document with a slice of content, but
  // preserve a range of the replaced content by moving it into the
  // slice.
  var ReplaceAroundStep = /*@__PURE__*/(function (Step) {
    function ReplaceAroundStep(from, to, gapFrom, gapTo, slice, insert, structure) {
      Step.call(this);
      // :: number
      // The start position of the replaced range.
      this.from = from;
      // :: number
      // The end position of the replaced range.
      this.to = to;
      // :: number
      // The start of preserved range.
      this.gapFrom = gapFrom;
      // :: number
      // The end of preserved range.
      this.gapTo = gapTo;
      // :: Slice
      // The slice to insert.
      this.slice = slice;
      // :: number
      // The position in the slice where the preserved range should be
      // inserted.
      this.insert = insert;
      this.structure = !!structure;
    }

    if ( Step ) { ReplaceAroundStep.__proto__ = Step; }
    ReplaceAroundStep.prototype = Object.create( Step && Step.prototype );
    ReplaceAroundStep.prototype.constructor = ReplaceAroundStep;

    ReplaceAroundStep.prototype.apply = function apply (doc) {
      if (this.structure && (contentBetween(doc, this.from, this.gapFrom) ||
                             contentBetween(doc, this.gapTo, this.to)))
        { return StepResult.fail("Structure gap-replace would overwrite content") }

      var gap = doc.slice(this.gapFrom, this.gapTo);
      if (gap.openStart || gap.openEnd)
        { return StepResult.fail("Gap is not a flat range") }
      var inserted = this.slice.insertAt(this.insert, gap.content);
      if (!inserted) { return StepResult.fail("Content does not fit in gap") }
      return StepResult.fromReplace(doc, this.from, this.to, inserted)
    };

    ReplaceAroundStep.prototype.getMap = function getMap () {
      return new StepMap([this.from, this.gapFrom - this.from, this.insert,
                          this.gapTo, this.to - this.gapTo, this.slice.size - this.insert])
    };

    ReplaceAroundStep.prototype.invert = function invert (doc) {
      var gap = this.gapTo - this.gapFrom;
      return new ReplaceAroundStep(this.from, this.from + this.slice.size + gap,
                                   this.from + this.insert, this.from + this.insert + gap,
                                   doc.slice(this.from, this.to).removeBetween(this.gapFrom - this.from, this.gapTo - this.from),
                                   this.gapFrom - this.from, this.structure)
    };

    ReplaceAroundStep.prototype.map = function map (mapping) {
      var from = mapping.mapResult(this.from, 1), to = mapping.mapResult(this.to, -1);
      var gapFrom = mapping.map(this.gapFrom, -1), gapTo = mapping.map(this.gapTo, 1);
      if ((from.deleted && to.deleted) || gapFrom < from.pos || gapTo > to.pos) { return null }
      return new ReplaceAroundStep(from.pos, to.pos, gapFrom, gapTo, this.slice, this.insert, this.structure)
    };

    ReplaceAroundStep.prototype.toJSON = function toJSON () {
      var json = {stepType: "replaceAround", from: this.from, to: this.to,
                  gapFrom: this.gapFrom, gapTo: this.gapTo, insert: this.insert};
      if (this.slice.size) { json.slice = this.slice.toJSON(); }
      if (this.structure) { json.structure = true; }
      return json
    };

    ReplaceAroundStep.fromJSON = function fromJSON (schema, json) {
      if (typeof json.from != "number" || typeof json.to != "number" ||
          typeof json.gapFrom != "number" || typeof json.gapTo != "number" || typeof json.insert != "number")
        { throw new RangeError("Invalid input for ReplaceAroundStep.fromJSON") }
      return new ReplaceAroundStep(json.from, json.to, json.gapFrom, json.gapTo,
                                   Slice$1.fromJSON(schema, json.slice), json.insert, !!json.structure)
    };

    return ReplaceAroundStep;
  }(Step));

  Step.jsonID("replaceAround", ReplaceAroundStep);

  function contentBetween(doc, from, to) {
    var $from = doc.resolve(from), dist = to - from, depth = $from.depth;
    while (dist > 0 && depth > 0 && $from.indexAfter(depth) == $from.node(depth).childCount) {
      depth--;
      dist--;
    }
    if (dist > 0) {
      var next = $from.node(depth).maybeChild($from.indexAfter(depth));
      while (dist > 0) {
        if (!next || next.isLeaf) { return true }
        next = next.firstChild;
        dist--;
      }
    }
    return false
  }

  function canCut(node, start, end) {
    return (start == 0 || node.canReplace(start, node.childCount)) &&
      (end == node.childCount || node.canReplace(0, end))
  }

  // :: (NodeRange) → ?number
  // Try to find a target depth to which the content in the given range
  // can be lifted. Will not go across
  // [isolating](#model.NodeSpec.isolating) parent nodes.
  function liftTarget(range) {
    var parent = range.parent;
    var content = parent.content.cutByIndex(range.startIndex, range.endIndex);
    for (var depth = range.depth;; --depth) {
      var node = range.$from.node(depth);
      var index = range.$from.index(depth), endIndex = range.$to.indexAfter(depth);
      if (depth < range.depth && node.canReplace(index, endIndex, content))
        { return depth }
      if (depth == 0 || node.type.spec.isolating || !canCut(node, index, endIndex)) { break }
    }
  }

  // :: (NodeRange, number) → this
  // Split the content in the given range off from its parent, if there
  // is sibling content before or after it, and move it up the tree to
  // the depth specified by `target`. You'll probably want to use
  // [`liftTarget`](#transform.liftTarget) to compute `target`, to make
  // sure the lift is valid.
  Transform.prototype.lift = function(range, target) {
    var $from = range.$from;
    var $to = range.$to;
    var depth = range.depth;

    var gapStart = $from.before(depth + 1), gapEnd = $to.after(depth + 1);
    var start = gapStart, end = gapEnd;

    var before = Fragment$1.empty, openStart = 0;
    for (var d = depth, splitting = false; d > target; d--)
      { if (splitting || $from.index(d) > 0) {
        splitting = true;
        before = Fragment$1.from($from.node(d).copy(before));
        openStart++;
      } else {
        start--;
      } }
    var after = Fragment$1.empty, openEnd = 0;
    for (var d$1 = depth, splitting$1 = false; d$1 > target; d$1--)
      { if (splitting$1 || $to.after(d$1 + 1) < $to.end(d$1)) {
        splitting$1 = true;
        after = Fragment$1.from($to.node(d$1).copy(after));
        openEnd++;
      } else {
        end++;
      } }

    return this.step(new ReplaceAroundStep(start, end, gapStart, gapEnd,
                                           new Slice$1(before.append(after), openStart, openEnd),
                                           before.size - openStart, true))
  };

  // :: (NodeRange, NodeType, ?Object, ?NodeRange) → ?[{type: NodeType, attrs: ?Object}]
  // Try to find a valid way to wrap the content in the given range in a
  // node of the given type. May introduce extra nodes around and inside
  // the wrapper node, if necessary. Returns null if no valid wrapping
  // could be found. When `innerRange` is given, that range's content is
  // used as the content to fit into the wrapping, instead of the
  // content of `range`.
  function findWrapping(range, nodeType, attrs, innerRange) {
    if ( innerRange === void 0 ) { innerRange = range; }

    var around = findWrappingOutside(range, nodeType);
    var inner = around && findWrappingInside(innerRange, nodeType);
    if (!inner) { return null }
    return around.map(withAttrs).concat({type: nodeType, attrs: attrs}).concat(inner.map(withAttrs))
  }

  function withAttrs(type) { return {type: type, attrs: null} }

  function findWrappingOutside(range, type) {
    var parent = range.parent;
    var startIndex = range.startIndex;
    var endIndex = range.endIndex;
    var around = parent.contentMatchAt(startIndex).findWrapping(type);
    if (!around) { return null }
    var outer = around.length ? around[0] : type;
    return parent.canReplaceWith(startIndex, endIndex, outer) ? around : null
  }

  function findWrappingInside(range, type) {
    var parent = range.parent;
    var startIndex = range.startIndex;
    var endIndex = range.endIndex;
    var inner = parent.child(startIndex);
    var inside = type.contentMatch.findWrapping(inner.type);
    if (!inside) { return null }
    var lastType = inside.length ? inside[inside.length - 1] : type;
    var innerMatch = lastType.contentMatch;
    for (var i = startIndex; innerMatch && i < endIndex; i++)
      { innerMatch = innerMatch.matchType(parent.child(i).type); }
    if (!innerMatch || !innerMatch.validEnd) { return null }
    return inside
  }

  // :: (NodeRange, [{type: NodeType, attrs: ?Object}]) → this
  // Wrap the given [range](#model.NodeRange) in the given set of wrappers.
  // The wrappers are assumed to be valid in this position, and should
  // probably be computed with [`findWrapping`](#transform.findWrapping).
  Transform.prototype.wrap = function(range, wrappers) {
    var content = Fragment$1.empty;
    for (var i = wrappers.length - 1; i >= 0; i--)
      { content = Fragment$1.from(wrappers[i].type.create(wrappers[i].attrs, content)); }

    var start = range.start, end = range.end;
    return this.step(new ReplaceAroundStep(start, end, start, end, new Slice$1(content, 0, 0), wrappers.length, true))
  };

  // :: (number, ?number, NodeType, ?Object) → this
  // Set the type of all textblocks (partly) between `from` and `to` to
  // the given node type with the given attributes.
  Transform.prototype.setBlockType = function(from, to, type, attrs) {
    var this$1$1 = this;
    if ( to === void 0 ) { to = from; }

    if (!type.isTextblock) { throw new RangeError("Type given to setBlockType should be a textblock") }
    var mapFrom = this.steps.length;
    this.doc.nodesBetween(from, to, function (node, pos) {
      if (node.isTextblock && !node.hasMarkup(type, attrs) && canChangeType(this$1$1.doc, this$1$1.mapping.slice(mapFrom).map(pos), type)) {
        // Ensure all markup that isn't allowed in the new node type is cleared
        this$1$1.clearIncompatible(this$1$1.mapping.slice(mapFrom).map(pos, 1), type);
        var mapping = this$1$1.mapping.slice(mapFrom);
        var startM = mapping.map(pos, 1), endM = mapping.map(pos + node.nodeSize, 1);
        this$1$1.step(new ReplaceAroundStep(startM, endM, startM + 1, endM - 1,
                                        new Slice$1(Fragment$1.from(type.create(attrs, null, node.marks)), 0, 0), 1, true));
        return false
      }
    });
    return this
  };

  function canChangeType(doc, pos, type) {
    var $pos = doc.resolve(pos), index = $pos.index();
    return $pos.parent.canReplaceWith(index, index + 1, type)
  }

  // :: (number, ?NodeType, ?Object, ?[Mark]) → this
  // Change the type, attributes, and/or marks of the node at `pos`.
  // When `type` isn't given, the existing node type is preserved,
  Transform.prototype.setNodeMarkup = function(pos, type, attrs, marks) {
    var node = this.doc.nodeAt(pos);
    if (!node) { throw new RangeError("No node at given position") }
    if (!type) { type = node.type; }
    var newNode = type.create(attrs, null, marks || node.marks);
    if (node.isLeaf)
      { return this.replaceWith(pos, pos + node.nodeSize, newNode) }

    if (!type.validContent(node.content))
      { throw new RangeError("Invalid content for node type " + type.name) }

    return this.step(new ReplaceAroundStep(pos, pos + node.nodeSize, pos + 1, pos + node.nodeSize - 1,
                                           new Slice$1(Fragment$1.from(newNode), 0, 0), 1, true))
  };

  // :: (Node, number, number, ?[?{type: NodeType, attrs: ?Object}]) → bool
  // Check whether splitting at the given position is allowed.
  function canSplit(doc, pos, depth, typesAfter) {
    if ( depth === void 0 ) { depth = 1; }

    var $pos = doc.resolve(pos), base = $pos.depth - depth;
    var innerType = (typesAfter && typesAfter[typesAfter.length - 1]) || $pos.parent;
    if (base < 0 || $pos.parent.type.spec.isolating ||
        !$pos.parent.canReplace($pos.index(), $pos.parent.childCount) ||
        !innerType.type.validContent($pos.parent.content.cutByIndex($pos.index(), $pos.parent.childCount)))
      { return false }
    for (var d = $pos.depth - 1, i = depth - 2; d > base; d--, i--) {
      var node = $pos.node(d), index$1 = $pos.index(d);
      if (node.type.spec.isolating) { return false }
      var rest = node.content.cutByIndex(index$1, node.childCount);
      var after = (typesAfter && typesAfter[i]) || node;
      if (after != node) { rest = rest.replaceChild(0, after.type.create(after.attrs)); }
      if (!node.canReplace(index$1 + 1, node.childCount) || !after.type.validContent(rest))
        { return false }
    }
    var index = $pos.indexAfter(base);
    var baseType = typesAfter && typesAfter[0];
    return $pos.node(base).canReplaceWith(index, index, baseType ? baseType.type : $pos.node(base + 1).type)
  }

  // :: (number, ?number, ?[?{type: NodeType, attrs: ?Object}]) → this
  // Split the node at the given position, and optionally, if `depth` is
  // greater than one, any number of nodes above that. By default, the
  // parts split off will inherit the node type of the original node.
  // This can be changed by passing an array of types and attributes to
  // use after the split.
  Transform.prototype.split = function(pos, depth, typesAfter) {
    if ( depth === void 0 ) { depth = 1; }

    var $pos = this.doc.resolve(pos), before = Fragment$1.empty, after = Fragment$1.empty;
    for (var d = $pos.depth, e = $pos.depth - depth, i = depth - 1; d > e; d--, i--) {
      before = Fragment$1.from($pos.node(d).copy(before));
      var typeAfter = typesAfter && typesAfter[i];
      after = Fragment$1.from(typeAfter ? typeAfter.type.create(typeAfter.attrs, after) : $pos.node(d).copy(after));
    }
    return this.step(new ReplaceStep(pos, pos, new Slice$1(before.append(after), depth, depth), true))
  };

  // :: (Node, number) → bool
  // Test whether the blocks before and after a given position can be
  // joined.
  function canJoin(doc, pos) {
    var $pos = doc.resolve(pos), index = $pos.index();
    return joinable$1($pos.nodeBefore, $pos.nodeAfter) &&
      $pos.parent.canReplace(index, index + 1)
  }

  function joinable$1(a, b) {
    return a && b && !a.isLeaf && a.canAppend(b)
  }

  // :: (Node, number, ?number) → ?number
  // Find an ancestor of the given position that can be joined to the
  // block before (or after if `dir` is positive). Returns the joinable
  // point, if any.
  function joinPoint(doc, pos, dir) {
    if ( dir === void 0 ) { dir = -1; }

    var $pos = doc.resolve(pos);
    for (var d = $pos.depth;; d--) {
      var before = (void 0), after = (void 0), index = $pos.index(d);
      if (d == $pos.depth) {
        before = $pos.nodeBefore;
        after = $pos.nodeAfter;
      } else if (dir > 0) {
        before = $pos.node(d + 1);
        index++;
        after = $pos.node(d).maybeChild(index);
      } else {
        before = $pos.node(d).maybeChild(index - 1);
        after = $pos.node(d + 1);
      }
      if (before && !before.isTextblock && joinable$1(before, after) &&
          $pos.node(d).canReplace(index, index + 1)) { return pos }
      if (d == 0) { break }
      pos = dir < 0 ? $pos.before(d) : $pos.after(d);
    }
  }

  // :: (number, ?number) → this
  // Join the blocks around the given position. If depth is 2, their
  // last and first siblings are also joined, and so on.
  Transform.prototype.join = function(pos, depth) {
    if ( depth === void 0 ) { depth = 1; }

    var step = new ReplaceStep(pos - depth, pos + depth, Slice$1.empty, true);
    return this.step(step)
  };

  // :: (Node, number, NodeType) → ?number
  // Try to find a point where a node of the given type can be inserted
  // near `pos`, by searching up the node hierarchy when `pos` itself
  // isn't a valid place but is at the start or end of a node. Return
  // null if no position was found.
  function insertPoint(doc, pos, nodeType) {
    var $pos = doc.resolve(pos);
    if ($pos.parent.canReplaceWith($pos.index(), $pos.index(), nodeType)) { return pos }

    if ($pos.parentOffset == 0)
      { for (var d = $pos.depth - 1; d >= 0; d--) {
        var index = $pos.index(d);
        if ($pos.node(d).canReplaceWith(index, index, nodeType)) { return $pos.before(d + 1) }
        if (index > 0) { return null }
      } }
    if ($pos.parentOffset == $pos.parent.content.size)
      { for (var d$1 = $pos.depth - 1; d$1 >= 0; d$1--) {
        var index$1 = $pos.indexAfter(d$1);
        if ($pos.node(d$1).canReplaceWith(index$1, index$1, nodeType)) { return $pos.after(d$1 + 1) }
        if (index$1 < $pos.node(d$1).childCount) { return null }
      } }
  }

  // :: (Node, number, Slice) → ?number
  // Finds a position at or around the given position where the given
  // slice can be inserted. Will look at parent nodes' nearest boundary
  // and try there, even if the original position wasn't directly at the
  // start or end of that node. Returns null when no position was found.
  function dropPoint(doc, pos, slice) {
    var $pos = doc.resolve(pos);
    if (!slice.content.size) { return pos }
    var content = slice.content;
    for (var i = 0; i < slice.openStart; i++) { content = content.firstChild.content; }
    for (var pass = 1; pass <= (slice.openStart == 0 && slice.size ? 2 : 1); pass++) {
      for (var d = $pos.depth; d >= 0; d--) {
        var bias = d == $pos.depth ? 0 : $pos.pos <= ($pos.start(d + 1) + $pos.end(d + 1)) / 2 ? -1 : 1;
        var insertPos = $pos.index(d) + (bias > 0 ? 1 : 0);
        var parent = $pos.node(d), fits = false;
        if (pass == 1) {
          fits = parent.canReplace(insertPos, insertPos, content);
        } else {
          var wrapping = parent.contentMatchAt(insertPos).findWrapping(content.firstChild.type);
          fits = wrapping && parent.canReplaceWith(insertPos, insertPos, wrapping[0]);
        }
        if (fits)
          { return bias == 0 ? $pos.pos : bias < 0 ? $pos.before(d + 1) : $pos.after(d + 1) }
      }
    }
    return null
  }

  function mapFragment(fragment, f, parent) {
    var mapped = [];
    for (var i = 0; i < fragment.childCount; i++) {
      var child = fragment.child(i);
      if (child.content.size) { child = child.copy(mapFragment(child.content, f, child)); }
      if (child.isInline) { child = f(child, parent, i); }
      mapped.push(child);
    }
    return Fragment$1.fromArray(mapped)
  }

  // ::- Add a mark to all inline content between two positions.
  var AddMarkStep = /*@__PURE__*/(function (Step) {
    function AddMarkStep(from, to, mark) {
      Step.call(this);
      // :: number
      // The start of the marked range.
      this.from = from;
      // :: number
      // The end of the marked range.
      this.to = to;
      // :: Mark
      // The mark to add.
      this.mark = mark;
    }

    if ( Step ) { AddMarkStep.__proto__ = Step; }
    AddMarkStep.prototype = Object.create( Step && Step.prototype );
    AddMarkStep.prototype.constructor = AddMarkStep;

    AddMarkStep.prototype.apply = function apply (doc) {
      var this$1$1 = this;

      var oldSlice = doc.slice(this.from, this.to), $from = doc.resolve(this.from);
      var parent = $from.node($from.sharedDepth(this.to));
      var slice = new Slice$1(mapFragment(oldSlice.content, function (node, parent) {
        if (!node.isAtom || !parent.type.allowsMarkType(this$1$1.mark.type)) { return node }
        return node.mark(this$1$1.mark.addToSet(node.marks))
      }, parent), oldSlice.openStart, oldSlice.openEnd);
      return StepResult.fromReplace(doc, this.from, this.to, slice)
    };

    AddMarkStep.prototype.invert = function invert () {
      return new RemoveMarkStep(this.from, this.to, this.mark)
    };

    AddMarkStep.prototype.map = function map (mapping) {
      var from = mapping.mapResult(this.from, 1), to = mapping.mapResult(this.to, -1);
      if (from.deleted && to.deleted || from.pos >= to.pos) { return null }
      return new AddMarkStep(from.pos, to.pos, this.mark)
    };

    AddMarkStep.prototype.merge = function merge (other) {
      if (other instanceof AddMarkStep &&
          other.mark.eq(this.mark) &&
          this.from <= other.to && this.to >= other.from)
        { return new AddMarkStep(Math.min(this.from, other.from),
                               Math.max(this.to, other.to), this.mark) }
    };

    AddMarkStep.prototype.toJSON = function toJSON () {
      return {stepType: "addMark", mark: this.mark.toJSON(),
              from: this.from, to: this.to}
    };

    AddMarkStep.fromJSON = function fromJSON (schema, json) {
      if (typeof json.from != "number" || typeof json.to != "number")
        { throw new RangeError("Invalid input for AddMarkStep.fromJSON") }
      return new AddMarkStep(json.from, json.to, schema.markFromJSON(json.mark))
    };

    return AddMarkStep;
  }(Step));

  Step.jsonID("addMark", AddMarkStep);

  // ::- Remove a mark from all inline content between two positions.
  var RemoveMarkStep = /*@__PURE__*/(function (Step) {
    function RemoveMarkStep(from, to, mark) {
      Step.call(this);
      // :: number
      // The start of the unmarked range.
      this.from = from;
      // :: number
      // The end of the unmarked range.
      this.to = to;
      // :: Mark
      // The mark to remove.
      this.mark = mark;
    }

    if ( Step ) { RemoveMarkStep.__proto__ = Step; }
    RemoveMarkStep.prototype = Object.create( Step && Step.prototype );
    RemoveMarkStep.prototype.constructor = RemoveMarkStep;

    RemoveMarkStep.prototype.apply = function apply (doc) {
      var this$1$1 = this;

      var oldSlice = doc.slice(this.from, this.to);
      var slice = new Slice$1(mapFragment(oldSlice.content, function (node) {
        return node.mark(this$1$1.mark.removeFromSet(node.marks))
      }), oldSlice.openStart, oldSlice.openEnd);
      return StepResult.fromReplace(doc, this.from, this.to, slice)
    };

    RemoveMarkStep.prototype.invert = function invert () {
      return new AddMarkStep(this.from, this.to, this.mark)
    };

    RemoveMarkStep.prototype.map = function map (mapping) {
      var from = mapping.mapResult(this.from, 1), to = mapping.mapResult(this.to, -1);
      if (from.deleted && to.deleted || from.pos >= to.pos) { return null }
      return new RemoveMarkStep(from.pos, to.pos, this.mark)
    };

    RemoveMarkStep.prototype.merge = function merge (other) {
      if (other instanceof RemoveMarkStep &&
          other.mark.eq(this.mark) &&
          this.from <= other.to && this.to >= other.from)
        { return new RemoveMarkStep(Math.min(this.from, other.from),
                                  Math.max(this.to, other.to), this.mark) }
    };

    RemoveMarkStep.prototype.toJSON = function toJSON () {
      return {stepType: "removeMark", mark: this.mark.toJSON(),
              from: this.from, to: this.to}
    };

    RemoveMarkStep.fromJSON = function fromJSON (schema, json) {
      if (typeof json.from != "number" || typeof json.to != "number")
        { throw new RangeError("Invalid input for RemoveMarkStep.fromJSON") }
      return new RemoveMarkStep(json.from, json.to, schema.markFromJSON(json.mark))
    };

    return RemoveMarkStep;
  }(Step));

  Step.jsonID("removeMark", RemoveMarkStep);

  // :: (number, number, Mark) → this
  // Add the given mark to the inline content between `from` and `to`.
  Transform.prototype.addMark = function(from, to, mark) {
    var this$1$1 = this;

    var removed = [], added = [], removing = null, adding = null;
    this.doc.nodesBetween(from, to, function (node, pos, parent) {
      if (!node.isInline) { return }
      var marks = node.marks;
      if (!mark.isInSet(marks) && parent.type.allowsMarkType(mark.type)) {
        var start = Math.max(pos, from), end = Math.min(pos + node.nodeSize, to);
        var newSet = mark.addToSet(marks);

        for (var i = 0; i < marks.length; i++) {
          if (!marks[i].isInSet(newSet)) {
            if (removing && removing.to == start && removing.mark.eq(marks[i]))
              { removing.to = end; }
            else
              { removed.push(removing = new RemoveMarkStep(start, end, marks[i])); }
          }
        }

        if (adding && adding.to == start)
          { adding.to = end; }
        else
          { added.push(adding = new AddMarkStep(start, end, mark)); }
      }
    });

    removed.forEach(function (s) { return this$1$1.step(s); });
    added.forEach(function (s) { return this$1$1.step(s); });
    return this
  };

  // :: (number, number, ?union<Mark, MarkType>) → this
  // Remove marks from inline nodes between `from` and `to`. When `mark`
  // is a single mark, remove precisely that mark. When it is a mark type,
  // remove all marks of that type. When it is null, remove all marks of
  // any type.
  Transform.prototype.removeMark = function(from, to, mark) {
    var this$1$1 = this;
    if ( mark === void 0 ) { mark = null; }

    var matched = [], step = 0;
    this.doc.nodesBetween(from, to, function (node, pos) {
      if (!node.isInline) { return }
      step++;
      var toRemove = null;
      if (mark instanceof MarkType) {
        var set = node.marks, found;
        while (found = mark.isInSet(set)) {
  (toRemove || (toRemove = [])).push(found);
          set = found.removeFromSet(set);
        }
      } else if (mark) {
        if (mark.isInSet(node.marks)) { toRemove = [mark]; }
      } else {
        toRemove = node.marks;
      }
      if (toRemove && toRemove.length) {
        var end = Math.min(pos + node.nodeSize, to);
        for (var i = 0; i < toRemove.length; i++) {
          var style = toRemove[i], found$1 = (void 0);
          for (var j = 0; j < matched.length; j++) {
            var m = matched[j];
            if (m.step == step - 1 && style.eq(matched[j].style)) { found$1 = m; }
          }
          if (found$1) {
            found$1.to = end;
            found$1.step = step;
          } else {
            matched.push({style: style, from: Math.max(pos, from), to: end, step: step});
          }
        }
      }
    });
    matched.forEach(function (m) { return this$1$1.step(new RemoveMarkStep(m.from, m.to, m.style)); });
    return this
  };

  // :: (number, NodeType, ?ContentMatch) → this
  // Removes all marks and nodes from the content of the node at `pos`
  // that don't match the given new parent node type. Accepts an
  // optional starting [content match](#model.ContentMatch) as third
  // argument.
  Transform.prototype.clearIncompatible = function(pos, parentType, match) {
    if ( match === void 0 ) { match = parentType.contentMatch; }

    var node = this.doc.nodeAt(pos);
    var delSteps = [], cur = pos + 1;
    for (var i = 0; i < node.childCount; i++) {
      var child = node.child(i), end = cur + child.nodeSize;
      var allowed = match.matchType(child.type, child.attrs);
      if (!allowed) {
        delSteps.push(new ReplaceStep(cur, end, Slice$1.empty));
      } else {
        match = allowed;
        for (var j = 0; j < child.marks.length; j++) { if (!parentType.allowsMarkType(child.marks[j].type))
          { this.step(new RemoveMarkStep(cur, end, child.marks[j])); } }
      }
      cur = end;
    }
    if (!match.validEnd) {
      var fill = match.fillBefore(Fragment$1.empty, true);
      this.replace(cur, cur, new Slice$1(fill, 0, 0));
    }
    for (var i$1 = delSteps.length - 1; i$1 >= 0; i$1--) { this.step(delSteps[i$1]); }
    return this
  };

  // :: (Node, number, ?number, ?Slice) → ?Step
  // ‘Fit’ a slice into a given position in the document, producing a
  // [step](#transform.Step) that inserts it. Will return null if
  // there's no meaningful way to insert the slice here, or inserting it
  // would be a no-op (an empty slice over an empty range).
  function replaceStep(doc, from, to, slice) {
    if ( to === void 0 ) { to = from; }
    if ( slice === void 0 ) { slice = Slice$1.empty; }

    if (from == to && !slice.size) { return null }

    var $from = doc.resolve(from), $to = doc.resolve(to);
    // Optimization -- avoid work if it's obvious that it's not needed.
    if (fitsTrivially($from, $to, slice)) { return new ReplaceStep(from, to, slice) }
    return new Fitter($from, $to, slice).fit()
  }

  // :: (number, ?number, ?Slice) → this
  // Replace the part of the document between `from` and `to` with the
  // given `slice`.
  Transform.prototype.replace = function(from, to, slice) {
    if ( to === void 0 ) { to = from; }
    if ( slice === void 0 ) { slice = Slice$1.empty; }

    var step = replaceStep(this.doc, from, to, slice);
    if (step) { this.step(step); }
    return this
  };

  // :: (number, number, union<Fragment, Node, [Node]>) → this
  // Replace the given range with the given content, which may be a
  // fragment, node, or array of nodes.
  Transform.prototype.replaceWith = function(from, to, content) {
    return this.replace(from, to, new Slice$1(Fragment$1.from(content), 0, 0))
  };

  // :: (number, number) → this
  // Delete the content between the given positions.
  Transform.prototype.delete = function(from, to) {
    return this.replace(from, to, Slice$1.empty)
  };

  // :: (number, union<Fragment, Node, [Node]>) → this
  // Insert the given content at the given position.
  Transform.prototype.insert = function(pos, content) {
    return this.replaceWith(pos, pos, content)
  };

  function fitsTrivially($from, $to, slice) {
    return !slice.openStart && !slice.openEnd && $from.start() == $to.start() &&
      $from.parent.canReplace($from.index(), $to.index(), slice.content)
  }

  // Algorithm for 'placing' the elements of a slice into a gap:
  //
  // We consider the content of each node that is open to the left to be
  // independently placeable. I.e. in <p("foo"), p("bar")>, when the
  // paragraph on the left is open, "foo" can be placed (somewhere on
  // the left side of the replacement gap) independently from p("bar").
  //
  // This class tracks the state of the placement progress in the
  // following properties:
  //
  //  - `frontier` holds a stack of `{type, match}` objects that
  //    represent the open side of the replacement. It starts at
  //    `$from`, then moves forward as content is placed, and is finally
  //    reconciled with `$to`.
  //
  //  - `unplaced` is a slice that represents the content that hasn't
  //    been placed yet.
  //
  //  - `placed` is a fragment of placed content. Its open-start value
  //    is implicit in `$from`, and its open-end value in `frontier`.
  var Fitter = function Fitter($from, $to, slice) {
    this.$to = $to;
    this.$from = $from;
    this.unplaced = slice;

    this.frontier = [];
    for (var i = 0; i <= $from.depth; i++) {
      var node = $from.node(i);
      this.frontier.push({
        type: node.type,
        match: node.contentMatchAt($from.indexAfter(i))
      });
    }

    this.placed = Fragment$1.empty;
    for (var i$1 = $from.depth; i$1 > 0; i$1--)
      { this.placed = Fragment$1.from($from.node(i$1).copy(this.placed)); }
  };

  var prototypeAccessors$1$4 = { depth: { configurable: true } };

  prototypeAccessors$1$4.depth.get = function () { return this.frontier.length - 1 };

  Fitter.prototype.fit = function fit () {
    // As long as there's unplaced content, try to place some of it.
    // If that fails, either increase the open score of the unplaced
    // slice, or drop nodes from it, and then try again.
    while (this.unplaced.size) {
      var fit = this.findFittable();
      if (fit) { this.placeNodes(fit); }
      else { this.openMore() || this.dropNode(); }
    }
    // When there's inline content directly after the frontier _and_
    // directly after `this.$to`, we must generate a `ReplaceAround`
    // step that pulls that content into the node after the frontier.
    // That means the fitting must be done to the end of the textblock
    // node after `this.$to`, not `this.$to` itself.
    var moveInline = this.mustMoveInline(), placedSize = this.placed.size - this.depth - this.$from.depth;
    var $from = this.$from, $to = this.close(moveInline < 0 ? this.$to : $from.doc.resolve(moveInline));
    if (!$to) { return null }

    // If closing to `$to` succeeded, create a step
    var content = this.placed, openStart = $from.depth, openEnd = $to.depth;
    while (openStart && openEnd && content.childCount == 1) { // Normalize by dropping open parent nodes
      content = content.firstChild.content;
      openStart--; openEnd--;
    }
    var slice = new Slice$1(content, openStart, openEnd);
    if (moveInline > -1)
      { return new ReplaceAroundStep($from.pos, moveInline, this.$to.pos, this.$to.end(), slice, placedSize) }
    if (slice.size || $from.pos != this.$to.pos) // Don't generate no-op steps
      { return new ReplaceStep($from.pos, $to.pos, slice) }
  };

  // Find a position on the start spine of `this.unplaced` that has
  // content that can be moved somewhere on the frontier. Returns two
  // depths, one for the slice and one for the frontier.
  Fitter.prototype.findFittable = function findFittable () {
    // Only try wrapping nodes (pass 2) after finding a place without
    // wrapping failed.
    for (var pass = 1; pass <= 2; pass++) {
      for (var sliceDepth = this.unplaced.openStart; sliceDepth >= 0; sliceDepth--) {
        var fragment = (void 0), parent = (void 0);
        if (sliceDepth) {
          parent = contentAt(this.unplaced.content, sliceDepth - 1).firstChild;
          fragment = parent.content;
        } else {
          fragment = this.unplaced.content;
        }
        var first = fragment.firstChild;
        for (var frontierDepth = this.depth; frontierDepth >= 0; frontierDepth--) {
          var ref = this.frontier[frontierDepth];
            var type = ref.type;
            var match = ref.match;
            var wrap = (void 0), inject = (void 0);
          // In pass 1, if the next node matches, or there is no next
          // node but the parents look compatible, we've found a
          // place.
          if (pass == 1 && (first ? match.matchType(first.type) || (inject = match.fillBefore(Fragment$1.from(first), false))
                            : type.compatibleContent(parent.type)))
            { return {sliceDepth: sliceDepth, frontierDepth: frontierDepth, parent: parent, inject: inject} }
          // In pass 2, look for a set of wrapping nodes that make
          // `first` fit here.
          else if (pass == 2 && first && (wrap = match.findWrapping(first.type)))
            { return {sliceDepth: sliceDepth, frontierDepth: frontierDepth, parent: parent, wrap: wrap} }
          // Don't continue looking further up if the parent node
          // would fit here.
          if (parent && match.matchType(parent.type)) { break }
        }
      }
    }
  };

  Fitter.prototype.openMore = function openMore () {
    var ref = this.unplaced;
      var content = ref.content;
      var openStart = ref.openStart;
      var openEnd = ref.openEnd;
    var inner = contentAt(content, openStart);
    if (!inner.childCount || inner.firstChild.isLeaf) { return false }
    this.unplaced = new Slice$1(content, openStart + 1,
                              Math.max(openEnd, inner.size + openStart >= content.size - openEnd ? openStart + 1 : 0));
    return true
  };

  Fitter.prototype.dropNode = function dropNode () {
    var ref = this.unplaced;
      var content = ref.content;
      var openStart = ref.openStart;
      var openEnd = ref.openEnd;
    var inner = contentAt(content, openStart);
    if (inner.childCount <= 1 && openStart > 0) {
      var openAtEnd = content.size - openStart <= openStart + inner.size;
      this.unplaced = new Slice$1(dropFromFragment(content, openStart - 1, 1), openStart - 1,
                                openAtEnd ? openStart - 1 : openEnd);
    } else {
      this.unplaced = new Slice$1(dropFromFragment(content, openStart, 1), openStart, openEnd);
    }
  };

  // : ({sliceDepth: number, frontierDepth: number, parent: ?Node, wrap: ?[NodeType], inject: ?Fragment})
  // Move content from the unplaced slice at `sliceDepth` to the
  // frontier node at `frontierDepth`. Close that frontier node when
  // applicable.
  Fitter.prototype.placeNodes = function placeNodes (ref) {
      var sliceDepth = ref.sliceDepth;
      var frontierDepth = ref.frontierDepth;
      var parent = ref.parent;
      var inject = ref.inject;
      var wrap = ref.wrap;

    while (this.depth > frontierDepth) { this.closeFrontierNode(); }
    if (wrap) { for (var i = 0; i < wrap.length; i++) { this.openFrontierNode(wrap[i]); } }

    var slice = this.unplaced, fragment = parent ? parent.content : slice.content;
    var openStart = slice.openStart - sliceDepth;
    var taken = 0, add = [];
    var ref$1 = this.frontier[frontierDepth];
      var match = ref$1.match;
      var type = ref$1.type;
    if (inject) {
      for (var i$1 = 0; i$1 < inject.childCount; i$1++) { add.push(inject.child(i$1)); }
      match = match.matchFragment(inject);
    }
    // Computes the amount of (end) open nodes at the end of the
    // fragment. When 0, the parent is open, but no more. When
    // negative, nothing is open.
    var openEndCount = (fragment.size + sliceDepth) - (slice.content.size - slice.openEnd);
    // Scan over the fragment, fitting as many child nodes as
    // possible.
    while (taken < fragment.childCount) {
      var next = fragment.child(taken), matches = match.matchType(next.type);
      if (!matches) { break }
      taken++;
      if (taken > 1 || openStart == 0 || next.content.size) { // Drop empty open nodes
        match = matches;
        add.push(closeNodeStart(next.mark(type.allowedMarks(next.marks)), taken == 1 ? openStart : 0,
                                taken == fragment.childCount ? openEndCount : -1));
      }
    }
    var toEnd = taken == fragment.childCount;
    if (!toEnd) { openEndCount = -1; }

    this.placed = addToFragment(this.placed, frontierDepth, Fragment$1.from(add));
    this.frontier[frontierDepth].match = match;

    // If the parent types match, and the entire node was moved, and
    // it's not open, close this frontier node right away.
    if (toEnd && openEndCount < 0 && parent && parent.type == this.frontier[this.depth].type && this.frontier.length > 1)
      { this.closeFrontierNode(); }

    // Add new frontier nodes for any open nodes at the end.
    for (var i$2 = 0, cur = fragment; i$2 < openEndCount; i$2++) {
      var node = cur.lastChild;
      this.frontier.push({type: node.type, match: node.contentMatchAt(node.childCount)});
      cur = node.content;
    }

    // Update `this.unplaced`. Drop the entire node from which we
    // placed it we got to its end, otherwise just drop the placed
    // nodes.
    this.unplaced = !toEnd ? new Slice$1(dropFromFragment(slice.content, sliceDepth, taken), slice.openStart, slice.openEnd)
      : sliceDepth == 0 ? Slice$1.empty
      : new Slice$1(dropFromFragment(slice.content, sliceDepth - 1, 1),
                  sliceDepth - 1, openEndCount < 0 ? slice.openEnd : sliceDepth - 1);
  };

  Fitter.prototype.mustMoveInline = function mustMoveInline () {
    if (!this.$to.parent.isTextblock || this.$to.end() == this.$to.pos) { return -1 }
    var top = this.frontier[this.depth], level;
    if (!top.type.isTextblock || !contentAfterFits(this.$to, this.$to.depth, top.type, top.match, false) ||
        (this.$to.depth == this.depth && (level = this.findCloseLevel(this.$to)) && level.depth == this.depth)) { return -1 }

    var ref = this.$to;
      var depth = ref.depth;
      var after = this.$to.after(depth);
    while (depth > 1 && after == this.$to.end(--depth)) { ++after; }
    return after
  };

  Fitter.prototype.findCloseLevel = function findCloseLevel ($to) {
    scan: for (var i = Math.min(this.depth, $to.depth); i >= 0; i--) {
      var ref = this.frontier[i];
        var match = ref.match;
        var type = ref.type;
      var dropInner = i < $to.depth && $to.end(i + 1) == $to.pos + ($to.depth - (i + 1));
      var fit = contentAfterFits($to, i, type, match, dropInner);
      if (!fit) { continue }
      for (var d = i - 1; d >= 0; d--) {
        var ref$1 = this.frontier[d];
          var match$1 = ref$1.match;
          var type$1 = ref$1.type;
        var matches = contentAfterFits($to, d, type$1, match$1, true);
        if (!matches || matches.childCount) { continue scan }
      }
      return {depth: i, fit: fit, move: dropInner ? $to.doc.resolve($to.after(i + 1)) : $to}
    }
  };

  Fitter.prototype.close = function close ($to) {
    var close = this.findCloseLevel($to);
    if (!close) { return null }

    while (this.depth > close.depth) { this.closeFrontierNode(); }
    if (close.fit.childCount) { this.placed = addToFragment(this.placed, close.depth, close.fit); }
    $to = close.move;
    for (var d = close.depth + 1; d <= $to.depth; d++) {
      var node = $to.node(d), add = node.type.contentMatch.fillBefore(node.content, true, $to.index(d));
      this.openFrontierNode(node.type, node.attrs, add);
    }
    return $to
  };

  Fitter.prototype.openFrontierNode = function openFrontierNode (type, attrs, content) {
    var top = this.frontier[this.depth];
    top.match = top.match.matchType(type);
    this.placed = addToFragment(this.placed, this.depth, Fragment$1.from(type.create(attrs, content)));
    this.frontier.push({type: type, match: type.contentMatch});
  };

  Fitter.prototype.closeFrontierNode = function closeFrontierNode () {
    var open = this.frontier.pop();
    var add = open.match.fillBefore(Fragment$1.empty, true);
    if (add.childCount) { this.placed = addToFragment(this.placed, this.frontier.length, add); }
  };

  Object.defineProperties( Fitter.prototype, prototypeAccessors$1$4 );

  function dropFromFragment(fragment, depth, count) {
    if (depth == 0) { return fragment.cutByIndex(count) }
    return fragment.replaceChild(0, fragment.firstChild.copy(dropFromFragment(fragment.firstChild.content, depth - 1, count)))
  }

  function addToFragment(fragment, depth, content) {
    if (depth == 0) { return fragment.append(content) }
    return fragment.replaceChild(fragment.childCount - 1,
                                 fragment.lastChild.copy(addToFragment(fragment.lastChild.content, depth - 1, content)))
  }

  function contentAt(fragment, depth) {
    for (var i = 0; i < depth; i++) { fragment = fragment.firstChild.content; }
    return fragment
  }

  function closeNodeStart(node, openStart, openEnd) {
    if (openStart <= 0) { return node }
    var frag = node.content;
    if (openStart > 1)
      { frag = frag.replaceChild(0, closeNodeStart(frag.firstChild, openStart - 1, frag.childCount == 1 ? openEnd - 1 : 0)); }
    if (openStart > 0) {
      frag = node.type.contentMatch.fillBefore(frag).append(frag);
      if (openEnd <= 0) { frag = frag.append(node.type.contentMatch.matchFragment(frag).fillBefore(Fragment$1.empty, true)); }
    }
    return node.copy(frag)
  }

  function contentAfterFits($to, depth, type, match, open) {
    var node = $to.node(depth), index = open ? $to.indexAfter(depth) : $to.index(depth);
    if (index == node.childCount && !type.compatibleContent(node.type)) { return null }
    var fit = match.fillBefore(node.content, true, index);
    return fit && !invalidMarks(type, node.content, index) ? fit : null
  }

  function invalidMarks(type, fragment, start) {
    for (var i = start; i < fragment.childCount; i++)
      { if (!type.allowsMarks(fragment.child(i).marks)) { return true } }
    return false
  }

  // :: (number, number, Slice) → this
  // Replace a range of the document with a given slice, using `from`,
  // `to`, and the slice's [`openStart`](#model.Slice.openStart) property
  // as hints, rather than fixed start and end points. This method may
  // grow the replaced area or close open nodes in the slice in order to
  // get a fit that is more in line with WYSIWYG expectations, by
  // dropping fully covered parent nodes of the replaced region when
  // they are marked [non-defining](#model.NodeSpec.defining), or
  // including an open parent node from the slice that _is_ marked as
  // [defining](#model.NodeSpec.defining).
  //
  // This is the method, for example, to handle paste. The similar
  // [`replace`](#transform.Transform.replace) method is a more
  // primitive tool which will _not_ move the start and end of its given
  // range, and is useful in situations where you need more precise
  // control over what happens.
  Transform.prototype.replaceRange = function(from, to, slice) {
    if (!slice.size) { return this.deleteRange(from, to) }

    var $from = this.doc.resolve(from), $to = this.doc.resolve(to);
    if (fitsTrivially($from, $to, slice))
      { return this.step(new ReplaceStep(from, to, slice)) }

    var targetDepths = coveredDepths($from, this.doc.resolve(to));
    // Can't replace the whole document, so remove 0 if it's present
    if (targetDepths[targetDepths.length - 1] == 0) { targetDepths.pop(); }
    // Negative numbers represent not expansion over the whole node at
    // that depth, but replacing from $from.before(-D) to $to.pos.
    var preferredTarget = -($from.depth + 1);
    targetDepths.unshift(preferredTarget);
    // This loop picks a preferred target depth, if one of the covering
    // depths is not outside of a defining node, and adds negative
    // depths for any depth that has $from at its start and does not
    // cross a defining node.
    for (var d = $from.depth, pos = $from.pos - 1; d > 0; d--, pos--) {
      var spec = $from.node(d).type.spec;
      if (spec.defining || spec.isolating) { break }
      if (targetDepths.indexOf(d) > -1) { preferredTarget = d; }
      else if ($from.before(d) == pos) { targetDepths.splice(1, 0, -d); }
    }
    // Try to fit each possible depth of the slice into each possible
    // target depth, starting with the preferred depths.
    var preferredTargetIndex = targetDepths.indexOf(preferredTarget);

    var leftNodes = [], preferredDepth = slice.openStart;
    for (var content = slice.content, i = 0;; i++) {
      var node = content.firstChild;
      leftNodes.push(node);
      if (i == slice.openStart) { break }
      content = node.content;
    }
    // Back up if the node directly above openStart, or the node above
    // that separated only by a non-defining textblock node, is defining.
    if (preferredDepth > 0 && leftNodes[preferredDepth - 1].type.spec.defining &&
        $from.node(preferredTargetIndex).type != leftNodes[preferredDepth - 1].type)
      { preferredDepth -= 1; }
    else if (preferredDepth >= 2 && leftNodes[preferredDepth - 1].isTextblock && leftNodes[preferredDepth - 2].type.spec.defining &&
             $from.node(preferredTargetIndex).type != leftNodes[preferredDepth - 2].type)
      { preferredDepth -= 2; }

    for (var j = slice.openStart; j >= 0; j--) {
      var openDepth = (j + preferredDepth + 1) % (slice.openStart + 1);
      var insert = leftNodes[openDepth];
      if (!insert) { continue }
      for (var i$1 = 0; i$1 < targetDepths.length; i$1++) {
        // Loop over possible expansion levels, starting with the
        // preferred one
        var targetDepth = targetDepths[(i$1 + preferredTargetIndex) % targetDepths.length], expand = true;
        if (targetDepth < 0) { expand = false; targetDepth = -targetDepth; }
        var parent = $from.node(targetDepth - 1), index = $from.index(targetDepth - 1);
        if (parent.canReplaceWith(index, index, insert.type, insert.marks))
          { return this.replace($from.before(targetDepth), expand ? $to.after(targetDepth) : to,
                              new Slice$1(closeFragment(slice.content, 0, slice.openStart, openDepth),
                                        openDepth, slice.openEnd)) }
      }
    }

    var startSteps = this.steps.length;
    for (var i$2 = targetDepths.length - 1; i$2 >= 0; i$2--) {
      this.replace(from, to, slice);
      if (this.steps.length > startSteps) { break }
      var depth = targetDepths[i$2];
      if (depth < 0) { continue }
      from = $from.before(depth); to = $to.after(depth);
    }
    return this
  };

  function closeFragment(fragment, depth, oldOpen, newOpen, parent) {
    if (depth < oldOpen) {
      var first = fragment.firstChild;
      fragment = fragment.replaceChild(0, first.copy(closeFragment(first.content, depth + 1, oldOpen, newOpen, first)));
    }
    if (depth > newOpen) {
      var match = parent.contentMatchAt(0);
      var start = match.fillBefore(fragment).append(fragment);
      fragment = start.append(match.matchFragment(start).fillBefore(Fragment$1.empty, true));
    }
    return fragment
  }

  // :: (number, number, Node) → this
  // Replace the given range with a node, but use `from` and `to` as
  // hints, rather than precise positions. When from and to are the same
  // and are at the start or end of a parent node in which the given
  // node doesn't fit, this method may _move_ them out towards a parent
  // that does allow the given node to be placed. When the given range
  // completely covers a parent node, this method may completely replace
  // that parent node.
  Transform.prototype.replaceRangeWith = function(from, to, node) {
    if (!node.isInline && from == to && this.doc.resolve(from).parent.content.size) {
      var point = insertPoint(this.doc, from, node.type);
      if (point != null) { from = to = point; }
    }
    return this.replaceRange(from, to, new Slice$1(Fragment$1.from(node), 0, 0))
  };

  // :: (number, number) → this
  // Delete the given range, expanding it to cover fully covered
  // parent nodes until a valid replace is found.
  Transform.prototype.deleteRange = function(from, to) {
    var $from = this.doc.resolve(from), $to = this.doc.resolve(to);
    var covered = coveredDepths($from, $to);
    for (var i = 0; i < covered.length; i++) {
      var depth = covered[i], last = i == covered.length - 1;
      if ((last && depth == 0) || $from.node(depth).type.contentMatch.validEnd)
        { return this.delete($from.start(depth), $to.end(depth)) }
      if (depth > 0 && (last || $from.node(depth - 1).canReplace($from.index(depth - 1), $to.indexAfter(depth - 1))))
        { return this.delete($from.before(depth), $to.after(depth)) }
    }
    for (var d = 1; d <= $from.depth && d <= $to.depth; d++) {
      if (from - $from.start(d) == $from.depth - d && to > $from.end(d) && $to.end(d) - to != $to.depth - d)
        { return this.delete($from.before(d), to) }
    }
    return this.delete(from, to)
  };

  // : (ResolvedPos, ResolvedPos) → [number]
  // Returns an array of all depths for which $from - $to spans the
  // whole content of the nodes at that depth.
  function coveredDepths($from, $to) {
    var result = [], minDepth = Math.min($from.depth, $to.depth);
    for (var d = minDepth; d >= 0; d--) {
      var start = $from.start(d);
      if (start < $from.pos - ($from.depth - d) ||
          $to.end(d) > $to.pos + ($to.depth - d) ||
          $from.node(d).type.spec.isolating ||
          $to.node(d).type.spec.isolating) { break }
      if (start == $to.start(d)) { result.push(d); }
    }
    return result
  }

  var transform = /*#__PURE__*/Object.freeze({
    __proto__: null,
    AddMarkStep: AddMarkStep,
    MapResult: MapResult,
    Mapping: Mapping,
    RemoveMarkStep: RemoveMarkStep,
    ReplaceAroundStep: ReplaceAroundStep,
    ReplaceStep: ReplaceStep,
    Step: Step,
    StepMap: StepMap,
    StepResult: StepResult,
    Transform: Transform,
    TransformError: TransformError,
    canJoin: canJoin,
    canSplit: canSplit,
    dropPoint: dropPoint,
    findWrapping: findWrapping,
    insertPoint: insertPoint,
    joinPoint: joinPoint,
    liftTarget: liftTarget,
    replaceStep: replaceStep
  });

  var classesById = Object.create(null);

  // ::- Superclass for editor selections. Every selection type should
  // extend this. Should not be instantiated directly.
  var Selection = function Selection($anchor, $head, ranges) {
    // :: [SelectionRange]
    // The ranges covered by the selection.
    this.ranges = ranges || [new SelectionRange($anchor.min($head), $anchor.max($head))];
    // :: ResolvedPos
    // The resolved anchor of the selection (the side that stays in
    // place when the selection is modified).
    this.$anchor = $anchor;
    // :: ResolvedPos
    // The resolved head of the selection (the side that moves when
    // the selection is modified).
    this.$head = $head;
  };

  var prototypeAccessors$5 = { anchor: { configurable: true },head: { configurable: true },from: { configurable: true },to: { configurable: true },$from: { configurable: true },$to: { configurable: true },empty: { configurable: true } };

  // :: number
  // The selection's anchor, as an unresolved position.
  prototypeAccessors$5.anchor.get = function () { return this.$anchor.pos };

  // :: number
  // The selection's head.
  prototypeAccessors$5.head.get = function () { return this.$head.pos };

  // :: number
  // The lower bound of the selection's main range.
  prototypeAccessors$5.from.get = function () { return this.$from.pos };

  // :: number
  // The upper bound of the selection's main range.
  prototypeAccessors$5.to.get = function () { return this.$to.pos };

  // :: ResolvedPos
  // The resolved lowerbound of the selection's main range.
  prototypeAccessors$5.$from.get = function () {
    return this.ranges[0].$from
  };

  // :: ResolvedPos
  // The resolved upper bound of the selection's main range.
  prototypeAccessors$5.$to.get = function () {
    return this.ranges[0].$to
  };

  // :: bool
  // Indicates whether the selection contains any content.
  prototypeAccessors$5.empty.get = function () {
    var ranges = this.ranges;
    for (var i = 0; i < ranges.length; i++)
      { if (ranges[i].$from.pos != ranges[i].$to.pos) { return false } }
    return true
  };

  // eq:: (Selection) → bool
  // Test whether the selection is the same as another selection.

  // map:: (doc: Node, mapping: Mappable) → Selection
  // Map this selection through a [mappable](#transform.Mappable) thing. `doc`
  // should be the new document to which we are mapping.

  // :: () → Slice
  // Get the content of this selection as a slice.
  Selection.prototype.content = function content () {
    return this.$from.node(0).slice(this.from, this.to, true)
  };

  // :: (Transaction, ?Slice)
  // Replace the selection with a slice or, if no slice is given,
  // delete the selection. Will append to the given transaction.
  Selection.prototype.replace = function replace (tr, content) {
      if ( content === void 0 ) { content = Slice$1.empty; }

    // Put the new selection at the position after the inserted
    // content. When that ended in an inline node, search backwards,
    // to get the position after that node. If not, search forward.
    var lastNode = content.content.lastChild, lastParent = null;
    for (var i = 0; i < content.openEnd; i++) {
      lastParent = lastNode;
      lastNode = lastNode.lastChild;
    }

    var mapFrom = tr.steps.length, ranges = this.ranges;
    for (var i$1 = 0; i$1 < ranges.length; i$1++) {
      var ref = ranges[i$1];
        var $from = ref.$from;
        var $to = ref.$to;
        var mapping = tr.mapping.slice(mapFrom);
      tr.replaceRange(mapping.map($from.pos), mapping.map($to.pos), i$1 ? Slice$1.empty : content);
      if (i$1 == 0)
        { selectionToInsertionEnd(tr, mapFrom, (lastNode ? lastNode.isInline : lastParent && lastParent.isTextblock) ? -1 : 1); }
    }
  };

  // :: (Transaction, Node)
  // Replace the selection with the given node, appending the changes
  // to the given transaction.
  Selection.prototype.replaceWith = function replaceWith (tr, node) {
    var mapFrom = tr.steps.length, ranges = this.ranges;
    for (var i = 0; i < ranges.length; i++) {
      var ref = ranges[i];
        var $from = ref.$from;
        var $to = ref.$to;
        var mapping = tr.mapping.slice(mapFrom);
      var from = mapping.map($from.pos), to = mapping.map($to.pos);
      if (i) {
        tr.deleteRange(from, to);
      } else {
        tr.replaceRangeWith(from, to, node);
        selectionToInsertionEnd(tr, mapFrom, node.isInline ? -1 : 1);
      }
    }
  };

  // toJSON:: () → Object
  // Convert the selection to a JSON representation. When implementing
  // this for a custom selection class, make sure to give the object a
  // `type` property whose value matches the ID under which you
  // [registered](#state.Selection^jsonID) your class.

  // :: (ResolvedPos, number, ?bool) → ?Selection
  // Find a valid cursor or leaf node selection starting at the given
  // position and searching back if `dir` is negative, and forward if
  // positive. When `textOnly` is true, only consider cursor
  // selections. Will return null when no valid selection position is
  // found.
  Selection.findFrom = function findFrom ($pos, dir, textOnly) {
    var inner = $pos.parent.inlineContent ? new TextSelection($pos)
        : findSelectionIn($pos.node(0), $pos.parent, $pos.pos, $pos.index(), dir, textOnly);
    if (inner) { return inner }

    for (var depth = $pos.depth - 1; depth >= 0; depth--) {
      var found = dir < 0
          ? findSelectionIn($pos.node(0), $pos.node(depth), $pos.before(depth + 1), $pos.index(depth), dir, textOnly)
          : findSelectionIn($pos.node(0), $pos.node(depth), $pos.after(depth + 1), $pos.index(depth) + 1, dir, textOnly);
      if (found) { return found }
    }
  };

  // :: (ResolvedPos, ?number) → Selection
  // Find a valid cursor or leaf node selection near the given
  // position. Searches forward first by default, but if `bias` is
  // negative, it will search backwards first.
  Selection.near = function near ($pos, bias) {
      if ( bias === void 0 ) { bias = 1; }

    return this.findFrom($pos, bias) || this.findFrom($pos, -bias) || new AllSelection($pos.node(0))
  };

  // :: (Node) → Selection
  // Find the cursor or leaf node selection closest to the start of
  // the given document. Will return an
  // [`AllSelection`](#state.AllSelection) if no valid position
  // exists.
  Selection.atStart = function atStart (doc) {
    return findSelectionIn(doc, doc, 0, 0, 1) || new AllSelection(doc)
  };

  // :: (Node) → Selection
  // Find the cursor or leaf node selection closest to the end of the
  // given document.
  Selection.atEnd = function atEnd (doc) {
    return findSelectionIn(doc, doc, doc.content.size, doc.childCount, -1) || new AllSelection(doc)
  };

  // :: (Node, Object) → Selection
  // Deserialize the JSON representation of a selection. Must be
  // implemented for custom classes (as a static class method).
  Selection.fromJSON = function fromJSON (doc, json) {
    if (!json || !json.type) { throw new RangeError("Invalid input for Selection.fromJSON") }
    var cls = classesById[json.type];
    if (!cls) { throw new RangeError(("No selection type " + (json.type) + " defined")) }
    return cls.fromJSON(doc, json)
  };

  // :: (string, constructor<Selection>)
  // To be able to deserialize selections from JSON, custom selection
  // classes must register themselves with an ID string, so that they
  // can be disambiguated. Try to pick something that's unlikely to
  // clash with classes from other modules.
  Selection.jsonID = function jsonID (id, selectionClass) {
    if (id in classesById) { throw new RangeError("Duplicate use of selection JSON ID " + id) }
    classesById[id] = selectionClass;
    selectionClass.prototype.jsonID = id;
    return selectionClass
  };

  // :: () → SelectionBookmark
  // Get a [bookmark](#state.SelectionBookmark) for this selection,
  // which is a value that can be mapped without having access to a
  // current document, and later resolved to a real selection for a
  // given document again. (This is used mostly by the history to
  // track and restore old selections.) The default implementation of
  // this method just converts the selection to a text selection and
  // returns the bookmark for that.
  Selection.prototype.getBookmark = function getBookmark () {
    return TextSelection.between(this.$anchor, this.$head).getBookmark()
  };

  Object.defineProperties( Selection.prototype, prototypeAccessors$5 );

  // :: bool
  // Controls whether, when a selection of this type is active in the
  // browser, the selected range should be visible to the user. Defaults
  // to `true`.
  Selection.prototype.visible = true;

  // SelectionBookmark:: interface
  // A lightweight, document-independent representation of a selection.
  // You can define a custom bookmark type for a custom selection class
  // to make the history handle it well.
  //
  //   map:: (mapping: Mapping) → SelectionBookmark
  //   Map the bookmark through a set of changes.
  //
  //   resolve:: (doc: Node) → Selection
  //   Resolve the bookmark to a real selection again. This may need to
  //   do some error checking and may fall back to a default (usually
  //   [`TextSelection.between`](#state.TextSelection^between)) if
  //   mapping made the bookmark invalid.

  // ::- Represents a selected range in a document.
  var SelectionRange = function SelectionRange($from, $to) {
    // :: ResolvedPos
    // The lower bound of the range.
    this.$from = $from;
    // :: ResolvedPos
    // The upper bound of the range.
    this.$to = $to;
  };

  // ::- A text selection represents a classical editor selection, with
  // a head (the moving side) and anchor (immobile side), both of which
  // point into textblock nodes. It can be empty (a regular cursor
  // position).
  var TextSelection = /*@__PURE__*/(function (Selection) {
    function TextSelection($anchor, $head) {
      if ( $head === void 0 ) { $head = $anchor; }

      Selection.call(this, $anchor, $head);
    }

    if ( Selection ) { TextSelection.__proto__ = Selection; }
    TextSelection.prototype = Object.create( Selection && Selection.prototype );
    TextSelection.prototype.constructor = TextSelection;

    var prototypeAccessors$1 = { $cursor: { configurable: true } };

    // :: ?ResolvedPos
    // Returns a resolved position if this is a cursor selection (an
    // empty text selection), and null otherwise.
    prototypeAccessors$1.$cursor.get = function () { return this.$anchor.pos == this.$head.pos ? this.$head : null };

    TextSelection.prototype.map = function map (doc, mapping) {
      var $head = doc.resolve(mapping.map(this.head));
      if (!$head.parent.inlineContent) { return Selection.near($head) }
      var $anchor = doc.resolve(mapping.map(this.anchor));
      return new TextSelection($anchor.parent.inlineContent ? $anchor : $head, $head)
    };

    TextSelection.prototype.replace = function replace (tr, content) {
      if ( content === void 0 ) { content = Slice$1.empty; }

      Selection.prototype.replace.call(this, tr, content);
      if (content == Slice$1.empty) {
        var marks = this.$from.marksAcross(this.$to);
        if (marks) { tr.ensureMarks(marks); }
      }
    };

    TextSelection.prototype.eq = function eq (other) {
      return other instanceof TextSelection && other.anchor == this.anchor && other.head == this.head
    };

    TextSelection.prototype.getBookmark = function getBookmark () {
      return new TextBookmark(this.anchor, this.head)
    };

    TextSelection.prototype.toJSON = function toJSON () {
      return {type: "text", anchor: this.anchor, head: this.head}
    };

    TextSelection.fromJSON = function fromJSON (doc, json) {
      if (typeof json.anchor != "number" || typeof json.head != "number")
        { throw new RangeError("Invalid input for TextSelection.fromJSON") }
      return new TextSelection(doc.resolve(json.anchor), doc.resolve(json.head))
    };

    // :: (Node, number, ?number) → TextSelection
    // Create a text selection from non-resolved positions.
    TextSelection.create = function create (doc, anchor, head) {
      if ( head === void 0 ) { head = anchor; }

      var $anchor = doc.resolve(anchor);
      return new this($anchor, head == anchor ? $anchor : doc.resolve(head))
    };

    // :: (ResolvedPos, ResolvedPos, ?number) → Selection
    // Return a text selection that spans the given positions or, if
    // they aren't text positions, find a text selection near them.
    // `bias` determines whether the method searches forward (default)
    // or backwards (negative number) first. Will fall back to calling
    // [`Selection.near`](#state.Selection^near) when the document
    // doesn't contain a valid text position.
    TextSelection.between = function between ($anchor, $head, bias) {
      var dPos = $anchor.pos - $head.pos;
      if (!bias || dPos) { bias = dPos >= 0 ? 1 : -1; }
      if (!$head.parent.inlineContent) {
        var found = Selection.findFrom($head, bias, true) || Selection.findFrom($head, -bias, true);
        if (found) { $head = found.$head; }
        else { return Selection.near($head, bias) }
      }
      if (!$anchor.parent.inlineContent) {
        if (dPos == 0) {
          $anchor = $head;
        } else {
          $anchor = (Selection.findFrom($anchor, -bias, true) || Selection.findFrom($anchor, bias, true)).$anchor;
          if (($anchor.pos < $head.pos) != (dPos < 0)) { $anchor = $head; }
        }
      }
      return new TextSelection($anchor, $head)
    };

    Object.defineProperties( TextSelection.prototype, prototypeAccessors$1 );

    return TextSelection;
  }(Selection));

  Selection.jsonID("text", TextSelection);

  var TextBookmark = function TextBookmark(anchor, head) {
    this.anchor = anchor;
    this.head = head;
  };
  TextBookmark.prototype.map = function map (mapping) {
    return new TextBookmark(mapping.map(this.anchor), mapping.map(this.head))
  };
  TextBookmark.prototype.resolve = function resolve (doc) {
    return TextSelection.between(doc.resolve(this.anchor), doc.resolve(this.head))
  };

  // ::- A node selection is a selection that points at a single node.
  // All nodes marked [selectable](#model.NodeSpec.selectable) can be
  // the target of a node selection. In such a selection, `from` and
  // `to` point directly before and after the selected node, `anchor`
  // equals `from`, and `head` equals `to`..
  var NodeSelection = /*@__PURE__*/(function (Selection) {
    function NodeSelection($pos) {
      var node = $pos.nodeAfter;
      var $end = $pos.node(0).resolve($pos.pos + node.nodeSize);
      Selection.call(this, $pos, $end);
      // :: Node The selected node.
      this.node = node;
    }

    if ( Selection ) { NodeSelection.__proto__ = Selection; }
    NodeSelection.prototype = Object.create( Selection && Selection.prototype );
    NodeSelection.prototype.constructor = NodeSelection;

    NodeSelection.prototype.map = function map (doc, mapping) {
      var ref = mapping.mapResult(this.anchor);
      var deleted = ref.deleted;
      var pos = ref.pos;
      var $pos = doc.resolve(pos);
      if (deleted) { return Selection.near($pos) }
      return new NodeSelection($pos)
    };

    NodeSelection.prototype.content = function content () {
      return new Slice$1(Fragment$1.from(this.node), 0, 0)
    };

    NodeSelection.prototype.eq = function eq (other) {
      return other instanceof NodeSelection && other.anchor == this.anchor
    };

    NodeSelection.prototype.toJSON = function toJSON () {
      return {type: "node", anchor: this.anchor}
    };

    NodeSelection.prototype.getBookmark = function getBookmark () { return new NodeBookmark(this.anchor) };

    NodeSelection.fromJSON = function fromJSON (doc, json) {
      if (typeof json.anchor != "number")
        { throw new RangeError("Invalid input for NodeSelection.fromJSON") }
      return new NodeSelection(doc.resolve(json.anchor))
    };

    // :: (Node, number) → NodeSelection
    // Create a node selection from non-resolved positions.
    NodeSelection.create = function create (doc, from) {
      return new this(doc.resolve(from))
    };

    // :: (Node) → bool
    // Determines whether the given node may be selected as a node
    // selection.
    NodeSelection.isSelectable = function isSelectable (node) {
      return !node.isText && node.type.spec.selectable !== false
    };

    return NodeSelection;
  }(Selection));

  NodeSelection.prototype.visible = false;

  Selection.jsonID("node", NodeSelection);

  var NodeBookmark = function NodeBookmark(anchor) {
    this.anchor = anchor;
  };
  NodeBookmark.prototype.map = function map (mapping) {
    var ref = mapping.mapResult(this.anchor);
      var deleted = ref.deleted;
      var pos = ref.pos;
    return deleted ? new TextBookmark(pos, pos) : new NodeBookmark(pos)
  };
  NodeBookmark.prototype.resolve = function resolve (doc) {
    var $pos = doc.resolve(this.anchor), node = $pos.nodeAfter;
    if (node && NodeSelection.isSelectable(node)) { return new NodeSelection($pos) }
    return Selection.near($pos)
  };

  // ::- A selection type that represents selecting the whole document
  // (which can not necessarily be expressed with a text selection, when
  // there are for example leaf block nodes at the start or end of the
  // document).
  var AllSelection = /*@__PURE__*/(function (Selection) {
    function AllSelection(doc) {
      Selection.call(this, doc.resolve(0), doc.resolve(doc.content.size));
    }

    if ( Selection ) { AllSelection.__proto__ = Selection; }
    AllSelection.prototype = Object.create( Selection && Selection.prototype );
    AllSelection.prototype.constructor = AllSelection;

    AllSelection.prototype.replace = function replace (tr, content) {
      if ( content === void 0 ) { content = Slice$1.empty; }

      if (content == Slice$1.empty) {
        tr.delete(0, tr.doc.content.size);
        var sel = Selection.atStart(tr.doc);
        if (!sel.eq(tr.selection)) { tr.setSelection(sel); }
      } else {
        Selection.prototype.replace.call(this, tr, content);
      }
    };

    AllSelection.prototype.toJSON = function toJSON () { return {type: "all"} };

    AllSelection.fromJSON = function fromJSON (doc) { return new AllSelection(doc) };

    AllSelection.prototype.map = function map (doc) { return new AllSelection(doc) };

    AllSelection.prototype.eq = function eq (other) { return other instanceof AllSelection };

    AllSelection.prototype.getBookmark = function getBookmark () { return AllBookmark };

    return AllSelection;
  }(Selection));

  Selection.jsonID("all", AllSelection);

  var AllBookmark = {
    map: function map() { return this },
    resolve: function resolve(doc) { return new AllSelection(doc) }
  };

  // FIXME we'll need some awareness of text direction when scanning for selections

  // Try to find a selection inside the given node. `pos` points at the
  // position where the search starts. When `text` is true, only return
  // text selections.
  function findSelectionIn(doc, node, pos, index, dir, text) {
    if (node.inlineContent) { return TextSelection.create(doc, pos) }
    for (var i = index - (dir > 0 ? 0 : 1); dir > 0 ? i < node.childCount : i >= 0; i += dir) {
      var child = node.child(i);
      if (!child.isAtom) {
        var inner = findSelectionIn(doc, child, pos + dir, dir < 0 ? child.childCount : 0, dir, text);
        if (inner) { return inner }
      } else if (!text && NodeSelection.isSelectable(child)) {
        return NodeSelection.create(doc, pos - (dir < 0 ? child.nodeSize : 0))
      }
      pos += child.nodeSize * dir;
    }
  }

  function selectionToInsertionEnd(tr, startLen, bias) {
    var last = tr.steps.length - 1;
    if (last < startLen) { return }
    var step = tr.steps[last];
    if (!(step instanceof ReplaceStep || step instanceof ReplaceAroundStep)) { return }
    var map = tr.mapping.maps[last], end;
    map.forEach(function (_from, _to, _newFrom, newTo) { if (end == null) { end = newTo; } });
    tr.setSelection(Selection.near(tr.doc.resolve(end), bias));
  }

  var UPDATED_SEL = 1, UPDATED_MARKS = 2, UPDATED_SCROLL = 4;

  // ::- An editor state transaction, which can be applied to a state to
  // create an updated state. Use
  // [`EditorState.tr`](#state.EditorState.tr) to create an instance.
  //
  // Transactions track changes to the document (they are a subclass of
  // [`Transform`](#transform.Transform)), but also other state changes,
  // like selection updates and adjustments of the set of [stored
  // marks](#state.EditorState.storedMarks). In addition, you can store
  // metadata properties in a transaction, which are extra pieces of
  // information that client code or plugins can use to describe what a
  // transacion represents, so that they can update their [own
  // state](#state.StateField) accordingly.
  //
  // The [editor view](#view.EditorView) uses a few metadata properties:
  // it will attach a property `"pointer"` with the value `true` to
  // selection transactions directly caused by mouse or touch input, and
  // a `"uiEvent"` property of that may be `"paste"`, `"cut"`, or `"drop"`.
  var Transaction = /*@__PURE__*/(function (Transform) {
    function Transaction(state) {
      Transform.call(this, state.doc);
      // :: number
      // The timestamp associated with this transaction, in the same
      // format as `Date.now()`.
      this.time = Date.now();
      this.curSelection = state.selection;
      // The step count for which the current selection is valid.
      this.curSelectionFor = 0;
      // :: ?[Mark]
      // The stored marks set by this transaction, if any.
      this.storedMarks = state.storedMarks;
      // Bitfield to track which aspects of the state were updated by
      // this transaction.
      this.updated = 0;
      // Object used to store metadata properties for the transaction.
      this.meta = Object.create(null);
    }

    if ( Transform ) { Transaction.__proto__ = Transform; }
    Transaction.prototype = Object.create( Transform && Transform.prototype );
    Transaction.prototype.constructor = Transaction;

    var prototypeAccessors = { selection: { configurable: true },selectionSet: { configurable: true },storedMarksSet: { configurable: true },isGeneric: { configurable: true },scrolledIntoView: { configurable: true } };

    // :: Selection
    // The transaction's current selection. This defaults to the editor
    // selection [mapped](#state.Selection.map) through the steps in the
    // transaction, but can be overwritten with
    // [`setSelection`](#state.Transaction.setSelection).
    prototypeAccessors.selection.get = function () {
      if (this.curSelectionFor < this.steps.length) {
        this.curSelection = this.curSelection.map(this.doc, this.mapping.slice(this.curSelectionFor));
        this.curSelectionFor = this.steps.length;
      }
      return this.curSelection
    };

    // :: (Selection) → Transaction
    // Update the transaction's current selection. Will determine the
    // selection that the editor gets when the transaction is applied.
    Transaction.prototype.setSelection = function setSelection (selection) {
      if (selection.$from.doc != this.doc)
        { throw new RangeError("Selection passed to setSelection must point at the current document") }
      this.curSelection = selection;
      this.curSelectionFor = this.steps.length;
      this.updated = (this.updated | UPDATED_SEL) & ~UPDATED_MARKS;
      this.storedMarks = null;
      return this
    };

    // :: bool
    // Whether the selection was explicitly updated by this transaction.
    prototypeAccessors.selectionSet.get = function () {
      return (this.updated & UPDATED_SEL) > 0
    };

    // :: (?[Mark]) → Transaction
    // Set the current stored marks.
    Transaction.prototype.setStoredMarks = function setStoredMarks (marks) {
      this.storedMarks = marks;
      this.updated |= UPDATED_MARKS;
      return this
    };

    // :: ([Mark]) → Transaction
    // Make sure the current stored marks or, if that is null, the marks
    // at the selection, match the given set of marks. Does nothing if
    // this is already the case.
    Transaction.prototype.ensureMarks = function ensureMarks (marks) {
      if (!Mark$1.sameSet(this.storedMarks || this.selection.$from.marks(), marks))
        { this.setStoredMarks(marks); }
      return this
    };

    // :: (Mark) → Transaction
    // Add a mark to the set of stored marks.
    Transaction.prototype.addStoredMark = function addStoredMark (mark) {
      return this.ensureMarks(mark.addToSet(this.storedMarks || this.selection.$head.marks()))
    };

    // :: (union<Mark, MarkType>) → Transaction
    // Remove a mark or mark type from the set of stored marks.
    Transaction.prototype.removeStoredMark = function removeStoredMark (mark) {
      return this.ensureMarks(mark.removeFromSet(this.storedMarks || this.selection.$head.marks()))
    };

    // :: bool
    // Whether the stored marks were explicitly set for this transaction.
    prototypeAccessors.storedMarksSet.get = function () {
      return (this.updated & UPDATED_MARKS) > 0
    };

    Transaction.prototype.addStep = function addStep (step, doc) {
      Transform.prototype.addStep.call(this, step, doc);
      this.updated = this.updated & ~UPDATED_MARKS;
      this.storedMarks = null;
    };

    // :: (number) → Transaction
    // Update the timestamp for the transaction.
    Transaction.prototype.setTime = function setTime (time) {
      this.time = time;
      return this
    };

    // :: (Slice) → Transaction
    // Replace the current selection with the given slice.
    Transaction.prototype.replaceSelection = function replaceSelection (slice) {
      this.selection.replace(this, slice);
      return this
    };

    // :: (Node, ?bool) → Transaction
    // Replace the selection with the given node. When `inheritMarks` is
    // true and the content is inline, it inherits the marks from the
    // place where it is inserted.
    Transaction.prototype.replaceSelectionWith = function replaceSelectionWith (node, inheritMarks) {
      var selection = this.selection;
      if (inheritMarks !== false)
        { node = node.mark(this.storedMarks || (selection.empty ? selection.$from.marks() : (selection.$from.marksAcross(selection.$to) || Mark$1.none))); }
      selection.replaceWith(this, node);
      return this
    };

    // :: () → Transaction
    // Delete the selection.
    Transaction.prototype.deleteSelection = function deleteSelection () {
      this.selection.replace(this);
      return this
    };

    // :: (string, from: ?number, to: ?number) → Transaction
    // Replace the given range, or the selection if no range is given,
    // with a text node containing the given string.
    Transaction.prototype.insertText = function insertText (text, from, to) {
      if ( to === void 0 ) { to = from; }

      var schema = this.doc.type.schema;
      if (from == null) {
        if (!text) { return this.deleteSelection() }
        return this.replaceSelectionWith(schema.text(text), true)
      } else {
        if (!text) { return this.deleteRange(from, to) }
        var marks = this.storedMarks;
        if (!marks) {
          var $from = this.doc.resolve(from);
          marks = to == from ? $from.marks() : $from.marksAcross(this.doc.resolve(to));
        }
        this.replaceRangeWith(from, to, schema.text(text, marks));
        if (!this.selection.empty) { this.setSelection(Selection.near(this.selection.$to)); }
        return this
      }
    };

    // :: (union<string, Plugin, PluginKey>, any) → Transaction
    // Store a metadata property in this transaction, keyed either by
    // name or by plugin.
    Transaction.prototype.setMeta = function setMeta (key, value) {
      this.meta[typeof key == "string" ? key : key.key] = value;
      return this
    };

    // :: (union<string, Plugin, PluginKey>) → any
    // Retrieve a metadata property for a given name or plugin.
    Transaction.prototype.getMeta = function getMeta (key) {
      return this.meta[typeof key == "string" ? key : key.key]
    };

    // :: bool
    // Returns true if this transaction doesn't contain any metadata,
    // and can thus safely be extended.
    prototypeAccessors.isGeneric.get = function () {
      for (var _ in this.meta) { return false }
      return true
    };

    // :: () → Transaction
    // Indicate that the editor should scroll the selection into view
    // when updated to the state produced by this transaction.
    Transaction.prototype.scrollIntoView = function scrollIntoView () {
      this.updated |= UPDATED_SCROLL;
      return this
    };

    prototypeAccessors.scrolledIntoView.get = function () {
      return (this.updated & UPDATED_SCROLL) > 0
    };

    Object.defineProperties( Transaction.prototype, prototypeAccessors );

    return Transaction;
  }(Transform));

  function bind(f, self) {
    return !self || !f ? f : f.bind(self)
  }

  var FieldDesc = function FieldDesc(name, desc, self) {
    this.name = name;
    this.init = bind(desc.init, self);
    this.apply = bind(desc.apply, self);
  };

  var baseFields = [
    new FieldDesc("doc", {
      init: function init(config) { return config.doc || config.schema.topNodeType.createAndFill() },
      apply: function apply(tr) { return tr.doc }
    }),

    new FieldDesc("selection", {
      init: function init(config, instance) { return config.selection || Selection.atStart(instance.doc) },
      apply: function apply(tr) { return tr.selection }
    }),

    new FieldDesc("storedMarks", {
      init: function init(config) { return config.storedMarks || null },
      apply: function apply(tr, _marks, _old, state) { return state.selection.$cursor ? tr.storedMarks : null }
    }),

    new FieldDesc("scrollToSelection", {
      init: function init() { return 0 },
      apply: function apply(tr, prev) { return tr.scrolledIntoView ? prev + 1 : prev }
    })
  ];

  // Object wrapping the part of a state object that stays the same
  // across transactions. Stored in the state's `config` property.
  var Configuration = function Configuration(schema, plugins) {
    var this$1$1 = this;

    this.schema = schema;
    this.fields = baseFields.concat();
    this.plugins = [];
    this.pluginsByKey = Object.create(null);
    if (plugins) { plugins.forEach(function (plugin) {
      if (this$1$1.pluginsByKey[plugin.key])
        { throw new RangeError("Adding different instances of a keyed plugin (" + plugin.key + ")") }
      this$1$1.plugins.push(plugin);
      this$1$1.pluginsByKey[plugin.key] = plugin;
      if (plugin.spec.state)
        { this$1$1.fields.push(new FieldDesc(plugin.key, plugin.spec.state, plugin)); }
    }); }
  };

  // ::- The state of a ProseMirror editor is represented by an object
  // of this type. A state is a persistent data structure—it isn't
  // updated, but rather a new state value is computed from an old one
  // using the [`apply`](#state.EditorState.apply) method.
  //
  // A state holds a number of built-in fields, and plugins can
  // [define](#state.PluginSpec.state) additional fields.
  var EditorState = function EditorState(config) {
    this.config = config;
  };

  var prototypeAccessors$1$3 = { schema: { configurable: true },plugins: { configurable: true },tr: { configurable: true } };

  // doc:: Node
  // The current document.

  // selection:: Selection
  // The selection.

  // storedMarks:: ?[Mark]
  // A set of marks to apply to the next input. Will be null when
  // no explicit marks have been set.

  // :: Schema
  // The schema of the state's document.
  prototypeAccessors$1$3.schema.get = function () {
    return this.config.schema
  };

  // :: [Plugin]
  // The plugins that are active in this state.
  prototypeAccessors$1$3.plugins.get = function () {
    return this.config.plugins
  };

  // :: (Transaction) → EditorState
  // Apply the given transaction to produce a new state.
  EditorState.prototype.apply = function apply (tr) {
    return this.applyTransaction(tr).state
  };

  // : (Transaction) → bool
  EditorState.prototype.filterTransaction = function filterTransaction (tr, ignore) {
      if ( ignore === void 0 ) { ignore = -1; }

    for (var i = 0; i < this.config.plugins.length; i++) { if (i != ignore) {
      var plugin = this.config.plugins[i];
      if (plugin.spec.filterTransaction && !plugin.spec.filterTransaction.call(plugin, tr, this))
        { return false }
    } }
    return true
  };

  // :: (Transaction) → {state: EditorState, transactions: [Transaction]}
  // Verbose variant of [`apply`](#state.EditorState.apply) that
  // returns the precise transactions that were applied (which might
  // be influenced by the [transaction
  // hooks](#state.PluginSpec.filterTransaction) of
  // plugins) along with the new state.
  EditorState.prototype.applyTransaction = function applyTransaction (rootTr) {
    if (!this.filterTransaction(rootTr)) { return {state: this, transactions: []} }

    var trs = [rootTr], newState = this.applyInner(rootTr), seen = null;
    // This loop repeatedly gives plugins a chance to respond to
    // transactions as new transactions are added, making sure to only
    // pass the transactions the plugin did not see before.
     for (;;) {
      var haveNew = false;
      for (var i = 0; i < this.config.plugins.length; i++) {
        var plugin = this.config.plugins[i];
        if (plugin.spec.appendTransaction) {
          var n = seen ? seen[i].n : 0, oldState = seen ? seen[i].state : this;
          var tr = n < trs.length &&
              plugin.spec.appendTransaction.call(plugin, n ? trs.slice(n) : trs, oldState, newState);
          if (tr && newState.filterTransaction(tr, i)) {
            tr.setMeta("appendedTransaction", rootTr);
            if (!seen) {
              seen = [];
              for (var j = 0; j < this.config.plugins.length; j++)
                { seen.push(j < i ? {state: newState, n: trs.length} : {state: this, n: 0}); }
            }
            trs.push(tr);
            newState = newState.applyInner(tr);
            haveNew = true;
          }
          if (seen) { seen[i] = {state: newState, n: trs.length}; }
        }
      }
      if (!haveNew) { return {state: newState, transactions: trs} }
    }
  };

  // : (Transaction) → EditorState
  EditorState.prototype.applyInner = function applyInner (tr) {
    if (!tr.before.eq(this.doc)) { throw new RangeError("Applying a mismatched transaction") }
    var newInstance = new EditorState(this.config), fields = this.config.fields;
    for (var i = 0; i < fields.length; i++) {
      var field = fields[i];
      newInstance[field.name] = field.apply(tr, this[field.name], this, newInstance);
    }
    for (var i$1 = 0; i$1 < applyListeners.length; i$1++) { applyListeners[i$1](this, tr, newInstance); }
    return newInstance
  };

  // :: Transaction
  // Start a [transaction](#state.Transaction) from this state.
  prototypeAccessors$1$3.tr.get = function () { return new Transaction(this) };

  // :: (Object) → EditorState
  // Create a new state.
  //
  // config::- Configuration options. Must contain `schema` or `doc` (or both).
  //
  //    schema:: ?Schema
  //    The schema to use (only relevant if no `doc` is specified).
  //
  //    doc:: ?Node
  //    The starting document.
  //
  //    selection:: ?Selection
  //    A valid selection in the document.
  //
  //    storedMarks:: ?[Mark]
  //    The initial set of [stored marks](#state.EditorState.storedMarks).
  //
  //    plugins:: ?[Plugin]
  //    The plugins that should be active in this state.
  EditorState.create = function create (config) {
    var $config = new Configuration(config.doc ? config.doc.type.schema : config.schema, config.plugins);
    var instance = new EditorState($config);
    for (var i = 0; i < $config.fields.length; i++)
      { instance[$config.fields[i].name] = $config.fields[i].init(config, instance); }
    return instance
  };

  // :: (Object) → EditorState
  // Create a new state based on this one, but with an adjusted set of
  // active plugins. State fields that exist in both sets of plugins
  // are kept unchanged. Those that no longer exist are dropped, and
  // those that are new are initialized using their
  // [`init`](#state.StateField.init) method, passing in the new
  // configuration object..
  //
  // config::- configuration options
  //
  //   plugins:: [Plugin]
  //   New set of active plugins.
  EditorState.prototype.reconfigure = function reconfigure (config) {
    var $config = new Configuration(this.schema, config.plugins);
    var fields = $config.fields, instance = new EditorState($config);
    for (var i = 0; i < fields.length; i++) {
      var name = fields[i].name;
      instance[name] = this.hasOwnProperty(name) ? this[name] : fields[i].init(config, instance);
    }
    return instance
  };

  // :: (?union<Object<Plugin>, string, number>) → Object
  // Serialize this state to JSON. If you want to serialize the state
  // of plugins, pass an object mapping property names to use in the
  // resulting JSON object to plugin objects. The argument may also be
  // a string or number, in which case it is ignored, to support the
  // way `JSON.stringify` calls `toString` methods.
  EditorState.prototype.toJSON = function toJSON (pluginFields) {
    var result = {doc: this.doc.toJSON(), selection: this.selection.toJSON()};
    if (this.storedMarks) { result.storedMarks = this.storedMarks.map(function (m) { return m.toJSON(); }); }
    if (pluginFields && typeof pluginFields == 'object') { for (var prop in pluginFields) {
      if (prop == "doc" || prop == "selection")
        { throw new RangeError("The JSON fields `doc` and `selection` are reserved") }
      var plugin = pluginFields[prop], state = plugin.spec.state;
      if (state && state.toJSON) { result[prop] = state.toJSON.call(plugin, this[plugin.key]); }
    } }
    return result
  };

  // :: (Object, Object, ?Object<Plugin>) → EditorState
  // Deserialize a JSON representation of a state. `config` should
  // have at least a `schema` field, and should contain array of
  // plugins to initialize the state with. `pluginFields` can be used
  // to deserialize the state of plugins, by associating plugin
  // instances with the property names they use in the JSON object.
  //
  // config::- configuration options
  //
  //   schema:: Schema
  //   The schema to use.
  //
  //   plugins:: ?[Plugin]
  //   The set of active plugins.
  EditorState.fromJSON = function fromJSON (config, json, pluginFields) {
    if (!json) { throw new RangeError("Invalid input for EditorState.fromJSON") }
    if (!config.schema) { throw new RangeError("Required config field 'schema' missing") }
    var $config = new Configuration(config.schema, config.plugins);
    var instance = new EditorState($config);
    $config.fields.forEach(function (field) {
      if (field.name == "doc") {
        instance.doc = Node$2.fromJSON(config.schema, json.doc);
      } else if (field.name == "selection") {
        instance.selection = Selection.fromJSON(instance.doc, json.selection);
      } else if (field.name == "storedMarks") {
        if (json.storedMarks) { instance.storedMarks = json.storedMarks.map(config.schema.markFromJSON); }
      } else {
        if (pluginFields) { for (var prop in pluginFields) {
          var plugin = pluginFields[prop], state = plugin.spec.state;
          if (plugin.key == field.name && state && state.fromJSON &&
              Object.prototype.hasOwnProperty.call(json, prop)) {
            // This field belongs to a plugin mapped to a JSON field, read it from there.
            instance[field.name] = state.fromJSON.call(plugin, config, json[prop], instance);
            return
          }
        } }
        instance[field.name] = field.init(config, instance);
      }
    });
    return instance
  };

  // Kludge to allow the view to track mappings between different
  // instances of a state.
  //
  // FIXME this is no longer needed as of prosemirror-view 1.9.0,
  // though due to backwards-compat we should probably keep it around
  // for a while (if only as a no-op)
  EditorState.addApplyListener = function addApplyListener (f) {
    applyListeners.push(f);
  };
  EditorState.removeApplyListener = function removeApplyListener (f) {
    var found = applyListeners.indexOf(f);
    if (found > -1) { applyListeners.splice(found, 1); }
  };

  Object.defineProperties( EditorState.prototype, prototypeAccessors$1$3 );

  var applyListeners = [];

  // PluginSpec:: interface
  //
  // This is the type passed to the [`Plugin`](#state.Plugin)
  // constructor. It provides a definition for a plugin.
  //
  //   props:: ?EditorProps
  //   The [view props](#view.EditorProps) added by this plugin. Props
  //   that are functions will be bound to have the plugin instance as
  //   their `this` binding.
  //
  //   state:: ?StateField<any>
  //   Allows a plugin to define a [state field](#state.StateField), an
  //   extra slot in the state object in which it can keep its own data.
  //
  //   key:: ?PluginKey
  //   Can be used to make this a keyed plugin. You can have only one
  //   plugin with a given key in a given state, but it is possible to
  //   access the plugin's configuration and state through the key,
  //   without having access to the plugin instance object.
  //
  //   view:: ?(EditorView) → Object
  //   When the plugin needs to interact with the editor view, or
  //   set something up in the DOM, use this field. The function
  //   will be called when the plugin's state is associated with an
  //   editor view.
  //
  //     return::-
  //     Should return an object with the following optional
  //     properties:
  //
  //       update:: ?(view: EditorView, prevState: EditorState)
  //       Called whenever the view's state is updated.
  //
  //       destroy:: ?()
  //       Called when the view is destroyed or receives a state
  //       with different plugins.
  //
  //   filterTransaction:: ?(Transaction, EditorState) → bool
  //   When present, this will be called before a transaction is
  //   applied by the state, allowing the plugin to cancel it (by
  //   returning false).
  //
  //   appendTransaction:: ?(transactions: [Transaction], oldState: EditorState, newState: EditorState) → ?Transaction
  //   Allows the plugin to append another transaction to be applied
  //   after the given array of transactions. When another plugin
  //   appends a transaction after this was called, it is called again
  //   with the new state and new transactions—but only the new
  //   transactions, i.e. it won't be passed transactions that it
  //   already saw.

  function bindProps(obj, self, target) {
    for (var prop in obj) {
      var val = obj[prop];
      if (val instanceof Function) { val = val.bind(self); }
      else if (prop == "handleDOMEvents") { val = bindProps(val, self, {}); }
      target[prop] = val;
    }
    return target
  }

  // ::- Plugins bundle functionality that can be added to an editor.
  // They are part of the [editor state](#state.EditorState) and
  // may influence that state and the view that contains it.
  var Plugin = function Plugin(spec) {
    // :: EditorProps
    // The [props](#view.EditorProps) exported by this plugin.
    this.props = {};
    if (spec.props) { bindProps(spec.props, this, this.props); }
    // :: Object
    // The plugin's [spec object](#state.PluginSpec).
    this.spec = spec;
    this.key = spec.key ? spec.key.key : createKey("plugin");
  };

  // :: (EditorState) → any
  // Extract the plugin's state field from an editor state.
  Plugin.prototype.getState = function getState (state) { return state[this.key] };

  // StateField:: interface<T>
  // A plugin spec may provide a state field (under its
  // [`state`](#state.PluginSpec.state) property) of this type, which
  // describes the state it wants to keep. Functions provided here are
  // always called with the plugin instance as their `this` binding.
  //
  //   init:: (config: Object, instance: EditorState) → T
  //   Initialize the value of the field. `config` will be the object
  //   passed to [`EditorState.create`](#state.EditorState^create). Note
  //   that `instance` is a half-initialized state instance, and will
  //   not have values for plugin fields initialized after this one.
  //
  //   apply:: (tr: Transaction, value: T, oldState: EditorState, newState: EditorState) → T
  //   Apply the given transaction to this state field, producing a new
  //   field value. Note that the `newState` argument is again a partially
  //   constructed state does not yet contain the state from plugins
  //   coming after this one.
  //
  //   toJSON:: ?(value: T) → *
  //   Convert this field to JSON. Optional, can be left off to disable
  //   JSON serialization for the field.
  //
  //   fromJSON:: ?(config: Object, value: *, state: EditorState) → T
  //   Deserialize the JSON representation of this field. Note that the
  //   `state` argument is again a half-initialized state.

  var keys = Object.create(null);

  function createKey(name) {
    if (name in keys) { return name + "$" + ++keys[name] }
    keys[name] = 0;
    return name + "$"
  }

  // ::- A key is used to [tag](#state.PluginSpec.key)
  // plugins in a way that makes it possible to find them, given an
  // editor state. Assigning a key does mean only one plugin of that
  // type can be active in a state.
  var PluginKey = function PluginKey(name) {
  if ( name === void 0 ) { name = "key"; }
   this.key = createKey(name); };

  // :: (EditorState) → ?Plugin
  // Get the active plugin with this key, if any, from an editor
  // state.
  PluginKey.prototype.get = function get (state) { return state.config.pluginsByKey[this.key] };

  // :: (EditorState) → ?any
  // Get the plugin's state from an editor state.
  PluginKey.prototype.getState = function getState (state) { return state[this.key] };

  var state = /*#__PURE__*/Object.freeze({
    __proto__: null,
    AllSelection: AllSelection,
    EditorState: EditorState,
    NodeSelection: NodeSelection,
    Plugin: Plugin,
    PluginKey: PluginKey,
    Selection: Selection,
    SelectionRange: SelectionRange,
    TextSelection: TextSelection,
    Transaction: Transaction
  });

  var result = {};

  if (typeof navigator != "undefined" && typeof document != "undefined") {
    var ie_edge = /Edge\/(\d+)/.exec(navigator.userAgent);
    var ie_upto10 = /MSIE \d/.test(navigator.userAgent);
    var ie_11up = /Trident\/(?:[7-9]|\d{2,})\..*rv:(\d+)/.exec(navigator.userAgent);

    result.mac = /Mac/.test(navigator.platform);
    var ie$1 = result.ie = !!(ie_upto10 || ie_11up || ie_edge);
    result.ie_version = ie_upto10 ? document.documentMode || 6 : ie_11up ? +ie_11up[1] : ie_edge ? +ie_edge[1] : null;
    result.gecko = !ie$1 && /gecko\/(\d+)/i.test(navigator.userAgent);
    result.gecko_version = result.gecko && +(/Firefox\/(\d+)/.exec(navigator.userAgent) || [0, 0])[1];
    var chrome$1 = !ie$1 && /Chrome\/(\d+)/.exec(navigator.userAgent);
    result.chrome = !!chrome$1;
    result.chrome_version = chrome$1 && +chrome$1[1];
    // Is true for both iOS and iPadOS for convenience
    result.safari = !ie$1 && /Apple Computer/.test(navigator.vendor);
    result.ios = result.safari && (/Mobile\/\w+/.test(navigator.userAgent) || navigator.maxTouchPoints > 2);
    result.android = /Android \d/.test(navigator.userAgent);
    result.webkit = "webkitFontSmoothing" in document.documentElement.style;
    result.webkit_version = result.webkit && +(/\bAppleWebKit\/(\d+)/.exec(navigator.userAgent) || [0, 0])[1];
  }

  var domIndex = function(node) {
    for (var index = 0;; index++) {
      node = node.previousSibling;
      if (!node) { return index }
    }
  };

  var parentNode = function(node) {
    var parent = node.assignedSlot || node.parentNode;
    return parent && parent.nodeType == 11 ? parent.host : parent
  };

  var reusedRange = null;

  // Note that this will always return the same range, because DOM range
  // objects are every expensive, and keep slowing down subsequent DOM
  // updates, for some reason.
  var textRange = function(node, from, to) {
    var range = reusedRange || (reusedRange = document.createRange());
    range.setEnd(node, to == null ? node.nodeValue.length : to);
    range.setStart(node, from || 0);
    return range
  };

  // Scans forward and backward through DOM positions equivalent to the
  // given one to see if the two are in the same place (i.e. after a
  // text node vs at the end of that text node)
  var isEquivalentPosition = function(node, off, targetNode, targetOff) {
    return targetNode && (scanFor(node, off, targetNode, targetOff, -1) ||
                          scanFor(node, off, targetNode, targetOff, 1))
  };

  var atomElements = /^(img|br|input|textarea|hr)$/i;

  function scanFor(node, off, targetNode, targetOff, dir) {
    for (;;) {
      if (node == targetNode && off == targetOff) { return true }
      if (off == (dir < 0 ? 0 : nodeSize(node))) {
        var parent = node.parentNode;
        if (parent.nodeType != 1 || hasBlockDesc(node) || atomElements.test(node.nodeName) || node.contentEditable == "false")
          { return false }
        off = domIndex(node) + (dir < 0 ? 0 : 1);
        node = parent;
      } else if (node.nodeType == 1) {
        node = node.childNodes[off + (dir < 0 ? -1 : 0)];
        if (node.contentEditable == "false") { return false }
        off = dir < 0 ? nodeSize(node) : 0;
      } else {
        return false
      }
    }
  }

  function nodeSize(node) {
    return node.nodeType == 3 ? node.nodeValue.length : node.childNodes.length
  }

  function isOnEdge(node, offset, parent) {
    for (var atStart = offset == 0, atEnd = offset == nodeSize(node); atStart || atEnd;) {
      if (node == parent) { return true }
      var index = domIndex(node);
      node = node.parentNode;
      if (!node) { return false }
      atStart = atStart && index == 0;
      atEnd = atEnd && index == nodeSize(node);
    }
  }

  function hasBlockDesc(dom) {
    var desc;
    for (var cur = dom; cur; cur = cur.parentNode) { if (desc = cur.pmViewDesc) { break } }
    return desc && desc.node && desc.node.isBlock && (desc.dom == dom || desc.contentDOM == dom)
  }

  // Work around Chrome issue https://bugs.chromium.org/p/chromium/issues/detail?id=447523
  // (isCollapsed inappropriately returns true in shadow dom)
  var selectionCollapsed = function(domSel) {
    var collapsed = domSel.isCollapsed;
    if (collapsed && result.chrome && domSel.rangeCount && !domSel.getRangeAt(0).collapsed)
      { collapsed = false; }
    return collapsed
  };

  function keyEvent(keyCode, key) {
    var event = document.createEvent("Event");
    event.initEvent("keydown", true, true);
    event.keyCode = keyCode;
    event.key = event.code = key;
    return event
  }

  function windowRect(doc) {
    return {left: 0, right: doc.documentElement.clientWidth,
            top: 0, bottom: doc.documentElement.clientHeight}
  }

  function getSide(value, side) {
    return typeof value == "number" ? value : value[side]
  }

  function clientRect(node) {
    var rect = node.getBoundingClientRect();
    // Adjust for elements with style "transform: scale()"
    var scaleX = (rect.width / node.offsetWidth) || 1;
    var scaleY = (rect.height / node.offsetHeight) || 1;
    // Make sure scrollbar width isn't included in the rectangle
    return {left: rect.left, right: rect.left + node.clientWidth * scaleX,
            top: rect.top, bottom: rect.top + node.clientHeight * scaleY}
  }

  function scrollRectIntoView(view, rect, startDOM) {
    var scrollThreshold = view.someProp("scrollThreshold") || 0, scrollMargin = view.someProp("scrollMargin") || 5;
    var doc = view.dom.ownerDocument;
    for (var parent = startDOM || view.dom;; parent = parentNode(parent)) {
      if (!parent) { break }
      if (parent.nodeType != 1) { continue }
      var atTop = parent == doc.body || parent.nodeType != 1;
      var bounding = atTop ? windowRect(doc) : clientRect(parent);
      var moveX = 0, moveY = 0;
      if (rect.top < bounding.top + getSide(scrollThreshold, "top"))
        { moveY = -(bounding.top - rect.top + getSide(scrollMargin, "top")); }
      else if (rect.bottom > bounding.bottom - getSide(scrollThreshold, "bottom"))
        { moveY = rect.bottom - bounding.bottom + getSide(scrollMargin, "bottom"); }
      if (rect.left < bounding.left + getSide(scrollThreshold, "left"))
        { moveX = -(bounding.left - rect.left + getSide(scrollMargin, "left")); }
      else if (rect.right > bounding.right - getSide(scrollThreshold, "right"))
        { moveX = rect.right - bounding.right + getSide(scrollMargin, "right"); }
      if (moveX || moveY) {
        if (atTop) {
          doc.defaultView.scrollBy(moveX, moveY);
        } else {
          var startX = parent.scrollLeft, startY = parent.scrollTop;
          if (moveY) { parent.scrollTop += moveY; }
          if (moveX) { parent.scrollLeft += moveX; }
          var dX = parent.scrollLeft - startX, dY = parent.scrollTop - startY;
          rect = {left: rect.left - dX, top: rect.top - dY, right: rect.right - dX, bottom: rect.bottom - dY};
        }
      }
      if (atTop) { break }
    }
  }

  // Store the scroll position of the editor's parent nodes, along with
  // the top position of an element near the top of the editor, which
  // will be used to make sure the visible viewport remains stable even
  // when the size of the content above changes.
  function storeScrollPos(view) {
    var rect = view.dom.getBoundingClientRect(), startY = Math.max(0, rect.top);
    var refDOM, refTop;
    for (var x = (rect.left + rect.right) / 2, y = startY + 1;
         y < Math.min(innerHeight, rect.bottom); y += 5) {
      var dom = view.root.elementFromPoint(x, y);
      if (dom == view.dom || !view.dom.contains(dom)) { continue }
      var localRect = dom.getBoundingClientRect();
      if (localRect.top >= startY - 20) {
        refDOM = dom;
        refTop = localRect.top;
        break
      }
    }
    return {refDOM: refDOM, refTop: refTop, stack: scrollStack(view.dom)}
  }

  function scrollStack(dom) {
    var stack = [], doc = dom.ownerDocument;
    for (; dom; dom = parentNode(dom)) {
      stack.push({dom: dom, top: dom.scrollTop, left: dom.scrollLeft});
      if (dom == doc) { break }
    }
    return stack
  }

  // Reset the scroll position of the editor's parent nodes to that what
  // it was before, when storeScrollPos was called.
  function resetScrollPos(ref) {
    var refDOM = ref.refDOM;
    var refTop = ref.refTop;
    var stack = ref.stack;

    var newRefTop = refDOM ? refDOM.getBoundingClientRect().top : 0;
    restoreScrollStack(stack, newRefTop == 0 ? 0 : newRefTop - refTop);
  }

  function restoreScrollStack(stack, dTop) {
    for (var i = 0; i < stack.length; i++) {
      var ref = stack[i];
      var dom = ref.dom;
      var top = ref.top;
      var left = ref.left;
      if (dom.scrollTop != top + dTop) { dom.scrollTop = top + dTop; }
      if (dom.scrollLeft != left) { dom.scrollLeft = left; }
    }
  }

  var preventScrollSupported = null;
  // Feature-detects support for .focus({preventScroll: true}), and uses
  // a fallback kludge when not supported.
  function focusPreventScroll(dom) {
    if (dom.setActive) { return dom.setActive() } // in IE
    if (preventScrollSupported) { return dom.focus(preventScrollSupported) }

    var stored = scrollStack(dom);
    dom.focus(preventScrollSupported == null ? {
      get preventScroll() {
        preventScrollSupported = {preventScroll: true};
        return true
      }
    } : undefined);
    if (!preventScrollSupported) {
      preventScrollSupported = false;
      restoreScrollStack(stored, 0);
    }
  }

  function findOffsetInNode(node, coords) {
    var closest, dxClosest = 2e8, coordsClosest, offset = 0;
    var rowBot = coords.top, rowTop = coords.top;
    for (var child = node.firstChild, childIndex = 0; child; child = child.nextSibling, childIndex++) {
      var rects = (void 0);
      if (child.nodeType == 1) { rects = child.getClientRects(); }
      else if (child.nodeType == 3) { rects = textRange(child).getClientRects(); }
      else { continue }

      for (var i = 0; i < rects.length; i++) {
        var rect = rects[i];
        if (rect.top <= rowBot && rect.bottom >= rowTop) {
          rowBot = Math.max(rect.bottom, rowBot);
          rowTop = Math.min(rect.top, rowTop);
          var dx = rect.left > coords.left ? rect.left - coords.left
              : rect.right < coords.left ? coords.left - rect.right : 0;
          if (dx < dxClosest) {
            closest = child;
            dxClosest = dx;
            coordsClosest = dx && closest.nodeType == 3 ? {left: rect.right < coords.left ? rect.right : rect.left, top: coords.top} : coords;
            if (child.nodeType == 1 && dx)
              { offset = childIndex + (coords.left >= (rect.left + rect.right) / 2 ? 1 : 0); }
            continue
          }
        }
        if (!closest && (coords.left >= rect.right && coords.top >= rect.top ||
                         coords.left >= rect.left && coords.top >= rect.bottom))
          { offset = childIndex + 1; }
      }
    }
    if (closest && closest.nodeType == 3) { return findOffsetInText(closest, coordsClosest) }
    if (!closest || (dxClosest && closest.nodeType == 1)) { return {node: node, offset: offset} }
    return findOffsetInNode(closest, coordsClosest)
  }

  function findOffsetInText(node, coords) {
    var len = node.nodeValue.length;
    var range = document.createRange();
    for (var i = 0; i < len; i++) {
      range.setEnd(node, i + 1);
      range.setStart(node, i);
      var rect = singleRect(range, 1);
      if (rect.top == rect.bottom) { continue }
      if (inRect(coords, rect))
        { return {node: node, offset: i + (coords.left >= (rect.left + rect.right) / 2 ? 1 : 0)} }
    }
    return {node: node, offset: 0}
  }

  function inRect(coords, rect) {
    return coords.left >= rect.left - 1 && coords.left <= rect.right + 1&&
      coords.top >= rect.top - 1 && coords.top <= rect.bottom + 1
  }

  function targetKludge(dom, coords) {
    var parent = dom.parentNode;
    if (parent && /^li$/i.test(parent.nodeName) && coords.left < dom.getBoundingClientRect().left)
      { return parent }
    return dom
  }

  function posFromElement(view, elt, coords) {
    var ref = findOffsetInNode(elt, coords);
    var node = ref.node;
    var offset = ref.offset;
    var bias = -1;
    if (node.nodeType == 1 && !node.firstChild) {
      var rect = node.getBoundingClientRect();
      bias = rect.left != rect.right && coords.left > (rect.left + rect.right) / 2 ? 1 : -1;
    }
    return view.docView.posFromDOM(node, offset, bias)
  }

  function posFromCaret(view, node, offset, coords) {
    // Browser (in caretPosition/RangeFromPoint) will agressively
    // normalize towards nearby inline nodes. Since we are interested in
    // positions between block nodes too, we first walk up the hierarchy
    // of nodes to see if there are block nodes that the coordinates
    // fall outside of. If so, we take the position before/after that
    // block. If not, we call `posFromDOM` on the raw node/offset.
    var outside = -1;
    for (var cur = node;;) {
      if (cur == view.dom) { break }
      var desc = view.docView.nearestDesc(cur, true);
      if (!desc) { return null }
      if (desc.node.isBlock && desc.parent) {
        var rect = desc.dom.getBoundingClientRect();
        if (rect.left > coords.left || rect.top > coords.top) { outside = desc.posBefore; }
        else if (rect.right < coords.left || rect.bottom < coords.top) { outside = desc.posAfter; }
        else { break }
      }
      cur = desc.dom.parentNode;
    }
    return outside > -1 ? outside : view.docView.posFromDOM(node, offset)
  }

  function elementFromPoint(element, coords, box) {
    var len = element.childNodes.length;
    if (len && box.top < box.bottom) {
      for (var startI = Math.max(0, Math.min(len - 1, Math.floor(len * (coords.top - box.top) / (box.bottom - box.top)) - 2)), i = startI;;) {
        var child = element.childNodes[i];
        if (child.nodeType == 1) {
          var rects = child.getClientRects();
          for (var j = 0; j < rects.length; j++) {
            var rect = rects[j];
            if (inRect(coords, rect)) { return elementFromPoint(child, coords, rect) }
          }
        }
        if ((i = (i + 1) % len) == startI) { break }
      }
    }
    return element
  }

  // Given an x,y position on the editor, get the position in the document.
  function posAtCoords(view, coords) {
    var assign, assign$1;

    var root = view.root, node, offset;
    if (root.caretPositionFromPoint) {
      try { // Firefox throws for this call in hard-to-predict circumstances (#994)
        var pos$1 = root.caretPositionFromPoint(coords.left, coords.top);
        if (pos$1) { ((assign = pos$1, node = assign.offsetNode, offset = assign.offset)); }
      } catch (_) {}
    }
    if (!node && root.caretRangeFromPoint) {
      var range = root.caretRangeFromPoint(coords.left, coords.top);
      if (range) { ((assign$1 = range, node = assign$1.startContainer, offset = assign$1.startOffset)); }
    }

    var elt = root.elementFromPoint(coords.left, coords.top + 1), pos;
    if (!elt || !view.dom.contains(elt.nodeType != 1 ? elt.parentNode : elt)) {
      var box = view.dom.getBoundingClientRect();
      if (!inRect(coords, box)) { return null }
      elt = elementFromPoint(view.dom, coords, box);
      if (!elt) { return null }
    }
    // Safari's caretRangeFromPoint returns nonsense when on a draggable element
    if (result.safari) {
      for (var p = elt; node && p; p = parentNode(p))
        { if (p.draggable) { node = offset = null; } }
    }
    elt = targetKludge(elt, coords);
    if (node) {
      if (result.gecko && node.nodeType == 1) {
        // Firefox will sometimes return offsets into <input> nodes, which
        // have no actual children, from caretPositionFromPoint (#953)
        offset = Math.min(offset, node.childNodes.length);
        // It'll also move the returned position before image nodes,
        // even if those are behind it.
        if (offset < node.childNodes.length) {
          var next = node.childNodes[offset], box$1;
          if (next.nodeName == "IMG" && (box$1 = next.getBoundingClientRect()).right <= coords.left &&
              box$1.bottom > coords.top)
            { offset++; }
        }
      }
      // Suspiciously specific kludge to work around caret*FromPoint
      // never returning a position at the end of the document
      if (node == view.dom && offset == node.childNodes.length - 1 && node.lastChild.nodeType == 1 &&
          coords.top > node.lastChild.getBoundingClientRect().bottom)
        { pos = view.state.doc.content.size; }
      // Ignore positions directly after a BR, since caret*FromPoint
      // 'round up' positions that would be more accurately placed
      // before the BR node.
      else if (offset == 0 || node.nodeType != 1 || node.childNodes[offset - 1].nodeName != "BR")
        { pos = posFromCaret(view, node, offset, coords); }
    }
    if (pos == null) { pos = posFromElement(view, elt, coords); }

    var desc = view.docView.nearestDesc(elt, true);
    return {pos: pos, inside: desc ? desc.posAtStart - desc.border : -1}
  }

  function singleRect(object, bias) {
    var rects = object.getClientRects();
    return !rects.length ? object.getBoundingClientRect() : rects[bias < 0 ? 0 : rects.length - 1]
  }

  var BIDI = /[\u0590-\u05f4\u0600-\u06ff\u0700-\u08ac]/;

  // : (EditorView, number, number) → {left: number, top: number, right: number, bottom: number}
  // Given a position in the document model, get a bounding box of the
  // character at that position, relative to the window.
  function coordsAtPos(view, pos, side) {
    var ref = view.docView.domFromPos(pos, side < 0 ? -1 : 1);
    var node = ref.node;
    var offset = ref.offset;

    var supportEmptyRange = result.webkit || result.gecko;
    if (node.nodeType == 3) {
      // These browsers support querying empty text ranges. Prefer that in
      // bidi context or when at the end of a node.
      if (supportEmptyRange && (BIDI.test(node.nodeValue) || (side < 0 ? !offset : offset == node.nodeValue.length))) {
        var rect = singleRect(textRange(node, offset, offset), side);
        // Firefox returns bad results (the position before the space)
        // when querying a position directly after line-broken
        // whitespace. Detect this situation and and kludge around it
        if (result.gecko && offset && /\s/.test(node.nodeValue[offset - 1]) && offset < node.nodeValue.length) {
          var rectBefore = singleRect(textRange(node, offset - 1, offset - 1), -1);
          if (rectBefore.top == rect.top) {
            var rectAfter = singleRect(textRange(node, offset, offset + 1), -1);
            if (rectAfter.top != rect.top)
              { return flattenV(rectAfter, rectAfter.left < rectBefore.left) }
          }
        }
        return rect
      } else {
        var from = offset, to = offset, takeSide = side < 0 ? 1 : -1;
        if (side < 0 && !offset) { to++; takeSide = -1; }
        else if (side >= 0 && offset == node.nodeValue.length) { from--; takeSide = 1; }
        else if (side < 0) { from--; }
        else { to ++; }
        return flattenV(singleRect(textRange(node, from, to), takeSide), takeSide < 0)
      }
    }

    // Return a horizontal line in block context
    if (!view.state.doc.resolve(pos).parent.inlineContent) {
      if (offset && (side < 0 || offset == nodeSize(node))) {
        var before = node.childNodes[offset - 1];
        if (before.nodeType == 1) { return flattenH(before.getBoundingClientRect(), false) }
      }
      if (offset < nodeSize(node)) {
        var after = node.childNodes[offset];
        if (after.nodeType == 1) { return flattenH(after.getBoundingClientRect(), true) }
      }
      return flattenH(node.getBoundingClientRect(), side >= 0)
    }

    // Inline, not in text node (this is not Bidi-safe)
    if (offset && (side < 0 || offset == nodeSize(node))) {
      var before$1 = node.childNodes[offset - 1];
      var target = before$1.nodeType == 3 ? textRange(before$1, nodeSize(before$1) - (supportEmptyRange ? 0 : 1))
          // BR nodes tend to only return the rectangle before them.
          // Only use them if they are the last element in their parent
          : before$1.nodeType == 1 && (before$1.nodeName != "BR" || !before$1.nextSibling) ? before$1 : null;
      if (target) { return flattenV(singleRect(target, 1), false) }
    }
    if (offset < nodeSize(node)) {
      var after$1 = node.childNodes[offset];
      var target$1 = after$1.nodeType == 3 ? textRange(after$1, 0, (supportEmptyRange ? 0 : 1))
          : after$1.nodeType == 1 ? after$1 : null;
      if (target$1) { return flattenV(singleRect(target$1, -1), true) }
    }
    // All else failed, just try to get a rectangle for the target node
    return flattenV(singleRect(node.nodeType == 3 ? textRange(node) : node, -side), side >= 0)
  }

  function flattenV(rect, left) {
    if (rect.width == 0) { return rect }
    var x = left ? rect.left : rect.right;
    return {top: rect.top, bottom: rect.bottom, left: x, right: x}
  }

  function flattenH(rect, top) {
    if (rect.height == 0) { return rect }
    var y = top ? rect.top : rect.bottom;
    return {top: y, bottom: y, left: rect.left, right: rect.right}
  }

  function withFlushedState(view, state, f) {
    var viewState = view.state, active = view.root.activeElement;
    if (viewState != state) { view.updateState(state); }
    if (active != view.dom) { view.focus(); }
    try {
      return f()
    } finally {
      if (viewState != state) { view.updateState(viewState); }
      if (active != view.dom && active) { active.focus(); }
    }
  }

  // : (EditorView, number, number)
  // Whether vertical position motion in a given direction
  // from a position would leave a text block.
  function endOfTextblockVertical(view, state, dir) {
    var sel = state.selection;
    var $pos = dir == "up" ? sel.$from : sel.$to;
    return withFlushedState(view, state, function () {
      var ref = view.docView.domFromPos($pos.pos, dir == "up" ? -1 : 1);
      var dom = ref.node;
      for (;;) {
        var nearest = view.docView.nearestDesc(dom, true);
        if (!nearest) { break }
        if (nearest.node.isBlock) { dom = nearest.dom; break }
        dom = nearest.dom.parentNode;
      }
      var coords = coordsAtPos(view, $pos.pos, 1);
      for (var child = dom.firstChild; child; child = child.nextSibling) {
        var boxes = (void 0);
        if (child.nodeType == 1) { boxes = child.getClientRects(); }
        else if (child.nodeType == 3) { boxes = textRange(child, 0, child.nodeValue.length).getClientRects(); }
        else { continue }
        for (var i = 0; i < boxes.length; i++) {
          var box = boxes[i];
          if (box.bottom > box.top && (dir == "up" ? box.bottom < coords.top + 1 : box.top > coords.bottom - 1))
            { return false }
        }
      }
      return true
    })
  }

  var maybeRTL = /[\u0590-\u08ac]/;

  function endOfTextblockHorizontal(view, state, dir) {
    var ref = state.selection;
    var $head = ref.$head;
    if (!$head.parent.isTextblock) { return false }
    var offset = $head.parentOffset, atStart = !offset, atEnd = offset == $head.parent.content.size;
    var sel = view.root.getSelection();
    // If the textblock is all LTR, or the browser doesn't support
    // Selection.modify (Edge), fall back to a primitive approach
    if (!maybeRTL.test($head.parent.textContent) || !sel.modify)
      { return dir == "left" || dir == "backward" ? atStart : atEnd }

    return withFlushedState(view, state, function () {
      // This is a huge hack, but appears to be the best we can
      // currently do: use `Selection.modify` to move the selection by
      // one character, and see if that moves the cursor out of the
      // textblock (or doesn't move it at all, when at the start/end of
      // the document).
      var oldRange = sel.getRangeAt(0), oldNode = sel.focusNode, oldOff = sel.focusOffset;
      var oldBidiLevel = sel.caretBidiLevel; // Only for Firefox
      sel.modify("move", dir, "character");
      var parentDOM = $head.depth ? view.docView.domAfterPos($head.before()) : view.dom;
      var result = !parentDOM.contains(sel.focusNode.nodeType == 1 ? sel.focusNode : sel.focusNode.parentNode) ||
          (oldNode == sel.focusNode && oldOff == sel.focusOffset);
      // Restore the previous selection
      sel.removeAllRanges();
      sel.addRange(oldRange);
      if (oldBidiLevel != null) { sel.caretBidiLevel = oldBidiLevel; }
      return result
    })
  }

  var cachedState = null, cachedDir = null, cachedResult = false;
  function endOfTextblock(view, state, dir) {
    if (cachedState == state && cachedDir == dir) { return cachedResult }
    cachedState = state; cachedDir = dir;
    return cachedResult = dir == "up" || dir == "down"
      ? endOfTextblockVertical(view, state, dir)
      : endOfTextblockHorizontal(view, state, dir)
  }

  // NodeView:: interface
  //
  // By default, document nodes are rendered using the result of the
  // [`toDOM`](#model.NodeSpec.toDOM) method of their spec, and managed
  // entirely by the editor. For some use cases, such as embedded
  // node-specific editing interfaces, you want more control over
  // the behavior of a node's in-editor representation, and need to
  // [define](#view.EditorProps.nodeViews) a custom node view.
  //
  // Mark views only support `dom` and `contentDOM`, and don't support
  // any of the node view methods.
  //
  // Objects returned as node views must conform to this interface.
  //
  //   dom:: ?dom.Node
  //   The outer DOM node that represents the document node. When not
  //   given, the default strategy is used to create a DOM node.
  //
  //   contentDOM:: ?dom.Node
  //   The DOM node that should hold the node's content. Only meaningful
  //   if the node view also defines a `dom` property and if its node
  //   type is not a leaf node type. When this is present, ProseMirror
  //   will take care of rendering the node's children into it. When it
  //   is not present, the node view itself is responsible for rendering
  //   (or deciding not to render) its child nodes.
  //
  //   update:: ?(node: Node, decorations: [Decoration], innerDecorations: DecorationSource) → bool
  //   When given, this will be called when the view is updating itself.
  //   It will be given a node (possibly of a different type), an array
  //   of active decorations around the node (which are automatically
  //   drawn, and the node view may ignore if it isn't interested in
  //   them), and a [decoration source](#view.DecorationSource) that
  //   represents any decorations that apply to the content of the node
  //   (which again may be ignored). It should return true if it was
  //   able to update to that node, and false otherwise. If the node
  //   view has a `contentDOM` property (or no `dom` property), updating
  //   its child nodes will be handled by ProseMirror.
  //
  //   selectNode:: ?()
  //   Can be used to override the way the node's selected status (as a
  //   node selection) is displayed.
  //
  //   deselectNode:: ?()
  //   When defining a `selectNode` method, you should also provide a
  //   `deselectNode` method to remove the effect again.
  //
  //   setSelection:: ?(anchor: number, head: number, root: dom.Document)
  //   This will be called to handle setting the selection inside the
  //   node. The `anchor` and `head` positions are relative to the start
  //   of the node. By default, a DOM selection will be created between
  //   the DOM positions corresponding to those positions, but if you
  //   override it you can do something else.
  //
  //   stopEvent:: ?(event: dom.Event) → bool
  //   Can be used to prevent the editor view from trying to handle some
  //   or all DOM events that bubble up from the node view. Events for
  //   which this returns true are not handled by the editor.
  //
  //   ignoreMutation:: ?(dom.MutationRecord) → bool
  //   Called when a DOM
  //   [mutation](https://developer.mozilla.org/en-US/docs/Web/API/MutationObserver)
  //   or a selection change happens within the view. When the change is
  //   a selection change, the record will have a `type` property of
  //   `"selection"` (which doesn't occur for native mutation records).
  //   Return false if the editor should re-read the selection or
  //   re-parse the range around the mutation, true if it can safely be
  //   ignored.
  //
  //   destroy:: ?()
  //   Called when the node view is removed from the editor or the whole
  //   editor is destroyed. (Not available for marks.)

  // View descriptions are data structures that describe the DOM that is
  // used to represent the editor's content. They are used for:
  //
  // - Incremental redrawing when the document changes
  //
  // - Figuring out what part of the document a given DOM position
  //   corresponds to
  //
  // - Wiring in custom implementations of the editing interface for a
  //   given node
  //
  // They form a doubly-linked mutable tree, starting at `view.docView`.

  var NOT_DIRTY = 0, CHILD_DIRTY = 1, CONTENT_DIRTY = 2, NODE_DIRTY = 3;

  // Superclass for the various kinds of descriptions. Defines their
  // basic structure and shared methods.
  var ViewDesc = function ViewDesc(parent, children, dom, contentDOM) {
    this.parent = parent;
    this.children = children;
    this.dom = dom;
    // An expando property on the DOM node provides a link back to its
    // description.
    dom.pmViewDesc = this;
    // This is the node that holds the child views. It may be null for
    // descs that don't have children.
    this.contentDOM = contentDOM;
    this.dirty = NOT_DIRTY;
  };

  var prototypeAccessors$4 = { beforePosition: { configurable: true },size: { configurable: true },border: { configurable: true },posBefore: { configurable: true },posAtStart: { configurable: true },posAfter: { configurable: true },posAtEnd: { configurable: true },contentLost: { configurable: true },domAtom: { configurable: true } };

  // Used to check whether a given description corresponds to a
  // widget/mark/node.
  ViewDesc.prototype.matchesWidget = function matchesWidget () { return false };
  ViewDesc.prototype.matchesMark = function matchesMark () { return false };
  ViewDesc.prototype.matchesNode = function matchesNode () { return false };
  ViewDesc.prototype.matchesHack = function matchesHack (_nodeName) { return false };

  prototypeAccessors$4.beforePosition.get = function () { return false };

  // : () → ?ParseRule
  // When parsing in-editor content (in domchange.js), we allow
  // descriptions to determine the parse rules that should be used to
  // parse them.
  ViewDesc.prototype.parseRule = function parseRule () { return null };

  // : (dom.Event) → bool
  // Used by the editor's event handler to ignore events that come
  // from certain descs.
  ViewDesc.prototype.stopEvent = function stopEvent () { return false };

  // The size of the content represented by this desc.
  prototypeAccessors$4.size.get = function () {
    var size = 0;
    for (var i = 0; i < this.children.length; i++) { size += this.children[i].size; }
    return size
  };

  // For block nodes, this represents the space taken up by their
  // start/end tokens.
  prototypeAccessors$4.border.get = function () { return 0 };

  ViewDesc.prototype.destroy = function destroy () {
    this.parent = null;
    if (this.dom.pmViewDesc == this) { this.dom.pmViewDesc = null; }
    for (var i = 0; i < this.children.length; i++)
      { this.children[i].destroy(); }
  };

  ViewDesc.prototype.posBeforeChild = function posBeforeChild (child) {
    for (var i = 0, pos = this.posAtStart; i < this.children.length; i++) {
      var cur = this.children[i];
      if (cur == child) { return pos }
      pos += cur.size;
    }
  };

  prototypeAccessors$4.posBefore.get = function () {
    return this.parent.posBeforeChild(this)
  };

  prototypeAccessors$4.posAtStart.get = function () {
    return this.parent ? this.parent.posBeforeChild(this) + this.border : 0
  };

  prototypeAccessors$4.posAfter.get = function () {
    return this.posBefore + this.size
  };

  prototypeAccessors$4.posAtEnd.get = function () {
    return this.posAtStart + this.size - 2 * this.border
  };

  // : (dom.Node, number, ?number) → number
  ViewDesc.prototype.localPosFromDOM = function localPosFromDOM (dom, offset, bias) {
    // If the DOM position is in the content, use the child desc after
    // it to figure out a position.
    if (this.contentDOM && this.contentDOM.contains(dom.nodeType == 1 ? dom : dom.parentNode)) {
      if (bias < 0) {
        var domBefore, desc;
        if (dom == this.contentDOM) {
          domBefore = dom.childNodes[offset - 1];
        } else {
          while (dom.parentNode != this.contentDOM) { dom = dom.parentNode; }
          domBefore = dom.previousSibling;
        }
        while (domBefore && !((desc = domBefore.pmViewDesc) && desc.parent == this)) { domBefore = domBefore.previousSibling; }
        return domBefore ? this.posBeforeChild(desc) + desc.size : this.posAtStart
      } else {
        var domAfter, desc$1;
        if (dom == this.contentDOM) {
          domAfter = dom.childNodes[offset];
        } else {
          while (dom.parentNode != this.contentDOM) { dom = dom.parentNode; }
          domAfter = dom.nextSibling;
        }
        while (domAfter && !((desc$1 = domAfter.pmViewDesc) && desc$1.parent == this)) { domAfter = domAfter.nextSibling; }
        return domAfter ? this.posBeforeChild(desc$1) : this.posAtEnd
      }
    }
    // Otherwise, use various heuristics, falling back on the bias
    // parameter, to determine whether to return the position at the
    // start or at the end of this view desc.
    var atEnd;
    if (dom == this.dom && this.contentDOM) {
      atEnd = offset > domIndex(this.contentDOM);
    } else if (this.contentDOM && this.contentDOM != this.dom && this.dom.contains(this.contentDOM)) {
      atEnd = dom.compareDocumentPosition(this.contentDOM) & 2;
    } else if (this.dom.firstChild) {
      if (offset == 0) { for (var search = dom;; search = search.parentNode) {
        if (search == this.dom) { atEnd = false; break }
        if (search.parentNode.firstChild != search) { break }
      } }
      if (atEnd == null && offset == dom.childNodes.length) { for (var search$1 = dom;; search$1 = search$1.parentNode) {
        if (search$1 == this.dom) { atEnd = true; break }
        if (search$1.parentNode.lastChild != search$1) { break }
      } }
    }
    return (atEnd == null ? bias > 0 : atEnd) ? this.posAtEnd : this.posAtStart
  };

  // Scan up the dom finding the first desc that is a descendant of
  // this one.
  ViewDesc.prototype.nearestDesc = function nearestDesc (dom, onlyNodes) {
    for (var first = true, cur = dom; cur; cur = cur.parentNode) {
      var desc = this.getDesc(cur);
      if (desc && (!onlyNodes || desc.node)) {
        // If dom is outside of this desc's nodeDOM, don't count it.
        if (first && desc.nodeDOM &&
            !(desc.nodeDOM.nodeType == 1 ? desc.nodeDOM.contains(dom.nodeType == 1 ? dom : dom.parentNode) : desc.nodeDOM == dom))
          { first = false; }
        else
          { return desc }
      }
    }
  };

  ViewDesc.prototype.getDesc = function getDesc (dom) {
    var desc = dom.pmViewDesc;
    for (var cur = desc; cur; cur = cur.parent) { if (cur == this) { return desc } }
  };

  ViewDesc.prototype.posFromDOM = function posFromDOM (dom, offset, bias) {
    for (var scan = dom; scan; scan = scan.parentNode) {
      var desc = this.getDesc(scan);
      if (desc) { return desc.localPosFromDOM(dom, offset, bias) }
    }
    return -1
  };

  // : (number) → ?NodeViewDesc
  // Find the desc for the node after the given pos, if any. (When a
  // parent node overrode rendering, there might not be one.)
  ViewDesc.prototype.descAt = function descAt (pos) {
    for (var i = 0, offset = 0; i < this.children.length; i++) {
      var child = this.children[i], end = offset + child.size;
      if (offset == pos && end != offset) {
        while (!child.border && child.children.length) { child = child.children[0]; }
        return child
      }
      if (pos < end) { return child.descAt(pos - offset - child.border) }
      offset = end;
    }
  };

  // : (number, number) → {node: dom.Node, offset: number}
  ViewDesc.prototype.domFromPos = function domFromPos (pos, side) {
    if (!this.contentDOM) { return {node: this.dom, offset: 0} }
    for (var offset = 0, i = 0, first = true;; i++, first = false) {
      // Skip removed or always-before children
      while (i < this.children.length && (this.children[i].beforePosition ||
                                          this.children[i].dom.parentNode != this.contentDOM))
        { offset += this.children[i++].size; }
      var child = i == this.children.length ? null : this.children[i];
      if (offset == pos && (side == 0 || !child || !child.size || child.border || (side < 0 && first)) ||
          child && child.domAtom && pos < offset + child.size) { return {
        node: this.contentDOM,
        offset: child ? domIndex(child.dom) : this.contentDOM.childNodes.length
      } }
      if (!child) { throw new Error("Invalid position " + pos) }
      var end = offset + child.size;
      if (!child.domAtom && (side < 0 && !child.border ? end >= pos : end > pos) &&
          (end > pos || i + 1 >= this.children.length || !this.children[i + 1].beforePosition))
        { return child.domFromPos(pos - offset - child.border, side) }
      offset = end;
    }
  };

  // Used to find a DOM range in a single parent for a given changed
  // range.
  ViewDesc.prototype.parseRange = function parseRange (from, to, base) {
      if ( base === void 0 ) { base = 0; }

    if (this.children.length == 0)
      { return {node: this.contentDOM, from: from, to: to, fromOffset: 0, toOffset: this.contentDOM.childNodes.length} }

    var fromOffset = -1, toOffset = -1;
    for (var offset = base, i = 0;; i++) {
      var child = this.children[i], end = offset + child.size;
      if (fromOffset == -1 && from <= end) {
        var childBase = offset + child.border;
        // FIXME maybe descend mark views to parse a narrower range?
        if (from >= childBase && to <= end - child.border && child.node &&
            child.contentDOM && this.contentDOM.contains(child.contentDOM))
          { return child.parseRange(from, to, childBase) }

        from = offset;
        for (var j = i; j > 0; j--) {
          var prev = this.children[j - 1];
          if (prev.size && prev.dom.parentNode == this.contentDOM && !prev.emptyChildAt(1)) {
            fromOffset = domIndex(prev.dom) + 1;
            break
          }
          from -= prev.size;
        }
        if (fromOffset == -1) { fromOffset = 0; }
      }
      if (fromOffset > -1 && (end > to || i == this.children.length - 1)) {
        to = end;
        for (var j$1 = i + 1; j$1 < this.children.length; j$1++) {
          var next = this.children[j$1];
          if (next.size && next.dom.parentNode == this.contentDOM && !next.emptyChildAt(-1)) {
            toOffset = domIndex(next.dom);
            break
          }
          to += next.size;
        }
        if (toOffset == -1) { toOffset = this.contentDOM.childNodes.length; }
        break
      }
      offset = end;
    }
    return {node: this.contentDOM, from: from, to: to, fromOffset: fromOffset, toOffset: toOffset}
  };

  ViewDesc.prototype.emptyChildAt = function emptyChildAt (side) {
    if (this.border || !this.contentDOM || !this.children.length) { return false }
    var child = this.children[side < 0 ? 0 : this.children.length - 1];
    return child.size == 0 || child.emptyChildAt(side)
  };

  // : (number) → dom.Node
  ViewDesc.prototype.domAfterPos = function domAfterPos (pos) {
    var ref = this.domFromPos(pos, 0);
      var node = ref.node;
      var offset = ref.offset;
    if (node.nodeType != 1 || offset == node.childNodes.length)
      { throw new RangeError("No node after pos " + pos) }
    return node.childNodes[offset]
  };

  // : (number, number, dom.Document)
  // View descs are responsible for setting any selection that falls
  // entirely inside of them, so that custom implementations can do
  // custom things with the selection. Note that this falls apart when
  // a selection starts in such a node and ends in another, in which
  // case we just use whatever domFromPos produces as a best effort.
  ViewDesc.prototype.setSelection = function setSelection (anchor, head, root, force) {
    // If the selection falls entirely in a child, give it to that child
    var from = Math.min(anchor, head), to = Math.max(anchor, head);
    for (var i = 0, offset = 0; i < this.children.length; i++) {
      var child = this.children[i], end = offset + child.size;
      if (from > offset && to < end)
        { return child.setSelection(anchor - offset - child.border, head - offset - child.border, root, force) }
      offset = end;
    }

    var anchorDOM = this.domFromPos(anchor, anchor ? -1 : 1);
    var headDOM = head == anchor ? anchorDOM : this.domFromPos(head, head ? -1 : 1);
    var domSel = root.getSelection();

    var brKludge = false;
    // On Firefox, using Selection.collapse to put the cursor after a
    // BR node for some reason doesn't always work (#1073). On Safari,
    // the cursor sometimes inexplicable visually lags behind its
    // reported position in such situations (#1092).
    if ((result.gecko || result.safari) && anchor == head) {
      var node = anchorDOM.node;
        var offset$1 = anchorDOM.offset;
      if (node.nodeType == 3) {
        brKludge = offset$1 && node.nodeValue[offset$1 - 1] == "\n";
        // Issue #1128
        if (brKludge && offset$1 == node.nodeValue.length) {
          for (var scan = node, after = (void 0); scan; scan = scan.parentNode) {
            if (after = scan.nextSibling) {
              if (after.nodeName == "BR")
                { anchorDOM = headDOM = {node: after.parentNode, offset: domIndex(after) + 1}; }
              break
            }
            var desc = scan.pmViewDesc;
            if (desc && desc.node && desc.node.isBlock) { break }
          }
        }
      } else {
        var prev = node.childNodes[offset$1 - 1];
        brKludge = prev && (prev.nodeName == "BR" || prev.contentEditable == "false");
      }
    }
    // Firefox can act strangely when the selection is in front of an
    // uneditable node. See #1163 and https://bugzilla.mozilla.org/show_bug.cgi?id=1709536
    if (result.gecko && domSel.focusNode && domSel.focusNode != headDOM.node && domSel.focusNode.nodeType == 1) {
      var after$1 = domSel.focusNode.childNodes[domSel.focusOffset];
      if (after$1 && after$1.contentEditable == "false") { force = true; }
    }

    if (!(force || brKludge && result.safari) &&
        isEquivalentPosition(anchorDOM.node, anchorDOM.offset, domSel.anchorNode, domSel.anchorOffset) &&
        isEquivalentPosition(headDOM.node, headDOM.offset, domSel.focusNode, domSel.focusOffset))
      { return }

    // Selection.extend can be used to create an 'inverted' selection
    // (one where the focus is before the anchor), but not all
    // browsers support it yet.
    var domSelExtended = false;
    if ((domSel.extend || anchor == head) && !brKludge) {
      domSel.collapse(anchorDOM.node, anchorDOM.offset);
      try {
        if (anchor != head) { domSel.extend(headDOM.node, headDOM.offset); }
        domSelExtended = true;
      } catch (err) {
        // In some cases with Chrome the selection is empty after calling
        // collapse, even when it should be valid. This appears to be a bug, but
        // it is difficult to isolate. If this happens fallback to the old path
        // without using extend.
        if (!(err instanceof DOMException)) { throw err }
        // declare global: DOMException
      }
    }
    if (!domSelExtended) {
      if (anchor > head) { var tmp = anchorDOM; anchorDOM = headDOM; headDOM = tmp; }
      var range = document.createRange();
      range.setEnd(headDOM.node, headDOM.offset);
      range.setStart(anchorDOM.node, anchorDOM.offset);
      domSel.removeAllRanges();
      domSel.addRange(range);
    }
  };

  // : (dom.MutationRecord) → bool
  ViewDesc.prototype.ignoreMutation = function ignoreMutation (mutation) {
    return !this.contentDOM && mutation.type != "selection"
  };

  prototypeAccessors$4.contentLost.get = function () {
    return this.contentDOM && this.contentDOM != this.dom && !this.dom.contains(this.contentDOM)
  };

  // Remove a subtree of the element tree that has been touched
  // by a DOM change, so that the next update will redraw it.
  ViewDesc.prototype.markDirty = function markDirty (from, to) {
    for (var offset = 0, i = 0; i < this.children.length; i++) {
      var child = this.children[i], end = offset + child.size;
      if (offset == end ? from <= end && to >= offset : from < end && to > offset) {
        var startInside = offset + child.border, endInside = end - child.border;
        if (from >= startInside && to <= endInside) {
          this.dirty = from == offset || to == end ? CONTENT_DIRTY : CHILD_DIRTY;
          if (from == startInside && to == endInside &&
              (child.contentLost || child.dom.parentNode != this.contentDOM)) { child.dirty = NODE_DIRTY; }
          else { child.markDirty(from - startInside, to - startInside); }
          return
        } else {
          child.dirty = child.dom == child.contentDOM && child.dom.parentNode == this.contentDOM ? CONTENT_DIRTY : NODE_DIRTY;
        }
      }
      offset = end;
    }
    this.dirty = CONTENT_DIRTY;
  };

  ViewDesc.prototype.markParentsDirty = function markParentsDirty () {
    var level = 1;
    for (var node = this.parent; node; node = node.parent, level++) {
      var dirty = level == 1 ? CONTENT_DIRTY : CHILD_DIRTY;
      if (node.dirty < dirty) { node.dirty = dirty; }
    }
  };

  prototypeAccessors$4.domAtom.get = function () { return false };

  Object.defineProperties( ViewDesc.prototype, prototypeAccessors$4 );

  // Reused array to avoid allocating fresh arrays for things that will
  // stay empty anyway.
  var nothing = [];

  // A widget desc represents a widget decoration, which is a DOM node
  // drawn between the document nodes.
  var WidgetViewDesc = /*@__PURE__*/(function (ViewDesc) {
    function WidgetViewDesc(parent, widget, view, pos) {
      var self, dom = widget.type.toDOM;
      if (typeof dom == "function") { dom = dom(view, function () {
        if (!self) { return pos }
        if (self.parent) { return self.parent.posBeforeChild(self) }
      }); }
      if (!widget.type.spec.raw) {
        if (dom.nodeType != 1) {
          var wrap = document.createElement("span");
          wrap.appendChild(dom);
          dom = wrap;
        }
        dom.contentEditable = false;
        dom.classList.add("ProseMirror-widget");
      }
      ViewDesc.call(this, parent, nothing, dom, null);
      this.widget = widget;
      self = this;
    }

    if ( ViewDesc ) { WidgetViewDesc.__proto__ = ViewDesc; }
    WidgetViewDesc.prototype = Object.create( ViewDesc && ViewDesc.prototype );
    WidgetViewDesc.prototype.constructor = WidgetViewDesc;

    var prototypeAccessors$1 = { beforePosition: { configurable: true },domAtom: { configurable: true } };

    prototypeAccessors$1.beforePosition.get = function () {
      return this.widget.type.side < 0
    };

    WidgetViewDesc.prototype.matchesWidget = function matchesWidget (widget) {
      return this.dirty == NOT_DIRTY && widget.type.eq(this.widget.type)
    };

    WidgetViewDesc.prototype.parseRule = function parseRule () { return {ignore: true} };

    WidgetViewDesc.prototype.stopEvent = function stopEvent (event) {
      var stop = this.widget.spec.stopEvent;
      return stop ? stop(event) : false
    };

    WidgetViewDesc.prototype.ignoreMutation = function ignoreMutation (mutation) {
      return mutation.type != "selection" || this.widget.spec.ignoreSelection
    };

    prototypeAccessors$1.domAtom.get = function () { return true };

    Object.defineProperties( WidgetViewDesc.prototype, prototypeAccessors$1 );

    return WidgetViewDesc;
  }(ViewDesc));

  var CompositionViewDesc = /*@__PURE__*/(function (ViewDesc) {
    function CompositionViewDesc(parent, dom, textDOM, text) {
      ViewDesc.call(this, parent, nothing, dom, null);
      this.textDOM = textDOM;
      this.text = text;
    }

    if ( ViewDesc ) { CompositionViewDesc.__proto__ = ViewDesc; }
    CompositionViewDesc.prototype = Object.create( ViewDesc && ViewDesc.prototype );
    CompositionViewDesc.prototype.constructor = CompositionViewDesc;

    var prototypeAccessors$2 = { size: { configurable: true } };

    prototypeAccessors$2.size.get = function () { return this.text.length };

    CompositionViewDesc.prototype.localPosFromDOM = function localPosFromDOM (dom, offset) {
      if (dom != this.textDOM) { return this.posAtStart + (offset ? this.size : 0) }
      return this.posAtStart + offset
    };

    CompositionViewDesc.prototype.domFromPos = function domFromPos (pos) {
      return {node: this.textDOM, offset: pos}
    };

    CompositionViewDesc.prototype.ignoreMutation = function ignoreMutation (mut) {
      return mut.type === 'characterData' && mut.target.nodeValue == mut.oldValue
     };

    Object.defineProperties( CompositionViewDesc.prototype, prototypeAccessors$2 );

    return CompositionViewDesc;
  }(ViewDesc));

  // A mark desc represents a mark. May have multiple children,
  // depending on how the mark is split. Note that marks are drawn using
  // a fixed nesting order, for simplicity and predictability, so in
  // some cases they will be split more often than would appear
  // necessary.
  var MarkViewDesc = /*@__PURE__*/(function (ViewDesc) {
    function MarkViewDesc(parent, mark, dom, contentDOM) {
      ViewDesc.call(this, parent, [], dom, contentDOM);
      this.mark = mark;
    }

    if ( ViewDesc ) { MarkViewDesc.__proto__ = ViewDesc; }
    MarkViewDesc.prototype = Object.create( ViewDesc && ViewDesc.prototype );
    MarkViewDesc.prototype.constructor = MarkViewDesc;

    MarkViewDesc.create = function create (parent, mark, inline, view) {
      var custom = view.nodeViews[mark.type.name];
      var spec = custom && custom(mark, view, inline);
      if (!spec || !spec.dom)
        { spec = DOMSerializer.renderSpec(document, mark.type.spec.toDOM(mark, inline)); }
      return new MarkViewDesc(parent, mark, spec.dom, spec.contentDOM || spec.dom)
    };

    MarkViewDesc.prototype.parseRule = function parseRule () { return {mark: this.mark.type.name, attrs: this.mark.attrs, contentElement: this.contentDOM} };

    MarkViewDesc.prototype.matchesMark = function matchesMark (mark) { return this.dirty != NODE_DIRTY && this.mark.eq(mark) };

    MarkViewDesc.prototype.markDirty = function markDirty (from, to) {
      ViewDesc.prototype.markDirty.call(this, from, to);
      // Move dirty info to nearest node view
      if (this.dirty != NOT_DIRTY) {
        var parent = this.parent;
        while (!parent.node) { parent = parent.parent; }
        if (parent.dirty < this.dirty) { parent.dirty = this.dirty; }
        this.dirty = NOT_DIRTY;
      }
    };

    MarkViewDesc.prototype.slice = function slice (from, to, view) {
      var copy = MarkViewDesc.create(this.parent, this.mark, true, view);
      var nodes = this.children, size = this.size;
      if (to < size) { nodes = replaceNodes(nodes, to, size, view); }
      if (from > 0) { nodes = replaceNodes(nodes, 0, from, view); }
      for (var i = 0; i < nodes.length; i++) { nodes[i].parent = copy; }
      copy.children = nodes;
      return copy
    };

    return MarkViewDesc;
  }(ViewDesc));

  // Node view descs are the main, most common type of view desc, and
  // correspond to an actual node in the document. Unlike mark descs,
  // they populate their child array themselves.
  var NodeViewDesc = /*@__PURE__*/(function (ViewDesc) {
    function NodeViewDesc(parent, node, outerDeco, innerDeco, dom, contentDOM, nodeDOM, view, pos) {
      ViewDesc.call(this, parent, node.isLeaf ? nothing : [], dom, contentDOM);
      this.nodeDOM = nodeDOM;
      this.node = node;
      this.outerDeco = outerDeco;
      this.innerDeco = innerDeco;
      if (contentDOM) { this.updateChildren(view, pos); }
    }

    if ( ViewDesc ) { NodeViewDesc.__proto__ = ViewDesc; }
    NodeViewDesc.prototype = Object.create( ViewDesc && ViewDesc.prototype );
    NodeViewDesc.prototype.constructor = NodeViewDesc;

    var prototypeAccessors$3 = { size: { configurable: true },border: { configurable: true },domAtom: { configurable: true } };

    // By default, a node is rendered using the `toDOM` method from the
    // node type spec. But client code can use the `nodeViews` spec to
    // supply a custom node view, which can influence various aspects of
    // the way the node works.
    //
    // (Using subclassing for this was intentionally decided against,
    // since it'd require exposing a whole slew of finicky
    // implementation details to the user code that they probably will
    // never need.)
    NodeViewDesc.create = function create (parent, node, outerDeco, innerDeco, view, pos) {
      var assign;

      var custom = view.nodeViews[node.type.name], descObj;
      var spec = custom && custom(node, view, function () {
        // (This is a function that allows the custom view to find its
        // own position)
        if (!descObj) { return pos }
        if (descObj.parent) { return descObj.parent.posBeforeChild(descObj) }
      }, outerDeco, innerDeco);

      var dom = spec && spec.dom, contentDOM = spec && spec.contentDOM;
      if (node.isText) {
        if (!dom) { dom = document.createTextNode(node.text); }
        else if (dom.nodeType != 3) { throw new RangeError("Text must be rendered as a DOM text node") }
      } else if (!dom) {
  ((assign = DOMSerializer.renderSpec(document, node.type.spec.toDOM(node)), dom = assign.dom, contentDOM = assign.contentDOM));
      }
      if (!contentDOM && !node.isText && dom.nodeName != "BR") { // Chrome gets confused by <br contenteditable=false>
        if (!dom.hasAttribute("contenteditable")) { dom.contentEditable = false; }
        if (node.type.spec.draggable) { dom.draggable = true; }
      }

      var nodeDOM = dom;
      dom = applyOuterDeco(dom, outerDeco, node);

      if (spec)
        { return descObj = new CustomNodeViewDesc(parent, node, outerDeco, innerDeco, dom, contentDOM, nodeDOM,
                                                spec, view, pos + 1) }
      else if (node.isText)
        { return new TextViewDesc(parent, node, outerDeco, innerDeco, dom, nodeDOM, view) }
      else
        { return new NodeViewDesc(parent, node, outerDeco, innerDeco, dom, contentDOM, nodeDOM, view, pos + 1) }
    };

    NodeViewDesc.prototype.parseRule = function parseRule () {
      var this$1$1 = this;

      // Experimental kludge to allow opt-in re-parsing of nodes
      if (this.node.type.spec.reparseInView) { return null }
      // FIXME the assumption that this can always return the current
      // attrs means that if the user somehow manages to change the
      // attrs in the dom, that won't be picked up. Not entirely sure
      // whether this is a problem
      var rule = {node: this.node.type.name, attrs: this.node.attrs};
      if (this.node.type.spec.code) { rule.preserveWhitespace = "full"; }
      if (this.contentDOM && !this.contentLost) { rule.contentElement = this.contentDOM; }
      else { rule.getContent = function () { return this$1$1.contentDOM ? Fragment$1.empty : this$1$1.node.content; }; }
      return rule
    };

    NodeViewDesc.prototype.matchesNode = function matchesNode (node, outerDeco, innerDeco) {
      return this.dirty == NOT_DIRTY && node.eq(this.node) &&
        sameOuterDeco(outerDeco, this.outerDeco) && innerDeco.eq(this.innerDeco)
    };

    prototypeAccessors$3.size.get = function () { return this.node.nodeSize };

    prototypeAccessors$3.border.get = function () { return this.node.isLeaf ? 0 : 1 };

    // Syncs `this.children` to match `this.node.content` and the local
    // decorations, possibly introducing nesting for marks. Then, in a
    // separate step, syncs the DOM inside `this.contentDOM` to
    // `this.children`.
    NodeViewDesc.prototype.updateChildren = function updateChildren (view, pos) {
      var this$1$1 = this;

      var inline = this.node.inlineContent, off = pos;
      var composition = view.composing && this.localCompositionInfo(view, pos);
      var localComposition = composition && composition.pos > -1 ? composition : null;
      var compositionInChild = composition && composition.pos < 0;
      var updater = new ViewTreeUpdater(this, localComposition && localComposition.node);
      iterDeco(this.node, this.innerDeco, function (widget, i, insideNode) {
        if (widget.spec.marks)
          { updater.syncToMarks(widget.spec.marks, inline, view); }
        else if (widget.type.side >= 0 && !insideNode)
          { updater.syncToMarks(i == this$1$1.node.childCount ? Mark$1.none : this$1$1.node.child(i).marks, inline, view); }
        // If the next node is a desc matching this widget, reuse it,
        // otherwise insert the widget as a new view desc.
        updater.placeWidget(widget, view, off);
      }, function (child, outerDeco, innerDeco, i) {
        // Make sure the wrapping mark descs match the node's marks.
        updater.syncToMarks(child.marks, inline, view);
        // Try several strategies for drawing this node
        var compIndex;
        if (updater.findNodeMatch(child, outerDeco, innerDeco, i)) ; else if (compositionInChild && view.state.selection.from > off &&
                   view.state.selection.to < off + child.nodeSize &&
                   (compIndex = updater.findIndexWithChild(composition.node)) > -1 &&
                   updater.updateNodeAt(child, outerDeco, innerDeco, compIndex, view)) ; else if (updater.updateNextNode(child, outerDeco, innerDeco, view, i)) ; else {
          // Add it as a new view
          updater.addNode(child, outerDeco, innerDeco, view, off);
        }
        off += child.nodeSize;
      });
      // Drop all remaining descs after the current position.
      updater.syncToMarks(nothing, inline, view);
      if (this.node.isTextblock) { updater.addTextblockHacks(); }
      updater.destroyRest();

      // Sync the DOM if anything changed
      if (updater.changed || this.dirty == CONTENT_DIRTY) {
        // May have to protect focused DOM from being changed if a composition is active
        if (localComposition) { this.protectLocalComposition(view, localComposition); }
        renderDescs(this.contentDOM, this.children, view);
        if (result.ios) { iosHacks(this.dom); }
      }
    };

    NodeViewDesc.prototype.localCompositionInfo = function localCompositionInfo (view, pos) {
      // Only do something if both the selection and a focused text node
      // are inside of this node
      var ref = view.state.selection;
      var from = ref.from;
      var to = ref.to;
      if (!(view.state.selection instanceof TextSelection) || from < pos || to > pos + this.node.content.size) { return }
      var sel = view.root.getSelection();
      var textNode = nearbyTextNode(sel.focusNode, sel.focusOffset);
      if (!textNode || !this.dom.contains(textNode.parentNode)) { return }

      if (this.node.inlineContent) {
        // Find the text in the focused node in the node, stop if it's not
        // there (may have been modified through other means, in which
        // case it should overwritten)
        var text = textNode.nodeValue;
        var textPos = findTextInFragment(this.node.content, text, from - pos, to - pos);
        return textPos < 0 ? null : {node: textNode, pos: textPos, text: text}
      } else {
        return {node: textNode, pos: -1}
      }
    };

    NodeViewDesc.prototype.protectLocalComposition = function protectLocalComposition (view, ref) {
      var node = ref.node;
      var pos = ref.pos;
      var text = ref.text;

      // The node is already part of a local view desc, leave it there
      if (this.getDesc(node)) { return }

      // Create a composition view for the orphaned nodes
      var topNode = node;
      for (;; topNode = topNode.parentNode) {
        if (topNode.parentNode == this.contentDOM) { break }
        while (topNode.previousSibling) { topNode.parentNode.removeChild(topNode.previousSibling); }
        while (topNode.nextSibling) { topNode.parentNode.removeChild(topNode.nextSibling); }
        if (topNode.pmViewDesc) { topNode.pmViewDesc = null; }
      }
      var desc = new CompositionViewDesc(this, topNode, node, text);
      view.compositionNodes.push(desc);

      // Patch up this.children to contain the composition view
      this.children = replaceNodes(this.children, pos, pos + text.length, view, desc);
    };

    // : (Node, [Decoration], DecorationSource, EditorView) → bool
    // If this desc be updated to match the given node decoration,
    // do so and return true.
    NodeViewDesc.prototype.update = function update (node, outerDeco, innerDeco, view) {
      if (this.dirty == NODE_DIRTY ||
          !node.sameMarkup(this.node)) { return false }
      this.updateInner(node, outerDeco, innerDeco, view);
      return true
    };

    NodeViewDesc.prototype.updateInner = function updateInner (node, outerDeco, innerDeco, view) {
      this.updateOuterDeco(outerDeco);
      this.node = node;
      this.innerDeco = innerDeco;
      if (this.contentDOM) { this.updateChildren(view, this.posAtStart); }
      this.dirty = NOT_DIRTY;
    };

    NodeViewDesc.prototype.updateOuterDeco = function updateOuterDeco (outerDeco) {
      if (sameOuterDeco(outerDeco, this.outerDeco)) { return }
      var needsWrap = this.nodeDOM.nodeType != 1;
      var oldDOM = this.dom;
      this.dom = patchOuterDeco(this.dom, this.nodeDOM,
                                computeOuterDeco(this.outerDeco, this.node, needsWrap),
                                computeOuterDeco(outerDeco, this.node, needsWrap));
      if (this.dom != oldDOM) {
        oldDOM.pmViewDesc = null;
        this.dom.pmViewDesc = this;
      }
      this.outerDeco = outerDeco;
    };

    // Mark this node as being the selected node.
    NodeViewDesc.prototype.selectNode = function selectNode () {
      this.nodeDOM.classList.add("ProseMirror-selectednode");
      if (this.contentDOM || !this.node.type.spec.draggable) { this.dom.draggable = true; }
    };

    // Remove selected node marking from this node.
    NodeViewDesc.prototype.deselectNode = function deselectNode () {
      this.nodeDOM.classList.remove("ProseMirror-selectednode");
      if (this.contentDOM || !this.node.type.spec.draggable) { this.dom.removeAttribute("draggable"); }
    };

    prototypeAccessors$3.domAtom.get = function () { return this.node.isAtom };

    Object.defineProperties( NodeViewDesc.prototype, prototypeAccessors$3 );

    return NodeViewDesc;
  }(ViewDesc));

  // Create a view desc for the top-level document node, to be exported
  // and used by the view class.
  function docViewDesc(doc, outerDeco, innerDeco, dom, view) {
    applyOuterDeco(dom, outerDeco, doc);
    return new NodeViewDesc(null, doc, outerDeco, innerDeco, dom, dom, dom, view, 0)
  }

  var TextViewDesc = /*@__PURE__*/(function (NodeViewDesc) {
    function TextViewDesc(parent, node, outerDeco, innerDeco, dom, nodeDOM, view) {
      NodeViewDesc.call(this, parent, node, outerDeco, innerDeco, dom, null, nodeDOM, view);
    }

    if ( NodeViewDesc ) { TextViewDesc.__proto__ = NodeViewDesc; }
    TextViewDesc.prototype = Object.create( NodeViewDesc && NodeViewDesc.prototype );
    TextViewDesc.prototype.constructor = TextViewDesc;

    var prototypeAccessors$4 = { domAtom: { configurable: true } };

    TextViewDesc.prototype.parseRule = function parseRule () {
      var skip = this.nodeDOM.parentNode;
      while (skip && skip != this.dom && !skip.pmIsDeco) { skip = skip.parentNode; }
      return {skip: skip || true}
    };

    TextViewDesc.prototype.update = function update (node, outerDeco, _, view) {
      if (this.dirty == NODE_DIRTY || (this.dirty != NOT_DIRTY && !this.inParent()) ||
          !node.sameMarkup(this.node)) { return false }
      this.updateOuterDeco(outerDeco);
      if ((this.dirty != NOT_DIRTY || node.text != this.node.text) && node.text != this.nodeDOM.nodeValue) {
        this.nodeDOM.nodeValue = node.text;
        if (view.trackWrites == this.nodeDOM) { view.trackWrites = null; }
      }
      this.node = node;
      this.dirty = NOT_DIRTY;
      return true
    };

    TextViewDesc.prototype.inParent = function inParent () {
      var parentDOM = this.parent.contentDOM;
      for (var n = this.nodeDOM; n; n = n.parentNode) { if (n == parentDOM) { return true } }
      return false
    };

    TextViewDesc.prototype.domFromPos = function domFromPos (pos) {
      return {node: this.nodeDOM, offset: pos}
    };

    TextViewDesc.prototype.localPosFromDOM = function localPosFromDOM (dom, offset, bias) {
      if (dom == this.nodeDOM) { return this.posAtStart + Math.min(offset, this.node.text.length) }
      return NodeViewDesc.prototype.localPosFromDOM.call(this, dom, offset, bias)
    };

    TextViewDesc.prototype.ignoreMutation = function ignoreMutation (mutation) {
      return mutation.type != "characterData" && mutation.type != "selection"
    };

    TextViewDesc.prototype.slice = function slice (from, to, view) {
      var node = this.node.cut(from, to), dom = document.createTextNode(node.text);
      return new TextViewDesc(this.parent, node, this.outerDeco, this.innerDeco, dom, dom, view)
    };

    TextViewDesc.prototype.markDirty = function markDirty (from, to) {
      NodeViewDesc.prototype.markDirty.call(this, from, to);
      if (this.dom != this.nodeDOM && (from == 0 || to == this.nodeDOM.nodeValue.length))
        { this.dirty = NODE_DIRTY; }
    };

    prototypeAccessors$4.domAtom.get = function () { return false };

    Object.defineProperties( TextViewDesc.prototype, prototypeAccessors$4 );

    return TextViewDesc;
  }(NodeViewDesc));

  // A dummy desc used to tag trailing BR or IMG nodes created to work
  // around contentEditable terribleness.
  var TrailingHackViewDesc = /*@__PURE__*/(function (ViewDesc) {
    function TrailingHackViewDesc () {
      ViewDesc.apply(this, arguments);
    }

    if ( ViewDesc ) { TrailingHackViewDesc.__proto__ = ViewDesc; }
    TrailingHackViewDesc.prototype = Object.create( ViewDesc && ViewDesc.prototype );
    TrailingHackViewDesc.prototype.constructor = TrailingHackViewDesc;

    var prototypeAccessors$5 = { domAtom: { configurable: true } };

    TrailingHackViewDesc.prototype.parseRule = function parseRule () { return {ignore: true} };
    TrailingHackViewDesc.prototype.matchesHack = function matchesHack (nodeName) { return this.dirty == NOT_DIRTY && this.dom.nodeName == nodeName };
    prototypeAccessors$5.domAtom.get = function () { return true };

    Object.defineProperties( TrailingHackViewDesc.prototype, prototypeAccessors$5 );

    return TrailingHackViewDesc;
  }(ViewDesc));

  // A separate subclass is used for customized node views, so that the
  // extra checks only have to be made for nodes that are actually
  // customized.
  var CustomNodeViewDesc = /*@__PURE__*/(function (NodeViewDesc) {
    function CustomNodeViewDesc(parent, node, outerDeco, innerDeco, dom, contentDOM, nodeDOM, spec, view, pos) {
      NodeViewDesc.call(this, parent, node, outerDeco, innerDeco, dom, contentDOM, nodeDOM, view, pos);
      this.spec = spec;
    }

    if ( NodeViewDesc ) { CustomNodeViewDesc.__proto__ = NodeViewDesc; }
    CustomNodeViewDesc.prototype = Object.create( NodeViewDesc && NodeViewDesc.prototype );
    CustomNodeViewDesc.prototype.constructor = CustomNodeViewDesc;

    // A custom `update` method gets to decide whether the update goes
    // through. If it does, and there's a `contentDOM` node, our logic
    // updates the children.
    CustomNodeViewDesc.prototype.update = function update (node, outerDeco, innerDeco, view) {
      if (this.dirty == NODE_DIRTY) { return false }
      if (this.spec.update) {
        var result = this.spec.update(node, outerDeco, innerDeco);
        if (result) { this.updateInner(node, outerDeco, innerDeco, view); }
        return result
      } else if (!this.contentDOM && !node.isLeaf) {
        return false
      } else {
        return NodeViewDesc.prototype.update.call(this, node, outerDeco, innerDeco, view)
      }
    };

    CustomNodeViewDesc.prototype.selectNode = function selectNode () {
      this.spec.selectNode ? this.spec.selectNode() : NodeViewDesc.prototype.selectNode.call(this);
    };

    CustomNodeViewDesc.prototype.deselectNode = function deselectNode () {
      this.spec.deselectNode ? this.spec.deselectNode() : NodeViewDesc.prototype.deselectNode.call(this);
    };

    CustomNodeViewDesc.prototype.setSelection = function setSelection (anchor, head, root, force) {
      this.spec.setSelection ? this.spec.setSelection(anchor, head, root)
        : NodeViewDesc.prototype.setSelection.call(this, anchor, head, root, force);
    };

    CustomNodeViewDesc.prototype.destroy = function destroy () {
      if (this.spec.destroy) { this.spec.destroy(); }
      NodeViewDesc.prototype.destroy.call(this);
    };

    CustomNodeViewDesc.prototype.stopEvent = function stopEvent (event) {
      return this.spec.stopEvent ? this.spec.stopEvent(event) : false
    };

    CustomNodeViewDesc.prototype.ignoreMutation = function ignoreMutation (mutation) {
      return this.spec.ignoreMutation ? this.spec.ignoreMutation(mutation) : NodeViewDesc.prototype.ignoreMutation.call(this, mutation)
    };

    return CustomNodeViewDesc;
  }(NodeViewDesc));

  // : (dom.Node, [ViewDesc])
  // Sync the content of the given DOM node with the nodes associated
  // with the given array of view descs, recursing into mark descs
  // because this should sync the subtree for a whole node at a time.
  function renderDescs(parentDOM, descs, view) {
    var dom = parentDOM.firstChild, written = false;
    for (var i = 0; i < descs.length; i++) {
      var desc = descs[i], childDOM = desc.dom;
      if (childDOM.parentNode == parentDOM) {
        while (childDOM != dom) { dom = rm(dom); written = true; }
        dom = dom.nextSibling;
      } else {
        written = true;
        parentDOM.insertBefore(childDOM, dom);
      }
      if (desc instanceof MarkViewDesc) {
        var pos = dom ? dom.previousSibling : parentDOM.lastChild;
        renderDescs(desc.contentDOM, desc.children, view);
        dom = pos ? pos.nextSibling : parentDOM.firstChild;
      }
    }
    while (dom) { dom = rm(dom); written = true; }
    if (written && view.trackWrites == parentDOM) { view.trackWrites = null; }
  }

  function OuterDecoLevel(nodeName) {
    if (nodeName) { this.nodeName = nodeName; }
  }
  OuterDecoLevel.prototype = Object.create(null);

  var noDeco = [new OuterDecoLevel];

  function computeOuterDeco(outerDeco, node, needsWrap) {
    if (outerDeco.length == 0) { return noDeco }

    var top = needsWrap ? noDeco[0] : new OuterDecoLevel, result = [top];

    for (var i = 0; i < outerDeco.length; i++) {
      var attrs = outerDeco[i].type.attrs;
      if (!attrs) { continue }
      if (attrs.nodeName)
        { result.push(top = new OuterDecoLevel(attrs.nodeName)); }

      for (var name in attrs) {
        var val = attrs[name];
        if (val == null) { continue }
        if (needsWrap && result.length == 1)
          { result.push(top = new OuterDecoLevel(node.isInline ? "span" : "div")); }
        if (name == "class") { top.class = (top.class ? top.class + " " : "") + val; }
        else if (name == "style") { top.style = (top.style ? top.style + ";" : "") + val; }
        else if (name != "nodeName") { top[name] = val; }
      }
    }

    return result
  }

  function patchOuterDeco(outerDOM, nodeDOM, prevComputed, curComputed) {
    // Shortcut for trivial case
    if (prevComputed == noDeco && curComputed == noDeco) { return nodeDOM }

    var curDOM = nodeDOM;
    for (var i = 0; i < curComputed.length; i++) {
      var deco = curComputed[i], prev = prevComputed[i];
      if (i) {
        var parent = (void 0);
        if (prev && prev.nodeName == deco.nodeName && curDOM != outerDOM &&
            (parent = curDOM.parentNode) && parent.tagName.toLowerCase() == deco.nodeName) {
          curDOM = parent;
        } else {
          parent = document.createElement(deco.nodeName);
          parent.pmIsDeco = true;
          parent.appendChild(curDOM);
          prev = noDeco[0];
          curDOM = parent;
        }
      }
      patchAttributes(curDOM, prev || noDeco[0], deco);
    }
    return curDOM
  }

  function patchAttributes(dom, prev, cur) {
    for (var name in prev)
      { if (name != "class" && name != "style" && name != "nodeName" && !(name in cur))
        { dom.removeAttribute(name); } }
    for (var name$1 in cur)
      { if (name$1 != "class" && name$1 != "style" && name$1 != "nodeName" && cur[name$1] != prev[name$1])
        { dom.setAttribute(name$1, cur[name$1]); } }
    if (prev.class != cur.class) {
      var prevList = prev.class ? prev.class.split(" ").filter(Boolean) : nothing;
      var curList = cur.class ? cur.class.split(" ").filter(Boolean) : nothing;
      for (var i = 0; i < prevList.length; i++) { if (curList.indexOf(prevList[i]) == -1)
        { dom.classList.remove(prevList[i]); } }
      for (var i$1 = 0; i$1 < curList.length; i$1++) { if (prevList.indexOf(curList[i$1]) == -1)
        { dom.classList.add(curList[i$1]); } }
    }
    if (prev.style != cur.style) {
      if (prev.style) {
        var prop = /\s*([\w\-\xa1-\uffff]+)\s*:(?:"(?:\\.|[^"])*"|'(?:\\.|[^'])*'|\(.*?\)|[^;])*/g, m;
        while (m = prop.exec(prev.style))
          { dom.style.removeProperty(m[1]); }
      }
      if (cur.style)
        { dom.style.cssText += cur.style; }
    }
  }

  function applyOuterDeco(dom, deco, node) {
    return patchOuterDeco(dom, dom, noDeco, computeOuterDeco(deco, node, dom.nodeType != 1))
  }

  // : ([Decoration], [Decoration]) → bool
  function sameOuterDeco(a, b) {
    if (a.length != b.length) { return false }
    for (var i = 0; i < a.length; i++) { if (!a[i].type.eq(b[i].type)) { return false } }
    return true
  }

  // Remove a DOM node and return its next sibling.
  function rm(dom) {
    var next = dom.nextSibling;
    dom.parentNode.removeChild(dom);
    return next
  }

  // Helper class for incrementally updating a tree of mark descs and
  // the widget and node descs inside of them.
  var ViewTreeUpdater = function ViewTreeUpdater(top, lockedNode) {
    this.top = top;
    this.lock = lockedNode;
    // Index into `this.top`'s child array, represents the current
    // update position.
    this.index = 0;
    // When entering a mark, the current top and index are pushed
    // onto this.
    this.stack = [];
    // Tracks whether anything was changed
    this.changed = false;

    this.preMatch = preMatch(top.node.content, top.children);
  };

  // Destroy and remove the children between the given indices in
  // `this.top`.
  ViewTreeUpdater.prototype.destroyBetween = function destroyBetween (start, end) {
    if (start == end) { return }
    for (var i = start; i < end; i++) { this.top.children[i].destroy(); }
    this.top.children.splice(start, end - start);
    this.changed = true;
  };

  // Destroy all remaining children in `this.top`.
  ViewTreeUpdater.prototype.destroyRest = function destroyRest () {
    this.destroyBetween(this.index, this.top.children.length);
  };

  // : ([Mark], EditorView)
  // Sync the current stack of mark descs with the given array of
  // marks, reusing existing mark descs when possible.
  ViewTreeUpdater.prototype.syncToMarks = function syncToMarks (marks, inline, view) {
    var keep = 0, depth = this.stack.length >> 1;
    var maxKeep = Math.min(depth, marks.length);
    while (keep < maxKeep &&
           (keep == depth - 1 ? this.top : this.stack[(keep + 1) << 1]).matchesMark(marks[keep]) && marks[keep].type.spec.spanning !== false)
      { keep++; }

    while (keep < depth) {
      this.destroyRest();
      this.top.dirty = NOT_DIRTY;
      this.index = this.stack.pop();
      this.top = this.stack.pop();
      depth--;
    }
    while (depth < marks.length) {
      this.stack.push(this.top, this.index + 1);
      var found = -1;
      for (var i = this.index; i < Math.min(this.index + 3, this.top.children.length); i++) {
        if (this.top.children[i].matchesMark(marks[depth])) { found = i; break }
      }
      if (found > -1) {
        if (found > this.index) {
          this.changed = true;
          this.destroyBetween(this.index, found);
        }
        this.top = this.top.children[this.index];
      } else {
        var markDesc = MarkViewDesc.create(this.top, marks[depth], inline, view);
        this.top.children.splice(this.index, 0, markDesc);
        this.top = markDesc;
        this.changed = true;
      }
      this.index = 0;
      depth++;
    }
  };

  // : (Node, [Decoration], DecorationSource) → bool
  // Try to find a node desc matching the given data. Skip over it and
  // return true when successful.
  ViewTreeUpdater.prototype.findNodeMatch = function findNodeMatch (node, outerDeco, innerDeco, index) {
    var children = this.top.children, found = -1;
    if (index >= this.preMatch.index) {
      for (var i = this.index; i < children.length; i++) { if (children[i].matchesNode(node, outerDeco, innerDeco)) {
        found = i;
        break
      } }
    } else {
      for (var i$1 = this.index, e = Math.min(children.length, i$1 + 1); i$1 < e; i$1++) {
        var child = children[i$1];
        if (child.matchesNode(node, outerDeco, innerDeco) && !this.preMatch.matched.has(child)) {
          found = i$1;
          break
        }
      }
    }
    if (found < 0) { return false }
    this.destroyBetween(this.index, found);
    this.index++;
    return true
  };

  ViewTreeUpdater.prototype.updateNodeAt = function updateNodeAt (node, outerDeco, innerDeco, index, view) {
    var child = this.top.children[index];
    if (!child.update(node, outerDeco, innerDeco, view)) { return false }
    this.destroyBetween(this.index, index);
    this.index = index + 1;
    return true
  };

  ViewTreeUpdater.prototype.findIndexWithChild = function findIndexWithChild (domNode) {
    for (;;) {
      var parent = domNode.parentNode;
      if (!parent) { return -1 }
      if (parent == this.top.contentDOM) {
        var desc = domNode.pmViewDesc;
        if (desc) { for (var i = this.index; i < this.top.children.length; i++) {
          if (this.top.children[i] == desc) { return i }
        } }
        return -1
      }
      domNode = parent;
    }
  };

  // : (Node, [Decoration], DecorationSource, EditorView, Fragment, number) → bool
  // Try to update the next node, if any, to the given data. Checks
  // pre-matches to avoid overwriting nodes that could still be used.
  ViewTreeUpdater.prototype.updateNextNode = function updateNextNode (node, outerDeco, innerDeco, view, index) {
    for (var i = this.index; i < this.top.children.length; i++) {
      var next = this.top.children[i];
      if (next instanceof NodeViewDesc) {
        var preMatch = this.preMatch.matched.get(next);
        if (preMatch != null && preMatch != index) { return false }
        var nextDOM = next.dom;

        // Can't update if nextDOM is or contains this.lock, except if
        // it's a text node whose content already matches the new text
        // and whose decorations match the new ones.
        var locked = this.lock && (nextDOM == this.lock || nextDOM.nodeType == 1 && nextDOM.contains(this.lock.parentNode)) &&
            !(node.isText && next.node && next.node.isText && next.nodeDOM.nodeValue == node.text &&
              next.dirty != NODE_DIRTY && sameOuterDeco(outerDeco, next.outerDeco));
        if (!locked && next.update(node, outerDeco, innerDeco, view)) {
          this.destroyBetween(this.index, i);
          if (next.dom != nextDOM) { this.changed = true; }
          this.index++;
          return true
        }
        break
      }
    }
    return false
  };

  // : (Node, [Decoration], DecorationSource, EditorView)
  // Insert the node as a newly created node desc.
  ViewTreeUpdater.prototype.addNode = function addNode (node, outerDeco, innerDeco, view, pos) {
    this.top.children.splice(this.index++, 0, NodeViewDesc.create(this.top, node, outerDeco, innerDeco, view, pos));
    this.changed = true;
  };

  ViewTreeUpdater.prototype.placeWidget = function placeWidget (widget, view, pos) {
    var next = this.index < this.top.children.length ? this.top.children[this.index] : null;
    if (next && next.matchesWidget(widget) && (widget == next.widget || !next.widget.type.toDOM.parentNode)) {
      this.index++;
    } else {
      var desc = new WidgetViewDesc(this.top, widget, view, pos);
      this.top.children.splice(this.index++, 0, desc);
      this.changed = true;
    }
  };

  // Make sure a textblock looks and behaves correctly in
  // contentEditable.
  ViewTreeUpdater.prototype.addTextblockHacks = function addTextblockHacks () {
    var lastChild = this.top.children[this.index - 1];
    while (lastChild instanceof MarkViewDesc) { lastChild = lastChild.children[lastChild.children.length - 1]; }

    if (!lastChild || // Empty textblock
        !(lastChild instanceof TextViewDesc) ||
        /\n$/.test(lastChild.node.text)) {
      // Avoid bugs in Safari's cursor drawing (#1165) and Chrome's mouse selection (#1152)
      if ((result.safari || result.chrome) && lastChild && lastChild.dom.contentEditable == "false")
        { this.addHackNode("IMG"); }
      this.addHackNode("BR");
    }
  };

  ViewTreeUpdater.prototype.addHackNode = function addHackNode (nodeName) {
    if (this.index < this.top.children.length && this.top.children[this.index].matchesHack(nodeName)) {
      this.index++;
    } else {
      var dom = document.createElement(nodeName);
      if (nodeName == "IMG") { dom.className = "ProseMirror-separator"; }
      this.top.children.splice(this.index++, 0, new TrailingHackViewDesc(this.top, nothing, dom, null));
      this.changed = true;
    }
  };

  // : (Fragment, [ViewDesc]) → {index: number, matched: Map<ViewDesc, number>}
  // Iterate from the end of the fragment and array of descs to find
  // directly matching ones, in order to avoid overeagerly reusing those
  // for other nodes. Returns the fragment index of the first node that
  // is part of the sequence of matched nodes at the end of the
  // fragment.
  function preMatch(frag, descs) {
    var fI = frag.childCount, dI = descs.length, matched = new Map;
    for (; fI > 0 && dI > 0; dI--) {
      var desc = descs[dI - 1], node = desc.node;
      if (!node) { continue }
      if (node != frag.child(fI - 1)) { break }
      --fI;
      matched.set(desc, fI);
    }
    return {index: fI, matched: matched}
  }

  function compareSide(a, b) { return a.type.side - b.type.side }

  // : (ViewDesc, DecorationSource, (Decoration, number), (Node, [Decoration], DecorationSource, number))
  // This function abstracts iterating over the nodes and decorations in
  // a fragment. Calls `onNode` for each node, with its local and child
  // decorations. Splits text nodes when there is a decoration starting
  // or ending inside of them. Calls `onWidget` for each widget.
  function iterDeco(parent, deco, onWidget, onNode) {
    var locals = deco.locals(parent), offset = 0;
    // Simple, cheap variant for when there are no local decorations
    if (locals.length == 0) {
      for (var i = 0; i < parent.childCount; i++) {
        var child = parent.child(i);
        onNode(child, locals, deco.forChild(offset, child), i);
        offset += child.nodeSize;
      }
      return
    }

    var decoIndex = 0, active = [], restNode = null;
    for (var parentIndex = 0;;) {
      if (decoIndex < locals.length && locals[decoIndex].to == offset) {
        var widget = locals[decoIndex++], widgets = (void 0);
        while (decoIndex < locals.length && locals[decoIndex].to == offset)
          { (widgets || (widgets = [widget])).push(locals[decoIndex++]); }
        if (widgets) {
          widgets.sort(compareSide);
          for (var i$1 = 0; i$1 < widgets.length; i$1++) { onWidget(widgets[i$1], parentIndex, !!restNode); }
        } else {
          onWidget(widget, parentIndex, !!restNode);
        }
      }

      var child$1 = (void 0), index = (void 0);
      if (restNode) {
        index = -1;
        child$1 = restNode;
        restNode = null;
      } else if (parentIndex < parent.childCount) {
        index = parentIndex;
        child$1 = parent.child(parentIndex++);
      } else {
        break
      }

      for (var i$2 = 0; i$2 < active.length; i$2++) { if (active[i$2].to <= offset) { active.splice(i$2--, 1); } }
      while (decoIndex < locals.length && locals[decoIndex].from <= offset && locals[decoIndex].to > offset)
        { active.push(locals[decoIndex++]); }

      var end = offset + child$1.nodeSize;
      if (child$1.isText) {
        var cutAt = end;
        if (decoIndex < locals.length && locals[decoIndex].from < cutAt) { cutAt = locals[decoIndex].from; }
        for (var i$3 = 0; i$3 < active.length; i$3++) { if (active[i$3].to < cutAt) { cutAt = active[i$3].to; } }
        if (cutAt < end) {
          restNode = child$1.cut(cutAt - offset);
          child$1 = child$1.cut(0, cutAt - offset);
          end = cutAt;
          index = -1;
        }
      }

      var outerDeco = !active.length ? nothing
          : child$1.isInline && !child$1.isLeaf ? active.filter(function (d) { return !d.inline; })
          : active.slice();
      onNode(child$1, outerDeco, deco.forChild(offset, child$1), index);
      offset = end;
    }
  }

  // List markers in Mobile Safari will mysteriously disappear
  // sometimes. This works around that.
  function iosHacks(dom) {
    if (dom.nodeName == "UL" || dom.nodeName == "OL") {
      var oldCSS = dom.style.cssText;
      dom.style.cssText = oldCSS + "; list-style: square !important";
      window.getComputedStyle(dom).listStyle;
      dom.style.cssText = oldCSS;
    }
  }

  function nearbyTextNode(node, offset) {
    for (;;) {
      if (node.nodeType == 3) { return node }
      if (node.nodeType == 1 && offset > 0) {
        if (node.childNodes.length > offset && node.childNodes[offset].nodeType == 3)
          { return node.childNodes[offset] }
        node = node.childNodes[offset - 1];
        offset = nodeSize(node);
      } else if (node.nodeType == 1 && offset < node.childNodes.length) {
        node = node.childNodes[offset];
        offset = 0;
      } else {
        return null
      }
    }
  }

  // Find a piece of text in an inline fragment, overlapping from-to
  function findTextInFragment(frag, text, from, to) {
    for (var i = 0, pos = 0; i < frag.childCount && pos <= to;) {
      var child = frag.child(i++), childStart = pos;
      pos += child.nodeSize;
      if (!child.isText) { continue }
      var str = child.text;
      while (i < frag.childCount) {
        var next = frag.child(i++);
        pos += next.nodeSize;
        if (!next.isText) { break }
        str += next.text;
      }
      if (pos >= from) {
        var found = str.lastIndexOf(text, to - childStart);
        if (found >= 0 && found + text.length + childStart >= from)
          { return childStart + found }
      }
    }
    return -1
  }

  // Replace range from-to in an array of view descs with replacement
  // (may be null to just delete). This goes very much against the grain
  // of the rest of this code, which tends to create nodes with the
  // right shape in one go, rather than messing with them after
  // creation, but is necessary in the composition hack.
  function replaceNodes(nodes, from, to, view, replacement) {
    var result = [];
    for (var i = 0, off = 0; i < nodes.length; i++) {
      var child = nodes[i], start = off, end = off += child.size;
      if (start >= to || end <= from) {
        result.push(child);
      } else {
        if (start < from) { result.push(child.slice(0, from - start, view)); }
        if (replacement) {
          result.push(replacement);
          replacement = null;
        }
        if (end > to) { result.push(child.slice(to - start, child.size, view)); }
      }
    }
    return result
  }

  function selectionFromDOM(view, origin) {
    var domSel = view.root.getSelection(), doc = view.state.doc;
    if (!domSel.focusNode) { return null }
    var nearestDesc = view.docView.nearestDesc(domSel.focusNode), inWidget = nearestDesc && nearestDesc.size == 0;
    var head = view.docView.posFromDOM(domSel.focusNode, domSel.focusOffset);
    if (head < 0) { return null }
    var $head = doc.resolve(head), $anchor, selection;
    if (selectionCollapsed(domSel)) {
      $anchor = $head;
      while (nearestDesc && !nearestDesc.node) { nearestDesc = nearestDesc.parent; }
      if (nearestDesc && nearestDesc.node.isAtom && NodeSelection.isSelectable(nearestDesc.node) && nearestDesc.parent
          && !(nearestDesc.node.isInline && isOnEdge(domSel.focusNode, domSel.focusOffset, nearestDesc.dom))) {
        var pos = nearestDesc.posBefore;
        selection = new NodeSelection(head == pos ? $head : doc.resolve(pos));
      }
    } else {
      var anchor = view.docView.posFromDOM(domSel.anchorNode, domSel.anchorOffset);
      if (anchor < 0) { return null }
      $anchor = doc.resolve(anchor);
    }

    if (!selection) {
      var bias = origin == "pointer" || (view.state.selection.head < $head.pos && !inWidget) ? 1 : -1;
      selection = selectionBetween(view, $anchor, $head, bias);
    }
    return selection
  }

  function editorOwnsSelection(view) {
    return view.editable ? view.hasFocus() :
      hasSelection(view) && document.activeElement && document.activeElement.contains(view.dom)
  }

  function selectionToDOM(view, force) {
    var sel = view.state.selection;
    syncNodeSelection(view, sel);

    if (!editorOwnsSelection(view)) { return }

    if (!force && view.mouseDown && view.mouseDown.allowDefault) {
      view.mouseDown.delayedSelectionSync = true;
      view.domObserver.setCurSelection();
      return
    }

    view.domObserver.disconnectSelection();

    if (view.cursorWrapper) {
      selectCursorWrapper(view);
    } else {
      var anchor = sel.anchor;
      var head = sel.head;
      var resetEditableFrom, resetEditableTo;
      if (brokenSelectBetweenUneditable && !(sel instanceof TextSelection)) {
        if (!sel.$from.parent.inlineContent)
          { resetEditableFrom = temporarilyEditableNear(view, sel.from); }
        if (!sel.empty && !sel.$from.parent.inlineContent)
          { resetEditableTo = temporarilyEditableNear(view, sel.to); }
      }
      view.docView.setSelection(anchor, head, view.root, force);
      if (brokenSelectBetweenUneditable) {
        if (resetEditableFrom) { resetEditable(resetEditableFrom); }
        if (resetEditableTo) { resetEditable(resetEditableTo); }
      }
      if (sel.visible) {
        view.dom.classList.remove("ProseMirror-hideselection");
      } else {
        view.dom.classList.add("ProseMirror-hideselection");
        if ("onselectionchange" in document) { removeClassOnSelectionChange(view); }
      }
    }

    view.domObserver.setCurSelection();
    view.domObserver.connectSelection();
  }

  // Kludge to work around Webkit not allowing a selection to start/end
  // between non-editable block nodes. We briefly make something
  // editable, set the selection, then set it uneditable again.

  var brokenSelectBetweenUneditable = result.safari || result.chrome && result.chrome_version < 63;

  function temporarilyEditableNear(view, pos) {
    var ref = view.docView.domFromPos(pos, 0);
    var node = ref.node;
    var offset = ref.offset;
    var after = offset < node.childNodes.length ? node.childNodes[offset] : null;
    var before = offset ? node.childNodes[offset - 1] : null;
    if (result.safari && after && after.contentEditable == "false") { return setEditable(after) }
    if ((!after || after.contentEditable == "false") && (!before || before.contentEditable == "false")) {
      if (after) { return setEditable(after) }
      else if (before) { return setEditable(before) }
    }
  }

  function setEditable(element) {
    element.contentEditable = "true";
    if (result.safari && element.draggable) { element.draggable = false; element.wasDraggable = true; }
    return element
  }

  function resetEditable(element) {
    element.contentEditable = "false";
    if (element.wasDraggable) { element.draggable = true; element.wasDraggable = null; }
  }

  function removeClassOnSelectionChange(view) {
    var doc = view.dom.ownerDocument;
    doc.removeEventListener("selectionchange", view.hideSelectionGuard);
    var domSel = view.root.getSelection();
    var node = domSel.anchorNode, offset = domSel.anchorOffset;
    doc.addEventListener("selectionchange", view.hideSelectionGuard = function () {
      if (domSel.anchorNode != node || domSel.anchorOffset != offset) {
        doc.removeEventListener("selectionchange", view.hideSelectionGuard);
        setTimeout(function () {
          if (!editorOwnsSelection(view) || view.state.selection.visible)
            { view.dom.classList.remove("ProseMirror-hideselection"); }
        }, 20);
      }
    });
  }

  function selectCursorWrapper(view) {
    var domSel = view.root.getSelection(), range = document.createRange();
    var node = view.cursorWrapper.dom, img = node.nodeName == "IMG";
    if (img) { range.setEnd(node.parentNode, domIndex(node) + 1); }
    else { range.setEnd(node, 0); }
    range.collapse(false);
    domSel.removeAllRanges();
    domSel.addRange(range);
    // Kludge to kill 'control selection' in IE11 when selecting an
    // invisible cursor wrapper, since that would result in those weird
    // resize handles and a selection that considers the absolutely
    // positioned wrapper, rather than the root editable node, the
    // focused element.
    if (!img && !view.state.selection.visible && result.ie && result.ie_version <= 11) {
      node.disabled = true;
      node.disabled = false;
    }
  }

  function syncNodeSelection(view, sel) {
    if (sel instanceof NodeSelection) {
      var desc = view.docView.descAt(sel.from);
      if (desc != view.lastSelectedViewDesc) {
        clearNodeSelection(view);
        if (desc) { desc.selectNode(); }
        view.lastSelectedViewDesc = desc;
      }
    } else {
      clearNodeSelection(view);
    }
  }

  // Clear all DOM statefulness of the last node selection.
  function clearNodeSelection(view) {
    if (view.lastSelectedViewDesc) {
      if (view.lastSelectedViewDesc.parent)
        { view.lastSelectedViewDesc.deselectNode(); }
      view.lastSelectedViewDesc = null;
    }
  }

  function selectionBetween(view, $anchor, $head, bias) {
    return view.someProp("createSelectionBetween", function (f) { return f(view, $anchor, $head); })
      || TextSelection.between($anchor, $head, bias)
  }

  function hasFocusAndSelection(view) {
    if (view.editable && view.root.activeElement != view.dom) { return false }
    return hasSelection(view)
  }

  function hasSelection(view) {
    var sel = view.root.getSelection();
    if (!sel.anchorNode) { return false }
    try {
      // Firefox will raise 'permission denied' errors when accessing
      // properties of `sel.anchorNode` when it's in a generated CSS
      // element.
      return view.dom.contains(sel.anchorNode.nodeType == 3 ? sel.anchorNode.parentNode : sel.anchorNode) &&
        (view.editable || view.dom.contains(sel.focusNode.nodeType == 3 ? sel.focusNode.parentNode : sel.focusNode))
    } catch(_) {
      return false
    }
  }

  function anchorInRightPlace(view) {
    var anchorDOM = view.docView.domFromPos(view.state.selection.anchor, 0);
    var domSel = view.root.getSelection();
    return isEquivalentPosition(anchorDOM.node, anchorDOM.offset, domSel.anchorNode, domSel.anchorOffset)
  }

  function moveSelectionBlock(state, dir) {
    var ref = state.selection;
    var $anchor = ref.$anchor;
    var $head = ref.$head;
    var $side = dir > 0 ? $anchor.max($head) : $anchor.min($head);
    var $start = !$side.parent.inlineContent ? $side : $side.depth ? state.doc.resolve(dir > 0 ? $side.after() : $side.before()) : null;
    return $start && Selection.findFrom($start, dir)
  }

  function apply(view, sel) {
    view.dispatch(view.state.tr.setSelection(sel).scrollIntoView());
    return true
  }

  function selectHorizontally(view, dir, mods) {
    var sel = view.state.selection;
    if (sel instanceof TextSelection) {
      if (!sel.empty || mods.indexOf("s") > -1) {
        return false
      } else if (view.endOfTextblock(dir > 0 ? "right" : "left")) {
        var next = moveSelectionBlock(view.state, dir);
        if (next && (next instanceof NodeSelection)) { return apply(view, next) }
        return false
      } else if (!(result.mac && mods.indexOf("m") > -1)) {
        var $head = sel.$head, node = $head.textOffset ? null : dir < 0 ? $head.nodeBefore : $head.nodeAfter, desc;
        if (!node || node.isText) { return false }
        var nodePos = dir < 0 ? $head.pos - node.nodeSize : $head.pos;
        if (!(node.isAtom || (desc = view.docView.descAt(nodePos)) && !desc.contentDOM)) { return false }
        if (NodeSelection.isSelectable(node)) {
          return apply(view, new NodeSelection(dir < 0 ? view.state.doc.resolve($head.pos - node.nodeSize) : $head))
        } else if (result.webkit) {
          // Chrome and Safari will introduce extra pointless cursor
          // positions around inline uneditable nodes, so we have to
          // take over and move the cursor past them (#937)
          return apply(view, new TextSelection(view.state.doc.resolve(dir < 0 ? nodePos : nodePos + node.nodeSize)))
        } else {
          return false
        }
      }
    } else if (sel instanceof NodeSelection && sel.node.isInline) {
      return apply(view, new TextSelection(dir > 0 ? sel.$to : sel.$from))
    } else {
      var next$1 = moveSelectionBlock(view.state, dir);
      if (next$1) { return apply(view, next$1) }
      return false
    }
  }

  function nodeLen(node) {
    return node.nodeType == 3 ? node.nodeValue.length : node.childNodes.length
  }

  function isIgnorable(dom) {
    var desc = dom.pmViewDesc;
    return desc && desc.size == 0 && (dom.nextSibling || dom.nodeName != "BR")
  }

  // Make sure the cursor isn't directly after one or more ignored
  // nodes, which will confuse the browser's cursor motion logic.
  function skipIgnoredNodesLeft(view) {
    var sel = view.root.getSelection();
    var node = sel.focusNode, offset = sel.focusOffset;
    if (!node) { return }
    var moveNode, moveOffset, force = false;
    // Gecko will do odd things when the selection is directly in front
    // of a non-editable node, so in that case, move it into the next
    // node if possible. Issue prosemirror/prosemirror#832.
    if (result.gecko && node.nodeType == 1 && offset < nodeLen(node) && isIgnorable(node.childNodes[offset])) { force = true; }
    for (;;) {
      if (offset > 0) {
        if (node.nodeType != 1) {
          break
        } else {
          var before = node.childNodes[offset - 1];
          if (isIgnorable(before)) {
            moveNode = node;
            moveOffset = --offset;
          } else if (before.nodeType == 3) {
            node = before;
            offset = node.nodeValue.length;
          } else { break }
        }
      } else if (isBlockNode(node)) {
        break
      } else {
        var prev = node.previousSibling;
        while (prev && isIgnorable(prev)) {
          moveNode = node.parentNode;
          moveOffset = domIndex(prev);
          prev = prev.previousSibling;
        }
        if (!prev) {
          node = node.parentNode;
          if (node == view.dom) { break }
          offset = 0;
        } else {
          node = prev;
          offset = nodeLen(node);
        }
      }
    }
    if (force) { setSelFocus(view, sel, node, offset); }
    else if (moveNode) { setSelFocus(view, sel, moveNode, moveOffset); }
  }

  // Make sure the cursor isn't directly before one or more ignored
  // nodes.
  function skipIgnoredNodesRight(view) {
    var sel = view.root.getSelection();
    var node = sel.focusNode, offset = sel.focusOffset;
    if (!node) { return }
    var len = nodeLen(node);
    var moveNode, moveOffset;
    for (;;) {
      if (offset < len) {
        if (node.nodeType != 1) { break }
        var after = node.childNodes[offset];
        if (isIgnorable(after)) {
          moveNode = node;
          moveOffset = ++offset;
        }
        else { break }
      } else if (isBlockNode(node)) {
        break
      } else {
        var next = node.nextSibling;
        while (next && isIgnorable(next)) {
          moveNode = next.parentNode;
          moveOffset = domIndex(next) + 1;
          next = next.nextSibling;
        }
        if (!next) {
          node = node.parentNode;
          if (node == view.dom) { break }
          offset = len = 0;
        } else {
          node = next;
          offset = 0;
          len = nodeLen(node);
        }
      }
    }
    if (moveNode) { setSelFocus(view, sel, moveNode, moveOffset); }
  }

  function isBlockNode(dom) {
    var desc = dom.pmViewDesc;
    return desc && desc.node && desc.node.isBlock
  }

  function setSelFocus(view, sel, node, offset) {
    if (selectionCollapsed(sel)) {
      var range = document.createRange();
      range.setEnd(node, offset);
      range.setStart(node, offset);
      sel.removeAllRanges();
      sel.addRange(range);
    } else if (sel.extend) {
      sel.extend(node, offset);
    }
    view.domObserver.setCurSelection();
    var state = view.state;
    // If no state update ends up happening, reset the selection.
    setTimeout(function () {
      if (view.state == state) { selectionToDOM(view); }
    }, 50);
  }

  // : (EditorState, number)
  // Check whether vertical selection motion would involve node
  // selections. If so, apply it (if not, the result is left to the
  // browser)
  function selectVertically(view, dir, mods) {
    var sel = view.state.selection;
    if (sel instanceof TextSelection && !sel.empty || mods.indexOf("s") > -1) { return false }
    if (result.mac && mods.indexOf("m") > -1) { return false }
    var $from = sel.$from;
    var $to = sel.$to;

    if (!$from.parent.inlineContent || view.endOfTextblock(dir < 0 ? "up" : "down")) {
      var next = moveSelectionBlock(view.state, dir);
      if (next && (next instanceof NodeSelection))
        { return apply(view, next) }
    }
    if (!$from.parent.inlineContent) {
      var side = dir < 0 ? $from : $to;
      var beyond = sel instanceof AllSelection ? Selection.near(side, dir) : Selection.findFrom(side, dir);
      return beyond ? apply(view, beyond) : false
    }
    return false
  }

  function stopNativeHorizontalDelete(view, dir) {
    if (!(view.state.selection instanceof TextSelection)) { return true }
    var ref = view.state.selection;
    var $head = ref.$head;
    var $anchor = ref.$anchor;
    var empty = ref.empty;
    if (!$head.sameParent($anchor)) { return true }
    if (!empty) { return false }
    if (view.endOfTextblock(dir > 0 ? "forward" : "backward")) { return true }
    var nextNode = !$head.textOffset && (dir < 0 ? $head.nodeBefore : $head.nodeAfter);
    if (nextNode && !nextNode.isText) {
      var tr = view.state.tr;
      if (dir < 0) { tr.delete($head.pos - nextNode.nodeSize, $head.pos); }
      else { tr.delete($head.pos, $head.pos + nextNode.nodeSize); }
      view.dispatch(tr);
      return true
    }
    return false
  }

  function switchEditable(view, node, state) {
    view.domObserver.stop();
    node.contentEditable = state;
    view.domObserver.start();
  }

  // Issue #867 / #1090 / https://bugs.chromium.org/p/chromium/issues/detail?id=903821
  // In which Safari (and at some point in the past, Chrome) does really
  // wrong things when the down arrow is pressed when the cursor is
  // directly at the start of a textblock and has an uneditable node
  // after it
  function safariDownArrowBug(view) {
    if (!result.safari || view.state.selection.$head.parentOffset > 0) { return }
    var ref = view.root.getSelection();
    var focusNode = ref.focusNode;
    var focusOffset = ref.focusOffset;
    if (focusNode && focusNode.nodeType == 1 && focusOffset == 0 &&
        focusNode.firstChild && focusNode.firstChild.contentEditable == "false") {
      var child = focusNode.firstChild;
      switchEditable(view, child, true);
      setTimeout(function () { return switchEditable(view, child, false); }, 20);
    }
  }

  // A backdrop key mapping used to make sure we always suppress keys
  // that have a dangerous default effect, even if the commands they are
  // bound to return false, and to make sure that cursor-motion keys
  // find a cursor (as opposed to a node selection) when pressed. For
  // cursor-motion keys, the code in the handlers also takes care of
  // block selections.

  function getMods(event) {
    var result = "";
    if (event.ctrlKey) { result += "c"; }
    if (event.metaKey) { result += "m"; }
    if (event.altKey) { result += "a"; }
    if (event.shiftKey) { result += "s"; }
    return result
  }

  function captureKeyDown(view, event) {
    var code = event.keyCode, mods = getMods(event);
    if (code == 8 || (result.mac && code == 72 && mods == "c")) { // Backspace, Ctrl-h on Mac
      return stopNativeHorizontalDelete(view, -1) || skipIgnoredNodesLeft(view)
    } else if (code == 46 || (result.mac && code == 68 && mods == "c")) { // Delete, Ctrl-d on Mac
      return stopNativeHorizontalDelete(view, 1) || skipIgnoredNodesRight(view)
    } else if (code == 13 || code == 27) { // Enter, Esc
      return true
    } else if (code == 37) { // Left arrow
      return selectHorizontally(view, -1, mods) || skipIgnoredNodesLeft(view)
    } else if (code == 39) { // Right arrow
      return selectHorizontally(view, 1, mods) || skipIgnoredNodesRight(view)
    } else if (code == 38) { // Up arrow
      return selectVertically(view, -1, mods) || skipIgnoredNodesLeft(view)
    } else if (code == 40) { // Down arrow
      return safariDownArrowBug(view) || selectVertically(view, 1, mods) || skipIgnoredNodesRight(view)
    } else if (mods == (result.mac ? "m" : "c") &&
               (code == 66 || code == 73 || code == 89 || code == 90)) { // Mod-[biyz]
      return true
    }
    return false
  }

  // Note that all referencing and parsing is done with the
  // start-of-operation selection and document, since that's the one
  // that the DOM represents. If any changes came in in the meantime,
  // the modification is mapped over those before it is applied, in
  // readDOMChange.

  function parseBetween(view, from_, to_) {
    var ref = view.docView.parseRange(from_, to_);
    var parent = ref.node;
    var fromOffset = ref.fromOffset;
    var toOffset = ref.toOffset;
    var from = ref.from;
    var to = ref.to;

    var domSel = view.root.getSelection(), find = null, anchor = domSel.anchorNode;
    if (anchor && view.dom.contains(anchor.nodeType == 1 ? anchor : anchor.parentNode)) {
      find = [{node: anchor, offset: domSel.anchorOffset}];
      if (!selectionCollapsed(domSel))
        { find.push({node: domSel.focusNode, offset: domSel.focusOffset}); }
    }
    // Work around issue in Chrome where backspacing sometimes replaces
    // the deleted content with a random BR node (issues #799, #831)
    if (result.chrome && view.lastKeyCode === 8) {
      for (var off = toOffset; off > fromOffset; off--) {
        var node = parent.childNodes[off - 1], desc = node.pmViewDesc;
        if (node.nodeName == "BR" && !desc) { toOffset = off; break }
        if (!desc || desc.size) { break }
      }
    }
    var startDoc = view.state.doc;
    var parser = view.someProp("domParser") || DOMParser.fromSchema(view.state.schema);
    var $from = startDoc.resolve(from);

    var sel = null, doc = parser.parse(parent, {
      topNode: $from.parent,
      topMatch: $from.parent.contentMatchAt($from.index()),
      topOpen: true,
      from: fromOffset,
      to: toOffset,
      preserveWhitespace: $from.parent.type.spec.code ? "full" : true,
      editableContent: true,
      findPositions: find,
      ruleFromNode: ruleFromNode,
      context: $from
    });
    if (find && find[0].pos != null) {
      var anchor$1 = find[0].pos, head = find[1] && find[1].pos;
      if (head == null) { head = anchor$1; }
      sel = {anchor: anchor$1 + from, head: head + from};
    }
    return {doc: doc, sel: sel, from: from, to: to}
  }

  function ruleFromNode(dom) {
    var desc = dom.pmViewDesc;
    if (desc) {
      return desc.parseRule()
    } else if (dom.nodeName == "BR" && dom.parentNode) {
      // Safari replaces the list item or table cell with a BR
      // directly in the list node (?!) if you delete the last
      // character in a list item or table cell (#708, #862)
      if (result.safari && /^(ul|ol)$/i.test(dom.parentNode.nodeName)) {
        var skip = document.createElement("div");
        skip.appendChild(document.createElement("li"));
        return {skip: skip}
      } else if (dom.parentNode.lastChild == dom || result.safari && /^(tr|table)$/i.test(dom.parentNode.nodeName)) {
        return {ignore: true}
      }
    } else if (dom.nodeName == "IMG" && dom.getAttribute("mark-placeholder")) {
      return {ignore: true}
    }
  }

  function readDOMChange(view, from, to, typeOver, addedNodes) {
    if (from < 0) {
      var origin = view.lastSelectionTime > Date.now() - 50 ? view.lastSelectionOrigin : null;
      var newSel = selectionFromDOM(view, origin);
      if (newSel && !view.state.selection.eq(newSel)) {
        var tr$1 = view.state.tr.setSelection(newSel);
        if (origin == "pointer") { tr$1.setMeta("pointer", true); }
        else if (origin == "key") { tr$1.scrollIntoView(); }
        view.dispatch(tr$1);
      }
      return
    }

    var $before = view.state.doc.resolve(from);
    var shared = $before.sharedDepth(to);
    from = $before.before(shared + 1);
    to = view.state.doc.resolve(to).after(shared + 1);

    var sel = view.state.selection;
    var parse = parseBetween(view, from, to);
    // Chrome sometimes leaves the cursor before the inserted text when
    // composing after a cursor wrapper. This moves it forward.
    if (result.chrome && view.cursorWrapper && parse.sel && parse.sel.anchor == view.cursorWrapper.deco.from) {
      var text = view.cursorWrapper.deco.type.toDOM.nextSibling;
      var size = text && text.nodeValue ? text.nodeValue.length : 1;
      parse.sel = {anchor: parse.sel.anchor + size, head: parse.sel.anchor + size};
    }

    var doc = view.state.doc, compare = doc.slice(parse.from, parse.to);
    var preferredPos, preferredSide;
    // Prefer anchoring to end when Backspace is pressed
    if (view.lastKeyCode === 8 && Date.now() - 100 < view.lastKeyCodeTime) {
      preferredPos = view.state.selection.to;
      preferredSide = "end";
    } else {
      preferredPos = view.state.selection.from;
      preferredSide = "start";
    }
    view.lastKeyCode = null;

    var change = findDiff(compare.content, parse.doc.content, parse.from, preferredPos, preferredSide);
    if (!change) {
      if (typeOver && sel instanceof TextSelection && !sel.empty && sel.$head.sameParent(sel.$anchor) &&
          !view.composing && !(parse.sel && parse.sel.anchor != parse.sel.head)) {
        change = {start: sel.from, endA: sel.to, endB: sel.to};
      } else if ((result.ios && view.lastIOSEnter > Date.now() - 225 || result.android) &&
                 addedNodes.some(function (n) { return n.nodeName == "DIV" || n.nodeName == "P"; }) &&
                 view.someProp("handleKeyDown", function (f) { return f(view, keyEvent(13, "Enter")); })) {
        view.lastIOSEnter = 0;
        return
      } else {
        if (parse.sel) {
          var sel$1 = resolveSelection(view, view.state.doc, parse.sel);
          if (sel$1 && !sel$1.eq(view.state.selection)) { view.dispatch(view.state.tr.setSelection(sel$1)); }
        }
        return
      }
    }
    view.domChangeCount++;
    // Handle the case where overwriting a selection by typing matches
    // the start or end of the selected content, creating a change
    // that's smaller than what was actually overwritten.
    if (view.state.selection.from < view.state.selection.to &&
        change.start == change.endB &&
        view.state.selection instanceof TextSelection) {
      if (change.start > view.state.selection.from && change.start <= view.state.selection.from + 2) {
        change.start = view.state.selection.from;
      } else if (change.endA < view.state.selection.to && change.endA >= view.state.selection.to - 2) {
        change.endB += (view.state.selection.to - change.endA);
        change.endA = view.state.selection.to;
      }
    }

    // IE11 will insert a non-breaking space _ahead_ of the space after
    // the cursor space when adding a space before another space. When
    // that happened, adjust the change to cover the space instead.
    if (result.ie && result.ie_version <= 11 && change.endB == change.start + 1 &&
        change.endA == change.start && change.start > parse.from &&
        parse.doc.textBetween(change.start - parse.from - 1, change.start - parse.from + 1) == " \u00a0") {
      change.start--;
      change.endA--;
      change.endB--;
    }

    var $from = parse.doc.resolveNoCache(change.start - parse.from);
    var $to = parse.doc.resolveNoCache(change.endB - parse.from);
    var inlineChange = $from.sameParent($to) && $from.parent.inlineContent;
    var nextSel;
    // If this looks like the effect of pressing Enter (or was recorded
    // as being an iOS enter press), just dispatch an Enter key instead.
    if (((result.ios && view.lastIOSEnter > Date.now() - 225 &&
          (!inlineChange || addedNodes.some(function (n) { return n.nodeName == "DIV" || n.nodeName == "P"; }))) ||
         (!inlineChange && $from.pos < parse.doc.content.size &&
          (nextSel = Selection.findFrom(parse.doc.resolve($from.pos + 1), 1, true)) &&
          nextSel.head == $to.pos)) &&
        view.someProp("handleKeyDown", function (f) { return f(view, keyEvent(13, "Enter")); })) {
      view.lastIOSEnter = 0;
      return
    }
    // Same for backspace
    if (view.state.selection.anchor > change.start &&
        looksLikeJoin(doc, change.start, change.endA, $from, $to) &&
        view.someProp("handleKeyDown", function (f) { return f(view, keyEvent(8, "Backspace")); })) {
      if (result.android && result.chrome) { view.domObserver.suppressSelectionUpdates(); } // #820
      return
    }

    // Chrome Android will occasionally, during composition, delete the
    // entire composition and then immediately insert it again. This is
    // used to detect that situation.
    if (result.chrome && result.android && change.toB == change.from)
      { view.lastAndroidDelete = Date.now(); }

    // This tries to detect Android virtual keyboard
    // enter-and-pick-suggestion action. That sometimes (see issue
    // #1059) first fires a DOM mutation, before moving the selection to
    // the newly created block. And then, because ProseMirror cleans up
    // the DOM selection, it gives up moving the selection entirely,
    // leaving the cursor in the wrong place. When that happens, we drop
    // the new paragraph from the initial change, and fire a simulated
    // enter key afterwards.
    if (result.android && !inlineChange && $from.start() != $to.start() && $to.parentOffset == 0 && $from.depth == $to.depth &&
        parse.sel && parse.sel.anchor == parse.sel.head && parse.sel.head == change.endA) {
      change.endB -= 2;
      $to = parse.doc.resolveNoCache(change.endB - parse.from);
      setTimeout(function () {
        view.someProp("handleKeyDown", function (f) { return f(view, keyEvent(13, "Enter")); });
      }, 20);
    }

    var chFrom = change.start, chTo = change.endA;

    var tr, storedMarks, markChange, $from1;
    if (inlineChange) {
      if ($from.pos == $to.pos) { // Deletion
        // IE11 sometimes weirdly moves the DOM selection around after
        // backspacing out the first element in a textblock
        if (result.ie && result.ie_version <= 11 && $from.parentOffset == 0) {
          view.domObserver.suppressSelectionUpdates();
          setTimeout(function () { return selectionToDOM(view); }, 20);
        }
        tr = view.state.tr.delete(chFrom, chTo);
        storedMarks = doc.resolve(change.start).marksAcross(doc.resolve(change.endA));
      } else if ( // Adding or removing a mark
        change.endA == change.endB && ($from1 = doc.resolve(change.start)) &&
        (markChange = isMarkChange($from.parent.content.cut($from.parentOffset, $to.parentOffset),
                                   $from1.parent.content.cut($from1.parentOffset, change.endA - $from1.start())))
      ) {
        tr = view.state.tr;
        if (markChange.type == "add") { tr.addMark(chFrom, chTo, markChange.mark); }
        else { tr.removeMark(chFrom, chTo, markChange.mark); }
      } else if ($from.parent.child($from.index()).isText && $from.index() == $to.index() - ($to.textOffset ? 0 : 1)) {
        // Both positions in the same text node -- simply insert text
        var text$1 = $from.parent.textBetween($from.parentOffset, $to.parentOffset);
        if (view.someProp("handleTextInput", function (f) { return f(view, chFrom, chTo, text$1); })) { return }
        tr = view.state.tr.insertText(text$1, chFrom, chTo);
      }
    }

    if (!tr)
      { tr = view.state.tr.replace(chFrom, chTo, parse.doc.slice(change.start - parse.from, change.endB - parse.from)); }
    if (parse.sel) {
      var sel$2 = resolveSelection(view, tr.doc, parse.sel);
      // Chrome Android will sometimes, during composition, report the
      // selection in the wrong place. If it looks like that is
      // happening, don't update the selection.
      // Edge just doesn't move the cursor forward when you start typing
      // in an empty block or between br nodes.
      if (sel$2 && !(result.chrome && result.android && view.composing && sel$2.empty &&
                   (change.start != change.endB || view.lastAndroidDelete < Date.now() - 100) &&
                   (sel$2.head == chFrom || sel$2.head == tr.mapping.map(chTo) - 1) ||
                   result.ie && sel$2.empty && sel$2.head == chFrom))
        { tr.setSelection(sel$2); }
    }
    if (storedMarks) { tr.ensureMarks(storedMarks); }
    view.dispatch(tr.scrollIntoView());
  }

  function resolveSelection(view, doc, parsedSel) {
    if (Math.max(parsedSel.anchor, parsedSel.head) > doc.content.size) { return null }
    return selectionBetween(view, doc.resolve(parsedSel.anchor), doc.resolve(parsedSel.head))
  }

  // : (Fragment, Fragment) → ?{mark: Mark, type: string}
  // Given two same-length, non-empty fragments of inline content,
  // determine whether the first could be created from the second by
  // removing or adding a single mark type.
  function isMarkChange(cur, prev) {
    var curMarks = cur.firstChild.marks, prevMarks = prev.firstChild.marks;
    var added = curMarks, removed = prevMarks, type, mark, update;
    for (var i = 0; i < prevMarks.length; i++) { added = prevMarks[i].removeFromSet(added); }
    for (var i$1 = 0; i$1 < curMarks.length; i$1++) { removed = curMarks[i$1].removeFromSet(removed); }
    if (added.length == 1 && removed.length == 0) {
      mark = added[0];
      type = "add";
      update = function (node) { return node.mark(mark.addToSet(node.marks)); };
    } else if (added.length == 0 && removed.length == 1) {
      mark = removed[0];
      type = "remove";
      update = function (node) { return node.mark(mark.removeFromSet(node.marks)); };
    } else {
      return null
    }
    var updated = [];
    for (var i$2 = 0; i$2 < prev.childCount; i$2++) { updated.push(update(prev.child(i$2))); }
    if (Fragment$1.from(updated).eq(cur)) { return {mark: mark, type: type} }
  }

  function looksLikeJoin(old, start, end, $newStart, $newEnd) {
    if (!$newStart.parent.isTextblock ||
        // The content must have shrunk
        end - start <= $newEnd.pos - $newStart.pos ||
        // newEnd must point directly at or after the end of the block that newStart points into
        skipClosingAndOpening($newStart, true, false) < $newEnd.pos)
      { return false }

    var $start = old.resolve(start);
    // Start must be at the end of a block
    if ($start.parentOffset < $start.parent.content.size || !$start.parent.isTextblock)
      { return false }
    var $next = old.resolve(skipClosingAndOpening($start, true, true));
    // The next textblock must start before end and end near it
    if (!$next.parent.isTextblock || $next.pos > end ||
        skipClosingAndOpening($next, true, false) < end)
      { return false }

    // The fragments after the join point must match
    return $newStart.parent.content.cut($newStart.parentOffset).eq($next.parent.content)
  }

  function skipClosingAndOpening($pos, fromEnd, mayOpen) {
    var depth = $pos.depth, end = fromEnd ? $pos.end() : $pos.pos;
    while (depth > 0 && (fromEnd || $pos.indexAfter(depth) == $pos.node(depth).childCount)) {
      depth--;
      end++;
      fromEnd = false;
    }
    if (mayOpen) {
      var next = $pos.node(depth).maybeChild($pos.indexAfter(depth));
      while (next && !next.isLeaf) {
        next = next.firstChild;
        end++;
      }
    }
    return end
  }

  function findDiff(a, b, pos, preferredPos, preferredSide) {
    var start = a.findDiffStart(b, pos);
    if (start == null) { return null }
    var ref = a.findDiffEnd(b, pos + a.size, pos + b.size);
    var endA = ref.a;
    var endB = ref.b;
    if (preferredSide == "end") {
      var adjust = Math.max(0, start - Math.min(endA, endB));
      preferredPos -= endA + adjust - start;
    }
    if (endA < start && a.size < b.size) {
      var move = preferredPos <= start && preferredPos >= endA ? start - preferredPos : 0;
      start -= move;
      endB = start + (endB - endA);
      endA = start;
    } else if (endB < start) {
      var move$1 = preferredPos <= start && preferredPos >= endB ? start - preferredPos : 0;
      start -= move$1;
      endA = start + (endA - endB);
      endB = start;
    }
    return {start: start, endA: endA, endB: endB}
  }

  function serializeForClipboard(view, slice) {
    var context = [];
    var content = slice.content;
    var openStart = slice.openStart;
    var openEnd = slice.openEnd;
    while (openStart > 1 && openEnd > 1 && content.childCount == 1 && content.firstChild.childCount == 1) {
      openStart--;
      openEnd--;
      var node = content.firstChild;
      context.push(node.type.name, node.attrs != node.type.defaultAttrs ? node.attrs : null);
      content = node.content;
    }

    var serializer = view.someProp("clipboardSerializer") || DOMSerializer.fromSchema(view.state.schema);
    var doc = detachedDoc(), wrap = doc.createElement("div");
    wrap.appendChild(serializer.serializeFragment(content, {document: doc}));

    var firstChild = wrap.firstChild, needsWrap;
    while (firstChild && firstChild.nodeType == 1 && (needsWrap = wrapMap[firstChild.nodeName.toLowerCase()])) {
      for (var i = needsWrap.length - 1; i >= 0; i--) {
        var wrapper = doc.createElement(needsWrap[i]);
        while (wrap.firstChild) { wrapper.appendChild(wrap.firstChild); }
        wrap.appendChild(wrapper);
        if (needsWrap[i] != "tbody") {
          openStart++;
          openEnd++;
        }
      }
      firstChild = wrap.firstChild;
    }

    if (firstChild && firstChild.nodeType == 1)
      { firstChild.setAttribute("data-pm-slice", (openStart + " " + openEnd + " " + (JSON.stringify(context)))); }

    var text = view.someProp("clipboardTextSerializer", function (f) { return f(slice); }) ||
        slice.content.textBetween(0, slice.content.size, "\n\n");

    return {dom: wrap, text: text}
  }

  // : (EditorView, string, string, ?bool, ResolvedPos) → ?Slice
  // Read a slice of content from the clipboard (or drop data).
  function parseFromClipboard(view, text, html, plainText, $context) {
    var dom, inCode = $context.parent.type.spec.code, slice;
    if (!html && !text) { return null }
    var asText = text && (plainText || inCode || !html);
    if (asText) {
      view.someProp("transformPastedText", function (f) { text = f(text, inCode || plainText); });
      if (inCode) { return new Slice$1(Fragment$1.from(view.state.schema.text(text.replace(/\r\n?/g, "\n"))), 0, 0) }
      var parsed = view.someProp("clipboardTextParser", function (f) { return f(text, $context, plainText); });
      if (parsed) {
        slice = parsed;
      } else {
        var marks = $context.marks();
        var ref = view.state;
        var schema = ref.schema;
        var serializer = DOMSerializer.fromSchema(schema);
        dom = document.createElement("div");
        text.trim().split(/(?:\r\n?|\n)+/).forEach(function (block) {
          dom.appendChild(document.createElement("p")).appendChild(serializer.serializeNode(schema.text(block, marks)));
        });
      }
    } else {
      view.someProp("transformPastedHTML", function (f) { html = f(html); });
      dom = readHTML(html);
      if (result.webkit) { restoreReplacedSpaces(dom); }
    }

    var contextNode = dom && dom.querySelector("[data-pm-slice]");
    var sliceData = contextNode && /^(\d+) (\d+) (.*)/.exec(contextNode.getAttribute("data-pm-slice"));
    if (!slice) {
      var parser = view.someProp("clipboardParser") || view.someProp("domParser") || DOMParser.fromSchema(view.state.schema);
      slice = parser.parseSlice(dom, {preserveWhitespace: !!(asText || sliceData), context: $context});
    }
    if (sliceData)
      { slice = addContext(closeSlice(slice, +sliceData[1], +sliceData[2]), sliceData[3]); }
    else // HTML wasn't created by ProseMirror. Make sure top-level siblings are coherent
      { slice = Slice$1.maxOpen(normalizeSiblings(slice.content, $context), false); }

    view.someProp("transformPasted", function (f) { slice = f(slice); });
    return slice
  }

  // Takes a slice parsed with parseSlice, which means there hasn't been
  // any content-expression checking done on the top nodes, tries to
  // find a parent node in the current context that might fit the nodes,
  // and if successful, rebuilds the slice so that it fits into that parent.
  //
  // This addresses the problem that Transform.replace expects a
  // coherent slice, and will fail to place a set of siblings that don't
  // fit anywhere in the schema.
  function normalizeSiblings(fragment, $context) {
    if (fragment.childCount < 2) { return fragment }
    var loop = function ( d ) {
      var parent = $context.node(d);
      var match = parent.contentMatchAt($context.index(d));
      var lastWrap = (void 0), result = [];
      fragment.forEach(function (node) {
        if (!result) { return }
        var wrap = match.findWrapping(node.type), inLast;
        if (!wrap) { return result = null }
        if (inLast = result.length && lastWrap.length && addToSibling(wrap, lastWrap, node, result[result.length - 1], 0)) {
          result[result.length - 1] = inLast;
        } else {
          if (result.length) { result[result.length - 1] = closeRight(result[result.length - 1], lastWrap.length); }
          var wrapped = withWrappers(node, wrap);
          result.push(wrapped);
          match = match.matchType(wrapped.type, wrapped.attrs);
          lastWrap = wrap;
        }
      });
      if (result) { return { v: Fragment$1.from(result) } }
    };

    for (var d = $context.depth; d >= 0; d--) {
      var returned = loop( d );

      if ( returned ) { return returned.v; }
    }
    return fragment
  }

  function withWrappers(node, wrap, from) {
    if ( from === void 0 ) { from = 0; }

    for (var i = wrap.length - 1; i >= from; i--)
      { node = wrap[i].create(null, Fragment$1.from(node)); }
    return node
  }

  // Used to group adjacent nodes wrapped in similar parents by
  // normalizeSiblings into the same parent node
  function addToSibling(wrap, lastWrap, node, sibling, depth) {
    if (depth < wrap.length && depth < lastWrap.length && wrap[depth] == lastWrap[depth]) {
      var inner = addToSibling(wrap, lastWrap, node, sibling.lastChild, depth + 1);
      if (inner) { return sibling.copy(sibling.content.replaceChild(sibling.childCount - 1, inner)) }
      var match = sibling.contentMatchAt(sibling.childCount);
      if (match.matchType(depth == wrap.length - 1 ? node.type : wrap[depth + 1]))
        { return sibling.copy(sibling.content.append(Fragment$1.from(withWrappers(node, wrap, depth + 1)))) }
    }
  }

  function closeRight(node, depth) {
    if (depth == 0) { return node }
    var fragment = node.content.replaceChild(node.childCount - 1, closeRight(node.lastChild, depth - 1));
    var fill = node.contentMatchAt(node.childCount).fillBefore(Fragment$1.empty, true);
    return node.copy(fragment.append(fill))
  }

  function closeRange(fragment, side, from, to, depth, openEnd) {
    var node = side < 0 ? fragment.firstChild : fragment.lastChild, inner = node.content;
    if (depth < to - 1) { inner = closeRange(inner, side, from, to, depth + 1, openEnd); }
    if (depth >= from)
      { inner = side < 0 ? node.contentMatchAt(0).fillBefore(inner, fragment.childCount > 1 || openEnd <= depth).append(inner)
        : inner.append(node.contentMatchAt(node.childCount).fillBefore(Fragment$1.empty, true)); }
    return fragment.replaceChild(side < 0 ? 0 : fragment.childCount - 1, node.copy(inner))
  }

  function closeSlice(slice, openStart, openEnd) {
    if (openStart < slice.openStart)
      { slice = new Slice$1(closeRange(slice.content, -1, openStart, slice.openStart, 0, slice.openEnd), openStart, slice.openEnd); }
    if (openEnd < slice.openEnd)
      { slice = new Slice$1(closeRange(slice.content, 1, openEnd, slice.openEnd, 0, 0), slice.openStart, openEnd); }
    return slice
  }

  // Trick from jQuery -- some elements must be wrapped in other
  // elements for innerHTML to work. I.e. if you do `div.innerHTML =
  // "<td>..</td>"` the table cells are ignored.
  var wrapMap = {
    thead: ["table"],
    tbody: ["table"],
    tfoot: ["table"],
    caption: ["table"],
    colgroup: ["table"],
    col: ["table", "colgroup"],
    tr: ["table", "tbody"],
    td: ["table", "tbody", "tr"],
    th: ["table", "tbody", "tr"]
  };

  var _detachedDoc = null;
  function detachedDoc() {
    return _detachedDoc || (_detachedDoc = document.implementation.createHTMLDocument("title"))
  }

  function readHTML(html) {
    var metas = /^(\s*<meta [^>]*>)*/.exec(html);
    if (metas) { html = html.slice(metas[0].length); }
    var elt = detachedDoc().createElement("div");
    var firstTag = /<([a-z][^>\s]+)/i.exec(html), wrap;
    if (wrap = firstTag && wrapMap[firstTag[1].toLowerCase()])
      { html = wrap.map(function (n) { return "<" + n + ">"; }).join("") + html + wrap.map(function (n) { return "</" + n + ">"; }).reverse().join(""); }
    elt.innerHTML = html;
    if (wrap) { for (var i = 0; i < wrap.length; i++) { elt = elt.querySelector(wrap[i]) || elt; } }
    return elt
  }

  // Webkit browsers do some hard-to-predict replacement of regular
  // spaces with non-breaking spaces when putting content on the
  // clipboard. This tries to convert such non-breaking spaces (which
  // will be wrapped in a plain span on Chrome, a span with class
  // Apple-converted-space on Safari) back to regular spaces.
  function restoreReplacedSpaces(dom) {
    var nodes = dom.querySelectorAll(result.chrome ? "span:not([class]):not([style])" : "span.Apple-converted-space");
    for (var i = 0; i < nodes.length; i++) {
      var node = nodes[i];
      if (node.childNodes.length == 1 && node.textContent == "\u00a0" && node.parentNode)
        { node.parentNode.replaceChild(dom.ownerDocument.createTextNode(" "), node); }
    }
  }

  function addContext(slice, context) {
    if (!slice.size) { return slice }
    var schema = slice.content.firstChild.type.schema, array;
    try { array = JSON.parse(context); }
    catch(e) { return slice }
    var content = slice.content;
    var openStart = slice.openStart;
    var openEnd = slice.openEnd;
    for (var i = array.length - 2; i >= 0; i -= 2) {
      var type = schema.nodes[array[i]];
      if (!type || type.hasRequiredAttrs()) { break }
      content = Fragment$1.from(type.create(array[i + 1], content));
      openStart++; openEnd++;
    }
    return new Slice$1(content, openStart, openEnd)
  }

  var observeOptions = {
    childList: true,
    characterData: true,
    characterDataOldValue: true,
    attributes: true,
    attributeOldValue: true,
    subtree: true
  };
  // IE11 has very broken mutation observers, so we also listen to DOMCharacterDataModified
  var useCharData = result.ie && result.ie_version <= 11;

  var SelectionState = function SelectionState() {
    this.anchorNode = this.anchorOffset = this.focusNode = this.focusOffset = null;
  };

  SelectionState.prototype.set = function set (sel) {
    this.anchorNode = sel.anchorNode; this.anchorOffset = sel.anchorOffset;
    this.focusNode = sel.focusNode; this.focusOffset = sel.focusOffset;
  };

  SelectionState.prototype.eq = function eq (sel) {
    return sel.anchorNode == this.anchorNode && sel.anchorOffset == this.anchorOffset &&
      sel.focusNode == this.focusNode && sel.focusOffset == this.focusOffset
  };

  var DOMObserver = function DOMObserver(view, handleDOMChange) {
    var this$1$1 = this;

    this.view = view;
    this.handleDOMChange = handleDOMChange;
    this.queue = [];
    this.flushingSoon = -1;
    this.observer = window.MutationObserver &&
      new window.MutationObserver(function (mutations) {
        for (var i = 0; i < mutations.length; i++) { this$1$1.queue.push(mutations[i]); }
        // IE11 will sometimes (on backspacing out a single character
        // text node after a BR node) call the observer callback
        // before actually updating the DOM, which will cause
        // ProseMirror to miss the change (see #930)
        if (result.ie && result.ie_version <= 11 && mutations.some(
          function (m) { return m.type == "childList" && m.removedNodes.length ||
               m.type == "characterData" && m.oldValue.length > m.target.nodeValue.length; }))
          { this$1$1.flushSoon(); }
        else
          { this$1$1.flush(); }
      });
    this.currentSelection = new SelectionState;
    if (useCharData) {
      this.onCharData = function (e) {
        this$1$1.queue.push({target: e.target, type: "characterData", oldValue: e.prevValue});
        this$1$1.flushSoon();
      };
    }
    this.onSelectionChange = this.onSelectionChange.bind(this);
    this.suppressingSelectionUpdates = false;
  };

  DOMObserver.prototype.flushSoon = function flushSoon () {
      var this$1$1 = this;

    if (this.flushingSoon < 0)
      { this.flushingSoon = window.setTimeout(function () { this$1$1.flushingSoon = -1; this$1$1.flush(); }, 20); }
  };

  DOMObserver.prototype.forceFlush = function forceFlush () {
    if (this.flushingSoon > -1) {
      window.clearTimeout(this.flushingSoon);
      this.flushingSoon = -1;
      this.flush();
    }
  };

  DOMObserver.prototype.start = function start () {
    if (this.observer)
      { this.observer.observe(this.view.dom, observeOptions); }
    if (useCharData)
      { this.view.dom.addEventListener("DOMCharacterDataModified", this.onCharData); }
    this.connectSelection();
  };

  DOMObserver.prototype.stop = function stop () {
      var this$1$1 = this;

    if (this.observer) {
      var take = this.observer.takeRecords();
      if (take.length) {
        for (var i = 0; i < take.length; i++) { this.queue.push(take[i]); }
        window.setTimeout(function () { return this$1$1.flush(); }, 20);
      }
      this.observer.disconnect();
    }
    if (useCharData) { this.view.dom.removeEventListener("DOMCharacterDataModified", this.onCharData); }
    this.disconnectSelection();
  };

  DOMObserver.prototype.connectSelection = function connectSelection () {
    this.view.dom.ownerDocument.addEventListener("selectionchange", this.onSelectionChange);
  };

  DOMObserver.prototype.disconnectSelection = function disconnectSelection () {
    this.view.dom.ownerDocument.removeEventListener("selectionchange", this.onSelectionChange);
  };

  DOMObserver.prototype.suppressSelectionUpdates = function suppressSelectionUpdates () {
      var this$1$1 = this;

    this.suppressingSelectionUpdates = true;
    setTimeout(function () { return this$1$1.suppressingSelectionUpdates = false; }, 50);
  };

  DOMObserver.prototype.onSelectionChange = function onSelectionChange () {
    if (!hasFocusAndSelection(this.view)) { return }
    if (this.suppressingSelectionUpdates) { return selectionToDOM(this.view) }
    // Deletions on IE11 fire their events in the wrong order, giving
    // us a selection change event before the DOM changes are
    // reported.
    if (result.ie && result.ie_version <= 11 && !this.view.state.selection.empty) {
      var sel = this.view.root.getSelection();
      // Selection.isCollapsed isn't reliable on IE
      if (sel.focusNode && isEquivalentPosition(sel.focusNode, sel.focusOffset, sel.anchorNode, sel.anchorOffset))
        { return this.flushSoon() }
    }
    this.flush();
  };

  DOMObserver.prototype.setCurSelection = function setCurSelection () {
    this.currentSelection.set(this.view.root.getSelection());
  };

  DOMObserver.prototype.ignoreSelectionChange = function ignoreSelectionChange (sel) {
    if (sel.rangeCount == 0) { return true }
    var container = sel.getRangeAt(0).commonAncestorContainer;
    var desc = this.view.docView.nearestDesc(container);
    if (desc && desc.ignoreMutation({type: "selection", target: container.nodeType == 3 ? container.parentNode : container})) {
      this.setCurSelection();
      return true
    }
  };

  DOMObserver.prototype.flush = function flush () {
    if (!this.view.docView || this.flushingSoon > -1) { return }
    var mutations = this.observer ? this.observer.takeRecords() : [];
    if (this.queue.length) {
      mutations = this.queue.concat(mutations);
      this.queue.length = 0;
    }

    var sel = this.view.root.getSelection();
    var newSel = !this.suppressingSelectionUpdates && !this.currentSelection.eq(sel) && hasSelection(this.view) && !this.ignoreSelectionChange(sel);

    var from = -1, to = -1, typeOver = false, added = [];
    if (this.view.editable) {
      for (var i = 0; i < mutations.length; i++) {
        var result$1 = this.registerMutation(mutations[i], added);
        if (result$1) {
          from = from < 0 ? result$1.from : Math.min(result$1.from, from);
          to = to < 0 ? result$1.to : Math.max(result$1.to, to);
          if (result$1.typeOver) { typeOver = true; }
        }
      }
    }

    if (result.gecko && added.length > 1) {
      var brs = added.filter(function (n) { return n.nodeName == "BR"; });
      if (brs.length == 2) {
        var a = brs[0];
          var b = brs[1];
        if (a.parentNode && a.parentNode.parentNode == b.parentNode) { b.remove(); }
        else { a.remove(); }
      }
    }

    if (from > -1 || newSel) {
      if (from > -1) {
        this.view.docView.markDirty(from, to);
        checkCSS(this.view);
      }
      this.handleDOMChange(from, to, typeOver, added);
      if (this.view.docView.dirty) { this.view.updateState(this.view.state); }
      else if (!this.currentSelection.eq(sel)) { selectionToDOM(this.view); }
      this.currentSelection.set(sel);
    }
  };

  DOMObserver.prototype.registerMutation = function registerMutation (mut, added) {
    // Ignore mutations inside nodes that were already noted as inserted
    if (added.indexOf(mut.target) > -1) { return null }
    var desc = this.view.docView.nearestDesc(mut.target);
    if (mut.type == "attributes" &&
        (desc == this.view.docView || mut.attributeName == "contenteditable" ||
         // Firefox sometimes fires spurious events for null/empty styles
         (mut.attributeName == "style" && !mut.oldValue && !mut.target.getAttribute("style"))))
      { return null }
    if (!desc || desc.ignoreMutation(mut)) { return null }

    if (mut.type == "childList") {
      for (var i = 0; i < mut.addedNodes.length; i++) { added.push(mut.addedNodes[i]); }
      if (desc.contentDOM && desc.contentDOM != desc.dom && !desc.contentDOM.contains(mut.target))
        { return {from: desc.posBefore, to: desc.posAfter} }
      var prev = mut.previousSibling, next = mut.nextSibling;
      if (result.ie && result.ie_version <= 11 && mut.addedNodes.length) {
        // IE11 gives us incorrect next/prev siblings for some
        // insertions, so if there are added nodes, recompute those
        for (var i$1 = 0; i$1 < mut.addedNodes.length; i$1++) {
          var ref = mut.addedNodes[i$1];
            var previousSibling = ref.previousSibling;
            var nextSibling = ref.nextSibling;
          if (!previousSibling || Array.prototype.indexOf.call(mut.addedNodes, previousSibling) < 0) { prev = previousSibling; }
          if (!nextSibling || Array.prototype.indexOf.call(mut.addedNodes, nextSibling) < 0) { next = nextSibling; }
        }
      }
      var fromOffset = prev && prev.parentNode == mut.target
          ? domIndex(prev) + 1 : 0;
      var from = desc.localPosFromDOM(mut.target, fromOffset, -1);
      var toOffset = next && next.parentNode == mut.target
          ? domIndex(next) : mut.target.childNodes.length;
      var to = desc.localPosFromDOM(mut.target, toOffset, 1);
      return {from: from, to: to}
    } else if (mut.type == "attributes") {
      return {from: desc.posAtStart - desc.border, to: desc.posAtEnd + desc.border}
    } else { // "characterData"
      return {
        from: desc.posAtStart,
        to: desc.posAtEnd,
        // An event was generated for a text change that didn't change
        // any text. Mark the dom change to fall back to assuming the
        // selection was typed over with an identical value if it can't
        // find another change.
        typeOver: mut.target.nodeValue == mut.oldValue
      }
    }
  };

  var cssChecked = false;

  function checkCSS(view) {
    if (cssChecked) { return }
    cssChecked = true;
    if (getComputedStyle(view.dom).whiteSpace == "normal")
      { console["warn"]("ProseMirror expects the CSS white-space property to be set, preferably to 'pre-wrap'. It is recommended to load style/prosemirror.css from the prosemirror-view package."); }
  }

  // A collection of DOM events that occur within the editor, and callback functions
  // to invoke when the event fires.
  var handlers = {}, editHandlers = {};

  function initInput(view) {
    view.shiftKey = false;
    view.mouseDown = null;
    view.lastKeyCode = null;
    view.lastKeyCodeTime = 0;
    view.lastClick = {time: 0, x: 0, y: 0, type: ""};
    view.lastSelectionOrigin = null;
    view.lastSelectionTime = 0;

    view.lastIOSEnter = 0;
    view.lastIOSEnterFallbackTimeout = null;
    view.lastAndroidDelete = 0;

    view.composing = false;
    view.composingTimeout = null;
    view.compositionNodes = [];
    view.compositionEndedAt = -2e8;

    view.domObserver = new DOMObserver(view, function (from, to, typeOver, added) { return readDOMChange(view, from, to, typeOver, added); });
    view.domObserver.start();
    // Used by hacks like the beforeinput handler to check whether anything happened in the DOM
    view.domChangeCount = 0;

    view.eventHandlers = Object.create(null);
    var loop = function ( event ) {
      var handler = handlers[event];
      view.dom.addEventListener(event, view.eventHandlers[event] = function (event) {
        if (eventBelongsToView(view, event) && !runCustomHandler(view, event) &&
            (view.editable || !(event.type in editHandlers)))
          { handler(view, event); }
      });
    };

    for (var event in handlers) { loop( event ); }
    // On Safari, for reasons beyond my understanding, adding an input
    // event handler makes an issue where the composition vanishes when
    // you press enter go away.
    if (result.safari) { view.dom.addEventListener("input", function () { return null; }); }

    ensureListeners(view);
  }

  function setSelectionOrigin(view, origin) {
    view.lastSelectionOrigin = origin;
    view.lastSelectionTime = Date.now();
  }

  function destroyInput(view) {
    view.domObserver.stop();
    for (var type in view.eventHandlers)
      { view.dom.removeEventListener(type, view.eventHandlers[type]); }
    clearTimeout(view.composingTimeout);
    clearTimeout(view.lastIOSEnterFallbackTimeout);
  }

  function ensureListeners(view) {
    view.someProp("handleDOMEvents", function (currentHandlers) {
      for (var type in currentHandlers) { if (!view.eventHandlers[type])
        { view.dom.addEventListener(type, view.eventHandlers[type] = function (event) { return runCustomHandler(view, event); }); } }
    });
  }

  function runCustomHandler(view, event) {
    return view.someProp("handleDOMEvents", function (handlers) {
      var handler = handlers[event.type];
      return handler ? handler(view, event) || event.defaultPrevented : false
    })
  }

  function eventBelongsToView(view, event) {
    if (!event.bubbles) { return true }
    if (event.defaultPrevented) { return false }
    for (var node = event.target; node != view.dom; node = node.parentNode)
      { if (!node || node.nodeType == 11 ||
          (node.pmViewDesc && node.pmViewDesc.stopEvent(event)))
        { return false } }
    return true
  }

  function dispatchEvent(view, event) {
    if (!runCustomHandler(view, event) && handlers[event.type] &&
        (view.editable || !(event.type in editHandlers)))
      { handlers[event.type](view, event); }
  }

  editHandlers.keydown = function (view, event) {
    view.shiftKey = event.keyCode == 16 || event.shiftKey;
    if (inOrNearComposition(view, event)) { return }
    if (event.keyCode != 229) { view.domObserver.forceFlush(); }
    view.lastKeyCode = event.keyCode;
    view.lastKeyCodeTime = Date.now();
    // On iOS, if we preventDefault enter key presses, the virtual
    // keyboard gets confused. So the hack here is to set a flag that
    // makes the DOM change code recognize that what just happens should
    // be replaced by whatever the Enter key handlers do.
    if (result.ios && event.keyCode == 13 && !event.ctrlKey && !event.altKey && !event.metaKey) {
      var now = Date.now();
      view.lastIOSEnter = now;
      view.lastIOSEnterFallbackTimeout = setTimeout(function () {
        if (view.lastIOSEnter == now) {
          view.someProp("handleKeyDown", function (f) { return f(view, keyEvent(13, "Enter")); });
          view.lastIOSEnter = 0;
        }
      }, 200);
    } else if (view.someProp("handleKeyDown", function (f) { return f(view, event); }) || captureKeyDown(view, event)) {
      event.preventDefault();
    } else {
      setSelectionOrigin(view, "key");
    }
  };

  editHandlers.keyup = function (view, e) {
    if (e.keyCode == 16) { view.shiftKey = false; }
  };

  editHandlers.keypress = function (view, event) {
    if (inOrNearComposition(view, event) || !event.charCode ||
        event.ctrlKey && !event.altKey || result.mac && event.metaKey) { return }

    if (view.someProp("handleKeyPress", function (f) { return f(view, event); })) {
      event.preventDefault();
      return
    }

    var sel = view.state.selection;
    if (!(sel instanceof TextSelection) || !sel.$from.sameParent(sel.$to)) {
      var text = String.fromCharCode(event.charCode);
      if (!view.someProp("handleTextInput", function (f) { return f(view, sel.$from.pos, sel.$to.pos, text); }))
        { view.dispatch(view.state.tr.insertText(text).scrollIntoView()); }
      event.preventDefault();
    }
  };

  function eventCoords(event) { return {left: event.clientX, top: event.clientY} }

  function isNear(event, click) {
    var dx = click.x - event.clientX, dy = click.y - event.clientY;
    return dx * dx + dy * dy < 100
  }

  function runHandlerOnContext(view, propName, pos, inside, event) {
    if (inside == -1) { return false }
    var $pos = view.state.doc.resolve(inside);
    var loop = function ( i ) {
      if (view.someProp(propName, function (f) { return i > $pos.depth ? f(view, pos, $pos.nodeAfter, $pos.before(i), event, true)
                                                      : f(view, pos, $pos.node(i), $pos.before(i), event, false); }))
        { return { v: true } }
    };

    for (var i = $pos.depth + 1; i > 0; i--) {
      var returned = loop( i );

      if ( returned ) { return returned.v; }
    }
    return false
  }

  function updateSelection(view, selection, origin) {
    if (!view.focused) { view.focus(); }
    var tr = view.state.tr.setSelection(selection);
    if (origin == "pointer") { tr.setMeta("pointer", true); }
    view.dispatch(tr);
  }

  function selectClickedLeaf(view, inside) {
    if (inside == -1) { return false }
    var $pos = view.state.doc.resolve(inside), node = $pos.nodeAfter;
    if (node && node.isAtom && NodeSelection.isSelectable(node)) {
      updateSelection(view, new NodeSelection($pos), "pointer");
      return true
    }
    return false
  }

  function selectClickedNode(view, inside) {
    if (inside == -1) { return false }
    var sel = view.state.selection, selectedNode, selectAt;
    if (sel instanceof NodeSelection) { selectedNode = sel.node; }

    var $pos = view.state.doc.resolve(inside);
    for (var i = $pos.depth + 1; i > 0; i--) {
      var node = i > $pos.depth ? $pos.nodeAfter : $pos.node(i);
      if (NodeSelection.isSelectable(node)) {
        if (selectedNode && sel.$from.depth > 0 &&
            i >= sel.$from.depth && $pos.before(sel.$from.depth + 1) == sel.$from.pos)
          { selectAt = $pos.before(sel.$from.depth); }
        else
          { selectAt = $pos.before(i); }
        break
      }
    }

    if (selectAt != null) {
      updateSelection(view, NodeSelection.create(view.state.doc, selectAt), "pointer");
      return true
    } else {
      return false
    }
  }

  function handleSingleClick(view, pos, inside, event, selectNode) {
    return runHandlerOnContext(view, "handleClickOn", pos, inside, event) ||
      view.someProp("handleClick", function (f) { return f(view, pos, event); }) ||
      (selectNode ? selectClickedNode(view, inside) : selectClickedLeaf(view, inside))
  }

  function handleDoubleClick(view, pos, inside, event) {
    return runHandlerOnContext(view, "handleDoubleClickOn", pos, inside, event) ||
      view.someProp("handleDoubleClick", function (f) { return f(view, pos, event); })
  }

  function handleTripleClick$1(view, pos, inside, event) {
    return runHandlerOnContext(view, "handleTripleClickOn", pos, inside, event) ||
      view.someProp("handleTripleClick", function (f) { return f(view, pos, event); }) ||
      defaultTripleClick(view, inside, event)
  }

  function defaultTripleClick(view, inside, event) {
    if (event.button != 0) { return false }
    var doc = view.state.doc;
    if (inside == -1) {
      if (doc.inlineContent) {
        updateSelection(view, TextSelection.create(doc, 0, doc.content.size), "pointer");
        return true
      }
      return false
    }

    var $pos = doc.resolve(inside);
    for (var i = $pos.depth + 1; i > 0; i--) {
      var node = i > $pos.depth ? $pos.nodeAfter : $pos.node(i);
      var nodePos = $pos.before(i);
      if (node.inlineContent)
        { updateSelection(view, TextSelection.create(doc, nodePos + 1, nodePos + 1 + node.content.size), "pointer"); }
      else if (NodeSelection.isSelectable(node))
        { updateSelection(view, NodeSelection.create(doc, nodePos), "pointer"); }
      else
        { continue }
      return true
    }
  }

  function forceDOMFlush(view) {
    return endComposition(view)
  }

  var selectNodeModifier = result.mac ? "metaKey" : "ctrlKey";

  handlers.mousedown = function (view, event) {
    view.shiftKey = event.shiftKey;
    var flushed = forceDOMFlush(view);
    var now = Date.now(), type = "singleClick";
    if (now - view.lastClick.time < 500 && isNear(event, view.lastClick) && !event[selectNodeModifier]) {
      if (view.lastClick.type == "singleClick") { type = "doubleClick"; }
      else if (view.lastClick.type == "doubleClick") { type = "tripleClick"; }
    }
    view.lastClick = {time: now, x: event.clientX, y: event.clientY, type: type};

    var pos = view.posAtCoords(eventCoords(event));
    if (!pos) { return }

    if (type == "singleClick") {
      if (view.mouseDown) { view.mouseDown.done(); }
      view.mouseDown = new MouseDown(view, pos, event, flushed);
    } else if ((type == "doubleClick" ? handleDoubleClick : handleTripleClick$1)(view, pos.pos, pos.inside, event)) {
      event.preventDefault();
    } else {
      setSelectionOrigin(view, "pointer");
    }
  };

  var MouseDown = function MouseDown(view, pos, event, flushed) {
    var this$1$1 = this;

    this.view = view;
    this.startDoc = view.state.doc;
    this.pos = pos;
    this.event = event;
    this.flushed = flushed;
    this.selectNode = event[selectNodeModifier];
    this.allowDefault = event.shiftKey;
    this.delayedSelectionSync = false;

    var targetNode, targetPos;
    if (pos.inside > -1) {
      targetNode = view.state.doc.nodeAt(pos.inside);
      targetPos = pos.inside;
    } else {
      var $pos = view.state.doc.resolve(pos.pos);
      targetNode = $pos.parent;
      targetPos = $pos.depth ? $pos.before() : 0;
    }

    this.mightDrag = null;

    var target = flushed ? null : event.target;
    var targetDesc = target ? view.docView.nearestDesc(target, true) : null;
    this.target = targetDesc ? targetDesc.dom : null;

    var ref = view.state;
    var selection = ref.selection;
    if (event.button == 0 &&
        targetNode.type.spec.draggable && targetNode.type.spec.selectable !== false ||
        selection instanceof NodeSelection && selection.from <= targetPos && selection.to > targetPos)
      { this.mightDrag = {node: targetNode,
                        pos: targetPos,
                        addAttr: this.target && !this.target.draggable,
                        setUneditable: this.target && result.gecko && !this.target.hasAttribute("contentEditable")}; }

    if (this.target && this.mightDrag && (this.mightDrag.addAttr || this.mightDrag.setUneditable)) {
      this.view.domObserver.stop();
      if (this.mightDrag.addAttr) { this.target.draggable = true; }
      if (this.mightDrag.setUneditable)
        { setTimeout(function () {
          if (this$1$1.view.mouseDown == this$1$1) { this$1$1.target.setAttribute("contentEditable", "false"); }
        }, 20); }
      this.view.domObserver.start();
    }

    view.root.addEventListener("mouseup", this.up = this.up.bind(this));
    view.root.addEventListener("mousemove", this.move = this.move.bind(this));
    setSelectionOrigin(view, "pointer");
  };

  MouseDown.prototype.done = function done () {
      var this$1$1 = this;

    this.view.root.removeEventListener("mouseup", this.up);
    this.view.root.removeEventListener("mousemove", this.move);
    if (this.mightDrag && this.target) {
      this.view.domObserver.stop();
      if (this.mightDrag.addAttr) { this.target.removeAttribute("draggable"); }
      if (this.mightDrag.setUneditable) { this.target.removeAttribute("contentEditable"); }
      this.view.domObserver.start();
    }
    if (this.delayedSelectionSync) { setTimeout(function () { return selectionToDOM(this$1$1.view); }); }
    this.view.mouseDown = null;
  };

  MouseDown.prototype.up = function up (event) {
    this.done();

    if (!this.view.dom.contains(event.target.nodeType == 3 ? event.target.parentNode : event.target))
      { return }

    var pos = this.pos;
    if (this.view.state.doc != this.startDoc) { pos = this.view.posAtCoords(eventCoords(event)); }

    if (this.allowDefault || !pos) {
      setSelectionOrigin(this.view, "pointer");
    } else if (handleSingleClick(this.view, pos.pos, pos.inside, event, this.selectNode)) {
      event.preventDefault();
    } else if (event.button == 0 &&
               (this.flushed ||
                // Safari ignores clicks on draggable elements
                (result.safari && this.mightDrag && !this.mightDrag.node.isAtom) ||
                // Chrome will sometimes treat a node selection as a
                // cursor, but still report that the node is selected
                // when asked through getSelection. You'll then get a
                // situation where clicking at the point where that
                // (hidden) cursor is doesn't change the selection, and
                // thus doesn't get a reaction from ProseMirror. This
                // works around that.
                (result.chrome && !(this.view.state.selection instanceof TextSelection) &&
                 Math.min(Math.abs(pos.pos - this.view.state.selection.from),
                          Math.abs(pos.pos - this.view.state.selection.to)) <= 2))) {
      updateSelection(this.view, Selection.near(this.view.state.doc.resolve(pos.pos)), "pointer");
      event.preventDefault();
    } else {
      setSelectionOrigin(this.view, "pointer");
    }
  };

  MouseDown.prototype.move = function move (event) {
    if (!this.allowDefault && (Math.abs(this.event.x - event.clientX) > 4 ||
                               Math.abs(this.event.y - event.clientY) > 4))
      { this.allowDefault = true; }
    setSelectionOrigin(this.view, "pointer");
    if (event.buttons == 0) { this.done(); }
  };

  handlers.touchdown = function (view) {
    forceDOMFlush(view);
    setSelectionOrigin(view, "pointer");
  };

  handlers.contextmenu = function (view) { return forceDOMFlush(view); };

  function inOrNearComposition(view, event) {
    if (view.composing) { return true }
    // See https://www.stum.de/2016/06/24/handling-ime-events-in-javascript/.
    // On Japanese input method editors (IMEs), the Enter key is used to confirm character
    // selection. On Safari, when Enter is pressed, compositionend and keydown events are
    // emitted. The keydown event triggers newline insertion, which we don't want.
    // This method returns true if the keydown event should be ignored.
    // We only ignore it once, as pressing Enter a second time *should* insert a newline.
    // Furthermore, the keydown event timestamp must be close to the compositionEndedAt timestamp.
    // This guards against the case where compositionend is triggered without the keyboard
    // (e.g. character confirmation may be done with the mouse), and keydown is triggered
    // afterwards- we wouldn't want to ignore the keydown event in this case.
    if (result.safari && Math.abs(event.timeStamp - view.compositionEndedAt) < 500) {
      view.compositionEndedAt = -2e8;
      return true
    }
    return false
  }

  // Drop active composition after 5 seconds of inactivity on Android
  var timeoutComposition = result.android ? 5000 : -1;

  editHandlers.compositionstart = editHandlers.compositionupdate = function (view) {
    if (!view.composing) {
      view.domObserver.flush();
      var state = view.state;
      var $pos = state.selection.$from;
      if (state.selection.empty &&
          (state.storedMarks ||
           (!$pos.textOffset && $pos.parentOffset && $pos.nodeBefore.marks.some(function (m) { return m.type.spec.inclusive === false; })))) {
        // Need to wrap the cursor in mark nodes different from the ones in the DOM context
        view.markCursor = view.state.storedMarks || $pos.marks();
        endComposition(view, true);
        view.markCursor = null;
      } else {
        endComposition(view);
        // In firefox, if the cursor is after but outside a marked node,
        // the inserted text won't inherit the marks. So this moves it
        // inside if necessary.
        if (result.gecko && state.selection.empty && $pos.parentOffset && !$pos.textOffset && $pos.nodeBefore.marks.length) {
          var sel = view.root.getSelection();
          for (var node = sel.focusNode, offset = sel.focusOffset; node && node.nodeType == 1 && offset != 0;) {
            var before = offset < 0 ? node.lastChild : node.childNodes[offset - 1];
            if (!before) { break }
            if (before.nodeType == 3) {
              sel.collapse(before, before.nodeValue.length);
              break
            } else {
              node = before;
              offset = -1;
            }
          }
        }
      }
      view.composing = true;
    }
    scheduleComposeEnd(view, timeoutComposition);
  };

  editHandlers.compositionend = function (view, event) {
    if (view.composing) {
      view.composing = false;
      view.compositionEndedAt = event.timeStamp;
      scheduleComposeEnd(view, 20);
    }
  };

  function scheduleComposeEnd(view, delay) {
    clearTimeout(view.composingTimeout);
    if (delay > -1) { view.composingTimeout = setTimeout(function () { return endComposition(view); }, delay); }
  }

  function clearComposition(view) {
    if (view.composing) {
      view.composing = false;
      view.compositionEndedAt = timestampFromCustomEvent();
    }
    while (view.compositionNodes.length > 0) { view.compositionNodes.pop().markParentsDirty(); }
  }

  function timestampFromCustomEvent() {
    var event = document.createEvent("Event");
    event.initEvent("event", true, true);
    return event.timeStamp
  }

  function endComposition(view, forceUpdate) {
    view.domObserver.forceFlush();
    clearComposition(view);
    if (forceUpdate || view.docView.dirty) {
      var sel = selectionFromDOM(view);
      if (sel && !sel.eq(view.state.selection)) { view.dispatch(view.state.tr.setSelection(sel)); }
      else { view.updateState(view.state); }
      return true
    }
    return false
  }

  function captureCopy(view, dom) {
    // The extra wrapper is somehow necessary on IE/Edge to prevent the
    // content from being mangled when it is put onto the clipboard
    if (!view.dom.parentNode) { return }
    var wrap = view.dom.parentNode.appendChild(document.createElement("div"));
    wrap.appendChild(dom);
    wrap.style.cssText = "position: fixed; left: -10000px; top: 10px";
    var sel = getSelection(), range = document.createRange();
    range.selectNodeContents(dom);
    // Done because IE will fire a selectionchange moving the selection
    // to its start when removeAllRanges is called and the editor still
    // has focus (which will mess up the editor's selection state).
    view.dom.blur();
    sel.removeAllRanges();
    sel.addRange(range);
    setTimeout(function () {
      if (wrap.parentNode) { wrap.parentNode.removeChild(wrap); }
      view.focus();
    }, 50);
  }

  // This is very crude, but unfortunately both these browsers _pretend_
  // that they have a clipboard API—all the objects and methods are
  // there, they just don't work, and they are hard to test.
  var brokenClipboardAPI = (result.ie && result.ie_version < 15) ||
        (result.ios && result.webkit_version < 604);

  handlers.copy = editHandlers.cut = function (view, e) {
    var sel = view.state.selection, cut = e.type == "cut";
    if (sel.empty) { return }

    // IE and Edge's clipboard interface is completely broken
    var data = brokenClipboardAPI ? null : e.clipboardData;
    var slice = sel.content();
    var ref = serializeForClipboard(view, slice);
    var dom = ref.dom;
    var text = ref.text;
    if (data) {
      e.preventDefault();
      data.clearData();
      data.setData("text/html", dom.innerHTML);
      data.setData("text/plain", text);
    } else {
      captureCopy(view, dom);
    }
    if (cut) { view.dispatch(view.state.tr.deleteSelection().scrollIntoView().setMeta("uiEvent", "cut")); }
  };

  function sliceSingleNode(slice) {
    return slice.openStart == 0 && slice.openEnd == 0 && slice.content.childCount == 1 ? slice.content.firstChild : null
  }

  function capturePaste(view, e) {
    if (!view.dom.parentNode) { return }
    var plainText = view.shiftKey || view.state.selection.$from.parent.type.spec.code;
    var target = view.dom.parentNode.appendChild(document.createElement(plainText ? "textarea" : "div"));
    if (!plainText) { target.contentEditable = "true"; }
    target.style.cssText = "position: fixed; left: -10000px; top: 10px";
    target.focus();
    setTimeout(function () {
      view.focus();
      if (target.parentNode) { target.parentNode.removeChild(target); }
      if (plainText) { doPaste(view, target.value, null, e); }
      else { doPaste(view, target.textContent, target.innerHTML, e); }
    }, 50);
  }

  function doPaste(view, text, html, e) {
    var slice = parseFromClipboard(view, text, html, view.shiftKey, view.state.selection.$from);
    if (view.someProp("handlePaste", function (f) { return f(view, e, slice || Slice$1.empty); })) { return true }
    if (!slice) { return false }

    var singleNode = sliceSingleNode(slice);
    var tr = singleNode ? view.state.tr.replaceSelectionWith(singleNode, view.shiftKey) : view.state.tr.replaceSelection(slice);
    view.dispatch(tr.scrollIntoView().setMeta("paste", true).setMeta("uiEvent", "paste"));
    return true
  }

  editHandlers.paste = function (view, e) {
    var data = brokenClipboardAPI ? null : e.clipboardData;
    if (data && doPaste(view, data.getData("text/plain"), data.getData("text/html"), e)) { e.preventDefault(); }
    else { capturePaste(view, e); }
  };

  var Dragging = function Dragging(slice, move) {
    this.slice = slice;
    this.move = move;
  };

  var dragCopyModifier = result.mac ? "altKey" : "ctrlKey";

  handlers.dragstart = function (view, e) {
    var mouseDown = view.mouseDown;
    if (mouseDown) { mouseDown.done(); }
    if (!e.dataTransfer) { return }

    var sel = view.state.selection;
    var pos = sel.empty ? null : view.posAtCoords(eventCoords(e));
    if (pos && pos.pos >= sel.from && pos.pos <= (sel instanceof NodeSelection ? sel.to - 1: sel.to)) ; else if (mouseDown && mouseDown.mightDrag) {
      view.dispatch(view.state.tr.setSelection(NodeSelection.create(view.state.doc, mouseDown.mightDrag.pos)));
    } else if (e.target && e.target.nodeType == 1) {
      var desc = view.docView.nearestDesc(e.target, true);
      if (desc && desc.node.type.spec.draggable && desc != view.docView)
        { view.dispatch(view.state.tr.setSelection(NodeSelection.create(view.state.doc, desc.posBefore))); }
    }
    var slice = view.state.selection.content();
    var ref = serializeForClipboard(view, slice);
    var dom = ref.dom;
    var text = ref.text;
    e.dataTransfer.clearData();
    e.dataTransfer.setData(brokenClipboardAPI ? "Text" : "text/html", dom.innerHTML);
    // See https://github.com/ProseMirror/prosemirror/issues/1156
    e.dataTransfer.effectAllowed = "copyMove";
    if (!brokenClipboardAPI) { e.dataTransfer.setData("text/plain", text); }
    view.dragging = new Dragging(slice, !e[dragCopyModifier]);
  };

  handlers.dragend = function (view) {
    var dragging = view.dragging;
    window.setTimeout(function () {
      if (view.dragging == dragging)  { view.dragging = null; }
    }, 50);
  };

  editHandlers.dragover = editHandlers.dragenter = function (_, e) { return e.preventDefault(); };

  editHandlers.drop = function (view, e) {
    var dragging = view.dragging;
    view.dragging = null;

    if (!e.dataTransfer) { return }

    var eventPos = view.posAtCoords(eventCoords(e));
    if (!eventPos) { return }
    var $mouse = view.state.doc.resolve(eventPos.pos);
    if (!$mouse) { return }
    var slice = dragging && dragging.slice;
    if (slice) {
      view.someProp("transformPasted", function (f) { slice = f(slice); });
    } else {
      slice = parseFromClipboard(view, e.dataTransfer.getData(brokenClipboardAPI ? "Text" : "text/plain"),
                                 brokenClipboardAPI ? null : e.dataTransfer.getData("text/html"), false, $mouse);
    }
    var move = dragging && !e[dragCopyModifier];
    if (view.someProp("handleDrop", function (f) { return f(view, e, slice || Slice$1.empty, move); })) {
      e.preventDefault();
      return
    }
    if (!slice) { return }

    e.preventDefault();
    var insertPos = slice ? dropPoint(view.state.doc, $mouse.pos, slice) : $mouse.pos;
    if (insertPos == null) { insertPos = $mouse.pos; }

    var tr = view.state.tr;
    if (move) { tr.deleteSelection(); }

    var pos = tr.mapping.map(insertPos);
    var isNode = slice.openStart == 0 && slice.openEnd == 0 && slice.content.childCount == 1;
    var beforeInsert = tr.doc;
    if (isNode)
      { tr.replaceRangeWith(pos, pos, slice.content.firstChild); }
    else
      { tr.replaceRange(pos, pos, slice); }
    if (tr.doc.eq(beforeInsert)) { return }

    var $pos = tr.doc.resolve(pos);
    if (isNode && NodeSelection.isSelectable(slice.content.firstChild) &&
        $pos.nodeAfter && $pos.nodeAfter.sameMarkup(slice.content.firstChild)) {
      tr.setSelection(new NodeSelection($pos));
    } else {
      var end = tr.mapping.map(insertPos);
      tr.mapping.maps[tr.mapping.maps.length - 1].forEach(function (_from, _to, _newFrom, newTo) { return end = newTo; });
      tr.setSelection(selectionBetween(view, $pos, tr.doc.resolve(end)));
    }
    view.focus();
    view.dispatch(tr.setMeta("uiEvent", "drop"));
  };

  handlers.focus = function (view) {
    if (!view.focused) {
      view.domObserver.stop();
      view.dom.classList.add("ProseMirror-focused");
      view.domObserver.start();
      view.focused = true;
      setTimeout(function () {
        if (view.docView && view.hasFocus() && !view.domObserver.currentSelection.eq(view.root.getSelection()))
          { selectionToDOM(view); }
      }, 20);
    }
  };

  handlers.blur = function (view, e) {
    if (view.focused) {
      view.domObserver.stop();
      view.dom.classList.remove("ProseMirror-focused");
      view.domObserver.start();
      if (e.relatedTarget && view.dom.contains(e.relatedTarget))
        { view.domObserver.currentSelection.set({}); }
      view.focused = false;
    }
  };

  handlers.beforeinput = function (view, event) {
    // We should probably do more with beforeinput events, but support
    // is so spotty that I'm still waiting to see where they are going.

    // Very specific hack to deal with backspace sometimes failing on
    // Chrome Android when after an uneditable node.
    if (result.chrome && result.android && event.inputType == "deleteContentBackward") {
      var domChangeCount = view.domChangeCount;
      setTimeout(function () {
        if (view.domChangeCount != domChangeCount) { return } // Event already had some effect
        // This bug tends to close the virtual keyboard, so we refocus
        view.dom.blur();
        view.focus();
        if (view.someProp("handleKeyDown", function (f) { return f(view, keyEvent(8, "Backspace")); })) { return }
        var ref = view.state.selection;
        var $cursor = ref.$cursor;
        // Crude approximation of backspace behavior when no command handled it
        if ($cursor && $cursor.pos > 0) { view.dispatch(view.state.tr.delete($cursor.pos - 1, $cursor.pos).scrollIntoView()); }
      }, 50);
    }
  };

  // Make sure all handlers get registered
  for (var prop$2 in editHandlers) { handlers[prop$2] = editHandlers[prop$2]; }

  function compareObjs(a, b) {
    if (a == b) { return true }
    for (var p in a) { if (a[p] !== b[p]) { return false } }
    for (var p$1 in b) { if (!(p$1 in a)) { return false } }
    return true
  }

  var WidgetType = function WidgetType(toDOM, spec) {
    this.spec = spec || noSpec;
    this.side = this.spec.side || 0;
    this.toDOM = toDOM;
  };

  WidgetType.prototype.map = function map (mapping, span, offset, oldOffset) {
    var ref = mapping.mapResult(span.from + oldOffset, this.side < 0 ? -1 : 1);
      var pos = ref.pos;
      var deleted = ref.deleted;
    return deleted ? null : new Decoration(pos - offset, pos - offset, this)
  };

  WidgetType.prototype.valid = function valid () { return true };

  WidgetType.prototype.eq = function eq (other) {
    return this == other ||
      (other instanceof WidgetType &&
       (this.spec.key && this.spec.key == other.spec.key ||
        this.toDOM == other.toDOM && compareObjs(this.spec, other.spec)))
  };

  var InlineType = function InlineType(attrs, spec) {
    this.spec = spec || noSpec;
    this.attrs = attrs;
  };

  InlineType.prototype.map = function map (mapping, span, offset, oldOffset) {
    var from = mapping.map(span.from + oldOffset, this.spec.inclusiveStart ? -1 : 1) - offset;
    var to = mapping.map(span.to + oldOffset, this.spec.inclusiveEnd ? 1 : -1) - offset;
    return from >= to ? null : new Decoration(from, to, this)
  };

  InlineType.prototype.valid = function valid (_, span) { return span.from < span.to };

  InlineType.prototype.eq = function eq (other) {
    return this == other ||
      (other instanceof InlineType && compareObjs(this.attrs, other.attrs) &&
       compareObjs(this.spec, other.spec))
  };

  InlineType.is = function is (span) { return span.type instanceof InlineType };

  var NodeType = function NodeType(attrs, spec) {
    this.spec = spec || noSpec;
    this.attrs = attrs;
  };

  NodeType.prototype.map = function map (mapping, span, offset, oldOffset) {
    var from = mapping.mapResult(span.from + oldOffset, 1);
    if (from.deleted) { return null }
    var to = mapping.mapResult(span.to + oldOffset, -1);
    if (to.deleted || to.pos <= from.pos) { return null }
    return new Decoration(from.pos - offset, to.pos - offset, this)
  };

  NodeType.prototype.valid = function valid (node, span) {
    var ref = node.content.findIndex(span.from);
      var index = ref.index;
      var offset = ref.offset;
    return offset == span.from && offset + node.child(index).nodeSize == span.to
  };

  NodeType.prototype.eq = function eq (other) {
    return this == other ||
      (other instanceof NodeType && compareObjs(this.attrs, other.attrs) &&
       compareObjs(this.spec, other.spec))
  };

  // ::- Decoration objects can be provided to the view through the
  // [`decorations` prop](#view.EditorProps.decorations). They come in
  // several variants—see the static members of this class for details.
  var Decoration = function Decoration(from, to, type) {
    // :: number
    // The start position of the decoration.
    this.from = from;
    // :: number
    // The end position. Will be the same as `from` for [widget
    // decorations](#view.Decoration^widget).
    this.to = to;
    this.type = type;
  };

  var prototypeAccessors$1$2 = { spec: { configurable: true },inline: { configurable: true } };

  Decoration.prototype.copy = function copy (from, to) {
    return new Decoration(from, to, this.type)
  };

  Decoration.prototype.eq = function eq (other, offset) {
      if ( offset === void 0 ) { offset = 0; }

    return this.type.eq(other.type) && this.from + offset == other.from && this.to + offset == other.to
  };

  Decoration.prototype.map = function map (mapping, offset, oldOffset) {
    return this.type.map(mapping, this, offset, oldOffset)
  };

  // :: (number, union<(view: EditorView, getPos: () → number) → dom.Node, dom.Node>, ?Object) → Decoration
  // Creates a widget decoration, which is a DOM node that's shown in
  // the document at the given position. It is recommended that you
  // delay rendering the widget by passing a function that will be
  // called when the widget is actually drawn in a view, but you can
  // also directly pass a DOM node. `getPos` can be used to find the
  // widget's current document position.
  //
  // spec::- These options are supported:
  //
  //   side:: ?number
  //   Controls which side of the document position this widget is
  //   associated with. When negative, it is drawn before a cursor
  //   at its position, and content inserted at that position ends
  //   up after the widget. When zero (the default) or positive, the
  //   widget is drawn after the cursor and content inserted there
  //   ends up before the widget.
  //
  //   When there are multiple widgets at a given position, their
  //   `side` values determine the order in which they appear. Those
  //   with lower values appear first. The ordering of widgets with
  //   the same `side` value is unspecified.
  //
  //   When `marks` is null, `side` also determines the marks that
  //   the widget is wrapped in—those of the node before when
  //   negative, those of the node after when positive.
  //
  //   marks:: ?[Mark]
  //   The precise set of marks to draw around the widget.
  //
  //   stopEvent:: ?(event: dom.Event) → bool
  //   Can be used to control which DOM events, when they bubble out
  //   of this widget, the editor view should ignore.
  //
  //   ignoreSelection:: ?bool
  //   When set (defaults to false), selection changes inside the
  //   widget are ignored, and don't cause ProseMirror to try and
  //   re-sync the selection with its selection state.
  //
  //   key:: ?string
  //   When comparing decorations of this type (in order to decide
  //   whether it needs to be redrawn), ProseMirror will by default
  //   compare the widget DOM node by identity. If you pass a key,
  //   that key will be compared instead, which can be useful when
  //   you generate decorations on the fly and don't want to store
  //   and reuse DOM nodes. Make sure that any widgets with the same
  //   key are interchangeable—if widgets differ in, for example,
  //   the behavior of some event handler, they should get
  //   different keys.
  Decoration.widget = function widget (pos, toDOM, spec) {
    return new Decoration(pos, pos, new WidgetType(toDOM, spec))
  };

  // :: (number, number, DecorationAttrs, ?Object) → Decoration
  // Creates an inline decoration, which adds the given attributes to
  // each inline node between `from` and `to`.
  //
  // spec::- These options are recognized:
  //
  //   inclusiveStart:: ?bool
  //   Determines how the left side of the decoration is
  //   [mapped](#transform.Position_Mapping) when content is
  //   inserted directly at that position. By default, the decoration
  //   won't include the new content, but you can set this to `true`
  //   to make it inclusive.
  //
  //   inclusiveEnd:: ?bool
  //   Determines how the right side of the decoration is mapped.
  //   See
  //   [`inclusiveStart`](#view.Decoration^inline^spec.inclusiveStart).
  Decoration.inline = function inline (from, to, attrs, spec) {
    return new Decoration(from, to, new InlineType(attrs, spec))
  };

  // :: (number, number, DecorationAttrs, ?Object) → Decoration
  // Creates a node decoration. `from` and `to` should point precisely
  // before and after a node in the document. That node, and only that
  // node, will receive the given attributes.
  //
  // spec::-
  //
  // Optional information to store with the decoration. It
  // is also used when comparing decorators for equality.
  Decoration.node = function node (from, to, attrs, spec) {
    return new Decoration(from, to, new NodeType(attrs, spec))
  };

  // :: Object
  // The spec provided when creating this decoration. Can be useful
  // if you've stored extra information in that object.
  prototypeAccessors$1$2.spec.get = function () { return this.type.spec };

  prototypeAccessors$1$2.inline.get = function () { return this.type instanceof InlineType };

  Object.defineProperties( Decoration.prototype, prototypeAccessors$1$2 );

  // DecorationAttrs:: interface
  // A set of attributes to add to a decorated node. Most properties
  // simply directly correspond to DOM attributes of the same name,
  // which will be set to the property's value. These are exceptions:
  //
  //   class:: ?string
  //   A CSS class name or a space-separated set of class names to be
  //   _added_ to the classes that the node already had.
  //
  //   style:: ?string
  //   A string of CSS to be _added_ to the node's existing `style` property.
  //
  //   nodeName:: ?string
  //   When non-null, the target node is wrapped in a DOM element of
  //   this type (and the other attributes are applied to this element).

  var none = [], noSpec = {};

  // :: class extends DecorationSource
  // A collection of [decorations](#view.Decoration), organized in
  // such a way that the drawing algorithm can efficiently use and
  // compare them. This is a persistent data structure—it is not
  // modified, updates create a new value.
  var DecorationSet = function DecorationSet(local, children) {
    this.local = local && local.length ? local : none;
    this.children = children && children.length ? children : none;
  };

  // :: (Node, [Decoration]) → DecorationSet
  // Create a set of decorations, using the structure of the given
  // document.
  DecorationSet.create = function create (doc, decorations) {
    return decorations.length ? buildTree(decorations, doc, 0, noSpec) : empty$2
  };

  // :: (?number, ?number, ?(spec: Object) → bool) → [Decoration]
  // Find all decorations in this set which touch the given range
  // (including decorations that start or end directly at the
  // boundaries) and match the given predicate on their spec. When
  // `start` and `end` are omitted, all decorations in the set are
  // considered. When `predicate` isn't given, all decorations are
  // assumed to match.
  DecorationSet.prototype.find = function find (start, end, predicate) {
    var result = [];
    this.findInner(start == null ? 0 : start, end == null ? 1e9 : end, result, 0, predicate);
    return result
  };

  DecorationSet.prototype.findInner = function findInner (start, end, result, offset, predicate) {
    for (var i = 0; i < this.local.length; i++) {
      var span = this.local[i];
      if (span.from <= end && span.to >= start && (!predicate || predicate(span.spec)))
        { result.push(span.copy(span.from + offset, span.to + offset)); }
    }
    for (var i$1 = 0; i$1 < this.children.length; i$1 += 3) {
      if (this.children[i$1] < end && this.children[i$1 + 1] > start) {
        var childOff = this.children[i$1] + 1;
        this.children[i$1 + 2].findInner(start - childOff, end - childOff, result, offset + childOff, predicate);
      }
    }
  };

  // :: (Mapping, Node, ?Object) → DecorationSet
  // Map the set of decorations in response to a change in the
  // document.
  //
  // options::- An optional set of options.
  //
  //   onRemove:: ?(decorationSpec: Object)
  //   When given, this function will be called for each decoration
  //   that gets dropped as a result of the mapping, passing the
  //   spec of that decoration.
  DecorationSet.prototype.map = function map (mapping, doc, options) {
    if (this == empty$2 || mapping.maps.length == 0) { return this }
    return this.mapInner(mapping, doc, 0, 0, options || noSpec)
  };

  DecorationSet.prototype.mapInner = function mapInner (mapping, node, offset, oldOffset, options) {
    var newLocal;
    for (var i = 0; i < this.local.length; i++) {
      var mapped = this.local[i].map(mapping, offset, oldOffset);
      if (mapped && mapped.type.valid(node, mapped)) { (newLocal || (newLocal = [])).push(mapped); }
      else if (options.onRemove) { options.onRemove(this.local[i].spec); }
    }

    if (this.children.length)
      { return mapChildren(this.children, newLocal, mapping, node, offset, oldOffset, options) }
    else
      { return newLocal ? new DecorationSet(newLocal.sort(byPos)) : empty$2 }
  };

  // :: (Node, [Decoration]) → DecorationSet
  // Add the given array of decorations to the ones in the set,
  // producing a new set. Needs access to the current document to
  // create the appropriate tree structure.
  DecorationSet.prototype.add = function add (doc, decorations) {
    if (!decorations.length) { return this }
    if (this == empty$2) { return DecorationSet.create(doc, decorations) }
    return this.addInner(doc, decorations, 0)
  };

  DecorationSet.prototype.addInner = function addInner (doc, decorations, offset) {
      var this$1$1 = this;

    var children, childIndex = 0;
    doc.forEach(function (childNode, childOffset) {
      var baseOffset = childOffset + offset, found;
      if (!(found = takeSpansForNode(decorations, childNode, baseOffset))) { return }

      if (!children) { children = this$1$1.children.slice(); }
      while (childIndex < children.length && children[childIndex] < childOffset) { childIndex += 3; }
      if (children[childIndex] == childOffset)
        { children[childIndex + 2] = children[childIndex + 2].addInner(childNode, found, baseOffset + 1); }
      else
        { children.splice(childIndex, 0, childOffset, childOffset + childNode.nodeSize, buildTree(found, childNode, baseOffset + 1, noSpec)); }
      childIndex += 3;
    });

    var local = moveSpans(childIndex ? withoutNulls(decorations) : decorations, -offset);
    for (var i = 0; i < local.length; i++) { if (!local[i].type.valid(doc, local[i])) { local.splice(i--, 1); } }

    return new DecorationSet(local.length ? this.local.concat(local).sort(byPos) : this.local,
                             children || this.children)
  };

  // :: ([Decoration]) → DecorationSet
  // Create a new set that contains the decorations in this set, minus
  // the ones in the given array.
  DecorationSet.prototype.remove = function remove (decorations) {
    if (decorations.length == 0 || this == empty$2) { return this }
    return this.removeInner(decorations, 0)
  };

  DecorationSet.prototype.removeInner = function removeInner (decorations, offset) {
    var children = this.children, local = this.local;
    for (var i = 0; i < children.length; i += 3) {
      var found = (void 0), from = children[i] + offset, to = children[i + 1] + offset;
      for (var j = 0, span = (void 0); j < decorations.length; j++) { if (span = decorations[j]) {
        if (span.from > from && span.to < to) {
          decorations[j] = null
          ;(found || (found = [])).push(span);
        }
      } }
      if (!found) { continue }
      if (children == this.children) { children = this.children.slice(); }
      var removed = children[i + 2].removeInner(found, from + 1);
      if (removed != empty$2) {
        children[i + 2] = removed;
      } else {
        children.splice(i, 3);
        i -= 3;
      }
    }
    if (local.length) { for (var i$1 = 0, span$1 = (void 0); i$1 < decorations.length; i$1++) { if (span$1 = decorations[i$1]) {
      for (var j$1 = 0; j$1 < local.length; j$1++) { if (local[j$1].eq(span$1, offset)) {
        if (local == this.local) { local = this.local.slice(); }
        local.splice(j$1--, 1);
      } }
    } } }
    if (children == this.children && local == this.local) { return this }
    return local.length || children.length ? new DecorationSet(local, children) : empty$2
  };

  DecorationSet.prototype.forChild = function forChild (offset, node) {
    if (this == empty$2) { return this }
    if (node.isLeaf) { return DecorationSet.empty }

    var child, local;
    for (var i = 0; i < this.children.length; i += 3) { if (this.children[i] >= offset) {
      if (this.children[i] == offset) { child = this.children[i + 2]; }
      break
    } }
    var start = offset + 1, end = start + node.content.size;
    for (var i$1 = 0; i$1 < this.local.length; i$1++) {
      var dec = this.local[i$1];
      if (dec.from < end && dec.to > start && (dec.type instanceof InlineType)) {
        var from = Math.max(start, dec.from) - start, to = Math.min(end, dec.to) - start;
        if (from < to) { (local || (local = [])).push(dec.copy(from, to)); }
      }
    }
    if (local) {
      var localSet = new DecorationSet(local.sort(byPos));
      return child ? new DecorationGroup([localSet, child]) : localSet
    }
    return child || empty$2
  };

  DecorationSet.prototype.eq = function eq (other) {
    if (this == other) { return true }
    if (!(other instanceof DecorationSet) ||
        this.local.length != other.local.length ||
        this.children.length != other.children.length) { return false }
    for (var i = 0; i < this.local.length; i++)
      { if (!this.local[i].eq(other.local[i])) { return false } }
    for (var i$1 = 0; i$1 < this.children.length; i$1 += 3)
      { if (this.children[i$1] != other.children[i$1] ||
          this.children[i$1 + 1] != other.children[i$1 + 1] ||
          !this.children[i$1 + 2].eq(other.children[i$1 + 2])) { return false } }
    return true
  };

  DecorationSet.prototype.locals = function locals (node) {
    return removeOverlap(this.localsInner(node))
  };

  DecorationSet.prototype.localsInner = function localsInner (node) {
    if (this == empty$2) { return none }
    if (node.inlineContent || !this.local.some(InlineType.is)) { return this.local }
    var result = [];
    for (var i = 0; i < this.local.length; i++) {
      if (!(this.local[i].type instanceof InlineType))
        { result.push(this.local[i]); }
    }
    return result
  };

  // DecorationSource:: interface
  // An object that can [provide](#view.EditorProps.decorations)
  // decorations. Implemented by [`DecorationSet`](#view.DecorationSet),
  // and passed to [node views](#view.EditorProps.nodeViews).
  //
  // map:: (Mapping, Node) → DecorationSource
  // Map the set of decorations in response to a change in the
  // document.

  var empty$2 = new DecorationSet();

  // :: DecorationSet
  // The empty set of decorations.
  DecorationSet.empty = empty$2;

  DecorationSet.removeOverlap = removeOverlap;

  // :- An abstraction that allows the code dealing with decorations to
  // treat multiple DecorationSet objects as if it were a single object
  // with (a subset of) the same interface.
  var DecorationGroup = function DecorationGroup(members) {
    this.members = members;
  };

  DecorationGroup.prototype.map = function map (mapping, doc) {
    var mappedDecos = this.members.map(
      function (member) { return member.map(mapping, doc, noSpec); }
    );
    return DecorationGroup.from(mappedDecos)
  };

  DecorationGroup.prototype.forChild = function forChild (offset, child) {
    if (child.isLeaf) { return DecorationSet.empty }
    var found = [];
    for (var i = 0; i < this.members.length; i++) {
      var result = this.members[i].forChild(offset, child);
      if (result == empty$2) { continue }
      if (result instanceof DecorationGroup) { found = found.concat(result.members); }
      else { found.push(result); }
    }
    return DecorationGroup.from(found)
  };

  DecorationGroup.prototype.eq = function eq (other) {
    if (!(other instanceof DecorationGroup) ||
        other.members.length != this.members.length) { return false }
    for (var i = 0; i < this.members.length; i++)
      { if (!this.members[i].eq(other.members[i])) { return false } }
    return true
  };

  DecorationGroup.prototype.locals = function locals (node) {
    var result, sorted = true;
    for (var i = 0; i < this.members.length; i++) {
      var locals = this.members[i].localsInner(node);
      if (!locals.length) { continue }
      if (!result) {
        result = locals;
      } else {
        if (sorted) {
          result = result.slice();
          sorted = false;
        }
        for (var j = 0; j < locals.length; j++) { result.push(locals[j]); }
      }
    }
    return result ? removeOverlap(sorted ? result : result.sort(byPos)) : none
  };

  // : ([DecorationSet]) → union<DecorationSet, DecorationGroup>
  // Create a group for the given array of decoration sets, or return
  // a single set when possible.
  DecorationGroup.from = function from (members) {
    switch (members.length) {
      case 0: return empty$2
      case 1: return members[0]
      default: return new DecorationGroup(members)
    }
  };

  function mapChildren(oldChildren, newLocal, mapping, node, offset, oldOffset, options) {
    var children = oldChildren.slice();

    // Mark the children that are directly touched by changes, and
    // move those that are after the changes.
    var shift = function (oldStart, oldEnd, newStart, newEnd) {
      for (var i = 0; i < children.length; i += 3) {
        var end = children[i + 1], dSize = (void 0);
        if (end == -1 || oldStart > end + oldOffset) { continue }
        if (oldEnd >= children[i] + oldOffset) {
          children[i + 1] = -1;
        } else if (newStart >= offset && (dSize = (newEnd - newStart) - (oldEnd - oldStart))) {
          children[i] += dSize;
          children[i + 1] += dSize;
        }
      }
    };
    for (var i = 0; i < mapping.maps.length; i++) { mapping.maps[i].forEach(shift); }

    // Find the child nodes that still correspond to a single node,
    // recursively call mapInner on them and update their positions.
    var mustRebuild = false;
    for (var i$1 = 0; i$1 < children.length; i$1 += 3) { if (children[i$1 + 1] == -1) { // Touched nodes
      var from = mapping.map(oldChildren[i$1] + oldOffset), fromLocal = from - offset;
      if (fromLocal < 0 || fromLocal >= node.content.size) {
        mustRebuild = true;
        continue
      }
      // Must read oldChildren because children was tagged with -1
      var to = mapping.map(oldChildren[i$1 + 1] + oldOffset, -1), toLocal = to - offset;
      var ref = node.content.findIndex(fromLocal);
      var index = ref.index;
      var childOffset = ref.offset;
      var childNode = node.maybeChild(index);
      if (childNode && childOffset == fromLocal && childOffset + childNode.nodeSize == toLocal) {
        var mapped = children[i$1 + 2].mapInner(mapping, childNode, from + 1, oldChildren[i$1] + oldOffset + 1, options);
        if (mapped != empty$2) {
          children[i$1] = fromLocal;
          children[i$1 + 1] = toLocal;
          children[i$1 + 2] = mapped;
        } else {
          children[i$1 + 1] = -2;
          mustRebuild = true;
        }
      } else {
        mustRebuild = true;
      }
    } }

    // Remaining children must be collected and rebuilt into the appropriate structure
    if (mustRebuild) {
      var decorations = mapAndGatherRemainingDecorations(children, oldChildren, newLocal || [], mapping,
                                                         offset, oldOffset, options);
      var built = buildTree(decorations, node, 0, options);
      newLocal = built.local;
      for (var i$2 = 0; i$2 < children.length; i$2 += 3) { if (children[i$2 + 1] < 0) {
        children.splice(i$2, 3);
        i$2 -= 3;
      } }
      for (var i$3 = 0, j = 0; i$3 < built.children.length; i$3 += 3) {
        var from$1 = built.children[i$3];
        while (j < children.length && children[j] < from$1) { j += 3; }
        children.splice(j, 0, built.children[i$3], built.children[i$3 + 1], built.children[i$3 + 2]);
      }
    }

    return new DecorationSet(newLocal && newLocal.sort(byPos), children)
  }

  function moveSpans(spans, offset) {
    if (!offset || !spans.length) { return spans }
    var result = [];
    for (var i = 0; i < spans.length; i++) {
      var span = spans[i];
      result.push(new Decoration(span.from + offset, span.to + offset, span.type));
    }
    return result
  }

  function mapAndGatherRemainingDecorations(children, oldChildren, decorations, mapping, offset, oldOffset, options) {
    // Gather all decorations from the remaining marked children
    function gather(set, oldOffset) {
      for (var i = 0; i < set.local.length; i++) {
        var mapped = set.local[i].map(mapping, offset, oldOffset);
        if (mapped) { decorations.push(mapped); }
        else if (options.onRemove) { options.onRemove(set.local[i].spec); }
      }
      for (var i$1 = 0; i$1 < set.children.length; i$1 += 3)
        { gather(set.children[i$1 + 2], set.children[i$1] + oldOffset + 1); }
    }
    for (var i = 0; i < children.length; i += 3) { if (children[i + 1] == -1)
      { gather(children[i + 2], oldChildren[i] + oldOffset + 1); } }

    return decorations
  }

  function takeSpansForNode(spans, node, offset) {
    if (node.isLeaf) { return null }
    var end = offset + node.nodeSize, found = null;
    for (var i = 0, span = (void 0); i < spans.length; i++) {
      if ((span = spans[i]) && span.from > offset && span.to < end) {
  (found || (found = [])).push(span);
        spans[i] = null;
      }
    }
    return found
  }

  function withoutNulls(array) {
    var result = [];
    for (var i = 0; i < array.length; i++)
      { if (array[i] != null) { result.push(array[i]); } }
    return result
  }

  // : ([Decoration], Node, number) → DecorationSet
  // Build up a tree that corresponds to a set of decorations. `offset`
  // is a base offset that should be subtracted from the `from` and `to`
  // positions in the spans (so that we don't have to allocate new spans
  // for recursive calls).
  function buildTree(spans, node, offset, options) {
    var children = [], hasNulls = false;
    node.forEach(function (childNode, localStart) {
      var found = takeSpansForNode(spans, childNode, localStart + offset);
      if (found) {
        hasNulls = true;
        var subtree = buildTree(found, childNode, offset + localStart + 1, options);
        if (subtree != empty$2)
          { children.push(localStart, localStart + childNode.nodeSize, subtree); }
      }
    });
    var locals = moveSpans(hasNulls ? withoutNulls(spans) : spans, -offset).sort(byPos);
    for (var i = 0; i < locals.length; i++) { if (!locals[i].type.valid(node, locals[i])) {
      if (options.onRemove) { options.onRemove(locals[i].spec); }
      locals.splice(i--, 1);
    } }
    return locals.length || children.length ? new DecorationSet(locals, children) : empty$2
  }

  // : (Decoration, Decoration) → number
  // Used to sort decorations so that ones with a low start position
  // come first, and within a set with the same start position, those
  // with an smaller end position come first.
  function byPos(a, b) {
    return a.from - b.from || a.to - b.to
  }

  // : ([Decoration]) → [Decoration]
  // Scan a sorted array of decorations for partially overlapping spans,
  // and split those so that only fully overlapping spans are left (to
  // make subsequent rendering easier). Will return the input array if
  // no partially overlapping spans are found (the common case).
  function removeOverlap(spans) {
    var working = spans;
    for (var i = 0; i < working.length - 1; i++) {
      var span = working[i];
      if (span.from != span.to) { for (var j = i + 1; j < working.length; j++) {
        var next = working[j];
        if (next.from == span.from) {
          if (next.to != span.to) {
            if (working == spans) { working = spans.slice(); }
            // Followed by a partially overlapping larger span. Split that
            // span.
            working[j] = next.copy(next.from, span.to);
            insertAhead(working, j + 1, next.copy(span.to, next.to));
          }
          continue
        } else {
          if (next.from < span.to) {
            if (working == spans) { working = spans.slice(); }
            // The end of this one overlaps with a subsequent span. Split
            // this one.
            working[i] = span.copy(span.from, next.from);
            insertAhead(working, j, span.copy(next.from, span.to));
          }
          break
        }
      } }
    }
    return working
  }

  function insertAhead(array, i, deco) {
    while (i < array.length && byPos(deco, array[i]) > 0) { i++; }
    array.splice(i, 0, deco);
  }

  // : (EditorView) → union<DecorationSet, DecorationGroup>
  // Get the decorations associated with the current props of a view.
  function viewDecorations(view) {
    var found = [];
    view.someProp("decorations", function (f) {
      var result = f(view.state);
      if (result && result != empty$2) { found.push(result); }
    });
    if (view.cursorWrapper)
      { found.push(DecorationSet.create(view.state.doc, [view.cursorWrapper.deco])); }
    return DecorationGroup.from(found)
  }

  // ::- An editor view manages the DOM structure that represents an
  // editable document. Its state and behavior are determined by its
  // [props](#view.DirectEditorProps).
  var EditorView = function EditorView(place, props) {
    this._props = props;
    // :: EditorState
    // The view's current [state](#state.EditorState).
    this.state = props.state;

    this.dispatch = this.dispatch.bind(this);

    this._root = null;
    this.focused = false;
    // Kludge used to work around a Chrome bug
    this.trackWrites = null;

    // :: dom.Element
    // An editable DOM node containing the document. (You probably
    // should not directly interfere with its content.)
    this.dom = (place && place.mount) || document.createElement("div");
    if (place) {
      if (place.appendChild) { place.appendChild(this.dom); }
      else if (place.apply) { place(this.dom); }
      else if (place.mount) { this.mounted = true; }
    }

    // :: bool
    // Indicates whether the editor is currently [editable](#view.EditorProps.editable).
    this.editable = getEditable(this);
    this.markCursor = null;
    this.cursorWrapper = null;
    updateCursorWrapper(this);
    this.nodeViews = buildNodeViews(this);
    this.docView = docViewDesc(this.state.doc, computeDocDeco(this), viewDecorations(this), this.dom, this);

    this.lastSelectedViewDesc = null;
    // :: ?{slice: Slice, move: bool}
    // When editor content is being dragged, this object contains
    // information about the dragged slice and whether it is being
    // copied or moved. At any other time, it is null.
    this.dragging = null;

    initInput(this);

    this.pluginViews = [];
    this.updatePluginViews();
  };

  var prototypeAccessors$2$1 = { props: { configurable: true },root: { configurable: true } };

  // composing:: boolean
  // Holds `true` when a
  // [composition](https://developer.mozilla.org/en-US/docs/Mozilla/IME_handling_guide)
  // is active.

  // :: DirectEditorProps
  // The view's current [props](#view.EditorProps).
  prototypeAccessors$2$1.props.get = function () {
    if (this._props.state != this.state) {
      var prev = this._props;
      this._props = {};
      for (var name in prev) { this._props[name] = prev[name]; }
      this._props.state = this.state;
    }
    return this._props
  };

  // :: (DirectEditorProps)
  // Update the view's props. Will immediately cause an update to
  // the DOM.
  EditorView.prototype.update = function update (props) {
    if (props.handleDOMEvents != this._props.handleDOMEvents) { ensureListeners(this); }
    this._props = props;
    this.updateStateInner(props.state, true);
  };

  // :: (DirectEditorProps)
  // Update the view by updating existing props object with the object
  // given as argument. Equivalent to `view.update(Object.assign({},
  // view.props, props))`.
  EditorView.prototype.setProps = function setProps (props) {
    var updated = {};
    for (var name in this._props) { updated[name] = this._props[name]; }
    updated.state = this.state;
    for (var name$1 in props) { updated[name$1] = props[name$1]; }
    this.update(updated);
  };

  // :: (EditorState)
  // Update the editor's `state` prop, without touching any of the
  // other props.
  EditorView.prototype.updateState = function updateState (state) {
    this.updateStateInner(state, this.state.plugins != state.plugins);
  };

  EditorView.prototype.updateStateInner = function updateStateInner (state, reconfigured) {
      var this$1$1 = this;

    var prev = this.state, redraw = false, updateSel = false;
    // When stored marks are added, stop composition, so that they can
    // be displayed.
    if (state.storedMarks && this.composing) {
      clearComposition(this);
      updateSel = true;
    }
    this.state = state;
    if (reconfigured) {
      var nodeViews = buildNodeViews(this);
      if (changedNodeViews(nodeViews, this.nodeViews)) {
        this.nodeViews = nodeViews;
        redraw = true;
      }
      ensureListeners(this);
    }

    this.editable = getEditable(this);
    updateCursorWrapper(this);
    var innerDeco = viewDecorations(this), outerDeco = computeDocDeco(this);

    var scroll = reconfigured ? "reset"
        : state.scrollToSelection > prev.scrollToSelection ? "to selection" : "preserve";
    var updateDoc = redraw || !this.docView.matchesNode(state.doc, outerDeco, innerDeco);
    if (updateDoc || !state.selection.eq(prev.selection)) { updateSel = true; }
    var oldScrollPos = scroll == "preserve" && updateSel && this.dom.style.overflowAnchor == null && storeScrollPos(this);

    if (updateSel) {
      this.domObserver.stop();
      // Work around an issue in Chrome, IE, and Edge where changing
      // the DOM around an active selection puts it into a broken
      // state where the thing the user sees differs from the
      // selection reported by the Selection object (#710, #973,
      // #1011, #1013, #1035).
      var forceSelUpdate = updateDoc && (result.ie || result.chrome) && !this.composing &&
          !prev.selection.empty && !state.selection.empty && selectionContextChanged(prev.selection, state.selection);
      if (updateDoc) {
        // If the node that the selection points into is written to,
        // Chrome sometimes starts misreporting the selection, so this
        // tracks that and forces a selection reset when our update
        // did write to the node.
        var chromeKludge = result.chrome ? (this.trackWrites = this.root.getSelection().focusNode) : null;
        if (redraw || !this.docView.update(state.doc, outerDeco, innerDeco, this)) {
          this.docView.updateOuterDeco([]);
          this.docView.destroy();
          this.docView = docViewDesc(state.doc, outerDeco, innerDeco, this.dom, this);
        }
        if (chromeKludge && !this.trackWrites) { forceSelUpdate = true; }
      }
      // Work around for an issue where an update arriving right between
      // a DOM selection change and the "selectionchange" event for it
      // can cause a spurious DOM selection update, disrupting mouse
      // drag selection.
      if (forceSelUpdate ||
          !(this.mouseDown && this.domObserver.currentSelection.eq(this.root.getSelection()) && anchorInRightPlace(this))) {
        selectionToDOM(this, forceSelUpdate);
      } else {
        syncNodeSelection(this, state.selection);
        this.domObserver.setCurSelection();
      }
      this.domObserver.start();
    }

    this.updatePluginViews(prev);

    if (scroll == "reset") {
      this.dom.scrollTop = 0;
    } else if (scroll == "to selection") {
      var startDOM = this.root.getSelection().focusNode;
      if (this.someProp("handleScrollToSelection", function (f) { return f(this$1$1); }))
        ; // Handled
      else if (state.selection instanceof NodeSelection)
        { scrollRectIntoView(this, this.docView.domAfterPos(state.selection.from).getBoundingClientRect(), startDOM); }
      else
        { scrollRectIntoView(this, this.coordsAtPos(state.selection.head, 1), startDOM); }
    } else if (oldScrollPos) {
      resetScrollPos(oldScrollPos);
    }
  };

  EditorView.prototype.destroyPluginViews = function destroyPluginViews () {
    var view;
    while (view = this.pluginViews.pop()) { if (view.destroy) { view.destroy(); } }
  };

  EditorView.prototype.updatePluginViews = function updatePluginViews (prevState) {
    if (!prevState || prevState.plugins != this.state.plugins) {
      this.destroyPluginViews();
      for (var i = 0; i < this.state.plugins.length; i++) {
        var plugin = this.state.plugins[i];
        if (plugin.spec.view) { this.pluginViews.push(plugin.spec.view(this)); }
      }
    } else {
      for (var i$1 = 0; i$1 < this.pluginViews.length; i$1++) {
        var pluginView = this.pluginViews[i$1];
        if (pluginView.update) { pluginView.update(this, prevState); }
      }
    }
  };

  // :: (string, ?(prop: *) → *) → *
  // Goes over the values of a prop, first those provided directly,
  // then those from plugins (in order), and calls `f` every time a
  // non-undefined value is found. When `f` returns a truthy value,
  // that is immediately returned. When `f` isn't provided, it is
  // treated as the identity function (the prop value is returned
  // directly).
  EditorView.prototype.someProp = function someProp (propName, f) {
    var prop = this._props && this._props[propName], value;
    if (prop != null && (value = f ? f(prop) : prop)) { return value }
    var plugins = this.state.plugins;
    if (plugins) { for (var i = 0; i < plugins.length; i++) {
      var prop$1 = plugins[i].props[propName];
      if (prop$1 != null && (value = f ? f(prop$1) : prop$1)) { return value }
    } }
  };

  // :: () → bool
  // Query whether the view has focus.
  EditorView.prototype.hasFocus = function hasFocus () {
    return this.root.activeElement == this.dom
  };

  // :: ()
  // Focus the editor.
  EditorView.prototype.focus = function focus () {
    this.domObserver.stop();
    if (this.editable) { focusPreventScroll(this.dom); }
    selectionToDOM(this);
    this.domObserver.start();
  };

  // :: union<dom.Document, dom.DocumentFragment>
  // Get the document root in which the editor exists. This will
  // usually be the top-level `document`, but might be a [shadow
  // DOM](https://developer.mozilla.org/en-US/docs/Web/Web_Components/Shadow_DOM)
  // root if the editor is inside one.
  prototypeAccessors$2$1.root.get = function () {
    var cached = this._root;
    if (cached == null) { for (var search = this.dom.parentNode; search; search = search.parentNode) {
      if (search.nodeType == 9 || (search.nodeType == 11 && search.host)) {
        if (!search.getSelection) { Object.getPrototypeOf(search).getSelection = function () { return document.getSelection(); }; }
        return this._root = search
      }
    } }
    return cached || document
  };

  // :: ({left: number, top: number}) → ?{pos: number, inside: number}
  // Given a pair of viewport coordinates, return the document
  // position that corresponds to them. May return null if the given
  // coordinates aren't inside of the editor. When an object is
  // returned, its `pos` property is the position nearest to the
  // coordinates, and its `inside` property holds the position of the
  // inner node that the position falls inside of, or -1 if it is at
  // the top level, not in any node.
  EditorView.prototype.posAtCoords = function posAtCoords$1 (coords) {
    return posAtCoords(this, coords)
  };

  // :: (number, number) → {left: number, right: number, top: number, bottom: number}
  // Returns the viewport rectangle at a given document position.
  // `left` and `right` will be the same number, as this returns a
  // flat cursor-ish rectangle. If the position is between two things
  // that aren't directly adjacent, `side` determines which element is
  // used. When < 0, the element before the position is used,
  // otherwise the element after.
  EditorView.prototype.coordsAtPos = function coordsAtPos$1 (pos, side) {
      if ( side === void 0 ) { side = 1; }

    return coordsAtPos(this, pos, side)
  };

  // :: (number, number) → {node: dom.Node, offset: number}
  // Find the DOM position that corresponds to the given document
  // position. When `side` is negative, find the position as close as
  // possible to the content before the position. When positive,
  // prefer positions close to the content after the position. When
  // zero, prefer as shallow a position as possible.
  //
  // Note that you should **not** mutate the editor's internal DOM,
  // only inspect it (and even that is usually not necessary).
  EditorView.prototype.domAtPos = function domAtPos (pos, side) {
      if ( side === void 0 ) { side = 0; }

    return this.docView.domFromPos(pos, side)
  };

  // :: (number) → ?dom.Node
  // Find the DOM node that represents the document node after the
  // given position. May return `null` when the position doesn't point
  // in front of a node or if the node is inside an opaque node view.
  //
  // This is intended to be able to call things like
  // `getBoundingClientRect` on that DOM node. Do **not** mutate the
  // editor DOM directly, or add styling this way, since that will be
  // immediately overriden by the editor as it redraws the node.
  EditorView.prototype.nodeDOM = function nodeDOM (pos) {
    var desc = this.docView.descAt(pos);
    return desc ? desc.nodeDOM : null
  };

  // :: (dom.Node, number, ?number) → number
  // Find the document position that corresponds to a given DOM
  // position. (Whenever possible, it is preferable to inspect the
  // document structure directly, rather than poking around in the
  // DOM, but sometimes—for example when interpreting an event
  // target—you don't have a choice.)
  //
  // The `bias` parameter can be used to influence which side of a DOM
  // node to use when the position is inside a leaf node.
  EditorView.prototype.posAtDOM = function posAtDOM (node, offset, bias) {
      if ( bias === void 0 ) { bias = -1; }

    var pos = this.docView.posFromDOM(node, offset, bias);
    if (pos == null) { throw new RangeError("DOM position not inside the editor") }
    return pos
  };

  // :: (union<"up", "down", "left", "right", "forward", "backward">, ?EditorState) → bool
  // Find out whether the selection is at the end of a textblock when
  // moving in a given direction. When, for example, given `"left"`,
  // it will return true if moving left from the current cursor
  // position would leave that position's parent textblock. Will apply
  // to the view's current state by default, but it is possible to
  // pass a different state.
  EditorView.prototype.endOfTextblock = function endOfTextblock$1 (dir, state) {
    return endOfTextblock(this, state || this.state, dir)
  };

  // :: ()
  // Removes the editor from the DOM and destroys all [node
  // views](#view.NodeView).
  EditorView.prototype.destroy = function destroy () {
    if (!this.docView) { return }
    destroyInput(this);
    this.destroyPluginViews();
    if (this.mounted) {
      this.docView.update(this.state.doc, [], viewDecorations(this), this);
      this.dom.textContent = "";
    } else if (this.dom.parentNode) {
      this.dom.parentNode.removeChild(this.dom);
    }
    this.docView.destroy();
    this.docView = null;
  };

  // Used for testing.
  EditorView.prototype.dispatchEvent = function dispatchEvent$1 (event) {
    return dispatchEvent(this, event)
  };

  // :: (Transaction)
  // Dispatch a transaction. Will call
  // [`dispatchTransaction`](#view.DirectEditorProps.dispatchTransaction)
  // when given, and otherwise defaults to applying the transaction to
  // the current state and calling
  // [`updateState`](#view.EditorView.updateState) with the result.
  // This method is bound to the view instance, so that it can be
  // easily passed around.
  EditorView.prototype.dispatch = function dispatch (tr) {
    var dispatchTransaction = this._props.dispatchTransaction;
    if (dispatchTransaction) { dispatchTransaction.call(this, tr); }
    else { this.updateState(this.state.apply(tr)); }
  };

  Object.defineProperties( EditorView.prototype, prototypeAccessors$2$1 );

  function computeDocDeco(view) {
    var attrs = Object.create(null);
    attrs.class = "ProseMirror";
    attrs.contenteditable = String(view.editable);
    attrs.translate = "no";

    view.someProp("attributes", function (value) {
      if (typeof value == "function") { value = value(view.state); }
      if (value) { for (var attr in value) {
        if (attr == "class")
          { attrs.class += " " + value[attr]; }
        else if (!attrs[attr] && attr != "contenteditable" && attr != "nodeName")
          { attrs[attr] = String(value[attr]); }
      } }
    });

    return [Decoration.node(0, view.state.doc.content.size, attrs)]
  }

  function updateCursorWrapper(view) {
    if (view.markCursor) {
      var dom = document.createElement("img");
      dom.className = "ProseMirror-separator";
      dom.setAttribute("mark-placeholder", "true");
      view.cursorWrapper = {dom: dom, deco: Decoration.widget(view.state.selection.head, dom, {raw: true, marks: view.markCursor})};
    } else {
      view.cursorWrapper = null;
    }
  }

  function getEditable(view) {
    return !view.someProp("editable", function (value) { return value(view.state) === false; })
  }

  function selectionContextChanged(sel1, sel2) {
    var depth = Math.min(sel1.$anchor.sharedDepth(sel1.head), sel2.$anchor.sharedDepth(sel2.head));
    return sel1.$anchor.start(depth) != sel2.$anchor.start(depth)
  }

  function buildNodeViews(view) {
    var result = {};
    view.someProp("nodeViews", function (obj) {
      for (var prop in obj) { if (!Object.prototype.hasOwnProperty.call(result, prop))
        { result[prop] = obj[prop]; } }
    });
    return result
  }

  function changedNodeViews(a, b) {
    var nA = 0, nB = 0;
    for (var prop in a) {
      if (a[prop] != b[prop]) { return true }
      nA++;
    }
    for (var _ in b) { nB++; }
    return nA != nB
  }

  var view = /*#__PURE__*/Object.freeze({
    __proto__: null,
    Decoration: Decoration,
    DecorationSet: DecorationSet,
    EditorView: EditorView,
    __endComposition: endComposition,
    __parseFromClipboard: parseFromClipboard,
    __serializeForClipboard: serializeForClipboard
  });

  // ::- Input rules are regular expressions describing a piece of text
  // that, when typed, causes something to happen. This might be
  // changing two dashes into an emdash, wrapping a paragraph starting
  // with `"> "` into a blockquote, or something entirely different.
  var InputRule = function InputRule(match, handler) {
    this.match = match;
    this.handler = typeof handler == "string" ? stringHandler(handler) : handler;
  };

  function stringHandler(string) {
    return function(state, match, start, end) {
      var insert = string;
      if (match[1]) {
        var offset = match[0].lastIndexOf(match[1]);
        insert += match[0].slice(offset + match[1].length);
        start += offset;
        var cutOff = start - end;
        if (cutOff > 0) {
          insert = match[0].slice(offset - cutOff, offset) + insert;
          start = end;
        }
      }
      return state.tr.insertText(insert, start, end)
    }
  }

  var MAX_MATCH = 500;

  // :: (config: {rules: [InputRule]}) → Plugin
  // Create an input rules plugin. When enabled, it will cause text
  // input that matches any of the given rules to trigger the rule's
  // action.
  function inputRules(ref) {
    var rules = ref.rules;

    var plugin = new Plugin({
      state: {
        init: function init() { return null },
        apply: function apply(tr, prev) {
          var stored = tr.getMeta(this);
          if (stored) { return stored }
          return tr.selectionSet || tr.docChanged ? null : prev
        }
      },

      props: {
        handleTextInput: function handleTextInput(view, from, to, text) {
          return run(view, from, to, text, rules, plugin)
        },
        handleDOMEvents: {
          compositionend: function (view) {
            setTimeout(function () {
              var ref = view.state.selection;
              var $cursor = ref.$cursor;
              if ($cursor) { run(view, $cursor.pos, $cursor.pos, "", rules, plugin); }
            });
          }
        }
      },

      isInputRules: true
    });
    return plugin
  }

  function run(view, from, to, text, rules, plugin) {
    if (view.composing) { return false }
    var state = view.state, $from = state.doc.resolve(from);
    if ($from.parent.type.spec.code) { return false }
    var textBefore = $from.parent.textBetween(Math.max(0, $from.parentOffset - MAX_MATCH), $from.parentOffset,
                                              null, "\ufffc") + text;
    for (var i = 0; i < rules.length; i++) {
      var match = rules[i].match.exec(textBefore);
      var tr = match && rules[i].handler(state, match, from - (match[0].length - text.length), to);
      if (!tr) { continue }
      view.dispatch(tr.setMeta(plugin, {transform: tr, from: from, to: to, text: text}));
      return true
    }
    return false
  }

  // :: (EditorState, ?(Transaction)) → bool
  // This is a command that will undo an input rule, if applying such a
  // rule was the last thing that the user did.
  function undoInputRule(state, dispatch) {
    var plugins = state.plugins;
    for (var i = 0; i < plugins.length; i++) {
      var plugin = plugins[i], undoable = (void 0);
      if (plugin.spec.isInputRules && (undoable = plugin.getState(state))) {
        if (dispatch) {
          var tr = state.tr, toUndo = undoable.transform;
          for (var j = toUndo.steps.length - 1; j >= 0; j--)
            { tr.step(toUndo.steps[j].invert(toUndo.docs[j])); }
          if (undoable.text) {
            var marks = tr.doc.resolve(undoable.from).marks();
            tr.replaceWith(undoable.from, undoable.to, state.schema.text(undoable.text, marks));
          } else {
            tr.delete(undoable.from, undoable.to);
          }
          dispatch(tr);
        }
        return true
      }
    }
    return false
  }

  // :: InputRule Converts double dashes to an emdash.
  var emDash = new InputRule(/--$/, "—");
  // :: InputRule Converts three dots to an ellipsis character.
  var ellipsis = new InputRule(/\.\.\.$/, "…");
  // :: InputRule “Smart” opening double quotes.
  var openDoubleQuote = new InputRule(/(?:^|[\s\{\[\(\<'"\u2018\u201C])(")$/, "“");
  // :: InputRule “Smart” closing double quotes.
  var closeDoubleQuote = new InputRule(/"$/, "”");
  // :: InputRule “Smart” opening single quotes.
  var openSingleQuote = new InputRule(/(?:^|[\s\{\[\(\<'"\u2018\u201C])(')$/, "‘");
  // :: InputRule “Smart” closing single quotes.
  var closeSingleQuote = new InputRule(/'$/, "’");

  // :: [InputRule] Smart-quote related input rules.
  var smartQuotes = [openDoubleQuote, closeDoubleQuote, openSingleQuote, closeSingleQuote];

  // :: (RegExp, NodeType, ?union<Object, ([string]) → ?Object>, ?([string], Node) → bool) → InputRule
  // Build an input rule for automatically wrapping a textblock when a
  // given string is typed. The `regexp` argument is
  // directly passed through to the `InputRule` constructor. You'll
  // probably want the regexp to start with `^`, so that the pattern can
  // only occur at the start of a textblock.
  //
  // `nodeType` is the type of node to wrap in. If it needs attributes,
  // you can either pass them directly, or pass a function that will
  // compute them from the regular expression match.
  //
  // By default, if there's a node with the same type above the newly
  // wrapped node, the rule will try to [join](#transform.Transform.join) those
  // two nodes. You can pass a join predicate, which takes a regular
  // expression match and the node before the wrapped node, and can
  // return a boolean to indicate whether a join should happen.
  function wrappingInputRule(regexp, nodeType, getAttrs, joinPredicate) {
    return new InputRule(regexp, function (state, match, start, end) {
      var attrs = getAttrs instanceof Function ? getAttrs(match) : getAttrs;
      var tr = state.tr.delete(start, end);
      var $start = tr.doc.resolve(start), range = $start.blockRange(), wrapping = range && findWrapping(range, nodeType, attrs);
      if (!wrapping) { return null }
      tr.wrap(range, wrapping);
      var before = tr.doc.resolve(start - 1).nodeBefore;
      if (before && before.type == nodeType && canJoin(tr.doc, start - 1) &&
          (!joinPredicate || joinPredicate(match, before)))
        { tr.join(start - 1); }
      return tr
    })
  }

  // :: (RegExp, NodeType, ?union<Object, ([string]) → ?Object>) → InputRule
  // Build an input rule that changes the type of a textblock when the
  // matched text is typed into it. You'll usually want to start your
  // regexp with `^` to that it is only matched at the start of a
  // textblock. The optional `getAttrs` parameter can be used to compute
  // the new node's attributes, and works the same as in the
  // `wrappingInputRule` function.
  function textblockTypeInputRule(regexp, nodeType, getAttrs) {
    return new InputRule(regexp, function (state, match, start, end) {
      var $start = state.doc.resolve(start);
      var attrs = getAttrs instanceof Function ? getAttrs(match) : getAttrs;
      if (!$start.node(-1).canReplaceWith($start.index(-1), $start.indexAfter(-1), nodeType)) { return null }
      return state.tr
        .delete(start, end)
        .setBlockType(start, start, nodeType, attrs)
    })
  }

  var inputRules$1 = /*#__PURE__*/Object.freeze({
    __proto__: null,
    InputRule: InputRule,
    closeDoubleQuote: closeDoubleQuote,
    closeSingleQuote: closeSingleQuote,
    ellipsis: ellipsis,
    emDash: emDash,
    inputRules: inputRules,
    openDoubleQuote: openDoubleQuote,
    openSingleQuote: openSingleQuote,
    smartQuotes: smartQuotes,
    textblockTypeInputRule: textblockTypeInputRule,
    undoInputRule: undoInputRule,
    wrappingInputRule: wrappingInputRule
  });

  // :: (EditorState, ?(tr: Transaction)) → bool
  // Delete the selection, if there is one.
  function deleteSelection(state, dispatch) {
    if (state.selection.empty) { return false }
    if (dispatch) { dispatch(state.tr.deleteSelection().scrollIntoView()); }
    return true
  }

  // :: (EditorState, ?(tr: Transaction), ?EditorView) → bool
  // If the selection is empty and at the start of a textblock, try to
  // reduce the distance between that block and the one before it—if
  // there's a block directly before it that can be joined, join them.
  // If not, try to move the selected block closer to the next one in
  // the document structure by lifting it out of its parent or moving it
  // into a parent of the previous block. Will use the view for accurate
  // (bidi-aware) start-of-textblock detection if given.
  function joinBackward(state, dispatch, view) {
    var ref = state.selection;
    var $cursor = ref.$cursor;
    if (!$cursor || (view ? !view.endOfTextblock("backward", state)
                          : $cursor.parentOffset > 0))
      { return false }

    var $cut = findCutBefore($cursor);

    // If there is no node before this, try to lift
    if (!$cut) {
      var range = $cursor.blockRange(), target = range && liftTarget(range);
      if (target == null) { return false }
      if (dispatch) { dispatch(state.tr.lift(range, target).scrollIntoView()); }
      return true
    }

    var before = $cut.nodeBefore;
    // Apply the joining algorithm
    if (!before.type.spec.isolating && deleteBarrier(state, $cut, dispatch))
      { return true }

    // If the node below has no content and the node above is
    // selectable, delete the node below and select the one above.
    if ($cursor.parent.content.size == 0 &&
        (textblockAt(before, "end") || NodeSelection.isSelectable(before))) {
      if (dispatch) {
        var tr = state.tr.deleteRange($cursor.before(), $cursor.after());
        tr.setSelection(textblockAt(before, "end") ? Selection.findFrom(tr.doc.resolve(tr.mapping.map($cut.pos, -1)), -1)
                        : NodeSelection.create(tr.doc, $cut.pos - before.nodeSize));
        dispatch(tr.scrollIntoView());
      }
      return true
    }

    // If the node before is an atom, delete it
    if (before.isAtom && $cut.depth == $cursor.depth - 1) {
      if (dispatch) { dispatch(state.tr.delete($cut.pos - before.nodeSize, $cut.pos).scrollIntoView()); }
      return true
    }

    return false
  }

  function textblockAt(node, side, only) {
    for (; node; node = (side == "start" ? node.firstChild : node.lastChild)) {
      if (node.isTextblock) { return true }
      if (only && node.childCount != 1) { return false }
    }
    return false
  }

  // :: (EditorState, ?(tr: Transaction), ?EditorView) → bool
  // When the selection is empty and at the start of a textblock, select
  // the node before that textblock, if possible. This is intended to be
  // bound to keys like backspace, after
  // [`joinBackward`](#commands.joinBackward) or other deleting
  // commands, as a fall-back behavior when the schema doesn't allow
  // deletion at the selected point.
  function selectNodeBackward(state, dispatch, view) {
    var ref = state.selection;
    var $head = ref.$head;
    var empty = ref.empty;
    var $cut = $head;
    if (!empty) { return false }

    if ($head.parent.isTextblock) {
      if (view ? !view.endOfTextblock("backward", state) : $head.parentOffset > 0) { return false }
      $cut = findCutBefore($head);
    }
    var node = $cut && $cut.nodeBefore;
    if (!node || !NodeSelection.isSelectable(node)) { return false }
    if (dispatch)
      { dispatch(state.tr.setSelection(NodeSelection.create(state.doc, $cut.pos - node.nodeSize)).scrollIntoView()); }
    return true
  }

  function findCutBefore($pos) {
    if (!$pos.parent.type.spec.isolating) { for (var i = $pos.depth - 1; i >= 0; i--) {
      if ($pos.index(i) > 0) { return $pos.doc.resolve($pos.before(i + 1)) }
      if ($pos.node(i).type.spec.isolating) { break }
    } }
    return null
  }

  // :: (EditorState, ?(tr: Transaction), ?EditorView) → bool
  // If the selection is empty and the cursor is at the end of a
  // textblock, try to reduce or remove the boundary between that block
  // and the one after it, either by joining them or by moving the other
  // block closer to this one in the tree structure. Will use the view
  // for accurate start-of-textblock detection if given.
  function joinForward(state, dispatch, view) {
    var ref = state.selection;
    var $cursor = ref.$cursor;
    if (!$cursor || (view ? !view.endOfTextblock("forward", state)
                          : $cursor.parentOffset < $cursor.parent.content.size))
      { return false }

    var $cut = findCutAfter($cursor);

    // If there is no node after this, there's nothing to do
    if (!$cut) { return false }

    var after = $cut.nodeAfter;
    // Try the joining algorithm
    if (deleteBarrier(state, $cut, dispatch)) { return true }

    // If the node above has no content and the node below is
    // selectable, delete the node above and select the one below.
    if ($cursor.parent.content.size == 0 &&
        (textblockAt(after, "start") || NodeSelection.isSelectable(after))) {
      if (dispatch) {
        var tr = state.tr.deleteRange($cursor.before(), $cursor.after());
        tr.setSelection(textblockAt(after, "start") ? Selection.findFrom(tr.doc.resolve(tr.mapping.map($cut.pos)), 1)
                        : NodeSelection.create(tr.doc, tr.mapping.map($cut.pos)));
        dispatch(tr.scrollIntoView());
      }
      return true
    }

    // If the next node is an atom, delete it
    if (after.isAtom && $cut.depth == $cursor.depth - 1) {
      if (dispatch) { dispatch(state.tr.delete($cut.pos, $cut.pos + after.nodeSize).scrollIntoView()); }
      return true
    }

    return false
  }

  // :: (EditorState, ?(tr: Transaction), ?EditorView) → bool
  // When the selection is empty and at the end of a textblock, select
  // the node coming after that textblock, if possible. This is intended
  // to be bound to keys like delete, after
  // [`joinForward`](#commands.joinForward) and similar deleting
  // commands, to provide a fall-back behavior when the schema doesn't
  // allow deletion at the selected point.
  function selectNodeForward(state, dispatch, view) {
    var ref = state.selection;
    var $head = ref.$head;
    var empty = ref.empty;
    var $cut = $head;
    if (!empty) { return false }
    if ($head.parent.isTextblock) {
      if (view ? !view.endOfTextblock("forward", state) : $head.parentOffset < $head.parent.content.size)
        { return false }
      $cut = findCutAfter($head);
    }
    var node = $cut && $cut.nodeAfter;
    if (!node || !NodeSelection.isSelectable(node)) { return false }
    if (dispatch)
      { dispatch(state.tr.setSelection(NodeSelection.create(state.doc, $cut.pos)).scrollIntoView()); }
    return true
  }

  function findCutAfter($pos) {
    if (!$pos.parent.type.spec.isolating) { for (var i = $pos.depth - 1; i >= 0; i--) {
      var parent = $pos.node(i);
      if ($pos.index(i) + 1 < parent.childCount) { return $pos.doc.resolve($pos.after(i + 1)) }
      if (parent.type.spec.isolating) { break }
    } }
    return null
  }

  // :: (EditorState, ?(tr: Transaction)) → bool
  // Join the selected block or, if there is a text selection, the
  // closest ancestor block of the selection that can be joined, with
  // the sibling above it.
  function joinUp(state, dispatch) {
    var sel = state.selection, nodeSel = sel instanceof NodeSelection, point;
    if (nodeSel) {
      if (sel.node.isTextblock || !canJoin(state.doc, sel.from)) { return false }
      point = sel.from;
    } else {
      point = joinPoint(state.doc, sel.from, -1);
      if (point == null) { return false }
    }
    if (dispatch) {
      var tr = state.tr.join(point);
      if (nodeSel) { tr.setSelection(NodeSelection.create(tr.doc, point - state.doc.resolve(point).nodeBefore.nodeSize)); }
      dispatch(tr.scrollIntoView());
    }
    return true
  }

  // :: (EditorState, ?(tr: Transaction)) → bool
  // Join the selected block, or the closest ancestor of the selection
  // that can be joined, with the sibling after it.
  function joinDown(state, dispatch) {
    var sel = state.selection, point;
    if (sel instanceof NodeSelection) {
      if (sel.node.isTextblock || !canJoin(state.doc, sel.to)) { return false }
      point = sel.to;
    } else {
      point = joinPoint(state.doc, sel.to, 1);
      if (point == null) { return false }
    }
    if (dispatch)
      { dispatch(state.tr.join(point).scrollIntoView()); }
    return true
  }

  // :: (EditorState, ?(tr: Transaction)) → bool
  // Lift the selected block, or the closest ancestor block of the
  // selection that can be lifted, out of its parent node.
  function lift$1(state, dispatch) {
    var ref = state.selection;
    var $from = ref.$from;
    var $to = ref.$to;
    var range = $from.blockRange($to), target = range && liftTarget(range);
    if (target == null) { return false }
    if (dispatch) { dispatch(state.tr.lift(range, target).scrollIntoView()); }
    return true
  }

  // :: (EditorState, ?(tr: Transaction)) → bool
  // If the selection is in a node whose type has a truthy
  // [`code`](#model.NodeSpec.code) property in its spec, replace the
  // selection with a newline character.
  function newlineInCode(state, dispatch) {
    var ref = state.selection;
    var $head = ref.$head;
    var $anchor = ref.$anchor;
    if (!$head.parent.type.spec.code || !$head.sameParent($anchor)) { return false }
    if (dispatch) { dispatch(state.tr.insertText("\n").scrollIntoView()); }
    return true
  }

  function defaultBlockAt(match) {
    for (var i = 0; i < match.edgeCount; i++) {
      var ref = match.edge(i);
      var type = ref.type;
      if (type.isTextblock && !type.hasRequiredAttrs()) { return type }
    }
    return null
  }

  // :: (EditorState, ?(tr: Transaction)) → bool
  // When the selection is in a node with a truthy
  // [`code`](#model.NodeSpec.code) property in its spec, create a
  // default block after the code block, and move the cursor there.
  function exitCode(state, dispatch) {
    var ref = state.selection;
    var $head = ref.$head;
    var $anchor = ref.$anchor;
    if (!$head.parent.type.spec.code || !$head.sameParent($anchor)) { return false }
    var above = $head.node(-1), after = $head.indexAfter(-1), type = defaultBlockAt(above.contentMatchAt(after));
    if (!above.canReplaceWith(after, after, type)) { return false }
    if (dispatch) {
      var pos = $head.after(), tr = state.tr.replaceWith(pos, pos, type.createAndFill());
      tr.setSelection(Selection.near(tr.doc.resolve(pos), 1));
      dispatch(tr.scrollIntoView());
    }
    return true
  }

  // :: (EditorState, ?(tr: Transaction)) → bool
  // If a block node is selected, create an empty paragraph before (if
  // it is its parent's first child) or after it.
  function createParagraphNear(state, dispatch) {
    var sel = state.selection;
    var $from = sel.$from;
    var $to = sel.$to;
    if (sel instanceof AllSelection || $from.parent.inlineContent || $to.parent.inlineContent) { return false }
    var type = defaultBlockAt($to.parent.contentMatchAt($to.indexAfter()));
    if (!type || !type.isTextblock) { return false }
    if (dispatch) {
      var side = (!$from.parentOffset && $to.index() < $to.parent.childCount ? $from : $to).pos;
      var tr = state.tr.insert(side, type.createAndFill());
      tr.setSelection(TextSelection.create(tr.doc, side + 1));
      dispatch(tr.scrollIntoView());
    }
    return true
  }

  // :: (EditorState, ?(tr: Transaction)) → bool
  // If the cursor is in an empty textblock that can be lifted, lift the
  // block.
  function liftEmptyBlock(state, dispatch) {
    var ref = state.selection;
    var $cursor = ref.$cursor;
    if (!$cursor || $cursor.parent.content.size) { return false }
    if ($cursor.depth > 1 && $cursor.after() != $cursor.end(-1)) {
      var before = $cursor.before();
      if (canSplit(state.doc, before)) {
        if (dispatch) { dispatch(state.tr.split(before).scrollIntoView()); }
        return true
      }
    }
    var range = $cursor.blockRange(), target = range && liftTarget(range);
    if (target == null) { return false }
    if (dispatch) { dispatch(state.tr.lift(range, target).scrollIntoView()); }
    return true
  }

  // :: (EditorState, ?(tr: Transaction)) → bool
  // Split the parent block of the selection. If the selection is a text
  // selection, also delete its content.
  function splitBlock(state, dispatch) {
    var ref = state.selection;
    var $from = ref.$from;
    var $to = ref.$to;
    if (state.selection instanceof NodeSelection && state.selection.node.isBlock) {
      if (!$from.parentOffset || !canSplit(state.doc, $from.pos)) { return false }
      if (dispatch) { dispatch(state.tr.split($from.pos).scrollIntoView()); }
      return true
    }

    if (!$from.parent.isBlock) { return false }

    if (dispatch) {
      var atEnd = $to.parentOffset == $to.parent.content.size;
      var tr = state.tr;
      if (state.selection instanceof TextSelection || state.selection instanceof AllSelection) { tr.deleteSelection(); }
      var deflt = $from.depth == 0 ? null : defaultBlockAt($from.node(-1).contentMatchAt($from.indexAfter(-1)));
      var types = atEnd && deflt ? [{type: deflt}] : null;
      var can = canSplit(tr.doc, tr.mapping.map($from.pos), 1, types);
      if (!types && !can && canSplit(tr.doc, tr.mapping.map($from.pos), 1, deflt && [{type: deflt}])) {
        types = [{type: deflt}];
        can = true;
      }
      if (can) {
        tr.split(tr.mapping.map($from.pos), 1, types);
        if (!atEnd && !$from.parentOffset && $from.parent.type != deflt) {
          var first = tr.mapping.map($from.before()), $first = tr.doc.resolve(first);
          if ($from.node(-1).canReplaceWith($first.index(), $first.index() + 1, deflt))
            { tr.setNodeMarkup(tr.mapping.map($from.before()), deflt); }
        }
      }
      dispatch(tr.scrollIntoView());
    }
    return true
  }

  // :: (EditorState, ?(tr: Transaction)) → bool
  // Acts like [`splitBlock`](#commands.splitBlock), but without
  // resetting the set of active marks at the cursor.
  function splitBlockKeepMarks(state, dispatch) {
    return splitBlock(state, dispatch && (function (tr) {
      var marks = state.storedMarks || (state.selection.$to.parentOffset && state.selection.$from.marks());
      if (marks) { tr.ensureMarks(marks); }
      dispatch(tr);
    }))
  }

  // :: (EditorState, ?(tr: Transaction)) → bool
  // Move the selection to the node wrapping the current selection, if
  // any. (Will not select the document node.)
  function selectParentNode(state, dispatch) {
    var ref = state.selection;
    var $from = ref.$from;
    var to = ref.to;
    var pos;
    var same = $from.sharedDepth(to);
    if (same == 0) { return false }
    pos = $from.before(same);
    if (dispatch) { dispatch(state.tr.setSelection(NodeSelection.create(state.doc, pos))); }
    return true
  }

  // :: (EditorState, ?(tr: Transaction)) → bool
  // Select the whole document.
  function selectAll(state, dispatch) {
    if (dispatch) { dispatch(state.tr.setSelection(new AllSelection(state.doc))); }
    return true
  }

  function joinMaybeClear(state, $pos, dispatch) {
    var before = $pos.nodeBefore, after = $pos.nodeAfter, index = $pos.index();
    if (!before || !after || !before.type.compatibleContent(after.type)) { return false }
    if (!before.content.size && $pos.parent.canReplace(index - 1, index)) {
      if (dispatch) { dispatch(state.tr.delete($pos.pos - before.nodeSize, $pos.pos).scrollIntoView()); }
      return true
    }
    if (!$pos.parent.canReplace(index, index + 1) || !(after.isTextblock || canJoin(state.doc, $pos.pos)))
      { return false }
    if (dispatch)
      { dispatch(state.tr
               .clearIncompatible($pos.pos, before.type, before.contentMatchAt(before.childCount))
               .join($pos.pos)
               .scrollIntoView()); }
    return true
  }

  function deleteBarrier(state, $cut, dispatch) {
    var before = $cut.nodeBefore, after = $cut.nodeAfter, conn, match;
    if (before.type.spec.isolating || after.type.spec.isolating) { return false }
    if (joinMaybeClear(state, $cut, dispatch)) { return true }

    var canDelAfter = $cut.parent.canReplace($cut.index(), $cut.index() + 1);
    if (canDelAfter &&
        (conn = (match = before.contentMatchAt(before.childCount)).findWrapping(after.type)) &&
        match.matchType(conn[0] || after.type).validEnd) {
      if (dispatch) {
        var end = $cut.pos + after.nodeSize, wrap = Fragment$1.empty;
        for (var i = conn.length - 1; i >= 0; i--)
          { wrap = Fragment$1.from(conn[i].create(null, wrap)); }
        wrap = Fragment$1.from(before.copy(wrap));
        var tr = state.tr.step(new ReplaceAroundStep($cut.pos - 1, end, $cut.pos, end, new Slice$1(wrap, 1, 0), conn.length, true));
        var joinAt = end + 2 * conn.length;
        if (canJoin(tr.doc, joinAt)) { tr.join(joinAt); }
        dispatch(tr.scrollIntoView());
      }
      return true
    }

    var selAfter = Selection.findFrom($cut, 1);
    var range = selAfter && selAfter.$from.blockRange(selAfter.$to), target = range && liftTarget(range);
    if (target != null && target >= $cut.depth) {
      if (dispatch) { dispatch(state.tr.lift(range, target).scrollIntoView()); }
      return true
    }

    if (canDelAfter && textblockAt(after, "start", true) && textblockAt(before, "end")) {
      var at = before, wrap$1 = [];
      for (;;) {
        wrap$1.push(at);
        if (at.isTextblock) { break }
        at = at.lastChild;
      }
      var afterText = after, afterDepth = 1;
      for (; !afterText.isTextblock; afterText = afterText.firstChild) { afterDepth++; }
      if (at.canReplace(at.childCount, at.childCount, afterText.content)) {
        if (dispatch) {
          var end$1 = Fragment$1.empty;
          for (var i$1 = wrap$1.length - 1; i$1 >= 0; i$1--) { end$1 = Fragment$1.from(wrap$1[i$1].copy(end$1)); }
          var tr$1 = state.tr.step(new ReplaceAroundStep($cut.pos - wrap$1.length, $cut.pos + after.nodeSize,
                                                       $cut.pos + afterDepth, $cut.pos + after.nodeSize - afterDepth,
                                                       new Slice$1(end$1, wrap$1.length, 0), 0, true));
          dispatch(tr$1.scrollIntoView());
        }
        return true
      }
    }

    return false
  }

  // Parameterized commands

  // :: (NodeType, ?Object) → (state: EditorState, dispatch: ?(tr: Transaction)) → bool
  // Wrap the selection in a node of the given type with the given
  // attributes.
  function wrapIn(nodeType, attrs) {
    return function(state, dispatch) {
      var ref = state.selection;
      var $from = ref.$from;
      var $to = ref.$to;
      var range = $from.blockRange($to), wrapping = range && findWrapping(range, nodeType, attrs);
      if (!wrapping) { return false }
      if (dispatch) { dispatch(state.tr.wrap(range, wrapping).scrollIntoView()); }
      return true
    }
  }

  // :: (NodeType, ?Object) → (state: EditorState, dispatch: ?(tr: Transaction)) → bool
  // Returns a command that tries to set the selected textblocks to the
  // given node type with the given attributes.
  function setBlockType(nodeType, attrs) {
    return function(state, dispatch) {
      var ref = state.selection;
      var from = ref.from;
      var to = ref.to;
      var applicable = false;
      state.doc.nodesBetween(from, to, function (node, pos) {
        if (applicable) { return false }
        if (!node.isTextblock || node.hasMarkup(nodeType, attrs)) { return }
        if (node.type == nodeType) {
          applicable = true;
        } else {
          var $pos = state.doc.resolve(pos), index = $pos.index();
          applicable = $pos.parent.canReplaceWith(index, index + 1, nodeType);
        }
      });
      if (!applicable) { return false }
      if (dispatch) { dispatch(state.tr.setBlockType(from, to, nodeType, attrs).scrollIntoView()); }
      return true
    }
  }

  function markApplies$1(doc, ranges, type) {
    var loop = function ( i ) {
      var ref = ranges[i];
      var $from = ref.$from;
      var $to = ref.$to;
      var can = $from.depth == 0 ? doc.type.allowsMarkType(type) : false;
      doc.nodesBetween($from.pos, $to.pos, function (node) {
        if (can) { return false }
        can = node.inlineContent && node.type.allowsMarkType(type);
      });
      if (can) { return { v: true } }
    };

    for (var i = 0; i < ranges.length; i++) {
      var returned = loop( i );

      if ( returned ) { return returned.v; }
    }
    return false
  }

  // :: (MarkType, ?Object) → (state: EditorState, dispatch: ?(tr: Transaction)) → bool
  // Create a command function that toggles the given mark with the
  // given attributes. Will return `false` when the current selection
  // doesn't support that mark. This will remove the mark if any marks
  // of that type exist in the selection, or add it otherwise. If the
  // selection is empty, this applies to the [stored
  // marks](#state.EditorState.storedMarks) instead of a range of the
  // document.
  function toggleMark(markType, attrs) {
    return function(state, dispatch) {
      var ref = state.selection;
      var empty = ref.empty;
      var $cursor = ref.$cursor;
      var ranges = ref.ranges;
      if ((empty && !$cursor) || !markApplies$1(state.doc, ranges, markType)) { return false }
      if (dispatch) {
        if ($cursor) {
          if (markType.isInSet(state.storedMarks || $cursor.marks()))
            { dispatch(state.tr.removeStoredMark(markType)); }
          else
            { dispatch(state.tr.addStoredMark(markType.create(attrs))); }
        } else {
          var has = false, tr = state.tr;
          for (var i = 0; !has && i < ranges.length; i++) {
            var ref$1 = ranges[i];
            var $from = ref$1.$from;
            var $to = ref$1.$to;
            has = state.doc.rangeHasMark($from.pos, $to.pos, markType);
          }
          for (var i$1 = 0; i$1 < ranges.length; i$1++) {
            var ref$2 = ranges[i$1];
            var $from$1 = ref$2.$from;
            var $to$1 = ref$2.$to;
            if (has) {
              tr.removeMark($from$1.pos, $to$1.pos, markType);
            } else {
              var from = $from$1.pos, to = $to$1.pos, start = $from$1.nodeAfter, end = $to$1.nodeBefore;
              var spaceStart = start && start.isText ? /^\s*/.exec(start.text)[0].length : 0;
              var spaceEnd = end && end.isText ? /\s*$/.exec(end.text)[0].length : 0;
              if (from + spaceStart < to) { from += spaceStart; to -= spaceEnd; }
              tr.addMark(from, to, markType.create(attrs));
            }
          }
          dispatch(tr.scrollIntoView());
        }
      }
      return true
    }
  }

  function wrapDispatchForJoin(dispatch, isJoinable) {
    return function (tr) {
      if (!tr.isGeneric) { return dispatch(tr) }

      var ranges = [];
      for (var i = 0; i < tr.mapping.maps.length; i++) {
        var map = tr.mapping.maps[i];
        for (var j = 0; j < ranges.length; j++)
          { ranges[j] = map.map(ranges[j]); }
        map.forEach(function (_s, _e, from, to) { return ranges.push(from, to); });
      }

      // Figure out which joinable points exist inside those ranges,
      // by checking all node boundaries in their parent nodes.
      var joinable = [];
      for (var i$1 = 0; i$1 < ranges.length; i$1 += 2) {
        var from = ranges[i$1], to = ranges[i$1 + 1];
        var $from = tr.doc.resolve(from), depth = $from.sharedDepth(to), parent = $from.node(depth);
        for (var index = $from.indexAfter(depth), pos = $from.after(depth + 1); pos <= to; ++index) {
          var after = parent.maybeChild(index);
          if (!after) { break }
          if (index && joinable.indexOf(pos) == -1) {
            var before = parent.child(index - 1);
            if (before.type == after.type && isJoinable(before, after))
              { joinable.push(pos); }
          }
          pos += after.nodeSize;
        }
      }
      // Join the joinable points
      joinable.sort(function (a, b) { return a - b; });
      for (var i$2 = joinable.length - 1; i$2 >= 0; i$2--) {
        if (canJoin(tr.doc, joinable[i$2])) { tr.join(joinable[i$2]); }
      }
      dispatch(tr);
    }
  }

  // :: ((state: EditorState, ?(tr: Transaction)) → bool, union<(before: Node, after: Node) → bool, [string]>) → (state: EditorState, ?(tr: Transaction)) → bool
  // Wrap a command so that, when it produces a transform that causes
  // two joinable nodes to end up next to each other, those are joined.
  // Nodes are considered joinable when they are of the same type and
  // when the `isJoinable` predicate returns true for them or, if an
  // array of strings was passed, if their node type name is in that
  // array.
  function autoJoin(command, isJoinable) {
    if (Array.isArray(isJoinable)) {
      var types = isJoinable;
      isJoinable = function (node) { return types.indexOf(node.type.name) > -1; };
    }
    return function (state, dispatch) { return command(state, dispatch && wrapDispatchForJoin(dispatch, isJoinable)); }
  }

  // :: (...[(EditorState, ?(tr: Transaction), ?EditorView) → bool]) → (EditorState, ?(tr: Transaction), ?EditorView) → bool
  // Combine a number of command functions into a single function (which
  // calls them one by one until one returns true).
  function chainCommands() {
    var arguments$1 = arguments;

    var commands = [], len = arguments.length;
    while ( len-- ) { commands[ len ] = arguments$1[ len ]; }

    return function(state, dispatch, view) {
      for (var i = 0; i < commands.length; i++)
        { if (commands[i](state, dispatch, view)) { return true } }
      return false
    }
  }

  var backspace = chainCommands(deleteSelection, joinBackward, selectNodeBackward);
  var del = chainCommands(deleteSelection, joinForward, selectNodeForward);

  // :: Object
  // A basic keymap containing bindings not specific to any schema.
  // Binds the following keys (when multiple commands are listed, they
  // are chained with [`chainCommands`](#commands.chainCommands)):
  //
  // * **Enter** to `newlineInCode`, `createParagraphNear`, `liftEmptyBlock`, `splitBlock`
  // * **Mod-Enter** to `exitCode`
  // * **Backspace** and **Mod-Backspace** to `deleteSelection`, `joinBackward`, `selectNodeBackward`
  // * **Delete** and **Mod-Delete** to `deleteSelection`, `joinForward`, `selectNodeForward`
  // * **Mod-Delete** to `deleteSelection`, `joinForward`, `selectNodeForward`
  // * **Mod-a** to `selectAll`
  var pcBaseKeymap = {
    "Enter": chainCommands(newlineInCode, createParagraphNear, liftEmptyBlock, splitBlock),
    "Mod-Enter": exitCode,
    "Backspace": backspace,
    "Mod-Backspace": backspace,
    "Delete": del,
    "Mod-Delete": del,
    "Mod-a": selectAll
  };

  // :: Object
  // A copy of `pcBaseKeymap` that also binds **Ctrl-h** like Backspace,
  // **Ctrl-d** like Delete, **Alt-Backspace** like Ctrl-Backspace, and
  // **Ctrl-Alt-Backspace**, **Alt-Delete**, and **Alt-d** like
  // Ctrl-Delete.
  var macBaseKeymap = {
    "Ctrl-h": pcBaseKeymap["Backspace"],
    "Alt-Backspace": pcBaseKeymap["Mod-Backspace"],
    "Ctrl-d": pcBaseKeymap["Delete"],
    "Ctrl-Alt-Backspace": pcBaseKeymap["Mod-Delete"],
    "Alt-Delete": pcBaseKeymap["Mod-Delete"],
    "Alt-d": pcBaseKeymap["Mod-Delete"]
  };
  for (var key$3 in pcBaseKeymap) { macBaseKeymap[key$3] = pcBaseKeymap[key$3]; }

  // declare global: os, navigator
  var mac$3 = typeof navigator != "undefined" ? /Mac/.test(navigator.platform)
            : typeof os != "undefined" ? os.platform() == "darwin" : false;

  // :: Object
  // Depending on the detected platform, this will hold
  // [`pcBasekeymap`](#commands.pcBaseKeymap) or
  // [`macBaseKeymap`](#commands.macBaseKeymap).
  var baseKeymap = mac$3 ? macBaseKeymap : pcBaseKeymap;

  var commands = /*#__PURE__*/Object.freeze({
    __proto__: null,
    autoJoin: autoJoin,
    baseKeymap: baseKeymap,
    chainCommands: chainCommands,
    createParagraphNear: createParagraphNear,
    deleteSelection: deleteSelection,
    exitCode: exitCode,
    joinBackward: joinBackward,
    joinDown: joinDown,
    joinForward: joinForward,
    joinUp: joinUp,
    lift: lift$1,
    liftEmptyBlock: liftEmptyBlock,
    macBaseKeymap: macBaseKeymap,
    newlineInCode: newlineInCode,
    pcBaseKeymap: pcBaseKeymap,
    selectAll: selectAll,
    selectNodeBackward: selectNodeBackward,
    selectNodeForward: selectNodeForward,
    selectParentNode: selectParentNode,
    setBlockType: setBlockType,
    splitBlock: splitBlock,
    splitBlockKeepMarks: splitBlockKeepMarks,
    toggleMark: toggleMark,
    wrapIn: wrapIn
  });

  var GOOD_LEAF_SIZE = 200;

  // :: class<T> A rope sequence is a persistent sequence data structure
  // that supports appending, prepending, and slicing without doing a
  // full copy. It is represented as a mostly-balanced tree.
  var RopeSequence = function RopeSequence () {};

  RopeSequence.prototype.append = function append (other) {
    if (!other.length) { return this }
    other = RopeSequence.from(other);

    return (!this.length && other) ||
      (other.length < GOOD_LEAF_SIZE && this.leafAppend(other)) ||
      (this.length < GOOD_LEAF_SIZE && other.leafPrepend(this)) ||
      this.appendInner(other)
  };

  // :: (union<[T], RopeSequence<T>>) → RopeSequence<T>
  // Prepend an array or other rope to this one, returning a new rope.
  RopeSequence.prototype.prepend = function prepend (other) {
    if (!other.length) { return this }
    return RopeSequence.from(other).append(this)
  };

  RopeSequence.prototype.appendInner = function appendInner (other) {
    return new Append(this, other)
  };

  // :: (?number, ?number) → RopeSequence<T>
  // Create a rope repesenting a sub-sequence of this rope.
  RopeSequence.prototype.slice = function slice (from, to) {
      if ( from === void 0 ) { from = 0; }
      if ( to === void 0 ) { to = this.length; }

    if (from >= to) { return RopeSequence.empty }
    return this.sliceInner(Math.max(0, from), Math.min(this.length, to))
  };

  // :: (number) → T
  // Retrieve the element at the given position from this rope.
  RopeSequence.prototype.get = function get (i) {
    if (i < 0 || i >= this.length) { return undefined }
    return this.getInner(i)
  };

  // :: ((element: T, index: number) → ?bool, ?number, ?number)
  // Call the given function for each element between the given
  // indices. This tends to be more efficient than looping over the
  // indices and calling `get`, because it doesn't have to descend the
  // tree for every element.
  RopeSequence.prototype.forEach = function forEach (f, from, to) {
      if ( from === void 0 ) { from = 0; }
      if ( to === void 0 ) { to = this.length; }

    if (from <= to)
      { this.forEachInner(f, from, to, 0); }
    else
      { this.forEachInvertedInner(f, from, to, 0); }
  };

  // :: ((element: T, index: number) → U, ?number, ?number) → [U]
  // Map the given functions over the elements of the rope, producing
  // a flat array.
  RopeSequence.prototype.map = function map (f, from, to) {
      if ( from === void 0 ) { from = 0; }
      if ( to === void 0 ) { to = this.length; }

    var result = [];
    this.forEach(function (elt, i) { return result.push(f(elt, i)); }, from, to);
    return result
  };

  // :: (?union<[T], RopeSequence<T>>) → RopeSequence<T>
  // Create a rope representing the given array, or return the rope
  // itself if a rope was given.
  RopeSequence.from = function from (values) {
    if (values instanceof RopeSequence) { return values }
    return values && values.length ? new Leaf(values) : RopeSequence.empty
  };

  var Leaf = /*@__PURE__*/(function (RopeSequence) {
    function Leaf(values) {
      RopeSequence.call(this);
      this.values = values;
    }

    if ( RopeSequence ) { Leaf.__proto__ = RopeSequence; }
    Leaf.prototype = Object.create( RopeSequence && RopeSequence.prototype );
    Leaf.prototype.constructor = Leaf;

    var prototypeAccessors = { length: { configurable: true },depth: { configurable: true } };

    Leaf.prototype.flatten = function flatten () {
      return this.values
    };

    Leaf.prototype.sliceInner = function sliceInner (from, to) {
      if (from == 0 && to == this.length) { return this }
      return new Leaf(this.values.slice(from, to))
    };

    Leaf.prototype.getInner = function getInner (i) {
      return this.values[i]
    };

    Leaf.prototype.forEachInner = function forEachInner (f, from, to, start) {
      for (var i = from; i < to; i++)
        { if (f(this.values[i], start + i) === false) { return false } }
    };

    Leaf.prototype.forEachInvertedInner = function forEachInvertedInner (f, from, to, start) {
      for (var i = from - 1; i >= to; i--)
        { if (f(this.values[i], start + i) === false) { return false } }
    };

    Leaf.prototype.leafAppend = function leafAppend (other) {
      if (this.length + other.length <= GOOD_LEAF_SIZE)
        { return new Leaf(this.values.concat(other.flatten())) }
    };

    Leaf.prototype.leafPrepend = function leafPrepend (other) {
      if (this.length + other.length <= GOOD_LEAF_SIZE)
        { return new Leaf(other.flatten().concat(this.values)) }
    };

    prototypeAccessors.length.get = function () { return this.values.length };

    prototypeAccessors.depth.get = function () { return 0 };

    Object.defineProperties( Leaf.prototype, prototypeAccessors );

    return Leaf;
  }(RopeSequence));

  // :: RopeSequence
  // The empty rope sequence.
  RopeSequence.empty = new Leaf([]);

  var Append = /*@__PURE__*/(function (RopeSequence) {
    function Append(left, right) {
      RopeSequence.call(this);
      this.left = left;
      this.right = right;
      this.length = left.length + right.length;
      this.depth = Math.max(left.depth, right.depth) + 1;
    }

    if ( RopeSequence ) { Append.__proto__ = RopeSequence; }
    Append.prototype = Object.create( RopeSequence && RopeSequence.prototype );
    Append.prototype.constructor = Append;

    Append.prototype.flatten = function flatten () {
      return this.left.flatten().concat(this.right.flatten())
    };

    Append.prototype.getInner = function getInner (i) {
      return i < this.left.length ? this.left.get(i) : this.right.get(i - this.left.length)
    };

    Append.prototype.forEachInner = function forEachInner (f, from, to, start) {
      var leftLen = this.left.length;
      if (from < leftLen &&
          this.left.forEachInner(f, from, Math.min(to, leftLen), start) === false)
        { return false }
      if (to > leftLen &&
          this.right.forEachInner(f, Math.max(from - leftLen, 0), Math.min(this.length, to) - leftLen, start + leftLen) === false)
        { return false }
    };

    Append.prototype.forEachInvertedInner = function forEachInvertedInner (f, from, to, start) {
      var leftLen = this.left.length;
      if (from > leftLen &&
          this.right.forEachInvertedInner(f, from - leftLen, Math.max(to, leftLen) - leftLen, start + leftLen) === false)
        { return false }
      if (to < leftLen &&
          this.left.forEachInvertedInner(f, Math.min(from, leftLen), to, start) === false)
        { return false }
    };

    Append.prototype.sliceInner = function sliceInner (from, to) {
      if (from == 0 && to == this.length) { return this }
      var leftLen = this.left.length;
      if (to <= leftLen) { return this.left.slice(from, to) }
      if (from >= leftLen) { return this.right.slice(from - leftLen, to - leftLen) }
      return this.left.slice(from, leftLen).append(this.right.slice(0, to - leftLen))
    };

    Append.prototype.leafAppend = function leafAppend (other) {
      var inner = this.right.leafAppend(other);
      if (inner) { return new Append(this.left, inner) }
    };

    Append.prototype.leafPrepend = function leafPrepend (other) {
      var inner = this.left.leafPrepend(other);
      if (inner) { return new Append(inner, this.right) }
    };

    Append.prototype.appendInner = function appendInner (other) {
      if (this.left.depth >= Math.max(this.right.depth, other.depth) + 1)
        { return new Append(this.left, new Append(this.right, other)) }
      return new Append(this, other)
    };

    return Append;
  }(RopeSequence));

  var ropeSequence = RopeSequence;

  // ProseMirror's history isn't simply a way to roll back to a previous
  // state, because ProseMirror supports applying changes without adding
  // them to the history (for example during collaboration).
  //
  // To this end, each 'Branch' (one for the undo history and one for
  // the redo history) keeps an array of 'Items', which can optionally
  // hold a step (an actual undoable change), and always hold a position
  // map (which is needed to move changes below them to apply to the
  // current document).
  //
  // An item that has both a step and a selection bookmark is the start
  // of an 'event' — a group of changes that will be undone or redone at
  // once. (It stores only the bookmark, since that way we don't have to
  // provide a document until the selection is actually applied, which
  // is useful when compressing.)

  // Used to schedule history compression
  var max_empty_items = 500;

  var Branch = function Branch(items, eventCount) {
    this.items = items;
    this.eventCount = eventCount;
  };

  // : (EditorState, bool) → ?{transform: Transform, selection: ?SelectionBookmark, remaining: Branch}
  // Pop the latest event off the branch's history and apply it
  // to a document transform.
  Branch.prototype.popEvent = function popEvent (state, preserveItems) {
      var this$1$1 = this;

    if (this.eventCount == 0) { return null }

    var end = this.items.length;
    for (;; end--) {
      var next = this.items.get(end - 1);
      if (next.selection) { --end; break }
    }

    var remap, mapFrom;
    if (preserveItems) {
      remap = this.remapping(end, this.items.length);
      mapFrom = remap.maps.length;
    }
    var transform = state.tr;
    var selection, remaining;
    var addAfter = [], addBefore = [];

    this.items.forEach(function (item, i) {
      if (!item.step) {
        if (!remap) {
          remap = this$1$1.remapping(end, i + 1);
          mapFrom = remap.maps.length;
        }
        mapFrom--;
        addBefore.push(item);
        return
      }

      if (remap) {
        addBefore.push(new Item(item.map));
        var step = item.step.map(remap.slice(mapFrom)), map;

        if (step && transform.maybeStep(step).doc) {
          map = transform.mapping.maps[transform.mapping.maps.length - 1];
          addAfter.push(new Item(map, null, null, addAfter.length + addBefore.length));
        }
        mapFrom--;
        if (map) { remap.appendMap(map, mapFrom); }
      } else {
        transform.maybeStep(item.step);
      }

      if (item.selection) {
        selection = remap ? item.selection.map(remap.slice(mapFrom)) : item.selection;
        remaining = new Branch(this$1$1.items.slice(0, end).append(addBefore.reverse().concat(addAfter)), this$1$1.eventCount - 1);
        return false
      }
    }, this.items.length, 0);

    return {remaining: remaining, transform: transform, selection: selection}
  };

  // : (Transform, ?SelectionBookmark, Object) → Branch
  // Create a new branch with the given transform added.
  Branch.prototype.addTransform = function addTransform (transform, selection, histOptions, preserveItems) {
    var newItems = [], eventCount = this.eventCount;
    var oldItems = this.items, lastItem = !preserveItems && oldItems.length ? oldItems.get(oldItems.length - 1) : null;

    for (var i = 0; i < transform.steps.length; i++) {
      var step = transform.steps[i].invert(transform.docs[i]);
      var item = new Item(transform.mapping.maps[i], step, selection), merged = (void 0);
      if (merged = lastItem && lastItem.merge(item)) {
        item = merged;
        if (i) { newItems.pop(); }
        else { oldItems = oldItems.slice(0, oldItems.length - 1); }
      }
      newItems.push(item);
      if (selection) {
        eventCount++;
        selection = null;
      }
      if (!preserveItems) { lastItem = item; }
    }
    var overflow = eventCount - histOptions.depth;
    if (overflow > DEPTH_OVERFLOW) {
      oldItems = cutOffEvents(oldItems, overflow);
      eventCount -= overflow;
    }
    return new Branch(oldItems.append(newItems), eventCount)
  };

  Branch.prototype.remapping = function remapping (from, to) {
    var maps = new Mapping;
    this.items.forEach(function (item, i) {
      var mirrorPos = item.mirrorOffset != null && i - item.mirrorOffset >= from
          ? maps.maps.length - item.mirrorOffset : null;
      maps.appendMap(item.map, mirrorPos);
    }, from, to);
    return maps
  };

  Branch.prototype.addMaps = function addMaps (array) {
    if (this.eventCount == 0) { return this }
    return new Branch(this.items.append(array.map(function (map) { return new Item(map); })), this.eventCount)
  };

  // : (Transform, number)
  // When the collab module receives remote changes, the history has
  // to know about those, so that it can adjust the steps that were
  // rebased on top of the remote changes, and include the position
  // maps for the remote changes in its array of items.
  Branch.prototype.rebased = function rebased (rebasedTransform, rebasedCount) {
    if (!this.eventCount) { return this }

    var rebasedItems = [], start = Math.max(0, this.items.length - rebasedCount);

    var mapping = rebasedTransform.mapping;
    var newUntil = rebasedTransform.steps.length;
    var eventCount = this.eventCount;
    this.items.forEach(function (item) { if (item.selection) { eventCount--; } }, start);

    var iRebased = rebasedCount;
    this.items.forEach(function (item) {
      var pos = mapping.getMirror(--iRebased);
      if (pos == null) { return }
      newUntil = Math.min(newUntil, pos);
      var map = mapping.maps[pos];
      if (item.step) {
        var step = rebasedTransform.steps[pos].invert(rebasedTransform.docs[pos]);
        var selection = item.selection && item.selection.map(mapping.slice(iRebased + 1, pos));
        if (selection) { eventCount++; }
        rebasedItems.push(new Item(map, step, selection));
      } else {
        rebasedItems.push(new Item(map));
      }
    }, start);

    var newMaps = [];
    for (var i = rebasedCount; i < newUntil; i++)
      { newMaps.push(new Item(mapping.maps[i])); }
    var items = this.items.slice(0, start).append(newMaps).append(rebasedItems);
    var branch = new Branch(items, eventCount);

    if (branch.emptyItemCount() > max_empty_items)
      { branch = branch.compress(this.items.length - rebasedItems.length); }
    return branch
  };

  Branch.prototype.emptyItemCount = function emptyItemCount () {
    var count = 0;
    this.items.forEach(function (item) { if (!item.step) { count++; } });
    return count
  };

  // Compressing a branch means rewriting it to push the air (map-only
  // items) out. During collaboration, these naturally accumulate
  // because each remote change adds one. The `upto` argument is used
  // to ensure that only the items below a given level are compressed,
  // because `rebased` relies on a clean, untouched set of items in
  // order to associate old items with rebased steps.
  Branch.prototype.compress = function compress (upto) {
      if ( upto === void 0 ) { upto = this.items.length; }

    var remap = this.remapping(0, upto), mapFrom = remap.maps.length;
    var items = [], events = 0;
    this.items.forEach(function (item, i) {
      if (i >= upto) {
        items.push(item);
        if (item.selection) { events++; }
      } else if (item.step) {
        var step = item.step.map(remap.slice(mapFrom)), map = step && step.getMap();
        mapFrom--;
        if (map) { remap.appendMap(map, mapFrom); }
        if (step) {
          var selection = item.selection && item.selection.map(remap.slice(mapFrom));
          if (selection) { events++; }
          var newItem = new Item(map.invert(), step, selection), merged, last = items.length - 1;
          if (merged = items.length && items[last].merge(newItem))
            { items[last] = merged; }
          else
            { items.push(newItem); }
        }
      } else if (item.map) {
        mapFrom--;
      }
    }, this.items.length, 0);
    return new Branch(ropeSequence.from(items.reverse()), events)
  };

  Branch.empty = new Branch(ropeSequence.empty, 0);

  function cutOffEvents(items, n) {
    var cutPoint;
    items.forEach(function (item, i) {
      if (item.selection && (n-- == 0)) {
        cutPoint = i;
        return false
      }
    });
    return items.slice(cutPoint)
  }

  var Item = function Item(map, step, selection, mirrorOffset) {
    // The (forward) step map for this item.
    this.map = map;
    // The inverted step
    this.step = step;
    // If this is non-null, this item is the start of a group, and
    // this selection is the starting selection for the group (the one
    // that was active before the first step was applied)
    this.selection = selection;
    // If this item is the inverse of a previous mapping on the stack,
    // this points at the inverse's offset
    this.mirrorOffset = mirrorOffset;
  };

  Item.prototype.merge = function merge (other) {
    if (this.step && other.step && !other.selection) {
      var step = other.step.merge(this.step);
      if (step) { return new Item(step.getMap().invert(), step, this.selection) }
    }
  };

  // The value of the state field that tracks undo/redo history for that
  // state. Will be stored in the plugin state when the history plugin
  // is active.
  var HistoryState = function HistoryState(done, undone, prevRanges, prevTime) {
    this.done = done;
    this.undone = undone;
    this.prevRanges = prevRanges;
    this.prevTime = prevTime;
  };

  var DEPTH_OVERFLOW = 20;

  // : (HistoryState, EditorState, Transaction, Object)
  // Record a transformation in undo history.
  function applyTransaction(history, state, tr, options) {
    var historyTr = tr.getMeta(historyKey), rebased;
    if (historyTr) { return historyTr.historyState }

    if (tr.getMeta(closeHistoryKey)) { history = new HistoryState(history.done, history.undone, null, 0); }

    var appended = tr.getMeta("appendedTransaction");

    if (tr.steps.length == 0) {
      return history
    } else if (appended && appended.getMeta(historyKey)) {
      if (appended.getMeta(historyKey).redo)
        { return new HistoryState(history.done.addTransform(tr, null, options, mustPreserveItems(state)),
                                history.undone, rangesFor(tr.mapping.maps[tr.steps.length - 1]), history.prevTime) }
      else
        { return new HistoryState(history.done, history.undone.addTransform(tr, null, options, mustPreserveItems(state)),
                                null, history.prevTime) }
    } else if (tr.getMeta("addToHistory") !== false && !(appended && appended.getMeta("addToHistory") === false)) {
      // Group transforms that occur in quick succession into one event.
      var newGroup = history.prevTime == 0 || !appended && (history.prevTime < (tr.time || 0) - options.newGroupDelay ||
                                                            !isAdjacentTo(tr, history.prevRanges));
      var prevRanges = appended ? mapRanges(history.prevRanges, tr.mapping) : rangesFor(tr.mapping.maps[tr.steps.length - 1]);
      return new HistoryState(history.done.addTransform(tr, newGroup ? state.selection.getBookmark() : null,
                                                        options, mustPreserveItems(state)),
                              Branch.empty, prevRanges, tr.time)
    } else if (rebased = tr.getMeta("rebased")) {
      // Used by the collab module to tell the history that some of its
      // content has been rebased.
      return new HistoryState(history.done.rebased(tr, rebased),
                              history.undone.rebased(tr, rebased),
                              mapRanges(history.prevRanges, tr.mapping), history.prevTime)
    } else {
      return new HistoryState(history.done.addMaps(tr.mapping.maps),
                              history.undone.addMaps(tr.mapping.maps),
                              mapRanges(history.prevRanges, tr.mapping), history.prevTime)
    }
  }

  function isAdjacentTo(transform, prevRanges) {
    if (!prevRanges) { return false }
    if (!transform.docChanged) { return true }
    var adjacent = false;
    transform.mapping.maps[0].forEach(function (start, end) {
      for (var i = 0; i < prevRanges.length; i += 2)
        { if (start <= prevRanges[i + 1] && end >= prevRanges[i])
          { adjacent = true; } }
    });
    return adjacent
  }

  function rangesFor(map) {
    var result = [];
    map.forEach(function (_from, _to, from, to) { return result.push(from, to); });
    return result
  }

  function mapRanges(ranges, mapping) {
    if (!ranges) { return null }
    var result = [];
    for (var i = 0; i < ranges.length; i += 2) {
      var from = mapping.map(ranges[i], 1), to = mapping.map(ranges[i + 1], -1);
      if (from <= to) { result.push(from, to); }
    }
    return result
  }

  // : (HistoryState, EditorState, (tr: Transaction), bool)
  // Apply the latest event from one branch to the document and shift the event
  // onto the other branch.
  function histTransaction(history, state, dispatch, redo) {
    var preserveItems = mustPreserveItems(state), histOptions = historyKey.get(state).spec.config;
    var pop = (redo ? history.undone : history.done).popEvent(state, preserveItems);
    if (!pop) { return }

    var selection = pop.selection.resolve(pop.transform.doc);
    var added = (redo ? history.done : history.undone).addTransform(pop.transform, state.selection.getBookmark(),
                                                                    histOptions, preserveItems);

    var newHist = new HistoryState(redo ? added : pop.remaining, redo ? pop.remaining : added, null, 0);
    dispatch(pop.transform.setSelection(selection).setMeta(historyKey, {redo: redo, historyState: newHist}).scrollIntoView());
  }

  var cachedPreserveItems = false, cachedPreserveItemsPlugins = null;
  // Check whether any plugin in the given state has a
  // `historyPreserveItems` property in its spec, in which case we must
  // preserve steps exactly as they came in, so that they can be
  // rebased.
  function mustPreserveItems(state) {
    var plugins = state.plugins;
    if (cachedPreserveItemsPlugins != plugins) {
      cachedPreserveItems = false;
      cachedPreserveItemsPlugins = plugins;
      for (var i = 0; i < plugins.length; i++) { if (plugins[i].spec.historyPreserveItems) {
        cachedPreserveItems = true;
        break
      } }
    }
    return cachedPreserveItems
  }

  // :: (Transaction) → Transaction
  // Set a flag on the given transaction that will prevent further steps
  // from being appended to an existing history event (so that they
  // require a separate undo command to undo).
  function closeHistory(tr) {
    return tr.setMeta(closeHistoryKey, true)
  }

  var historyKey = new PluginKey("history");
  var closeHistoryKey = new PluginKey("closeHistory");

  // :: (?Object) → Plugin
  // Returns a plugin that enables the undo history for an editor. The
  // plugin will track undo and redo stacks, which can be used with the
  // [`undo`](#history.undo) and [`redo`](#history.redo) commands.
  //
  // You can set an `"addToHistory"` [metadata
  // property](#state.Transaction.setMeta) of `false` on a transaction
  // to prevent it from being rolled back by undo.
  //
  //   config::-
  //   Supports the following configuration options:
  //
  //     depth:: ?number
  //     The amount of history events that are collected before the
  //     oldest events are discarded. Defaults to 100.
  //
  //     newGroupDelay:: ?number
  //     The delay between changes after which a new group should be
  //     started. Defaults to 500 (milliseconds). Note that when changes
  //     aren't adjacent, a new group is always started.
  function history(config) {
    config = {depth: config && config.depth || 100,
              newGroupDelay: config && config.newGroupDelay || 500};
    return new Plugin({
      key: historyKey,

      state: {
        init: function init() {
          return new HistoryState(Branch.empty, Branch.empty, null, 0)
        },
        apply: function apply(tr, hist, state) {
          return applyTransaction(hist, state, tr, config)
        }
      },

      config: config,

      props: {
        handleDOMEvents: {
          beforeinput: function beforeinput(view, e) {
            var handled = e.inputType == "historyUndo" ? undo(view.state, view.dispatch) :
                e.inputType == "historyRedo" ? redo(view.state, view.dispatch) : false;
            if (handled) { e.preventDefault(); }
            return handled
          }
        }
      }
    })
  }

  // :: (EditorState, ?(tr: Transaction)) → bool
  // A command function that undoes the last change, if any.
  function undo(state, dispatch) {
    var hist = historyKey.getState(state);
    if (!hist || hist.done.eventCount == 0) { return false }
    if (dispatch) { histTransaction(hist, state, dispatch, false); }
    return true
  }

  // :: (EditorState, ?(tr: Transaction)) → bool
  // A command function that redoes the last undone change, if any.
  function redo(state, dispatch) {
    var hist = historyKey.getState(state);
    if (!hist || hist.undone.eventCount == 0) { return false }
    if (dispatch) { histTransaction(hist, state, dispatch, true); }
    return true
  }

  // :: (EditorState) → number
  // The amount of undoable events available in a given state.
  function undoDepth(state) {
    var hist = historyKey.getState(state);
    return hist ? hist.done.eventCount : 0
  }

  // :: (EditorState) → number
  // The amount of redoable events available in a given editor state.
  function redoDepth(state) {
    var hist = historyKey.getState(state);
    return hist ? hist.undone.eventCount : 0
  }

  var history$1 = /*#__PURE__*/Object.freeze({
    __proto__: null,
    HistoryState: HistoryState,
    closeHistory: closeHistory,
    history: history,
    redo: redo,
    redoDepth: redoDepth,
    undo: undo,
    undoDepth: undoDepth
  });

  var base$1 = {
    8: "Backspace",
    9: "Tab",
    10: "Enter",
    12: "NumLock",
    13: "Enter",
    16: "Shift",
    17: "Control",
    18: "Alt",
    20: "CapsLock",
    27: "Escape",
    32: " ",
    33: "PageUp",
    34: "PageDown",
    35: "End",
    36: "Home",
    37: "ArrowLeft",
    38: "ArrowUp",
    39: "ArrowRight",
    40: "ArrowDown",
    44: "PrintScreen",
    45: "Insert",
    46: "Delete",
    59: ";",
    61: "=",
    91: "Meta",
    92: "Meta",
    106: "*",
    107: "+",
    108: ",",
    109: "-",
    110: ".",
    111: "/",
    144: "NumLock",
    145: "ScrollLock",
    160: "Shift",
    161: "Shift",
    162: "Control",
    163: "Control",
    164: "Alt",
    165: "Alt",
    173: "-",
    186: ";",
    187: "=",
    188: ",",
    189: "-",
    190: ".",
    191: "/",
    192: "`",
    219: "[",
    220: "\\",
    221: "]",
    222: "'",
    229: "q"
  };

  var shift = {
    48: ")",
    49: "!",
    50: "@",
    51: "#",
    52: "$",
    53: "%",
    54: "^",
    55: "&",
    56: "*",
    57: "(",
    59: ":",
    61: "+",
    173: "_",
    186: ":",
    187: "+",
    188: "<",
    189: "_",
    190: ">",
    191: "?",
    192: "~",
    219: "{",
    220: "|",
    221: "}",
    222: "\"",
    229: "Q"
  };

  var chrome = typeof navigator != "undefined" && /Chrome\/(\d+)/.exec(navigator.userAgent);
  var safari = typeof navigator != "undefined" && /Apple Computer/.test(navigator.vendor);
  var gecko = typeof navigator != "undefined" && /Gecko\/\d+/.test(navigator.userAgent);
  var mac$2 = typeof navigator != "undefined" && /Mac/.test(navigator.platform);
  var ie = typeof navigator != "undefined" && /MSIE \d|Trident\/(?:[7-9]|\d{2,})\..*rv:(\d+)/.exec(navigator.userAgent);
  var brokenModifierNames = chrome && (mac$2 || +chrome[1] < 57) || gecko && mac$2;

  // Fill in the digit keys
  for (var i$2 = 0; i$2 < 10; i$2++) { base$1[48 + i$2] = base$1[96 + i$2] = String(i$2); }

  // The function keys
  for (var i$2 = 1; i$2 <= 24; i$2++) { base$1[i$2 + 111] = "F" + i$2; }

  // And the alphabetic keys
  for (var i$2 = 65; i$2 <= 90; i$2++) {
    base$1[i$2] = String.fromCharCode(i$2 + 32);
    shift[i$2] = String.fromCharCode(i$2);
  }

  // For each code that doesn't have a shift-equivalent, copy the base name
  for (var code$3 in base$1) { if (!shift.hasOwnProperty(code$3)) { shift[code$3] = base$1[code$3]; } }

  function keyName(event) {
    // Don't trust event.key in Chrome when there are modifiers until
    // they fix https://bugs.chromium.org/p/chromium/issues/detail?id=633838
    var ignoreKey = brokenModifierNames && (event.ctrlKey || event.altKey || event.metaKey) ||
      (safari || ie) && event.shiftKey && event.key && event.key.length == 1;
    var name = (!ignoreKey && event.key) ||
      (event.shiftKey ? shift : base$1)[event.keyCode] ||
      event.key || "Unidentified";
    // Edge sometimes produces wrong names (Issue #3)
    if (name == "Esc") { name = "Escape"; }
    if (name == "Del") { name = "Delete"; }
    // https://developer.microsoft.com/en-us/microsoft-edge/platform/issues/8860571/
    if (name == "Left") { name = "ArrowLeft"; }
    if (name == "Up") { name = "ArrowUp"; }
    if (name == "Right") { name = "ArrowRight"; }
    if (name == "Down") { name = "ArrowDown"; }
    return name
  }

  // declare global: navigator

  var mac$1 = typeof navigator != "undefined" ? /Mac/.test(navigator.platform) : false;

  function normalizeKeyName(name) {
    var parts = name.split(/-(?!$)/), result = parts[parts.length - 1];
    if (result == "Space") { result = " "; }
    var alt, ctrl, shift, meta;
    for (var i = 0; i < parts.length - 1; i++) {
      var mod = parts[i];
      if (/^(cmd|meta|m)$/i.test(mod)) { meta = true; }
      else if (/^a(lt)?$/i.test(mod)) { alt = true; }
      else if (/^(c|ctrl|control)$/i.test(mod)) { ctrl = true; }
      else if (/^s(hift)?$/i.test(mod)) { shift = true; }
      else if (/^mod$/i.test(mod)) { if (mac$1) { meta = true; } else { ctrl = true; } }
      else { throw new Error("Unrecognized modifier name: " + mod) }
    }
    if (alt) { result = "Alt-" + result; }
    if (ctrl) { result = "Ctrl-" + result; }
    if (meta) { result = "Meta-" + result; }
    if (shift) { result = "Shift-" + result; }
    return result
  }

  function normalize$2(map) {
    var copy = Object.create(null);
    for (var prop in map) { copy[normalizeKeyName(prop)] = map[prop]; }
    return copy
  }

  function modifiers(name, event, shift) {
    if (event.altKey) { name = "Alt-" + name; }
    if (event.ctrlKey) { name = "Ctrl-" + name; }
    if (event.metaKey) { name = "Meta-" + name; }
    if (shift !== false && event.shiftKey) { name = "Shift-" + name; }
    return name
  }

  // :: (Object) → Plugin
  // Create a keymap plugin for the given set of bindings.
  //
  // Bindings should map key names to [command](#commands)-style
  // functions, which will be called with `(EditorState, dispatch,
  // EditorView)` arguments, and should return true when they've handled
  // the key. Note that the view argument isn't part of the command
  // protocol, but can be used as an escape hatch if a binding needs to
  // directly interact with the UI.
  //
  // Key names may be strings like `"Shift-Ctrl-Enter"`—a key
  // identifier prefixed with zero or more modifiers. Key identifiers
  // are based on the strings that can appear in
  // [`KeyEvent.key`](https://developer.mozilla.org/en-US/docs/Web/API/KeyboardEvent/key).
  // Use lowercase letters to refer to letter keys (or uppercase letters
  // if you want shift to be held). You may use `"Space"` as an alias
  // for the `" "` name.
  //
  // Modifiers can be given in any order. `Shift-` (or `s-`), `Alt-` (or
  // `a-`), `Ctrl-` (or `c-` or `Control-`) and `Cmd-` (or `m-` or
  // `Meta-`) are recognized. For characters that are created by holding
  // shift, the `Shift-` prefix is implied, and should not be added
  // explicitly.
  //
  // You can use `Mod-` as a shorthand for `Cmd-` on Mac and `Ctrl-` on
  // other platforms.
  //
  // You can add multiple keymap plugins to an editor. The order in
  // which they appear determines their precedence (the ones early in
  // the array get to dispatch first).
  function keymap$2(bindings) {
    return new Plugin({props: {handleKeyDown: keydownHandler(bindings)}})
  }

  // :: (Object) → (view: EditorView, event: dom.Event) → bool
  // Given a set of bindings (using the same format as
  // [`keymap`](#keymap.keymap), return a [keydown
  // handler](#view.EditorProps.handleKeyDown) that handles them.
  function keydownHandler(bindings) {
    var map = normalize$2(bindings);
    return function(view, event) {
      var name = keyName(event), isChar = name.length == 1 && name != " ", baseName;
      var direct = map[modifiers(name, event, !isChar)];
      if (direct && direct(view.state, view.dispatch, view)) { return true }
      if (isChar && (event.shiftKey || event.altKey || event.metaKey || name.charCodeAt(0) > 127) &&
          (baseName = base$1[event.keyCode]) && baseName != name) {
        // Try falling back to the keyCode when there's a modifier
        // active or the character produced isn't ASCII, and our table
        // produces a different name from the the keyCode. See #668,
        // #1060
        var fromCode = map[modifiers(baseName, event, true)];
        if (fromCode && fromCode(view.state, view.dispatch, view)) { return true }
      } else if (isChar && event.shiftKey) {
        // Otherwise, if shift is active, also try the binding with the
        // Shift- prefix enabled. See #997
        var withShift = map[modifiers(name, event, true)];
        if (withShift && withShift(view.state, view.dispatch, view)) { return true }
      }
      return false
    }
  }

  var keymap$3 = /*#__PURE__*/Object.freeze({
    __proto__: null,
    keydownHandler: keydownHandler,
    keymap: keymap$2
  });

  function crelt() {
    var arguments$1 = arguments;

    var elt = arguments[0];
    if (typeof elt == "string") { elt = document.createElement(elt); }
    var i = 1, next = arguments[1];
    if (next && typeof next == "object" && next.nodeType == null && !Array.isArray(next)) {
      for (var name in next) { if (Object.prototype.hasOwnProperty.call(next, name)) {
        var value = next[name];
        if (typeof value == "string") { elt.setAttribute(name, value); }
        else if (value != null) { elt[name] = value; }
      } }
      i++;
    }
    for (; i < arguments.length; i++) { add(elt, arguments$1[i]); }
    return elt
  }

  function add(elt, child) {
    if (typeof child == "string") {
      elt.appendChild(document.createTextNode(child));
    } else if (child == null) ; else if (child.nodeType != null) {
      elt.appendChild(child);
    } else if (Array.isArray(child)) {
      for (var i = 0; i < child.length; i++) { add(elt, child[i]); }
    } else {
      throw new RangeError("Unsupported child node: " + child)
    }
  }

  var SVG$1 = "http://www.w3.org/2000/svg";
  var XLINK$1 = "http://www.w3.org/1999/xlink";

  var prefix$5 = "ProseMirror-icon";

  function hashPath$1(path) {
      if (Array.isArray(path)) {
          var paths = path;
          path = '';
          paths.forEach(function (pathItem) {
              path += pathItem;
          });
      }

      var hash = 0;
      for (var i = 0; i < path.length; i++)
          { hash = (((hash << 5) - hash) + path.charCodeAt(i)) | 0; }
      return hash
  }

  function getIcon$1(icon) {
      var node = document.createElement("div");
      node.className = prefix$5;
      if (icon.path) {
          var name = "pm-icon-" + hashPath$1(icon.path).toString(16);
          if (!document.getElementById(name)) { buildSVG$1(name, icon); }
          var svg = node.appendChild(document.createElementNS(SVG$1, "svg"));
          svg.style.width = (icon.width / icon.height) + "em";
          var use = svg.appendChild(document.createElementNS(SVG$1, "use"));
          use.setAttributeNS(XLINK$1, "href", /([^#]*)/.exec(document.location)[1] + "#" + name);
      } else if (icon.dom) {
          node.appendChild(icon.dom.cloneNode(true));
      } else {
          node.appendChild(document.createElement("span")).textContent = icon.text || '';
          if (icon.css) { node.firstChild.style.cssText = icon.css; }
      }
      return node
  }

  function buildSVG$1(name, data) {
      var collection = document.getElementById(prefix$5 + "-collection");
      if (!collection) {
          collection = document.createElementNS(SVG$1, "svg");
          collection.id = prefix$5 + "-collection";
          collection.style.display = "none";
          document.body.insertBefore(collection, document.body.firstChild);
      }
      var sym = document.createElementNS(SVG$1, "symbol");
      sym.id = name;
      sym.setAttribute("viewBox", "0 0 " + data.width + " " + data.height);

      var pathData = Array.isArray(data.path) ? data.path : [data.path];

      pathData.forEach(function (path) {
          var pathDom = sym.appendChild(document.createElementNS(SVG$1, "path"));
          pathDom.setAttribute("d", path);
          collection.appendChild(sym);
      });

  }

  var prefix$4 = "ProseMirror-menu";

  // Helpers to create specific types of items
  function cmdItem(cmd, options) {
      var passedOptions = {
          label: options.title,
          run: cmd
      };
      for (var prop in options) { passedOptions[prop] = options[prop]; }
      if ((!options.enable || options.enable === true) && !options.select)
          { passedOptions[options.enable ? "enable" : "select"] = function (state) { return cmd(state); }; }

      return new MenuItem$1(passedOptions)
  }

  function markItem(markType, options) {
      var passedOptions = {
          active: function active(state) {
              return markActive(state, markType)
          },
          enable: true
      };
      for (var prop in options) { passedOptions[prop] = options[prop]; }
      return cmdItem(toggleMark(markType), passedOptions)
  }

  function markActive(state, type) {
      var ref = state.selection;
      var from = ref.from;
      var $from = ref.$from;
      var to = ref.to;
      var empty = ref.empty;
      if (empty) { return type.isInSet(state.storedMarks || $from.marks()) }
      else { return state.doc.rangeHasMark(from, to, type) }
  }

  function wrapListItem(nodeType, options) {
      return cmdItem(wrapInList(nodeType, options.attrs), options)
  }

  // ::- An icon or label that, when clicked, executes a command.
  var MenuItem$1 = function MenuItem(options) {
      // :: MenuItemSpec
      // The options used to create the menu item.
      this.options = options || {};
      this.sortOrder = this.options.sortOrder;
  };

  // :: (EditorView) → {dom: dom.Node, update: (EditorState) → bool}
  // Renders the icon according to its [display
  // options](#menu.MenuItemSpec.display), and adds an event handler which
  // executes the command when the representation is clicked.
  MenuItem$1.prototype.render = function render (view) {
          var this$1$1 = this;

      var options = this.options;

      if (typeof this.options.render === 'function') {
          return this.options.render.apply(this, [options]);
      }

      this.dom = options.icon ? getIcon$1(options.icon)
          : options.label ? $('<div>').html(translate$1(view, options.label))[0]
              : null;

      if(this.options.id) {
          this.dom.classList.add(prefix$4+'-'+this.options.id);
      }

      if (!this.dom) { throw new RangeError("MenuItem without icon or label property"); }

      if (options.title !== 'undefined') {
          var title = (typeof options.title === "function" ? options.title(view.state) : options.title);
          this.dom.setAttribute("title", translate$1(view, title));
      }

      if (options.class) { this.dom.classList.add(options.class); }
      if (options.css) { this.dom.style.cssText += options.css; }

      $(this.dom).on("mousedown", function (e) {
          e.preventDefault();
          if (!$(this$1$1.dom).hasClass(prefix$4 + "-disabled")) {
              options.run.call(this$1$1, view.state, view.dispatch, view, e);
          }
      });

      return this.dom;
  };

  MenuItem$1.prototype.switchIcon = function switchIcon (icon, title) {
      if(title) {
          $(this.dom).attr('title', title);
      }
      $(this.dom).find('svg').replaceWith($(getIcon$1(icon)).find('svg'));
  };

  MenuItem$1.prototype.update = function update (state) {
      this.adoptItemState(state);
      return this.selected;
  };

  MenuItem$1.prototype.adoptItemState = function adoptItemState (state, forceEnable, forceActive) {
      this.setEnabledItemState(state, forceEnable);
      this.setActiveItemState(state, forceActive);
      this.setSelectedItemState(state, forceEnable);
  };

  MenuItem$1.prototype.setActiveItemState = function setActiveItemState (state, forceActive) {
      this.active = false;
      if (this.options.active) {
          this.active = (this.options.active(state) || forceActive) || false;
          setClass$1(this.dom, prefix$4 + "-active", this.active);
      }
  };

  MenuItem$1.prototype.setEnabledItemState = function setEnabledItemState (state, forceEnable) {
      this.enabled = true;
      if (this.options.enable) {
          this.enabled = this.options.enable(state) || forceEnable || false;
          setClass$1(this.dom, prefix$4 + "-disabled", !this.enabled);
      }
  };

  MenuItem$1.prototype.setSelectedItemState = function setSelectedItemState (state, forceEnable) {
      this.selected = true;
      if (this.options.select) {
          this.selected = this.options.select(state);
          this.dom.style.display = this.selected || forceEnable ? "" : "none";

          if(!this.selected) {
              this.dom.classList.add('hidden');
          } else {
              this.dom.classList.remove('hidden');
          }
          if (!this.selected) { return false }
      }
  };

  function translate$1(view, text) {
      return view._props.translate ? view._props.translate(text) : text
  }

  // MenuItemSpec:: interface
  // The configuration object passed to the `MenuItem` constructor.
  //
  //   run:: (EditorState, (Transaction), EditorView, dom.Event)
  //   The function to execute when the menu item is activated.
  //
  //   select:: ?(EditorState) → bool
  //   Optional function that is used to determine whether the item is
  //   appropriate at the moment. Deselected items will be hidden.
  //
  //   enable:: ?(EditorState) → bool
  //   Function that is used to determine if the item is enabled. If
  //   given and returning false, the item will be given a disabled
  //   styling.
  //
  //   active:: ?(EditorState) → bool
  //   A predicate function to determine whether the item is 'active' (for
  //   example, the item for toggling the strong mark might be active then
  //   the cursor is in strong text).
  //
  //   render:: ?(EditorView) → dom.Node
  //   A function that renders the item. You must provide either this,
  //   [`icon`](#menu.MenuItemSpec.icon), or [`label`](#MenuItemSpec.label).
  //
  //   icon:: ?Object
  //   Describes an icon to show for this item. The object may specify
  //   an SVG icon, in which case its `path` property should be an [SVG
  //   path
  //   spec](https://developer.mozilla.org/en-US/docs/Web/SVG/Attribute/d),
  //   and `width` and `height` should provide the viewbox in which that
  //   path exists. Alternatively, it may have a `text` property
  //   specifying a string of text that makes up the icon, with an
  //   optional `css` property giving additional CSS styling for the
  //   text. _Or_ it may contain `dom` property containing a DOM node.
  //
  //   label:: ?string
  //   Makes the item show up as a text label. Mostly useful for items
  //   wrapped in a [drop-down](#menu.Dropdown) or similar menu. The object
  //   should have a `label` property providing the text to display.
  //
  //   title:: ?union<string, (EditorState) → string>
  //   Defines DOM title (mouseover) text for the item.
  //
  //   class:: string
  //   Optionally adds a CSS class to the item's DOM representation.
  //
  //   css:: string
  //   Optionally adds a string of inline CSS to the item's DOM
  //   representation.
  //
  //   execEvent:: string
  //   Defines which event on the command's DOM representation should
  //   trigger the execution of the command. Defaults to mousedown.

  var lastMenuEvent$1 = {time: 0, node: null};

  function markMenuEvent$1(e) {
      lastMenuEvent$1.time = Date.now();
      lastMenuEvent$1.node = e.target;
  }

  function isMenuEvent$1(wrapper) {
      return Date.now() - 100 < lastMenuEvent$1.time &&
          lastMenuEvent$1.node && wrapper.contains(lastMenuEvent$1.node)
  }

  function sort(items) {
      var result = [];
      items.forEach(function (item) {
          if (item && item.type && item.type === 'dropdown') {
              result.push(new Dropdown$1(sort(item.items), item));
          } else if (item && item.type && item.type === 'group') {
              result.push(new MenuItemGroup(sort(item.items), item));
          } else if (item) {
              result.push(item);
          }
      });

      return result.sort(function (a, b) {
          if (typeof a.sortOrder === 'undefined') {
              return 1;
          }
          if (typeof b.sortOrder === 'undefined') {
              return -1;
          }
          return a.sortOrder - b.sortOrder;
      });
  }


  var MenuItemGroup = /*@__PURE__*/(function (MenuItem) {
      function MenuItemGroup(content, options) {
          var this$1$1 = this;

          MenuItem.call(this, options);
          this.content = {
              items: sort(Array.isArray(content) ? content : [content]),
              update: function (state) {

                  var result = false;

                  sort(this$1$1.content.items).forEach(function (item, i) {
                      var updateResult = item.update(state);
                      var $item = $(item.dom);

                      if(!updateResult) {
                          $item.hide();
                      } else {
                          $item.show();
                      }

                      if((i === this$1$1.content.items.length - 1)) {
                          $item.addClass('last');
                      }

                      result = result || updateResult;
                  });
                  return result;
              }
          };
      }

      if ( MenuItem ) MenuItemGroup.__proto__ = MenuItem;
      MenuItemGroup.prototype = Object.create( MenuItem && MenuItem.prototype );
      MenuItemGroup.prototype.constructor = MenuItemGroup;

      MenuItemGroup.prototype.render = function render (view) {
          var $dom = $('<div>').addClass(prefix$4 + '-group');

          if(this.options.id) {
              $dom.addClass(this.options.id);
          }

          this.renderItems(view).forEach(function (itemDom) {
              $dom.append(itemDom);
          });

          return this.dom = $dom[0];
      };

      MenuItemGroup.prototype.update = function update (state) {
          return this.content.update(state);
      };

      MenuItemGroup.prototype.renderItems = function renderItems (view) {
          var rendered = [];

          this.content.items.forEach(function (item) {
              var dom = item.render(view);
              rendered.push(crelt("div", {class: prefix$4 + "item"}, dom));
          });

          return rendered;
      };

      return MenuItemGroup;
  }(MenuItem$1));

  // ::- A drop-down menu, displayed as a label with a downwards-pointing
  // triangle to the right of it.
  var Dropdown$1 = /*@__PURE__*/(function (MenuItemGroup) {
      function Dropdown(content, options) {
          var this$1$1 = this;

          MenuItemGroup.call(this, content, options);
          this.content.update = function (state) {
              var result = false;
              this$1$1.content.items.forEach(function (item) {
                  var updateResult = item.update(state);
                  item.dom.style.display = updateResult ? "" : "none";
                  result = result || updateResult;
              });
              return result;
          };
      }

      if ( MenuItemGroup ) Dropdown.__proto__ = MenuItemGroup;
      Dropdown.prototype = Object.create( MenuItemGroup && MenuItemGroup.prototype );
      Dropdown.prototype.constructor = Dropdown;

      // :: (EditorView) → {dom: dom.Node, update: (EditorState)}
      // Render the dropdown menu and sub-items.
      Dropdown.prototype.render = function render (view) {
          var this$1$1 = this;

          var contentDom = this.renderItems(view);

          var innerDom = this.options.icon ? getIcon$1(this.options.icon)
              : this.options.label ? crelt("div", {style: this.options.css}, translate$1(view, this.options.label))
                  : null;

          if (!innerDom) {
              throw new RangeError("Dropdown without icon or label property")
          }

          innerDom.className += " " + prefix$4 + "-dropdown " + (this.options.class || "");

          if (this.options.title) {
              innerDom.setAttribute("title", translate$1(view, this.options.title));
          }

          if(this.options.id) {
              innerDom.classList.add(this.options.id);
          }

          this.dom = crelt("div", {class: prefix$4 + "-dropdown-wrap"}, innerDom);

          if(this.options.seperator) {
              this.dom.className += ' seperator';
          }

          var open = null, listeningOnClose = null;
          var close = function () {
              if (open && open.close()) {
                  open = null;
                  window.removeEventListener("mousedown", listeningOnClose);
              }
          };

          innerDom.addEventListener("mousedown", function (e) {
              e.preventDefault();
              if (!this$1$1.selected || !this$1$1.enabled) { return; }
              markMenuEvent$1(e);
              if (open) {
                  close();
              } else {
                  open = this$1$1.expand(this$1$1.dom, contentDom);
                  window.addEventListener("mousedown", listeningOnClose = function () {
                      if (!isMenuEvent$1(this$1$1.dom)) { close(); }
                  });
              }
          });

          return this.dom;
      };

      Dropdown.prototype.renderItems = function renderItems (view) {
          var rendered = [];

          this.content.items.forEach(function (item) {
              var dom = item.render(view);
              rendered.push(crelt("div", {class: prefix$4 + "-dropdown-item"}, dom));
          });

          return rendered;
      };

      Dropdown.prototype.update = function update (state) {
          var contentUpdateResult = this.content.update(state);
          this.dom.style.display = contentUpdateResult ? "" : "none";

          var innerEnabled = false;
          var innerActive = false;

          this.content.items.forEach(function (item) {
              innerEnabled = innerEnabled || item.enabled;
              innerActive = innerActive || item.active;
          });

          this.adoptItemState(state, innerEnabled, innerActive);
          return contentUpdateResult;
      };

      Dropdown.prototype.expand = function expand (dom, contentDom) {
          var menuDOM = crelt("div", {class: prefix$4 + "-dropdown-menu " + (this.options.class || "")}, contentDom);



          var done = false;

          function close() {
              if (done) { return; }
              done = true;
              dom.removeChild(menuDOM);
              return true
          }

          dom.appendChild(menuDOM);

          var $menuDom = $(menuDOM);
          var right = $menuDom.offset().left + $menuDom.width() ;

          if(right > $(window).width() / 2) {
              $menuDom.addClass(prefix$4 + "-dropdown-right");
          } else {
              $menuDom.removeClass(prefix$4 + "-dropdown-right");
          }

          return {close: close, node: menuDOM}
      };

      return Dropdown;
  }(MenuItemGroup));

  // ::- Represents a submenu wrapping a group of elements that start
  // hidden and expand to the right when hovered over or tapped.
  var DropdownSubmenu$1 = /*@__PURE__*/(function (Dropdown) {
      function DropdownSubmenu(content, options) {
          Dropdown.call(this, content, options);
      }

      if ( Dropdown ) DropdownSubmenu.__proto__ = Dropdown;
      DropdownSubmenu.prototype = Object.create( Dropdown && Dropdown.prototype );
      DropdownSubmenu.prototype.constructor = DropdownSubmenu;

      // :: (EditorView) → {dom: dom.Node, update: (EditorState) → bool}
      // Renders the submenu.
      DropdownSubmenu.prototype.render = function render (view) {
          var this$1$1 = this;

          var itemDom = this.renderItems(view);

          var innerDom = $('<div>').addClass(prefix$4 + "-submenu-label").html(translate$1(view, this.options.label))[0];

          //let innerDom = crelt("div", {class: prefix + "-submenu-label"}, translate(view, this.options.label));
          this.dom = crelt("div", {class: prefix$4 + "-submenu-wrap"}, innerDom,
              crelt("div", {class: prefix$4 + "-submenu"}, itemDom));
          var listeningOnClose = null;

          innerDom.addEventListener("mousedown", function (e) {
              e.preventDefault();
              markMenuEvent$1(e);
              setClass$1(this$1$1.dom, prefix$4 + "-submenu-wrap-active");
              if (!listeningOnClose)
                  { window.addEventListener("mousedown", listeningOnClose = function () {
                      if (!isMenuEvent$1(this$1$1.dom)) {
                          this$1$1.dom.classList.remove(prefix$4 + "-submenu-wrap-active");
                          window.removeEventListener("mousedown", listeningOnClose);
                          listeningOnClose = null;
                      }
                  }); }
          });

          return this.dom;
      };

      DropdownSubmenu.prototype.update = function update (state) {
          var contentUpdateResult = this.content.update(state);
          this.dom.style.display = contentUpdateResult ? "" : "none";
          return contentUpdateResult;
      };

      return DropdownSubmenu;
  }(Dropdown$1));

  // :: Object
  // A set of basic editor-related icons. Contains the properties
  // `join`, `lift`, `selectParentNode`, `undo`, `redo`, `strong`, `em`,
  // `code`, `link`, `bulletList`, `orderedList`, and `blockquote`, each
  // holding an object that can be used as the `icon` option to
  // `MenuItem`.
  var icons$1 = {
      headline: {
          width: 27, height: 27,
          path: "M26.281 26c-1.375 0-2.766-0.109-4.156-0.109-1.375 0-2.75 0.109-4.125 0.109-0.531 0-0.781-0.578-0.781-1.031 0-1.391 1.563-0.797 2.375-1.328 0.516-0.328 0.516-1.641 0.516-2.188l-0.016-6.109c0-0.172 0-0.328-0.016-0.484-0.25-0.078-0.531-0.063-0.781-0.063h-10.547c-0.266 0-0.547-0.016-0.797 0.063-0.016 0.156-0.016 0.313-0.016 0.484l-0.016 5.797c0 0.594 0 2.219 0.578 2.562 0.812 0.5 2.656-0.203 2.656 1.203 0 0.469-0.219 1.094-0.766 1.094-1.453 0-2.906-0.109-4.344-0.109-1.328 0-2.656 0.109-3.984 0.109-0.516 0-0.75-0.594-0.75-1.031 0-1.359 1.437-0.797 2.203-1.328 0.5-0.344 0.516-1.687 0.516-2.234l-0.016-0.891v-12.703c0-0.75 0.109-3.156-0.594-3.578-0.781-0.484-2.453 0.266-2.453-1.141 0-0.453 0.203-1.094 0.75-1.094 1.437 0 2.891 0.109 4.328 0.109 1.313 0 2.641-0.109 3.953-0.109 0.562 0 0.781 0.625 0.781 1.094 0 1.344-1.547 0.688-2.312 1.172-0.547 0.328-0.547 1.937-0.547 2.5l0.016 5c0 0.172 0 0.328 0.016 0.5 0.203 0.047 0.406 0.047 0.609 0.047h10.922c0.187 0 0.391 0 0.594-0.047 0.016-0.172 0.016-0.328 0.016-0.5l0.016-5c0-0.578 0-2.172-0.547-2.5-0.781-0.469-2.344 0.156-2.344-1.172 0-0.469 0.219-1.094 0.781-1.094 1.375 0 2.75 0.109 4.125 0.109 1.344 0 2.688-0.109 4.031-0.109 0.562 0 0.781 0.625 0.781 1.094 0 1.359-1.609 0.672-2.391 1.156-0.531 0.344-0.547 1.953-0.547 2.516l0.016 14.734c0 0.516 0.031 1.875 0.531 2.188 0.797 0.5 2.484-0.141 2.484 1.219 0 0.453-0.203 1.094-0.75 1.094z"
      },
      plus: {
          width: 32, height: 32,
          path: "M31 12h-11v-11c0-0.552-0.448-1-1-1h-6c-0.552 0-1 0.448-1 1v11h-11c-0.552 0-1 0.448-1 1v6c0 0.552 0.448 1 1 1h11v11c0 0.552 0.448 1 1 1h6c0.552 0 1-0.448 1-1v-11h11c0.552 0 1-0.448 1-1v-6c0-0.552-0.448-1-1-1z"
      },
      table: {
          width: 32, height: 32,
          path: "M0 2v28h32v-28h-32zM12 20v-6h8v6h-8zM20 22v6h-8v-6h8zM20 6v6h-8v-6h8zM10 6v6h-8v-6h8zM2 14h8v6h-8v-6zM22 14h8v6h-8v-6zM22 12v-6h8v6h-8zM2 22h8v6h-8v-6zM22 28v-6h8v6h-8z"
      },
      join: {
          width: 800, height: 900,
          path: "M0 75h800v125h-800z M0 825h800v-125h-800z M250 400h100v-100h100v100h100v100h-100v100h-100v-100h-100z"
      },
      lift: {
          width: 1024, height: 1024,
          path: "M219 310v329q0 7-5 12t-12 5q-8 0-13-5l-164-164q-5-5-5-13t5-13l164-164q5-5 13-5 7 0 12 5t5 12zM1024 749v109q0 7-5 12t-12 5h-987q-7 0-12-5t-5-12v-109q0-7 5-12t12-5h987q7 0 12 5t5 12zM1024 530v109q0 7-5 12t-12 5h-621q-7 0-12-5t-5-12v-109q0-7 5-12t12-5h621q7 0 12 5t5 12zM1024 310v109q0 7-5 12t-12 5h-621q-7 0-12-5t-5-12v-109q0-7 5-12t12-5h621q7 0 12 5t5 12zM1024 91v109q0 7-5 12t-12 5h-987q-7 0-12-5t-5-12v-109q0-7 5-12t12-5h987q7 0 12 5t5 12z"
      },
      indent: {
          width: 28, height: 28,
          path: "M5.5 13c0 0.125-0.047 0.266-0.141 0.359l-4.5 4.5c-0.094 0.094-0.234 0.141-0.359 0.141-0.266 0-0.5-0.234-0.5-0.5v-9c0-0.266 0.234-0.5 0.5-0.5 0.125 0 0.266 0.047 0.359 0.141l4.5 4.5c0.094 0.094 0.141 0.234 0.141 0.359zM28 20.5v3c0 0.266-0.234 0.5-0.5 0.5h-27c-0.266 0-0.5-0.234-0.5-0.5v-3c0-0.266 0.234-0.5 0.5-0.5h27c0.266 0 0.5 0.234 0.5 0.5zM28 14.5v3c0 0.266-0.234 0.5-0.5 0.5h-17c-0.266 0-0.5-0.234-0.5-0.5v-3c0-0.266 0.234-0.5 0.5-0.5h17c0.266 0 0.5 0.234 0.5 0.5zM28 8.5v3c0 0.266-0.234 0.5-0.5 0.5h-17c-0.266 0-0.5-0.234-0.5-0.5v-3c0-0.266 0.234-0.5 0.5-0.5h17c0.266 0 0.5 0.234 0.5 0.5zM28 2.5v3c0 0.266-0.234 0.5-0.5 0.5h-27c-0.266 0-0.5-0.234-0.5-0.5v-3c0-0.266 0.234-0.5 0.5-0.5h27c0.266 0 0.5 0.234 0.5 0.5z"
      },
      outdent: {
          width: 28, height: 28,
          path: "M6 8.5v9c0 0.266-0.234 0.5-0.5 0.5-0.125 0-0.266-0.047-0.359-0.141l-4.5-4.5c-0.094-0.094-0.141-0.234-0.141-0.359s0.047-0.266 0.141-0.359l4.5-4.5c0.094-0.094 0.234-0.141 0.359-0.141 0.266 0 0.5 0.234 0.5 0.5zM28 20.5v3c0 0.266-0.234 0.5-0.5 0.5h-27c-0.266 0-0.5-0.234-0.5-0.5v-3c0-0.266 0.234-0.5 0.5-0.5h27c0.266 0 0.5 0.234 0.5 0.5zM28 14.5v3c0 0.266-0.234 0.5-0.5 0.5h-17c-0.266 0-0.5-0.234-0.5-0.5v-3c0-0.266 0.234-0.5 0.5-0.5h17c0.266 0 0.5 0.234 0.5 0.5zM28 8.5v3c0 0.266-0.234 0.5-0.5 0.5h-17c-0.266 0-0.5-0.234-0.5-0.5v-3c0-0.266 0.234-0.5 0.5-0.5h17c0.266 0 0.5 0.234 0.5 0.5zM28 2.5v3c0 0.266-0.234 0.5-0.5 0.5h-27c-0.266 0-0.5-0.234-0.5-0.5v-3c0-0.266 0.234-0.5 0.5-0.5h27c0.266 0 0.5 0.234 0.5 0.5z"
      },
      selectParentNode: {text: "\u2b1a", css: "font-weight: bold"},
      undo: {
          width: 1024, height: 1024,
          path: "M761 1024c113-206 132-520-313-509v253l-384-384 384-384v248c534-13 594 472 313 775z"
      },
      redo: {
          width: 1024, height: 1024,
          path: "M576 248v-248l384 384-384 384v-253c-446-10-427 303-313 509-280-303-221-789 313-775z"
      },
      strong: {
          width: 805, height: 1024,
          path: "M317 869q42 18 80 18 214 0 214-191 0-65-23-102-15-25-35-42t-38-26-46-14-48-6-54-1q-41 0-57 5 0 30-0 90t-0 90q0 4-0 38t-0 55 2 47 6 38zM309 442q24 4 62 4 46 0 81-7t62-25 42-51 14-81q0-40-16-70t-45-46-61-24-70-8q-28 0-74 7 0 28 2 86t2 86q0 15-0 45t-0 45q0 26 0 39zM0 950l1-53q8-2 48-9t60-15q4-6 7-15t4-19 3-18 1-21 0-19v-37q0-561-12-585-2-4-12-8t-25-6-28-4-27-2-17-1l-2-47q56-1 194-6t213-5q13 0 39 0t38 0q40 0 78 7t73 24 61 40 42 59 16 78q0 29-9 54t-22 41-36 32-41 25-48 22q88 20 146 76t58 141q0 57-20 102t-53 74-78 48-93 27-100 8q-25 0-75-1t-75-1q-60 0-175 6t-132 6z"
      },
      em: {
          width: 585, height: 1024,
          path: "M0 949l9-48q3-1 46-12t63-21q16-20 23-57 0-4 35-165t65-310 29-169v-14q-13-7-31-10t-39-4-33-3l10-58q18 1 68 3t85 4 68 1q27 0 56-1t69-4 56-3q-2 22-10 50-17 5-58 16t-62 19q-4 10-8 24t-5 22-4 26-3 24q-15 84-50 239t-44 203q-1 5-7 33t-11 51-9 47-3 32l0 10q9 2 105 17-1 25-9 56-6 0-18 0t-18 0q-16 0-49-5t-49-5q-78-1-117-1-29 0-81 5t-69 6z"
      },
      emoji: {
          width: 20, height: 20,
          path: "M10 0.4c-5.302 0-9.6 4.298-9.6 9.6s4.298 9.6 9.6 9.6c5.301 0 9.6-4.298 9.6-9.601 0-5.301-4.299-9.599-9.6-9.599zM10 17.599c-4.197 0-7.6-3.402-7.6-7.6s3.402-7.599 7.6-7.599c4.197 0 7.601 3.402 7.601 7.6s-3.404 7.599-7.601 7.599zM7.501 9.75c0.828 0 1.499-0.783 1.499-1.75s-0.672-1.75-1.5-1.75-1.5 0.783-1.5 1.75 0.672 1.75 1.501 1.75zM12.5 9.75c0.829 0 1.5-0.783 1.5-1.75s-0.672-1.75-1.5-1.75-1.5 0.784-1.5 1.75 0.672 1.75 1.5 1.75zM14.341 11.336c-0.363-0.186-0.815-0.043-1.008 0.32-0.034 0.066-0.869 1.593-3.332 1.593-2.451 0-3.291-1.513-3.333-1.592-0.188-0.365-0.632-0.514-1.004-0.329-0.37 0.186-0.52 0.636-0.335 1.007 0.050 0.099 1.248 2.414 4.672 2.414 3.425 0 4.621-2.316 4.67-2.415 0.184-0.367 0.036-0.81-0.33-0.998z"
      },
      code: {
          width: 896, height: 1024,
          path: "M608 192l-96 96 224 224-224 224 96 96 288-320-288-320zM288 192l-288 320 288 320 96-96-224-224 224-224-96-96z"
      },
      embed: {
          width: 40, height: 32,
          path: [
              'M26 23l3 3 10-10-10-10-3 3 7 7z',
              'M14 9l-3-3-10 10 10 10 3-3-7-7z',
              'M21.916 4.704l2.171 0.592-6 22.001-2.171-0.592 6-22.001z'
          ]
      },
      text: {
          width: 768, height: 768,
          path: [
              'M688.5 288v96h-96v223.5h-96v-223.5h-96v-96h288z',
              'M79.5 127.5h417v96h-160.5v384h-96v-384h-160.5v-96z'
          ]
      },
      image: {
          width: 512, height: 512,
          path: [
              'M479.942 64c0.020 0.017 0.041 0.038 0.058 0.058v383.885c-0.017 0.020-0.038 0.041-0.058 0.058h-447.885c-0.020-0.017-0.041-0.038-0.057-0.058v-383.886c0.017-0.020 0.038-0.041 0.057-0.057h447.885zM480 32h-448c-17.6 0-32 14.4-32 32v384c0 17.6 14.4 32 32 32h448c17.6 0 32-14.4 32-32v-384c0-17.6-14.4-32-32-32v0z',
              'M416 144c0 26.51-21.49 48-48 48s-48-21.49-48-48 21.49-48 48-48 48 21.49 48 48z',
              'M448 416h-384v-64l112-192 128 160h32l112-96z'
          ]
      },
      add: {
          width: 22, height: 28,
          path: "M18 12.5v1c0 0.281-0.219 0.5-0.5 0.5h-5.5v5.5c0 0.281-0.219 0.5-0.5 0.5h-1c-0.281 0-0.5-0.219-0.5-0.5v-5.5h-5.5c-0.281 0-0.5-0.219-0.5-0.5v-1c0-0.281 0.219-0.5 0.5-0.5h5.5v-5.5c0-0.281 0.219-0.5 0.5-0.5h1c0.281 0 0.5 0.219 0.5 0.5v5.5h5.5c0.281 0 0.5 0.219 0.5 0.5zM20 19.5v-13c0-1.375-1.125-2.5-2.5-2.5h-13c-1.375 0-2.5 1.125-2.5 2.5v13c0 1.375 1.125 2.5 2.5 2.5h13c1.375 0 2.5-1.125 2.5-2.5zM22 6.5v13c0 2.484-2.016 4.5-4.5 4.5h-13c-2.484 0-4.5-2.016-4.5-4.5v-13c0-2.484 2.016-4.5 4.5-4.5h13c2.484 0 4.5 2.016 4.5 4.5z"
      },
      link: {
          width: 951, height: 1024,
          path: "M832 694q0-22-16-38l-118-118q-16-16-38-16-24 0-41 18 1 1 10 10t12 12 8 10 7 14 2 15q0 22-16 38t-38 16q-8 0-15-2t-14-7-10-8-12-12-10-10q-18 17-18 41 0 22 16 38l117 118q15 15 38 15 22 0 38-14l84-83q16-16 16-38zM430 292q0-22-16-38l-117-118q-16-16-38-16-22 0-38 15l-84 83q-16 16-16 38 0 22 16 38l118 118q15 15 38 15 24 0 41-17-1-1-10-10t-12-12-8-10-7-14-2-15q0-22 16-38t38-16q8 0 15 2t14 7 10 8 12 12 10 10q18-17 18-41zM941 694q0 68-48 116l-84 83q-47 47-116 47-69 0-116-48l-117-118q-47-47-47-116 0-70 50-119l-50-50q-49 50-118 50-68 0-116-48l-118-118q-48-48-48-116t48-116l84-83q47-47 116-47 69 0 116 48l117 118q47 47 47 116 0 70-50 119l50 50q49-50 118-50 68 0 116 48l118 118q48 48 48 116z"
      },
      bulletList: {
          width: 768, height: 896,
          path: "M0 512h128v-128h-128v128zM0 256h128v-128h-128v128zM0 768h128v-128h-128v128zM256 512h512v-128h-512v128zM256 256h512v-128h-512v128zM256 768h512v-128h-512v128z"
      },
      orderedList: {
          width: 768, height: 896,
          path: "M320 512h448v-128h-448v128zM320 768h448v-128h-448v128zM320 128v128h448v-128h-448zM79 384h78v-256h-36l-85 23v50l43-2v185zM189 590c0-36-12-78-96-78-33 0-64 6-83 16l1 66c21-10 42-15 67-15s32 11 32 28c0 26-30 58-110 112v50h192v-67l-91 2c49-30 87-66 87-113l1-1z"
      },
      blockquote: {
          width: 640, height: 896,
          path: "M0 448v256h256v-256h-128c0 0 0-128 128-128v-128c0 0-256 0-256 256zM640 320v-128c0 0-256 0-256 256v256h256v-256h-128c0 0 0-128 128-128z"
      },
      strikethrough: {
          width: 28, height: 28,
          path: "M27.5 14c0.281 0 0.5 0.219 0.5 0.5v1c0 0.281-0.219 0.5-0.5 0.5h-27c-0.281 0-0.5-0.219-0.5-0.5v-1c0-0.281 0.219-0.5 0.5-0.5h27zM7.547 13c-0.297-0.375-0.562-0.797-0.797-1.25-0.5-1.016-0.75-2-0.75-2.938 0-1.906 0.703-3.5 2.094-4.828s3.437-1.984 6.141-1.984c0.594 0 1.453 0.109 2.609 0.297 0.688 0.125 1.609 0.375 2.766 0.75 0.109 0.406 0.219 1.031 0.328 1.844 0.141 1.234 0.219 2.187 0.219 2.859 0 0.219-0.031 0.453-0.078 0.703l-0.187 0.047-1.313-0.094-0.219-0.031c-0.531-1.578-1.078-2.641-1.609-3.203-0.922-0.953-2.031-1.422-3.281-1.422-1.188 0-2.141 0.313-2.844 0.922s-1.047 1.375-1.047 2.281c0 0.766 0.344 1.484 1.031 2.188s2.141 1.375 4.359 2.016c0.75 0.219 1.641 0.562 2.703 1.031 0.562 0.266 1.062 0.531 1.484 0.812h-11.609zM15.469 17h6.422c0.078 0.438 0.109 0.922 0.109 1.437 0 1.125-0.203 2.234-0.641 3.313-0.234 0.578-0.594 1.109-1.109 1.625-0.375 0.359-0.938 0.781-1.703 1.266-0.781 0.469-1.563 0.828-2.391 1.031-0.828 0.219-1.875 0.328-3.172 0.328-0.859 0-1.891-0.031-3.047-0.359l-2.188-0.625c-0.609-0.172-0.969-0.313-1.125-0.438-0.063-0.063-0.125-0.172-0.125-0.344v-0.203c0-0.125 0.031-0.938-0.031-2.438-0.031-0.781 0.031-1.328 0.031-1.641v-0.688l1.594-0.031c0.578 1.328 0.844 2.125 1.016 2.406 0.375 0.609 0.797 1.094 1.25 1.469s1 0.672 1.641 0.891c0.625 0.234 1.328 0.344 2.063 0.344 0.656 0 1.391-0.141 2.172-0.422 0.797-0.266 1.437-0.719 1.906-1.344 0.484-0.625 0.734-1.297 0.734-2.016 0-0.875-0.422-1.687-1.266-2.453-0.344-0.297-1.062-0.672-2.141-1.109z"
      },
      enlarge: {
          width:32, height: 32,
          path: "M32 0v13l-5-5-6 6-3-3 6-6-5-5zM14 21l-6 6 5 5h-13v-13l5 5 6-6z"
      },
      angleDoubleRight: {
          width:16, height: 28,
          path: "M9.297 15c0 0.125-0.063 0.266-0.156 0.359l-7.281 7.281c-0.094 0.094-0.234 0.156-0.359 0.156s-0.266-0.063-0.359-0.156l-0.781-0.781c-0.094-0.094-0.156-0.234-0.156-0.359s0.063-0.266 0.156-0.359l6.141-6.141-6.141-6.141c-0.094-0.094-0.156-0.234-0.156-0.359s0.063-0.266 0.156-0.359l0.781-0.781c0.094-0.094 0.234-0.156 0.359-0.156s0.266 0.063 0.359 0.156l7.281 7.281c0.094 0.094 0.156 0.234 0.156 0.359zM15.297 15c0 0.125-0.063 0.266-0.156 0.359l-7.281 7.281c-0.094 0.094-0.234 0.156-0.359 0.156s-0.266-0.063-0.359-0.156l-0.781-0.781c-0.094-0.094-0.156-0.234-0.156-0.359s0.063-0.266 0.156-0.359l6.141-6.141-6.141-6.141c-0.094-0.094-0.156-0.234-0.156-0.359s0.063-0.266 0.156-0.359l0.781-0.781c0.094-0.094 0.234-0.156 0.359-0.156s0.266 0.063 0.359 0.156l7.281 7.281c0.094 0.094 0.156 0.234 0.156 0.359z"
      },
      angleDoubleLeft: {
          width:16, height: 28,
          path: "M9.797 21.5c0 0.125-0.063 0.266-0.156 0.359l-0.781 0.781c-0.094 0.094-0.234 0.156-0.359 0.156s-0.266-0.063-0.359-0.156l-7.281-7.281c-0.094-0.094-0.156-0.234-0.156-0.359s0.063-0.266 0.156-0.359l7.281-7.281c0.094-0.094 0.234-0.156 0.359-0.156s0.266 0.063 0.359 0.156l0.781 0.781c0.094 0.094 0.156 0.234 0.156 0.359s-0.063 0.266-0.156 0.359l-6.141 6.141 6.141 6.141c0.094 0.094 0.156 0.234 0.156 0.359zM15.797 21.5c0 0.125-0.063 0.266-0.156 0.359l-0.781 0.781c-0.094 0.094-0.234 0.156-0.359 0.156s-0.266-0.063-0.359-0.156l-7.281-7.281c-0.094-0.094-0.156-0.234-0.156-0.359s0.063-0.266 0.156-0.359l7.281-7.281c0.094-0.094 0.234-0.156 0.359-0.156s0.266 0.063 0.359 0.156l0.781 0.781c0.094 0.094 0.156 0.234 0.156 0.359s-0.063 0.266-0.156 0.359l-6.141 6.141 6.141 6.141c0.094 0.094 0.156 0.234 0.156 0.359z"
      },
      shrink: {
          width:32, height: 32,
          path: "M14 18v13l-5-5-6 6-3-3 6-6-5-5zM32 3l-6 6 5 5h-13v-13l5 5 6-6z"
      }
  };

  // :: MenuItem
  // Menu item for the `joinUp` command.
  var joinUpItem$1 = function () {
      return new MenuItem$1({
          title: "Join with above block",
          run: joinUp,
          select: function (state) { return joinUp(state); },
          icon: icons$1.join
      });
  };

  // :: MenuItem
  // Menu item for the `lift` command.
  var liftItem$1 = function () {
      return new MenuItem$1({
          title: "Lift out of enclosing block",
          run: lift,
          select: function (state) { return lift(state); },
          icon: icons$1.outdent
      });
  };

  function lift(state, dispatch) {
      var ref = state.selection;
      var $from = ref.$from;
      var $to = ref.$to;

      var inList = $from.blockRange($to, function (node) {
          return node.childCount && node.firstChild.type.name === 'list_item';
      });

      if(inList) {
          return false;
      }

      var range = $from.blockRange($to), target = range && liftTarget(range);
      if (target == null) { return false }
      if (dispatch) { dispatch(state.tr.lift(range, target).scrollIntoView()); }
      return true
  }

  // :: MenuItem
  // Menu item for the `selectParentNode` command.
  var selectParentNodeItem$1 = function () {
      return new MenuItem$1({
          title: "Select parent node",
          run: selectParentNode,
          select: function (state) { return selectParentNode(state); },
          icon: icons$1.selectParentNode
      });
  };

  // :: (Object) → MenuItem
  // Menu item for the `undo` command.
  var undoItem$1 = function () {
      return new MenuItem$1({
          title: "Undo last change",
          run: undo,
          enable: function (state) { return undo(state); },
          icon: icons$1.undo
      });
  };

  // :: (Object) → MenuItem
  // Menu item for the `redo` command.
  var redoItem$1 = function () {
      return new MenuItem$1({
          title: "Redo last undone change",
          run: redo,
          enable: function (state) { return redo(state); },
          icon: icons$1.redo
      });
  };

  // :: (NodeType, Object) → MenuItem
  // Build a menu item for wrapping the selection in a given node type.
  // Adds `run` and `select` properties to the ones present in
  // `options`. `options.attrs` may be an object or a function, as in
  // `toggleMarkItem`.
  function wrapItem$1(nodeType, options) {
      var passedOptions = {
          run: function run(state, dispatch) {
              // FIXME if (options.attrs instanceof Function) options.attrs(state, attrs => wrapIn(nodeType, attrs)(state))
              return wrapIn(nodeType, options.attrs)(state, dispatch)
          },
          select: function select(state) {
              return wrapIn(nodeType, options.attrs instanceof Function ? null : options.attrs)(state)
          }
      };
      for (var prop in options) { passedOptions[prop] = options[prop]; }
      return new MenuItem$1(passedOptions)
  }

  // :: (NodeType, Object) → MenuItem
  // Build a menu item for changing the type of the textblock around the
  // selection to the given type. Provides `run`, `active`, and `select`
  // properties. Others must be given in `options`. `options.attrs` may
  // be an object to provide the attributes for the textblock node.
  function blockTypeItem$1(nodeType, options) {
      var command = setBlockType(nodeType, options.attrs);
      var passedOptions = {
          run: command,
          enable: function enable(state) {
              return command(state)
          },
          active: function active(state) {
              var ref = state.selection;
              var $from = ref.$from;
              var to = ref.to;
              var node = ref.node;
              if (node) { return node.hasMarkup(nodeType, options.attrs) }
              return to <= $from.end() && $from.parent.hasMarkup(nodeType, options.attrs)
          }
      };
      for (var prop in options) { passedOptions[prop] = options[prop]; }
      return new MenuItem$1(passedOptions)
  }

  // Work around classList.toggle being broken in IE11
  function setClass$1(dom, cls, on) {
      if (on) { dom.classList.add(cls); }
      else { dom.classList.remove(cls); }
  }

  function canInsert(state, nodeType) {
      var $from = state.selection.$from;
      for (var d = $from.depth; d >= 0; d--) {
          var index = $from.index(d);
          if ($from.node(d).canReplaceWith(index, index, nodeType)) { return true }
      }
      return false
  }

  function canInsertLink(state) {
      var allowLink = true;
      state.doc.nodesBetween(state.selection.$from.pos, state.selection.$to.pos, function(node) {
          if(node.type.spec.code) {
              allowLink = false;
          } else {
              node.marks.forEach(function(mark) {
                  var spec = mark.type.spec;
                  if(spec.preventMarks && $.inArray('link', spec.preventMarks) >= 0) {
                      allowLink = false;
                  }
              });
          }
      });

      return allowLink;
  }

  var prefix$3 = "ProseMirror-menubar";

  function buildMenuItems(context) {
      var groups = {
          types:  {type: 'dropdown', sortOrder: 100, label: context.translate("Type"), seperator: true, icon: icons$1.text, items: []},
          marks:  {type: 'group', id: 'marks-group', sortOrder: 200, items: []},
          format:  {type: 'group', id: 'format-group',  sortOrder: 300, items: [liftItem$1()]},
          insert: {type: 'dropdown', id: 'insert-dropdown',  sortOrder: 400, label: context.translate("Insert"), seperator: true, icon: icons$1.image, items: []},
          helper:  {type: 'group', id: 'helper-group', sortOrder: 500, items: [undoItem$1(), redoItem$1()]},
          resize:  {type: 'group', id: 'resize-group', sortOrder: 600, items: []},
      };

      var definitions = [groups.types, groups.insert, groups.marks, groups.format, groups.helper, groups.resize];

      context.plugins.forEach(function (plugin) {
          if(plugin.menu) {
              plugin.menu(context).forEach(function(menuDefinition) {
                  if(checkMenuDefinition(context, menuDefinition)) {
                      menuDefinition.item.options.id = menuDefinition.id;

                      if(menuDefinition.group && groups[menuDefinition.group]) {
                          groups[menuDefinition.group].items.push(menuDefinition.item);
                      } else if(!menuDefinition.group) {
                          definitions.push(menuDefinition.item);
                      }
                  }
              });
          }
      });

      //selectParentNodeItem -> don't know if we should add this one

      // TODO: fire event
      return definitions;
  }

  function checkMenuDefinition(context, menuDefinition) {
      if(!menuDefinition || menuDefinition.node && !context.schema.nodes[menuDefinition.node]) {
          return false;
      }

      if(menuDefinition.mark && !context.schema.marks[menuDefinition.mark]) {
          return false;
      }

      if(context.options.menu && Array.isArray(context.options.menu.exclude) && context.options.menu.exclude[menuDefinition.id]) {
          return false;
      }

      return true;
  }

  function buildMenuBar(context) {
      context.menu = menuBar$1({
          content: buildMenuItems(context),
          floating: false,
          context: context
      });

      return context.menu;
  }


  function isIOS$1() {
      if (typeof navigator == "undefined") { return false }
      var agent = navigator.userAgent;
      return !/Edge\/\d/.test(agent) && /AppleWebKit/.test(agent) && /Mobile\/\w+/.test(agent)
  }

  // :: (Object) → Plugin
  // A plugin that will place a menu bar above the editor. Note that
  // this involves wrapping the editor in an additional `<div>`.
  //
  //   options::-
  //   Supports the following options:
  //
  //     content:: [[MenuElement]]
  //     Provides the content of the menu, as a nested array to be
  //     passed to `renderGrouped`.
  //
  //     floating:: ?bool
  //     Determines whether the menu floats, i.e. whether it sticks to
  //     the top of the viewport when the editor is partially scrolled
  //     out of view.
  function menuBar$1(options) {
      return new Plugin({
          view: function view(editorView) {
              options.context.menu = new MenuBarView$1(editorView, options);
              options.context.event.trigger('afterMenuBarInit', options.context.menu);
              return options.context.menu;
          }
      })
  }

  var MenuBarView$1 = function MenuBarView(editorView, options) {
      var this$1$1 = this;

      this.editorView = editorView;
      this.options = options;
      this.context = this.options.context;

      this.wrapper = crelt("div", {class: prefix$3 + "-wrapper"});
      this.menu = this.wrapper.appendChild(crelt("div", {class: prefix$3}));
      this.menu.className = prefix$3;
      this.spacer = null;

      editorView.dom.parentNode.replaceChild(this.wrapper, editorView.dom);
      this.wrapper.appendChild(editorView.dom);

      this.maxHeight = 0;
      this.widthForMaxHeight = 0;
      this.floating = false;

      this.groupItem = new MenuItemGroup(this.options.content);
      var dom = this.groupItem.render(this.editorView);

      this.menu.appendChild(dom);

      $(this.menu).on('mousedown', function(evt) {
          // Prevent focusout if we click outside of a menu item, but still inside menu container
          evt.preventDefault();
      });

      this.update();

      if (options.floating && !isIOS$1()) {
          this.updateFloat();
          this.scrollFunc = function () {
              var root = this$1$1.editorView.root;
              if (!(root.body || root).contains(this$1$1.wrapper))
                  { window.removeEventListener("scroll", this$1$1.scrollFunc); }
              else
                  { this$1$1.updateFloat(); }
          };
          window.addEventListener("scroll", this.scrollFunc);
      }
  };

  MenuBarView$1.prototype.update = function update () {
      this.groupItem.update(this.editorView.state);

      var $mainGroup = $(this.menu).find('.'+prefix$3+'-menu-group:first');
      $mainGroup.find('');

      if (this.floating) {
          this.updateScrollCursor();
      } else {
          if (this.menu.offsetWidth != this.widthForMaxHeight) {
              this.widthForMaxHeight = this.menu.offsetWidth;
              this.maxHeight = 0;
          }
          if (this.menu.offsetHeight > this.maxHeight) {
              this.maxHeight = this.menu.offsetHeight;
          }
      }
      this.context.event.trigger('afterMenuBarUpdate', this);
  };

  MenuBarView$1.prototype.updateScrollCursor = function updateScrollCursor () {
      var selection = this.editorView.root.getSelection();
      if (!selection.focusNode) { return; }
      var rects = selection.getRangeAt(0).getClientRects();
      var selRect = rects[selectionIsInverted$1(selection) ? 0 : rects.length - 1];
      if (!selRect) { return; }
      var menuRect = this.menu.getBoundingClientRect();
      if (selRect.top < menuRect.bottom && selRect.bottom > menuRect.top) {
          var scrollable = findWrappingScrollable$1(this.wrapper);
          if (scrollable) { scrollable.scrollTop -= (menuRect.bottom - selRect.top); }
      }
  };

  MenuBarView$1.prototype.updateFloat = function updateFloat () {
      var parent = this.wrapper, editorRect = parent.getBoundingClientRect();
      if (this.floating) {
          if (editorRect.top >= 0 || editorRect.bottom < this.menu.offsetHeight + 10) {
              this.floating = false;
              this.menu.style.position = this.menu.style.left = this.menu.style.width = "";
              this.menu.style.display = "";
              this.spacer.parentNode.removeChild(this.spacer);
              this.spacer = null;
          } else {
              var border = (parent.offsetWidth - parent.clientWidth) / 2;
              this.menu.style.left = (editorRect.left + border) + "px";
              this.menu.style.display = (editorRect.top > window.innerHeight ? "none" : "");
          }
      } else {
          if (editorRect.top < 0 && editorRect.bottom >= this.menu.offsetHeight + 10) {
              this.floating = true;
              var menuRect = this.menu.getBoundingClientRect();
              this.menu.style.left = menuRect.left + "px";
              this.menu.style.width = menuRect.width + "px";
              this.menu.style.position = "fixed";
              this.spacer = crel("div", {class: prefix$3 + "-spacer", style: ("height: " + (menuRect.height) + "px")});
              parent.insertBefore(this.spacer, this.menu);
          }
      }
  };

  MenuBarView$1.prototype.destroy = function destroy () {
      if (this.wrapper.parentNode)
          { this.wrapper.parentNode.replaceChild(this.editorView.dom, this.wrapper); }
  };

  // Not precise, but close enough
  function selectionIsInverted$1(selection) {
      if (selection.anchorNode == selection.focusNode) { return selection.anchorOffset > selection.focusOffset; }
      return selection.anchorNode.compareDocumentPosition(selection.focusNode) == Node.DOCUMENT_POSITION_FOLLOWING;
  }

  function findWrappingScrollable$1(node) {
      for (var cur = node.parentNode; cur; cur = cur.parentNode)
          { if (cur.scrollHeight > cur.clientHeight) { return cur } }
  }

  var menu$j = /*#__PURE__*/Object.freeze({
    __proto__: null,
    menuBar: menuBar$1,
    cmdItem: cmdItem,
    markItem: markItem,
    markActive: markActive,
    wrapListItem: wrapListItem,
    MenuItem: MenuItem$1,
    MenuItemGroup: MenuItemGroup,
    Dropdown: Dropdown$1,
    DropdownSubmenu: DropdownSubmenu$1,
    icons: icons$1,
    joinUpItem: joinUpItem$1,
    liftItem: liftItem$1,
    selectParentNodeItem: selectParentNodeItem$1,
    undoItem: undoItem$1,
    redoItem: redoItem$1,
    wrapItem: wrapItem$1,
    blockTypeItem: blockTypeItem$1,
    canInsert: canInsert,
    canInsertLink: canInsertLink
  });

  var SVG = "http://www.w3.org/2000/svg";
  var XLINK = "http://www.w3.org/1999/xlink";

  var prefix$1 = "ProseMirror-icon";

  function hashPath(path) {
    var hash = 0;
    for (var i = 0; i < path.length; i++)
      { hash = (((hash << 5) - hash) + path.charCodeAt(i)) | 0; }
    return hash
  }

  function getIcon(icon) {
    var node = document.createElement("div");
    node.className = prefix$1;
    if (icon.path) {
      var name = "pm-icon-" + hashPath(icon.path).toString(16);
      if (!document.getElementById(name)) { buildSVG(name, icon); }
      var svg = node.appendChild(document.createElementNS(SVG, "svg"));
      svg.style.width = (icon.width / icon.height) + "em";
      var use = svg.appendChild(document.createElementNS(SVG, "use"));
      use.setAttributeNS(XLINK, "href", /([^#]*)/.exec(document.location)[1] + "#" + name);
    } else if (icon.dom) {
      node.appendChild(icon.dom.cloneNode(true));
    } else {
      node.appendChild(document.createElement("span")).textContent = icon.text || '';
      if (icon.css) { node.firstChild.style.cssText = icon.css; }
    }
    return node
  }

  function buildSVG(name, data) {
    var collection = document.getElementById(prefix$1 + "-collection");
    if (!collection) {
      collection = document.createElementNS(SVG, "svg");
      collection.id = prefix$1 + "-collection";
      collection.style.display = "none";
      document.body.insertBefore(collection, document.body.firstChild);
    }
    var sym = document.createElementNS(SVG, "symbol");
    sym.id = name;
    sym.setAttribute("viewBox", "0 0 " + data.width + " " + data.height);
    var path = sym.appendChild(document.createElementNS(SVG, "path"));
    path.setAttribute("d", data.path);
    collection.appendChild(sym);
  }

  var prefix$1$1 = "ProseMirror-menu";

  // ::- An icon or label that, when clicked, executes a command.
  var MenuItem = function MenuItem(spec) {
    // :: MenuItemSpec
    // The spec used to create the menu item.
    this.spec = spec;
  };

  // :: (EditorView) → {dom: dom.Node, update: (EditorState) → bool}
  // Renders the icon according to its [display
  // spec](#menu.MenuItemSpec.display), and adds an event handler which
  // executes the command when the representation is clicked.
  MenuItem.prototype.render = function render (view) {
    var spec = this.spec;
    var dom = spec.render ? spec.render(view)
        : spec.icon ? getIcon(spec.icon)
        : spec.label ? crelt("div", null, translate(view, spec.label))
        : null;
    if (!dom) { throw new RangeError("MenuItem without icon or label property") }
    if (spec.title) {
      var title = (typeof spec.title === "function" ? spec.title(view.state) : spec.title);
      dom.setAttribute("title", translate(view, title));
    }
    if (spec.class) { dom.classList.add(spec.class); }
    if (spec.css) { dom.style.cssText += spec.css; }

    dom.addEventListener("mousedown", function (e) {
      e.preventDefault();
      if (!dom.classList.contains(prefix$1$1 + "-disabled"))
        { spec.run(view.state, view.dispatch, view, e); }
    });

    function update(state) {
      if (spec.select) {
        var selected = spec.select(state);
        dom.style.display = selected ? "" : "none";
        if (!selected) { return false }
      }
      var enabled = true;
      if (spec.enable) {
        enabled = spec.enable(state) || false;
        setClass(dom, prefix$1$1 + "-disabled", !enabled);
      }
      if (spec.active) {
        var active = enabled && spec.active(state) || false;
        setClass(dom, prefix$1$1 + "-active", active);
      }
      return true
    }

    return {dom: dom, update: update}
  };

  function translate(view, text) {
    return view._props.translate ? view._props.translate(text) : text
  }

  // MenuItemSpec:: interface
  // The configuration object passed to the `MenuItem` constructor.
  //
  //   run:: (EditorState, (Transaction), EditorView, dom.Event)
  //   The function to execute when the menu item is activated.
  //
  //   select:: ?(EditorState) → bool
  //   Optional function that is used to determine whether the item is
  //   appropriate at the moment. Deselected items will be hidden.
  //
  //   enable:: ?(EditorState) → bool
  //   Function that is used to determine if the item is enabled. If
  //   given and returning false, the item will be given a disabled
  //   styling.
  //
  //   active:: ?(EditorState) → bool
  //   A predicate function to determine whether the item is 'active' (for
  //   example, the item for toggling the strong mark might be active then
  //   the cursor is in strong text).
  //
  //   render:: ?(EditorView) → dom.Node
  //   A function that renders the item. You must provide either this,
  //   [`icon`](#menu.MenuItemSpec.icon), or [`label`](#MenuItemSpec.label).
  //
  //   icon:: ?Object
  //   Describes an icon to show for this item. The object may specify
  //   an SVG icon, in which case its `path` property should be an [SVG
  //   path
  //   spec](https://developer.mozilla.org/en-US/docs/Web/SVG/Attribute/d),
  //   and `width` and `height` should provide the viewbox in which that
  //   path exists. Alternatively, it may have a `text` property
  //   specifying a string of text that makes up the icon, with an
  //   optional `css` property giving additional CSS styling for the
  //   text. _Or_ it may contain `dom` property containing a DOM node.
  //
  //   label:: ?string
  //   Makes the item show up as a text label. Mostly useful for items
  //   wrapped in a [drop-down](#menu.Dropdown) or similar menu. The object
  //   should have a `label` property providing the text to display.
  //
  //   title:: ?union<string, (EditorState) → string>
  //   Defines DOM title (mouseover) text for the item.
  //
  //   class:: ?string
  //   Optionally adds a CSS class to the item's DOM representation.
  //
  //   css:: ?string
  //   Optionally adds a string of inline CSS to the item's DOM
  //   representation.

  var lastMenuEvent = {time: 0, node: null};
  function markMenuEvent(e) {
    lastMenuEvent.time = Date.now();
    lastMenuEvent.node = e.target;
  }
  function isMenuEvent(wrapper) {
    return Date.now() - 100 < lastMenuEvent.time &&
      lastMenuEvent.node && wrapper.contains(lastMenuEvent.node)
  }

  // ::- A drop-down menu, displayed as a label with a downwards-pointing
  // triangle to the right of it.
  var Dropdown = function Dropdown(content, options) {
    this.options = options || {};
    this.content = Array.isArray(content) ? content : [content];
  };

  // :: (EditorView) → {dom: dom.Node, update: (EditorState)}
  // Render the dropdown menu and sub-items.
  Dropdown.prototype.render = function render (view) {
      var this$1$1 = this;

    var content = renderDropdownItems(this.content, view);

    var label = crelt("div", {class: prefix$1$1 + "-dropdown " + (this.options.class || ""),
                             style: this.options.css},
                     translate(view, this.options.label));
    if (this.options.title) { label.setAttribute("title", translate(view, this.options.title)); }
    var wrap = crelt("div", {class: prefix$1$1 + "-dropdown-wrap"}, label);
    var open = null, listeningOnClose = null;
    var close = function () {
      if (open && open.close()) {
        open = null;
        window.removeEventListener("mousedown", listeningOnClose);
      }
    };
    label.addEventListener("mousedown", function (e) {
      e.preventDefault();
      markMenuEvent(e);
      if (open) {
        close();
      } else {
        open = this$1$1.expand(wrap, content.dom);
        window.addEventListener("mousedown", listeningOnClose = function () {
          if (!isMenuEvent(wrap)) { close(); }
        });
      }
    });

    function update(state) {
      var inner = content.update(state);
      wrap.style.display = inner ? "" : "none";
      return inner
    }

    return {dom: wrap, update: update}
  };

  Dropdown.prototype.expand = function expand (dom, items) {
    var menuDOM = crelt("div", {class: prefix$1$1 + "-dropdown-menu " + (this.options.class || "")}, items);

    var done = false;
    function close() {
      if (done) { return }
      done = true;
      dom.removeChild(menuDOM);
      return true
    }
    dom.appendChild(menuDOM);
    return {close: close, node: menuDOM}
  };

  function renderDropdownItems(items, view) {
    var rendered = [], updates = [];
    for (var i = 0; i < items.length; i++) {
      var ref = items[i].render(view);
      var dom = ref.dom;
      var update = ref.update;
      rendered.push(crelt("div", {class: prefix$1$1 + "-dropdown-item"}, dom));
      updates.push(update);
    }
    return {dom: rendered, update: combineUpdates(updates, rendered)}
  }

  function combineUpdates(updates, nodes) {
    return function (state) {
      var something = false;
      for (var i = 0; i < updates.length; i++) {
        var up = updates[i](state);
        nodes[i].style.display = up ? "" : "none";
        if (up) { something = true; }
      }
      return something
    }
  }

  // ::- Represents a submenu wrapping a group of elements that start
  // hidden and expand to the right when hovered over or tapped.
  var DropdownSubmenu = function DropdownSubmenu(content, options) {
    this.options = options || {};
    this.content = Array.isArray(content) ? content : [content];
  };

  // :: (EditorView) → {dom: dom.Node, update: (EditorState) → bool}
  // Renders the submenu.
  DropdownSubmenu.prototype.render = function render (view) {
    var items = renderDropdownItems(this.content, view);

    var label = crelt("div", {class: prefix$1$1 + "-submenu-label"}, translate(view, this.options.label));
    var wrap = crelt("div", {class: prefix$1$1 + "-submenu-wrap"}, label,
                   crelt("div", {class: prefix$1$1 + "-submenu"}, items.dom));
    var listeningOnClose = null;
    label.addEventListener("mousedown", function (e) {
      e.preventDefault();
      markMenuEvent(e);
      setClass(wrap, prefix$1$1 + "-submenu-wrap-active");
      if (!listeningOnClose)
        { window.addEventListener("mousedown", listeningOnClose = function () {
          if (!isMenuEvent(wrap)) {
            wrap.classList.remove(prefix$1$1 + "-submenu-wrap-active");
            window.removeEventListener("mousedown", listeningOnClose);
            listeningOnClose = null;
          }
        }); }
    });

    function update(state) {
      var inner = items.update(state);
      wrap.style.display = inner ? "" : "none";
      return inner
    }
    return {dom: wrap, update: update}
  };

  // :: (EditorView, [union<MenuElement, [MenuElement]>]) → {dom: ?dom.DocumentFragment, update: (EditorState) → bool}
  // Render the given, possibly nested, array of menu elements into a
  // document fragment, placing separators between them (and ensuring no
  // superfluous separators appear when some of the groups turn out to
  // be empty).
  function renderGrouped(view, content) {
    var result = document.createDocumentFragment();
    var updates = [], separators = [];
    for (var i = 0; i < content.length; i++) {
      var items = content[i], localUpdates = [], localNodes = [];
      for (var j = 0; j < items.length; j++) {
        var ref = items[j].render(view);
        var dom = ref.dom;
        var update$1 = ref.update;
        var span = crelt("span", {class: prefix$1$1 + "item"}, dom);
        result.appendChild(span);
        localNodes.push(span);
        localUpdates.push(update$1);
      }
      if (localUpdates.length) {
        updates.push(combineUpdates(localUpdates, localNodes));
        if (i < content.length - 1)
          { separators.push(result.appendChild(separator())); }
      }
    }

    function update(state) {
      var something = false, needSep = false;
      for (var i = 0; i < updates.length; i++) {
        var hasContent = updates[i](state);
        if (i) { separators[i - 1].style.display = needSep && hasContent ? "" : "none"; }
        needSep = hasContent;
        if (hasContent) { something = true; }
      }
      return something
    }
    return {dom: result, update: update}
  }

  function separator() {
    return crelt("span", {class: prefix$1$1 + "separator"})
  }

  // :: Object
  // A set of basic editor-related icons. Contains the properties
  // `join`, `lift`, `selectParentNode`, `undo`, `redo`, `strong`, `em`,
  // `code`, `link`, `bulletList`, `orderedList`, and `blockquote`, each
  // holding an object that can be used as the `icon` option to
  // `MenuItem`.
  var icons = {
    join: {
      width: 800, height: 900,
      path: "M0 75h800v125h-800z M0 825h800v-125h-800z M250 400h100v-100h100v100h100v100h-100v100h-100v-100h-100z"
    },
    lift: {
      width: 1024, height: 1024,
      path: "M219 310v329q0 7-5 12t-12 5q-8 0-13-5l-164-164q-5-5-5-13t5-13l164-164q5-5 13-5 7 0 12 5t5 12zM1024 749v109q0 7-5 12t-12 5h-987q-7 0-12-5t-5-12v-109q0-7 5-12t12-5h987q7 0 12 5t5 12zM1024 530v109q0 7-5 12t-12 5h-621q-7 0-12-5t-5-12v-109q0-7 5-12t12-5h621q7 0 12 5t5 12zM1024 310v109q0 7-5 12t-12 5h-621q-7 0-12-5t-5-12v-109q0-7 5-12t12-5h621q7 0 12 5t5 12zM1024 91v109q0 7-5 12t-12 5h-987q-7 0-12-5t-5-12v-109q0-7 5-12t12-5h987q7 0 12 5t5 12z"
    },
    selectParentNode: {text: "\u2b1a", css: "font-weight: bold"},
    undo: {
      width: 1024, height: 1024,
      path: "M761 1024c113-206 132-520-313-509v253l-384-384 384-384v248c534-13 594 472 313 775z"
    },
    redo: {
      width: 1024, height: 1024,
      path: "M576 248v-248l384 384-384 384v-253c-446-10-427 303-313 509-280-303-221-789 313-775z"
    },
    strong: {
      width: 805, height: 1024,
      path: "M317 869q42 18 80 18 214 0 214-191 0-65-23-102-15-25-35-42t-38-26-46-14-48-6-54-1q-41 0-57 5 0 30-0 90t-0 90q0 4-0 38t-0 55 2 47 6 38zM309 442q24 4 62 4 46 0 81-7t62-25 42-51 14-81q0-40-16-70t-45-46-61-24-70-8q-28 0-74 7 0 28 2 86t2 86q0 15-0 45t-0 45q0 26 0 39zM0 950l1-53q8-2 48-9t60-15q4-6 7-15t4-19 3-18 1-21 0-19v-37q0-561-12-585-2-4-12-8t-25-6-28-4-27-2-17-1l-2-47q56-1 194-6t213-5q13 0 39 0t38 0q40 0 78 7t73 24 61 40 42 59 16 78q0 29-9 54t-22 41-36 32-41 25-48 22q88 20 146 76t58 141q0 57-20 102t-53 74-78 48-93 27-100 8q-25 0-75-1t-75-1q-60 0-175 6t-132 6z"
    },
    em: {
      width: 585, height: 1024,
      path: "M0 949l9-48q3-1 46-12t63-21q16-20 23-57 0-4 35-165t65-310 29-169v-14q-13-7-31-10t-39-4-33-3l10-58q18 1 68 3t85 4 68 1q27 0 56-1t69-4 56-3q-2 22-10 50-17 5-58 16t-62 19q-4 10-8 24t-5 22-4 26-3 24q-15 84-50 239t-44 203q-1 5-7 33t-11 51-9 47-3 32l0 10q9 2 105 17-1 25-9 56-6 0-18 0t-18 0q-16 0-49-5t-49-5q-78-1-117-1-29 0-81 5t-69 6z"
    },
    code: {
      width: 896, height: 1024,
      path: "M608 192l-96 96 224 224-224 224 96 96 288-320-288-320zM288 192l-288 320 288 320 96-96-224-224 224-224-96-96z"
    },
    link: {
      width: 951, height: 1024,
      path: "M832 694q0-22-16-38l-118-118q-16-16-38-16-24 0-41 18 1 1 10 10t12 12 8 10 7 14 2 15q0 22-16 38t-38 16q-8 0-15-2t-14-7-10-8-12-12-10-10q-18 17-18 41 0 22 16 38l117 118q15 15 38 15 22 0 38-14l84-83q16-16 16-38zM430 292q0-22-16-38l-117-118q-16-16-38-16-22 0-38 15l-84 83q-16 16-16 38 0 22 16 38l118 118q15 15 38 15 24 0 41-17-1-1-10-10t-12-12-8-10-7-14-2-15q0-22 16-38t38-16q8 0 15 2t14 7 10 8 12 12 10 10q18-17 18-41zM941 694q0 68-48 116l-84 83q-47 47-116 47-69 0-116-48l-117-118q-47-47-47-116 0-70 50-119l-50-50q-49 50-118 50-68 0-116-48l-118-118q-48-48-48-116t48-116l84-83q47-47 116-47 69 0 116 48l117 118q47 47 47 116 0 70-50 119l50 50q49-50 118-50 68 0 116 48l118 118q48 48 48 116z"
    },
    bulletList: {
      width: 768, height: 896,
      path: "M0 512h128v-128h-128v128zM0 256h128v-128h-128v128zM0 768h128v-128h-128v128zM256 512h512v-128h-512v128zM256 256h512v-128h-512v128zM256 768h512v-128h-512v128z"
    },
    orderedList: {
      width: 768, height: 896,
      path: "M320 512h448v-128h-448v128zM320 768h448v-128h-448v128zM320 128v128h448v-128h-448zM79 384h78v-256h-36l-85 23v50l43-2v185zM189 590c0-36-12-78-96-78-33 0-64 6-83 16l1 66c21-10 42-15 67-15s32 11 32 28c0 26-30 58-110 112v50h192v-67l-91 2c49-30 87-66 87-113l1-1z"
    },
    blockquote: {
      width: 640, height: 896,
      path: "M0 448v256h256v-256h-128c0 0 0-128 128-128v-128c0 0-256 0-256 256zM640 320v-128c0 0-256 0-256 256v256h256v-256h-128c0 0 0-128 128-128z"
    }
  };

  // :: MenuItem
  // Menu item for the `joinUp` command.
  var joinUpItem = new MenuItem({
    title: "Join with above block",
    run: joinUp,
    select: function (state) { return joinUp(state); },
    icon: icons.join
  });

  // :: MenuItem
  // Menu item for the `lift` command.
  var liftItem = new MenuItem({
    title: "Lift out of enclosing block",
    run: lift$1,
    select: function (state) { return lift$1(state); },
    icon: icons.lift
  });

  // :: MenuItem
  // Menu item for the `selectParentNode` command.
  var selectParentNodeItem = new MenuItem({
    title: "Select parent node",
    run: selectParentNode,
    select: function (state) { return selectParentNode(state); },
    icon: icons.selectParentNode
  });

  // :: MenuItem
  // Menu item for the `undo` command.
  var undoItem = new MenuItem({
    title: "Undo last change",
    run: undo,
    enable: function (state) { return undo(state); },
    icon: icons.undo
  });

  // :: MenuItem
  // Menu item for the `redo` command.
  var redoItem = new MenuItem({
    title: "Redo last undone change",
    run: redo,
    enable: function (state) { return redo(state); },
    icon: icons.redo
  });

  // :: (NodeType, Object) → MenuItem
  // Build a menu item for wrapping the selection in a given node type.
  // Adds `run` and `select` properties to the ones present in
  // `options`. `options.attrs` may be an object or a function.
  function wrapItem(nodeType, options) {
    var passedOptions = {
      run: function run(state, dispatch) {
        // FIXME if (options.attrs instanceof Function) options.attrs(state, attrs => wrapIn(nodeType, attrs)(state))
        return wrapIn(nodeType, options.attrs)(state, dispatch)
      },
      select: function select(state) {
        return wrapIn(nodeType, options.attrs instanceof Function ? null : options.attrs)(state)
      }
    };
    for (var prop in options) { passedOptions[prop] = options[prop]; }
    return new MenuItem(passedOptions)
  }

  // :: (NodeType, Object) → MenuItem
  // Build a menu item for changing the type of the textblock around the
  // selection to the given type. Provides `run`, `active`, and `select`
  // properties. Others must be given in `options`. `options.attrs` may
  // be an object to provide the attributes for the textblock node.
  function blockTypeItem(nodeType, options) {
    var command = setBlockType(nodeType, options.attrs);
    var passedOptions = {
      run: command,
      enable: function enable(state) { return command(state) },
      active: function active(state) {
        var ref = state.selection;
        var $from = ref.$from;
        var to = ref.to;
        var node = ref.node;
        if (node) { return node.hasMarkup(nodeType, options.attrs) }
        return to <= $from.end() && $from.parent.hasMarkup(nodeType, options.attrs)
      }
    };
    for (var prop in options) { passedOptions[prop] = options[prop]; }
    return new MenuItem(passedOptions)
  }

  // Work around classList.toggle being broken in IE11
  function setClass(dom, cls, on) {
    if (on) { dom.classList.add(cls); }
    else { dom.classList.remove(cls); }
  }

  var prefix$2 = "ProseMirror-menubar";

  function isIOS() {
    if (typeof navigator == "undefined") { return false }
    var agent = navigator.userAgent;
    return !/Edge\/\d/.test(agent) && /AppleWebKit/.test(agent) && /Mobile\/\w+/.test(agent)
  }

  // :: (Object) → Plugin
  // A plugin that will place a menu bar above the editor. Note that
  // this involves wrapping the editor in an additional `<div>`.
  //
  //   options::-
  //   Supports the following options:
  //
  //     content:: [[MenuElement]]
  //     Provides the content of the menu, as a nested array to be
  //     passed to `renderGrouped`.
  //
  //     floating:: ?bool
  //     Determines whether the menu floats, i.e. whether it sticks to
  //     the top of the viewport when the editor is partially scrolled
  //     out of view.
  function menuBar(options) {
    return new Plugin({
      view: function view(editorView) { return new MenuBarView(editorView, options) }
    })
  }

  var MenuBarView = function MenuBarView(editorView, options) {
    var this$1$1 = this;

    this.editorView = editorView;
    this.options = options;

    this.wrapper = crelt("div", {class: prefix$2 + "-wrapper"});
    this.menu = this.wrapper.appendChild(crelt("div", {class: prefix$2}));
    this.menu.className = prefix$2;
    this.spacer = null;

    editorView.dom.parentNode.replaceChild(this.wrapper, editorView.dom);
    this.wrapper.appendChild(editorView.dom);

    this.maxHeight = 0;
    this.widthForMaxHeight = 0;
    this.floating = false;

    var ref = renderGrouped(this.editorView, this.options.content);
    var dom = ref.dom;
    var update = ref.update;
    this.contentUpdate = update;
    this.menu.appendChild(dom);
    this.update();

    if (options.floating && !isIOS()) {
      this.updateFloat();
      var potentialScrollers = getAllWrapping(this.wrapper);
      this.scrollFunc = function (e) {
        var root = this$1$1.editorView.root;
        if (!(root.body || root).contains(this$1$1.wrapper)) {
            potentialScrollers.forEach(function (el) { return el.removeEventListener("scroll", this$1$1.scrollFunc); });
        } else {
            this$1$1.updateFloat(e.target.getBoundingClientRect && e.target);
        }
      };
      potentialScrollers.forEach(function (el) { return el.addEventListener('scroll', this$1$1.scrollFunc); });
    }
  };

  MenuBarView.prototype.update = function update () {
    this.contentUpdate(this.editorView.state);

    if (this.floating) {
      this.updateScrollCursor();
    } else {
      if (this.menu.offsetWidth != this.widthForMaxHeight) {
        this.widthForMaxHeight = this.menu.offsetWidth;
        this.maxHeight = 0;
      }
      if (this.menu.offsetHeight > this.maxHeight) {
        this.maxHeight = this.menu.offsetHeight;
        this.menu.style.minHeight = this.maxHeight + "px";
      }
    }
  };

  MenuBarView.prototype.updateScrollCursor = function updateScrollCursor () {
    var selection = this.editorView.root.getSelection();
    if (!selection.focusNode) { return }
    var rects = selection.getRangeAt(0).getClientRects();
    var selRect = rects[selectionIsInverted(selection) ? 0 : rects.length - 1];
    if (!selRect) { return }
    var menuRect = this.menu.getBoundingClientRect();
    if (selRect.top < menuRect.bottom && selRect.bottom > menuRect.top) {
      var scrollable = findWrappingScrollable(this.wrapper);
      if (scrollable) { scrollable.scrollTop -= (menuRect.bottom - selRect.top); }
    }
  };

  MenuBarView.prototype.updateFloat = function updateFloat (scrollAncestor) {
    var parent = this.wrapper, editorRect = parent.getBoundingClientRect(),
        top = scrollAncestor ? Math.max(0, scrollAncestor.getBoundingClientRect().top) : 0;

    if (this.floating) {
      if (editorRect.top >= top || editorRect.bottom < this.menu.offsetHeight + 10) {
        this.floating = false;
        this.menu.style.position = this.menu.style.left = this.menu.style.top = this.menu.style.width = "";
        this.menu.style.display = "";
        this.spacer.parentNode.removeChild(this.spacer);
        this.spacer = null;
      } else {
        var border = (parent.offsetWidth - parent.clientWidth) / 2;
        this.menu.style.left = (editorRect.left + border) + "px";
        this.menu.style.display = (editorRect.top > window.innerHeight ? "none" : "");
        if (scrollAncestor) { this.menu.style.top = top + "px"; }
      }
    } else {
      if (editorRect.top < top && editorRect.bottom >= this.menu.offsetHeight + 10) {
        this.floating = true;
        var menuRect = this.menu.getBoundingClientRect();
        this.menu.style.left = menuRect.left + "px";
        this.menu.style.width = menuRect.width + "px";
        if (scrollAncestor) { this.menu.style.top = top + "px"; }
        this.menu.style.position = "fixed";
        this.spacer = crelt("div", {class: prefix$2 + "-spacer", style: ("height: " + (menuRect.height) + "px")});
        parent.insertBefore(this.spacer, this.menu);
      }
    }
  };

  MenuBarView.prototype.destroy = function destroy () {
    if (this.wrapper.parentNode)
      { this.wrapper.parentNode.replaceChild(this.editorView.dom, this.wrapper); }
  };

  // Not precise, but close enough
  function selectionIsInverted(selection) {
    if (selection.anchorNode == selection.focusNode) { return selection.anchorOffset > selection.focusOffset }
    return selection.anchorNode.compareDocumentPosition(selection.focusNode) == Node.DOCUMENT_POSITION_FOLLOWING
  }

  function findWrappingScrollable(node) {
    for (var cur = node.parentNode; cur; cur = cur.parentNode)
      { if (cur.scrollHeight > cur.clientHeight) { return cur } }
  }

  function getAllWrapping(node) {
      var res = [window];
      for (var cur = node.parentNode; cur; cur = cur.parentNode)
          { res.push(cur); }
      return res
  }

  var pmmenu = /*#__PURE__*/Object.freeze({
    __proto__: null,
    Dropdown: Dropdown,
    DropdownSubmenu: DropdownSubmenu,
    MenuItem: MenuItem,
    blockTypeItem: blockTypeItem,
    icons: icons,
    joinUpItem: joinUpItem,
    liftItem: liftItem,
    menuBar: menuBar,
    redoItem: redoItem,
    renderGrouped: renderGrouped,
    selectParentNodeItem: selectParentNodeItem,
    undoItem: undoItem,
    wrapItem: wrapItem
  });

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */


  var loaderPlugin = function (context) {
      return new Plugin({
          state: {
              init: function init() {
                  return DecorationSet.empty
              },
              apply: function apply(tr, set) {
                  // Adjust decoration positions to changes made by the transaction
                  set = set.map(tr.mapping, tr.doc);
                  // See if the transaction adds or removes any placeholders
                  var action = tr.getMeta(this);
                  if (action && action.add) {
                      var widget = humhub.require('ui.loader').set($('<span class="ProseMirror-placeholder">'), {
                          span: true,
                          size: '8px',
                          css: {
                              padding: '0px',
                              width: '60px'
                          }
                      })[0];
                      var deco = Decoration.widget(action.add.pos, widget, {id: action.add.id, content: true});
                      set = set.add(tr.doc, [deco]);
                      context.addContentDecoration('loader');
                  } else if (action && action.remove) {
                      set = set.remove(set.find(null, null, function (spec) { return spec.id === action.remove.id; }));
                      context.removeContentDecoration('loader');
                  }
                  return set
              }
          },
          props: {
              decorations: function decorations(state) {
                  return this.getState(state)
              }
          }
      });
  };

  function findLoader(context, id) {
      var decos = context.getProsemirrorPlugins('loader')[0].getState(context.editor.view.state);
      var found = decos.find(null, null, function (spec) { return spec.id === id; });
      return found.length ? found[0].from : null
  }

  function loaderStart(context, id, dispatch) {
      var view = context.editor.view;
      var tr = view.state.tr;

      if (!tr.selection.empty) {
          tr.deleteSelection();
      }

      tr.setMeta(context.getProsemirrorPlugins('loader')[0], {add: {id: id, pos: tr.selection.from}});

      if(dispatch) {
          view.dispatch(tr);
      }

      return tr;
  }

  function replaceLoader(context, id, content, dispatch) {
      var view = context.editor.view;
      var pos = findLoader(context, id);

      // If the content around the placeholder has been deleted, drop the image
      if (pos === null) {
          return;
      }

      var tr = view.state.tr.replaceWith(pos, pos, content).setMeta(context.getProsemirrorPlugins('loader')[0], {remove: {id: id}});

      if(dispatch) {
          view.dispatch(tr);
      }

      return tr;
  }

  function removeLoader(context, id, dispatch) {
      var view = context.editor.view;
      var pos = findLoader(context, id);

      // Focus the editor in order to synchronized changes into hidden textarea
      // for case when before file uploading the editor was not focused
      view.focus();

      // If the content around the placeholder has been deleted, drop the image
      if (pos === null) {
          return;
      }

      var tr = view.state.tr.setMeta(context.getProsemirrorPlugins('loader')[0], {remove: {id: id}});

      if(dispatch) {
          view.dispatch(tr);
      }

      return tr;
  }

  var loader$1 = /*#__PURE__*/Object.freeze({
    __proto__: null,
    loaderPlugin: loaderPlugin,
    loaderStart: loaderStart,
    replaceLoader: replaceLoader,
    removeLoader: removeLoader
  });

  // Because working with row and column-spanning cells is not quite
  // trivial, this code builds up a descriptive structure for a given
  // table node. The structures are cached with the (persistent) table
  // nodes as key, so that they only have to be recomputed when the
  // content of the table changes.
  //
  // This does mean that they have to store table-relative, not
  // document-relative positions. So code that uses them will typically
  // compute the start position of the table and offset positions passed
  // to or gotten from this structure by that amount.

  var readFromCache, addToCache;
  // Prefer using a weak map to cache table maps. Fall back on a
  // fixed-size cache if that's not supported.
  if (typeof WeakMap != "undefined") {
    var cache = new WeakMap;
    readFromCache = function (key) { return cache.get(key); };
    addToCache = function (key, value) {
      cache.set(key, value);
      return value
    };
  } else {
    var cache$1 = [], cacheSize = 10, cachePos = 0;
    readFromCache = function (key) {
      for (var i = 0; i < cache$1.length; i += 2)
        { if (cache$1[i] == key) { return cache$1[i + 1] } }
    };
    addToCache = function (key, value) {
      if (cachePos == cacheSize) { cachePos = 0; }
      cache$1[cachePos++] = key;
      return cache$1[cachePos++] = value
    };
  }

  var Rect = function Rect(left, top, right, bottom) {
    this.left = left; this.top = top; this.right = right; this.bottom = bottom;
  };

  // ::- A table map describes the structore of a given table. To avoid
  // recomputing them all the time, they are cached per table node. To
  // be able to do that, positions saved in the map are relative to the
  // start of the table, rather than the start of the document.
  var TableMap = function TableMap(width, height, map, problems) {
    // :: number The width of the table
    this.width = width;
    // :: number The table's height
    this.height = height;
    // :: [number] A width * height array with the start position of
    // the cell covering that part of the table in each slot
    this.map = map;
    // An optional array of problems (cell overlap or non-rectangular
    // shape) for the table, used by the table normalizer.
    this.problems = problems;
  };

  // :: (number) → Rect
  // Find the dimensions of the cell at the given position.
  TableMap.prototype.findCell = function findCell (pos) {
    for (var i = 0; i < this.map.length; i++) {
      var curPos = this.map[i];
      if (curPos != pos) { continue }
      var left = i % this.width, top = (i / this.width) | 0;
      var right = left + 1, bottom = top + 1;
      for (var j = 1; right < this.width && this.map[i + j] == curPos; j++) { right++; }
      for (var j$1 = 1; bottom < this.height && this.map[i + (this.width * j$1)] == curPos; j$1++) { bottom++; }
      return new Rect(left, top, right, bottom)
    }
    throw new RangeError("No cell with offset " + pos + " found")
  };

  // :: (number) → number
  // Find the left side of the cell at the given position.
  TableMap.prototype.colCount = function colCount (pos) {
    for (var i = 0; i < this.map.length; i++)
      { if (this.map[i] == pos) { return i % this.width } }
    throw new RangeError("No cell with offset " + pos + " found")
  };

  // :: (number, string, number) → ?number
  // Find the next cell in the given direction, starting from the cell
  // at `pos`, if any.
  TableMap.prototype.nextCell = function nextCell (pos, axis, dir) {
    var ref = this.findCell(pos);
      var left = ref.left;
      var right = ref.right;
      var top = ref.top;
      var bottom = ref.bottom;
    if (axis == "horiz") {
      if (dir < 0 ? left == 0 : right == this.width) { return null }
      return this.map[top * this.width + (dir < 0 ? left - 1 : right)]
    } else {
      if (dir < 0 ? top == 0 : bottom == this.height) { return null }
      return this.map[left + this.width * (dir < 0 ? top - 1 : bottom)]
    }
  };

  // :: (number, number) → Rect
  // Get the rectangle spanning the two given cells.
  TableMap.prototype.rectBetween = function rectBetween (a, b) {
    var ref = this.findCell(a);
      var leftA = ref.left;
      var rightA = ref.right;
      var topA = ref.top;
      var bottomA = ref.bottom;
    var ref$1 = this.findCell(b);
      var leftB = ref$1.left;
      var rightB = ref$1.right;
      var topB = ref$1.top;
      var bottomB = ref$1.bottom;
    return new Rect(Math.min(leftA, leftB), Math.min(topA, topB),
                    Math.max(rightA, rightB), Math.max(bottomA, bottomB))
  };

  // :: (Rect) → [number]
  // Return the position of all cells that have the top left corner in
  // the given rectangle.
  TableMap.prototype.cellsInRect = function cellsInRect (rect) {
    var result = [], seen = {};
    for (var row = rect.top; row < rect.bottom; row++) {
      for (var col = rect.left; col < rect.right; col++) {
        var index = row * this.width + col, pos = this.map[index];
        if (seen[pos]) { continue }
        seen[pos] = true;
        if ((col != rect.left || !col || this.map[index - 1] != pos) &&
            (row != rect.top || !row || this.map[index - this.width] != pos))
          { result.push(pos); }
      }
    }
    return result
  };

  // :: (number, number, Node) → number
  // Return the position at which the cell at the given row and column
  // starts, or would start, if a cell started there.
  TableMap.prototype.positionAt = function positionAt (row, col, table) {
    for (var i = 0, rowStart = 0;; i++) {
      var rowEnd = rowStart + table.child(i).nodeSize;
      if (i == row) {
        var index = col + row * this.width, rowEndIndex = (row + 1) * this.width;
        // Skip past cells from previous rows (via rowspan)
        while (index < rowEndIndex && this.map[index] < rowStart) { index++; }
        return index == rowEndIndex ? rowEnd - 1 : this.map[index]
      }
      rowStart = rowEnd;
    }
  };

  // :: (Node) → TableMap
  // Find the table map for the given table node.
  TableMap.get = function get (table) {
    return readFromCache(table) || addToCache(table, computeMap(table))
  };

  // Compute a table map.
  function computeMap(table) {
    if (table.type.spec.tableRole != "table") { throw new RangeError("Not a table node: " + table.type.name) }
    var width = findWidth(table), height = table.childCount;
    var map = [], mapPos = 0, problems = null, colWidths = [];
    for (var i = 0, e = width * height; i < e; i++) { map[i] = 0; }

    for (var row = 0, pos = 0; row < height; row++) {
      var rowNode = table.child(row);
      pos++;
      for (var i$1 = 0;; i$1++) {
        while (mapPos < map.length && map[mapPos] != 0) { mapPos++; }
        if (i$1 == rowNode.childCount) { break }
        var cellNode = rowNode.child(i$1);
        var ref = cellNode.attrs;
        var colspan = ref.colspan;
        var rowspan = ref.rowspan;
        var colwidth = ref.colwidth;
        for (var h = 0; h < rowspan; h++) {
          if (h + row >= height) {
            (problems || (problems = [])).push({type: "overlong_rowspan", pos: pos, n: rowspan - h});
            break
          }
          var start = mapPos + (h * width);
          for (var w = 0; w < colspan; w++) {
            if (map[start + w] == 0)
              { map[start + w] = pos; }
            else
              { (problems || (problems = [])).push({type: "collision", row: row, pos: pos, n: colspan - w}); }
            var colW = colwidth && colwidth[w];
            if (colW) {
              var widthIndex = ((start + w) % width) * 2, prev = colWidths[widthIndex];
              if (prev == null || (prev != colW && colWidths[widthIndex + 1] == 1)) {
                colWidths[widthIndex] = colW;
                colWidths[widthIndex + 1] = 1;
              } else if (prev == colW) {
                colWidths[widthIndex + 1]++;
              }
            }
          }
        }
        mapPos += colspan;
        pos += cellNode.nodeSize;
      }
      var expectedPos = (row + 1) * width, missing = 0;
      while (mapPos < expectedPos) { if (map[mapPos++] == 0) { missing++; } }
      if (missing) { (problems || (problems = [])).push({type: "missing", row: row, n: missing}); }
      pos++;
    }

    var tableMap = new TableMap(width, height, map, problems), badWidths = false;

    // For columns that have defined widths, but whose widths disagree
    // between rows, fix up the cells whose width doesn't match the
    // computed one.
    for (var i$2 = 0; !badWidths && i$2 < colWidths.length; i$2 += 2)
      { if (colWidths[i$2] != null && colWidths[i$2 + 1] < height) { badWidths = true; } }
    if (badWidths) { findBadColWidths(tableMap, colWidths, table); }

    return tableMap
  }

  function findWidth(table) {
    var width = -1, hasRowSpan = false;
    for (var row = 0; row < table.childCount; row++) {
      var rowNode = table.child(row), rowWidth = 0;
      if (hasRowSpan) { for (var j = 0; j < row; j++) {
        var prevRow = table.child(j);
        for (var i = 0; i < prevRow.childCount; i++) {
          var cell = prevRow.child(i);
          if (j + cell.attrs.rowspan > row) { rowWidth += cell.attrs.colspan; }
        }
      } }
      for (var i$1 = 0; i$1 < rowNode.childCount; i$1++) {
        var cell$1 = rowNode.child(i$1);
        rowWidth += cell$1.attrs.colspan;
        if (cell$1.attrs.rowspan > 1) { hasRowSpan = true; }
      }
      if (width == -1)
        { width = rowWidth; }
      else if (width != rowWidth)
        { width = Math.max(width, rowWidth); }
    }
    return width
  }

  function findBadColWidths(map, colWidths, table) {
    if (!map.problems) { map.problems = []; }
    for (var i = 0, seen = {}; i < map.map.length; i++) {
      var pos = map.map[i];
      if (seen[pos]) { continue }
      seen[pos] = true;
      var node = table.nodeAt(pos), updated = null;
      for (var j = 0; j < node.attrs.colspan; j++) {
        var col = (i + j) % map.width, colWidth = colWidths[col * 2];
        if (colWidth != null && (!node.attrs.colwidth || node.attrs.colwidth[j] != colWidth))
          { (updated || (updated = freshColWidth(node.attrs)))[j] = colWidth; }
      }
      if (updated) { map.problems.unshift({type: "colwidth mismatch", pos: pos, colwidth: updated}); }
    }
  }

  function freshColWidth(attrs) {
    if (attrs.colwidth) { return attrs.colwidth.slice() }
    var result = [];
    for (var i = 0; i < attrs.colspan; i++) { result.push(0); }
    return result
  }

  // Helper for creating a schema that supports tables.

  function getCellAttrs(dom, extraAttrs) {
    var widthAttr = dom.getAttribute("data-colwidth");
    var widths = widthAttr && /^\d+(,\d+)*$/.test(widthAttr) ? widthAttr.split(",").map(function (s) { return Number(s); }) : null;
    var colspan = Number(dom.getAttribute("colspan") || 1);
    var result = {
      colspan: colspan,
      rowspan: Number(dom.getAttribute("rowspan") || 1),
      colwidth: widths && widths.length == colspan ? widths : null
    };
    for (var prop in extraAttrs) {
      var getter = extraAttrs[prop].getFromDOM;
      var value = getter && getter(dom);
      if (value != null) { result[prop] = value; }
    }
    return result
  }

  function setCellAttrs(node, extraAttrs) {
    var attrs = {};
    if (node.attrs.colspan != 1) { attrs.colspan = node.attrs.colspan; }
    if (node.attrs.rowspan != 1) { attrs.rowspan = node.attrs.rowspan; }
    if (node.attrs.colwidth)
      { attrs["data-colwidth"] = node.attrs.colwidth.join(","); }
    for (var prop in extraAttrs) {
      var setter = extraAttrs[prop].setDOMAttr;
      if (setter) { setter(node.attrs[prop], attrs); }
    }
    return attrs
  }

  // :: (Object) → Object
  //
  // This function creates a set of [node
  // specs](http://prosemirror.net/docs/ref/#model.SchemaSpec.nodes) for
  // `table`, `table_row`, and `table_cell` nodes types as used by this
  // module. The result can then be added to the set of nodes when
  // creating a a schema.
  //
  //   options::- The following options are understood:
  //
  //     tableGroup:: ?string
  //     A group name (something like `"block"`) to add to the table
  //     node type.
  //
  //     cellContent:: string
  //     The content expression for table cells.
  //
  //     cellAttributes:: ?Object
  //     Additional attributes to add to cells. Maps attribute names to
  //     objects with the following properties:
  //
  //       default:: any
  //       The attribute's default value.
  //
  //       getFromDOM:: ?(dom.Node) → any
  //       A function to read the attribute's value from a DOM node.
  //
  //       setDOMAttr:: ?(value: any, attrs: Object)
  //       A function to add the attribute's value to an attribute
  //       object that's used to render the cell's DOM.
  function tableNodes(options) {
    var extraAttrs = options.cellAttributes || {};
    var cellAttrs = {
      colspan: {default: 1},
      rowspan: {default: 1},
      colwidth: {default: null}
    };
    for (var prop in extraAttrs)
      { cellAttrs[prop] = {default: extraAttrs[prop].default}; }

    return {
      table: {
        content: "table_row+",
        tableRole: "table",
        isolating: true,
        group: options.tableGroup,
        parseDOM: [{tag: "table"}],
        toDOM: function toDOM() { return ["table", ["tbody", 0]] }
      },
      table_row: {
        content: "(table_cell | table_header)*",
        tableRole: "row",
        parseDOM: [{tag: "tr"}],
        toDOM: function toDOM() { return ["tr", 0] }
      },
      table_cell: {
        content: options.cellContent,
        attrs: cellAttrs,
        tableRole: "cell",
        isolating: true,
        parseDOM: [{tag: "td", getAttrs: function (dom) { return getCellAttrs(dom, extraAttrs); }}],
        toDOM: function toDOM(node) { return ["td", setCellAttrs(node, extraAttrs), 0] }
      },
      table_header: {
        content: options.cellContent,
        attrs: cellAttrs,
        tableRole: "header_cell",
        isolating: true,
        parseDOM: [{tag: "th", getAttrs: function (dom) { return getCellAttrs(dom, extraAttrs); }}],
        toDOM: function toDOM(node) { return ["th", setCellAttrs(node, extraAttrs), 0] }
      }
    }
  }

  function tableNodeTypes(schema) {
    var result = schema.cached.tableNodeTypes;
    if (!result) {
      result = schema.cached.tableNodeTypes = {};
      for (var name in schema.nodes) {
        var type = schema.nodes[name], role = type.spec.tableRole;
        if (role) { result[role] = type; }
      }
    }
    return result
  }

  // Various helper function for working with tables

  var key$2 = new PluginKey("selectingCells");

  function cellAround($pos) {
    for (var d = $pos.depth - 1; d > 0; d--)
      { if ($pos.node(d).type.spec.tableRole == "row") { return $pos.node(0).resolve($pos.before(d + 1)) } }
    return null
  }

  function isInTable(state) {
    var $head = state.selection.$head;
    for (var d = $head.depth; d > 0; d--) { if ($head.node(d).type.spec.tableRole == "row") { return true } }
    return false
  }

  function selectionCell(state) {
    var sel = state.selection;
    if (sel.$anchorCell) {
      return sel.$anchorCell.pos > sel.$headCell.pos ? sel.$anchorCell : sel.$headCell;
    } else if (sel.node && sel.node.type.spec.tableRole == "cell") {
      return sel.$anchor
    }
    return cellAround(sel.$head) || cellNear(sel.$head)
  }

  function cellNear($pos) {
    for (var after = $pos.nodeAfter, pos = $pos.pos; after; after = after.firstChild, pos++) {
      var role = after.type.spec.tableRole;
      if (role == "cell" || role == "header_cell") { return $pos.doc.resolve(pos) }
    }
    for (var before = $pos.nodeBefore, pos$1 = $pos.pos; before; before = before.lastChild, pos$1--) {
      var role$1 = before.type.spec.tableRole;
      if (role$1 == "cell" || role$1 == "header_cell") { return $pos.doc.resolve(pos$1 - before.nodeSize) }
    }
  }

  function pointsAtCell($pos) {
    return $pos.parent.type.spec.tableRole == "row" && $pos.nodeAfter
  }

  function moveCellForward($pos) {
    return $pos.node(0).resolve($pos.pos + $pos.nodeAfter.nodeSize)
  }

  function inSameTable($a, $b) {
    return $a.depth == $b.depth && $a.pos >= $b.start(-1) && $a.pos <= $b.end(-1)
  }

  function nextCell($pos, axis, dir) {
    var start = $pos.start(-1), map = TableMap.get($pos.node(-1));
    var moved = map.nextCell($pos.pos - start, axis, dir);
    return moved == null ? null : $pos.node(0).resolve(start + moved)
  }

  function setAttr(attrs, name, value) {
    var result = {};
    for (var prop in attrs) { result[prop] = attrs[prop]; }
    result[name] = value;
    return result
  }

  function removeColSpan(attrs, pos, n) {
    if ( n === void 0 ) { n=1; }

    var result = setAttr(attrs, "colspan", attrs.colspan - n);
    if (result.colwidth) {
      result.colwidth = result.colwidth.slice();
      result.colwidth.splice(pos, n);
      if (!result.colwidth.some(function (w) { return w > 0; })) { result.colwidth = null; }
    }
    return result
  }

  function addColSpan(attrs, pos, n) {
    if ( n === void 0 ) { n=1; }

    var result = setAttr(attrs, "colspan", attrs.colspan + n);
    if (result.colwidth) {
      result.colwidth = result.colwidth.slice();
      for (var i = 0; i < n; i++) { result.colwidth.splice(pos, 0, 0); }
    }
    return result
  }

  function columnIsHeader(map, table, col) {
    var headerCell = tableNodeTypes(table.type.schema).header_cell;
    for (var row = 0; row < map.height; row++)
      { if (table.nodeAt(map.map[col + row * map.width]).type != headerCell)
        { return false } }
    return true
  }

  // This file defines a ProseMirror selection subclass that models

  // ::- A [`Selection`](http://prosemirror.net/docs/ref/#state.Selection)
  // subclass that represents a cell selection spanning part of a table.
  // With the plugin enabled, these will be created when the user
  // selects across cells, and will be drawn by giving selected cells a
  // `selectedCell` CSS class.
  var CellSelection = /*@__PURE__*/(function (Selection) {
    function CellSelection($anchorCell, $headCell) {
      if ( $headCell === void 0 ) { $headCell = $anchorCell; }

      var table = $anchorCell.node(-1), map = TableMap.get(table), start = $anchorCell.start(-1);
      var rect = map.rectBetween($anchorCell.pos - start, $headCell.pos - start);
      var doc = $anchorCell.node(0);
      var cells = map.cellsInRect(rect).filter(function (p) { return p != $headCell.pos - start; });
      // Make the head cell the first range, so that it counts as the
      // primary part of the selection
      cells.unshift($headCell.pos - start);
      var ranges = cells.map(function (pos) {
        var cell = table.nodeAt(pos), from = pos + start + 1;
        return new SelectionRange(doc.resolve(from), doc.resolve(from + cell.content.size))
      });
      Selection.call(this, ranges[0].$from, ranges[0].$to, ranges);
      // :: ResolvedPos
      // A resolved position pointing _in front of_ the anchor cell (the one
      // that doesn't move when extending the selection).
      this.$anchorCell = $anchorCell;
      // :: ResolvedPos
      // A resolved position pointing in front of the head cell (the one
      // moves when extending the selection).
      this.$headCell = $headCell;
    }

    if ( Selection ) { CellSelection.__proto__ = Selection; }
    CellSelection.prototype = Object.create( Selection && Selection.prototype );
    CellSelection.prototype.constructor = CellSelection;

    CellSelection.prototype.map = function map (doc, mapping) {
      var $anchorCell = doc.resolve(mapping.map(this.$anchorCell.pos));
      var $headCell = doc.resolve(mapping.map(this.$headCell.pos));
      if (pointsAtCell($anchorCell) && pointsAtCell($headCell) && inSameTable($anchorCell, $headCell)) {
        var tableChanged = this.$anchorCell.node(-1) != $anchorCell.node(-1);
        if (tableChanged && this.isRowSelection())
          { return CellSelection.rowSelection($anchorCell, $headCell) }
        else if (tableChanged && this.isColSelection())
          { return CellSelection.colSelection($anchorCell, $headCell) }
        else
          { return new CellSelection($anchorCell, $headCell) }
      }
      return TextSelection.between($anchorCell, $headCell)
    };

    // :: () → Slice
    // Returns a rectangular slice of table rows containing the selected
    // cells.
    CellSelection.prototype.content = function content () {
      var table = this.$anchorCell.node(-1), map = TableMap.get(table), start = this.$anchorCell.start(-1);
      var rect = map.rectBetween(this.$anchorCell.pos - start, this.$headCell.pos - start);
      var seen = {}, rows = [];
      for (var row = rect.top; row < rect.bottom; row++) {
        var rowContent = [];
        for (var index = row * map.width + rect.left, col = rect.left; col < rect.right; col++, index++) {
          var pos = map.map[index];
          if (!seen[pos]) {
            seen[pos] = true;
            var cellRect = map.findCell(pos), cell = table.nodeAt(pos);
            var extraLeft = rect.left - cellRect.left, extraRight = cellRect.right - rect.right;
            if (extraLeft > 0 || extraRight > 0) {
              var attrs = cell.attrs;
              if (extraLeft > 0) { attrs = removeColSpan(attrs, 0, extraLeft); }
              if (extraRight > 0) { attrs = removeColSpan(attrs, attrs.colspan - extraRight, extraRight); }
              if (cellRect.left < rect.left) { cell = cell.type.createAndFill(attrs); }
              else { cell = cell.type.create(attrs, cell.content); }
            }
            if (cellRect.top < rect.top || cellRect.bottom > rect.bottom) {
              var attrs$1 = setAttr(cell.attrs, "rowspan", Math.min(cellRect.bottom, rect.bottom) - Math.max(cellRect.top, rect.top));
              if (cellRect.top < rect.top) { cell = cell.type.createAndFill(attrs$1); }
              else { cell = cell.type.create(attrs$1, cell.content); }
            }
            rowContent.push(cell);
          }
        }
        rows.push(table.child(row).copy(Fragment$1.from(rowContent)));
      }

      var fragment = this.isColSelection() && this.isRowSelection() ? table : rows;
      return new Slice$1(Fragment$1.from(fragment), 1, 1)
    };

    CellSelection.prototype.replace = function replace (tr, content) {
      if ( content === void 0 ) { content = Slice$1.empty; }

      var mapFrom = tr.steps.length, ranges = this.ranges;
      for (var i = 0; i < ranges.length; i++) {
        var ref = ranges[i];
        var $from = ref.$from;
        var $to = ref.$to;
        var mapping = tr.mapping.slice(mapFrom);
        tr.replace(mapping.map($from.pos), mapping.map($to.pos), i ? Slice$1.empty : content);
      }
      var sel = Selection.findFrom(tr.doc.resolve(tr.mapping.slice(mapFrom).map(this.to)), -1);
      if (sel) { tr.setSelection(sel); }
    };

    CellSelection.prototype.replaceWith = function replaceWith (tr, node) {
      this.replace(tr, new Slice$1(Fragment$1.from(node), 0, 0));
    };

    CellSelection.prototype.forEachCell = function forEachCell (f) {
      var table = this.$anchorCell.node(-1), map = TableMap.get(table), start = this.$anchorCell.start(-1);
      var cells = map.cellsInRect(map.rectBetween(this.$anchorCell.pos - start, this.$headCell.pos - start));
      for (var i = 0; i < cells.length; i++)
        { f(table.nodeAt(cells[i]), start + cells[i]); }
    };

    // :: () → bool
    // True if this selection goes all the way from the top to the
    // bottom of the table.
    CellSelection.prototype.isColSelection = function isColSelection () {
      var anchorTop = this.$anchorCell.index(-1), headTop = this.$headCell.index(-1);
      if (Math.min(anchorTop, headTop) > 0) { return false }
      var anchorBot = anchorTop + this.$anchorCell.nodeAfter.attrs.rowspan,
          headBot = headTop + this.$headCell.nodeAfter.attrs.rowspan;
      return Math.max(anchorBot, headBot) == this.$headCell.node(-1).childCount
    };

    // :: (ResolvedPos, ?ResolvedPos) → CellSelection
    // Returns the smallest column selection that covers the given anchor
    // and head cell.
    CellSelection.colSelection = function colSelection ($anchorCell, $headCell) {
      if ( $headCell === void 0 ) { $headCell = $anchorCell; }

      var map = TableMap.get($anchorCell.node(-1)), start = $anchorCell.start(-1);
      var anchorRect = map.findCell($anchorCell.pos - start), headRect = map.findCell($headCell.pos - start);
      var doc = $anchorCell.node(0);
      if (anchorRect.top <= headRect.top) {
        if (anchorRect.top > 0)
          { $anchorCell = doc.resolve(start + map.map[anchorRect.left]); }
        if (headRect.bottom < map.height)
          { $headCell = doc.resolve(start + map.map[map.width * (map.height - 1) + headRect.right - 1]); }
      } else {
        if (headRect.top > 0)
          { $headCell = doc.resolve(start + map.map[headRect.left]); }
        if (anchorRect.bottom < map.height)
          { $anchorCell = doc.resolve(start + map.map[map.width * (map.height - 1) + anchorRect.right - 1]); }
      }
      return new CellSelection($anchorCell, $headCell)
    };

    // :: () → bool
    // True if this selection goes all the way from the left to the
    // right of the table.
    CellSelection.prototype.isRowSelection = function isRowSelection () {
      var map = TableMap.get(this.$anchorCell.node(-1)), start = this.$anchorCell.start(-1);
      var anchorLeft = map.colCount(this.$anchorCell.pos - start),
          headLeft = map.colCount(this.$headCell.pos - start);
      if (Math.min(anchorLeft, headLeft) > 0) { return false }
      var anchorRight = anchorLeft + this.$anchorCell.nodeAfter.attrs.colspan,
          headRight = headLeft + this.$headCell.nodeAfter.attrs.colspan;
      return Math.max(anchorRight, headRight) == map.width
    };

    CellSelection.prototype.eq = function eq (other) {
      return other instanceof CellSelection && other.$anchorCell.pos == this.$anchorCell.pos &&
        other.$headCell.pos == this.$headCell.pos
    };

    // :: (ResolvedPos, ?ResolvedPos) → CellSelection
    // Returns the smallest row selection that covers the given anchor
    // and head cell.
    CellSelection.rowSelection = function rowSelection ($anchorCell, $headCell) {
      if ( $headCell === void 0 ) { $headCell = $anchorCell; }

      var map = TableMap.get($anchorCell.node(-1)), start = $anchorCell.start(-1);
      var anchorRect = map.findCell($anchorCell.pos - start), headRect = map.findCell($headCell.pos - start);
      var doc = $anchorCell.node(0);
      if (anchorRect.left <= headRect.left) {
        if (anchorRect.left > 0)
          { $anchorCell = doc.resolve(start + map.map[anchorRect.top * map.width]); }
        if (headRect.right < map.width)
          { $headCell = doc.resolve(start + map.map[map.width * (headRect.top + 1) - 1]); }
      } else {
        if (headRect.left > 0)
          { $headCell = doc.resolve(start + map.map[headRect.top * map.width]); }
        if (anchorRect.right < map.width)
          { $anchorCell = doc.resolve(start + map.map[map.width * (anchorRect.top + 1) - 1]); }
      }
      return new CellSelection($anchorCell, $headCell)
    };

    CellSelection.prototype.toJSON = function toJSON () {
      return {type: "cell", anchor: this.$anchorCell.pos, head: this.$headCell.pos}
    };

    CellSelection.fromJSON = function fromJSON (doc, json) {
      return new CellSelection(doc.resolve(json.anchor), doc.resolve(json.head))
    };

    // :: (Node, number, ?number) → CellSelection
    CellSelection.create = function create (doc, anchorCell, headCell) {
      if ( headCell === void 0 ) { headCell = anchorCell; }

      return new CellSelection(doc.resolve(anchorCell), doc.resolve(headCell))
    };

    CellSelection.prototype.getBookmark = function getBookmark () { return new CellBookmark(this.$anchorCell.pos, this.$headCell.pos) };

    return CellSelection;
  }(Selection));

  CellSelection.prototype.visible = false;

  Selection.jsonID("cell", CellSelection);

  var CellBookmark = function CellBookmark(anchor, head) {
    this.anchor = anchor;
    this.head = head;
  };
  CellBookmark.prototype.map = function map (mapping) {
    return new CellBookmark(mapping.map(this.anchor), mapping.map(this.head))
  };
  CellBookmark.prototype.resolve = function resolve (doc) {
    var $anchorCell = doc.resolve(this.anchor), $headCell = doc.resolve(this.head);
    if ($anchorCell.parent.type.spec.tableRole == "row" &&
        $headCell.parent.type.spec.tableRole == "row" &&
        $anchorCell.index() < $anchorCell.parent.childCount &&
        $headCell.index() < $headCell.parent.childCount &&
        inSameTable($anchorCell, $headCell))
      { return new CellSelection($anchorCell, $headCell) }
    else
      { return Selection.near($headCell, 1) }
  };

  function drawCellSelection(state) {
    if (!(state.selection instanceof CellSelection)) { return null }
    var cells = [];
    state.selection.forEachCell(function (node, pos) {
      cells.push(Decoration.node(pos, pos + node.nodeSize, {class: "selectedCell"}));
    });
    return DecorationSet.create(state.doc, cells)
  }

  function isCellBoundarySelection(ref) {
    var $from = ref.$from;
    var $to = ref.$to;

    if ($from.pos == $to.pos || $from.pos < $from.pos - 6) { return false } // Cheap elimination
    var afterFrom = $from.pos, beforeTo = $to.pos, depth = $from.depth;
    for (; depth >= 0; depth--, afterFrom++)
      { if ($from.after(depth + 1) < $from.end(depth)) { break } }
    for (var d = $to.depth; d >= 0; d--, beforeTo--)
      { if ($to.before(d + 1) > $to.start(d)) { break } }
    return afterFrom == beforeTo && /row|table/.test($from.node(depth).type.spec.tableRole)
  }

  function isTextSelectionAcrossCells(ref) {
    var $from = ref.$from;
    var $to = ref.$to;

    var fromCellBoundaryNode;
    var toCellBoundaryNode;

    for (var i = $from.depth; i > 0; i--) {
      var node = $from.node(i);
      if (node.type.spec.tableRole === 'cell' || node.type.spec.tableRole === 'header_cell') {
        fromCellBoundaryNode = node;
        break;
      }
    }

    for (var i$1 = $to.depth; i$1 > 0; i$1--) {
      var node$1 = $to.node(i$1);
      if (node$1.type.spec.tableRole === 'cell' || node$1.type.spec.tableRole === 'header_cell') {
        toCellBoundaryNode = node$1;
        break;
      }
    }

    return fromCellBoundaryNode !== toCellBoundaryNode && $to.parentOffset === 0
  }

  function normalizeSelection(state, tr, allowTableNodeSelection) {
    var sel = (tr || state).selection, doc = (tr || state).doc, normalize, role;
    if (sel instanceof NodeSelection && (role = sel.node.type.spec.tableRole)) {
      if (role == "cell" || role == "header_cell") {
        normalize = CellSelection.create(doc, sel.from);
      } else if (role == "row") {
        var $cell = doc.resolve(sel.from + 1);
        normalize = CellSelection.rowSelection($cell, $cell);
      } else if (!allowTableNodeSelection) {
        var map = TableMap.get(sel.node), start = sel.from + 1;
        var lastCell = start + map.map[map.width * map.height - 1];
        normalize = CellSelection.create(doc, start + 1, lastCell);
      }
    } else if (sel instanceof TextSelection && isCellBoundarySelection(sel)) {
      normalize = TextSelection.create(doc, sel.from);
    } else if (sel instanceof TextSelection && isTextSelectionAcrossCells(sel)) {
      normalize = TextSelection.create(doc, sel.$from.start(), sel.$from.end());
    }
    if (normalize)
      { (tr || (tr = state.tr)).setSelection(normalize); }
    return tr
  }

  // Utilities used for copy/paste handling.

  // Utilities to help with copying and pasting table cells

  // : (Slice) → ?{width: number, height: number, rows: [Fragment]}
  // Get a rectangular area of cells from a slice, or null if the outer
  // nodes of the slice aren't table cells or rows.
  function pastedCells(slice) {
    if (!slice.size) { return null }
    var content = slice.content;
    var openStart = slice.openStart;
    var openEnd = slice.openEnd;
    while (content.childCount == 1 && (openStart > 0 && openEnd > 0 || content.firstChild.type.spec.tableRole == "table")) {
      openStart--;
      openEnd--;
      content = content.firstChild.content;
    }
    var first = content.firstChild, role = first.type.spec.tableRole;
    var schema = first.type.schema, rows = [];
    if (role == "row") {
      for (var i = 0; i < content.childCount; i++) {
        var cells = content.child(i).content;
        var left = i ? 0 : Math.max(0, openStart - 1);
        var right = i < content.childCount - 1 ? 0 : Math.max(0, openEnd - 1);
        if (left || right) { cells = fitSlice(tableNodeTypes(schema).row, new Slice$1(cells, left, right)).content; }
        rows.push(cells);
      }
    } else if (role == "cell" || role == "header_cell") {
      rows.push(openStart || openEnd ? fitSlice(tableNodeTypes(schema).row, new Slice$1(content, openStart, openEnd)).content : content);
    } else {
      return null
    }
    return ensureRectangular(schema, rows)
  }

  // : (Schema, [Fragment]) → {width: number, height: number, rows: [Fragment]}
  // Compute the width and height of a set of cells, and make sure each
  // row has the same number of cells.
  function ensureRectangular(schema, rows) {
    var widths = [];
    for (var i = 0; i < rows.length; i++) {
      var row = rows[i];
      for (var j = row.childCount - 1; j >= 0; j--) {
        var ref = row.child(j).attrs;
        var rowspan = ref.rowspan;
        var colspan = ref.colspan;
        for (var r = i; r < i + rowspan; r++)
          { widths[r] = (widths[r] || 0) + colspan; }
      }
    }
    var width = 0;
    for (var r$1 = 0; r$1 < widths.length; r$1++) { width = Math.max(width, widths[r$1]); }
    for (var r$2 = 0; r$2 < widths.length; r$2++) {
      if (r$2 >= rows.length) { rows.push(Fragment$1.empty); }
      if (widths[r$2] < width) {
        var empty = tableNodeTypes(schema).cell.createAndFill(), cells = [];
        for (var i$1 = widths[r$2]; i$1 < width; i$1++) { cells.push(empty); }
        rows[r$2] = rows[r$2].append(Fragment$1.from(cells));
      }
    }
    return {height: rows.length, width: width, rows: rows}
  }

  function fitSlice(nodeType, slice) {
    var node = nodeType.createAndFill();
    var tr = new Transform(node).replace(0, node.content.size, slice);
    return tr.doc
  }

  // : ({width: number, height: number, rows: [Fragment]}, number, number) → {width: number, height: number, rows: [Fragment]}
  // Clip or extend (repeat) the given set of cells to cover the given
  // width and height. Will clip rowspan/colspan cells at the edges when
  // they stick out.
  function clipCells(ref, newWidth, newHeight) {
    var width = ref.width;
    var height = ref.height;
    var rows = ref.rows;

    if (width != newWidth) {
      var added = [], newRows = [];
      for (var row = 0; row < rows.length; row++) {
        var frag = rows[row], cells = [];
        for (var col = added[row] || 0, i = 0; col < newWidth; i++) {
          var cell = frag.child(i % frag.childCount);
          if (col + cell.attrs.colspan > newWidth)
            { cell = cell.type.create(removeColSpan(cell.attrs, cell.attrs.colspan, col + cell.attrs.colspan - newWidth), cell.content); }
          cells.push(cell);
          col += cell.attrs.colspan;
          for (var j = 1; j < cell.attrs.rowspan; j++)
            { added[row + j] = (added[row + j] || 0) + cell.attrs.colspan; }
        }
        newRows.push(Fragment$1.from(cells));
      }
      rows = newRows;
      width = newWidth;
    }

    if (height != newHeight) {
      var newRows$1 = [];
      for (var row$1 = 0, i$1 = 0; row$1 < newHeight; row$1++, i$1++) {
        var cells$1 = [], source = rows[i$1 % height];
        for (var j$1 = 0; j$1 < source.childCount; j$1++) {
          var cell$1 = source.child(j$1);
          if (row$1 + cell$1.attrs.rowspan > newHeight)
            { cell$1 = cell$1.type.create(setAttr(cell$1.attrs, "rowspan", Math.max(1, newHeight - cell$1.attrs.rowspan)), cell$1.content); }
          cells$1.push(cell$1);
        }
        newRows$1.push(Fragment$1.from(cells$1));
      }
      rows = newRows$1;
      height = newHeight;
    }

    return {width: width, height: height, rows: rows}
  }

  // Make sure a table has at least the given width and height. Return
  // true if something was changed.
  function growTable(tr, map, table, start, width, height, mapFrom) {
    var schema = tr.doc.type.schema, types = tableNodeTypes(schema), empty, emptyHead;
    if (width > map.width) {
      for (var row = 0, rowEnd = 0; row < map.height; row++) {
        var rowNode = table.child(row);
        rowEnd += rowNode.nodeSize;
        var cells = [], add = (void 0);
        if (rowNode.lastChild == null || rowNode.lastChild.type == types.cell)
          { add = empty || (empty = types.cell.createAndFill()); }
        else
          { add = emptyHead || (emptyHead = types.header_cell.createAndFill()); }
        for (var i = map.width; i < width; i++) { cells.push(add); }
        tr.insert(tr.mapping.slice(mapFrom).map(rowEnd - 1 + start), cells);
      }
    }
    if (height > map.height) {
      var cells$1 = [];
      for (var i$1 = 0, start$1 = (map.height - 1) * map.width; i$1 < Math.max(map.width, width); i$1++) {
        var header = i$1 >= map.width ? false :
            table.nodeAt(map.map[start$1 + i$1]).type == types.header_cell;
        cells$1.push(header
                   ? (emptyHead || (emptyHead = types.header_cell.createAndFill()))
                   : (empty || (empty = types.cell.createAndFill())));
      }

      var emptyRow = types.row.create(null, Fragment$1.from(cells$1)), rows = [];
      for (var i$2 = map.height; i$2 < height; i$2++) { rows.push(emptyRow); }
      tr.insert(tr.mapping.slice(mapFrom).map(start + table.nodeSize - 2), rows);
    }
    return !!(empty || emptyHead)
  }

  // Make sure the given line (left, top) to (right, top) doesn't cross
  // any rowspan cells by splitting cells that cross it. Return true if
  // something changed.
  function isolateHorizontal(tr, map, table, start, left, right, top, mapFrom) {
    if (top == 0 || top == map.height) { return false }
    var found = false;
    for (var col = left; col < right; col++) {
      var index = top * map.width + col, pos = map.map[index];
      if (map.map[index - map.width] == pos) {
        found = true;
        var cell = table.nodeAt(pos);
        var ref = map.findCell(pos);
        var cellTop = ref.top;
        var cellLeft = ref.left;
        tr.setNodeMarkup(tr.mapping.slice(mapFrom).map(pos + start), null, setAttr(cell.attrs, "rowspan", top - cellTop));
        tr.insert(tr.mapping.slice(mapFrom).map(map.positionAt(top, cellLeft, table)),
                  cell.type.createAndFill(setAttr(cell.attrs, "rowspan", (cellTop + cell.attrs.rowspan) - top)));
        col += cell.attrs.colspan - 1;
      }
    }
    return found
  }

  // Make sure the given line (left, top) to (left, bottom) doesn't
  // cross any colspan cells by splitting cells that cross it. Return
  // true if something changed.
  function isolateVertical(tr, map, table, start, top, bottom, left, mapFrom) {
    if (left == 0 || left == map.width) { return false }
    var found = false;
    for (var row = top; row < bottom; row++) {
      var index = row * map.width + left, pos = map.map[index];
      if (map.map[index - 1] == pos) {
        found = true;
        var cell = table.nodeAt(pos), cellLeft = map.colCount(pos);
        var updatePos = tr.mapping.slice(mapFrom).map(pos + start);
        tr.setNodeMarkup(updatePos, null, removeColSpan(cell.attrs, left - cellLeft, cell.attrs.colspan - (left - cellLeft)));
        tr.insert(updatePos + cell.nodeSize, cell.type.createAndFill(removeColSpan(cell.attrs, 0, left - cellLeft)));
        row += cell.attrs.rowspan - 1;
      }
    }
    return found
  }

  // Insert the given set of cells (as returned by `pastedCells`) into a
  // table, at the position pointed at by rect.
  function insertCells(state, dispatch, tableStart, rect, cells) {
    var table = tableStart ? state.doc.nodeAt(tableStart - 1) : state.doc, map = TableMap.get(table);
    var top = rect.top;
    var left = rect.left;
    var right = left + cells.width, bottom = top + cells.height;
    var tr = state.tr, mapFrom = 0;
    function recomp() {
      table = tableStart ? tr.doc.nodeAt(tableStart - 1) : tr.doc;
      map = TableMap.get(table);
      mapFrom = tr.mapping.maps.length;
    }
    // Prepare the table to be large enough and not have any cells
    // crossing the boundaries of the rectangle that we want to
    // insert into. If anything about it changes, recompute the table
    // map so that subsequent operations can see the current shape.
    if (growTable(tr, map, table, tableStart, right, bottom, mapFrom)) { recomp(); }
    if (isolateHorizontal(tr, map, table, tableStart, left, right, top, mapFrom)) { recomp(); }
    if (isolateHorizontal(tr, map, table, tableStart, left, right, bottom, mapFrom)) { recomp(); }
    if (isolateVertical(tr, map, table, tableStart, top, bottom, left, mapFrom)) { recomp(); }
    if (isolateVertical(tr, map, table, tableStart, top, bottom, right, mapFrom)) { recomp(); }

    for (var row = top; row < bottom; row++) {
      var from = map.positionAt(row, left, table), to = map.positionAt(row, right, table);
      tr.replace(tr.mapping.slice(mapFrom).map(from + tableStart), tr.mapping.slice(mapFrom).map(to + tableStart),
                 new Slice$1(cells.rows[row - top], 0, 0));
    }
    recomp();
    tr.setSelection(new CellSelection(tr.doc.resolve(tableStart + map.positionAt(top, left, table)),
                                      tr.doc.resolve(tableStart + map.positionAt(bottom - 1, right - 1, table))));
    dispatch(tr);
  }

  // This file defines a number of helpers for wiring up user input to

  var handleKeyDown$1 = keydownHandler({
    "ArrowLeft": arrow$1("horiz", -1),
    "ArrowRight": arrow$1("horiz", 1),
    "ArrowUp": arrow$1("vert", -1),
    "ArrowDown": arrow$1("vert", 1),

    "Shift-ArrowLeft": shiftArrow("horiz", -1),
    "Shift-ArrowRight": shiftArrow("horiz", 1),
    "Shift-ArrowUp": shiftArrow("vert", -1),
    "Shift-ArrowDown": shiftArrow("vert", 1),

    "Backspace": deleteCellSelection,
    "Mod-Backspace": deleteCellSelection,
    "Delete": deleteCellSelection,
    "Mod-Delete": deleteCellSelection
  });

  function maybeSetSelection(state, dispatch, selection) {
    if (selection.eq(state.selection)) { return false }
    if (dispatch) { dispatch(state.tr.setSelection(selection).scrollIntoView()); }
    return true
  }

  function arrow$1(axis, dir) {
    return function (state, dispatch, view) {
      var sel = state.selection;
      if (sel instanceof CellSelection) {
        return maybeSetSelection(state, dispatch, Selection.near(sel.$headCell, dir))
      }
      if (axis != "horiz" && !sel.empty) { return false }
      var end = atEndOfCell(view, axis, dir);
      if (end == null) { return false }
      if (axis == "horiz") {
        return maybeSetSelection(state, dispatch, Selection.near(state.doc.resolve(sel.head + dir), dir))
      } else {
        var $cell = state.doc.resolve(end), $next = nextCell($cell, axis, dir), newSel;
        if ($next) { newSel = Selection.near($next, 1); }
        else if (dir < 0) { newSel = Selection.near(state.doc.resolve($cell.before(-1)), -1); }
        else { newSel = Selection.near(state.doc.resolve($cell.after(-1)), 1); }
        return maybeSetSelection(state, dispatch, newSel)
      }
    }
  }

  function shiftArrow(axis, dir) {
    return function (state, dispatch, view) {
      var sel = state.selection;
      if (!(sel instanceof CellSelection)) {
        var end = atEndOfCell(view, axis, dir);
        if (end == null) { return false }
        sel = new CellSelection(state.doc.resolve(end));
      }
      var $head = nextCell(sel.$headCell, axis, dir);
      if (!$head) { return false }
      return maybeSetSelection(state, dispatch, new CellSelection(sel.$anchorCell, $head))
    }
  }

  function deleteCellSelection(state, dispatch) {
    var sel = state.selection;
    if (!(sel instanceof CellSelection)) { return false }
    if (dispatch) {
      var tr = state.tr, baseContent = tableNodeTypes(state.schema).cell.createAndFill().content;
      sel.forEachCell(function (cell, pos) {
        if (!cell.content.eq(baseContent))
          { tr.replace(tr.mapping.map(pos + 1), tr.mapping.map(pos + cell.nodeSize - 1),
                     new Slice$1(baseContent, 0, 0)); }
      });
      if (tr.docChanged) { dispatch(tr); }
    }
    return true
  }

  function handleTripleClick(view, pos) {
    var doc = view.state.doc, $cell = cellAround(doc.resolve(pos));
    if (!$cell) { return false }
    view.dispatch(view.state.tr.setSelection(new CellSelection($cell)));
    return true
  }

  function handlePaste(view, _, slice) {
    if (!isInTable(view.state)) { return false }
    var cells = pastedCells(slice), sel = view.state.selection;
    if (sel instanceof CellSelection) {
      if (!cells) { cells = {width: 1, height: 1, rows: [Fragment$1.from(fitSlice(tableNodeTypes(view.state.schema).cell, slice))]}; }
      var table = sel.$anchorCell.node(-1), start = sel.$anchorCell.start(-1);
      var rect = TableMap.get(table).rectBetween(sel.$anchorCell.pos - start, sel.$headCell.pos - start);
      cells = clipCells(cells, rect.right - rect.left, rect.bottom - rect.top);
      insertCells(view.state, view.dispatch, start, rect, cells);
      return true
    } else if (cells) {
      var $cell = selectionCell(view.state), start$1 = $cell.start(-1);
      insertCells(view.state, view.dispatch, start$1, TableMap.get($cell.node(-1)).findCell($cell.pos - start$1), cells);
      return true
    } else {
      return false
    }
  }

  function handleMouseDown(view, startEvent) {
    if (startEvent.ctrlKey || startEvent.metaKey) { return }

    var startDOMCell = domInCell(view, startEvent.target), $anchor;
    if (startEvent.shiftKey && (view.state.selection instanceof CellSelection)) {
      // Adding to an existing cell selection
      setCellSelection(view.state.selection.$anchorCell, startEvent);
      startEvent.preventDefault();
    } else if (startEvent.shiftKey && startDOMCell &&
               ($anchor = cellAround(view.state.selection.$anchor)) != null &&
               cellUnderMouse(view, startEvent).pos != $anchor.pos) {
      // Adding to a selection that starts in another cell (causing a
      // cell selection to be created).
      setCellSelection($anchor, startEvent);
      startEvent.preventDefault();
    } else if (!startDOMCell) {
      // Not in a cell, let the default behavior happen.
      return
    }

    // Create and dispatch a cell selection between the given anchor and
    // the position under the mouse.
    function setCellSelection($anchor, event) {
      var $head = cellUnderMouse(view, event);
      var starting = key$2.getState(view.state) == null;
      if (!$head || !inSameTable($anchor, $head)) {
        if (starting) { $head = $anchor; }
        else { return }
      }
      var selection = new CellSelection($anchor, $head);
      if (starting || !view.state.selection.eq(selection)) {
        var tr = view.state.tr.setSelection(selection);
        if (starting) { tr.setMeta(key$2, $anchor.pos); }
        view.dispatch(tr);
      }
    }

    // Stop listening to mouse motion events.
    function stop() {
      view.root.removeEventListener("mouseup", stop);
      view.root.removeEventListener("dragstart", stop);
      view.root.removeEventListener("mousemove", move);
      if (key$2.getState(view.state) != null) { view.dispatch(view.state.tr.setMeta(key$2, -1)); }
    }

    function move(event) {
      var anchor = key$2.getState(view.state), $anchor;
      if (anchor != null) {
        // Continuing an existing cross-cell selection
        $anchor = view.state.doc.resolve(anchor);
      } else if (domInCell(view, event.target) != startDOMCell) {
        // Moving out of the initial cell -- start a new cell selection
        $anchor = cellUnderMouse(view, startEvent);
        if (!$anchor) { return stop() }
      }
      if ($anchor) { setCellSelection($anchor, event); }
    }
    view.root.addEventListener("mouseup", stop);
    view.root.addEventListener("dragstart", stop);
    view.root.addEventListener("mousemove", move);
  }

  // Check whether the cursor is at the end of a cell (so that further
  // motion would move out of the cell)
  function atEndOfCell(view, axis, dir) {
    if (!(view.state.selection instanceof TextSelection)) { return null }
    var ref = view.state.selection;
    var $head = ref.$head;
    for (var d = $head.depth - 1; d >= 0; d--) {
      var parent = $head.node(d), index = dir < 0 ? $head.index(d) : $head.indexAfter(d);
      if (index != (dir < 0 ? 0 : parent.childCount)) { return null }
      if (parent.type.spec.tableRole == "cell" || parent.type.spec.tableRole == "header_cell") {
        var cellPos = $head.before(d);
        var dirStr = axis == "vert" ? (dir > 0 ? "down" : "up") : (dir > 0 ? "right" : "left");
        return view.endOfTextblock(dirStr) ? cellPos : null
      }
    }
    return null
  }

  function domInCell(view, dom) {
    for (; dom && dom != view.dom; dom = dom.parentNode)
      { if (dom.nodeName == "TD" || dom.nodeName == "TH") { return dom } }
  }

  function cellUnderMouse(view, event) {
    var mousePos = view.posAtCoords({left: event.clientX, top: event.clientY});
    if (!mousePos) { return null }
    return mousePos ? cellAround(view.state.doc.resolve(mousePos.pos)) : null
  }

  // This file defines helpers for normalizing tables, making sure no

  var fixTablesKey = new PluginKey("fix-tables");

  // Helper for iterating through the nodes in a document that changed
  // compared to the given previous document. Useful for avoiding
  // duplicate work on each transaction.
  function changedDescendants(old, cur, offset, f) {
    var oldSize = old.childCount, curSize = cur.childCount;
    outer: for (var i = 0, j = 0; i < curSize; i++) {
      var child = cur.child(i);
      for (var scan = j, e = Math.min(oldSize, i + 3); scan < e; scan++) {
        if (old.child(scan) == child) {
          j = scan + 1;
          offset += child.nodeSize;
          continue outer
        }
      }
      f(child, offset);
      if (j < oldSize && old.child(j).sameMarkup(child))
        { changedDescendants(old.child(j), child, offset + 1, f); }
      else
        { child.nodesBetween(0, child.content.size, f, offset + 1); }
      offset += child.nodeSize;
    }
  }

  // :: (EditorState, ?EditorState) → ?Transaction
  // Inspect all tables in the given state's document and return a
  // transaction that fixes them, if necessary. If `oldState` was
  // provided, that is assumed to hold a previous, known-good state,
  // which will be used to avoid re-scanning unchanged parts of the
  // document.
  function fixTables(state, oldState) {
    var tr, check = function (node, pos) {
      if (node.type.spec.tableRole == "table") { tr = fixTable(state, node, pos, tr); }
    };
    if (!oldState) { state.doc.descendants(check); }
    else if (oldState.doc != state.doc) { changedDescendants(oldState.doc, state.doc, 0, check); }
    return tr
  }

  // : (EditorState, Node, number, ?Transaction) → ?Transaction
  // Fix the given table, if necessary. Will append to the transaction
  // it was given, if non-null, or create a new one if necessary.
  function fixTable(state, table, tablePos, tr) {
    var map = TableMap.get(table);
    if (!map.problems) { return tr }
    if (!tr) { tr = state.tr; }

    // Track which rows we must add cells to, so that we can adjust that
    // when fixing collisions.
    var mustAdd = [];
    for (var i = 0; i < map.height; i++) { mustAdd.push(0); }
    for (var i$1 = 0; i$1 < map.problems.length; i$1++) {
      var prob = map.problems[i$1];
      if (prob.type == "collision") {
        var cell = table.nodeAt(prob.pos);
        for (var j = 0; j < cell.attrs.rowspan; j++) { mustAdd[prob.row + j] += prob.n; }
        tr.setNodeMarkup(tr.mapping.map(tablePos + 1 + prob.pos), null, removeColSpan(cell.attrs, cell.attrs.colspan - prob.n, prob.n));
      } else if (prob.type == "missing") {
        mustAdd[prob.row] += prob.n;
      } else if (prob.type == "overlong_rowspan") {
        var cell$1 = table.nodeAt(prob.pos);
        tr.setNodeMarkup(tr.mapping.map(tablePos + 1 + prob.pos), null, setAttr(cell$1.attrs, "rowspan", cell$1.attrs.rowspan - prob.n));
      } else if (prob.type == "colwidth mismatch") {
        var cell$2 = table.nodeAt(prob.pos);
        tr.setNodeMarkup(tr.mapping.map(tablePos + 1 + prob.pos), null, setAttr(cell$2.attrs, "colwidth", prob.colwidth));
      }
    }
    var first, last;
    for (var i$2 = 0; i$2 < mustAdd.length; i$2++) { if (mustAdd[i$2]) {
      if (first == null) { first = i$2; }
      last = i$2;
    } }
    // Add the necessary cells, using a heuristic for whether to add the
    // cells at the start or end of the rows (if it looks like a 'bite'
    // was taken out of the table, add cells at the start of the row
    // after the bite. Otherwise add them at the end).
    for (var i$3 = 0, pos = tablePos + 1; i$3 < map.height; i$3++) {
      var row = table.child(i$3);
      var end = pos + row.nodeSize;
      var add = mustAdd[i$3];
      if (add > 0) {
        var tableNodeType = 'cell';
        if (row.firstChild) {
          tableNodeType = row.firstChild.type.spec.tableRole;
        }
        var nodes = [];
        for (var j$1 = 0; j$1 < add; j$1++)
          { nodes.push(tableNodeTypes(state.schema)[tableNodeType].createAndFill()); }
        var side = (i$3 == 0 || first == i$3 - 1) && last == i$3 ? pos + 1 : end - 1;
        tr.insert(tr.mapping.map(side), nodes);
      }
      pos = end;
    }
    return tr.setMeta(fixTablesKey, { fixTables: true })
  }

  // This file defines a number of table-related commands.

  // Helper to get the selected rectangle in a table, if any. Adds table
  // map, table node, and table start offset to the object for
  // convenience.
  function selectedRect(state) {
    var sel = state.selection, $pos = selectionCell(state);
    var table = $pos.node(-1), tableStart = $pos.start(-1), map = TableMap.get(table);
    var rect;
    if (sel instanceof CellSelection)
      { rect = map.rectBetween(sel.$anchorCell.pos - tableStart, sel.$headCell.pos - tableStart); }
    else
      { rect = map.findCell($pos.pos - tableStart); }
    rect.tableStart = tableStart;
    rect.map = map;
    rect.table = table;
    return rect
  }

  // Add a column at the given position in a table.
  function addColumn(tr, ref, col) {
    var map = ref.map;
    var tableStart = ref.tableStart;
    var table = ref.table;

    var refColumn = col > 0 ? -1 : 0;
    if (columnIsHeader(map, table, col + refColumn))
      { refColumn = col == 0 || col == map.width ? null : 0; }

    for (var row = 0; row < map.height; row++) {
      var index = row * map.width + col;
      // If this position falls inside a col-spanning cell
      if (col > 0 && col < map.width && map.map[index - 1] == map.map[index]) {
        var pos = map.map[index], cell = table.nodeAt(pos);
        tr.setNodeMarkup(tr.mapping.map(tableStart + pos), null,
                         addColSpan(cell.attrs, col - map.colCount(pos)));
        // Skip ahead if rowspan > 1
        row += cell.attrs.rowspan - 1;
      } else {
        var type = refColumn == null ? tableNodeTypes(table.type.schema).cell
            : table.nodeAt(map.map[index + refColumn]).type;
        var pos$1 = map.positionAt(row, col, table);
        tr.insert(tr.mapping.map(tableStart + pos$1), type.createAndFill());
      }
    }
    return tr
  }

  // :: (EditorState, dispatch: ?(tr: Transaction)) → bool
  // Command to add a column before the column with the selection.
  function addColumnBefore(state, dispatch) {
    if (!isInTable(state)) { return false }
    if (dispatch) {
      var rect = selectedRect(state);
      dispatch(addColumn(state.tr, rect, rect.left));
    }
    return true
  }

  // :: (EditorState, dispatch: ?(tr: Transaction)) → bool
  // Command to add a column after the column with the selection.
  function addColumnAfter(state, dispatch) {
    if (!isInTable(state)) { return false }
    if (dispatch) {
      var rect = selectedRect(state);
      dispatch(addColumn(state.tr, rect, rect.right));
    }
    return true
  }

  function removeColumn(tr, ref, col) {
    var map = ref.map;
    var table = ref.table;
    var tableStart = ref.tableStart;

    var mapStart = tr.mapping.maps.length;
    for (var row = 0; row < map.height;) {
      var index = row * map.width + col, pos = map.map[index], cell = table.nodeAt(pos);
      // If this is part of a col-spanning cell
      if ((col > 0 && map.map[index - 1] == pos) || (col < map.width - 1 && map.map[index + 1] == pos)) {
        tr.setNodeMarkup(tr.mapping.slice(mapStart).map(tableStart + pos), null,
                         removeColSpan(cell.attrs, col - map.colCount(pos)));
      } else {
        var start = tr.mapping.slice(mapStart).map(tableStart + pos);
        tr.delete(start, start + cell.nodeSize);
      }
      row += cell.attrs.rowspan;
    }
  }

  // :: (EditorState, dispatch: ?(tr: Transaction)) → bool
  // Command function that removes the selected columns from a table.
  function deleteColumn(state, dispatch) {
    if (!isInTable(state)) { return false }
    if (dispatch) {
      var rect = selectedRect(state), tr = state.tr;
      if (rect.left == 0 && rect.right == rect.map.width) { return false }
      for (var i = rect.right - 1;; i--) {
        removeColumn(tr, rect, i);
        if (i == rect.left) { break }
        rect.table = rect.tableStart ? tr.doc.nodeAt(rect.tableStart - 1) : tr.doc;
        rect.map = TableMap.get(rect.table);
      }
      dispatch(tr);
    }
    return true
  }

  function rowIsHeader(map, table, row) {
    var headerCell = tableNodeTypes(table.type.schema).header_cell;
    for (var col = 0; col < map.width; col++)
      { if (table.nodeAt(map.map[col + row * map.width]).type != headerCell)
        { return false } }
    return true
  }

  function addRow(tr, ref, row) {
    var map = ref.map;
    var tableStart = ref.tableStart;
    var table = ref.table;

    var rowPos = tableStart;
    for (var i = 0; i < row; i++) { rowPos += table.child(i).nodeSize; }
    var cells = [], refRow = row > 0 ? -1 : 0;
    if (rowIsHeader(map, table, row + refRow))
      { refRow = row == 0 || row == map.height ? null : 0; }
    for (var col = 0, index = map.width * row; col < map.width; col++, index++) {
      // Covered by a rowspan cell
      if (row > 0 && row < map.height && map.map[index] == map.map[index - map.width]) {
        var pos = map.map[index], attrs = table.nodeAt(pos).attrs;
        tr.setNodeMarkup(tableStart + pos, null, setAttr(attrs, "rowspan", attrs.rowspan + 1));
        col += attrs.colspan - 1;
      } else {
        var type = refRow == null ? tableNodeTypes(table.type.schema).cell
            : table.nodeAt(map.map[index + refRow * map.width]).type;
        cells.push(type.createAndFill());
      }
    }
    tr.insert(rowPos, tableNodeTypes(table.type.schema).row.create(null, cells));
    return tr
  }

  // :: (EditorState, dispatch: ?(tr: Transaction)) → bool
  // Add a table row before the selection.
  function addRowBefore(state, dispatch) {
    if (!isInTable(state)) { return false }
    if (dispatch) {
      var rect = selectedRect(state);
      dispatch(addRow(state.tr, rect, rect.top));
    }
    return true
  }

  // :: (EditorState, dispatch: ?(tr: Transaction)) → bool
  // Add a table row after the selection.
  function addRowAfter(state, dispatch) {
    if (!isInTable(state)) { return false }
    if (dispatch) {
      var rect = selectedRect(state);
      dispatch(addRow(state.tr, rect, rect.bottom));
    }
    return true
  }

  function removeRow(tr, ref, row) {
    var map = ref.map;
    var table = ref.table;
    var tableStart = ref.tableStart;

    var rowPos = 0;
    for (var i = 0; i < row; i++) { rowPos += table.child(i).nodeSize; }
    var nextRow = rowPos + table.child(row).nodeSize;

    var mapFrom = tr.mapping.maps.length;
    tr.delete(rowPos + tableStart, nextRow + tableStart);

    for (var col = 0, index = row * map.width; col < map.width; col++, index++) {
      var pos = map.map[index];
      if (row > 0 && pos == map.map[index - map.width]) {
        // If this cell starts in the row above, simply reduce its rowspan
        var attrs = table.nodeAt(pos).attrs;
        tr.setNodeMarkup(tr.mapping.slice(mapFrom).map(pos + tableStart), null, setAttr(attrs, "rowspan", attrs.rowspan - 1));
        col += attrs.colspan - 1;
      } else if (row < map.width && pos == map.map[index + map.width]) {
        // Else, if it continues in the row below, it has to be moved down
        var cell = table.nodeAt(pos);
        var copy = cell.type.create(setAttr(cell.attrs, "rowspan", cell.attrs.rowspan - 1), cell.content);
        var newPos = map.positionAt(row + 1, col, table);
        tr.insert(tr.mapping.slice(mapFrom).map(tableStart + newPos), copy);
        col += cell.attrs.colspan - 1;
      }
    }
  }

  // :: (EditorState, dispatch: ?(tr: Transaction)) → bool
  // Remove the selected rows from a table.
  function deleteRow(state, dispatch) {
    if (!isInTable(state)) { return false }
    if (dispatch) {
      var rect = selectedRect(state), tr = state.tr;
      if (rect.top == 0 && rect.bottom == rect.map.height) { return false }
      for (var i = rect.bottom - 1;; i--) {
        removeRow(tr, rect, i);
        if (i == rect.top) { break }
        rect.table = rect.tableStart ? tr.doc.nodeAt(rect.tableStart - 1) : tr.doc;
        rect.map = TableMap.get(rect.table);
      }
      dispatch(tr);
    }
    return true
  }

  function deprecated_toggleHeader(type) {
    return function(state, dispatch) {
      if (!isInTable(state)) { return false }
      if (dispatch) {
        var types = tableNodeTypes(state.schema);
        var rect = selectedRect(state), tr = state.tr;
        var cells = rect.map.cellsInRect(type == "column" ? new Rect(rect.left, 0, rect.right, rect.map.height) :
                                         type == "row" ? new Rect(0, rect.top, rect.map.width, rect.bottom) : rect);
        var nodes = cells.map(function (pos) { return rect.table.nodeAt(pos); });
        for (var i = 0; i < cells.length; i++) // Remove headers, if any
          { if (nodes[i].type == types.header_cell)
            { tr.setNodeMarkup(rect.tableStart + cells[i], types.cell, nodes[i].attrs); } }
        if (tr.steps.length == 0) { for (var i$1 = 0; i$1 < cells.length; i$1++) // No headers removed, add instead
          { tr.setNodeMarkup(rect.tableStart + cells[i$1], types.header_cell, nodes[i$1].attrs); } }
        dispatch(tr);
      }
      return true
    }
  }

  function isHeaderEnabledByType(type, rect, types) {
    // Get cell positions for first row or first column
    var cellPositions = rect.map.cellsInRect({
      left: 0,
      top: 0,
      right: type == "row" ? rect.map.width : 1,
      bottom: type == "column" ? rect.map.height : 1,
    });

    for (var i = 0; i < cellPositions.length; i++) {
      var cell = rect.table.nodeAt(cellPositions[i]);
      if (cell && cell.type !== types.header_cell) {
        return false
      }
    }

    return true
  }

  // :: (string, ?{ useDeprecatedLogic: bool }) → (EditorState, dispatch: ?(tr: Transaction)) → bool
  // Toggles between row/column header and normal cells (Only applies to first row/column).
  // For deprecated behavior pass `useDeprecatedLogic` in options with true.
  function toggleHeader(type, options) {
    options = options || { useDeprecatedLogic: false };

    if (options.useDeprecatedLogic)
      { return deprecated_toggleHeader(type) }

    return function(state, dispatch) {
      if (!isInTable(state)) { return false }
      if (dispatch) {
        var types = tableNodeTypes(state.schema);
        var rect = selectedRect(state), tr = state.tr;

        var isHeaderRowEnabled = isHeaderEnabledByType("row", rect, types);
        var isHeaderColumnEnabled = isHeaderEnabledByType("column", rect, types);

        var isHeaderEnabled = type === "column" ? isHeaderRowEnabled :
                              type === "row"    ? isHeaderColumnEnabled : false;

        var selectionStartsAt = isHeaderEnabled ? 1 : 0;

        var cellsRect = type == "column" ? new Rect(0, selectionStartsAt, 1, rect.map.height) :
                        type == "row" ? new Rect(selectionStartsAt, 0, rect.map.width, 1) : rect;

        var newType = type == "column" ? isHeaderColumnEnabled ? types.cell : types.header_cell :
                      type == "row" ? isHeaderRowEnabled ? types.cell : types.header_cell : types.cell;

        rect.map.cellsInRect(cellsRect).forEach(function (relativeCellPos) {
          var cellPos = relativeCellPos + rect.tableStart;
          var cell = tr.doc.nodeAt(cellPos);

          if (cell) {
            tr.setNodeMarkup(cellPos, newType, cell.attrs);
          }
        });

        dispatch(tr);
      }
      return true
    }
  }

  // :: (EditorState, dispatch: ?(tr: Transaction)) → bool
  // Toggles whether the selected row contains header cells.
  var toggleHeaderRow = toggleHeader("row", { useDeprecatedLogic: true });

  // :: (EditorState, dispatch: ?(tr: Transaction)) → bool
  // Toggles whether the selected column contains header cells.
  toggleHeader("column", { useDeprecatedLogic: true });

  // :: (EditorState, dispatch: ?(tr: Transaction)) → bool
  // Toggles whether the selected cells are header cells.
  toggleHeader("cell", { useDeprecatedLogic: true });

  function findNextCell($cell, dir) {
    if (dir < 0) {
      var before = $cell.nodeBefore;
      if (before) { return $cell.pos - before.nodeSize }
      for (var row = $cell.index(-1) - 1, rowEnd = $cell.before(); row >= 0; row--) {
        var rowNode = $cell.node(-1).child(row);
        if (rowNode.childCount) { return rowEnd - 1 - rowNode.lastChild.nodeSize }
        rowEnd -= rowNode.nodeSize;
      }
    } else {
      if ($cell.index() < $cell.parent.childCount - 1) { return $cell.pos + $cell.nodeAfter.nodeSize }
      var table = $cell.node(-1);
      for (var row$1 = $cell.indexAfter(-1), rowStart = $cell.after(); row$1 < table.childCount; row$1++) {
        var rowNode$1 = table.child(row$1);
        if (rowNode$1.childCount) { return rowStart + 1 }
        rowStart += rowNode$1.nodeSize;
      }
    }
  }

  // :: (number) → (EditorState, dispatch: ?(tr: Transaction)) → bool
  // Returns a command for selecting the next (direction=1) or previous
  // (direction=-1) cell in a table.
  function goToNextCell(direction) {
    return function(state, dispatch) {
      if (!isInTable(state)) { return false }
      var cell = findNextCell(selectionCell(state), direction);
      if (cell == null) { return }
      if (dispatch) {
        var $cell = state.doc.resolve(cell);
        dispatch(state.tr.setSelection(TextSelection.between($cell, moveCellForward($cell))).scrollIntoView());
      }
      return true
    }
  }

  // :: (EditorState, ?(tr: Transaction)) → bool
  // Deletes the table around the selection, if any.
  function deleteTable(state, dispatch) {
    var $pos = state.selection.$anchor;
    for (var d = $pos.depth; d > 0; d--) {
      var node = $pos.node(d);
      if (node.type.spec.tableRole == "table") {
        if (dispatch) { dispatch(state.tr.delete($pos.before(d), $pos.after(d)).scrollIntoView()); }
        return true
      }
    }
    return false
  }

  new PluginKey("tableColumnResizing");

  // This file defines a plugin that handles the drawing of cell

  // :: () → Plugin
  //
  // Creates a [plugin](http://prosemirror.net/docs/ref/#state.Plugin)
  // that, when added to an editor, enables cell-selection, handles
  // cell-based copy/paste, and makes sure tables stay well-formed (each
  // row has the same width, and cells don't overlap).
  //
  // You should probably put this plugin near the end of your array of
  // plugins, since it handles mouse and arrow key events in tables
  // rather broadly, and other plugins, like the gap cursor or the
  // column-width dragging plugin, might want to get a turn first to
  // perform more specific behavior.
  function tableEditing(ref) {
    if ( ref === void 0 ) { ref = {}; }
    var allowTableNodeSelection = ref.allowTableNodeSelection; if ( allowTableNodeSelection === void 0 ) { allowTableNodeSelection = false; }

    return new Plugin({
      key: key$2,

      // This piece of state is used to remember when a mouse-drag
      // cell-selection is happening, so that it can continue even as
      // transactions (which might move its anchor cell) come in.
      state: {
        init: function init() { return null },
        apply: function apply(tr, cur) {
          var set = tr.getMeta(key$2);
          if (set != null) { return set == -1 ? null : set }
          if (cur == null || !tr.docChanged) { return cur }
          var ref = tr.mapping.mapResult(cur);
          var deleted = ref.deleted;
          var pos = ref.pos;
          return deleted ? null : pos
        }
      },

      props: {
        decorations: drawCellSelection,

        handleDOMEvents: {
          mousedown: handleMouseDown
        },

        createSelectionBetween: function createSelectionBetween(view) {
          if (key$2.getState(view.state) != null) { return view.state.selection }
        },

        handleTripleClick: handleTripleClick,

        handleKeyDown: handleKeyDown$1,

        handlePaste: handlePaste
      },

      appendTransaction: function appendTransaction(_, oldState, state) {
        return normalizeSelection(state, fixTables(state, oldState), allowTableNodeSelection)
      }
    })
  }

  var prefix = "ProseMirror-prompt";

  var Promt = function Promt(options) {
      this.options = options;
      this.render();
  };

  Promt.prototype.render = function render () {
      $('.ProseMirror-prompt').remove();
      this.$wrapper = $('<div>').addClass(prefix).appendTo($('body'));
      this.buildForm();
      this.initEvents();
      this.initWrapper();
  };

  Promt.prototype.initWrapper = function initWrapper () {
      var box = this.$wrapper[0].getBoundingClientRect();
      this.$wrapper.css({
          top: ((window.innerHeight - box.height) / 2)+ "px",
          left: ((window.innerWidth - box.width) / 2) + "px"
      });

      this.$wrapper.find('select:visible, input[type="text"]:visible, textarea:visible, [contenteditable="true"]:visible').first().focus();
  };

  Promt.prototype.buildForm = function buildForm () {
          var this$1$1 = this;

      this.$form = $('<form>').appendTo(this.$wrapper);

      if (this.options.title) {
          this.$form.append('<h5>'+this.options.title+'</h5>');
      }

      this.buildFormFields();

      this.domFields.forEach(function (field) {
          this$1$1.$form.append(field);
      });

      this.$form.on('submit', function (e) {
          e.preventDefault();
          this$1$1.submit();
      });

      this.buildButtons();
  };

  Promt.prototype.buildFormFields = function buildFormFields () {
      this.domFields = [];
      for (var name in this.options.fields) {
          var field = this.options.fields[name];
          var $field = $('<div>').append('<label>'+(field.options.label || name)+':</label>').append(this.options.fields[name].render());
          this.domFields.push($field[0]);
      }
  };

  Promt.prototype.buildButtons = function buildButtons () {
          var this$1$1 = this;

      this.$buttons = $('<div>').addClass(prefix + "-buttons");
      // TODO: translate text!
      $('<button type="submit" class="btn btn-primary">').addClass(prefix + "-submit").text('OK').appendTo(this.$buttons);

      this.$buttons.append(document.createTextNode(' '));

      $('<button type="button" class="btn btn-default">').addClass(prefix + "-cancel").text('Cancel').appendTo(this.$buttons)
          .on('click', function () {this$1$1.close();});

      this.$form.append(this.$buttons);
  };

  Promt.prototype.submit = function submit () {
      var params = this.getValues();
      if (params) {
          this.close();
          this.options.callback(params);
      }
  };

  Promt.prototype.getValues = function getValues () {
      var result = Object.create(null);
      var i = 0;

      for (var name in this.options.fields) {
          var field = this.options.fields[name];
          var dom = this.domFields[i++];

          var value = field.read(dom);
          var bad = field.validate(value);

          if (bad) {
              this.reportInvalid(dom, bad);
              return null
          }
          result[name] = field.clean(value);
      }

      return result
  };

  Promt.prototype.reportInvalid = function reportInvalid (dom, message) {
      // FIXME this is awful and needs a lot more work
      var parent = dom.parentNode;
      var msg = parent.appendChild(document.createElement("div"));
      msg.style.left = (dom.offsetLeft + dom.offsetWidth + 2) + "px";
      msg.style.top = (dom.offsetTop - 5) + "px";
      msg.className = "ProseMirror-invalid";
      msg.textContent = message;
      setTimeout(function () { return parent.removeChild(msg); }, 1500);
  };

  Promt.prototype.initEvents = function initEvents () {
          var this$1$1 = this;

      this.$form.on("keydown", function (e) {
          if (e.keyCode == 27) {
              e.preventDefault();
              this$1$1.close();
          } else if (e.keyCode == 13 && !(e.ctrlKey || e.metaKey || e.shiftKey)) {
              e.preventDefault();
              this$1$1.submit();
          } else if (e.keyCode == 9) {
              window.setTimeout(function () {
                  if (!$.contains(this$1$1.$wrapper[0], document.activeElement)) {
                      this$1$1.close();
                  }
              }, 500);
          }
      }).on('mousedown', function (e) {
          if (!$.contains(this$1$1.$wrapper[0], e.target)) {
              this$1$1.close();
          }
      });
  };

  Promt.prototype.close = function close () {
      this.$wrapper.remove();
  };

  function openPrompt(options) {
      return new Promt(options);
  }

  // ::- The type of field that `FieldPrompt` expects to be passed to it.
  var Field = function Field(options) {
      this.options = options;
  };

  // render:: (state: EditorState, props: Object) → dom.Node
  // Render the field to the DOM. Should be implemented by all subclasses.

  // :: (dom.Node) → any
  // Read the field's value from its DOM node.
  Field.prototype.read = function read (dom) {
      if(dom.value) {
          return dom.value;
      } else {
          return $(dom).find('input, select')[0].value;
      }
  };

  // :: (any) → ?string
  // A field-type-specific validation function.
  Field.prototype.validateType = function validateType (_value) {
  };

  Field.prototype.validate = function validate (value) {
      if (!value && this.options.required)
          { return "Required field" }
      return this.validateType(value) || (this.options.validate && this.options.validate(value))
  };

  Field.prototype.clean = function clean (value) {
      return this.options.clean ? this.options.clean(value) : value
  };

  // ::- A field class for single-line text fields.
  var TextField = /*@__PURE__*/(function (Field) {
      function TextField () {
          Field.apply(this, arguments);
      }

      if ( Field ) TextField.__proto__ = Field;
      TextField.prototype = Object.create( Field && Field.prototype );
      TextField.prototype.constructor = TextField;

      TextField.prototype.render = function render () {
          var input = document.createElement("input");
          input.type = "text";
          input.className = 'form-control';
          input.value = this.options.value || "";
          input.autocomplete = "off";
          return input
      };

      return TextField;
  }(Field));


  // ::- A field class for dropdown fields based on a plain `<select>`
  // tag. Expects an option `options`, which should be an array of
  // `{value: string, label: string}` objects, or a function taking a
  // `ProseMirror` instance and returning such an array.
  var SelectField = /*@__PURE__*/(function (Field) {
      function SelectField () {
          Field.apply(this, arguments);
      }

      if ( Field ) SelectField.__proto__ = Field;
      SelectField.prototype = Object.create( Field && Field.prototype );
      SelectField.prototype.constructor = SelectField;

      SelectField.prototype.render = function render () {
          var this$1$1 = this;

          var select = document.createElement("select");
          select.className = 'form-control';

          this.options.options.forEach(function (o) {
              var opt = select.appendChild(document.createElement("option"));
              opt.value = o.value;
              opt.selected = o.value == this$1$1.options.value;
              opt.label = o.label;
          });
          return select
      };

      return SelectField;
  }(Field));

  var prompt = /*#__PURE__*/Object.freeze({
    __proto__: null,
    openPrompt: openPrompt,
    Field: Field,
    TextField: TextField,
    SelectField: SelectField
  });

  function createCommonjsModule(fn, module) {
  	return module = { exports: {} }, fn(module, module.exports), module.exports;
  }

  function getCjsExportFromNamespace (n) {
  	return n && n['default'] || n;
  }

  var Aacute$1 = "Á";
  var aacute$1 = "á";
  var Abreve$1 = "Ă";
  var abreve$1 = "ă";
  var ac$1 = "∾";
  var acd$1 = "∿";
  var acE$1 = "∾̳";
  var Acirc$1 = "Â";
  var acirc$1 = "â";
  var acute$1 = "´";
  var Acy$1 = "А";
  var acy$1 = "а";
  var AElig$1 = "Æ";
  var aelig$1 = "æ";
  var af$1 = "⁡";
  var Afr$1 = "𝔄";
  var afr$1 = "𝔞";
  var Agrave$1 = "À";
  var agrave$1 = "à";
  var alefsym$1 = "ℵ";
  var aleph$1 = "ℵ";
  var Alpha$1 = "Α";
  var alpha$1 = "α";
  var Amacr$1 = "Ā";
  var amacr$1 = "ā";
  var amalg$1 = "⨿";
  var amp$1 = "&";
  var AMP$1 = "&";
  var andand$1 = "⩕";
  var And$1 = "⩓";
  var and$1 = "∧";
  var andd$1 = "⩜";
  var andslope$1 = "⩘";
  var andv$1 = "⩚";
  var ang$1 = "∠";
  var ange$1 = "⦤";
  var angle$1 = "∠";
  var angmsdaa$1 = "⦨";
  var angmsdab$1 = "⦩";
  var angmsdac$1 = "⦪";
  var angmsdad$1 = "⦫";
  var angmsdae$1 = "⦬";
  var angmsdaf$1 = "⦭";
  var angmsdag$1 = "⦮";
  var angmsdah$1 = "⦯";
  var angmsd$1 = "∡";
  var angrt$1 = "∟";
  var angrtvb$1 = "⊾";
  var angrtvbd$1 = "⦝";
  var angsph$1 = "∢";
  var angst$1 = "Å";
  var angzarr$1 = "⍼";
  var Aogon$1 = "Ą";
  var aogon$1 = "ą";
  var Aopf$1 = "𝔸";
  var aopf$1 = "𝕒";
  var apacir$1 = "⩯";
  var ap$1 = "≈";
  var apE$1 = "⩰";
  var ape$1 = "≊";
  var apid$1 = "≋";
  var apos$1 = "'";
  var ApplyFunction$1 = "⁡";
  var approx$1 = "≈";
  var approxeq$1 = "≊";
  var Aring$1 = "Å";
  var aring$1 = "å";
  var Ascr$1 = "𝒜";
  var ascr$1 = "𝒶";
  var Assign$1 = "≔";
  var ast$1 = "*";
  var asymp$1 = "≈";
  var asympeq$1 = "≍";
  var Atilde$1 = "Ã";
  var atilde$1 = "ã";
  var Auml$1 = "Ä";
  var auml$1 = "ä";
  var awconint$1 = "∳";
  var awint$1 = "⨑";
  var backcong$1 = "≌";
  var backepsilon$1 = "϶";
  var backprime$1 = "‵";
  var backsim$1 = "∽";
  var backsimeq$1 = "⋍";
  var Backslash$1 = "∖";
  var Barv$1 = "⫧";
  var barvee$1 = "⊽";
  var barwed$1 = "⌅";
  var Barwed$1 = "⌆";
  var barwedge$1 = "⌅";
  var bbrk$1 = "⎵";
  var bbrktbrk$1 = "⎶";
  var bcong$1 = "≌";
  var Bcy$1 = "Б";
  var bcy$1 = "б";
  var bdquo$1 = "„";
  var becaus$1 = "∵";
  var because$1 = "∵";
  var Because$1 = "∵";
  var bemptyv$1 = "⦰";
  var bepsi$1 = "϶";
  var bernou$1 = "ℬ";
  var Bernoullis$1 = "ℬ";
  var Beta$1 = "Β";
  var beta$1 = "β";
  var beth$1 = "ℶ";
  var between$1 = "≬";
  var Bfr$1 = "𝔅";
  var bfr$1 = "𝔟";
  var bigcap$1 = "⋂";
  var bigcirc$1 = "◯";
  var bigcup$1 = "⋃";
  var bigodot$1 = "⨀";
  var bigoplus$1 = "⨁";
  var bigotimes$1 = "⨂";
  var bigsqcup$1 = "⨆";
  var bigstar$1 = "★";
  var bigtriangledown$1 = "▽";
  var bigtriangleup$1 = "△";
  var biguplus$1 = "⨄";
  var bigvee$1 = "⋁";
  var bigwedge$1 = "⋀";
  var bkarow$1 = "⤍";
  var blacklozenge$1 = "⧫";
  var blacksquare$1 = "▪";
  var blacktriangle$1 = "▴";
  var blacktriangledown$1 = "▾";
  var blacktriangleleft$1 = "◂";
  var blacktriangleright$1 = "▸";
  var blank$1 = "␣";
  var blk12$1 = "▒";
  var blk14$1 = "░";
  var blk34$1 = "▓";
  var block$3 = "█";
  var bne$1 = "=⃥";
  var bnequiv$1 = "≡⃥";
  var bNot$1 = "⫭";
  var bnot$1 = "⌐";
  var Bopf$1 = "𝔹";
  var bopf$1 = "𝕓";
  var bot$1 = "⊥";
  var bottom$1 = "⊥";
  var bowtie$1 = "⋈";
  var boxbox$1 = "⧉";
  var boxdl$1 = "┐";
  var boxdL$1 = "╕";
  var boxDl$1 = "╖";
  var boxDL$1 = "╗";
  var boxdr$1 = "┌";
  var boxdR$1 = "╒";
  var boxDr$1 = "╓";
  var boxDR$1 = "╔";
  var boxh$1 = "─";
  var boxH$1 = "═";
  var boxhd$1 = "┬";
  var boxHd$1 = "╤";
  var boxhD$1 = "╥";
  var boxHD$1 = "╦";
  var boxhu$1 = "┴";
  var boxHu$1 = "╧";
  var boxhU$1 = "╨";
  var boxHU$1 = "╩";
  var boxminus$1 = "⊟";
  var boxplus$1 = "⊞";
  var boxtimes$1 = "⊠";
  var boxul$1 = "┘";
  var boxuL$1 = "╛";
  var boxUl$1 = "╜";
  var boxUL$1 = "╝";
  var boxur$1 = "└";
  var boxuR$1 = "╘";
  var boxUr$1 = "╙";
  var boxUR$1 = "╚";
  var boxv$1 = "│";
  var boxV$1 = "║";
  var boxvh$1 = "┼";
  var boxvH$1 = "╪";
  var boxVh$1 = "╫";
  var boxVH$1 = "╬";
  var boxvl$1 = "┤";
  var boxvL$1 = "╡";
  var boxVl$1 = "╢";
  var boxVL$1 = "╣";
  var boxvr$1 = "├";
  var boxvR$1 = "╞";
  var boxVr$1 = "╟";
  var boxVR$1 = "╠";
  var bprime$1 = "‵";
  var breve$1 = "˘";
  var Breve$1 = "˘";
  var brvbar$1 = "¦";
  var bscr$1 = "𝒷";
  var Bscr$1 = "ℬ";
  var bsemi$1 = "⁏";
  var bsim$1 = "∽";
  var bsime$1 = "⋍";
  var bsolb$1 = "⧅";
  var bsol$1 = "\\";
  var bsolhsub$1 = "⟈";
  var bull$1 = "•";
  var bullet$1 = "•";
  var bump$1 = "≎";
  var bumpE$1 = "⪮";
  var bumpe$1 = "≏";
  var Bumpeq$1 = "≎";
  var bumpeq$1 = "≏";
  var Cacute$1 = "Ć";
  var cacute$1 = "ć";
  var capand$1 = "⩄";
  var capbrcup$1 = "⩉";
  var capcap$1 = "⩋";
  var cap$1 = "∩";
  var Cap$1 = "⋒";
  var capcup$1 = "⩇";
  var capdot$1 = "⩀";
  var CapitalDifferentialD$1 = "ⅅ";
  var caps$1 = "∩︀";
  var caret$1 = "⁁";
  var caron$1 = "ˇ";
  var Cayleys$1 = "ℭ";
  var ccaps$1 = "⩍";
  var Ccaron$1 = "Č";
  var ccaron$1 = "č";
  var Ccedil$1 = "Ç";
  var ccedil$1 = "ç";
  var Ccirc$1 = "Ĉ";
  var ccirc$1 = "ĉ";
  var Cconint$1 = "∰";
  var ccups$1 = "⩌";
  var ccupssm$1 = "⩐";
  var Cdot$1 = "Ċ";
  var cdot$1 = "ċ";
  var cedil$1 = "¸";
  var Cedilla$1 = "¸";
  var cemptyv$1 = "⦲";
  var cent$1 = "¢";
  var centerdot$1 = "·";
  var CenterDot$1 = "·";
  var cfr$1 = "𝔠";
  var Cfr$1 = "ℭ";
  var CHcy$1 = "Ч";
  var chcy$1 = "ч";
  var check$1 = "✓";
  var checkmark$1 = "✓";
  var Chi$1 = "Χ";
  var chi$1 = "χ";
  var circ$1 = "ˆ";
  var circeq$1 = "≗";
  var circlearrowleft$1 = "↺";
  var circlearrowright$1 = "↻";
  var circledast$1 = "⊛";
  var circledcirc$1 = "⊚";
  var circleddash$1 = "⊝";
  var CircleDot$1 = "⊙";
  var circledR$1 = "®";
  var circledS$1 = "Ⓢ";
  var CircleMinus$1 = "⊖";
  var CirclePlus$1 = "⊕";
  var CircleTimes$1 = "⊗";
  var cir$1 = "○";
  var cirE$1 = "⧃";
  var cire$1 = "≗";
  var cirfnint$1 = "⨐";
  var cirmid$1 = "⫯";
  var cirscir$1 = "⧂";
  var ClockwiseContourIntegral$1 = "∲";
  var CloseCurlyDoubleQuote$1 = "”";
  var CloseCurlyQuote$1 = "’";
  var clubs$3 = "♣";
  var clubsuit$1 = "♣";
  var colon$1 = ":";
  var Colon$1 = "∷";
  var Colone$1 = "⩴";
  var colone$1 = "≔";
  var coloneq$1 = "≔";
  var comma$1 = ",";
  var commat$1 = "@";
  var comp$1 = "∁";
  var compfn$1 = "∘";
  var complement$1 = "∁";
  var complexes$1 = "ℂ";
  var cong$1 = "≅";
  var congdot$1 = "⩭";
  var Congruent$1 = "≡";
  var conint$1 = "∮";
  var Conint$1 = "∯";
  var ContourIntegral$1 = "∮";
  var copf$1 = "𝕔";
  var Copf$1 = "ℂ";
  var coprod$1 = "∐";
  var Coproduct$1 = "∐";
  var copy$1 = "©";
  var COPY$1 = "©";
  var copysr$1 = "℗";
  var CounterClockwiseContourIntegral$1 = "∳";
  var crarr$1 = "↵";
  var cross$1 = "✗";
  var Cross$1 = "⨯";
  var Cscr$1 = "𝒞";
  var cscr$1 = "𝒸";
  var csub$1 = "⫏";
  var csube$1 = "⫑";
  var csup$1 = "⫐";
  var csupe$1 = "⫒";
  var ctdot$1 = "⋯";
  var cudarrl$1 = "⤸";
  var cudarrr$1 = "⤵";
  var cuepr$1 = "⋞";
  var cuesc$1 = "⋟";
  var cularr$1 = "↶";
  var cularrp$1 = "⤽";
  var cupbrcap$1 = "⩈";
  var cupcap$1 = "⩆";
  var CupCap$1 = "≍";
  var cup$1 = "∪";
  var Cup$1 = "⋓";
  var cupcup$1 = "⩊";
  var cupdot$1 = "⊍";
  var cupor$1 = "⩅";
  var cups$1 = "∪︀";
  var curarr$1 = "↷";
  var curarrm$1 = "⤼";
  var curlyeqprec$1 = "⋞";
  var curlyeqsucc$1 = "⋟";
  var curlyvee$1 = "⋎";
  var curlywedge$1 = "⋏";
  var curren$1 = "¤";
  var curvearrowleft$1 = "↶";
  var curvearrowright$1 = "↷";
  var cuvee$1 = "⋎";
  var cuwed$1 = "⋏";
  var cwconint$1 = "∲";
  var cwint$1 = "∱";
  var cylcty$1 = "⌭";
  var dagger$3 = "†";
  var Dagger$1 = "‡";
  var daleth$1 = "ℸ";
  var darr$1 = "↓";
  var Darr$1 = "↡";
  var dArr$1 = "⇓";
  var dash$3 = "‐";
  var Dashv$1 = "⫤";
  var dashv$1 = "⊣";
  var dbkarow$1 = "⤏";
  var dblac$1 = "˝";
  var Dcaron$1 = "Ď";
  var dcaron$1 = "ď";
  var Dcy$1 = "Д";
  var dcy$1 = "д";
  var ddagger$1 = "‡";
  var ddarr$1 = "⇊";
  var DD$1 = "ⅅ";
  var dd$1 = "ⅆ";
  var DDotrahd$1 = "⤑";
  var ddotseq$1 = "⩷";
  var deg$1 = "°";
  var Del$1 = "∇";
  var Delta$1 = "Δ";
  var delta$1 = "δ";
  var demptyv$1 = "⦱";
  var dfisht$1 = "⥿";
  var Dfr$1 = "𝔇";
  var dfr$1 = "𝔡";
  var dHar$1 = "⥥";
  var dharl$1 = "⇃";
  var dharr$1 = "⇂";
  var DiacriticalAcute$1 = "´";
  var DiacriticalDot$1 = "˙";
  var DiacriticalDoubleAcute$1 = "˝";
  var DiacriticalGrave$1 = "`";
  var DiacriticalTilde$1 = "˜";
  var diam$1 = "⋄";
  var diamond$1 = "⋄";
  var Diamond$1 = "⋄";
  var diamondsuit$1 = "♦";
  var diams$1 = "♦";
  var die$1 = "¨";
  var DifferentialD$1 = "ⅆ";
  var digamma$1 = "ϝ";
  var disin$1 = "⋲";
  var div$1 = "÷";
  var divide$1 = "÷";
  var divideontimes$1 = "⋇";
  var divonx$1 = "⋇";
  var DJcy$1 = "Ђ";
  var djcy$1 = "ђ";
  var dlcorn$1 = "⌞";
  var dlcrop$1 = "⌍";
  var dollar$3 = "$";
  var Dopf$1 = "𝔻";
  var dopf$1 = "𝕕";
  var Dot$1 = "¨";
  var dot$1 = "˙";
  var DotDot$1 = "⃜";
  var doteq$1 = "≐";
  var doteqdot$1 = "≑";
  var DotEqual$1 = "≐";
  var dotminus$1 = "∸";
  var dotplus$1 = "∔";
  var dotsquare$1 = "⊡";
  var doublebarwedge$1 = "⌆";
  var DoubleContourIntegral$1 = "∯";
  var DoubleDot$1 = "¨";
  var DoubleDownArrow$1 = "⇓";
  var DoubleLeftArrow$1 = "⇐";
  var DoubleLeftRightArrow$1 = "⇔";
  var DoubleLeftTee$1 = "⫤";
  var DoubleLongLeftArrow$1 = "⟸";
  var DoubleLongLeftRightArrow$1 = "⟺";
  var DoubleLongRightArrow$1 = "⟹";
  var DoubleRightArrow$1 = "⇒";
  var DoubleRightTee$1 = "⊨";
  var DoubleUpArrow$1 = "⇑";
  var DoubleUpDownArrow$1 = "⇕";
  var DoubleVerticalBar$1 = "∥";
  var DownArrowBar$1 = "⤓";
  var downarrow$1 = "↓";
  var DownArrow$1 = "↓";
  var Downarrow$1 = "⇓";
  var DownArrowUpArrow$1 = "⇵";
  var DownBreve$1 = "̑";
  var downdownarrows$1 = "⇊";
  var downharpoonleft$1 = "⇃";
  var downharpoonright$1 = "⇂";
  var DownLeftRightVector$1 = "⥐";
  var DownLeftTeeVector$1 = "⥞";
  var DownLeftVectorBar$1 = "⥖";
  var DownLeftVector$1 = "↽";
  var DownRightTeeVector$1 = "⥟";
  var DownRightVectorBar$1 = "⥗";
  var DownRightVector$1 = "⇁";
  var DownTeeArrow$1 = "↧";
  var DownTee$1 = "⊤";
  var drbkarow$1 = "⤐";
  var drcorn$1 = "⌟";
  var drcrop$1 = "⌌";
  var Dscr$1 = "𝒟";
  var dscr$1 = "𝒹";
  var DScy$1 = "Ѕ";
  var dscy$1 = "ѕ";
  var dsol$1 = "⧶";
  var Dstrok$1 = "Đ";
  var dstrok$1 = "đ";
  var dtdot$1 = "⋱";
  var dtri$1 = "▿";
  var dtrif$1 = "▾";
  var duarr$1 = "⇵";
  var duhar$1 = "⥯";
  var dwangle$1 = "⦦";
  var DZcy$1 = "Џ";
  var dzcy$1 = "џ";
  var dzigrarr$1 = "⟿";
  var Eacute$1 = "É";
  var eacute$1 = "é";
  var easter$1 = "⩮";
  var Ecaron$1 = "Ě";
  var ecaron$1 = "ě";
  var Ecirc$1 = "Ê";
  var ecirc$1 = "ê";
  var ecir$1 = "≖";
  var ecolon$1 = "≕";
  var Ecy$1 = "Э";
  var ecy$1 = "э";
  var eDDot$1 = "⩷";
  var Edot$1 = "Ė";
  var edot$1 = "ė";
  var eDot$1 = "≑";
  var ee$1 = "ⅇ";
  var efDot$1 = "≒";
  var Efr$1 = "𝔈";
  var efr$1 = "𝔢";
  var eg$1 = "⪚";
  var Egrave$1 = "È";
  var egrave$1 = "è";
  var egs$1 = "⪖";
  var egsdot$1 = "⪘";
  var el$1 = "⪙";
  var Element$1 = "∈";
  var elinters$1 = "⏧";
  var ell$1 = "ℓ";
  var els$1 = "⪕";
  var elsdot$1 = "⪗";
  var Emacr$1 = "Ē";
  var emacr$1 = "ē";
  var empty$1 = "∅";
  var emptyset$1 = "∅";
  var EmptySmallSquare$1 = "◻";
  var emptyv$1 = "∅";
  var EmptyVerySmallSquare$1 = "▫";
  var emsp13$1 = " ";
  var emsp14$1 = " ";
  var emsp$1 = " ";
  var ENG$1 = "Ŋ";
  var eng$1 = "ŋ";
  var ensp$1 = " ";
  var Eogon$1 = "Ę";
  var eogon$1 = "ę";
  var Eopf$1 = "𝔼";
  var eopf$1 = "𝕖";
  var epar$1 = "⋕";
  var eparsl$1 = "⧣";
  var eplus$1 = "⩱";
  var epsi$1 = "ε";
  var Epsilon$1 = "Ε";
  var epsilon$1 = "ε";
  var epsiv$1 = "ϵ";
  var eqcirc$1 = "≖";
  var eqcolon$1 = "≕";
  var eqsim$1 = "≂";
  var eqslantgtr$1 = "⪖";
  var eqslantless$1 = "⪕";
  var Equal$1 = "⩵";
  var equals$1 = "=";
  var EqualTilde$1 = "≂";
  var equest$1 = "≟";
  var Equilibrium$1 = "⇌";
  var equiv$1 = "≡";
  var equivDD$1 = "⩸";
  var eqvparsl$1 = "⧥";
  var erarr$1 = "⥱";
  var erDot$1 = "≓";
  var escr$1 = "ℯ";
  var Escr$1 = "ℰ";
  var esdot$1 = "≐";
  var Esim$1 = "⩳";
  var esim$1 = "≂";
  var Eta$1 = "Η";
  var eta$1 = "η";
  var ETH$1 = "Ð";
  var eth$1 = "ð";
  var Euml$1 = "Ë";
  var euml$1 = "ë";
  var euro$3 = "€";
  var excl$1 = "!";
  var exist$1 = "∃";
  var Exists$1 = "∃";
  var expectation$1 = "ℰ";
  var exponentiale$1 = "ⅇ";
  var ExponentialE$1 = "ⅇ";
  var fallingdotseq$1 = "≒";
  var Fcy$1 = "Ф";
  var fcy$1 = "ф";
  var female$1 = "♀";
  var ffilig$1 = "ﬃ";
  var fflig$1 = "ﬀ";
  var ffllig$1 = "ﬄ";
  var Ffr$1 = "𝔉";
  var ffr$1 = "𝔣";
  var filig$1 = "ﬁ";
  var FilledSmallSquare$1 = "◼";
  var FilledVerySmallSquare$1 = "▪";
  var fjlig$1 = "fj";
  var flat$1 = "♭";
  var fllig$1 = "ﬂ";
  var fltns$1 = "▱";
  var fnof$1 = "ƒ";
  var Fopf$1 = "𝔽";
  var fopf$1 = "𝕗";
  var forall$1 = "∀";
  var ForAll$1 = "∀";
  var fork$1 = "⋔";
  var forkv$1 = "⫙";
  var Fouriertrf$1 = "ℱ";
  var fpartint$1 = "⨍";
  var frac12$1 = "½";
  var frac13$1 = "⅓";
  var frac14$1 = "¼";
  var frac15$1 = "⅕";
  var frac16$1 = "⅙";
  var frac18$1 = "⅛";
  var frac23$1 = "⅔";
  var frac25$1 = "⅖";
  var frac34$1 = "¾";
  var frac35$1 = "⅗";
  var frac38$1 = "⅜";
  var frac45$1 = "⅘";
  var frac56$1 = "⅚";
  var frac58$1 = "⅝";
  var frac78$1 = "⅞";
  var frasl$1 = "⁄";
  var frown$1 = "⌢";
  var fscr$1 = "𝒻";
  var Fscr$1 = "ℱ";
  var gacute$1 = "ǵ";
  var Gamma$1 = "Γ";
  var gamma$1 = "γ";
  var Gammad$1 = "Ϝ";
  var gammad$1 = "ϝ";
  var gap$1 = "⪆";
  var Gbreve$1 = "Ğ";
  var gbreve$1 = "ğ";
  var Gcedil$1 = "Ģ";
  var Gcirc$1 = "Ĝ";
  var gcirc$1 = "ĝ";
  var Gcy$1 = "Г";
  var gcy$1 = "г";
  var Gdot$1 = "Ġ";
  var gdot$1 = "ġ";
  var ge$1 = "≥";
  var gE$1 = "≧";
  var gEl$1 = "⪌";
  var gel$1 = "⋛";
  var geq$1 = "≥";
  var geqq$1 = "≧";
  var geqslant$1 = "⩾";
  var gescc$1 = "⪩";
  var ges$1 = "⩾";
  var gesdot$1 = "⪀";
  var gesdoto$1 = "⪂";
  var gesdotol$1 = "⪄";
  var gesl$1 = "⋛︀";
  var gesles$1 = "⪔";
  var Gfr$1 = "𝔊";
  var gfr$1 = "𝔤";
  var gg$1 = "≫";
  var Gg$1 = "⋙";
  var ggg$1 = "⋙";
  var gimel$1 = "ℷ";
  var GJcy$1 = "Ѓ";
  var gjcy$1 = "ѓ";
  var gla$1 = "⪥";
  var gl$1 = "≷";
  var glE$1 = "⪒";
  var glj$1 = "⪤";
  var gnap$1 = "⪊";
  var gnapprox$1 = "⪊";
  var gne$1 = "⪈";
  var gnE$1 = "≩";
  var gneq$1 = "⪈";
  var gneqq$1 = "≩";
  var gnsim$1 = "⋧";
  var Gopf$1 = "𝔾";
  var gopf$1 = "𝕘";
  var grave$1 = "`";
  var GreaterEqual$1 = "≥";
  var GreaterEqualLess$1 = "⋛";
  var GreaterFullEqual$1 = "≧";
  var GreaterGreater$1 = "⪢";
  var GreaterLess$1 = "≷";
  var GreaterSlantEqual$1 = "⩾";
  var GreaterTilde$1 = "≳";
  var Gscr$1 = "𝒢";
  var gscr$1 = "ℊ";
  var gsim$1 = "≳";
  var gsime$1 = "⪎";
  var gsiml$1 = "⪐";
  var gtcc$1 = "⪧";
  var gtcir$1 = "⩺";
  var gt$1 = ">";
  var GT$1 = ">";
  var Gt$1 = "≫";
  var gtdot$1 = "⋗";
  var gtlPar$1 = "⦕";
  var gtquest$1 = "⩼";
  var gtrapprox$1 = "⪆";
  var gtrarr$1 = "⥸";
  var gtrdot$1 = "⋗";
  var gtreqless$1 = "⋛";
  var gtreqqless$1 = "⪌";
  var gtrless$1 = "≷";
  var gtrsim$1 = "≳";
  var gvertneqq$1 = "≩︀";
  var gvnE$1 = "≩︀";
  var Hacek$1 = "ˇ";
  var hairsp$1 = " ";
  var half$1 = "½";
  var hamilt$1 = "ℋ";
  var HARDcy$1 = "Ъ";
  var hardcy$1 = "ъ";
  var harrcir$1 = "⥈";
  var harr$1 = "↔";
  var hArr$1 = "⇔";
  var harrw$1 = "↭";
  var Hat$1 = "^";
  var hbar$1 = "ℏ";
  var Hcirc$1 = "Ĥ";
  var hcirc$1 = "ĥ";
  var hearts$3 = "♥";
  var heartsuit$1 = "♥";
  var hellip$1 = "…";
  var hercon$1 = "⊹";
  var hfr$1 = "𝔥";
  var Hfr$1 = "ℌ";
  var HilbertSpace$1 = "ℋ";
  var hksearow$1 = "⤥";
  var hkswarow$1 = "⤦";
  var hoarr$1 = "⇿";
  var homtht$1 = "∻";
  var hookleftarrow$1 = "↩";
  var hookrightarrow$1 = "↪";
  var hopf$1 = "𝕙";
  var Hopf$1 = "ℍ";
  var horbar$1 = "―";
  var HorizontalLine$1 = "─";
  var hscr$1 = "𝒽";
  var Hscr$1 = "ℋ";
  var hslash$1 = "ℏ";
  var Hstrok$1 = "Ħ";
  var hstrok$1 = "ħ";
  var HumpDownHump$1 = "≎";
  var HumpEqual$1 = "≏";
  var hybull$1 = "⁃";
  var hyphen$1 = "‐";
  var Iacute$1 = "Í";
  var iacute$1 = "í";
  var ic$1 = "⁣";
  var Icirc$1 = "Î";
  var icirc$1 = "î";
  var Icy$1 = "И";
  var icy$1 = "и";
  var Idot$1 = "İ";
  var IEcy$1 = "Е";
  var iecy$1 = "е";
  var iexcl$1 = "¡";
  var iff$1 = "⇔";
  var ifr$1 = "𝔦";
  var Ifr$1 = "ℑ";
  var Igrave$1 = "Ì";
  var igrave$1 = "ì";
  var ii$1 = "ⅈ";
  var iiiint$1 = "⨌";
  var iiint$1 = "∭";
  var iinfin$1 = "⧜";
  var iiota$1 = "℩";
  var IJlig$1 = "Ĳ";
  var ijlig$1 = "ĳ";
  var Imacr$1 = "Ī";
  var imacr$1 = "ī";
  var image$4 = "ℑ";
  var ImaginaryI$1 = "ⅈ";
  var imagline$1 = "ℐ";
  var imagpart$1 = "ℑ";
  var imath$1 = "ı";
  var Im$1 = "ℑ";
  var imof$1 = "⊷";
  var imped$1 = "Ƶ";
  var Implies$1 = "⇒";
  var incare$1 = "℅";
  var infin$1 = "∞";
  var infintie$1 = "⧝";
  var inodot$1 = "ı";
  var intcal$1 = "⊺";
  var int$1 = "∫";
  var Int$1 = "∬";
  var integers$1 = "ℤ";
  var Integral$1 = "∫";
  var intercal$1 = "⊺";
  var Intersection$1 = "⋂";
  var intlarhk$1 = "⨗";
  var intprod$1 = "⨼";
  var InvisibleComma$1 = "⁣";
  var InvisibleTimes$1 = "⁢";
  var IOcy$1 = "Ё";
  var iocy$1 = "ё";
  var Iogon$1 = "Į";
  var iogon$1 = "į";
  var Iopf$1 = "𝕀";
  var iopf$1 = "𝕚";
  var Iota$1 = "Ι";
  var iota$1 = "ι";
  var iprod$1 = "⨼";
  var iquest$1 = "¿";
  var iscr$1 = "𝒾";
  var Iscr$1 = "ℐ";
  var isin$1 = "∈";
  var isindot$1 = "⋵";
  var isinE$1 = "⋹";
  var isins$1 = "⋴";
  var isinsv$1 = "⋳";
  var isinv$1 = "∈";
  var it$3 = "⁢";
  var Itilde$1 = "Ĩ";
  var itilde$1 = "ĩ";
  var Iukcy$1 = "І";
  var iukcy$1 = "і";
  var Iuml$1 = "Ï";
  var iuml$1 = "ï";
  var Jcirc$1 = "Ĵ";
  var jcirc$1 = "ĵ";
  var Jcy$1 = "Й";
  var jcy$1 = "й";
  var Jfr$1 = "𝔍";
  var jfr$1 = "𝔧";
  var jmath$1 = "ȷ";
  var Jopf$1 = "𝕁";
  var jopf$1 = "𝕛";
  var Jscr$1 = "𝒥";
  var jscr$1 = "𝒿";
  var Jsercy$1 = "Ј";
  var jsercy$1 = "ј";
  var Jukcy$1 = "Є";
  var jukcy$1 = "є";
  var Kappa$1 = "Κ";
  var kappa$1 = "κ";
  var kappav$1 = "ϰ";
  var Kcedil$1 = "Ķ";
  var kcedil$1 = "ķ";
  var Kcy$1 = "К";
  var kcy$1 = "к";
  var Kfr$1 = "𝔎";
  var kfr$1 = "𝔨";
  var kgreen$1 = "ĸ";
  var KHcy$1 = "Х";
  var khcy$1 = "х";
  var KJcy$1 = "Ќ";
  var kjcy$1 = "ќ";
  var Kopf$1 = "𝕂";
  var kopf$1 = "𝕜";
  var Kscr$1 = "𝒦";
  var kscr$1 = "𝓀";
  var lAarr$1 = "⇚";
  var Lacute$1 = "Ĺ";
  var lacute$1 = "ĺ";
  var laemptyv$1 = "⦴";
  var lagran$1 = "ℒ";
  var Lambda$1 = "Λ";
  var lambda$1 = "λ";
  var lang$1 = "⟨";
  var Lang$1 = "⟪";
  var langd$1 = "⦑";
  var langle$1 = "⟨";
  var lap$1 = "⪅";
  var Laplacetrf$1 = "ℒ";
  var laquo$1 = "«";
  var larrb$1 = "⇤";
  var larrbfs$1 = "⤟";
  var larr$1 = "←";
  var Larr$1 = "↞";
  var lArr$1 = "⇐";
  var larrfs$1 = "⤝";
  var larrhk$1 = "↩";
  var larrlp$1 = "↫";
  var larrpl$1 = "⤹";
  var larrsim$1 = "⥳";
  var larrtl$1 = "↢";
  var latail$1 = "⤙";
  var lAtail$1 = "⤛";
  var lat$1 = "⪫";
  var late$1 = "⪭";
  var lates$1 = "⪭︀";
  var lbarr$1 = "⤌";
  var lBarr$1 = "⤎";
  var lbbrk$1 = "❲";
  var lbrace$1 = "{";
  var lbrack$1 = "[";
  var lbrke$1 = "⦋";
  var lbrksld$1 = "⦏";
  var lbrkslu$1 = "⦍";
  var Lcaron$1 = "Ľ";
  var lcaron$1 = "ľ";
  var Lcedil$1 = "Ļ";
  var lcedil$1 = "ļ";
  var lceil$1 = "⌈";
  var lcub$1 = "{";
  var Lcy$1 = "Л";
  var lcy$1 = "л";
  var ldca$1 = "⤶";
  var ldquo$1 = "“";
  var ldquor$1 = "„";
  var ldrdhar$1 = "⥧";
  var ldrushar$1 = "⥋";
  var ldsh$1 = "↲";
  var le$1 = "≤";
  var lE$1 = "≦";
  var LeftAngleBracket$1 = "⟨";
  var LeftArrowBar$1 = "⇤";
  var leftarrow$1 = "←";
  var LeftArrow$1 = "←";
  var Leftarrow$1 = "⇐";
  var LeftArrowRightArrow$1 = "⇆";
  var leftarrowtail$1 = "↢";
  var LeftCeiling$1 = "⌈";
  var LeftDoubleBracket$1 = "⟦";
  var LeftDownTeeVector$1 = "⥡";
  var LeftDownVectorBar$1 = "⥙";
  var LeftDownVector$1 = "⇃";
  var LeftFloor$1 = "⌊";
  var leftharpoondown$1 = "↽";
  var leftharpoonup$1 = "↼";
  var leftleftarrows$1 = "⇇";
  var leftrightarrow$1 = "↔";
  var LeftRightArrow$1 = "↔";
  var Leftrightarrow$1 = "⇔";
  var leftrightarrows$1 = "⇆";
  var leftrightharpoons$1 = "⇋";
  var leftrightsquigarrow$1 = "↭";
  var LeftRightVector$1 = "⥎";
  var LeftTeeArrow$1 = "↤";
  var LeftTee$1 = "⊣";
  var LeftTeeVector$1 = "⥚";
  var leftthreetimes$1 = "⋋";
  var LeftTriangleBar$1 = "⧏";
  var LeftTriangle$1 = "⊲";
  var LeftTriangleEqual$1 = "⊴";
  var LeftUpDownVector$1 = "⥑";
  var LeftUpTeeVector$1 = "⥠";
  var LeftUpVectorBar$1 = "⥘";
  var LeftUpVector$1 = "↿";
  var LeftVectorBar$1 = "⥒";
  var LeftVector$1 = "↼";
  var lEg$1 = "⪋";
  var leg$3 = "⋚";
  var leq$1 = "≤";
  var leqq$1 = "≦";
  var leqslant$1 = "⩽";
  var lescc$1 = "⪨";
  var les$1 = "⩽";
  var lesdot$1 = "⩿";
  var lesdoto$1 = "⪁";
  var lesdotor$1 = "⪃";
  var lesg$1 = "⋚︀";
  var lesges$1 = "⪓";
  var lessapprox$1 = "⪅";
  var lessdot$1 = "⋖";
  var lesseqgtr$1 = "⋚";
  var lesseqqgtr$1 = "⪋";
  var LessEqualGreater$1 = "⋚";
  var LessFullEqual$1 = "≦";
  var LessGreater$1 = "≶";
  var lessgtr$1 = "≶";
  var LessLess$1 = "⪡";
  var lesssim$1 = "≲";
  var LessSlantEqual$1 = "⩽";
  var LessTilde$1 = "≲";
  var lfisht$1 = "⥼";
  var lfloor$1 = "⌊";
  var Lfr$1 = "𝔏";
  var lfr$1 = "𝔩";
  var lg$1 = "≶";
  var lgE$1 = "⪑";
  var lHar$1 = "⥢";
  var lhard$1 = "↽";
  var lharu$1 = "↼";
  var lharul$1 = "⥪";
  var lhblk$1 = "▄";
  var LJcy$1 = "Љ";
  var ljcy$1 = "љ";
  var llarr$1 = "⇇";
  var ll$1 = "≪";
  var Ll$1 = "⋘";
  var llcorner$1 = "⌞";
  var Lleftarrow$1 = "⇚";
  var llhard$1 = "⥫";
  var lltri$1 = "◺";
  var Lmidot$1 = "Ŀ";
  var lmidot$1 = "ŀ";
  var lmoustache$1 = "⎰";
  var lmoust$1 = "⎰";
  var lnap$1 = "⪉";
  var lnapprox$1 = "⪉";
  var lne$1 = "⪇";
  var lnE$1 = "≨";
  var lneq$1 = "⪇";
  var lneqq$1 = "≨";
  var lnsim$1 = "⋦";
  var loang$1 = "⟬";
  var loarr$1 = "⇽";
  var lobrk$1 = "⟦";
  var longleftarrow$1 = "⟵";
  var LongLeftArrow$1 = "⟵";
  var Longleftarrow$1 = "⟸";
  var longleftrightarrow$1 = "⟷";
  var LongLeftRightArrow$1 = "⟷";
  var Longleftrightarrow$1 = "⟺";
  var longmapsto$1 = "⟼";
  var longrightarrow$1 = "⟶";
  var LongRightArrow$1 = "⟶";
  var Longrightarrow$1 = "⟹";
  var looparrowleft$1 = "↫";
  var looparrowright$1 = "↬";
  var lopar$1 = "⦅";
  var Lopf$1 = "𝕃";
  var lopf$1 = "𝕝";
  var loplus$1 = "⨭";
  var lotimes$1 = "⨴";
  var lowast$1 = "∗";
  var lowbar$1 = "_";
  var LowerLeftArrow$1 = "↙";
  var LowerRightArrow$1 = "↘";
  var loz$1 = "◊";
  var lozenge$1 = "◊";
  var lozf$1 = "⧫";
  var lpar$1 = "(";
  var lparlt$1 = "⦓";
  var lrarr$1 = "⇆";
  var lrcorner$1 = "⌟";
  var lrhar$1 = "⇋";
  var lrhard$1 = "⥭";
  var lrm$1 = "‎";
  var lrtri$1 = "⊿";
  var lsaquo$1 = "‹";
  var lscr$1 = "𝓁";
  var Lscr$1 = "ℒ";
  var lsh$1 = "↰";
  var Lsh$1 = "↰";
  var lsim$1 = "≲";
  var lsime$1 = "⪍";
  var lsimg$1 = "⪏";
  var lsqb$1 = "[";
  var lsquo$1 = "‘";
  var lsquor$1 = "‚";
  var Lstrok$1 = "Ł";
  var lstrok$1 = "ł";
  var ltcc$1 = "⪦";
  var ltcir$1 = "⩹";
  var lt$1 = "<";
  var LT$1 = "<";
  var Lt$1 = "≪";
  var ltdot$1 = "⋖";
  var lthree$1 = "⋋";
  var ltimes$1 = "⋉";
  var ltlarr$1 = "⥶";
  var ltquest$1 = "⩻";
  var ltri$1 = "◃";
  var ltrie$1 = "⊴";
  var ltrif$1 = "◂";
  var ltrPar$1 = "⦖";
  var lurdshar$1 = "⥊";
  var luruhar$1 = "⥦";
  var lvertneqq$1 = "≨︀";
  var lvnE$1 = "≨︀";
  var macr$1 = "¯";
  var male$1 = "♂";
  var malt$1 = "✠";
  var maltese$1 = "✠";
  var map$2 = "↦";
  var mapsto$1 = "↦";
  var mapstodown$1 = "↧";
  var mapstoleft$1 = "↤";
  var mapstoup$1 = "↥";
  var marker$1 = "▮";
  var mcomma$1 = "⨩";
  var Mcy$1 = "М";
  var mcy$1 = "м";
  var mdash$1 = "—";
  var mDDot$1 = "∺";
  var measuredangle$1 = "∡";
  var MediumSpace$1 = " ";
  var Mellintrf$1 = "ℳ";
  var Mfr$1 = "𝔐";
  var mfr$1 = "𝔪";
  var mho$1 = "℧";
  var micro$1 = "µ";
  var midast$1 = "*";
  var midcir$1 = "⫰";
  var mid$1 = "∣";
  var middot$1 = "·";
  var minusb$1 = "⊟";
  var minus$1 = "−";
  var minusd$1 = "∸";
  var minusdu$1 = "⨪";
  var MinusPlus$1 = "∓";
  var mlcp$1 = "⫛";
  var mldr$1 = "…";
  var mnplus$1 = "∓";
  var models$1 = "⊧";
  var Mopf$1 = "𝕄";
  var mopf$1 = "𝕞";
  var mp$1 = "∓";
  var mscr$1 = "𝓂";
  var Mscr$1 = "ℳ";
  var mstpos$1 = "∾";
  var Mu$1 = "Μ";
  var mu$1 = "μ";
  var multimap$1 = "⊸";
  var mumap$1 = "⊸";
  var nabla$1 = "∇";
  var Nacute$1 = "Ń";
  var nacute$1 = "ń";
  var nang$1 = "∠⃒";
  var nap$1 = "≉";
  var napE$1 = "⩰̸";
  var napid$1 = "≋̸";
  var napos$1 = "ŉ";
  var napprox$1 = "≉";
  var natural$1 = "♮";
  var naturals$1 = "ℕ";
  var natur$1 = "♮";
  var nbsp$1 = " ";
  var nbump$1 = "≎̸";
  var nbumpe$1 = "≏̸";
  var ncap$1 = "⩃";
  var Ncaron$1 = "Ň";
  var ncaron$1 = "ň";
  var Ncedil$1 = "Ņ";
  var ncedil$1 = "ņ";
  var ncong$1 = "≇";
  var ncongdot$1 = "⩭̸";
  var ncup$1 = "⩂";
  var Ncy$1 = "Н";
  var ncy$1 = "н";
  var ndash$1 = "–";
  var nearhk$1 = "⤤";
  var nearr$1 = "↗";
  var neArr$1 = "⇗";
  var nearrow$1 = "↗";
  var ne$1 = "≠";
  var nedot$1 = "≐̸";
  var NegativeMediumSpace$1 = "​";
  var NegativeThickSpace$1 = "​";
  var NegativeThinSpace$1 = "​";
  var NegativeVeryThinSpace$1 = "​";
  var nequiv$1 = "≢";
  var nesear$1 = "⤨";
  var nesim$1 = "≂̸";
  var NestedGreaterGreater$1 = "≫";
  var NestedLessLess$1 = "≪";
  var NewLine$1 = "\n";
  var nexist$1 = "∄";
  var nexists$1 = "∄";
  var Nfr$1 = "𝔑";
  var nfr$1 = "𝔫";
  var ngE$1 = "≧̸";
  var nge$1 = "≱";
  var ngeq$1 = "≱";
  var ngeqq$1 = "≧̸";
  var ngeqslant$1 = "⩾̸";
  var nges$1 = "⩾̸";
  var nGg$1 = "⋙̸";
  var ngsim$1 = "≵";
  var nGt$1 = "≫⃒";
  var ngt$1 = "≯";
  var ngtr$1 = "≯";
  var nGtv$1 = "≫̸";
  var nharr$1 = "↮";
  var nhArr$1 = "⇎";
  var nhpar$1 = "⫲";
  var ni$1 = "∋";
  var nis$1 = "⋼";
  var nisd$1 = "⋺";
  var niv$1 = "∋";
  var NJcy$1 = "Њ";
  var njcy$1 = "њ";
  var nlarr$1 = "↚";
  var nlArr$1 = "⇍";
  var nldr$1 = "‥";
  var nlE$1 = "≦̸";
  var nle$1 = "≰";
  var nleftarrow$1 = "↚";
  var nLeftarrow$1 = "⇍";
  var nleftrightarrow$1 = "↮";
  var nLeftrightarrow$1 = "⇎";
  var nleq$1 = "≰";
  var nleqq$1 = "≦̸";
  var nleqslant$1 = "⩽̸";
  var nles$1 = "⩽̸";
  var nless$1 = "≮";
  var nLl$1 = "⋘̸";
  var nlsim$1 = "≴";
  var nLt$1 = "≪⃒";
  var nlt$1 = "≮";
  var nltri$1 = "⋪";
  var nltrie$1 = "⋬";
  var nLtv$1 = "≪̸";
  var nmid$1 = "∤";
  var NoBreak$1 = "⁠";
  var NonBreakingSpace$1 = " ";
  var nopf$1 = "𝕟";
  var Nopf$1 = "ℕ";
  var Not$1 = "⫬";
  var not$1 = "¬";
  var NotCongruent$1 = "≢";
  var NotCupCap$1 = "≭";
  var NotDoubleVerticalBar$1 = "∦";
  var NotElement$1 = "∉";
  var NotEqual$1 = "≠";
  var NotEqualTilde$1 = "≂̸";
  var NotExists$1 = "∄";
  var NotGreater$1 = "≯";
  var NotGreaterEqual$1 = "≱";
  var NotGreaterFullEqual$1 = "≧̸";
  var NotGreaterGreater$1 = "≫̸";
  var NotGreaterLess$1 = "≹";
  var NotGreaterSlantEqual$1 = "⩾̸";
  var NotGreaterTilde$1 = "≵";
  var NotHumpDownHump$1 = "≎̸";
  var NotHumpEqual$1 = "≏̸";
  var notin$1 = "∉";
  var notindot$1 = "⋵̸";
  var notinE$1 = "⋹̸";
  var notinva$1 = "∉";
  var notinvb$1 = "⋷";
  var notinvc$1 = "⋶";
  var NotLeftTriangleBar$1 = "⧏̸";
  var NotLeftTriangle$1 = "⋪";
  var NotLeftTriangleEqual$1 = "⋬";
  var NotLess$1 = "≮";
  var NotLessEqual$1 = "≰";
  var NotLessGreater$1 = "≸";
  var NotLessLess$1 = "≪̸";
  var NotLessSlantEqual$1 = "⩽̸";
  var NotLessTilde$1 = "≴";
  var NotNestedGreaterGreater$1 = "⪢̸";
  var NotNestedLessLess$1 = "⪡̸";
  var notni$1 = "∌";
  var notniva$1 = "∌";
  var notnivb$1 = "⋾";
  var notnivc$1 = "⋽";
  var NotPrecedes$1 = "⊀";
  var NotPrecedesEqual$1 = "⪯̸";
  var NotPrecedesSlantEqual$1 = "⋠";
  var NotReverseElement$1 = "∌";
  var NotRightTriangleBar$1 = "⧐̸";
  var NotRightTriangle$1 = "⋫";
  var NotRightTriangleEqual$1 = "⋭";
  var NotSquareSubset$1 = "⊏̸";
  var NotSquareSubsetEqual$1 = "⋢";
  var NotSquareSuperset$1 = "⊐̸";
  var NotSquareSupersetEqual$1 = "⋣";
  var NotSubset$1 = "⊂⃒";
  var NotSubsetEqual$1 = "⊈";
  var NotSucceeds$1 = "⊁";
  var NotSucceedsEqual$1 = "⪰̸";
  var NotSucceedsSlantEqual$1 = "⋡";
  var NotSucceedsTilde$1 = "≿̸";
  var NotSuperset$1 = "⊃⃒";
  var NotSupersetEqual$1 = "⊉";
  var NotTilde$1 = "≁";
  var NotTildeEqual$1 = "≄";
  var NotTildeFullEqual$1 = "≇";
  var NotTildeTilde$1 = "≉";
  var NotVerticalBar$1 = "∤";
  var nparallel$1 = "∦";
  var npar$1 = "∦";
  var nparsl$1 = "⫽⃥";
  var npart$1 = "∂̸";
  var npolint$1 = "⨔";
  var npr$1 = "⊀";
  var nprcue$1 = "⋠";
  var nprec$1 = "⊀";
  var npreceq$1 = "⪯̸";
  var npre$1 = "⪯̸";
  var nrarrc$1 = "⤳̸";
  var nrarr$1 = "↛";
  var nrArr$1 = "⇏";
  var nrarrw$1 = "↝̸";
  var nrightarrow$1 = "↛";
  var nRightarrow$1 = "⇏";
  var nrtri$1 = "⋫";
  var nrtrie$1 = "⋭";
  var nsc$1 = "⊁";
  var nsccue$1 = "⋡";
  var nsce$1 = "⪰̸";
  var Nscr$1 = "𝒩";
  var nscr$1 = "𝓃";
  var nshortmid$1 = "∤";
  var nshortparallel$1 = "∦";
  var nsim$1 = "≁";
  var nsime$1 = "≄";
  var nsimeq$1 = "≄";
  var nsmid$1 = "∤";
  var nspar$1 = "∦";
  var nsqsube$1 = "⋢";
  var nsqsupe$1 = "⋣";
  var nsub$1 = "⊄";
  var nsubE$1 = "⫅̸";
  var nsube$1 = "⊈";
  var nsubset$1 = "⊂⃒";
  var nsubseteq$1 = "⊈";
  var nsubseteqq$1 = "⫅̸";
  var nsucc$1 = "⊁";
  var nsucceq$1 = "⪰̸";
  var nsup$1 = "⊅";
  var nsupE$1 = "⫆̸";
  var nsupe$1 = "⊉";
  var nsupset$1 = "⊃⃒";
  var nsupseteq$1 = "⊉";
  var nsupseteqq$1 = "⫆̸";
  var ntgl$1 = "≹";
  var Ntilde$1 = "Ñ";
  var ntilde$1 = "ñ";
  var ntlg$1 = "≸";
  var ntriangleleft$1 = "⋪";
  var ntrianglelefteq$1 = "⋬";
  var ntriangleright$1 = "⋫";
  var ntrianglerighteq$1 = "⋭";
  var Nu$1 = "Ν";
  var nu$1 = "ν";
  var num$1 = "#";
  var numero$1 = "№";
  var numsp$1 = " ";
  var nvap$1 = "≍⃒";
  var nvdash$1 = "⊬";
  var nvDash$1 = "⊭";
  var nVdash$1 = "⊮";
  var nVDash$1 = "⊯";
  var nvge$1 = "≥⃒";
  var nvgt$1 = ">⃒";
  var nvHarr$1 = "⤄";
  var nvinfin$1 = "⧞";
  var nvlArr$1 = "⤂";
  var nvle$1 = "≤⃒";
  var nvlt$1 = "<⃒";
  var nvltrie$1 = "⊴⃒";
  var nvrArr$1 = "⤃";
  var nvrtrie$1 = "⊵⃒";
  var nvsim$1 = "∼⃒";
  var nwarhk$1 = "⤣";
  var nwarr$1 = "↖";
  var nwArr$1 = "⇖";
  var nwarrow$1 = "↖";
  var nwnear$1 = "⤧";
  var Oacute$1 = "Ó";
  var oacute$1 = "ó";
  var oast$1 = "⊛";
  var Ocirc$1 = "Ô";
  var ocirc$1 = "ô";
  var ocir$1 = "⊚";
  var Ocy$1 = "О";
  var ocy$1 = "о";
  var odash$1 = "⊝";
  var Odblac$1 = "Ő";
  var odblac$1 = "ő";
  var odiv$1 = "⨸";
  var odot$1 = "⊙";
  var odsold$1 = "⦼";
  var OElig$1 = "Œ";
  var oelig$1 = "œ";
  var ofcir$1 = "⦿";
  var Ofr$1 = "𝔒";
  var ofr$1 = "𝔬";
  var ogon$1 = "˛";
  var Ograve$1 = "Ò";
  var ograve$1 = "ò";
  var ogt$1 = "⧁";
  var ohbar$1 = "⦵";
  var ohm$1 = "Ω";
  var oint$1 = "∮";
  var olarr$1 = "↺";
  var olcir$1 = "⦾";
  var olcross$1 = "⦻";
  var oline$1 = "‾";
  var olt$1 = "⧀";
  var Omacr$1 = "Ō";
  var omacr$1 = "ō";
  var Omega$1 = "Ω";
  var omega$1 = "ω";
  var Omicron$1 = "Ο";
  var omicron$1 = "ο";
  var omid$1 = "⦶";
  var ominus$1 = "⊖";
  var Oopf$1 = "𝕆";
  var oopf$1 = "𝕠";
  var opar$1 = "⦷";
  var OpenCurlyDoubleQuote$1 = "“";
  var OpenCurlyQuote$1 = "‘";
  var operp$1 = "⦹";
  var oplus$1 = "⊕";
  var orarr$1 = "↻";
  var Or$1 = "⩔";
  var or$1 = "∨";
  var ord$1 = "⩝";
  var order$1 = "ℴ";
  var orderof$1 = "ℴ";
  var ordf$1 = "ª";
  var ordm$1 = "º";
  var origof$1 = "⊶";
  var oror$1 = "⩖";
  var orslope$1 = "⩗";
  var orv$1 = "⩛";
  var oS$1 = "Ⓢ";
  var Oscr$1 = "𝒪";
  var oscr$1 = "ℴ";
  var Oslash$1 = "Ø";
  var oslash$1 = "ø";
  var osol$1 = "⊘";
  var Otilde$1 = "Õ";
  var otilde$1 = "õ";
  var otimesas$1 = "⨶";
  var Otimes$1 = "⨷";
  var otimes$1 = "⊗";
  var Ouml$1 = "Ö";
  var ouml$1 = "ö";
  var ovbar$1 = "⌽";
  var OverBar$1 = "‾";
  var OverBrace$1 = "⏞";
  var OverBracket$1 = "⎴";
  var OverParenthesis$1 = "⏜";
  var para$1 = "¶";
  var parallel$1 = "∥";
  var par$1 = "∥";
  var parsim$1 = "⫳";
  var parsl$1 = "⫽";
  var part$1 = "∂";
  var PartialD$1 = "∂";
  var Pcy$1 = "П";
  var pcy$1 = "п";
  var percnt$1 = "%";
  var period$1 = ".";
  var permil$1 = "‰";
  var perp$1 = "⊥";
  var pertenk$1 = "‱";
  var Pfr$1 = "𝔓";
  var pfr$1 = "𝔭";
  var Phi$1 = "Φ";
  var phi$1 = "φ";
  var phiv$1 = "ϕ";
  var phmmat$1 = "ℳ";
  var phone$3 = "☎";
  var Pi$1 = "Π";
  var pi$1 = "π";
  var pitchfork$1 = "⋔";
  var piv$1 = "ϖ";
  var planck$1 = "ℏ";
  var planckh$1 = "ℎ";
  var plankv$1 = "ℏ";
  var plusacir$1 = "⨣";
  var plusb$1 = "⊞";
  var pluscir$1 = "⨢";
  var plus$1 = "+";
  var plusdo$1 = "∔";
  var plusdu$1 = "⨥";
  var pluse$1 = "⩲";
  var PlusMinus$1 = "±";
  var plusmn$1 = "±";
  var plussim$1 = "⨦";
  var plustwo$1 = "⨧";
  var pm$1 = "±";
  var Poincareplane$1 = "ℌ";
  var pointint$1 = "⨕";
  var popf$1 = "𝕡";
  var Popf$1 = "ℙ";
  var pound$3 = "£";
  var prap$1 = "⪷";
  var Pr$1 = "⪻";
  var pr$1 = "≺";
  var prcue$1 = "≼";
  var precapprox$1 = "⪷";
  var prec$1 = "≺";
  var preccurlyeq$1 = "≼";
  var Precedes$1 = "≺";
  var PrecedesEqual$1 = "⪯";
  var PrecedesSlantEqual$1 = "≼";
  var PrecedesTilde$1 = "≾";
  var preceq$1 = "⪯";
  var precnapprox$1 = "⪹";
  var precneqq$1 = "⪵";
  var precnsim$1 = "⋨";
  var pre$1 = "⪯";
  var prE$1 = "⪳";
  var precsim$1 = "≾";
  var prime$1 = "′";
  var Prime$1 = "″";
  var primes$1 = "ℙ";
  var prnap$1 = "⪹";
  var prnE$1 = "⪵";
  var prnsim$1 = "⋨";
  var prod$1 = "∏";
  var Product$1 = "∏";
  var profalar$1 = "⌮";
  var profline$1 = "⌒";
  var profsurf$1 = "⌓";
  var prop$1 = "∝";
  var Proportional$1 = "∝";
  var Proportion$1 = "∷";
  var propto$1 = "∝";
  var prsim$1 = "≾";
  var prurel$1 = "⊰";
  var Pscr$1 = "𝒫";
  var pscr$1 = "𝓅";
  var Psi$1 = "Ψ";
  var psi$1 = "ψ";
  var puncsp$1 = " ";
  var Qfr$1 = "𝔔";
  var qfr$1 = "𝔮";
  var qint$1 = "⨌";
  var qopf$1 = "𝕢";
  var Qopf$1 = "ℚ";
  var qprime$1 = "⁗";
  var Qscr$1 = "𝒬";
  var qscr$1 = "𝓆";
  var quaternions$1 = "ℍ";
  var quatint$1 = "⨖";
  var quest$1 = "?";
  var questeq$1 = "≟";
  var quot$1 = "\"";
  var QUOT$1 = "\"";
  var rAarr$1 = "⇛";
  var race$1 = "∽̱";
  var Racute$1 = "Ŕ";
  var racute$1 = "ŕ";
  var radic$1 = "√";
  var raemptyv$1 = "⦳";
  var rang$1 = "⟩";
  var Rang$1 = "⟫";
  var rangd$1 = "⦒";
  var range$1 = "⦥";
  var rangle$1 = "⟩";
  var raquo$1 = "»";
  var rarrap$1 = "⥵";
  var rarrb$1 = "⇥";
  var rarrbfs$1 = "⤠";
  var rarrc$1 = "⤳";
  var rarr$1 = "→";
  var Rarr$1 = "↠";
  var rArr$1 = "⇒";
  var rarrfs$1 = "⤞";
  var rarrhk$1 = "↪";
  var rarrlp$1 = "↬";
  var rarrpl$1 = "⥅";
  var rarrsim$1 = "⥴";
  var Rarrtl$1 = "⤖";
  var rarrtl$1 = "↣";
  var rarrw$1 = "↝";
  var ratail$1 = "⤚";
  var rAtail$1 = "⤜";
  var ratio$1 = "∶";
  var rationals$1 = "ℚ";
  var rbarr$1 = "⤍";
  var rBarr$1 = "⤏";
  var RBarr$1 = "⤐";
  var rbbrk$1 = "❳";
  var rbrace$1 = "}";
  var rbrack$1 = "]";
  var rbrke$1 = "⦌";
  var rbrksld$1 = "⦎";
  var rbrkslu$1 = "⦐";
  var Rcaron$1 = "Ř";
  var rcaron$1 = "ř";
  var Rcedil$1 = "Ŗ";
  var rcedil$1 = "ŗ";
  var rceil$1 = "⌉";
  var rcub$1 = "}";
  var Rcy$1 = "Р";
  var rcy$1 = "р";
  var rdca$1 = "⤷";
  var rdldhar$1 = "⥩";
  var rdquo$1 = "”";
  var rdquor$1 = "”";
  var rdsh$1 = "↳";
  var real$1 = "ℜ";
  var realine$1 = "ℛ";
  var realpart$1 = "ℜ";
  var reals$1 = "ℝ";
  var Re$1 = "ℜ";
  var rect$1 = "▭";
  var reg$1 = "®";
  var REG$1 = "®";
  var ReverseElement$1 = "∋";
  var ReverseEquilibrium$1 = "⇋";
  var ReverseUpEquilibrium$1 = "⥯";
  var rfisht$1 = "⥽";
  var rfloor$1 = "⌋";
  var rfr$1 = "𝔯";
  var Rfr$1 = "ℜ";
  var rHar$1 = "⥤";
  var rhard$1 = "⇁";
  var rharu$1 = "⇀";
  var rharul$1 = "⥬";
  var Rho$1 = "Ρ";
  var rho$1 = "ρ";
  var rhov$1 = "ϱ";
  var RightAngleBracket$1 = "⟩";
  var RightArrowBar$1 = "⇥";
  var rightarrow$1 = "→";
  var RightArrow$1 = "→";
  var Rightarrow$1 = "⇒";
  var RightArrowLeftArrow$1 = "⇄";
  var rightarrowtail$1 = "↣";
  var RightCeiling$1 = "⌉";
  var RightDoubleBracket$1 = "⟧";
  var RightDownTeeVector$1 = "⥝";
  var RightDownVectorBar$1 = "⥕";
  var RightDownVector$1 = "⇂";
  var RightFloor$1 = "⌋";
  var rightharpoondown$1 = "⇁";
  var rightharpoonup$1 = "⇀";
  var rightleftarrows$1 = "⇄";
  var rightleftharpoons$1 = "⇌";
  var rightrightarrows$1 = "⇉";
  var rightsquigarrow$1 = "↝";
  var RightTeeArrow$1 = "↦";
  var RightTee$1 = "⊢";
  var RightTeeVector$1 = "⥛";
  var rightthreetimes$1 = "⋌";
  var RightTriangleBar$1 = "⧐";
  var RightTriangle$1 = "⊳";
  var RightTriangleEqual$1 = "⊵";
  var RightUpDownVector$1 = "⥏";
  var RightUpTeeVector$1 = "⥜";
  var RightUpVectorBar$1 = "⥔";
  var RightUpVector$1 = "↾";
  var RightVectorBar$1 = "⥓";
  var RightVector$1 = "⇀";
  var ring$3 = "˚";
  var risingdotseq$1 = "≓";
  var rlarr$1 = "⇄";
  var rlhar$1 = "⇌";
  var rlm$1 = "‏";
  var rmoustache$1 = "⎱";
  var rmoust$1 = "⎱";
  var rnmid$1 = "⫮";
  var roang$1 = "⟭";
  var roarr$1 = "⇾";
  var robrk$1 = "⟧";
  var ropar$1 = "⦆";
  var ropf$1 = "𝕣";
  var Ropf$1 = "ℝ";
  var roplus$1 = "⨮";
  var rotimes$1 = "⨵";
  var RoundImplies$1 = "⥰";
  var rpar$1 = ")";
  var rpargt$1 = "⦔";
  var rppolint$1 = "⨒";
  var rrarr$1 = "⇉";
  var Rrightarrow$1 = "⇛";
  var rsaquo$1 = "›";
  var rscr$1 = "𝓇";
  var Rscr$1 = "ℛ";
  var rsh$1 = "↱";
  var Rsh$1 = "↱";
  var rsqb$1 = "]";
  var rsquo$1 = "’";
  var rsquor$1 = "’";
  var rthree$1 = "⋌";
  var rtimes$1 = "⋊";
  var rtri$1 = "▹";
  var rtrie$1 = "⊵";
  var rtrif$1 = "▸";
  var rtriltri$1 = "⧎";
  var RuleDelayed$1 = "⧴";
  var ruluhar$1 = "⥨";
  var rx$1 = "℞";
  var Sacute$1 = "Ś";
  var sacute$1 = "ś";
  var sbquo$1 = "‚";
  var scap$1 = "⪸";
  var Scaron$1 = "Š";
  var scaron$1 = "š";
  var Sc$1 = "⪼";
  var sc$1 = "≻";
  var sccue$1 = "≽";
  var sce$1 = "⪰";
  var scE$1 = "⪴";
  var Scedil$1 = "Ş";
  var scedil$1 = "ş";
  var Scirc$1 = "Ŝ";
  var scirc$1 = "ŝ";
  var scnap$1 = "⪺";
  var scnE$1 = "⪶";
  var scnsim$1 = "⋩";
  var scpolint$1 = "⨓";
  var scsim$1 = "≿";
  var Scy$1 = "С";
  var scy$1 = "с";
  var sdotb$1 = "⊡";
  var sdot$1 = "⋅";
  var sdote$1 = "⩦";
  var searhk$1 = "⤥";
  var searr$1 = "↘";
  var seArr$1 = "⇘";
  var searrow$1 = "↘";
  var sect$1 = "§";
  var semi$1 = ";";
  var seswar$1 = "⤩";
  var setminus$1 = "∖";
  var setmn$1 = "∖";
  var sext$1 = "✶";
  var Sfr$1 = "𝔖";
  var sfr$1 = "𝔰";
  var sfrown$1 = "⌢";
  var sharp$1 = "♯";
  var SHCHcy$1 = "Щ";
  var shchcy$1 = "щ";
  var SHcy$1 = "Ш";
  var shcy$1 = "ш";
  var ShortDownArrow$1 = "↓";
  var ShortLeftArrow$1 = "←";
  var shortmid$1 = "∣";
  var shortparallel$1 = "∥";
  var ShortRightArrow$1 = "→";
  var ShortUpArrow$1 = "↑";
  var shy$1 = "­";
  var Sigma$1 = "Σ";
  var sigma$1 = "σ";
  var sigmaf$1 = "ς";
  var sigmav$1 = "ς";
  var sim$1 = "∼";
  var simdot$1 = "⩪";
  var sime$1 = "≃";
  var simeq$1 = "≃";
  var simg$1 = "⪞";
  var simgE$1 = "⪠";
  var siml$1 = "⪝";
  var simlE$1 = "⪟";
  var simne$1 = "≆";
  var simplus$1 = "⨤";
  var simrarr$1 = "⥲";
  var slarr$1 = "←";
  var SmallCircle$1 = "∘";
  var smallsetminus$1 = "∖";
  var smashp$1 = "⨳";
  var smeparsl$1 = "⧤";
  var smid$1 = "∣";
  var smile$3 = "⌣";
  var smt$1 = "⪪";
  var smte$1 = "⪬";
  var smtes$1 = "⪬︀";
  var SOFTcy$1 = "Ь";
  var softcy$1 = "ь";
  var solbar$1 = "⌿";
  var solb$1 = "⧄";
  var sol$1 = "/";
  var Sopf$1 = "𝕊";
  var sopf$1 = "𝕤";
  var spades$3 = "♠";
  var spadesuit$1 = "♠";
  var spar$1 = "∥";
  var sqcap$1 = "⊓";
  var sqcaps$1 = "⊓︀";
  var sqcup$1 = "⊔";
  var sqcups$1 = "⊔︀";
  var Sqrt$1 = "√";
  var sqsub$1 = "⊏";
  var sqsube$1 = "⊑";
  var sqsubset$1 = "⊏";
  var sqsubseteq$1 = "⊑";
  var sqsup$1 = "⊐";
  var sqsupe$1 = "⊒";
  var sqsupset$1 = "⊐";
  var sqsupseteq$1 = "⊒";
  var square$1 = "□";
  var Square$1 = "□";
  var SquareIntersection$1 = "⊓";
  var SquareSubset$1 = "⊏";
  var SquareSubsetEqual$1 = "⊑";
  var SquareSuperset$1 = "⊐";
  var SquareSupersetEqual$1 = "⊒";
  var SquareUnion$1 = "⊔";
  var squarf$1 = "▪";
  var squ$1 = "□";
  var squf$1 = "▪";
  var srarr$1 = "→";
  var Sscr$1 = "𝒮";
  var sscr$1 = "𝓈";
  var ssetmn$1 = "∖";
  var ssmile$1 = "⌣";
  var sstarf$1 = "⋆";
  var Star$1 = "⋆";
  var star$3 = "☆";
  var starf$1 = "★";
  var straightepsilon$1 = "ϵ";
  var straightphi$1 = "ϕ";
  var strns$1 = "¯";
  var sub$1 = "⊂";
  var Sub$1 = "⋐";
  var subdot$1 = "⪽";
  var subE$1 = "⫅";
  var sube$1 = "⊆";
  var subedot$1 = "⫃";
  var submult$1 = "⫁";
  var subnE$1 = "⫋";
  var subne$1 = "⊊";
  var subplus$1 = "⪿";
  var subrarr$1 = "⥹";
  var subset$1 = "⊂";
  var Subset$1 = "⋐";
  var subseteq$1 = "⊆";
  var subseteqq$1 = "⫅";
  var SubsetEqual$1 = "⊆";
  var subsetneq$1 = "⊊";
  var subsetneqq$1 = "⫋";
  var subsim$1 = "⫇";
  var subsub$1 = "⫕";
  var subsup$1 = "⫓";
  var succapprox$1 = "⪸";
  var succ$1 = "≻";
  var succcurlyeq$1 = "≽";
  var Succeeds$1 = "≻";
  var SucceedsEqual$1 = "⪰";
  var SucceedsSlantEqual$1 = "≽";
  var SucceedsTilde$1 = "≿";
  var succeq$1 = "⪰";
  var succnapprox$1 = "⪺";
  var succneqq$1 = "⪶";
  var succnsim$1 = "⋩";
  var succsim$1 = "≿";
  var SuchThat$1 = "∋";
  var sum$1 = "∑";
  var Sum$1 = "∑";
  var sung$1 = "♪";
  var sup1$1 = "¹";
  var sup2$1 = "²";
  var sup3$1 = "³";
  var sup$1 = "⊃";
  var Sup$1 = "⋑";
  var supdot$1 = "⪾";
  var supdsub$1 = "⫘";
  var supE$1 = "⫆";
  var supe$1 = "⊇";
  var supedot$1 = "⫄";
  var Superset$1 = "⊃";
  var SupersetEqual$1 = "⊇";
  var suphsol$1 = "⟉";
  var suphsub$1 = "⫗";
  var suplarr$1 = "⥻";
  var supmult$1 = "⫂";
  var supnE$1 = "⫌";
  var supne$1 = "⊋";
  var supplus$1 = "⫀";
  var supset$1 = "⊃";
  var Supset$1 = "⋑";
  var supseteq$1 = "⊇";
  var supseteqq$1 = "⫆";
  var supsetneq$1 = "⊋";
  var supsetneqq$1 = "⫌";
  var supsim$1 = "⫈";
  var supsub$1 = "⫔";
  var supsup$1 = "⫖";
  var swarhk$1 = "⤦";
  var swarr$1 = "↙";
  var swArr$1 = "⇙";
  var swarrow$1 = "↙";
  var swnwar$1 = "⤪";
  var szlig$1 = "ß";
  var Tab$1 = "\t";
  var target$1 = "⌖";
  var Tau$1 = "Τ";
  var tau$1 = "τ";
  var tbrk$1 = "⎴";
  var Tcaron$1 = "Ť";
  var tcaron$1 = "ť";
  var Tcedil$1 = "Ţ";
  var tcedil$1 = "ţ";
  var Tcy$1 = "Т";
  var tcy$1 = "т";
  var tdot$1 = "⃛";
  var telrec$1 = "⌕";
  var Tfr$1 = "𝔗";
  var tfr$1 = "𝔱";
  var there4$1 = "∴";
  var therefore$1 = "∴";
  var Therefore$1 = "∴";
  var Theta$1 = "Θ";
  var theta$1 = "θ";
  var thetasym$1 = "ϑ";
  var thetav$1 = "ϑ";
  var thickapprox$1 = "≈";
  var thicksim$1 = "∼";
  var ThickSpace$1 = "  ";
  var ThinSpace$1 = " ";
  var thinsp$1 = " ";
  var thkap$1 = "≈";
  var thksim$1 = "∼";
  var THORN$1 = "Þ";
  var thorn$1 = "þ";
  var tilde$1 = "˜";
  var Tilde$1 = "∼";
  var TildeEqual$1 = "≃";
  var TildeFullEqual$1 = "≅";
  var TildeTilde$1 = "≈";
  var timesbar$1 = "⨱";
  var timesb$1 = "⊠";
  var times$1 = "×";
  var timesd$1 = "⨰";
  var tint$1 = "∭";
  var toea$1 = "⤨";
  var topbot$1 = "⌶";
  var topcir$1 = "⫱";
  var top$3 = "⊤";
  var Topf$1 = "𝕋";
  var topf$1 = "𝕥";
  var topfork$1 = "⫚";
  var tosa$1 = "⤩";
  var tprime$1 = "‴";
  var trade$1 = "™";
  var TRADE$1 = "™";
  var triangle$1 = "▵";
  var triangledown$1 = "▿";
  var triangleleft$1 = "◃";
  var trianglelefteq$1 = "⊴";
  var triangleq$1 = "≜";
  var triangleright$1 = "▹";
  var trianglerighteq$1 = "⊵";
  var tridot$1 = "◬";
  var trie$1 = "≜";
  var triminus$1 = "⨺";
  var TripleDot$1 = "⃛";
  var triplus$1 = "⨹";
  var trisb$1 = "⧍";
  var tritime$1 = "⨻";
  var trpezium$1 = "⏢";
  var Tscr$1 = "𝒯";
  var tscr$1 = "𝓉";
  var TScy$1 = "Ц";
  var tscy$1 = "ц";
  var TSHcy$1 = "Ћ";
  var tshcy$1 = "ћ";
  var Tstrok$1 = "Ŧ";
  var tstrok$1 = "ŧ";
  var twixt$1 = "≬";
  var twoheadleftarrow$1 = "↞";
  var twoheadrightarrow$1 = "↠";
  var Uacute$1 = "Ú";
  var uacute$1 = "ú";
  var uarr$1 = "↑";
  var Uarr$1 = "↟";
  var uArr$1 = "⇑";
  var Uarrocir$1 = "⥉";
  var Ubrcy$1 = "Ў";
  var ubrcy$1 = "ў";
  var Ubreve$1 = "Ŭ";
  var ubreve$1 = "ŭ";
  var Ucirc$1 = "Û";
  var ucirc$1 = "û";
  var Ucy$1 = "У";
  var ucy$1 = "у";
  var udarr$1 = "⇅";
  var Udblac$1 = "Ű";
  var udblac$1 = "ű";
  var udhar$1 = "⥮";
  var ufisht$1 = "⥾";
  var Ufr$1 = "𝔘";
  var ufr$1 = "𝔲";
  var Ugrave$1 = "Ù";
  var ugrave$1 = "ù";
  var uHar$1 = "⥣";
  var uharl$1 = "↿";
  var uharr$1 = "↾";
  var uhblk$1 = "▀";
  var ulcorn$1 = "⌜";
  var ulcorner$1 = "⌜";
  var ulcrop$1 = "⌏";
  var ultri$1 = "◸";
  var Umacr$1 = "Ū";
  var umacr$1 = "ū";
  var uml$1 = "¨";
  var UnderBar$1 = "_";
  var UnderBrace$1 = "⏟";
  var UnderBracket$1 = "⎵";
  var UnderParenthesis$1 = "⏝";
  var Union$1 = "⋃";
  var UnionPlus$1 = "⊎";
  var Uogon$1 = "Ų";
  var uogon$1 = "ų";
  var Uopf$1 = "𝕌";
  var uopf$1 = "𝕦";
  var UpArrowBar$1 = "⤒";
  var uparrow$1 = "↑";
  var UpArrow$1 = "↑";
  var Uparrow$1 = "⇑";
  var UpArrowDownArrow$1 = "⇅";
  var updownarrow$1 = "↕";
  var UpDownArrow$1 = "↕";
  var Updownarrow$1 = "⇕";
  var UpEquilibrium$1 = "⥮";
  var upharpoonleft$1 = "↿";
  var upharpoonright$1 = "↾";
  var uplus$1 = "⊎";
  var UpperLeftArrow$1 = "↖";
  var UpperRightArrow$1 = "↗";
  var upsi$1 = "υ";
  var Upsi$1 = "ϒ";
  var upsih$1 = "ϒ";
  var Upsilon$1 = "Υ";
  var upsilon$1 = "υ";
  var UpTeeArrow$1 = "↥";
  var UpTee$1 = "⊥";
  var upuparrows$1 = "⇈";
  var urcorn$1 = "⌝";
  var urcorner$1 = "⌝";
  var urcrop$1 = "⌎";
  var Uring$1 = "Ů";
  var uring$1 = "ů";
  var urtri$1 = "◹";
  var Uscr$1 = "𝒰";
  var uscr$1 = "𝓊";
  var utdot$1 = "⋰";
  var Utilde$1 = "Ũ";
  var utilde$1 = "ũ";
  var utri$1 = "▵";
  var utrif$1 = "▴";
  var uuarr$1 = "⇈";
  var Uuml$1 = "Ü";
  var uuml$1 = "ü";
  var uwangle$1 = "⦧";
  var vangrt$1 = "⦜";
  var varepsilon$1 = "ϵ";
  var varkappa$1 = "ϰ";
  var varnothing$1 = "∅";
  var varphi$1 = "ϕ";
  var varpi$1 = "ϖ";
  var varpropto$1 = "∝";
  var varr$1 = "↕";
  var vArr$1 = "⇕";
  var varrho$1 = "ϱ";
  var varsigma$1 = "ς";
  var varsubsetneq$1 = "⊊︀";
  var varsubsetneqq$1 = "⫋︀";
  var varsupsetneq$1 = "⊋︀";
  var varsupsetneqq$1 = "⫌︀";
  var vartheta$1 = "ϑ";
  var vartriangleleft$1 = "⊲";
  var vartriangleright$1 = "⊳";
  var vBar$1 = "⫨";
  var Vbar$1 = "⫫";
  var vBarv$1 = "⫩";
  var Vcy$1 = "В";
  var vcy$1 = "в";
  var vdash$1 = "⊢";
  var vDash$1 = "⊨";
  var Vdash$1 = "⊩";
  var VDash$1 = "⊫";
  var Vdashl$1 = "⫦";
  var veebar$1 = "⊻";
  var vee$1 = "∨";
  var Vee$1 = "⋁";
  var veeeq$1 = "≚";
  var vellip$1 = "⋮";
  var verbar$1 = "|";
  var Verbar$1 = "‖";
  var vert$1 = "|";
  var Vert$1 = "‖";
  var VerticalBar$1 = "∣";
  var VerticalLine$1 = "|";
  var VerticalSeparator$1 = "❘";
  var VerticalTilde$1 = "≀";
  var VeryThinSpace$1 = " ";
  var Vfr$1 = "𝔙";
  var vfr$1 = "𝔳";
  var vltri$1 = "⊲";
  var vnsub$1 = "⊂⃒";
  var vnsup$1 = "⊃⃒";
  var Vopf$1 = "𝕍";
  var vopf$1 = "𝕧";
  var vprop$1 = "∝";
  var vrtri$1 = "⊳";
  var Vscr$1 = "𝒱";
  var vscr$1 = "𝓋";
  var vsubnE$1 = "⫋︀";
  var vsubne$1 = "⊊︀";
  var vsupnE$1 = "⫌︀";
  var vsupne$1 = "⊋︀";
  var Vvdash$1 = "⊪";
  var vzigzag$1 = "⦚";
  var Wcirc$1 = "Ŵ";
  var wcirc$1 = "ŵ";
  var wedbar$1 = "⩟";
  var wedge$1 = "∧";
  var Wedge$1 = "⋀";
  var wedgeq$1 = "≙";
  var weierp$1 = "℘";
  var Wfr$1 = "𝔚";
  var wfr$1 = "𝔴";
  var Wopf$1 = "𝕎";
  var wopf$1 = "𝕨";
  var wp$1 = "℘";
  var wr$1 = "≀";
  var wreath$1 = "≀";
  var Wscr$1 = "𝒲";
  var wscr$1 = "𝓌";
  var xcap$1 = "⋂";
  var xcirc$1 = "◯";
  var xcup$1 = "⋃";
  var xdtri$1 = "▽";
  var Xfr$1 = "𝔛";
  var xfr$1 = "𝔵";
  var xharr$1 = "⟷";
  var xhArr$1 = "⟺";
  var Xi$1 = "Ξ";
  var xi$1 = "ξ";
  var xlarr$1 = "⟵";
  var xlArr$1 = "⟸";
  var xmap$1 = "⟼";
  var xnis$1 = "⋻";
  var xodot$1 = "⨀";
  var Xopf$1 = "𝕏";
  var xopf$1 = "𝕩";
  var xoplus$1 = "⨁";
  var xotime$1 = "⨂";
  var xrarr$1 = "⟶";
  var xrArr$1 = "⟹";
  var Xscr$1 = "𝒳";
  var xscr$1 = "𝓍";
  var xsqcup$1 = "⨆";
  var xuplus$1 = "⨄";
  var xutri$1 = "△";
  var xvee$1 = "⋁";
  var xwedge$1 = "⋀";
  var Yacute$1 = "Ý";
  var yacute$1 = "ý";
  var YAcy$1 = "Я";
  var yacy$1 = "я";
  var Ycirc$1 = "Ŷ";
  var ycirc$1 = "ŷ";
  var Ycy$1 = "Ы";
  var ycy$1 = "ы";
  var yen$3 = "¥";
  var Yfr$1 = "𝔜";
  var yfr$1 = "𝔶";
  var YIcy$1 = "Ї";
  var yicy$1 = "ї";
  var Yopf$1 = "𝕐";
  var yopf$1 = "𝕪";
  var Yscr$1 = "𝒴";
  var yscr$1 = "𝓎";
  var YUcy$1 = "Ю";
  var yucy$1 = "ю";
  var yuml$1 = "ÿ";
  var Yuml$1 = "Ÿ";
  var Zacute$1 = "Ź";
  var zacute$1 = "ź";
  var Zcaron$1 = "Ž";
  var zcaron$1 = "ž";
  var Zcy$1 = "З";
  var zcy$1 = "з";
  var Zdot$1 = "Ż";
  var zdot$1 = "ż";
  var zeetrf$1 = "ℨ";
  var ZeroWidthSpace$1 = "​";
  var Zeta$1 = "Ζ";
  var zeta$1 = "ζ";
  var zfr$1 = "𝔷";
  var Zfr$1 = "ℨ";
  var ZHcy$1 = "Ж";
  var zhcy$1 = "ж";
  var zigrarr$1 = "⇝";
  var zopf$1 = "𝕫";
  var Zopf$1 = "ℤ";
  var Zscr$1 = "𝒵";
  var zscr$1 = "𝓏";
  var zwj$1 = "‍";
  var zwnj$1 = "‌";
  var entities$4 = {
  	Aacute: Aacute$1,
  	aacute: aacute$1,
  	Abreve: Abreve$1,
  	abreve: abreve$1,
  	ac: ac$1,
  	acd: acd$1,
  	acE: acE$1,
  	Acirc: Acirc$1,
  	acirc: acirc$1,
  	acute: acute$1,
  	Acy: Acy$1,
  	acy: acy$1,
  	AElig: AElig$1,
  	aelig: aelig$1,
  	af: af$1,
  	Afr: Afr$1,
  	afr: afr$1,
  	Agrave: Agrave$1,
  	agrave: agrave$1,
  	alefsym: alefsym$1,
  	aleph: aleph$1,
  	Alpha: Alpha$1,
  	alpha: alpha$1,
  	Amacr: Amacr$1,
  	amacr: amacr$1,
  	amalg: amalg$1,
  	amp: amp$1,
  	AMP: AMP$1,
  	andand: andand$1,
  	And: And$1,
  	and: and$1,
  	andd: andd$1,
  	andslope: andslope$1,
  	andv: andv$1,
  	ang: ang$1,
  	ange: ange$1,
  	angle: angle$1,
  	angmsdaa: angmsdaa$1,
  	angmsdab: angmsdab$1,
  	angmsdac: angmsdac$1,
  	angmsdad: angmsdad$1,
  	angmsdae: angmsdae$1,
  	angmsdaf: angmsdaf$1,
  	angmsdag: angmsdag$1,
  	angmsdah: angmsdah$1,
  	angmsd: angmsd$1,
  	angrt: angrt$1,
  	angrtvb: angrtvb$1,
  	angrtvbd: angrtvbd$1,
  	angsph: angsph$1,
  	angst: angst$1,
  	angzarr: angzarr$1,
  	Aogon: Aogon$1,
  	aogon: aogon$1,
  	Aopf: Aopf$1,
  	aopf: aopf$1,
  	apacir: apacir$1,
  	ap: ap$1,
  	apE: apE$1,
  	ape: ape$1,
  	apid: apid$1,
  	apos: apos$1,
  	ApplyFunction: ApplyFunction$1,
  	approx: approx$1,
  	approxeq: approxeq$1,
  	Aring: Aring$1,
  	aring: aring$1,
  	Ascr: Ascr$1,
  	ascr: ascr$1,
  	Assign: Assign$1,
  	ast: ast$1,
  	asymp: asymp$1,
  	asympeq: asympeq$1,
  	Atilde: Atilde$1,
  	atilde: atilde$1,
  	Auml: Auml$1,
  	auml: auml$1,
  	awconint: awconint$1,
  	awint: awint$1,
  	backcong: backcong$1,
  	backepsilon: backepsilon$1,
  	backprime: backprime$1,
  	backsim: backsim$1,
  	backsimeq: backsimeq$1,
  	Backslash: Backslash$1,
  	Barv: Barv$1,
  	barvee: barvee$1,
  	barwed: barwed$1,
  	Barwed: Barwed$1,
  	barwedge: barwedge$1,
  	bbrk: bbrk$1,
  	bbrktbrk: bbrktbrk$1,
  	bcong: bcong$1,
  	Bcy: Bcy$1,
  	bcy: bcy$1,
  	bdquo: bdquo$1,
  	becaus: becaus$1,
  	because: because$1,
  	Because: Because$1,
  	bemptyv: bemptyv$1,
  	bepsi: bepsi$1,
  	bernou: bernou$1,
  	Bernoullis: Bernoullis$1,
  	Beta: Beta$1,
  	beta: beta$1,
  	beth: beth$1,
  	between: between$1,
  	Bfr: Bfr$1,
  	bfr: bfr$1,
  	bigcap: bigcap$1,
  	bigcirc: bigcirc$1,
  	bigcup: bigcup$1,
  	bigodot: bigodot$1,
  	bigoplus: bigoplus$1,
  	bigotimes: bigotimes$1,
  	bigsqcup: bigsqcup$1,
  	bigstar: bigstar$1,
  	bigtriangledown: bigtriangledown$1,
  	bigtriangleup: bigtriangleup$1,
  	biguplus: biguplus$1,
  	bigvee: bigvee$1,
  	bigwedge: bigwedge$1,
  	bkarow: bkarow$1,
  	blacklozenge: blacklozenge$1,
  	blacksquare: blacksquare$1,
  	blacktriangle: blacktriangle$1,
  	blacktriangledown: blacktriangledown$1,
  	blacktriangleleft: blacktriangleleft$1,
  	blacktriangleright: blacktriangleright$1,
  	blank: blank$1,
  	blk12: blk12$1,
  	blk14: blk14$1,
  	blk34: blk34$1,
  	block: block$3,
  	bne: bne$1,
  	bnequiv: bnequiv$1,
  	bNot: bNot$1,
  	bnot: bnot$1,
  	Bopf: Bopf$1,
  	bopf: bopf$1,
  	bot: bot$1,
  	bottom: bottom$1,
  	bowtie: bowtie$1,
  	boxbox: boxbox$1,
  	boxdl: boxdl$1,
  	boxdL: boxdL$1,
  	boxDl: boxDl$1,
  	boxDL: boxDL$1,
  	boxdr: boxdr$1,
  	boxdR: boxdR$1,
  	boxDr: boxDr$1,
  	boxDR: boxDR$1,
  	boxh: boxh$1,
  	boxH: boxH$1,
  	boxhd: boxhd$1,
  	boxHd: boxHd$1,
  	boxhD: boxhD$1,
  	boxHD: boxHD$1,
  	boxhu: boxhu$1,
  	boxHu: boxHu$1,
  	boxhU: boxhU$1,
  	boxHU: boxHU$1,
  	boxminus: boxminus$1,
  	boxplus: boxplus$1,
  	boxtimes: boxtimes$1,
  	boxul: boxul$1,
  	boxuL: boxuL$1,
  	boxUl: boxUl$1,
  	boxUL: boxUL$1,
  	boxur: boxur$1,
  	boxuR: boxuR$1,
  	boxUr: boxUr$1,
  	boxUR: boxUR$1,
  	boxv: boxv$1,
  	boxV: boxV$1,
  	boxvh: boxvh$1,
  	boxvH: boxvH$1,
  	boxVh: boxVh$1,
  	boxVH: boxVH$1,
  	boxvl: boxvl$1,
  	boxvL: boxvL$1,
  	boxVl: boxVl$1,
  	boxVL: boxVL$1,
  	boxvr: boxvr$1,
  	boxvR: boxvR$1,
  	boxVr: boxVr$1,
  	boxVR: boxVR$1,
  	bprime: bprime$1,
  	breve: breve$1,
  	Breve: Breve$1,
  	brvbar: brvbar$1,
  	bscr: bscr$1,
  	Bscr: Bscr$1,
  	bsemi: bsemi$1,
  	bsim: bsim$1,
  	bsime: bsime$1,
  	bsolb: bsolb$1,
  	bsol: bsol$1,
  	bsolhsub: bsolhsub$1,
  	bull: bull$1,
  	bullet: bullet$1,
  	bump: bump$1,
  	bumpE: bumpE$1,
  	bumpe: bumpe$1,
  	Bumpeq: Bumpeq$1,
  	bumpeq: bumpeq$1,
  	Cacute: Cacute$1,
  	cacute: cacute$1,
  	capand: capand$1,
  	capbrcup: capbrcup$1,
  	capcap: capcap$1,
  	cap: cap$1,
  	Cap: Cap$1,
  	capcup: capcup$1,
  	capdot: capdot$1,
  	CapitalDifferentialD: CapitalDifferentialD$1,
  	caps: caps$1,
  	caret: caret$1,
  	caron: caron$1,
  	Cayleys: Cayleys$1,
  	ccaps: ccaps$1,
  	Ccaron: Ccaron$1,
  	ccaron: ccaron$1,
  	Ccedil: Ccedil$1,
  	ccedil: ccedil$1,
  	Ccirc: Ccirc$1,
  	ccirc: ccirc$1,
  	Cconint: Cconint$1,
  	ccups: ccups$1,
  	ccupssm: ccupssm$1,
  	Cdot: Cdot$1,
  	cdot: cdot$1,
  	cedil: cedil$1,
  	Cedilla: Cedilla$1,
  	cemptyv: cemptyv$1,
  	cent: cent$1,
  	centerdot: centerdot$1,
  	CenterDot: CenterDot$1,
  	cfr: cfr$1,
  	Cfr: Cfr$1,
  	CHcy: CHcy$1,
  	chcy: chcy$1,
  	check: check$1,
  	checkmark: checkmark$1,
  	Chi: Chi$1,
  	chi: chi$1,
  	circ: circ$1,
  	circeq: circeq$1,
  	circlearrowleft: circlearrowleft$1,
  	circlearrowright: circlearrowright$1,
  	circledast: circledast$1,
  	circledcirc: circledcirc$1,
  	circleddash: circleddash$1,
  	CircleDot: CircleDot$1,
  	circledR: circledR$1,
  	circledS: circledS$1,
  	CircleMinus: CircleMinus$1,
  	CirclePlus: CirclePlus$1,
  	CircleTimes: CircleTimes$1,
  	cir: cir$1,
  	cirE: cirE$1,
  	cire: cire$1,
  	cirfnint: cirfnint$1,
  	cirmid: cirmid$1,
  	cirscir: cirscir$1,
  	ClockwiseContourIntegral: ClockwiseContourIntegral$1,
  	CloseCurlyDoubleQuote: CloseCurlyDoubleQuote$1,
  	CloseCurlyQuote: CloseCurlyQuote$1,
  	clubs: clubs$3,
  	clubsuit: clubsuit$1,
  	colon: colon$1,
  	Colon: Colon$1,
  	Colone: Colone$1,
  	colone: colone$1,
  	coloneq: coloneq$1,
  	comma: comma$1,
  	commat: commat$1,
  	comp: comp$1,
  	compfn: compfn$1,
  	complement: complement$1,
  	complexes: complexes$1,
  	cong: cong$1,
  	congdot: congdot$1,
  	Congruent: Congruent$1,
  	conint: conint$1,
  	Conint: Conint$1,
  	ContourIntegral: ContourIntegral$1,
  	copf: copf$1,
  	Copf: Copf$1,
  	coprod: coprod$1,
  	Coproduct: Coproduct$1,
  	copy: copy$1,
  	COPY: COPY$1,
  	copysr: copysr$1,
  	CounterClockwiseContourIntegral: CounterClockwiseContourIntegral$1,
  	crarr: crarr$1,
  	cross: cross$1,
  	Cross: Cross$1,
  	Cscr: Cscr$1,
  	cscr: cscr$1,
  	csub: csub$1,
  	csube: csube$1,
  	csup: csup$1,
  	csupe: csupe$1,
  	ctdot: ctdot$1,
  	cudarrl: cudarrl$1,
  	cudarrr: cudarrr$1,
  	cuepr: cuepr$1,
  	cuesc: cuesc$1,
  	cularr: cularr$1,
  	cularrp: cularrp$1,
  	cupbrcap: cupbrcap$1,
  	cupcap: cupcap$1,
  	CupCap: CupCap$1,
  	cup: cup$1,
  	Cup: Cup$1,
  	cupcup: cupcup$1,
  	cupdot: cupdot$1,
  	cupor: cupor$1,
  	cups: cups$1,
  	curarr: curarr$1,
  	curarrm: curarrm$1,
  	curlyeqprec: curlyeqprec$1,
  	curlyeqsucc: curlyeqsucc$1,
  	curlyvee: curlyvee$1,
  	curlywedge: curlywedge$1,
  	curren: curren$1,
  	curvearrowleft: curvearrowleft$1,
  	curvearrowright: curvearrowright$1,
  	cuvee: cuvee$1,
  	cuwed: cuwed$1,
  	cwconint: cwconint$1,
  	cwint: cwint$1,
  	cylcty: cylcty$1,
  	dagger: dagger$3,
  	Dagger: Dagger$1,
  	daleth: daleth$1,
  	darr: darr$1,
  	Darr: Darr$1,
  	dArr: dArr$1,
  	dash: dash$3,
  	Dashv: Dashv$1,
  	dashv: dashv$1,
  	dbkarow: dbkarow$1,
  	dblac: dblac$1,
  	Dcaron: Dcaron$1,
  	dcaron: dcaron$1,
  	Dcy: Dcy$1,
  	dcy: dcy$1,
  	ddagger: ddagger$1,
  	ddarr: ddarr$1,
  	DD: DD$1,
  	dd: dd$1,
  	DDotrahd: DDotrahd$1,
  	ddotseq: ddotseq$1,
  	deg: deg$1,
  	Del: Del$1,
  	Delta: Delta$1,
  	delta: delta$1,
  	demptyv: demptyv$1,
  	dfisht: dfisht$1,
  	Dfr: Dfr$1,
  	dfr: dfr$1,
  	dHar: dHar$1,
  	dharl: dharl$1,
  	dharr: dharr$1,
  	DiacriticalAcute: DiacriticalAcute$1,
  	DiacriticalDot: DiacriticalDot$1,
  	DiacriticalDoubleAcute: DiacriticalDoubleAcute$1,
  	DiacriticalGrave: DiacriticalGrave$1,
  	DiacriticalTilde: DiacriticalTilde$1,
  	diam: diam$1,
  	diamond: diamond$1,
  	Diamond: Diamond$1,
  	diamondsuit: diamondsuit$1,
  	diams: diams$1,
  	die: die$1,
  	DifferentialD: DifferentialD$1,
  	digamma: digamma$1,
  	disin: disin$1,
  	div: div$1,
  	divide: divide$1,
  	divideontimes: divideontimes$1,
  	divonx: divonx$1,
  	DJcy: DJcy$1,
  	djcy: djcy$1,
  	dlcorn: dlcorn$1,
  	dlcrop: dlcrop$1,
  	dollar: dollar$3,
  	Dopf: Dopf$1,
  	dopf: dopf$1,
  	Dot: Dot$1,
  	dot: dot$1,
  	DotDot: DotDot$1,
  	doteq: doteq$1,
  	doteqdot: doteqdot$1,
  	DotEqual: DotEqual$1,
  	dotminus: dotminus$1,
  	dotplus: dotplus$1,
  	dotsquare: dotsquare$1,
  	doublebarwedge: doublebarwedge$1,
  	DoubleContourIntegral: DoubleContourIntegral$1,
  	DoubleDot: DoubleDot$1,
  	DoubleDownArrow: DoubleDownArrow$1,
  	DoubleLeftArrow: DoubleLeftArrow$1,
  	DoubleLeftRightArrow: DoubleLeftRightArrow$1,
  	DoubleLeftTee: DoubleLeftTee$1,
  	DoubleLongLeftArrow: DoubleLongLeftArrow$1,
  	DoubleLongLeftRightArrow: DoubleLongLeftRightArrow$1,
  	DoubleLongRightArrow: DoubleLongRightArrow$1,
  	DoubleRightArrow: DoubleRightArrow$1,
  	DoubleRightTee: DoubleRightTee$1,
  	DoubleUpArrow: DoubleUpArrow$1,
  	DoubleUpDownArrow: DoubleUpDownArrow$1,
  	DoubleVerticalBar: DoubleVerticalBar$1,
  	DownArrowBar: DownArrowBar$1,
  	downarrow: downarrow$1,
  	DownArrow: DownArrow$1,
  	Downarrow: Downarrow$1,
  	DownArrowUpArrow: DownArrowUpArrow$1,
  	DownBreve: DownBreve$1,
  	downdownarrows: downdownarrows$1,
  	downharpoonleft: downharpoonleft$1,
  	downharpoonright: downharpoonright$1,
  	DownLeftRightVector: DownLeftRightVector$1,
  	DownLeftTeeVector: DownLeftTeeVector$1,
  	DownLeftVectorBar: DownLeftVectorBar$1,
  	DownLeftVector: DownLeftVector$1,
  	DownRightTeeVector: DownRightTeeVector$1,
  	DownRightVectorBar: DownRightVectorBar$1,
  	DownRightVector: DownRightVector$1,
  	DownTeeArrow: DownTeeArrow$1,
  	DownTee: DownTee$1,
  	drbkarow: drbkarow$1,
  	drcorn: drcorn$1,
  	drcrop: drcrop$1,
  	Dscr: Dscr$1,
  	dscr: dscr$1,
  	DScy: DScy$1,
  	dscy: dscy$1,
  	dsol: dsol$1,
  	Dstrok: Dstrok$1,
  	dstrok: dstrok$1,
  	dtdot: dtdot$1,
  	dtri: dtri$1,
  	dtrif: dtrif$1,
  	duarr: duarr$1,
  	duhar: duhar$1,
  	dwangle: dwangle$1,
  	DZcy: DZcy$1,
  	dzcy: dzcy$1,
  	dzigrarr: dzigrarr$1,
  	Eacute: Eacute$1,
  	eacute: eacute$1,
  	easter: easter$1,
  	Ecaron: Ecaron$1,
  	ecaron: ecaron$1,
  	Ecirc: Ecirc$1,
  	ecirc: ecirc$1,
  	ecir: ecir$1,
  	ecolon: ecolon$1,
  	Ecy: Ecy$1,
  	ecy: ecy$1,
  	eDDot: eDDot$1,
  	Edot: Edot$1,
  	edot: edot$1,
  	eDot: eDot$1,
  	ee: ee$1,
  	efDot: efDot$1,
  	Efr: Efr$1,
  	efr: efr$1,
  	eg: eg$1,
  	Egrave: Egrave$1,
  	egrave: egrave$1,
  	egs: egs$1,
  	egsdot: egsdot$1,
  	el: el$1,
  	Element: Element$1,
  	elinters: elinters$1,
  	ell: ell$1,
  	els: els$1,
  	elsdot: elsdot$1,
  	Emacr: Emacr$1,
  	emacr: emacr$1,
  	empty: empty$1,
  	emptyset: emptyset$1,
  	EmptySmallSquare: EmptySmallSquare$1,
  	emptyv: emptyv$1,
  	EmptyVerySmallSquare: EmptyVerySmallSquare$1,
  	emsp13: emsp13$1,
  	emsp14: emsp14$1,
  	emsp: emsp$1,
  	ENG: ENG$1,
  	eng: eng$1,
  	ensp: ensp$1,
  	Eogon: Eogon$1,
  	eogon: eogon$1,
  	Eopf: Eopf$1,
  	eopf: eopf$1,
  	epar: epar$1,
  	eparsl: eparsl$1,
  	eplus: eplus$1,
  	epsi: epsi$1,
  	Epsilon: Epsilon$1,
  	epsilon: epsilon$1,
  	epsiv: epsiv$1,
  	eqcirc: eqcirc$1,
  	eqcolon: eqcolon$1,
  	eqsim: eqsim$1,
  	eqslantgtr: eqslantgtr$1,
  	eqslantless: eqslantless$1,
  	Equal: Equal$1,
  	equals: equals$1,
  	EqualTilde: EqualTilde$1,
  	equest: equest$1,
  	Equilibrium: Equilibrium$1,
  	equiv: equiv$1,
  	equivDD: equivDD$1,
  	eqvparsl: eqvparsl$1,
  	erarr: erarr$1,
  	erDot: erDot$1,
  	escr: escr$1,
  	Escr: Escr$1,
  	esdot: esdot$1,
  	Esim: Esim$1,
  	esim: esim$1,
  	Eta: Eta$1,
  	eta: eta$1,
  	ETH: ETH$1,
  	eth: eth$1,
  	Euml: Euml$1,
  	euml: euml$1,
  	euro: euro$3,
  	excl: excl$1,
  	exist: exist$1,
  	Exists: Exists$1,
  	expectation: expectation$1,
  	exponentiale: exponentiale$1,
  	ExponentialE: ExponentialE$1,
  	fallingdotseq: fallingdotseq$1,
  	Fcy: Fcy$1,
  	fcy: fcy$1,
  	female: female$1,
  	ffilig: ffilig$1,
  	fflig: fflig$1,
  	ffllig: ffllig$1,
  	Ffr: Ffr$1,
  	ffr: ffr$1,
  	filig: filig$1,
  	FilledSmallSquare: FilledSmallSquare$1,
  	FilledVerySmallSquare: FilledVerySmallSquare$1,
  	fjlig: fjlig$1,
  	flat: flat$1,
  	fllig: fllig$1,
  	fltns: fltns$1,
  	fnof: fnof$1,
  	Fopf: Fopf$1,
  	fopf: fopf$1,
  	forall: forall$1,
  	ForAll: ForAll$1,
  	fork: fork$1,
  	forkv: forkv$1,
  	Fouriertrf: Fouriertrf$1,
  	fpartint: fpartint$1,
  	frac12: frac12$1,
  	frac13: frac13$1,
  	frac14: frac14$1,
  	frac15: frac15$1,
  	frac16: frac16$1,
  	frac18: frac18$1,
  	frac23: frac23$1,
  	frac25: frac25$1,
  	frac34: frac34$1,
  	frac35: frac35$1,
  	frac38: frac38$1,
  	frac45: frac45$1,
  	frac56: frac56$1,
  	frac58: frac58$1,
  	frac78: frac78$1,
  	frasl: frasl$1,
  	frown: frown$1,
  	fscr: fscr$1,
  	Fscr: Fscr$1,
  	gacute: gacute$1,
  	Gamma: Gamma$1,
  	gamma: gamma$1,
  	Gammad: Gammad$1,
  	gammad: gammad$1,
  	gap: gap$1,
  	Gbreve: Gbreve$1,
  	gbreve: gbreve$1,
  	Gcedil: Gcedil$1,
  	Gcirc: Gcirc$1,
  	gcirc: gcirc$1,
  	Gcy: Gcy$1,
  	gcy: gcy$1,
  	Gdot: Gdot$1,
  	gdot: gdot$1,
  	ge: ge$1,
  	gE: gE$1,
  	gEl: gEl$1,
  	gel: gel$1,
  	geq: geq$1,
  	geqq: geqq$1,
  	geqslant: geqslant$1,
  	gescc: gescc$1,
  	ges: ges$1,
  	gesdot: gesdot$1,
  	gesdoto: gesdoto$1,
  	gesdotol: gesdotol$1,
  	gesl: gesl$1,
  	gesles: gesles$1,
  	Gfr: Gfr$1,
  	gfr: gfr$1,
  	gg: gg$1,
  	Gg: Gg$1,
  	ggg: ggg$1,
  	gimel: gimel$1,
  	GJcy: GJcy$1,
  	gjcy: gjcy$1,
  	gla: gla$1,
  	gl: gl$1,
  	glE: glE$1,
  	glj: glj$1,
  	gnap: gnap$1,
  	gnapprox: gnapprox$1,
  	gne: gne$1,
  	gnE: gnE$1,
  	gneq: gneq$1,
  	gneqq: gneqq$1,
  	gnsim: gnsim$1,
  	Gopf: Gopf$1,
  	gopf: gopf$1,
  	grave: grave$1,
  	GreaterEqual: GreaterEqual$1,
  	GreaterEqualLess: GreaterEqualLess$1,
  	GreaterFullEqual: GreaterFullEqual$1,
  	GreaterGreater: GreaterGreater$1,
  	GreaterLess: GreaterLess$1,
  	GreaterSlantEqual: GreaterSlantEqual$1,
  	GreaterTilde: GreaterTilde$1,
  	Gscr: Gscr$1,
  	gscr: gscr$1,
  	gsim: gsim$1,
  	gsime: gsime$1,
  	gsiml: gsiml$1,
  	gtcc: gtcc$1,
  	gtcir: gtcir$1,
  	gt: gt$1,
  	GT: GT$1,
  	Gt: Gt$1,
  	gtdot: gtdot$1,
  	gtlPar: gtlPar$1,
  	gtquest: gtquest$1,
  	gtrapprox: gtrapprox$1,
  	gtrarr: gtrarr$1,
  	gtrdot: gtrdot$1,
  	gtreqless: gtreqless$1,
  	gtreqqless: gtreqqless$1,
  	gtrless: gtrless$1,
  	gtrsim: gtrsim$1,
  	gvertneqq: gvertneqq$1,
  	gvnE: gvnE$1,
  	Hacek: Hacek$1,
  	hairsp: hairsp$1,
  	half: half$1,
  	hamilt: hamilt$1,
  	HARDcy: HARDcy$1,
  	hardcy: hardcy$1,
  	harrcir: harrcir$1,
  	harr: harr$1,
  	hArr: hArr$1,
  	harrw: harrw$1,
  	Hat: Hat$1,
  	hbar: hbar$1,
  	Hcirc: Hcirc$1,
  	hcirc: hcirc$1,
  	hearts: hearts$3,
  	heartsuit: heartsuit$1,
  	hellip: hellip$1,
  	hercon: hercon$1,
  	hfr: hfr$1,
  	Hfr: Hfr$1,
  	HilbertSpace: HilbertSpace$1,
  	hksearow: hksearow$1,
  	hkswarow: hkswarow$1,
  	hoarr: hoarr$1,
  	homtht: homtht$1,
  	hookleftarrow: hookleftarrow$1,
  	hookrightarrow: hookrightarrow$1,
  	hopf: hopf$1,
  	Hopf: Hopf$1,
  	horbar: horbar$1,
  	HorizontalLine: HorizontalLine$1,
  	hscr: hscr$1,
  	Hscr: Hscr$1,
  	hslash: hslash$1,
  	Hstrok: Hstrok$1,
  	hstrok: hstrok$1,
  	HumpDownHump: HumpDownHump$1,
  	HumpEqual: HumpEqual$1,
  	hybull: hybull$1,
  	hyphen: hyphen$1,
  	Iacute: Iacute$1,
  	iacute: iacute$1,
  	ic: ic$1,
  	Icirc: Icirc$1,
  	icirc: icirc$1,
  	Icy: Icy$1,
  	icy: icy$1,
  	Idot: Idot$1,
  	IEcy: IEcy$1,
  	iecy: iecy$1,
  	iexcl: iexcl$1,
  	iff: iff$1,
  	ifr: ifr$1,
  	Ifr: Ifr$1,
  	Igrave: Igrave$1,
  	igrave: igrave$1,
  	ii: ii$1,
  	iiiint: iiiint$1,
  	iiint: iiint$1,
  	iinfin: iinfin$1,
  	iiota: iiota$1,
  	IJlig: IJlig$1,
  	ijlig: ijlig$1,
  	Imacr: Imacr$1,
  	imacr: imacr$1,
  	image: image$4,
  	ImaginaryI: ImaginaryI$1,
  	imagline: imagline$1,
  	imagpart: imagpart$1,
  	imath: imath$1,
  	Im: Im$1,
  	imof: imof$1,
  	imped: imped$1,
  	Implies: Implies$1,
  	incare: incare$1,
  	"in": "∈",
  	infin: infin$1,
  	infintie: infintie$1,
  	inodot: inodot$1,
  	intcal: intcal$1,
  	int: int$1,
  	Int: Int$1,
  	integers: integers$1,
  	Integral: Integral$1,
  	intercal: intercal$1,
  	Intersection: Intersection$1,
  	intlarhk: intlarhk$1,
  	intprod: intprod$1,
  	InvisibleComma: InvisibleComma$1,
  	InvisibleTimes: InvisibleTimes$1,
  	IOcy: IOcy$1,
  	iocy: iocy$1,
  	Iogon: Iogon$1,
  	iogon: iogon$1,
  	Iopf: Iopf$1,
  	iopf: iopf$1,
  	Iota: Iota$1,
  	iota: iota$1,
  	iprod: iprod$1,
  	iquest: iquest$1,
  	iscr: iscr$1,
  	Iscr: Iscr$1,
  	isin: isin$1,
  	isindot: isindot$1,
  	isinE: isinE$1,
  	isins: isins$1,
  	isinsv: isinsv$1,
  	isinv: isinv$1,
  	it: it$3,
  	Itilde: Itilde$1,
  	itilde: itilde$1,
  	Iukcy: Iukcy$1,
  	iukcy: iukcy$1,
  	Iuml: Iuml$1,
  	iuml: iuml$1,
  	Jcirc: Jcirc$1,
  	jcirc: jcirc$1,
  	Jcy: Jcy$1,
  	jcy: jcy$1,
  	Jfr: Jfr$1,
  	jfr: jfr$1,
  	jmath: jmath$1,
  	Jopf: Jopf$1,
  	jopf: jopf$1,
  	Jscr: Jscr$1,
  	jscr: jscr$1,
  	Jsercy: Jsercy$1,
  	jsercy: jsercy$1,
  	Jukcy: Jukcy$1,
  	jukcy: jukcy$1,
  	Kappa: Kappa$1,
  	kappa: kappa$1,
  	kappav: kappav$1,
  	Kcedil: Kcedil$1,
  	kcedil: kcedil$1,
  	Kcy: Kcy$1,
  	kcy: kcy$1,
  	Kfr: Kfr$1,
  	kfr: kfr$1,
  	kgreen: kgreen$1,
  	KHcy: KHcy$1,
  	khcy: khcy$1,
  	KJcy: KJcy$1,
  	kjcy: kjcy$1,
  	Kopf: Kopf$1,
  	kopf: kopf$1,
  	Kscr: Kscr$1,
  	kscr: kscr$1,
  	lAarr: lAarr$1,
  	Lacute: Lacute$1,
  	lacute: lacute$1,
  	laemptyv: laemptyv$1,
  	lagran: lagran$1,
  	Lambda: Lambda$1,
  	lambda: lambda$1,
  	lang: lang$1,
  	Lang: Lang$1,
  	langd: langd$1,
  	langle: langle$1,
  	lap: lap$1,
  	Laplacetrf: Laplacetrf$1,
  	laquo: laquo$1,
  	larrb: larrb$1,
  	larrbfs: larrbfs$1,
  	larr: larr$1,
  	Larr: Larr$1,
  	lArr: lArr$1,
  	larrfs: larrfs$1,
  	larrhk: larrhk$1,
  	larrlp: larrlp$1,
  	larrpl: larrpl$1,
  	larrsim: larrsim$1,
  	larrtl: larrtl$1,
  	latail: latail$1,
  	lAtail: lAtail$1,
  	lat: lat$1,
  	late: late$1,
  	lates: lates$1,
  	lbarr: lbarr$1,
  	lBarr: lBarr$1,
  	lbbrk: lbbrk$1,
  	lbrace: lbrace$1,
  	lbrack: lbrack$1,
  	lbrke: lbrke$1,
  	lbrksld: lbrksld$1,
  	lbrkslu: lbrkslu$1,
  	Lcaron: Lcaron$1,
  	lcaron: lcaron$1,
  	Lcedil: Lcedil$1,
  	lcedil: lcedil$1,
  	lceil: lceil$1,
  	lcub: lcub$1,
  	Lcy: Lcy$1,
  	lcy: lcy$1,
  	ldca: ldca$1,
  	ldquo: ldquo$1,
  	ldquor: ldquor$1,
  	ldrdhar: ldrdhar$1,
  	ldrushar: ldrushar$1,
  	ldsh: ldsh$1,
  	le: le$1,
  	lE: lE$1,
  	LeftAngleBracket: LeftAngleBracket$1,
  	LeftArrowBar: LeftArrowBar$1,
  	leftarrow: leftarrow$1,
  	LeftArrow: LeftArrow$1,
  	Leftarrow: Leftarrow$1,
  	LeftArrowRightArrow: LeftArrowRightArrow$1,
  	leftarrowtail: leftarrowtail$1,
  	LeftCeiling: LeftCeiling$1,
  	LeftDoubleBracket: LeftDoubleBracket$1,
  	LeftDownTeeVector: LeftDownTeeVector$1,
  	LeftDownVectorBar: LeftDownVectorBar$1,
  	LeftDownVector: LeftDownVector$1,
  	LeftFloor: LeftFloor$1,
  	leftharpoondown: leftharpoondown$1,
  	leftharpoonup: leftharpoonup$1,
  	leftleftarrows: leftleftarrows$1,
  	leftrightarrow: leftrightarrow$1,
  	LeftRightArrow: LeftRightArrow$1,
  	Leftrightarrow: Leftrightarrow$1,
  	leftrightarrows: leftrightarrows$1,
  	leftrightharpoons: leftrightharpoons$1,
  	leftrightsquigarrow: leftrightsquigarrow$1,
  	LeftRightVector: LeftRightVector$1,
  	LeftTeeArrow: LeftTeeArrow$1,
  	LeftTee: LeftTee$1,
  	LeftTeeVector: LeftTeeVector$1,
  	leftthreetimes: leftthreetimes$1,
  	LeftTriangleBar: LeftTriangleBar$1,
  	LeftTriangle: LeftTriangle$1,
  	LeftTriangleEqual: LeftTriangleEqual$1,
  	LeftUpDownVector: LeftUpDownVector$1,
  	LeftUpTeeVector: LeftUpTeeVector$1,
  	LeftUpVectorBar: LeftUpVectorBar$1,
  	LeftUpVector: LeftUpVector$1,
  	LeftVectorBar: LeftVectorBar$1,
  	LeftVector: LeftVector$1,
  	lEg: lEg$1,
  	leg: leg$3,
  	leq: leq$1,
  	leqq: leqq$1,
  	leqslant: leqslant$1,
  	lescc: lescc$1,
  	les: les$1,
  	lesdot: lesdot$1,
  	lesdoto: lesdoto$1,
  	lesdotor: lesdotor$1,
  	lesg: lesg$1,
  	lesges: lesges$1,
  	lessapprox: lessapprox$1,
  	lessdot: lessdot$1,
  	lesseqgtr: lesseqgtr$1,
  	lesseqqgtr: lesseqqgtr$1,
  	LessEqualGreater: LessEqualGreater$1,
  	LessFullEqual: LessFullEqual$1,
  	LessGreater: LessGreater$1,
  	lessgtr: lessgtr$1,
  	LessLess: LessLess$1,
  	lesssim: lesssim$1,
  	LessSlantEqual: LessSlantEqual$1,
  	LessTilde: LessTilde$1,
  	lfisht: lfisht$1,
  	lfloor: lfloor$1,
  	Lfr: Lfr$1,
  	lfr: lfr$1,
  	lg: lg$1,
  	lgE: lgE$1,
  	lHar: lHar$1,
  	lhard: lhard$1,
  	lharu: lharu$1,
  	lharul: lharul$1,
  	lhblk: lhblk$1,
  	LJcy: LJcy$1,
  	ljcy: ljcy$1,
  	llarr: llarr$1,
  	ll: ll$1,
  	Ll: Ll$1,
  	llcorner: llcorner$1,
  	Lleftarrow: Lleftarrow$1,
  	llhard: llhard$1,
  	lltri: lltri$1,
  	Lmidot: Lmidot$1,
  	lmidot: lmidot$1,
  	lmoustache: lmoustache$1,
  	lmoust: lmoust$1,
  	lnap: lnap$1,
  	lnapprox: lnapprox$1,
  	lne: lne$1,
  	lnE: lnE$1,
  	lneq: lneq$1,
  	lneqq: lneqq$1,
  	lnsim: lnsim$1,
  	loang: loang$1,
  	loarr: loarr$1,
  	lobrk: lobrk$1,
  	longleftarrow: longleftarrow$1,
  	LongLeftArrow: LongLeftArrow$1,
  	Longleftarrow: Longleftarrow$1,
  	longleftrightarrow: longleftrightarrow$1,
  	LongLeftRightArrow: LongLeftRightArrow$1,
  	Longleftrightarrow: Longleftrightarrow$1,
  	longmapsto: longmapsto$1,
  	longrightarrow: longrightarrow$1,
  	LongRightArrow: LongRightArrow$1,
  	Longrightarrow: Longrightarrow$1,
  	looparrowleft: looparrowleft$1,
  	looparrowright: looparrowright$1,
  	lopar: lopar$1,
  	Lopf: Lopf$1,
  	lopf: lopf$1,
  	loplus: loplus$1,
  	lotimes: lotimes$1,
  	lowast: lowast$1,
  	lowbar: lowbar$1,
  	LowerLeftArrow: LowerLeftArrow$1,
  	LowerRightArrow: LowerRightArrow$1,
  	loz: loz$1,
  	lozenge: lozenge$1,
  	lozf: lozf$1,
  	lpar: lpar$1,
  	lparlt: lparlt$1,
  	lrarr: lrarr$1,
  	lrcorner: lrcorner$1,
  	lrhar: lrhar$1,
  	lrhard: lrhard$1,
  	lrm: lrm$1,
  	lrtri: lrtri$1,
  	lsaquo: lsaquo$1,
  	lscr: lscr$1,
  	Lscr: Lscr$1,
  	lsh: lsh$1,
  	Lsh: Lsh$1,
  	lsim: lsim$1,
  	lsime: lsime$1,
  	lsimg: lsimg$1,
  	lsqb: lsqb$1,
  	lsquo: lsquo$1,
  	lsquor: lsquor$1,
  	Lstrok: Lstrok$1,
  	lstrok: lstrok$1,
  	ltcc: ltcc$1,
  	ltcir: ltcir$1,
  	lt: lt$1,
  	LT: LT$1,
  	Lt: Lt$1,
  	ltdot: ltdot$1,
  	lthree: lthree$1,
  	ltimes: ltimes$1,
  	ltlarr: ltlarr$1,
  	ltquest: ltquest$1,
  	ltri: ltri$1,
  	ltrie: ltrie$1,
  	ltrif: ltrif$1,
  	ltrPar: ltrPar$1,
  	lurdshar: lurdshar$1,
  	luruhar: luruhar$1,
  	lvertneqq: lvertneqq$1,
  	lvnE: lvnE$1,
  	macr: macr$1,
  	male: male$1,
  	malt: malt$1,
  	maltese: maltese$1,
  	"Map": "⤅",
  	map: map$2,
  	mapsto: mapsto$1,
  	mapstodown: mapstodown$1,
  	mapstoleft: mapstoleft$1,
  	mapstoup: mapstoup$1,
  	marker: marker$1,
  	mcomma: mcomma$1,
  	Mcy: Mcy$1,
  	mcy: mcy$1,
  	mdash: mdash$1,
  	mDDot: mDDot$1,
  	measuredangle: measuredangle$1,
  	MediumSpace: MediumSpace$1,
  	Mellintrf: Mellintrf$1,
  	Mfr: Mfr$1,
  	mfr: mfr$1,
  	mho: mho$1,
  	micro: micro$1,
  	midast: midast$1,
  	midcir: midcir$1,
  	mid: mid$1,
  	middot: middot$1,
  	minusb: minusb$1,
  	minus: minus$1,
  	minusd: minusd$1,
  	minusdu: minusdu$1,
  	MinusPlus: MinusPlus$1,
  	mlcp: mlcp$1,
  	mldr: mldr$1,
  	mnplus: mnplus$1,
  	models: models$1,
  	Mopf: Mopf$1,
  	mopf: mopf$1,
  	mp: mp$1,
  	mscr: mscr$1,
  	Mscr: Mscr$1,
  	mstpos: mstpos$1,
  	Mu: Mu$1,
  	mu: mu$1,
  	multimap: multimap$1,
  	mumap: mumap$1,
  	nabla: nabla$1,
  	Nacute: Nacute$1,
  	nacute: nacute$1,
  	nang: nang$1,
  	nap: nap$1,
  	napE: napE$1,
  	napid: napid$1,
  	napos: napos$1,
  	napprox: napprox$1,
  	natural: natural$1,
  	naturals: naturals$1,
  	natur: natur$1,
  	nbsp: nbsp$1,
  	nbump: nbump$1,
  	nbumpe: nbumpe$1,
  	ncap: ncap$1,
  	Ncaron: Ncaron$1,
  	ncaron: ncaron$1,
  	Ncedil: Ncedil$1,
  	ncedil: ncedil$1,
  	ncong: ncong$1,
  	ncongdot: ncongdot$1,
  	ncup: ncup$1,
  	Ncy: Ncy$1,
  	ncy: ncy$1,
  	ndash: ndash$1,
  	nearhk: nearhk$1,
  	nearr: nearr$1,
  	neArr: neArr$1,
  	nearrow: nearrow$1,
  	ne: ne$1,
  	nedot: nedot$1,
  	NegativeMediumSpace: NegativeMediumSpace$1,
  	NegativeThickSpace: NegativeThickSpace$1,
  	NegativeThinSpace: NegativeThinSpace$1,
  	NegativeVeryThinSpace: NegativeVeryThinSpace$1,
  	nequiv: nequiv$1,
  	nesear: nesear$1,
  	nesim: nesim$1,
  	NestedGreaterGreater: NestedGreaterGreater$1,
  	NestedLessLess: NestedLessLess$1,
  	NewLine: NewLine$1,
  	nexist: nexist$1,
  	nexists: nexists$1,
  	Nfr: Nfr$1,
  	nfr: nfr$1,
  	ngE: ngE$1,
  	nge: nge$1,
  	ngeq: ngeq$1,
  	ngeqq: ngeqq$1,
  	ngeqslant: ngeqslant$1,
  	nges: nges$1,
  	nGg: nGg$1,
  	ngsim: ngsim$1,
  	nGt: nGt$1,
  	ngt: ngt$1,
  	ngtr: ngtr$1,
  	nGtv: nGtv$1,
  	nharr: nharr$1,
  	nhArr: nhArr$1,
  	nhpar: nhpar$1,
  	ni: ni$1,
  	nis: nis$1,
  	nisd: nisd$1,
  	niv: niv$1,
  	NJcy: NJcy$1,
  	njcy: njcy$1,
  	nlarr: nlarr$1,
  	nlArr: nlArr$1,
  	nldr: nldr$1,
  	nlE: nlE$1,
  	nle: nle$1,
  	nleftarrow: nleftarrow$1,
  	nLeftarrow: nLeftarrow$1,
  	nleftrightarrow: nleftrightarrow$1,
  	nLeftrightarrow: nLeftrightarrow$1,
  	nleq: nleq$1,
  	nleqq: nleqq$1,
  	nleqslant: nleqslant$1,
  	nles: nles$1,
  	nless: nless$1,
  	nLl: nLl$1,
  	nlsim: nlsim$1,
  	nLt: nLt$1,
  	nlt: nlt$1,
  	nltri: nltri$1,
  	nltrie: nltrie$1,
  	nLtv: nLtv$1,
  	nmid: nmid$1,
  	NoBreak: NoBreak$1,
  	NonBreakingSpace: NonBreakingSpace$1,
  	nopf: nopf$1,
  	Nopf: Nopf$1,
  	Not: Not$1,
  	not: not$1,
  	NotCongruent: NotCongruent$1,
  	NotCupCap: NotCupCap$1,
  	NotDoubleVerticalBar: NotDoubleVerticalBar$1,
  	NotElement: NotElement$1,
  	NotEqual: NotEqual$1,
  	NotEqualTilde: NotEqualTilde$1,
  	NotExists: NotExists$1,
  	NotGreater: NotGreater$1,
  	NotGreaterEqual: NotGreaterEqual$1,
  	NotGreaterFullEqual: NotGreaterFullEqual$1,
  	NotGreaterGreater: NotGreaterGreater$1,
  	NotGreaterLess: NotGreaterLess$1,
  	NotGreaterSlantEqual: NotGreaterSlantEqual$1,
  	NotGreaterTilde: NotGreaterTilde$1,
  	NotHumpDownHump: NotHumpDownHump$1,
  	NotHumpEqual: NotHumpEqual$1,
  	notin: notin$1,
  	notindot: notindot$1,
  	notinE: notinE$1,
  	notinva: notinva$1,
  	notinvb: notinvb$1,
  	notinvc: notinvc$1,
  	NotLeftTriangleBar: NotLeftTriangleBar$1,
  	NotLeftTriangle: NotLeftTriangle$1,
  	NotLeftTriangleEqual: NotLeftTriangleEqual$1,
  	NotLess: NotLess$1,
  	NotLessEqual: NotLessEqual$1,
  	NotLessGreater: NotLessGreater$1,
  	NotLessLess: NotLessLess$1,
  	NotLessSlantEqual: NotLessSlantEqual$1,
  	NotLessTilde: NotLessTilde$1,
  	NotNestedGreaterGreater: NotNestedGreaterGreater$1,
  	NotNestedLessLess: NotNestedLessLess$1,
  	notni: notni$1,
  	notniva: notniva$1,
  	notnivb: notnivb$1,
  	notnivc: notnivc$1,
  	NotPrecedes: NotPrecedes$1,
  	NotPrecedesEqual: NotPrecedesEqual$1,
  	NotPrecedesSlantEqual: NotPrecedesSlantEqual$1,
  	NotReverseElement: NotReverseElement$1,
  	NotRightTriangleBar: NotRightTriangleBar$1,
  	NotRightTriangle: NotRightTriangle$1,
  	NotRightTriangleEqual: NotRightTriangleEqual$1,
  	NotSquareSubset: NotSquareSubset$1,
  	NotSquareSubsetEqual: NotSquareSubsetEqual$1,
  	NotSquareSuperset: NotSquareSuperset$1,
  	NotSquareSupersetEqual: NotSquareSupersetEqual$1,
  	NotSubset: NotSubset$1,
  	NotSubsetEqual: NotSubsetEqual$1,
  	NotSucceeds: NotSucceeds$1,
  	NotSucceedsEqual: NotSucceedsEqual$1,
  	NotSucceedsSlantEqual: NotSucceedsSlantEqual$1,
  	NotSucceedsTilde: NotSucceedsTilde$1,
  	NotSuperset: NotSuperset$1,
  	NotSupersetEqual: NotSupersetEqual$1,
  	NotTilde: NotTilde$1,
  	NotTildeEqual: NotTildeEqual$1,
  	NotTildeFullEqual: NotTildeFullEqual$1,
  	NotTildeTilde: NotTildeTilde$1,
  	NotVerticalBar: NotVerticalBar$1,
  	nparallel: nparallel$1,
  	npar: npar$1,
  	nparsl: nparsl$1,
  	npart: npart$1,
  	npolint: npolint$1,
  	npr: npr$1,
  	nprcue: nprcue$1,
  	nprec: nprec$1,
  	npreceq: npreceq$1,
  	npre: npre$1,
  	nrarrc: nrarrc$1,
  	nrarr: nrarr$1,
  	nrArr: nrArr$1,
  	nrarrw: nrarrw$1,
  	nrightarrow: nrightarrow$1,
  	nRightarrow: nRightarrow$1,
  	nrtri: nrtri$1,
  	nrtrie: nrtrie$1,
  	nsc: nsc$1,
  	nsccue: nsccue$1,
  	nsce: nsce$1,
  	Nscr: Nscr$1,
  	nscr: nscr$1,
  	nshortmid: nshortmid$1,
  	nshortparallel: nshortparallel$1,
  	nsim: nsim$1,
  	nsime: nsime$1,
  	nsimeq: nsimeq$1,
  	nsmid: nsmid$1,
  	nspar: nspar$1,
  	nsqsube: nsqsube$1,
  	nsqsupe: nsqsupe$1,
  	nsub: nsub$1,
  	nsubE: nsubE$1,
  	nsube: nsube$1,
  	nsubset: nsubset$1,
  	nsubseteq: nsubseteq$1,
  	nsubseteqq: nsubseteqq$1,
  	nsucc: nsucc$1,
  	nsucceq: nsucceq$1,
  	nsup: nsup$1,
  	nsupE: nsupE$1,
  	nsupe: nsupe$1,
  	nsupset: nsupset$1,
  	nsupseteq: nsupseteq$1,
  	nsupseteqq: nsupseteqq$1,
  	ntgl: ntgl$1,
  	Ntilde: Ntilde$1,
  	ntilde: ntilde$1,
  	ntlg: ntlg$1,
  	ntriangleleft: ntriangleleft$1,
  	ntrianglelefteq: ntrianglelefteq$1,
  	ntriangleright: ntriangleright$1,
  	ntrianglerighteq: ntrianglerighteq$1,
  	Nu: Nu$1,
  	nu: nu$1,
  	num: num$1,
  	numero: numero$1,
  	numsp: numsp$1,
  	nvap: nvap$1,
  	nvdash: nvdash$1,
  	nvDash: nvDash$1,
  	nVdash: nVdash$1,
  	nVDash: nVDash$1,
  	nvge: nvge$1,
  	nvgt: nvgt$1,
  	nvHarr: nvHarr$1,
  	nvinfin: nvinfin$1,
  	nvlArr: nvlArr$1,
  	nvle: nvle$1,
  	nvlt: nvlt$1,
  	nvltrie: nvltrie$1,
  	nvrArr: nvrArr$1,
  	nvrtrie: nvrtrie$1,
  	nvsim: nvsim$1,
  	nwarhk: nwarhk$1,
  	nwarr: nwarr$1,
  	nwArr: nwArr$1,
  	nwarrow: nwarrow$1,
  	nwnear: nwnear$1,
  	Oacute: Oacute$1,
  	oacute: oacute$1,
  	oast: oast$1,
  	Ocirc: Ocirc$1,
  	ocirc: ocirc$1,
  	ocir: ocir$1,
  	Ocy: Ocy$1,
  	ocy: ocy$1,
  	odash: odash$1,
  	Odblac: Odblac$1,
  	odblac: odblac$1,
  	odiv: odiv$1,
  	odot: odot$1,
  	odsold: odsold$1,
  	OElig: OElig$1,
  	oelig: oelig$1,
  	ofcir: ofcir$1,
  	Ofr: Ofr$1,
  	ofr: ofr$1,
  	ogon: ogon$1,
  	Ograve: Ograve$1,
  	ograve: ograve$1,
  	ogt: ogt$1,
  	ohbar: ohbar$1,
  	ohm: ohm$1,
  	oint: oint$1,
  	olarr: olarr$1,
  	olcir: olcir$1,
  	olcross: olcross$1,
  	oline: oline$1,
  	olt: olt$1,
  	Omacr: Omacr$1,
  	omacr: omacr$1,
  	Omega: Omega$1,
  	omega: omega$1,
  	Omicron: Omicron$1,
  	omicron: omicron$1,
  	omid: omid$1,
  	ominus: ominus$1,
  	Oopf: Oopf$1,
  	oopf: oopf$1,
  	opar: opar$1,
  	OpenCurlyDoubleQuote: OpenCurlyDoubleQuote$1,
  	OpenCurlyQuote: OpenCurlyQuote$1,
  	operp: operp$1,
  	oplus: oplus$1,
  	orarr: orarr$1,
  	Or: Or$1,
  	or: or$1,
  	ord: ord$1,
  	order: order$1,
  	orderof: orderof$1,
  	ordf: ordf$1,
  	ordm: ordm$1,
  	origof: origof$1,
  	oror: oror$1,
  	orslope: orslope$1,
  	orv: orv$1,
  	oS: oS$1,
  	Oscr: Oscr$1,
  	oscr: oscr$1,
  	Oslash: Oslash$1,
  	oslash: oslash$1,
  	osol: osol$1,
  	Otilde: Otilde$1,
  	otilde: otilde$1,
  	otimesas: otimesas$1,
  	Otimes: Otimes$1,
  	otimes: otimes$1,
  	Ouml: Ouml$1,
  	ouml: ouml$1,
  	ovbar: ovbar$1,
  	OverBar: OverBar$1,
  	OverBrace: OverBrace$1,
  	OverBracket: OverBracket$1,
  	OverParenthesis: OverParenthesis$1,
  	para: para$1,
  	parallel: parallel$1,
  	par: par$1,
  	parsim: parsim$1,
  	parsl: parsl$1,
  	part: part$1,
  	PartialD: PartialD$1,
  	Pcy: Pcy$1,
  	pcy: pcy$1,
  	percnt: percnt$1,
  	period: period$1,
  	permil: permil$1,
  	perp: perp$1,
  	pertenk: pertenk$1,
  	Pfr: Pfr$1,
  	pfr: pfr$1,
  	Phi: Phi$1,
  	phi: phi$1,
  	phiv: phiv$1,
  	phmmat: phmmat$1,
  	phone: phone$3,
  	Pi: Pi$1,
  	pi: pi$1,
  	pitchfork: pitchfork$1,
  	piv: piv$1,
  	planck: planck$1,
  	planckh: planckh$1,
  	plankv: plankv$1,
  	plusacir: plusacir$1,
  	plusb: plusb$1,
  	pluscir: pluscir$1,
  	plus: plus$1,
  	plusdo: plusdo$1,
  	plusdu: plusdu$1,
  	pluse: pluse$1,
  	PlusMinus: PlusMinus$1,
  	plusmn: plusmn$1,
  	plussim: plussim$1,
  	plustwo: plustwo$1,
  	pm: pm$1,
  	Poincareplane: Poincareplane$1,
  	pointint: pointint$1,
  	popf: popf$1,
  	Popf: Popf$1,
  	pound: pound$3,
  	prap: prap$1,
  	Pr: Pr$1,
  	pr: pr$1,
  	prcue: prcue$1,
  	precapprox: precapprox$1,
  	prec: prec$1,
  	preccurlyeq: preccurlyeq$1,
  	Precedes: Precedes$1,
  	PrecedesEqual: PrecedesEqual$1,
  	PrecedesSlantEqual: PrecedesSlantEqual$1,
  	PrecedesTilde: PrecedesTilde$1,
  	preceq: preceq$1,
  	precnapprox: precnapprox$1,
  	precneqq: precneqq$1,
  	precnsim: precnsim$1,
  	pre: pre$1,
  	prE: prE$1,
  	precsim: precsim$1,
  	prime: prime$1,
  	Prime: Prime$1,
  	primes: primes$1,
  	prnap: prnap$1,
  	prnE: prnE$1,
  	prnsim: prnsim$1,
  	prod: prod$1,
  	Product: Product$1,
  	profalar: profalar$1,
  	profline: profline$1,
  	profsurf: profsurf$1,
  	prop: prop$1,
  	Proportional: Proportional$1,
  	Proportion: Proportion$1,
  	propto: propto$1,
  	prsim: prsim$1,
  	prurel: prurel$1,
  	Pscr: Pscr$1,
  	pscr: pscr$1,
  	Psi: Psi$1,
  	psi: psi$1,
  	puncsp: puncsp$1,
  	Qfr: Qfr$1,
  	qfr: qfr$1,
  	qint: qint$1,
  	qopf: qopf$1,
  	Qopf: Qopf$1,
  	qprime: qprime$1,
  	Qscr: Qscr$1,
  	qscr: qscr$1,
  	quaternions: quaternions$1,
  	quatint: quatint$1,
  	quest: quest$1,
  	questeq: questeq$1,
  	quot: quot$1,
  	QUOT: QUOT$1,
  	rAarr: rAarr$1,
  	race: race$1,
  	Racute: Racute$1,
  	racute: racute$1,
  	radic: radic$1,
  	raemptyv: raemptyv$1,
  	rang: rang$1,
  	Rang: Rang$1,
  	rangd: rangd$1,
  	range: range$1,
  	rangle: rangle$1,
  	raquo: raquo$1,
  	rarrap: rarrap$1,
  	rarrb: rarrb$1,
  	rarrbfs: rarrbfs$1,
  	rarrc: rarrc$1,
  	rarr: rarr$1,
  	Rarr: Rarr$1,
  	rArr: rArr$1,
  	rarrfs: rarrfs$1,
  	rarrhk: rarrhk$1,
  	rarrlp: rarrlp$1,
  	rarrpl: rarrpl$1,
  	rarrsim: rarrsim$1,
  	Rarrtl: Rarrtl$1,
  	rarrtl: rarrtl$1,
  	rarrw: rarrw$1,
  	ratail: ratail$1,
  	rAtail: rAtail$1,
  	ratio: ratio$1,
  	rationals: rationals$1,
  	rbarr: rbarr$1,
  	rBarr: rBarr$1,
  	RBarr: RBarr$1,
  	rbbrk: rbbrk$1,
  	rbrace: rbrace$1,
  	rbrack: rbrack$1,
  	rbrke: rbrke$1,
  	rbrksld: rbrksld$1,
  	rbrkslu: rbrkslu$1,
  	Rcaron: Rcaron$1,
  	rcaron: rcaron$1,
  	Rcedil: Rcedil$1,
  	rcedil: rcedil$1,
  	rceil: rceil$1,
  	rcub: rcub$1,
  	Rcy: Rcy$1,
  	rcy: rcy$1,
  	rdca: rdca$1,
  	rdldhar: rdldhar$1,
  	rdquo: rdquo$1,
  	rdquor: rdquor$1,
  	rdsh: rdsh$1,
  	real: real$1,
  	realine: realine$1,
  	realpart: realpart$1,
  	reals: reals$1,
  	Re: Re$1,
  	rect: rect$1,
  	reg: reg$1,
  	REG: REG$1,
  	ReverseElement: ReverseElement$1,
  	ReverseEquilibrium: ReverseEquilibrium$1,
  	ReverseUpEquilibrium: ReverseUpEquilibrium$1,
  	rfisht: rfisht$1,
  	rfloor: rfloor$1,
  	rfr: rfr$1,
  	Rfr: Rfr$1,
  	rHar: rHar$1,
  	rhard: rhard$1,
  	rharu: rharu$1,
  	rharul: rharul$1,
  	Rho: Rho$1,
  	rho: rho$1,
  	rhov: rhov$1,
  	RightAngleBracket: RightAngleBracket$1,
  	RightArrowBar: RightArrowBar$1,
  	rightarrow: rightarrow$1,
  	RightArrow: RightArrow$1,
  	Rightarrow: Rightarrow$1,
  	RightArrowLeftArrow: RightArrowLeftArrow$1,
  	rightarrowtail: rightarrowtail$1,
  	RightCeiling: RightCeiling$1,
  	RightDoubleBracket: RightDoubleBracket$1,
  	RightDownTeeVector: RightDownTeeVector$1,
  	RightDownVectorBar: RightDownVectorBar$1,
  	RightDownVector: RightDownVector$1,
  	RightFloor: RightFloor$1,
  	rightharpoondown: rightharpoondown$1,
  	rightharpoonup: rightharpoonup$1,
  	rightleftarrows: rightleftarrows$1,
  	rightleftharpoons: rightleftharpoons$1,
  	rightrightarrows: rightrightarrows$1,
  	rightsquigarrow: rightsquigarrow$1,
  	RightTeeArrow: RightTeeArrow$1,
  	RightTee: RightTee$1,
  	RightTeeVector: RightTeeVector$1,
  	rightthreetimes: rightthreetimes$1,
  	RightTriangleBar: RightTriangleBar$1,
  	RightTriangle: RightTriangle$1,
  	RightTriangleEqual: RightTriangleEqual$1,
  	RightUpDownVector: RightUpDownVector$1,
  	RightUpTeeVector: RightUpTeeVector$1,
  	RightUpVectorBar: RightUpVectorBar$1,
  	RightUpVector: RightUpVector$1,
  	RightVectorBar: RightVectorBar$1,
  	RightVector: RightVector$1,
  	ring: ring$3,
  	risingdotseq: risingdotseq$1,
  	rlarr: rlarr$1,
  	rlhar: rlhar$1,
  	rlm: rlm$1,
  	rmoustache: rmoustache$1,
  	rmoust: rmoust$1,
  	rnmid: rnmid$1,
  	roang: roang$1,
  	roarr: roarr$1,
  	robrk: robrk$1,
  	ropar: ropar$1,
  	ropf: ropf$1,
  	Ropf: Ropf$1,
  	roplus: roplus$1,
  	rotimes: rotimes$1,
  	RoundImplies: RoundImplies$1,
  	rpar: rpar$1,
  	rpargt: rpargt$1,
  	rppolint: rppolint$1,
  	rrarr: rrarr$1,
  	Rrightarrow: Rrightarrow$1,
  	rsaquo: rsaquo$1,
  	rscr: rscr$1,
  	Rscr: Rscr$1,
  	rsh: rsh$1,
  	Rsh: Rsh$1,
  	rsqb: rsqb$1,
  	rsquo: rsquo$1,
  	rsquor: rsquor$1,
  	rthree: rthree$1,
  	rtimes: rtimes$1,
  	rtri: rtri$1,
  	rtrie: rtrie$1,
  	rtrif: rtrif$1,
  	rtriltri: rtriltri$1,
  	RuleDelayed: RuleDelayed$1,
  	ruluhar: ruluhar$1,
  	rx: rx$1,
  	Sacute: Sacute$1,
  	sacute: sacute$1,
  	sbquo: sbquo$1,
  	scap: scap$1,
  	Scaron: Scaron$1,
  	scaron: scaron$1,
  	Sc: Sc$1,
  	sc: sc$1,
  	sccue: sccue$1,
  	sce: sce$1,
  	scE: scE$1,
  	Scedil: Scedil$1,
  	scedil: scedil$1,
  	Scirc: Scirc$1,
  	scirc: scirc$1,
  	scnap: scnap$1,
  	scnE: scnE$1,
  	scnsim: scnsim$1,
  	scpolint: scpolint$1,
  	scsim: scsim$1,
  	Scy: Scy$1,
  	scy: scy$1,
  	sdotb: sdotb$1,
  	sdot: sdot$1,
  	sdote: sdote$1,
  	searhk: searhk$1,
  	searr: searr$1,
  	seArr: seArr$1,
  	searrow: searrow$1,
  	sect: sect$1,
  	semi: semi$1,
  	seswar: seswar$1,
  	setminus: setminus$1,
  	setmn: setmn$1,
  	sext: sext$1,
  	Sfr: Sfr$1,
  	sfr: sfr$1,
  	sfrown: sfrown$1,
  	sharp: sharp$1,
  	SHCHcy: SHCHcy$1,
  	shchcy: shchcy$1,
  	SHcy: SHcy$1,
  	shcy: shcy$1,
  	ShortDownArrow: ShortDownArrow$1,
  	ShortLeftArrow: ShortLeftArrow$1,
  	shortmid: shortmid$1,
  	shortparallel: shortparallel$1,
  	ShortRightArrow: ShortRightArrow$1,
  	ShortUpArrow: ShortUpArrow$1,
  	shy: shy$1,
  	Sigma: Sigma$1,
  	sigma: sigma$1,
  	sigmaf: sigmaf$1,
  	sigmav: sigmav$1,
  	sim: sim$1,
  	simdot: simdot$1,
  	sime: sime$1,
  	simeq: simeq$1,
  	simg: simg$1,
  	simgE: simgE$1,
  	siml: siml$1,
  	simlE: simlE$1,
  	simne: simne$1,
  	simplus: simplus$1,
  	simrarr: simrarr$1,
  	slarr: slarr$1,
  	SmallCircle: SmallCircle$1,
  	smallsetminus: smallsetminus$1,
  	smashp: smashp$1,
  	smeparsl: smeparsl$1,
  	smid: smid$1,
  	smile: smile$3,
  	smt: smt$1,
  	smte: smte$1,
  	smtes: smtes$1,
  	SOFTcy: SOFTcy$1,
  	softcy: softcy$1,
  	solbar: solbar$1,
  	solb: solb$1,
  	sol: sol$1,
  	Sopf: Sopf$1,
  	sopf: sopf$1,
  	spades: spades$3,
  	spadesuit: spadesuit$1,
  	spar: spar$1,
  	sqcap: sqcap$1,
  	sqcaps: sqcaps$1,
  	sqcup: sqcup$1,
  	sqcups: sqcups$1,
  	Sqrt: Sqrt$1,
  	sqsub: sqsub$1,
  	sqsube: sqsube$1,
  	sqsubset: sqsubset$1,
  	sqsubseteq: sqsubseteq$1,
  	sqsup: sqsup$1,
  	sqsupe: sqsupe$1,
  	sqsupset: sqsupset$1,
  	sqsupseteq: sqsupseteq$1,
  	square: square$1,
  	Square: Square$1,
  	SquareIntersection: SquareIntersection$1,
  	SquareSubset: SquareSubset$1,
  	SquareSubsetEqual: SquareSubsetEqual$1,
  	SquareSuperset: SquareSuperset$1,
  	SquareSupersetEqual: SquareSupersetEqual$1,
  	SquareUnion: SquareUnion$1,
  	squarf: squarf$1,
  	squ: squ$1,
  	squf: squf$1,
  	srarr: srarr$1,
  	Sscr: Sscr$1,
  	sscr: sscr$1,
  	ssetmn: ssetmn$1,
  	ssmile: ssmile$1,
  	sstarf: sstarf$1,
  	Star: Star$1,
  	star: star$3,
  	starf: starf$1,
  	straightepsilon: straightepsilon$1,
  	straightphi: straightphi$1,
  	strns: strns$1,
  	sub: sub$1,
  	Sub: Sub$1,
  	subdot: subdot$1,
  	subE: subE$1,
  	sube: sube$1,
  	subedot: subedot$1,
  	submult: submult$1,
  	subnE: subnE$1,
  	subne: subne$1,
  	subplus: subplus$1,
  	subrarr: subrarr$1,
  	subset: subset$1,
  	Subset: Subset$1,
  	subseteq: subseteq$1,
  	subseteqq: subseteqq$1,
  	SubsetEqual: SubsetEqual$1,
  	subsetneq: subsetneq$1,
  	subsetneqq: subsetneqq$1,
  	subsim: subsim$1,
  	subsub: subsub$1,
  	subsup: subsup$1,
  	succapprox: succapprox$1,
  	succ: succ$1,
  	succcurlyeq: succcurlyeq$1,
  	Succeeds: Succeeds$1,
  	SucceedsEqual: SucceedsEqual$1,
  	SucceedsSlantEqual: SucceedsSlantEqual$1,
  	SucceedsTilde: SucceedsTilde$1,
  	succeq: succeq$1,
  	succnapprox: succnapprox$1,
  	succneqq: succneqq$1,
  	succnsim: succnsim$1,
  	succsim: succsim$1,
  	SuchThat: SuchThat$1,
  	sum: sum$1,
  	Sum: Sum$1,
  	sung: sung$1,
  	sup1: sup1$1,
  	sup2: sup2$1,
  	sup3: sup3$1,
  	sup: sup$1,
  	Sup: Sup$1,
  	supdot: supdot$1,
  	supdsub: supdsub$1,
  	supE: supE$1,
  	supe: supe$1,
  	supedot: supedot$1,
  	Superset: Superset$1,
  	SupersetEqual: SupersetEqual$1,
  	suphsol: suphsol$1,
  	suphsub: suphsub$1,
  	suplarr: suplarr$1,
  	supmult: supmult$1,
  	supnE: supnE$1,
  	supne: supne$1,
  	supplus: supplus$1,
  	supset: supset$1,
  	Supset: Supset$1,
  	supseteq: supseteq$1,
  	supseteqq: supseteqq$1,
  	supsetneq: supsetneq$1,
  	supsetneqq: supsetneqq$1,
  	supsim: supsim$1,
  	supsub: supsub$1,
  	supsup: supsup$1,
  	swarhk: swarhk$1,
  	swarr: swarr$1,
  	swArr: swArr$1,
  	swarrow: swarrow$1,
  	swnwar: swnwar$1,
  	szlig: szlig$1,
  	Tab: Tab$1,
  	target: target$1,
  	Tau: Tau$1,
  	tau: tau$1,
  	tbrk: tbrk$1,
  	Tcaron: Tcaron$1,
  	tcaron: tcaron$1,
  	Tcedil: Tcedil$1,
  	tcedil: tcedil$1,
  	Tcy: Tcy$1,
  	tcy: tcy$1,
  	tdot: tdot$1,
  	telrec: telrec$1,
  	Tfr: Tfr$1,
  	tfr: tfr$1,
  	there4: there4$1,
  	therefore: therefore$1,
  	Therefore: Therefore$1,
  	Theta: Theta$1,
  	theta: theta$1,
  	thetasym: thetasym$1,
  	thetav: thetav$1,
  	thickapprox: thickapprox$1,
  	thicksim: thicksim$1,
  	ThickSpace: ThickSpace$1,
  	ThinSpace: ThinSpace$1,
  	thinsp: thinsp$1,
  	thkap: thkap$1,
  	thksim: thksim$1,
  	THORN: THORN$1,
  	thorn: thorn$1,
  	tilde: tilde$1,
  	Tilde: Tilde$1,
  	TildeEqual: TildeEqual$1,
  	TildeFullEqual: TildeFullEqual$1,
  	TildeTilde: TildeTilde$1,
  	timesbar: timesbar$1,
  	timesb: timesb$1,
  	times: times$1,
  	timesd: timesd$1,
  	tint: tint$1,
  	toea: toea$1,
  	topbot: topbot$1,
  	topcir: topcir$1,
  	top: top$3,
  	Topf: Topf$1,
  	topf: topf$1,
  	topfork: topfork$1,
  	tosa: tosa$1,
  	tprime: tprime$1,
  	trade: trade$1,
  	TRADE: TRADE$1,
  	triangle: triangle$1,
  	triangledown: triangledown$1,
  	triangleleft: triangleleft$1,
  	trianglelefteq: trianglelefteq$1,
  	triangleq: triangleq$1,
  	triangleright: triangleright$1,
  	trianglerighteq: trianglerighteq$1,
  	tridot: tridot$1,
  	trie: trie$1,
  	triminus: triminus$1,
  	TripleDot: TripleDot$1,
  	triplus: triplus$1,
  	trisb: trisb$1,
  	tritime: tritime$1,
  	trpezium: trpezium$1,
  	Tscr: Tscr$1,
  	tscr: tscr$1,
  	TScy: TScy$1,
  	tscy: tscy$1,
  	TSHcy: TSHcy$1,
  	tshcy: tshcy$1,
  	Tstrok: Tstrok$1,
  	tstrok: tstrok$1,
  	twixt: twixt$1,
  	twoheadleftarrow: twoheadleftarrow$1,
  	twoheadrightarrow: twoheadrightarrow$1,
  	Uacute: Uacute$1,
  	uacute: uacute$1,
  	uarr: uarr$1,
  	Uarr: Uarr$1,
  	uArr: uArr$1,
  	Uarrocir: Uarrocir$1,
  	Ubrcy: Ubrcy$1,
  	ubrcy: ubrcy$1,
  	Ubreve: Ubreve$1,
  	ubreve: ubreve$1,
  	Ucirc: Ucirc$1,
  	ucirc: ucirc$1,
  	Ucy: Ucy$1,
  	ucy: ucy$1,
  	udarr: udarr$1,
  	Udblac: Udblac$1,
  	udblac: udblac$1,
  	udhar: udhar$1,
  	ufisht: ufisht$1,
  	Ufr: Ufr$1,
  	ufr: ufr$1,
  	Ugrave: Ugrave$1,
  	ugrave: ugrave$1,
  	uHar: uHar$1,
  	uharl: uharl$1,
  	uharr: uharr$1,
  	uhblk: uhblk$1,
  	ulcorn: ulcorn$1,
  	ulcorner: ulcorner$1,
  	ulcrop: ulcrop$1,
  	ultri: ultri$1,
  	Umacr: Umacr$1,
  	umacr: umacr$1,
  	uml: uml$1,
  	UnderBar: UnderBar$1,
  	UnderBrace: UnderBrace$1,
  	UnderBracket: UnderBracket$1,
  	UnderParenthesis: UnderParenthesis$1,
  	Union: Union$1,
  	UnionPlus: UnionPlus$1,
  	Uogon: Uogon$1,
  	uogon: uogon$1,
  	Uopf: Uopf$1,
  	uopf: uopf$1,
  	UpArrowBar: UpArrowBar$1,
  	uparrow: uparrow$1,
  	UpArrow: UpArrow$1,
  	Uparrow: Uparrow$1,
  	UpArrowDownArrow: UpArrowDownArrow$1,
  	updownarrow: updownarrow$1,
  	UpDownArrow: UpDownArrow$1,
  	Updownarrow: Updownarrow$1,
  	UpEquilibrium: UpEquilibrium$1,
  	upharpoonleft: upharpoonleft$1,
  	upharpoonright: upharpoonright$1,
  	uplus: uplus$1,
  	UpperLeftArrow: UpperLeftArrow$1,
  	UpperRightArrow: UpperRightArrow$1,
  	upsi: upsi$1,
  	Upsi: Upsi$1,
  	upsih: upsih$1,
  	Upsilon: Upsilon$1,
  	upsilon: upsilon$1,
  	UpTeeArrow: UpTeeArrow$1,
  	UpTee: UpTee$1,
  	upuparrows: upuparrows$1,
  	urcorn: urcorn$1,
  	urcorner: urcorner$1,
  	urcrop: urcrop$1,
  	Uring: Uring$1,
  	uring: uring$1,
  	urtri: urtri$1,
  	Uscr: Uscr$1,
  	uscr: uscr$1,
  	utdot: utdot$1,
  	Utilde: Utilde$1,
  	utilde: utilde$1,
  	utri: utri$1,
  	utrif: utrif$1,
  	uuarr: uuarr$1,
  	Uuml: Uuml$1,
  	uuml: uuml$1,
  	uwangle: uwangle$1,
  	vangrt: vangrt$1,
  	varepsilon: varepsilon$1,
  	varkappa: varkappa$1,
  	varnothing: varnothing$1,
  	varphi: varphi$1,
  	varpi: varpi$1,
  	varpropto: varpropto$1,
  	varr: varr$1,
  	vArr: vArr$1,
  	varrho: varrho$1,
  	varsigma: varsigma$1,
  	varsubsetneq: varsubsetneq$1,
  	varsubsetneqq: varsubsetneqq$1,
  	varsupsetneq: varsupsetneq$1,
  	varsupsetneqq: varsupsetneqq$1,
  	vartheta: vartheta$1,
  	vartriangleleft: vartriangleleft$1,
  	vartriangleright: vartriangleright$1,
  	vBar: vBar$1,
  	Vbar: Vbar$1,
  	vBarv: vBarv$1,
  	Vcy: Vcy$1,
  	vcy: vcy$1,
  	vdash: vdash$1,
  	vDash: vDash$1,
  	Vdash: Vdash$1,
  	VDash: VDash$1,
  	Vdashl: Vdashl$1,
  	veebar: veebar$1,
  	vee: vee$1,
  	Vee: Vee$1,
  	veeeq: veeeq$1,
  	vellip: vellip$1,
  	verbar: verbar$1,
  	Verbar: Verbar$1,
  	vert: vert$1,
  	Vert: Vert$1,
  	VerticalBar: VerticalBar$1,
  	VerticalLine: VerticalLine$1,
  	VerticalSeparator: VerticalSeparator$1,
  	VerticalTilde: VerticalTilde$1,
  	VeryThinSpace: VeryThinSpace$1,
  	Vfr: Vfr$1,
  	vfr: vfr$1,
  	vltri: vltri$1,
  	vnsub: vnsub$1,
  	vnsup: vnsup$1,
  	Vopf: Vopf$1,
  	vopf: vopf$1,
  	vprop: vprop$1,
  	vrtri: vrtri$1,
  	Vscr: Vscr$1,
  	vscr: vscr$1,
  	vsubnE: vsubnE$1,
  	vsubne: vsubne$1,
  	vsupnE: vsupnE$1,
  	vsupne: vsupne$1,
  	Vvdash: Vvdash$1,
  	vzigzag: vzigzag$1,
  	Wcirc: Wcirc$1,
  	wcirc: wcirc$1,
  	wedbar: wedbar$1,
  	wedge: wedge$1,
  	Wedge: Wedge$1,
  	wedgeq: wedgeq$1,
  	weierp: weierp$1,
  	Wfr: Wfr$1,
  	wfr: wfr$1,
  	Wopf: Wopf$1,
  	wopf: wopf$1,
  	wp: wp$1,
  	wr: wr$1,
  	wreath: wreath$1,
  	Wscr: Wscr$1,
  	wscr: wscr$1,
  	xcap: xcap$1,
  	xcirc: xcirc$1,
  	xcup: xcup$1,
  	xdtri: xdtri$1,
  	Xfr: Xfr$1,
  	xfr: xfr$1,
  	xharr: xharr$1,
  	xhArr: xhArr$1,
  	Xi: Xi$1,
  	xi: xi$1,
  	xlarr: xlarr$1,
  	xlArr: xlArr$1,
  	xmap: xmap$1,
  	xnis: xnis$1,
  	xodot: xodot$1,
  	Xopf: Xopf$1,
  	xopf: xopf$1,
  	xoplus: xoplus$1,
  	xotime: xotime$1,
  	xrarr: xrarr$1,
  	xrArr: xrArr$1,
  	Xscr: Xscr$1,
  	xscr: xscr$1,
  	xsqcup: xsqcup$1,
  	xuplus: xuplus$1,
  	xutri: xutri$1,
  	xvee: xvee$1,
  	xwedge: xwedge$1,
  	Yacute: Yacute$1,
  	yacute: yacute$1,
  	YAcy: YAcy$1,
  	yacy: yacy$1,
  	Ycirc: Ycirc$1,
  	ycirc: ycirc$1,
  	Ycy: Ycy$1,
  	ycy: ycy$1,
  	yen: yen$3,
  	Yfr: Yfr$1,
  	yfr: yfr$1,
  	YIcy: YIcy$1,
  	yicy: yicy$1,
  	Yopf: Yopf$1,
  	yopf: yopf$1,
  	Yscr: Yscr$1,
  	yscr: yscr$1,
  	YUcy: YUcy$1,
  	yucy: yucy$1,
  	yuml: yuml$1,
  	Yuml: Yuml$1,
  	Zacute: Zacute$1,
  	zacute: zacute$1,
  	Zcaron: Zcaron$1,
  	zcaron: zcaron$1,
  	Zcy: Zcy$1,
  	zcy: zcy$1,
  	Zdot: Zdot$1,
  	zdot: zdot$1,
  	zeetrf: zeetrf$1,
  	ZeroWidthSpace: ZeroWidthSpace$1,
  	Zeta: Zeta$1,
  	zeta: zeta$1,
  	zfr: zfr$1,
  	Zfr: Zfr$1,
  	ZHcy: ZHcy$1,
  	zhcy: zhcy$1,
  	zigrarr: zigrarr$1,
  	zopf: zopf$1,
  	Zopf: Zopf$1,
  	Zscr: Zscr$1,
  	zscr: zscr$1,
  	zwj: zwj$1,
  	zwnj: zwnj$1
  };

  var entities$5 = /*#__PURE__*/Object.freeze({
    __proto__: null,
    Aacute: Aacute$1,
    aacute: aacute$1,
    Abreve: Abreve$1,
    abreve: abreve$1,
    ac: ac$1,
    acd: acd$1,
    acE: acE$1,
    Acirc: Acirc$1,
    acirc: acirc$1,
    acute: acute$1,
    Acy: Acy$1,
    acy: acy$1,
    AElig: AElig$1,
    aelig: aelig$1,
    af: af$1,
    Afr: Afr$1,
    afr: afr$1,
    Agrave: Agrave$1,
    agrave: agrave$1,
    alefsym: alefsym$1,
    aleph: aleph$1,
    Alpha: Alpha$1,
    alpha: alpha$1,
    Amacr: Amacr$1,
    amacr: amacr$1,
    amalg: amalg$1,
    amp: amp$1,
    AMP: AMP$1,
    andand: andand$1,
    And: And$1,
    and: and$1,
    andd: andd$1,
    andslope: andslope$1,
    andv: andv$1,
    ang: ang$1,
    ange: ange$1,
    angle: angle$1,
    angmsdaa: angmsdaa$1,
    angmsdab: angmsdab$1,
    angmsdac: angmsdac$1,
    angmsdad: angmsdad$1,
    angmsdae: angmsdae$1,
    angmsdaf: angmsdaf$1,
    angmsdag: angmsdag$1,
    angmsdah: angmsdah$1,
    angmsd: angmsd$1,
    angrt: angrt$1,
    angrtvb: angrtvb$1,
    angrtvbd: angrtvbd$1,
    angsph: angsph$1,
    angst: angst$1,
    angzarr: angzarr$1,
    Aogon: Aogon$1,
    aogon: aogon$1,
    Aopf: Aopf$1,
    aopf: aopf$1,
    apacir: apacir$1,
    ap: ap$1,
    apE: apE$1,
    ape: ape$1,
    apid: apid$1,
    apos: apos$1,
    ApplyFunction: ApplyFunction$1,
    approx: approx$1,
    approxeq: approxeq$1,
    Aring: Aring$1,
    aring: aring$1,
    Ascr: Ascr$1,
    ascr: ascr$1,
    Assign: Assign$1,
    ast: ast$1,
    asymp: asymp$1,
    asympeq: asympeq$1,
    Atilde: Atilde$1,
    atilde: atilde$1,
    Auml: Auml$1,
    auml: auml$1,
    awconint: awconint$1,
    awint: awint$1,
    backcong: backcong$1,
    backepsilon: backepsilon$1,
    backprime: backprime$1,
    backsim: backsim$1,
    backsimeq: backsimeq$1,
    Backslash: Backslash$1,
    Barv: Barv$1,
    barvee: barvee$1,
    barwed: barwed$1,
    Barwed: Barwed$1,
    barwedge: barwedge$1,
    bbrk: bbrk$1,
    bbrktbrk: bbrktbrk$1,
    bcong: bcong$1,
    Bcy: Bcy$1,
    bcy: bcy$1,
    bdquo: bdquo$1,
    becaus: becaus$1,
    because: because$1,
    Because: Because$1,
    bemptyv: bemptyv$1,
    bepsi: bepsi$1,
    bernou: bernou$1,
    Bernoullis: Bernoullis$1,
    Beta: Beta$1,
    beta: beta$1,
    beth: beth$1,
    between: between$1,
    Bfr: Bfr$1,
    bfr: bfr$1,
    bigcap: bigcap$1,
    bigcirc: bigcirc$1,
    bigcup: bigcup$1,
    bigodot: bigodot$1,
    bigoplus: bigoplus$1,
    bigotimes: bigotimes$1,
    bigsqcup: bigsqcup$1,
    bigstar: bigstar$1,
    bigtriangledown: bigtriangledown$1,
    bigtriangleup: bigtriangleup$1,
    biguplus: biguplus$1,
    bigvee: bigvee$1,
    bigwedge: bigwedge$1,
    bkarow: bkarow$1,
    blacklozenge: blacklozenge$1,
    blacksquare: blacksquare$1,
    blacktriangle: blacktriangle$1,
    blacktriangledown: blacktriangledown$1,
    blacktriangleleft: blacktriangleleft$1,
    blacktriangleright: blacktriangleright$1,
    blank: blank$1,
    blk12: blk12$1,
    blk14: blk14$1,
    blk34: blk34$1,
    block: block$3,
    bne: bne$1,
    bnequiv: bnequiv$1,
    bNot: bNot$1,
    bnot: bnot$1,
    Bopf: Bopf$1,
    bopf: bopf$1,
    bot: bot$1,
    bottom: bottom$1,
    bowtie: bowtie$1,
    boxbox: boxbox$1,
    boxdl: boxdl$1,
    boxdL: boxdL$1,
    boxDl: boxDl$1,
    boxDL: boxDL$1,
    boxdr: boxdr$1,
    boxdR: boxdR$1,
    boxDr: boxDr$1,
    boxDR: boxDR$1,
    boxh: boxh$1,
    boxH: boxH$1,
    boxhd: boxhd$1,
    boxHd: boxHd$1,
    boxhD: boxhD$1,
    boxHD: boxHD$1,
    boxhu: boxhu$1,
    boxHu: boxHu$1,
    boxhU: boxhU$1,
    boxHU: boxHU$1,
    boxminus: boxminus$1,
    boxplus: boxplus$1,
    boxtimes: boxtimes$1,
    boxul: boxul$1,
    boxuL: boxuL$1,
    boxUl: boxUl$1,
    boxUL: boxUL$1,
    boxur: boxur$1,
    boxuR: boxuR$1,
    boxUr: boxUr$1,
    boxUR: boxUR$1,
    boxv: boxv$1,
    boxV: boxV$1,
    boxvh: boxvh$1,
    boxvH: boxvH$1,
    boxVh: boxVh$1,
    boxVH: boxVH$1,
    boxvl: boxvl$1,
    boxvL: boxvL$1,
    boxVl: boxVl$1,
    boxVL: boxVL$1,
    boxvr: boxvr$1,
    boxvR: boxvR$1,
    boxVr: boxVr$1,
    boxVR: boxVR$1,
    bprime: bprime$1,
    breve: breve$1,
    Breve: Breve$1,
    brvbar: brvbar$1,
    bscr: bscr$1,
    Bscr: Bscr$1,
    bsemi: bsemi$1,
    bsim: bsim$1,
    bsime: bsime$1,
    bsolb: bsolb$1,
    bsol: bsol$1,
    bsolhsub: bsolhsub$1,
    bull: bull$1,
    bullet: bullet$1,
    bump: bump$1,
    bumpE: bumpE$1,
    bumpe: bumpe$1,
    Bumpeq: Bumpeq$1,
    bumpeq: bumpeq$1,
    Cacute: Cacute$1,
    cacute: cacute$1,
    capand: capand$1,
    capbrcup: capbrcup$1,
    capcap: capcap$1,
    cap: cap$1,
    Cap: Cap$1,
    capcup: capcup$1,
    capdot: capdot$1,
    CapitalDifferentialD: CapitalDifferentialD$1,
    caps: caps$1,
    caret: caret$1,
    caron: caron$1,
    Cayleys: Cayleys$1,
    ccaps: ccaps$1,
    Ccaron: Ccaron$1,
    ccaron: ccaron$1,
    Ccedil: Ccedil$1,
    ccedil: ccedil$1,
    Ccirc: Ccirc$1,
    ccirc: ccirc$1,
    Cconint: Cconint$1,
    ccups: ccups$1,
    ccupssm: ccupssm$1,
    Cdot: Cdot$1,
    cdot: cdot$1,
    cedil: cedil$1,
    Cedilla: Cedilla$1,
    cemptyv: cemptyv$1,
    cent: cent$1,
    centerdot: centerdot$1,
    CenterDot: CenterDot$1,
    cfr: cfr$1,
    Cfr: Cfr$1,
    CHcy: CHcy$1,
    chcy: chcy$1,
    check: check$1,
    checkmark: checkmark$1,
    Chi: Chi$1,
    chi: chi$1,
    circ: circ$1,
    circeq: circeq$1,
    circlearrowleft: circlearrowleft$1,
    circlearrowright: circlearrowright$1,
    circledast: circledast$1,
    circledcirc: circledcirc$1,
    circleddash: circleddash$1,
    CircleDot: CircleDot$1,
    circledR: circledR$1,
    circledS: circledS$1,
    CircleMinus: CircleMinus$1,
    CirclePlus: CirclePlus$1,
    CircleTimes: CircleTimes$1,
    cir: cir$1,
    cirE: cirE$1,
    cire: cire$1,
    cirfnint: cirfnint$1,
    cirmid: cirmid$1,
    cirscir: cirscir$1,
    ClockwiseContourIntegral: ClockwiseContourIntegral$1,
    CloseCurlyDoubleQuote: CloseCurlyDoubleQuote$1,
    CloseCurlyQuote: CloseCurlyQuote$1,
    clubs: clubs$3,
    clubsuit: clubsuit$1,
    colon: colon$1,
    Colon: Colon$1,
    Colone: Colone$1,
    colone: colone$1,
    coloneq: coloneq$1,
    comma: comma$1,
    commat: commat$1,
    comp: comp$1,
    compfn: compfn$1,
    complement: complement$1,
    complexes: complexes$1,
    cong: cong$1,
    congdot: congdot$1,
    Congruent: Congruent$1,
    conint: conint$1,
    Conint: Conint$1,
    ContourIntegral: ContourIntegral$1,
    copf: copf$1,
    Copf: Copf$1,
    coprod: coprod$1,
    Coproduct: Coproduct$1,
    copy: copy$1,
    COPY: COPY$1,
    copysr: copysr$1,
    CounterClockwiseContourIntegral: CounterClockwiseContourIntegral$1,
    crarr: crarr$1,
    cross: cross$1,
    Cross: Cross$1,
    Cscr: Cscr$1,
    cscr: cscr$1,
    csub: csub$1,
    csube: csube$1,
    csup: csup$1,
    csupe: csupe$1,
    ctdot: ctdot$1,
    cudarrl: cudarrl$1,
    cudarrr: cudarrr$1,
    cuepr: cuepr$1,
    cuesc: cuesc$1,
    cularr: cularr$1,
    cularrp: cularrp$1,
    cupbrcap: cupbrcap$1,
    cupcap: cupcap$1,
    CupCap: CupCap$1,
    cup: cup$1,
    Cup: Cup$1,
    cupcup: cupcup$1,
    cupdot: cupdot$1,
    cupor: cupor$1,
    cups: cups$1,
    curarr: curarr$1,
    curarrm: curarrm$1,
    curlyeqprec: curlyeqprec$1,
    curlyeqsucc: curlyeqsucc$1,
    curlyvee: curlyvee$1,
    curlywedge: curlywedge$1,
    curren: curren$1,
    curvearrowleft: curvearrowleft$1,
    curvearrowright: curvearrowright$1,
    cuvee: cuvee$1,
    cuwed: cuwed$1,
    cwconint: cwconint$1,
    cwint: cwint$1,
    cylcty: cylcty$1,
    dagger: dagger$3,
    Dagger: Dagger$1,
    daleth: daleth$1,
    darr: darr$1,
    Darr: Darr$1,
    dArr: dArr$1,
    dash: dash$3,
    Dashv: Dashv$1,
    dashv: dashv$1,
    dbkarow: dbkarow$1,
    dblac: dblac$1,
    Dcaron: Dcaron$1,
    dcaron: dcaron$1,
    Dcy: Dcy$1,
    dcy: dcy$1,
    ddagger: ddagger$1,
    ddarr: ddarr$1,
    DD: DD$1,
    dd: dd$1,
    DDotrahd: DDotrahd$1,
    ddotseq: ddotseq$1,
    deg: deg$1,
    Del: Del$1,
    Delta: Delta$1,
    delta: delta$1,
    demptyv: demptyv$1,
    dfisht: dfisht$1,
    Dfr: Dfr$1,
    dfr: dfr$1,
    dHar: dHar$1,
    dharl: dharl$1,
    dharr: dharr$1,
    DiacriticalAcute: DiacriticalAcute$1,
    DiacriticalDot: DiacriticalDot$1,
    DiacriticalDoubleAcute: DiacriticalDoubleAcute$1,
    DiacriticalGrave: DiacriticalGrave$1,
    DiacriticalTilde: DiacriticalTilde$1,
    diam: diam$1,
    diamond: diamond$1,
    Diamond: Diamond$1,
    diamondsuit: diamondsuit$1,
    diams: diams$1,
    die: die$1,
    DifferentialD: DifferentialD$1,
    digamma: digamma$1,
    disin: disin$1,
    div: div$1,
    divide: divide$1,
    divideontimes: divideontimes$1,
    divonx: divonx$1,
    DJcy: DJcy$1,
    djcy: djcy$1,
    dlcorn: dlcorn$1,
    dlcrop: dlcrop$1,
    dollar: dollar$3,
    Dopf: Dopf$1,
    dopf: dopf$1,
    Dot: Dot$1,
    dot: dot$1,
    DotDot: DotDot$1,
    doteq: doteq$1,
    doteqdot: doteqdot$1,
    DotEqual: DotEqual$1,
    dotminus: dotminus$1,
    dotplus: dotplus$1,
    dotsquare: dotsquare$1,
    doublebarwedge: doublebarwedge$1,
    DoubleContourIntegral: DoubleContourIntegral$1,
    DoubleDot: DoubleDot$1,
    DoubleDownArrow: DoubleDownArrow$1,
    DoubleLeftArrow: DoubleLeftArrow$1,
    DoubleLeftRightArrow: DoubleLeftRightArrow$1,
    DoubleLeftTee: DoubleLeftTee$1,
    DoubleLongLeftArrow: DoubleLongLeftArrow$1,
    DoubleLongLeftRightArrow: DoubleLongLeftRightArrow$1,
    DoubleLongRightArrow: DoubleLongRightArrow$1,
    DoubleRightArrow: DoubleRightArrow$1,
    DoubleRightTee: DoubleRightTee$1,
    DoubleUpArrow: DoubleUpArrow$1,
    DoubleUpDownArrow: DoubleUpDownArrow$1,
    DoubleVerticalBar: DoubleVerticalBar$1,
    DownArrowBar: DownArrowBar$1,
    downarrow: downarrow$1,
    DownArrow: DownArrow$1,
    Downarrow: Downarrow$1,
    DownArrowUpArrow: DownArrowUpArrow$1,
    DownBreve: DownBreve$1,
    downdownarrows: downdownarrows$1,
    downharpoonleft: downharpoonleft$1,
    downharpoonright: downharpoonright$1,
    DownLeftRightVector: DownLeftRightVector$1,
    DownLeftTeeVector: DownLeftTeeVector$1,
    DownLeftVectorBar: DownLeftVectorBar$1,
    DownLeftVector: DownLeftVector$1,
    DownRightTeeVector: DownRightTeeVector$1,
    DownRightVectorBar: DownRightVectorBar$1,
    DownRightVector: DownRightVector$1,
    DownTeeArrow: DownTeeArrow$1,
    DownTee: DownTee$1,
    drbkarow: drbkarow$1,
    drcorn: drcorn$1,
    drcrop: drcrop$1,
    Dscr: Dscr$1,
    dscr: dscr$1,
    DScy: DScy$1,
    dscy: dscy$1,
    dsol: dsol$1,
    Dstrok: Dstrok$1,
    dstrok: dstrok$1,
    dtdot: dtdot$1,
    dtri: dtri$1,
    dtrif: dtrif$1,
    duarr: duarr$1,
    duhar: duhar$1,
    dwangle: dwangle$1,
    DZcy: DZcy$1,
    dzcy: dzcy$1,
    dzigrarr: dzigrarr$1,
    Eacute: Eacute$1,
    eacute: eacute$1,
    easter: easter$1,
    Ecaron: Ecaron$1,
    ecaron: ecaron$1,
    Ecirc: Ecirc$1,
    ecirc: ecirc$1,
    ecir: ecir$1,
    ecolon: ecolon$1,
    Ecy: Ecy$1,
    ecy: ecy$1,
    eDDot: eDDot$1,
    Edot: Edot$1,
    edot: edot$1,
    eDot: eDot$1,
    ee: ee$1,
    efDot: efDot$1,
    Efr: Efr$1,
    efr: efr$1,
    eg: eg$1,
    Egrave: Egrave$1,
    egrave: egrave$1,
    egs: egs$1,
    egsdot: egsdot$1,
    el: el$1,
    Element: Element$1,
    elinters: elinters$1,
    ell: ell$1,
    els: els$1,
    elsdot: elsdot$1,
    Emacr: Emacr$1,
    emacr: emacr$1,
    empty: empty$1,
    emptyset: emptyset$1,
    EmptySmallSquare: EmptySmallSquare$1,
    emptyv: emptyv$1,
    EmptyVerySmallSquare: EmptyVerySmallSquare$1,
    emsp13: emsp13$1,
    emsp14: emsp14$1,
    emsp: emsp$1,
    ENG: ENG$1,
    eng: eng$1,
    ensp: ensp$1,
    Eogon: Eogon$1,
    eogon: eogon$1,
    Eopf: Eopf$1,
    eopf: eopf$1,
    epar: epar$1,
    eparsl: eparsl$1,
    eplus: eplus$1,
    epsi: epsi$1,
    Epsilon: Epsilon$1,
    epsilon: epsilon$1,
    epsiv: epsiv$1,
    eqcirc: eqcirc$1,
    eqcolon: eqcolon$1,
    eqsim: eqsim$1,
    eqslantgtr: eqslantgtr$1,
    eqslantless: eqslantless$1,
    Equal: Equal$1,
    equals: equals$1,
    EqualTilde: EqualTilde$1,
    equest: equest$1,
    Equilibrium: Equilibrium$1,
    equiv: equiv$1,
    equivDD: equivDD$1,
    eqvparsl: eqvparsl$1,
    erarr: erarr$1,
    erDot: erDot$1,
    escr: escr$1,
    Escr: Escr$1,
    esdot: esdot$1,
    Esim: Esim$1,
    esim: esim$1,
    Eta: Eta$1,
    eta: eta$1,
    ETH: ETH$1,
    eth: eth$1,
    Euml: Euml$1,
    euml: euml$1,
    euro: euro$3,
    excl: excl$1,
    exist: exist$1,
    Exists: Exists$1,
    expectation: expectation$1,
    exponentiale: exponentiale$1,
    ExponentialE: ExponentialE$1,
    fallingdotseq: fallingdotseq$1,
    Fcy: Fcy$1,
    fcy: fcy$1,
    female: female$1,
    ffilig: ffilig$1,
    fflig: fflig$1,
    ffllig: ffllig$1,
    Ffr: Ffr$1,
    ffr: ffr$1,
    filig: filig$1,
    FilledSmallSquare: FilledSmallSquare$1,
    FilledVerySmallSquare: FilledVerySmallSquare$1,
    fjlig: fjlig$1,
    flat: flat$1,
    fllig: fllig$1,
    fltns: fltns$1,
    fnof: fnof$1,
    Fopf: Fopf$1,
    fopf: fopf$1,
    forall: forall$1,
    ForAll: ForAll$1,
    fork: fork$1,
    forkv: forkv$1,
    Fouriertrf: Fouriertrf$1,
    fpartint: fpartint$1,
    frac12: frac12$1,
    frac13: frac13$1,
    frac14: frac14$1,
    frac15: frac15$1,
    frac16: frac16$1,
    frac18: frac18$1,
    frac23: frac23$1,
    frac25: frac25$1,
    frac34: frac34$1,
    frac35: frac35$1,
    frac38: frac38$1,
    frac45: frac45$1,
    frac56: frac56$1,
    frac58: frac58$1,
    frac78: frac78$1,
    frasl: frasl$1,
    frown: frown$1,
    fscr: fscr$1,
    Fscr: Fscr$1,
    gacute: gacute$1,
    Gamma: Gamma$1,
    gamma: gamma$1,
    Gammad: Gammad$1,
    gammad: gammad$1,
    gap: gap$1,
    Gbreve: Gbreve$1,
    gbreve: gbreve$1,
    Gcedil: Gcedil$1,
    Gcirc: Gcirc$1,
    gcirc: gcirc$1,
    Gcy: Gcy$1,
    gcy: gcy$1,
    Gdot: Gdot$1,
    gdot: gdot$1,
    ge: ge$1,
    gE: gE$1,
    gEl: gEl$1,
    gel: gel$1,
    geq: geq$1,
    geqq: geqq$1,
    geqslant: geqslant$1,
    gescc: gescc$1,
    ges: ges$1,
    gesdot: gesdot$1,
    gesdoto: gesdoto$1,
    gesdotol: gesdotol$1,
    gesl: gesl$1,
    gesles: gesles$1,
    Gfr: Gfr$1,
    gfr: gfr$1,
    gg: gg$1,
    Gg: Gg$1,
    ggg: ggg$1,
    gimel: gimel$1,
    GJcy: GJcy$1,
    gjcy: gjcy$1,
    gla: gla$1,
    gl: gl$1,
    glE: glE$1,
    glj: glj$1,
    gnap: gnap$1,
    gnapprox: gnapprox$1,
    gne: gne$1,
    gnE: gnE$1,
    gneq: gneq$1,
    gneqq: gneqq$1,
    gnsim: gnsim$1,
    Gopf: Gopf$1,
    gopf: gopf$1,
    grave: grave$1,
    GreaterEqual: GreaterEqual$1,
    GreaterEqualLess: GreaterEqualLess$1,
    GreaterFullEqual: GreaterFullEqual$1,
    GreaterGreater: GreaterGreater$1,
    GreaterLess: GreaterLess$1,
    GreaterSlantEqual: GreaterSlantEqual$1,
    GreaterTilde: GreaterTilde$1,
    Gscr: Gscr$1,
    gscr: gscr$1,
    gsim: gsim$1,
    gsime: gsime$1,
    gsiml: gsiml$1,
    gtcc: gtcc$1,
    gtcir: gtcir$1,
    gt: gt$1,
    GT: GT$1,
    Gt: Gt$1,
    gtdot: gtdot$1,
    gtlPar: gtlPar$1,
    gtquest: gtquest$1,
    gtrapprox: gtrapprox$1,
    gtrarr: gtrarr$1,
    gtrdot: gtrdot$1,
    gtreqless: gtreqless$1,
    gtreqqless: gtreqqless$1,
    gtrless: gtrless$1,
    gtrsim: gtrsim$1,
    gvertneqq: gvertneqq$1,
    gvnE: gvnE$1,
    Hacek: Hacek$1,
    hairsp: hairsp$1,
    half: half$1,
    hamilt: hamilt$1,
    HARDcy: HARDcy$1,
    hardcy: hardcy$1,
    harrcir: harrcir$1,
    harr: harr$1,
    hArr: hArr$1,
    harrw: harrw$1,
    Hat: Hat$1,
    hbar: hbar$1,
    Hcirc: Hcirc$1,
    hcirc: hcirc$1,
    hearts: hearts$3,
    heartsuit: heartsuit$1,
    hellip: hellip$1,
    hercon: hercon$1,
    hfr: hfr$1,
    Hfr: Hfr$1,
    HilbertSpace: HilbertSpace$1,
    hksearow: hksearow$1,
    hkswarow: hkswarow$1,
    hoarr: hoarr$1,
    homtht: homtht$1,
    hookleftarrow: hookleftarrow$1,
    hookrightarrow: hookrightarrow$1,
    hopf: hopf$1,
    Hopf: Hopf$1,
    horbar: horbar$1,
    HorizontalLine: HorizontalLine$1,
    hscr: hscr$1,
    Hscr: Hscr$1,
    hslash: hslash$1,
    Hstrok: Hstrok$1,
    hstrok: hstrok$1,
    HumpDownHump: HumpDownHump$1,
    HumpEqual: HumpEqual$1,
    hybull: hybull$1,
    hyphen: hyphen$1,
    Iacute: Iacute$1,
    iacute: iacute$1,
    ic: ic$1,
    Icirc: Icirc$1,
    icirc: icirc$1,
    Icy: Icy$1,
    icy: icy$1,
    Idot: Idot$1,
    IEcy: IEcy$1,
    iecy: iecy$1,
    iexcl: iexcl$1,
    iff: iff$1,
    ifr: ifr$1,
    Ifr: Ifr$1,
    Igrave: Igrave$1,
    igrave: igrave$1,
    ii: ii$1,
    iiiint: iiiint$1,
    iiint: iiint$1,
    iinfin: iinfin$1,
    iiota: iiota$1,
    IJlig: IJlig$1,
    ijlig: ijlig$1,
    Imacr: Imacr$1,
    imacr: imacr$1,
    image: image$4,
    ImaginaryI: ImaginaryI$1,
    imagline: imagline$1,
    imagpart: imagpart$1,
    imath: imath$1,
    Im: Im$1,
    imof: imof$1,
    imped: imped$1,
    Implies: Implies$1,
    incare: incare$1,
    infin: infin$1,
    infintie: infintie$1,
    inodot: inodot$1,
    intcal: intcal$1,
    int: int$1,
    Int: Int$1,
    integers: integers$1,
    Integral: Integral$1,
    intercal: intercal$1,
    Intersection: Intersection$1,
    intlarhk: intlarhk$1,
    intprod: intprod$1,
    InvisibleComma: InvisibleComma$1,
    InvisibleTimes: InvisibleTimes$1,
    IOcy: IOcy$1,
    iocy: iocy$1,
    Iogon: Iogon$1,
    iogon: iogon$1,
    Iopf: Iopf$1,
    iopf: iopf$1,
    Iota: Iota$1,
    iota: iota$1,
    iprod: iprod$1,
    iquest: iquest$1,
    iscr: iscr$1,
    Iscr: Iscr$1,
    isin: isin$1,
    isindot: isindot$1,
    isinE: isinE$1,
    isins: isins$1,
    isinsv: isinsv$1,
    isinv: isinv$1,
    it: it$3,
    Itilde: Itilde$1,
    itilde: itilde$1,
    Iukcy: Iukcy$1,
    iukcy: iukcy$1,
    Iuml: Iuml$1,
    iuml: iuml$1,
    Jcirc: Jcirc$1,
    jcirc: jcirc$1,
    Jcy: Jcy$1,
    jcy: jcy$1,
    Jfr: Jfr$1,
    jfr: jfr$1,
    jmath: jmath$1,
    Jopf: Jopf$1,
    jopf: jopf$1,
    Jscr: Jscr$1,
    jscr: jscr$1,
    Jsercy: Jsercy$1,
    jsercy: jsercy$1,
    Jukcy: Jukcy$1,
    jukcy: jukcy$1,
    Kappa: Kappa$1,
    kappa: kappa$1,
    kappav: kappav$1,
    Kcedil: Kcedil$1,
    kcedil: kcedil$1,
    Kcy: Kcy$1,
    kcy: kcy$1,
    Kfr: Kfr$1,
    kfr: kfr$1,
    kgreen: kgreen$1,
    KHcy: KHcy$1,
    khcy: khcy$1,
    KJcy: KJcy$1,
    kjcy: kjcy$1,
    Kopf: Kopf$1,
    kopf: kopf$1,
    Kscr: Kscr$1,
    kscr: kscr$1,
    lAarr: lAarr$1,
    Lacute: Lacute$1,
    lacute: lacute$1,
    laemptyv: laemptyv$1,
    lagran: lagran$1,
    Lambda: Lambda$1,
    lambda: lambda$1,
    lang: lang$1,
    Lang: Lang$1,
    langd: langd$1,
    langle: langle$1,
    lap: lap$1,
    Laplacetrf: Laplacetrf$1,
    laquo: laquo$1,
    larrb: larrb$1,
    larrbfs: larrbfs$1,
    larr: larr$1,
    Larr: Larr$1,
    lArr: lArr$1,
    larrfs: larrfs$1,
    larrhk: larrhk$1,
    larrlp: larrlp$1,
    larrpl: larrpl$1,
    larrsim: larrsim$1,
    larrtl: larrtl$1,
    latail: latail$1,
    lAtail: lAtail$1,
    lat: lat$1,
    late: late$1,
    lates: lates$1,
    lbarr: lbarr$1,
    lBarr: lBarr$1,
    lbbrk: lbbrk$1,
    lbrace: lbrace$1,
    lbrack: lbrack$1,
    lbrke: lbrke$1,
    lbrksld: lbrksld$1,
    lbrkslu: lbrkslu$1,
    Lcaron: Lcaron$1,
    lcaron: lcaron$1,
    Lcedil: Lcedil$1,
    lcedil: lcedil$1,
    lceil: lceil$1,
    lcub: lcub$1,
    Lcy: Lcy$1,
    lcy: lcy$1,
    ldca: ldca$1,
    ldquo: ldquo$1,
    ldquor: ldquor$1,
    ldrdhar: ldrdhar$1,
    ldrushar: ldrushar$1,
    ldsh: ldsh$1,
    le: le$1,
    lE: lE$1,
    LeftAngleBracket: LeftAngleBracket$1,
    LeftArrowBar: LeftArrowBar$1,
    leftarrow: leftarrow$1,
    LeftArrow: LeftArrow$1,
    Leftarrow: Leftarrow$1,
    LeftArrowRightArrow: LeftArrowRightArrow$1,
    leftarrowtail: leftarrowtail$1,
    LeftCeiling: LeftCeiling$1,
    LeftDoubleBracket: LeftDoubleBracket$1,
    LeftDownTeeVector: LeftDownTeeVector$1,
    LeftDownVectorBar: LeftDownVectorBar$1,
    LeftDownVector: LeftDownVector$1,
    LeftFloor: LeftFloor$1,
    leftharpoondown: leftharpoondown$1,
    leftharpoonup: leftharpoonup$1,
    leftleftarrows: leftleftarrows$1,
    leftrightarrow: leftrightarrow$1,
    LeftRightArrow: LeftRightArrow$1,
    Leftrightarrow: Leftrightarrow$1,
    leftrightarrows: leftrightarrows$1,
    leftrightharpoons: leftrightharpoons$1,
    leftrightsquigarrow: leftrightsquigarrow$1,
    LeftRightVector: LeftRightVector$1,
    LeftTeeArrow: LeftTeeArrow$1,
    LeftTee: LeftTee$1,
    LeftTeeVector: LeftTeeVector$1,
    leftthreetimes: leftthreetimes$1,
    LeftTriangleBar: LeftTriangleBar$1,
    LeftTriangle: LeftTriangle$1,
    LeftTriangleEqual: LeftTriangleEqual$1,
    LeftUpDownVector: LeftUpDownVector$1,
    LeftUpTeeVector: LeftUpTeeVector$1,
    LeftUpVectorBar: LeftUpVectorBar$1,
    LeftUpVector: LeftUpVector$1,
    LeftVectorBar: LeftVectorBar$1,
    LeftVector: LeftVector$1,
    lEg: lEg$1,
    leg: leg$3,
    leq: leq$1,
    leqq: leqq$1,
    leqslant: leqslant$1,
    lescc: lescc$1,
    les: les$1,
    lesdot: lesdot$1,
    lesdoto: lesdoto$1,
    lesdotor: lesdotor$1,
    lesg: lesg$1,
    lesges: lesges$1,
    lessapprox: lessapprox$1,
    lessdot: lessdot$1,
    lesseqgtr: lesseqgtr$1,
    lesseqqgtr: lesseqqgtr$1,
    LessEqualGreater: LessEqualGreater$1,
    LessFullEqual: LessFullEqual$1,
    LessGreater: LessGreater$1,
    lessgtr: lessgtr$1,
    LessLess: LessLess$1,
    lesssim: lesssim$1,
    LessSlantEqual: LessSlantEqual$1,
    LessTilde: LessTilde$1,
    lfisht: lfisht$1,
    lfloor: lfloor$1,
    Lfr: Lfr$1,
    lfr: lfr$1,
    lg: lg$1,
    lgE: lgE$1,
    lHar: lHar$1,
    lhard: lhard$1,
    lharu: lharu$1,
    lharul: lharul$1,
    lhblk: lhblk$1,
    LJcy: LJcy$1,
    ljcy: ljcy$1,
    llarr: llarr$1,
    ll: ll$1,
    Ll: Ll$1,
    llcorner: llcorner$1,
    Lleftarrow: Lleftarrow$1,
    llhard: llhard$1,
    lltri: lltri$1,
    Lmidot: Lmidot$1,
    lmidot: lmidot$1,
    lmoustache: lmoustache$1,
    lmoust: lmoust$1,
    lnap: lnap$1,
    lnapprox: lnapprox$1,
    lne: lne$1,
    lnE: lnE$1,
    lneq: lneq$1,
    lneqq: lneqq$1,
    lnsim: lnsim$1,
    loang: loang$1,
    loarr: loarr$1,
    lobrk: lobrk$1,
    longleftarrow: longleftarrow$1,
    LongLeftArrow: LongLeftArrow$1,
    Longleftarrow: Longleftarrow$1,
    longleftrightarrow: longleftrightarrow$1,
    LongLeftRightArrow: LongLeftRightArrow$1,
    Longleftrightarrow: Longleftrightarrow$1,
    longmapsto: longmapsto$1,
    longrightarrow: longrightarrow$1,
    LongRightArrow: LongRightArrow$1,
    Longrightarrow: Longrightarrow$1,
    looparrowleft: looparrowleft$1,
    looparrowright: looparrowright$1,
    lopar: lopar$1,
    Lopf: Lopf$1,
    lopf: lopf$1,
    loplus: loplus$1,
    lotimes: lotimes$1,
    lowast: lowast$1,
    lowbar: lowbar$1,
    LowerLeftArrow: LowerLeftArrow$1,
    LowerRightArrow: LowerRightArrow$1,
    loz: loz$1,
    lozenge: lozenge$1,
    lozf: lozf$1,
    lpar: lpar$1,
    lparlt: lparlt$1,
    lrarr: lrarr$1,
    lrcorner: lrcorner$1,
    lrhar: lrhar$1,
    lrhard: lrhard$1,
    lrm: lrm$1,
    lrtri: lrtri$1,
    lsaquo: lsaquo$1,
    lscr: lscr$1,
    Lscr: Lscr$1,
    lsh: lsh$1,
    Lsh: Lsh$1,
    lsim: lsim$1,
    lsime: lsime$1,
    lsimg: lsimg$1,
    lsqb: lsqb$1,
    lsquo: lsquo$1,
    lsquor: lsquor$1,
    Lstrok: Lstrok$1,
    lstrok: lstrok$1,
    ltcc: ltcc$1,
    ltcir: ltcir$1,
    lt: lt$1,
    LT: LT$1,
    Lt: Lt$1,
    ltdot: ltdot$1,
    lthree: lthree$1,
    ltimes: ltimes$1,
    ltlarr: ltlarr$1,
    ltquest: ltquest$1,
    ltri: ltri$1,
    ltrie: ltrie$1,
    ltrif: ltrif$1,
    ltrPar: ltrPar$1,
    lurdshar: lurdshar$1,
    luruhar: luruhar$1,
    lvertneqq: lvertneqq$1,
    lvnE: lvnE$1,
    macr: macr$1,
    male: male$1,
    malt: malt$1,
    maltese: maltese$1,
    map: map$2,
    mapsto: mapsto$1,
    mapstodown: mapstodown$1,
    mapstoleft: mapstoleft$1,
    mapstoup: mapstoup$1,
    marker: marker$1,
    mcomma: mcomma$1,
    Mcy: Mcy$1,
    mcy: mcy$1,
    mdash: mdash$1,
    mDDot: mDDot$1,
    measuredangle: measuredangle$1,
    MediumSpace: MediumSpace$1,
    Mellintrf: Mellintrf$1,
    Mfr: Mfr$1,
    mfr: mfr$1,
    mho: mho$1,
    micro: micro$1,
    midast: midast$1,
    midcir: midcir$1,
    mid: mid$1,
    middot: middot$1,
    minusb: minusb$1,
    minus: minus$1,
    minusd: minusd$1,
    minusdu: minusdu$1,
    MinusPlus: MinusPlus$1,
    mlcp: mlcp$1,
    mldr: mldr$1,
    mnplus: mnplus$1,
    models: models$1,
    Mopf: Mopf$1,
    mopf: mopf$1,
    mp: mp$1,
    mscr: mscr$1,
    Mscr: Mscr$1,
    mstpos: mstpos$1,
    Mu: Mu$1,
    mu: mu$1,
    multimap: multimap$1,
    mumap: mumap$1,
    nabla: nabla$1,
    Nacute: Nacute$1,
    nacute: nacute$1,
    nang: nang$1,
    nap: nap$1,
    napE: napE$1,
    napid: napid$1,
    napos: napos$1,
    napprox: napprox$1,
    natural: natural$1,
    naturals: naturals$1,
    natur: natur$1,
    nbsp: nbsp$1,
    nbump: nbump$1,
    nbumpe: nbumpe$1,
    ncap: ncap$1,
    Ncaron: Ncaron$1,
    ncaron: ncaron$1,
    Ncedil: Ncedil$1,
    ncedil: ncedil$1,
    ncong: ncong$1,
    ncongdot: ncongdot$1,
    ncup: ncup$1,
    Ncy: Ncy$1,
    ncy: ncy$1,
    ndash: ndash$1,
    nearhk: nearhk$1,
    nearr: nearr$1,
    neArr: neArr$1,
    nearrow: nearrow$1,
    ne: ne$1,
    nedot: nedot$1,
    NegativeMediumSpace: NegativeMediumSpace$1,
    NegativeThickSpace: NegativeThickSpace$1,
    NegativeThinSpace: NegativeThinSpace$1,
    NegativeVeryThinSpace: NegativeVeryThinSpace$1,
    nequiv: nequiv$1,
    nesear: nesear$1,
    nesim: nesim$1,
    NestedGreaterGreater: NestedGreaterGreater$1,
    NestedLessLess: NestedLessLess$1,
    NewLine: NewLine$1,
    nexist: nexist$1,
    nexists: nexists$1,
    Nfr: Nfr$1,
    nfr: nfr$1,
    ngE: ngE$1,
    nge: nge$1,
    ngeq: ngeq$1,
    ngeqq: ngeqq$1,
    ngeqslant: ngeqslant$1,
    nges: nges$1,
    nGg: nGg$1,
    ngsim: ngsim$1,
    nGt: nGt$1,
    ngt: ngt$1,
    ngtr: ngtr$1,
    nGtv: nGtv$1,
    nharr: nharr$1,
    nhArr: nhArr$1,
    nhpar: nhpar$1,
    ni: ni$1,
    nis: nis$1,
    nisd: nisd$1,
    niv: niv$1,
    NJcy: NJcy$1,
    njcy: njcy$1,
    nlarr: nlarr$1,
    nlArr: nlArr$1,
    nldr: nldr$1,
    nlE: nlE$1,
    nle: nle$1,
    nleftarrow: nleftarrow$1,
    nLeftarrow: nLeftarrow$1,
    nleftrightarrow: nleftrightarrow$1,
    nLeftrightarrow: nLeftrightarrow$1,
    nleq: nleq$1,
    nleqq: nleqq$1,
    nleqslant: nleqslant$1,
    nles: nles$1,
    nless: nless$1,
    nLl: nLl$1,
    nlsim: nlsim$1,
    nLt: nLt$1,
    nlt: nlt$1,
    nltri: nltri$1,
    nltrie: nltrie$1,
    nLtv: nLtv$1,
    nmid: nmid$1,
    NoBreak: NoBreak$1,
    NonBreakingSpace: NonBreakingSpace$1,
    nopf: nopf$1,
    Nopf: Nopf$1,
    Not: Not$1,
    not: not$1,
    NotCongruent: NotCongruent$1,
    NotCupCap: NotCupCap$1,
    NotDoubleVerticalBar: NotDoubleVerticalBar$1,
    NotElement: NotElement$1,
    NotEqual: NotEqual$1,
    NotEqualTilde: NotEqualTilde$1,
    NotExists: NotExists$1,
    NotGreater: NotGreater$1,
    NotGreaterEqual: NotGreaterEqual$1,
    NotGreaterFullEqual: NotGreaterFullEqual$1,
    NotGreaterGreater: NotGreaterGreater$1,
    NotGreaterLess: NotGreaterLess$1,
    NotGreaterSlantEqual: NotGreaterSlantEqual$1,
    NotGreaterTilde: NotGreaterTilde$1,
    NotHumpDownHump: NotHumpDownHump$1,
    NotHumpEqual: NotHumpEqual$1,
    notin: notin$1,
    notindot: notindot$1,
    notinE: notinE$1,
    notinva: notinva$1,
    notinvb: notinvb$1,
    notinvc: notinvc$1,
    NotLeftTriangleBar: NotLeftTriangleBar$1,
    NotLeftTriangle: NotLeftTriangle$1,
    NotLeftTriangleEqual: NotLeftTriangleEqual$1,
    NotLess: NotLess$1,
    NotLessEqual: NotLessEqual$1,
    NotLessGreater: NotLessGreater$1,
    NotLessLess: NotLessLess$1,
    NotLessSlantEqual: NotLessSlantEqual$1,
    NotLessTilde: NotLessTilde$1,
    NotNestedGreaterGreater: NotNestedGreaterGreater$1,
    NotNestedLessLess: NotNestedLessLess$1,
    notni: notni$1,
    notniva: notniva$1,
    notnivb: notnivb$1,
    notnivc: notnivc$1,
    NotPrecedes: NotPrecedes$1,
    NotPrecedesEqual: NotPrecedesEqual$1,
    NotPrecedesSlantEqual: NotPrecedesSlantEqual$1,
    NotReverseElement: NotReverseElement$1,
    NotRightTriangleBar: NotRightTriangleBar$1,
    NotRightTriangle: NotRightTriangle$1,
    NotRightTriangleEqual: NotRightTriangleEqual$1,
    NotSquareSubset: NotSquareSubset$1,
    NotSquareSubsetEqual: NotSquareSubsetEqual$1,
    NotSquareSuperset: NotSquareSuperset$1,
    NotSquareSupersetEqual: NotSquareSupersetEqual$1,
    NotSubset: NotSubset$1,
    NotSubsetEqual: NotSubsetEqual$1,
    NotSucceeds: NotSucceeds$1,
    NotSucceedsEqual: NotSucceedsEqual$1,
    NotSucceedsSlantEqual: NotSucceedsSlantEqual$1,
    NotSucceedsTilde: NotSucceedsTilde$1,
    NotSuperset: NotSuperset$1,
    NotSupersetEqual: NotSupersetEqual$1,
    NotTilde: NotTilde$1,
    NotTildeEqual: NotTildeEqual$1,
    NotTildeFullEqual: NotTildeFullEqual$1,
    NotTildeTilde: NotTildeTilde$1,
    NotVerticalBar: NotVerticalBar$1,
    nparallel: nparallel$1,
    npar: npar$1,
    nparsl: nparsl$1,
    npart: npart$1,
    npolint: npolint$1,
    npr: npr$1,
    nprcue: nprcue$1,
    nprec: nprec$1,
    npreceq: npreceq$1,
    npre: npre$1,
    nrarrc: nrarrc$1,
    nrarr: nrarr$1,
    nrArr: nrArr$1,
    nrarrw: nrarrw$1,
    nrightarrow: nrightarrow$1,
    nRightarrow: nRightarrow$1,
    nrtri: nrtri$1,
    nrtrie: nrtrie$1,
    nsc: nsc$1,
    nsccue: nsccue$1,
    nsce: nsce$1,
    Nscr: Nscr$1,
    nscr: nscr$1,
    nshortmid: nshortmid$1,
    nshortparallel: nshortparallel$1,
    nsim: nsim$1,
    nsime: nsime$1,
    nsimeq: nsimeq$1,
    nsmid: nsmid$1,
    nspar: nspar$1,
    nsqsube: nsqsube$1,
    nsqsupe: nsqsupe$1,
    nsub: nsub$1,
    nsubE: nsubE$1,
    nsube: nsube$1,
    nsubset: nsubset$1,
    nsubseteq: nsubseteq$1,
    nsubseteqq: nsubseteqq$1,
    nsucc: nsucc$1,
    nsucceq: nsucceq$1,
    nsup: nsup$1,
    nsupE: nsupE$1,
    nsupe: nsupe$1,
    nsupset: nsupset$1,
    nsupseteq: nsupseteq$1,
    nsupseteqq: nsupseteqq$1,
    ntgl: ntgl$1,
    Ntilde: Ntilde$1,
    ntilde: ntilde$1,
    ntlg: ntlg$1,
    ntriangleleft: ntriangleleft$1,
    ntrianglelefteq: ntrianglelefteq$1,
    ntriangleright: ntriangleright$1,
    ntrianglerighteq: ntrianglerighteq$1,
    Nu: Nu$1,
    nu: nu$1,
    num: num$1,
    numero: numero$1,
    numsp: numsp$1,
    nvap: nvap$1,
    nvdash: nvdash$1,
    nvDash: nvDash$1,
    nVdash: nVdash$1,
    nVDash: nVDash$1,
    nvge: nvge$1,
    nvgt: nvgt$1,
    nvHarr: nvHarr$1,
    nvinfin: nvinfin$1,
    nvlArr: nvlArr$1,
    nvle: nvle$1,
    nvlt: nvlt$1,
    nvltrie: nvltrie$1,
    nvrArr: nvrArr$1,
    nvrtrie: nvrtrie$1,
    nvsim: nvsim$1,
    nwarhk: nwarhk$1,
    nwarr: nwarr$1,
    nwArr: nwArr$1,
    nwarrow: nwarrow$1,
    nwnear: nwnear$1,
    Oacute: Oacute$1,
    oacute: oacute$1,
    oast: oast$1,
    Ocirc: Ocirc$1,
    ocirc: ocirc$1,
    ocir: ocir$1,
    Ocy: Ocy$1,
    ocy: ocy$1,
    odash: odash$1,
    Odblac: Odblac$1,
    odblac: odblac$1,
    odiv: odiv$1,
    odot: odot$1,
    odsold: odsold$1,
    OElig: OElig$1,
    oelig: oelig$1,
    ofcir: ofcir$1,
    Ofr: Ofr$1,
    ofr: ofr$1,
    ogon: ogon$1,
    Ograve: Ograve$1,
    ograve: ograve$1,
    ogt: ogt$1,
    ohbar: ohbar$1,
    ohm: ohm$1,
    oint: oint$1,
    olarr: olarr$1,
    olcir: olcir$1,
    olcross: olcross$1,
    oline: oline$1,
    olt: olt$1,
    Omacr: Omacr$1,
    omacr: omacr$1,
    Omega: Omega$1,
    omega: omega$1,
    Omicron: Omicron$1,
    omicron: omicron$1,
    omid: omid$1,
    ominus: ominus$1,
    Oopf: Oopf$1,
    oopf: oopf$1,
    opar: opar$1,
    OpenCurlyDoubleQuote: OpenCurlyDoubleQuote$1,
    OpenCurlyQuote: OpenCurlyQuote$1,
    operp: operp$1,
    oplus: oplus$1,
    orarr: orarr$1,
    Or: Or$1,
    or: or$1,
    ord: ord$1,
    order: order$1,
    orderof: orderof$1,
    ordf: ordf$1,
    ordm: ordm$1,
    origof: origof$1,
    oror: oror$1,
    orslope: orslope$1,
    orv: orv$1,
    oS: oS$1,
    Oscr: Oscr$1,
    oscr: oscr$1,
    Oslash: Oslash$1,
    oslash: oslash$1,
    osol: osol$1,
    Otilde: Otilde$1,
    otilde: otilde$1,
    otimesas: otimesas$1,
    Otimes: Otimes$1,
    otimes: otimes$1,
    Ouml: Ouml$1,
    ouml: ouml$1,
    ovbar: ovbar$1,
    OverBar: OverBar$1,
    OverBrace: OverBrace$1,
    OverBracket: OverBracket$1,
    OverParenthesis: OverParenthesis$1,
    para: para$1,
    parallel: parallel$1,
    par: par$1,
    parsim: parsim$1,
    parsl: parsl$1,
    part: part$1,
    PartialD: PartialD$1,
    Pcy: Pcy$1,
    pcy: pcy$1,
    percnt: percnt$1,
    period: period$1,
    permil: permil$1,
    perp: perp$1,
    pertenk: pertenk$1,
    Pfr: Pfr$1,
    pfr: pfr$1,
    Phi: Phi$1,
    phi: phi$1,
    phiv: phiv$1,
    phmmat: phmmat$1,
    phone: phone$3,
    Pi: Pi$1,
    pi: pi$1,
    pitchfork: pitchfork$1,
    piv: piv$1,
    planck: planck$1,
    planckh: planckh$1,
    plankv: plankv$1,
    plusacir: plusacir$1,
    plusb: plusb$1,
    pluscir: pluscir$1,
    plus: plus$1,
    plusdo: plusdo$1,
    plusdu: plusdu$1,
    pluse: pluse$1,
    PlusMinus: PlusMinus$1,
    plusmn: plusmn$1,
    plussim: plussim$1,
    plustwo: plustwo$1,
    pm: pm$1,
    Poincareplane: Poincareplane$1,
    pointint: pointint$1,
    popf: popf$1,
    Popf: Popf$1,
    pound: pound$3,
    prap: prap$1,
    Pr: Pr$1,
    pr: pr$1,
    prcue: prcue$1,
    precapprox: precapprox$1,
    prec: prec$1,
    preccurlyeq: preccurlyeq$1,
    Precedes: Precedes$1,
    PrecedesEqual: PrecedesEqual$1,
    PrecedesSlantEqual: PrecedesSlantEqual$1,
    PrecedesTilde: PrecedesTilde$1,
    preceq: preceq$1,
    precnapprox: precnapprox$1,
    precneqq: precneqq$1,
    precnsim: precnsim$1,
    pre: pre$1,
    prE: prE$1,
    precsim: precsim$1,
    prime: prime$1,
    Prime: Prime$1,
    primes: primes$1,
    prnap: prnap$1,
    prnE: prnE$1,
    prnsim: prnsim$1,
    prod: prod$1,
    Product: Product$1,
    profalar: profalar$1,
    profline: profline$1,
    profsurf: profsurf$1,
    prop: prop$1,
    Proportional: Proportional$1,
    Proportion: Proportion$1,
    propto: propto$1,
    prsim: prsim$1,
    prurel: prurel$1,
    Pscr: Pscr$1,
    pscr: pscr$1,
    Psi: Psi$1,
    psi: psi$1,
    puncsp: puncsp$1,
    Qfr: Qfr$1,
    qfr: qfr$1,
    qint: qint$1,
    qopf: qopf$1,
    Qopf: Qopf$1,
    qprime: qprime$1,
    Qscr: Qscr$1,
    qscr: qscr$1,
    quaternions: quaternions$1,
    quatint: quatint$1,
    quest: quest$1,
    questeq: questeq$1,
    quot: quot$1,
    QUOT: QUOT$1,
    rAarr: rAarr$1,
    race: race$1,
    Racute: Racute$1,
    racute: racute$1,
    radic: radic$1,
    raemptyv: raemptyv$1,
    rang: rang$1,
    Rang: Rang$1,
    rangd: rangd$1,
    range: range$1,
    rangle: rangle$1,
    raquo: raquo$1,
    rarrap: rarrap$1,
    rarrb: rarrb$1,
    rarrbfs: rarrbfs$1,
    rarrc: rarrc$1,
    rarr: rarr$1,
    Rarr: Rarr$1,
    rArr: rArr$1,
    rarrfs: rarrfs$1,
    rarrhk: rarrhk$1,
    rarrlp: rarrlp$1,
    rarrpl: rarrpl$1,
    rarrsim: rarrsim$1,
    Rarrtl: Rarrtl$1,
    rarrtl: rarrtl$1,
    rarrw: rarrw$1,
    ratail: ratail$1,
    rAtail: rAtail$1,
    ratio: ratio$1,
    rationals: rationals$1,
    rbarr: rbarr$1,
    rBarr: rBarr$1,
    RBarr: RBarr$1,
    rbbrk: rbbrk$1,
    rbrace: rbrace$1,
    rbrack: rbrack$1,
    rbrke: rbrke$1,
    rbrksld: rbrksld$1,
    rbrkslu: rbrkslu$1,
    Rcaron: Rcaron$1,
    rcaron: rcaron$1,
    Rcedil: Rcedil$1,
    rcedil: rcedil$1,
    rceil: rceil$1,
    rcub: rcub$1,
    Rcy: Rcy$1,
    rcy: rcy$1,
    rdca: rdca$1,
    rdldhar: rdldhar$1,
    rdquo: rdquo$1,
    rdquor: rdquor$1,
    rdsh: rdsh$1,
    real: real$1,
    realine: realine$1,
    realpart: realpart$1,
    reals: reals$1,
    Re: Re$1,
    rect: rect$1,
    reg: reg$1,
    REG: REG$1,
    ReverseElement: ReverseElement$1,
    ReverseEquilibrium: ReverseEquilibrium$1,
    ReverseUpEquilibrium: ReverseUpEquilibrium$1,
    rfisht: rfisht$1,
    rfloor: rfloor$1,
    rfr: rfr$1,
    Rfr: Rfr$1,
    rHar: rHar$1,
    rhard: rhard$1,
    rharu: rharu$1,
    rharul: rharul$1,
    Rho: Rho$1,
    rho: rho$1,
    rhov: rhov$1,
    RightAngleBracket: RightAngleBracket$1,
    RightArrowBar: RightArrowBar$1,
    rightarrow: rightarrow$1,
    RightArrow: RightArrow$1,
    Rightarrow: Rightarrow$1,
    RightArrowLeftArrow: RightArrowLeftArrow$1,
    rightarrowtail: rightarrowtail$1,
    RightCeiling: RightCeiling$1,
    RightDoubleBracket: RightDoubleBracket$1,
    RightDownTeeVector: RightDownTeeVector$1,
    RightDownVectorBar: RightDownVectorBar$1,
    RightDownVector: RightDownVector$1,
    RightFloor: RightFloor$1,
    rightharpoondown: rightharpoondown$1,
    rightharpoonup: rightharpoonup$1,
    rightleftarrows: rightleftarrows$1,
    rightleftharpoons: rightleftharpoons$1,
    rightrightarrows: rightrightarrows$1,
    rightsquigarrow: rightsquigarrow$1,
    RightTeeArrow: RightTeeArrow$1,
    RightTee: RightTee$1,
    RightTeeVector: RightTeeVector$1,
    rightthreetimes: rightthreetimes$1,
    RightTriangleBar: RightTriangleBar$1,
    RightTriangle: RightTriangle$1,
    RightTriangleEqual: RightTriangleEqual$1,
    RightUpDownVector: RightUpDownVector$1,
    RightUpTeeVector: RightUpTeeVector$1,
    RightUpVectorBar: RightUpVectorBar$1,
    RightUpVector: RightUpVector$1,
    RightVectorBar: RightVectorBar$1,
    RightVector: RightVector$1,
    ring: ring$3,
    risingdotseq: risingdotseq$1,
    rlarr: rlarr$1,
    rlhar: rlhar$1,
    rlm: rlm$1,
    rmoustache: rmoustache$1,
    rmoust: rmoust$1,
    rnmid: rnmid$1,
    roang: roang$1,
    roarr: roarr$1,
    robrk: robrk$1,
    ropar: ropar$1,
    ropf: ropf$1,
    Ropf: Ropf$1,
    roplus: roplus$1,
    rotimes: rotimes$1,
    RoundImplies: RoundImplies$1,
    rpar: rpar$1,
    rpargt: rpargt$1,
    rppolint: rppolint$1,
    rrarr: rrarr$1,
    Rrightarrow: Rrightarrow$1,
    rsaquo: rsaquo$1,
    rscr: rscr$1,
    Rscr: Rscr$1,
    rsh: rsh$1,
    Rsh: Rsh$1,
    rsqb: rsqb$1,
    rsquo: rsquo$1,
    rsquor: rsquor$1,
    rthree: rthree$1,
    rtimes: rtimes$1,
    rtri: rtri$1,
    rtrie: rtrie$1,
    rtrif: rtrif$1,
    rtriltri: rtriltri$1,
    RuleDelayed: RuleDelayed$1,
    ruluhar: ruluhar$1,
    rx: rx$1,
    Sacute: Sacute$1,
    sacute: sacute$1,
    sbquo: sbquo$1,
    scap: scap$1,
    Scaron: Scaron$1,
    scaron: scaron$1,
    Sc: Sc$1,
    sc: sc$1,
    sccue: sccue$1,
    sce: sce$1,
    scE: scE$1,
    Scedil: Scedil$1,
    scedil: scedil$1,
    Scirc: Scirc$1,
    scirc: scirc$1,
    scnap: scnap$1,
    scnE: scnE$1,
    scnsim: scnsim$1,
    scpolint: scpolint$1,
    scsim: scsim$1,
    Scy: Scy$1,
    scy: scy$1,
    sdotb: sdotb$1,
    sdot: sdot$1,
    sdote: sdote$1,
    searhk: searhk$1,
    searr: searr$1,
    seArr: seArr$1,
    searrow: searrow$1,
    sect: sect$1,
    semi: semi$1,
    seswar: seswar$1,
    setminus: setminus$1,
    setmn: setmn$1,
    sext: sext$1,
    Sfr: Sfr$1,
    sfr: sfr$1,
    sfrown: sfrown$1,
    sharp: sharp$1,
    SHCHcy: SHCHcy$1,
    shchcy: shchcy$1,
    SHcy: SHcy$1,
    shcy: shcy$1,
    ShortDownArrow: ShortDownArrow$1,
    ShortLeftArrow: ShortLeftArrow$1,
    shortmid: shortmid$1,
    shortparallel: shortparallel$1,
    ShortRightArrow: ShortRightArrow$1,
    ShortUpArrow: ShortUpArrow$1,
    shy: shy$1,
    Sigma: Sigma$1,
    sigma: sigma$1,
    sigmaf: sigmaf$1,
    sigmav: sigmav$1,
    sim: sim$1,
    simdot: simdot$1,
    sime: sime$1,
    simeq: simeq$1,
    simg: simg$1,
    simgE: simgE$1,
    siml: siml$1,
    simlE: simlE$1,
    simne: simne$1,
    simplus: simplus$1,
    simrarr: simrarr$1,
    slarr: slarr$1,
    SmallCircle: SmallCircle$1,
    smallsetminus: smallsetminus$1,
    smashp: smashp$1,
    smeparsl: smeparsl$1,
    smid: smid$1,
    smile: smile$3,
    smt: smt$1,
    smte: smte$1,
    smtes: smtes$1,
    SOFTcy: SOFTcy$1,
    softcy: softcy$1,
    solbar: solbar$1,
    solb: solb$1,
    sol: sol$1,
    Sopf: Sopf$1,
    sopf: sopf$1,
    spades: spades$3,
    spadesuit: spadesuit$1,
    spar: spar$1,
    sqcap: sqcap$1,
    sqcaps: sqcaps$1,
    sqcup: sqcup$1,
    sqcups: sqcups$1,
    Sqrt: Sqrt$1,
    sqsub: sqsub$1,
    sqsube: sqsube$1,
    sqsubset: sqsubset$1,
    sqsubseteq: sqsubseteq$1,
    sqsup: sqsup$1,
    sqsupe: sqsupe$1,
    sqsupset: sqsupset$1,
    sqsupseteq: sqsupseteq$1,
    square: square$1,
    Square: Square$1,
    SquareIntersection: SquareIntersection$1,
    SquareSubset: SquareSubset$1,
    SquareSubsetEqual: SquareSubsetEqual$1,
    SquareSuperset: SquareSuperset$1,
    SquareSupersetEqual: SquareSupersetEqual$1,
    SquareUnion: SquareUnion$1,
    squarf: squarf$1,
    squ: squ$1,
    squf: squf$1,
    srarr: srarr$1,
    Sscr: Sscr$1,
    sscr: sscr$1,
    ssetmn: ssetmn$1,
    ssmile: ssmile$1,
    sstarf: sstarf$1,
    Star: Star$1,
    star: star$3,
    starf: starf$1,
    straightepsilon: straightepsilon$1,
    straightphi: straightphi$1,
    strns: strns$1,
    sub: sub$1,
    Sub: Sub$1,
    subdot: subdot$1,
    subE: subE$1,
    sube: sube$1,
    subedot: subedot$1,
    submult: submult$1,
    subnE: subnE$1,
    subne: subne$1,
    subplus: subplus$1,
    subrarr: subrarr$1,
    subset: subset$1,
    Subset: Subset$1,
    subseteq: subseteq$1,
    subseteqq: subseteqq$1,
    SubsetEqual: SubsetEqual$1,
    subsetneq: subsetneq$1,
    subsetneqq: subsetneqq$1,
    subsim: subsim$1,
    subsub: subsub$1,
    subsup: subsup$1,
    succapprox: succapprox$1,
    succ: succ$1,
    succcurlyeq: succcurlyeq$1,
    Succeeds: Succeeds$1,
    SucceedsEqual: SucceedsEqual$1,
    SucceedsSlantEqual: SucceedsSlantEqual$1,
    SucceedsTilde: SucceedsTilde$1,
    succeq: succeq$1,
    succnapprox: succnapprox$1,
    succneqq: succneqq$1,
    succnsim: succnsim$1,
    succsim: succsim$1,
    SuchThat: SuchThat$1,
    sum: sum$1,
    Sum: Sum$1,
    sung: sung$1,
    sup1: sup1$1,
    sup2: sup2$1,
    sup3: sup3$1,
    sup: sup$1,
    Sup: Sup$1,
    supdot: supdot$1,
    supdsub: supdsub$1,
    supE: supE$1,
    supe: supe$1,
    supedot: supedot$1,
    Superset: Superset$1,
    SupersetEqual: SupersetEqual$1,
    suphsol: suphsol$1,
    suphsub: suphsub$1,
    suplarr: suplarr$1,
    supmult: supmult$1,
    supnE: supnE$1,
    supne: supne$1,
    supplus: supplus$1,
    supset: supset$1,
    Supset: Supset$1,
    supseteq: supseteq$1,
    supseteqq: supseteqq$1,
    supsetneq: supsetneq$1,
    supsetneqq: supsetneqq$1,
    supsim: supsim$1,
    supsub: supsub$1,
    supsup: supsup$1,
    swarhk: swarhk$1,
    swarr: swarr$1,
    swArr: swArr$1,
    swarrow: swarrow$1,
    swnwar: swnwar$1,
    szlig: szlig$1,
    Tab: Tab$1,
    target: target$1,
    Tau: Tau$1,
    tau: tau$1,
    tbrk: tbrk$1,
    Tcaron: Tcaron$1,
    tcaron: tcaron$1,
    Tcedil: Tcedil$1,
    tcedil: tcedil$1,
    Tcy: Tcy$1,
    tcy: tcy$1,
    tdot: tdot$1,
    telrec: telrec$1,
    Tfr: Tfr$1,
    tfr: tfr$1,
    there4: there4$1,
    therefore: therefore$1,
    Therefore: Therefore$1,
    Theta: Theta$1,
    theta: theta$1,
    thetasym: thetasym$1,
    thetav: thetav$1,
    thickapprox: thickapprox$1,
    thicksim: thicksim$1,
    ThickSpace: ThickSpace$1,
    ThinSpace: ThinSpace$1,
    thinsp: thinsp$1,
    thkap: thkap$1,
    thksim: thksim$1,
    THORN: THORN$1,
    thorn: thorn$1,
    tilde: tilde$1,
    Tilde: Tilde$1,
    TildeEqual: TildeEqual$1,
    TildeFullEqual: TildeFullEqual$1,
    TildeTilde: TildeTilde$1,
    timesbar: timesbar$1,
    timesb: timesb$1,
    times: times$1,
    timesd: timesd$1,
    tint: tint$1,
    toea: toea$1,
    topbot: topbot$1,
    topcir: topcir$1,
    top: top$3,
    Topf: Topf$1,
    topf: topf$1,
    topfork: topfork$1,
    tosa: tosa$1,
    tprime: tprime$1,
    trade: trade$1,
    TRADE: TRADE$1,
    triangle: triangle$1,
    triangledown: triangledown$1,
    triangleleft: triangleleft$1,
    trianglelefteq: trianglelefteq$1,
    triangleq: triangleq$1,
    triangleright: triangleright$1,
    trianglerighteq: trianglerighteq$1,
    tridot: tridot$1,
    trie: trie$1,
    triminus: triminus$1,
    TripleDot: TripleDot$1,
    triplus: triplus$1,
    trisb: trisb$1,
    tritime: tritime$1,
    trpezium: trpezium$1,
    Tscr: Tscr$1,
    tscr: tscr$1,
    TScy: TScy$1,
    tscy: tscy$1,
    TSHcy: TSHcy$1,
    tshcy: tshcy$1,
    Tstrok: Tstrok$1,
    tstrok: tstrok$1,
    twixt: twixt$1,
    twoheadleftarrow: twoheadleftarrow$1,
    twoheadrightarrow: twoheadrightarrow$1,
    Uacute: Uacute$1,
    uacute: uacute$1,
    uarr: uarr$1,
    Uarr: Uarr$1,
    uArr: uArr$1,
    Uarrocir: Uarrocir$1,
    Ubrcy: Ubrcy$1,
    ubrcy: ubrcy$1,
    Ubreve: Ubreve$1,
    ubreve: ubreve$1,
    Ucirc: Ucirc$1,
    ucirc: ucirc$1,
    Ucy: Ucy$1,
    ucy: ucy$1,
    udarr: udarr$1,
    Udblac: Udblac$1,
    udblac: udblac$1,
    udhar: udhar$1,
    ufisht: ufisht$1,
    Ufr: Ufr$1,
    ufr: ufr$1,
    Ugrave: Ugrave$1,
    ugrave: ugrave$1,
    uHar: uHar$1,
    uharl: uharl$1,
    uharr: uharr$1,
    uhblk: uhblk$1,
    ulcorn: ulcorn$1,
    ulcorner: ulcorner$1,
    ulcrop: ulcrop$1,
    ultri: ultri$1,
    Umacr: Umacr$1,
    umacr: umacr$1,
    uml: uml$1,
    UnderBar: UnderBar$1,
    UnderBrace: UnderBrace$1,
    UnderBracket: UnderBracket$1,
    UnderParenthesis: UnderParenthesis$1,
    Union: Union$1,
    UnionPlus: UnionPlus$1,
    Uogon: Uogon$1,
    uogon: uogon$1,
    Uopf: Uopf$1,
    uopf: uopf$1,
    UpArrowBar: UpArrowBar$1,
    uparrow: uparrow$1,
    UpArrow: UpArrow$1,
    Uparrow: Uparrow$1,
    UpArrowDownArrow: UpArrowDownArrow$1,
    updownarrow: updownarrow$1,
    UpDownArrow: UpDownArrow$1,
    Updownarrow: Updownarrow$1,
    UpEquilibrium: UpEquilibrium$1,
    upharpoonleft: upharpoonleft$1,
    upharpoonright: upharpoonright$1,
    uplus: uplus$1,
    UpperLeftArrow: UpperLeftArrow$1,
    UpperRightArrow: UpperRightArrow$1,
    upsi: upsi$1,
    Upsi: Upsi$1,
    upsih: upsih$1,
    Upsilon: Upsilon$1,
    upsilon: upsilon$1,
    UpTeeArrow: UpTeeArrow$1,
    UpTee: UpTee$1,
    upuparrows: upuparrows$1,
    urcorn: urcorn$1,
    urcorner: urcorner$1,
    urcrop: urcrop$1,
    Uring: Uring$1,
    uring: uring$1,
    urtri: urtri$1,
    Uscr: Uscr$1,
    uscr: uscr$1,
    utdot: utdot$1,
    Utilde: Utilde$1,
    utilde: utilde$1,
    utri: utri$1,
    utrif: utrif$1,
    uuarr: uuarr$1,
    Uuml: Uuml$1,
    uuml: uuml$1,
    uwangle: uwangle$1,
    vangrt: vangrt$1,
    varepsilon: varepsilon$1,
    varkappa: varkappa$1,
    varnothing: varnothing$1,
    varphi: varphi$1,
    varpi: varpi$1,
    varpropto: varpropto$1,
    varr: varr$1,
    vArr: vArr$1,
    varrho: varrho$1,
    varsigma: varsigma$1,
    varsubsetneq: varsubsetneq$1,
    varsubsetneqq: varsubsetneqq$1,
    varsupsetneq: varsupsetneq$1,
    varsupsetneqq: varsupsetneqq$1,
    vartheta: vartheta$1,
    vartriangleleft: vartriangleleft$1,
    vartriangleright: vartriangleright$1,
    vBar: vBar$1,
    Vbar: Vbar$1,
    vBarv: vBarv$1,
    Vcy: Vcy$1,
    vcy: vcy$1,
    vdash: vdash$1,
    vDash: vDash$1,
    Vdash: Vdash$1,
    VDash: VDash$1,
    Vdashl: Vdashl$1,
    veebar: veebar$1,
    vee: vee$1,
    Vee: Vee$1,
    veeeq: veeeq$1,
    vellip: vellip$1,
    verbar: verbar$1,
    Verbar: Verbar$1,
    vert: vert$1,
    Vert: Vert$1,
    VerticalBar: VerticalBar$1,
    VerticalLine: VerticalLine$1,
    VerticalSeparator: VerticalSeparator$1,
    VerticalTilde: VerticalTilde$1,
    VeryThinSpace: VeryThinSpace$1,
    Vfr: Vfr$1,
    vfr: vfr$1,
    vltri: vltri$1,
    vnsub: vnsub$1,
    vnsup: vnsup$1,
    Vopf: Vopf$1,
    vopf: vopf$1,
    vprop: vprop$1,
    vrtri: vrtri$1,
    Vscr: Vscr$1,
    vscr: vscr$1,
    vsubnE: vsubnE$1,
    vsubne: vsubne$1,
    vsupnE: vsupnE$1,
    vsupne: vsupne$1,
    Vvdash: Vvdash$1,
    vzigzag: vzigzag$1,
    Wcirc: Wcirc$1,
    wcirc: wcirc$1,
    wedbar: wedbar$1,
    wedge: wedge$1,
    Wedge: Wedge$1,
    wedgeq: wedgeq$1,
    weierp: weierp$1,
    Wfr: Wfr$1,
    wfr: wfr$1,
    Wopf: Wopf$1,
    wopf: wopf$1,
    wp: wp$1,
    wr: wr$1,
    wreath: wreath$1,
    Wscr: Wscr$1,
    wscr: wscr$1,
    xcap: xcap$1,
    xcirc: xcirc$1,
    xcup: xcup$1,
    xdtri: xdtri$1,
    Xfr: Xfr$1,
    xfr: xfr$1,
    xharr: xharr$1,
    xhArr: xhArr$1,
    Xi: Xi$1,
    xi: xi$1,
    xlarr: xlarr$1,
    xlArr: xlArr$1,
    xmap: xmap$1,
    xnis: xnis$1,
    xodot: xodot$1,
    Xopf: Xopf$1,
    xopf: xopf$1,
    xoplus: xoplus$1,
    xotime: xotime$1,
    xrarr: xrarr$1,
    xrArr: xrArr$1,
    Xscr: Xscr$1,
    xscr: xscr$1,
    xsqcup: xsqcup$1,
    xuplus: xuplus$1,
    xutri: xutri$1,
    xvee: xvee$1,
    xwedge: xwedge$1,
    Yacute: Yacute$1,
    yacute: yacute$1,
    YAcy: YAcy$1,
    yacy: yacy$1,
    Ycirc: Ycirc$1,
    ycirc: ycirc$1,
    Ycy: Ycy$1,
    ycy: ycy$1,
    yen: yen$3,
    Yfr: Yfr$1,
    yfr: yfr$1,
    YIcy: YIcy$1,
    yicy: yicy$1,
    Yopf: Yopf$1,
    yopf: yopf$1,
    Yscr: Yscr$1,
    yscr: yscr$1,
    YUcy: YUcy$1,
    yucy: yucy$1,
    yuml: yuml$1,
    Yuml: Yuml$1,
    Zacute: Zacute$1,
    zacute: zacute$1,
    Zcaron: Zcaron$1,
    zcaron: zcaron$1,
    Zcy: Zcy$1,
    zcy: zcy$1,
    Zdot: Zdot$1,
    zdot: zdot$1,
    zeetrf: zeetrf$1,
    ZeroWidthSpace: ZeroWidthSpace$1,
    Zeta: Zeta$1,
    zeta: zeta$1,
    zfr: zfr$1,
    Zfr: Zfr$1,
    ZHcy: ZHcy$1,
    zhcy: zhcy$1,
    zigrarr: zigrarr$1,
    zopf: zopf$1,
    Zopf: Zopf$1,
    Zscr: Zscr$1,
    zscr: zscr$1,
    zwj: zwj$1,
    zwnj: zwnj$1,
    'default': entities$4
  });

  var require$$0$2 = getCjsExportFromNamespace(entities$5);

  /*eslint quotes:0*/
  var entities$3 = require$$0$2;

  var regex$4=/[!-#%-\*,-\/:;\?@\[-\]_\{\}\xA1\xA7\xAB\xB6\xB7\xBB\xBF\u037E\u0387\u055A-\u055F\u0589\u058A\u05BE\u05C0\u05C3\u05C6\u05F3\u05F4\u0609\u060A\u060C\u060D\u061B\u061E\u061F\u066A-\u066D\u06D4\u0700-\u070D\u07F7-\u07F9\u0830-\u083E\u085E\u0964\u0965\u0970\u09FD\u0A76\u0AF0\u0C84\u0DF4\u0E4F\u0E5A\u0E5B\u0F04-\u0F12\u0F14\u0F3A-\u0F3D\u0F85\u0FD0-\u0FD4\u0FD9\u0FDA\u104A-\u104F\u10FB\u1360-\u1368\u1400\u166D\u166E\u169B\u169C\u16EB-\u16ED\u1735\u1736\u17D4-\u17D6\u17D8-\u17DA\u1800-\u180A\u1944\u1945\u1A1E\u1A1F\u1AA0-\u1AA6\u1AA8-\u1AAD\u1B5A-\u1B60\u1BFC-\u1BFF\u1C3B-\u1C3F\u1C7E\u1C7F\u1CC0-\u1CC7\u1CD3\u2010-\u2027\u2030-\u2043\u2045-\u2051\u2053-\u205E\u207D\u207E\u208D\u208E\u2308-\u230B\u2329\u232A\u2768-\u2775\u27C5\u27C6\u27E6-\u27EF\u2983-\u2998\u29D8-\u29DB\u29FC\u29FD\u2CF9-\u2CFC\u2CFE\u2CFF\u2D70\u2E00-\u2E2E\u2E30-\u2E4E\u3001-\u3003\u3008-\u3011\u3014-\u301F\u3030\u303D\u30A0\u30FB\uA4FE\uA4FF\uA60D-\uA60F\uA673\uA67E\uA6F2-\uA6F7\uA874-\uA877\uA8CE\uA8CF\uA8F8-\uA8FA\uA8FC\uA92E\uA92F\uA95F\uA9C1-\uA9CD\uA9DE\uA9DF\uAA5C-\uAA5F\uAADE\uAADF\uAAF0\uAAF1\uABEB\uFD3E\uFD3F\uFE10-\uFE19\uFE30-\uFE52\uFE54-\uFE61\uFE63\uFE68\uFE6A\uFE6B\uFF01-\uFF03\uFF05-\uFF0A\uFF0C-\uFF0F\uFF1A\uFF1B\uFF1F\uFF20\uFF3B-\uFF3D\uFF3F\uFF5B\uFF5D\uFF5F-\uFF65]|\uD800[\uDD00-\uDD02\uDF9F\uDFD0]|\uD801\uDD6F|\uD802[\uDC57\uDD1F\uDD3F\uDE50-\uDE58\uDE7F\uDEF0-\uDEF6\uDF39-\uDF3F\uDF99-\uDF9C]|\uD803[\uDF55-\uDF59]|\uD804[\uDC47-\uDC4D\uDCBB\uDCBC\uDCBE-\uDCC1\uDD40-\uDD43\uDD74\uDD75\uDDC5-\uDDC8\uDDCD\uDDDB\uDDDD-\uDDDF\uDE38-\uDE3D\uDEA9]|\uD805[\uDC4B-\uDC4F\uDC5B\uDC5D\uDCC6\uDDC1-\uDDD7\uDE41-\uDE43\uDE60-\uDE6C\uDF3C-\uDF3E]|\uD806[\uDC3B\uDE3F-\uDE46\uDE9A-\uDE9C\uDE9E-\uDEA2]|\uD807[\uDC41-\uDC45\uDC70\uDC71\uDEF7\uDEF8]|\uD809[\uDC70-\uDC74]|\uD81A[\uDE6E\uDE6F\uDEF5\uDF37-\uDF3B\uDF44]|\uD81B[\uDE97-\uDE9A]|\uD82F\uDC9F|\uD836[\uDE87-\uDE8B]|\uD83A[\uDD5E\uDD5F]/;

  var encodeCache = {};


  // Create a lookup array where anything but characters in `chars` string
  // and alphanumeric chars is percent-encoded.
  //
  function getEncodeCache(exclude) {
    var i, ch, cache = encodeCache[exclude];
    if (cache) { return cache; }

    cache = encodeCache[exclude] = [];

    for (i = 0; i < 128; i++) {
      ch = String.fromCharCode(i);

      if (/^[0-9a-z]$/i.test(ch)) {
        // always allow unencoded alphanumeric characters
        cache.push(ch);
      } else {
        cache.push('%' + ('0' + i.toString(16).toUpperCase()).slice(-2));
      }
    }

    for (i = 0; i < exclude.length; i++) {
      cache[exclude.charCodeAt(i)] = exclude[i];
    }

    return cache;
  }


  // Encode unsafe characters with percent-encoding, skipping already
  // encoded sequences.
  //
  //  - string       - string to encode
  //  - exclude      - list of characters to ignore (in addition to a-zA-Z0-9)
  //  - keepEscaped  - don't encode '%' in a correct escape sequence (default: true)
  //
  function encode$2(string, exclude, keepEscaped) {
    var i, l, code, nextCode, cache,
        result = '';

    if (typeof exclude !== 'string') {
      // encode(string, keepEscaped)
      keepEscaped  = exclude;
      exclude = encode$2.defaultChars;
    }

    if (typeof keepEscaped === 'undefined') {
      keepEscaped = true;
    }

    cache = getEncodeCache(exclude);

    for (i = 0, l = string.length; i < l; i++) {
      code = string.charCodeAt(i);

      if (keepEscaped && code === 0x25 /* % */ && i + 2 < l) {
        if (/^[0-9a-f]{2}$/i.test(string.slice(i + 1, i + 3))) {
          result += string.slice(i, i + 3);
          i += 2;
          continue;
        }
      }

      if (code < 128) {
        result += cache[code];
        continue;
      }

      if (code >= 0xD800 && code <= 0xDFFF) {
        if (code >= 0xD800 && code <= 0xDBFF && i + 1 < l) {
          nextCode = string.charCodeAt(i + 1);
          if (nextCode >= 0xDC00 && nextCode <= 0xDFFF) {
            result += encodeURIComponent(string[i] + string[i + 1]);
            i++;
            continue;
          }
        }
        result += '%EF%BF%BD';
        continue;
      }

      result += encodeURIComponent(string[i]);
    }

    return result;
  }

  encode$2.defaultChars   = ";/?:@&=+$,-_.!~*'()#";
  encode$2.componentChars = "-_.!~*'()";


  var encode_1 = encode$2;

  /* eslint-disable no-bitwise */

  var decodeCache = {};

  function getDecodeCache(exclude) {
    var i, ch, cache = decodeCache[exclude];
    if (cache) { return cache; }

    cache = decodeCache[exclude] = [];

    for (i = 0; i < 128; i++) {
      ch = String.fromCharCode(i);
      cache.push(ch);
    }

    for (i = 0; i < exclude.length; i++) {
      ch = exclude.charCodeAt(i);
      cache[ch] = '%' + ('0' + ch.toString(16).toUpperCase()).slice(-2);
    }

    return cache;
  }


  // Decode percent-encoded string.
  //
  function decode$2(string, exclude) {
    var cache;

    if (typeof exclude !== 'string') {
      exclude = decode$2.defaultChars;
    }

    cache = getDecodeCache(exclude);

    return string.replace(/(%[a-f0-9]{2})+/gi, function(seq) {
      var i, l, b1, b2, b3, b4, chr,
          result = '';

      for (i = 0, l = seq.length; i < l; i += 3) {
        b1 = parseInt(seq.slice(i + 1, i + 3), 16);

        if (b1 < 0x80) {
          result += cache[b1];
          continue;
        }

        if ((b1 & 0xE0) === 0xC0 && (i + 3 < l)) {
          // 110xxxxx 10xxxxxx
          b2 = parseInt(seq.slice(i + 4, i + 6), 16);

          if ((b2 & 0xC0) === 0x80) {
            chr = ((b1 << 6) & 0x7C0) | (b2 & 0x3F);

            if (chr < 0x80) {
              result += '\ufffd\ufffd';
            } else {
              result += String.fromCharCode(chr);
            }

            i += 3;
            continue;
          }
        }

        if ((b1 & 0xF0) === 0xE0 && (i + 6 < l)) {
          // 1110xxxx 10xxxxxx 10xxxxxx
          b2 = parseInt(seq.slice(i + 4, i + 6), 16);
          b3 = parseInt(seq.slice(i + 7, i + 9), 16);

          if ((b2 & 0xC0) === 0x80 && (b3 & 0xC0) === 0x80) {
            chr = ((b1 << 12) & 0xF000) | ((b2 << 6) & 0xFC0) | (b3 & 0x3F);

            if (chr < 0x800 || (chr >= 0xD800 && chr <= 0xDFFF)) {
              result += '\ufffd\ufffd\ufffd';
            } else {
              result += String.fromCharCode(chr);
            }

            i += 6;
            continue;
          }
        }

        if ((b1 & 0xF8) === 0xF0 && (i + 9 < l)) {
          // 111110xx 10xxxxxx 10xxxxxx 10xxxxxx
          b2 = parseInt(seq.slice(i + 4, i + 6), 16);
          b3 = parseInt(seq.slice(i + 7, i + 9), 16);
          b4 = parseInt(seq.slice(i + 10, i + 12), 16);

          if ((b2 & 0xC0) === 0x80 && (b3 & 0xC0) === 0x80 && (b4 & 0xC0) === 0x80) {
            chr = ((b1 << 18) & 0x1C0000) | ((b2 << 12) & 0x3F000) | ((b3 << 6) & 0xFC0) | (b4 & 0x3F);

            if (chr < 0x10000 || chr > 0x10FFFF) {
              result += '\ufffd\ufffd\ufffd\ufffd';
            } else {
              chr -= 0x10000;
              result += String.fromCharCode(0xD800 + (chr >> 10), 0xDC00 + (chr & 0x3FF));
            }

            i += 9;
            continue;
          }
        }

        result += '\ufffd';
      }

      return result;
    });
  }


  decode$2.defaultChars   = ';/?:@&=+$,#';
  decode$2.componentChars = '';


  var decode_1 = decode$2;

  var format$1 = function format(url) {
    var result = '';

    result += url.protocol || '';
    result += url.slashes ? '//' : '';
    result += url.auth ? url.auth + '@' : '';

    if (url.hostname && url.hostname.indexOf(':') !== -1) {
      // ipv6 address
      result += '[' + url.hostname + ']';
    } else {
      result += url.hostname || '';
    }

    result += url.port ? ':' + url.port : '';
    result += url.pathname || '';
    result += url.search || '';
    result += url.hash || '';

    return result;
  };

  // Copyright Joyent, Inc. and other Node contributors.

  //
  // Changes from joyent/node:
  //
  // 1. No leading slash in paths,
  //    e.g. in `url.parse('http://foo?bar')` pathname is ``, not `/`
  //
  // 2. Backslashes are not replaced with slashes,
  //    so `http:\\example.org\` is treated like a relative path
  //
  // 3. Trailing colon is treated like a part of the path,
  //    i.e. in `http://example.org:foo` pathname is `:foo`
  //
  // 4. Nothing is URL-encoded in the resulting object,
  //    (in joyent/node some chars in auth and paths are encoded)
  //
  // 5. `url.parse()` does not have `parseQueryString` argument
  //
  // 6. Removed extraneous result properties: `host`, `path`, `query`, etc.,
  //    which can be constructed using other parts of the url.
  //


  function Url() {
    this.protocol = null;
    this.slashes = null;
    this.auth = null;
    this.port = null;
    this.hostname = null;
    this.hash = null;
    this.search = null;
    this.pathname = null;
  }

  // Reference: RFC 3986, RFC 1808, RFC 2396

  // define these here so at least they only have to be
  // compiled once on the first module load.
  var protocolPattern = /^([a-z0-9.+-]+:)/i,
      portPattern = /:[0-9]*$/,

      // Special case for a simple path URL
      simplePathPattern = /^(\/\/?(?!\/)[^\?\s]*)(\?[^\s]*)?$/,

      // RFC 2396: characters reserved for delimiting URLs.
      // We actually just auto-escape these.
      delims = [ '<', '>', '"', '`', ' ', '\r', '\n', '\t' ],

      // RFC 2396: characters not allowed for various reasons.
      unwise = [ '{', '}', '|', '\\', '^', '`' ].concat(delims),

      // Allowed by RFCs, but cause of XSS attacks.  Always escape these.
      autoEscape = [ '\'' ].concat(unwise),
      // Characters that are never ever allowed in a hostname.
      // Note that any invalid chars are also handled, but these
      // are the ones that are *expected* to be seen, so we fast-path
      // them.
      nonHostChars = [ '%', '/', '?', ';', '#' ].concat(autoEscape),
      hostEndingChars = [ '/', '?', '#' ],
      hostnameMaxLen = 255,
      hostnamePartPattern = /^[+a-z0-9A-Z_-]{0,63}$/,
      hostnamePartStart = /^([+a-z0-9A-Z_-]{0,63})(.*)$/,
      // protocols that can allow "unsafe" and "unwise" chars.
      /* eslint-disable no-script-url */
      // protocols that never have a hostname.
      hostlessProtocol = {
        'javascript': true,
        'javascript:': true
      },
      // protocols that always contain a // bit.
      slashedProtocol = {
        'http': true,
        'https': true,
        'ftp': true,
        'gopher': true,
        'file': true,
        'http:': true,
        'https:': true,
        'ftp:': true,
        'gopher:': true,
        'file:': true
      };
      /* eslint-enable no-script-url */

  function urlParse(url, slashesDenoteHost) {
    if (url && url instanceof Url) { return url; }

    var u = new Url();
    u.parse(url, slashesDenoteHost);
    return u;
  }

  Url.prototype.parse = function(url, slashesDenoteHost) {
    var i, l, lowerProto, hec, slashes,
        rest = url;

    // trim before proceeding.
    // This is to support parse stuff like "  http://foo.com  \n"
    rest = rest.trim();

    if (!slashesDenoteHost && url.split('#').length === 1) {
      // Try fast path regexp
      var simplePath = simplePathPattern.exec(rest);
      if (simplePath) {
        this.pathname = simplePath[1];
        if (simplePath[2]) {
          this.search = simplePath[2];
        }
        return this;
      }
    }

    var proto = protocolPattern.exec(rest);
    if (proto) {
      proto = proto[0];
      lowerProto = proto.toLowerCase();
      this.protocol = proto;
      rest = rest.substr(proto.length);
    }

    // figure out if it's got a host
    // user@server is *always* interpreted as a hostname, and url
    // resolution will treat //foo/bar as host=foo,path=bar because that's
    // how the browser resolves relative URLs.
    if (slashesDenoteHost || proto || rest.match(/^\/\/[^@\/]+@[^@\/]+/)) {
      slashes = rest.substr(0, 2) === '//';
      if (slashes && !(proto && hostlessProtocol[proto])) {
        rest = rest.substr(2);
        this.slashes = true;
      }
    }

    if (!hostlessProtocol[proto] &&
        (slashes || (proto && !slashedProtocol[proto]))) {

      // there's a hostname.
      // the first instance of /, ?, ;, or # ends the host.
      //
      // If there is an @ in the hostname, then non-host chars *are* allowed
      // to the left of the last @ sign, unless some host-ending character
      // comes *before* the @-sign.
      // URLs are obnoxious.
      //
      // ex:
      // http://a@b@c/ => user:a@b host:c
      // http://a@b?@c => user:a host:c path:/?@c

      // v0.12 TODO(isaacs): This is not quite how Chrome does things.
      // Review our test case against browsers more comprehensively.

      // find the first instance of any hostEndingChars
      var hostEnd = -1;
      for (i = 0; i < hostEndingChars.length; i++) {
        hec = rest.indexOf(hostEndingChars[i]);
        if (hec !== -1 && (hostEnd === -1 || hec < hostEnd)) {
          hostEnd = hec;
        }
      }

      // at this point, either we have an explicit point where the
      // auth portion cannot go past, or the last @ char is the decider.
      var auth, atSign;
      if (hostEnd === -1) {
        // atSign can be anywhere.
        atSign = rest.lastIndexOf('@');
      } else {
        // atSign must be in auth portion.
        // http://a@b/c@d => host:b auth:a path:/c@d
        atSign = rest.lastIndexOf('@', hostEnd);
      }

      // Now we have a portion which is definitely the auth.
      // Pull that off.
      if (atSign !== -1) {
        auth = rest.slice(0, atSign);
        rest = rest.slice(atSign + 1);
        this.auth = auth;
      }

      // the host is the remaining to the left of the first non-host char
      hostEnd = -1;
      for (i = 0; i < nonHostChars.length; i++) {
        hec = rest.indexOf(nonHostChars[i]);
        if (hec !== -1 && (hostEnd === -1 || hec < hostEnd)) {
          hostEnd = hec;
        }
      }
      // if we still have not hit it, then the entire thing is a host.
      if (hostEnd === -1) {
        hostEnd = rest.length;
      }

      if (rest[hostEnd - 1] === ':') { hostEnd--; }
      var host = rest.slice(0, hostEnd);
      rest = rest.slice(hostEnd);

      // pull out port.
      this.parseHost(host);

      // we've indicated that there is a hostname,
      // so even if it's empty, it has to be present.
      this.hostname = this.hostname || '';

      // if hostname begins with [ and ends with ]
      // assume that it's an IPv6 address.
      var ipv6Hostname = this.hostname[0] === '[' &&
          this.hostname[this.hostname.length - 1] === ']';

      // validate a little.
      if (!ipv6Hostname) {
        var hostparts = this.hostname.split(/\./);
        for (i = 0, l = hostparts.length; i < l; i++) {
          var part = hostparts[i];
          if (!part) { continue; }
          if (!part.match(hostnamePartPattern)) {
            var newpart = '';
            for (var j = 0, k = part.length; j < k; j++) {
              if (part.charCodeAt(j) > 127) {
                // we replace non-ASCII char with a temporary placeholder
                // we need this to make sure size of hostname is not
                // broken by replacing non-ASCII by nothing
                newpart += 'x';
              } else {
                newpart += part[j];
              }
            }
            // we test again with ASCII char only
            if (!newpart.match(hostnamePartPattern)) {
              var validParts = hostparts.slice(0, i);
              var notHost = hostparts.slice(i + 1);
              var bit = part.match(hostnamePartStart);
              if (bit) {
                validParts.push(bit[1]);
                notHost.unshift(bit[2]);
              }
              if (notHost.length) {
                rest = notHost.join('.') + rest;
              }
              this.hostname = validParts.join('.');
              break;
            }
          }
        }
      }

      if (this.hostname.length > hostnameMaxLen) {
        this.hostname = '';
      }

      // strip [ and ] from the hostname
      // the host field still retains them, though
      if (ipv6Hostname) {
        this.hostname = this.hostname.substr(1, this.hostname.length - 2);
      }
    }

    // chop off from the tail first.
    var hash = rest.indexOf('#');
    if (hash !== -1) {
      // got a fragment string.
      this.hash = rest.substr(hash);
      rest = rest.slice(0, hash);
    }
    var qm = rest.indexOf('?');
    if (qm !== -1) {
      this.search = rest.substr(qm);
      rest = rest.slice(0, qm);
    }
    if (rest) { this.pathname = rest; }
    if (slashedProtocol[lowerProto] &&
        this.hostname && !this.pathname) {
      this.pathname = '';
    }

    return this;
  };

  Url.prototype.parseHost = function(host) {
    var port = portPattern.exec(host);
    if (port) {
      port = port[0];
      if (port !== ':') {
        this.port = port.substr(1);
      }
      host = host.substr(0, host.length - port.length);
    }
    if (host) { this.hostname = host; }
  };

  var parse$1 = urlParse;

  var encode$1 = encode_1;
  var decode$1 = decode_1;
  var format = format$1;
  var parse  = parse$1;

  var mdurl = {
  	encode: encode$1,
  	decode: decode$1,
  	format: format,
  	parse: parse
  };

  var regex$3=/[\0-\uD7FF\uE000-\uFFFF]|[\uD800-\uDBFF][\uDC00-\uDFFF]|[\uD800-\uDBFF](?![\uDC00-\uDFFF])|(?:[^\uD800-\uDBFF]|^)[\uDC00-\uDFFF]/;

  var regex$2=/[\0-\x1F\x7F-\x9F]/;

  var regex$1=/[\xAD\u0600-\u0605\u061C\u06DD\u070F\u08E2\u180E\u200B-\u200F\u202A-\u202E\u2060-\u2064\u2066-\u206F\uFEFF\uFFF9-\uFFFB]|\uD804[\uDCBD\uDCCD]|\uD82F[\uDCA0-\uDCA3]|\uD834[\uDD73-\uDD7A]|\uDB40[\uDC01\uDC20-\uDC7F]/;

  var regex=/[ \xA0\u1680\u2000-\u200A\u2028\u2029\u202F\u205F\u3000]/;

  var Any = regex$3;
  var Cc  = regex$2;
  var Cf  = regex$1;
  var P   = regex$4;
  var Z   = regex;

  var uc_micro = {
  	Any: Any,
  	Cc: Cc,
  	Cf: Cf,
  	P: P,
  	Z: Z
  };

  var utils$1 = createCommonjsModule(function (module, exports) {


  function _class(obj) { return Object.prototype.toString.call(obj); }

  function isString(obj) { return _class(obj) === '[object String]'; }

  var _hasOwnProperty = Object.prototype.hasOwnProperty;

  function has(object, key) {
    return _hasOwnProperty.call(object, key);
  }

  // Merge objects
  //
  function assign(obj /*from1, from2, from3, ...*/) {
    var sources = Array.prototype.slice.call(arguments, 1);

    sources.forEach(function (source) {
      if (!source) { return; }

      if (typeof source !== 'object') {
        throw new TypeError(source + 'must be object');
      }

      Object.keys(source).forEach(function (key) {
        obj[key] = source[key];
      });
    });

    return obj;
  }

  // Remove element from array and put another array at those position.
  // Useful for some operations with tokens
  function arrayReplaceAt(src, pos, newElements) {
    return [].concat(src.slice(0, pos), newElements, src.slice(pos + 1));
  }

  ////////////////////////////////////////////////////////////////////////////////

  function isValidEntityCode(c) {
    /*eslint no-bitwise:0*/
    // broken sequence
    if (c >= 0xD800 && c <= 0xDFFF) { return false; }
    // never used
    if (c >= 0xFDD0 && c <= 0xFDEF) { return false; }
    if ((c & 0xFFFF) === 0xFFFF || (c & 0xFFFF) === 0xFFFE) { return false; }
    // control codes
    if (c >= 0x00 && c <= 0x08) { return false; }
    if (c === 0x0B) { return false; }
    if (c >= 0x0E && c <= 0x1F) { return false; }
    if (c >= 0x7F && c <= 0x9F) { return false; }
    // out of range
    if (c > 0x10FFFF) { return false; }
    return true;
  }

  function fromCodePoint(c) {
    /*eslint no-bitwise:0*/
    if (c > 0xffff) {
      c -= 0x10000;
      var surrogate1 = 0xd800 + (c >> 10),
          surrogate2 = 0xdc00 + (c & 0x3ff);

      return String.fromCharCode(surrogate1, surrogate2);
    }
    return String.fromCharCode(c);
  }


  var UNESCAPE_MD_RE  = /\\([!"#$%&'()*+,\-.\/:;<=>?@[\\\]^_`{|}~])/g;
  var ENTITY_RE       = /&([a-z#][a-z0-9]{1,31});/gi;
  var UNESCAPE_ALL_RE = new RegExp(UNESCAPE_MD_RE.source + '|' + ENTITY_RE.source, 'gi');

  var DIGITAL_ENTITY_TEST_RE = /^#((?:x[a-f0-9]{1,8}|[0-9]{1,8}))/i;



  function replaceEntityPattern(match, name) {
    var code = 0;

    if (has(entities$3, name)) {
      return entities$3[name];
    }

    if (name.charCodeAt(0) === 0x23/* # */ && DIGITAL_ENTITY_TEST_RE.test(name)) {
      code = name[1].toLowerCase() === 'x' ?
        parseInt(name.slice(2), 16) : parseInt(name.slice(1), 10);

      if (isValidEntityCode(code)) {
        return fromCodePoint(code);
      }
    }

    return match;
  }

  /*function replaceEntities(str) {
    if (str.indexOf('&') < 0) { return str; }

    return str.replace(ENTITY_RE, replaceEntityPattern);
  }*/

  function unescapeMd(str) {
    if (str.indexOf('\\') < 0) { return str; }
    return str.replace(UNESCAPE_MD_RE, '$1');
  }

  function unescapeAll(str) {
    if (str.indexOf('\\') < 0 && str.indexOf('&') < 0) { return str; }

    return str.replace(UNESCAPE_ALL_RE, function (match, escaped, entity) {
      if (escaped) { return escaped; }
      return replaceEntityPattern(match, entity);
    });
  }

  ////////////////////////////////////////////////////////////////////////////////

  var HTML_ESCAPE_TEST_RE = /[&<>"]/;
  var HTML_ESCAPE_REPLACE_RE = /[&<>"]/g;
  var HTML_REPLACEMENTS = {
    '&': '&amp;',
    '<': '&lt;',
    '>': '&gt;',
    '"': '&quot;'
  };

  function replaceUnsafeChar(ch) {
    return HTML_REPLACEMENTS[ch];
  }

  function escapeHtml(str) {
    if (HTML_ESCAPE_TEST_RE.test(str)) {
      return str.replace(HTML_ESCAPE_REPLACE_RE, replaceUnsafeChar);
    }
    return str;
  }

  ////////////////////////////////////////////////////////////////////////////////

  var REGEXP_ESCAPE_RE = /[.?*+^$[\]\\(){}|-]/g;

  function escapeRE(str) {
    return str.replace(REGEXP_ESCAPE_RE, '\\$&');
  }

  ////////////////////////////////////////////////////////////////////////////////

  function isSpace(code) {
    switch (code) {
      case 0x09:
      case 0x20:
        return true;
    }
    return false;
  }

  // Zs (unicode class) || [\t\f\v\r\n]
  function isWhiteSpace(code) {
    if (code >= 0x2000 && code <= 0x200A) { return true; }
    switch (code) {
      case 0x09: // \t
      case 0x0A: // \n
      case 0x0B: // \v
      case 0x0C: // \f
      case 0x0D: // \r
      case 0x20:
      case 0xA0:
      case 0x1680:
      case 0x202F:
      case 0x205F:
      case 0x3000:
        return true;
    }
    return false;
  }

  ////////////////////////////////////////////////////////////////////////////////

  /*eslint-disable max-len*/


  // Currently without astral characters support.
  function isPunctChar(ch) {
    return regex$4.test(ch);
  }


  // Markdown ASCII punctuation characters.
  //
  // !, ", #, $, %, &, ', (, ), *, +, ,, -, ., /, :, ;, <, =, >, ?, @, [, \, ], ^, _, `, {, |, }, or ~
  // http://spec.commonmark.org/0.15/#ascii-punctuation-character
  //
  // Don't confuse with unicode punctuation !!! It lacks some chars in ascii range.
  //
  function isMdAsciiPunct(ch) {
    switch (ch) {
      case 0x21/* ! */:
      case 0x22/* " */:
      case 0x23/* # */:
      case 0x24/* $ */:
      case 0x25/* % */:
      case 0x26/* & */:
      case 0x27/* ' */:
      case 0x28/* ( */:
      case 0x29/* ) */:
      case 0x2A/* * */:
      case 0x2B/* + */:
      case 0x2C/* , */:
      case 0x2D/* - */:
      case 0x2E/* . */:
      case 0x2F/* / */:
      case 0x3A/* : */:
      case 0x3B/* ; */:
      case 0x3C/* < */:
      case 0x3D/* = */:
      case 0x3E/* > */:
      case 0x3F/* ? */:
      case 0x40/* @ */:
      case 0x5B/* [ */:
      case 0x5C/* \ */:
      case 0x5D/* ] */:
      case 0x5E/* ^ */:
      case 0x5F/* _ */:
      case 0x60/* ` */:
      case 0x7B/* { */:
      case 0x7C/* | */:
      case 0x7D/* } */:
      case 0x7E/* ~ */:
        return true;
      default:
        return false;
    }
  }

  // Hepler to unify [reference labels].
  //
  function normalizeReference(str) {
    // Trim and collapse whitespace
    //
    str = str.trim().replace(/\s+/g, ' ');

    // In node v10 'ẞ'.toLowerCase() === 'Ṿ', which is presumed to be a bug
    // fixed in v12 (couldn't find any details).
    //
    // So treat this one as a special case
    // (remove this when node v10 is no longer supported).
    //
    if ('ẞ'.toLowerCase() === 'Ṿ') {
      str = str.replace(/ẞ/g, 'ß');
    }

    // .toLowerCase().toUpperCase() should get rid of all differences
    // between letter variants.
    //
    // Simple .toLowerCase() doesn't normalize 125 code points correctly,
    // and .toUpperCase doesn't normalize 6 of them (list of exceptions:
    // İ, ϴ, ẞ, Ω, K, Å - those are already uppercased, but have differently
    // uppercased versions).
    //
    // Here's an example showing how it happens. Lets take greek letter omega:
    // uppercase U+0398 (Θ), U+03f4 (ϴ) and lowercase U+03b8 (θ), U+03d1 (ϑ)
    //
    // Unicode entries:
    // 0398;GREEK CAPITAL LETTER THETA;Lu;0;L;;;;;N;;;;03B8;
    // 03B8;GREEK SMALL LETTER THETA;Ll;0;L;;;;;N;;;0398;;0398
    // 03D1;GREEK THETA SYMBOL;Ll;0;L;<compat> 03B8;;;;N;GREEK SMALL LETTER SCRIPT THETA;;0398;;0398
    // 03F4;GREEK CAPITAL THETA SYMBOL;Lu;0;L;<compat> 0398;;;;N;;;;03B8;
    //
    // Case-insensitive comparison should treat all of them as equivalent.
    //
    // But .toLowerCase() doesn't change ϑ (it's already lowercase),
    // and .toUpperCase() doesn't change ϴ (already uppercase).
    //
    // Applying first lower then upper case normalizes any character:
    // '\u0398\u03f4\u03b8\u03d1'.toLowerCase().toUpperCase() === '\u0398\u0398\u0398\u0398'
    //
    // Note: this is equivalent to unicode case folding; unicode normalization
    // is a different step that is not required here.
    //
    // Final result should be uppercased, because it's later stored in an object
    // (this avoid a conflict with Object.prototype members,
    // most notably, `__proto__`)
    //
    return str.toLowerCase().toUpperCase();
  }

  ////////////////////////////////////////////////////////////////////////////////

  // Re-export libraries commonly used in both markdown-it and its plugins,
  // so plugins won't have to depend on them explicitly, which reduces their
  // bundled size (e.g. a browser build).
  //
  exports.lib                 = {};
  exports.lib.mdurl           = mdurl;
  exports.lib.ucmicro         = uc_micro;

  exports.assign              = assign;
  exports.isString            = isString;
  exports.has                 = has;
  exports.unescapeMd          = unescapeMd;
  exports.unescapeAll         = unescapeAll;
  exports.isValidEntityCode   = isValidEntityCode;
  exports.fromCodePoint       = fromCodePoint;
  // exports.replaceEntities     = replaceEntities;
  exports.escapeHtml          = escapeHtml;
  exports.arrayReplaceAt      = arrayReplaceAt;
  exports.isSpace             = isSpace;
  exports.isWhiteSpace        = isWhiteSpace;
  exports.isMdAsciiPunct      = isMdAsciiPunct;
  exports.isPunctChar         = isPunctChar;
  exports.escapeRE            = escapeRE;
  exports.normalizeReference  = normalizeReference;
  });
  utils$1.lib;
  utils$1.assign;
  utils$1.isString;
  utils$1.has;
  utils$1.unescapeMd;
  utils$1.unescapeAll;
  utils$1.isValidEntityCode;
  utils$1.fromCodePoint;
  utils$1.escapeHtml;
  utils$1.arrayReplaceAt;
  utils$1.isSpace;
  utils$1.isWhiteSpace;
  utils$1.isMdAsciiPunct;
  utils$1.isPunctChar;
  utils$1.escapeRE;
  utils$1.normalizeReference;

  // Parse link label

  var parse_link_label$1 = function parseLinkLabel(state, start, disableNested) {
    var level, found, marker, prevPos,
        labelEnd = -1,
        max = state.posMax,
        oldPos = state.pos;

    state.pos = start + 1;
    level = 1;

    while (state.pos < max) {
      marker = state.src.charCodeAt(state.pos);
      if (marker === 0x5D /* ] */) {
        level--;
        if (level === 0) {
          found = true;
          break;
        }
      }

      prevPos = state.pos;
      state.md.inline.skipToken(state);
      if (marker === 0x5B /* [ */) {
        if (prevPos === state.pos - 1) {
          // increase level if we find text `[`, which is not a part of any token
          level++;
        } else if (disableNested) {
          state.pos = oldPos;
          return -1;
        }
      }
    }

    if (found) {
      labelEnd = state.pos;
    }

    // restore old state
    state.pos = oldPos;

    return labelEnd;
  };

  var unescapeAll$5 = utils$1.unescapeAll;


  var parse_link_destination$1 = function parseLinkDestination(str, pos, max) {
    var code, level,
        lines = 0,
        start = pos,
        result = {
          ok: false,
          pos: 0,
          lines: 0,
          str: ''
        };

    if (str.charCodeAt(pos) === 0x3C /* < */) {
      pos++;
      while (pos < max) {
        code = str.charCodeAt(pos);
        if (code === 0x0A /* \n */) { return result; }
        if (code === 0x3E /* > */) {
          result.pos = pos + 1;
          result.str = unescapeAll$5(str.slice(start + 1, pos));
          result.ok = true;
          return result;
        }
        if (code === 0x5C /* \ */ && pos + 1 < max) {
          pos += 2;
          continue;
        }

        pos++;
      }

      // no closing '>'
      return result;
    }

    // this should be ... } else { ... branch

    level = 0;
    while (pos < max) {
      code = str.charCodeAt(pos);

      if (code === 0x20) { break; }

      // ascii control characters
      if (code < 0x20 || code === 0x7F) { break; }

      if (code === 0x5C /* \ */ && pos + 1 < max) {
        pos += 2;
        continue;
      }

      if (code === 0x28 /* ( */) {
        level++;
      }

      if (code === 0x29 /* ) */) {
        if (level === 0) { break; }
        level--;
      }

      pos++;
    }

    if (start === pos) { return result; }
    if (level !== 0) { return result; }

    result.str = unescapeAll$5(str.slice(start, pos));
    result.lines = lines;
    result.pos = pos;
    result.ok = true;
    return result;
  };

  var unescapeAll$4 = utils$1.unescapeAll;


  var parse_link_title$1 = function parseLinkTitle(str, pos, max) {
    var code,
        marker,
        lines = 0,
        start = pos,
        result = {
          ok: false,
          pos: 0,
          lines: 0,
          str: ''
        };

    if (pos >= max) { return result; }

    marker = str.charCodeAt(pos);

    if (marker !== 0x22 /* " */ && marker !== 0x27 /* ' */ && marker !== 0x28 /* ( */) { return result; }

    pos++;

    // if opening marker is "(", switch it to closing marker ")"
    if (marker === 0x28) { marker = 0x29; }

    while (pos < max) {
      code = str.charCodeAt(pos);
      if (code === marker) {
        result.pos = pos + 1;
        result.lines = lines;
        result.str = unescapeAll$4(str.slice(start + 1, pos));
        result.ok = true;
        return result;
      } else if (code === 0x0A) {
        lines++;
      } else if (code === 0x5C /* \ */ && pos + 1 < max) {
        pos++;
        if (str.charCodeAt(pos) === 0x0A) {
          lines++;
        }
      }

      pos++;
    }

    return result;
  };

  var parseLinkLabel$1       = parse_link_label$1;
  var parseLinkDestination$1 = parse_link_destination$1;
  var parseLinkTitle$1       = parse_link_title$1;

  var helpers$1 = {
  	parseLinkLabel: parseLinkLabel$1,
  	parseLinkDestination: parseLinkDestination$1,
  	parseLinkTitle: parseLinkTitle$1
  };

  var assign$3          = utils$1.assign;
  var unescapeAll$3     = utils$1.unescapeAll;
  var escapeHtml$1      = utils$1.escapeHtml;


  ////////////////////////////////////////////////////////////////////////////////

  var default_rules$1 = {};


  default_rules$1.code_inline = function (tokens, idx, options, env, slf) {
    var token = tokens[idx];

    return  '<code' + slf.renderAttrs(token) + '>' +
            escapeHtml$1(tokens[idx].content) +
            '</code>';
  };


  default_rules$1.code_block = function (tokens, idx, options, env, slf) {
    var token = tokens[idx];

    return  '<pre' + slf.renderAttrs(token) + '><code>' +
            escapeHtml$1(tokens[idx].content) +
            '</code></pre>\n';
  };


  default_rules$1.fence = function (tokens, idx, options, env, slf) {
    var token = tokens[idx],
        info = token.info ? unescapeAll$3(token.info).trim() : '',
        langName = '',
        highlighted, i, tmpAttrs, tmpToken;

    if (info) {
      langName = info.split(/\s+/g)[0];
    }

    if (options.highlight) {
      highlighted = options.highlight(token.content, langName) || escapeHtml$1(token.content);
    } else {
      highlighted = escapeHtml$1(token.content);
    }

    if (highlighted.indexOf('<pre') === 0) {
      return highlighted + '\n';
    }

    // If language exists, inject class gently, without modifying original token.
    // May be, one day we will add .clone() for token and simplify this part, but
    // now we prefer to keep things local.
    if (info) {
      i        = token.attrIndex('class');
      tmpAttrs = token.attrs ? token.attrs.slice() : [];

      if (i < 0) {
        tmpAttrs.push([ 'class', options.langPrefix + langName ]);
      } else {
        tmpAttrs[i][1] += ' ' + options.langPrefix + langName;
      }

      // Fake token just to render attributes
      tmpToken = {
        attrs: tmpAttrs
      };

      return  '<pre><code' + slf.renderAttrs(tmpToken) + '>'
            + highlighted
            + '</code></pre>\n';
    }


    return  '<pre><code' + slf.renderAttrs(token) + '>'
          + highlighted
          + '</code></pre>\n';
  };


  default_rules$1.image = function (tokens, idx, options, env, slf) {
    var token = tokens[idx];

    // "alt" attr MUST be set, even if empty. Because it's mandatory and
    // should be placed on proper position for tests.
    //
    // Replace content with actual value

    token.attrs[token.attrIndex('alt')][1] =
      slf.renderInlineAsText(token.children, options, env);

    return slf.renderToken(tokens, idx, options);
  };


  default_rules$1.hardbreak = function (tokens, idx, options /*, env */) {
    return options.xhtmlOut ? '<br />\n' : '<br>\n';
  };
  default_rules$1.softbreak = function (tokens, idx, options /*, env */) {
    return options.breaks ? (options.xhtmlOut ? '<br />\n' : '<br>\n') : '\n';
  };


  default_rules$1.text = function (tokens, idx /*, options, env */) {
    return escapeHtml$1(tokens[idx].content);
  };


  default_rules$1.html_block = function (tokens, idx /*, options, env */) {
    return tokens[idx].content;
  };
  default_rules$1.html_inline = function (tokens, idx /*, options, env */) {
    return tokens[idx].content;
  };


  /**
   * new Renderer()
   *
   * Creates new [[Renderer]] instance and fill [[Renderer#rules]] with defaults.
   **/
  function Renderer$1() {

    /**
     * Renderer#rules -> Object
     *
     * Contains render rules for tokens. Can be updated and extended.
     *
     * ##### Example
     *
     * ```javascript
     * var md = require('markdown-it')();
     *
     * md.renderer.rules.strong_open  = function () { return '<b>'; };
     * md.renderer.rules.strong_close = function () { return '</b>'; };
     *
     * var result = md.renderInline(...);
     * ```
     *
     * Each rule is called as independent static function with fixed signature:
     *
     * ```javascript
     * function my_token_render(tokens, idx, options, env, renderer) {
     *   // ...
     *   return renderedHTML;
     * }
     * ```
     *
     * See [source code](https://github.com/markdown-it/markdown-it/blob/master/lib/renderer.js)
     * for more details and examples.
     **/
    this.rules = assign$3({}, default_rules$1);
  }


  /**
   * Renderer.renderAttrs(token) -> String
   *
   * Render token attributes to string.
   **/
  Renderer$1.prototype.renderAttrs = function renderAttrs(token) {
    var i, l, result;

    if (!token.attrs) { return ''; }

    result = '';

    for (i = 0, l = token.attrs.length; i < l; i++) {
      result += ' ' + escapeHtml$1(token.attrs[i][0]) + '="' + escapeHtml$1(token.attrs[i][1]) + '"';
    }

    return result;
  };


  /**
   * Renderer.renderToken(tokens, idx, options) -> String
   * - tokens (Array): list of tokens
   * - idx (Numbed): token index to render
   * - options (Object): params of parser instance
   *
   * Default token renderer. Can be overriden by custom function
   * in [[Renderer#rules]].
   **/
  Renderer$1.prototype.renderToken = function renderToken(tokens, idx, options) {
    var nextToken,
        result = '',
        needLf = false,
        token = tokens[idx];

    // Tight list paragraphs
    if (token.hidden) {
      return '';
    }

    // Insert a newline between hidden paragraph and subsequent opening
    // block-level tag.
    //
    // For example, here we should insert a newline before blockquote:
    //  - a
    //    >
    //
    if (token.block && token.nesting !== -1 && idx && tokens[idx - 1].hidden) {
      result += '\n';
    }

    // Add token name, e.g. `<img`
    result += (token.nesting === -1 ? '</' : '<') + token.tag;

    // Encode attributes, e.g. `<img src="foo"`
    result += this.renderAttrs(token);

    // Add a slash for self-closing tags, e.g. `<img src="foo" /`
    if (token.nesting === 0 && options.xhtmlOut) {
      result += ' /';
    }

    // Check if we need to add a newline after this tag
    if (token.block) {
      needLf = true;

      if (token.nesting === 1) {
        if (idx + 1 < tokens.length) {
          nextToken = tokens[idx + 1];

          if (nextToken.type === 'inline' || nextToken.hidden) {
            // Block-level tag containing an inline tag.
            //
            needLf = false;

          } else if (nextToken.nesting === -1 && nextToken.tag === token.tag) {
            // Opening tag + closing tag of the same type. E.g. `<li></li>`.
            //
            needLf = false;
          }
        }
      }
    }

    result += needLf ? '>\n' : '>';

    return result;
  };


  /**
   * Renderer.renderInline(tokens, options, env) -> String
   * - tokens (Array): list on block tokens to renter
   * - options (Object): params of parser instance
   * - env (Object): additional data from parsed input (references, for example)
   *
   * The same as [[Renderer.render]], but for single token of `inline` type.
   **/
  Renderer$1.prototype.renderInline = function (tokens, options, env) {
    var type,
        result = '',
        rules = this.rules;

    for (var i = 0, len = tokens.length; i < len; i++) {
      type = tokens[i].type;

      if (typeof rules[type] !== 'undefined') {
        result += rules[type](tokens, i, options, env, this);
      } else {
        result += this.renderToken(tokens, i, options);
      }
    }

    return result;
  };


  /** internal
   * Renderer.renderInlineAsText(tokens, options, env) -> String
   * - tokens (Array): list on block tokens to renter
   * - options (Object): params of parser instance
   * - env (Object): additional data from parsed input (references, for example)
   *
   * Special kludge for image `alt` attributes to conform CommonMark spec.
   * Don't try to use it! Spec requires to show `alt` content with stripped markup,
   * instead of simple escaping.
   **/
  Renderer$1.prototype.renderInlineAsText = function (tokens, options, env) {
    var result = '';

    for (var i = 0, len = tokens.length; i < len; i++) {
      if (tokens[i].type === 'text') {
        result += tokens[i].content;
      } else if (tokens[i].type === 'image') {
        result += this.renderInlineAsText(tokens[i].children, options, env);
      }
    }

    return result;
  };


  /**
   * Renderer.render(tokens, options, env) -> String
   * - tokens (Array): list on block tokens to renter
   * - options (Object): params of parser instance
   * - env (Object): additional data from parsed input (references, for example)
   *
   * Takes token stream and generates HTML. Probably, you will never need to call
   * this method directly.
   **/
  Renderer$1.prototype.render = function (tokens, options, env) {
    var i, len, type,
        result = '',
        rules = this.rules;

    for (i = 0, len = tokens.length; i < len; i++) {
      type = tokens[i].type;

      if (type === 'inline') {
        result += this.renderInline(tokens[i].children, options, env);
      } else if (typeof rules[type] !== 'undefined') {
        result += rules[tokens[i].type](tokens, i, options, env, this);
      } else {
        result += this.renderToken(tokens, i, options, env);
      }
    }

    return result;
  };

  var renderer$1 = Renderer$1;

  /**
   * class Ruler
   *
   * Helper class, used by [[MarkdownIt#core]], [[MarkdownIt#block]] and
   * [[MarkdownIt#inline]] to manage sequences of functions (rules):
   *
   * - keep rules in defined order
   * - assign the name to each rule
   * - enable/disable rules
   * - add/replace rules
   * - allow assign rules to additional named chains (in the same)
   * - cacheing lists of active rules
   *
   * You will not need use this class directly until write plugins. For simple
   * rules control use [[MarkdownIt.disable]], [[MarkdownIt.enable]] and
   * [[MarkdownIt.use]].
   **/


  /**
   * new Ruler()
   **/
  function Ruler$1() {
    // List of added rules. Each element is:
    //
    // {
    //   name: XXX,
    //   enabled: Boolean,
    //   fn: Function(),
    //   alt: [ name2, name3 ]
    // }
    //
    this.__rules__ = [];

    // Cached rule chains.
    //
    // First level - chain name, '' for default.
    // Second level - diginal anchor for fast filtering by charcodes.
    //
    this.__cache__ = null;
  }

  ////////////////////////////////////////////////////////////////////////////////
  // Helper methods, should not be used directly


  // Find rule index by name
  //
  Ruler$1.prototype.__find__ = function (name) {
    for (var i = 0; i < this.__rules__.length; i++) {
      if (this.__rules__[i].name === name) {
        return i;
      }
    }
    return -1;
  };


  // Build rules lookup cache
  //
  Ruler$1.prototype.__compile__ = function () {
    var self = this;
    var chains = [ '' ];

    // collect unique names
    self.__rules__.forEach(function (rule) {
      if (!rule.enabled) { return; }

      rule.alt.forEach(function (altName) {
        if (chains.indexOf(altName) < 0) {
          chains.push(altName);
        }
      });
    });

    self.__cache__ = {};

    chains.forEach(function (chain) {
      self.__cache__[chain] = [];
      self.__rules__.forEach(function (rule) {
        if (!rule.enabled) { return; }

        if (chain && rule.alt.indexOf(chain) < 0) { return; }

        self.__cache__[chain].push(rule.fn);
      });
    });
  };


  /**
   * Ruler.at(name, fn [, options])
   * - name (String): rule name to replace.
   * - fn (Function): new rule function.
   * - options (Object): new rule options (not mandatory).
   *
   * Replace rule by name with new function & options. Throws error if name not
   * found.
   *
   * ##### Options:
   *
   * - __alt__ - array with names of "alternate" chains.
   *
   * ##### Example
   *
   * Replace existing typographer replacement rule with new one:
   *
   * ```javascript
   * var md = require('markdown-it')();
   *
   * md.core.ruler.at('replacements', function replace(state) {
   *   //...
   * });
   * ```
   **/
  Ruler$1.prototype.at = function (name, fn, options) {
    var index = this.__find__(name);
    var opt = options || {};

    if (index === -1) { throw new Error('Parser rule not found: ' + name); }

    this.__rules__[index].fn = fn;
    this.__rules__[index].alt = opt.alt || [];
    this.__cache__ = null;
  };


  /**
   * Ruler.before(beforeName, ruleName, fn [, options])
   * - beforeName (String): new rule will be added before this one.
   * - ruleName (String): name of added rule.
   * - fn (Function): rule function.
   * - options (Object): rule options (not mandatory).
   *
   * Add new rule to chain before one with given name. See also
   * [[Ruler.after]], [[Ruler.push]].
   *
   * ##### Options:
   *
   * - __alt__ - array with names of "alternate" chains.
   *
   * ##### Example
   *
   * ```javascript
   * var md = require('markdown-it')();
   *
   * md.block.ruler.before('paragraph', 'my_rule', function replace(state) {
   *   //...
   * });
   * ```
   **/
  Ruler$1.prototype.before = function (beforeName, ruleName, fn, options) {
    var index = this.__find__(beforeName);
    var opt = options || {};

    if (index === -1) { throw new Error('Parser rule not found: ' + beforeName); }

    this.__rules__.splice(index, 0, {
      name: ruleName,
      enabled: true,
      fn: fn,
      alt: opt.alt || []
    });

    this.__cache__ = null;
  };


  /**
   * Ruler.after(afterName, ruleName, fn [, options])
   * - afterName (String): new rule will be added after this one.
   * - ruleName (String): name of added rule.
   * - fn (Function): rule function.
   * - options (Object): rule options (not mandatory).
   *
   * Add new rule to chain after one with given name. See also
   * [[Ruler.before]], [[Ruler.push]].
   *
   * ##### Options:
   *
   * - __alt__ - array with names of "alternate" chains.
   *
   * ##### Example
   *
   * ```javascript
   * var md = require('markdown-it')();
   *
   * md.inline.ruler.after('text', 'my_rule', function replace(state) {
   *   //...
   * });
   * ```
   **/
  Ruler$1.prototype.after = function (afterName, ruleName, fn, options) {
    var index = this.__find__(afterName);
    var opt = options || {};

    if (index === -1) { throw new Error('Parser rule not found: ' + afterName); }

    this.__rules__.splice(index + 1, 0, {
      name: ruleName,
      enabled: true,
      fn: fn,
      alt: opt.alt || []
    });

    this.__cache__ = null;
  };

  /**
   * Ruler.push(ruleName, fn [, options])
   * - ruleName (String): name of added rule.
   * - fn (Function): rule function.
   * - options (Object): rule options (not mandatory).
   *
   * Push new rule to the end of chain. See also
   * [[Ruler.before]], [[Ruler.after]].
   *
   * ##### Options:
   *
   * - __alt__ - array with names of "alternate" chains.
   *
   * ##### Example
   *
   * ```javascript
   * var md = require('markdown-it')();
   *
   * md.core.ruler.push('my_rule', function replace(state) {
   *   //...
   * });
   * ```
   **/
  Ruler$1.prototype.push = function (ruleName, fn, options) {
    var opt = options || {};

    this.__rules__.push({
      name: ruleName,
      enabled: true,
      fn: fn,
      alt: opt.alt || []
    });

    this.__cache__ = null;
  };


  /**
   * Ruler.enable(list [, ignoreInvalid]) -> Array
   * - list (String|Array): list of rule names to enable.
   * - ignoreInvalid (Boolean): set `true` to ignore errors when rule not found.
   *
   * Enable rules with given names. If any rule name not found - throw Error.
   * Errors can be disabled by second param.
   *
   * Returns list of found rule names (if no exception happened).
   *
   * See also [[Ruler.disable]], [[Ruler.enableOnly]].
   **/
  Ruler$1.prototype.enable = function (list, ignoreInvalid) {
    if (!Array.isArray(list)) { list = [ list ]; }

    var result = [];

    // Search by name and enable
    list.forEach(function (name) {
      var idx = this.__find__(name);

      if (idx < 0) {
        if (ignoreInvalid) { return; }
        throw new Error('Rules manager: invalid rule name ' + name);
      }
      this.__rules__[idx].enabled = true;
      result.push(name);
    }, this);

    this.__cache__ = null;
    return result;
  };


  /**
   * Ruler.enableOnly(list [, ignoreInvalid])
   * - list (String|Array): list of rule names to enable (whitelist).
   * - ignoreInvalid (Boolean): set `true` to ignore errors when rule not found.
   *
   * Enable rules with given names, and disable everything else. If any rule name
   * not found - throw Error. Errors can be disabled by second param.
   *
   * See also [[Ruler.disable]], [[Ruler.enable]].
   **/
  Ruler$1.prototype.enableOnly = function (list, ignoreInvalid) {
    if (!Array.isArray(list)) { list = [ list ]; }

    this.__rules__.forEach(function (rule) { rule.enabled = false; });

    this.enable(list, ignoreInvalid);
  };


  /**
   * Ruler.disable(list [, ignoreInvalid]) -> Array
   * - list (String|Array): list of rule names to disable.
   * - ignoreInvalid (Boolean): set `true` to ignore errors when rule not found.
   *
   * Disable rules with given names. If any rule name not found - throw Error.
   * Errors can be disabled by second param.
   *
   * Returns list of found rule names (if no exception happened).
   *
   * See also [[Ruler.enable]], [[Ruler.enableOnly]].
   **/
  Ruler$1.prototype.disable = function (list, ignoreInvalid) {
    if (!Array.isArray(list)) { list = [ list ]; }

    var result = [];

    // Search by name and disable
    list.forEach(function (name) {
      var idx = this.__find__(name);

      if (idx < 0) {
        if (ignoreInvalid) { return; }
        throw new Error('Rules manager: invalid rule name ' + name);
      }
      this.__rules__[idx].enabled = false;
      result.push(name);
    }, this);

    this.__cache__ = null;
    return result;
  };


  /**
   * Ruler.getRules(chainName) -> Array
   *
   * Return array of active functions (rules) for given chain name. It analyzes
   * rules configuration, compiles caches if not exists and returns result.
   *
   * Default chain name is `''` (empty string). It can't be skipped. That's
   * done intentionally, to keep signature monomorphic for high speed.
   **/
  Ruler$1.prototype.getRules = function (chainName) {
    if (this.__cache__ === null) {
      this.__compile__();
    }

    // Chain can be empty, if rules disabled. But we still have to return Array.
    return this.__cache__[chainName] || [];
  };

  var ruler$1 = Ruler$1;

  // Normalize input string


  // https://spec.commonmark.org/0.29/#line-ending
  var NEWLINES_RE$1  = /\r\n?|\n/g;
  var NULL_RE$1      = /\0/g;


  var normalize$1 = function normalize(state) {
    var str;

    // Normalize newlines
    str = state.src.replace(NEWLINES_RE$1, '\n');

    // Replace NULL characters
    str = str.replace(NULL_RE$1, '\uFFFD');

    state.src = str;
  };

  var block$2 = function block(state) {
    var token;

    if (state.inlineMode) {
      token          = new state.Token('inline', '', 0);
      token.content  = state.src;
      token.map      = [ 0, 1 ];
      token.children = [];
      state.tokens.push(token);
    } else {
      state.md.block.parse(state.src, state.md, state.env, state.tokens);
    }
  };

  var inline$1 = function inline(state) {
    var tokens = state.tokens, tok, i, l;

    // Parse inlines
    for (i = 0, l = tokens.length; i < l; i++) {
      tok = tokens[i];
      if (tok.type === 'inline') {
        state.md.inline.parse(tok.content, state.md, state.env, tok.children);
      }
    }
  };

  var arrayReplaceAt$1 = utils$1.arrayReplaceAt;


  function isLinkOpen$1(str) {
    return /^<a[>\s]/i.test(str);
  }
  function isLinkClose$1(str) {
    return /^<\/a\s*>/i.test(str);
  }


  var linkify$2 = function linkify(state) {
    var i, j, l, tokens, token, currentToken, nodes, ln, text, pos, lastPos,
        level, htmlLinkLevel, url, fullUrl, urlText,
        blockTokens = state.tokens,
        links;

    if (!state.md.options.linkify) { return; }

    for (j = 0, l = blockTokens.length; j < l; j++) {
      if (blockTokens[j].type !== 'inline' ||
          !state.md.linkify.pretest(blockTokens[j].content)) {
        continue;
      }

      tokens = blockTokens[j].children;

      htmlLinkLevel = 0;

      // We scan from the end, to keep position when new tags added.
      // Use reversed logic in links start/end match
      for (i = tokens.length - 1; i >= 0; i--) {
        currentToken = tokens[i];

        // Skip content of markdown links
        if (currentToken.type === 'link_close') {
          i--;
          while (tokens[i].level !== currentToken.level && tokens[i].type !== 'link_open') {
            i--;
          }
          continue;
        }

        // Skip content of html tag links
        if (currentToken.type === 'html_inline') {
          if (isLinkOpen$1(currentToken.content) && htmlLinkLevel > 0) {
            htmlLinkLevel--;
          }
          if (isLinkClose$1(currentToken.content)) {
            htmlLinkLevel++;
          }
        }
        if (htmlLinkLevel > 0) { continue; }

        if (currentToken.type === 'text' && state.md.linkify.test(currentToken.content)) {

          text = currentToken.content;
          links = state.md.linkify.match(text);

          // Now split string to nodes
          nodes = [];
          level = currentToken.level;
          lastPos = 0;

          for (ln = 0; ln < links.length; ln++) {

            url = links[ln].url;
            fullUrl = state.md.normalizeLink(url);
            if (!state.md.validateLink(fullUrl)) { continue; }

            urlText = links[ln].text;

            // Linkifier might send raw hostnames like "example.com", where url
            // starts with domain name. So we prepend http:// in those cases,
            // and remove it afterwards.
            //
            if (!links[ln].schema) {
              urlText = state.md.normalizeLinkText('http://' + urlText).replace(/^http:\/\//, '');
            } else if (links[ln].schema === 'mailto:' && !/^mailto:/i.test(urlText)) {
              urlText = state.md.normalizeLinkText('mailto:' + urlText).replace(/^mailto:/, '');
            } else {
              urlText = state.md.normalizeLinkText(urlText);
            }

            pos = links[ln].index;

            if (pos > lastPos) {
              token         = new state.Token('text', '', 0);
              token.content = text.slice(lastPos, pos);
              token.level   = level;
              nodes.push(token);
            }

            token         = new state.Token('link_open', 'a', 1);
            token.attrs   = [ [ 'href', fullUrl ] ];
            token.level   = level++;
            token.markup  = 'linkify';
            token.info    = 'auto';
            nodes.push(token);

            token         = new state.Token('text', '', 0);
            token.content = urlText;
            token.level   = level;
            nodes.push(token);

            token         = new state.Token('link_close', 'a', -1);
            token.level   = --level;
            token.markup  = 'linkify';
            token.info    = 'auto';
            nodes.push(token);

            lastPos = links[ln].lastIndex;
          }
          if (lastPos < text.length) {
            token         = new state.Token('text', '', 0);
            token.content = text.slice(lastPos);
            token.level   = level;
            nodes.push(token);
          }

          // replace current node
          blockTokens[j].children = tokens = arrayReplaceAt$1(tokens, i, nodes);
        }
      }
    }
  };

  // Simple typographic replacements

  // TODO:
  // - fractionals 1/2, 1/4, 3/4 -> ½, ¼, ¾
  // - miltiplication 2 x 4 -> 2 × 4

  var RARE_RE$1 = /\+-|\.\.|\?\?\?\?|!!!!|,,|--/;

  // Workaround for phantomjs - need regex without /g flag,
  // or root check will fail every second time
  var SCOPED_ABBR_TEST_RE$1 = /\((c|tm|r|p)\)/i;

  var SCOPED_ABBR_RE$1 = /\((c|tm|r|p)\)/ig;
  var SCOPED_ABBR$1 = {
    c: '©',
    r: '®',
    p: '§',
    tm: '™'
  };

  function replaceFn$1(match, name) {
    return SCOPED_ABBR$1[name.toLowerCase()];
  }

  function replace_scoped$1(inlineTokens) {
    var i, token, inside_autolink = 0;

    for (i = inlineTokens.length - 1; i >= 0; i--) {
      token = inlineTokens[i];

      if (token.type === 'text' && !inside_autolink) {
        token.content = token.content.replace(SCOPED_ABBR_RE$1, replaceFn$1);
      }

      if (token.type === 'link_open' && token.info === 'auto') {
        inside_autolink--;
      }

      if (token.type === 'link_close' && token.info === 'auto') {
        inside_autolink++;
      }
    }
  }

  function replace_rare$1(inlineTokens) {
    var i, token, inside_autolink = 0;

    for (i = inlineTokens.length - 1; i >= 0; i--) {
      token = inlineTokens[i];

      if (token.type === 'text' && !inside_autolink) {
        if (RARE_RE$1.test(token.content)) {
          token.content = token.content
            .replace(/\+-/g, '±')
            // .., ..., ....... -> …
            // but ?..... & !..... -> ?.. & !..
            .replace(/\.{2,}/g, '…').replace(/([?!])…/g, '$1..')
            .replace(/([?!]){4,}/g, '$1$1$1').replace(/,{2,}/g, ',')
            // em-dash
            .replace(/(^|[^-])---([^-]|$)/mg, '$1\u2014$2')
            // en-dash
            .replace(/(^|\s)--(\s|$)/mg, '$1\u2013$2')
            .replace(/(^|[^-\s])--([^-\s]|$)/mg, '$1\u2013$2');
        }
      }

      if (token.type === 'link_open' && token.info === 'auto') {
        inside_autolink--;
      }

      if (token.type === 'link_close' && token.info === 'auto') {
        inside_autolink++;
      }
    }
  }


  var replacements$1 = function replace(state) {
    var blkIdx;

    if (!state.md.options.typographer) { return; }

    for (blkIdx = state.tokens.length - 1; blkIdx >= 0; blkIdx--) {

      if (state.tokens[blkIdx].type !== 'inline') { continue; }

      if (SCOPED_ABBR_TEST_RE$1.test(state.tokens[blkIdx].content)) {
        replace_scoped$1(state.tokens[blkIdx].children);
      }

      if (RARE_RE$1.test(state.tokens[blkIdx].content)) {
        replace_rare$1(state.tokens[blkIdx].children);
      }

    }
  };

  var isWhiteSpace$3   = utils$1.isWhiteSpace;
  var isPunctChar$3    = utils$1.isPunctChar;
  var isMdAsciiPunct$3 = utils$1.isMdAsciiPunct;

  var QUOTE_TEST_RE$1 = /['"]/;
  var QUOTE_RE$1 = /['"]/g;
  var APOSTROPHE$1 = '\u2019'; /* ’ */


  function replaceAt$1(str, index, ch) {
    return str.substr(0, index) + ch + str.substr(index + 1);
  }

  function process_inlines$1(tokens, state) {
    var i, token, text, t, pos, max, thisLevel, item, lastChar, nextChar,
        isLastPunctChar, isNextPunctChar, isLastWhiteSpace, isNextWhiteSpace,
        canOpen, canClose, j, isSingle, stack, openQuote, closeQuote;

    stack = [];

    for (i = 0; i < tokens.length; i++) {
      token = tokens[i];

      thisLevel = tokens[i].level;

      for (j = stack.length - 1; j >= 0; j--) {
        if (stack[j].level <= thisLevel) { break; }
      }
      stack.length = j + 1;

      if (token.type !== 'text') { continue; }

      text = token.content;
      pos = 0;
      max = text.length;

      /*eslint no-labels:0,block-scoped-var:0*/
      OUTER:
      while (pos < max) {
        QUOTE_RE$1.lastIndex = pos;
        t = QUOTE_RE$1.exec(text);
        if (!t) { break; }

        canOpen = canClose = true;
        pos = t.index + 1;
        isSingle = (t[0] === "'");

        // Find previous character,
        // default to space if it's the beginning of the line
        //
        lastChar = 0x20;

        if (t.index - 1 >= 0) {
          lastChar = text.charCodeAt(t.index - 1);
        } else {
          for (j = i - 1; j >= 0; j--) {
            if (tokens[j].type === 'softbreak' || tokens[j].type === 'hardbreak') { break; } // lastChar defaults to 0x20
            if (tokens[j].type !== 'text') { continue; }

            lastChar = tokens[j].content.charCodeAt(tokens[j].content.length - 1);
            break;
          }
        }

        // Find next character,
        // default to space if it's the end of the line
        //
        nextChar = 0x20;

        if (pos < max) {
          nextChar = text.charCodeAt(pos);
        } else {
          for (j = i + 1; j < tokens.length; j++) {
            if (tokens[j].type === 'softbreak' || tokens[j].type === 'hardbreak') { break; } // nextChar defaults to 0x20
            if (tokens[j].type !== 'text') { continue; }

            nextChar = tokens[j].content.charCodeAt(0);
            break;
          }
        }

        isLastPunctChar = isMdAsciiPunct$3(lastChar) || isPunctChar$3(String.fromCharCode(lastChar));
        isNextPunctChar = isMdAsciiPunct$3(nextChar) || isPunctChar$3(String.fromCharCode(nextChar));

        isLastWhiteSpace = isWhiteSpace$3(lastChar);
        isNextWhiteSpace = isWhiteSpace$3(nextChar);

        if (isNextWhiteSpace) {
          canOpen = false;
        } else if (isNextPunctChar) {
          if (!(isLastWhiteSpace || isLastPunctChar)) {
            canOpen = false;
          }
        }

        if (isLastWhiteSpace) {
          canClose = false;
        } else if (isLastPunctChar) {
          if (!(isNextWhiteSpace || isNextPunctChar)) {
            canClose = false;
          }
        }

        if (nextChar === 0x22 /* " */ && t[0] === '"') {
          if (lastChar >= 0x30 /* 0 */ && lastChar <= 0x39 /* 9 */) {
            // special case: 1"" - count first quote as an inch
            canClose = canOpen = false;
          }
        }

        if (canOpen && canClose) {
          // treat this as the middle of the word
          canOpen = false;
          canClose = isNextPunctChar;
        }

        if (!canOpen && !canClose) {
          // middle of word
          if (isSingle) {
            token.content = replaceAt$1(token.content, t.index, APOSTROPHE$1);
          }
          continue;
        }

        if (canClose) {
          // this could be a closing quote, rewind the stack to get a match
          for (j = stack.length - 1; j >= 0; j--) {
            item = stack[j];
            if (stack[j].level < thisLevel) { break; }
            if (item.single === isSingle && stack[j].level === thisLevel) {
              item = stack[j];

              if (isSingle) {
                openQuote = state.md.options.quotes[2];
                closeQuote = state.md.options.quotes[3];
              } else {
                openQuote = state.md.options.quotes[0];
                closeQuote = state.md.options.quotes[1];
              }

              // replace token.content *before* tokens[item.token].content,
              // because, if they are pointing at the same token, replaceAt
              // could mess up indices when quote length != 1
              token.content = replaceAt$1(token.content, t.index, closeQuote);
              tokens[item.token].content = replaceAt$1(
                tokens[item.token].content, item.pos, openQuote);

              pos += closeQuote.length - 1;
              if (item.token === i) { pos += openQuote.length - 1; }

              text = token.content;
              max = text.length;

              stack.length = j;
              continue OUTER;
            }
          }
        }

        if (canOpen) {
          stack.push({
            token: i,
            pos: t.index,
            single: isSingle,
            level: thisLevel
          });
        } else if (canClose && isSingle) {
          token.content = replaceAt$1(token.content, t.index, APOSTROPHE$1);
        }
      }
    }
  }


  var smartquotes$1 = function smartquotes(state) {
    /*eslint max-depth:0*/
    var blkIdx;

    if (!state.md.options.typographer) { return; }

    for (blkIdx = state.tokens.length - 1; blkIdx >= 0; blkIdx--) {

      if (state.tokens[blkIdx].type !== 'inline' ||
          !QUOTE_TEST_RE$1.test(state.tokens[blkIdx].content)) {
        continue;
      }

      process_inlines$1(state.tokens[blkIdx].children, state);
    }
  };

  // Token class


  /**
   * class Token
   **/

  /**
   * new Token(type, tag, nesting)
   *
   * Create new token and fill passed properties.
   **/
  function Token$1(type, tag, nesting) {
    /**
     * Token#type -> String
     *
     * Type of the token (string, e.g. "paragraph_open")
     **/
    this.type     = type;

    /**
     * Token#tag -> String
     *
     * html tag name, e.g. "p"
     **/
    this.tag      = tag;

    /**
     * Token#attrs -> Array
     *
     * Html attributes. Format: `[ [ name1, value1 ], [ name2, value2 ] ]`
     **/
    this.attrs    = null;

    /**
     * Token#map -> Array
     *
     * Source map info. Format: `[ line_begin, line_end ]`
     **/
    this.map      = null;

    /**
     * Token#nesting -> Number
     *
     * Level change (number in {-1, 0, 1} set), where:
     *
     * -  `1` means the tag is opening
     * -  `0` means the tag is self-closing
     * - `-1` means the tag is closing
     **/
    this.nesting  = nesting;

    /**
     * Token#level -> Number
     *
     * nesting level, the same as `state.level`
     **/
    this.level    = 0;

    /**
     * Token#children -> Array
     *
     * An array of child nodes (inline and img tokens)
     **/
    this.children = null;

    /**
     * Token#content -> String
     *
     * In a case of self-closing tag (code, html, fence, etc.),
     * it has contents of this tag.
     **/
    this.content  = '';

    /**
     * Token#markup -> String
     *
     * '*' or '_' for emphasis, fence string for fence, etc.
     **/
    this.markup   = '';

    /**
     * Token#info -> String
     *
     * fence infostring
     **/
    this.info     = '';

    /**
     * Token#meta -> Object
     *
     * A place for plugins to store an arbitrary data
     **/
    this.meta     = null;

    /**
     * Token#block -> Boolean
     *
     * True for block-level tokens, false for inline tokens.
     * Used in renderer to calculate line breaks
     **/
    this.block    = false;

    /**
     * Token#hidden -> Boolean
     *
     * If it's true, ignore this element when rendering. Used for tight lists
     * to hide paragraphs.
     **/
    this.hidden   = false;
  }


  /**
   * Token.attrIndex(name) -> Number
   *
   * Search attribute index by name.
   **/
  Token$1.prototype.attrIndex = function attrIndex(name) {
    var attrs, i, len;

    if (!this.attrs) { return -1; }

    attrs = this.attrs;

    for (i = 0, len = attrs.length; i < len; i++) {
      if (attrs[i][0] === name) { return i; }
    }
    return -1;
  };


  /**
   * Token.attrPush(attrData)
   *
   * Add `[ name, value ]` attribute to list. Init attrs if necessary
   **/
  Token$1.prototype.attrPush = function attrPush(attrData) {
    if (this.attrs) {
      this.attrs.push(attrData);
    } else {
      this.attrs = [ attrData ];
    }
  };


  /**
   * Token.attrSet(name, value)
   *
   * Set `name` attribute to `value`. Override old value if exists.
   **/
  Token$1.prototype.attrSet = function attrSet(name, value) {
    var idx = this.attrIndex(name),
        attrData = [ name, value ];

    if (idx < 0) {
      this.attrPush(attrData);
    } else {
      this.attrs[idx] = attrData;
    }
  };


  /**
   * Token.attrGet(name)
   *
   * Get the value of attribute `name`, or null if it does not exist.
   **/
  Token$1.prototype.attrGet = function attrGet(name) {
    var idx = this.attrIndex(name), value = null;
    if (idx >= 0) {
      value = this.attrs[idx][1];
    }
    return value;
  };


  /**
   * Token.attrJoin(name, value)
   *
   * Join value to existing attribute via space. Or create new attribute if not
   * exists. Useful to operate with token classes.
   **/
  Token$1.prototype.attrJoin = function attrJoin(name, value) {
    var idx = this.attrIndex(name);

    if (idx < 0) {
      this.attrPush([ name, value ]);
    } else {
      this.attrs[idx][1] = this.attrs[idx][1] + ' ' + value;
    }
  };


  var token$1 = Token$1;

  function StateCore$1(src, md, env) {
    this.src = src;
    this.env = env;
    this.tokens = [];
    this.inlineMode = false;
    this.md = md; // link to parser instance
  }

  // re-export Token class to use in core rules
  StateCore$1.prototype.Token = token$1;


  var state_core$1 = StateCore$1;

  var _rules$5 = [
    [ 'normalize',      normalize$1      ],
    [ 'block',          block$2          ],
    [ 'inline',         inline$1         ],
    [ 'linkify',        linkify$2        ],
    [ 'replacements',   replacements$1   ],
    [ 'smartquotes',    smartquotes$1    ]
  ];


  /**
   * new Core()
   **/
  function Core$1() {
    /**
     * Core#ruler -> Ruler
     *
     * [[Ruler]] instance. Keep configuration of core rules.
     **/
    this.ruler = new ruler$1();

    for (var i = 0; i < _rules$5.length; i++) {
      this.ruler.push(_rules$5[i][0], _rules$5[i][1]);
    }
  }


  /**
   * Core.process(state)
   *
   * Executes core chain rules.
   **/
  Core$1.prototype.process = function (state) {
    var i, l, rules;

    rules = this.ruler.getRules('');

    for (i = 0, l = rules.length; i < l; i++) {
      rules[i](state);
    }
  };

  Core$1.prototype.State = state_core$1;


  var parser_core$1 = Core$1;

  var isSpace$o = utils$1.isSpace;


  function getLine$2(state, line) {
    var pos = state.bMarks[line] + state.blkIndent,
        max = state.eMarks[line];

    return state.src.substr(pos, max - pos);
  }

  function escapedSplit$2(str) {
    var result = [],
        pos = 0,
        max = str.length,
        ch,
        escapes = 0,
        lastPos = 0,
        backTicked = false,
        lastBackTick = 0;

    ch  = str.charCodeAt(pos);

    while (pos < max) {
      if (ch === 0x60/* ` */) {
        if (backTicked) {
          // make \` close code sequence, but not open it;
          // the reason is: `\` is correct code block
          backTicked = false;
          lastBackTick = pos;
        } else if (escapes % 2 === 0) {
          backTicked = true;
          lastBackTick = pos;
        }
      } else if (ch === 0x7c/* | */ && (escapes % 2 === 0) && !backTicked) {
        result.push(str.substring(lastPos, pos));
        lastPos = pos + 1;
      }

      if (ch === 0x5c/* \ */) {
        escapes++;
      } else {
        escapes = 0;
      }

      pos++;

      // If there was an un-closed backtick, go back to just after
      // the last backtick, but as if it was a normal character
      if (pos === max && backTicked) {
        backTicked = false;
        pos = lastBackTick + 1;
      }

      ch = str.charCodeAt(pos);
    }

    result.push(str.substring(lastPos));

    return result;
  }


  var table$2 = function table(state, startLine, endLine, silent) {
    var ch, lineText, pos, i, nextLine, columns, columnCount, token,
        aligns, t, tableLines, tbodyLines;

    // should have at least two lines
    if (startLine + 2 > endLine) { return false; }

    nextLine = startLine + 1;

    if (state.sCount[nextLine] < state.blkIndent) { return false; }

    // if it's indented more than 3 spaces, it should be a code block
    if (state.sCount[nextLine] - state.blkIndent >= 4) { return false; }

    // first character of the second line should be '|', '-', ':',
    // and no other characters are allowed but spaces;
    // basically, this is the equivalent of /^[-:|][-:|\s]*$/ regexp

    pos = state.bMarks[nextLine] + state.tShift[nextLine];
    if (pos >= state.eMarks[nextLine]) { return false; }

    ch = state.src.charCodeAt(pos++);
    if (ch !== 0x7C/* | */ && ch !== 0x2D/* - */ && ch !== 0x3A/* : */) { return false; }

    while (pos < state.eMarks[nextLine]) {
      ch = state.src.charCodeAt(pos);

      if (ch !== 0x7C/* | */ && ch !== 0x2D/* - */ && ch !== 0x3A/* : */ && !isSpace$o(ch)) { return false; }

      pos++;
    }

    lineText = getLine$2(state, startLine + 1);

    columns = lineText.split('|');
    aligns = [];
    for (i = 0; i < columns.length; i++) {
      t = columns[i].trim();
      if (!t) {
        // allow empty columns before and after table, but not in between columns;
        // e.g. allow ` |---| `, disallow ` ---||--- `
        if (i === 0 || i === columns.length - 1) {
          continue;
        } else {
          return false;
        }
      }

      if (!/^:?-+:?$/.test(t)) { return false; }
      if (t.charCodeAt(t.length - 1) === 0x3A/* : */) {
        aligns.push(t.charCodeAt(0) === 0x3A/* : */ ? 'center' : 'right');
      } else if (t.charCodeAt(0) === 0x3A/* : */) {
        aligns.push('left');
      } else {
        aligns.push('');
      }
    }

    lineText = getLine$2(state, startLine).trim();
    if (lineText.indexOf('|') === -1) { return false; }
    if (state.sCount[startLine] - state.blkIndent >= 4) { return false; }
    columns = escapedSplit$2(lineText.replace(/^\||\|$/g, ''));

    // header row will define an amount of columns in the entire table,
    // and align row shouldn't be smaller than that (the rest of the rows can)
    columnCount = columns.length;
    if (columnCount > aligns.length) { return false; }

    if (silent) { return true; }

    token     = state.push('table_open', 'table', 1);
    token.map = tableLines = [ startLine, 0 ];

    token     = state.push('thead_open', 'thead', 1);
    token.map = [ startLine, startLine + 1 ];

    token     = state.push('tr_open', 'tr', 1);
    token.map = [ startLine, startLine + 1 ];

    for (i = 0; i < columns.length; i++) {
      token          = state.push('th_open', 'th', 1);
      token.map      = [ startLine, startLine + 1 ];
      if (aligns[i]) {
        token.attrs  = [ [ 'style', 'text-align:' + aligns[i] ] ];
      }

      token          = state.push('inline', '', 0);
      token.content  = columns[i].trim();
      token.map      = [ startLine, startLine + 1 ];
      token.children = [];

      token          = state.push('th_close', 'th', -1);
    }

    token     = state.push('tr_close', 'tr', -1);
    token     = state.push('thead_close', 'thead', -1);

    token     = state.push('tbody_open', 'tbody', 1);
    token.map = tbodyLines = [ startLine + 2, 0 ];

    for (nextLine = startLine + 2; nextLine < endLine; nextLine++) {
      if (state.sCount[nextLine] < state.blkIndent) { break; }

      lineText = getLine$2(state, nextLine).trim();
      if (lineText.indexOf('|') === -1) { break; }
      if (state.sCount[nextLine] - state.blkIndent >= 4) { break; }
      columns = escapedSplit$2(lineText.replace(/^\||\|$/g, ''));

      token = state.push('tr_open', 'tr', 1);
      for (i = 0; i < columnCount; i++) {
        token          = state.push('td_open', 'td', 1);
        if (aligns[i]) {
          token.attrs  = [ [ 'style', 'text-align:' + aligns[i] ] ];
        }

        token          = state.push('inline', '', 0);
        token.content  = columns[i] ? columns[i].trim() : '';
        token.children = [];

        token          = state.push('td_close', 'td', -1);
      }
      token = state.push('tr_close', 'tr', -1);
    }
    token = state.push('tbody_close', 'tbody', -1);
    token = state.push('table_close', 'table', -1);

    tableLines[1] = tbodyLines[1] = nextLine;
    state.line = nextLine;
    return true;
  };

  // Code block (4 spaces padded)


  var code$2 = function code(state, startLine, endLine/*, silent*/) {
    var nextLine, last, token;

    if (state.sCount[startLine] - state.blkIndent < 4) { return false; }

    last = nextLine = startLine + 1;

    while (nextLine < endLine) {
      if (state.isEmpty(nextLine)) {
        nextLine++;
        continue;
      }

      if (state.sCount[nextLine] - state.blkIndent >= 4) {
        nextLine++;
        last = nextLine;
        continue;
      }
      break;
    }

    state.line = last;

    token         = state.push('code_block', 'code', 0);
    token.content = state.getLines(startLine, last, 4 + state.blkIndent, true);
    token.map     = [ startLine, state.line ];

    return true;
  };

  // fences (``` lang, ~~~ lang)


  var fence$2 = function fence(state, startLine, endLine, silent) {
    var marker, len, params, nextLine, mem, token, markup,
        haveEndMarker = false,
        pos = state.bMarks[startLine] + state.tShift[startLine],
        max = state.eMarks[startLine];

    // if it's indented more than 3 spaces, it should be a code block
    if (state.sCount[startLine] - state.blkIndent >= 4) { return false; }

    if (pos + 3 > max) { return false; }

    marker = state.src.charCodeAt(pos);

    if (marker !== 0x7E/* ~ */ && marker !== 0x60 /* ` */) {
      return false;
    }

    // scan marker length
    mem = pos;
    pos = state.skipChars(pos, marker);

    len = pos - mem;

    if (len < 3) { return false; }

    markup = state.src.slice(mem, pos);
    params = state.src.slice(pos, max);

    if (marker === 0x60 /* ` */) {
      if (params.indexOf(String.fromCharCode(marker)) >= 0) {
        return false;
      }
    }

    // Since start is found, we can report success here in validation mode
    if (silent) { return true; }

    // search end of block
    nextLine = startLine;

    for (;;) {
      nextLine++;
      if (nextLine >= endLine) {
        // unclosed block should be autoclosed by end of document.
        // also block seems to be autoclosed by end of parent
        break;
      }

      pos = mem = state.bMarks[nextLine] + state.tShift[nextLine];
      max = state.eMarks[nextLine];

      if (pos < max && state.sCount[nextLine] < state.blkIndent) {
        // non-empty line with negative indent should stop the list:
        // - ```
        //  test
        break;
      }

      if (state.src.charCodeAt(pos) !== marker) { continue; }

      if (state.sCount[nextLine] - state.blkIndent >= 4) {
        // closing fence should be indented less than 4 spaces
        continue;
      }

      pos = state.skipChars(pos, marker);

      // closing code fence must be at least as long as the opening one
      if (pos - mem < len) { continue; }

      // make sure tail has spaces only
      pos = state.skipSpaces(pos);

      if (pos < max) { continue; }

      haveEndMarker = true;
      // found!
      break;
    }

    // If a fence has heading spaces, they should be removed from its inner block
    len = state.sCount[startLine];

    state.line = nextLine + (haveEndMarker ? 1 : 0);

    token         = state.push('fence', 'code', 0);
    token.info    = params;
    token.content = state.getLines(startLine + 1, nextLine, len, true);
    token.markup  = markup;
    token.map     = [ startLine, state.line ];

    return true;
  };

  var isSpace$n = utils$1.isSpace;


  var blockquote$2 = function blockquote(state, startLine, endLine, silent) {
    var adjustTab,
        ch,
        i,
        initial,
        l,
        lastLineEmpty,
        lines,
        nextLine,
        offset,
        oldBMarks,
        oldBSCount,
        oldIndent,
        oldParentType,
        oldSCount,
        oldTShift,
        spaceAfterMarker,
        terminate,
        terminatorRules,
        token,
        wasOutdented,
        oldLineMax = state.lineMax,
        pos = state.bMarks[startLine] + state.tShift[startLine],
        max = state.eMarks[startLine];

    // if it's indented more than 3 spaces, it should be a code block
    if (state.sCount[startLine] - state.blkIndent >= 4) { return false; }

    // check the block quote marker
    if (state.src.charCodeAt(pos++) !== 0x3E/* > */) { return false; }

    // we know that it's going to be a valid blockquote,
    // so no point trying to find the end of it in silent mode
    if (silent) { return true; }

    // skip spaces after ">" and re-calculate offset
    initial = offset = state.sCount[startLine] + pos - (state.bMarks[startLine] + state.tShift[startLine]);

    // skip one optional space after '>'
    if (state.src.charCodeAt(pos) === 0x20 /* space */) {
      // ' >   test '
      //     ^ -- position start of line here:
      pos++;
      initial++;
      offset++;
      adjustTab = false;
      spaceAfterMarker = true;
    } else if (state.src.charCodeAt(pos) === 0x09 /* tab */) {
      spaceAfterMarker = true;

      if ((state.bsCount[startLine] + offset) % 4 === 3) {
        // '  >\t  test '
        //       ^ -- position start of line here (tab has width===1)
        pos++;
        initial++;
        offset++;
        adjustTab = false;
      } else {
        // ' >\t  test '
        //    ^ -- position start of line here + shift bsCount slightly
        //         to make extra space appear
        adjustTab = true;
      }
    } else {
      spaceAfterMarker = false;
    }

    oldBMarks = [ state.bMarks[startLine] ];
    state.bMarks[startLine] = pos;

    while (pos < max) {
      ch = state.src.charCodeAt(pos);

      if (isSpace$n(ch)) {
        if (ch === 0x09) {
          offset += 4 - (offset + state.bsCount[startLine] + (adjustTab ? 1 : 0)) % 4;
        } else {
          offset++;
        }
      } else {
        break;
      }

      pos++;
    }

    oldBSCount = [ state.bsCount[startLine] ];
    state.bsCount[startLine] = state.sCount[startLine] + 1 + (spaceAfterMarker ? 1 : 0);

    lastLineEmpty = pos >= max;

    oldSCount = [ state.sCount[startLine] ];
    state.sCount[startLine] = offset - initial;

    oldTShift = [ state.tShift[startLine] ];
    state.tShift[startLine] = pos - state.bMarks[startLine];

    terminatorRules = state.md.block.ruler.getRules('blockquote');

    oldParentType = state.parentType;
    state.parentType = 'blockquote';
    wasOutdented = false;

    // Search the end of the block
    //
    // Block ends with either:
    //  1. an empty line outside:
    //     ```
    //     > test
    //
    //     ```
    //  2. an empty line inside:
    //     ```
    //     >
    //     test
    //     ```
    //  3. another tag:
    //     ```
    //     > test
    //      - - -
    //     ```
    for (nextLine = startLine + 1; nextLine < endLine; nextLine++) {
      // check if it's outdented, i.e. it's inside list item and indented
      // less than said list item:
      //
      // ```
      // 1. anything
      //    > current blockquote
      // 2. checking this line
      // ```
      if (state.sCount[nextLine] < state.blkIndent) { wasOutdented = true; }

      pos = state.bMarks[nextLine] + state.tShift[nextLine];
      max = state.eMarks[nextLine];

      if (pos >= max) {
        // Case 1: line is not inside the blockquote, and this line is empty.
        break;
      }

      if (state.src.charCodeAt(pos++) === 0x3E/* > */ && !wasOutdented) {
        // This line is inside the blockquote.

        // skip spaces after ">" and re-calculate offset
        initial = offset = state.sCount[nextLine] + pos - (state.bMarks[nextLine] + state.tShift[nextLine]);

        // skip one optional space after '>'
        if (state.src.charCodeAt(pos) === 0x20 /* space */) {
          // ' >   test '
          //     ^ -- position start of line here:
          pos++;
          initial++;
          offset++;
          adjustTab = false;
          spaceAfterMarker = true;
        } else if (state.src.charCodeAt(pos) === 0x09 /* tab */) {
          spaceAfterMarker = true;

          if ((state.bsCount[nextLine] + offset) % 4 === 3) {
            // '  >\t  test '
            //       ^ -- position start of line here (tab has width===1)
            pos++;
            initial++;
            offset++;
            adjustTab = false;
          } else {
            // ' >\t  test '
            //    ^ -- position start of line here + shift bsCount slightly
            //         to make extra space appear
            adjustTab = true;
          }
        } else {
          spaceAfterMarker = false;
        }

        oldBMarks.push(state.bMarks[nextLine]);
        state.bMarks[nextLine] = pos;

        while (pos < max) {
          ch = state.src.charCodeAt(pos);

          if (isSpace$n(ch)) {
            if (ch === 0x09) {
              offset += 4 - (offset + state.bsCount[nextLine] + (adjustTab ? 1 : 0)) % 4;
            } else {
              offset++;
            }
          } else {
            break;
          }

          pos++;
        }

        lastLineEmpty = pos >= max;

        oldBSCount.push(state.bsCount[nextLine]);
        state.bsCount[nextLine] = state.sCount[nextLine] + 1 + (spaceAfterMarker ? 1 : 0);

        oldSCount.push(state.sCount[nextLine]);
        state.sCount[nextLine] = offset - initial;

        oldTShift.push(state.tShift[nextLine]);
        state.tShift[nextLine] = pos - state.bMarks[nextLine];
        continue;
      }

      // Case 2: line is not inside the blockquote, and the last line was empty.
      if (lastLineEmpty) { break; }

      // Case 3: another tag found.
      terminate = false;
      for (i = 0, l = terminatorRules.length; i < l; i++) {
        if (terminatorRules[i](state, nextLine, endLine, true)) {
          terminate = true;
          break;
        }
      }

      if (terminate) {
        // Quirk to enforce "hard termination mode" for paragraphs;
        // normally if you call `tokenize(state, startLine, nextLine)`,
        // paragraphs will look below nextLine for paragraph continuation,
        // but if blockquote is terminated by another tag, they shouldn't
        state.lineMax = nextLine;

        if (state.blkIndent !== 0) {
          // state.blkIndent was non-zero, we now set it to zero,
          // so we need to re-calculate all offsets to appear as
          // if indent wasn't changed
          oldBMarks.push(state.bMarks[nextLine]);
          oldBSCount.push(state.bsCount[nextLine]);
          oldTShift.push(state.tShift[nextLine]);
          oldSCount.push(state.sCount[nextLine]);
          state.sCount[nextLine] -= state.blkIndent;
        }

        break;
      }

      oldBMarks.push(state.bMarks[nextLine]);
      oldBSCount.push(state.bsCount[nextLine]);
      oldTShift.push(state.tShift[nextLine]);
      oldSCount.push(state.sCount[nextLine]);

      // A negative indentation means that this is a paragraph continuation
      //
      state.sCount[nextLine] = -1;
    }

    oldIndent = state.blkIndent;
    state.blkIndent = 0;

    token        = state.push('blockquote_open', 'blockquote', 1);
    token.markup = '>';
    token.map    = lines = [ startLine, 0 ];

    state.md.block.tokenize(state, startLine, nextLine);

    token        = state.push('blockquote_close', 'blockquote', -1);
    token.markup = '>';

    state.lineMax = oldLineMax;
    state.parentType = oldParentType;
    lines[1] = state.line;

    // Restore original tShift; this might not be necessary since the parser
    // has already been here, but just to make sure we can do that.
    for (i = 0; i < oldTShift.length; i++) {
      state.bMarks[i + startLine] = oldBMarks[i];
      state.tShift[i + startLine] = oldTShift[i];
      state.sCount[i + startLine] = oldSCount[i];
      state.bsCount[i + startLine] = oldBSCount[i];
    }
    state.blkIndent = oldIndent;

    return true;
  };

  var isSpace$m = utils$1.isSpace;


  var hr$1 = function hr(state, startLine, endLine, silent) {
    var marker, cnt, ch, token,
        pos = state.bMarks[startLine] + state.tShift[startLine],
        max = state.eMarks[startLine];

    // if it's indented more than 3 spaces, it should be a code block
    if (state.sCount[startLine] - state.blkIndent >= 4) { return false; }

    marker = state.src.charCodeAt(pos++);

    // Check hr marker
    if (marker !== 0x2A/* * */ &&
        marker !== 0x2D/* - */ &&
        marker !== 0x5F/* _ */) {
      return false;
    }

    // markers can be mixed with spaces, but there should be at least 3 of them

    cnt = 1;
    while (pos < max) {
      ch = state.src.charCodeAt(pos++);
      if (ch !== marker && !isSpace$m(ch)) { return false; }
      if (ch === marker) { cnt++; }
    }

    if (cnt < 3) { return false; }

    if (silent) { return true; }

    state.line = startLine + 1;

    token        = state.push('hr', 'hr', 0);
    token.map    = [ startLine, state.line ];
    token.markup = Array(cnt + 1).join(String.fromCharCode(marker));

    return true;
  };

  var isSpace$l = utils$1.isSpace;


  // Search `[-+*][\n ]`, returns next pos after marker on success
  // or -1 on fail.
  function skipBulletListMarker$1(state, startLine) {
    var marker, pos, max, ch;

    pos = state.bMarks[startLine] + state.tShift[startLine];
    max = state.eMarks[startLine];

    marker = state.src.charCodeAt(pos++);
    // Check bullet
    if (marker !== 0x2A/* * */ &&
        marker !== 0x2D/* - */ &&
        marker !== 0x2B/* + */) {
      return -1;
    }

    if (pos < max) {
      ch = state.src.charCodeAt(pos);

      if (!isSpace$l(ch)) {
        // " -test " - is not a list item
        return -1;
      }
    }

    return pos;
  }

  // Search `\d+[.)][\n ]`, returns next pos after marker on success
  // or -1 on fail.
  function skipOrderedListMarker$1(state, startLine) {
    var ch,
        start = state.bMarks[startLine] + state.tShift[startLine],
        pos = start,
        max = state.eMarks[startLine];

    // List marker should have at least 2 chars (digit + dot)
    if (pos + 1 >= max) { return -1; }

    ch = state.src.charCodeAt(pos++);

    if (ch < 0x30/* 0 */ || ch > 0x39/* 9 */) { return -1; }

    for (;;) {
      // EOL -> fail
      if (pos >= max) { return -1; }

      ch = state.src.charCodeAt(pos++);

      if (ch >= 0x30/* 0 */ && ch <= 0x39/* 9 */) {

        // List marker should have no more than 9 digits
        // (prevents integer overflow in browsers)
        if (pos - start >= 10) { return -1; }

        continue;
      }

      // found valid marker
      if (ch === 0x29/* ) */ || ch === 0x2e/* . */) {
        break;
      }

      return -1;
    }


    if (pos < max) {
      ch = state.src.charCodeAt(pos);

      if (!isSpace$l(ch)) {
        // " 1.test " - is not a list item
        return -1;
      }
    }
    return pos;
  }

  function markTightParagraphs$1(state, idx) {
    var i, l,
        level = state.level + 2;

    for (i = idx + 2, l = state.tokens.length - 2; i < l; i++) {
      if (state.tokens[i].level === level && state.tokens[i].type === 'paragraph_open') {
        state.tokens[i + 2].hidden = true;
        state.tokens[i].hidden = true;
        i += 2;
      }
    }
  }


  var list$1 = function list(state, startLine, endLine, silent) {
    var ch,
        contentStart,
        i,
        indent,
        indentAfterMarker,
        initial,
        isOrdered,
        itemLines,
        l,
        listLines,
        listTokIdx,
        markerCharCode,
        markerValue,
        max,
        nextLine,
        offset,
        oldListIndent,
        oldParentType,
        oldSCount,
        oldTShift,
        oldTight,
        pos,
        posAfterMarker,
        prevEmptyEnd,
        start,
        terminate,
        terminatorRules,
        token,
        isTerminatingParagraph = false,
        tight = true;

    // if it's indented more than 3 spaces, it should be a code block
    if (state.sCount[startLine] - state.blkIndent >= 4) { return false; }

    // Special case:
    //  - item 1
    //   - item 2
    //    - item 3
    //     - item 4
    //      - this one is a paragraph continuation
    if (state.listIndent >= 0 &&
        state.sCount[startLine] - state.listIndent >= 4 &&
        state.sCount[startLine] < state.blkIndent) {
      return false;
    }

    // limit conditions when list can interrupt
    // a paragraph (validation mode only)
    if (silent && state.parentType === 'paragraph') {
      // Next list item should still terminate previous list item;
      //
      // This code can fail if plugins use blkIndent as well as lists,
      // but I hope the spec gets fixed long before that happens.
      //
      if (state.tShift[startLine] >= state.blkIndent) {
        isTerminatingParagraph = true;
      }
    }

    // Detect list type and position after marker
    if ((posAfterMarker = skipOrderedListMarker$1(state, startLine)) >= 0) {
      isOrdered = true;
      start = state.bMarks[startLine] + state.tShift[startLine];
      markerValue = Number(state.src.substr(start, posAfterMarker - start - 1));

      // If we're starting a new ordered list right after
      // a paragraph, it should start with 1.
      if (isTerminatingParagraph && markerValue !== 1) { return false; }

    } else if ((posAfterMarker = skipBulletListMarker$1(state, startLine)) >= 0) {
      isOrdered = false;

    } else {
      return false;
    }

    // If we're starting a new unordered list right after
    // a paragraph, first line should not be empty.
    if (isTerminatingParagraph) {
      if (state.skipSpaces(posAfterMarker) >= state.eMarks[startLine]) { return false; }
    }

    // We should terminate list on style change. Remember first one to compare.
    markerCharCode = state.src.charCodeAt(posAfterMarker - 1);

    // For validation mode we can terminate immediately
    if (silent) { return true; }

    // Start list
    listTokIdx = state.tokens.length;

    if (isOrdered) {
      token       = state.push('ordered_list_open', 'ol', 1);
      if (markerValue !== 1) {
        token.attrs = [ [ 'start', markerValue ] ];
      }

    } else {
      token       = state.push('bullet_list_open', 'ul', 1);
    }

    token.map    = listLines = [ startLine, 0 ];
    token.markup = String.fromCharCode(markerCharCode);

    //
    // Iterate list items
    //

    nextLine = startLine;
    prevEmptyEnd = false;
    terminatorRules = state.md.block.ruler.getRules('list');

    oldParentType = state.parentType;
    state.parentType = 'list';

    while (nextLine < endLine) {
      pos = posAfterMarker;
      max = state.eMarks[nextLine];

      initial = offset = state.sCount[nextLine] + posAfterMarker - (state.bMarks[startLine] + state.tShift[startLine]);

      while (pos < max) {
        ch = state.src.charCodeAt(pos);

        if (ch === 0x09) {
          offset += 4 - (offset + state.bsCount[nextLine]) % 4;
        } else if (ch === 0x20) {
          offset++;
        } else {
          break;
        }

        pos++;
      }

      contentStart = pos;

      if (contentStart >= max) {
        // trimming space in "-    \n  3" case, indent is 1 here
        indentAfterMarker = 1;
      } else {
        indentAfterMarker = offset - initial;
      }

      // If we have more than 4 spaces, the indent is 1
      // (the rest is just indented code block)
      if (indentAfterMarker > 4) { indentAfterMarker = 1; }

      // "  -  test"
      //  ^^^^^ - calculating total length of this thing
      indent = initial + indentAfterMarker;

      // Run subparser & write tokens
      token        = state.push('list_item_open', 'li', 1);
      token.markup = String.fromCharCode(markerCharCode);
      token.map    = itemLines = [ startLine, 0 ];

      // change current state, then restore it after parser subcall
      oldTight = state.tight;
      oldTShift = state.tShift[startLine];
      oldSCount = state.sCount[startLine];

      //  - example list
      // ^ listIndent position will be here
      //   ^ blkIndent position will be here
      //
      oldListIndent = state.listIndent;
      state.listIndent = state.blkIndent;
      state.blkIndent = indent;

      state.tight = true;
      state.tShift[startLine] = contentStart - state.bMarks[startLine];
      state.sCount[startLine] = offset;

      if (contentStart >= max && state.isEmpty(startLine + 1)) {
        // workaround for this case
        // (list item is empty, list terminates before "foo"):
        // ~~~~~~~~
        //   -
        //
        //     foo
        // ~~~~~~~~
        state.line = Math.min(state.line + 2, endLine);
      } else {
        state.md.block.tokenize(state, startLine, endLine, true);
      }

      // If any of list item is tight, mark list as tight
      if (!state.tight || prevEmptyEnd) {
        tight = false;
      }
      // Item become loose if finish with empty line,
      // but we should filter last element, because it means list finish
      prevEmptyEnd = (state.line - startLine) > 1 && state.isEmpty(state.line - 1);

      state.blkIndent = state.listIndent;
      state.listIndent = oldListIndent;
      state.tShift[startLine] = oldTShift;
      state.sCount[startLine] = oldSCount;
      state.tight = oldTight;

      token        = state.push('list_item_close', 'li', -1);
      token.markup = String.fromCharCode(markerCharCode);

      nextLine = startLine = state.line;
      itemLines[1] = nextLine;
      contentStart = state.bMarks[startLine];

      if (nextLine >= endLine) { break; }

      //
      // Try to check if list is terminated or continued.
      //
      if (state.sCount[nextLine] < state.blkIndent) { break; }

      // if it's indented more than 3 spaces, it should be a code block
      if (state.sCount[startLine] - state.blkIndent >= 4) { break; }

      // fail if terminating block found
      terminate = false;
      for (i = 0, l = terminatorRules.length; i < l; i++) {
        if (terminatorRules[i](state, nextLine, endLine, true)) {
          terminate = true;
          break;
        }
      }
      if (terminate) { break; }

      // fail if list has another type
      if (isOrdered) {
        posAfterMarker = skipOrderedListMarker$1(state, nextLine);
        if (posAfterMarker < 0) { break; }
      } else {
        posAfterMarker = skipBulletListMarker$1(state, nextLine);
        if (posAfterMarker < 0) { break; }
      }

      if (markerCharCode !== state.src.charCodeAt(posAfterMarker - 1)) { break; }
    }

    // Finalize list
    if (isOrdered) {
      token = state.push('ordered_list_close', 'ol', -1);
    } else {
      token = state.push('bullet_list_close', 'ul', -1);
    }
    token.markup = String.fromCharCode(markerCharCode);

    listLines[1] = nextLine;
    state.line = nextLine;

    state.parentType = oldParentType;

    // mark paragraphs tight if needed
    if (tight) {
      markTightParagraphs$1(state, listTokIdx);
    }

    return true;
  };

  var normalizeReference$5   = utils$1.normalizeReference;
  var isSpace$k              = utils$1.isSpace;


  var reference$1 = function reference(state, startLine, _endLine, silent) {
    var ch,
        destEndPos,
        destEndLineNo,
        endLine,
        href,
        i,
        l,
        label,
        labelEnd,
        oldParentType,
        res,
        start,
        str,
        terminate,
        terminatorRules,
        title,
        lines = 0,
        pos = state.bMarks[startLine] + state.tShift[startLine],
        max = state.eMarks[startLine],
        nextLine = startLine + 1;

    // if it's indented more than 3 spaces, it should be a code block
    if (state.sCount[startLine] - state.blkIndent >= 4) { return false; }

    if (state.src.charCodeAt(pos) !== 0x5B/* [ */) { return false; }

    // Simple check to quickly interrupt scan on [link](url) at the start of line.
    // Can be useful on practice: https://github.com/markdown-it/markdown-it/issues/54
    while (++pos < max) {
      if (state.src.charCodeAt(pos) === 0x5D /* ] */ &&
          state.src.charCodeAt(pos - 1) !== 0x5C/* \ */) {
        if (pos + 1 === max) { return false; }
        if (state.src.charCodeAt(pos + 1) !== 0x3A/* : */) { return false; }
        break;
      }
    }

    endLine = state.lineMax;

    // jump line-by-line until empty one or EOF
    terminatorRules = state.md.block.ruler.getRules('reference');

    oldParentType = state.parentType;
    state.parentType = 'reference';

    for (; nextLine < endLine && !state.isEmpty(nextLine); nextLine++) {
      // this would be a code block normally, but after paragraph
      // it's considered a lazy continuation regardless of what's there
      if (state.sCount[nextLine] - state.blkIndent > 3) { continue; }

      // quirk for blockquotes, this line should already be checked by that rule
      if (state.sCount[nextLine] < 0) { continue; }

      // Some tags can terminate paragraph without empty line.
      terminate = false;
      for (i = 0, l = terminatorRules.length; i < l; i++) {
        if (terminatorRules[i](state, nextLine, endLine, true)) {
          terminate = true;
          break;
        }
      }
      if (terminate) { break; }
    }

    str = state.getLines(startLine, nextLine, state.blkIndent, false).trim();
    max = str.length;

    for (pos = 1; pos < max; pos++) {
      ch = str.charCodeAt(pos);
      if (ch === 0x5B /* [ */) {
        return false;
      } else if (ch === 0x5D /* ] */) {
        labelEnd = pos;
        break;
      } else if (ch === 0x0A /* \n */) {
        lines++;
      } else if (ch === 0x5C /* \ */) {
        pos++;
        if (pos < max && str.charCodeAt(pos) === 0x0A) {
          lines++;
        }
      }
    }

    if (labelEnd < 0 || str.charCodeAt(labelEnd + 1) !== 0x3A/* : */) { return false; }

    // [label]:   destination   'title'
    //         ^^^ skip optional whitespace here
    for (pos = labelEnd + 2; pos < max; pos++) {
      ch = str.charCodeAt(pos);
      if (ch === 0x0A) {
        lines++;
      } else if (isSpace$k(ch)) ; else {
        break;
      }
    }

    // [label]:   destination   'title'
    //            ^^^^^^^^^^^ parse this
    res = state.md.helpers.parseLinkDestination(str, pos, max);
    if (!res.ok) { return false; }

    href = state.md.normalizeLink(res.str);
    if (!state.md.validateLink(href)) { return false; }

    pos = res.pos;
    lines += res.lines;

    // save cursor state, we could require to rollback later
    destEndPos = pos;
    destEndLineNo = lines;

    // [label]:   destination   'title'
    //                       ^^^ skipping those spaces
    start = pos;
    for (; pos < max; pos++) {
      ch = str.charCodeAt(pos);
      if (ch === 0x0A) {
        lines++;
      } else if (isSpace$k(ch)) ; else {
        break;
      }
    }

    // [label]:   destination   'title'
    //                          ^^^^^^^ parse this
    res = state.md.helpers.parseLinkTitle(str, pos, max);
    if (pos < max && start !== pos && res.ok) {
      title = res.str;
      pos = res.pos;
      lines += res.lines;
    } else {
      title = '';
      pos = destEndPos;
      lines = destEndLineNo;
    }

    // skip trailing spaces until the rest of the line
    while (pos < max) {
      ch = str.charCodeAt(pos);
      if (!isSpace$k(ch)) { break; }
      pos++;
    }

    if (pos < max && str.charCodeAt(pos) !== 0x0A) {
      if (title) {
        // garbage at the end of the line after title,
        // but it could still be a valid reference if we roll back
        title = '';
        pos = destEndPos;
        lines = destEndLineNo;
        while (pos < max) {
          ch = str.charCodeAt(pos);
          if (!isSpace$k(ch)) { break; }
          pos++;
        }
      }
    }

    if (pos < max && str.charCodeAt(pos) !== 0x0A) {
      // garbage at the end of the line
      return false;
    }

    label = normalizeReference$5(str.slice(1, labelEnd));
    if (!label) {
      // CommonMark 0.20 disallows empty labels
      return false;
    }

    // Reference can not terminate anything. This check is for safety only.
    /*istanbul ignore if*/
    if (silent) { return true; }

    if (typeof state.env.references === 'undefined') {
      state.env.references = {};
    }
    if (typeof state.env.references[label] === 'undefined') {
      state.env.references[label] = { title: title, href: href };
    }

    state.parentType = oldParentType;

    state.line = startLine + lines + 1;
    return true;
  };

  var isSpace$j = utils$1.isSpace;


  var heading$2 = function heading(state, startLine, endLine, silent) {
    var ch, level, tmp, token,
        pos = state.bMarks[startLine] + state.tShift[startLine],
        max = state.eMarks[startLine];

    // if it's indented more than 3 spaces, it should be a code block
    if (state.sCount[startLine] - state.blkIndent >= 4) { return false; }

    ch  = state.src.charCodeAt(pos);

    if (ch !== 0x23/* # */ || pos >= max) { return false; }

    // count heading level
    level = 1;
    ch = state.src.charCodeAt(++pos);
    while (ch === 0x23/* # */ && pos < max && level <= 6) {
      level++;
      ch = state.src.charCodeAt(++pos);
    }

    if (level > 6 || (pos < max && !isSpace$j(ch))) { return false; }

    if (silent) { return true; }

    // Let's cut tails like '    ###  ' from the end of string

    max = state.skipSpacesBack(max, pos);
    tmp = state.skipCharsBack(max, 0x23, pos); // #
    if (tmp > pos && isSpace$j(state.src.charCodeAt(tmp - 1))) {
      max = tmp;
    }

    state.line = startLine + 1;

    token        = state.push('heading_open', 'h' + String(level), 1);
    token.markup = '########'.slice(0, level);
    token.map    = [ startLine, state.line ];

    token          = state.push('inline', '', 0);
    token.content  = state.src.slice(pos, max).trim();
    token.map      = [ startLine, state.line ];
    token.children = [];

    token        = state.push('heading_close', 'h' + String(level), -1);
    token.markup = '########'.slice(0, level);

    return true;
  };

  // lheading (---, ===)


  var lheading$1 = function lheading(state, startLine, endLine/*, silent*/) {
    var content, terminate, i, l, token, pos, max, level, marker,
        nextLine = startLine + 1, oldParentType,
        terminatorRules = state.md.block.ruler.getRules('paragraph');

    // if it's indented more than 3 spaces, it should be a code block
    if (state.sCount[startLine] - state.blkIndent >= 4) { return false; }

    oldParentType = state.parentType;
    state.parentType = 'paragraph'; // use paragraph to match terminatorRules

    // jump line-by-line until empty one or EOF
    for (; nextLine < endLine && !state.isEmpty(nextLine); nextLine++) {
      // this would be a code block normally, but after paragraph
      // it's considered a lazy continuation regardless of what's there
      if (state.sCount[nextLine] - state.blkIndent > 3) { continue; }

      //
      // Check for underline in setext header
      //
      if (state.sCount[nextLine] >= state.blkIndent) {
        pos = state.bMarks[nextLine] + state.tShift[nextLine];
        max = state.eMarks[nextLine];

        if (pos < max) {
          marker = state.src.charCodeAt(pos);

          if (marker === 0x2D/* - */ || marker === 0x3D/* = */) {
            pos = state.skipChars(pos, marker);
            pos = state.skipSpaces(pos);

            if (pos >= max) {
              level = (marker === 0x3D/* = */ ? 1 : 2);
              break;
            }
          }
        }
      }

      // quirk for blockquotes, this line should already be checked by that rule
      if (state.sCount[nextLine] < 0) { continue; }

      // Some tags can terminate paragraph without empty line.
      terminate = false;
      for (i = 0, l = terminatorRules.length; i < l; i++) {
        if (terminatorRules[i](state, nextLine, endLine, true)) {
          terminate = true;
          break;
        }
      }
      if (terminate) { break; }
    }

    if (!level) {
      // Didn't find valid underline
      return false;
    }

    content = state.getLines(startLine, nextLine, state.blkIndent, false).trim();

    state.line = nextLine + 1;

    token          = state.push('heading_open', 'h' + String(level), 1);
    token.markup   = String.fromCharCode(marker);
    token.map      = [ startLine, state.line ];

    token          = state.push('inline', '', 0);
    token.content  = content;
    token.map      = [ startLine, state.line - 1 ];
    token.children = [];

    token          = state.push('heading_close', 'h' + String(level), -1);
    token.markup   = String.fromCharCode(marker);

    state.parentType = oldParentType;

    return true;
  };

  // List of valid html blocks names, accorting to commonmark spec


  var html_blocks$1 = [
    'address',
    'article',
    'aside',
    'base',
    'basefont',
    'blockquote',
    'body',
    'caption',
    'center',
    'col',
    'colgroup',
    'dd',
    'details',
    'dialog',
    'dir',
    'div',
    'dl',
    'dt',
    'fieldset',
    'figcaption',
    'figure',
    'footer',
    'form',
    'frame',
    'frameset',
    'h1',
    'h2',
    'h3',
    'h4',
    'h5',
    'h6',
    'head',
    'header',
    'hr',
    'html',
    'iframe',
    'legend',
    'li',
    'link',
    'main',
    'menu',
    'menuitem',
    'meta',
    'nav',
    'noframes',
    'ol',
    'optgroup',
    'option',
    'p',
    'param',
    'section',
    'source',
    'summary',
    'table',
    'tbody',
    'td',
    'tfoot',
    'th',
    'thead',
    'title',
    'tr',
    'track',
    'ul'
  ];

  // Regexps to match html elements

  var attr_name$1     = '[a-zA-Z_:][a-zA-Z0-9:._-]*';

  var unquoted$1      = '[^"\'=<>`\\x00-\\x20]+';
  var single_quoted$1 = "'[^']*'";
  var double_quoted$1 = '"[^"]*"';

  var attr_value$1  = '(?:' + unquoted$1 + '|' + single_quoted$1 + '|' + double_quoted$1 + ')';

  var attribute$1   = '(?:\\s+' + attr_name$1 + '(?:\\s*=\\s*' + attr_value$1 + ')?)';

  var open_tag$1    = '<[A-Za-z][A-Za-z0-9\\-]*' + attribute$1 + '*\\s*\\/?>';

  var close_tag$1   = '<\\/[A-Za-z][A-Za-z0-9\\-]*\\s*>';
  var comment$1     = '<!---->|<!--(?:-?[^>-])(?:-?[^-])*-->';
  var processing$1  = '<[?].*?[?]>';
  var declaration$1 = '<![A-Z]+\\s+[^>]*>';
  var cdata$1       = '<!\\[CDATA\\[[\\s\\S]*?\\]\\]>';

  var HTML_TAG_RE$3 = new RegExp('^(?:' + open_tag$1 + '|' + close_tag$1 + '|' + comment$1 +
                          '|' + processing$1 + '|' + declaration$1 + '|' + cdata$1 + ')');
  var HTML_OPEN_CLOSE_TAG_RE$3 = new RegExp('^(?:' + open_tag$1 + '|' + close_tag$1 + ')');

  var HTML_TAG_RE_1$1 = HTML_TAG_RE$3;
  var HTML_OPEN_CLOSE_TAG_RE_1$1 = HTML_OPEN_CLOSE_TAG_RE$3;

  var html_re$1 = {
  	HTML_TAG_RE: HTML_TAG_RE_1$1,
  	HTML_OPEN_CLOSE_TAG_RE: HTML_OPEN_CLOSE_TAG_RE_1$1
  };

  var HTML_OPEN_CLOSE_TAG_RE$2 = html_re$1.HTML_OPEN_CLOSE_TAG_RE;

  // An array of opening and corresponding closing sequences for html tags,
  // last argument defines whether it can terminate a paragraph or not
  //
  var HTML_SEQUENCES$1 = [
    [ /^<(script|pre|style)(?=(\s|>|$))/i, /<\/(script|pre|style)>/i, true ],
    [ /^<!--/,        /-->/,   true ],
    [ /^<\?/,         /\?>/,   true ],
    [ /^<![A-Z]/,     />/,     true ],
    [ /^<!\[CDATA\[/, /\]\]>/, true ],
    [ new RegExp('^</?(' + html_blocks$1.join('|') + ')(?=(\\s|/?>|$))', 'i'), /^$/, true ],
    [ new RegExp(HTML_OPEN_CLOSE_TAG_RE$2.source + '\\s*$'),  /^$/, false ]
  ];


  var html_block$1 = function html_block(state, startLine, endLine, silent) {
    var i, nextLine, token, lineText,
        pos = state.bMarks[startLine] + state.tShift[startLine],
        max = state.eMarks[startLine];

    // if it's indented more than 3 spaces, it should be a code block
    if (state.sCount[startLine] - state.blkIndent >= 4) { return false; }

    if (!state.md.options.html) { return false; }

    if (state.src.charCodeAt(pos) !== 0x3C/* < */) { return false; }

    lineText = state.src.slice(pos, max);

    for (i = 0; i < HTML_SEQUENCES$1.length; i++) {
      if (HTML_SEQUENCES$1[i][0].test(lineText)) { break; }
    }

    if (i === HTML_SEQUENCES$1.length) { return false; }

    if (silent) {
      // true if this sequence can be a terminator, false otherwise
      return HTML_SEQUENCES$1[i][2];
    }

    nextLine = startLine + 1;

    // If we are here - we detected HTML block.
    // Let's roll down till block end.
    if (!HTML_SEQUENCES$1[i][1].test(lineText)) {
      for (; nextLine < endLine; nextLine++) {
        if (state.sCount[nextLine] < state.blkIndent) { break; }

        pos = state.bMarks[nextLine] + state.tShift[nextLine];
        max = state.eMarks[nextLine];
        lineText = state.src.slice(pos, max);

        if (HTML_SEQUENCES$1[i][1].test(lineText)) {
          if (lineText.length !== 0) { nextLine++; }
          break;
        }
      }
    }

    state.line = nextLine;

    token         = state.push('html_block', '', 0);
    token.map     = [ startLine, nextLine ];
    token.content = state.getLines(startLine, nextLine, state.blkIndent, true);

    return true;
  };

  // Paragraph


  var paragraph$2 = function paragraph(state, startLine/*, endLine*/) {
    var content, terminate, i, l, token, oldParentType,
        nextLine = startLine + 1,
        terminatorRules = state.md.block.ruler.getRules('paragraph'),
        endLine = state.lineMax;

    oldParentType = state.parentType;
    state.parentType = 'paragraph';

    // jump line-by-line until empty one or EOF
    for (; nextLine < endLine && !state.isEmpty(nextLine); nextLine++) {
      // this would be a code block normally, but after paragraph
      // it's considered a lazy continuation regardless of what's there
      if (state.sCount[nextLine] - state.blkIndent > 3) { continue; }

      // quirk for blockquotes, this line should already be checked by that rule
      if (state.sCount[nextLine] < 0) { continue; }

      // Some tags can terminate paragraph without empty line.
      terminate = false;
      for (i = 0, l = terminatorRules.length; i < l; i++) {
        if (terminatorRules[i](state, nextLine, endLine, true)) {
          terminate = true;
          break;
        }
      }
      if (terminate) { break; }
    }

    content = state.getLines(startLine, nextLine, state.blkIndent, false).trim();

    state.line = nextLine;

    token          = state.push('paragraph_open', 'p', 1);
    token.map      = [ startLine, state.line ];

    token          = state.push('inline', '', 0);
    token.content  = content;
    token.map      = [ startLine, state.line ];
    token.children = [];

    token          = state.push('paragraph_close', 'p', -1);

    state.parentType = oldParentType;

    return true;
  };

  var isSpace$i = utils$1.isSpace;


  function StateBlock$1(src, md, env, tokens) {
    var ch, s, start, pos, len, indent, offset, indent_found;

    this.src = src;

    // link to parser instance
    this.md     = md;

    this.env = env;

    //
    // Internal state vartiables
    //

    this.tokens = tokens;

    this.bMarks = [];  // line begin offsets for fast jumps
    this.eMarks = [];  // line end offsets for fast jumps
    this.tShift = [];  // offsets of the first non-space characters (tabs not expanded)
    this.sCount = [];  // indents for each line (tabs expanded)

    // An amount of virtual spaces (tabs expanded) between beginning
    // of each line (bMarks) and real beginning of that line.
    //
    // It exists only as a hack because blockquotes override bMarks
    // losing information in the process.
    //
    // It's used only when expanding tabs, you can think about it as
    // an initial tab length, e.g. bsCount=21 applied to string `\t123`
    // means first tab should be expanded to 4-21%4 === 3 spaces.
    //
    this.bsCount = [];

    // block parser variables
    this.blkIndent  = 0; // required block content indent (for example, if we are
                         // inside a list, it would be positioned after list marker)
    this.line       = 0; // line index in src
    this.lineMax    = 0; // lines count
    this.tight      = false;  // loose/tight mode for lists
    this.ddIndent   = -1; // indent of the current dd block (-1 if there isn't any)
    this.listIndent = -1; // indent of the current list block (-1 if there isn't any)

    // can be 'blockquote', 'list', 'root', 'paragraph' or 'reference'
    // used in lists to determine if they interrupt a paragraph
    this.parentType = 'root';

    this.level = 0;

    // renderer
    this.result = '';

    // Create caches
    // Generate markers.
    s = this.src;
    indent_found = false;

    for (start = pos = indent = offset = 0, len = s.length; pos < len; pos++) {
      ch = s.charCodeAt(pos);

      if (!indent_found) {
        if (isSpace$i(ch)) {
          indent++;

          if (ch === 0x09) {
            offset += 4 - offset % 4;
          } else {
            offset++;
          }
          continue;
        } else {
          indent_found = true;
        }
      }

      if (ch === 0x0A || pos === len - 1) {
        if (ch !== 0x0A) { pos++; }
        this.bMarks.push(start);
        this.eMarks.push(pos);
        this.tShift.push(indent);
        this.sCount.push(offset);
        this.bsCount.push(0);

        indent_found = false;
        indent = 0;
        offset = 0;
        start = pos + 1;
      }
    }

    // Push fake entry to simplify cache bounds checks
    this.bMarks.push(s.length);
    this.eMarks.push(s.length);
    this.tShift.push(0);
    this.sCount.push(0);
    this.bsCount.push(0);

    this.lineMax = this.bMarks.length - 1; // don't count last fake line
  }

  // Push new token to "stream".
  //
  StateBlock$1.prototype.push = function (type, tag, nesting) {
    var token = new token$1(type, tag, nesting);
    token.block = true;

    if (nesting < 0) { this.level--; } // closing tag
    token.level = this.level;
    if (nesting > 0) { this.level++; } // opening tag

    this.tokens.push(token);
    return token;
  };

  StateBlock$1.prototype.isEmpty = function isEmpty(line) {
    return this.bMarks[line] + this.tShift[line] >= this.eMarks[line];
  };

  StateBlock$1.prototype.skipEmptyLines = function skipEmptyLines(from) {
    for (var max = this.lineMax; from < max; from++) {
      if (this.bMarks[from] + this.tShift[from] < this.eMarks[from]) {
        break;
      }
    }
    return from;
  };

  // Skip spaces from given position.
  StateBlock$1.prototype.skipSpaces = function skipSpaces(pos) {
    var ch;

    for (var max = this.src.length; pos < max; pos++) {
      ch = this.src.charCodeAt(pos);
      if (!isSpace$i(ch)) { break; }
    }
    return pos;
  };

  // Skip spaces from given position in reverse.
  StateBlock$1.prototype.skipSpacesBack = function skipSpacesBack(pos, min) {
    if (pos <= min) { return pos; }

    while (pos > min) {
      if (!isSpace$i(this.src.charCodeAt(--pos))) { return pos + 1; }
    }
    return pos;
  };

  // Skip char codes from given position
  StateBlock$1.prototype.skipChars = function skipChars(pos, code) {
    for (var max = this.src.length; pos < max; pos++) {
      if (this.src.charCodeAt(pos) !== code) { break; }
    }
    return pos;
  };

  // Skip char codes reverse from given position - 1
  StateBlock$1.prototype.skipCharsBack = function skipCharsBack(pos, code, min) {
    if (pos <= min) { return pos; }

    while (pos > min) {
      if (code !== this.src.charCodeAt(--pos)) { return pos + 1; }
    }
    return pos;
  };

  // cut lines range from source.
  StateBlock$1.prototype.getLines = function getLines(begin, end, indent, keepLastLF) {
    var i, lineIndent, ch, first, last, queue, lineStart,
        line = begin;

    if (begin >= end) {
      return '';
    }

    queue = new Array(end - begin);

    for (i = 0; line < end; line++, i++) {
      lineIndent = 0;
      lineStart = first = this.bMarks[line];

      if (line + 1 < end || keepLastLF) {
        // No need for bounds check because we have fake entry on tail.
        last = this.eMarks[line] + 1;
      } else {
        last = this.eMarks[line];
      }

      while (first < last && lineIndent < indent) {
        ch = this.src.charCodeAt(first);

        if (isSpace$i(ch)) {
          if (ch === 0x09) {
            lineIndent += 4 - (lineIndent + this.bsCount[line]) % 4;
          } else {
            lineIndent++;
          }
        } else if (first - lineStart < this.tShift[line]) {
          // patched tShift masked characters to look like spaces (blockquotes, list markers)
          lineIndent++;
        } else {
          break;
        }

        first++;
      }

      if (lineIndent > indent) {
        // partially expanding tabs in code blocks, e.g '\t\tfoobar'
        // with indent=2 becomes '  \tfoobar'
        queue[i] = new Array(lineIndent - indent + 1).join(' ') + this.src.slice(first, last);
      } else {
        queue[i] = this.src.slice(first, last);
      }
    }

    return queue.join('');
  };

  // re-export Token class to use in block rules
  StateBlock$1.prototype.Token = token$1;


  var state_block$1 = StateBlock$1;

  var _rules$4 = [
    // First 2 params - rule name & source. Secondary array - list of rules,
    // which can be terminated by this one.
    [ 'table',      table$2,      [ 'paragraph', 'reference' ] ],
    [ 'code',       code$2 ],
    [ 'fence',      fence$2,      [ 'paragraph', 'reference', 'blockquote', 'list' ] ],
    [ 'blockquote', blockquote$2, [ 'paragraph', 'reference', 'blockquote', 'list' ] ],
    [ 'hr',         hr$1,         [ 'paragraph', 'reference', 'blockquote', 'list' ] ],
    [ 'list',       list$1,       [ 'paragraph', 'reference', 'blockquote' ] ],
    [ 'reference',  reference$1 ],
    [ 'heading',    heading$2,    [ 'paragraph', 'reference', 'blockquote' ] ],
    [ 'lheading',   lheading$1 ],
    [ 'html_block', html_block$1, [ 'paragraph', 'reference', 'blockquote' ] ],
    [ 'paragraph',  paragraph$2 ]
  ];


  /**
   * new ParserBlock()
   **/
  function ParserBlock$1() {
    /**
     * ParserBlock#ruler -> Ruler
     *
     * [[Ruler]] instance. Keep configuration of block rules.
     **/
    this.ruler = new ruler$1();

    for (var i = 0; i < _rules$4.length; i++) {
      this.ruler.push(_rules$4[i][0], _rules$4[i][1], { alt: (_rules$4[i][2] || []).slice() });
    }
  }


  // Generate tokens for input range
  //
  ParserBlock$1.prototype.tokenize = function (state, startLine, endLine) {
    var ok, i,
        rules = this.ruler.getRules(''),
        len = rules.length,
        line = startLine,
        hasEmptyLines = false,
        maxNesting = state.md.options.maxNesting;

    while (line < endLine) {
      state.line = line = state.skipEmptyLines(line);
      if (line >= endLine) { break; }

      // Termination condition for nested calls.
      // Nested calls currently used for blockquotes & lists
      if (state.sCount[line] < state.blkIndent) { break; }

      // If nesting level exceeded - skip tail to the end. That's not ordinary
      // situation and we should not care about content.
      if (state.level >= maxNesting) {
        state.line = endLine;
        break;
      }

      // Try all possible rules.
      // On success, rule should:
      //
      // - update `state.line`
      // - update `state.tokens`
      // - return true

      for (i = 0; i < len; i++) {
        ok = rules[i](state, line, endLine, false);
        if (ok) { break; }
      }

      // set state.tight if we had an empty line before current tag
      // i.e. latest empty line should not count
      state.tight = !hasEmptyLines;

      // paragraph might "eat" one newline after it in nested lists
      if (state.isEmpty(state.line - 1)) {
        hasEmptyLines = true;
      }

      line = state.line;

      if (line < endLine && state.isEmpty(line)) {
        hasEmptyLines = true;
        line++;
        state.line = line;
      }
    }
  };


  /**
   * ParserBlock.parse(str, md, env, outTokens)
   *
   * Process input string and push block tokens into `outTokens`
   **/
  ParserBlock$1.prototype.parse = function (src, md, env, outTokens) {
    var state;

    if (!src) { return; }

    state = new this.State(src, md, env, outTokens);

    this.tokenize(state, state.line, state.lineMax);
  };


  ParserBlock$1.prototype.State = state_block$1;


  var parser_block$1 = ParserBlock$1;

  // Skip text characters for text token, place those to pending buffer


  // Rule to skip pure text
  // '{}$%@~+=:' reserved for extentions

  // !, ", #, $, %, &, ', (, ), *, +, ,, -, ., /, :, ;, <, =, >, ?, @, [, \, ], ^, _, `, {, |, }, or ~

  // !!!! Don't confuse with "Markdown ASCII Punctuation" chars
  // http://spec.commonmark.org/0.15/#ascii-punctuation-character
  function isTerminatorChar$1(ch) {
    switch (ch) {
      case 0x0A/* \n */:
      case 0x21/* ! */:
      case 0x23/* # */:
      case 0x24/* $ */:
      case 0x25/* % */:
      case 0x26/* & */:
      case 0x2A/* * */:
      case 0x2B/* + */:
      case 0x2D/* - */:
      case 0x3A/* : */:
      case 0x3C/* < */:
      case 0x3D/* = */:
      case 0x3E/* > */:
      case 0x40/* @ */:
      case 0x5B/* [ */:
      case 0x5C/* \ */:
      case 0x5D/* ] */:
      case 0x5E/* ^ */:
      case 0x5F/* _ */:
      case 0x60/* ` */:
      case 0x7B/* { */:
      case 0x7D/* } */:
      case 0x7E/* ~ */:
        return true;
      default:
        return false;
    }
  }

  var text$2 = function text(state, silent) {
    var pos = state.pos;

    while (pos < state.posMax && !isTerminatorChar$1(state.src.charCodeAt(pos))) {
      pos++;
    }

    if (pos === state.pos) { return false; }

    if (!silent) { state.pending += state.src.slice(state.pos, pos); }

    state.pos = pos;

    return true;
  };

  var isSpace$h = utils$1.isSpace;


  var newline$1 = function newline(state, silent) {
    var pmax, max, pos = state.pos;

    if (state.src.charCodeAt(pos) !== 0x0A/* \n */) { return false; }

    pmax = state.pending.length - 1;
    max = state.posMax;

    // '  \n' -> hardbreak
    // Lookup in pending chars is bad practice! Don't copy to other rules!
    // Pending string is stored in concat mode, indexed lookups will cause
    // convertion to flat mode.
    if (!silent) {
      if (pmax >= 0 && state.pending.charCodeAt(pmax) === 0x20) {
        if (pmax >= 1 && state.pending.charCodeAt(pmax - 1) === 0x20) {
          state.pending = state.pending.replace(/ +$/, '');
          state.push('hardbreak', 'br', 0);
        } else {
          state.pending = state.pending.slice(0, -1);
          state.push('softbreak', 'br', 0);
        }

      } else {
        state.push('softbreak', 'br', 0);
      }
    }

    pos++;

    // skip heading spaces for next line
    while (pos < max && isSpace$h(state.src.charCodeAt(pos))) { pos++; }

    state.pos = pos;
    return true;
  };

  var isSpace$g = utils$1.isSpace;

  var ESCAPED$1 = [];

  for (var i$1 = 0; i$1 < 256; i$1++) { ESCAPED$1.push(0); }

  '\\!"#$%&\'()*+,./:;<=>?@[]^_`{|}~-'
    .split('').forEach(function (ch) { ESCAPED$1[ch.charCodeAt(0)] = 1; });


  var _escape$1 = function escape(state, silent) {
    var ch, pos = state.pos, max = state.posMax;

    if (state.src.charCodeAt(pos) !== 0x5C/* \ */) { return false; }

    pos++;

    if (pos < max) {
      ch = state.src.charCodeAt(pos);

      if (ch < 256 && ESCAPED$1[ch] !== 0) {
        if (!silent) { state.pending += state.src[pos]; }
        state.pos += 2;
        return true;
      }

      if (ch === 0x0A) {
        if (!silent) {
          state.push('hardbreak', 'br', 0);
        }

        pos++;
        // skip leading whitespaces from next line
        while (pos < max) {
          ch = state.src.charCodeAt(pos);
          if (!isSpace$g(ch)) { break; }
          pos++;
        }

        state.pos = pos;
        return true;
      }
    }

    if (!silent) { state.pending += '\\'; }
    state.pos++;
    return true;
  };

  // Parse backticks

  var backticks$1 = function backtick(state, silent) {
    var start, max, marker, matchStart, matchEnd, token,
        pos = state.pos,
        ch = state.src.charCodeAt(pos);

    if (ch !== 0x60/* ` */) { return false; }

    start = pos;
    pos++;
    max = state.posMax;

    while (pos < max && state.src.charCodeAt(pos) === 0x60/* ` */) { pos++; }

    marker = state.src.slice(start, pos);

    matchStart = matchEnd = pos;

    while ((matchStart = state.src.indexOf('`', matchEnd)) !== -1) {
      matchEnd = matchStart + 1;

      while (matchEnd < max && state.src.charCodeAt(matchEnd) === 0x60/* ` */) { matchEnd++; }

      if (matchEnd - matchStart === marker.length) {
        if (!silent) {
          token         = state.push('code_inline', 'code', 0);
          token.markup  = marker;
          token.content = state.src.slice(pos, matchStart)
            .replace(/\n/g, ' ')
            .replace(/^ (.+) $/, '$1');
        }
        state.pos = matchEnd;
        return true;
      }
    }

    if (!silent) { state.pending += marker; }
    state.pos += marker.length;
    return true;
  };

  // ~~strike through~~


  // Insert each marker as a separate text token, and add it to delimiter list
  //
  var tokenize$3 = function strikethrough(state, silent) {
    var i, scanned, token, len, ch,
        start = state.pos,
        marker = state.src.charCodeAt(start);

    if (silent) { return false; }

    if (marker !== 0x7E/* ~ */) { return false; }

    scanned = state.scanDelims(state.pos, true);
    len = scanned.length;
    ch = String.fromCharCode(marker);

    if (len < 2) { return false; }

    if (len % 2) {
      token         = state.push('text', '', 0);
      token.content = ch;
      len--;
    }

    for (i = 0; i < len; i += 2) {
      token         = state.push('text', '', 0);
      token.content = ch + ch;

      state.delimiters.push({
        marker: marker,
        length: 0, // disable "rule of 3" length checks meant for emphasis
        jump:   i,
        token:  state.tokens.length - 1,
        end:    -1,
        open:   scanned.can_open,
        close:  scanned.can_close
      });
    }

    state.pos += scanned.length;

    return true;
  };


  function postProcess$3(state, delimiters) {
    var i, j,
        startDelim,
        endDelim,
        token,
        loneMarkers = [],
        max = delimiters.length;

    for (i = 0; i < max; i++) {
      startDelim = delimiters[i];

      if (startDelim.marker !== 0x7E/* ~ */) {
        continue;
      }

      if (startDelim.end === -1) {
        continue;
      }

      endDelim = delimiters[startDelim.end];

      token         = state.tokens[startDelim.token];
      token.type    = 's_open';
      token.tag     = 's';
      token.nesting = 1;
      token.markup  = '~~';
      token.content = '';

      token         = state.tokens[endDelim.token];
      token.type    = 's_close';
      token.tag     = 's';
      token.nesting = -1;
      token.markup  = '~~';
      token.content = '';

      if (state.tokens[endDelim.token - 1].type === 'text' &&
          state.tokens[endDelim.token - 1].content === '~') {

        loneMarkers.push(endDelim.token - 1);
      }
    }

    // If a marker sequence has an odd number of characters, it's splitted
    // like this: `~~~~~` -> `~` + `~~` + `~~`, leaving one marker at the
    // start of the sequence.
    //
    // So, we have to move all those markers after subsequent s_close tags.
    //
    while (loneMarkers.length) {
      i = loneMarkers.pop();
      j = i + 1;

      while (j < state.tokens.length && state.tokens[j].type === 's_close') {
        j++;
      }

      j--;

      if (i !== j) {
        token = state.tokens[j];
        state.tokens[j] = state.tokens[i];
        state.tokens[i] = token;
      }
    }
  }


  // Walk through delimiter list and replace text tokens with tags
  //
  var postProcess_1$3 = function strikethrough(state) {
    var curr,
        tokens_meta = state.tokens_meta,
        max = state.tokens_meta.length;

    postProcess$3(state, state.delimiters);

    for (curr = 0; curr < max; curr++) {
      if (tokens_meta[curr] && tokens_meta[curr].delimiters) {
        postProcess$3(state, tokens_meta[curr].delimiters);
      }
    }
  };

  var strikethrough$2 = {
  	tokenize: tokenize$3,
  	postProcess: postProcess_1$3
  };

  // Process *this* and _that_


  // Insert each marker as a separate text token, and add it to delimiter list
  //
  var tokenize$2 = function emphasis(state, silent) {
    var i, scanned, token,
        start = state.pos,
        marker = state.src.charCodeAt(start);

    if (silent) { return false; }

    if (marker !== 0x5F /* _ */ && marker !== 0x2A /* * */) { return false; }

    scanned = state.scanDelims(state.pos, marker === 0x2A);

    for (i = 0; i < scanned.length; i++) {
      token         = state.push('text', '', 0);
      token.content = String.fromCharCode(marker);

      state.delimiters.push({
        // Char code of the starting marker (number).
        //
        marker: marker,

        // Total length of these series of delimiters.
        //
        length: scanned.length,

        // An amount of characters before this one that's equivalent to
        // current one. In plain English: if this delimiter does not open
        // an emphasis, neither do previous `jump` characters.
        //
        // Used to skip sequences like "*****" in one step, for 1st asterisk
        // value will be 0, for 2nd it's 1 and so on.
        //
        jump:   i,

        // A position of the token this delimiter corresponds to.
        //
        token:  state.tokens.length - 1,

        // If this delimiter is matched as a valid opener, `end` will be
        // equal to its position, otherwise it's `-1`.
        //
        end:    -1,

        // Boolean flags that determine if this delimiter could open or close
        // an emphasis.
        //
        open:   scanned.can_open,
        close:  scanned.can_close
      });
    }

    state.pos += scanned.length;

    return true;
  };


  function postProcess$2(state, delimiters) {
    var i,
        startDelim,
        endDelim,
        token,
        ch,
        isStrong,
        max = delimiters.length;

    for (i = max - 1; i >= 0; i--) {
      startDelim = delimiters[i];

      if (startDelim.marker !== 0x5F/* _ */ && startDelim.marker !== 0x2A/* * */) {
        continue;
      }

      // Process only opening markers
      if (startDelim.end === -1) {
        continue;
      }

      endDelim = delimiters[startDelim.end];

      // If the previous delimiter has the same marker and is adjacent to this one,
      // merge those into one strong delimiter.
      //
      // `<em><em>whatever</em></em>` -> `<strong>whatever</strong>`
      //
      isStrong = i > 0 &&
                 delimiters[i - 1].end === startDelim.end + 1 &&
                 delimiters[i - 1].token === startDelim.token - 1 &&
                 delimiters[startDelim.end + 1].token === endDelim.token + 1 &&
                 delimiters[i - 1].marker === startDelim.marker;

      ch = String.fromCharCode(startDelim.marker);

      token         = state.tokens[startDelim.token];
      token.type    = isStrong ? 'strong_open' : 'em_open';
      token.tag     = isStrong ? 'strong' : 'em';
      token.nesting = 1;
      token.markup  = isStrong ? ch + ch : ch;
      token.content = '';

      token         = state.tokens[endDelim.token];
      token.type    = isStrong ? 'strong_close' : 'em_close';
      token.tag     = isStrong ? 'strong' : 'em';
      token.nesting = -1;
      token.markup  = isStrong ? ch + ch : ch;
      token.content = '';

      if (isStrong) {
        state.tokens[delimiters[i - 1].token].content = '';
        state.tokens[delimiters[startDelim.end + 1].token].content = '';
        i--;
      }
    }
  }


  // Walk through delimiter list and replace text tokens with tags
  //
  var postProcess_1$2 = function emphasis(state) {
    var curr,
        tokens_meta = state.tokens_meta,
        max = state.tokens_meta.length;

    postProcess$2(state, state.delimiters);

    for (curr = 0; curr < max; curr++) {
      if (tokens_meta[curr] && tokens_meta[curr].delimiters) {
        postProcess$2(state, tokens_meta[curr].delimiters);
      }
    }
  };

  var emphasis$1 = {
  	tokenize: tokenize$2,
  	postProcess: postProcess_1$2
  };

  var normalizeReference$4   = utils$1.normalizeReference;
  var isSpace$f              = utils$1.isSpace;


  var link$4 = function link(state, silent) {
    var attrs,
        code,
        label,
        labelEnd,
        labelStart,
        pos,
        res,
        ref,
        title,
        token,
        href = '',
        oldPos = state.pos,
        max = state.posMax,
        start = state.pos,
        parseReference = true;

    if (state.src.charCodeAt(state.pos) !== 0x5B/* [ */) { return false; }

    labelStart = state.pos + 1;
    labelEnd = state.md.helpers.parseLinkLabel(state, state.pos, true);

    // parser failed to find ']', so it's not a valid link
    if (labelEnd < 0) { return false; }

    pos = labelEnd + 1;
    if (pos < max && state.src.charCodeAt(pos) === 0x28/* ( */) {
      //
      // Inline link
      //

      // might have found a valid shortcut link, disable reference parsing
      parseReference = false;

      // [link](  <href>  "title"  )
      //        ^^ skipping these spaces
      pos++;
      for (; pos < max; pos++) {
        code = state.src.charCodeAt(pos);
        if (!isSpace$f(code) && code !== 0x0A) { break; }
      }
      if (pos >= max) { return false; }

      // [link](  <href>  "title"  )
      //          ^^^^^^ parsing link destination
      start = pos;
      res = state.md.helpers.parseLinkDestination(state.src, pos, state.posMax);
      if (res.ok) {
        href = state.md.normalizeLink(res.str);
        if (state.md.validateLink(href)) {
          pos = res.pos;
        } else {
          href = '';
        }
      }

      // [link](  <href>  "title"  )
      //                ^^ skipping these spaces
      start = pos;
      for (; pos < max; pos++) {
        code = state.src.charCodeAt(pos);
        if (!isSpace$f(code) && code !== 0x0A) { break; }
      }

      // [link](  <href>  "title"  )
      //                  ^^^^^^^ parsing link title
      res = state.md.helpers.parseLinkTitle(state.src, pos, state.posMax);
      if (pos < max && start !== pos && res.ok) {
        title = res.str;
        pos = res.pos;

        // [link](  <href>  "title"  )
        //                         ^^ skipping these spaces
        for (; pos < max; pos++) {
          code = state.src.charCodeAt(pos);
          if (!isSpace$f(code) && code !== 0x0A) { break; }
        }
      } else {
        title = '';
      }

      if (pos >= max || state.src.charCodeAt(pos) !== 0x29/* ) */) {
        // parsing a valid shortcut link failed, fallback to reference
        parseReference = true;
      }
      pos++;
    }

    if (parseReference) {
      //
      // Link reference
      //
      if (typeof state.env.references === 'undefined') { return false; }

      if (pos < max && state.src.charCodeAt(pos) === 0x5B/* [ */) {
        start = pos + 1;
        pos = state.md.helpers.parseLinkLabel(state, pos);
        if (pos >= 0) {
          label = state.src.slice(start, pos++);
        } else {
          pos = labelEnd + 1;
        }
      } else {
        pos = labelEnd + 1;
      }

      // covers label === '' and label === undefined
      // (collapsed reference link and shortcut reference link respectively)
      if (!label) { label = state.src.slice(labelStart, labelEnd); }

      ref = state.env.references[normalizeReference$4(label)];
      if (!ref) {
        state.pos = oldPos;
        return false;
      }
      href = ref.href;
      title = ref.title;
    }

    //
    // We found the end of the link, and know for a fact it's a valid link;
    // so all that's left to do is to call tokenizer.
    //
    if (!silent) {
      state.pos = labelStart;
      state.posMax = labelEnd;

      token        = state.push('link_open', 'a', 1);
      token.attrs  = attrs = [ [ 'href', href ] ];
      if (title) {
        attrs.push([ 'title', title ]);
      }

      state.md.inline.tokenize(state);

      token        = state.push('link_close', 'a', -1);
    }

    state.pos = pos;
    state.posMax = max;
    return true;
  };

  var normalizeReference$3   = utils$1.normalizeReference;
  var isSpace$e              = utils$1.isSpace;


  var image$3 = function image(state, silent) {
    var attrs,
        code,
        content,
        label,
        labelEnd,
        labelStart,
        pos,
        ref,
        res,
        title,
        token,
        tokens,
        start,
        href = '',
        oldPos = state.pos,
        max = state.posMax;

    if (state.src.charCodeAt(state.pos) !== 0x21/* ! */) { return false; }
    if (state.src.charCodeAt(state.pos + 1) !== 0x5B/* [ */) { return false; }

    labelStart = state.pos + 2;
    labelEnd = state.md.helpers.parseLinkLabel(state, state.pos + 1, false);

    // parser failed to find ']', so it's not a valid link
    if (labelEnd < 0) { return false; }

    pos = labelEnd + 1;
    if (pos < max && state.src.charCodeAt(pos) === 0x28/* ( */) {
      //
      // Inline link
      //

      // [link](  <href>  "title"  )
      //        ^^ skipping these spaces
      pos++;
      for (; pos < max; pos++) {
        code = state.src.charCodeAt(pos);
        if (!isSpace$e(code) && code !== 0x0A) { break; }
      }
      if (pos >= max) { return false; }

      // [link](  <href>  "title"  )
      //          ^^^^^^ parsing link destination
      start = pos;
      res = state.md.helpers.parseLinkDestination(state.src, pos, state.posMax);
      if (res.ok) {
        href = state.md.normalizeLink(res.str);
        if (state.md.validateLink(href)) {
          pos = res.pos;
        } else {
          href = '';
        }
      }

      // [link](  <href>  "title"  )
      //                ^^ skipping these spaces
      start = pos;
      for (; pos < max; pos++) {
        code = state.src.charCodeAt(pos);
        if (!isSpace$e(code) && code !== 0x0A) { break; }
      }

      // [link](  <href>  "title"  )
      //                  ^^^^^^^ parsing link title
      res = state.md.helpers.parseLinkTitle(state.src, pos, state.posMax);
      if (pos < max && start !== pos && res.ok) {
        title = res.str;
        pos = res.pos;

        // [link](  <href>  "title"  )
        //                         ^^ skipping these spaces
        for (; pos < max; pos++) {
          code = state.src.charCodeAt(pos);
          if (!isSpace$e(code) && code !== 0x0A) { break; }
        }
      } else {
        title = '';
      }

      if (pos >= max || state.src.charCodeAt(pos) !== 0x29/* ) */) {
        state.pos = oldPos;
        return false;
      }
      pos++;
    } else {
      //
      // Link reference
      //
      if (typeof state.env.references === 'undefined') { return false; }

      if (pos < max && state.src.charCodeAt(pos) === 0x5B/* [ */) {
        start = pos + 1;
        pos = state.md.helpers.parseLinkLabel(state, pos);
        if (pos >= 0) {
          label = state.src.slice(start, pos++);
        } else {
          pos = labelEnd + 1;
        }
      } else {
        pos = labelEnd + 1;
      }

      // covers label === '' and label === undefined
      // (collapsed reference link and shortcut reference link respectively)
      if (!label) { label = state.src.slice(labelStart, labelEnd); }

      ref = state.env.references[normalizeReference$3(label)];
      if (!ref) {
        state.pos = oldPos;
        return false;
      }
      href = ref.href;
      title = ref.title;
    }

    //
    // We found the end of the link, and know for a fact it's a valid link;
    // so all that's left to do is to call tokenizer.
    //
    if (!silent) {
      content = state.src.slice(labelStart, labelEnd);

      state.md.inline.parse(
        content,
        state.md,
        state.env,
        tokens = []
      );

      token          = state.push('image', 'img', 0);
      token.attrs    = attrs = [ [ 'src', href ], [ 'alt', '' ] ];
      token.children = tokens;
      token.content  = content;

      if (title) {
        attrs.push([ 'title', title ]);
      }
    }

    state.pos = pos;
    state.posMax = max;
    return true;
  };

  // Process autolinks '<protocol:...>'


  /*eslint max-len:0*/
  var EMAIL_RE$1    = /^<([a-zA-Z0-9.!#$%&'*+\/=?^_`{|}~-]+@[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?(?:\.[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)*)>/;
  var AUTOLINK_RE$1 = /^<([a-zA-Z][a-zA-Z0-9+.\-]{1,31}):([^<>\x00-\x20]*)>/;


  var autolink$1 = function autolink(state, silent) {
    var tail, linkMatch, emailMatch, url, fullUrl, token,
        pos = state.pos;

    if (state.src.charCodeAt(pos) !== 0x3C/* < */) { return false; }

    tail = state.src.slice(pos);

    if (tail.indexOf('>') < 0) { return false; }

    if (AUTOLINK_RE$1.test(tail)) {
      linkMatch = tail.match(AUTOLINK_RE$1);

      url = linkMatch[0].slice(1, -1);
      fullUrl = state.md.normalizeLink(url);
      if (!state.md.validateLink(fullUrl)) { return false; }

      if (!silent) {
        token         = state.push('link_open', 'a', 1);
        token.attrs   = [ [ 'href', fullUrl ] ];
        token.markup  = 'autolink';
        token.info    = 'auto';

        token         = state.push('text', '', 0);
        token.content = state.md.normalizeLinkText(url);

        token         = state.push('link_close', 'a', -1);
        token.markup  = 'autolink';
        token.info    = 'auto';
      }

      state.pos += linkMatch[0].length;
      return true;
    }

    if (EMAIL_RE$1.test(tail)) {
      emailMatch = tail.match(EMAIL_RE$1);

      url = emailMatch[0].slice(1, -1);
      fullUrl = state.md.normalizeLink('mailto:' + url);
      if (!state.md.validateLink(fullUrl)) { return false; }

      if (!silent) {
        token         = state.push('link_open', 'a', 1);
        token.attrs   = [ [ 'href', fullUrl ] ];
        token.markup  = 'autolink';
        token.info    = 'auto';

        token         = state.push('text', '', 0);
        token.content = state.md.normalizeLinkText(url);

        token         = state.push('link_close', 'a', -1);
        token.markup  = 'autolink';
        token.info    = 'auto';
      }

      state.pos += emailMatch[0].length;
      return true;
    }

    return false;
  };

  var HTML_TAG_RE$2 = html_re$1.HTML_TAG_RE;


  function isLetter$1(ch) {
    /*eslint no-bitwise:0*/
    var lc = ch | 0x20; // to lower case
    return (lc >= 0x61/* a */) && (lc <= 0x7a/* z */);
  }


  var html_inline$1 = function html_inline(state, silent) {
    var ch, match, max, token,
        pos = state.pos;

    if (!state.md.options.html) { return false; }

    // Check start
    max = state.posMax;
    if (state.src.charCodeAt(pos) !== 0x3C/* < */ ||
        pos + 2 >= max) {
      return false;
    }

    // Quick fail on second char
    ch = state.src.charCodeAt(pos + 1);
    if (ch !== 0x21/* ! */ &&
        ch !== 0x3F/* ? */ &&
        ch !== 0x2F/* / */ &&
        !isLetter$1(ch)) {
      return false;
    }

    match = state.src.slice(pos).match(HTML_TAG_RE$2);
    if (!match) { return false; }

    if (!silent) {
      token         = state.push('html_inline', '', 0);
      token.content = state.src.slice(pos, pos + match[0].length);
    }
    state.pos += match[0].length;
    return true;
  };

  var has$1               = utils$1.has;
  var isValidEntityCode$1 = utils$1.isValidEntityCode;
  var fromCodePoint$1     = utils$1.fromCodePoint;


  var DIGITAL_RE$1 = /^&#((?:x[a-f0-9]{1,6}|[0-9]{1,7}));/i;
  var NAMED_RE$1   = /^&([a-z][a-z0-9]{1,31});/i;


  var entity$1 = function entity(state, silent) {
    var ch, code, match, pos = state.pos, max = state.posMax;

    if (state.src.charCodeAt(pos) !== 0x26/* & */) { return false; }

    if (pos + 1 < max) {
      ch = state.src.charCodeAt(pos + 1);

      if (ch === 0x23 /* # */) {
        match = state.src.slice(pos).match(DIGITAL_RE$1);
        if (match) {
          if (!silent) {
            code = match[1][0].toLowerCase() === 'x' ? parseInt(match[1].slice(1), 16) : parseInt(match[1], 10);
            state.pending += isValidEntityCode$1(code) ? fromCodePoint$1(code) : fromCodePoint$1(0xFFFD);
          }
          state.pos += match[0].length;
          return true;
        }
      } else {
        match = state.src.slice(pos).match(NAMED_RE$1);
        if (match) {
          if (has$1(entities$3, match[1])) {
            if (!silent) { state.pending += entities$3[match[1]]; }
            state.pos += match[0].length;
            return true;
          }
        }
      }
    }

    if (!silent) { state.pending += '&'; }
    state.pos++;
    return true;
  };

  // For each opening emphasis-like marker find a matching closing one


  function processDelimiters$1(state, delimiters) {
    var closerIdx, openerIdx, closer, opener, minOpenerIdx, newMinOpenerIdx,
        isOddMatch, lastJump,
        openersBottom = {},
        max = delimiters.length;

    for (closerIdx = 0; closerIdx < max; closerIdx++) {
      closer = delimiters[closerIdx];

      // Length is only used for emphasis-specific "rule of 3",
      // if it's not defined (in strikethrough or 3rd party plugins),
      // we can default it to 0 to disable those checks.
      //
      closer.length = closer.length || 0;

      if (!closer.close) { continue; }

      // Previously calculated lower bounds (previous fails)
      // for each marker and each delimiter length modulo 3.
      if (!openersBottom.hasOwnProperty(closer.marker)) {
        openersBottom[closer.marker] = [ -1, -1, -1 ];
      }

      minOpenerIdx = openersBottom[closer.marker][closer.length % 3];
      newMinOpenerIdx = -1;

      openerIdx = closerIdx - closer.jump - 1;

      for (; openerIdx > minOpenerIdx; openerIdx -= opener.jump + 1) {
        opener = delimiters[openerIdx];

        if (opener.marker !== closer.marker) { continue; }

        if (newMinOpenerIdx === -1) { newMinOpenerIdx = openerIdx; }

        if (opener.open &&
            opener.end < 0 &&
            opener.level === closer.level) {

          isOddMatch = false;

          // from spec:
          //
          // If one of the delimiters can both open and close emphasis, then the
          // sum of the lengths of the delimiter runs containing the opening and
          // closing delimiters must not be a multiple of 3 unless both lengths
          // are multiples of 3.
          //
          if (opener.close || closer.open) {
            if ((opener.length + closer.length) % 3 === 0) {
              if (opener.length % 3 !== 0 || closer.length % 3 !== 0) {
                isOddMatch = true;
              }
            }
          }

          if (!isOddMatch) {
            // If previous delimiter cannot be an opener, we can safely skip
            // the entire sequence in future checks. This is required to make
            // sure algorithm has linear complexity (see *_*_*_*_*_... case).
            //
            lastJump = openerIdx > 0 && !delimiters[openerIdx - 1].open ?
              delimiters[openerIdx - 1].jump + 1 :
              0;

            closer.jump  = closerIdx - openerIdx + lastJump;
            closer.open  = false;
            opener.end   = closerIdx;
            opener.jump  = lastJump;
            opener.close = false;
            newMinOpenerIdx = -1;
            break;
          }
        }
      }

      if (newMinOpenerIdx !== -1) {
        // If match for this delimiter run failed, we want to set lower bound for
        // future lookups. This is required to make sure algorithm has linear
        // complexity.
        //
        // See details here:
        // https://github.com/commonmark/cmark/issues/178#issuecomment-270417442
        //
        openersBottom[closer.marker][(closer.length || 0) % 3] = newMinOpenerIdx;
      }
    }
  }


  var balance_pairs$1 = function link_pairs(state) {
    var curr,
        tokens_meta = state.tokens_meta,
        max = state.tokens_meta.length;

    processDelimiters$1(state, state.delimiters);

    for (curr = 0; curr < max; curr++) {
      if (tokens_meta[curr] && tokens_meta[curr].delimiters) {
        processDelimiters$1(state, tokens_meta[curr].delimiters);
      }
    }
  };

  // Clean up tokens after emphasis and strikethrough postprocessing:


  var text_collapse$1 = function text_collapse(state) {
    var curr, last,
        level = 0,
        tokens = state.tokens,
        max = state.tokens.length;

    for (curr = last = 0; curr < max; curr++) {
      // re-calculate levels after emphasis/strikethrough turns some text nodes
      // into opening/closing tags
      if (tokens[curr].nesting < 0) { level--; } // closing tag
      tokens[curr].level = level;
      if (tokens[curr].nesting > 0) { level++; } // opening tag

      if (tokens[curr].type === 'text' &&
          curr + 1 < max &&
          tokens[curr + 1].type === 'text') {

        // collapse two adjacent text nodes
        tokens[curr + 1].content = tokens[curr].content + tokens[curr + 1].content;
      } else {
        if (curr !== last) { tokens[last] = tokens[curr]; }

        last++;
      }
    }

    if (curr !== last) {
      tokens.length = last;
    }
  };

  var isWhiteSpace$2   = utils$1.isWhiteSpace;
  var isPunctChar$2    = utils$1.isPunctChar;
  var isMdAsciiPunct$2 = utils$1.isMdAsciiPunct;


  function StateInline$1(src, md, env, outTokens) {
    this.src = src;
    this.env = env;
    this.md = md;
    this.tokens = outTokens;
    this.tokens_meta = Array(outTokens.length);

    this.pos = 0;
    this.posMax = this.src.length;
    this.level = 0;
    this.pending = '';
    this.pendingLevel = 0;

    // Stores { start: end } pairs. Useful for backtrack
    // optimization of pairs parse (emphasis, strikes).
    this.cache = {};

    // List of emphasis-like delimiters for current tag
    this.delimiters = [];

    // Stack of delimiter lists for upper level tags
    this._prev_delimiters = [];
  }


  // Flush pending text
  //
  StateInline$1.prototype.pushPending = function () {
    var token = new token$1('text', '', 0);
    token.content = this.pending;
    token.level = this.pendingLevel;
    this.tokens.push(token);
    this.pending = '';
    return token;
  };


  // Push new token to "stream".
  // If pending text exists - flush it as text token
  //
  StateInline$1.prototype.push = function (type, tag, nesting) {
    if (this.pending) {
      this.pushPending();
    }

    var token = new token$1(type, tag, nesting);
    var token_meta = null;

    if (nesting < 0) {
      // closing tag
      this.level--;
      this.delimiters = this._prev_delimiters.pop();
    }

    token.level = this.level;

    if (nesting > 0) {
      // opening tag
      this.level++;
      this._prev_delimiters.push(this.delimiters);
      this.delimiters = [];
      token_meta = { delimiters: this.delimiters };
    }

    this.pendingLevel = this.level;
    this.tokens.push(token);
    this.tokens_meta.push(token_meta);
    return token;
  };


  // Scan a sequence of emphasis-like markers, and determine whether
  // it can start an emphasis sequence or end an emphasis sequence.
  //
  //  - start - position to scan from (it should point at a valid marker);
  //  - canSplitWord - determine if these markers can be found inside a word
  //
  StateInline$1.prototype.scanDelims = function (start, canSplitWord) {
    var pos = start, lastChar, nextChar, count, can_open, can_close,
        isLastWhiteSpace, isLastPunctChar,
        isNextWhiteSpace, isNextPunctChar,
        left_flanking = true,
        right_flanking = true,
        max = this.posMax,
        marker = this.src.charCodeAt(start);

    // treat beginning of the line as a whitespace
    lastChar = start > 0 ? this.src.charCodeAt(start - 1) : 0x20;

    while (pos < max && this.src.charCodeAt(pos) === marker) { pos++; }

    count = pos - start;

    // treat end of the line as a whitespace
    nextChar = pos < max ? this.src.charCodeAt(pos) : 0x20;

    isLastPunctChar = isMdAsciiPunct$2(lastChar) || isPunctChar$2(String.fromCharCode(lastChar));
    isNextPunctChar = isMdAsciiPunct$2(nextChar) || isPunctChar$2(String.fromCharCode(nextChar));

    isLastWhiteSpace = isWhiteSpace$2(lastChar);
    isNextWhiteSpace = isWhiteSpace$2(nextChar);

    if (isNextWhiteSpace) {
      left_flanking = false;
    } else if (isNextPunctChar) {
      if (!(isLastWhiteSpace || isLastPunctChar)) {
        left_flanking = false;
      }
    }

    if (isLastWhiteSpace) {
      right_flanking = false;
    } else if (isLastPunctChar) {
      if (!(isNextWhiteSpace || isNextPunctChar)) {
        right_flanking = false;
      }
    }

    if (!canSplitWord) {
      can_open  = left_flanking  && (!right_flanking || isLastPunctChar);
      can_close = right_flanking && (!left_flanking  || isNextPunctChar);
    } else {
      can_open  = left_flanking;
      can_close = right_flanking;
    }

    return {
      can_open:  can_open,
      can_close: can_close,
      length:    count
    };
  };


  // re-export Token class to use in block rules
  StateInline$1.prototype.Token = token$1;


  var state_inline$1 = StateInline$1;

  ////////////////////////////////////////////////////////////////////////////////
  // Parser rules

  var _rules$3 = [
    [ 'text',            text$2 ],
    [ 'newline',         newline$1 ],
    [ 'escape',          _escape$1 ],
    [ 'backticks',       backticks$1 ],
    [ 'strikethrough',   strikethrough$2.tokenize ],
    [ 'emphasis',        emphasis$1.tokenize ],
    [ 'link',            link$4 ],
    [ 'image',           image$3 ],
    [ 'autolink',        autolink$1 ],
    [ 'html_inline',     html_inline$1 ],
    [ 'entity',          entity$1 ]
  ];

  var _rules2$1 = [
    [ 'balance_pairs',   balance_pairs$1 ],
    [ 'strikethrough',   strikethrough$2.postProcess ],
    [ 'emphasis',        emphasis$1.postProcess ],
    [ 'text_collapse',   text_collapse$1 ]
  ];


  /**
   * new ParserInline()
   **/
  function ParserInline$1() {
    var i;

    /**
     * ParserInline#ruler -> Ruler
     *
     * [[Ruler]] instance. Keep configuration of inline rules.
     **/
    this.ruler = new ruler$1();

    for (i = 0; i < _rules$3.length; i++) {
      this.ruler.push(_rules$3[i][0], _rules$3[i][1]);
    }

    /**
     * ParserInline#ruler2 -> Ruler
     *
     * [[Ruler]] instance. Second ruler used for post-processing
     * (e.g. in emphasis-like rules).
     **/
    this.ruler2 = new ruler$1();

    for (i = 0; i < _rules2$1.length; i++) {
      this.ruler2.push(_rules2$1[i][0], _rules2$1[i][1]);
    }
  }


  // Skip single token by running all rules in validation mode;
  // returns `true` if any rule reported success
  //
  ParserInline$1.prototype.skipToken = function (state) {
    var ok, i, pos = state.pos,
        rules = this.ruler.getRules(''),
        len = rules.length,
        maxNesting = state.md.options.maxNesting,
        cache = state.cache;


    if (typeof cache[pos] !== 'undefined') {
      state.pos = cache[pos];
      return;
    }

    if (state.level < maxNesting) {
      for (i = 0; i < len; i++) {
        // Increment state.level and decrement it later to limit recursion.
        // It's harmless to do here, because no tokens are created. But ideally,
        // we'd need a separate private state variable for this purpose.
        //
        state.level++;
        ok = rules[i](state, true);
        state.level--;

        if (ok) { break; }
      }
    } else {
      // Too much nesting, just skip until the end of the paragraph.
      //
      // NOTE: this will cause links to behave incorrectly in the following case,
      //       when an amount of `[` is exactly equal to `maxNesting + 1`:
      //
      //       [[[[[[[[[[[[[[[[[[[[[foo]()
      //
      // TODO: remove this workaround when CM standard will allow nested links
      //       (we can replace it by preventing links from being parsed in
      //       validation mode)
      //
      state.pos = state.posMax;
    }

    if (!ok) { state.pos++; }
    cache[pos] = state.pos;
  };


  // Generate tokens for input range
  //
  ParserInline$1.prototype.tokenize = function (state) {
    var ok, i,
        rules = this.ruler.getRules(''),
        len = rules.length,
        end = state.posMax,
        maxNesting = state.md.options.maxNesting;

    while (state.pos < end) {
      // Try all possible rules.
      // On success, rule should:
      //
      // - update `state.pos`
      // - update `state.tokens`
      // - return true

      if (state.level < maxNesting) {
        for (i = 0; i < len; i++) {
          ok = rules[i](state, false);
          if (ok) { break; }
        }
      }

      if (ok) {
        if (state.pos >= end) { break; }
        continue;
      }

      state.pending += state.src[state.pos++];
    }

    if (state.pending) {
      state.pushPending();
    }
  };


  /**
   * ParserInline.parse(str, md, env, outTokens)
   *
   * Process input string and push inline tokens into `outTokens`
   **/
  ParserInline$1.prototype.parse = function (str, md, env, outTokens) {
    var i, rules, len;
    var state = new this.State(str, md, env, outTokens);

    this.tokenize(state);

    rules = this.ruler2.getRules('');
    len = rules.length;

    for (i = 0; i < len; i++) {
      rules[i](state);
    }
  };


  ParserInline$1.prototype.State = state_inline$1;


  var parser_inline$1 = ParserInline$1;

  var re$1 = function (opts) {
    var re = {};

    // Use direct extract instead of `regenerate` to reduse browserified size
    re.src_Any = regex$3.source;
    re.src_Cc  = regex$2.source;
    re.src_Z   = regex.source;
    re.src_P   = regex$4.source;

    // \p{\Z\P\Cc\CF} (white spaces + control + format + punctuation)
    re.src_ZPCc = [ re.src_Z, re.src_P, re.src_Cc ].join('|');

    // \p{\Z\Cc} (white spaces + control)
    re.src_ZCc = [ re.src_Z, re.src_Cc ].join('|');

    // Experimental. List of chars, completely prohibited in links
    // because can separate it from other part of text
    var text_separators = '[><\uff5c]';

    // All possible word characters (everything without punctuation, spaces & controls)
    // Defined via punctuation & spaces to save space
    // Should be something like \p{\L\N\S\M} (\w but without `_`)
    re.src_pseudo_letter       = '(?:(?!' + text_separators + '|' + re.src_ZPCc + ')' + re.src_Any + ')';
    // The same as abothe but without [0-9]
    // var src_pseudo_letter_non_d = '(?:(?![0-9]|' + src_ZPCc + ')' + src_Any + ')';

    ////////////////////////////////////////////////////////////////////////////////

    re.src_ip4 =

      '(?:(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)';

    // Prohibit any of "@/[]()" in user/pass to avoid wrong domain fetch.
    re.src_auth    = '(?:(?:(?!' + re.src_ZCc + '|[@/\\[\\]()]).)+@)?';

    re.src_port =

      '(?::(?:6(?:[0-4]\\d{3}|5(?:[0-4]\\d{2}|5(?:[0-2]\\d|3[0-5])))|[1-5]?\\d{1,4}))?';

    re.src_host_terminator =

      '(?=$|' + text_separators + '|' + re.src_ZPCc + ')(?!-|_|:\\d|\\.-|\\.(?!$|' + re.src_ZPCc + '))';

    re.src_path =

      '(?:' +
        '[/?#]' +
          '(?:' +
            '(?!' + re.src_ZCc + '|' + text_separators + '|[()[\\]{}.,"\'?!\\-]).|' +
            '\\[(?:(?!' + re.src_ZCc + '|\\]).)*\\]|' +
            '\\((?:(?!' + re.src_ZCc + '|[)]).)*\\)|' +
            '\\{(?:(?!' + re.src_ZCc + '|[}]).)*\\}|' +
            '\\"(?:(?!' + re.src_ZCc + '|["]).)+\\"|' +
            "\\'(?:(?!" + re.src_ZCc + "|[']).)+\\'|" +
            "\\'(?=" + re.src_pseudo_letter + '|[-]).|' +  // allow `I'm_king` if no pair found
            '\\.{2,4}[a-zA-Z0-9%/]|' + // github has ... in commit range links,
                                       // google has .... in links (issue #66)
                                       // Restrict to
                                       // - english
                                       // - percent-encoded
                                       // - parts of file path
                                       // until more examples found.
            '\\.(?!' + re.src_ZCc + '|[.]).|' +
            (opts && opts['---'] ?
              '\\-(?!--(?:[^-]|$))(?:-*)|' // `---` => long dash, terminate
              :
              '\\-+|'
            ) +
            '\\,(?!' + re.src_ZCc + ').|' +      // allow `,,,` in paths
            '\\!(?!' + re.src_ZCc + '|[!]).|' +
            '\\?(?!' + re.src_ZCc + '|[?]).' +
          ')+' +
        '|\\/' +
      ')?';

    // Allow anything in markdown spec, forbid quote (") at the first position
    // because emails enclosed in quotes are far more common
    re.src_email_name =

      '[\\-;:&=\\+\\$,\\.a-zA-Z0-9_][\\-;:&=\\+\\$,\\"\\.a-zA-Z0-9_]*';

    re.src_xn =

      'xn--[a-z0-9\\-]{1,59}';

    // More to read about domain names
    // http://serverfault.com/questions/638260/

    re.src_domain_root =

      // Allow letters & digits (http://test1)
      '(?:' +
        re.src_xn +
        '|' +
        re.src_pseudo_letter + '{1,63}' +
      ')';

    re.src_domain =

      '(?:' +
        re.src_xn +
        '|' +
        '(?:' + re.src_pseudo_letter + ')' +
        '|' +
        '(?:' + re.src_pseudo_letter + '(?:-|' + re.src_pseudo_letter + '){0,61}' + re.src_pseudo_letter + ')' +
      ')';

    re.src_host =

      '(?:' +
      // Don't need IP check, because digits are already allowed in normal domain names
      //   src_ip4 +
      // '|' +
        '(?:(?:(?:' + re.src_domain + ')\\.)*' + re.src_domain/*_root*/ + ')' +
      ')';

    re.tpl_host_fuzzy =

      '(?:' +
        re.src_ip4 +
      '|' +
        '(?:(?:(?:' + re.src_domain + ')\\.)+(?:%TLDS%))' +
      ')';

    re.tpl_host_no_ip_fuzzy =

      '(?:(?:(?:' + re.src_domain + ')\\.)+(?:%TLDS%))';

    re.src_host_strict =

      re.src_host + re.src_host_terminator;

    re.tpl_host_fuzzy_strict =

      re.tpl_host_fuzzy + re.src_host_terminator;

    re.src_host_port_strict =

      re.src_host + re.src_port + re.src_host_terminator;

    re.tpl_host_port_fuzzy_strict =

      re.tpl_host_fuzzy + re.src_port + re.src_host_terminator;

    re.tpl_host_port_no_ip_fuzzy_strict =

      re.tpl_host_no_ip_fuzzy + re.src_port + re.src_host_terminator;


    ////////////////////////////////////////////////////////////////////////////////
    // Main rules

    // Rude test fuzzy links by host, for quick deny
    re.tpl_host_fuzzy_test =

      'localhost|www\\.|\\.\\d{1,3}\\.|(?:\\.(?:%TLDS%)(?:' + re.src_ZPCc + '|>|$))';

    re.tpl_email_fuzzy =

        '(^|' + text_separators + '|"|\\(|' + re.src_ZCc + ')' +
        '(' + re.src_email_name + '@' + re.tpl_host_fuzzy_strict + ')';

    re.tpl_link_fuzzy =
        // Fuzzy link can't be prepended with .:/\- and non punctuation.
        // but can start with > (markdown blockquote)
        '(^|(?![.:/\\-_@])(?:[$+<=>^`|\uff5c]|' + re.src_ZPCc + '))' +
        '((?![$+<=>^`|\uff5c])' + re.tpl_host_port_fuzzy_strict + re.src_path + ')';

    re.tpl_link_no_ip_fuzzy =
        // Fuzzy link can't be prepended with .:/\- and non punctuation.
        // but can start with > (markdown blockquote)
        '(^|(?![.:/\\-_@])(?:[$+<=>^`|\uff5c]|' + re.src_ZPCc + '))' +
        '((?![$+<=>^`|\uff5c])' + re.tpl_host_port_no_ip_fuzzy_strict + re.src_path + ')';

    return re;
  };

  ////////////////////////////////////////////////////////////////////////////////
  // Helpers

  // Merge objects
  //
  function assign$2(obj /*from1, from2, from3, ...*/) {
    var sources = Array.prototype.slice.call(arguments, 1);

    sources.forEach(function (source) {
      if (!source) { return; }

      Object.keys(source).forEach(function (key) {
        obj[key] = source[key];
      });
    });

    return obj;
  }

  function _class$1(obj) { return Object.prototype.toString.call(obj); }
  function isString$1(obj) { return _class$1(obj) === '[object String]'; }
  function isObject$1(obj) { return _class$1(obj) === '[object Object]'; }
  function isRegExp$1(obj) { return _class$1(obj) === '[object RegExp]'; }
  function isFunction$1(obj) { return _class$1(obj) === '[object Function]'; }


  function escapeRE$1(str) { return str.replace(/[.?*+^$[\]\\(){}|-]/g, '\\$&'); }

  ////////////////////////////////////////////////////////////////////////////////


  var defaultOptions$1 = {
    fuzzyLink: true,
    fuzzyEmail: true,
    fuzzyIP: false
  };


  function isOptionsObj$1(obj) {
    return Object.keys(obj || {}).reduce(function (acc, k) {
      return acc || defaultOptions$1.hasOwnProperty(k);
    }, false);
  }


  var defaultSchemas$1 = {
    'http:': {
      validate: function (text, pos, self) {
        var tail = text.slice(pos);

        if (!self.re.http) {
          // compile lazily, because "host"-containing variables can change on tlds update.
          self.re.http =  new RegExp(
            '^\\/\\/' + self.re.src_auth + self.re.src_host_port_strict + self.re.src_path, 'i'
          );
        }
        if (self.re.http.test(tail)) {
          return tail.match(self.re.http)[0].length;
        }
        return 0;
      }
    },
    'https:':  'http:',
    'ftp:':    'http:',
    '//':      {
      validate: function (text, pos, self) {
        var tail = text.slice(pos);

        if (!self.re.no_http) {
        // compile lazily, because "host"-containing variables can change on tlds update.
          self.re.no_http =  new RegExp(
            '^' +
            self.re.src_auth +
            // Don't allow single-level domains, because of false positives like '//test'
            // with code comments
            '(?:localhost|(?:(?:' + self.re.src_domain + ')\\.)+' + self.re.src_domain_root + ')' +
            self.re.src_port +
            self.re.src_host_terminator +
            self.re.src_path,

            'i'
          );
        }

        if (self.re.no_http.test(tail)) {
          // should not be `://` & `///`, that protects from errors in protocol name
          if (pos >= 3 && text[pos - 3] === ':') { return 0; }
          if (pos >= 3 && text[pos - 3] === '/') { return 0; }
          return tail.match(self.re.no_http)[0].length;
        }
        return 0;
      }
    },
    'mailto:': {
      validate: function (text, pos, self) {
        var tail = text.slice(pos);

        if (!self.re.mailto) {
          self.re.mailto =  new RegExp(
            '^' + self.re.src_email_name + '@' + self.re.src_host_strict, 'i'
          );
        }
        if (self.re.mailto.test(tail)) {
          return tail.match(self.re.mailto)[0].length;
        }
        return 0;
      }
    }
  };

  /*eslint-disable max-len*/

  // RE pattern for 2-character tlds (autogenerated by ./support/tlds_2char_gen.js)
  var tlds_2ch_src_re$1 = 'a[cdefgilmnoqrstuwxz]|b[abdefghijmnorstvwyz]|c[acdfghiklmnoruvwxyz]|d[ejkmoz]|e[cegrstu]|f[ijkmor]|g[abdefghilmnpqrstuwy]|h[kmnrtu]|i[delmnoqrst]|j[emop]|k[eghimnprwyz]|l[abcikrstuvy]|m[acdeghklmnopqrstuvwxyz]|n[acefgilopruz]|om|p[aefghklmnrstwy]|qa|r[eosuw]|s[abcdeghijklmnortuvxyz]|t[cdfghjklmnortvwz]|u[agksyz]|v[aceginu]|w[fs]|y[et]|z[amw]';

  // DON'T try to make PRs with changes. Extend TLDs with LinkifyIt.tlds() instead
  var tlds_default$1 = 'biz|com|edu|gov|net|org|pro|web|xxx|aero|asia|coop|info|museum|name|shop|рф'.split('|');

  /*eslint-enable max-len*/

  ////////////////////////////////////////////////////////////////////////////////

  function resetScanCache$1(self) {
    self.__index__ = -1;
    self.__text_cache__   = '';
  }

  function createValidator$1(re) {
    return function (text, pos) {
      var tail = text.slice(pos);

      if (re.test(tail)) {
        return tail.match(re)[0].length;
      }
      return 0;
    };
  }

  function createNormalizer$1() {
    return function (match, self) {
      self.normalize(match);
    };
  }

  // Schemas compiler. Build regexps.
  //
  function compile$1(self) {

    // Load & clone RE patterns.
    var re = self.re = re$1(self.__opts__);

    // Define dynamic patterns
    var tlds = self.__tlds__.slice();

    self.onCompile();

    if (!self.__tlds_replaced__) {
      tlds.push(tlds_2ch_src_re$1);
    }
    tlds.push(re.src_xn);

    re.src_tlds = tlds.join('|');

    function untpl(tpl) { return tpl.replace('%TLDS%', re.src_tlds); }

    re.email_fuzzy      = RegExp(untpl(re.tpl_email_fuzzy), 'i');
    re.link_fuzzy       = RegExp(untpl(re.tpl_link_fuzzy), 'i');
    re.link_no_ip_fuzzy = RegExp(untpl(re.tpl_link_no_ip_fuzzy), 'i');
    re.host_fuzzy_test  = RegExp(untpl(re.tpl_host_fuzzy_test), 'i');

    //
    // Compile each schema
    //

    var aliases = [];

    self.__compiled__ = {}; // Reset compiled data

    function schemaError(name, val) {
      throw new Error('(LinkifyIt) Invalid schema "' + name + '": ' + val);
    }

    Object.keys(self.__schemas__).forEach(function (name) {
      var val = self.__schemas__[name];

      // skip disabled methods
      if (val === null) { return; }

      var compiled = { validate: null, link: null };

      self.__compiled__[name] = compiled;

      if (isObject$1(val)) {
        if (isRegExp$1(val.validate)) {
          compiled.validate = createValidator$1(val.validate);
        } else if (isFunction$1(val.validate)) {
          compiled.validate = val.validate;
        } else {
          schemaError(name, val);
        }

        if (isFunction$1(val.normalize)) {
          compiled.normalize = val.normalize;
        } else if (!val.normalize) {
          compiled.normalize = createNormalizer$1();
        } else {
          schemaError(name, val);
        }

        return;
      }

      if (isString$1(val)) {
        aliases.push(name);
        return;
      }

      schemaError(name, val);
    });

    //
    // Compile postponed aliases
    //

    aliases.forEach(function (alias) {
      if (!self.__compiled__[self.__schemas__[alias]]) {
        // Silently fail on missed schemas to avoid errons on disable.
        // schemaError(alias, self.__schemas__[alias]);
        return;
      }

      self.__compiled__[alias].validate =
        self.__compiled__[self.__schemas__[alias]].validate;
      self.__compiled__[alias].normalize =
        self.__compiled__[self.__schemas__[alias]].normalize;
    });

    //
    // Fake record for guessed links
    //
    self.__compiled__[''] = { validate: null, normalize: createNormalizer$1() };

    //
    // Build schema condition
    //
    var slist = Object.keys(self.__compiled__)
                        .filter(function (name) {
                          // Filter disabled & fake schemas
                          return name.length > 0 && self.__compiled__[name];
                        })
                        .map(escapeRE$1)
                        .join('|');
    // (?!_) cause 1.5x slowdown
    self.re.schema_test   = RegExp('(^|(?!_)(?:[><\uff5c]|' + re.src_ZPCc + '))(' + slist + ')', 'i');
    self.re.schema_search = RegExp('(^|(?!_)(?:[><\uff5c]|' + re.src_ZPCc + '))(' + slist + ')', 'ig');

    self.re.pretest = RegExp(
      '(' + self.re.schema_test.source + ')|(' + self.re.host_fuzzy_test.source + ')|@',
      'i'
    );

    //
    // Cleanup
    //

    resetScanCache$1(self);
  }

  /**
   * class Match
   *
   * Match result. Single element of array, returned by [[LinkifyIt#match]]
   **/
  function Match$1(self, shift) {
    var start = self.__index__,
        end   = self.__last_index__,
        text  = self.__text_cache__.slice(start, end);

    /**
     * Match#schema -> String
     *
     * Prefix (protocol) for matched string.
     **/
    this.schema    = self.__schema__.toLowerCase();
    /**
     * Match#index -> Number
     *
     * First position of matched string.
     **/
    this.index     = start + shift;
    /**
     * Match#lastIndex -> Number
     *
     * Next position after matched string.
     **/
    this.lastIndex = end + shift;
    /**
     * Match#raw -> String
     *
     * Matched string.
     **/
    this.raw       = text;
    /**
     * Match#text -> String
     *
     * Notmalized text of matched string.
     **/
    this.text      = text;
    /**
     * Match#url -> String
     *
     * Normalized url of matched string.
     **/
    this.url       = text;
  }

  function createMatch$1(self, shift) {
    var match = new Match$1(self, shift);

    self.__compiled__[match.schema].normalize(match, self);

    return match;
  }


  /**
   * class LinkifyIt
   **/

  /**
   * new LinkifyIt(schemas, options)
   * - schemas (Object): Optional. Additional schemas to validate (prefix/validator)
   * - options (Object): { fuzzyLink|fuzzyEmail|fuzzyIP: true|false }
   *
   * Creates new linkifier instance with optional additional schemas.
   * Can be called without `new` keyword for convenience.
   *
   * By default understands:
   *
   * - `http(s)://...` , `ftp://...`, `mailto:...` & `//...` links
   * - "fuzzy" links and emails (example.com, foo@bar.com).
   *
   * `schemas` is an object, where each key/value describes protocol/rule:
   *
   * - __key__ - link prefix (usually, protocol name with `:` at the end, `skype:`
   *   for example). `linkify-it` makes shure that prefix is not preceeded with
   *   alphanumeric char and symbols. Only whitespaces and punctuation allowed.
   * - __value__ - rule to check tail after link prefix
   *   - _String_ - just alias to existing rule
   *   - _Object_
   *     - _validate_ - validator function (should return matched length on success),
   *       or `RegExp`.
   *     - _normalize_ - optional function to normalize text & url of matched result
   *       (for example, for @twitter mentions).
   *
   * `options`:
   *
   * - __fuzzyLink__ - recognige URL-s without `http(s):` prefix. Default `true`.
   * - __fuzzyIP__ - allow IPs in fuzzy links above. Can conflict with some texts
   *   like version numbers. Default `false`.
   * - __fuzzyEmail__ - recognize emails without `mailto:` prefix.
   *
   **/
  function LinkifyIt$1(schemas, options) {
    if (!(this instanceof LinkifyIt$1)) {
      return new LinkifyIt$1(schemas, options);
    }

    if (!options) {
      if (isOptionsObj$1(schemas)) {
        options = schemas;
        schemas = {};
      }
    }

    this.__opts__           = assign$2({}, defaultOptions$1, options);

    // Cache last tested result. Used to skip repeating steps on next `match` call.
    this.__index__          = -1;
    this.__last_index__     = -1; // Next scan position
    this.__schema__         = '';
    this.__text_cache__     = '';

    this.__schemas__        = assign$2({}, defaultSchemas$1, schemas);
    this.__compiled__       = {};

    this.__tlds__           = tlds_default$1;
    this.__tlds_replaced__  = false;

    this.re = {};

    compile$1(this);
  }


  /** chainable
   * LinkifyIt#add(schema, definition)
   * - schema (String): rule name (fixed pattern prefix)
   * - definition (String|RegExp|Object): schema definition
   *
   * Add new rule definition. See constructor description for details.
   **/
  LinkifyIt$1.prototype.add = function add(schema, definition) {
    this.__schemas__[schema] = definition;
    compile$1(this);
    return this;
  };


  /** chainable
   * LinkifyIt#set(options)
   * - options (Object): { fuzzyLink|fuzzyEmail|fuzzyIP: true|false }
   *
   * Set recognition options for links without schema.
   **/
  LinkifyIt$1.prototype.set = function set(options) {
    this.__opts__ = assign$2(this.__opts__, options);
    return this;
  };


  /**
   * LinkifyIt#test(text) -> Boolean
   *
   * Searches linkifiable pattern and returns `true` on success or `false` on fail.
   **/
  LinkifyIt$1.prototype.test = function test(text) {
    // Reset scan cache
    this.__text_cache__ = text;
    this.__index__      = -1;

    if (!text.length) { return false; }

    var m, ml, me, len, shift, next, re, tld_pos, at_pos;

    // try to scan for link with schema - that's the most simple rule
    if (this.re.schema_test.test(text)) {
      re = this.re.schema_search;
      re.lastIndex = 0;
      while ((m = re.exec(text)) !== null) {
        len = this.testSchemaAt(text, m[2], re.lastIndex);
        if (len) {
          this.__schema__     = m[2];
          this.__index__      = m.index + m[1].length;
          this.__last_index__ = m.index + m[0].length + len;
          break;
        }
      }
    }

    if (this.__opts__.fuzzyLink && this.__compiled__['http:']) {
      // guess schemaless links
      tld_pos = text.search(this.re.host_fuzzy_test);
      if (tld_pos >= 0) {
        // if tld is located after found link - no need to check fuzzy pattern
        if (this.__index__ < 0 || tld_pos < this.__index__) {
          if ((ml = text.match(this.__opts__.fuzzyIP ? this.re.link_fuzzy : this.re.link_no_ip_fuzzy)) !== null) {

            shift = ml.index + ml[1].length;

            if (this.__index__ < 0 || shift < this.__index__) {
              this.__schema__     = '';
              this.__index__      = shift;
              this.__last_index__ = ml.index + ml[0].length;
            }
          }
        }
      }
    }

    if (this.__opts__.fuzzyEmail && this.__compiled__['mailto:']) {
      // guess schemaless emails
      at_pos = text.indexOf('@');
      if (at_pos >= 0) {
        // We can't skip this check, because this cases are possible:
        // 192.168.1.1@gmail.com, my.in@example.com
        if ((me = text.match(this.re.email_fuzzy)) !== null) {

          shift = me.index + me[1].length;
          next  = me.index + me[0].length;

          if (this.__index__ < 0 || shift < this.__index__ ||
              (shift === this.__index__ && next > this.__last_index__)) {
            this.__schema__     = 'mailto:';
            this.__index__      = shift;
            this.__last_index__ = next;
          }
        }
      }
    }

    return this.__index__ >= 0;
  };


  /**
   * LinkifyIt#pretest(text) -> Boolean
   *
   * Very quick check, that can give false positives. Returns true if link MAY BE
   * can exists. Can be used for speed optimization, when you need to check that
   * link NOT exists.
   **/
  LinkifyIt$1.prototype.pretest = function pretest(text) {
    return this.re.pretest.test(text);
  };


  /**
   * LinkifyIt#testSchemaAt(text, name, position) -> Number
   * - text (String): text to scan
   * - name (String): rule (schema) name
   * - position (Number): text offset to check from
   *
   * Similar to [[LinkifyIt#test]] but checks only specific protocol tail exactly
   * at given position. Returns length of found pattern (0 on fail).
   **/
  LinkifyIt$1.prototype.testSchemaAt = function testSchemaAt(text, schema, pos) {
    // If not supported schema check requested - terminate
    if (!this.__compiled__[schema.toLowerCase()]) {
      return 0;
    }
    return this.__compiled__[schema.toLowerCase()].validate(text, pos, this);
  };


  /**
   * LinkifyIt#match(text) -> Array|null
   *
   * Returns array of found link descriptions or `null` on fail. We strongly
   * recommend to use [[LinkifyIt#test]] first, for best speed.
   *
   * ##### Result match description
   *
   * - __schema__ - link schema, can be empty for fuzzy links, or `//` for
   *   protocol-neutral  links.
   * - __index__ - offset of matched text
   * - __lastIndex__ - index of next char after mathch end
   * - __raw__ - matched text
   * - __text__ - normalized text
   * - __url__ - link, generated from matched text
   **/
  LinkifyIt$1.prototype.match = function match(text) {
    var shift = 0, result = [];

    // Try to take previous element from cache, if .test() called before
    if (this.__index__ >= 0 && this.__text_cache__ === text) {
      result.push(createMatch$1(this, shift));
      shift = this.__last_index__;
    }

    // Cut head if cache was used
    var tail = shift ? text.slice(shift) : text;

    // Scan string until end reached
    while (this.test(tail)) {
      result.push(createMatch$1(this, shift));

      tail = tail.slice(this.__last_index__);
      shift += this.__last_index__;
    }

    if (result.length) {
      return result;
    }

    return null;
  };


  /** chainable
   * LinkifyIt#tlds(list [, keepOld]) -> this
   * - list (Array): list of tlds
   * - keepOld (Boolean): merge with current list if `true` (`false` by default)
   *
   * Load (or merge) new tlds list. Those are user for fuzzy links (without prefix)
   * to avoid false positives. By default this algorythm used:
   *
   * - hostname with any 2-letter root zones are ok.
   * - biz|com|edu|gov|net|org|pro|web|xxx|aero|asia|coop|info|museum|name|shop|рф
   *   are ok.
   * - encoded (`xn--...`) root zones are ok.
   *
   * If list is replaced, then exact match for 2-chars root zones will be checked.
   **/
  LinkifyIt$1.prototype.tlds = function tlds(list, keepOld) {
    list = Array.isArray(list) ? list : [ list ];

    if (!keepOld) {
      this.__tlds__ = list.slice();
      this.__tlds_replaced__ = true;
      compile$1(this);
      return this;
    }

    this.__tlds__ = this.__tlds__.concat(list)
                                    .sort()
                                    .filter(function (el, idx, arr) {
                                      return el !== arr[idx - 1];
                                    })
                                    .reverse();

    compile$1(this);
    return this;
  };

  /**
   * LinkifyIt#normalize(match)
   *
   * Default normalizer (if schema does not define it's own).
   **/
  LinkifyIt$1.prototype.normalize = function normalize(match) {

    // Do minimal possible changes by default. Need to collect feedback prior
    // to move forward https://github.com/markdown-it/linkify-it/issues/1

    if (!match.schema) { match.url = 'http://' + match.url; }

    if (match.schema === 'mailto:' && !/^mailto:/i.test(match.url)) {
      match.url = 'mailto:' + match.url;
    }
  };


  /**
   * LinkifyIt#onCompile()
   *
   * Override to modify basic RegExp-s.
   **/
  LinkifyIt$1.prototype.onCompile = function onCompile() {
  };


  var linkifyIt$1 = LinkifyIt$1;

  /*! https://mths.be/punycode v1.4.1 by @mathias */


  /** Highest positive signed 32-bit float value */
  var maxInt = 2147483647; // aka. 0x7FFFFFFF or 2^31-1

  /** Bootstring parameters */
  var base = 36;
  var tMin = 1;
  var tMax = 26;
  var skew = 38;
  var damp = 700;
  var initialBias = 72;
  var initialN = 128; // 0x80
  var delimiter = '-'; // '\x2D'

  /** Regular expressions */
  var regexPunycode = /^xn--/;
  var regexNonASCII = /[^\x20-\x7E]/; // unprintable ASCII chars + non-ASCII chars
  var regexSeparators = /[\x2E\u3002\uFF0E\uFF61]/g; // RFC 3490 separators

  /** Error messages */
  var errors = {
    'overflow': 'Overflow: input needs wider integers to process',
    'not-basic': 'Illegal input >= 0x80 (not a basic code point)',
    'invalid-input': 'Invalid input'
  };

  /** Convenience shortcuts */
  var baseMinusTMin = base - tMin;
  var floor = Math.floor;
  var stringFromCharCode = String.fromCharCode;

  /*--------------------------------------------------------------------------*/

  /**
   * A generic error utility function.
   * @private
   * @param {String} type The error type.
   * @returns {Error} Throws a `RangeError` with the applicable error message.
   */
  function error(type) {
    throw new RangeError(errors[type]);
  }

  /**
   * A generic `Array#map` utility function.
   * @private
   * @param {Array} array The array to iterate over.
   * @param {Function} callback The function that gets called for every array
   * item.
   * @returns {Array} A new array of values returned by the callback function.
   */
  function map$1(array, fn) {
    var length = array.length;
    var result = [];
    while (length--) {
      result[length] = fn(array[length]);
    }
    return result;
  }

  /**
   * A simple `Array#map`-like wrapper to work with domain name strings or email
   * addresses.
   * @private
   * @param {String} domain The domain name or email address.
   * @param {Function} callback The function that gets called for every
   * character.
   * @returns {Array} A new string of characters returned by the callback
   * function.
   */
  function mapDomain(string, fn) {
    var parts = string.split('@');
    var result = '';
    if (parts.length > 1) {
      // In email addresses, only the domain name should be punycoded. Leave
      // the local part (i.e. everything up to `@`) intact.
      result = parts[0] + '@';
      string = parts[1];
    }
    // Avoid `split(regex)` for IE8 compatibility. See #17.
    string = string.replace(regexSeparators, '\x2E');
    var labels = string.split('.');
    var encoded = map$1(labels, fn).join('.');
    return result + encoded;
  }

  /**
   * Creates an array containing the numeric code points of each Unicode
   * character in the string. While JavaScript uses UCS-2 internally,
   * this function will convert a pair of surrogate halves (each of which
   * UCS-2 exposes as separate characters) into a single code point,
   * matching UTF-16.
   * @see `punycode.ucs2.encode`
   * @see <https://mathiasbynens.be/notes/javascript-encoding>
   * @memberOf punycode.ucs2
   * @name decode
   * @param {String} string The Unicode input string (UCS-2).
   * @returns {Array} The new array of code points.
   */
  function ucs2decode(string) {
    var output = [],
      counter = 0,
      length = string.length,
      value,
      extra;
    while (counter < length) {
      value = string.charCodeAt(counter++);
      if (value >= 0xD800 && value <= 0xDBFF && counter < length) {
        // high surrogate, and there is a next character
        extra = string.charCodeAt(counter++);
        if ((extra & 0xFC00) == 0xDC00) { // low surrogate
          output.push(((value & 0x3FF) << 10) + (extra & 0x3FF) + 0x10000);
        } else {
          // unmatched surrogate; only append this code unit, in case the next
          // code unit is the high surrogate of a surrogate pair
          output.push(value);
          counter--;
        }
      } else {
        output.push(value);
      }
    }
    return output;
  }

  /**
   * Creates a string based on an array of numeric code points.
   * @see `punycode.ucs2.decode`
   * @memberOf punycode.ucs2
   * @name encode
   * @param {Array} codePoints The array of numeric code points.
   * @returns {String} The new Unicode string (UCS-2).
   */
  function ucs2encode(array) {
    return map$1(array, function(value) {
      var output = '';
      if (value > 0xFFFF) {
        value -= 0x10000;
        output += stringFromCharCode(value >>> 10 & 0x3FF | 0xD800);
        value = 0xDC00 | value & 0x3FF;
      }
      output += stringFromCharCode(value);
      return output;
    }).join('');
  }

  /**
   * Converts a basic code point into a digit/integer.
   * @see `digitToBasic()`
   * @private
   * @param {Number} codePoint The basic numeric code point value.
   * @returns {Number} The numeric value of a basic code point (for use in
   * representing integers) in the range `0` to `base - 1`, or `base` if
   * the code point does not represent a value.
   */
  function basicToDigit(codePoint) {
    if (codePoint - 48 < 10) {
      return codePoint - 22;
    }
    if (codePoint - 65 < 26) {
      return codePoint - 65;
    }
    if (codePoint - 97 < 26) {
      return codePoint - 97;
    }
    return base;
  }

  /**
   * Converts a digit/integer into a basic code point.
   * @see `basicToDigit()`
   * @private
   * @param {Number} digit The numeric value of a basic code point.
   * @returns {Number} The basic code point whose value (when used for
   * representing integers) is `digit`, which needs to be in the range
   * `0` to `base - 1`. If `flag` is non-zero, the uppercase form is
   * used; else, the lowercase form is used. The behavior is undefined
   * if `flag` is non-zero and `digit` has no uppercase form.
   */
  function digitToBasic(digit, flag) {
    //  0..25 map to ASCII a..z or A..Z
    // 26..35 map to ASCII 0..9
    return digit + 22 + 75 * (digit < 26) - ((flag != 0) << 5);
  }

  /**
   * Bias adaptation function as per section 3.4 of RFC 3492.
   * https://tools.ietf.org/html/rfc3492#section-3.4
   * @private
   */
  function adapt(delta, numPoints, firstTime) {
    var k = 0;
    delta = firstTime ? floor(delta / damp) : delta >> 1;
    delta += floor(delta / numPoints);
    for ( /* no initialization */ ; delta > baseMinusTMin * tMax >> 1; k += base) {
      delta = floor(delta / baseMinusTMin);
    }
    return floor(k + (baseMinusTMin + 1) * delta / (delta + skew));
  }

  /**
   * Converts a Punycode string of ASCII-only symbols to a string of Unicode
   * symbols.
   * @memberOf punycode
   * @param {String} input The Punycode string of ASCII-only symbols.
   * @returns {String} The resulting string of Unicode symbols.
   */
  function decode(input) {
    // Don't use UCS-2
    var output = [],
      inputLength = input.length,
      out,
      i = 0,
      n = initialN,
      bias = initialBias,
      basic,
      j,
      index,
      oldi,
      w,
      k,
      digit,
      t,
      /** Cached calculation results */
      baseMinusT;

    // Handle the basic code points: let `basic` be the number of input code
    // points before the last delimiter, or `0` if there is none, then copy
    // the first basic code points to the output.

    basic = input.lastIndexOf(delimiter);
    if (basic < 0) {
      basic = 0;
    }

    for (j = 0; j < basic; ++j) {
      // if it's not a basic code point
      if (input.charCodeAt(j) >= 0x80) {
        error('not-basic');
      }
      output.push(input.charCodeAt(j));
    }

    // Main decoding loop: start just after the last delimiter if any basic code
    // points were copied; start at the beginning otherwise.

    for (index = basic > 0 ? basic + 1 : 0; index < inputLength; /* no final expression */ ) {

      // `index` is the index of the next character to be consumed.
      // Decode a generalized variable-length integer into `delta`,
      // which gets added to `i`. The overflow checking is easier
      // if we increase `i` as we go, then subtract off its starting
      // value at the end to obtain `delta`.
      for (oldi = i, w = 1, k = base; /* no condition */ ; k += base) {

        if (index >= inputLength) {
          error('invalid-input');
        }

        digit = basicToDigit(input.charCodeAt(index++));

        if (digit >= base || digit > floor((maxInt - i) / w)) {
          error('overflow');
        }

        i += digit * w;
        t = k <= bias ? tMin : (k >= bias + tMax ? tMax : k - bias);

        if (digit < t) {
          break;
        }

        baseMinusT = base - t;
        if (w > floor(maxInt / baseMinusT)) {
          error('overflow');
        }

        w *= baseMinusT;

      }

      out = output.length + 1;
      bias = adapt(i - oldi, out, oldi == 0);

      // `i` was supposed to wrap around from `out` to `0`,
      // incrementing `n` each time, so we'll fix that now:
      if (floor(i / out) > maxInt - n) {
        error('overflow');
      }

      n += floor(i / out);
      i %= out;

      // Insert `n` at position `i` of the output
      output.splice(i++, 0, n);

    }

    return ucs2encode(output);
  }

  /**
   * Converts a string of Unicode symbols (e.g. a domain name label) to a
   * Punycode string of ASCII-only symbols.
   * @memberOf punycode
   * @param {String} input The string of Unicode symbols.
   * @returns {String} The resulting Punycode string of ASCII-only symbols.
   */
  function encode(input) {
    var n,
      delta,
      handledCPCount,
      basicLength,
      bias,
      j,
      m,
      q,
      k,
      t,
      currentValue,
      output = [],
      /** `inputLength` will hold the number of code points in `input`. */
      inputLength,
      /** Cached calculation results */
      handledCPCountPlusOne,
      baseMinusT,
      qMinusT;

    // Convert the input in UCS-2 to Unicode
    input = ucs2decode(input);

    // Cache the length
    inputLength = input.length;

    // Initialize the state
    n = initialN;
    delta = 0;
    bias = initialBias;

    // Handle the basic code points
    for (j = 0; j < inputLength; ++j) {
      currentValue = input[j];
      if (currentValue < 0x80) {
        output.push(stringFromCharCode(currentValue));
      }
    }

    handledCPCount = basicLength = output.length;

    // `handledCPCount` is the number of code points that have been handled;
    // `basicLength` is the number of basic code points.

    // Finish the basic string - if it is not empty - with a delimiter
    if (basicLength) {
      output.push(delimiter);
    }

    // Main encoding loop:
    while (handledCPCount < inputLength) {

      // All non-basic code points < n have been handled already. Find the next
      // larger one:
      for (m = maxInt, j = 0; j < inputLength; ++j) {
        currentValue = input[j];
        if (currentValue >= n && currentValue < m) {
          m = currentValue;
        }
      }

      // Increase `delta` enough to advance the decoder's <n,i> state to <m,0>,
      // but guard against overflow
      handledCPCountPlusOne = handledCPCount + 1;
      if (m - n > floor((maxInt - delta) / handledCPCountPlusOne)) {
        error('overflow');
      }

      delta += (m - n) * handledCPCountPlusOne;
      n = m;

      for (j = 0; j < inputLength; ++j) {
        currentValue = input[j];

        if (currentValue < n && ++delta > maxInt) {
          error('overflow');
        }

        if (currentValue == n) {
          // Represent delta as a generalized variable-length integer
          for (q = delta, k = base; /* no condition */ ; k += base) {
            t = k <= bias ? tMin : (k >= bias + tMax ? tMax : k - bias);
            if (q < t) {
              break;
            }
            qMinusT = q - t;
            baseMinusT = base - t;
            output.push(
              stringFromCharCode(digitToBasic(t + qMinusT % baseMinusT, 0))
            );
            q = floor(qMinusT / baseMinusT);
          }

          output.push(stringFromCharCode(digitToBasic(q, 0)));
          bias = adapt(delta, handledCPCountPlusOne, handledCPCount == basicLength);
          delta = 0;
          ++handledCPCount;
        }
      }

      ++delta;
      ++n;

    }
    return output.join('');
  }

  /**
   * Converts a Punycode string representing a domain name or an email address
   * to Unicode. Only the Punycoded parts of the input will be converted, i.e.
   * it doesn't matter if you call it on a string that has already been
   * converted to Unicode.
   * @memberOf punycode
   * @param {String} input The Punycoded domain name or email address to
   * convert to Unicode.
   * @returns {String} The Unicode representation of the given Punycode
   * string.
   */
  function toUnicode(input) {
    return mapDomain(input, function(string) {
      return regexPunycode.test(string) ?
        decode(string.slice(4).toLowerCase()) :
        string;
    });
  }

  /**
   * Converts a Unicode string representing a domain name or an email address to
   * Punycode. Only the non-ASCII parts of the domain name will be converted,
   * i.e. it doesn't matter if you call it with a domain that's already in
   * ASCII.
   * @memberOf punycode
   * @param {String} input The domain name or email address to convert, as a
   * Unicode string.
   * @returns {String} The Punycode representation of the given domain name or
   * email address.
   */
  function toASCII(input) {
    return mapDomain(input, function(string) {
      return regexNonASCII.test(string) ?
        'xn--' + encode(string) :
        string;
    });
  }
  var version = '1.4.1';
  /**
   * An object of methods to convert from JavaScript's internal character
   * representation (UCS-2) to Unicode code points, and back.
   * @see <https://mathiasbynens.be/notes/javascript-encoding>
   * @memberOf punycode
   * @type Object
   */

  var ucs2 = {
    decode: ucs2decode,
    encode: ucs2encode
  };
  var punycode = {
    version: version,
    ucs2: ucs2,
    toASCII: toASCII,
    toUnicode: toUnicode,
    encode: encode,
    decode: decode
  };

  // markdown-it default options


  var _default$1 = {
    options: {
      html:         false,        // Enable HTML tags in source
      xhtmlOut:     false,        // Use '/' to close single tags (<br />)
      breaks:       false,        // Convert '\n' in paragraphs into <br>
      langPrefix:   'language-',  // CSS language prefix for fenced blocks
      linkify:      false,        // autoconvert URL-like texts to links

      // Enable some language-neutral replacements + quotes beautification
      typographer:  false,

      // Double + single quotes replacement pairs, when typographer enabled,
      // and smartquotes on. Could be either a String or an Array.
      //
      // For example, you can use '«»„“' for Russian, '„“‚‘' for German,
      // and ['«\xA0', '\xA0»', '‹\xA0', '\xA0›'] for French (including nbsp).
      quotes: '\u201c\u201d\u2018\u2019', /* “”‘’ */

      // Highlighter function. Should return escaped HTML,
      // or '' if the source string is not changed and should be escaped externaly.
      // If result starts with <pre... internal wrapper is skipped.
      //
      // function (/*str, lang*/) { return ''; }
      //
      highlight: null,

      maxNesting:   100            // Internal protection, recursion limit
    },

    components: {

      core: {},
      block: {},
      inline: {}
    }
  };
  _default$1.options;
  _default$1.components;

  // "Zero" preset, with nothing enabled. Useful for manual configuring of simple


  var zero$3 = {
    options: {
      html:         false,        // Enable HTML tags in source
      xhtmlOut:     false,        // Use '/' to close single tags (<br />)
      breaks:       false,        // Convert '\n' in paragraphs into <br>
      langPrefix:   'language-',  // CSS language prefix for fenced blocks
      linkify:      false,        // autoconvert URL-like texts to links

      // Enable some language-neutral replacements + quotes beautification
      typographer:  false,

      // Double + single quotes replacement pairs, when typographer enabled,
      // and smartquotes on. Could be either a String or an Array.
      //
      // For example, you can use '«»„“' for Russian, '„“‚‘' for German,
      // and ['«\xA0', '\xA0»', '‹\xA0', '\xA0›'] for French (including nbsp).
      quotes: '\u201c\u201d\u2018\u2019', /* “”‘’ */

      // Highlighter function. Should return escaped HTML,
      // or '' if the source string is not changed and should be escaped externaly.
      // If result starts with <pre... internal wrapper is skipped.
      //
      // function (/*str, lang*/) { return ''; }
      //
      highlight: null,

      maxNesting:   20            // Internal protection, recursion limit
    },

    components: {

      core: {
        rules: [
          'normalize',
          'block',
          'inline'
        ]
      },

      block: {
        rules: [
          'paragraph'
        ]
      },

      inline: {
        rules: [
          'text'
        ],
        rules2: [
          'balance_pairs',
          'text_collapse'
        ]
      }
    }
  };
  zero$3.options;
  zero$3.components;

  // Commonmark default options


  var commonmark$1 = {
    options: {
      html:         true,         // Enable HTML tags in source
      xhtmlOut:     true,         // Use '/' to close single tags (<br />)
      breaks:       false,        // Convert '\n' in paragraphs into <br>
      langPrefix:   'language-',  // CSS language prefix for fenced blocks
      linkify:      false,        // autoconvert URL-like texts to links

      // Enable some language-neutral replacements + quotes beautification
      typographer:  false,

      // Double + single quotes replacement pairs, when typographer enabled,
      // and smartquotes on. Could be either a String or an Array.
      //
      // For example, you can use '«»„“' for Russian, '„“‚‘' for German,
      // and ['«\xA0', '\xA0»', '‹\xA0', '\xA0›'] for French (including nbsp).
      quotes: '\u201c\u201d\u2018\u2019', /* “”‘’ */

      // Highlighter function. Should return escaped HTML,
      // or '' if the source string is not changed and should be escaped externaly.
      // If result starts with <pre... internal wrapper is skipped.
      //
      // function (/*str, lang*/) { return ''; }
      //
      highlight: null,

      maxNesting:   20            // Internal protection, recursion limit
    },

    components: {

      core: {
        rules: [
          'normalize',
          'block',
          'inline'
        ]
      },

      block: {
        rules: [
          'blockquote',
          'code',
          'fence',
          'heading',
          'hr',
          'html_block',
          'lheading',
          'list',
          'reference',
          'paragraph'
        ]
      },

      inline: {
        rules: [
          'autolink',
          'backticks',
          'emphasis',
          'entity',
          'escape',
          'html_inline',
          'image',
          'link',
          'newline',
          'text'
        ],
        rules2: [
          'balance_pairs',
          'emphasis',
          'text_collapse'
        ]
      }
    }
  };
  commonmark$1.options;
  commonmark$1.components;

  var config$1 = {
    'default': _default$1,
    zero: zero$3,
    commonmark: commonmark$1
  };

  ////////////////////////////////////////////////////////////////////////////////
  //
  // This validator can prohibit more than really needed to prevent XSS. It's a
  // tradeoff to keep code simple and to be secure by default.
  //
  // If you need different setup - override validator method as you wish. Or
  // replace it with dummy function and use external sanitizer.
  //

  var BAD_PROTO_RE$1 = /^(vbscript|javascript|file|data):/;
  var GOOD_DATA_RE$1 = /^data:image\/(gif|png|jpeg|webp);/;

  function validateLink$1(url) {
    // url should be normalized at this point, and existing entities are decoded
    var str = url.trim().toLowerCase();

    return BAD_PROTO_RE$1.test(str) ? (GOOD_DATA_RE$1.test(str) ? true : false) : true;
  }

  ////////////////////////////////////////////////////////////////////////////////


  var RECODE_HOSTNAME_FOR$1 = [ 'http:', 'https:', 'mailto:' ];

  function normalizeLink$1(url) {
    var parsed = mdurl.parse(url, true);

    if (parsed.hostname) {
      // Encode hostnames in urls like:
      // `http://host/`, `https://host/`, `mailto:user@host`, `//host/`
      //
      // We don't encode unknown schemas, because it's likely that we encode
      // something we shouldn't (e.g. `skype:name` treated as `skype:host`)
      //
      if (!parsed.protocol || RECODE_HOSTNAME_FOR$1.indexOf(parsed.protocol) >= 0) {
        try {
          parsed.hostname = punycode.toASCII(parsed.hostname);
        } catch (er) { /**/ }
      }
    }

    return mdurl.encode(mdurl.format(parsed));
  }

  function normalizeLinkText$1(url) {
    var parsed = mdurl.parse(url, true);

    if (parsed.hostname) {
      // Encode hostnames in urls like:
      // `http://host/`, `https://host/`, `mailto:user@host`, `//host/`
      //
      // We don't encode unknown schemas, because it's likely that we encode
      // something we shouldn't (e.g. `skype:name` treated as `skype:host`)
      //
      if (!parsed.protocol || RECODE_HOSTNAME_FOR$1.indexOf(parsed.protocol) >= 0) {
        try {
          parsed.hostname = punycode.toUnicode(parsed.hostname);
        } catch (er) { /**/ }
      }
    }

    return mdurl.decode(mdurl.format(parsed));
  }


  /**
   * class MarkdownIt
   *
   * Main parser/renderer class.
   *
   * ##### Usage
   *
   * ```javascript
   * // node.js, "classic" way:
   * var MarkdownIt = require('markdown-it'),
   *     md = new MarkdownIt();
   * var result = md.render('# markdown-it rulezz!');
   *
   * // node.js, the same, but with sugar:
   * var md = require('markdown-it')();
   * var result = md.render('# markdown-it rulezz!');
   *
   * // browser without AMD, added to "window" on script load
   * // Note, there are no dash.
   * var md = window.markdownit();
   * var result = md.render('# markdown-it rulezz!');
   * ```
   *
   * Single line rendering, without paragraph wrap:
   *
   * ```javascript
   * var md = require('markdown-it')();
   * var result = md.renderInline('__markdown-it__ rulezz!');
   * ```
   **/

  /**
   * new MarkdownIt([presetName, options])
   * - presetName (String): optional, `commonmark` / `zero`
   * - options (Object)
   *
   * Creates parser instanse with given config. Can be called without `new`.
   *
   * ##### presetName
   *
   * MarkdownIt provides named presets as a convenience to quickly
   * enable/disable active syntax rules and options for common use cases.
   *
   * - ["commonmark"](https://github.com/markdown-it/markdown-it/blob/master/lib/presets/commonmark.js) -
   *   configures parser to strict [CommonMark](http://commonmark.org/) mode.
   * - [default](https://github.com/markdown-it/markdown-it/blob/master/lib/presets/default.js) -
   *   similar to GFM, used when no preset name given. Enables all available rules,
   *   but still without html, typographer & autolinker.
   * - ["zero"](https://github.com/markdown-it/markdown-it/blob/master/lib/presets/zero.js) -
   *   all rules disabled. Useful to quickly setup your config via `.enable()`.
   *   For example, when you need only `bold` and `italic` markup and nothing else.
   *
   * ##### options:
   *
   * - __html__ - `false`. Set `true` to enable HTML tags in source. Be careful!
   *   That's not safe! You may need external sanitizer to protect output from XSS.
   *   It's better to extend features via plugins, instead of enabling HTML.
   * - __xhtmlOut__ - `false`. Set `true` to add '/' when closing single tags
   *   (`<br />`). This is needed only for full CommonMark compatibility. In real
   *   world you will need HTML output.
   * - __breaks__ - `false`. Set `true` to convert `\n` in paragraphs into `<br>`.
   * - __langPrefix__ - `language-`. CSS language class prefix for fenced blocks.
   *   Can be useful for external highlighters.
   * - __linkify__ - `false`. Set `true` to autoconvert URL-like text to links.
   * - __typographer__  - `false`. Set `true` to enable [some language-neutral
   *   replacement](https://github.com/markdown-it/markdown-it/blob/master/lib/rules_core/replacements.js) +
   *   quotes beautification (smartquotes).
   * - __quotes__ - `“”‘’`, String or Array. Double + single quotes replacement
   *   pairs, when typographer enabled and smartquotes on. For example, you can
   *   use `'«»„“'` for Russian, `'„“‚‘'` for German, and
   *   `['«\xA0', '\xA0»', '‹\xA0', '\xA0›']` for French (including nbsp).
   * - __highlight__ - `null`. Highlighter function for fenced code blocks.
   *   Highlighter `function (str, lang)` should return escaped HTML. It can also
   *   return empty string if the source was not changed and should be escaped
   *   externaly. If result starts with <pre... internal wrapper is skipped.
   *
   * ##### Example
   *
   * ```javascript
   * // commonmark mode
   * var md = require('markdown-it')('commonmark');
   *
   * // default mode
   * var md = require('markdown-it')();
   *
   * // enable everything
   * var md = require('markdown-it')({
   *   html: true,
   *   linkify: true,
   *   typographer: true
   * });
   * ```
   *
   * ##### Syntax highlighting
   *
   * ```js
   * var hljs = require('highlight.js') // https://highlightjs.org/
   *
   * var md = require('markdown-it')({
   *   highlight: function (str, lang) {
   *     if (lang && hljs.getLanguage(lang)) {
   *       try {
   *         return hljs.highlight(lang, str, true).value;
   *       } catch (__) {}
   *     }
   *
   *     return ''; // use external default escaping
   *   }
   * });
   * ```
   *
   * Or with full wrapper override (if you need assign class to `<pre>`):
   *
   * ```javascript
   * var hljs = require('highlight.js') // https://highlightjs.org/
   *
   * // Actual default values
   * var md = require('markdown-it')({
   *   highlight: function (str, lang) {
   *     if (lang && hljs.getLanguage(lang)) {
   *       try {
   *         return '<pre class="hljs"><code>' +
   *                hljs.highlight(lang, str, true).value +
   *                '</code></pre>';
   *       } catch (__) {}
   *     }
   *
   *     return '<pre class="hljs"><code>' + md.utils.escapeHtml(str) + '</code></pre>';
   *   }
   * });
   * ```
   *
   **/
  function MarkdownIt$1(presetName, options) {
    if (!(this instanceof MarkdownIt$1)) {
      return new MarkdownIt$1(presetName, options);
    }

    if (!options) {
      if (!utils$1.isString(presetName)) {
        options = presetName || {};
        presetName = 'default';
      }
    }

    /**
     * MarkdownIt#inline -> ParserInline
     *
     * Instance of [[ParserInline]]. You may need it to add new rules when
     * writing plugins. For simple rules control use [[MarkdownIt.disable]] and
     * [[MarkdownIt.enable]].
     **/
    this.inline = new parser_inline$1();

    /**
     * MarkdownIt#block -> ParserBlock
     *
     * Instance of [[ParserBlock]]. You may need it to add new rules when
     * writing plugins. For simple rules control use [[MarkdownIt.disable]] and
     * [[MarkdownIt.enable]].
     **/
    this.block = new parser_block$1();

    /**
     * MarkdownIt#core -> Core
     *
     * Instance of [[Core]] chain executor. You may need it to add new rules when
     * writing plugins. For simple rules control use [[MarkdownIt.disable]] and
     * [[MarkdownIt.enable]].
     **/
    this.core = new parser_core$1();

    /**
     * MarkdownIt#renderer -> Renderer
     *
     * Instance of [[Renderer]]. Use it to modify output look. Or to add rendering
     * rules for new token types, generated by plugins.
     *
     * ##### Example
     *
     * ```javascript
     * var md = require('markdown-it')();
     *
     * function myToken(tokens, idx, options, env, self) {
     *   //...
     *   return result;
     * };
     *
     * md.renderer.rules['my_token'] = myToken
     * ```
     *
     * See [[Renderer]] docs and [source code](https://github.com/markdown-it/markdown-it/blob/master/lib/renderer.js).
     **/
    this.renderer = new renderer$1();

    /**
     * MarkdownIt#linkify -> LinkifyIt
     *
     * [linkify-it](https://github.com/markdown-it/linkify-it) instance.
     * Used by [linkify](https://github.com/markdown-it/markdown-it/blob/master/lib/rules_core/linkify.js)
     * rule.
     **/
    this.linkify = new linkifyIt$1();

    /**
     * MarkdownIt#validateLink(url) -> Boolean
     *
     * Link validation function. CommonMark allows too much in links. By default
     * we disable `javascript:`, `vbscript:`, `file:` schemas, and almost all `data:...` schemas
     * except some embedded image types.
     *
     * You can change this behaviour:
     *
     * ```javascript
     * var md = require('markdown-it')();
     * // enable everything
     * md.validateLink = function () { return true; }
     * ```
     **/
    this.validateLink = validateLink$1;

    /**
     * MarkdownIt#normalizeLink(url) -> String
     *
     * Function used to encode link url to a machine-readable format,
     * which includes url-encoding, punycode, etc.
     **/
    this.normalizeLink = normalizeLink$1;

    /**
     * MarkdownIt#normalizeLinkText(url) -> String
     *
     * Function used to decode link url to a human-readable format`
     **/
    this.normalizeLinkText = normalizeLinkText$1;


    // Expose utils & helpers for easy acces from plugins

    /**
     * MarkdownIt#utils -> utils
     *
     * Assorted utility functions, useful to write plugins. See details
     * [here](https://github.com/markdown-it/markdown-it/blob/master/lib/common/utils.js).
     **/
    this.utils = utils$1;

    /**
     * MarkdownIt#helpers -> helpers
     *
     * Link components parser functions, useful to write plugins. See details
     * [here](https://github.com/markdown-it/markdown-it/blob/master/lib/helpers).
     **/
    this.helpers = utils$1.assign({}, helpers$1);


    this.options = {};
    this.configure(presetName);

    if (options) { this.set(options); }
  }


  /** chainable
   * MarkdownIt.set(options)
   *
   * Set parser options (in the same format as in constructor). Probably, you
   * will never need it, but you can change options after constructor call.
   *
   * ##### Example
   *
   * ```javascript
   * var md = require('markdown-it')()
   *             .set({ html: true, breaks: true })
   *             .set({ typographer, true });
   * ```
   *
   * __Note:__ To achieve the best possible performance, don't modify a
   * `markdown-it` instance options on the fly. If you need multiple configurations
   * it's best to create multiple instances and initialize each with separate
   * config.
   **/
  MarkdownIt$1.prototype.set = function (options) {
    utils$1.assign(this.options, options);
    return this;
  };


  /** chainable, internal
   * MarkdownIt.configure(presets)
   *
   * Batch load of all options and compenent settings. This is internal method,
   * and you probably will not need it. But if you with - see available presets
   * and data structure [here](https://github.com/markdown-it/markdown-it/tree/master/lib/presets)
   *
   * We strongly recommend to use presets instead of direct config loads. That
   * will give better compatibility with next versions.
   **/
  MarkdownIt$1.prototype.configure = function (presets) {
    var self = this, presetName;

    if (utils$1.isString(presets)) {
      presetName = presets;
      presets = config$1[presetName];
      if (!presets) { throw new Error('Wrong `markdown-it` preset "' + presetName + '", check name'); }
    }

    if (!presets) { throw new Error('Wrong `markdown-it` preset, can\'t be empty'); }

    if (presets.options) { self.set(presets.options); }

    if (presets.components) {
      Object.keys(presets.components).forEach(function (name) {
        if (presets.components[name].rules) {
          self[name].ruler.enableOnly(presets.components[name].rules);
        }
        if (presets.components[name].rules2) {
          self[name].ruler2.enableOnly(presets.components[name].rules2);
        }
      });
    }
    return this;
  };


  /** chainable
   * MarkdownIt.enable(list, ignoreInvalid)
   * - list (String|Array): rule name or list of rule names to enable
   * - ignoreInvalid (Boolean): set `true` to ignore errors when rule not found.
   *
   * Enable list or rules. It will automatically find appropriate components,
   * containing rules with given names. If rule not found, and `ignoreInvalid`
   * not set - throws exception.
   *
   * ##### Example
   *
   * ```javascript
   * var md = require('markdown-it')()
   *             .enable(['sub', 'sup'])
   *             .disable('smartquotes');
   * ```
   **/
  MarkdownIt$1.prototype.enable = function (list, ignoreInvalid) {
    var result = [];

    if (!Array.isArray(list)) { list = [ list ]; }

    [ 'core', 'block', 'inline' ].forEach(function (chain) {
      result = result.concat(this[chain].ruler.enable(list, true));
    }, this);

    result = result.concat(this.inline.ruler2.enable(list, true));

    var missed = list.filter(function (name) { return result.indexOf(name) < 0; });

    if (missed.length && !ignoreInvalid) {
      throw new Error('MarkdownIt. Failed to enable unknown rule(s): ' + missed);
    }

    return this;
  };


  /** chainable
   * MarkdownIt.disable(list, ignoreInvalid)
   * - list (String|Array): rule name or list of rule names to disable.
   * - ignoreInvalid (Boolean): set `true` to ignore errors when rule not found.
   *
   * The same as [[MarkdownIt.enable]], but turn specified rules off.
   **/
  MarkdownIt$1.prototype.disable = function (list, ignoreInvalid) {
    var result = [];

    if (!Array.isArray(list)) { list = [ list ]; }

    [ 'core', 'block', 'inline' ].forEach(function (chain) {
      result = result.concat(this[chain].ruler.disable(list, true));
    }, this);

    result = result.concat(this.inline.ruler2.disable(list, true));

    var missed = list.filter(function (name) { return result.indexOf(name) < 0; });

    if (missed.length && !ignoreInvalid) {
      throw new Error('MarkdownIt. Failed to disable unknown rule(s): ' + missed);
    }
    return this;
  };


  /** chainable
   * MarkdownIt.use(plugin, params)
   *
   * Load specified plugin with given params into current parser instance.
   * It's just a sugar to call `plugin(md, params)` with curring.
   *
   * ##### Example
   *
   * ```javascript
   * var iterator = require('markdown-it-for-inline');
   * var md = require('markdown-it')()
   *             .use(iterator, 'foo_replace', 'text', function (tokens, idx) {
   *               tokens[idx].content = tokens[idx].content.replace(/foo/g, 'bar');
   *             });
   * ```
   **/
  MarkdownIt$1.prototype.use = function (plugin /*, params, ... */) {
    var args = [ this ].concat(Array.prototype.slice.call(arguments, 1));
    plugin.apply(plugin, args);
    return this;
  };


  /** internal
   * MarkdownIt.parse(src, env) -> Array
   * - src (String): source string
   * - env (Object): environment sandbox
   *
   * Parse input string and returns list of block tokens (special token type
   * "inline" will contain list of inline tokens). You should not call this
   * method directly, until you write custom renderer (for example, to produce
   * AST).
   *
   * `env` is used to pass data between "distributed" rules and return additional
   * metadata like reference info, needed for the renderer. It also can be used to
   * inject data in specific cases. Usually, you will be ok to pass `{}`,
   * and then pass updated object to renderer.
   **/
  MarkdownIt$1.prototype.parse = function (src, env) {
    if (typeof src !== 'string') {
      throw new Error('Input data should be a String');
    }

    var state = new this.core.State(src, this, env);

    this.core.process(state);

    return state.tokens;
  };


  /**
   * MarkdownIt.render(src [, env]) -> String
   * - src (String): source string
   * - env (Object): environment sandbox
   *
   * Render markdown string into html. It does all magic for you :).
   *
   * `env` can be used to inject additional metadata (`{}` by default).
   * But you will not need it with high probability. See also comment
   * in [[MarkdownIt.parse]].
   **/
  MarkdownIt$1.prototype.render = function (src, env) {
    env = env || {};

    return this.renderer.render(this.parse(src, env), this.options, env);
  };


  /** internal
   * MarkdownIt.parseInline(src, env) -> Array
   * - src (String): source string
   * - env (Object): environment sandbox
   *
   * The same as [[MarkdownIt.parse]] but skip all block rules. It returns the
   * block tokens list with the single `inline` element, containing parsed inline
   * tokens in `children` property. Also updates `env` object.
   **/
  MarkdownIt$1.prototype.parseInline = function (src, env) {
    var state = new this.core.State(src, this, env);

    state.inlineMode = true;
    this.core.process(state);

    return state.tokens;
  };


  /**
   * MarkdownIt.renderInline(src [, env]) -> String
   * - src (String): source string
   * - env (Object): environment sandbox
   *
   * Similar to [[MarkdownIt.render]] but for single paragraph content. Result
   * will NOT be wrapped into `<p>` tags.
   **/
  MarkdownIt$1.prototype.renderInline = function (src, env) {
    env = env || {};

    return this.renderer.render(this.parseInline(src, env), this.options, env);
  };


  var lib$4 = MarkdownIt$1;

  var markdownIt$1 = lib$4;

  // ::Schema Document schema for the data model used by CommonMark.
  var schema$l = new Schema({
    nodes: {
      doc: {
        content: "block+"
      },

      paragraph: {
        content: "inline*",
        group: "block",
        parseDOM: [{tag: "p"}],
        toDOM: function toDOM() { return ["p", 0] }
      },

      blockquote: {
        content: "block+",
        group: "block",
        parseDOM: [{tag: "blockquote"}],
        toDOM: function toDOM() { return ["blockquote", 0] }
      },

      horizontal_rule: {
        group: "block",
        parseDOM: [{tag: "hr"}],
        toDOM: function toDOM() { return ["div", ["hr"]] }
      },

      heading: {
        attrs: {level: {default: 1}},
        content: "(text | image)*",
        group: "block",
        defining: true,
        parseDOM: [{tag: "h1", attrs: {level: 1}},
                   {tag: "h2", attrs: {level: 2}},
                   {tag: "h3", attrs: {level: 3}},
                   {tag: "h4", attrs: {level: 4}},
                   {tag: "h5", attrs: {level: 5}},
                   {tag: "h6", attrs: {level: 6}}],
        toDOM: function toDOM(node) { return ["h" + node.attrs.level, 0] }
      },

      code_block: {
        content: "text*",
        group: "block",
        code: true,
        defining: true,
        marks: "",
        attrs: {params: {default: ""}},
        parseDOM: [{tag: "pre", preserveWhitespace: "full", getAttrs: function (node) { return (
          {params: node.getAttribute("data-params") || ""}
        ); }}],
        toDOM: function toDOM(node) { return ["pre", node.attrs.params ? {"data-params": node.attrs.params} : {}, ["code", 0]] }
      },

      ordered_list: {
        content: "list_item+",
        group: "block",
        attrs: {order: {default: 1}, tight: {default: false}},
        parseDOM: [{tag: "ol", getAttrs: function getAttrs(dom) {
          return {order: dom.hasAttribute("start") ? +dom.getAttribute("start") : 1,
                  tight: dom.hasAttribute("data-tight")}
        }}],
        toDOM: function toDOM(node) {
          return ["ol", {start: node.attrs.order == 1 ? null : node.attrs.order,
                         "data-tight": node.attrs.tight ? "true" : null}, 0]
        }
      },

      bullet_list: {
        content: "list_item+",
        group: "block",
        attrs: {tight: {default: false}},
        parseDOM: [{tag: "ul", getAttrs: function (dom) { return ({tight: dom.hasAttribute("data-tight")}); }}],
        toDOM: function toDOM(node) { return ["ul", {"data-tight": node.attrs.tight ? "true" : null}, 0] }
      },

      list_item: {
        content: "paragraph block*",
        defining: true,
        parseDOM: [{tag: "li"}],
        toDOM: function toDOM() { return ["li", 0] }
      },

      text: {
        group: "inline"
      },

      image: {
        inline: true,
        attrs: {
          src: {},
          alt: {default: null},
          title: {default: null}
        },
        group: "inline",
        draggable: true,
        parseDOM: [{tag: "img[src]", getAttrs: function getAttrs(dom) {
          return {
            src: dom.getAttribute("src"),
            title: dom.getAttribute("title"),
            alt: dom.getAttribute("alt")
          }
        }}],
        toDOM: function toDOM(node) { return ["img", node.attrs] }
      },

      hard_break: {
        inline: true,
        group: "inline",
        selectable: false,
        parseDOM: [{tag: "br"}],
        toDOM: function toDOM() { return ["br"] }
      }
    },

    marks: {
      em: {
        parseDOM: [{tag: "i"}, {tag: "em"},
                   {style: "font-style", getAttrs: function (value) { return value == "italic" && null; }}],
        toDOM: function toDOM() { return ["em"] }
      },

      strong: {
        parseDOM: [{tag: "b"}, {tag: "strong"},
                   {style: "font-weight", getAttrs: function (value) { return /^(bold(er)?|[5-9]\d{2,})$/.test(value) && null; }}],
        toDOM: function toDOM() { return ["strong"] }
      },

      link: {
        attrs: {
          href: {},
          title: {default: null}
        },
        inclusive: false,
        parseDOM: [{tag: "a[href]", getAttrs: function getAttrs(dom) {
          return {href: dom.getAttribute("href"), title: dom.getAttribute("title")}
        }}],
        toDOM: function toDOM(node) { return ["a", node.attrs] }
      },

      code: {
        parseDOM: [{tag: "code"}],
        toDOM: function toDOM() { return ["code"] }
      }
    }
  });

  function maybeMerge(a, b) {
    if (a.isText && b.isText && Mark$1.sameSet(a.marks, b.marks))
      { return a.withText(a.text + b.text) }
  }

  // Object used to track the context of a running parse.
  var MarkdownParseState = function MarkdownParseState(schema, tokenHandlers) {
    this.schema = schema;
    this.stack = [{type: schema.topNodeType, content: []}];
    this.marks = Mark$1.none;
    this.tokenHandlers = tokenHandlers;
  };

  MarkdownParseState.prototype.top = function top () {
    return this.stack[this.stack.length - 1]
  };

  MarkdownParseState.prototype.push = function push (elt) {
    if (this.stack.length) { this.top().content.push(elt); }
  };

  // : (string)
  // Adds the given text to the current position in the document,
  // using the current marks as styling.
  MarkdownParseState.prototype.addText = function addText (text) {
    if (!text) { return }
    var nodes = this.top().content, last = nodes[nodes.length - 1];
    var node = this.schema.text(text, this.marks), merged;
    if (last && (merged = maybeMerge(last, node))) { nodes[nodes.length - 1] = merged; }
    else { nodes.push(node); }
  };

  // : (Mark)
  // Adds the given mark to the set of active marks.
  MarkdownParseState.prototype.openMark = function openMark (mark) {
    this.marks = mark.addToSet(this.marks);
  };

  // : (Mark)
  // Removes the given mark from the set of active marks.
  MarkdownParseState.prototype.closeMark = function closeMark (mark) {
    this.marks = mark.removeFromSet(this.marks);
  };

  MarkdownParseState.prototype.parseTokens = function parseTokens (toks) {
    for (var i = 0; i < toks.length; i++) {
      var tok = toks[i];
      var handler = this.tokenHandlers[tok.type];
      if (!handler)
        { throw new Error("Token type `" + tok.type + "` not supported by Markdown parser") }
      handler(this, tok, toks, i);
    }
  };

  // : (NodeType, ?Object, ?[Node]) → ?Node
  // Add a node at the current position.
  MarkdownParseState.prototype.addNode = function addNode (type, attrs, content) {
    var node = type.createAndFill(attrs, content, this.marks);
    if (!node) { return null }
    this.push(node);
    return node
  };

  // : (NodeType, ?Object)
  // Wrap subsequent content in a node of the given type.
  MarkdownParseState.prototype.openNode = function openNode (type, attrs) {
    this.stack.push({type: type, attrs: attrs, content: []});
  };

  // : () → ?Node
  // Close and return the node that is currently on top of the stack.
  MarkdownParseState.prototype.closeNode = function closeNode () {
    if (this.marks.length) { this.marks = Mark$1.none; }
    var info = this.stack.pop();
    return this.addNode(info.type, info.attrs, info.content)
  };

  function attrs(spec, token, tokens, i) {
    if (spec.getAttrs) { return spec.getAttrs(token, tokens, i) }
    // For backwards compatibility when `attrs` is a Function
    else if (spec.attrs instanceof Function) { return spec.attrs(token) }
    else { return spec.attrs }
  }

  // Code content is represented as a single token with a `content`
  // property in Markdown-it.
  function noCloseToken(spec, type) {
    return spec.noCloseToken || type == "code_inline" || type == "code_block" || type == "fence"
  }

  function withoutTrailingNewline(str) {
    return str[str.length - 1] == "\n" ? str.slice(0, str.length - 1) : str
  }

  function noOp() {}

  function tokenHandlers(schema, tokens) {
    var handlers = Object.create(null);
    var loop = function ( type ) {
      var spec = tokens[type];
      if (spec.block) {
        var nodeType = schema.nodeType(spec.block);
        if (noCloseToken(spec, type)) {
          handlers[type] = function (state, tok, tokens, i) {
            state.openNode(nodeType, attrs(spec, tok, tokens, i));
            state.addText(withoutTrailingNewline(tok.content));
            state.closeNode();
          };
        } else {
          handlers[type + "_open"] = function (state, tok, tokens, i) { return state.openNode(nodeType, attrs(spec, tok, tokens, i)); };
          handlers[type + "_close"] = function (state) { return state.closeNode(); };
        }
      } else if (spec.node) {
        var nodeType$1 = schema.nodeType(spec.node);
        handlers[type] = function (state, tok, tokens, i) { return state.addNode(nodeType$1, attrs(spec, tok, tokens, i)); };
      } else if (spec.mark) {
        var markType = schema.marks[spec.mark];
        if (noCloseToken(spec, type)) {
          handlers[type] = function (state, tok, tokens, i) {
            state.openMark(markType.create(attrs(spec, tok, tokens, i)));
            state.addText(withoutTrailingNewline(tok.content));
            state.closeMark(markType);
          };
        } else {
          handlers[type + "_open"] = function (state, tok, tokens, i) { return state.openMark(markType.create(attrs(spec, tok, tokens, i))); };
          handlers[type + "_close"] = function (state) { return state.closeMark(markType); };
        }
      } else if (spec.ignore) {
        if (noCloseToken(spec, type)) {
          handlers[type] = noOp;
        } else {
          handlers[type + '_open'] = noOp;
          handlers[type + '_close'] = noOp;
        }
      } else {
        throw new RangeError("Unrecognized parsing spec " + JSON.stringify(spec))
      }
    };

    for (var type in tokens) { loop( type ); }

    handlers.text = function (state, tok) { return state.addText(tok.content); };
    handlers.inline = function (state, tok) { return state.parseTokens(tok.children); };
    handlers.softbreak = handlers.softbreak || (function (state) { return state.addText("\n"); });

    return handlers
  }

  // ::- A configuration of a Markdown parser. Such a parser uses
  // [markdown-it](https://github.com/markdown-it/markdown-it) to
  // tokenize a file, and then runs the custom rules it is given over
  // the tokens to create a ProseMirror document tree.
  var MarkdownParser = function MarkdownParser(schema, tokenizer, tokens) {
    // :: Object The value of the `tokens` object used to construct
    // this parser. Can be useful to copy and modify to base other
    // parsers on.
    this.tokens = tokens;
    this.schema = schema;
    this.tokenizer = tokenizer;
    this.tokenHandlers = tokenHandlers(schema, tokens);
  };

  // :: (string) → Node
  // Parse a string as [CommonMark](http://commonmark.org/) markup,
  // and create a ProseMirror document as prescribed by this parser's
  // rules.
  MarkdownParser.prototype.parse = function parse (text) {
    var state = new MarkdownParseState(this.schema, this.tokenHandlers), doc;
    state.parseTokens(this.tokenizer.parse(text, {}));
    do { doc = state.closeNode(); } while (state.stack.length)
    return doc
  };

  function listIsTight(tokens, i) {
    while (++i < tokens.length)
      { if (tokens[i].type != "list_item_open") { return tokens[i].hidden } }
    return false
  }

  // :: MarkdownParser
  // A parser parsing unextended [CommonMark](http://commonmark.org/),
  // without inline HTML, and producing a document in the basic schema.
  var defaultMarkdownParser = new MarkdownParser(schema$l, markdownIt$1("commonmark", {html: false}), {
    blockquote: {block: "blockquote"},
    paragraph: {block: "paragraph"},
    list_item: {block: "list_item"},
    bullet_list: {block: "bullet_list", getAttrs: function (_, tokens, i) { return ({tight: listIsTight(tokens, i)}); }},
    ordered_list: {block: "ordered_list", getAttrs: function (tok, tokens, i) { return ({
      order: +tok.attrGet("start") || 1,
      tight: listIsTight(tokens, i)
    }); }},
    heading: {block: "heading", getAttrs: function (tok) { return ({level: +tok.tag.slice(1)}); }},
    code_block: {block: "code_block", noCloseToken: true},
    fence: {block: "code_block", getAttrs: function (tok) { return ({params: tok.info || ""}); }, noCloseToken: true},
    hr: {node: "horizontal_rule"},
    image: {node: "image", getAttrs: function (tok) { return ({
      src: tok.attrGet("src"),
      title: tok.attrGet("title") || null,
      alt: tok.children[0] && tok.children[0].content || null
    }); }},
    hardbreak: {node: "hard_break"},

    em: {mark: "em"},
    strong: {mark: "strong"},
    link: {mark: "link", getAttrs: function (tok) { return ({
      href: tok.attrGet("href"),
      title: tok.attrGet("title") || null
    }); }},
    code_inline: {mark: "code", noCloseToken: true}
  });

  // ::- A specification for serializing a ProseMirror document as
  // Markdown/CommonMark text.
  var MarkdownSerializer = function MarkdownSerializer(nodes, marks) {
    // :: Object<(MarkdownSerializerState, Node)> The node serializer
    // functions for this serializer.
    this.nodes = nodes;
    // :: Object The mark serializer info.
    this.marks = marks;
  };

  // :: (Node, ?Object) → string
  // Serialize the content of the given node to
  // [CommonMark](http://commonmark.org/).
  MarkdownSerializer.prototype.serialize = function serialize (content, options) {
    var state = new MarkdownSerializerState(this.nodes, this.marks, options);
    state.renderContent(content);
    return state.out
  };

  // :: MarkdownSerializer
  // A serializer for the [basic schema](#schema).
  var defaultMarkdownSerializer = new MarkdownSerializer({
    blockquote: function blockquote(state, node) {
      state.wrapBlock("> ", null, node, function () { return state.renderContent(node); });
    },
    code_block: function code_block(state, node) {
      state.write("```" + (node.attrs.params || "") + "\n");
      state.text(node.textContent, false);
      state.ensureNewLine();
      state.write("```");
      state.closeBlock(node);
    },
    heading: function heading(state, node) {
      state.write(state.repeat("#", node.attrs.level) + " ");
      state.renderInline(node);
      state.closeBlock(node);
    },
    horizontal_rule: function horizontal_rule(state, node) {
      state.write(node.attrs.markup || "---");
      state.closeBlock(node);
    },
    bullet_list: function bullet_list(state, node) {
      state.renderList(node, "  ", function () { return (node.attrs.bullet || "*") + " "; });
    },
    ordered_list: function ordered_list(state, node) {
      var start = node.attrs.order || 1;
      var maxW = String(start + node.childCount - 1).length;
      var space = state.repeat(" ", maxW + 2);
      state.renderList(node, space, function (i) {
        var nStr = String(start + i);
        return state.repeat(" ", maxW - nStr.length) + nStr + ". "
      });
    },
    list_item: function list_item(state, node) {
      state.renderContent(node);
    },
    paragraph: function paragraph(state, node) {
      state.renderInline(node);
      state.closeBlock(node);
    },

    image: function image(state, node) {
      state.write("![" + state.esc(node.attrs.alt || "") + "](" + state.esc(node.attrs.src) +
                  (node.attrs.title ? " " + state.quote(node.attrs.title) : "") + ")");
    },
    hard_break: function hard_break(state, node, parent, index) {
      for (var i = index + 1; i < parent.childCount; i++)
        { if (parent.child(i).type != node.type) {
          state.write("\\\n");
          return
        } }
    },
    text: function text(state, node) {
      state.text(node.text);
    }
  }, {
    em: {open: "*", close: "*", mixable: true, expelEnclosingWhitespace: true},
    strong: {open: "**", close: "**", mixable: true, expelEnclosingWhitespace: true},
    link: {
      open: function open(_state, mark, parent, index) {
        return isPlainURL(mark, parent, index, 1) ? "<" : "["
      },
      close: function close(state, mark, parent, index) {
        return isPlainURL(mark, parent, index, -1) ? ">"
          : "](" + state.esc(mark.attrs.href) + (mark.attrs.title ? " " + state.quote(mark.attrs.title) : "") + ")"
      }
    },
    code: {open: function open(_state, _mark, parent, index) { return backticksFor(parent.child(index), -1) },
           close: function close(_state, _mark, parent, index) { return backticksFor(parent.child(index - 1), 1) },
           escape: false}
  });

  function backticksFor(node, side) {
    var ticks = /`+/g, m, len = 0;
    if (node.isText) { while (m = ticks.exec(node.text)) { len = Math.max(len, m[0].length); } }
    var result = len > 0 && side > 0 ? " `" : "`";
    for (var i = 0; i < len; i++) { result += "`"; }
    if (len > 0 && side < 0) { result += " "; }
    return result
  }

  function isPlainURL(link, parent, index, side) {
    if (link.attrs.title || !/^\w+:/.test(link.attrs.href)) { return false }
    var content = parent.child(index + (side < 0 ? -1 : 0));
    if (!content.isText || content.text != link.attrs.href || content.marks[content.marks.length - 1] != link) { return false }
    if (index == (side < 0 ? 1 : parent.childCount - 1)) { return true }
    var next = parent.child(index + (side < 0 ? -2 : 1));
    return !link.isInSet(next.marks)
  }

  // ::- This is an object used to track state and expose
  // methods related to markdown serialization. Instances are passed to
  // node and mark serialization methods (see `toMarkdown`).
  var MarkdownSerializerState = function MarkdownSerializerState(nodes, marks, options) {
    this.nodes = nodes;
    this.marks = marks;
    this.delim = this.out = "";
    this.closed = false;
    this.inTightList = false;
    // :: Object
    // The options passed to the serializer.
    // tightLists:: ?bool
    // Whether to render lists in a tight style. This can be overridden
    // on a node level by specifying a tight attribute on the node.
    // Defaults to false.
    this.options = options || {};
    if (typeof this.options.tightLists == "undefined")
      { this.options.tightLists = false; }
  };

  MarkdownSerializerState.prototype.flushClose = function flushClose (size) {
    if (this.closed) {
      if (!this.atBlank()) { this.out += "\n"; }
      if (size == null) { size = 2; }
      if (size > 1) {
        var delimMin = this.delim;
        var trim = /\s+$/.exec(delimMin);
        if (trim) { delimMin = delimMin.slice(0, delimMin.length - trim[0].length); }
        for (var i = 1; i < size; i++)
          { this.out += delimMin + "\n"; }
      }
      this.closed = false;
    }
  };

  // :: (string, ?string, Node, ())
  // Render a block, prefixing each line with `delim`, and the first
  // line in `firstDelim`. `node` should be the node that is closed at
  // the end of the block, and `f` is a function that renders the
  // content of the block.
  MarkdownSerializerState.prototype.wrapBlock = function wrapBlock (delim, firstDelim, node, f) {
    var old = this.delim;
    this.write(firstDelim || delim);
    this.delim += delim;
    f();
    this.delim = old;
    this.closeBlock(node);
  };

  MarkdownSerializerState.prototype.atBlank = function atBlank () {
    return /(^|\n)$/.test(this.out)
  };

  // :: ()
  // Ensure the current content ends with a newline.
  MarkdownSerializerState.prototype.ensureNewLine = function ensureNewLine () {
    if (!this.atBlank()) { this.out += "\n"; }
  };

  // :: (?string)
  // Prepare the state for writing output (closing closed paragraphs,
  // adding delimiters, and so on), and then optionally add content
  // (unescaped) to the output.
  MarkdownSerializerState.prototype.write = function write (content) {
    this.flushClose();
    if (this.delim && this.atBlank())
      { this.out += this.delim; }
    if (content) { this.out += content; }
  };

  // :: (Node)
  // Close the block for the given node.
  MarkdownSerializerState.prototype.closeBlock = function closeBlock (node) {
    this.closed = node;
  };

  // :: (string, ?bool)
  // Add the given text to the document. When escape is not `false`,
  // it will be escaped.
  MarkdownSerializerState.prototype.text = function text (text$1, escape) {
    var lines = text$1.split("\n");
    for (var i = 0; i < lines.length; i++) {
      var startOfLine = this.atBlank() || this.closed;
      this.write();
      this.out += escape !== false ? this.esc(lines[i], startOfLine) : lines[i];
      if (i != lines.length - 1) { this.out += "\n"; }
    }
  };

  // :: (Node)
  // Render the given node as a block.
  MarkdownSerializerState.prototype.render = function render (node, parent, index) {
    if (typeof parent == "number") { throw new Error("!") }
    if (!this.nodes[node.type.name]) { throw new Error("Token type `" + node.type.name + "` not supported by Markdown renderer") }
    this.nodes[node.type.name](this, node, parent, index);
  };

  // :: (Node)
  // Render the contents of `parent` as block nodes.
  MarkdownSerializerState.prototype.renderContent = function renderContent (parent) {
      var this$1$1 = this;

    parent.forEach(function (node, _, i) { return this$1$1.render(node, parent, i); });
  };

  // :: (Node)
  // Render the contents of `parent` as inline content.
  MarkdownSerializerState.prototype.renderInline = function renderInline (parent) {
      var this$1$1 = this;

    var active = [], trailing = "";
    var progress = function (node, _, index) {
      var marks = node ? node.marks : [];

      // Remove marks from `hard_break` that are the last node inside
      // that mark to prevent parser edge cases with new lines just
      // before closing marks.
      // (FIXME it'd be nice if we had a schema-agnostic way to
      // identify nodes that serialize as hard breaks)
      if (node && node.type.name === "hard_break")
        { marks = marks.filter(function (m) {
          if (index + 1 == parent.childCount) { return false }
          var next = parent.child(index + 1);
          return m.isInSet(next.marks) && (!next.isText || /\S/.test(next.text))
        }); }

      var leading = trailing;
      trailing = "";
      // If whitespace has to be expelled from the node, adjust
      // leading and trailing accordingly.
      if (node && node.isText && marks.some(function (mark) {
        var info = this$1$1.marks[mark.type.name];
        return info && info.expelEnclosingWhitespace
      })) {
        var ref = /^(\s*)(.*?)(\s*)$/m.exec(node.text);
          ref[0];
          var lead = ref[1];
          var inner$1 = ref[2];
          var trail = ref[3];
        leading += lead;
        trailing = trail;
        if (lead || trail) {
          node = inner$1 ? node.withText(inner$1) : null;
          if (!node) { marks = active; }
        }
      }

      var inner = marks.length && marks[marks.length - 1], noEsc = inner && this$1$1.marks[inner.type.name].escape === false;
      var len = marks.length - (noEsc ? 1 : 0);

      // Try to reorder 'mixable' marks, such as em and strong, which
      // in Markdown may be opened and closed in different order, so
      // that order of the marks for the token matches the order in
      // active.
      outer: for (var i = 0; i < len; i++) {
        var mark = marks[i];
        if (!this$1$1.marks[mark.type.name].mixable) { break }
        for (var j = 0; j < active.length; j++) {
          var other = active[j];
          if (!this$1$1.marks[other.type.name].mixable) { break }
          if (mark.eq(other)) {
            if (i > j)
              { marks = marks.slice(0, j).concat(mark).concat(marks.slice(j, i)).concat(marks.slice(i + 1, len)); }
            else if (j > i)
              { marks = marks.slice(0, i).concat(marks.slice(i + 1, j)).concat(mark).concat(marks.slice(j, len)); }
            continue outer
          }
        }
      }

      // Find the prefix of the mark set that didn't change
      var keep = 0;
      while (keep < Math.min(active.length, len) && marks[keep].eq(active[keep])) { ++keep; }

      // Close the marks that need to be closed
      while (keep < active.length)
        { this$1$1.text(this$1$1.markString(active.pop(), false, parent, index), false); }

      // Output any previously expelled trailing whitespace outside the marks
      if (leading) { this$1$1.text(leading); }

      // Open the marks that need to be opened
      if (node) {
        while (active.length < len) {
          var add = marks[active.length];
          active.push(add);
          this$1$1.text(this$1$1.markString(add, true, parent, index), false);
        }

        // Render the node. Special case code marks, since their content
        // may not be escaped.
        if (noEsc && node.isText)
          { this$1$1.text(this$1$1.markString(inner, true, parent, index) + node.text +
                    this$1$1.markString(inner, false, parent, index + 1), false); }
        else
          { this$1$1.render(node, parent, index); }
      }
    };
    parent.forEach(progress);
    progress(null, null, parent.childCount);
  };

  // :: (Node, string, (number) → string)
  // Render a node's content as a list. `delim` should be the extra
  // indentation added to all lines except the first in an item,
  // `firstDelim` is a function going from an item index to a
  // delimiter for the first line of the item.
  MarkdownSerializerState.prototype.renderList = function renderList (node, delim, firstDelim) {
      var this$1$1 = this;

    if (this.closed && this.closed.type == node.type)
      { this.flushClose(3); }
    else if (this.inTightList)
      { this.flushClose(1); }

    var isTight = typeof node.attrs.tight != "undefined" ? node.attrs.tight : this.options.tightLists;
    var prevTight = this.inTightList;
    this.inTightList = isTight;
    node.forEach(function (child, _, i) {
      if (i && isTight) { this$1$1.flushClose(1); }
      this$1$1.wrapBlock(delim, firstDelim(i), node, function () { return this$1$1.render(child, node, i); });
    });
    this.inTightList = prevTight;
  };

  // :: (string, ?bool) → string
  // Escape the given string so that it can safely appear in Markdown
  // content. If `startOfLine` is true, also escape characters that
  // have special meaning only at the start of the line.
  MarkdownSerializerState.prototype.esc = function esc (str, startOfLine) {
    str = str.replace(/[`*\\~\[\]]/g, "\\$&");
    if (startOfLine) { str = str.replace(/^[:#\-*+]/, "\\$&").replace(/^(\s*\d+)\./, "$1\\."); }
    return str
  };

  MarkdownSerializerState.prototype.quote = function quote (str) {
    var wrap = str.indexOf('"') == -1 ? '""' : str.indexOf("'") == -1 ? "''" : "()";
    return wrap[0] + str + wrap[1]
  };

  // :: (string, number) → string
  // Repeat the given string `n` times.
  MarkdownSerializerState.prototype.repeat = function repeat (str, n) {
    var out = "";
    for (var i = 0; i < n; i++) { out += str; }
    return out
  };

  // : (Mark, bool, string?) → string
  // Get the markdown string for a given opening or closing mark.
  MarkdownSerializerState.prototype.markString = function markString (mark, open, parent, index) {
    var info = this.marks[mark.type.name];
    var value = open ? info.open : info.close;
    return typeof value == "string" ? value : value(this, mark, parent, index)
  };

  // :: (string) → { leading: ?string, trailing: ?string }
  // Get leading and trailing whitespace from a string. Values of
  // leading or trailing property of the return object will be undefined
  // if there is no match.
  MarkdownSerializerState.prototype.getEnclosingWhitespace = function getEnclosingWhitespace (text) {
    return {
      leading: (text.match(/^(\s+)/) || [])[0],
      trailing: (text.match(/(\s+)$/) || [])[0]
    }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var schema$k = {
      nodes: {
          doc: {
              sortOrder: 0,
              content: "block+"
          }
      }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var doc = {
      id: 'doc',
      schema: schema$k
  };

  var schema$j = {
      nodes: {
          blockquote : {
              sortOrder: 200,
              content: "block+",
              group: "block",
              marks: "",
              parseDOM: [{tag: "blockquote"}],
              toDOM: function toDOM() {
                  return ["blockquote", 0]
              },
              parseMarkdown: {block: "blockquote"},
              toMarkdown: function (state, node) {
                  if(state.table) { return state.renderContent(node); }
                  state.wrapBlock("> ", null, node, function () { return state.renderContent(node); });
              }
          }
      }
  };

  // : (NodeType) → InputRule
  // Given a blockquote node type, returns an input rule that turns `"> "`
  // at the start of a textblock into a blockquote.
  var blockquoteRule = function (schema) {
      return wrappingInputRule(/^\s*>\s$/, schema.nodes.blockquote)
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  function wrapBlockQuote(context) {
      return wrapItem$1(context.schema.nodes.blockquote, {
          title: context.translate("Wrap in block quote"),
          icon: icons$1.blockquote,
          sortOrder: 300
      });
  }

  function menu$i(context) {
      return [
          {
              id: 'wrapBlockQuote',
              node: 'blockquote',
              group: 'format',
              item: wrapBlockQuote(context)
          }
      ]
  }

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var blockquote$1 = {
      id: 'blockquote',
      schema: schema$j,
      menu: function (context) { return menu$i(context); },
      inputRules: function (schema) {return [blockquoteRule(schema)]}
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */
  var schema$i = {
      nodes: {
          bullet_list: {
              sortOrder: 700,
              content: "list_item+",
              group: "block",
              attrs: {tight: {default: true}},
              parseDOM: [{
                  tag: "ul", getAttrs: function (dom) {
                      return ({tight: dom.hasAttribute("data-tight")});
                  }
              }],
              toDOM: function (node) {
                  return ["ul", {"data-tight": node.attrs.tight ? "true" : null}, 0]
              },
              parseMarkdown: {block: "bullet_list"},
              toMarkdown: function (state, node) {
                  state.renderList(node, "  ", function () { return (node.attrs.bullet || "*") + " "; });
              }
          }
      }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  // : (NodeType) → InputRule
  // Given a blockquote node type, returns an input rule that turns `"> "`
  // at the start of a textblock into a blockquote.
  var bulletListRule = function (schema) {
      return wrappingInputRule(/^\s*([-+*])\s$/, schema.nodes.bullet_list)
  };

  // :: (NodeType, ?Object) → (state: EditorState, dispatch: ?(tr: Transaction)) → bool
  // Returns a command function that wraps the selection in a list with
  // the given type an attributes. If `dispatch` is null, only return a
  // value to indicate whether this is possible, but don't actually
  // perform the change.
  function wrapInList$1(listType, attrs) {
    return function(state, dispatch) {
      var ref = state.selection;
      var $from = ref.$from;
      var $to = ref.$to;
      var range = $from.blockRange($to), doJoin = false, outerRange = range;
      if (!range) { return false }
      // This is at the top of an existing list item
      if (range.depth >= 2 && $from.node(range.depth - 1).type.compatibleContent(listType) && range.startIndex == 0) {
        // Don't do anything if this is the top of the list
        if ($from.index(range.depth - 1) == 0) { return false }
        var $insert = state.doc.resolve(range.start - 2);
        outerRange = new NodeRange$1($insert, $insert, range.depth);
        if (range.endIndex < range.parent.childCount)
          { range = new NodeRange$1($from, state.doc.resolve($to.end(range.depth)), range.depth); }
        doJoin = true;
      }
      var wrap = findWrapping(outerRange, listType, attrs, range);
      if (!wrap) { return false }
      if (dispatch) { dispatch(doWrapInList(state.tr, range, wrap, doJoin, listType).scrollIntoView()); }
      return true
    }
  }

  function doWrapInList(tr, range, wrappers, joinBefore, listType) {
    var content = Fragment$1.empty;
    for (var i = wrappers.length - 1; i >= 0; i--)
      { content = Fragment$1.from(wrappers[i].type.create(wrappers[i].attrs, content)); }

    tr.step(new ReplaceAroundStep(range.start - (joinBefore ? 2 : 0), range.end, range.start, range.end,
                                  new Slice$1(content, 0, 0), wrappers.length, true));

    var found = 0;
    for (var i$1 = 0; i$1 < wrappers.length; i$1++) { if (wrappers[i$1].type == listType) { found = i$1 + 1; } }
    var splitDepth = wrappers.length - found;

    var splitPos = range.start + wrappers.length - (joinBefore ? 2 : 0), parent = range.parent;
    for (var i$2 = range.startIndex, e = range.endIndex, first = true; i$2 < e; i$2++, first = false) {
      if (!first && canSplit(tr.doc, splitPos, splitDepth)) {
        tr.split(splitPos, splitDepth);
        splitPos += 2 * splitDepth;
      }
      splitPos += parent.child(i$2).nodeSize;
    }
    return tr
  }

  // :: (NodeType) → (state: EditorState, dispatch: ?(tr: Transaction)) → bool
  // Build a command that splits a non-empty textblock at the top level
  // of a list item by also splitting that list item.
  function splitListItem(itemType) {
    return function(state, dispatch) {
      var ref = state.selection;
      var $from = ref.$from;
      var $to = ref.$to;
      var node = ref.node;
      if ((node && node.isBlock) || $from.depth < 2 || !$from.sameParent($to)) { return false }
      var grandParent = $from.node(-1);
      if (grandParent.type != itemType) { return false }
      if ($from.parent.content.size == 0 && $from.node(-1).childCount == $from.indexAfter(-1)) {
        // In an empty block. If this is a nested list, the wrapping
        // list item should be split. Otherwise, bail out and let next
        // command handle lifting.
        if ($from.depth == 2 || $from.node(-3).type != itemType ||
            $from.index(-2) != $from.node(-2).childCount - 1) { return false }
        if (dispatch) {
          var wrap = Fragment$1.empty;
          var depthBefore = $from.index(-1) ? 1 : $from.index(-2) ? 2 : 3;
          // Build a fragment containing empty versions of the structure
          // from the outer list item to the parent node of the cursor
          for (var d = $from.depth - depthBefore; d >= $from.depth - 3; d--)
            { wrap = Fragment$1.from($from.node(d).copy(wrap)); }
          var depthAfter = $from.indexAfter(-1) < $from.node(-2).childCount ? 1
              : $from.indexAfter(-2) < $from.node(-3).childCount ? 2 : 3;
          // Add a second list item with an empty default start node
          wrap = wrap.append(Fragment$1.from(itemType.createAndFill()));
          var start = $from.before($from.depth - (depthBefore - 1));
          var tr$1 = state.tr.replace(start, $from.after(-depthAfter), new Slice$1(wrap, 4 - depthBefore, 0));
          var sel = -1;
          tr$1.doc.nodesBetween(start, tr$1.doc.content.size, function (node, pos) {
            if (sel > -1) { return false }
            if (node.isTextblock && node.content.size == 0) { sel = pos + 1; }
          });
          if (sel > -1) { tr$1.setSelection(state.selection.constructor.near(tr$1.doc.resolve(sel))); }
          dispatch(tr$1.scrollIntoView());
        }
        return true
      }
      var nextType = $to.pos == $from.end() ? grandParent.contentMatchAt(0).defaultType : null;
      var tr = state.tr.delete($from.pos, $to.pos);
      var types = nextType && [null, {type: nextType}];
      if (!canSplit(tr.doc, $from.pos, 2, types)) { return false }
      if (dispatch) { dispatch(tr.split($from.pos, 2, types).scrollIntoView()); }
      return true
    }
  }

  // :: (NodeType) → (state: EditorState, dispatch: ?(tr: Transaction)) → bool
  // Create a command to lift the list item around the selection up into
  // a wrapping list.
  function liftListItem(itemType) {
    return function(state, dispatch) {
      var ref = state.selection;
      var $from = ref.$from;
      var $to = ref.$to;
      var range = $from.blockRange($to, function (node) { return node.childCount && node.firstChild.type == itemType; });
      if (!range) { return false }
      if (!dispatch) { return true }
      if ($from.node(range.depth - 1).type == itemType) // Inside a parent list
        { return liftToOuterList(state, dispatch, itemType, range) }
      else // Outer list node
        { return liftOutOfList(state, dispatch, range) }
    }
  }

  function liftToOuterList(state, dispatch, itemType, range) {
    var tr = state.tr, end = range.end, endOfList = range.$to.end(range.depth);
    if (end < endOfList) {
      // There are siblings after the lifted items, which must become
      // children of the last item
      tr.step(new ReplaceAroundStep(end - 1, endOfList, end, endOfList,
                                    new Slice$1(Fragment$1.from(itemType.create(null, range.parent.copy())), 1, 0), 1, true));
      range = new NodeRange$1(tr.doc.resolve(range.$from.pos), tr.doc.resolve(endOfList), range.depth);
    }
    dispatch(tr.lift(range, liftTarget(range)).scrollIntoView());
    return true
  }

  function liftOutOfList(state, dispatch, range) {
    var tr = state.tr, list = range.parent;
    // Merge the list items into a single big item
    for (var pos = range.end, i = range.endIndex - 1, e = range.startIndex; i > e; i--) {
      pos -= list.child(i).nodeSize;
      tr.delete(pos - 1, pos + 1);
    }
    var $start = tr.doc.resolve(range.start), item = $start.nodeAfter;
    var atStart = range.startIndex == 0, atEnd = range.endIndex == list.childCount;
    var parent = $start.node(-1), indexBefore = $start.index(-1);
    if (!parent.canReplace(indexBefore + (atStart ? 0 : 1), indexBefore + 1,
                           item.content.append(atEnd ? Fragment$1.empty : Fragment$1.from(list))))
      { return false }
    var start = $start.pos, end = start + item.nodeSize;
    // Strip off the surrounding list. At the sides where we're not at
    // the end of the list, the existing list is closed. At sides where
    // this is the end, it is overwritten to its end.
    tr.step(new ReplaceAroundStep(start - (atStart ? 1 : 0), end + (atEnd ? 1 : 0), start + 1, end - 1,
                                  new Slice$1((atStart ? Fragment$1.empty : Fragment$1.from(list.copy(Fragment$1.empty)))
                                            .append(atEnd ? Fragment$1.empty : Fragment$1.from(list.copy(Fragment$1.empty))),
                                            atStart ? 0 : 1, atEnd ? 0 : 1), atStart ? 0 : 1));
    dispatch(tr.scrollIntoView());
    return true
  }

  // :: (NodeType) → (state: EditorState, dispatch: ?(tr: Transaction)) → bool
  // Create a command to sink the list item around the selection down
  // into an inner list.
  function sinkListItem(itemType) {
    return function(state, dispatch) {
      var ref = state.selection;
      var $from = ref.$from;
      var $to = ref.$to;
      var range = $from.blockRange($to, function (node) { return node.childCount && node.firstChild.type == itemType; });
      if (!range) { return false }
      var startIndex = range.startIndex;
      if (startIndex == 0) { return false }
      var parent = range.parent, nodeBefore = parent.child(startIndex - 1);
      if (nodeBefore.type != itemType) { return false }

      if (dispatch) {
        var nestedBefore = nodeBefore.lastChild && nodeBefore.lastChild.type == parent.type;
        var inner = Fragment$1.from(nestedBefore ? itemType.create() : null);
        var slice = new Slice$1(Fragment$1.from(itemType.create(null, Fragment$1.from(parent.type.create(null, inner)))),
                              nestedBefore ? 3 : 1, 0);
        var before = range.start, after = range.end;
        dispatch(state.tr.step(new ReplaceAroundStep(before - (nestedBefore ? 3 : 1), after,
                                                     before, after, slice, 1, true))
                 .scrollIntoView());
      }
      return true
    }
  }

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  function wrapBulletList(context) {
      return cmdItem(wrapInList$1(context.schema.nodes.bullet_list), {
          title: context.translate("Wrap in bullet list"),
          icon: icons$1.bulletList,
          sortOrder: 100
      });
  }

  function menu$h(context) {
      return [
          {
              id: 'wrapBulletList',
              node: 'bullet_list',
              group: 'format',
              item: wrapBulletList(context)
          }
      ]
  }

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var bullet_list = {
      id: 'bullet_list',
      schema: schema$i,
      menu: function (context) {
          return menu$h(context);
      },
      inputRules: function (schema) {return [bulletListRule(schema)]}
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var schema$h = {
      marks: {
          code:{
              isCode: true,
              sortOrder: 400,
              preventMarks: ['link'],
              parseDOM: [{tag: "code"}],
              toDOM: function () {
                  return ["code"]
              },
              parseMarkdown:  {code_inline: {mark: "code"}},
              toMarkdown: {open: "`", close: "`"}
          }
      }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */


  function markCode(context) {
      return markItem(context.schema.marks.code, {
          title: context.translate("Toggle code font"),
          icon: icons$1.code,
          sortOrder: 400
      });
  }

  function menu$g(context) {
      return [
          {
              id: 'markCode',
              mark: 'code',
              group: 'marks',
              item: markCode(context)
          }
      ]
  }

  var codeRules = function (schema) {
      return [
          markInputRuleOpen(/(?:`)([^`]+)$/, schema.marks.code),
          markInputRuleClosed(/(?:`)([^`]+)(?:`)$/, schema.marks.code)
      ]
  };

  function markInputRuleOpen(regexp, markType, getAttrs) {
      return new InputRule(regexp, function (state, match, start, end) {
          var attrs = getAttrs instanceof Function ? getAttrs(match) : getAttrs;
          var tr = state.tr;

          var nodeAfter = state.selection.$to.nodeAfter;
          if(nodeAfter && nodeAfter.isText && nodeAfter.text.indexOf('`') === 0) {
              tr.delete(start, end + 1);
              tr.addStoredMark(markType.create(attrs));
              tr.insertText(match[1], start);
              return tr;
          }

          return null;
      })
  }

  function markInputRuleClosed(regexp, markType, getAttrs) {
      return new InputRule(regexp, function (state, match, start, end) {
          var attrs = getAttrs instanceof Function ? getAttrs(match) : getAttrs;
          var tr = state.tr;

          if (match[1]) {
              var textStart = start + match[0].indexOf(match[1]);
              var textEnd = textStart + match[1].length;
              if (textEnd < end) { tr.delete(textEnd, end); }
              if (textStart > start) { tr.delete(start, textStart); }
              end = start + match[1].length;
              tr.addMark(start, end, markType.create(attrs));
              tr.removeStoredMark(markType); // Do not continue with mark.
              return tr;
          }

          return null;
      })
  }

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var code$1 = {
      id: 'code',
      schema: schema$h,
      menu: function (context) { return menu$g(context); },
      inputRules: function (schema) { return codeRules(schema); },
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */
  var code_block$1 =  {
      sortOrder: 500,
      content: "text*",
      group: "block",
      code: true,
      defining: true,
      marks: "",
      attrs: {params: {default: ""}},
      parseDOM: [{
          tag: "pre", preserveWhitespace: true, getAttrs: function (node) {
              return ({params: node.getAttribute("data-params") || ""});
          }
      }],
      toDOM: function toDOM(node) {
          return ["pre", node.attrs.params ? {"data-params": node.attrs.params} : {}, ["code", 0]]
      },
      parseMarkdown: {block: "code_block"},
      toMarkdown: function (state, node) {
          if(state.table) {
              state.wrapBlock("`", "`", node, function () { return state.text(node.textContent, false); });
          } else if (!node.attrs.params) {
              state.write("```\n");
              state.text(node.textContent, false);
              state.ensureNewLine();
              state.write("```");
              state.closeBlock(node);
          } else {
              state.write("```" + node.attrs.params + "\n");
              state.text(node.textContent, false);
              state.ensureNewLine();
              state.write("```");
              state.closeBlock(node);
          }
      }
  };

  var fence$1 = {
      parseMarkdown:  {block: "code_block", getAttrs: function (tok) { return ({params: tok.info || ""}); }}
  };

  var schema$g = {
      nodes: {
          code_block: code_block$1,
          fence: fence$1
      }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  // : (NodeType) → InputRule
  // Given a blockquote node type, returns an input rule that turns `"> "`
  // at the start of a textblock into a blockquote.
  var codeBlockRule = function (schema) {
      return textblockTypeInputRule(/^```$/, schema.nodes.code_block)
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  function makeCodeBlock(context) {
      return blockTypeItem$1(context.schema.nodes.code_block, {
          title: context.translate("Change to code block"),
          label: context.translate("Code")
      })
  }

  function menu$f(context) {
      return [
          {
              id: 'makeCodeBlock',
              node: 'code_block',
              group: 'types',
              item: makeCodeBlock(context)
          }
      ]
  }

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var code_block = {
      id: 'code_block',
      schema: schema$g,
      menu: function (context) { return menu$f(context); },
      inputRules: function (schema) {return [codeBlockRule(schema)]}
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var schema$f = {
      marks: {
          em: {
              sortOrder: 100,
              parseDOM: [{tag: "i"}, {tag: "em"},
                  {
                      style: "font-style", getAttrs: function (value) {
                      return value == "italic" && null;
                  }
                  }],
              toDOM: function () {
                  return ["em"]
              },
              parseMarkdown: {mark: "em"},
              toMarkdown: {open: "*", close: "*", mixable: true, expelEnclosingWhitespace: true}
          }
      }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */


  function markEm(context) {
      return markItem(context.schema.marks.em, {
          title: context.translate("Toggle emphasis"),
          icon: icons$1.em,
          sortOrder: 200});
  }

  function menu$e(context) {
      return [
          {
              id: 'markEm',
              mark: 'em',
              group: 'marks',
              item: markEm(context)
          }
      ]
  }

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var em = {
      id: 'em',
      schema: schema$f,
      menu: function (context) { return menu$e(context); }
  };

  /*! Copyright Twitter Inc. and other contributors. Licensed under MIT */
  var twemoji=function(){var twemoji={base:"https://twemoji.maxcdn.com/v/13.1.0/",ext:".png",size:"72x72",className:"emoji",convert:{fromCodePoint:fromCodePoint,toCodePoint:toCodePoint},onerror:function onerror(){if(this.parentNode){this.parentNode.replaceChild(createText(this.alt,false),this);}},parse:parse,replace:replace,test:test},escaper={"&":"&amp;","<":"&lt;",">":"&gt;","'":"&#39;",'"':"&quot;"},re=/(?:\ud83d\udc68\ud83c\udffb\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc68\ud83c[\udffb-\udfff]|\ud83d\udc68\ud83c\udffc\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc68\ud83c[\udffb-\udfff]|\ud83d\udc68\ud83c\udffd\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc68\ud83c[\udffb-\udfff]|\ud83d\udc68\ud83c\udffe\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc68\ud83c[\udffb-\udfff]|\ud83d\udc68\ud83c\udfff\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc68\ud83c[\udffb-\udfff]|\ud83d\udc69\ud83c\udffb\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc68\ud83c[\udffb-\udfff]|\ud83d\udc69\ud83c\udffb\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc69\ud83c[\udffb-\udfff]|\ud83d\udc69\ud83c\udffc\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc68\ud83c[\udffb-\udfff]|\ud83d\udc69\ud83c\udffc\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc69\ud83c[\udffb-\udfff]|\ud83d\udc69\ud83c\udffd\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc68\ud83c[\udffb-\udfff]|\ud83d\udc69\ud83c\udffd\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc69\ud83c[\udffb-\udfff]|\ud83d\udc69\ud83c\udffe\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc68\ud83c[\udffb-\udfff]|\ud83d\udc69\ud83c\udffe\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc69\ud83c[\udffb-\udfff]|\ud83d\udc69\ud83c\udfff\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc68\ud83c[\udffb-\udfff]|\ud83d\udc69\ud83c\udfff\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc69\ud83c[\udffb-\udfff]|\ud83e\uddd1\ud83c\udffb\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83e\uddd1\ud83c[\udffc-\udfff]|\ud83e\uddd1\ud83c\udffc\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83e\uddd1\ud83c[\udffb\udffd-\udfff]|\ud83e\uddd1\ud83c\udffd\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83e\uddd1\ud83c[\udffb\udffc\udffe\udfff]|\ud83e\uddd1\ud83c\udffe\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83e\uddd1\ud83c[\udffb-\udffd\udfff]|\ud83e\uddd1\ud83c\udfff\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83e\uddd1\ud83c[\udffb-\udffe]|\ud83d\udc68\ud83c\udffb\u200d\u2764\ufe0f\u200d\ud83d\udc68\ud83c[\udffb-\udfff]|\ud83d\udc68\ud83c\udffb\u200d\ud83e\udd1d\u200d\ud83d\udc68\ud83c[\udffc-\udfff]|\ud83d\udc68\ud83c\udffc\u200d\u2764\ufe0f\u200d\ud83d\udc68\ud83c[\udffb-\udfff]|\ud83d\udc68\ud83c\udffc\u200d\ud83e\udd1d\u200d\ud83d\udc68\ud83c[\udffb\udffd-\udfff]|\ud83d\udc68\ud83c\udffd\u200d\u2764\ufe0f\u200d\ud83d\udc68\ud83c[\udffb-\udfff]|\ud83d\udc68\ud83c\udffd\u200d\ud83e\udd1d\u200d\ud83d\udc68\ud83c[\udffb\udffc\udffe\udfff]|\ud83d\udc68\ud83c\udffe\u200d\u2764\ufe0f\u200d\ud83d\udc68\ud83c[\udffb-\udfff]|\ud83d\udc68\ud83c\udffe\u200d\ud83e\udd1d\u200d\ud83d\udc68\ud83c[\udffb-\udffd\udfff]|\ud83d\udc68\ud83c\udfff\u200d\u2764\ufe0f\u200d\ud83d\udc68\ud83c[\udffb-\udfff]|\ud83d\udc68\ud83c\udfff\u200d\ud83e\udd1d\u200d\ud83d\udc68\ud83c[\udffb-\udffe]|\ud83d\udc69\ud83c\udffb\u200d\u2764\ufe0f\u200d\ud83d\udc68\ud83c[\udffb-\udfff]|\ud83d\udc69\ud83c\udffb\u200d\u2764\ufe0f\u200d\ud83d\udc69\ud83c[\udffb-\udfff]|\ud83d\udc69\ud83c\udffb\u200d\ud83e\udd1d\u200d\ud83d\udc68\ud83c[\udffc-\udfff]|\ud83d\udc69\ud83c\udffb\u200d\ud83e\udd1d\u200d\ud83d\udc69\ud83c[\udffc-\udfff]|\ud83d\udc69\ud83c\udffc\u200d\u2764\ufe0f\u200d\ud83d\udc68\ud83c[\udffb-\udfff]|\ud83d\udc69\ud83c\udffc\u200d\u2764\ufe0f\u200d\ud83d\udc69\ud83c[\udffb-\udfff]|\ud83d\udc69\ud83c\udffc\u200d\ud83e\udd1d\u200d\ud83d\udc68\ud83c[\udffb\udffd-\udfff]|\ud83d\udc69\ud83c\udffc\u200d\ud83e\udd1d\u200d\ud83d\udc69\ud83c[\udffb\udffd-\udfff]|\ud83d\udc69\ud83c\udffd\u200d\u2764\ufe0f\u200d\ud83d\udc68\ud83c[\udffb-\udfff]|\ud83d\udc69\ud83c\udffd\u200d\u2764\ufe0f\u200d\ud83d\udc69\ud83c[\udffb-\udfff]|\ud83d\udc69\ud83c\udffd\u200d\ud83e\udd1d\u200d\ud83d\udc68\ud83c[\udffb\udffc\udffe\udfff]|\ud83d\udc69\ud83c\udffd\u200d\ud83e\udd1d\u200d\ud83d\udc69\ud83c[\udffb\udffc\udffe\udfff]|\ud83d\udc69\ud83c\udffe\u200d\u2764\ufe0f\u200d\ud83d\udc68\ud83c[\udffb-\udfff]|\ud83d\udc69\ud83c\udffe\u200d\u2764\ufe0f\u200d\ud83d\udc69\ud83c[\udffb-\udfff]|\ud83d\udc69\ud83c\udffe\u200d\ud83e\udd1d\u200d\ud83d\udc68\ud83c[\udffb-\udffd\udfff]|\ud83d\udc69\ud83c\udffe\u200d\ud83e\udd1d\u200d\ud83d\udc69\ud83c[\udffb-\udffd\udfff]|\ud83d\udc69\ud83c\udfff\u200d\u2764\ufe0f\u200d\ud83d\udc68\ud83c[\udffb-\udfff]|\ud83d\udc69\ud83c\udfff\u200d\u2764\ufe0f\u200d\ud83d\udc69\ud83c[\udffb-\udfff]|\ud83d\udc69\ud83c\udfff\u200d\ud83e\udd1d\u200d\ud83d\udc68\ud83c[\udffb-\udffe]|\ud83d\udc69\ud83c\udfff\u200d\ud83e\udd1d\u200d\ud83d\udc69\ud83c[\udffb-\udffe]|\ud83e\uddd1\ud83c\udffb\u200d\u2764\ufe0f\u200d\ud83e\uddd1\ud83c[\udffc-\udfff]|\ud83e\uddd1\ud83c\udffb\u200d\ud83e\udd1d\u200d\ud83e\uddd1\ud83c[\udffb-\udfff]|\ud83e\uddd1\ud83c\udffc\u200d\u2764\ufe0f\u200d\ud83e\uddd1\ud83c[\udffb\udffd-\udfff]|\ud83e\uddd1\ud83c\udffc\u200d\ud83e\udd1d\u200d\ud83e\uddd1\ud83c[\udffb-\udfff]|\ud83e\uddd1\ud83c\udffd\u200d\u2764\ufe0f\u200d\ud83e\uddd1\ud83c[\udffb\udffc\udffe\udfff]|\ud83e\uddd1\ud83c\udffd\u200d\ud83e\udd1d\u200d\ud83e\uddd1\ud83c[\udffb-\udfff]|\ud83e\uddd1\ud83c\udffe\u200d\u2764\ufe0f\u200d\ud83e\uddd1\ud83c[\udffb-\udffd\udfff]|\ud83e\uddd1\ud83c\udffe\u200d\ud83e\udd1d\u200d\ud83e\uddd1\ud83c[\udffb-\udfff]|\ud83e\uddd1\ud83c\udfff\u200d\u2764\ufe0f\u200d\ud83e\uddd1\ud83c[\udffb-\udffe]|\ud83e\uddd1\ud83c\udfff\u200d\ud83e\udd1d\u200d\ud83e\uddd1\ud83c[\udffb-\udfff]|\ud83d\udc68\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc68|\ud83d\udc69\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d[\udc68\udc69]|\ud83d\udc68\u200d\u2764\ufe0f\u200d\ud83d\udc68|\ud83d\udc69\u200d\u2764\ufe0f\u200d\ud83d[\udc68\udc69]|\ud83e\uddd1\u200d\ud83e\udd1d\u200d\ud83e\uddd1|\ud83d\udc6b\ud83c[\udffb-\udfff]|\ud83d\udc6c\ud83c[\udffb-\udfff]|\ud83d\udc6d\ud83c[\udffb-\udfff]|\ud83d\udc8f\ud83c[\udffb-\udfff]|\ud83d\udc91\ud83c[\udffb-\udfff]|\ud83d[\udc6b-\udc6d\udc8f\udc91])|(?:\ud83d[\udc68\udc69]|\ud83e\uddd1)(?:\ud83c[\udffb-\udfff])?\u200d(?:\u2695\ufe0f|\u2696\ufe0f|\u2708\ufe0f|\ud83c[\udf3e\udf73\udf7c\udf84\udf93\udfa4\udfa8\udfeb\udfed]|\ud83d[\udcbb\udcbc\udd27\udd2c\ude80\ude92]|\ud83e[\uddaf-\uddb3\uddbc\uddbd])|(?:\ud83c[\udfcb\udfcc]|\ud83d[\udd74\udd75]|\u26f9)((?:\ud83c[\udffb-\udfff]|\ufe0f)\u200d[\u2640\u2642]\ufe0f)|(?:\ud83c[\udfc3\udfc4\udfca]|\ud83d[\udc6e\udc70\udc71\udc73\udc77\udc81\udc82\udc86\udc87\ude45-\ude47\ude4b\ude4d\ude4e\udea3\udeb4-\udeb6]|\ud83e[\udd26\udd35\udd37-\udd39\udd3d\udd3e\uddb8\uddb9\uddcd-\uddcf\uddd4\uddd6-\udddd])(?:\ud83c[\udffb-\udfff])?\u200d[\u2640\u2642]\ufe0f|(?:\ud83d\udc68\u200d\ud83d\udc68\u200d\ud83d\udc66\u200d\ud83d\udc66|\ud83d\udc68\u200d\ud83d\udc68\u200d\ud83d\udc67\u200d\ud83d[\udc66\udc67]|\ud83d\udc68\u200d\ud83d\udc69\u200d\ud83d\udc66\u200d\ud83d\udc66|\ud83d\udc68\u200d\ud83d\udc69\u200d\ud83d\udc67\u200d\ud83d[\udc66\udc67]|\ud83d\udc69\u200d\ud83d\udc69\u200d\ud83d\udc66\u200d\ud83d\udc66|\ud83d\udc69\u200d\ud83d\udc69\u200d\ud83d\udc67\u200d\ud83d[\udc66\udc67]|\ud83d\udc68\u200d\ud83d\udc66\u200d\ud83d\udc66|\ud83d\udc68\u200d\ud83d\udc67\u200d\ud83d[\udc66\udc67]|\ud83d\udc68\u200d\ud83d\udc68\u200d\ud83d[\udc66\udc67]|\ud83d\udc68\u200d\ud83d\udc69\u200d\ud83d[\udc66\udc67]|\ud83d\udc69\u200d\ud83d\udc66\u200d\ud83d\udc66|\ud83d\udc69\u200d\ud83d\udc67\u200d\ud83d[\udc66\udc67]|\ud83d\udc69\u200d\ud83d\udc69\u200d\ud83d[\udc66\udc67]|\ud83c\udff3\ufe0f\u200d\u26a7\ufe0f|\ud83c\udff3\ufe0f\u200d\ud83c\udf08|\ud83d\ude36\u200d\ud83c\udf2b\ufe0f|\u2764\ufe0f\u200d\ud83d\udd25|\u2764\ufe0f\u200d\ud83e\ude79|\ud83c\udff4\u200d\u2620\ufe0f|\ud83d\udc15\u200d\ud83e\uddba|\ud83d\udc3b\u200d\u2744\ufe0f|\ud83d\udc41\u200d\ud83d\udde8|\ud83d\udc68\u200d\ud83d[\udc66\udc67]|\ud83d\udc69\u200d\ud83d[\udc66\udc67]|\ud83d\udc6f\u200d\u2640\ufe0f|\ud83d\udc6f\u200d\u2642\ufe0f|\ud83d\ude2e\u200d\ud83d\udca8|\ud83d\ude35\u200d\ud83d\udcab|\ud83e\udd3c\u200d\u2640\ufe0f|\ud83e\udd3c\u200d\u2642\ufe0f|\ud83e\uddde\u200d\u2640\ufe0f|\ud83e\uddde\u200d\u2642\ufe0f|\ud83e\udddf\u200d\u2640\ufe0f|\ud83e\udddf\u200d\u2642\ufe0f|\ud83d\udc08\u200d\u2b1b)|[#*0-9]\ufe0f?\u20e3|(?:[©®\u2122\u265f]\ufe0f)|(?:\ud83c[\udc04\udd70\udd71\udd7e\udd7f\ude02\ude1a\ude2f\ude37\udf21\udf24-\udf2c\udf36\udf7d\udf96\udf97\udf99-\udf9b\udf9e\udf9f\udfcd\udfce\udfd4-\udfdf\udff3\udff5\udff7]|\ud83d[\udc3f\udc41\udcfd\udd49\udd4a\udd6f\udd70\udd73\udd76-\udd79\udd87\udd8a-\udd8d\udda5\udda8\uddb1\uddb2\uddbc\uddc2-\uddc4\uddd1-\uddd3\udddc-\uddde\udde1\udde3\udde8\uddef\uddf3\uddfa\udecb\udecd-\udecf\udee0-\udee5\udee9\udef0\udef3]|[\u203c\u2049\u2139\u2194-\u2199\u21a9\u21aa\u231a\u231b\u2328\u23cf\u23ed-\u23ef\u23f1\u23f2\u23f8-\u23fa\u24c2\u25aa\u25ab\u25b6\u25c0\u25fb-\u25fe\u2600-\u2604\u260e\u2611\u2614\u2615\u2618\u2620\u2622\u2623\u2626\u262a\u262e\u262f\u2638-\u263a\u2640\u2642\u2648-\u2653\u2660\u2663\u2665\u2666\u2668\u267b\u267f\u2692-\u2697\u2699\u269b\u269c\u26a0\u26a1\u26a7\u26aa\u26ab\u26b0\u26b1\u26bd\u26be\u26c4\u26c5\u26c8\u26cf\u26d1\u26d3\u26d4\u26e9\u26ea\u26f0-\u26f5\u26f8\u26fa\u26fd\u2702\u2708\u2709\u270f\u2712\u2714\u2716\u271d\u2721\u2733\u2734\u2744\u2747\u2757\u2763\u2764\u27a1\u2934\u2935\u2b05-\u2b07\u2b1b\u2b1c\u2b50\u2b55\u3030\u303d\u3297\u3299])(?:\ufe0f|(?!\ufe0e))|(?:(?:\ud83c[\udfcb\udfcc]|\ud83d[\udd74\udd75\udd90]|[\u261d\u26f7\u26f9\u270c\u270d])(?:\ufe0f|(?!\ufe0e))|(?:\ud83c[\udf85\udfc2-\udfc4\udfc7\udfca]|\ud83d[\udc42\udc43\udc46-\udc50\udc66-\udc69\udc6e\udc70-\udc78\udc7c\udc81-\udc83\udc85-\udc87\udcaa\udd7a\udd95\udd96\ude45-\ude47\ude4b-\ude4f\udea3\udeb4-\udeb6\udec0\udecc]|\ud83e[\udd0c\udd0f\udd18-\udd1c\udd1e\udd1f\udd26\udd30-\udd39\udd3d\udd3e\udd77\uddb5\uddb6\uddb8\uddb9\uddbb\uddcd-\uddcf\uddd1-\udddd]|[\u270a\u270b]))(?:\ud83c[\udffb-\udfff])?|(?:\ud83c\udff4\udb40\udc67\udb40\udc62\udb40\udc65\udb40\udc6e\udb40\udc67\udb40\udc7f|\ud83c\udff4\udb40\udc67\udb40\udc62\udb40\udc73\udb40\udc63\udb40\udc74\udb40\udc7f|\ud83c\udff4\udb40\udc67\udb40\udc62\udb40\udc77\udb40\udc6c\udb40\udc73\udb40\udc7f|\ud83c\udde6\ud83c[\udde8-\uddec\uddee\uddf1\uddf2\uddf4\uddf6-\uddfa\uddfc\uddfd\uddff]|\ud83c\udde7\ud83c[\udde6\udde7\udde9-\uddef\uddf1-\uddf4\uddf6-\uddf9\uddfb\uddfc\uddfe\uddff]|\ud83c\udde8\ud83c[\udde6\udde8\udde9\uddeb-\uddee\uddf0-\uddf5\uddf7\uddfa-\uddff]|\ud83c\udde9\ud83c[\uddea\uddec\uddef\uddf0\uddf2\uddf4\uddff]|\ud83c\uddea\ud83c[\udde6\udde8\uddea\uddec\udded\uddf7-\uddfa]|\ud83c\uddeb\ud83c[\uddee-\uddf0\uddf2\uddf4\uddf7]|\ud83c\uddec\ud83c[\udde6\udde7\udde9-\uddee\uddf1-\uddf3\uddf5-\uddfa\uddfc\uddfe]|\ud83c\udded\ud83c[\uddf0\uddf2\uddf3\uddf7\uddf9\uddfa]|\ud83c\uddee\ud83c[\udde8-\uddea\uddf1-\uddf4\uddf6-\uddf9]|\ud83c\uddef\ud83c[\uddea\uddf2\uddf4\uddf5]|\ud83c\uddf0\ud83c[\uddea\uddec-\uddee\uddf2\uddf3\uddf5\uddf7\uddfc\uddfe\uddff]|\ud83c\uddf1\ud83c[\udde6-\udde8\uddee\uddf0\uddf7-\uddfb\uddfe]|\ud83c\uddf2\ud83c[\udde6\udde8-\udded\uddf0-\uddff]|\ud83c\uddf3\ud83c[\udde6\udde8\uddea-\uddec\uddee\uddf1\uddf4\uddf5\uddf7\uddfa\uddff]|\ud83c\uddf4\ud83c\uddf2|\ud83c\uddf5\ud83c[\udde6\uddea-\udded\uddf0-\uddf3\uddf7-\uddf9\uddfc\uddfe]|\ud83c\uddf6\ud83c\udde6|\ud83c\uddf7\ud83c[\uddea\uddf4\uddf8\uddfa\uddfc]|\ud83c\uddf8\ud83c[\udde6-\uddea\uddec-\uddf4\uddf7-\uddf9\uddfb\uddfd-\uddff]|\ud83c\uddf9\ud83c[\udde6\udde8\udde9\uddeb-\udded\uddef-\uddf4\uddf7\uddf9\uddfb\uddfc\uddff]|\ud83c\uddfa\ud83c[\udde6\uddec\uddf2\uddf3\uddf8\uddfe\uddff]|\ud83c\uddfb\ud83c[\udde6\udde8\uddea\uddec\uddee\uddf3\uddfa]|\ud83c\uddfc\ud83c[\uddeb\uddf8]|\ud83c\uddfd\ud83c\uddf0|\ud83c\uddfe\ud83c[\uddea\uddf9]|\ud83c\uddff\ud83c[\udde6\uddf2\uddfc]|\ud83c[\udccf\udd8e\udd91-\udd9a\udde6-\uddff\ude01\ude32-\ude36\ude38-\ude3a\ude50\ude51\udf00-\udf20\udf2d-\udf35\udf37-\udf7c\udf7e-\udf84\udf86-\udf93\udfa0-\udfc1\udfc5\udfc6\udfc8\udfc9\udfcf-\udfd3\udfe0-\udff0\udff4\udff8-\udfff]|\ud83d[\udc00-\udc3e\udc40\udc44\udc45\udc51-\udc65\udc6a\udc6f\udc79-\udc7b\udc7d-\udc80\udc84\udc88-\udc8e\udc90\udc92-\udca9\udcab-\udcfc\udcff-\udd3d\udd4b-\udd4e\udd50-\udd67\udda4\uddfb-\ude44\ude48-\ude4a\ude80-\udea2\udea4-\udeb3\udeb7-\udebf\udec1-\udec5\uded0-\uded2\uded5-\uded7\udeeb\udeec\udef4-\udefc\udfe0-\udfeb]|\ud83e[\udd0d\udd0e\udd10-\udd17\udd1d\udd20-\udd25\udd27-\udd2f\udd3a\udd3c\udd3f-\udd45\udd47-\udd76\udd78\udd7a-\uddb4\uddb7\uddba\uddbc-\uddcb\uddd0\uddde-\uddff\ude70-\ude74\ude78-\ude7a\ude80-\ude86\ude90-\udea8\udeb0-\udeb6\udec0-\udec2\uded0-\uded6]|[\u23e9-\u23ec\u23f0\u23f3\u267e\u26ce\u2705\u2728\u274c\u274e\u2753-\u2755\u2795-\u2797\u27b0\u27bf\ue50a])|\ufe0f/g,UFE0Fg=/\uFE0F/g,U200D=String.fromCharCode(8205),rescaper=/[&<>'"]/g,shouldntBeParsed=/^(?:iframe|noframes|noscript|script|select|style|textarea)$/,fromCharCode=String.fromCharCode;return twemoji;function createText(text,clean){return document.createTextNode(clean?text.replace(UFE0Fg,""):text)}function escapeHTML(s){return s.replace(rescaper,replacer)}function defaultImageSrcGenerator(icon,options){return "".concat(options.base,options.size,"/",icon,options.ext)}function grabAllTextNodes(node,allText){var childNodes=node.childNodes,length=childNodes.length,subnode,nodeType;while(length--){subnode=childNodes[length];nodeType=subnode.nodeType;if(nodeType===3){allText.push(subnode);}else if(nodeType===1&&!("ownerSVGElement"in subnode)&&!shouldntBeParsed.test(subnode.nodeName.toLowerCase())){grabAllTextNodes(subnode,allText);}}return allText}function grabTheRightIcon(rawText){return toCodePoint(rawText.indexOf(U200D)<0?rawText.replace(UFE0Fg,""):rawText)}function parseNode(node,options){var allText=grabAllTextNodes(node,[]),length=allText.length,attrib,attrname,modified,fragment,subnode,text,match,i,index,img,rawText,iconId,src;while(length--){modified=false;fragment=document.createDocumentFragment();subnode=allText[length];text=subnode.nodeValue;i=0;while(match=re.exec(text)){index=match.index;if(index!==i){fragment.appendChild(createText(text.slice(i,index),true));}rawText=match[0];iconId=grabTheRightIcon(rawText);i=index+rawText.length;src=options.callback(iconId,options);if(iconId&&src){img=new Image;img.onerror=options.onerror;img.setAttribute("draggable","false");attrib=options.attributes(rawText,iconId);for(attrname in attrib){if(attrib.hasOwnProperty(attrname)&&attrname.indexOf("on")!==0&&!img.hasAttribute(attrname)){img.setAttribute(attrname,attrib[attrname]);}}img.className=options.className;img.alt=rawText;img.src=src;modified=true;fragment.appendChild(img);}if(!img){ fragment.appendChild(createText(rawText,false)); }img=null;}if(modified){if(i<text.length){fragment.appendChild(createText(text.slice(i),true));}subnode.parentNode.replaceChild(fragment,subnode);}}return node}function parseString(str,options){return replace(str,function(rawText){var ret=rawText,iconId=grabTheRightIcon(rawText),src=options.callback(iconId,options),attrib,attrname;if(iconId&&src){ret="<img ".concat('class="',options.className,'" ','draggable="false" ','alt="',rawText,'"',' src="',src,'"');attrib=options.attributes(rawText,iconId);for(attrname in attrib){if(attrib.hasOwnProperty(attrname)&&attrname.indexOf("on")!==0&&ret.indexOf(" "+attrname+"=")===-1){ret=ret.concat(" ",attrname,'="',escapeHTML(attrib[attrname]),'"');}}ret=ret.concat("/>");}return ret})}function replacer(m){return escaper[m]}function returnNull(){return null}function toSizeSquaredAsset(value){return typeof value==="number"?value+"x"+value:value}function fromCodePoint(codepoint){var code=typeof codepoint==="string"?parseInt(codepoint,16):codepoint;if(code<65536){return fromCharCode(code)}code-=65536;return fromCharCode(55296+(code>>10),56320+(code&1023))}function parse(what,how){if(!how||typeof how==="function"){how={callback:how};}return (typeof what==="string"?parseString:parseNode)(what,{callback:how.callback||defaultImageSrcGenerator,attributes:typeof how.attributes==="function"?how.attributes:returnNull,base:typeof how.base==="string"?how.base:twemoji.base,ext:how.ext||twemoji.ext,size:how.folder||toSizeSquaredAsset(how.size||twemoji.size),className:how.className||twemoji.className,onerror:how.onerror||twemoji.onerror})}function replace(text,callback){return String(text).replace(re,callback)}function test(text){re.lastIndex=0;var result=re.test(text);re.lastIndex=0;return result}function toCodePoint(unicodeSurrogates,sep){var r=[],c=0,p=0,i=0;while(i<unicodeSurrogates.length){c=unicodeSurrogates.charCodeAt(i++);if(p){r.push((65536+(p-55296<<10)+(c-56320)).toString(16));p=0;}else if(55296<=c&&c<=56319){p=c;}else {r.push(c.toString(16));}}return r.join(sep||"-")}}();

  var schema$e = {
      nodes: {
          emoji:  {
              attrs: {
                  class: {default: 'emoji'},
                  draggable: {default: 'false'},
                  width: {default: '16'},
                  height: {default: '16'},
                  'data-name': {default: null},
                  alt: {default: null},
                  src: {default: null},
              },
              inline: true,
              group: "inline",
              parseDOM: [{
                  tag: "img.emoji", getAttrs: function getAttrs(dom) {
                      return {
                          src: dom.getAttribute("src"),
                          alt: dom.getAttribute("alt"),
                          'data-name': String(dom.getAttribute('data-name'))
                      }
                  }
              }],
              toDOM: function toDOM(node) {
                  return ['img', node.attrs]
              },
              parseMarkdown:  {
                  node: "emoji", getAttrs: function (tok) {

                      // Workaround, since the context is not available here, so we can't use context.getPluginOption('emoji', 'twemoji');
                      var options = (humhub && humhub.config) ? humhub.config.get('ui.richtext.prosemirror', 'emoji')['twemoji'] : null;

                      var $dom = $(twemoji.parse(tok.content, options));
                      return ({
                          'data-name': String(tok.markup),
                          alt: $dom.attr('alt'),
                          src: $dom.attr('src')
                      })
                  }
              },
              toMarkdown: function (state, node) {
                  var result;

                  if(!node.attrs['data-name']) {
                      result = (state.alt) ? state.esc(state.alt) : '';
                  } else {
                      result = ':'+state.esc(node.attrs['data-name'])+':';
                  }

                  state.write(result);
              }
          }
      },
      marks: {
          emojiQuery: {
              excludes: "_",
              inclusive: true,
              parseDOM: [
                  { tag: 'span[data-emoji-query]' }
              ],
              toDOM: function toDOM(node) {
                  return ['span', {
                      'data-emoji-query': true,
                  }];
              }
          }
      }
  };

  // Emoticons -> Emoji mapping.

  var shortcuts$1 = {
    angry:            [ '>:(', '>:-(' ],
    blush:            [ ':")', ':-")' ],
    broken_heart:     [ '</3', '<\\3' ],
    // :\ and :-\ not used because of conflict with markdown escaping
    confused:         [ ':/', ':-/' ], // twemoji shows question
    cry:              [ ":'(", ":'-(", ':,(', ':,-(' ],
    frowning:         [ ':(', ':-(' ],
    heart:            [ '<3' ],
    imp:              [ ']:(', ']:-(' ],
    innocent:         [ 'o:)', 'O:)', 'o:-)', 'O:-)', '0:)', '0:-)' ],
    joy:              [ ":')", ":'-)", ':,)', ':,-)', ":'D", ":'-D", ':,D', ':,-D' ],
    kissing:          [ ':*', ':-*' ],
    laughing:         [ 'x-)', 'X-)' ],
    neutral_face:     [ ':|', ':-|' ],
    open_mouth:       [ ':o', ':-o', ':O', ':-O' ],
    rage:             [ ':@', ':-@' ],
    smile:            [ ':D', ':-D' ],
    smiley:           [ ':)', ':-)' ],
    smiling_imp:      [ ']:)', ']:-)' ],
    sob:              [ ":,'(", ":,'-(", ';(', ';-(' ],
    stuck_out_tongue: [ ':P', ':-P' ],
    sunglasses:       [ '8-)', 'B-)' ],
    sweat:            [ ',:(', ',:-(' ],
    sweat_smile:      [ ',:)', ',:-)' ],
    unamused:         [ ':s', ':-S', ':z', ':-Z', ':$', ':-$' ],
    wink:             [ ';)', ';-)' ]
  };
  shortcuts$1.angry;
  shortcuts$1.blush;
  shortcuts$1.broken_heart;
  shortcuts$1.confused;
  shortcuts$1.cry;
  shortcuts$1.frowning;
  shortcuts$1.heart;
  shortcuts$1.imp;
  shortcuts$1.innocent;
  shortcuts$1.joy;
  shortcuts$1.kissing;
  shortcuts$1.laughing;
  shortcuts$1.neutral_face;
  shortcuts$1.open_mouth;
  shortcuts$1.rage;
  shortcuts$1.smile;
  shortcuts$1.smiley;
  shortcuts$1.smiling_imp;
  shortcuts$1.sob;
  shortcuts$1.stuck_out_tongue;
  shortcuts$1.sunglasses;
  shortcuts$1.sweat;
  shortcuts$1.sweat_smile;
  shortcuts$1.unamused;
  shortcuts$1.wink;

  var grinning$1 = {
  	keywords: [
  		"face",
  		"smile",
  		"happy",
  		"joy",
  		":D",
  		"grin"
  	],
  	char: "😀",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var grimacing$1 = {
  	keywords: [
  		"face",
  		"grimace",
  		"teeth"
  	],
  	char: "😬",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var grin$1 = {
  	keywords: [
  		"face",
  		"happy",
  		"smile",
  		"joy",
  		"kawaii"
  	],
  	char: "😁",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var joy$1 = {
  	keywords: [
  		"face",
  		"cry",
  		"tears",
  		"weep",
  		"happy",
  		"happytears",
  		"haha"
  	],
  	char: "😂",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var rofl$1 = {
  	keywords: [
  		"face",
  		"rolling",
  		"floor",
  		"laughing",
  		"lol",
  		"haha"
  	],
  	char: "🤣",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var partying = {
  	keywords: [
  		"face",
  		"celebration",
  		"woohoo"
  	],
  	char: "🥳",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var smiley$1 = {
  	keywords: [
  		"face",
  		"happy",
  		"joy",
  		"haha",
  		":D",
  		":)",
  		"smile",
  		"funny"
  	],
  	char: "😃",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var smile$2 = {
  	keywords: [
  		"face",
  		"happy",
  		"joy",
  		"funny",
  		"haha",
  		"laugh",
  		"like",
  		":D",
  		":)"
  	],
  	char: "😄",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var sweat_smile$1 = {
  	keywords: [
  		"face",
  		"hot",
  		"happy",
  		"laugh",
  		"sweat",
  		"smile",
  		"relief"
  	],
  	char: "😅",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var laughing$1 = {
  	keywords: [
  		"happy",
  		"joy",
  		"lol",
  		"satisfied",
  		"haha",
  		"face",
  		"glad",
  		"XD",
  		"laugh"
  	],
  	char: "😆",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var innocent$1 = {
  	keywords: [
  		"face",
  		"angel",
  		"heaven",
  		"halo"
  	],
  	char: "😇",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var wink$1 = {
  	keywords: [
  		"face",
  		"happy",
  		"mischievous",
  		"secret",
  		";)",
  		"smile",
  		"eye"
  	],
  	char: "😉",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var blush$1 = {
  	keywords: [
  		"face",
  		"smile",
  		"happy",
  		"flushed",
  		"crush",
  		"embarrassed",
  		"shy",
  		"joy"
  	],
  	char: "😊",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var slightly_smiling_face$1 = {
  	keywords: [
  		"face",
  		"smile"
  	],
  	char: "🙂",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var upside_down_face$1 = {
  	keywords: [
  		"face",
  		"flipped",
  		"silly",
  		"smile"
  	],
  	char: "🙃",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var relaxed$1 = {
  	keywords: [
  		"face",
  		"blush",
  		"massage",
  		"happiness"
  	],
  	char: "☺️",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var yum$1 = {
  	keywords: [
  		"happy",
  		"joy",
  		"tongue",
  		"smile",
  		"face",
  		"silly",
  		"yummy",
  		"nom",
  		"delicious",
  		"savouring"
  	],
  	char: "😋",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var relieved$1 = {
  	keywords: [
  		"face",
  		"relaxed",
  		"phew",
  		"massage",
  		"happiness"
  	],
  	char: "😌",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var heart_eyes$1 = {
  	keywords: [
  		"face",
  		"love",
  		"like",
  		"affection",
  		"valentines",
  		"infatuation",
  		"crush",
  		"heart"
  	],
  	char: "😍",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var smiling_face_with_three_hearts$1 = {
  	keywords: [
  		"face",
  		"love",
  		"like",
  		"affection",
  		"valentines",
  		"infatuation",
  		"crush",
  		"hearts",
  		"adore"
  	],
  	char: "🥰",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var kissing_heart$1 = {
  	keywords: [
  		"face",
  		"love",
  		"like",
  		"affection",
  		"valentines",
  		"infatuation",
  		"kiss"
  	],
  	char: "😘",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var kissing$1 = {
  	keywords: [
  		"love",
  		"like",
  		"face",
  		"3",
  		"valentines",
  		"infatuation",
  		"kiss"
  	],
  	char: "😗",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var kissing_smiling_eyes$1 = {
  	keywords: [
  		"face",
  		"affection",
  		"valentines",
  		"infatuation",
  		"kiss"
  	],
  	char: "😙",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var kissing_closed_eyes$1 = {
  	keywords: [
  		"face",
  		"love",
  		"like",
  		"affection",
  		"valentines",
  		"infatuation",
  		"kiss"
  	],
  	char: "😚",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var stuck_out_tongue_winking_eye$1 = {
  	keywords: [
  		"face",
  		"prank",
  		"childish",
  		"playful",
  		"mischievous",
  		"smile",
  		"wink",
  		"tongue"
  	],
  	char: "😜",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var zany = {
  	keywords: [
  		"face",
  		"goofy",
  		"crazy"
  	],
  	char: "🤪",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var raised_eyebrow$1 = {
  	keywords: [
  		"face",
  		"distrust",
  		"scepticism",
  		"disapproval",
  		"disbelief",
  		"surprise"
  	],
  	char: "🤨",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var monocle = {
  	keywords: [
  		"face",
  		"stuffy",
  		"wealthy"
  	],
  	char: "🧐",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var stuck_out_tongue_closed_eyes$1 = {
  	keywords: [
  		"face",
  		"prank",
  		"playful",
  		"mischievous",
  		"smile",
  		"tongue"
  	],
  	char: "😝",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var stuck_out_tongue$1 = {
  	keywords: [
  		"face",
  		"prank",
  		"childish",
  		"playful",
  		"mischievous",
  		"smile",
  		"tongue"
  	],
  	char: "😛",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var money_mouth_face$1 = {
  	keywords: [
  		"face",
  		"rich",
  		"dollar",
  		"money"
  	],
  	char: "🤑",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var nerd_face$1 = {
  	keywords: [
  		"face",
  		"nerdy",
  		"geek",
  		"dork"
  	],
  	char: "🤓",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var sunglasses$1 = {
  	keywords: [
  		"face",
  		"cool",
  		"smile",
  		"summer",
  		"beach",
  		"sunglass"
  	],
  	char: "😎",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var star_struck$1 = {
  	keywords: [
  		"face",
  		"smile",
  		"starry",
  		"eyes",
  		"grinning"
  	],
  	char: "🤩",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var clown_face$1 = {
  	keywords: [
  		"face"
  	],
  	char: "🤡",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var cowboy_hat_face$1 = {
  	keywords: [
  		"face",
  		"cowgirl",
  		"hat"
  	],
  	char: "🤠",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var hugs$1 = {
  	keywords: [
  		"face",
  		"smile",
  		"hug"
  	],
  	char: "🤗",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var smirk$1 = {
  	keywords: [
  		"face",
  		"smile",
  		"mean",
  		"prank",
  		"smug",
  		"sarcasm"
  	],
  	char: "😏",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var no_mouth$1 = {
  	keywords: [
  		"face",
  		"hellokitty"
  	],
  	char: "😶",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var neutral_face$1 = {
  	keywords: [
  		"indifference",
  		"meh",
  		":|",
  		"neutral"
  	],
  	char: "😐",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var expressionless$1 = {
  	keywords: [
  		"face",
  		"indifferent",
  		"-_-",
  		"meh",
  		"deadpan"
  	],
  	char: "😑",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var unamused$1 = {
  	keywords: [
  		"indifference",
  		"bored",
  		"straight face",
  		"serious",
  		"sarcasm",
  		"unimpressed",
  		"skeptical",
  		"dubious",
  		"side_eye"
  	],
  	char: "😒",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var roll_eyes$1 = {
  	keywords: [
  		"face",
  		"eyeroll",
  		"frustrated"
  	],
  	char: "🙄",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var thinking$1 = {
  	keywords: [
  		"face",
  		"hmmm",
  		"think",
  		"consider"
  	],
  	char: "🤔",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var lying_face$1 = {
  	keywords: [
  		"face",
  		"lie",
  		"pinocchio"
  	],
  	char: "🤥",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var hand_over_mouth$1 = {
  	keywords: [
  		"face",
  		"whoops",
  		"shock",
  		"surprise"
  	],
  	char: "🤭",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var shushing = {
  	keywords: [
  		"face",
  		"quiet",
  		"shhh"
  	],
  	char: "🤫",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var symbols_over_mouth = {
  	keywords: [
  		"face",
  		"swearing",
  		"cursing",
  		"cussing",
  		"profanity",
  		"expletive"
  	],
  	char: "🤬",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var exploding_head$1 = {
  	keywords: [
  		"face",
  		"shocked",
  		"mind",
  		"blown"
  	],
  	char: "🤯",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var flushed$1 = {
  	keywords: [
  		"face",
  		"blush",
  		"shy",
  		"flattered"
  	],
  	char: "😳",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var disappointed$1 = {
  	keywords: [
  		"face",
  		"sad",
  		"upset",
  		"depressed",
  		":("
  	],
  	char: "😞",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var worried$1 = {
  	keywords: [
  		"face",
  		"concern",
  		"nervous",
  		":("
  	],
  	char: "😟",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var angry$1 = {
  	keywords: [
  		"mad",
  		"face",
  		"annoyed",
  		"frustrated"
  	],
  	char: "😠",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var rage$1 = {
  	keywords: [
  		"angry",
  		"mad",
  		"hate",
  		"despise"
  	],
  	char: "😡",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var pensive$1 = {
  	keywords: [
  		"face",
  		"sad",
  		"depressed",
  		"upset"
  	],
  	char: "😔",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var confused$1 = {
  	keywords: [
  		"face",
  		"indifference",
  		"huh",
  		"weird",
  		"hmmm",
  		":/"
  	],
  	char: "😕",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var slightly_frowning_face$1 = {
  	keywords: [
  		"face",
  		"frowning",
  		"disappointed",
  		"sad",
  		"upset"
  	],
  	char: "🙁",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var frowning_face$1 = {
  	keywords: [
  		"face",
  		"sad",
  		"upset",
  		"frown"
  	],
  	char: "☹",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var persevere$1 = {
  	keywords: [
  		"face",
  		"sick",
  		"no",
  		"upset",
  		"oops"
  	],
  	char: "😣",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var confounded$1 = {
  	keywords: [
  		"face",
  		"confused",
  		"sick",
  		"unwell",
  		"oops",
  		":S"
  	],
  	char: "😖",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var tired_face$1 = {
  	keywords: [
  		"sick",
  		"whine",
  		"upset",
  		"frustrated"
  	],
  	char: "😫",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var weary$1 = {
  	keywords: [
  		"face",
  		"tired",
  		"sleepy",
  		"sad",
  		"frustrated",
  		"upset"
  	],
  	char: "😩",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var pleading = {
  	keywords: [
  		"face",
  		"begging",
  		"mercy"
  	],
  	char: "🥺",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var triumph$1 = {
  	keywords: [
  		"face",
  		"gas",
  		"phew",
  		"proud",
  		"pride"
  	],
  	char: "😤",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var open_mouth$1 = {
  	keywords: [
  		"face",
  		"surprise",
  		"impressed",
  		"wow",
  		"whoa",
  		":O"
  	],
  	char: "😮",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var scream$1 = {
  	keywords: [
  		"face",
  		"munch",
  		"scared",
  		"omg"
  	],
  	char: "😱",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var fearful$1 = {
  	keywords: [
  		"face",
  		"scared",
  		"terrified",
  		"nervous",
  		"oops",
  		"huh"
  	],
  	char: "😨",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var cold_sweat$1 = {
  	keywords: [
  		"face",
  		"nervous",
  		"sweat"
  	],
  	char: "😰",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var hushed$1 = {
  	keywords: [
  		"face",
  		"woo",
  		"shh"
  	],
  	char: "😯",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var frowning$1 = {
  	keywords: [
  		"face",
  		"aw",
  		"what"
  	],
  	char: "😦",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var anguished$1 = {
  	keywords: [
  		"face",
  		"stunned",
  		"nervous"
  	],
  	char: "😧",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var cry$1 = {
  	keywords: [
  		"face",
  		"tears",
  		"sad",
  		"depressed",
  		"upset",
  		":'("
  	],
  	char: "😢",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var disappointed_relieved$1 = {
  	keywords: [
  		"face",
  		"phew",
  		"sweat",
  		"nervous"
  	],
  	char: "😥",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var drooling_face$1 = {
  	keywords: [
  		"face"
  	],
  	char: "🤤",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var sleepy$1 = {
  	keywords: [
  		"face",
  		"tired",
  		"rest",
  		"nap"
  	],
  	char: "😪",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var sweat$1 = {
  	keywords: [
  		"face",
  		"hot",
  		"sad",
  		"tired",
  		"exercise"
  	],
  	char: "😓",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var hot = {
  	keywords: [
  		"face",
  		"feverish",
  		"heat",
  		"red",
  		"sweating"
  	],
  	char: "🥵",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var cold = {
  	keywords: [
  		"face",
  		"blue",
  		"freezing",
  		"frozen",
  		"frostbite",
  		"icicles"
  	],
  	char: "🥶",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var sob$1 = {
  	keywords: [
  		"face",
  		"cry",
  		"tears",
  		"sad",
  		"upset",
  		"depressed"
  	],
  	char: "😭",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var dizzy_face$1 = {
  	keywords: [
  		"spent",
  		"unconscious",
  		"xox",
  		"dizzy"
  	],
  	char: "😵",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var astonished$1 = {
  	keywords: [
  		"face",
  		"xox",
  		"surprised",
  		"poisoned"
  	],
  	char: "😲",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var zipper_mouth_face$1 = {
  	keywords: [
  		"face",
  		"sealed",
  		"zipper",
  		"secret"
  	],
  	char: "🤐",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var nauseated_face$1 = {
  	keywords: [
  		"face",
  		"vomit",
  		"gross",
  		"green",
  		"sick",
  		"throw up",
  		"ill"
  	],
  	char: "🤢",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var sneezing_face$1 = {
  	keywords: [
  		"face",
  		"gesundheit",
  		"sneeze",
  		"sick",
  		"allergy"
  	],
  	char: "🤧",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var vomiting = {
  	keywords: [
  		"face",
  		"sick"
  	],
  	char: "🤮",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var mask$1 = {
  	keywords: [
  		"face",
  		"sick",
  		"ill",
  		"disease"
  	],
  	char: "😷",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var face_with_thermometer$1 = {
  	keywords: [
  		"sick",
  		"temperature",
  		"thermometer",
  		"cold",
  		"fever"
  	],
  	char: "🤒",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var face_with_head_bandage$1 = {
  	keywords: [
  		"injured",
  		"clumsy",
  		"bandage",
  		"hurt"
  	],
  	char: "🤕",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var woozy = {
  	keywords: [
  		"face",
  		"dizzy",
  		"intoxicated",
  		"tipsy",
  		"wavy"
  	],
  	char: "🥴",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var sleeping$1 = {
  	keywords: [
  		"face",
  		"tired",
  		"sleepy",
  		"night",
  		"zzz"
  	],
  	char: "😴",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var zzz$1 = {
  	keywords: [
  		"sleepy",
  		"tired",
  		"dream"
  	],
  	char: "💤",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var poop$1 = {
  	keywords: [
  		"hankey",
  		"shitface",
  		"fail",
  		"turd",
  		"shit"
  	],
  	char: "💩",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var smiling_imp$1 = {
  	keywords: [
  		"devil",
  		"horns"
  	],
  	char: "😈",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var imp$1 = {
  	keywords: [
  		"devil",
  		"angry",
  		"horns"
  	],
  	char: "👿",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var japanese_ogre$1 = {
  	keywords: [
  		"monster",
  		"red",
  		"mask",
  		"halloween",
  		"scary",
  		"creepy",
  		"devil",
  		"demon",
  		"japanese",
  		"ogre"
  	],
  	char: "👹",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var japanese_goblin$1 = {
  	keywords: [
  		"red",
  		"evil",
  		"mask",
  		"monster",
  		"scary",
  		"creepy",
  		"japanese",
  		"goblin"
  	],
  	char: "👺",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var skull$1 = {
  	keywords: [
  		"dead",
  		"skeleton",
  		"creepy",
  		"death"
  	],
  	char: "💀",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var ghost$1 = {
  	keywords: [
  		"halloween",
  		"spooky",
  		"scary"
  	],
  	char: "👻",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var alien$1 = {
  	keywords: [
  		"UFO",
  		"paul",
  		"weird",
  		"outer_space"
  	],
  	char: "👽",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var robot$1 = {
  	keywords: [
  		"computer",
  		"machine",
  		"bot"
  	],
  	char: "🤖",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var smiley_cat$1 = {
  	keywords: [
  		"animal",
  		"cats",
  		"happy",
  		"smile"
  	],
  	char: "😺",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var smile_cat$1 = {
  	keywords: [
  		"animal",
  		"cats",
  		"smile"
  	],
  	char: "😸",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var joy_cat$1 = {
  	keywords: [
  		"animal",
  		"cats",
  		"haha",
  		"happy",
  		"tears"
  	],
  	char: "😹",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var heart_eyes_cat$1 = {
  	keywords: [
  		"animal",
  		"love",
  		"like",
  		"affection",
  		"cats",
  		"valentines",
  		"heart"
  	],
  	char: "😻",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var smirk_cat$1 = {
  	keywords: [
  		"animal",
  		"cats",
  		"smirk"
  	],
  	char: "😼",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var kissing_cat$1 = {
  	keywords: [
  		"animal",
  		"cats",
  		"kiss"
  	],
  	char: "😽",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var scream_cat$1 = {
  	keywords: [
  		"animal",
  		"cats",
  		"munch",
  		"scared",
  		"scream"
  	],
  	char: "🙀",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var crying_cat_face$1 = {
  	keywords: [
  		"animal",
  		"tears",
  		"weep",
  		"sad",
  		"cats",
  		"upset",
  		"cry"
  	],
  	char: "😿",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var pouting_cat$1 = {
  	keywords: [
  		"animal",
  		"cats"
  	],
  	char: "😾",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var palms_up = {
  	keywords: [
  		"hands",
  		"gesture",
  		"cupped",
  		"prayer"
  	],
  	char: "🤲",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var raised_hands$1 = {
  	keywords: [
  		"gesture",
  		"hooray",
  		"yea",
  		"celebration",
  		"hands"
  	],
  	char: "🙌",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var clap$1 = {
  	keywords: [
  		"hands",
  		"praise",
  		"applause",
  		"congrats",
  		"yay"
  	],
  	char: "👏",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var wave$1 = {
  	keywords: [
  		"hands",
  		"gesture",
  		"goodbye",
  		"solong",
  		"farewell",
  		"hello",
  		"hi",
  		"palm"
  	],
  	char: "👋",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var call_me_hand$1 = {
  	keywords: [
  		"hands",
  		"gesture"
  	],
  	char: "🤙",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var facepunch$1 = {
  	keywords: [
  		"angry",
  		"violence",
  		"fist",
  		"hit",
  		"attack",
  		"hand"
  	],
  	char: "👊",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var fist$1 = {
  	keywords: [
  		"fingers",
  		"hand",
  		"grasp"
  	],
  	char: "✊",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var fist_left$1 = {
  	keywords: [
  		"hand",
  		"fistbump"
  	],
  	char: "🤛",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var fist_right$1 = {
  	keywords: [
  		"hand",
  		"fistbump"
  	],
  	char: "🤜",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var v$1 = {
  	keywords: [
  		"fingers",
  		"ohyeah",
  		"hand",
  		"peace",
  		"victory",
  		"two"
  	],
  	char: "✌",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var ok_hand$1 = {
  	keywords: [
  		"fingers",
  		"limbs",
  		"perfect",
  		"ok",
  		"okay"
  	],
  	char: "👌",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var raised_hand$1 = {
  	keywords: [
  		"fingers",
  		"stop",
  		"highfive",
  		"palm",
  		"ban"
  	],
  	char: "✋",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var raised_back_of_hand$1 = {
  	keywords: [
  		"fingers",
  		"raised",
  		"backhand"
  	],
  	char: "🤚",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var open_hands$1 = {
  	keywords: [
  		"fingers",
  		"butterfly",
  		"hands",
  		"open"
  	],
  	char: "👐",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var muscle$1 = {
  	keywords: [
  		"arm",
  		"flex",
  		"hand",
  		"summer",
  		"strong",
  		"biceps"
  	],
  	char: "💪",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var pray$1 = {
  	keywords: [
  		"please",
  		"hope",
  		"wish",
  		"namaste",
  		"highfive"
  	],
  	char: "🙏",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var foot$1 = {
  	keywords: [
  		"kick",
  		"stomp"
  	],
  	char: "🦶",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var leg$2 = {
  	keywords: [
  		"kick",
  		"limb"
  	],
  	char: "🦵",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var handshake$1 = {
  	keywords: [
  		"agreement",
  		"shake"
  	],
  	char: "🤝",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var point_up$1 = {
  	keywords: [
  		"hand",
  		"fingers",
  		"direction",
  		"up"
  	],
  	char: "☝",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var point_up_2$1 = {
  	keywords: [
  		"fingers",
  		"hand",
  		"direction",
  		"up"
  	],
  	char: "👆",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var point_down$1 = {
  	keywords: [
  		"fingers",
  		"hand",
  		"direction",
  		"down"
  	],
  	char: "👇",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var point_left$1 = {
  	keywords: [
  		"direction",
  		"fingers",
  		"hand",
  		"left"
  	],
  	char: "👈",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var point_right$1 = {
  	keywords: [
  		"fingers",
  		"hand",
  		"direction",
  		"right"
  	],
  	char: "👉",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var fu$1 = {
  	keywords: [
  		"hand",
  		"fingers",
  		"rude",
  		"middle",
  		"flipping"
  	],
  	char: "🖕",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var raised_hand_with_fingers_splayed$1 = {
  	keywords: [
  		"hand",
  		"fingers",
  		"palm"
  	],
  	char: "🖐",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var love_you = {
  	keywords: [
  		"hand",
  		"fingers",
  		"gesture"
  	],
  	char: "🤟",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var metal$1 = {
  	keywords: [
  		"hand",
  		"fingers",
  		"evil_eye",
  		"sign_of_horns",
  		"rock_on"
  	],
  	char: "🤘",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var crossed_fingers$1 = {
  	keywords: [
  		"good",
  		"lucky"
  	],
  	char: "🤞",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var vulcan_salute$1 = {
  	keywords: [
  		"hand",
  		"fingers",
  		"spock",
  		"star trek"
  	],
  	char: "🖖",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var writing_hand$1 = {
  	keywords: [
  		"lower_left_ballpoint_pen",
  		"stationery",
  		"write",
  		"compose"
  	],
  	char: "✍",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var selfie$1 = {
  	keywords: [
  		"camera",
  		"phone"
  	],
  	char: "🤳",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var nail_care$1 = {
  	keywords: [
  		"beauty",
  		"manicure",
  		"finger",
  		"fashion",
  		"nail"
  	],
  	char: "💅",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var lips$1 = {
  	keywords: [
  		"mouth",
  		"kiss"
  	],
  	char: "👄",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var tooth$1 = {
  	keywords: [
  		"teeth",
  		"dentist"
  	],
  	char: "🦷",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var tongue$1 = {
  	keywords: [
  		"mouth",
  		"playful"
  	],
  	char: "👅",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var ear$1 = {
  	keywords: [
  		"face",
  		"hear",
  		"sound",
  		"listen"
  	],
  	char: "👂",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var nose$1 = {
  	keywords: [
  		"smell",
  		"sniff"
  	],
  	char: "👃",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var eye$1 = {
  	keywords: [
  		"face",
  		"look",
  		"see",
  		"watch",
  		"stare"
  	],
  	char: "👁",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var eyes$1 = {
  	keywords: [
  		"look",
  		"watch",
  		"stalk",
  		"peek",
  		"see"
  	],
  	char: "👀",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var brain$1 = {
  	keywords: [
  		"smart",
  		"intelligent"
  	],
  	char: "🧠",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var bust_in_silhouette$1 = {
  	keywords: [
  		"user",
  		"person",
  		"human"
  	],
  	char: "👤",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var busts_in_silhouette$1 = {
  	keywords: [
  		"user",
  		"person",
  		"human",
  		"group",
  		"team"
  	],
  	char: "👥",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var speaking_head$1 = {
  	keywords: [
  		"user",
  		"person",
  		"human",
  		"sing",
  		"say",
  		"talk"
  	],
  	char: "🗣",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var baby$1 = {
  	keywords: [
  		"child",
  		"boy",
  		"girl",
  		"toddler"
  	],
  	char: "👶",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var child$1 = {
  	keywords: [
  		"gender-neutral",
  		"young"
  	],
  	char: "🧒",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var boy$1 = {
  	keywords: [
  		"man",
  		"male",
  		"guy",
  		"teenager"
  	],
  	char: "👦",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var girl$1 = {
  	keywords: [
  		"female",
  		"woman",
  		"teenager"
  	],
  	char: "👧",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var adult$1 = {
  	keywords: [
  		"gender-neutral",
  		"person"
  	],
  	char: "🧑",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man$1 = {
  	keywords: [
  		"mustache",
  		"father",
  		"dad",
  		"guy",
  		"classy",
  		"sir",
  		"moustache"
  	],
  	char: "👨",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman$1 = {
  	keywords: [
  		"female",
  		"girls",
  		"lady"
  	],
  	char: "👩",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var blonde_woman$1 = {
  	keywords: [
  		"woman",
  		"female",
  		"girl",
  		"blonde",
  		"person"
  	],
  	char: "👱‍♀️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var blonde_man = {
  	keywords: [
  		"man",
  		"male",
  		"boy",
  		"blonde",
  		"guy",
  		"person"
  	],
  	char: "👱",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var bearded_person$1 = {
  	keywords: [
  		"person",
  		"bewhiskered"
  	],
  	char: "🧔",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var older_adult$1 = {
  	keywords: [
  		"human",
  		"elder",
  		"senior",
  		"gender-neutral"
  	],
  	char: "🧓",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var older_man$1 = {
  	keywords: [
  		"human",
  		"male",
  		"men",
  		"old",
  		"elder",
  		"senior"
  	],
  	char: "👴",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var older_woman$1 = {
  	keywords: [
  		"human",
  		"female",
  		"women",
  		"lady",
  		"old",
  		"elder",
  		"senior"
  	],
  	char: "👵",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_with_gua_pi_mao$1 = {
  	keywords: [
  		"male",
  		"boy",
  		"chinese"
  	],
  	char: "👲",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_with_headscarf$1 = {
  	keywords: [
  		"female",
  		"hijab",
  		"mantilla",
  		"tichel"
  	],
  	char: "🧕",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_with_turban$1 = {
  	keywords: [
  		"female",
  		"indian",
  		"hinduism",
  		"arabs",
  		"woman"
  	],
  	char: "👳‍♀️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_with_turban$1 = {
  	keywords: [
  		"male",
  		"indian",
  		"hinduism",
  		"arabs"
  	],
  	char: "👳",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var policewoman$1 = {
  	keywords: [
  		"woman",
  		"police",
  		"law",
  		"legal",
  		"enforcement",
  		"arrest",
  		"911",
  		"female"
  	],
  	char: "👮‍♀️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var policeman$1 = {
  	keywords: [
  		"man",
  		"police",
  		"law",
  		"legal",
  		"enforcement",
  		"arrest",
  		"911"
  	],
  	char: "👮",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var construction_worker_woman$1 = {
  	keywords: [
  		"female",
  		"human",
  		"wip",
  		"build",
  		"construction",
  		"worker",
  		"labor",
  		"woman"
  	],
  	char: "👷‍♀️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var construction_worker_man$1 = {
  	keywords: [
  		"male",
  		"human",
  		"wip",
  		"guy",
  		"build",
  		"construction",
  		"worker",
  		"labor"
  	],
  	char: "👷",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var guardswoman$1 = {
  	keywords: [
  		"uk",
  		"gb",
  		"british",
  		"female",
  		"royal",
  		"woman"
  	],
  	char: "💂‍♀️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var guardsman$1 = {
  	keywords: [
  		"uk",
  		"gb",
  		"british",
  		"male",
  		"guy",
  		"royal"
  	],
  	char: "💂",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var female_detective$1 = {
  	keywords: [
  		"human",
  		"spy",
  		"detective",
  		"female",
  		"woman"
  	],
  	char: "🕵️‍♀️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var male_detective$1 = {
  	keywords: [
  		"human",
  		"spy",
  		"detective"
  	],
  	char: "🕵",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_health_worker$1 = {
  	keywords: [
  		"doctor",
  		"nurse",
  		"therapist",
  		"healthcare",
  		"woman",
  		"human"
  	],
  	char: "👩‍⚕️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_health_worker$1 = {
  	keywords: [
  		"doctor",
  		"nurse",
  		"therapist",
  		"healthcare",
  		"man",
  		"human"
  	],
  	char: "👨‍⚕️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_farmer$1 = {
  	keywords: [
  		"rancher",
  		"gardener",
  		"woman",
  		"human"
  	],
  	char: "👩‍🌾",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_farmer$1 = {
  	keywords: [
  		"rancher",
  		"gardener",
  		"man",
  		"human"
  	],
  	char: "👨‍🌾",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_cook$1 = {
  	keywords: [
  		"chef",
  		"woman",
  		"human"
  	],
  	char: "👩‍🍳",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_cook$1 = {
  	keywords: [
  		"chef",
  		"man",
  		"human"
  	],
  	char: "👨‍🍳",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_student$1 = {
  	keywords: [
  		"graduate",
  		"woman",
  		"human"
  	],
  	char: "👩‍🎓",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_student$1 = {
  	keywords: [
  		"graduate",
  		"man",
  		"human"
  	],
  	char: "👨‍🎓",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_singer$1 = {
  	keywords: [
  		"rockstar",
  		"entertainer",
  		"woman",
  		"human"
  	],
  	char: "👩‍🎤",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_singer$1 = {
  	keywords: [
  		"rockstar",
  		"entertainer",
  		"man",
  		"human"
  	],
  	char: "👨‍🎤",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_teacher$1 = {
  	keywords: [
  		"instructor",
  		"professor",
  		"woman",
  		"human"
  	],
  	char: "👩‍🏫",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_teacher$1 = {
  	keywords: [
  		"instructor",
  		"professor",
  		"man",
  		"human"
  	],
  	char: "👨‍🏫",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_factory_worker$1 = {
  	keywords: [
  		"assembly",
  		"industrial",
  		"woman",
  		"human"
  	],
  	char: "👩‍🏭",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_factory_worker$1 = {
  	keywords: [
  		"assembly",
  		"industrial",
  		"man",
  		"human"
  	],
  	char: "👨‍🏭",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_technologist$1 = {
  	keywords: [
  		"coder",
  		"developer",
  		"engineer",
  		"programmer",
  		"software",
  		"woman",
  		"human",
  		"laptop",
  		"computer"
  	],
  	char: "👩‍💻",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_technologist$1 = {
  	keywords: [
  		"coder",
  		"developer",
  		"engineer",
  		"programmer",
  		"software",
  		"man",
  		"human",
  		"laptop",
  		"computer"
  	],
  	char: "👨‍💻",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_office_worker$1 = {
  	keywords: [
  		"business",
  		"manager",
  		"woman",
  		"human"
  	],
  	char: "👩‍💼",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_office_worker$1 = {
  	keywords: [
  		"business",
  		"manager",
  		"man",
  		"human"
  	],
  	char: "👨‍💼",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_mechanic$1 = {
  	keywords: [
  		"plumber",
  		"woman",
  		"human",
  		"wrench"
  	],
  	char: "👩‍🔧",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_mechanic$1 = {
  	keywords: [
  		"plumber",
  		"man",
  		"human",
  		"wrench"
  	],
  	char: "👨‍🔧",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_scientist$1 = {
  	keywords: [
  		"biologist",
  		"chemist",
  		"engineer",
  		"physicist",
  		"woman",
  		"human"
  	],
  	char: "👩‍🔬",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_scientist$1 = {
  	keywords: [
  		"biologist",
  		"chemist",
  		"engineer",
  		"physicist",
  		"man",
  		"human"
  	],
  	char: "👨‍🔬",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_artist$1 = {
  	keywords: [
  		"painter",
  		"woman",
  		"human"
  	],
  	char: "👩‍🎨",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_artist$1 = {
  	keywords: [
  		"painter",
  		"man",
  		"human"
  	],
  	char: "👨‍🎨",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_firefighter$1 = {
  	keywords: [
  		"fireman",
  		"woman",
  		"human"
  	],
  	char: "👩‍🚒",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_firefighter$1 = {
  	keywords: [
  		"fireman",
  		"man",
  		"human"
  	],
  	char: "👨‍🚒",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_pilot$1 = {
  	keywords: [
  		"aviator",
  		"plane",
  		"woman",
  		"human"
  	],
  	char: "👩‍✈️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_pilot$1 = {
  	keywords: [
  		"aviator",
  		"plane",
  		"man",
  		"human"
  	],
  	char: "👨‍✈️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_astronaut$1 = {
  	keywords: [
  		"space",
  		"rocket",
  		"woman",
  		"human"
  	],
  	char: "👩‍🚀",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_astronaut$1 = {
  	keywords: [
  		"space",
  		"rocket",
  		"man",
  		"human"
  	],
  	char: "👨‍🚀",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_judge$1 = {
  	keywords: [
  		"justice",
  		"court",
  		"woman",
  		"human"
  	],
  	char: "👩‍⚖️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_judge$1 = {
  	keywords: [
  		"justice",
  		"court",
  		"man",
  		"human"
  	],
  	char: "👨‍⚖️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_superhero = {
  	keywords: [
  		"woman",
  		"female",
  		"good",
  		"heroine",
  		"superpowers"
  	],
  	char: "🦸‍♀️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_superhero = {
  	keywords: [
  		"man",
  		"male",
  		"good",
  		"hero",
  		"superpowers"
  	],
  	char: "🦸‍♂️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_supervillain = {
  	keywords: [
  		"woman",
  		"female",
  		"evil",
  		"bad",
  		"criminal",
  		"heroine",
  		"superpowers"
  	],
  	char: "🦹‍♀️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_supervillain = {
  	keywords: [
  		"man",
  		"male",
  		"evil",
  		"bad",
  		"criminal",
  		"hero",
  		"superpowers"
  	],
  	char: "🦹‍♂️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var mrs_claus$1 = {
  	keywords: [
  		"woman",
  		"female",
  		"xmas",
  		"mother christmas"
  	],
  	char: "🤶",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var santa$1 = {
  	keywords: [
  		"festival",
  		"man",
  		"male",
  		"xmas",
  		"father christmas"
  	],
  	char: "🎅",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var sorceress = {
  	keywords: [
  		"woman",
  		"female",
  		"mage",
  		"witch"
  	],
  	char: "🧙‍♀️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var wizard = {
  	keywords: [
  		"man",
  		"male",
  		"mage",
  		"sorcerer"
  	],
  	char: "🧙‍♂️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_elf = {
  	keywords: [
  		"woman",
  		"female"
  	],
  	char: "🧝‍♀️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_elf = {
  	keywords: [
  		"man",
  		"male"
  	],
  	char: "🧝‍♂️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_vampire = {
  	keywords: [
  		"woman",
  		"female"
  	],
  	char: "🧛‍♀️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_vampire = {
  	keywords: [
  		"man",
  		"male",
  		"dracula"
  	],
  	char: "🧛‍♂️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_zombie = {
  	keywords: [
  		"woman",
  		"female",
  		"undead",
  		"walking dead"
  	],
  	char: "🧟‍♀️",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var man_zombie = {
  	keywords: [
  		"man",
  		"male",
  		"dracula",
  		"undead",
  		"walking dead"
  	],
  	char: "🧟‍♂️",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var woman_genie = {
  	keywords: [
  		"woman",
  		"female"
  	],
  	char: "🧞‍♀️",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var man_genie = {
  	keywords: [
  		"man",
  		"male"
  	],
  	char: "🧞‍♂️",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var mermaid$1 = {
  	keywords: [
  		"woman",
  		"female",
  		"merwoman",
  		"ariel"
  	],
  	char: "🧜‍♀️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var merman$1 = {
  	keywords: [
  		"man",
  		"male",
  		"triton"
  	],
  	char: "🧜‍♂️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_fairy = {
  	keywords: [
  		"woman",
  		"female"
  	],
  	char: "🧚‍♀️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_fairy = {
  	keywords: [
  		"man",
  		"male"
  	],
  	char: "🧚‍♂️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var angel$1 = {
  	keywords: [
  		"heaven",
  		"wings",
  		"halo"
  	],
  	char: "👼",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var pregnant_woman$1 = {
  	keywords: [
  		"baby"
  	],
  	char: "🤰",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var breastfeeding = {
  	keywords: [
  		"nursing",
  		"baby"
  	],
  	char: "🤱",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var princess$1 = {
  	keywords: [
  		"girl",
  		"woman",
  		"female",
  		"blond",
  		"crown",
  		"royal",
  		"queen"
  	],
  	char: "👸",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var prince$1 = {
  	keywords: [
  		"boy",
  		"man",
  		"male",
  		"crown",
  		"royal",
  		"king"
  	],
  	char: "🤴",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var bride_with_veil$1 = {
  	keywords: [
  		"couple",
  		"marriage",
  		"wedding",
  		"woman",
  		"bride"
  	],
  	char: "👰",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_in_tuxedo$1 = {
  	keywords: [
  		"couple",
  		"marriage",
  		"wedding",
  		"groom"
  	],
  	char: "🤵",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var running_woman$1 = {
  	keywords: [
  		"woman",
  		"walking",
  		"exercise",
  		"race",
  		"running",
  		"female"
  	],
  	char: "🏃‍♀️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var running_man$1 = {
  	keywords: [
  		"man",
  		"walking",
  		"exercise",
  		"race",
  		"running"
  	],
  	char: "🏃",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var walking_woman$1 = {
  	keywords: [
  		"human",
  		"feet",
  		"steps",
  		"woman",
  		"female"
  	],
  	char: "🚶‍♀️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var walking_man$1 = {
  	keywords: [
  		"human",
  		"feet",
  		"steps"
  	],
  	char: "🚶",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var dancer$1 = {
  	keywords: [
  		"female",
  		"girl",
  		"woman",
  		"fun"
  	],
  	char: "💃",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_dancing$1 = {
  	keywords: [
  		"male",
  		"boy",
  		"fun",
  		"dancer"
  	],
  	char: "🕺",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var dancing_women$1 = {
  	keywords: [
  		"female",
  		"bunny",
  		"women",
  		"girls"
  	],
  	char: "👯",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var dancing_men$1 = {
  	keywords: [
  		"male",
  		"bunny",
  		"men",
  		"boys"
  	],
  	char: "👯‍♂️",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var couple$1 = {
  	keywords: [
  		"pair",
  		"people",
  		"human",
  		"love",
  		"date",
  		"dating",
  		"like",
  		"affection",
  		"valentines",
  		"marriage"
  	],
  	char: "👫",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var two_men_holding_hands$1 = {
  	keywords: [
  		"pair",
  		"couple",
  		"love",
  		"like",
  		"bromance",
  		"friendship",
  		"people",
  		"human"
  	],
  	char: "👬",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var two_women_holding_hands$1 = {
  	keywords: [
  		"pair",
  		"friendship",
  		"couple",
  		"love",
  		"like",
  		"female",
  		"people",
  		"human"
  	],
  	char: "👭",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var bowing_woman$1 = {
  	keywords: [
  		"woman",
  		"female",
  		"girl"
  	],
  	char: "🙇‍♀️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var bowing_man$1 = {
  	keywords: [
  		"man",
  		"male",
  		"boy"
  	],
  	char: "🙇",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_facepalming$1 = {
  	keywords: [
  		"man",
  		"male",
  		"boy",
  		"disbelief"
  	],
  	char: "🤦‍♂️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_facepalming$1 = {
  	keywords: [
  		"woman",
  		"female",
  		"girl",
  		"disbelief"
  	],
  	char: "🤦‍♀️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_shrugging$1 = {
  	keywords: [
  		"woman",
  		"female",
  		"girl",
  		"confused",
  		"indifferent",
  		"doubt"
  	],
  	char: "🤷",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_shrugging$1 = {
  	keywords: [
  		"man",
  		"male",
  		"boy",
  		"confused",
  		"indifferent",
  		"doubt"
  	],
  	char: "🤷‍♂️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var tipping_hand_woman$1 = {
  	keywords: [
  		"female",
  		"girl",
  		"woman",
  		"human",
  		"information"
  	],
  	char: "💁",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var tipping_hand_man$1 = {
  	keywords: [
  		"male",
  		"boy",
  		"man",
  		"human",
  		"information"
  	],
  	char: "💁‍♂️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var no_good_woman$1 = {
  	keywords: [
  		"female",
  		"girl",
  		"woman",
  		"nope"
  	],
  	char: "🙅",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var no_good_man$1 = {
  	keywords: [
  		"male",
  		"boy",
  		"man",
  		"nope"
  	],
  	char: "🙅‍♂️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var ok_woman$1 = {
  	keywords: [
  		"women",
  		"girl",
  		"female",
  		"pink",
  		"human",
  		"woman"
  	],
  	char: "🙆",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var ok_man$1 = {
  	keywords: [
  		"men",
  		"boy",
  		"male",
  		"blue",
  		"human",
  		"man"
  	],
  	char: "🙆‍♂️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var raising_hand_woman$1 = {
  	keywords: [
  		"female",
  		"girl",
  		"woman"
  	],
  	char: "🙋",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var raising_hand_man$1 = {
  	keywords: [
  		"male",
  		"boy",
  		"man"
  	],
  	char: "🙋‍♂️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var pouting_woman$1 = {
  	keywords: [
  		"female",
  		"girl",
  		"woman"
  	],
  	char: "🙎",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var pouting_man$1 = {
  	keywords: [
  		"male",
  		"boy",
  		"man"
  	],
  	char: "🙎‍♂️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var frowning_woman$1 = {
  	keywords: [
  		"female",
  		"girl",
  		"woman",
  		"sad",
  		"depressed",
  		"discouraged",
  		"unhappy"
  	],
  	char: "🙍",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var frowning_man$1 = {
  	keywords: [
  		"male",
  		"boy",
  		"man",
  		"sad",
  		"depressed",
  		"discouraged",
  		"unhappy"
  	],
  	char: "🙍‍♂️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var haircut_woman$1 = {
  	keywords: [
  		"female",
  		"girl",
  		"woman"
  	],
  	char: "💇",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var haircut_man$1 = {
  	keywords: [
  		"male",
  		"boy",
  		"man"
  	],
  	char: "💇‍♂️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var massage_woman$1 = {
  	keywords: [
  		"female",
  		"girl",
  		"woman",
  		"head"
  	],
  	char: "💆",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var massage_man$1 = {
  	keywords: [
  		"male",
  		"boy",
  		"man",
  		"head"
  	],
  	char: "💆‍♂️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var woman_in_steamy_room = {
  	keywords: [
  		"female",
  		"woman",
  		"spa",
  		"steamroom",
  		"sauna"
  	],
  	char: "🧖‍♀️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var man_in_steamy_room = {
  	keywords: [
  		"male",
  		"man",
  		"spa",
  		"steamroom",
  		"sauna"
  	],
  	char: "🧖‍♂️",
  	fitzpatrick_scale: true,
  	category: "people"
  };
  var couple_with_heart_woman_man$1 = {
  	keywords: [
  		"pair",
  		"love",
  		"like",
  		"affection",
  		"human",
  		"dating",
  		"valentines",
  		"marriage"
  	],
  	char: "💑",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var couple_with_heart_woman_woman$1 = {
  	keywords: [
  		"pair",
  		"love",
  		"like",
  		"affection",
  		"human",
  		"dating",
  		"valentines",
  		"marriage"
  	],
  	char: "👩‍❤️‍👩",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var couple_with_heart_man_man$1 = {
  	keywords: [
  		"pair",
  		"love",
  		"like",
  		"affection",
  		"human",
  		"dating",
  		"valentines",
  		"marriage"
  	],
  	char: "👨‍❤️‍👨",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var couplekiss_man_woman$1 = {
  	keywords: [
  		"pair",
  		"valentines",
  		"love",
  		"like",
  		"dating",
  		"marriage"
  	],
  	char: "💏",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var couplekiss_woman_woman$1 = {
  	keywords: [
  		"pair",
  		"valentines",
  		"love",
  		"like",
  		"dating",
  		"marriage"
  	],
  	char: "👩‍❤️‍💋‍👩",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var couplekiss_man_man$1 = {
  	keywords: [
  		"pair",
  		"valentines",
  		"love",
  		"like",
  		"dating",
  		"marriage"
  	],
  	char: "👨‍❤️‍💋‍👨",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_man_woman_boy$1 = {
  	keywords: [
  		"home",
  		"parents",
  		"child",
  		"mom",
  		"dad",
  		"father",
  		"mother",
  		"people",
  		"human"
  	],
  	char: "👪",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_man_woman_girl$1 = {
  	keywords: [
  		"home",
  		"parents",
  		"people",
  		"human",
  		"child"
  	],
  	char: "👨‍👩‍👧",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_man_woman_girl_boy$1 = {
  	keywords: [
  		"home",
  		"parents",
  		"people",
  		"human",
  		"children"
  	],
  	char: "👨‍👩‍👧‍👦",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_man_woman_boy_boy$1 = {
  	keywords: [
  		"home",
  		"parents",
  		"people",
  		"human",
  		"children"
  	],
  	char: "👨‍👩‍👦‍👦",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_man_woman_girl_girl$1 = {
  	keywords: [
  		"home",
  		"parents",
  		"people",
  		"human",
  		"children"
  	],
  	char: "👨‍👩‍👧‍👧",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_woman_woman_boy$1 = {
  	keywords: [
  		"home",
  		"parents",
  		"people",
  		"human",
  		"children"
  	],
  	char: "👩‍👩‍👦",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_woman_woman_girl$1 = {
  	keywords: [
  		"home",
  		"parents",
  		"people",
  		"human",
  		"children"
  	],
  	char: "👩‍👩‍👧",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_woman_woman_girl_boy$1 = {
  	keywords: [
  		"home",
  		"parents",
  		"people",
  		"human",
  		"children"
  	],
  	char: "👩‍👩‍👧‍👦",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_woman_woman_boy_boy$1 = {
  	keywords: [
  		"home",
  		"parents",
  		"people",
  		"human",
  		"children"
  	],
  	char: "👩‍👩‍👦‍👦",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_woman_woman_girl_girl$1 = {
  	keywords: [
  		"home",
  		"parents",
  		"people",
  		"human",
  		"children"
  	],
  	char: "👩‍👩‍👧‍👧",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_man_man_boy$1 = {
  	keywords: [
  		"home",
  		"parents",
  		"people",
  		"human",
  		"children"
  	],
  	char: "👨‍👨‍👦",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_man_man_girl$1 = {
  	keywords: [
  		"home",
  		"parents",
  		"people",
  		"human",
  		"children"
  	],
  	char: "👨‍👨‍👧",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_man_man_girl_boy$1 = {
  	keywords: [
  		"home",
  		"parents",
  		"people",
  		"human",
  		"children"
  	],
  	char: "👨‍👨‍👧‍👦",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_man_man_boy_boy$1 = {
  	keywords: [
  		"home",
  		"parents",
  		"people",
  		"human",
  		"children"
  	],
  	char: "👨‍👨‍👦‍👦",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_man_man_girl_girl$1 = {
  	keywords: [
  		"home",
  		"parents",
  		"people",
  		"human",
  		"children"
  	],
  	char: "👨‍👨‍👧‍👧",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_woman_boy$1 = {
  	keywords: [
  		"home",
  		"parent",
  		"people",
  		"human",
  		"child"
  	],
  	char: "👩‍👦",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_woman_girl$1 = {
  	keywords: [
  		"home",
  		"parent",
  		"people",
  		"human",
  		"child"
  	],
  	char: "👩‍👧",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_woman_girl_boy$1 = {
  	keywords: [
  		"home",
  		"parent",
  		"people",
  		"human",
  		"children"
  	],
  	char: "👩‍👧‍👦",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_woman_boy_boy$1 = {
  	keywords: [
  		"home",
  		"parent",
  		"people",
  		"human",
  		"children"
  	],
  	char: "👩‍👦‍👦",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_woman_girl_girl$1 = {
  	keywords: [
  		"home",
  		"parent",
  		"people",
  		"human",
  		"children"
  	],
  	char: "👩‍👧‍👧",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_man_boy$1 = {
  	keywords: [
  		"home",
  		"parent",
  		"people",
  		"human",
  		"child"
  	],
  	char: "👨‍👦",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_man_girl$1 = {
  	keywords: [
  		"home",
  		"parent",
  		"people",
  		"human",
  		"child"
  	],
  	char: "👨‍👧",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_man_girl_boy$1 = {
  	keywords: [
  		"home",
  		"parent",
  		"people",
  		"human",
  		"children"
  	],
  	char: "👨‍👧‍👦",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_man_boy_boy$1 = {
  	keywords: [
  		"home",
  		"parent",
  		"people",
  		"human",
  		"children"
  	],
  	char: "👨‍👦‍👦",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var family_man_girl_girl$1 = {
  	keywords: [
  		"home",
  		"parent",
  		"people",
  		"human",
  		"children"
  	],
  	char: "👨‍👧‍👧",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var yarn$1 = {
  	keywords: [
  		"ball",
  		"crochet",
  		"knit"
  	],
  	char: "🧶",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var thread$1 = {
  	keywords: [
  		"needle",
  		"sewing",
  		"spool",
  		"string"
  	],
  	char: "🧵",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var coat$1 = {
  	keywords: [
  		"jacket"
  	],
  	char: "🧥",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var labcoat = {
  	keywords: [
  		"doctor",
  		"experiment",
  		"scientist",
  		"chemist"
  	],
  	char: "🥼",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var womans_clothes$1 = {
  	keywords: [
  		"fashion",
  		"shopping_bags",
  		"female"
  	],
  	char: "👚",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var tshirt$1 = {
  	keywords: [
  		"fashion",
  		"cloth",
  		"casual",
  		"shirt",
  		"tee"
  	],
  	char: "👕",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var jeans$1 = {
  	keywords: [
  		"fashion",
  		"shopping"
  	],
  	char: "👖",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var necktie$1 = {
  	keywords: [
  		"shirt",
  		"suitup",
  		"formal",
  		"fashion",
  		"cloth",
  		"business"
  	],
  	char: "👔",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var dress$1 = {
  	keywords: [
  		"clothes",
  		"fashion",
  		"shopping"
  	],
  	char: "👗",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var bikini$1 = {
  	keywords: [
  		"swimming",
  		"female",
  		"woman",
  		"girl",
  		"fashion",
  		"beach",
  		"summer"
  	],
  	char: "👙",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var kimono$1 = {
  	keywords: [
  		"dress",
  		"fashion",
  		"women",
  		"female",
  		"japanese"
  	],
  	char: "👘",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var lipstick$1 = {
  	keywords: [
  		"female",
  		"girl",
  		"fashion",
  		"woman"
  	],
  	char: "💄",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var kiss$1 = {
  	keywords: [
  		"face",
  		"lips",
  		"love",
  		"like",
  		"affection",
  		"valentines"
  	],
  	char: "💋",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var footprints$1 = {
  	keywords: [
  		"feet",
  		"tracking",
  		"walking",
  		"beach"
  	],
  	char: "👣",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var flat_shoe$1 = {
  	keywords: [
  		"ballet",
  		"slip-on",
  		"slipper"
  	],
  	char: "🥿",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var high_heel$1 = {
  	keywords: [
  		"fashion",
  		"shoes",
  		"female",
  		"pumps",
  		"stiletto"
  	],
  	char: "👠",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var sandal$1 = {
  	keywords: [
  		"shoes",
  		"fashion",
  		"flip flops"
  	],
  	char: "👡",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var boot$1 = {
  	keywords: [
  		"shoes",
  		"fashion"
  	],
  	char: "👢",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var mans_shoe$1 = {
  	keywords: [
  		"fashion",
  		"male"
  	],
  	char: "👞",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var athletic_shoe$1 = {
  	keywords: [
  		"shoes",
  		"sports",
  		"sneakers"
  	],
  	char: "👟",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var hiking_boot$1 = {
  	keywords: [
  		"backpacking",
  		"camping",
  		"hiking"
  	],
  	char: "🥾",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var socks$1 = {
  	keywords: [
  		"stockings",
  		"clothes"
  	],
  	char: "🧦",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var gloves$1 = {
  	keywords: [
  		"hands",
  		"winter",
  		"clothes"
  	],
  	char: "🧤",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var scarf$1 = {
  	keywords: [
  		"neck",
  		"winter",
  		"clothes"
  	],
  	char: "🧣",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var womans_hat$1 = {
  	keywords: [
  		"fashion",
  		"accessories",
  		"female",
  		"lady",
  		"spring"
  	],
  	char: "👒",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var tophat$1 = {
  	keywords: [
  		"magic",
  		"gentleman",
  		"classy",
  		"circus"
  	],
  	char: "🎩",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var billed_hat = {
  	keywords: [
  		"cap",
  		"baseball"
  	],
  	char: "🧢",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var rescue_worker_helmet$1 = {
  	keywords: [
  		"construction",
  		"build"
  	],
  	char: "⛑",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var mortar_board$1 = {
  	keywords: [
  		"school",
  		"college",
  		"degree",
  		"university",
  		"graduation",
  		"cap",
  		"hat",
  		"legal",
  		"learn",
  		"education"
  	],
  	char: "🎓",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var crown$1 = {
  	keywords: [
  		"king",
  		"kod",
  		"leader",
  		"royalty",
  		"lord"
  	],
  	char: "👑",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var school_satchel$1 = {
  	keywords: [
  		"student",
  		"education",
  		"bag",
  		"backpack"
  	],
  	char: "🎒",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var luggage$1 = {
  	keywords: [
  		"packing",
  		"travel"
  	],
  	char: "🧳",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var pouch$1 = {
  	keywords: [
  		"bag",
  		"accessories",
  		"shopping"
  	],
  	char: "👝",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var purse$1 = {
  	keywords: [
  		"fashion",
  		"accessories",
  		"money",
  		"sales",
  		"shopping"
  	],
  	char: "👛",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var handbag$1 = {
  	keywords: [
  		"fashion",
  		"accessory",
  		"accessories",
  		"shopping"
  	],
  	char: "👜",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var briefcase$1 = {
  	keywords: [
  		"business",
  		"documents",
  		"work",
  		"law",
  		"legal",
  		"job",
  		"career"
  	],
  	char: "💼",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var eyeglasses$1 = {
  	keywords: [
  		"fashion",
  		"accessories",
  		"eyesight",
  		"nerdy",
  		"dork",
  		"geek"
  	],
  	char: "👓",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var dark_sunglasses$1 = {
  	keywords: [
  		"face",
  		"cool",
  		"accessories"
  	],
  	char: "🕶",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var goggles$1 = {
  	keywords: [
  		"eyes",
  		"protection",
  		"safety"
  	],
  	char: "🥽",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var ring$2 = {
  	keywords: [
  		"wedding",
  		"propose",
  		"marriage",
  		"valentines",
  		"diamond",
  		"fashion",
  		"jewelry",
  		"gem",
  		"engagement"
  	],
  	char: "💍",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var closed_umbrella$1 = {
  	keywords: [
  		"weather",
  		"rain",
  		"drizzle"
  	],
  	char: "🌂",
  	fitzpatrick_scale: false,
  	category: "people"
  };
  var dog$1 = {
  	keywords: [
  		"animal",
  		"friend",
  		"nature",
  		"woof",
  		"puppy",
  		"pet",
  		"faithful"
  	],
  	char: "🐶",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var cat$1 = {
  	keywords: [
  		"animal",
  		"meow",
  		"nature",
  		"pet",
  		"kitten"
  	],
  	char: "🐱",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var mouse$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"cheese_wedge",
  		"rodent"
  	],
  	char: "🐭",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var hamster$1 = {
  	keywords: [
  		"animal",
  		"nature"
  	],
  	char: "🐹",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var rabbit$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"pet",
  		"spring",
  		"magic",
  		"bunny"
  	],
  	char: "🐰",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var fox_face$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"face"
  	],
  	char: "🦊",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var bear$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"wild"
  	],
  	char: "🐻",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var panda_face$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"panda"
  	],
  	char: "🐼",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var koala$1 = {
  	keywords: [
  		"animal",
  		"nature"
  	],
  	char: "🐨",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var tiger$1 = {
  	keywords: [
  		"animal",
  		"cat",
  		"danger",
  		"wild",
  		"nature",
  		"roar"
  	],
  	char: "🐯",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var lion$1 = {
  	keywords: [
  		"animal",
  		"nature"
  	],
  	char: "🦁",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var cow$1 = {
  	keywords: [
  		"beef",
  		"ox",
  		"animal",
  		"nature",
  		"moo",
  		"milk"
  	],
  	char: "🐮",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var pig$1 = {
  	keywords: [
  		"animal",
  		"oink",
  		"nature"
  	],
  	char: "🐷",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var pig_nose$1 = {
  	keywords: [
  		"animal",
  		"oink"
  	],
  	char: "🐽",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var frog$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"croak",
  		"toad"
  	],
  	char: "🐸",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var squid$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"ocean",
  		"sea"
  	],
  	char: "🦑",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var octopus$1 = {
  	keywords: [
  		"animal",
  		"creature",
  		"ocean",
  		"sea",
  		"nature",
  		"beach"
  	],
  	char: "🐙",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var shrimp$1 = {
  	keywords: [
  		"animal",
  		"ocean",
  		"nature",
  		"seafood"
  	],
  	char: "🦐",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var monkey_face$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"circus"
  	],
  	char: "🐵",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var gorilla$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"circus"
  	],
  	char: "🦍",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var see_no_evil$1 = {
  	keywords: [
  		"monkey",
  		"animal",
  		"nature",
  		"haha"
  	],
  	char: "🙈",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var hear_no_evil$1 = {
  	keywords: [
  		"animal",
  		"monkey",
  		"nature"
  	],
  	char: "🙉",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var speak_no_evil$1 = {
  	keywords: [
  		"monkey",
  		"animal",
  		"nature",
  		"omg"
  	],
  	char: "🙊",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var monkey$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"banana",
  		"circus"
  	],
  	char: "🐒",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var chicken$1 = {
  	keywords: [
  		"animal",
  		"cluck",
  		"nature",
  		"bird"
  	],
  	char: "🐔",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var penguin$1 = {
  	keywords: [
  		"animal",
  		"nature"
  	],
  	char: "🐧",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var bird$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"fly",
  		"tweet",
  		"spring"
  	],
  	char: "🐦",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var baby_chick$1 = {
  	keywords: [
  		"animal",
  		"chicken",
  		"bird"
  	],
  	char: "🐤",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var hatching_chick$1 = {
  	keywords: [
  		"animal",
  		"chicken",
  		"egg",
  		"born",
  		"baby",
  		"bird"
  	],
  	char: "🐣",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var hatched_chick$1 = {
  	keywords: [
  		"animal",
  		"chicken",
  		"baby",
  		"bird"
  	],
  	char: "🐥",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var duck$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"bird",
  		"mallard"
  	],
  	char: "🦆",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var eagle$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"bird"
  	],
  	char: "🦅",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var owl$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"bird",
  		"hoot"
  	],
  	char: "🦉",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var bat$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"blind",
  		"vampire"
  	],
  	char: "🦇",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var wolf$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"wild"
  	],
  	char: "🐺",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var boar$1 = {
  	keywords: [
  		"animal",
  		"nature"
  	],
  	char: "🐗",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var horse$1 = {
  	keywords: [
  		"animal",
  		"brown",
  		"nature"
  	],
  	char: "🐴",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var unicorn$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"mystical"
  	],
  	char: "🦄",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var honeybee$1 = {
  	keywords: [
  		"animal",
  		"insect",
  		"nature",
  		"bug",
  		"spring",
  		"honey"
  	],
  	char: "🐝",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var bug$1 = {
  	keywords: [
  		"animal",
  		"insect",
  		"nature",
  		"worm"
  	],
  	char: "🐛",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var butterfly$1 = {
  	keywords: [
  		"animal",
  		"insect",
  		"nature",
  		"caterpillar"
  	],
  	char: "🦋",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var snail$1 = {
  	keywords: [
  		"slow",
  		"animal",
  		"shell"
  	],
  	char: "🐌",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var beetle$1 = {
  	keywords: [
  		"animal",
  		"insect",
  		"nature",
  		"ladybug"
  	],
  	char: "🐞",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var ant$1 = {
  	keywords: [
  		"animal",
  		"insect",
  		"nature",
  		"bug"
  	],
  	char: "🐜",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var grasshopper = {
  	keywords: [
  		"animal",
  		"cricket",
  		"chirp"
  	],
  	char: "🦗",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var spider$1 = {
  	keywords: [
  		"animal",
  		"arachnid"
  	],
  	char: "🕷",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var scorpion$1 = {
  	keywords: [
  		"animal",
  		"arachnid"
  	],
  	char: "🦂",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var crab$1 = {
  	keywords: [
  		"animal",
  		"crustacean"
  	],
  	char: "🦀",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var snake$1 = {
  	keywords: [
  		"animal",
  		"evil",
  		"nature",
  		"hiss",
  		"python"
  	],
  	char: "🐍",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var lizard$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"reptile"
  	],
  	char: "🦎",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var sauropod$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"dinosaur",
  		"brachiosaurus",
  		"brontosaurus",
  		"diplodocus",
  		"extinct"
  	],
  	char: "🦕",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var turtle$1 = {
  	keywords: [
  		"animal",
  		"slow",
  		"nature",
  		"tortoise"
  	],
  	char: "🐢",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var tropical_fish$1 = {
  	keywords: [
  		"animal",
  		"swim",
  		"ocean",
  		"beach",
  		"nemo"
  	],
  	char: "🐠",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var fish$1 = {
  	keywords: [
  		"animal",
  		"food",
  		"nature"
  	],
  	char: "🐟",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var blowfish$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"food",
  		"sea",
  		"ocean"
  	],
  	char: "🐡",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var dolphin$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"fish",
  		"sea",
  		"ocean",
  		"flipper",
  		"fins",
  		"beach"
  	],
  	char: "🐬",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var shark$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"fish",
  		"sea",
  		"ocean",
  		"jaws",
  		"fins",
  		"beach"
  	],
  	char: "🦈",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var whale$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"sea",
  		"ocean"
  	],
  	char: "🐳",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var whale2$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"sea",
  		"ocean"
  	],
  	char: "🐋",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var crocodile$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"reptile",
  		"lizard",
  		"alligator"
  	],
  	char: "🐊",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var leopard$1 = {
  	keywords: [
  		"animal",
  		"nature"
  	],
  	char: "🐆",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var zebra$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"stripes",
  		"safari"
  	],
  	char: "🦓",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var tiger2$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"roar"
  	],
  	char: "🐅",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var water_buffalo$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"ox",
  		"cow"
  	],
  	char: "🐃",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var ox$1 = {
  	keywords: [
  		"animal",
  		"cow",
  		"beef"
  	],
  	char: "🐂",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var cow2$1 = {
  	keywords: [
  		"beef",
  		"ox",
  		"animal",
  		"nature",
  		"moo",
  		"milk"
  	],
  	char: "🐄",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var deer$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"horns",
  		"venison"
  	],
  	char: "🦌",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var dromedary_camel$1 = {
  	keywords: [
  		"animal",
  		"hot",
  		"desert",
  		"hump"
  	],
  	char: "🐪",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var camel$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"hot",
  		"desert",
  		"hump"
  	],
  	char: "🐫",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var giraffe$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"spots",
  		"safari"
  	],
  	char: "🦒",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var elephant$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"nose",
  		"th",
  		"circus"
  	],
  	char: "🐘",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var rhinoceros$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"horn"
  	],
  	char: "🦏",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var goat$1 = {
  	keywords: [
  		"animal",
  		"nature"
  	],
  	char: "🐐",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var ram$1 = {
  	keywords: [
  		"animal",
  		"sheep",
  		"nature"
  	],
  	char: "🐏",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var sheep$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"wool",
  		"shipit"
  	],
  	char: "🐑",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var racehorse$1 = {
  	keywords: [
  		"animal",
  		"gamble",
  		"luck"
  	],
  	char: "🐎",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var pig2$1 = {
  	keywords: [
  		"animal",
  		"nature"
  	],
  	char: "🐖",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var rat$1 = {
  	keywords: [
  		"animal",
  		"mouse",
  		"rodent"
  	],
  	char: "🐀",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var mouse2$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"rodent"
  	],
  	char: "🐁",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var rooster$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"chicken"
  	],
  	char: "🐓",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var turkey$1 = {
  	keywords: [
  		"animal",
  		"bird"
  	],
  	char: "🦃",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var dove$1 = {
  	keywords: [
  		"animal",
  		"bird"
  	],
  	char: "🕊",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var dog2$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"friend",
  		"doge",
  		"pet",
  		"faithful"
  	],
  	char: "🐕",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var poodle$1 = {
  	keywords: [
  		"dog",
  		"animal",
  		"101",
  		"nature",
  		"pet"
  	],
  	char: "🐩",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var cat2$1 = {
  	keywords: [
  		"animal",
  		"meow",
  		"pet",
  		"cats"
  	],
  	char: "🐈",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var rabbit2$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"pet",
  		"magic",
  		"spring"
  	],
  	char: "🐇",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var chipmunk$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"rodent",
  		"squirrel"
  	],
  	char: "🐿",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var hedgehog$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"spiny"
  	],
  	char: "🦔",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var raccoon$1 = {
  	keywords: [
  		"animal",
  		"nature"
  	],
  	char: "🦝",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var llama$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"alpaca"
  	],
  	char: "🦙",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var hippopotamus$1 = {
  	keywords: [
  		"animal",
  		"nature"
  	],
  	char: "🦛",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var kangaroo$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"australia",
  		"joey",
  		"hop",
  		"marsupial"
  	],
  	char: "🦘",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var badger$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"honey"
  	],
  	char: "🦡",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var swan$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"bird"
  	],
  	char: "🦢",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var peacock$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"peahen",
  		"bird"
  	],
  	char: "🦚",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var parrot$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"bird",
  		"pirate",
  		"talk"
  	],
  	char: "🦜",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var lobster$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"bisque",
  		"claws",
  		"seafood"
  	],
  	char: "🦞",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var mosquito$1 = {
  	keywords: [
  		"animal",
  		"nature",
  		"insect",
  		"malaria"
  	],
  	char: "🦟",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var paw_prints$1 = {
  	keywords: [
  		"animal",
  		"tracking",
  		"footprints",
  		"dog",
  		"cat",
  		"pet",
  		"feet"
  	],
  	char: "🐾",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var dragon$1 = {
  	keywords: [
  		"animal",
  		"myth",
  		"nature",
  		"chinese",
  		"green"
  	],
  	char: "🐉",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var dragon_face$1 = {
  	keywords: [
  		"animal",
  		"myth",
  		"nature",
  		"chinese",
  		"green"
  	],
  	char: "🐲",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var cactus$1 = {
  	keywords: [
  		"vegetable",
  		"plant",
  		"nature"
  	],
  	char: "🌵",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var christmas_tree$1 = {
  	keywords: [
  		"festival",
  		"vacation",
  		"december",
  		"xmas",
  		"celebration"
  	],
  	char: "🎄",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var evergreen_tree$1 = {
  	keywords: [
  		"plant",
  		"nature"
  	],
  	char: "🌲",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var deciduous_tree$1 = {
  	keywords: [
  		"plant",
  		"nature"
  	],
  	char: "🌳",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var palm_tree$1 = {
  	keywords: [
  		"plant",
  		"vegetable",
  		"nature",
  		"summer",
  		"beach",
  		"mojito",
  		"tropical"
  	],
  	char: "🌴",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var seedling$1 = {
  	keywords: [
  		"plant",
  		"nature",
  		"grass",
  		"lawn",
  		"spring"
  	],
  	char: "🌱",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var herb$1 = {
  	keywords: [
  		"vegetable",
  		"plant",
  		"medicine",
  		"weed",
  		"grass",
  		"lawn"
  	],
  	char: "🌿",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var shamrock$1 = {
  	keywords: [
  		"vegetable",
  		"plant",
  		"nature",
  		"irish",
  		"clover"
  	],
  	char: "☘",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var four_leaf_clover$1 = {
  	keywords: [
  		"vegetable",
  		"plant",
  		"nature",
  		"lucky",
  		"irish"
  	],
  	char: "🍀",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var bamboo$1 = {
  	keywords: [
  		"plant",
  		"nature",
  		"vegetable",
  		"panda",
  		"pine_decoration"
  	],
  	char: "🎍",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var tanabata_tree$1 = {
  	keywords: [
  		"plant",
  		"nature",
  		"branch",
  		"summer"
  	],
  	char: "🎋",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var leaves$1 = {
  	keywords: [
  		"nature",
  		"plant",
  		"tree",
  		"vegetable",
  		"grass",
  		"lawn",
  		"spring"
  	],
  	char: "🍃",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var fallen_leaf$1 = {
  	keywords: [
  		"nature",
  		"plant",
  		"vegetable",
  		"leaves"
  	],
  	char: "🍂",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var maple_leaf$1 = {
  	keywords: [
  		"nature",
  		"plant",
  		"vegetable",
  		"ca",
  		"fall"
  	],
  	char: "🍁",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var ear_of_rice$1 = {
  	keywords: [
  		"nature",
  		"plant"
  	],
  	char: "🌾",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var hibiscus$1 = {
  	keywords: [
  		"plant",
  		"vegetable",
  		"flowers",
  		"beach"
  	],
  	char: "🌺",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var sunflower$1 = {
  	keywords: [
  		"nature",
  		"plant",
  		"fall"
  	],
  	char: "🌻",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var rose$1 = {
  	keywords: [
  		"flowers",
  		"valentines",
  		"love",
  		"spring"
  	],
  	char: "🌹",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var wilted_flower$1 = {
  	keywords: [
  		"plant",
  		"nature",
  		"flower"
  	],
  	char: "🥀",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var tulip$1 = {
  	keywords: [
  		"flowers",
  		"plant",
  		"nature",
  		"summer",
  		"spring"
  	],
  	char: "🌷",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var blossom$1 = {
  	keywords: [
  		"nature",
  		"flowers",
  		"yellow"
  	],
  	char: "🌼",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var cherry_blossom$1 = {
  	keywords: [
  		"nature",
  		"plant",
  		"spring",
  		"flower"
  	],
  	char: "🌸",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var bouquet$1 = {
  	keywords: [
  		"flowers",
  		"nature",
  		"spring"
  	],
  	char: "💐",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var mushroom$1 = {
  	keywords: [
  		"plant",
  		"vegetable"
  	],
  	char: "🍄",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var chestnut$1 = {
  	keywords: [
  		"food",
  		"squirrel"
  	],
  	char: "🌰",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var jack_o_lantern$1 = {
  	keywords: [
  		"halloween",
  		"light",
  		"pumpkin",
  		"creepy",
  		"fall"
  	],
  	char: "🎃",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var shell$1 = {
  	keywords: [
  		"nature",
  		"sea",
  		"beach"
  	],
  	char: "🐚",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var spider_web$1 = {
  	keywords: [
  		"animal",
  		"insect",
  		"arachnid",
  		"silk"
  	],
  	char: "🕸",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var earth_americas$1 = {
  	keywords: [
  		"globe",
  		"world",
  		"USA",
  		"international"
  	],
  	char: "🌎",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var earth_africa$1 = {
  	keywords: [
  		"globe",
  		"world",
  		"international"
  	],
  	char: "🌍",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var earth_asia$1 = {
  	keywords: [
  		"globe",
  		"world",
  		"east",
  		"international"
  	],
  	char: "🌏",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var full_moon$1 = {
  	keywords: [
  		"nature",
  		"yellow",
  		"twilight",
  		"planet",
  		"space",
  		"night",
  		"evening",
  		"sleep"
  	],
  	char: "🌕",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var waning_gibbous_moon$1 = {
  	keywords: [
  		"nature",
  		"twilight",
  		"planet",
  		"space",
  		"night",
  		"evening",
  		"sleep",
  		"waxing_gibbous_moon"
  	],
  	char: "🌖",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var last_quarter_moon$1 = {
  	keywords: [
  		"nature",
  		"twilight",
  		"planet",
  		"space",
  		"night",
  		"evening",
  		"sleep"
  	],
  	char: "🌗",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var waning_crescent_moon$1 = {
  	keywords: [
  		"nature",
  		"twilight",
  		"planet",
  		"space",
  		"night",
  		"evening",
  		"sleep"
  	],
  	char: "🌘",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var new_moon$1 = {
  	keywords: [
  		"nature",
  		"twilight",
  		"planet",
  		"space",
  		"night",
  		"evening",
  		"sleep"
  	],
  	char: "🌑",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var waxing_crescent_moon$1 = {
  	keywords: [
  		"nature",
  		"twilight",
  		"planet",
  		"space",
  		"night",
  		"evening",
  		"sleep"
  	],
  	char: "🌒",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var first_quarter_moon$1 = {
  	keywords: [
  		"nature",
  		"twilight",
  		"planet",
  		"space",
  		"night",
  		"evening",
  		"sleep"
  	],
  	char: "🌓",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var waxing_gibbous_moon$1 = {
  	keywords: [
  		"nature",
  		"night",
  		"sky",
  		"gray",
  		"twilight",
  		"planet",
  		"space",
  		"evening",
  		"sleep"
  	],
  	char: "🌔",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var new_moon_with_face$1 = {
  	keywords: [
  		"nature",
  		"twilight",
  		"planet",
  		"space",
  		"night",
  		"evening",
  		"sleep"
  	],
  	char: "🌚",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var full_moon_with_face$1 = {
  	keywords: [
  		"nature",
  		"twilight",
  		"planet",
  		"space",
  		"night",
  		"evening",
  		"sleep"
  	],
  	char: "🌝",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var first_quarter_moon_with_face$1 = {
  	keywords: [
  		"nature",
  		"twilight",
  		"planet",
  		"space",
  		"night",
  		"evening",
  		"sleep"
  	],
  	char: "🌛",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var last_quarter_moon_with_face$1 = {
  	keywords: [
  		"nature",
  		"twilight",
  		"planet",
  		"space",
  		"night",
  		"evening",
  		"sleep"
  	],
  	char: "🌜",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var sun_with_face$1 = {
  	keywords: [
  		"nature",
  		"morning",
  		"sky"
  	],
  	char: "🌞",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var crescent_moon$1 = {
  	keywords: [
  		"night",
  		"sleep",
  		"sky",
  		"evening",
  		"magic"
  	],
  	char: "🌙",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var star$2 = {
  	keywords: [
  		"night",
  		"yellow"
  	],
  	char: "⭐",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var star2$1 = {
  	keywords: [
  		"night",
  		"sparkle",
  		"awesome",
  		"good",
  		"magic"
  	],
  	char: "🌟",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var dizzy$1 = {
  	keywords: [
  		"star",
  		"sparkle",
  		"shoot",
  		"magic"
  	],
  	char: "💫",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var sparkles$1 = {
  	keywords: [
  		"stars",
  		"shine",
  		"shiny",
  		"cool",
  		"awesome",
  		"good",
  		"magic"
  	],
  	char: "✨",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var comet$1 = {
  	keywords: [
  		"space"
  	],
  	char: "☄",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var sunny$1 = {
  	keywords: [
  		"weather",
  		"nature",
  		"brightness",
  		"summer",
  		"beach",
  		"spring"
  	],
  	char: "☀️",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var sun_behind_small_cloud$1 = {
  	keywords: [
  		"weather"
  	],
  	char: "🌤",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var partly_sunny$1 = {
  	keywords: [
  		"weather",
  		"nature",
  		"cloudy",
  		"morning",
  		"fall",
  		"spring"
  	],
  	char: "⛅",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var sun_behind_large_cloud$1 = {
  	keywords: [
  		"weather"
  	],
  	char: "🌥",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var sun_behind_rain_cloud$1 = {
  	keywords: [
  		"weather"
  	],
  	char: "🌦",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var cloud$1 = {
  	keywords: [
  		"weather",
  		"sky"
  	],
  	char: "☁️",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var cloud_with_rain$1 = {
  	keywords: [
  		"weather"
  	],
  	char: "🌧",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var cloud_with_lightning_and_rain$1 = {
  	keywords: [
  		"weather",
  		"lightning"
  	],
  	char: "⛈",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var cloud_with_lightning$1 = {
  	keywords: [
  		"weather",
  		"thunder"
  	],
  	char: "🌩",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var zap$1 = {
  	keywords: [
  		"thunder",
  		"weather",
  		"lightning bolt",
  		"fast"
  	],
  	char: "⚡",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var fire$1 = {
  	keywords: [
  		"hot",
  		"cook",
  		"flame"
  	],
  	char: "🔥",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var boom$1 = {
  	keywords: [
  		"bomb",
  		"explode",
  		"explosion",
  		"collision",
  		"blown"
  	],
  	char: "💥",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var snowflake$1 = {
  	keywords: [
  		"winter",
  		"season",
  		"cold",
  		"weather",
  		"christmas",
  		"xmas"
  	],
  	char: "❄️",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var cloud_with_snow$1 = {
  	keywords: [
  		"weather"
  	],
  	char: "🌨",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var snowman$1 = {
  	keywords: [
  		"winter",
  		"season",
  		"cold",
  		"weather",
  		"christmas",
  		"xmas",
  		"frozen",
  		"without_snow"
  	],
  	char: "⛄",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var snowman_with_snow$1 = {
  	keywords: [
  		"winter",
  		"season",
  		"cold",
  		"weather",
  		"christmas",
  		"xmas",
  		"frozen"
  	],
  	char: "☃",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var wind_face$1 = {
  	keywords: [
  		"gust",
  		"air"
  	],
  	char: "🌬",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var dash$2 = {
  	keywords: [
  		"wind",
  		"air",
  		"fast",
  		"shoo",
  		"fart",
  		"smoke",
  		"puff"
  	],
  	char: "💨",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var tornado$1 = {
  	keywords: [
  		"weather",
  		"cyclone",
  		"twister"
  	],
  	char: "🌪",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var fog$1 = {
  	keywords: [
  		"weather"
  	],
  	char: "🌫",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var open_umbrella$1 = {
  	keywords: [
  		"weather",
  		"spring"
  	],
  	char: "☂",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var umbrella$1 = {
  	keywords: [
  		"rainy",
  		"weather",
  		"spring"
  	],
  	char: "☔",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var droplet$1 = {
  	keywords: [
  		"water",
  		"drip",
  		"faucet",
  		"spring"
  	],
  	char: "💧",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var sweat_drops$1 = {
  	keywords: [
  		"water",
  		"drip",
  		"oops"
  	],
  	char: "💦",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var ocean$1 = {
  	keywords: [
  		"sea",
  		"water",
  		"wave",
  		"nature",
  		"tsunami",
  		"disaster"
  	],
  	char: "🌊",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  };
  var green_apple$1 = {
  	keywords: [
  		"fruit",
  		"nature"
  	],
  	char: "🍏",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var apple$1 = {
  	keywords: [
  		"fruit",
  		"mac",
  		"school"
  	],
  	char: "🍎",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var pear$1 = {
  	keywords: [
  		"fruit",
  		"nature",
  		"food"
  	],
  	char: "🍐",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var tangerine$1 = {
  	keywords: [
  		"food",
  		"fruit",
  		"nature",
  		"orange"
  	],
  	char: "🍊",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var lemon$1 = {
  	keywords: [
  		"fruit",
  		"nature"
  	],
  	char: "🍋",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var banana$1 = {
  	keywords: [
  		"fruit",
  		"food",
  		"monkey"
  	],
  	char: "🍌",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var watermelon$1 = {
  	keywords: [
  		"fruit",
  		"food",
  		"picnic",
  		"summer"
  	],
  	char: "🍉",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var grapes$1 = {
  	keywords: [
  		"fruit",
  		"food",
  		"wine"
  	],
  	char: "🍇",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var strawberry$1 = {
  	keywords: [
  		"fruit",
  		"food",
  		"nature"
  	],
  	char: "🍓",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var melon$1 = {
  	keywords: [
  		"fruit",
  		"nature",
  		"food"
  	],
  	char: "🍈",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var cherries$1 = {
  	keywords: [
  		"food",
  		"fruit"
  	],
  	char: "🍒",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var peach$1 = {
  	keywords: [
  		"fruit",
  		"nature",
  		"food"
  	],
  	char: "🍑",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var pineapple$1 = {
  	keywords: [
  		"fruit",
  		"nature",
  		"food"
  	],
  	char: "🍍",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var coconut$1 = {
  	keywords: [
  		"fruit",
  		"nature",
  		"food",
  		"palm"
  	],
  	char: "🥥",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var kiwi_fruit$1 = {
  	keywords: [
  		"fruit",
  		"food"
  	],
  	char: "🥝",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var mango$1 = {
  	keywords: [
  		"fruit",
  		"food",
  		"tropical"
  	],
  	char: "🥭",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var avocado$1 = {
  	keywords: [
  		"fruit",
  		"food"
  	],
  	char: "🥑",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var broccoli$1 = {
  	keywords: [
  		"fruit",
  		"food",
  		"vegetable"
  	],
  	char: "🥦",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var tomato$1 = {
  	keywords: [
  		"fruit",
  		"vegetable",
  		"nature",
  		"food"
  	],
  	char: "🍅",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var eggplant$1 = {
  	keywords: [
  		"vegetable",
  		"nature",
  		"food",
  		"aubergine"
  	],
  	char: "🍆",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var cucumber$1 = {
  	keywords: [
  		"fruit",
  		"food",
  		"pickle"
  	],
  	char: "🥒",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var carrot$1 = {
  	keywords: [
  		"vegetable",
  		"food",
  		"orange"
  	],
  	char: "🥕",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var hot_pepper$1 = {
  	keywords: [
  		"food",
  		"spicy",
  		"chilli",
  		"chili"
  	],
  	char: "🌶",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var potato$1 = {
  	keywords: [
  		"food",
  		"tuber",
  		"vegatable",
  		"starch"
  	],
  	char: "🥔",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var corn$1 = {
  	keywords: [
  		"food",
  		"vegetable",
  		"plant"
  	],
  	char: "🌽",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var leafy_greens = {
  	keywords: [
  		"food",
  		"vegetable",
  		"plant",
  		"bok choy",
  		"cabbage",
  		"kale",
  		"lettuce"
  	],
  	char: "🥬",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var sweet_potato$1 = {
  	keywords: [
  		"food",
  		"nature"
  	],
  	char: "🍠",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var peanuts$1 = {
  	keywords: [
  		"food",
  		"nut"
  	],
  	char: "🥜",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var honey_pot$1 = {
  	keywords: [
  		"bees",
  		"sweet",
  		"kitchen"
  	],
  	char: "🍯",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var croissant$1 = {
  	keywords: [
  		"food",
  		"bread",
  		"french"
  	],
  	char: "🥐",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var bread$1 = {
  	keywords: [
  		"food",
  		"wheat",
  		"breakfast",
  		"toast"
  	],
  	char: "🍞",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var baguette_bread$1 = {
  	keywords: [
  		"food",
  		"bread",
  		"french"
  	],
  	char: "🥖",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var bagel$1 = {
  	keywords: [
  		"food",
  		"bread",
  		"bakery",
  		"schmear"
  	],
  	char: "🥯",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var pretzel$1 = {
  	keywords: [
  		"food",
  		"bread",
  		"twisted"
  	],
  	char: "🥨",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var cheese$1 = {
  	keywords: [
  		"food",
  		"chadder"
  	],
  	char: "🧀",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var egg$1 = {
  	keywords: [
  		"food",
  		"chicken",
  		"breakfast"
  	],
  	char: "🥚",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var bacon$1 = {
  	keywords: [
  		"food",
  		"breakfast",
  		"pork",
  		"pig",
  		"meat"
  	],
  	char: "🥓",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var steak = {
  	keywords: [
  		"food",
  		"cow",
  		"meat",
  		"cut",
  		"chop",
  		"lambchop",
  		"porkchop"
  	],
  	char: "🥩",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var pancakes$1 = {
  	keywords: [
  		"food",
  		"breakfast",
  		"flapjacks",
  		"hotcakes"
  	],
  	char: "🥞",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var poultry_leg$1 = {
  	keywords: [
  		"food",
  		"meat",
  		"drumstick",
  		"bird",
  		"chicken",
  		"turkey"
  	],
  	char: "🍗",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var meat_on_bone$1 = {
  	keywords: [
  		"good",
  		"food",
  		"drumstick"
  	],
  	char: "🍖",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var bone$1 = {
  	keywords: [
  		"skeleton"
  	],
  	char: "🦴",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var fried_shrimp$1 = {
  	keywords: [
  		"food",
  		"animal",
  		"appetizer",
  		"summer"
  	],
  	char: "🍤",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var fried_egg$1 = {
  	keywords: [
  		"food",
  		"breakfast",
  		"kitchen",
  		"egg"
  	],
  	char: "🍳",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var hamburger$1 = {
  	keywords: [
  		"meat",
  		"fast food",
  		"beef",
  		"cheeseburger",
  		"mcdonalds",
  		"burger king"
  	],
  	char: "🍔",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var fries$1 = {
  	keywords: [
  		"chips",
  		"snack",
  		"fast food"
  	],
  	char: "🍟",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var stuffed_flatbread$1 = {
  	keywords: [
  		"food",
  		"flatbread",
  		"stuffed",
  		"gyro"
  	],
  	char: "🥙",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var hotdog$1 = {
  	keywords: [
  		"food",
  		"frankfurter"
  	],
  	char: "🌭",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var pizza$1 = {
  	keywords: [
  		"food",
  		"party"
  	],
  	char: "🍕",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var sandwich$1 = {
  	keywords: [
  		"food",
  		"lunch",
  		"bread"
  	],
  	char: "🥪",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var canned_food$1 = {
  	keywords: [
  		"food",
  		"soup"
  	],
  	char: "🥫",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var spaghetti$1 = {
  	keywords: [
  		"food",
  		"italian",
  		"noodle"
  	],
  	char: "🍝",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var taco$1 = {
  	keywords: [
  		"food",
  		"mexican"
  	],
  	char: "🌮",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var burrito$1 = {
  	keywords: [
  		"food",
  		"mexican"
  	],
  	char: "🌯",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var green_salad$1 = {
  	keywords: [
  		"food",
  		"healthy",
  		"lettuce"
  	],
  	char: "🥗",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var shallow_pan_of_food$1 = {
  	keywords: [
  		"food",
  		"cooking",
  		"casserole",
  		"paella"
  	],
  	char: "🥘",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var ramen$1 = {
  	keywords: [
  		"food",
  		"japanese",
  		"noodle",
  		"chopsticks"
  	],
  	char: "🍜",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var stew$1 = {
  	keywords: [
  		"food",
  		"meat",
  		"soup"
  	],
  	char: "🍲",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var fish_cake$1 = {
  	keywords: [
  		"food",
  		"japan",
  		"sea",
  		"beach",
  		"narutomaki",
  		"pink",
  		"swirl",
  		"kamaboko",
  		"surimi",
  		"ramen"
  	],
  	char: "🍥",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var fortune_cookie$1 = {
  	keywords: [
  		"food",
  		"prophecy"
  	],
  	char: "🥠",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var sushi$1 = {
  	keywords: [
  		"food",
  		"fish",
  		"japanese",
  		"rice"
  	],
  	char: "🍣",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var bento$1 = {
  	keywords: [
  		"food",
  		"japanese",
  		"box"
  	],
  	char: "🍱",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var curry$1 = {
  	keywords: [
  		"food",
  		"spicy",
  		"hot",
  		"indian"
  	],
  	char: "🍛",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var rice_ball$1 = {
  	keywords: [
  		"food",
  		"japanese"
  	],
  	char: "🍙",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var rice$1 = {
  	keywords: [
  		"food",
  		"china",
  		"asian"
  	],
  	char: "🍚",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var rice_cracker$1 = {
  	keywords: [
  		"food",
  		"japanese"
  	],
  	char: "🍘",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var oden$1 = {
  	keywords: [
  		"food",
  		"japanese"
  	],
  	char: "🍢",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var dango$1 = {
  	keywords: [
  		"food",
  		"dessert",
  		"sweet",
  		"japanese",
  		"barbecue",
  		"meat"
  	],
  	char: "🍡",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var shaved_ice$1 = {
  	keywords: [
  		"hot",
  		"dessert",
  		"summer"
  	],
  	char: "🍧",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var ice_cream$1 = {
  	keywords: [
  		"food",
  		"hot",
  		"dessert"
  	],
  	char: "🍨",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var icecream$1 = {
  	keywords: [
  		"food",
  		"hot",
  		"dessert",
  		"summer"
  	],
  	char: "🍦",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var pie$1 = {
  	keywords: [
  		"food",
  		"dessert",
  		"pastry"
  	],
  	char: "🥧",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var cake$1 = {
  	keywords: [
  		"food",
  		"dessert"
  	],
  	char: "🍰",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var cupcake$1 = {
  	keywords: [
  		"food",
  		"dessert",
  		"bakery",
  		"sweet"
  	],
  	char: "🧁",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var moon_cake$1 = {
  	keywords: [
  		"food",
  		"autumn"
  	],
  	char: "🥮",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var birthday$1 = {
  	keywords: [
  		"food",
  		"dessert",
  		"cake"
  	],
  	char: "🎂",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var custard$1 = {
  	keywords: [
  		"dessert",
  		"food"
  	],
  	char: "🍮",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var candy$1 = {
  	keywords: [
  		"snack",
  		"dessert",
  		"sweet",
  		"lolly"
  	],
  	char: "🍬",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var lollipop$1 = {
  	keywords: [
  		"food",
  		"snack",
  		"candy",
  		"sweet"
  	],
  	char: "🍭",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var chocolate_bar$1 = {
  	keywords: [
  		"food",
  		"snack",
  		"dessert",
  		"sweet"
  	],
  	char: "🍫",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var popcorn$1 = {
  	keywords: [
  		"food",
  		"movie theater",
  		"films",
  		"snack"
  	],
  	char: "🍿",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var dumpling$1 = {
  	keywords: [
  		"food",
  		"empanada",
  		"pierogi",
  		"potsticker"
  	],
  	char: "🥟",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var doughnut$1 = {
  	keywords: [
  		"food",
  		"dessert",
  		"snack",
  		"sweet",
  		"donut"
  	],
  	char: "🍩",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var cookie$1 = {
  	keywords: [
  		"food",
  		"snack",
  		"oreo",
  		"chocolate",
  		"sweet",
  		"dessert"
  	],
  	char: "🍪",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var milk_glass$1 = {
  	keywords: [
  		"beverage",
  		"drink",
  		"cow"
  	],
  	char: "🥛",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var beer$1 = {
  	keywords: [
  		"relax",
  		"beverage",
  		"drink",
  		"drunk",
  		"party",
  		"pub",
  		"summer",
  		"alcohol",
  		"booze"
  	],
  	char: "🍺",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var beers$1 = {
  	keywords: [
  		"relax",
  		"beverage",
  		"drink",
  		"drunk",
  		"party",
  		"pub",
  		"summer",
  		"alcohol",
  		"booze"
  	],
  	char: "🍻",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var clinking_glasses$1 = {
  	keywords: [
  		"beverage",
  		"drink",
  		"party",
  		"alcohol",
  		"celebrate",
  		"cheers",
  		"wine",
  		"champagne",
  		"toast"
  	],
  	char: "🥂",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var wine_glass$1 = {
  	keywords: [
  		"drink",
  		"beverage",
  		"drunk",
  		"alcohol",
  		"booze"
  	],
  	char: "🍷",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var tumbler_glass$1 = {
  	keywords: [
  		"drink",
  		"beverage",
  		"drunk",
  		"alcohol",
  		"liquor",
  		"booze",
  		"bourbon",
  		"scotch",
  		"whisky",
  		"glass",
  		"shot"
  	],
  	char: "🥃",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var cocktail$1 = {
  	keywords: [
  		"drink",
  		"drunk",
  		"alcohol",
  		"beverage",
  		"booze",
  		"mojito"
  	],
  	char: "🍸",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var tropical_drink$1 = {
  	keywords: [
  		"beverage",
  		"cocktail",
  		"summer",
  		"beach",
  		"alcohol",
  		"booze",
  		"mojito"
  	],
  	char: "🍹",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var champagne$1 = {
  	keywords: [
  		"drink",
  		"wine",
  		"bottle",
  		"celebration"
  	],
  	char: "🍾",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var sake$1 = {
  	keywords: [
  		"wine",
  		"drink",
  		"drunk",
  		"beverage",
  		"japanese",
  		"alcohol",
  		"booze"
  	],
  	char: "🍶",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var tea$1 = {
  	keywords: [
  		"drink",
  		"bowl",
  		"breakfast",
  		"green",
  		"british"
  	],
  	char: "🍵",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var cup_with_straw$1 = {
  	keywords: [
  		"drink",
  		"soda"
  	],
  	char: "🥤",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var coffee$1 = {
  	keywords: [
  		"beverage",
  		"caffeine",
  		"latte",
  		"espresso"
  	],
  	char: "☕",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var baby_bottle$1 = {
  	keywords: [
  		"food",
  		"container",
  		"milk"
  	],
  	char: "🍼",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var salt$1 = {
  	keywords: [
  		"condiment",
  		"shaker"
  	],
  	char: "🧂",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var spoon$1 = {
  	keywords: [
  		"cutlery",
  		"kitchen",
  		"tableware"
  	],
  	char: "🥄",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var fork_and_knife$1 = {
  	keywords: [
  		"cutlery",
  		"kitchen"
  	],
  	char: "🍴",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var plate_with_cutlery$1 = {
  	keywords: [
  		"food",
  		"eat",
  		"meal",
  		"lunch",
  		"dinner",
  		"restaurant"
  	],
  	char: "🍽",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var bowl_with_spoon$1 = {
  	keywords: [
  		"food",
  		"breakfast",
  		"cereal",
  		"oatmeal",
  		"porridge"
  	],
  	char: "🥣",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var takeout_box$1 = {
  	keywords: [
  		"food",
  		"leftovers"
  	],
  	char: "🥡",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var chopsticks$1 = {
  	keywords: [
  		"food"
  	],
  	char: "🥢",
  	fitzpatrick_scale: false,
  	category: "food_and_drink"
  };
  var soccer$1 = {
  	keywords: [
  		"sports",
  		"football"
  	],
  	char: "⚽",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var basketball$1 = {
  	keywords: [
  		"sports",
  		"balls",
  		"NBA"
  	],
  	char: "🏀",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var football$1 = {
  	keywords: [
  		"sports",
  		"balls",
  		"NFL"
  	],
  	char: "🏈",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var baseball$1 = {
  	keywords: [
  		"sports",
  		"balls"
  	],
  	char: "⚾",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var softball$1 = {
  	keywords: [
  		"sports",
  		"balls"
  	],
  	char: "🥎",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var tennis$1 = {
  	keywords: [
  		"sports",
  		"balls",
  		"green"
  	],
  	char: "🎾",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var volleyball$1 = {
  	keywords: [
  		"sports",
  		"balls"
  	],
  	char: "🏐",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var rugby_football$1 = {
  	keywords: [
  		"sports",
  		"team"
  	],
  	char: "🏉",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var flying_disc$1 = {
  	keywords: [
  		"sports",
  		"frisbee",
  		"ultimate"
  	],
  	char: "🥏",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var golf$1 = {
  	keywords: [
  		"sports",
  		"business",
  		"flag",
  		"hole",
  		"summer"
  	],
  	char: "⛳",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var golfing_woman$1 = {
  	keywords: [
  		"sports",
  		"business",
  		"woman",
  		"female"
  	],
  	char: "🏌️‍♀️",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var golfing_man$1 = {
  	keywords: [
  		"sports",
  		"business"
  	],
  	char: "🏌",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var ping_pong$1 = {
  	keywords: [
  		"sports",
  		"pingpong"
  	],
  	char: "🏓",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var badminton$1 = {
  	keywords: [
  		"sports"
  	],
  	char: "🏸",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var goal_net$1 = {
  	keywords: [
  		"sports"
  	],
  	char: "🥅",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var ice_hockey$1 = {
  	keywords: [
  		"sports"
  	],
  	char: "🏒",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var field_hockey$1 = {
  	keywords: [
  		"sports"
  	],
  	char: "🏑",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var lacrosse$1 = {
  	keywords: [
  		"sports",
  		"ball",
  		"stick"
  	],
  	char: "🥍",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var cricket$1 = {
  	keywords: [
  		"sports"
  	],
  	char: "🏏",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var ski$1 = {
  	keywords: [
  		"sports",
  		"winter",
  		"cold",
  		"snow"
  	],
  	char: "🎿",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var skier$1 = {
  	keywords: [
  		"sports",
  		"winter",
  		"snow"
  	],
  	char: "⛷",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var snowboarder$1 = {
  	keywords: [
  		"sports",
  		"winter"
  	],
  	char: "🏂",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var person_fencing$1 = {
  	keywords: [
  		"sports",
  		"fencing",
  		"sword"
  	],
  	char: "🤺",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var women_wrestling$1 = {
  	keywords: [
  		"sports",
  		"wrestlers"
  	],
  	char: "🤼‍♀️",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var men_wrestling$1 = {
  	keywords: [
  		"sports",
  		"wrestlers"
  	],
  	char: "🤼‍♂️",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var woman_cartwheeling$1 = {
  	keywords: [
  		"gymnastics"
  	],
  	char: "🤸‍♀️",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var man_cartwheeling$1 = {
  	keywords: [
  		"gymnastics"
  	],
  	char: "🤸‍♂️",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var woman_playing_handball$1 = {
  	keywords: [
  		"sports"
  	],
  	char: "🤾‍♀️",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var man_playing_handball$1 = {
  	keywords: [
  		"sports"
  	],
  	char: "🤾‍♂️",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var ice_skate$1 = {
  	keywords: [
  		"sports"
  	],
  	char: "⛸",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var curling_stone$1 = {
  	keywords: [
  		"sports"
  	],
  	char: "🥌",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var skateboard$1 = {
  	keywords: [
  		"board"
  	],
  	char: "🛹",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var sled$1 = {
  	keywords: [
  		"sleigh",
  		"luge",
  		"toboggan"
  	],
  	char: "🛷",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var bow_and_arrow$1 = {
  	keywords: [
  		"sports"
  	],
  	char: "🏹",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var fishing_pole_and_fish$1 = {
  	keywords: [
  		"food",
  		"hobby",
  		"summer"
  	],
  	char: "🎣",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var boxing_glove$1 = {
  	keywords: [
  		"sports",
  		"fighting"
  	],
  	char: "🥊",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var martial_arts_uniform$1 = {
  	keywords: [
  		"judo",
  		"karate",
  		"taekwondo"
  	],
  	char: "🥋",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var rowing_woman$1 = {
  	keywords: [
  		"sports",
  		"hobby",
  		"water",
  		"ship",
  		"woman",
  		"female"
  	],
  	char: "🚣‍♀️",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var rowing_man$1 = {
  	keywords: [
  		"sports",
  		"hobby",
  		"water",
  		"ship"
  	],
  	char: "🚣",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var climbing_woman$1 = {
  	keywords: [
  		"sports",
  		"hobby",
  		"woman",
  		"female",
  		"rock"
  	],
  	char: "🧗‍♀️",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var climbing_man$1 = {
  	keywords: [
  		"sports",
  		"hobby",
  		"man",
  		"male",
  		"rock"
  	],
  	char: "🧗‍♂️",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var swimming_woman$1 = {
  	keywords: [
  		"sports",
  		"exercise",
  		"human",
  		"athlete",
  		"water",
  		"summer",
  		"woman",
  		"female"
  	],
  	char: "🏊‍♀️",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var swimming_man$1 = {
  	keywords: [
  		"sports",
  		"exercise",
  		"human",
  		"athlete",
  		"water",
  		"summer"
  	],
  	char: "🏊",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var woman_playing_water_polo$1 = {
  	keywords: [
  		"sports",
  		"pool"
  	],
  	char: "🤽‍♀️",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var man_playing_water_polo$1 = {
  	keywords: [
  		"sports",
  		"pool"
  	],
  	char: "🤽‍♂️",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var woman_in_lotus_position = {
  	keywords: [
  		"woman",
  		"female",
  		"meditation",
  		"yoga",
  		"serenity",
  		"zen",
  		"mindfulness"
  	],
  	char: "🧘‍♀️",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var man_in_lotus_position = {
  	keywords: [
  		"man",
  		"male",
  		"meditation",
  		"yoga",
  		"serenity",
  		"zen",
  		"mindfulness"
  	],
  	char: "🧘‍♂️",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var surfing_woman$1 = {
  	keywords: [
  		"sports",
  		"ocean",
  		"sea",
  		"summer",
  		"beach",
  		"woman",
  		"female"
  	],
  	char: "🏄‍♀️",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var surfing_man$1 = {
  	keywords: [
  		"sports",
  		"ocean",
  		"sea",
  		"summer",
  		"beach"
  	],
  	char: "🏄",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var bath$1 = {
  	keywords: [
  		"clean",
  		"shower",
  		"bathroom"
  	],
  	char: "🛀",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var basketball_woman$1 = {
  	keywords: [
  		"sports",
  		"human",
  		"woman",
  		"female"
  	],
  	char: "⛹️‍♀️",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var basketball_man$1 = {
  	keywords: [
  		"sports",
  		"human"
  	],
  	char: "⛹",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var weight_lifting_woman$1 = {
  	keywords: [
  		"sports",
  		"training",
  		"exercise",
  		"woman",
  		"female"
  	],
  	char: "🏋️‍♀️",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var weight_lifting_man$1 = {
  	keywords: [
  		"sports",
  		"training",
  		"exercise"
  	],
  	char: "🏋",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var biking_woman$1 = {
  	keywords: [
  		"sports",
  		"bike",
  		"exercise",
  		"hipster",
  		"woman",
  		"female"
  	],
  	char: "🚴‍♀️",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var biking_man$1 = {
  	keywords: [
  		"sports",
  		"bike",
  		"exercise",
  		"hipster"
  	],
  	char: "🚴",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var mountain_biking_woman$1 = {
  	keywords: [
  		"transportation",
  		"sports",
  		"human",
  		"race",
  		"bike",
  		"woman",
  		"female"
  	],
  	char: "🚵‍♀️",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var mountain_biking_man$1 = {
  	keywords: [
  		"transportation",
  		"sports",
  		"human",
  		"race",
  		"bike"
  	],
  	char: "🚵",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var horse_racing$1 = {
  	keywords: [
  		"animal",
  		"betting",
  		"competition",
  		"gambling",
  		"luck"
  	],
  	char: "🏇",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var business_suit_levitating$1 = {
  	keywords: [
  		"suit",
  		"business",
  		"levitate",
  		"hover",
  		"jump"
  	],
  	char: "🕴",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var trophy$1 = {
  	keywords: [
  		"win",
  		"award",
  		"contest",
  		"place",
  		"ftw",
  		"ceremony"
  	],
  	char: "🏆",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var running_shirt_with_sash$1 = {
  	keywords: [
  		"play",
  		"pageant"
  	],
  	char: "🎽",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var medal_sports$1 = {
  	keywords: [
  		"award",
  		"winning"
  	],
  	char: "🏅",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var medal_military$1 = {
  	keywords: [
  		"award",
  		"winning",
  		"army"
  	],
  	char: "🎖",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var reminder_ribbon$1 = {
  	keywords: [
  		"sports",
  		"cause",
  		"support",
  		"awareness"
  	],
  	char: "🎗",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var rosette$1 = {
  	keywords: [
  		"flower",
  		"decoration",
  		"military"
  	],
  	char: "🏵",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var ticket$1 = {
  	keywords: [
  		"event",
  		"concert",
  		"pass"
  	],
  	char: "🎫",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var tickets$1 = {
  	keywords: [
  		"sports",
  		"concert",
  		"entrance"
  	],
  	char: "🎟",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var performing_arts$1 = {
  	keywords: [
  		"acting",
  		"theater",
  		"drama"
  	],
  	char: "🎭",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var art$1 = {
  	keywords: [
  		"design",
  		"paint",
  		"draw",
  		"colors"
  	],
  	char: "🎨",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var circus_tent$1 = {
  	keywords: [
  		"festival",
  		"carnival",
  		"party"
  	],
  	char: "🎪",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var woman_juggling$1 = {
  	keywords: [
  		"juggle",
  		"balance",
  		"skill",
  		"multitask"
  	],
  	char: "🤹‍♀️",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var man_juggling$1 = {
  	keywords: [
  		"juggle",
  		"balance",
  		"skill",
  		"multitask"
  	],
  	char: "🤹‍♂️",
  	fitzpatrick_scale: true,
  	category: "activity"
  };
  var microphone$1 = {
  	keywords: [
  		"sound",
  		"music",
  		"PA",
  		"sing",
  		"talkshow"
  	],
  	char: "🎤",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var headphones$1 = {
  	keywords: [
  		"music",
  		"score",
  		"gadgets"
  	],
  	char: "🎧",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var musical_score$1 = {
  	keywords: [
  		"treble",
  		"clef",
  		"compose"
  	],
  	char: "🎼",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var musical_keyboard$1 = {
  	keywords: [
  		"piano",
  		"instrument",
  		"compose"
  	],
  	char: "🎹",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var drum$1 = {
  	keywords: [
  		"music",
  		"instrument",
  		"drumsticks",
  		"snare"
  	],
  	char: "🥁",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var saxophone$1 = {
  	keywords: [
  		"music",
  		"instrument",
  		"jazz",
  		"blues"
  	],
  	char: "🎷",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var trumpet$1 = {
  	keywords: [
  		"music",
  		"brass"
  	],
  	char: "🎺",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var guitar$1 = {
  	keywords: [
  		"music",
  		"instrument"
  	],
  	char: "🎸",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var violin$1 = {
  	keywords: [
  		"music",
  		"instrument",
  		"orchestra",
  		"symphony"
  	],
  	char: "🎻",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var clapper$1 = {
  	keywords: [
  		"movie",
  		"film",
  		"record"
  	],
  	char: "🎬",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var video_game$1 = {
  	keywords: [
  		"play",
  		"console",
  		"PS4",
  		"controller"
  	],
  	char: "🎮",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var space_invader$1 = {
  	keywords: [
  		"game",
  		"arcade",
  		"play"
  	],
  	char: "👾",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var dart$1 = {
  	keywords: [
  		"game",
  		"play",
  		"bar",
  		"target",
  		"bullseye"
  	],
  	char: "🎯",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var game_die$1 = {
  	keywords: [
  		"dice",
  		"random",
  		"tabletop",
  		"play",
  		"luck"
  	],
  	char: "🎲",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var chess_pawn$1 = {
  	keywords: [
  		"expendable"
  	],
  	char: "♟",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var slot_machine$1 = {
  	keywords: [
  		"bet",
  		"gamble",
  		"vegas",
  		"fruit machine",
  		"luck",
  		"casino"
  	],
  	char: "🎰",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var jigsaw$1 = {
  	keywords: [
  		"interlocking",
  		"puzzle",
  		"piece"
  	],
  	char: "🧩",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var bowling$1 = {
  	keywords: [
  		"sports",
  		"fun",
  		"play"
  	],
  	char: "🎳",
  	fitzpatrick_scale: false,
  	category: "activity"
  };
  var red_car$1 = {
  	keywords: [
  		"red",
  		"transportation",
  		"vehicle"
  	],
  	char: "🚗",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var taxi$1 = {
  	keywords: [
  		"uber",
  		"vehicle",
  		"cars",
  		"transportation"
  	],
  	char: "🚕",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var blue_car$1 = {
  	keywords: [
  		"transportation",
  		"vehicle"
  	],
  	char: "🚙",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var bus$1 = {
  	keywords: [
  		"car",
  		"vehicle",
  		"transportation"
  	],
  	char: "🚌",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var trolleybus$1 = {
  	keywords: [
  		"bart",
  		"transportation",
  		"vehicle"
  	],
  	char: "🚎",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var racing_car$1 = {
  	keywords: [
  		"sports",
  		"race",
  		"fast",
  		"formula",
  		"f1"
  	],
  	char: "🏎",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var police_car$1 = {
  	keywords: [
  		"vehicle",
  		"cars",
  		"transportation",
  		"law",
  		"legal",
  		"enforcement"
  	],
  	char: "🚓",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var ambulance$1 = {
  	keywords: [
  		"health",
  		"911",
  		"hospital"
  	],
  	char: "🚑",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var fire_engine$1 = {
  	keywords: [
  		"transportation",
  		"cars",
  		"vehicle"
  	],
  	char: "🚒",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var minibus$1 = {
  	keywords: [
  		"vehicle",
  		"car",
  		"transportation"
  	],
  	char: "🚐",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var truck$1 = {
  	keywords: [
  		"cars",
  		"transportation"
  	],
  	char: "🚚",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var articulated_lorry$1 = {
  	keywords: [
  		"vehicle",
  		"cars",
  		"transportation",
  		"express"
  	],
  	char: "🚛",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var tractor$1 = {
  	keywords: [
  		"vehicle",
  		"car",
  		"farming",
  		"agriculture"
  	],
  	char: "🚜",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var kick_scooter$1 = {
  	keywords: [
  		"vehicle",
  		"kick",
  		"razor"
  	],
  	char: "🛴",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var motorcycle$1 = {
  	keywords: [
  		"race",
  		"sports",
  		"fast"
  	],
  	char: "🏍",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var bike$1 = {
  	keywords: [
  		"sports",
  		"bicycle",
  		"exercise",
  		"hipster"
  	],
  	char: "🚲",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var motor_scooter$1 = {
  	keywords: [
  		"vehicle",
  		"vespa",
  		"sasha"
  	],
  	char: "🛵",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var rotating_light$1 = {
  	keywords: [
  		"police",
  		"ambulance",
  		"911",
  		"emergency",
  		"alert",
  		"error",
  		"pinged",
  		"law",
  		"legal"
  	],
  	char: "🚨",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var oncoming_police_car$1 = {
  	keywords: [
  		"vehicle",
  		"law",
  		"legal",
  		"enforcement",
  		"911"
  	],
  	char: "🚔",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var oncoming_bus$1 = {
  	keywords: [
  		"vehicle",
  		"transportation"
  	],
  	char: "🚍",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var oncoming_automobile$1 = {
  	keywords: [
  		"car",
  		"vehicle",
  		"transportation"
  	],
  	char: "🚘",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var oncoming_taxi$1 = {
  	keywords: [
  		"vehicle",
  		"cars",
  		"uber"
  	],
  	char: "🚖",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var aerial_tramway$1 = {
  	keywords: [
  		"transportation",
  		"vehicle",
  		"ski"
  	],
  	char: "🚡",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var mountain_cableway$1 = {
  	keywords: [
  		"transportation",
  		"vehicle",
  		"ski"
  	],
  	char: "🚠",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var suspension_railway$1 = {
  	keywords: [
  		"vehicle",
  		"transportation"
  	],
  	char: "🚟",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var railway_car$1 = {
  	keywords: [
  		"transportation",
  		"vehicle"
  	],
  	char: "🚃",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var train$1 = {
  	keywords: [
  		"transportation",
  		"vehicle",
  		"carriage",
  		"public",
  		"travel"
  	],
  	char: "🚋",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var monorail$1 = {
  	keywords: [
  		"transportation",
  		"vehicle"
  	],
  	char: "🚝",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var bullettrain_side$1 = {
  	keywords: [
  		"transportation",
  		"vehicle"
  	],
  	char: "🚄",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var bullettrain_front$1 = {
  	keywords: [
  		"transportation",
  		"vehicle",
  		"speed",
  		"fast",
  		"public",
  		"travel"
  	],
  	char: "🚅",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var light_rail$1 = {
  	keywords: [
  		"transportation",
  		"vehicle"
  	],
  	char: "🚈",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var mountain_railway$1 = {
  	keywords: [
  		"transportation",
  		"vehicle"
  	],
  	char: "🚞",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var steam_locomotive$1 = {
  	keywords: [
  		"transportation",
  		"vehicle",
  		"train"
  	],
  	char: "🚂",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var train2$1 = {
  	keywords: [
  		"transportation",
  		"vehicle"
  	],
  	char: "🚆",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var metro$1 = {
  	keywords: [
  		"transportation",
  		"blue-square",
  		"mrt",
  		"underground",
  		"tube"
  	],
  	char: "🚇",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var tram$1 = {
  	keywords: [
  		"transportation",
  		"vehicle"
  	],
  	char: "🚊",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var station$1 = {
  	keywords: [
  		"transportation",
  		"vehicle",
  		"public"
  	],
  	char: "🚉",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var flying_saucer$1 = {
  	keywords: [
  		"transportation",
  		"vehicle",
  		"ufo"
  	],
  	char: "🛸",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var helicopter$1 = {
  	keywords: [
  		"transportation",
  		"vehicle",
  		"fly"
  	],
  	char: "🚁",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var small_airplane$1 = {
  	keywords: [
  		"flight",
  		"transportation",
  		"fly",
  		"vehicle"
  	],
  	char: "🛩",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var airplane$1 = {
  	keywords: [
  		"vehicle",
  		"transportation",
  		"flight",
  		"fly"
  	],
  	char: "✈️",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var flight_departure$1 = {
  	keywords: [
  		"airport",
  		"flight",
  		"landing"
  	],
  	char: "🛫",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var flight_arrival$1 = {
  	keywords: [
  		"airport",
  		"flight",
  		"boarding"
  	],
  	char: "🛬",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var sailboat$1 = {
  	keywords: [
  		"ship",
  		"summer",
  		"transportation",
  		"water",
  		"sailing"
  	],
  	char: "⛵",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var motor_boat$1 = {
  	keywords: [
  		"ship"
  	],
  	char: "🛥",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var speedboat$1 = {
  	keywords: [
  		"ship",
  		"transportation",
  		"vehicle",
  		"summer"
  	],
  	char: "🚤",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var ferry$1 = {
  	keywords: [
  		"boat",
  		"ship",
  		"yacht"
  	],
  	char: "⛴",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var passenger_ship$1 = {
  	keywords: [
  		"yacht",
  		"cruise",
  		"ferry"
  	],
  	char: "🛳",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var rocket$1 = {
  	keywords: [
  		"launch",
  		"ship",
  		"staffmode",
  		"NASA",
  		"outer space",
  		"outer_space",
  		"fly"
  	],
  	char: "🚀",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var artificial_satellite$1 = {
  	keywords: [
  		"communication",
  		"gps",
  		"orbit",
  		"spaceflight",
  		"NASA",
  		"ISS"
  	],
  	char: "🛰",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var seat$1 = {
  	keywords: [
  		"sit",
  		"airplane",
  		"transport",
  		"bus",
  		"flight",
  		"fly"
  	],
  	char: "💺",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var canoe$1 = {
  	keywords: [
  		"boat",
  		"paddle",
  		"water",
  		"ship"
  	],
  	char: "🛶",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var anchor$1 = {
  	keywords: [
  		"ship",
  		"ferry",
  		"sea",
  		"boat"
  	],
  	char: "⚓",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var construction$1 = {
  	keywords: [
  		"wip",
  		"progress",
  		"caution",
  		"warning"
  	],
  	char: "🚧",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var fuelpump$1 = {
  	keywords: [
  		"gas station",
  		"petroleum"
  	],
  	char: "⛽",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var busstop$1 = {
  	keywords: [
  		"transportation",
  		"wait"
  	],
  	char: "🚏",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var vertical_traffic_light$1 = {
  	keywords: [
  		"transportation",
  		"driving"
  	],
  	char: "🚦",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var traffic_light$1 = {
  	keywords: [
  		"transportation",
  		"signal"
  	],
  	char: "🚥",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var checkered_flag$1 = {
  	keywords: [
  		"contest",
  		"finishline",
  		"race",
  		"gokart"
  	],
  	char: "🏁",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var ship$1 = {
  	keywords: [
  		"transportation",
  		"titanic",
  		"deploy"
  	],
  	char: "🚢",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var ferris_wheel$1 = {
  	keywords: [
  		"photo",
  		"carnival",
  		"londoneye"
  	],
  	char: "🎡",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var roller_coaster$1 = {
  	keywords: [
  		"carnival",
  		"playground",
  		"photo",
  		"fun"
  	],
  	char: "🎢",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var carousel_horse$1 = {
  	keywords: [
  		"photo",
  		"carnival"
  	],
  	char: "🎠",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var building_construction$1 = {
  	keywords: [
  		"wip",
  		"working",
  		"progress"
  	],
  	char: "🏗",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var foggy$1 = {
  	keywords: [
  		"photo",
  		"mountain"
  	],
  	char: "🌁",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var tokyo_tower$1 = {
  	keywords: [
  		"photo",
  		"japanese"
  	],
  	char: "🗼",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var factory$1 = {
  	keywords: [
  		"building",
  		"industry",
  		"pollution",
  		"smoke"
  	],
  	char: "🏭",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var fountain$1 = {
  	keywords: [
  		"photo",
  		"summer",
  		"water",
  		"fresh"
  	],
  	char: "⛲",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var rice_scene$1 = {
  	keywords: [
  		"photo",
  		"japan",
  		"asia",
  		"tsukimi"
  	],
  	char: "🎑",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var mountain$1 = {
  	keywords: [
  		"photo",
  		"nature",
  		"environment"
  	],
  	char: "⛰",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var mountain_snow$1 = {
  	keywords: [
  		"photo",
  		"nature",
  		"environment",
  		"winter",
  		"cold"
  	],
  	char: "🏔",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var mount_fuji$1 = {
  	keywords: [
  		"photo",
  		"mountain",
  		"nature",
  		"japanese"
  	],
  	char: "🗻",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var volcano$1 = {
  	keywords: [
  		"photo",
  		"nature",
  		"disaster"
  	],
  	char: "🌋",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var japan$1 = {
  	keywords: [
  		"nation",
  		"country",
  		"japanese",
  		"asia"
  	],
  	char: "🗾",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var camping$1 = {
  	keywords: [
  		"photo",
  		"outdoors",
  		"tent"
  	],
  	char: "🏕",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var tent$1 = {
  	keywords: [
  		"photo",
  		"camping",
  		"outdoors"
  	],
  	char: "⛺",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var national_park$1 = {
  	keywords: [
  		"photo",
  		"environment",
  		"nature"
  	],
  	char: "🏞",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var motorway$1 = {
  	keywords: [
  		"road",
  		"cupertino",
  		"interstate",
  		"highway"
  	],
  	char: "🛣",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var railway_track$1 = {
  	keywords: [
  		"train",
  		"transportation"
  	],
  	char: "🛤",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var sunrise$1 = {
  	keywords: [
  		"morning",
  		"view",
  		"vacation",
  		"photo"
  	],
  	char: "🌅",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var sunrise_over_mountains$1 = {
  	keywords: [
  		"view",
  		"vacation",
  		"photo"
  	],
  	char: "🌄",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var desert$1 = {
  	keywords: [
  		"photo",
  		"warm",
  		"saharah"
  	],
  	char: "🏜",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var beach_umbrella$1 = {
  	keywords: [
  		"weather",
  		"summer",
  		"sunny",
  		"sand",
  		"mojito"
  	],
  	char: "🏖",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var desert_island$1 = {
  	keywords: [
  		"photo",
  		"tropical",
  		"mojito"
  	],
  	char: "🏝",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var city_sunrise$1 = {
  	keywords: [
  		"photo",
  		"good morning",
  		"dawn"
  	],
  	char: "🌇",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var city_sunset$1 = {
  	keywords: [
  		"photo",
  		"evening",
  		"sky",
  		"buildings"
  	],
  	char: "🌆",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var cityscape$1 = {
  	keywords: [
  		"photo",
  		"night life",
  		"urban"
  	],
  	char: "🏙",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var night_with_stars$1 = {
  	keywords: [
  		"evening",
  		"city",
  		"downtown"
  	],
  	char: "🌃",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var bridge_at_night$1 = {
  	keywords: [
  		"photo",
  		"sanfrancisco"
  	],
  	char: "🌉",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var milky_way$1 = {
  	keywords: [
  		"photo",
  		"space",
  		"stars"
  	],
  	char: "🌌",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var stars$1 = {
  	keywords: [
  		"night",
  		"photo"
  	],
  	char: "🌠",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var sparkler$1 = {
  	keywords: [
  		"stars",
  		"night",
  		"shine"
  	],
  	char: "🎇",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var fireworks$1 = {
  	keywords: [
  		"photo",
  		"festival",
  		"carnival",
  		"congratulations"
  	],
  	char: "🎆",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var rainbow$1 = {
  	keywords: [
  		"nature",
  		"happy",
  		"unicorn_face",
  		"photo",
  		"sky",
  		"spring"
  	],
  	char: "🌈",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var houses$1 = {
  	keywords: [
  		"buildings",
  		"photo"
  	],
  	char: "🏘",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var european_castle$1 = {
  	keywords: [
  		"building",
  		"royalty",
  		"history"
  	],
  	char: "🏰",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var japanese_castle$1 = {
  	keywords: [
  		"photo",
  		"building"
  	],
  	char: "🏯",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var stadium$1 = {
  	keywords: [
  		"photo",
  		"place",
  		"sports",
  		"concert",
  		"venue"
  	],
  	char: "🏟",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var statue_of_liberty$1 = {
  	keywords: [
  		"american",
  		"newyork"
  	],
  	char: "🗽",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var house$1 = {
  	keywords: [
  		"building",
  		"home"
  	],
  	char: "🏠",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var house_with_garden$1 = {
  	keywords: [
  		"home",
  		"plant",
  		"nature"
  	],
  	char: "🏡",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var derelict_house$1 = {
  	keywords: [
  		"abandon",
  		"evict",
  		"broken",
  		"building"
  	],
  	char: "🏚",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var office$1 = {
  	keywords: [
  		"building",
  		"bureau",
  		"work"
  	],
  	char: "🏢",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var department_store$1 = {
  	keywords: [
  		"building",
  		"shopping",
  		"mall"
  	],
  	char: "🏬",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var post_office$1 = {
  	keywords: [
  		"building",
  		"envelope",
  		"communication"
  	],
  	char: "🏣",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var european_post_office$1 = {
  	keywords: [
  		"building",
  		"email"
  	],
  	char: "🏤",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var hospital$1 = {
  	keywords: [
  		"building",
  		"health",
  		"surgery",
  		"doctor"
  	],
  	char: "🏥",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var bank$1 = {
  	keywords: [
  		"building",
  		"money",
  		"sales",
  		"cash",
  		"business",
  		"enterprise"
  	],
  	char: "🏦",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var hotel$1 = {
  	keywords: [
  		"building",
  		"accomodation",
  		"checkin"
  	],
  	char: "🏨",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var convenience_store$1 = {
  	keywords: [
  		"building",
  		"shopping",
  		"groceries"
  	],
  	char: "🏪",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var school$1 = {
  	keywords: [
  		"building",
  		"student",
  		"education",
  		"learn",
  		"teach"
  	],
  	char: "🏫",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var love_hotel$1 = {
  	keywords: [
  		"like",
  		"affection",
  		"dating"
  	],
  	char: "🏩",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var wedding$1 = {
  	keywords: [
  		"love",
  		"like",
  		"affection",
  		"couple",
  		"marriage",
  		"bride",
  		"groom"
  	],
  	char: "💒",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var classical_building$1 = {
  	keywords: [
  		"art",
  		"culture",
  		"history"
  	],
  	char: "🏛",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var church$1 = {
  	keywords: [
  		"building",
  		"religion",
  		"christ"
  	],
  	char: "⛪",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var mosque$1 = {
  	keywords: [
  		"islam",
  		"worship",
  		"minaret"
  	],
  	char: "🕌",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var synagogue$1 = {
  	keywords: [
  		"judaism",
  		"worship",
  		"temple",
  		"jewish"
  	],
  	char: "🕍",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var kaaba$1 = {
  	keywords: [
  		"mecca",
  		"mosque",
  		"islam"
  	],
  	char: "🕋",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var shinto_shrine$1 = {
  	keywords: [
  		"temple",
  		"japan",
  		"kyoto"
  	],
  	char: "⛩",
  	fitzpatrick_scale: false,
  	category: "travel_and_places"
  };
  var watch$1 = {
  	keywords: [
  		"time",
  		"accessories"
  	],
  	char: "⌚",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var iphone$1 = {
  	keywords: [
  		"technology",
  		"apple",
  		"gadgets",
  		"dial"
  	],
  	char: "📱",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var calling$1 = {
  	keywords: [
  		"iphone",
  		"incoming"
  	],
  	char: "📲",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var computer$1 = {
  	keywords: [
  		"technology",
  		"laptop",
  		"screen",
  		"display",
  		"monitor"
  	],
  	char: "💻",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var keyboard$1 = {
  	keywords: [
  		"technology",
  		"computer",
  		"type",
  		"input",
  		"text"
  	],
  	char: "⌨",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var desktop_computer$1 = {
  	keywords: [
  		"technology",
  		"computing",
  		"screen"
  	],
  	char: "🖥",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var printer$1 = {
  	keywords: [
  		"paper",
  		"ink"
  	],
  	char: "🖨",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var computer_mouse$1 = {
  	keywords: [
  		"click"
  	],
  	char: "🖱",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var trackball$1 = {
  	keywords: [
  		"technology",
  		"trackpad"
  	],
  	char: "🖲",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var joystick$1 = {
  	keywords: [
  		"game",
  		"play"
  	],
  	char: "🕹",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var clamp$1 = {
  	keywords: [
  		"tool"
  	],
  	char: "🗜",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var minidisc$1 = {
  	keywords: [
  		"technology",
  		"record",
  		"data",
  		"disk",
  		"90s"
  	],
  	char: "💽",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var floppy_disk$1 = {
  	keywords: [
  		"oldschool",
  		"technology",
  		"save",
  		"90s",
  		"80s"
  	],
  	char: "💾",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var cd$1 = {
  	keywords: [
  		"technology",
  		"dvd",
  		"disk",
  		"disc",
  		"90s"
  	],
  	char: "💿",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var dvd$1 = {
  	keywords: [
  		"cd",
  		"disk",
  		"disc"
  	],
  	char: "📀",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var vhs$1 = {
  	keywords: [
  		"record",
  		"video",
  		"oldschool",
  		"90s",
  		"80s"
  	],
  	char: "📼",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var camera$1 = {
  	keywords: [
  		"gadgets",
  		"photography"
  	],
  	char: "📷",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var camera_flash$1 = {
  	keywords: [
  		"photography",
  		"gadgets"
  	],
  	char: "📸",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var video_camera$1 = {
  	keywords: [
  		"film",
  		"record"
  	],
  	char: "📹",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var movie_camera$1 = {
  	keywords: [
  		"film",
  		"record"
  	],
  	char: "🎥",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var film_projector$1 = {
  	keywords: [
  		"video",
  		"tape",
  		"record",
  		"movie"
  	],
  	char: "📽",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var film_strip$1 = {
  	keywords: [
  		"movie"
  	],
  	char: "🎞",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var telephone_receiver$1 = {
  	keywords: [
  		"technology",
  		"communication",
  		"dial"
  	],
  	char: "📞",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var phone$2 = {
  	keywords: [
  		"technology",
  		"communication",
  		"dial",
  		"telephone"
  	],
  	char: "☎️",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var pager$1 = {
  	keywords: [
  		"bbcall",
  		"oldschool",
  		"90s"
  	],
  	char: "📟",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var fax$1 = {
  	keywords: [
  		"communication",
  		"technology"
  	],
  	char: "📠",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var tv$1 = {
  	keywords: [
  		"technology",
  		"program",
  		"oldschool",
  		"show",
  		"television"
  	],
  	char: "📺",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var radio$1 = {
  	keywords: [
  		"communication",
  		"music",
  		"podcast",
  		"program"
  	],
  	char: "📻",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var studio_microphone$1 = {
  	keywords: [
  		"sing",
  		"recording",
  		"artist",
  		"talkshow"
  	],
  	char: "🎙",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var level_slider$1 = {
  	keywords: [
  		"scale"
  	],
  	char: "🎚",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var control_knobs$1 = {
  	keywords: [
  		"dial"
  	],
  	char: "🎛",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var compass$1 = {
  	keywords: [
  		"magnetic",
  		"navigation",
  		"orienteering"
  	],
  	char: "🧭",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var stopwatch$1 = {
  	keywords: [
  		"time",
  		"deadline"
  	],
  	char: "⏱",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var timer_clock$1 = {
  	keywords: [
  		"alarm"
  	],
  	char: "⏲",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var alarm_clock$1 = {
  	keywords: [
  		"time",
  		"wake"
  	],
  	char: "⏰",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var mantelpiece_clock$1 = {
  	keywords: [
  		"time"
  	],
  	char: "🕰",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var hourglass_flowing_sand$1 = {
  	keywords: [
  		"oldschool",
  		"time",
  		"countdown"
  	],
  	char: "⏳",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var hourglass$1 = {
  	keywords: [
  		"time",
  		"clock",
  		"oldschool",
  		"limit",
  		"exam",
  		"quiz",
  		"test"
  	],
  	char: "⌛",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var satellite$1 = {
  	keywords: [
  		"communication",
  		"future",
  		"radio",
  		"space"
  	],
  	char: "📡",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var battery$1 = {
  	keywords: [
  		"power",
  		"energy",
  		"sustain"
  	],
  	char: "🔋",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var electric_plug$1 = {
  	keywords: [
  		"charger",
  		"power"
  	],
  	char: "🔌",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var bulb$1 = {
  	keywords: [
  		"light",
  		"electricity",
  		"idea"
  	],
  	char: "💡",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var flashlight$1 = {
  	keywords: [
  		"dark",
  		"camping",
  		"sight",
  		"night"
  	],
  	char: "🔦",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var candle$1 = {
  	keywords: [
  		"fire",
  		"wax"
  	],
  	char: "🕯",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var fire_extinguisher$1 = {
  	keywords: [
  		"quench"
  	],
  	char: "🧯",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var wastebasket$1 = {
  	keywords: [
  		"bin",
  		"trash",
  		"rubbish",
  		"garbage",
  		"toss"
  	],
  	char: "🗑",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var oil_drum$1 = {
  	keywords: [
  		"barrell"
  	],
  	char: "🛢",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var money_with_wings$1 = {
  	keywords: [
  		"dollar",
  		"bills",
  		"payment",
  		"sale"
  	],
  	char: "💸",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var dollar$2 = {
  	keywords: [
  		"money",
  		"sales",
  		"bill",
  		"currency"
  	],
  	char: "💵",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var yen$2 = {
  	keywords: [
  		"money",
  		"sales",
  		"japanese",
  		"dollar",
  		"currency"
  	],
  	char: "💴",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var euro$2 = {
  	keywords: [
  		"money",
  		"sales",
  		"dollar",
  		"currency"
  	],
  	char: "💶",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var pound$2 = {
  	keywords: [
  		"british",
  		"sterling",
  		"money",
  		"sales",
  		"bills",
  		"uk",
  		"england",
  		"currency"
  	],
  	char: "💷",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var moneybag$1 = {
  	keywords: [
  		"dollar",
  		"payment",
  		"coins",
  		"sale"
  	],
  	char: "💰",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var credit_card$1 = {
  	keywords: [
  		"money",
  		"sales",
  		"dollar",
  		"bill",
  		"payment",
  		"shopping"
  	],
  	char: "💳",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var gem$1 = {
  	keywords: [
  		"blue",
  		"ruby",
  		"diamond",
  		"jewelry"
  	],
  	char: "💎",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var balance_scale$1 = {
  	keywords: [
  		"law",
  		"fairness",
  		"weight"
  	],
  	char: "⚖",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var toolbox$1 = {
  	keywords: [
  		"tools",
  		"diy",
  		"fix",
  		"maintainer",
  		"mechanic"
  	],
  	char: "🧰",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var wrench$1 = {
  	keywords: [
  		"tools",
  		"diy",
  		"ikea",
  		"fix",
  		"maintainer"
  	],
  	char: "🔧",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var hammer$1 = {
  	keywords: [
  		"tools",
  		"build",
  		"create"
  	],
  	char: "🔨",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var hammer_and_pick$1 = {
  	keywords: [
  		"tools",
  		"build",
  		"create"
  	],
  	char: "⚒",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var hammer_and_wrench$1 = {
  	keywords: [
  		"tools",
  		"build",
  		"create"
  	],
  	char: "🛠",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var pick$1 = {
  	keywords: [
  		"tools",
  		"dig"
  	],
  	char: "⛏",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var nut_and_bolt$1 = {
  	keywords: [
  		"handy",
  		"tools",
  		"fix"
  	],
  	char: "🔩",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var gear$1 = {
  	keywords: [
  		"cog"
  	],
  	char: "⚙",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var brick = {
  	keywords: [
  		"bricks"
  	],
  	char: "🧱",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var chains$1 = {
  	keywords: [
  		"lock",
  		"arrest"
  	],
  	char: "⛓",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var magnet$1 = {
  	keywords: [
  		"attraction",
  		"magnetic"
  	],
  	char: "🧲",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var gun$1 = {
  	keywords: [
  		"violence",
  		"weapon",
  		"pistol",
  		"revolver"
  	],
  	char: "🔫",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var bomb$1 = {
  	keywords: [
  		"boom",
  		"explode",
  		"explosion",
  		"terrorism"
  	],
  	char: "💣",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var firecracker$1 = {
  	keywords: [
  		"dynamite",
  		"boom",
  		"explode",
  		"explosion",
  		"explosive"
  	],
  	char: "🧨",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var hocho$1 = {
  	keywords: [
  		"knife",
  		"blade",
  		"cutlery",
  		"kitchen",
  		"weapon"
  	],
  	char: "🔪",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var dagger$2 = {
  	keywords: [
  		"weapon"
  	],
  	char: "🗡",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var crossed_swords$1 = {
  	keywords: [
  		"weapon"
  	],
  	char: "⚔",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var shield$1 = {
  	keywords: [
  		"protection",
  		"security"
  	],
  	char: "🛡",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var smoking$1 = {
  	keywords: [
  		"kills",
  		"tobacco",
  		"cigarette",
  		"joint",
  		"smoke"
  	],
  	char: "🚬",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var skull_and_crossbones$1 = {
  	keywords: [
  		"poison",
  		"danger",
  		"deadly",
  		"scary",
  		"death",
  		"pirate",
  		"evil"
  	],
  	char: "☠",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var coffin$1 = {
  	keywords: [
  		"vampire",
  		"dead",
  		"die",
  		"death",
  		"rip",
  		"graveyard",
  		"cemetery",
  		"casket",
  		"funeral",
  		"box"
  	],
  	char: "⚰",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var funeral_urn$1 = {
  	keywords: [
  		"dead",
  		"die",
  		"death",
  		"rip",
  		"ashes"
  	],
  	char: "⚱",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var amphora$1 = {
  	keywords: [
  		"vase",
  		"jar"
  	],
  	char: "🏺",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var crystal_ball$1 = {
  	keywords: [
  		"disco",
  		"party",
  		"magic",
  		"circus",
  		"fortune_teller"
  	],
  	char: "🔮",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var prayer_beads$1 = {
  	keywords: [
  		"dhikr",
  		"religious"
  	],
  	char: "📿",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var nazar_amulet$1 = {
  	keywords: [
  		"bead",
  		"charm"
  	],
  	char: "🧿",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var barber$1 = {
  	keywords: [
  		"hair",
  		"salon",
  		"style"
  	],
  	char: "💈",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var alembic$1 = {
  	keywords: [
  		"distilling",
  		"science",
  		"experiment",
  		"chemistry"
  	],
  	char: "⚗",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var telescope$1 = {
  	keywords: [
  		"stars",
  		"space",
  		"zoom",
  		"science",
  		"astronomy"
  	],
  	char: "🔭",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var microscope$1 = {
  	keywords: [
  		"laboratory",
  		"experiment",
  		"zoomin",
  		"science",
  		"study"
  	],
  	char: "🔬",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var hole$1 = {
  	keywords: [
  		"embarrassing"
  	],
  	char: "🕳",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var pill$1 = {
  	keywords: [
  		"health",
  		"medicine",
  		"doctor",
  		"pharmacy",
  		"drug"
  	],
  	char: "💊",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var syringe$1 = {
  	keywords: [
  		"health",
  		"hospital",
  		"drugs",
  		"blood",
  		"medicine",
  		"needle",
  		"doctor",
  		"nurse"
  	],
  	char: "💉",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var dna$1 = {
  	keywords: [
  		"biologist",
  		"genetics",
  		"life"
  	],
  	char: "🧬",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var microbe$1 = {
  	keywords: [
  		"amoeba",
  		"bacteria",
  		"germs"
  	],
  	char: "🦠",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var petri_dish$1 = {
  	keywords: [
  		"bacteria",
  		"biology",
  		"culture",
  		"lab"
  	],
  	char: "🧫",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var test_tube$1 = {
  	keywords: [
  		"chemistry",
  		"experiment",
  		"lab",
  		"science"
  	],
  	char: "🧪",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var thermometer$1 = {
  	keywords: [
  		"weather",
  		"temperature",
  		"hot",
  		"cold"
  	],
  	char: "🌡",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var broom$1 = {
  	keywords: [
  		"cleaning",
  		"sweeping",
  		"witch"
  	],
  	char: "🧹",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var basket$1 = {
  	keywords: [
  		"laundry"
  	],
  	char: "🧺",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var toilet_paper = {
  	keywords: [
  		"roll"
  	],
  	char: "🧻",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var label$1 = {
  	keywords: [
  		"sale",
  		"tag"
  	],
  	char: "🏷",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var bookmark$1 = {
  	keywords: [
  		"favorite",
  		"label",
  		"save"
  	],
  	char: "🔖",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var toilet$1 = {
  	keywords: [
  		"restroom",
  		"wc",
  		"washroom",
  		"bathroom",
  		"potty"
  	],
  	char: "🚽",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var shower$1 = {
  	keywords: [
  		"clean",
  		"water",
  		"bathroom"
  	],
  	char: "🚿",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var bathtub$1 = {
  	keywords: [
  		"clean",
  		"shower",
  		"bathroom"
  	],
  	char: "🛁",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var soap$1 = {
  	keywords: [
  		"bar",
  		"bathing",
  		"cleaning",
  		"lather"
  	],
  	char: "🧼",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var sponge$1 = {
  	keywords: [
  		"absorbing",
  		"cleaning",
  		"porous"
  	],
  	char: "🧽",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var lotion_bottle$1 = {
  	keywords: [
  		"moisturizer",
  		"sunscreen"
  	],
  	char: "🧴",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var key$1 = {
  	keywords: [
  		"lock",
  		"door",
  		"password"
  	],
  	char: "🔑",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var old_key$1 = {
  	keywords: [
  		"lock",
  		"door",
  		"password"
  	],
  	char: "🗝",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var couch_and_lamp$1 = {
  	keywords: [
  		"read",
  		"chill"
  	],
  	char: "🛋",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var sleeping_bed$1 = {
  	keywords: [
  		"bed",
  		"rest"
  	],
  	char: "🛌",
  	fitzpatrick_scale: true,
  	category: "objects"
  };
  var bed$1 = {
  	keywords: [
  		"sleep",
  		"rest"
  	],
  	char: "🛏",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var door$1 = {
  	keywords: [
  		"house",
  		"entry",
  		"exit"
  	],
  	char: "🚪",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var bellhop_bell$1 = {
  	keywords: [
  		"service"
  	],
  	char: "🛎",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var teddy_bear$1 = {
  	keywords: [
  		"plush",
  		"stuffed"
  	],
  	char: "🧸",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var framed_picture$1 = {
  	keywords: [
  		"photography"
  	],
  	char: "🖼",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var world_map$1 = {
  	keywords: [
  		"location",
  		"direction"
  	],
  	char: "🗺",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var parasol_on_ground$1 = {
  	keywords: [
  		"weather",
  		"summer"
  	],
  	char: "⛱",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var moyai$1 = {
  	keywords: [
  		"rock",
  		"easter island",
  		"moai"
  	],
  	char: "🗿",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var shopping$1 = {
  	keywords: [
  		"mall",
  		"buy",
  		"purchase"
  	],
  	char: "🛍",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var shopping_cart$1 = {
  	keywords: [
  		"trolley"
  	],
  	char: "🛒",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var balloon$1 = {
  	keywords: [
  		"party",
  		"celebration",
  		"birthday",
  		"circus"
  	],
  	char: "🎈",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var flags$1 = {
  	keywords: [
  		"fish",
  		"japanese",
  		"koinobori",
  		"carp",
  		"banner"
  	],
  	char: "🎏",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var ribbon$1 = {
  	keywords: [
  		"decoration",
  		"pink",
  		"girl",
  		"bowtie"
  	],
  	char: "🎀",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var gift$1 = {
  	keywords: [
  		"present",
  		"birthday",
  		"christmas",
  		"xmas"
  	],
  	char: "🎁",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var confetti_ball$1 = {
  	keywords: [
  		"festival",
  		"party",
  		"birthday",
  		"circus"
  	],
  	char: "🎊",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var tada$1 = {
  	keywords: [
  		"party",
  		"congratulations",
  		"birthday",
  		"magic",
  		"circus",
  		"celebration"
  	],
  	char: "🎉",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var dolls$1 = {
  	keywords: [
  		"japanese",
  		"toy",
  		"kimono"
  	],
  	char: "🎎",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var wind_chime$1 = {
  	keywords: [
  		"nature",
  		"ding",
  		"spring",
  		"bell"
  	],
  	char: "🎐",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var crossed_flags$1 = {
  	keywords: [
  		"japanese",
  		"nation",
  		"country",
  		"border"
  	],
  	char: "🎌",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var izakaya_lantern$1 = {
  	keywords: [
  		"light",
  		"paper",
  		"halloween",
  		"spooky"
  	],
  	char: "🏮",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var red_envelope$1 = {
  	keywords: [
  		"gift"
  	],
  	char: "🧧",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var email$1 = {
  	keywords: [
  		"letter",
  		"postal",
  		"inbox",
  		"communication"
  	],
  	char: "✉️",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var envelope_with_arrow$1 = {
  	keywords: [
  		"email",
  		"communication"
  	],
  	char: "📩",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var incoming_envelope$1 = {
  	keywords: [
  		"email",
  		"inbox"
  	],
  	char: "📨",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var love_letter$1 = {
  	keywords: [
  		"email",
  		"like",
  		"affection",
  		"envelope",
  		"valentines"
  	],
  	char: "💌",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var postbox$1 = {
  	keywords: [
  		"email",
  		"letter",
  		"envelope"
  	],
  	char: "📮",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var mailbox_closed$1 = {
  	keywords: [
  		"email",
  		"communication",
  		"inbox"
  	],
  	char: "📪",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var mailbox$1 = {
  	keywords: [
  		"email",
  		"inbox",
  		"communication"
  	],
  	char: "📫",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var mailbox_with_mail$1 = {
  	keywords: [
  		"email",
  		"inbox",
  		"communication"
  	],
  	char: "📬",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var mailbox_with_no_mail$1 = {
  	keywords: [
  		"email",
  		"inbox"
  	],
  	char: "📭",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var postal_horn$1 = {
  	keywords: [
  		"instrument",
  		"music"
  	],
  	char: "📯",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var inbox_tray$1 = {
  	keywords: [
  		"email",
  		"documents"
  	],
  	char: "📥",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var outbox_tray$1 = {
  	keywords: [
  		"inbox",
  		"email"
  	],
  	char: "📤",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var scroll$1 = {
  	keywords: [
  		"documents",
  		"ancient",
  		"history",
  		"paper"
  	],
  	char: "📜",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var page_with_curl$1 = {
  	keywords: [
  		"documents",
  		"office",
  		"paper"
  	],
  	char: "📃",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var bookmark_tabs$1 = {
  	keywords: [
  		"favorite",
  		"save",
  		"order",
  		"tidy"
  	],
  	char: "📑",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var receipt$1 = {
  	keywords: [
  		"accounting",
  		"expenses"
  	],
  	char: "🧾",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var bar_chart$1 = {
  	keywords: [
  		"graph",
  		"presentation",
  		"stats"
  	],
  	char: "📊",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var chart_with_upwards_trend$1 = {
  	keywords: [
  		"graph",
  		"presentation",
  		"stats",
  		"recovery",
  		"business",
  		"economics",
  		"money",
  		"sales",
  		"good",
  		"success"
  	],
  	char: "📈",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var chart_with_downwards_trend$1 = {
  	keywords: [
  		"graph",
  		"presentation",
  		"stats",
  		"recession",
  		"business",
  		"economics",
  		"money",
  		"sales",
  		"bad",
  		"failure"
  	],
  	char: "📉",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var page_facing_up$1 = {
  	keywords: [
  		"documents",
  		"office",
  		"paper",
  		"information"
  	],
  	char: "📄",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var date$1 = {
  	keywords: [
  		"calendar",
  		"schedule"
  	],
  	char: "📅",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var calendar$1 = {
  	keywords: [
  		"schedule",
  		"date",
  		"planning"
  	],
  	char: "📆",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var spiral_calendar$1 = {
  	keywords: [
  		"date",
  		"schedule",
  		"planning"
  	],
  	char: "🗓",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var card_index$1 = {
  	keywords: [
  		"business",
  		"stationery"
  	],
  	char: "📇",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var card_file_box$1 = {
  	keywords: [
  		"business",
  		"stationery"
  	],
  	char: "🗃",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var ballot_box$1 = {
  	keywords: [
  		"election",
  		"vote"
  	],
  	char: "🗳",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var file_cabinet$1 = {
  	keywords: [
  		"filing",
  		"organizing"
  	],
  	char: "🗄",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var clipboard$2 = {
  	keywords: [
  		"stationery",
  		"documents"
  	],
  	char: "📋",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var spiral_notepad$1 = {
  	keywords: [
  		"memo",
  		"stationery"
  	],
  	char: "🗒",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var file_folder$1 = {
  	keywords: [
  		"documents",
  		"business",
  		"office"
  	],
  	char: "📁",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var open_file_folder$1 = {
  	keywords: [
  		"documents",
  		"load"
  	],
  	char: "📂",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var card_index_dividers$1 = {
  	keywords: [
  		"organizing",
  		"business",
  		"stationery"
  	],
  	char: "🗂",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var newspaper_roll$1 = {
  	keywords: [
  		"press",
  		"headline"
  	],
  	char: "🗞",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var newspaper$1 = {
  	keywords: [
  		"press",
  		"headline"
  	],
  	char: "📰",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var notebook$1 = {
  	keywords: [
  		"stationery",
  		"record",
  		"notes",
  		"paper",
  		"study"
  	],
  	char: "📓",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var closed_book$1 = {
  	keywords: [
  		"read",
  		"library",
  		"knowledge",
  		"textbook",
  		"learn"
  	],
  	char: "📕",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var green_book$1 = {
  	keywords: [
  		"read",
  		"library",
  		"knowledge",
  		"study"
  	],
  	char: "📗",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var blue_book$1 = {
  	keywords: [
  		"read",
  		"library",
  		"knowledge",
  		"learn",
  		"study"
  	],
  	char: "📘",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var orange_book$1 = {
  	keywords: [
  		"read",
  		"library",
  		"knowledge",
  		"textbook",
  		"study"
  	],
  	char: "📙",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var notebook_with_decorative_cover$1 = {
  	keywords: [
  		"classroom",
  		"notes",
  		"record",
  		"paper",
  		"study"
  	],
  	char: "📔",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var ledger$1 = {
  	keywords: [
  		"notes",
  		"paper"
  	],
  	char: "📒",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var books$1 = {
  	keywords: [
  		"literature",
  		"library",
  		"study"
  	],
  	char: "📚",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var open_book$1 = {
  	keywords: [
  		"book",
  		"read",
  		"library",
  		"knowledge",
  		"literature",
  		"learn",
  		"study"
  	],
  	char: "📖",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var safety_pin$1 = {
  	keywords: [
  		"diaper"
  	],
  	char: "🧷",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var link$3 = {
  	keywords: [
  		"rings",
  		"url"
  	],
  	char: "🔗",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var paperclip$1 = {
  	keywords: [
  		"documents",
  		"stationery"
  	],
  	char: "📎",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var paperclips$1 = {
  	keywords: [
  		"documents",
  		"stationery"
  	],
  	char: "🖇",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var scissors$1 = {
  	keywords: [
  		"stationery",
  		"cut"
  	],
  	char: "✂️",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var triangular_ruler$1 = {
  	keywords: [
  		"stationery",
  		"math",
  		"architect",
  		"sketch"
  	],
  	char: "📐",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var straight_ruler$1 = {
  	keywords: [
  		"stationery",
  		"calculate",
  		"length",
  		"math",
  		"school",
  		"drawing",
  		"architect",
  		"sketch"
  	],
  	char: "📏",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var abacus$1 = {
  	keywords: [
  		"calculation"
  	],
  	char: "🧮",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var pushpin$1 = {
  	keywords: [
  		"stationery",
  		"mark",
  		"here"
  	],
  	char: "📌",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var round_pushpin$1 = {
  	keywords: [
  		"stationery",
  		"location",
  		"map",
  		"here"
  	],
  	char: "📍",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var triangular_flag_on_post$1 = {
  	keywords: [
  		"mark",
  		"milestone",
  		"place"
  	],
  	char: "🚩",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var white_flag$1 = {
  	keywords: [
  		"losing",
  		"loser",
  		"lost",
  		"surrender",
  		"give up",
  		"fail"
  	],
  	char: "🏳",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var black_flag$1 = {
  	keywords: [
  		"pirate"
  	],
  	char: "🏴",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var rainbow_flag$1 = {
  	keywords: [
  		"flag",
  		"rainbow",
  		"pride",
  		"gay",
  		"lgbt",
  		"glbt",
  		"queer",
  		"homosexual",
  		"lesbian",
  		"bisexual",
  		"transgender"
  	],
  	char: "🏳️‍🌈",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var closed_lock_with_key$1 = {
  	keywords: [
  		"security",
  		"privacy"
  	],
  	char: "🔐",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var lock$1 = {
  	keywords: [
  		"security",
  		"password",
  		"padlock"
  	],
  	char: "🔒",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var unlock$1 = {
  	keywords: [
  		"privacy",
  		"security"
  	],
  	char: "🔓",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var lock_with_ink_pen$1 = {
  	keywords: [
  		"security",
  		"secret"
  	],
  	char: "🔏",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var pen$1 = {
  	keywords: [
  		"stationery",
  		"writing",
  		"write"
  	],
  	char: "🖊",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var fountain_pen$1 = {
  	keywords: [
  		"stationery",
  		"writing",
  		"write"
  	],
  	char: "🖋",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var black_nib$1 = {
  	keywords: [
  		"pen",
  		"stationery",
  		"writing",
  		"write"
  	],
  	char: "✒️",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var memo$1 = {
  	keywords: [
  		"write",
  		"documents",
  		"stationery",
  		"pencil",
  		"paper",
  		"writing",
  		"legal",
  		"exam",
  		"quiz",
  		"test",
  		"study",
  		"compose"
  	],
  	char: "📝",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var pencil2$1 = {
  	keywords: [
  		"stationery",
  		"write",
  		"paper",
  		"writing",
  		"school",
  		"study"
  	],
  	char: "✏️",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var crayon$1 = {
  	keywords: [
  		"drawing",
  		"creativity"
  	],
  	char: "🖍",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var paintbrush$1 = {
  	keywords: [
  		"drawing",
  		"creativity",
  		"art"
  	],
  	char: "🖌",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var mag$1 = {
  	keywords: [
  		"search",
  		"zoom",
  		"find",
  		"detective"
  	],
  	char: "🔍",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var mag_right$1 = {
  	keywords: [
  		"search",
  		"zoom",
  		"find",
  		"detective"
  	],
  	char: "🔎",
  	fitzpatrick_scale: false,
  	category: "objects"
  };
  var heart$1 = {
  	keywords: [
  		"love",
  		"like",
  		"valentines"
  	],
  	char: "❤️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var orange_heart$1 = {
  	keywords: [
  		"love",
  		"like",
  		"affection",
  		"valentines"
  	],
  	char: "🧡",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var yellow_heart$1 = {
  	keywords: [
  		"love",
  		"like",
  		"affection",
  		"valentines"
  	],
  	char: "💛",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var green_heart$1 = {
  	keywords: [
  		"love",
  		"like",
  		"affection",
  		"valentines"
  	],
  	char: "💚",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var blue_heart$1 = {
  	keywords: [
  		"love",
  		"like",
  		"affection",
  		"valentines"
  	],
  	char: "💙",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var purple_heart$1 = {
  	keywords: [
  		"love",
  		"like",
  		"affection",
  		"valentines"
  	],
  	char: "💜",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var black_heart$1 = {
  	keywords: [
  		"evil"
  	],
  	char: "🖤",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var broken_heart$1 = {
  	keywords: [
  		"sad",
  		"sorry",
  		"break",
  		"heart",
  		"heartbreak"
  	],
  	char: "💔",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var heavy_heart_exclamation$1 = {
  	keywords: [
  		"decoration",
  		"love"
  	],
  	char: "❣",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var two_hearts$1 = {
  	keywords: [
  		"love",
  		"like",
  		"affection",
  		"valentines",
  		"heart"
  	],
  	char: "💕",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var revolving_hearts$1 = {
  	keywords: [
  		"love",
  		"like",
  		"affection",
  		"valentines"
  	],
  	char: "💞",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var heartbeat$1 = {
  	keywords: [
  		"love",
  		"like",
  		"affection",
  		"valentines",
  		"pink",
  		"heart"
  	],
  	char: "💓",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var heartpulse$1 = {
  	keywords: [
  		"like",
  		"love",
  		"affection",
  		"valentines",
  		"pink"
  	],
  	char: "💗",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var sparkling_heart$1 = {
  	keywords: [
  		"love",
  		"like",
  		"affection",
  		"valentines"
  	],
  	char: "💖",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var cupid$1 = {
  	keywords: [
  		"love",
  		"like",
  		"heart",
  		"affection",
  		"valentines"
  	],
  	char: "💘",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var gift_heart$1 = {
  	keywords: [
  		"love",
  		"valentines"
  	],
  	char: "💝",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var heart_decoration$1 = {
  	keywords: [
  		"purple-square",
  		"love",
  		"like"
  	],
  	char: "💟",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var peace_symbol$1 = {
  	keywords: [
  		"hippie"
  	],
  	char: "☮",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var latin_cross$1 = {
  	keywords: [
  		"christianity"
  	],
  	char: "✝",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var star_and_crescent$1 = {
  	keywords: [
  		"islam"
  	],
  	char: "☪",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var om$1 = {
  	keywords: [
  		"hinduism",
  		"buddhism",
  		"sikhism",
  		"jainism"
  	],
  	char: "🕉",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var wheel_of_dharma$1 = {
  	keywords: [
  		"hinduism",
  		"buddhism",
  		"sikhism",
  		"jainism"
  	],
  	char: "☸",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var star_of_david$1 = {
  	keywords: [
  		"judaism"
  	],
  	char: "✡",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var six_pointed_star$1 = {
  	keywords: [
  		"purple-square",
  		"religion",
  		"jewish",
  		"hexagram"
  	],
  	char: "🔯",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var menorah$1 = {
  	keywords: [
  		"hanukkah",
  		"candles",
  		"jewish"
  	],
  	char: "🕎",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var yin_yang$1 = {
  	keywords: [
  		"balance"
  	],
  	char: "☯",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var orthodox_cross$1 = {
  	keywords: [
  		"suppedaneum",
  		"religion"
  	],
  	char: "☦",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var place_of_worship$1 = {
  	keywords: [
  		"religion",
  		"church",
  		"temple",
  		"prayer"
  	],
  	char: "🛐",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var ophiuchus$1 = {
  	keywords: [
  		"sign",
  		"purple-square",
  		"constellation",
  		"astrology"
  	],
  	char: "⛎",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var aries$1 = {
  	keywords: [
  		"sign",
  		"purple-square",
  		"zodiac",
  		"astrology"
  	],
  	char: "♈",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var taurus$1 = {
  	keywords: [
  		"purple-square",
  		"sign",
  		"zodiac",
  		"astrology"
  	],
  	char: "♉",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var gemini$1 = {
  	keywords: [
  		"sign",
  		"zodiac",
  		"purple-square",
  		"astrology"
  	],
  	char: "♊",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var cancer$1 = {
  	keywords: [
  		"sign",
  		"zodiac",
  		"purple-square",
  		"astrology"
  	],
  	char: "♋",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var leo$1 = {
  	keywords: [
  		"sign",
  		"purple-square",
  		"zodiac",
  		"astrology"
  	],
  	char: "♌",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var virgo$1 = {
  	keywords: [
  		"sign",
  		"zodiac",
  		"purple-square",
  		"astrology"
  	],
  	char: "♍",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var libra$1 = {
  	keywords: [
  		"sign",
  		"purple-square",
  		"zodiac",
  		"astrology"
  	],
  	char: "♎",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var scorpius$1 = {
  	keywords: [
  		"sign",
  		"zodiac",
  		"purple-square",
  		"astrology",
  		"scorpio"
  	],
  	char: "♏",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var sagittarius$1 = {
  	keywords: [
  		"sign",
  		"zodiac",
  		"purple-square",
  		"astrology"
  	],
  	char: "♐",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var capricorn$1 = {
  	keywords: [
  		"sign",
  		"zodiac",
  		"purple-square",
  		"astrology"
  	],
  	char: "♑",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var aquarius$1 = {
  	keywords: [
  		"sign",
  		"purple-square",
  		"zodiac",
  		"astrology"
  	],
  	char: "♒",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var pisces$1 = {
  	keywords: [
  		"purple-square",
  		"sign",
  		"zodiac",
  		"astrology"
  	],
  	char: "♓",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var id$1 = {
  	keywords: [
  		"purple-square",
  		"words"
  	],
  	char: "🆔",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var atom_symbol$1 = {
  	keywords: [
  		"science",
  		"physics",
  		"chemistry"
  	],
  	char: "⚛",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var u7a7a = {
  	keywords: [
  		"kanji",
  		"japanese",
  		"chinese",
  		"empty",
  		"sky",
  		"blue-square"
  	],
  	char: "🈳",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var u5272 = {
  	keywords: [
  		"cut",
  		"divide",
  		"chinese",
  		"kanji",
  		"pink-square"
  	],
  	char: "🈹",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var radioactive$1 = {
  	keywords: [
  		"nuclear",
  		"danger"
  	],
  	char: "☢",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var biohazard$1 = {
  	keywords: [
  		"danger"
  	],
  	char: "☣",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var mobile_phone_off$1 = {
  	keywords: [
  		"mute",
  		"orange-square",
  		"silence",
  		"quiet"
  	],
  	char: "📴",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var vibration_mode$1 = {
  	keywords: [
  		"orange-square",
  		"phone"
  	],
  	char: "📳",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var u6709 = {
  	keywords: [
  		"orange-square",
  		"chinese",
  		"have",
  		"kanji"
  	],
  	char: "🈶",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var u7121 = {
  	keywords: [
  		"nothing",
  		"chinese",
  		"kanji",
  		"japanese",
  		"orange-square"
  	],
  	char: "🈚",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var u7533 = {
  	keywords: [
  		"chinese",
  		"japanese",
  		"kanji",
  		"orange-square"
  	],
  	char: "🈸",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var u55b6 = {
  	keywords: [
  		"japanese",
  		"opening hours",
  		"orange-square"
  	],
  	char: "🈺",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var u6708 = {
  	keywords: [
  		"chinese",
  		"month",
  		"moon",
  		"japanese",
  		"orange-square",
  		"kanji"
  	],
  	char: "🈷️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var eight_pointed_black_star$1 = {
  	keywords: [
  		"orange-square",
  		"shape",
  		"polygon"
  	],
  	char: "✴️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var vs$1 = {
  	keywords: [
  		"words",
  		"orange-square"
  	],
  	char: "🆚",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var accept$1 = {
  	keywords: [
  		"ok",
  		"good",
  		"chinese",
  		"kanji",
  		"agree",
  		"yes",
  		"orange-circle"
  	],
  	char: "🉑",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var white_flower$1 = {
  	keywords: [
  		"japanese",
  		"spring"
  	],
  	char: "💮",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var ideograph_advantage$1 = {
  	keywords: [
  		"chinese",
  		"kanji",
  		"obtain",
  		"get",
  		"circle"
  	],
  	char: "🉐",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var secret$1 = {
  	keywords: [
  		"privacy",
  		"chinese",
  		"sshh",
  		"kanji",
  		"red-circle"
  	],
  	char: "㊙️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var congratulations$1 = {
  	keywords: [
  		"chinese",
  		"kanji",
  		"japanese",
  		"red-circle"
  	],
  	char: "㊗️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var u5408 = {
  	keywords: [
  		"japanese",
  		"chinese",
  		"join",
  		"kanji",
  		"red-square"
  	],
  	char: "🈴",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var u6e80$1 = {
  	keywords: [
  		"full",
  		"chinese",
  		"japanese",
  		"red-square",
  		"kanji"
  	],
  	char: "🈵",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var u7981 = {
  	keywords: [
  		"kanji",
  		"japanese",
  		"chinese",
  		"forbidden",
  		"limit",
  		"restricted",
  		"red-square"
  	],
  	char: "🈲",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var a$1 = {
  	keywords: [
  		"red-square",
  		"alphabet",
  		"letter"
  	],
  	char: "🅰️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var b$1 = {
  	keywords: [
  		"red-square",
  		"alphabet",
  		"letter"
  	],
  	char: "🅱️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var ab$1 = {
  	keywords: [
  		"red-square",
  		"alphabet"
  	],
  	char: "🆎",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var cl$1 = {
  	keywords: [
  		"alphabet",
  		"words",
  		"red-square"
  	],
  	char: "🆑",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var o2$1 = {
  	keywords: [
  		"alphabet",
  		"red-square",
  		"letter"
  	],
  	char: "🅾️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var sos$1 = {
  	keywords: [
  		"help",
  		"red-square",
  		"words",
  		"emergency",
  		"911"
  	],
  	char: "🆘",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var no_entry$1 = {
  	keywords: [
  		"limit",
  		"security",
  		"privacy",
  		"bad",
  		"denied",
  		"stop",
  		"circle"
  	],
  	char: "⛔",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var name_badge$1 = {
  	keywords: [
  		"fire",
  		"forbid"
  	],
  	char: "📛",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var no_entry_sign$1 = {
  	keywords: [
  		"forbid",
  		"stop",
  		"limit",
  		"denied",
  		"disallow",
  		"circle"
  	],
  	char: "🚫",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var x$1 = {
  	keywords: [
  		"no",
  		"delete",
  		"remove",
  		"cancel",
  		"red"
  	],
  	char: "❌",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var o$1 = {
  	keywords: [
  		"circle",
  		"round"
  	],
  	char: "⭕",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var stop_sign$1 = {
  	keywords: [
  		"stop"
  	],
  	char: "🛑",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var anger$1 = {
  	keywords: [
  		"angry",
  		"mad"
  	],
  	char: "💢",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var hotsprings$1 = {
  	keywords: [
  		"bath",
  		"warm",
  		"relax"
  	],
  	char: "♨️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var no_pedestrians$1 = {
  	keywords: [
  		"rules",
  		"crossing",
  		"walking",
  		"circle"
  	],
  	char: "🚷",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var do_not_litter$1 = {
  	keywords: [
  		"trash",
  		"bin",
  		"garbage",
  		"circle"
  	],
  	char: "🚯",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var no_bicycles$1 = {
  	keywords: [
  		"cyclist",
  		"prohibited",
  		"circle"
  	],
  	char: "🚳",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var underage$1 = {
  	keywords: [
  		"18",
  		"drink",
  		"pub",
  		"night",
  		"minor",
  		"circle"
  	],
  	char: "🔞",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var no_mobile_phones$1 = {
  	keywords: [
  		"iphone",
  		"mute",
  		"circle"
  	],
  	char: "📵",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var exclamation$1 = {
  	keywords: [
  		"heavy_exclamation_mark",
  		"danger",
  		"surprise",
  		"punctuation",
  		"wow",
  		"warning"
  	],
  	char: "❗",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var grey_exclamation$1 = {
  	keywords: [
  		"surprise",
  		"punctuation",
  		"gray",
  		"wow",
  		"warning"
  	],
  	char: "❕",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var question$1 = {
  	keywords: [
  		"doubt",
  		"confused"
  	],
  	char: "❓",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var grey_question$1 = {
  	keywords: [
  		"doubts",
  		"gray",
  		"huh",
  		"confused"
  	],
  	char: "❔",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var bangbang$1 = {
  	keywords: [
  		"exclamation",
  		"surprise"
  	],
  	char: "‼️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var interrobang$1 = {
  	keywords: [
  		"wat",
  		"punctuation",
  		"surprise"
  	],
  	char: "⁉️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var low_brightness$1 = {
  	keywords: [
  		"sun",
  		"afternoon",
  		"warm",
  		"summer"
  	],
  	char: "🔅",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var high_brightness$1 = {
  	keywords: [
  		"sun",
  		"light"
  	],
  	char: "🔆",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var trident$1 = {
  	keywords: [
  		"weapon",
  		"spear"
  	],
  	char: "🔱",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var fleur_de_lis$1 = {
  	keywords: [
  		"decorative",
  		"scout"
  	],
  	char: "⚜",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var part_alternation_mark$1 = {
  	keywords: [
  		"graph",
  		"presentation",
  		"stats",
  		"business",
  		"economics",
  		"bad"
  	],
  	char: "〽️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var warning$1 = {
  	keywords: [
  		"exclamation",
  		"wip",
  		"alert",
  		"error",
  		"problem",
  		"issue"
  	],
  	char: "⚠️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var children_crossing$1 = {
  	keywords: [
  		"school",
  		"warning",
  		"danger",
  		"sign",
  		"driving",
  		"yellow-diamond"
  	],
  	char: "🚸",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var beginner$1 = {
  	keywords: [
  		"badge",
  		"shield"
  	],
  	char: "🔰",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var recycle$1 = {
  	keywords: [
  		"arrow",
  		"environment",
  		"garbage",
  		"trash"
  	],
  	char: "♻️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var u6307 = {
  	keywords: [
  		"chinese",
  		"point",
  		"green-square",
  		"kanji"
  	],
  	char: "🈯",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var chart$1 = {
  	keywords: [
  		"green-square",
  		"graph",
  		"presentation",
  		"stats"
  	],
  	char: "💹",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var sparkle$1 = {
  	keywords: [
  		"stars",
  		"green-square",
  		"awesome",
  		"good",
  		"fireworks"
  	],
  	char: "❇️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var eight_spoked_asterisk$1 = {
  	keywords: [
  		"star",
  		"sparkle",
  		"green-square"
  	],
  	char: "✳️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var negative_squared_cross_mark$1 = {
  	keywords: [
  		"x",
  		"green-square",
  		"no",
  		"deny"
  	],
  	char: "❎",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var white_check_mark$1 = {
  	keywords: [
  		"green-square",
  		"ok",
  		"agree",
  		"vote",
  		"election",
  		"answer",
  		"tick"
  	],
  	char: "✅",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var diamond_shape_with_a_dot_inside$1 = {
  	keywords: [
  		"jewel",
  		"blue",
  		"gem",
  		"crystal",
  		"fancy"
  	],
  	char: "💠",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var cyclone$1 = {
  	keywords: [
  		"weather",
  		"swirl",
  		"blue",
  		"cloud",
  		"vortex",
  		"spiral",
  		"whirlpool",
  		"spin",
  		"tornado",
  		"hurricane",
  		"typhoon"
  	],
  	char: "🌀",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var loop$1 = {
  	keywords: [
  		"tape",
  		"cassette"
  	],
  	char: "➿",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var globe_with_meridians$1 = {
  	keywords: [
  		"earth",
  		"international",
  		"world",
  		"internet",
  		"interweb",
  		"i18n"
  	],
  	char: "🌐",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var m$1 = {
  	keywords: [
  		"alphabet",
  		"blue-circle",
  		"letter"
  	],
  	char: "Ⓜ️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var atm$1 = {
  	keywords: [
  		"money",
  		"sales",
  		"cash",
  		"blue-square",
  		"payment",
  		"bank"
  	],
  	char: "🏧",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var sa$1 = {
  	keywords: [
  		"japanese",
  		"blue-square",
  		"katakana"
  	],
  	char: "🈂️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var passport_control$1 = {
  	keywords: [
  		"custom",
  		"blue-square"
  	],
  	char: "🛂",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var customs$1 = {
  	keywords: [
  		"passport",
  		"border",
  		"blue-square"
  	],
  	char: "🛃",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var baggage_claim$1 = {
  	keywords: [
  		"blue-square",
  		"airport",
  		"transport"
  	],
  	char: "🛄",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var left_luggage$1 = {
  	keywords: [
  		"blue-square",
  		"travel"
  	],
  	char: "🛅",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var wheelchair$1 = {
  	keywords: [
  		"blue-square",
  		"disabled",
  		"a11y",
  		"accessibility"
  	],
  	char: "♿",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var no_smoking$1 = {
  	keywords: [
  		"cigarette",
  		"blue-square",
  		"smell",
  		"smoke"
  	],
  	char: "🚭",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var wc$1 = {
  	keywords: [
  		"toilet",
  		"restroom",
  		"blue-square"
  	],
  	char: "🚾",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var parking$1 = {
  	keywords: [
  		"cars",
  		"blue-square",
  		"alphabet",
  		"letter"
  	],
  	char: "🅿️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var potable_water$1 = {
  	keywords: [
  		"blue-square",
  		"liquid",
  		"restroom",
  		"cleaning",
  		"faucet"
  	],
  	char: "🚰",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var mens$1 = {
  	keywords: [
  		"toilet",
  		"restroom",
  		"wc",
  		"blue-square",
  		"gender",
  		"male"
  	],
  	char: "🚹",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var womens$1 = {
  	keywords: [
  		"purple-square",
  		"woman",
  		"female",
  		"toilet",
  		"loo",
  		"restroom",
  		"gender"
  	],
  	char: "🚺",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var baby_symbol$1 = {
  	keywords: [
  		"orange-square",
  		"child"
  	],
  	char: "🚼",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var restroom$1 = {
  	keywords: [
  		"blue-square",
  		"toilet",
  		"refresh",
  		"wc",
  		"gender"
  	],
  	char: "🚻",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var put_litter_in_its_place$1 = {
  	keywords: [
  		"blue-square",
  		"sign",
  		"human",
  		"info"
  	],
  	char: "🚮",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var cinema$1 = {
  	keywords: [
  		"blue-square",
  		"record",
  		"film",
  		"movie",
  		"curtain",
  		"stage",
  		"theater"
  	],
  	char: "🎦",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var signal_strength$1 = {
  	keywords: [
  		"blue-square",
  		"reception",
  		"phone",
  		"internet",
  		"connection",
  		"wifi",
  		"bluetooth",
  		"bars"
  	],
  	char: "📶",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var koko$1 = {
  	keywords: [
  		"blue-square",
  		"here",
  		"katakana",
  		"japanese",
  		"destination"
  	],
  	char: "🈁",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var ng$1 = {
  	keywords: [
  		"blue-square",
  		"words",
  		"shape",
  		"icon"
  	],
  	char: "🆖",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var ok$1 = {
  	keywords: [
  		"good",
  		"agree",
  		"yes",
  		"blue-square"
  	],
  	char: "🆗",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var up$1 = {
  	keywords: [
  		"blue-square",
  		"above",
  		"high"
  	],
  	char: "🆙",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var cool$1 = {
  	keywords: [
  		"words",
  		"blue-square"
  	],
  	char: "🆒",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var free$1 = {
  	keywords: [
  		"blue-square",
  		"words"
  	],
  	char: "🆓",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var zero$2 = {
  	keywords: [
  		"0",
  		"numbers",
  		"blue-square",
  		"null"
  	],
  	char: "0️⃣",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var one$1 = {
  	keywords: [
  		"blue-square",
  		"numbers",
  		"1"
  	],
  	char: "1️⃣",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var two$1 = {
  	keywords: [
  		"numbers",
  		"2",
  		"prime",
  		"blue-square"
  	],
  	char: "2️⃣",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var three$1 = {
  	keywords: [
  		"3",
  		"numbers",
  		"prime",
  		"blue-square"
  	],
  	char: "3️⃣",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var four$1 = {
  	keywords: [
  		"4",
  		"numbers",
  		"blue-square"
  	],
  	char: "4️⃣",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var five$1 = {
  	keywords: [
  		"5",
  		"numbers",
  		"blue-square",
  		"prime"
  	],
  	char: "5️⃣",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var six$1 = {
  	keywords: [
  		"6",
  		"numbers",
  		"blue-square"
  	],
  	char: "6️⃣",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var seven$1 = {
  	keywords: [
  		"7",
  		"numbers",
  		"blue-square",
  		"prime"
  	],
  	char: "7️⃣",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var eight$1 = {
  	keywords: [
  		"8",
  		"blue-square",
  		"numbers"
  	],
  	char: "8️⃣",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var nine$1 = {
  	keywords: [
  		"blue-square",
  		"numbers",
  		"9"
  	],
  	char: "9️⃣",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var keycap_ten$1 = {
  	keywords: [
  		"numbers",
  		"10",
  		"blue-square"
  	],
  	char: "🔟",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var asterisk$1 = {
  	keywords: [
  		"star",
  		"keycap"
  	],
  	char: "*⃣",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var eject_button$1 = {
  	keywords: [
  		"blue-square"
  	],
  	char: "⏏️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var arrow_forward$1 = {
  	keywords: [
  		"blue-square",
  		"right",
  		"direction",
  		"play"
  	],
  	char: "▶️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var pause_button$1 = {
  	keywords: [
  		"pause",
  		"blue-square"
  	],
  	char: "⏸",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var next_track_button$1 = {
  	keywords: [
  		"forward",
  		"next",
  		"blue-square"
  	],
  	char: "⏭",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var stop_button$1 = {
  	keywords: [
  		"blue-square"
  	],
  	char: "⏹",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var record_button$1 = {
  	keywords: [
  		"blue-square"
  	],
  	char: "⏺",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var play_or_pause_button$1 = {
  	keywords: [
  		"blue-square",
  		"play",
  		"pause"
  	],
  	char: "⏯",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var previous_track_button$1 = {
  	keywords: [
  		"backward"
  	],
  	char: "⏮",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var fast_forward$1 = {
  	keywords: [
  		"blue-square",
  		"play",
  		"speed",
  		"continue"
  	],
  	char: "⏩",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var rewind$1 = {
  	keywords: [
  		"play",
  		"blue-square"
  	],
  	char: "⏪",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var twisted_rightwards_arrows$1 = {
  	keywords: [
  		"blue-square",
  		"shuffle",
  		"music",
  		"random"
  	],
  	char: "🔀",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var repeat$1 = {
  	keywords: [
  		"loop",
  		"record"
  	],
  	char: "🔁",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var repeat_one$1 = {
  	keywords: [
  		"blue-square",
  		"loop"
  	],
  	char: "🔂",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var arrow_backward$1 = {
  	keywords: [
  		"blue-square",
  		"left",
  		"direction"
  	],
  	char: "◀️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var arrow_up_small$1 = {
  	keywords: [
  		"blue-square",
  		"triangle",
  		"direction",
  		"point",
  		"forward",
  		"top"
  	],
  	char: "🔼",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var arrow_down_small$1 = {
  	keywords: [
  		"blue-square",
  		"direction",
  		"bottom"
  	],
  	char: "🔽",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var arrow_double_up$1 = {
  	keywords: [
  		"blue-square",
  		"direction",
  		"top"
  	],
  	char: "⏫",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var arrow_double_down$1 = {
  	keywords: [
  		"blue-square",
  		"direction",
  		"bottom"
  	],
  	char: "⏬",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var arrow_right$1 = {
  	keywords: [
  		"blue-square",
  		"next"
  	],
  	char: "➡️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var arrow_left$1 = {
  	keywords: [
  		"blue-square",
  		"previous",
  		"back"
  	],
  	char: "⬅️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var arrow_up$1 = {
  	keywords: [
  		"blue-square",
  		"continue",
  		"top",
  		"direction"
  	],
  	char: "⬆️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var arrow_down$1 = {
  	keywords: [
  		"blue-square",
  		"direction",
  		"bottom"
  	],
  	char: "⬇️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var arrow_upper_right$1 = {
  	keywords: [
  		"blue-square",
  		"point",
  		"direction",
  		"diagonal",
  		"northeast"
  	],
  	char: "↗️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var arrow_lower_right$1 = {
  	keywords: [
  		"blue-square",
  		"direction",
  		"diagonal",
  		"southeast"
  	],
  	char: "↘️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var arrow_lower_left$1 = {
  	keywords: [
  		"blue-square",
  		"direction",
  		"diagonal",
  		"southwest"
  	],
  	char: "↙️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var arrow_upper_left$1 = {
  	keywords: [
  		"blue-square",
  		"point",
  		"direction",
  		"diagonal",
  		"northwest"
  	],
  	char: "↖️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var arrow_up_down$1 = {
  	keywords: [
  		"blue-square",
  		"direction",
  		"way",
  		"vertical"
  	],
  	char: "↕️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var left_right_arrow$1 = {
  	keywords: [
  		"shape",
  		"direction",
  		"horizontal",
  		"sideways"
  	],
  	char: "↔️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var arrows_counterclockwise$1 = {
  	keywords: [
  		"blue-square",
  		"sync",
  		"cycle"
  	],
  	char: "🔄",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var arrow_right_hook$1 = {
  	keywords: [
  		"blue-square",
  		"return",
  		"rotate",
  		"direction"
  	],
  	char: "↪️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var leftwards_arrow_with_hook$1 = {
  	keywords: [
  		"back",
  		"return",
  		"blue-square",
  		"undo",
  		"enter"
  	],
  	char: "↩️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var arrow_heading_up$1 = {
  	keywords: [
  		"blue-square",
  		"direction",
  		"top"
  	],
  	char: "⤴️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var arrow_heading_down$1 = {
  	keywords: [
  		"blue-square",
  		"direction",
  		"bottom"
  	],
  	char: "⤵️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var hash$1 = {
  	keywords: [
  		"symbol",
  		"blue-square",
  		"twitter"
  	],
  	char: "#️⃣",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var information_source$1 = {
  	keywords: [
  		"blue-square",
  		"alphabet",
  		"letter"
  	],
  	char: "ℹ️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var abc$1 = {
  	keywords: [
  		"blue-square",
  		"alphabet"
  	],
  	char: "🔤",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var abcd$1 = {
  	keywords: [
  		"blue-square",
  		"alphabet"
  	],
  	char: "🔡",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var capital_abcd$1 = {
  	keywords: [
  		"alphabet",
  		"words",
  		"blue-square"
  	],
  	char: "🔠",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var symbols$1 = {
  	keywords: [
  		"blue-square",
  		"music",
  		"note",
  		"ampersand",
  		"percent",
  		"glyphs",
  		"characters"
  	],
  	char: "🔣",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var musical_note$1 = {
  	keywords: [
  		"score",
  		"tone",
  		"sound"
  	],
  	char: "🎵",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var notes$1 = {
  	keywords: [
  		"music",
  		"score"
  	],
  	char: "🎶",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var wavy_dash$1 = {
  	keywords: [
  		"draw",
  		"line",
  		"moustache",
  		"mustache",
  		"squiggle",
  		"scribble"
  	],
  	char: "〰️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var curly_loop$1 = {
  	keywords: [
  		"scribble",
  		"draw",
  		"shape",
  		"squiggle"
  	],
  	char: "➰",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var heavy_check_mark$1 = {
  	keywords: [
  		"ok",
  		"nike",
  		"answer",
  		"yes",
  		"tick"
  	],
  	char: "✔️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var arrows_clockwise$1 = {
  	keywords: [
  		"sync",
  		"cycle",
  		"round",
  		"repeat"
  	],
  	char: "🔃",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var heavy_plus_sign$1 = {
  	keywords: [
  		"math",
  		"calculation",
  		"addition",
  		"more",
  		"increase"
  	],
  	char: "➕",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var heavy_minus_sign$1 = {
  	keywords: [
  		"math",
  		"calculation",
  		"subtract",
  		"less"
  	],
  	char: "➖",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var heavy_division_sign$1 = {
  	keywords: [
  		"divide",
  		"math",
  		"calculation"
  	],
  	char: "➗",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var heavy_multiplication_x$1 = {
  	keywords: [
  		"math",
  		"calculation"
  	],
  	char: "✖️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var infinity$1 = {
  	keywords: [
  		"forever"
  	],
  	char: "♾",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var heavy_dollar_sign$1 = {
  	keywords: [
  		"money",
  		"sales",
  		"payment",
  		"currency",
  		"buck"
  	],
  	char: "💲",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var currency_exchange$1 = {
  	keywords: [
  		"money",
  		"sales",
  		"dollar",
  		"travel"
  	],
  	char: "💱",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var copyright$1 = {
  	keywords: [
  		"ip",
  		"license",
  		"circle",
  		"law",
  		"legal"
  	],
  	char: "©️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var registered$1 = {
  	keywords: [
  		"alphabet",
  		"circle"
  	],
  	char: "®️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var tm$1 = {
  	keywords: [
  		"trademark",
  		"brand",
  		"law",
  		"legal"
  	],
  	char: "™️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var end$1 = {
  	keywords: [
  		"words",
  		"arrow"
  	],
  	char: "🔚",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var back$1 = {
  	keywords: [
  		"arrow",
  		"words",
  		"return"
  	],
  	char: "🔙",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var on$1 = {
  	keywords: [
  		"arrow",
  		"words"
  	],
  	char: "🔛",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var top$2 = {
  	keywords: [
  		"words",
  		"blue-square"
  	],
  	char: "🔝",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var soon$1 = {
  	keywords: [
  		"arrow",
  		"words"
  	],
  	char: "🔜",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var ballot_box_with_check$1 = {
  	keywords: [
  		"ok",
  		"agree",
  		"confirm",
  		"black-square",
  		"vote",
  		"election",
  		"yes",
  		"tick"
  	],
  	char: "☑️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var radio_button$1 = {
  	keywords: [
  		"input",
  		"old",
  		"music",
  		"circle"
  	],
  	char: "🔘",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var white_circle$1 = {
  	keywords: [
  		"shape",
  		"round"
  	],
  	char: "⚪",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var black_circle$1 = {
  	keywords: [
  		"shape",
  		"button",
  		"round"
  	],
  	char: "⚫",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var red_circle$1 = {
  	keywords: [
  		"shape",
  		"error",
  		"danger"
  	],
  	char: "🔴",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var large_blue_circle$1 = {
  	keywords: [
  		"shape",
  		"icon",
  		"button"
  	],
  	char: "🔵",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var small_orange_diamond$1 = {
  	keywords: [
  		"shape",
  		"jewel",
  		"gem"
  	],
  	char: "🔸",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var small_blue_diamond$1 = {
  	keywords: [
  		"shape",
  		"jewel",
  		"gem"
  	],
  	char: "🔹",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var large_orange_diamond$1 = {
  	keywords: [
  		"shape",
  		"jewel",
  		"gem"
  	],
  	char: "🔶",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var large_blue_diamond$1 = {
  	keywords: [
  		"shape",
  		"jewel",
  		"gem"
  	],
  	char: "🔷",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var small_red_triangle$1 = {
  	keywords: [
  		"shape",
  		"direction",
  		"up",
  		"top"
  	],
  	char: "🔺",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var black_small_square$1 = {
  	keywords: [
  		"shape",
  		"icon"
  	],
  	char: "▪️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var white_small_square$1 = {
  	keywords: [
  		"shape",
  		"icon"
  	],
  	char: "▫️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var black_large_square$1 = {
  	keywords: [
  		"shape",
  		"icon",
  		"button"
  	],
  	char: "⬛",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var white_large_square$1 = {
  	keywords: [
  		"shape",
  		"icon",
  		"stone",
  		"button"
  	],
  	char: "⬜",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var small_red_triangle_down$1 = {
  	keywords: [
  		"shape",
  		"direction",
  		"bottom"
  	],
  	char: "🔻",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var black_medium_square$1 = {
  	keywords: [
  		"shape",
  		"button",
  		"icon"
  	],
  	char: "◼️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var white_medium_square$1 = {
  	keywords: [
  		"shape",
  		"stone",
  		"icon"
  	],
  	char: "◻️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var black_medium_small_square$1 = {
  	keywords: [
  		"icon",
  		"shape",
  		"button"
  	],
  	char: "◾",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var white_medium_small_square$1 = {
  	keywords: [
  		"shape",
  		"stone",
  		"icon",
  		"button"
  	],
  	char: "◽",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var black_square_button$1 = {
  	keywords: [
  		"shape",
  		"input",
  		"frame"
  	],
  	char: "🔲",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var white_square_button$1 = {
  	keywords: [
  		"shape",
  		"input"
  	],
  	char: "🔳",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var speaker$1 = {
  	keywords: [
  		"sound",
  		"volume",
  		"silence",
  		"broadcast"
  	],
  	char: "🔈",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var sound$1 = {
  	keywords: [
  		"volume",
  		"speaker",
  		"broadcast"
  	],
  	char: "🔉",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var loud_sound$1 = {
  	keywords: [
  		"volume",
  		"noise",
  		"noisy",
  		"speaker",
  		"broadcast"
  	],
  	char: "🔊",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var mute$1 = {
  	keywords: [
  		"sound",
  		"volume",
  		"silence",
  		"quiet"
  	],
  	char: "🔇",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var mega$1 = {
  	keywords: [
  		"sound",
  		"speaker",
  		"volume"
  	],
  	char: "📣",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var loudspeaker$1 = {
  	keywords: [
  		"volume",
  		"sound"
  	],
  	char: "📢",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var bell$1 = {
  	keywords: [
  		"sound",
  		"notification",
  		"christmas",
  		"xmas",
  		"chime"
  	],
  	char: "🔔",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var no_bell$1 = {
  	keywords: [
  		"sound",
  		"volume",
  		"mute",
  		"quiet",
  		"silent"
  	],
  	char: "🔕",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var black_joker$1 = {
  	keywords: [
  		"poker",
  		"cards",
  		"game",
  		"play",
  		"magic"
  	],
  	char: "🃏",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var mahjong$1 = {
  	keywords: [
  		"game",
  		"play",
  		"chinese",
  		"kanji"
  	],
  	char: "🀄",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var spades$2 = {
  	keywords: [
  		"poker",
  		"cards",
  		"suits",
  		"magic"
  	],
  	char: "♠️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clubs$2 = {
  	keywords: [
  		"poker",
  		"cards",
  		"magic",
  		"suits"
  	],
  	char: "♣️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var hearts$2 = {
  	keywords: [
  		"poker",
  		"cards",
  		"magic",
  		"suits"
  	],
  	char: "♥️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var diamonds$1 = {
  	keywords: [
  		"poker",
  		"cards",
  		"magic",
  		"suits"
  	],
  	char: "♦️",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var flower_playing_cards$1 = {
  	keywords: [
  		"game",
  		"sunset",
  		"red"
  	],
  	char: "🎴",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var thought_balloon$1 = {
  	keywords: [
  		"bubble",
  		"cloud",
  		"speech",
  		"thinking",
  		"dream"
  	],
  	char: "💭",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var right_anger_bubble$1 = {
  	keywords: [
  		"caption",
  		"speech",
  		"thinking",
  		"mad"
  	],
  	char: "🗯",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var speech_balloon$1 = {
  	keywords: [
  		"bubble",
  		"words",
  		"message",
  		"talk",
  		"chatting"
  	],
  	char: "💬",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var left_speech_bubble$1 = {
  	keywords: [
  		"words",
  		"message",
  		"talk",
  		"chatting"
  	],
  	char: "🗨",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clock1$1 = {
  	keywords: [
  		"time",
  		"late",
  		"early",
  		"schedule"
  	],
  	char: "🕐",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clock2$1 = {
  	keywords: [
  		"time",
  		"late",
  		"early",
  		"schedule"
  	],
  	char: "🕑",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clock3$1 = {
  	keywords: [
  		"time",
  		"late",
  		"early",
  		"schedule"
  	],
  	char: "🕒",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clock4$1 = {
  	keywords: [
  		"time",
  		"late",
  		"early",
  		"schedule"
  	],
  	char: "🕓",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clock5$1 = {
  	keywords: [
  		"time",
  		"late",
  		"early",
  		"schedule"
  	],
  	char: "🕔",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clock6$1 = {
  	keywords: [
  		"time",
  		"late",
  		"early",
  		"schedule",
  		"dawn",
  		"dusk"
  	],
  	char: "🕕",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clock7$1 = {
  	keywords: [
  		"time",
  		"late",
  		"early",
  		"schedule"
  	],
  	char: "🕖",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clock8$1 = {
  	keywords: [
  		"time",
  		"late",
  		"early",
  		"schedule"
  	],
  	char: "🕗",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clock9$1 = {
  	keywords: [
  		"time",
  		"late",
  		"early",
  		"schedule"
  	],
  	char: "🕘",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clock10$1 = {
  	keywords: [
  		"time",
  		"late",
  		"early",
  		"schedule"
  	],
  	char: "🕙",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clock11$1 = {
  	keywords: [
  		"time",
  		"late",
  		"early",
  		"schedule"
  	],
  	char: "🕚",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clock12$1 = {
  	keywords: [
  		"time",
  		"noon",
  		"midnight",
  		"midday",
  		"late",
  		"early",
  		"schedule"
  	],
  	char: "🕛",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clock130$1 = {
  	keywords: [
  		"time",
  		"late",
  		"early",
  		"schedule"
  	],
  	char: "🕜",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clock230$1 = {
  	keywords: [
  		"time",
  		"late",
  		"early",
  		"schedule"
  	],
  	char: "🕝",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clock330$1 = {
  	keywords: [
  		"time",
  		"late",
  		"early",
  		"schedule"
  	],
  	char: "🕞",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clock430$1 = {
  	keywords: [
  		"time",
  		"late",
  		"early",
  		"schedule"
  	],
  	char: "🕟",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clock530$1 = {
  	keywords: [
  		"time",
  		"late",
  		"early",
  		"schedule"
  	],
  	char: "🕠",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clock630$1 = {
  	keywords: [
  		"time",
  		"late",
  		"early",
  		"schedule"
  	],
  	char: "🕡",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clock730$1 = {
  	keywords: [
  		"time",
  		"late",
  		"early",
  		"schedule"
  	],
  	char: "🕢",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clock830$1 = {
  	keywords: [
  		"time",
  		"late",
  		"early",
  		"schedule"
  	],
  	char: "🕣",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clock930$1 = {
  	keywords: [
  		"time",
  		"late",
  		"early",
  		"schedule"
  	],
  	char: "🕤",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clock1030$1 = {
  	keywords: [
  		"time",
  		"late",
  		"early",
  		"schedule"
  	],
  	char: "🕥",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clock1130$1 = {
  	keywords: [
  		"time",
  		"late",
  		"early",
  		"schedule"
  	],
  	char: "🕦",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var clock1230$1 = {
  	keywords: [
  		"time",
  		"late",
  		"early",
  		"schedule"
  	],
  	char: "🕧",
  	fitzpatrick_scale: false,
  	category: "symbols"
  };
  var afghanistan$1 = {
  	keywords: [
  		"af",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇦🇫",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var aland_islands$1 = {
  	keywords: [
  		"Åland",
  		"islands",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇦🇽",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var albania$1 = {
  	keywords: [
  		"al",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇦🇱",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var algeria$1 = {
  	keywords: [
  		"dz",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇩🇿",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var american_samoa$1 = {
  	keywords: [
  		"american",
  		"ws",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇦🇸",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var andorra$1 = {
  	keywords: [
  		"ad",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇦🇩",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var angola$1 = {
  	keywords: [
  		"ao",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇦🇴",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var anguilla$1 = {
  	keywords: [
  		"ai",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇦🇮",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var antarctica$1 = {
  	keywords: [
  		"aq",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇦🇶",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var antigua_barbuda$1 = {
  	keywords: [
  		"antigua",
  		"barbuda",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇦🇬",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var argentina$1 = {
  	keywords: [
  		"ar",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇦🇷",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var armenia$1 = {
  	keywords: [
  		"am",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇦🇲",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var aruba$1 = {
  	keywords: [
  		"aw",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇦🇼",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var australia$1 = {
  	keywords: [
  		"au",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇦🇺",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var austria$1 = {
  	keywords: [
  		"at",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇦🇹",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var azerbaijan$1 = {
  	keywords: [
  		"az",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇦🇿",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var bahamas$1 = {
  	keywords: [
  		"bs",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇧🇸",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var bahrain$1 = {
  	keywords: [
  		"bh",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇧🇭",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var bangladesh$1 = {
  	keywords: [
  		"bd",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇧🇩",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var barbados$1 = {
  	keywords: [
  		"bb",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇧🇧",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var belarus$1 = {
  	keywords: [
  		"by",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇧🇾",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var belgium$1 = {
  	keywords: [
  		"be",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇧🇪",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var belize$1 = {
  	keywords: [
  		"bz",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇧🇿",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var benin$1 = {
  	keywords: [
  		"bj",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇧🇯",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var bermuda$1 = {
  	keywords: [
  		"bm",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇧🇲",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var bhutan$1 = {
  	keywords: [
  		"bt",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇧🇹",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var bolivia$1 = {
  	keywords: [
  		"bo",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇧🇴",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var caribbean_netherlands$1 = {
  	keywords: [
  		"bonaire",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇧🇶",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var bosnia_herzegovina$1 = {
  	keywords: [
  		"bosnia",
  		"herzegovina",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇧🇦",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var botswana$1 = {
  	keywords: [
  		"bw",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇧🇼",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var brazil$1 = {
  	keywords: [
  		"br",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇧🇷",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var british_indian_ocean_territory$1 = {
  	keywords: [
  		"british",
  		"indian",
  		"ocean",
  		"territory",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇮🇴",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var british_virgin_islands$1 = {
  	keywords: [
  		"british",
  		"virgin",
  		"islands",
  		"bvi",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇻🇬",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var brunei$1 = {
  	keywords: [
  		"bn",
  		"darussalam",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇧🇳",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var bulgaria$1 = {
  	keywords: [
  		"bg",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇧🇬",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var burkina_faso$1 = {
  	keywords: [
  		"burkina",
  		"faso",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇧🇫",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var burundi$1 = {
  	keywords: [
  		"bi",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇧🇮",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var cape_verde$1 = {
  	keywords: [
  		"cabo",
  		"verde",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇨🇻",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var cambodia$1 = {
  	keywords: [
  		"kh",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇰🇭",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var cameroon$1 = {
  	keywords: [
  		"cm",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇨🇲",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var canada$1 = {
  	keywords: [
  		"ca",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇨🇦",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var canary_islands$1 = {
  	keywords: [
  		"canary",
  		"islands",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇮🇨",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var cayman_islands$1 = {
  	keywords: [
  		"cayman",
  		"islands",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇰🇾",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var central_african_republic$1 = {
  	keywords: [
  		"central",
  		"african",
  		"republic",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇨🇫",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var chad$1 = {
  	keywords: [
  		"td",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇹🇩",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var chile$1 = {
  	keywords: [
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇨🇱",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var cn$1 = {
  	keywords: [
  		"china",
  		"chinese",
  		"prc",
  		"flag",
  		"country",
  		"nation",
  		"banner"
  	],
  	char: "🇨🇳",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var christmas_island$1 = {
  	keywords: [
  		"christmas",
  		"island",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇨🇽",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var cocos_islands$1 = {
  	keywords: [
  		"cocos",
  		"keeling",
  		"islands",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇨🇨",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var colombia$1 = {
  	keywords: [
  		"co",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇨🇴",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var comoros$1 = {
  	keywords: [
  		"km",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇰🇲",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var congo_brazzaville$1 = {
  	keywords: [
  		"congo",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇨🇬",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var congo_kinshasa$1 = {
  	keywords: [
  		"congo",
  		"democratic",
  		"republic",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇨🇩",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var cook_islands$1 = {
  	keywords: [
  		"cook",
  		"islands",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇨🇰",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var costa_rica$1 = {
  	keywords: [
  		"costa",
  		"rica",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇨🇷",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var croatia$1 = {
  	keywords: [
  		"hr",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇭🇷",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var cuba$1 = {
  	keywords: [
  		"cu",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇨🇺",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var curacao$1 = {
  	keywords: [
  		"curaçao",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇨🇼",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var cyprus$1 = {
  	keywords: [
  		"cy",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇨🇾",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var czech_republic$1 = {
  	keywords: [
  		"cz",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇨🇿",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var denmark$1 = {
  	keywords: [
  		"dk",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇩🇰",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var djibouti$1 = {
  	keywords: [
  		"dj",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇩🇯",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var dominica$1 = {
  	keywords: [
  		"dm",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇩🇲",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var dominican_republic$1 = {
  	keywords: [
  		"dominican",
  		"republic",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇩🇴",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var ecuador$1 = {
  	keywords: [
  		"ec",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇪🇨",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var egypt$1 = {
  	keywords: [
  		"eg",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇪🇬",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var el_salvador$1 = {
  	keywords: [
  		"el",
  		"salvador",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇸🇻",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var equatorial_guinea$1 = {
  	keywords: [
  		"equatorial",
  		"gn",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇬🇶",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var eritrea$1 = {
  	keywords: [
  		"er",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇪🇷",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var estonia$1 = {
  	keywords: [
  		"ee",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇪🇪",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var ethiopia$1 = {
  	keywords: [
  		"et",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇪🇹",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var eu$1 = {
  	keywords: [
  		"european",
  		"union",
  		"flag",
  		"banner"
  	],
  	char: "🇪🇺",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var falkland_islands$1 = {
  	keywords: [
  		"falkland",
  		"islands",
  		"malvinas",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇫🇰",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var faroe_islands$1 = {
  	keywords: [
  		"faroe",
  		"islands",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇫🇴",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var fiji$1 = {
  	keywords: [
  		"fj",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇫🇯",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var finland$1 = {
  	keywords: [
  		"fi",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇫🇮",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var fr$1 = {
  	keywords: [
  		"banner",
  		"flag",
  		"nation",
  		"france",
  		"french",
  		"country"
  	],
  	char: "🇫🇷",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var french_guiana$1 = {
  	keywords: [
  		"french",
  		"guiana",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇬🇫",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var french_polynesia$1 = {
  	keywords: [
  		"french",
  		"polynesia",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇵🇫",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var french_southern_territories$1 = {
  	keywords: [
  		"french",
  		"southern",
  		"territories",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇹🇫",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var gabon$1 = {
  	keywords: [
  		"ga",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇬🇦",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var gambia$1 = {
  	keywords: [
  		"gm",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇬🇲",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var georgia$1 = {
  	keywords: [
  		"ge",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇬🇪",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var de$1 = {
  	keywords: [
  		"german",
  		"nation",
  		"flag",
  		"country",
  		"banner"
  	],
  	char: "🇩🇪",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var ghana$1 = {
  	keywords: [
  		"gh",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇬🇭",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var gibraltar$1 = {
  	keywords: [
  		"gi",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇬🇮",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var greece$1 = {
  	keywords: [
  		"gr",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇬🇷",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var greenland$1 = {
  	keywords: [
  		"gl",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇬🇱",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var grenada$1 = {
  	keywords: [
  		"gd",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇬🇩",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var guadeloupe$1 = {
  	keywords: [
  		"gp",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇬🇵",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var guam$1 = {
  	keywords: [
  		"gu",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇬🇺",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var guatemala$1 = {
  	keywords: [
  		"gt",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇬🇹",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var guernsey$1 = {
  	keywords: [
  		"gg",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇬🇬",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var guinea$1 = {
  	keywords: [
  		"gn",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇬🇳",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var guinea_bissau$1 = {
  	keywords: [
  		"gw",
  		"bissau",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇬🇼",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var guyana$1 = {
  	keywords: [
  		"gy",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇬🇾",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var haiti$1 = {
  	keywords: [
  		"ht",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇭🇹",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var honduras$1 = {
  	keywords: [
  		"hn",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇭🇳",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var hong_kong$1 = {
  	keywords: [
  		"hong",
  		"kong",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇭🇰",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var hungary$1 = {
  	keywords: [
  		"hu",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇭🇺",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var iceland$1 = {
  	keywords: [
  		"is",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇮🇸",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var india$1 = {
  	keywords: [
  		"in",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇮🇳",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var indonesia$1 = {
  	keywords: [
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇮🇩",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var iran$1 = {
  	keywords: [
  		"iran,",
  		"islamic",
  		"republic",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇮🇷",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var iraq$1 = {
  	keywords: [
  		"iq",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇮🇶",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var ireland$1 = {
  	keywords: [
  		"ie",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇮🇪",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var isle_of_man$1 = {
  	keywords: [
  		"isle",
  		"man",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇮🇲",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var israel$1 = {
  	keywords: [
  		"il",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇮🇱",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var it$2 = {
  	keywords: [
  		"italy",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇮🇹",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var cote_divoire$1 = {
  	keywords: [
  		"ivory",
  		"coast",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇨🇮",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var jamaica$1 = {
  	keywords: [
  		"jm",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇯🇲",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var jp$1 = {
  	keywords: [
  		"japanese",
  		"nation",
  		"flag",
  		"country",
  		"banner"
  	],
  	char: "🇯🇵",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var jersey$1 = {
  	keywords: [
  		"je",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇯🇪",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var jordan$1 = {
  	keywords: [
  		"jo",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇯🇴",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var kazakhstan$1 = {
  	keywords: [
  		"kz",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇰🇿",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var kenya$1 = {
  	keywords: [
  		"ke",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇰🇪",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var kiribati$1 = {
  	keywords: [
  		"ki",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇰🇮",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var kosovo$1 = {
  	keywords: [
  		"xk",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇽🇰",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var kuwait$1 = {
  	keywords: [
  		"kw",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇰🇼",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var kyrgyzstan$1 = {
  	keywords: [
  		"kg",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇰🇬",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var laos$1 = {
  	keywords: [
  		"lao",
  		"democratic",
  		"republic",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇱🇦",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var latvia$1 = {
  	keywords: [
  		"lv",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇱🇻",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var lebanon$1 = {
  	keywords: [
  		"lb",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇱🇧",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var lesotho$1 = {
  	keywords: [
  		"ls",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇱🇸",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var liberia$1 = {
  	keywords: [
  		"lr",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇱🇷",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var libya$1 = {
  	keywords: [
  		"ly",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇱🇾",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var liechtenstein$1 = {
  	keywords: [
  		"li",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇱🇮",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var lithuania$1 = {
  	keywords: [
  		"lt",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇱🇹",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var luxembourg$1 = {
  	keywords: [
  		"lu",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇱🇺",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var macau$1 = {
  	keywords: [
  		"macao",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇲🇴",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var macedonia$1 = {
  	keywords: [
  		"macedonia,",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇲🇰",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var madagascar$1 = {
  	keywords: [
  		"mg",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇲🇬",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var malawi$1 = {
  	keywords: [
  		"mw",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇲🇼",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var malaysia$1 = {
  	keywords: [
  		"my",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇲🇾",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var maldives$1 = {
  	keywords: [
  		"mv",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇲🇻",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var mali$1 = {
  	keywords: [
  		"ml",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇲🇱",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var malta$1 = {
  	keywords: [
  		"mt",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇲🇹",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var marshall_islands$1 = {
  	keywords: [
  		"marshall",
  		"islands",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇲🇭",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var martinique$1 = {
  	keywords: [
  		"mq",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇲🇶",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var mauritania$1 = {
  	keywords: [
  		"mr",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇲🇷",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var mauritius$1 = {
  	keywords: [
  		"mu",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇲🇺",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var mayotte$1 = {
  	keywords: [
  		"yt",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇾🇹",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var mexico$1 = {
  	keywords: [
  		"mx",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇲🇽",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var micronesia$1 = {
  	keywords: [
  		"micronesia,",
  		"federated",
  		"states",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇫🇲",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var moldova$1 = {
  	keywords: [
  		"moldova,",
  		"republic",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇲🇩",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var monaco$1 = {
  	keywords: [
  		"mc",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇲🇨",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var mongolia$1 = {
  	keywords: [
  		"mn",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇲🇳",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var montenegro$1 = {
  	keywords: [
  		"me",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇲🇪",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var montserrat$1 = {
  	keywords: [
  		"ms",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇲🇸",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var morocco$1 = {
  	keywords: [
  		"ma",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇲🇦",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var mozambique$1 = {
  	keywords: [
  		"mz",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇲🇿",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var myanmar$1 = {
  	keywords: [
  		"mm",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇲🇲",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var namibia$1 = {
  	keywords: [
  		"na",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇳🇦",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var nauru$1 = {
  	keywords: [
  		"nr",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇳🇷",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var nepal$1 = {
  	keywords: [
  		"np",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇳🇵",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var netherlands$1 = {
  	keywords: [
  		"nl",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇳🇱",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var new_caledonia$1 = {
  	keywords: [
  		"new",
  		"caledonia",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇳🇨",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var new_zealand$1 = {
  	keywords: [
  		"new",
  		"zealand",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇳🇿",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var nicaragua$1 = {
  	keywords: [
  		"ni",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇳🇮",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var niger$1 = {
  	keywords: [
  		"ne",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇳🇪",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var nigeria$1 = {
  	keywords: [
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇳🇬",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var niue$1 = {
  	keywords: [
  		"nu",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇳🇺",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var norfolk_island$1 = {
  	keywords: [
  		"norfolk",
  		"island",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇳🇫",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var northern_mariana_islands$1 = {
  	keywords: [
  		"northern",
  		"mariana",
  		"islands",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇲🇵",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var north_korea$1 = {
  	keywords: [
  		"north",
  		"korea",
  		"nation",
  		"flag",
  		"country",
  		"banner"
  	],
  	char: "🇰🇵",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var norway$1 = {
  	keywords: [
  		"no",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇳🇴",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var oman$1 = {
  	keywords: [
  		"om_symbol",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇴🇲",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var pakistan$1 = {
  	keywords: [
  		"pk",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇵🇰",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var palau$1 = {
  	keywords: [
  		"pw",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇵🇼",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var palestinian_territories$1 = {
  	keywords: [
  		"palestine",
  		"palestinian",
  		"territories",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇵🇸",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var panama$1 = {
  	keywords: [
  		"pa",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇵🇦",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var papua_new_guinea$1 = {
  	keywords: [
  		"papua",
  		"new",
  		"guinea",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇵🇬",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var paraguay$1 = {
  	keywords: [
  		"py",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇵🇾",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var peru$1 = {
  	keywords: [
  		"pe",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇵🇪",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var philippines$1 = {
  	keywords: [
  		"ph",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇵🇭",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var pitcairn_islands$1 = {
  	keywords: [
  		"pitcairn",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇵🇳",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var poland$1 = {
  	keywords: [
  		"pl",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇵🇱",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var portugal$1 = {
  	keywords: [
  		"pt",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇵🇹",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var puerto_rico$1 = {
  	keywords: [
  		"puerto",
  		"rico",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇵🇷",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var qatar$1 = {
  	keywords: [
  		"qa",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇶🇦",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var reunion$1 = {
  	keywords: [
  		"réunion",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇷🇪",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var romania$1 = {
  	keywords: [
  		"ro",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇷🇴",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var ru$1 = {
  	keywords: [
  		"russian",
  		"federation",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇷🇺",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var rwanda$1 = {
  	keywords: [
  		"rw",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇷🇼",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var st_barthelemy$1 = {
  	keywords: [
  		"saint",
  		"barthélemy",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇧🇱",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var st_helena$1 = {
  	keywords: [
  		"saint",
  		"helena",
  		"ascension",
  		"tristan",
  		"cunha",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇸🇭",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var st_kitts_nevis$1 = {
  	keywords: [
  		"saint",
  		"kitts",
  		"nevis",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇰🇳",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var st_lucia$1 = {
  	keywords: [
  		"saint",
  		"lucia",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇱🇨",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var st_pierre_miquelon$1 = {
  	keywords: [
  		"saint",
  		"pierre",
  		"miquelon",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇵🇲",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var st_vincent_grenadines$1 = {
  	keywords: [
  		"saint",
  		"vincent",
  		"grenadines",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇻🇨",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var samoa$1 = {
  	keywords: [
  		"ws",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇼🇸",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var san_marino$1 = {
  	keywords: [
  		"san",
  		"marino",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇸🇲",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var sao_tome_principe$1 = {
  	keywords: [
  		"sao",
  		"tome",
  		"principe",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇸🇹",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var saudi_arabia$1 = {
  	keywords: [
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇸🇦",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var senegal$1 = {
  	keywords: [
  		"sn",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇸🇳",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var serbia$1 = {
  	keywords: [
  		"rs",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇷🇸",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var seychelles$1 = {
  	keywords: [
  		"sc",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇸🇨",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var sierra_leone$1 = {
  	keywords: [
  		"sierra",
  		"leone",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇸🇱",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var singapore$1 = {
  	keywords: [
  		"sg",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇸🇬",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var sint_maarten$1 = {
  	keywords: [
  		"sint",
  		"maarten",
  		"dutch",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇸🇽",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var slovakia$1 = {
  	keywords: [
  		"sk",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇸🇰",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var slovenia$1 = {
  	keywords: [
  		"si",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇸🇮",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var solomon_islands$1 = {
  	keywords: [
  		"solomon",
  		"islands",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇸🇧",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var somalia$1 = {
  	keywords: [
  		"so",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇸🇴",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var south_africa$1 = {
  	keywords: [
  		"south",
  		"africa",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇿🇦",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var south_georgia_south_sandwich_islands$1 = {
  	keywords: [
  		"south",
  		"georgia",
  		"sandwich",
  		"islands",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇬🇸",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var kr$1 = {
  	keywords: [
  		"south",
  		"korea",
  		"nation",
  		"flag",
  		"country",
  		"banner"
  	],
  	char: "🇰🇷",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var south_sudan$1 = {
  	keywords: [
  		"south",
  		"sd",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇸🇸",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var es$1 = {
  	keywords: [
  		"spain",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇪🇸",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var sri_lanka$1 = {
  	keywords: [
  		"sri",
  		"lanka",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇱🇰",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var sudan$1 = {
  	keywords: [
  		"sd",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇸🇩",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var suriname$1 = {
  	keywords: [
  		"sr",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇸🇷",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var swaziland$1 = {
  	keywords: [
  		"sz",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇸🇿",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var sweden$1 = {
  	keywords: [
  		"se",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇸🇪",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var switzerland$1 = {
  	keywords: [
  		"ch",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇨🇭",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var syria$1 = {
  	keywords: [
  		"syrian",
  		"arab",
  		"republic",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇸🇾",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var taiwan$1 = {
  	keywords: [
  		"tw",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇹🇼",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var tajikistan$1 = {
  	keywords: [
  		"tj",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇹🇯",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var tanzania$1 = {
  	keywords: [
  		"tanzania,",
  		"united",
  		"republic",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇹🇿",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var thailand$1 = {
  	keywords: [
  		"th",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇹🇭",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var timor_leste$1 = {
  	keywords: [
  		"timor",
  		"leste",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇹🇱",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var togo$1 = {
  	keywords: [
  		"tg",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇹🇬",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var tokelau$1 = {
  	keywords: [
  		"tk",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇹🇰",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var tonga$1 = {
  	keywords: [
  		"to",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇹🇴",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var trinidad_tobago$1 = {
  	keywords: [
  		"trinidad",
  		"tobago",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇹🇹",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var tunisia$1 = {
  	keywords: [
  		"tn",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇹🇳",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var tr$1 = {
  	keywords: [
  		"turkey",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇹🇷",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var turkmenistan$1 = {
  	keywords: [
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇹🇲",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var turks_caicos_islands$1 = {
  	keywords: [
  		"turks",
  		"caicos",
  		"islands",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇹🇨",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var tuvalu$1 = {
  	keywords: [
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇹🇻",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var uganda$1 = {
  	keywords: [
  		"ug",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇺🇬",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var ukraine$1 = {
  	keywords: [
  		"ua",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇺🇦",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var united_arab_emirates$1 = {
  	keywords: [
  		"united",
  		"arab",
  		"emirates",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇦🇪",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var uk$1 = {
  	keywords: [
  		"united",
  		"kingdom",
  		"great",
  		"britain",
  		"northern",
  		"ireland",
  		"flag",
  		"nation",
  		"country",
  		"banner",
  		"british",
  		"UK",
  		"english",
  		"england",
  		"union jack"
  	],
  	char: "🇬🇧",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var england$1 = {
  	keywords: [
  		"flag",
  		"english"
  	],
  	char: "🏴󠁧󠁢󠁥󠁮󠁧󠁿",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var scotland$1 = {
  	keywords: [
  		"flag",
  		"scottish"
  	],
  	char: "🏴󠁧󠁢󠁳󠁣󠁴󠁿",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var wales$1 = {
  	keywords: [
  		"flag",
  		"welsh"
  	],
  	char: "🏴󠁧󠁢󠁷󠁬󠁳󠁿",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var us$1 = {
  	keywords: [
  		"united",
  		"states",
  		"america",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇺🇸",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var us_virgin_islands$1 = {
  	keywords: [
  		"virgin",
  		"islands",
  		"us",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇻🇮",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var uruguay$1 = {
  	keywords: [
  		"uy",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇺🇾",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var uzbekistan$1 = {
  	keywords: [
  		"uz",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇺🇿",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var vanuatu$1 = {
  	keywords: [
  		"vu",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇻🇺",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var vatican_city$1 = {
  	keywords: [
  		"vatican",
  		"city",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇻🇦",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var venezuela$1 = {
  	keywords: [
  		"ve",
  		"bolivarian",
  		"republic",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇻🇪",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var vietnam$1 = {
  	keywords: [
  		"viet",
  		"nam",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇻🇳",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var wallis_futuna$1 = {
  	keywords: [
  		"wallis",
  		"futuna",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇼🇫",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var western_sahara$1 = {
  	keywords: [
  		"western",
  		"sahara",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇪🇭",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var yemen$1 = {
  	keywords: [
  		"ye",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇾🇪",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var zambia$1 = {
  	keywords: [
  		"zm",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇿🇲",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var zimbabwe$1 = {
  	keywords: [
  		"zw",
  		"flag",
  		"nation",
  		"country",
  		"banner"
  	],
  	char: "🇿🇼",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var united_nations$1 = {
  	keywords: [
  		"un",
  		"flag",
  		"banner"
  	],
  	char: "🇺🇳",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var pirate_flag$1 = {
  	keywords: [
  		"skull",
  		"crossbones",
  		"flag",
  		"banner"
  	],
  	char: "🏴‍☠️",
  	fitzpatrick_scale: false,
  	category: "flags"
  };
  var emojis = {
  	"100": {
  	keywords: [
  		"score",
  		"perfect",
  		"numbers",
  		"century",
  		"exam",
  		"quiz",
  		"test",
  		"pass",
  		"hundred"
  	],
  	char: "💯",
  	fitzpatrick_scale: false,
  	category: "symbols"
  },
  	"1234": {
  	keywords: [
  		"numbers",
  		"blue-square"
  	],
  	char: "🔢",
  	fitzpatrick_scale: false,
  	category: "symbols"
  },
  	grinning: grinning$1,
  	grimacing: grimacing$1,
  	grin: grin$1,
  	joy: joy$1,
  	rofl: rofl$1,
  	partying: partying,
  	smiley: smiley$1,
  	smile: smile$2,
  	sweat_smile: sweat_smile$1,
  	laughing: laughing$1,
  	innocent: innocent$1,
  	wink: wink$1,
  	blush: blush$1,
  	slightly_smiling_face: slightly_smiling_face$1,
  	upside_down_face: upside_down_face$1,
  	relaxed: relaxed$1,
  	yum: yum$1,
  	relieved: relieved$1,
  	heart_eyes: heart_eyes$1,
  	smiling_face_with_three_hearts: smiling_face_with_three_hearts$1,
  	kissing_heart: kissing_heart$1,
  	kissing: kissing$1,
  	kissing_smiling_eyes: kissing_smiling_eyes$1,
  	kissing_closed_eyes: kissing_closed_eyes$1,
  	stuck_out_tongue_winking_eye: stuck_out_tongue_winking_eye$1,
  	zany: zany,
  	raised_eyebrow: raised_eyebrow$1,
  	monocle: monocle,
  	stuck_out_tongue_closed_eyes: stuck_out_tongue_closed_eyes$1,
  	stuck_out_tongue: stuck_out_tongue$1,
  	money_mouth_face: money_mouth_face$1,
  	nerd_face: nerd_face$1,
  	sunglasses: sunglasses$1,
  	star_struck: star_struck$1,
  	clown_face: clown_face$1,
  	cowboy_hat_face: cowboy_hat_face$1,
  	hugs: hugs$1,
  	smirk: smirk$1,
  	no_mouth: no_mouth$1,
  	neutral_face: neutral_face$1,
  	expressionless: expressionless$1,
  	unamused: unamused$1,
  	roll_eyes: roll_eyes$1,
  	thinking: thinking$1,
  	lying_face: lying_face$1,
  	hand_over_mouth: hand_over_mouth$1,
  	shushing: shushing,
  	symbols_over_mouth: symbols_over_mouth,
  	exploding_head: exploding_head$1,
  	flushed: flushed$1,
  	disappointed: disappointed$1,
  	worried: worried$1,
  	angry: angry$1,
  	rage: rage$1,
  	pensive: pensive$1,
  	confused: confused$1,
  	slightly_frowning_face: slightly_frowning_face$1,
  	frowning_face: frowning_face$1,
  	persevere: persevere$1,
  	confounded: confounded$1,
  	tired_face: tired_face$1,
  	weary: weary$1,
  	pleading: pleading,
  	triumph: triumph$1,
  	open_mouth: open_mouth$1,
  	scream: scream$1,
  	fearful: fearful$1,
  	cold_sweat: cold_sweat$1,
  	hushed: hushed$1,
  	frowning: frowning$1,
  	anguished: anguished$1,
  	cry: cry$1,
  	disappointed_relieved: disappointed_relieved$1,
  	drooling_face: drooling_face$1,
  	sleepy: sleepy$1,
  	sweat: sweat$1,
  	hot: hot,
  	cold: cold,
  	sob: sob$1,
  	dizzy_face: dizzy_face$1,
  	astonished: astonished$1,
  	zipper_mouth_face: zipper_mouth_face$1,
  	nauseated_face: nauseated_face$1,
  	sneezing_face: sneezing_face$1,
  	vomiting: vomiting,
  	mask: mask$1,
  	face_with_thermometer: face_with_thermometer$1,
  	face_with_head_bandage: face_with_head_bandage$1,
  	woozy: woozy,
  	sleeping: sleeping$1,
  	zzz: zzz$1,
  	poop: poop$1,
  	smiling_imp: smiling_imp$1,
  	imp: imp$1,
  	japanese_ogre: japanese_ogre$1,
  	japanese_goblin: japanese_goblin$1,
  	skull: skull$1,
  	ghost: ghost$1,
  	alien: alien$1,
  	robot: robot$1,
  	smiley_cat: smiley_cat$1,
  	smile_cat: smile_cat$1,
  	joy_cat: joy_cat$1,
  	heart_eyes_cat: heart_eyes_cat$1,
  	smirk_cat: smirk_cat$1,
  	kissing_cat: kissing_cat$1,
  	scream_cat: scream_cat$1,
  	crying_cat_face: crying_cat_face$1,
  	pouting_cat: pouting_cat$1,
  	palms_up: palms_up,
  	raised_hands: raised_hands$1,
  	clap: clap$1,
  	wave: wave$1,
  	call_me_hand: call_me_hand$1,
  	"+1": {
  	keywords: [
  		"thumbsup",
  		"yes",
  		"awesome",
  		"good",
  		"agree",
  		"accept",
  		"cool",
  		"hand",
  		"like"
  	],
  	char: "👍",
  	fitzpatrick_scale: true,
  	category: "people"
  },
  	"-1": {
  	keywords: [
  		"thumbsdown",
  		"no",
  		"dislike",
  		"hand"
  	],
  	char: "👎",
  	fitzpatrick_scale: true,
  	category: "people"
  },
  	facepunch: facepunch$1,
  	fist: fist$1,
  	fist_left: fist_left$1,
  	fist_right: fist_right$1,
  	v: v$1,
  	ok_hand: ok_hand$1,
  	raised_hand: raised_hand$1,
  	raised_back_of_hand: raised_back_of_hand$1,
  	open_hands: open_hands$1,
  	muscle: muscle$1,
  	pray: pray$1,
  	foot: foot$1,
  	leg: leg$2,
  	handshake: handshake$1,
  	point_up: point_up$1,
  	point_up_2: point_up_2$1,
  	point_down: point_down$1,
  	point_left: point_left$1,
  	point_right: point_right$1,
  	fu: fu$1,
  	raised_hand_with_fingers_splayed: raised_hand_with_fingers_splayed$1,
  	love_you: love_you,
  	metal: metal$1,
  	crossed_fingers: crossed_fingers$1,
  	vulcan_salute: vulcan_salute$1,
  	writing_hand: writing_hand$1,
  	selfie: selfie$1,
  	nail_care: nail_care$1,
  	lips: lips$1,
  	tooth: tooth$1,
  	tongue: tongue$1,
  	ear: ear$1,
  	nose: nose$1,
  	eye: eye$1,
  	eyes: eyes$1,
  	brain: brain$1,
  	bust_in_silhouette: bust_in_silhouette$1,
  	busts_in_silhouette: busts_in_silhouette$1,
  	speaking_head: speaking_head$1,
  	baby: baby$1,
  	child: child$1,
  	boy: boy$1,
  	girl: girl$1,
  	adult: adult$1,
  	man: man$1,
  	woman: woman$1,
  	blonde_woman: blonde_woman$1,
  	blonde_man: blonde_man,
  	bearded_person: bearded_person$1,
  	older_adult: older_adult$1,
  	older_man: older_man$1,
  	older_woman: older_woman$1,
  	man_with_gua_pi_mao: man_with_gua_pi_mao$1,
  	woman_with_headscarf: woman_with_headscarf$1,
  	woman_with_turban: woman_with_turban$1,
  	man_with_turban: man_with_turban$1,
  	policewoman: policewoman$1,
  	policeman: policeman$1,
  	construction_worker_woman: construction_worker_woman$1,
  	construction_worker_man: construction_worker_man$1,
  	guardswoman: guardswoman$1,
  	guardsman: guardsman$1,
  	female_detective: female_detective$1,
  	male_detective: male_detective$1,
  	woman_health_worker: woman_health_worker$1,
  	man_health_worker: man_health_worker$1,
  	woman_farmer: woman_farmer$1,
  	man_farmer: man_farmer$1,
  	woman_cook: woman_cook$1,
  	man_cook: man_cook$1,
  	woman_student: woman_student$1,
  	man_student: man_student$1,
  	woman_singer: woman_singer$1,
  	man_singer: man_singer$1,
  	woman_teacher: woman_teacher$1,
  	man_teacher: man_teacher$1,
  	woman_factory_worker: woman_factory_worker$1,
  	man_factory_worker: man_factory_worker$1,
  	woman_technologist: woman_technologist$1,
  	man_technologist: man_technologist$1,
  	woman_office_worker: woman_office_worker$1,
  	man_office_worker: man_office_worker$1,
  	woman_mechanic: woman_mechanic$1,
  	man_mechanic: man_mechanic$1,
  	woman_scientist: woman_scientist$1,
  	man_scientist: man_scientist$1,
  	woman_artist: woman_artist$1,
  	man_artist: man_artist$1,
  	woman_firefighter: woman_firefighter$1,
  	man_firefighter: man_firefighter$1,
  	woman_pilot: woman_pilot$1,
  	man_pilot: man_pilot$1,
  	woman_astronaut: woman_astronaut$1,
  	man_astronaut: man_astronaut$1,
  	woman_judge: woman_judge$1,
  	man_judge: man_judge$1,
  	woman_superhero: woman_superhero,
  	man_superhero: man_superhero,
  	woman_supervillain: woman_supervillain,
  	man_supervillain: man_supervillain,
  	mrs_claus: mrs_claus$1,
  	santa: santa$1,
  	sorceress: sorceress,
  	wizard: wizard,
  	woman_elf: woman_elf,
  	man_elf: man_elf,
  	woman_vampire: woman_vampire,
  	man_vampire: man_vampire,
  	woman_zombie: woman_zombie,
  	man_zombie: man_zombie,
  	woman_genie: woman_genie,
  	man_genie: man_genie,
  	mermaid: mermaid$1,
  	merman: merman$1,
  	woman_fairy: woman_fairy,
  	man_fairy: man_fairy,
  	angel: angel$1,
  	pregnant_woman: pregnant_woman$1,
  	breastfeeding: breastfeeding,
  	princess: princess$1,
  	prince: prince$1,
  	bride_with_veil: bride_with_veil$1,
  	man_in_tuxedo: man_in_tuxedo$1,
  	running_woman: running_woman$1,
  	running_man: running_man$1,
  	walking_woman: walking_woman$1,
  	walking_man: walking_man$1,
  	dancer: dancer$1,
  	man_dancing: man_dancing$1,
  	dancing_women: dancing_women$1,
  	dancing_men: dancing_men$1,
  	couple: couple$1,
  	two_men_holding_hands: two_men_holding_hands$1,
  	two_women_holding_hands: two_women_holding_hands$1,
  	bowing_woman: bowing_woman$1,
  	bowing_man: bowing_man$1,
  	man_facepalming: man_facepalming$1,
  	woman_facepalming: woman_facepalming$1,
  	woman_shrugging: woman_shrugging$1,
  	man_shrugging: man_shrugging$1,
  	tipping_hand_woman: tipping_hand_woman$1,
  	tipping_hand_man: tipping_hand_man$1,
  	no_good_woman: no_good_woman$1,
  	no_good_man: no_good_man$1,
  	ok_woman: ok_woman$1,
  	ok_man: ok_man$1,
  	raising_hand_woman: raising_hand_woman$1,
  	raising_hand_man: raising_hand_man$1,
  	pouting_woman: pouting_woman$1,
  	pouting_man: pouting_man$1,
  	frowning_woman: frowning_woman$1,
  	frowning_man: frowning_man$1,
  	haircut_woman: haircut_woman$1,
  	haircut_man: haircut_man$1,
  	massage_woman: massage_woman$1,
  	massage_man: massage_man$1,
  	woman_in_steamy_room: woman_in_steamy_room,
  	man_in_steamy_room: man_in_steamy_room,
  	couple_with_heart_woman_man: couple_with_heart_woman_man$1,
  	couple_with_heart_woman_woman: couple_with_heart_woman_woman$1,
  	couple_with_heart_man_man: couple_with_heart_man_man$1,
  	couplekiss_man_woman: couplekiss_man_woman$1,
  	couplekiss_woman_woman: couplekiss_woman_woman$1,
  	couplekiss_man_man: couplekiss_man_man$1,
  	family_man_woman_boy: family_man_woman_boy$1,
  	family_man_woman_girl: family_man_woman_girl$1,
  	family_man_woman_girl_boy: family_man_woman_girl_boy$1,
  	family_man_woman_boy_boy: family_man_woman_boy_boy$1,
  	family_man_woman_girl_girl: family_man_woman_girl_girl$1,
  	family_woman_woman_boy: family_woman_woman_boy$1,
  	family_woman_woman_girl: family_woman_woman_girl$1,
  	family_woman_woman_girl_boy: family_woman_woman_girl_boy$1,
  	family_woman_woman_boy_boy: family_woman_woman_boy_boy$1,
  	family_woman_woman_girl_girl: family_woman_woman_girl_girl$1,
  	family_man_man_boy: family_man_man_boy$1,
  	family_man_man_girl: family_man_man_girl$1,
  	family_man_man_girl_boy: family_man_man_girl_boy$1,
  	family_man_man_boy_boy: family_man_man_boy_boy$1,
  	family_man_man_girl_girl: family_man_man_girl_girl$1,
  	family_woman_boy: family_woman_boy$1,
  	family_woman_girl: family_woman_girl$1,
  	family_woman_girl_boy: family_woman_girl_boy$1,
  	family_woman_boy_boy: family_woman_boy_boy$1,
  	family_woman_girl_girl: family_woman_girl_girl$1,
  	family_man_boy: family_man_boy$1,
  	family_man_girl: family_man_girl$1,
  	family_man_girl_boy: family_man_girl_boy$1,
  	family_man_boy_boy: family_man_boy_boy$1,
  	family_man_girl_girl: family_man_girl_girl$1,
  	yarn: yarn$1,
  	thread: thread$1,
  	coat: coat$1,
  	labcoat: labcoat,
  	womans_clothes: womans_clothes$1,
  	tshirt: tshirt$1,
  	jeans: jeans$1,
  	necktie: necktie$1,
  	dress: dress$1,
  	bikini: bikini$1,
  	kimono: kimono$1,
  	lipstick: lipstick$1,
  	kiss: kiss$1,
  	footprints: footprints$1,
  	flat_shoe: flat_shoe$1,
  	high_heel: high_heel$1,
  	sandal: sandal$1,
  	boot: boot$1,
  	mans_shoe: mans_shoe$1,
  	athletic_shoe: athletic_shoe$1,
  	hiking_boot: hiking_boot$1,
  	socks: socks$1,
  	gloves: gloves$1,
  	scarf: scarf$1,
  	womans_hat: womans_hat$1,
  	tophat: tophat$1,
  	billed_hat: billed_hat,
  	rescue_worker_helmet: rescue_worker_helmet$1,
  	mortar_board: mortar_board$1,
  	crown: crown$1,
  	school_satchel: school_satchel$1,
  	luggage: luggage$1,
  	pouch: pouch$1,
  	purse: purse$1,
  	handbag: handbag$1,
  	briefcase: briefcase$1,
  	eyeglasses: eyeglasses$1,
  	dark_sunglasses: dark_sunglasses$1,
  	goggles: goggles$1,
  	ring: ring$2,
  	closed_umbrella: closed_umbrella$1,
  	dog: dog$1,
  	cat: cat$1,
  	mouse: mouse$1,
  	hamster: hamster$1,
  	rabbit: rabbit$1,
  	fox_face: fox_face$1,
  	bear: bear$1,
  	panda_face: panda_face$1,
  	koala: koala$1,
  	tiger: tiger$1,
  	lion: lion$1,
  	cow: cow$1,
  	pig: pig$1,
  	pig_nose: pig_nose$1,
  	frog: frog$1,
  	squid: squid$1,
  	octopus: octopus$1,
  	shrimp: shrimp$1,
  	monkey_face: monkey_face$1,
  	gorilla: gorilla$1,
  	see_no_evil: see_no_evil$1,
  	hear_no_evil: hear_no_evil$1,
  	speak_no_evil: speak_no_evil$1,
  	monkey: monkey$1,
  	chicken: chicken$1,
  	penguin: penguin$1,
  	bird: bird$1,
  	baby_chick: baby_chick$1,
  	hatching_chick: hatching_chick$1,
  	hatched_chick: hatched_chick$1,
  	duck: duck$1,
  	eagle: eagle$1,
  	owl: owl$1,
  	bat: bat$1,
  	wolf: wolf$1,
  	boar: boar$1,
  	horse: horse$1,
  	unicorn: unicorn$1,
  	honeybee: honeybee$1,
  	bug: bug$1,
  	butterfly: butterfly$1,
  	snail: snail$1,
  	beetle: beetle$1,
  	ant: ant$1,
  	grasshopper: grasshopper,
  	spider: spider$1,
  	scorpion: scorpion$1,
  	crab: crab$1,
  	snake: snake$1,
  	lizard: lizard$1,
  	"t-rex": {
  	keywords: [
  		"animal",
  		"nature",
  		"dinosaur",
  		"tyrannosaurus",
  		"extinct"
  	],
  	char: "🦖",
  	fitzpatrick_scale: false,
  	category: "animals_and_nature"
  },
  	sauropod: sauropod$1,
  	turtle: turtle$1,
  	tropical_fish: tropical_fish$1,
  	fish: fish$1,
  	blowfish: blowfish$1,
  	dolphin: dolphin$1,
  	shark: shark$1,
  	whale: whale$1,
  	whale2: whale2$1,
  	crocodile: crocodile$1,
  	leopard: leopard$1,
  	zebra: zebra$1,
  	tiger2: tiger2$1,
  	water_buffalo: water_buffalo$1,
  	ox: ox$1,
  	cow2: cow2$1,
  	deer: deer$1,
  	dromedary_camel: dromedary_camel$1,
  	camel: camel$1,
  	giraffe: giraffe$1,
  	elephant: elephant$1,
  	rhinoceros: rhinoceros$1,
  	goat: goat$1,
  	ram: ram$1,
  	sheep: sheep$1,
  	racehorse: racehorse$1,
  	pig2: pig2$1,
  	rat: rat$1,
  	mouse2: mouse2$1,
  	rooster: rooster$1,
  	turkey: turkey$1,
  	dove: dove$1,
  	dog2: dog2$1,
  	poodle: poodle$1,
  	cat2: cat2$1,
  	rabbit2: rabbit2$1,
  	chipmunk: chipmunk$1,
  	hedgehog: hedgehog$1,
  	raccoon: raccoon$1,
  	llama: llama$1,
  	hippopotamus: hippopotamus$1,
  	kangaroo: kangaroo$1,
  	badger: badger$1,
  	swan: swan$1,
  	peacock: peacock$1,
  	parrot: parrot$1,
  	lobster: lobster$1,
  	mosquito: mosquito$1,
  	paw_prints: paw_prints$1,
  	dragon: dragon$1,
  	dragon_face: dragon_face$1,
  	cactus: cactus$1,
  	christmas_tree: christmas_tree$1,
  	evergreen_tree: evergreen_tree$1,
  	deciduous_tree: deciduous_tree$1,
  	palm_tree: palm_tree$1,
  	seedling: seedling$1,
  	herb: herb$1,
  	shamrock: shamrock$1,
  	four_leaf_clover: four_leaf_clover$1,
  	bamboo: bamboo$1,
  	tanabata_tree: tanabata_tree$1,
  	leaves: leaves$1,
  	fallen_leaf: fallen_leaf$1,
  	maple_leaf: maple_leaf$1,
  	ear_of_rice: ear_of_rice$1,
  	hibiscus: hibiscus$1,
  	sunflower: sunflower$1,
  	rose: rose$1,
  	wilted_flower: wilted_flower$1,
  	tulip: tulip$1,
  	blossom: blossom$1,
  	cherry_blossom: cherry_blossom$1,
  	bouquet: bouquet$1,
  	mushroom: mushroom$1,
  	chestnut: chestnut$1,
  	jack_o_lantern: jack_o_lantern$1,
  	shell: shell$1,
  	spider_web: spider_web$1,
  	earth_americas: earth_americas$1,
  	earth_africa: earth_africa$1,
  	earth_asia: earth_asia$1,
  	full_moon: full_moon$1,
  	waning_gibbous_moon: waning_gibbous_moon$1,
  	last_quarter_moon: last_quarter_moon$1,
  	waning_crescent_moon: waning_crescent_moon$1,
  	new_moon: new_moon$1,
  	waxing_crescent_moon: waxing_crescent_moon$1,
  	first_quarter_moon: first_quarter_moon$1,
  	waxing_gibbous_moon: waxing_gibbous_moon$1,
  	new_moon_with_face: new_moon_with_face$1,
  	full_moon_with_face: full_moon_with_face$1,
  	first_quarter_moon_with_face: first_quarter_moon_with_face$1,
  	last_quarter_moon_with_face: last_quarter_moon_with_face$1,
  	sun_with_face: sun_with_face$1,
  	crescent_moon: crescent_moon$1,
  	star: star$2,
  	star2: star2$1,
  	dizzy: dizzy$1,
  	sparkles: sparkles$1,
  	comet: comet$1,
  	sunny: sunny$1,
  	sun_behind_small_cloud: sun_behind_small_cloud$1,
  	partly_sunny: partly_sunny$1,
  	sun_behind_large_cloud: sun_behind_large_cloud$1,
  	sun_behind_rain_cloud: sun_behind_rain_cloud$1,
  	cloud: cloud$1,
  	cloud_with_rain: cloud_with_rain$1,
  	cloud_with_lightning_and_rain: cloud_with_lightning_and_rain$1,
  	cloud_with_lightning: cloud_with_lightning$1,
  	zap: zap$1,
  	fire: fire$1,
  	boom: boom$1,
  	snowflake: snowflake$1,
  	cloud_with_snow: cloud_with_snow$1,
  	snowman: snowman$1,
  	snowman_with_snow: snowman_with_snow$1,
  	wind_face: wind_face$1,
  	dash: dash$2,
  	tornado: tornado$1,
  	fog: fog$1,
  	open_umbrella: open_umbrella$1,
  	umbrella: umbrella$1,
  	droplet: droplet$1,
  	sweat_drops: sweat_drops$1,
  	ocean: ocean$1,
  	green_apple: green_apple$1,
  	apple: apple$1,
  	pear: pear$1,
  	tangerine: tangerine$1,
  	lemon: lemon$1,
  	banana: banana$1,
  	watermelon: watermelon$1,
  	grapes: grapes$1,
  	strawberry: strawberry$1,
  	melon: melon$1,
  	cherries: cherries$1,
  	peach: peach$1,
  	pineapple: pineapple$1,
  	coconut: coconut$1,
  	kiwi_fruit: kiwi_fruit$1,
  	mango: mango$1,
  	avocado: avocado$1,
  	broccoli: broccoli$1,
  	tomato: tomato$1,
  	eggplant: eggplant$1,
  	cucumber: cucumber$1,
  	carrot: carrot$1,
  	hot_pepper: hot_pepper$1,
  	potato: potato$1,
  	corn: corn$1,
  	leafy_greens: leafy_greens,
  	sweet_potato: sweet_potato$1,
  	peanuts: peanuts$1,
  	honey_pot: honey_pot$1,
  	croissant: croissant$1,
  	bread: bread$1,
  	baguette_bread: baguette_bread$1,
  	bagel: bagel$1,
  	pretzel: pretzel$1,
  	cheese: cheese$1,
  	egg: egg$1,
  	bacon: bacon$1,
  	steak: steak,
  	pancakes: pancakes$1,
  	poultry_leg: poultry_leg$1,
  	meat_on_bone: meat_on_bone$1,
  	bone: bone$1,
  	fried_shrimp: fried_shrimp$1,
  	fried_egg: fried_egg$1,
  	hamburger: hamburger$1,
  	fries: fries$1,
  	stuffed_flatbread: stuffed_flatbread$1,
  	hotdog: hotdog$1,
  	pizza: pizza$1,
  	sandwich: sandwich$1,
  	canned_food: canned_food$1,
  	spaghetti: spaghetti$1,
  	taco: taco$1,
  	burrito: burrito$1,
  	green_salad: green_salad$1,
  	shallow_pan_of_food: shallow_pan_of_food$1,
  	ramen: ramen$1,
  	stew: stew$1,
  	fish_cake: fish_cake$1,
  	fortune_cookie: fortune_cookie$1,
  	sushi: sushi$1,
  	bento: bento$1,
  	curry: curry$1,
  	rice_ball: rice_ball$1,
  	rice: rice$1,
  	rice_cracker: rice_cracker$1,
  	oden: oden$1,
  	dango: dango$1,
  	shaved_ice: shaved_ice$1,
  	ice_cream: ice_cream$1,
  	icecream: icecream$1,
  	pie: pie$1,
  	cake: cake$1,
  	cupcake: cupcake$1,
  	moon_cake: moon_cake$1,
  	birthday: birthday$1,
  	custard: custard$1,
  	candy: candy$1,
  	lollipop: lollipop$1,
  	chocolate_bar: chocolate_bar$1,
  	popcorn: popcorn$1,
  	dumpling: dumpling$1,
  	doughnut: doughnut$1,
  	cookie: cookie$1,
  	milk_glass: milk_glass$1,
  	beer: beer$1,
  	beers: beers$1,
  	clinking_glasses: clinking_glasses$1,
  	wine_glass: wine_glass$1,
  	tumbler_glass: tumbler_glass$1,
  	cocktail: cocktail$1,
  	tropical_drink: tropical_drink$1,
  	champagne: champagne$1,
  	sake: sake$1,
  	tea: tea$1,
  	cup_with_straw: cup_with_straw$1,
  	coffee: coffee$1,
  	baby_bottle: baby_bottle$1,
  	salt: salt$1,
  	spoon: spoon$1,
  	fork_and_knife: fork_and_knife$1,
  	plate_with_cutlery: plate_with_cutlery$1,
  	bowl_with_spoon: bowl_with_spoon$1,
  	takeout_box: takeout_box$1,
  	chopsticks: chopsticks$1,
  	soccer: soccer$1,
  	basketball: basketball$1,
  	football: football$1,
  	baseball: baseball$1,
  	softball: softball$1,
  	tennis: tennis$1,
  	volleyball: volleyball$1,
  	rugby_football: rugby_football$1,
  	flying_disc: flying_disc$1,
  	"8ball": {
  	keywords: [
  		"pool",
  		"hobby",
  		"game",
  		"luck",
  		"magic"
  	],
  	char: "🎱",
  	fitzpatrick_scale: false,
  	category: "activity"
  },
  	golf: golf$1,
  	golfing_woman: golfing_woman$1,
  	golfing_man: golfing_man$1,
  	ping_pong: ping_pong$1,
  	badminton: badminton$1,
  	goal_net: goal_net$1,
  	ice_hockey: ice_hockey$1,
  	field_hockey: field_hockey$1,
  	lacrosse: lacrosse$1,
  	cricket: cricket$1,
  	ski: ski$1,
  	skier: skier$1,
  	snowboarder: snowboarder$1,
  	person_fencing: person_fencing$1,
  	women_wrestling: women_wrestling$1,
  	men_wrestling: men_wrestling$1,
  	woman_cartwheeling: woman_cartwheeling$1,
  	man_cartwheeling: man_cartwheeling$1,
  	woman_playing_handball: woman_playing_handball$1,
  	man_playing_handball: man_playing_handball$1,
  	ice_skate: ice_skate$1,
  	curling_stone: curling_stone$1,
  	skateboard: skateboard$1,
  	sled: sled$1,
  	bow_and_arrow: bow_and_arrow$1,
  	fishing_pole_and_fish: fishing_pole_and_fish$1,
  	boxing_glove: boxing_glove$1,
  	martial_arts_uniform: martial_arts_uniform$1,
  	rowing_woman: rowing_woman$1,
  	rowing_man: rowing_man$1,
  	climbing_woman: climbing_woman$1,
  	climbing_man: climbing_man$1,
  	swimming_woman: swimming_woman$1,
  	swimming_man: swimming_man$1,
  	woman_playing_water_polo: woman_playing_water_polo$1,
  	man_playing_water_polo: man_playing_water_polo$1,
  	woman_in_lotus_position: woman_in_lotus_position,
  	man_in_lotus_position: man_in_lotus_position,
  	surfing_woman: surfing_woman$1,
  	surfing_man: surfing_man$1,
  	bath: bath$1,
  	basketball_woman: basketball_woman$1,
  	basketball_man: basketball_man$1,
  	weight_lifting_woman: weight_lifting_woman$1,
  	weight_lifting_man: weight_lifting_man$1,
  	biking_woman: biking_woman$1,
  	biking_man: biking_man$1,
  	mountain_biking_woman: mountain_biking_woman$1,
  	mountain_biking_man: mountain_biking_man$1,
  	horse_racing: horse_racing$1,
  	business_suit_levitating: business_suit_levitating$1,
  	trophy: trophy$1,
  	running_shirt_with_sash: running_shirt_with_sash$1,
  	medal_sports: medal_sports$1,
  	medal_military: medal_military$1,
  	"1st_place_medal": {
  	keywords: [
  		"award",
  		"winning",
  		"first"
  	],
  	char: "🥇",
  	fitzpatrick_scale: false,
  	category: "activity"
  },
  	"2nd_place_medal": {
  	keywords: [
  		"award",
  		"second"
  	],
  	char: "🥈",
  	fitzpatrick_scale: false,
  	category: "activity"
  },
  	"3rd_place_medal": {
  	keywords: [
  		"award",
  		"third"
  	],
  	char: "🥉",
  	fitzpatrick_scale: false,
  	category: "activity"
  },
  	reminder_ribbon: reminder_ribbon$1,
  	rosette: rosette$1,
  	ticket: ticket$1,
  	tickets: tickets$1,
  	performing_arts: performing_arts$1,
  	art: art$1,
  	circus_tent: circus_tent$1,
  	woman_juggling: woman_juggling$1,
  	man_juggling: man_juggling$1,
  	microphone: microphone$1,
  	headphones: headphones$1,
  	musical_score: musical_score$1,
  	musical_keyboard: musical_keyboard$1,
  	drum: drum$1,
  	saxophone: saxophone$1,
  	trumpet: trumpet$1,
  	guitar: guitar$1,
  	violin: violin$1,
  	clapper: clapper$1,
  	video_game: video_game$1,
  	space_invader: space_invader$1,
  	dart: dart$1,
  	game_die: game_die$1,
  	chess_pawn: chess_pawn$1,
  	slot_machine: slot_machine$1,
  	jigsaw: jigsaw$1,
  	bowling: bowling$1,
  	red_car: red_car$1,
  	taxi: taxi$1,
  	blue_car: blue_car$1,
  	bus: bus$1,
  	trolleybus: trolleybus$1,
  	racing_car: racing_car$1,
  	police_car: police_car$1,
  	ambulance: ambulance$1,
  	fire_engine: fire_engine$1,
  	minibus: minibus$1,
  	truck: truck$1,
  	articulated_lorry: articulated_lorry$1,
  	tractor: tractor$1,
  	kick_scooter: kick_scooter$1,
  	motorcycle: motorcycle$1,
  	bike: bike$1,
  	motor_scooter: motor_scooter$1,
  	rotating_light: rotating_light$1,
  	oncoming_police_car: oncoming_police_car$1,
  	oncoming_bus: oncoming_bus$1,
  	oncoming_automobile: oncoming_automobile$1,
  	oncoming_taxi: oncoming_taxi$1,
  	aerial_tramway: aerial_tramway$1,
  	mountain_cableway: mountain_cableway$1,
  	suspension_railway: suspension_railway$1,
  	railway_car: railway_car$1,
  	train: train$1,
  	monorail: monorail$1,
  	bullettrain_side: bullettrain_side$1,
  	bullettrain_front: bullettrain_front$1,
  	light_rail: light_rail$1,
  	mountain_railway: mountain_railway$1,
  	steam_locomotive: steam_locomotive$1,
  	train2: train2$1,
  	metro: metro$1,
  	tram: tram$1,
  	station: station$1,
  	flying_saucer: flying_saucer$1,
  	helicopter: helicopter$1,
  	small_airplane: small_airplane$1,
  	airplane: airplane$1,
  	flight_departure: flight_departure$1,
  	flight_arrival: flight_arrival$1,
  	sailboat: sailboat$1,
  	motor_boat: motor_boat$1,
  	speedboat: speedboat$1,
  	ferry: ferry$1,
  	passenger_ship: passenger_ship$1,
  	rocket: rocket$1,
  	artificial_satellite: artificial_satellite$1,
  	seat: seat$1,
  	canoe: canoe$1,
  	anchor: anchor$1,
  	construction: construction$1,
  	fuelpump: fuelpump$1,
  	busstop: busstop$1,
  	vertical_traffic_light: vertical_traffic_light$1,
  	traffic_light: traffic_light$1,
  	checkered_flag: checkered_flag$1,
  	ship: ship$1,
  	ferris_wheel: ferris_wheel$1,
  	roller_coaster: roller_coaster$1,
  	carousel_horse: carousel_horse$1,
  	building_construction: building_construction$1,
  	foggy: foggy$1,
  	tokyo_tower: tokyo_tower$1,
  	factory: factory$1,
  	fountain: fountain$1,
  	rice_scene: rice_scene$1,
  	mountain: mountain$1,
  	mountain_snow: mountain_snow$1,
  	mount_fuji: mount_fuji$1,
  	volcano: volcano$1,
  	japan: japan$1,
  	camping: camping$1,
  	tent: tent$1,
  	national_park: national_park$1,
  	motorway: motorway$1,
  	railway_track: railway_track$1,
  	sunrise: sunrise$1,
  	sunrise_over_mountains: sunrise_over_mountains$1,
  	desert: desert$1,
  	beach_umbrella: beach_umbrella$1,
  	desert_island: desert_island$1,
  	city_sunrise: city_sunrise$1,
  	city_sunset: city_sunset$1,
  	cityscape: cityscape$1,
  	night_with_stars: night_with_stars$1,
  	bridge_at_night: bridge_at_night$1,
  	milky_way: milky_way$1,
  	stars: stars$1,
  	sparkler: sparkler$1,
  	fireworks: fireworks$1,
  	rainbow: rainbow$1,
  	houses: houses$1,
  	european_castle: european_castle$1,
  	japanese_castle: japanese_castle$1,
  	stadium: stadium$1,
  	statue_of_liberty: statue_of_liberty$1,
  	house: house$1,
  	house_with_garden: house_with_garden$1,
  	derelict_house: derelict_house$1,
  	office: office$1,
  	department_store: department_store$1,
  	post_office: post_office$1,
  	european_post_office: european_post_office$1,
  	hospital: hospital$1,
  	bank: bank$1,
  	hotel: hotel$1,
  	convenience_store: convenience_store$1,
  	school: school$1,
  	love_hotel: love_hotel$1,
  	wedding: wedding$1,
  	classical_building: classical_building$1,
  	church: church$1,
  	mosque: mosque$1,
  	synagogue: synagogue$1,
  	kaaba: kaaba$1,
  	shinto_shrine: shinto_shrine$1,
  	watch: watch$1,
  	iphone: iphone$1,
  	calling: calling$1,
  	computer: computer$1,
  	keyboard: keyboard$1,
  	desktop_computer: desktop_computer$1,
  	printer: printer$1,
  	computer_mouse: computer_mouse$1,
  	trackball: trackball$1,
  	joystick: joystick$1,
  	clamp: clamp$1,
  	minidisc: minidisc$1,
  	floppy_disk: floppy_disk$1,
  	cd: cd$1,
  	dvd: dvd$1,
  	vhs: vhs$1,
  	camera: camera$1,
  	camera_flash: camera_flash$1,
  	video_camera: video_camera$1,
  	movie_camera: movie_camera$1,
  	film_projector: film_projector$1,
  	film_strip: film_strip$1,
  	telephone_receiver: telephone_receiver$1,
  	phone: phone$2,
  	pager: pager$1,
  	fax: fax$1,
  	tv: tv$1,
  	radio: radio$1,
  	studio_microphone: studio_microphone$1,
  	level_slider: level_slider$1,
  	control_knobs: control_knobs$1,
  	compass: compass$1,
  	stopwatch: stopwatch$1,
  	timer_clock: timer_clock$1,
  	alarm_clock: alarm_clock$1,
  	mantelpiece_clock: mantelpiece_clock$1,
  	hourglass_flowing_sand: hourglass_flowing_sand$1,
  	hourglass: hourglass$1,
  	satellite: satellite$1,
  	battery: battery$1,
  	electric_plug: electric_plug$1,
  	bulb: bulb$1,
  	flashlight: flashlight$1,
  	candle: candle$1,
  	fire_extinguisher: fire_extinguisher$1,
  	wastebasket: wastebasket$1,
  	oil_drum: oil_drum$1,
  	money_with_wings: money_with_wings$1,
  	dollar: dollar$2,
  	yen: yen$2,
  	euro: euro$2,
  	pound: pound$2,
  	moneybag: moneybag$1,
  	credit_card: credit_card$1,
  	gem: gem$1,
  	balance_scale: balance_scale$1,
  	toolbox: toolbox$1,
  	wrench: wrench$1,
  	hammer: hammer$1,
  	hammer_and_pick: hammer_and_pick$1,
  	hammer_and_wrench: hammer_and_wrench$1,
  	pick: pick$1,
  	nut_and_bolt: nut_and_bolt$1,
  	gear: gear$1,
  	brick: brick,
  	chains: chains$1,
  	magnet: magnet$1,
  	gun: gun$1,
  	bomb: bomb$1,
  	firecracker: firecracker$1,
  	hocho: hocho$1,
  	dagger: dagger$2,
  	crossed_swords: crossed_swords$1,
  	shield: shield$1,
  	smoking: smoking$1,
  	skull_and_crossbones: skull_and_crossbones$1,
  	coffin: coffin$1,
  	funeral_urn: funeral_urn$1,
  	amphora: amphora$1,
  	crystal_ball: crystal_ball$1,
  	prayer_beads: prayer_beads$1,
  	nazar_amulet: nazar_amulet$1,
  	barber: barber$1,
  	alembic: alembic$1,
  	telescope: telescope$1,
  	microscope: microscope$1,
  	hole: hole$1,
  	pill: pill$1,
  	syringe: syringe$1,
  	dna: dna$1,
  	microbe: microbe$1,
  	petri_dish: petri_dish$1,
  	test_tube: test_tube$1,
  	thermometer: thermometer$1,
  	broom: broom$1,
  	basket: basket$1,
  	toilet_paper: toilet_paper,
  	label: label$1,
  	bookmark: bookmark$1,
  	toilet: toilet$1,
  	shower: shower$1,
  	bathtub: bathtub$1,
  	soap: soap$1,
  	sponge: sponge$1,
  	lotion_bottle: lotion_bottle$1,
  	key: key$1,
  	old_key: old_key$1,
  	couch_and_lamp: couch_and_lamp$1,
  	sleeping_bed: sleeping_bed$1,
  	bed: bed$1,
  	door: door$1,
  	bellhop_bell: bellhop_bell$1,
  	teddy_bear: teddy_bear$1,
  	framed_picture: framed_picture$1,
  	world_map: world_map$1,
  	parasol_on_ground: parasol_on_ground$1,
  	moyai: moyai$1,
  	shopping: shopping$1,
  	shopping_cart: shopping_cart$1,
  	balloon: balloon$1,
  	flags: flags$1,
  	ribbon: ribbon$1,
  	gift: gift$1,
  	confetti_ball: confetti_ball$1,
  	tada: tada$1,
  	dolls: dolls$1,
  	wind_chime: wind_chime$1,
  	crossed_flags: crossed_flags$1,
  	izakaya_lantern: izakaya_lantern$1,
  	red_envelope: red_envelope$1,
  	email: email$1,
  	envelope_with_arrow: envelope_with_arrow$1,
  	incoming_envelope: incoming_envelope$1,
  	"e-mail": {
  	keywords: [
  		"communication",
  		"inbox"
  	],
  	char: "📧",
  	fitzpatrick_scale: false,
  	category: "objects"
  },
  	love_letter: love_letter$1,
  	postbox: postbox$1,
  	mailbox_closed: mailbox_closed$1,
  	mailbox: mailbox$1,
  	mailbox_with_mail: mailbox_with_mail$1,
  	mailbox_with_no_mail: mailbox_with_no_mail$1,
  	"package": {
  	keywords: [
  		"mail",
  		"gift",
  		"cardboard",
  		"box",
  		"moving"
  	],
  	char: "📦",
  	fitzpatrick_scale: false,
  	category: "objects"
  },
  	postal_horn: postal_horn$1,
  	inbox_tray: inbox_tray$1,
  	outbox_tray: outbox_tray$1,
  	scroll: scroll$1,
  	page_with_curl: page_with_curl$1,
  	bookmark_tabs: bookmark_tabs$1,
  	receipt: receipt$1,
  	bar_chart: bar_chart$1,
  	chart_with_upwards_trend: chart_with_upwards_trend$1,
  	chart_with_downwards_trend: chart_with_downwards_trend$1,
  	page_facing_up: page_facing_up$1,
  	date: date$1,
  	calendar: calendar$1,
  	spiral_calendar: spiral_calendar$1,
  	card_index: card_index$1,
  	card_file_box: card_file_box$1,
  	ballot_box: ballot_box$1,
  	file_cabinet: file_cabinet$1,
  	clipboard: clipboard$2,
  	spiral_notepad: spiral_notepad$1,
  	file_folder: file_folder$1,
  	open_file_folder: open_file_folder$1,
  	card_index_dividers: card_index_dividers$1,
  	newspaper_roll: newspaper_roll$1,
  	newspaper: newspaper$1,
  	notebook: notebook$1,
  	closed_book: closed_book$1,
  	green_book: green_book$1,
  	blue_book: blue_book$1,
  	orange_book: orange_book$1,
  	notebook_with_decorative_cover: notebook_with_decorative_cover$1,
  	ledger: ledger$1,
  	books: books$1,
  	open_book: open_book$1,
  	safety_pin: safety_pin$1,
  	link: link$3,
  	paperclip: paperclip$1,
  	paperclips: paperclips$1,
  	scissors: scissors$1,
  	triangular_ruler: triangular_ruler$1,
  	straight_ruler: straight_ruler$1,
  	abacus: abacus$1,
  	pushpin: pushpin$1,
  	round_pushpin: round_pushpin$1,
  	triangular_flag_on_post: triangular_flag_on_post$1,
  	white_flag: white_flag$1,
  	black_flag: black_flag$1,
  	rainbow_flag: rainbow_flag$1,
  	closed_lock_with_key: closed_lock_with_key$1,
  	lock: lock$1,
  	unlock: unlock$1,
  	lock_with_ink_pen: lock_with_ink_pen$1,
  	pen: pen$1,
  	fountain_pen: fountain_pen$1,
  	black_nib: black_nib$1,
  	memo: memo$1,
  	pencil2: pencil2$1,
  	crayon: crayon$1,
  	paintbrush: paintbrush$1,
  	mag: mag$1,
  	mag_right: mag_right$1,
  	heart: heart$1,
  	orange_heart: orange_heart$1,
  	yellow_heart: yellow_heart$1,
  	green_heart: green_heart$1,
  	blue_heart: blue_heart$1,
  	purple_heart: purple_heart$1,
  	black_heart: black_heart$1,
  	broken_heart: broken_heart$1,
  	heavy_heart_exclamation: heavy_heart_exclamation$1,
  	two_hearts: two_hearts$1,
  	revolving_hearts: revolving_hearts$1,
  	heartbeat: heartbeat$1,
  	heartpulse: heartpulse$1,
  	sparkling_heart: sparkling_heart$1,
  	cupid: cupid$1,
  	gift_heart: gift_heart$1,
  	heart_decoration: heart_decoration$1,
  	peace_symbol: peace_symbol$1,
  	latin_cross: latin_cross$1,
  	star_and_crescent: star_and_crescent$1,
  	om: om$1,
  	wheel_of_dharma: wheel_of_dharma$1,
  	star_of_david: star_of_david$1,
  	six_pointed_star: six_pointed_star$1,
  	menorah: menorah$1,
  	yin_yang: yin_yang$1,
  	orthodox_cross: orthodox_cross$1,
  	place_of_worship: place_of_worship$1,
  	ophiuchus: ophiuchus$1,
  	aries: aries$1,
  	taurus: taurus$1,
  	gemini: gemini$1,
  	cancer: cancer$1,
  	leo: leo$1,
  	virgo: virgo$1,
  	libra: libra$1,
  	scorpius: scorpius$1,
  	sagittarius: sagittarius$1,
  	capricorn: capricorn$1,
  	aquarius: aquarius$1,
  	pisces: pisces$1,
  	id: id$1,
  	atom_symbol: atom_symbol$1,
  	u7a7a: u7a7a,
  	u5272: u5272,
  	radioactive: radioactive$1,
  	biohazard: biohazard$1,
  	mobile_phone_off: mobile_phone_off$1,
  	vibration_mode: vibration_mode$1,
  	u6709: u6709,
  	u7121: u7121,
  	u7533: u7533,
  	u55b6: u55b6,
  	u6708: u6708,
  	eight_pointed_black_star: eight_pointed_black_star$1,
  	vs: vs$1,
  	accept: accept$1,
  	white_flower: white_flower$1,
  	ideograph_advantage: ideograph_advantage$1,
  	secret: secret$1,
  	congratulations: congratulations$1,
  	u5408: u5408,
  	u6e80: u6e80$1,
  	u7981: u7981,
  	a: a$1,
  	b: b$1,
  	ab: ab$1,
  	cl: cl$1,
  	o2: o2$1,
  	sos: sos$1,
  	no_entry: no_entry$1,
  	name_badge: name_badge$1,
  	no_entry_sign: no_entry_sign$1,
  	x: x$1,
  	o: o$1,
  	stop_sign: stop_sign$1,
  	anger: anger$1,
  	hotsprings: hotsprings$1,
  	no_pedestrians: no_pedestrians$1,
  	do_not_litter: do_not_litter$1,
  	no_bicycles: no_bicycles$1,
  	"non-potable_water": {
  	keywords: [
  		"drink",
  		"faucet",
  		"tap",
  		"circle"
  	],
  	char: "🚱",
  	fitzpatrick_scale: false,
  	category: "symbols"
  },
  	underage: underage$1,
  	no_mobile_phones: no_mobile_phones$1,
  	exclamation: exclamation$1,
  	grey_exclamation: grey_exclamation$1,
  	question: question$1,
  	grey_question: grey_question$1,
  	bangbang: bangbang$1,
  	interrobang: interrobang$1,
  	low_brightness: low_brightness$1,
  	high_brightness: high_brightness$1,
  	trident: trident$1,
  	fleur_de_lis: fleur_de_lis$1,
  	part_alternation_mark: part_alternation_mark$1,
  	warning: warning$1,
  	children_crossing: children_crossing$1,
  	beginner: beginner$1,
  	recycle: recycle$1,
  	u6307: u6307,
  	chart: chart$1,
  	sparkle: sparkle$1,
  	eight_spoked_asterisk: eight_spoked_asterisk$1,
  	negative_squared_cross_mark: negative_squared_cross_mark$1,
  	white_check_mark: white_check_mark$1,
  	diamond_shape_with_a_dot_inside: diamond_shape_with_a_dot_inside$1,
  	cyclone: cyclone$1,
  	loop: loop$1,
  	globe_with_meridians: globe_with_meridians$1,
  	m: m$1,
  	atm: atm$1,
  	sa: sa$1,
  	passport_control: passport_control$1,
  	customs: customs$1,
  	baggage_claim: baggage_claim$1,
  	left_luggage: left_luggage$1,
  	wheelchair: wheelchair$1,
  	no_smoking: no_smoking$1,
  	wc: wc$1,
  	parking: parking$1,
  	potable_water: potable_water$1,
  	mens: mens$1,
  	womens: womens$1,
  	baby_symbol: baby_symbol$1,
  	restroom: restroom$1,
  	put_litter_in_its_place: put_litter_in_its_place$1,
  	cinema: cinema$1,
  	signal_strength: signal_strength$1,
  	koko: koko$1,
  	ng: ng$1,
  	ok: ok$1,
  	up: up$1,
  	cool: cool$1,
  	"new": {
  	keywords: [
  		"blue-square",
  		"words",
  		"start"
  	],
  	char: "🆕",
  	fitzpatrick_scale: false,
  	category: "symbols"
  },
  	free: free$1,
  	zero: zero$2,
  	one: one$1,
  	two: two$1,
  	three: three$1,
  	four: four$1,
  	five: five$1,
  	six: six$1,
  	seven: seven$1,
  	eight: eight$1,
  	nine: nine$1,
  	keycap_ten: keycap_ten$1,
  	asterisk: asterisk$1,
  	eject_button: eject_button$1,
  	arrow_forward: arrow_forward$1,
  	pause_button: pause_button$1,
  	next_track_button: next_track_button$1,
  	stop_button: stop_button$1,
  	record_button: record_button$1,
  	play_or_pause_button: play_or_pause_button$1,
  	previous_track_button: previous_track_button$1,
  	fast_forward: fast_forward$1,
  	rewind: rewind$1,
  	twisted_rightwards_arrows: twisted_rightwards_arrows$1,
  	repeat: repeat$1,
  	repeat_one: repeat_one$1,
  	arrow_backward: arrow_backward$1,
  	arrow_up_small: arrow_up_small$1,
  	arrow_down_small: arrow_down_small$1,
  	arrow_double_up: arrow_double_up$1,
  	arrow_double_down: arrow_double_down$1,
  	arrow_right: arrow_right$1,
  	arrow_left: arrow_left$1,
  	arrow_up: arrow_up$1,
  	arrow_down: arrow_down$1,
  	arrow_upper_right: arrow_upper_right$1,
  	arrow_lower_right: arrow_lower_right$1,
  	arrow_lower_left: arrow_lower_left$1,
  	arrow_upper_left: arrow_upper_left$1,
  	arrow_up_down: arrow_up_down$1,
  	left_right_arrow: left_right_arrow$1,
  	arrows_counterclockwise: arrows_counterclockwise$1,
  	arrow_right_hook: arrow_right_hook$1,
  	leftwards_arrow_with_hook: leftwards_arrow_with_hook$1,
  	arrow_heading_up: arrow_heading_up$1,
  	arrow_heading_down: arrow_heading_down$1,
  	hash: hash$1,
  	information_source: information_source$1,
  	abc: abc$1,
  	abcd: abcd$1,
  	capital_abcd: capital_abcd$1,
  	symbols: symbols$1,
  	musical_note: musical_note$1,
  	notes: notes$1,
  	wavy_dash: wavy_dash$1,
  	curly_loop: curly_loop$1,
  	heavy_check_mark: heavy_check_mark$1,
  	arrows_clockwise: arrows_clockwise$1,
  	heavy_plus_sign: heavy_plus_sign$1,
  	heavy_minus_sign: heavy_minus_sign$1,
  	heavy_division_sign: heavy_division_sign$1,
  	heavy_multiplication_x: heavy_multiplication_x$1,
  	infinity: infinity$1,
  	heavy_dollar_sign: heavy_dollar_sign$1,
  	currency_exchange: currency_exchange$1,
  	copyright: copyright$1,
  	registered: registered$1,
  	tm: tm$1,
  	end: end$1,
  	back: back$1,
  	on: on$1,
  	top: top$2,
  	soon: soon$1,
  	ballot_box_with_check: ballot_box_with_check$1,
  	radio_button: radio_button$1,
  	white_circle: white_circle$1,
  	black_circle: black_circle$1,
  	red_circle: red_circle$1,
  	large_blue_circle: large_blue_circle$1,
  	small_orange_diamond: small_orange_diamond$1,
  	small_blue_diamond: small_blue_diamond$1,
  	large_orange_diamond: large_orange_diamond$1,
  	large_blue_diamond: large_blue_diamond$1,
  	small_red_triangle: small_red_triangle$1,
  	black_small_square: black_small_square$1,
  	white_small_square: white_small_square$1,
  	black_large_square: black_large_square$1,
  	white_large_square: white_large_square$1,
  	small_red_triangle_down: small_red_triangle_down$1,
  	black_medium_square: black_medium_square$1,
  	white_medium_square: white_medium_square$1,
  	black_medium_small_square: black_medium_small_square$1,
  	white_medium_small_square: white_medium_small_square$1,
  	black_square_button: black_square_button$1,
  	white_square_button: white_square_button$1,
  	speaker: speaker$1,
  	sound: sound$1,
  	loud_sound: loud_sound$1,
  	mute: mute$1,
  	mega: mega$1,
  	loudspeaker: loudspeaker$1,
  	bell: bell$1,
  	no_bell: no_bell$1,
  	black_joker: black_joker$1,
  	mahjong: mahjong$1,
  	spades: spades$2,
  	clubs: clubs$2,
  	hearts: hearts$2,
  	diamonds: diamonds$1,
  	flower_playing_cards: flower_playing_cards$1,
  	thought_balloon: thought_balloon$1,
  	right_anger_bubble: right_anger_bubble$1,
  	speech_balloon: speech_balloon$1,
  	left_speech_bubble: left_speech_bubble$1,
  	clock1: clock1$1,
  	clock2: clock2$1,
  	clock3: clock3$1,
  	clock4: clock4$1,
  	clock5: clock5$1,
  	clock6: clock6$1,
  	clock7: clock7$1,
  	clock8: clock8$1,
  	clock9: clock9$1,
  	clock10: clock10$1,
  	clock11: clock11$1,
  	clock12: clock12$1,
  	clock130: clock130$1,
  	clock230: clock230$1,
  	clock330: clock330$1,
  	clock430: clock430$1,
  	clock530: clock530$1,
  	clock630: clock630$1,
  	clock730: clock730$1,
  	clock830: clock830$1,
  	clock930: clock930$1,
  	clock1030: clock1030$1,
  	clock1130: clock1130$1,
  	clock1230: clock1230$1,
  	afghanistan: afghanistan$1,
  	aland_islands: aland_islands$1,
  	albania: albania$1,
  	algeria: algeria$1,
  	american_samoa: american_samoa$1,
  	andorra: andorra$1,
  	angola: angola$1,
  	anguilla: anguilla$1,
  	antarctica: antarctica$1,
  	antigua_barbuda: antigua_barbuda$1,
  	argentina: argentina$1,
  	armenia: armenia$1,
  	aruba: aruba$1,
  	australia: australia$1,
  	austria: austria$1,
  	azerbaijan: azerbaijan$1,
  	bahamas: bahamas$1,
  	bahrain: bahrain$1,
  	bangladesh: bangladesh$1,
  	barbados: barbados$1,
  	belarus: belarus$1,
  	belgium: belgium$1,
  	belize: belize$1,
  	benin: benin$1,
  	bermuda: bermuda$1,
  	bhutan: bhutan$1,
  	bolivia: bolivia$1,
  	caribbean_netherlands: caribbean_netherlands$1,
  	bosnia_herzegovina: bosnia_herzegovina$1,
  	botswana: botswana$1,
  	brazil: brazil$1,
  	british_indian_ocean_territory: british_indian_ocean_territory$1,
  	british_virgin_islands: british_virgin_islands$1,
  	brunei: brunei$1,
  	bulgaria: bulgaria$1,
  	burkina_faso: burkina_faso$1,
  	burundi: burundi$1,
  	cape_verde: cape_verde$1,
  	cambodia: cambodia$1,
  	cameroon: cameroon$1,
  	canada: canada$1,
  	canary_islands: canary_islands$1,
  	cayman_islands: cayman_islands$1,
  	central_african_republic: central_african_republic$1,
  	chad: chad$1,
  	chile: chile$1,
  	cn: cn$1,
  	christmas_island: christmas_island$1,
  	cocos_islands: cocos_islands$1,
  	colombia: colombia$1,
  	comoros: comoros$1,
  	congo_brazzaville: congo_brazzaville$1,
  	congo_kinshasa: congo_kinshasa$1,
  	cook_islands: cook_islands$1,
  	costa_rica: costa_rica$1,
  	croatia: croatia$1,
  	cuba: cuba$1,
  	curacao: curacao$1,
  	cyprus: cyprus$1,
  	czech_republic: czech_republic$1,
  	denmark: denmark$1,
  	djibouti: djibouti$1,
  	dominica: dominica$1,
  	dominican_republic: dominican_republic$1,
  	ecuador: ecuador$1,
  	egypt: egypt$1,
  	el_salvador: el_salvador$1,
  	equatorial_guinea: equatorial_guinea$1,
  	eritrea: eritrea$1,
  	estonia: estonia$1,
  	ethiopia: ethiopia$1,
  	eu: eu$1,
  	falkland_islands: falkland_islands$1,
  	faroe_islands: faroe_islands$1,
  	fiji: fiji$1,
  	finland: finland$1,
  	fr: fr$1,
  	french_guiana: french_guiana$1,
  	french_polynesia: french_polynesia$1,
  	french_southern_territories: french_southern_territories$1,
  	gabon: gabon$1,
  	gambia: gambia$1,
  	georgia: georgia$1,
  	de: de$1,
  	ghana: ghana$1,
  	gibraltar: gibraltar$1,
  	greece: greece$1,
  	greenland: greenland$1,
  	grenada: grenada$1,
  	guadeloupe: guadeloupe$1,
  	guam: guam$1,
  	guatemala: guatemala$1,
  	guernsey: guernsey$1,
  	guinea: guinea$1,
  	guinea_bissau: guinea_bissau$1,
  	guyana: guyana$1,
  	haiti: haiti$1,
  	honduras: honduras$1,
  	hong_kong: hong_kong$1,
  	hungary: hungary$1,
  	iceland: iceland$1,
  	india: india$1,
  	indonesia: indonesia$1,
  	iran: iran$1,
  	iraq: iraq$1,
  	ireland: ireland$1,
  	isle_of_man: isle_of_man$1,
  	israel: israel$1,
  	it: it$2,
  	cote_divoire: cote_divoire$1,
  	jamaica: jamaica$1,
  	jp: jp$1,
  	jersey: jersey$1,
  	jordan: jordan$1,
  	kazakhstan: kazakhstan$1,
  	kenya: kenya$1,
  	kiribati: kiribati$1,
  	kosovo: kosovo$1,
  	kuwait: kuwait$1,
  	kyrgyzstan: kyrgyzstan$1,
  	laos: laos$1,
  	latvia: latvia$1,
  	lebanon: lebanon$1,
  	lesotho: lesotho$1,
  	liberia: liberia$1,
  	libya: libya$1,
  	liechtenstein: liechtenstein$1,
  	lithuania: lithuania$1,
  	luxembourg: luxembourg$1,
  	macau: macau$1,
  	macedonia: macedonia$1,
  	madagascar: madagascar$1,
  	malawi: malawi$1,
  	malaysia: malaysia$1,
  	maldives: maldives$1,
  	mali: mali$1,
  	malta: malta$1,
  	marshall_islands: marshall_islands$1,
  	martinique: martinique$1,
  	mauritania: mauritania$1,
  	mauritius: mauritius$1,
  	mayotte: mayotte$1,
  	mexico: mexico$1,
  	micronesia: micronesia$1,
  	moldova: moldova$1,
  	monaco: monaco$1,
  	mongolia: mongolia$1,
  	montenegro: montenegro$1,
  	montserrat: montserrat$1,
  	morocco: morocco$1,
  	mozambique: mozambique$1,
  	myanmar: myanmar$1,
  	namibia: namibia$1,
  	nauru: nauru$1,
  	nepal: nepal$1,
  	netherlands: netherlands$1,
  	new_caledonia: new_caledonia$1,
  	new_zealand: new_zealand$1,
  	nicaragua: nicaragua$1,
  	niger: niger$1,
  	nigeria: nigeria$1,
  	niue: niue$1,
  	norfolk_island: norfolk_island$1,
  	northern_mariana_islands: northern_mariana_islands$1,
  	north_korea: north_korea$1,
  	norway: norway$1,
  	oman: oman$1,
  	pakistan: pakistan$1,
  	palau: palau$1,
  	palestinian_territories: palestinian_territories$1,
  	panama: panama$1,
  	papua_new_guinea: papua_new_guinea$1,
  	paraguay: paraguay$1,
  	peru: peru$1,
  	philippines: philippines$1,
  	pitcairn_islands: pitcairn_islands$1,
  	poland: poland$1,
  	portugal: portugal$1,
  	puerto_rico: puerto_rico$1,
  	qatar: qatar$1,
  	reunion: reunion$1,
  	romania: romania$1,
  	ru: ru$1,
  	rwanda: rwanda$1,
  	st_barthelemy: st_barthelemy$1,
  	st_helena: st_helena$1,
  	st_kitts_nevis: st_kitts_nevis$1,
  	st_lucia: st_lucia$1,
  	st_pierre_miquelon: st_pierre_miquelon$1,
  	st_vincent_grenadines: st_vincent_grenadines$1,
  	samoa: samoa$1,
  	san_marino: san_marino$1,
  	sao_tome_principe: sao_tome_principe$1,
  	saudi_arabia: saudi_arabia$1,
  	senegal: senegal$1,
  	serbia: serbia$1,
  	seychelles: seychelles$1,
  	sierra_leone: sierra_leone$1,
  	singapore: singapore$1,
  	sint_maarten: sint_maarten$1,
  	slovakia: slovakia$1,
  	slovenia: slovenia$1,
  	solomon_islands: solomon_islands$1,
  	somalia: somalia$1,
  	south_africa: south_africa$1,
  	south_georgia_south_sandwich_islands: south_georgia_south_sandwich_islands$1,
  	kr: kr$1,
  	south_sudan: south_sudan$1,
  	es: es$1,
  	sri_lanka: sri_lanka$1,
  	sudan: sudan$1,
  	suriname: suriname$1,
  	swaziland: swaziland$1,
  	sweden: sweden$1,
  	switzerland: switzerland$1,
  	syria: syria$1,
  	taiwan: taiwan$1,
  	tajikistan: tajikistan$1,
  	tanzania: tanzania$1,
  	thailand: thailand$1,
  	timor_leste: timor_leste$1,
  	togo: togo$1,
  	tokelau: tokelau$1,
  	tonga: tonga$1,
  	trinidad_tobago: trinidad_tobago$1,
  	tunisia: tunisia$1,
  	tr: tr$1,
  	turkmenistan: turkmenistan$1,
  	turks_caicos_islands: turks_caicos_islands$1,
  	tuvalu: tuvalu$1,
  	uganda: uganda$1,
  	ukraine: ukraine$1,
  	united_arab_emirates: united_arab_emirates$1,
  	uk: uk$1,
  	england: england$1,
  	scotland: scotland$1,
  	wales: wales$1,
  	us: us$1,
  	us_virgin_islands: us_virgin_islands$1,
  	uruguay: uruguay$1,
  	uzbekistan: uzbekistan$1,
  	vanuatu: vanuatu$1,
  	vatican_city: vatican_city$1,
  	venezuela: venezuela$1,
  	vietnam: vietnam$1,
  	wallis_futuna: wallis_futuna$1,
  	western_sahara: western_sahara$1,
  	yemen: yemen$1,
  	zambia: zambia$1,
  	zimbabwe: zimbabwe$1,
  	united_nations: united_nations$1,
  	pirate_flag: pirate_flag$1
  };

  var emojis$1 = /*#__PURE__*/Object.freeze({
    __proto__: null,
    grinning: grinning$1,
    grimacing: grimacing$1,
    grin: grin$1,
    joy: joy$1,
    rofl: rofl$1,
    partying: partying,
    smiley: smiley$1,
    smile: smile$2,
    sweat_smile: sweat_smile$1,
    laughing: laughing$1,
    innocent: innocent$1,
    wink: wink$1,
    blush: blush$1,
    slightly_smiling_face: slightly_smiling_face$1,
    upside_down_face: upside_down_face$1,
    relaxed: relaxed$1,
    yum: yum$1,
    relieved: relieved$1,
    heart_eyes: heart_eyes$1,
    smiling_face_with_three_hearts: smiling_face_with_three_hearts$1,
    kissing_heart: kissing_heart$1,
    kissing: kissing$1,
    kissing_smiling_eyes: kissing_smiling_eyes$1,
    kissing_closed_eyes: kissing_closed_eyes$1,
    stuck_out_tongue_winking_eye: stuck_out_tongue_winking_eye$1,
    zany: zany,
    raised_eyebrow: raised_eyebrow$1,
    monocle: monocle,
    stuck_out_tongue_closed_eyes: stuck_out_tongue_closed_eyes$1,
    stuck_out_tongue: stuck_out_tongue$1,
    money_mouth_face: money_mouth_face$1,
    nerd_face: nerd_face$1,
    sunglasses: sunglasses$1,
    star_struck: star_struck$1,
    clown_face: clown_face$1,
    cowboy_hat_face: cowboy_hat_face$1,
    hugs: hugs$1,
    smirk: smirk$1,
    no_mouth: no_mouth$1,
    neutral_face: neutral_face$1,
    expressionless: expressionless$1,
    unamused: unamused$1,
    roll_eyes: roll_eyes$1,
    thinking: thinking$1,
    lying_face: lying_face$1,
    hand_over_mouth: hand_over_mouth$1,
    shushing: shushing,
    symbols_over_mouth: symbols_over_mouth,
    exploding_head: exploding_head$1,
    flushed: flushed$1,
    disappointed: disappointed$1,
    worried: worried$1,
    angry: angry$1,
    rage: rage$1,
    pensive: pensive$1,
    confused: confused$1,
    slightly_frowning_face: slightly_frowning_face$1,
    frowning_face: frowning_face$1,
    persevere: persevere$1,
    confounded: confounded$1,
    tired_face: tired_face$1,
    weary: weary$1,
    pleading: pleading,
    triumph: triumph$1,
    open_mouth: open_mouth$1,
    scream: scream$1,
    fearful: fearful$1,
    cold_sweat: cold_sweat$1,
    hushed: hushed$1,
    frowning: frowning$1,
    anguished: anguished$1,
    cry: cry$1,
    disappointed_relieved: disappointed_relieved$1,
    drooling_face: drooling_face$1,
    sleepy: sleepy$1,
    sweat: sweat$1,
    hot: hot,
    cold: cold,
    sob: sob$1,
    dizzy_face: dizzy_face$1,
    astonished: astonished$1,
    zipper_mouth_face: zipper_mouth_face$1,
    nauseated_face: nauseated_face$1,
    sneezing_face: sneezing_face$1,
    vomiting: vomiting,
    mask: mask$1,
    face_with_thermometer: face_with_thermometer$1,
    face_with_head_bandage: face_with_head_bandage$1,
    woozy: woozy,
    sleeping: sleeping$1,
    zzz: zzz$1,
    poop: poop$1,
    smiling_imp: smiling_imp$1,
    imp: imp$1,
    japanese_ogre: japanese_ogre$1,
    japanese_goblin: japanese_goblin$1,
    skull: skull$1,
    ghost: ghost$1,
    alien: alien$1,
    robot: robot$1,
    smiley_cat: smiley_cat$1,
    smile_cat: smile_cat$1,
    joy_cat: joy_cat$1,
    heart_eyes_cat: heart_eyes_cat$1,
    smirk_cat: smirk_cat$1,
    kissing_cat: kissing_cat$1,
    scream_cat: scream_cat$1,
    crying_cat_face: crying_cat_face$1,
    pouting_cat: pouting_cat$1,
    palms_up: palms_up,
    raised_hands: raised_hands$1,
    clap: clap$1,
    wave: wave$1,
    call_me_hand: call_me_hand$1,
    facepunch: facepunch$1,
    fist: fist$1,
    fist_left: fist_left$1,
    fist_right: fist_right$1,
    v: v$1,
    ok_hand: ok_hand$1,
    raised_hand: raised_hand$1,
    raised_back_of_hand: raised_back_of_hand$1,
    open_hands: open_hands$1,
    muscle: muscle$1,
    pray: pray$1,
    foot: foot$1,
    leg: leg$2,
    handshake: handshake$1,
    point_up: point_up$1,
    point_up_2: point_up_2$1,
    point_down: point_down$1,
    point_left: point_left$1,
    point_right: point_right$1,
    fu: fu$1,
    raised_hand_with_fingers_splayed: raised_hand_with_fingers_splayed$1,
    love_you: love_you,
    metal: metal$1,
    crossed_fingers: crossed_fingers$1,
    vulcan_salute: vulcan_salute$1,
    writing_hand: writing_hand$1,
    selfie: selfie$1,
    nail_care: nail_care$1,
    lips: lips$1,
    tooth: tooth$1,
    tongue: tongue$1,
    ear: ear$1,
    nose: nose$1,
    eye: eye$1,
    eyes: eyes$1,
    brain: brain$1,
    bust_in_silhouette: bust_in_silhouette$1,
    busts_in_silhouette: busts_in_silhouette$1,
    speaking_head: speaking_head$1,
    baby: baby$1,
    child: child$1,
    boy: boy$1,
    girl: girl$1,
    adult: adult$1,
    man: man$1,
    woman: woman$1,
    blonde_woman: blonde_woman$1,
    blonde_man: blonde_man,
    bearded_person: bearded_person$1,
    older_adult: older_adult$1,
    older_man: older_man$1,
    older_woman: older_woman$1,
    man_with_gua_pi_mao: man_with_gua_pi_mao$1,
    woman_with_headscarf: woman_with_headscarf$1,
    woman_with_turban: woman_with_turban$1,
    man_with_turban: man_with_turban$1,
    policewoman: policewoman$1,
    policeman: policeman$1,
    construction_worker_woman: construction_worker_woman$1,
    construction_worker_man: construction_worker_man$1,
    guardswoman: guardswoman$1,
    guardsman: guardsman$1,
    female_detective: female_detective$1,
    male_detective: male_detective$1,
    woman_health_worker: woman_health_worker$1,
    man_health_worker: man_health_worker$1,
    woman_farmer: woman_farmer$1,
    man_farmer: man_farmer$1,
    woman_cook: woman_cook$1,
    man_cook: man_cook$1,
    woman_student: woman_student$1,
    man_student: man_student$1,
    woman_singer: woman_singer$1,
    man_singer: man_singer$1,
    woman_teacher: woman_teacher$1,
    man_teacher: man_teacher$1,
    woman_factory_worker: woman_factory_worker$1,
    man_factory_worker: man_factory_worker$1,
    woman_technologist: woman_technologist$1,
    man_technologist: man_technologist$1,
    woman_office_worker: woman_office_worker$1,
    man_office_worker: man_office_worker$1,
    woman_mechanic: woman_mechanic$1,
    man_mechanic: man_mechanic$1,
    woman_scientist: woman_scientist$1,
    man_scientist: man_scientist$1,
    woman_artist: woman_artist$1,
    man_artist: man_artist$1,
    woman_firefighter: woman_firefighter$1,
    man_firefighter: man_firefighter$1,
    woman_pilot: woman_pilot$1,
    man_pilot: man_pilot$1,
    woman_astronaut: woman_astronaut$1,
    man_astronaut: man_astronaut$1,
    woman_judge: woman_judge$1,
    man_judge: man_judge$1,
    woman_superhero: woman_superhero,
    man_superhero: man_superhero,
    woman_supervillain: woman_supervillain,
    man_supervillain: man_supervillain,
    mrs_claus: mrs_claus$1,
    santa: santa$1,
    sorceress: sorceress,
    wizard: wizard,
    woman_elf: woman_elf,
    man_elf: man_elf,
    woman_vampire: woman_vampire,
    man_vampire: man_vampire,
    woman_zombie: woman_zombie,
    man_zombie: man_zombie,
    woman_genie: woman_genie,
    man_genie: man_genie,
    mermaid: mermaid$1,
    merman: merman$1,
    woman_fairy: woman_fairy,
    man_fairy: man_fairy,
    angel: angel$1,
    pregnant_woman: pregnant_woman$1,
    breastfeeding: breastfeeding,
    princess: princess$1,
    prince: prince$1,
    bride_with_veil: bride_with_veil$1,
    man_in_tuxedo: man_in_tuxedo$1,
    running_woman: running_woman$1,
    running_man: running_man$1,
    walking_woman: walking_woman$1,
    walking_man: walking_man$1,
    dancer: dancer$1,
    man_dancing: man_dancing$1,
    dancing_women: dancing_women$1,
    dancing_men: dancing_men$1,
    couple: couple$1,
    two_men_holding_hands: two_men_holding_hands$1,
    two_women_holding_hands: two_women_holding_hands$1,
    bowing_woman: bowing_woman$1,
    bowing_man: bowing_man$1,
    man_facepalming: man_facepalming$1,
    woman_facepalming: woman_facepalming$1,
    woman_shrugging: woman_shrugging$1,
    man_shrugging: man_shrugging$1,
    tipping_hand_woman: tipping_hand_woman$1,
    tipping_hand_man: tipping_hand_man$1,
    no_good_woman: no_good_woman$1,
    no_good_man: no_good_man$1,
    ok_woman: ok_woman$1,
    ok_man: ok_man$1,
    raising_hand_woman: raising_hand_woman$1,
    raising_hand_man: raising_hand_man$1,
    pouting_woman: pouting_woman$1,
    pouting_man: pouting_man$1,
    frowning_woman: frowning_woman$1,
    frowning_man: frowning_man$1,
    haircut_woman: haircut_woman$1,
    haircut_man: haircut_man$1,
    massage_woman: massage_woman$1,
    massage_man: massage_man$1,
    woman_in_steamy_room: woman_in_steamy_room,
    man_in_steamy_room: man_in_steamy_room,
    couple_with_heart_woman_man: couple_with_heart_woman_man$1,
    couple_with_heart_woman_woman: couple_with_heart_woman_woman$1,
    couple_with_heart_man_man: couple_with_heart_man_man$1,
    couplekiss_man_woman: couplekiss_man_woman$1,
    couplekiss_woman_woman: couplekiss_woman_woman$1,
    couplekiss_man_man: couplekiss_man_man$1,
    family_man_woman_boy: family_man_woman_boy$1,
    family_man_woman_girl: family_man_woman_girl$1,
    family_man_woman_girl_boy: family_man_woman_girl_boy$1,
    family_man_woman_boy_boy: family_man_woman_boy_boy$1,
    family_man_woman_girl_girl: family_man_woman_girl_girl$1,
    family_woman_woman_boy: family_woman_woman_boy$1,
    family_woman_woman_girl: family_woman_woman_girl$1,
    family_woman_woman_girl_boy: family_woman_woman_girl_boy$1,
    family_woman_woman_boy_boy: family_woman_woman_boy_boy$1,
    family_woman_woman_girl_girl: family_woman_woman_girl_girl$1,
    family_man_man_boy: family_man_man_boy$1,
    family_man_man_girl: family_man_man_girl$1,
    family_man_man_girl_boy: family_man_man_girl_boy$1,
    family_man_man_boy_boy: family_man_man_boy_boy$1,
    family_man_man_girl_girl: family_man_man_girl_girl$1,
    family_woman_boy: family_woman_boy$1,
    family_woman_girl: family_woman_girl$1,
    family_woman_girl_boy: family_woman_girl_boy$1,
    family_woman_boy_boy: family_woman_boy_boy$1,
    family_woman_girl_girl: family_woman_girl_girl$1,
    family_man_boy: family_man_boy$1,
    family_man_girl: family_man_girl$1,
    family_man_girl_boy: family_man_girl_boy$1,
    family_man_boy_boy: family_man_boy_boy$1,
    family_man_girl_girl: family_man_girl_girl$1,
    yarn: yarn$1,
    thread: thread$1,
    coat: coat$1,
    labcoat: labcoat,
    womans_clothes: womans_clothes$1,
    tshirt: tshirt$1,
    jeans: jeans$1,
    necktie: necktie$1,
    dress: dress$1,
    bikini: bikini$1,
    kimono: kimono$1,
    lipstick: lipstick$1,
    kiss: kiss$1,
    footprints: footprints$1,
    flat_shoe: flat_shoe$1,
    high_heel: high_heel$1,
    sandal: sandal$1,
    boot: boot$1,
    mans_shoe: mans_shoe$1,
    athletic_shoe: athletic_shoe$1,
    hiking_boot: hiking_boot$1,
    socks: socks$1,
    gloves: gloves$1,
    scarf: scarf$1,
    womans_hat: womans_hat$1,
    tophat: tophat$1,
    billed_hat: billed_hat,
    rescue_worker_helmet: rescue_worker_helmet$1,
    mortar_board: mortar_board$1,
    crown: crown$1,
    school_satchel: school_satchel$1,
    luggage: luggage$1,
    pouch: pouch$1,
    purse: purse$1,
    handbag: handbag$1,
    briefcase: briefcase$1,
    eyeglasses: eyeglasses$1,
    dark_sunglasses: dark_sunglasses$1,
    goggles: goggles$1,
    ring: ring$2,
    closed_umbrella: closed_umbrella$1,
    dog: dog$1,
    cat: cat$1,
    mouse: mouse$1,
    hamster: hamster$1,
    rabbit: rabbit$1,
    fox_face: fox_face$1,
    bear: bear$1,
    panda_face: panda_face$1,
    koala: koala$1,
    tiger: tiger$1,
    lion: lion$1,
    cow: cow$1,
    pig: pig$1,
    pig_nose: pig_nose$1,
    frog: frog$1,
    squid: squid$1,
    octopus: octopus$1,
    shrimp: shrimp$1,
    monkey_face: monkey_face$1,
    gorilla: gorilla$1,
    see_no_evil: see_no_evil$1,
    hear_no_evil: hear_no_evil$1,
    speak_no_evil: speak_no_evil$1,
    monkey: monkey$1,
    chicken: chicken$1,
    penguin: penguin$1,
    bird: bird$1,
    baby_chick: baby_chick$1,
    hatching_chick: hatching_chick$1,
    hatched_chick: hatched_chick$1,
    duck: duck$1,
    eagle: eagle$1,
    owl: owl$1,
    bat: bat$1,
    wolf: wolf$1,
    boar: boar$1,
    horse: horse$1,
    unicorn: unicorn$1,
    honeybee: honeybee$1,
    bug: bug$1,
    butterfly: butterfly$1,
    snail: snail$1,
    beetle: beetle$1,
    ant: ant$1,
    grasshopper: grasshopper,
    spider: spider$1,
    scorpion: scorpion$1,
    crab: crab$1,
    snake: snake$1,
    lizard: lizard$1,
    sauropod: sauropod$1,
    turtle: turtle$1,
    tropical_fish: tropical_fish$1,
    fish: fish$1,
    blowfish: blowfish$1,
    dolphin: dolphin$1,
    shark: shark$1,
    whale: whale$1,
    whale2: whale2$1,
    crocodile: crocodile$1,
    leopard: leopard$1,
    zebra: zebra$1,
    tiger2: tiger2$1,
    water_buffalo: water_buffalo$1,
    ox: ox$1,
    cow2: cow2$1,
    deer: deer$1,
    dromedary_camel: dromedary_camel$1,
    camel: camel$1,
    giraffe: giraffe$1,
    elephant: elephant$1,
    rhinoceros: rhinoceros$1,
    goat: goat$1,
    ram: ram$1,
    sheep: sheep$1,
    racehorse: racehorse$1,
    pig2: pig2$1,
    rat: rat$1,
    mouse2: mouse2$1,
    rooster: rooster$1,
    turkey: turkey$1,
    dove: dove$1,
    dog2: dog2$1,
    poodle: poodle$1,
    cat2: cat2$1,
    rabbit2: rabbit2$1,
    chipmunk: chipmunk$1,
    hedgehog: hedgehog$1,
    raccoon: raccoon$1,
    llama: llama$1,
    hippopotamus: hippopotamus$1,
    kangaroo: kangaroo$1,
    badger: badger$1,
    swan: swan$1,
    peacock: peacock$1,
    parrot: parrot$1,
    lobster: lobster$1,
    mosquito: mosquito$1,
    paw_prints: paw_prints$1,
    dragon: dragon$1,
    dragon_face: dragon_face$1,
    cactus: cactus$1,
    christmas_tree: christmas_tree$1,
    evergreen_tree: evergreen_tree$1,
    deciduous_tree: deciduous_tree$1,
    palm_tree: palm_tree$1,
    seedling: seedling$1,
    herb: herb$1,
    shamrock: shamrock$1,
    four_leaf_clover: four_leaf_clover$1,
    bamboo: bamboo$1,
    tanabata_tree: tanabata_tree$1,
    leaves: leaves$1,
    fallen_leaf: fallen_leaf$1,
    maple_leaf: maple_leaf$1,
    ear_of_rice: ear_of_rice$1,
    hibiscus: hibiscus$1,
    sunflower: sunflower$1,
    rose: rose$1,
    wilted_flower: wilted_flower$1,
    tulip: tulip$1,
    blossom: blossom$1,
    cherry_blossom: cherry_blossom$1,
    bouquet: bouquet$1,
    mushroom: mushroom$1,
    chestnut: chestnut$1,
    jack_o_lantern: jack_o_lantern$1,
    shell: shell$1,
    spider_web: spider_web$1,
    earth_americas: earth_americas$1,
    earth_africa: earth_africa$1,
    earth_asia: earth_asia$1,
    full_moon: full_moon$1,
    waning_gibbous_moon: waning_gibbous_moon$1,
    last_quarter_moon: last_quarter_moon$1,
    waning_crescent_moon: waning_crescent_moon$1,
    new_moon: new_moon$1,
    waxing_crescent_moon: waxing_crescent_moon$1,
    first_quarter_moon: first_quarter_moon$1,
    waxing_gibbous_moon: waxing_gibbous_moon$1,
    new_moon_with_face: new_moon_with_face$1,
    full_moon_with_face: full_moon_with_face$1,
    first_quarter_moon_with_face: first_quarter_moon_with_face$1,
    last_quarter_moon_with_face: last_quarter_moon_with_face$1,
    sun_with_face: sun_with_face$1,
    crescent_moon: crescent_moon$1,
    star: star$2,
    star2: star2$1,
    dizzy: dizzy$1,
    sparkles: sparkles$1,
    comet: comet$1,
    sunny: sunny$1,
    sun_behind_small_cloud: sun_behind_small_cloud$1,
    partly_sunny: partly_sunny$1,
    sun_behind_large_cloud: sun_behind_large_cloud$1,
    sun_behind_rain_cloud: sun_behind_rain_cloud$1,
    cloud: cloud$1,
    cloud_with_rain: cloud_with_rain$1,
    cloud_with_lightning_and_rain: cloud_with_lightning_and_rain$1,
    cloud_with_lightning: cloud_with_lightning$1,
    zap: zap$1,
    fire: fire$1,
    boom: boom$1,
    snowflake: snowflake$1,
    cloud_with_snow: cloud_with_snow$1,
    snowman: snowman$1,
    snowman_with_snow: snowman_with_snow$1,
    wind_face: wind_face$1,
    dash: dash$2,
    tornado: tornado$1,
    fog: fog$1,
    open_umbrella: open_umbrella$1,
    umbrella: umbrella$1,
    droplet: droplet$1,
    sweat_drops: sweat_drops$1,
    ocean: ocean$1,
    green_apple: green_apple$1,
    apple: apple$1,
    pear: pear$1,
    tangerine: tangerine$1,
    lemon: lemon$1,
    banana: banana$1,
    watermelon: watermelon$1,
    grapes: grapes$1,
    strawberry: strawberry$1,
    melon: melon$1,
    cherries: cherries$1,
    peach: peach$1,
    pineapple: pineapple$1,
    coconut: coconut$1,
    kiwi_fruit: kiwi_fruit$1,
    mango: mango$1,
    avocado: avocado$1,
    broccoli: broccoli$1,
    tomato: tomato$1,
    eggplant: eggplant$1,
    cucumber: cucumber$1,
    carrot: carrot$1,
    hot_pepper: hot_pepper$1,
    potato: potato$1,
    corn: corn$1,
    leafy_greens: leafy_greens,
    sweet_potato: sweet_potato$1,
    peanuts: peanuts$1,
    honey_pot: honey_pot$1,
    croissant: croissant$1,
    bread: bread$1,
    baguette_bread: baguette_bread$1,
    bagel: bagel$1,
    pretzel: pretzel$1,
    cheese: cheese$1,
    egg: egg$1,
    bacon: bacon$1,
    steak: steak,
    pancakes: pancakes$1,
    poultry_leg: poultry_leg$1,
    meat_on_bone: meat_on_bone$1,
    bone: bone$1,
    fried_shrimp: fried_shrimp$1,
    fried_egg: fried_egg$1,
    hamburger: hamburger$1,
    fries: fries$1,
    stuffed_flatbread: stuffed_flatbread$1,
    hotdog: hotdog$1,
    pizza: pizza$1,
    sandwich: sandwich$1,
    canned_food: canned_food$1,
    spaghetti: spaghetti$1,
    taco: taco$1,
    burrito: burrito$1,
    green_salad: green_salad$1,
    shallow_pan_of_food: shallow_pan_of_food$1,
    ramen: ramen$1,
    stew: stew$1,
    fish_cake: fish_cake$1,
    fortune_cookie: fortune_cookie$1,
    sushi: sushi$1,
    bento: bento$1,
    curry: curry$1,
    rice_ball: rice_ball$1,
    rice: rice$1,
    rice_cracker: rice_cracker$1,
    oden: oden$1,
    dango: dango$1,
    shaved_ice: shaved_ice$1,
    ice_cream: ice_cream$1,
    icecream: icecream$1,
    pie: pie$1,
    cake: cake$1,
    cupcake: cupcake$1,
    moon_cake: moon_cake$1,
    birthday: birthday$1,
    custard: custard$1,
    candy: candy$1,
    lollipop: lollipop$1,
    chocolate_bar: chocolate_bar$1,
    popcorn: popcorn$1,
    dumpling: dumpling$1,
    doughnut: doughnut$1,
    cookie: cookie$1,
    milk_glass: milk_glass$1,
    beer: beer$1,
    beers: beers$1,
    clinking_glasses: clinking_glasses$1,
    wine_glass: wine_glass$1,
    tumbler_glass: tumbler_glass$1,
    cocktail: cocktail$1,
    tropical_drink: tropical_drink$1,
    champagne: champagne$1,
    sake: sake$1,
    tea: tea$1,
    cup_with_straw: cup_with_straw$1,
    coffee: coffee$1,
    baby_bottle: baby_bottle$1,
    salt: salt$1,
    spoon: spoon$1,
    fork_and_knife: fork_and_knife$1,
    plate_with_cutlery: plate_with_cutlery$1,
    bowl_with_spoon: bowl_with_spoon$1,
    takeout_box: takeout_box$1,
    chopsticks: chopsticks$1,
    soccer: soccer$1,
    basketball: basketball$1,
    football: football$1,
    baseball: baseball$1,
    softball: softball$1,
    tennis: tennis$1,
    volleyball: volleyball$1,
    rugby_football: rugby_football$1,
    flying_disc: flying_disc$1,
    golf: golf$1,
    golfing_woman: golfing_woman$1,
    golfing_man: golfing_man$1,
    ping_pong: ping_pong$1,
    badminton: badminton$1,
    goal_net: goal_net$1,
    ice_hockey: ice_hockey$1,
    field_hockey: field_hockey$1,
    lacrosse: lacrosse$1,
    cricket: cricket$1,
    ski: ski$1,
    skier: skier$1,
    snowboarder: snowboarder$1,
    person_fencing: person_fencing$1,
    women_wrestling: women_wrestling$1,
    men_wrestling: men_wrestling$1,
    woman_cartwheeling: woman_cartwheeling$1,
    man_cartwheeling: man_cartwheeling$1,
    woman_playing_handball: woman_playing_handball$1,
    man_playing_handball: man_playing_handball$1,
    ice_skate: ice_skate$1,
    curling_stone: curling_stone$1,
    skateboard: skateboard$1,
    sled: sled$1,
    bow_and_arrow: bow_and_arrow$1,
    fishing_pole_and_fish: fishing_pole_and_fish$1,
    boxing_glove: boxing_glove$1,
    martial_arts_uniform: martial_arts_uniform$1,
    rowing_woman: rowing_woman$1,
    rowing_man: rowing_man$1,
    climbing_woman: climbing_woman$1,
    climbing_man: climbing_man$1,
    swimming_woman: swimming_woman$1,
    swimming_man: swimming_man$1,
    woman_playing_water_polo: woman_playing_water_polo$1,
    man_playing_water_polo: man_playing_water_polo$1,
    woman_in_lotus_position: woman_in_lotus_position,
    man_in_lotus_position: man_in_lotus_position,
    surfing_woman: surfing_woman$1,
    surfing_man: surfing_man$1,
    bath: bath$1,
    basketball_woman: basketball_woman$1,
    basketball_man: basketball_man$1,
    weight_lifting_woman: weight_lifting_woman$1,
    weight_lifting_man: weight_lifting_man$1,
    biking_woman: biking_woman$1,
    biking_man: biking_man$1,
    mountain_biking_woman: mountain_biking_woman$1,
    mountain_biking_man: mountain_biking_man$1,
    horse_racing: horse_racing$1,
    business_suit_levitating: business_suit_levitating$1,
    trophy: trophy$1,
    running_shirt_with_sash: running_shirt_with_sash$1,
    medal_sports: medal_sports$1,
    medal_military: medal_military$1,
    reminder_ribbon: reminder_ribbon$1,
    rosette: rosette$1,
    ticket: ticket$1,
    tickets: tickets$1,
    performing_arts: performing_arts$1,
    art: art$1,
    circus_tent: circus_tent$1,
    woman_juggling: woman_juggling$1,
    man_juggling: man_juggling$1,
    microphone: microphone$1,
    headphones: headphones$1,
    musical_score: musical_score$1,
    musical_keyboard: musical_keyboard$1,
    drum: drum$1,
    saxophone: saxophone$1,
    trumpet: trumpet$1,
    guitar: guitar$1,
    violin: violin$1,
    clapper: clapper$1,
    video_game: video_game$1,
    space_invader: space_invader$1,
    dart: dart$1,
    game_die: game_die$1,
    chess_pawn: chess_pawn$1,
    slot_machine: slot_machine$1,
    jigsaw: jigsaw$1,
    bowling: bowling$1,
    red_car: red_car$1,
    taxi: taxi$1,
    blue_car: blue_car$1,
    bus: bus$1,
    trolleybus: trolleybus$1,
    racing_car: racing_car$1,
    police_car: police_car$1,
    ambulance: ambulance$1,
    fire_engine: fire_engine$1,
    minibus: minibus$1,
    truck: truck$1,
    articulated_lorry: articulated_lorry$1,
    tractor: tractor$1,
    kick_scooter: kick_scooter$1,
    motorcycle: motorcycle$1,
    bike: bike$1,
    motor_scooter: motor_scooter$1,
    rotating_light: rotating_light$1,
    oncoming_police_car: oncoming_police_car$1,
    oncoming_bus: oncoming_bus$1,
    oncoming_automobile: oncoming_automobile$1,
    oncoming_taxi: oncoming_taxi$1,
    aerial_tramway: aerial_tramway$1,
    mountain_cableway: mountain_cableway$1,
    suspension_railway: suspension_railway$1,
    railway_car: railway_car$1,
    train: train$1,
    monorail: monorail$1,
    bullettrain_side: bullettrain_side$1,
    bullettrain_front: bullettrain_front$1,
    light_rail: light_rail$1,
    mountain_railway: mountain_railway$1,
    steam_locomotive: steam_locomotive$1,
    train2: train2$1,
    metro: metro$1,
    tram: tram$1,
    station: station$1,
    flying_saucer: flying_saucer$1,
    helicopter: helicopter$1,
    small_airplane: small_airplane$1,
    airplane: airplane$1,
    flight_departure: flight_departure$1,
    flight_arrival: flight_arrival$1,
    sailboat: sailboat$1,
    motor_boat: motor_boat$1,
    speedboat: speedboat$1,
    ferry: ferry$1,
    passenger_ship: passenger_ship$1,
    rocket: rocket$1,
    artificial_satellite: artificial_satellite$1,
    seat: seat$1,
    canoe: canoe$1,
    anchor: anchor$1,
    construction: construction$1,
    fuelpump: fuelpump$1,
    busstop: busstop$1,
    vertical_traffic_light: vertical_traffic_light$1,
    traffic_light: traffic_light$1,
    checkered_flag: checkered_flag$1,
    ship: ship$1,
    ferris_wheel: ferris_wheel$1,
    roller_coaster: roller_coaster$1,
    carousel_horse: carousel_horse$1,
    building_construction: building_construction$1,
    foggy: foggy$1,
    tokyo_tower: tokyo_tower$1,
    factory: factory$1,
    fountain: fountain$1,
    rice_scene: rice_scene$1,
    mountain: mountain$1,
    mountain_snow: mountain_snow$1,
    mount_fuji: mount_fuji$1,
    volcano: volcano$1,
    japan: japan$1,
    camping: camping$1,
    tent: tent$1,
    national_park: national_park$1,
    motorway: motorway$1,
    railway_track: railway_track$1,
    sunrise: sunrise$1,
    sunrise_over_mountains: sunrise_over_mountains$1,
    desert: desert$1,
    beach_umbrella: beach_umbrella$1,
    desert_island: desert_island$1,
    city_sunrise: city_sunrise$1,
    city_sunset: city_sunset$1,
    cityscape: cityscape$1,
    night_with_stars: night_with_stars$1,
    bridge_at_night: bridge_at_night$1,
    milky_way: milky_way$1,
    stars: stars$1,
    sparkler: sparkler$1,
    fireworks: fireworks$1,
    rainbow: rainbow$1,
    houses: houses$1,
    european_castle: european_castle$1,
    japanese_castle: japanese_castle$1,
    stadium: stadium$1,
    statue_of_liberty: statue_of_liberty$1,
    house: house$1,
    house_with_garden: house_with_garden$1,
    derelict_house: derelict_house$1,
    office: office$1,
    department_store: department_store$1,
    post_office: post_office$1,
    european_post_office: european_post_office$1,
    hospital: hospital$1,
    bank: bank$1,
    hotel: hotel$1,
    convenience_store: convenience_store$1,
    school: school$1,
    love_hotel: love_hotel$1,
    wedding: wedding$1,
    classical_building: classical_building$1,
    church: church$1,
    mosque: mosque$1,
    synagogue: synagogue$1,
    kaaba: kaaba$1,
    shinto_shrine: shinto_shrine$1,
    watch: watch$1,
    iphone: iphone$1,
    calling: calling$1,
    computer: computer$1,
    keyboard: keyboard$1,
    desktop_computer: desktop_computer$1,
    printer: printer$1,
    computer_mouse: computer_mouse$1,
    trackball: trackball$1,
    joystick: joystick$1,
    clamp: clamp$1,
    minidisc: minidisc$1,
    floppy_disk: floppy_disk$1,
    cd: cd$1,
    dvd: dvd$1,
    vhs: vhs$1,
    camera: camera$1,
    camera_flash: camera_flash$1,
    video_camera: video_camera$1,
    movie_camera: movie_camera$1,
    film_projector: film_projector$1,
    film_strip: film_strip$1,
    telephone_receiver: telephone_receiver$1,
    phone: phone$2,
    pager: pager$1,
    fax: fax$1,
    tv: tv$1,
    radio: radio$1,
    studio_microphone: studio_microphone$1,
    level_slider: level_slider$1,
    control_knobs: control_knobs$1,
    compass: compass$1,
    stopwatch: stopwatch$1,
    timer_clock: timer_clock$1,
    alarm_clock: alarm_clock$1,
    mantelpiece_clock: mantelpiece_clock$1,
    hourglass_flowing_sand: hourglass_flowing_sand$1,
    hourglass: hourglass$1,
    satellite: satellite$1,
    battery: battery$1,
    electric_plug: electric_plug$1,
    bulb: bulb$1,
    flashlight: flashlight$1,
    candle: candle$1,
    fire_extinguisher: fire_extinguisher$1,
    wastebasket: wastebasket$1,
    oil_drum: oil_drum$1,
    money_with_wings: money_with_wings$1,
    dollar: dollar$2,
    yen: yen$2,
    euro: euro$2,
    pound: pound$2,
    moneybag: moneybag$1,
    credit_card: credit_card$1,
    gem: gem$1,
    balance_scale: balance_scale$1,
    toolbox: toolbox$1,
    wrench: wrench$1,
    hammer: hammer$1,
    hammer_and_pick: hammer_and_pick$1,
    hammer_and_wrench: hammer_and_wrench$1,
    pick: pick$1,
    nut_and_bolt: nut_and_bolt$1,
    gear: gear$1,
    brick: brick,
    chains: chains$1,
    magnet: magnet$1,
    gun: gun$1,
    bomb: bomb$1,
    firecracker: firecracker$1,
    hocho: hocho$1,
    dagger: dagger$2,
    crossed_swords: crossed_swords$1,
    shield: shield$1,
    smoking: smoking$1,
    skull_and_crossbones: skull_and_crossbones$1,
    coffin: coffin$1,
    funeral_urn: funeral_urn$1,
    amphora: amphora$1,
    crystal_ball: crystal_ball$1,
    prayer_beads: prayer_beads$1,
    nazar_amulet: nazar_amulet$1,
    barber: barber$1,
    alembic: alembic$1,
    telescope: telescope$1,
    microscope: microscope$1,
    hole: hole$1,
    pill: pill$1,
    syringe: syringe$1,
    dna: dna$1,
    microbe: microbe$1,
    petri_dish: petri_dish$1,
    test_tube: test_tube$1,
    thermometer: thermometer$1,
    broom: broom$1,
    basket: basket$1,
    toilet_paper: toilet_paper,
    label: label$1,
    bookmark: bookmark$1,
    toilet: toilet$1,
    shower: shower$1,
    bathtub: bathtub$1,
    soap: soap$1,
    sponge: sponge$1,
    lotion_bottle: lotion_bottle$1,
    key: key$1,
    old_key: old_key$1,
    couch_and_lamp: couch_and_lamp$1,
    sleeping_bed: sleeping_bed$1,
    bed: bed$1,
    door: door$1,
    bellhop_bell: bellhop_bell$1,
    teddy_bear: teddy_bear$1,
    framed_picture: framed_picture$1,
    world_map: world_map$1,
    parasol_on_ground: parasol_on_ground$1,
    moyai: moyai$1,
    shopping: shopping$1,
    shopping_cart: shopping_cart$1,
    balloon: balloon$1,
    flags: flags$1,
    ribbon: ribbon$1,
    gift: gift$1,
    confetti_ball: confetti_ball$1,
    tada: tada$1,
    dolls: dolls$1,
    wind_chime: wind_chime$1,
    crossed_flags: crossed_flags$1,
    izakaya_lantern: izakaya_lantern$1,
    red_envelope: red_envelope$1,
    email: email$1,
    envelope_with_arrow: envelope_with_arrow$1,
    incoming_envelope: incoming_envelope$1,
    love_letter: love_letter$1,
    postbox: postbox$1,
    mailbox_closed: mailbox_closed$1,
    mailbox: mailbox$1,
    mailbox_with_mail: mailbox_with_mail$1,
    mailbox_with_no_mail: mailbox_with_no_mail$1,
    postal_horn: postal_horn$1,
    inbox_tray: inbox_tray$1,
    outbox_tray: outbox_tray$1,
    scroll: scroll$1,
    page_with_curl: page_with_curl$1,
    bookmark_tabs: bookmark_tabs$1,
    receipt: receipt$1,
    bar_chart: bar_chart$1,
    chart_with_upwards_trend: chart_with_upwards_trend$1,
    chart_with_downwards_trend: chart_with_downwards_trend$1,
    page_facing_up: page_facing_up$1,
    date: date$1,
    calendar: calendar$1,
    spiral_calendar: spiral_calendar$1,
    card_index: card_index$1,
    card_file_box: card_file_box$1,
    ballot_box: ballot_box$1,
    file_cabinet: file_cabinet$1,
    clipboard: clipboard$2,
    spiral_notepad: spiral_notepad$1,
    file_folder: file_folder$1,
    open_file_folder: open_file_folder$1,
    card_index_dividers: card_index_dividers$1,
    newspaper_roll: newspaper_roll$1,
    newspaper: newspaper$1,
    notebook: notebook$1,
    closed_book: closed_book$1,
    green_book: green_book$1,
    blue_book: blue_book$1,
    orange_book: orange_book$1,
    notebook_with_decorative_cover: notebook_with_decorative_cover$1,
    ledger: ledger$1,
    books: books$1,
    open_book: open_book$1,
    safety_pin: safety_pin$1,
    link: link$3,
    paperclip: paperclip$1,
    paperclips: paperclips$1,
    scissors: scissors$1,
    triangular_ruler: triangular_ruler$1,
    straight_ruler: straight_ruler$1,
    abacus: abacus$1,
    pushpin: pushpin$1,
    round_pushpin: round_pushpin$1,
    triangular_flag_on_post: triangular_flag_on_post$1,
    white_flag: white_flag$1,
    black_flag: black_flag$1,
    rainbow_flag: rainbow_flag$1,
    closed_lock_with_key: closed_lock_with_key$1,
    lock: lock$1,
    unlock: unlock$1,
    lock_with_ink_pen: lock_with_ink_pen$1,
    pen: pen$1,
    fountain_pen: fountain_pen$1,
    black_nib: black_nib$1,
    memo: memo$1,
    pencil2: pencil2$1,
    crayon: crayon$1,
    paintbrush: paintbrush$1,
    mag: mag$1,
    mag_right: mag_right$1,
    heart: heart$1,
    orange_heart: orange_heart$1,
    yellow_heart: yellow_heart$1,
    green_heart: green_heart$1,
    blue_heart: blue_heart$1,
    purple_heart: purple_heart$1,
    black_heart: black_heart$1,
    broken_heart: broken_heart$1,
    heavy_heart_exclamation: heavy_heart_exclamation$1,
    two_hearts: two_hearts$1,
    revolving_hearts: revolving_hearts$1,
    heartbeat: heartbeat$1,
    heartpulse: heartpulse$1,
    sparkling_heart: sparkling_heart$1,
    cupid: cupid$1,
    gift_heart: gift_heart$1,
    heart_decoration: heart_decoration$1,
    peace_symbol: peace_symbol$1,
    latin_cross: latin_cross$1,
    star_and_crescent: star_and_crescent$1,
    om: om$1,
    wheel_of_dharma: wheel_of_dharma$1,
    star_of_david: star_of_david$1,
    six_pointed_star: six_pointed_star$1,
    menorah: menorah$1,
    yin_yang: yin_yang$1,
    orthodox_cross: orthodox_cross$1,
    place_of_worship: place_of_worship$1,
    ophiuchus: ophiuchus$1,
    aries: aries$1,
    taurus: taurus$1,
    gemini: gemini$1,
    cancer: cancer$1,
    leo: leo$1,
    virgo: virgo$1,
    libra: libra$1,
    scorpius: scorpius$1,
    sagittarius: sagittarius$1,
    capricorn: capricorn$1,
    aquarius: aquarius$1,
    pisces: pisces$1,
    id: id$1,
    atom_symbol: atom_symbol$1,
    u7a7a: u7a7a,
    u5272: u5272,
    radioactive: radioactive$1,
    biohazard: biohazard$1,
    mobile_phone_off: mobile_phone_off$1,
    vibration_mode: vibration_mode$1,
    u6709: u6709,
    u7121: u7121,
    u7533: u7533,
    u55b6: u55b6,
    u6708: u6708,
    eight_pointed_black_star: eight_pointed_black_star$1,
    vs: vs$1,
    accept: accept$1,
    white_flower: white_flower$1,
    ideograph_advantage: ideograph_advantage$1,
    secret: secret$1,
    congratulations: congratulations$1,
    u5408: u5408,
    u6e80: u6e80$1,
    u7981: u7981,
    a: a$1,
    b: b$1,
    ab: ab$1,
    cl: cl$1,
    o2: o2$1,
    sos: sos$1,
    no_entry: no_entry$1,
    name_badge: name_badge$1,
    no_entry_sign: no_entry_sign$1,
    x: x$1,
    o: o$1,
    stop_sign: stop_sign$1,
    anger: anger$1,
    hotsprings: hotsprings$1,
    no_pedestrians: no_pedestrians$1,
    do_not_litter: do_not_litter$1,
    no_bicycles: no_bicycles$1,
    underage: underage$1,
    no_mobile_phones: no_mobile_phones$1,
    exclamation: exclamation$1,
    grey_exclamation: grey_exclamation$1,
    question: question$1,
    grey_question: grey_question$1,
    bangbang: bangbang$1,
    interrobang: interrobang$1,
    low_brightness: low_brightness$1,
    high_brightness: high_brightness$1,
    trident: trident$1,
    fleur_de_lis: fleur_de_lis$1,
    part_alternation_mark: part_alternation_mark$1,
    warning: warning$1,
    children_crossing: children_crossing$1,
    beginner: beginner$1,
    recycle: recycle$1,
    u6307: u6307,
    chart: chart$1,
    sparkle: sparkle$1,
    eight_spoked_asterisk: eight_spoked_asterisk$1,
    negative_squared_cross_mark: negative_squared_cross_mark$1,
    white_check_mark: white_check_mark$1,
    diamond_shape_with_a_dot_inside: diamond_shape_with_a_dot_inside$1,
    cyclone: cyclone$1,
    loop: loop$1,
    globe_with_meridians: globe_with_meridians$1,
    m: m$1,
    atm: atm$1,
    sa: sa$1,
    passport_control: passport_control$1,
    customs: customs$1,
    baggage_claim: baggage_claim$1,
    left_luggage: left_luggage$1,
    wheelchair: wheelchair$1,
    no_smoking: no_smoking$1,
    wc: wc$1,
    parking: parking$1,
    potable_water: potable_water$1,
    mens: mens$1,
    womens: womens$1,
    baby_symbol: baby_symbol$1,
    restroom: restroom$1,
    put_litter_in_its_place: put_litter_in_its_place$1,
    cinema: cinema$1,
    signal_strength: signal_strength$1,
    koko: koko$1,
    ng: ng$1,
    ok: ok$1,
    up: up$1,
    cool: cool$1,
    free: free$1,
    zero: zero$2,
    one: one$1,
    two: two$1,
    three: three$1,
    four: four$1,
    five: five$1,
    six: six$1,
    seven: seven$1,
    eight: eight$1,
    nine: nine$1,
    keycap_ten: keycap_ten$1,
    asterisk: asterisk$1,
    eject_button: eject_button$1,
    arrow_forward: arrow_forward$1,
    pause_button: pause_button$1,
    next_track_button: next_track_button$1,
    stop_button: stop_button$1,
    record_button: record_button$1,
    play_or_pause_button: play_or_pause_button$1,
    previous_track_button: previous_track_button$1,
    fast_forward: fast_forward$1,
    rewind: rewind$1,
    twisted_rightwards_arrows: twisted_rightwards_arrows$1,
    repeat: repeat$1,
    repeat_one: repeat_one$1,
    arrow_backward: arrow_backward$1,
    arrow_up_small: arrow_up_small$1,
    arrow_down_small: arrow_down_small$1,
    arrow_double_up: arrow_double_up$1,
    arrow_double_down: arrow_double_down$1,
    arrow_right: arrow_right$1,
    arrow_left: arrow_left$1,
    arrow_up: arrow_up$1,
    arrow_down: arrow_down$1,
    arrow_upper_right: arrow_upper_right$1,
    arrow_lower_right: arrow_lower_right$1,
    arrow_lower_left: arrow_lower_left$1,
    arrow_upper_left: arrow_upper_left$1,
    arrow_up_down: arrow_up_down$1,
    left_right_arrow: left_right_arrow$1,
    arrows_counterclockwise: arrows_counterclockwise$1,
    arrow_right_hook: arrow_right_hook$1,
    leftwards_arrow_with_hook: leftwards_arrow_with_hook$1,
    arrow_heading_up: arrow_heading_up$1,
    arrow_heading_down: arrow_heading_down$1,
    hash: hash$1,
    information_source: information_source$1,
    abc: abc$1,
    abcd: abcd$1,
    capital_abcd: capital_abcd$1,
    symbols: symbols$1,
    musical_note: musical_note$1,
    notes: notes$1,
    wavy_dash: wavy_dash$1,
    curly_loop: curly_loop$1,
    heavy_check_mark: heavy_check_mark$1,
    arrows_clockwise: arrows_clockwise$1,
    heavy_plus_sign: heavy_plus_sign$1,
    heavy_minus_sign: heavy_minus_sign$1,
    heavy_division_sign: heavy_division_sign$1,
    heavy_multiplication_x: heavy_multiplication_x$1,
    infinity: infinity$1,
    heavy_dollar_sign: heavy_dollar_sign$1,
    currency_exchange: currency_exchange$1,
    copyright: copyright$1,
    registered: registered$1,
    tm: tm$1,
    end: end$1,
    back: back$1,
    on: on$1,
    top: top$2,
    soon: soon$1,
    ballot_box_with_check: ballot_box_with_check$1,
    radio_button: radio_button$1,
    white_circle: white_circle$1,
    black_circle: black_circle$1,
    red_circle: red_circle$1,
    large_blue_circle: large_blue_circle$1,
    small_orange_diamond: small_orange_diamond$1,
    small_blue_diamond: small_blue_diamond$1,
    large_orange_diamond: large_orange_diamond$1,
    large_blue_diamond: large_blue_diamond$1,
    small_red_triangle: small_red_triangle$1,
    black_small_square: black_small_square$1,
    white_small_square: white_small_square$1,
    black_large_square: black_large_square$1,
    white_large_square: white_large_square$1,
    small_red_triangle_down: small_red_triangle_down$1,
    black_medium_square: black_medium_square$1,
    white_medium_square: white_medium_square$1,
    black_medium_small_square: black_medium_small_square$1,
    white_medium_small_square: white_medium_small_square$1,
    black_square_button: black_square_button$1,
    white_square_button: white_square_button$1,
    speaker: speaker$1,
    sound: sound$1,
    loud_sound: loud_sound$1,
    mute: mute$1,
    mega: mega$1,
    loudspeaker: loudspeaker$1,
    bell: bell$1,
    no_bell: no_bell$1,
    black_joker: black_joker$1,
    mahjong: mahjong$1,
    spades: spades$2,
    clubs: clubs$2,
    hearts: hearts$2,
    diamonds: diamonds$1,
    flower_playing_cards: flower_playing_cards$1,
    thought_balloon: thought_balloon$1,
    right_anger_bubble: right_anger_bubble$1,
    speech_balloon: speech_balloon$1,
    left_speech_bubble: left_speech_bubble$1,
    clock1: clock1$1,
    clock2: clock2$1,
    clock3: clock3$1,
    clock4: clock4$1,
    clock5: clock5$1,
    clock6: clock6$1,
    clock7: clock7$1,
    clock8: clock8$1,
    clock9: clock9$1,
    clock10: clock10$1,
    clock11: clock11$1,
    clock12: clock12$1,
    clock130: clock130$1,
    clock230: clock230$1,
    clock330: clock330$1,
    clock430: clock430$1,
    clock530: clock530$1,
    clock630: clock630$1,
    clock730: clock730$1,
    clock830: clock830$1,
    clock930: clock930$1,
    clock1030: clock1030$1,
    clock1130: clock1130$1,
    clock1230: clock1230$1,
    afghanistan: afghanistan$1,
    aland_islands: aland_islands$1,
    albania: albania$1,
    algeria: algeria$1,
    american_samoa: american_samoa$1,
    andorra: andorra$1,
    angola: angola$1,
    anguilla: anguilla$1,
    antarctica: antarctica$1,
    antigua_barbuda: antigua_barbuda$1,
    argentina: argentina$1,
    armenia: armenia$1,
    aruba: aruba$1,
    australia: australia$1,
    austria: austria$1,
    azerbaijan: azerbaijan$1,
    bahamas: bahamas$1,
    bahrain: bahrain$1,
    bangladesh: bangladesh$1,
    barbados: barbados$1,
    belarus: belarus$1,
    belgium: belgium$1,
    belize: belize$1,
    benin: benin$1,
    bermuda: bermuda$1,
    bhutan: bhutan$1,
    bolivia: bolivia$1,
    caribbean_netherlands: caribbean_netherlands$1,
    bosnia_herzegovina: bosnia_herzegovina$1,
    botswana: botswana$1,
    brazil: brazil$1,
    british_indian_ocean_territory: british_indian_ocean_territory$1,
    british_virgin_islands: british_virgin_islands$1,
    brunei: brunei$1,
    bulgaria: bulgaria$1,
    burkina_faso: burkina_faso$1,
    burundi: burundi$1,
    cape_verde: cape_verde$1,
    cambodia: cambodia$1,
    cameroon: cameroon$1,
    canada: canada$1,
    canary_islands: canary_islands$1,
    cayman_islands: cayman_islands$1,
    central_african_republic: central_african_republic$1,
    chad: chad$1,
    chile: chile$1,
    cn: cn$1,
    christmas_island: christmas_island$1,
    cocos_islands: cocos_islands$1,
    colombia: colombia$1,
    comoros: comoros$1,
    congo_brazzaville: congo_brazzaville$1,
    congo_kinshasa: congo_kinshasa$1,
    cook_islands: cook_islands$1,
    costa_rica: costa_rica$1,
    croatia: croatia$1,
    cuba: cuba$1,
    curacao: curacao$1,
    cyprus: cyprus$1,
    czech_republic: czech_republic$1,
    denmark: denmark$1,
    djibouti: djibouti$1,
    dominica: dominica$1,
    dominican_republic: dominican_republic$1,
    ecuador: ecuador$1,
    egypt: egypt$1,
    el_salvador: el_salvador$1,
    equatorial_guinea: equatorial_guinea$1,
    eritrea: eritrea$1,
    estonia: estonia$1,
    ethiopia: ethiopia$1,
    eu: eu$1,
    falkland_islands: falkland_islands$1,
    faroe_islands: faroe_islands$1,
    fiji: fiji$1,
    finland: finland$1,
    fr: fr$1,
    french_guiana: french_guiana$1,
    french_polynesia: french_polynesia$1,
    french_southern_territories: french_southern_territories$1,
    gabon: gabon$1,
    gambia: gambia$1,
    georgia: georgia$1,
    de: de$1,
    ghana: ghana$1,
    gibraltar: gibraltar$1,
    greece: greece$1,
    greenland: greenland$1,
    grenada: grenada$1,
    guadeloupe: guadeloupe$1,
    guam: guam$1,
    guatemala: guatemala$1,
    guernsey: guernsey$1,
    guinea: guinea$1,
    guinea_bissau: guinea_bissau$1,
    guyana: guyana$1,
    haiti: haiti$1,
    honduras: honduras$1,
    hong_kong: hong_kong$1,
    hungary: hungary$1,
    iceland: iceland$1,
    india: india$1,
    indonesia: indonesia$1,
    iran: iran$1,
    iraq: iraq$1,
    ireland: ireland$1,
    isle_of_man: isle_of_man$1,
    israel: israel$1,
    it: it$2,
    cote_divoire: cote_divoire$1,
    jamaica: jamaica$1,
    jp: jp$1,
    jersey: jersey$1,
    jordan: jordan$1,
    kazakhstan: kazakhstan$1,
    kenya: kenya$1,
    kiribati: kiribati$1,
    kosovo: kosovo$1,
    kuwait: kuwait$1,
    kyrgyzstan: kyrgyzstan$1,
    laos: laos$1,
    latvia: latvia$1,
    lebanon: lebanon$1,
    lesotho: lesotho$1,
    liberia: liberia$1,
    libya: libya$1,
    liechtenstein: liechtenstein$1,
    lithuania: lithuania$1,
    luxembourg: luxembourg$1,
    macau: macau$1,
    macedonia: macedonia$1,
    madagascar: madagascar$1,
    malawi: malawi$1,
    malaysia: malaysia$1,
    maldives: maldives$1,
    mali: mali$1,
    malta: malta$1,
    marshall_islands: marshall_islands$1,
    martinique: martinique$1,
    mauritania: mauritania$1,
    mauritius: mauritius$1,
    mayotte: mayotte$1,
    mexico: mexico$1,
    micronesia: micronesia$1,
    moldova: moldova$1,
    monaco: monaco$1,
    mongolia: mongolia$1,
    montenegro: montenegro$1,
    montserrat: montserrat$1,
    morocco: morocco$1,
    mozambique: mozambique$1,
    myanmar: myanmar$1,
    namibia: namibia$1,
    nauru: nauru$1,
    nepal: nepal$1,
    netherlands: netherlands$1,
    new_caledonia: new_caledonia$1,
    new_zealand: new_zealand$1,
    nicaragua: nicaragua$1,
    niger: niger$1,
    nigeria: nigeria$1,
    niue: niue$1,
    norfolk_island: norfolk_island$1,
    northern_mariana_islands: northern_mariana_islands$1,
    north_korea: north_korea$1,
    norway: norway$1,
    oman: oman$1,
    pakistan: pakistan$1,
    palau: palau$1,
    palestinian_territories: palestinian_territories$1,
    panama: panama$1,
    papua_new_guinea: papua_new_guinea$1,
    paraguay: paraguay$1,
    peru: peru$1,
    philippines: philippines$1,
    pitcairn_islands: pitcairn_islands$1,
    poland: poland$1,
    portugal: portugal$1,
    puerto_rico: puerto_rico$1,
    qatar: qatar$1,
    reunion: reunion$1,
    romania: romania$1,
    ru: ru$1,
    rwanda: rwanda$1,
    st_barthelemy: st_barthelemy$1,
    st_helena: st_helena$1,
    st_kitts_nevis: st_kitts_nevis$1,
    st_lucia: st_lucia$1,
    st_pierre_miquelon: st_pierre_miquelon$1,
    st_vincent_grenadines: st_vincent_grenadines$1,
    samoa: samoa$1,
    san_marino: san_marino$1,
    sao_tome_principe: sao_tome_principe$1,
    saudi_arabia: saudi_arabia$1,
    senegal: senegal$1,
    serbia: serbia$1,
    seychelles: seychelles$1,
    sierra_leone: sierra_leone$1,
    singapore: singapore$1,
    sint_maarten: sint_maarten$1,
    slovakia: slovakia$1,
    slovenia: slovenia$1,
    solomon_islands: solomon_islands$1,
    somalia: somalia$1,
    south_africa: south_africa$1,
    south_georgia_south_sandwich_islands: south_georgia_south_sandwich_islands$1,
    kr: kr$1,
    south_sudan: south_sudan$1,
    es: es$1,
    sri_lanka: sri_lanka$1,
    sudan: sudan$1,
    suriname: suriname$1,
    swaziland: swaziland$1,
    sweden: sweden$1,
    switzerland: switzerland$1,
    syria: syria$1,
    taiwan: taiwan$1,
    tajikistan: tajikistan$1,
    tanzania: tanzania$1,
    thailand: thailand$1,
    timor_leste: timor_leste$1,
    togo: togo$1,
    tokelau: tokelau$1,
    tonga: tonga$1,
    trinidad_tobago: trinidad_tobago$1,
    tunisia: tunisia$1,
    tr: tr$1,
    turkmenistan: turkmenistan$1,
    turks_caicos_islands: turks_caicos_islands$1,
    tuvalu: tuvalu$1,
    uganda: uganda$1,
    ukraine: ukraine$1,
    united_arab_emirates: united_arab_emirates$1,
    uk: uk$1,
    england: england$1,
    scotland: scotland$1,
    wales: wales$1,
    us: us$1,
    us_virgin_islands: us_virgin_islands$1,
    uruguay: uruguay$1,
    uzbekistan: uzbekistan$1,
    vanuatu: vanuatu$1,
    vatican_city: vatican_city$1,
    venezuela: venezuela$1,
    vietnam: vietnam$1,
    wallis_futuna: wallis_futuna$1,
    western_sahara: western_sahara$1,
    yemen: yemen$1,
    zambia: zambia$1,
    zimbabwe: zimbabwe$1,
    united_nations: united_nations$1,
    pirate_flag: pirate_flag$1,
    'default': emojis
  });

  var ordered = [
  	"grinning",
  	"smiley",
  	"smile",
  	"grin",
  	"laughing",
  	"sweat_smile",
  	"joy",
  	"rofl",
  	"relaxed",
  	"blush",
  	"innocent",
  	"slightly_smiling_face",
  	"upside_down_face",
  	"wink",
  	"relieved",
  	"heart_eyes",
  	"smiling_face_with_three_hearts",
  	"kissing_heart",
  	"kissing",
  	"kissing_smiling_eyes",
  	"kissing_closed_eyes",
  	"yum",
  	"stuck_out_tongue",
  	"stuck_out_tongue_closed_eyes",
  	"stuck_out_tongue_winking_eye",
  	"zany",
  	"raised_eyebrow",
  	"monocle",
  	"nerd_face",
  	"sunglasses",
  	"star_struck",
  	"partying",
  	"smirk",
  	"unamused",
  	"disappointed",
  	"pensive",
  	"worried",
  	"confused",
  	"slightly_frowning_face",
  	"frowning_face",
  	"persevere",
  	"confounded",
  	"tired_face",
  	"weary",
  	"pleading",
  	"cry",
  	"sob",
  	"triumph",
  	"angry",
  	"rage",
  	"symbols_over_mouth",
  	"exploding_head",
  	"flushed",
  	"hot",
  	"cold",
  	"scream",
  	"fearful",
  	"cold_sweat",
  	"disappointed_relieved",
  	"sweat",
  	"hugs",
  	"thinking",
  	"hand_over_mouth",
  	"shushing",
  	"lying_face",
  	"no_mouth",
  	"neutral_face",
  	"expressionless",
  	"grimacing",
  	"roll_eyes",
  	"hushed",
  	"frowning",
  	"anguished",
  	"open_mouth",
  	"astonished",
  	"sleeping",
  	"drooling_face",
  	"sleepy",
  	"dizzy_face",
  	"zipper_mouth_face",
  	"woozy",
  	"nauseated_face",
  	"vomiting",
  	"sneezing_face",
  	"mask",
  	"face_with_thermometer",
  	"face_with_head_bandage",
  	"money_mouth_face",
  	"cowboy_hat_face",
  	"smiling_imp",
  	"imp",
  	"japanese_ogre",
  	"japanese_goblin",
  	"clown_face",
  	"poop",
  	"ghost",
  	"skull",
  	"skull_and_crossbones",
  	"alien",
  	"space_invader",
  	"robot",
  	"jack_o_lantern",
  	"smiley_cat",
  	"smile_cat",
  	"joy_cat",
  	"heart_eyes_cat",
  	"smirk_cat",
  	"kissing_cat",
  	"scream_cat",
  	"crying_cat_face",
  	"pouting_cat",
  	"palms_up",
  	"open_hands",
  	"raised_hands",
  	"clap",
  	"handshake",
  	"+1",
  	"-1",
  	"facepunch",
  	"fist",
  	"fist_left",
  	"fist_right",
  	"crossed_fingers",
  	"v",
  	"love_you",
  	"metal",
  	"ok_hand",
  	"point_left",
  	"point_right",
  	"point_up",
  	"point_down",
  	"point_up_2",
  	"raised_hand",
  	"raised_back_of_hand",
  	"raised_hand_with_fingers_splayed",
  	"vulcan_salute",
  	"wave",
  	"call_me_hand",
  	"muscle",
  	"fu",
  	"writing_hand",
  	"pray",
  	"foot",
  	"leg",
  	"ring",
  	"lipstick",
  	"kiss",
  	"lips",
  	"tooth",
  	"tongue",
  	"ear",
  	"nose",
  	"footprints",
  	"eye",
  	"eyes",
  	"brain",
  	"speaking_head",
  	"bust_in_silhouette",
  	"busts_in_silhouette",
  	"baby",
  	"girl",
  	"child",
  	"boy",
  	"woman",
  	"adult",
  	"man",
  	"blonde_woman",
  	"blonde_man",
  	"bearded_person",
  	"older_woman",
  	"older_adult",
  	"older_man",
  	"man_with_gua_pi_mao",
  	"woman_with_headscarf",
  	"woman_with_turban",
  	"man_with_turban",
  	"policewoman",
  	"policeman",
  	"construction_worker_woman",
  	"construction_worker_man",
  	"guardswoman",
  	"guardsman",
  	"female_detective",
  	"male_detective",
  	"woman_health_worker",
  	"man_health_worker",
  	"woman_farmer",
  	"man_farmer",
  	"woman_cook",
  	"man_cook",
  	"woman_student",
  	"man_student",
  	"woman_singer",
  	"man_singer",
  	"woman_teacher",
  	"man_teacher",
  	"woman_factory_worker",
  	"man_factory_worker",
  	"woman_technologist",
  	"man_technologist",
  	"woman_office_worker",
  	"man_office_worker",
  	"woman_mechanic",
  	"man_mechanic",
  	"woman_scientist",
  	"man_scientist",
  	"woman_artist",
  	"man_artist",
  	"woman_firefighter",
  	"man_firefighter",
  	"woman_pilot",
  	"man_pilot",
  	"woman_astronaut",
  	"man_astronaut",
  	"woman_judge",
  	"man_judge",
  	"bride_with_veil",
  	"man_in_tuxedo",
  	"princess",
  	"prince",
  	"woman_superhero",
  	"man_superhero",
  	"woman_supervillain",
  	"man_supervillain",
  	"mrs_claus",
  	"santa",
  	"sorceress",
  	"wizard",
  	"woman_elf",
  	"man_elf",
  	"woman_vampire",
  	"man_vampire",
  	"woman_zombie",
  	"man_zombie",
  	"woman_genie",
  	"man_genie",
  	"mermaid",
  	"merman",
  	"woman_fairy",
  	"man_fairy",
  	"angel",
  	"pregnant_woman",
  	"breastfeeding",
  	"bowing_woman",
  	"bowing_man",
  	"tipping_hand_woman",
  	"tipping_hand_man",
  	"no_good_woman",
  	"no_good_man",
  	"ok_woman",
  	"ok_man",
  	"raising_hand_woman",
  	"raising_hand_man",
  	"woman_facepalming",
  	"man_facepalming",
  	"woman_shrugging",
  	"man_shrugging",
  	"pouting_woman",
  	"pouting_man",
  	"frowning_woman",
  	"frowning_man",
  	"haircut_woman",
  	"haircut_man",
  	"massage_woman",
  	"massage_man",
  	"woman_in_steamy_room",
  	"man_in_steamy_room",
  	"nail_care",
  	"selfie",
  	"dancer",
  	"man_dancing",
  	"dancing_women",
  	"dancing_men",
  	"business_suit_levitating",
  	"walking_woman",
  	"walking_man",
  	"running_woman",
  	"running_man",
  	"couple",
  	"two_women_holding_hands",
  	"two_men_holding_hands",
  	"couple_with_heart_woman_man",
  	"couple_with_heart_woman_woman",
  	"couple_with_heart_man_man",
  	"couplekiss_man_woman",
  	"couplekiss_woman_woman",
  	"couplekiss_man_man",
  	"family_man_woman_boy",
  	"family_man_woman_girl",
  	"family_man_woman_girl_boy",
  	"family_man_woman_boy_boy",
  	"family_man_woman_girl_girl",
  	"family_woman_woman_boy",
  	"family_woman_woman_girl",
  	"family_woman_woman_girl_boy",
  	"family_woman_woman_boy_boy",
  	"family_woman_woman_girl_girl",
  	"family_man_man_boy",
  	"family_man_man_girl",
  	"family_man_man_girl_boy",
  	"family_man_man_boy_boy",
  	"family_man_man_girl_girl",
  	"family_woman_boy",
  	"family_woman_girl",
  	"family_woman_girl_boy",
  	"family_woman_boy_boy",
  	"family_woman_girl_girl",
  	"family_man_boy",
  	"family_man_girl",
  	"family_man_girl_boy",
  	"family_man_boy_boy",
  	"family_man_girl_girl",
  	"yarn",
  	"thread",
  	"coat",
  	"labcoat",
  	"womans_clothes",
  	"tshirt",
  	"jeans",
  	"necktie",
  	"dress",
  	"bikini",
  	"kimono",
  	"flat_shoe",
  	"high_heel",
  	"sandal",
  	"boot",
  	"mans_shoe",
  	"athletic_shoe",
  	"hiking_boot",
  	"socks",
  	"gloves",
  	"scarf",
  	"tophat",
  	"billed_hat",
  	"womans_hat",
  	"mortar_board",
  	"rescue_worker_helmet",
  	"crown",
  	"pouch",
  	"purse",
  	"handbag",
  	"briefcase",
  	"school_satchel",
  	"luggage",
  	"eyeglasses",
  	"dark_sunglasses",
  	"goggles",
  	"closed_umbrella",
  	"dog",
  	"cat",
  	"mouse",
  	"hamster",
  	"rabbit",
  	"fox_face",
  	"bear",
  	"panda_face",
  	"koala",
  	"tiger",
  	"lion",
  	"cow",
  	"pig",
  	"pig_nose",
  	"frog",
  	"monkey_face",
  	"see_no_evil",
  	"hear_no_evil",
  	"speak_no_evil",
  	"monkey",
  	"chicken",
  	"penguin",
  	"bird",
  	"baby_chick",
  	"hatching_chick",
  	"hatched_chick",
  	"duck",
  	"eagle",
  	"owl",
  	"bat",
  	"wolf",
  	"boar",
  	"horse",
  	"unicorn",
  	"honeybee",
  	"bug",
  	"butterfly",
  	"snail",
  	"shell",
  	"beetle",
  	"ant",
  	"mosquito",
  	"grasshopper",
  	"spider",
  	"spider_web",
  	"scorpion",
  	"turtle",
  	"snake",
  	"lizard",
  	"t-rex",
  	"sauropod",
  	"octopus",
  	"squid",
  	"shrimp",
  	"lobster",
  	"crab",
  	"blowfish",
  	"tropical_fish",
  	"fish",
  	"dolphin",
  	"whale",
  	"whale2",
  	"shark",
  	"crocodile",
  	"tiger2",
  	"leopard",
  	"zebra",
  	"gorilla",
  	"elephant",
  	"hippopotamus",
  	"rhinoceros",
  	"dromedary_camel",
  	"giraffe",
  	"kangaroo",
  	"camel",
  	"water_buffalo",
  	"ox",
  	"cow2",
  	"racehorse",
  	"pig2",
  	"ram",
  	"sheep",
  	"llama",
  	"goat",
  	"deer",
  	"dog2",
  	"poodle",
  	"cat2",
  	"rooster",
  	"turkey",
  	"peacock",
  	"parrot",
  	"swan",
  	"dove",
  	"rabbit2",
  	"raccoon",
  	"badger",
  	"rat",
  	"mouse2",
  	"chipmunk",
  	"hedgehog",
  	"paw_prints",
  	"dragon",
  	"dragon_face",
  	"cactus",
  	"christmas_tree",
  	"evergreen_tree",
  	"deciduous_tree",
  	"palm_tree",
  	"seedling",
  	"herb",
  	"shamrock",
  	"four_leaf_clover",
  	"bamboo",
  	"tanabata_tree",
  	"leaves",
  	"fallen_leaf",
  	"maple_leaf",
  	"ear_of_rice",
  	"hibiscus",
  	"sunflower",
  	"rose",
  	"wilted_flower",
  	"tulip",
  	"blossom",
  	"cherry_blossom",
  	"bouquet",
  	"mushroom",
  	"earth_americas",
  	"earth_africa",
  	"earth_asia",
  	"full_moon",
  	"waning_gibbous_moon",
  	"last_quarter_moon",
  	"waning_crescent_moon",
  	"new_moon",
  	"waxing_crescent_moon",
  	"first_quarter_moon",
  	"waxing_gibbous_moon",
  	"new_moon_with_face",
  	"full_moon_with_face",
  	"first_quarter_moon_with_face",
  	"last_quarter_moon_with_face",
  	"sun_with_face",
  	"crescent_moon",
  	"star",
  	"star2",
  	"dizzy",
  	"sparkles",
  	"comet",
  	"sunny",
  	"sun_behind_small_cloud",
  	"partly_sunny",
  	"sun_behind_large_cloud",
  	"sun_behind_rain_cloud",
  	"cloud",
  	"cloud_with_rain",
  	"cloud_with_lightning_and_rain",
  	"cloud_with_lightning",
  	"zap",
  	"fire",
  	"boom",
  	"snowflake",
  	"cloud_with_snow",
  	"snowman",
  	"snowman_with_snow",
  	"wind_face",
  	"dash",
  	"tornado",
  	"fog",
  	"open_umbrella",
  	"umbrella",
  	"droplet",
  	"sweat_drops",
  	"ocean",
  	"green_apple",
  	"apple",
  	"pear",
  	"tangerine",
  	"lemon",
  	"banana",
  	"watermelon",
  	"grapes",
  	"strawberry",
  	"melon",
  	"cherries",
  	"peach",
  	"mango",
  	"pineapple",
  	"coconut",
  	"kiwi_fruit",
  	"tomato",
  	"eggplant",
  	"avocado",
  	"broccoli",
  	"leafy_greens",
  	"cucumber",
  	"hot_pepper",
  	"corn",
  	"carrot",
  	"potato",
  	"sweet_potato",
  	"croissant",
  	"bagel",
  	"bread",
  	"baguette_bread",
  	"pretzel",
  	"cheese",
  	"egg",
  	"fried_egg",
  	"pancakes",
  	"bacon",
  	"steak",
  	"poultry_leg",
  	"meat_on_bone",
  	"bone",
  	"hotdog",
  	"hamburger",
  	"fries",
  	"pizza",
  	"sandwich",
  	"stuffed_flatbread",
  	"taco",
  	"burrito",
  	"green_salad",
  	"shallow_pan_of_food",
  	"canned_food",
  	"spaghetti",
  	"ramen",
  	"stew",
  	"curry",
  	"sushi",
  	"bento",
  	"fried_shrimp",
  	"rice_ball",
  	"rice",
  	"rice_cracker",
  	"fish_cake",
  	"fortune_cookie",
  	"moon_cake",
  	"oden",
  	"dango",
  	"shaved_ice",
  	"ice_cream",
  	"icecream",
  	"pie",
  	"cupcake",
  	"cake",
  	"birthday",
  	"custard",
  	"lollipop",
  	"candy",
  	"chocolate_bar",
  	"popcorn",
  	"doughnut",
  	"dumpling",
  	"cookie",
  	"chestnut",
  	"peanuts",
  	"honey_pot",
  	"milk_glass",
  	"baby_bottle",
  	"coffee",
  	"tea",
  	"cup_with_straw",
  	"sake",
  	"beer",
  	"beers",
  	"clinking_glasses",
  	"wine_glass",
  	"tumbler_glass",
  	"cocktail",
  	"tropical_drink",
  	"champagne",
  	"spoon",
  	"fork_and_knife",
  	"plate_with_cutlery",
  	"bowl_with_spoon",
  	"takeout_box",
  	"chopsticks",
  	"salt",
  	"soccer",
  	"basketball",
  	"football",
  	"baseball",
  	"softball",
  	"tennis",
  	"volleyball",
  	"rugby_football",
  	"flying_disc",
  	"8ball",
  	"golf",
  	"golfing_woman",
  	"golfing_man",
  	"ping_pong",
  	"badminton",
  	"goal_net",
  	"ice_hockey",
  	"field_hockey",
  	"lacrosse",
  	"cricket",
  	"ski",
  	"skier",
  	"snowboarder",
  	"person_fencing",
  	"women_wrestling",
  	"men_wrestling",
  	"woman_cartwheeling",
  	"man_cartwheeling",
  	"woman_playing_handball",
  	"man_playing_handball",
  	"ice_skate",
  	"curling_stone",
  	"skateboard",
  	"sled",
  	"bow_and_arrow",
  	"fishing_pole_and_fish",
  	"boxing_glove",
  	"martial_arts_uniform",
  	"rowing_woman",
  	"rowing_man",
  	"climbing_woman",
  	"climbing_man",
  	"swimming_woman",
  	"swimming_man",
  	"woman_playing_water_polo",
  	"man_playing_water_polo",
  	"woman_in_lotus_position",
  	"man_in_lotus_position",
  	"surfing_woman",
  	"surfing_man",
  	"basketball_woman",
  	"basketball_man",
  	"weight_lifting_woman",
  	"weight_lifting_man",
  	"biking_woman",
  	"biking_man",
  	"mountain_biking_woman",
  	"mountain_biking_man",
  	"horse_racing",
  	"trophy",
  	"running_shirt_with_sash",
  	"medal_sports",
  	"medal_military",
  	"1st_place_medal",
  	"2nd_place_medal",
  	"3rd_place_medal",
  	"reminder_ribbon",
  	"rosette",
  	"ticket",
  	"tickets",
  	"performing_arts",
  	"art",
  	"circus_tent",
  	"woman_juggling",
  	"man_juggling",
  	"microphone",
  	"headphones",
  	"musical_score",
  	"musical_keyboard",
  	"drum",
  	"saxophone",
  	"trumpet",
  	"guitar",
  	"violin",
  	"clapper",
  	"video_game",
  	"dart",
  	"game_die",
  	"chess_pawn",
  	"slot_machine",
  	"jigsaw",
  	"bowling",
  	"red_car",
  	"taxi",
  	"blue_car",
  	"bus",
  	"trolleybus",
  	"racing_car",
  	"police_car",
  	"ambulance",
  	"fire_engine",
  	"minibus",
  	"truck",
  	"articulated_lorry",
  	"tractor",
  	"kick_scooter",
  	"motorcycle",
  	"bike",
  	"motor_scooter",
  	"rotating_light",
  	"oncoming_police_car",
  	"oncoming_bus",
  	"oncoming_automobile",
  	"oncoming_taxi",
  	"aerial_tramway",
  	"mountain_cableway",
  	"suspension_railway",
  	"railway_car",
  	"train",
  	"monorail",
  	"bullettrain_side",
  	"bullettrain_front",
  	"light_rail",
  	"mountain_railway",
  	"steam_locomotive",
  	"train2",
  	"metro",
  	"tram",
  	"station",
  	"flying_saucer",
  	"helicopter",
  	"small_airplane",
  	"airplane",
  	"flight_departure",
  	"flight_arrival",
  	"sailboat",
  	"motor_boat",
  	"speedboat",
  	"ferry",
  	"passenger_ship",
  	"rocket",
  	"artificial_satellite",
  	"seat",
  	"canoe",
  	"anchor",
  	"construction",
  	"fuelpump",
  	"busstop",
  	"vertical_traffic_light",
  	"traffic_light",
  	"ship",
  	"ferris_wheel",
  	"roller_coaster",
  	"carousel_horse",
  	"building_construction",
  	"foggy",
  	"tokyo_tower",
  	"factory",
  	"fountain",
  	"rice_scene",
  	"mountain",
  	"mountain_snow",
  	"mount_fuji",
  	"volcano",
  	"japan",
  	"camping",
  	"tent",
  	"national_park",
  	"motorway",
  	"railway_track",
  	"sunrise",
  	"sunrise_over_mountains",
  	"desert",
  	"beach_umbrella",
  	"desert_island",
  	"city_sunrise",
  	"city_sunset",
  	"cityscape",
  	"night_with_stars",
  	"bridge_at_night",
  	"milky_way",
  	"stars",
  	"sparkler",
  	"fireworks",
  	"rainbow",
  	"houses",
  	"european_castle",
  	"japanese_castle",
  	"stadium",
  	"statue_of_liberty",
  	"house",
  	"house_with_garden",
  	"derelict_house",
  	"office",
  	"department_store",
  	"post_office",
  	"european_post_office",
  	"hospital",
  	"bank",
  	"hotel",
  	"convenience_store",
  	"school",
  	"love_hotel",
  	"wedding",
  	"classical_building",
  	"church",
  	"mosque",
  	"synagogue",
  	"kaaba",
  	"shinto_shrine",
  	"watch",
  	"iphone",
  	"calling",
  	"computer",
  	"keyboard",
  	"desktop_computer",
  	"printer",
  	"computer_mouse",
  	"trackball",
  	"joystick",
  	"clamp",
  	"minidisc",
  	"floppy_disk",
  	"cd",
  	"dvd",
  	"vhs",
  	"camera",
  	"camera_flash",
  	"video_camera",
  	"movie_camera",
  	"film_projector",
  	"film_strip",
  	"telephone_receiver",
  	"phone",
  	"pager",
  	"fax",
  	"tv",
  	"radio",
  	"studio_microphone",
  	"level_slider",
  	"control_knobs",
  	"compass",
  	"stopwatch",
  	"timer_clock",
  	"alarm_clock",
  	"mantelpiece_clock",
  	"hourglass_flowing_sand",
  	"hourglass",
  	"satellite",
  	"battery",
  	"electric_plug",
  	"bulb",
  	"flashlight",
  	"candle",
  	"fire_extinguisher",
  	"wastebasket",
  	"oil_drum",
  	"money_with_wings",
  	"dollar",
  	"yen",
  	"euro",
  	"pound",
  	"moneybag",
  	"credit_card",
  	"gem",
  	"balance_scale",
  	"toolbox",
  	"wrench",
  	"hammer",
  	"hammer_and_pick",
  	"hammer_and_wrench",
  	"pick",
  	"nut_and_bolt",
  	"gear",
  	"brick",
  	"chains",
  	"magnet",
  	"gun",
  	"bomb",
  	"firecracker",
  	"hocho",
  	"dagger",
  	"crossed_swords",
  	"shield",
  	"smoking",
  	"coffin",
  	"funeral_urn",
  	"amphora",
  	"crystal_ball",
  	"prayer_beads",
  	"nazar_amulet",
  	"barber",
  	"alembic",
  	"telescope",
  	"microscope",
  	"hole",
  	"pill",
  	"syringe",
  	"dna",
  	"microbe",
  	"petri_dish",
  	"test_tube",
  	"thermometer",
  	"broom",
  	"basket",
  	"toilet_paper",
  	"label",
  	"bookmark",
  	"toilet",
  	"shower",
  	"bathtub",
  	"bath",
  	"soap",
  	"sponge",
  	"lotion_bottle",
  	"key",
  	"old_key",
  	"couch_and_lamp",
  	"sleeping_bed",
  	"bed",
  	"door",
  	"bellhop_bell",
  	"teddy_bear",
  	"framed_picture",
  	"world_map",
  	"parasol_on_ground",
  	"moyai",
  	"shopping",
  	"shopping_cart",
  	"balloon",
  	"flags",
  	"ribbon",
  	"gift",
  	"confetti_ball",
  	"tada",
  	"dolls",
  	"wind_chime",
  	"crossed_flags",
  	"izakaya_lantern",
  	"red_envelope",
  	"email",
  	"envelope_with_arrow",
  	"incoming_envelope",
  	"e-mail",
  	"love_letter",
  	"postbox",
  	"mailbox_closed",
  	"mailbox",
  	"mailbox_with_mail",
  	"mailbox_with_no_mail",
  	"package",
  	"postal_horn",
  	"inbox_tray",
  	"outbox_tray",
  	"scroll",
  	"page_with_curl",
  	"bookmark_tabs",
  	"receipt",
  	"bar_chart",
  	"chart_with_upwards_trend",
  	"chart_with_downwards_trend",
  	"page_facing_up",
  	"date",
  	"calendar",
  	"spiral_calendar",
  	"card_index",
  	"card_file_box",
  	"ballot_box",
  	"file_cabinet",
  	"clipboard",
  	"spiral_notepad",
  	"file_folder",
  	"open_file_folder",
  	"card_index_dividers",
  	"newspaper_roll",
  	"newspaper",
  	"notebook",
  	"closed_book",
  	"green_book",
  	"blue_book",
  	"orange_book",
  	"notebook_with_decorative_cover",
  	"ledger",
  	"books",
  	"open_book",
  	"safety_pin",
  	"link",
  	"paperclip",
  	"paperclips",
  	"scissors",
  	"triangular_ruler",
  	"straight_ruler",
  	"abacus",
  	"pushpin",
  	"round_pushpin",
  	"closed_lock_with_key",
  	"lock",
  	"unlock",
  	"lock_with_ink_pen",
  	"pen",
  	"fountain_pen",
  	"black_nib",
  	"memo",
  	"pencil2",
  	"crayon",
  	"paintbrush",
  	"mag",
  	"mag_right",
  	"heart",
  	"orange_heart",
  	"yellow_heart",
  	"green_heart",
  	"blue_heart",
  	"purple_heart",
  	"black_heart",
  	"broken_heart",
  	"heavy_heart_exclamation",
  	"two_hearts",
  	"revolving_hearts",
  	"heartbeat",
  	"heartpulse",
  	"sparkling_heart",
  	"cupid",
  	"gift_heart",
  	"heart_decoration",
  	"peace_symbol",
  	"latin_cross",
  	"star_and_crescent",
  	"om",
  	"wheel_of_dharma",
  	"star_of_david",
  	"six_pointed_star",
  	"menorah",
  	"yin_yang",
  	"orthodox_cross",
  	"place_of_worship",
  	"ophiuchus",
  	"aries",
  	"taurus",
  	"gemini",
  	"cancer",
  	"leo",
  	"virgo",
  	"libra",
  	"scorpius",
  	"sagittarius",
  	"capricorn",
  	"aquarius",
  	"pisces",
  	"id",
  	"atom_symbol",
  	"u7a7a",
  	"u5272",
  	"radioactive",
  	"biohazard",
  	"mobile_phone_off",
  	"vibration_mode",
  	"u6709",
  	"u7121",
  	"u7533",
  	"u55b6",
  	"u6708",
  	"eight_pointed_black_star",
  	"vs",
  	"accept",
  	"white_flower",
  	"ideograph_advantage",
  	"secret",
  	"congratulations",
  	"u5408",
  	"u6e80",
  	"u7981",
  	"a",
  	"b",
  	"ab",
  	"cl",
  	"o2",
  	"sos",
  	"no_entry",
  	"name_badge",
  	"no_entry_sign",
  	"x",
  	"o",
  	"stop_sign",
  	"anger",
  	"hotsprings",
  	"no_pedestrians",
  	"do_not_litter",
  	"no_bicycles",
  	"non-potable_water",
  	"underage",
  	"no_mobile_phones",
  	"exclamation",
  	"grey_exclamation",
  	"question",
  	"grey_question",
  	"bangbang",
  	"interrobang",
  	"100",
  	"low_brightness",
  	"high_brightness",
  	"trident",
  	"fleur_de_lis",
  	"part_alternation_mark",
  	"warning",
  	"children_crossing",
  	"beginner",
  	"recycle",
  	"u6307",
  	"chart",
  	"sparkle",
  	"eight_spoked_asterisk",
  	"negative_squared_cross_mark",
  	"white_check_mark",
  	"diamond_shape_with_a_dot_inside",
  	"cyclone",
  	"loop",
  	"globe_with_meridians",
  	"m",
  	"atm",
  	"zzz",
  	"sa",
  	"passport_control",
  	"customs",
  	"baggage_claim",
  	"left_luggage",
  	"wheelchair",
  	"no_smoking",
  	"wc",
  	"parking",
  	"potable_water",
  	"mens",
  	"womens",
  	"baby_symbol",
  	"restroom",
  	"put_litter_in_its_place",
  	"cinema",
  	"signal_strength",
  	"koko",
  	"ng",
  	"ok",
  	"up",
  	"cool",
  	"new",
  	"free",
  	"zero",
  	"one",
  	"two",
  	"three",
  	"four",
  	"five",
  	"six",
  	"seven",
  	"eight",
  	"nine",
  	"keycap_ten",
  	"asterisk",
  	"1234",
  	"eject_button",
  	"arrow_forward",
  	"pause_button",
  	"next_track_button",
  	"stop_button",
  	"record_button",
  	"play_or_pause_button",
  	"previous_track_button",
  	"fast_forward",
  	"rewind",
  	"twisted_rightwards_arrows",
  	"repeat",
  	"repeat_one",
  	"arrow_backward",
  	"arrow_up_small",
  	"arrow_down_small",
  	"arrow_double_up",
  	"arrow_double_down",
  	"arrow_right",
  	"arrow_left",
  	"arrow_up",
  	"arrow_down",
  	"arrow_upper_right",
  	"arrow_lower_right",
  	"arrow_lower_left",
  	"arrow_upper_left",
  	"arrow_up_down",
  	"left_right_arrow",
  	"arrows_counterclockwise",
  	"arrow_right_hook",
  	"leftwards_arrow_with_hook",
  	"arrow_heading_up",
  	"arrow_heading_down",
  	"hash",
  	"information_source",
  	"abc",
  	"abcd",
  	"capital_abcd",
  	"symbols",
  	"musical_note",
  	"notes",
  	"wavy_dash",
  	"curly_loop",
  	"heavy_check_mark",
  	"arrows_clockwise",
  	"heavy_plus_sign",
  	"heavy_minus_sign",
  	"heavy_division_sign",
  	"heavy_multiplication_x",
  	"infinity",
  	"heavy_dollar_sign",
  	"currency_exchange",
  	"copyright",
  	"registered",
  	"tm",
  	"end",
  	"back",
  	"on",
  	"top",
  	"soon",
  	"ballot_box_with_check",
  	"radio_button",
  	"white_circle",
  	"black_circle",
  	"red_circle",
  	"large_blue_circle",
  	"small_orange_diamond",
  	"small_blue_diamond",
  	"large_orange_diamond",
  	"large_blue_diamond",
  	"small_red_triangle",
  	"black_small_square",
  	"white_small_square",
  	"black_large_square",
  	"white_large_square",
  	"small_red_triangle_down",
  	"black_medium_square",
  	"white_medium_square",
  	"black_medium_small_square",
  	"white_medium_small_square",
  	"black_square_button",
  	"white_square_button",
  	"speaker",
  	"sound",
  	"loud_sound",
  	"mute",
  	"mega",
  	"loudspeaker",
  	"bell",
  	"no_bell",
  	"black_joker",
  	"mahjong",
  	"spades",
  	"clubs",
  	"hearts",
  	"diamonds",
  	"flower_playing_cards",
  	"thought_balloon",
  	"right_anger_bubble",
  	"speech_balloon",
  	"left_speech_bubble",
  	"clock1",
  	"clock2",
  	"clock3",
  	"clock4",
  	"clock5",
  	"clock6",
  	"clock7",
  	"clock8",
  	"clock9",
  	"clock10",
  	"clock11",
  	"clock12",
  	"clock130",
  	"clock230",
  	"clock330",
  	"clock430",
  	"clock530",
  	"clock630",
  	"clock730",
  	"clock830",
  	"clock930",
  	"clock1030",
  	"clock1130",
  	"clock1230",
  	"white_flag",
  	"black_flag",
  	"pirate_flag",
  	"checkered_flag",
  	"triangular_flag_on_post",
  	"rainbow_flag",
  	"united_nations",
  	"afghanistan",
  	"aland_islands",
  	"albania",
  	"algeria",
  	"american_samoa",
  	"andorra",
  	"angola",
  	"anguilla",
  	"antarctica",
  	"antigua_barbuda",
  	"argentina",
  	"armenia",
  	"aruba",
  	"australia",
  	"austria",
  	"azerbaijan",
  	"bahamas",
  	"bahrain",
  	"bangladesh",
  	"barbados",
  	"belarus",
  	"belgium",
  	"belize",
  	"benin",
  	"bermuda",
  	"bhutan",
  	"bolivia",
  	"caribbean_netherlands",
  	"bosnia_herzegovina",
  	"botswana",
  	"brazil",
  	"british_indian_ocean_territory",
  	"british_virgin_islands",
  	"brunei",
  	"bulgaria",
  	"burkina_faso",
  	"burundi",
  	"cape_verde",
  	"cambodia",
  	"cameroon",
  	"canada",
  	"canary_islands",
  	"cayman_islands",
  	"central_african_republic",
  	"chad",
  	"chile",
  	"cn",
  	"christmas_island",
  	"cocos_islands",
  	"colombia",
  	"comoros",
  	"congo_brazzaville",
  	"congo_kinshasa",
  	"cook_islands",
  	"costa_rica",
  	"croatia",
  	"cuba",
  	"curacao",
  	"cyprus",
  	"czech_republic",
  	"denmark",
  	"djibouti",
  	"dominica",
  	"dominican_republic",
  	"ecuador",
  	"egypt",
  	"el_salvador",
  	"equatorial_guinea",
  	"eritrea",
  	"estonia",
  	"ethiopia",
  	"eu",
  	"falkland_islands",
  	"faroe_islands",
  	"fiji",
  	"finland",
  	"fr",
  	"french_guiana",
  	"french_polynesia",
  	"french_southern_territories",
  	"gabon",
  	"gambia",
  	"georgia",
  	"de",
  	"ghana",
  	"gibraltar",
  	"greece",
  	"greenland",
  	"grenada",
  	"guadeloupe",
  	"guam",
  	"guatemala",
  	"guernsey",
  	"guinea",
  	"guinea_bissau",
  	"guyana",
  	"haiti",
  	"honduras",
  	"hong_kong",
  	"hungary",
  	"iceland",
  	"india",
  	"indonesia",
  	"iran",
  	"iraq",
  	"ireland",
  	"isle_of_man",
  	"israel",
  	"it",
  	"cote_divoire",
  	"jamaica",
  	"jp",
  	"jersey",
  	"jordan",
  	"kazakhstan",
  	"kenya",
  	"kiribati",
  	"kosovo",
  	"kuwait",
  	"kyrgyzstan",
  	"laos",
  	"latvia",
  	"lebanon",
  	"lesotho",
  	"liberia",
  	"libya",
  	"liechtenstein",
  	"lithuania",
  	"luxembourg",
  	"macau",
  	"macedonia",
  	"madagascar",
  	"malawi",
  	"malaysia",
  	"maldives",
  	"mali",
  	"malta",
  	"marshall_islands",
  	"martinique",
  	"mauritania",
  	"mauritius",
  	"mayotte",
  	"mexico",
  	"micronesia",
  	"moldova",
  	"monaco",
  	"mongolia",
  	"montenegro",
  	"montserrat",
  	"morocco",
  	"mozambique",
  	"myanmar",
  	"namibia",
  	"nauru",
  	"nepal",
  	"netherlands",
  	"new_caledonia",
  	"new_zealand",
  	"nicaragua",
  	"niger",
  	"nigeria",
  	"niue",
  	"norfolk_island",
  	"northern_mariana_islands",
  	"north_korea",
  	"norway",
  	"oman",
  	"pakistan",
  	"palau",
  	"palestinian_territories",
  	"panama",
  	"papua_new_guinea",
  	"paraguay",
  	"peru",
  	"philippines",
  	"pitcairn_islands",
  	"poland",
  	"portugal",
  	"puerto_rico",
  	"qatar",
  	"reunion",
  	"romania",
  	"ru",
  	"rwanda",
  	"st_barthelemy",
  	"st_helena",
  	"st_kitts_nevis",
  	"st_lucia",
  	"st_pierre_miquelon",
  	"st_vincent_grenadines",
  	"samoa",
  	"san_marino",
  	"sao_tome_principe",
  	"saudi_arabia",
  	"senegal",
  	"serbia",
  	"seychelles",
  	"sierra_leone",
  	"singapore",
  	"sint_maarten",
  	"slovakia",
  	"slovenia",
  	"solomon_islands",
  	"somalia",
  	"south_africa",
  	"south_georgia_south_sandwich_islands",
  	"kr",
  	"south_sudan",
  	"es",
  	"sri_lanka",
  	"sudan",
  	"suriname",
  	"swaziland",
  	"sweden",
  	"switzerland",
  	"syria",
  	"taiwan",
  	"tajikistan",
  	"tanzania",
  	"thailand",
  	"timor_leste",
  	"togo",
  	"tokelau",
  	"tonga",
  	"trinidad_tobago",
  	"tunisia",
  	"tr",
  	"turkmenistan",
  	"turks_caicos_islands",
  	"tuvalu",
  	"uganda",
  	"ukraine",
  	"united_arab_emirates",
  	"uk",
  	"england",
  	"scotland",
  	"wales",
  	"us",
  	"us_virgin_islands",
  	"uruguay",
  	"uzbekistan",
  	"vanuatu",
  	"vatican_city",
  	"venezuela",
  	"vietnam",
  	"wallis_futuna",
  	"western_sahara",
  	"yemen",
  	"zambia",
  	"zimbabwe"
  ];

  var ordered$1 = /*#__PURE__*/Object.freeze({
    __proto__: null,
    'default': ordered
  });

  var require$$0$1 = getCjsExportFromNamespace(emojis$1);

  var require$$1 = getCjsExportFromNamespace(ordered$1);

  var emojilib = {
    lib: require$$0$1,
    ordered: require$$1,
    fitzpatrick_scale_modifiers: ["🏻", "🏼", "🏽", "🏾", "🏿"]
  };
  emojilib.lib;
  emojilib.ordered;
  emojilib.fitzpatrick_scale_modifiers;

  var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

  /**
   * iterateObject
   * Iterates an object. Note the object field order may differ.
   *
   * @name iterateObject
   * @function
   * @param {Object} obj The input object.
   * @param {Function} fn A function that will be called with the current value, field name and provided object.
   * @return {Function} The `iterateObject` function.
   */
  function iterateObject(obj, fn) {
      var i = 0,
          keys = [];

      if (Array.isArray(obj)) {
          for (; i < obj.length; ++i) {
              if (fn(obj[i], i, obj) === false) {
                  break;
              }
          }
      } else if ((typeof obj === "undefined" ? "undefined" : _typeof(obj)) === "object" && obj !== null) {
          keys = Object.keys(obj);
          for (; i < keys.length; ++i) {
              if (fn(obj[keys[i]], keys[i], obj) === false) {
                  break;
              }
          }
      }
  }

  var lib$3 = iterateObject;

  // Dependencies


  /**
   * mapObject
   * Array-map like for objects.
   *
   * @name mapObject
   * @function
   * @param {Object} obj The input object.
   * @param {Function} fn A function returning the field values.
   * @param {Boolean|Object} clone If `true`, the input object will be cloned.
   * If `clone` is an object, it will be used as target object.
   * @return {Object} The modified object.
   */
  function mapObject(obj, fn, clone) {
      var dst = clone === true ? {} : clone ? clone : obj;
      lib$3(obj, function (v, n, o) {
          dst[n] = fn(v, n, o);
      });
      return dst;
  }

  var lib$2 = mapObject;

  var lib$1 = createCommonjsModule(function (module) {



  var emoji = emojilib;

  var nameMap = module.exports = {};
  nameMap.emoji = lib$2(emoji.lib, function (value) {
      return value.char;
  }, true);
  lib$3(nameMap.emoji, function (value, name, obj) {
      return !value && delete obj[name] || true;
  });

  /**
   * get
   * Gets the emoji character (unicode) by providing the name.
   *
   * @name get
   * @function
   * @param {String} name The emoji name.
   * @return {String} The emoji character (unicode).
   */
  nameMap.get = function (name) {
      if (name.charAt(0) === ":") {
          name = name.slice(1, -1);
      }
      return this.emoji[name];
  };

  emoji = null;
  });

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2018 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var emoji_markdown_it_defs = {};

  var emoji_defs_by_char = (function() {
      var result = {};
      $.each(emojilib.lib, function(name, def) {
          result[def['char']] = name;
          emoji_markdown_it_defs[name] = def['char'];
      });

      return result;
  })();

  // Flatten shortcuts to simple object: { alias: emoji_name }
  var shortcuts = Object.keys(shortcuts$1).reduce(function (acc, key) {
      if (Array.isArray(shortcuts$1[key])) {
          shortcuts$1[key].forEach(function (alias) {
              acc[alias] = key;
          });
          return acc;
      }

      acc[shortcuts$1[key]] = key;
      return acc;
  }, {});

  var getEmojiDefinitionByShortcut = function(shortcut) {
      var result = {
          name: getNameByShortcut(shortcut)
      };

      if(result.name) {
          result.emoji = getCharByName(result.name);
      }

      if(result.emoji) {
          result.$dom = getCharToDom(result.emoji);
      }

      return result;
  };

  var getNameByShortcut = function(shortcut) {
      return String(shortcuts[shortcut]);
  };

  var getCharByName = function(name) {
      return lib$1.get(name);
  };

  var getNameByChar = function(emojiChar) {
      return String(emoji_defs_by_char[emojiChar]);
  };

  var getCharToDom = function(emojiChar, name) {
      name =(typeof name !== 'undefined') ? name : emoji_defs_by_char[emojiChar];
      name = String(name);

      var config = humhub.config.get('ui.richtext.prosemirror', 'emoji');
      var twemojiConfig = config.twemoji || {};
      twemojiConfig.attributes = function (icon, variant) {
          return {
              'data-name': name,
              'style': 'width:16px'
          }
      };

      var parsed = twemoji.parse(emojiChar, twemojiConfig);

      if(parsed && parsed.length) {
          try {
              return $(parsed);
          } catch(e) {
              console.error(e);
          }
          return '';
      }
  };


  var byCategory = undefined;

  var getByCategory = function(category) {
      if(!byCategory) {
          byCategory = {};
          emojilib.ordered.forEach(function (name) {
              var emojiDef = emojilib.lib[name];
              emojiDef.name = String(name);
              byCategory[emojiDef.category] = byCategory[emojiDef.category] || [];
              byCategory[emojiDef.category].push(emojiDef);
          });
      }

      return byCategory[category];
  };

  var getMarkdownItOpts = function() {
      return {defs: emoji_markdown_it_defs};
  };

  var grinning = "😀";
  var smiley = "😃";
  var smile$1 = "😄";
  var grin = "😁";
  var laughing = "😆";
  var satisfied = "😆";
  var sweat_smile = "😅";
  var rofl = "🤣";
  var joy = "😂";
  var slightly_smiling_face = "🙂";
  var upside_down_face = "🙃";
  var wink = "😉";
  var blush = "😊";
  var innocent = "😇";
  var smiling_face_with_three_hearts = "🥰";
  var heart_eyes = "😍";
  var star_struck = "🤩";
  var kissing_heart = "😘";
  var kissing = "😗";
  var relaxed = "☺️";
  var kissing_closed_eyes = "😚";
  var kissing_smiling_eyes = "😙";
  var smiling_face_with_tear = "🥲";
  var yum = "😋";
  var stuck_out_tongue = "😛";
  var stuck_out_tongue_winking_eye = "😜";
  var zany_face = "🤪";
  var stuck_out_tongue_closed_eyes = "😝";
  var money_mouth_face = "🤑";
  var hugs = "🤗";
  var hand_over_mouth = "🤭";
  var shushing_face = "🤫";
  var thinking = "🤔";
  var zipper_mouth_face = "🤐";
  var raised_eyebrow = "🤨";
  var neutral_face = "😐";
  var expressionless = "😑";
  var no_mouth = "😶";
  var smirk = "😏";
  var unamused = "😒";
  var roll_eyes = "🙄";
  var grimacing = "😬";
  var lying_face = "🤥";
  var relieved = "😌";
  var pensive = "😔";
  var sleepy = "😪";
  var drooling_face = "🤤";
  var sleeping = "😴";
  var mask = "😷";
  var face_with_thermometer = "🤒";
  var face_with_head_bandage = "🤕";
  var nauseated_face = "🤢";
  var vomiting_face = "🤮";
  var sneezing_face = "🤧";
  var hot_face = "🥵";
  var cold_face = "🥶";
  var woozy_face = "🥴";
  var dizzy_face = "😵";
  var exploding_head = "🤯";
  var cowboy_hat_face = "🤠";
  var partying_face = "🥳";
  var disguised_face = "🥸";
  var sunglasses = "😎";
  var nerd_face = "🤓";
  var monocle_face = "🧐";
  var confused = "😕";
  var worried = "😟";
  var slightly_frowning_face = "🙁";
  var frowning_face = "☹️";
  var open_mouth = "😮";
  var hushed = "😯";
  var astonished = "😲";
  var flushed = "😳";
  var pleading_face = "🥺";
  var frowning = "😦";
  var anguished = "😧";
  var fearful = "😨";
  var cold_sweat = "😰";
  var disappointed_relieved = "😥";
  var cry = "😢";
  var sob = "😭";
  var scream = "😱";
  var confounded = "😖";
  var persevere = "😣";
  var disappointed = "😞";
  var sweat = "😓";
  var weary = "😩";
  var tired_face = "😫";
  var yawning_face = "🥱";
  var triumph = "😤";
  var rage = "😡";
  var pout = "😡";
  var angry = "😠";
  var cursing_face = "🤬";
  var smiling_imp = "😈";
  var imp = "👿";
  var skull = "💀";
  var skull_and_crossbones = "☠️";
  var hankey = "💩";
  var poop = "💩";
  var shit = "💩";
  var clown_face = "🤡";
  var japanese_ogre = "👹";
  var japanese_goblin = "👺";
  var ghost = "👻";
  var alien = "👽";
  var space_invader = "👾";
  var robot = "🤖";
  var smiley_cat = "😺";
  var smile_cat = "😸";
  var joy_cat = "😹";
  var heart_eyes_cat = "😻";
  var smirk_cat = "😼";
  var kissing_cat = "😽";
  var scream_cat = "🙀";
  var crying_cat_face = "😿";
  var pouting_cat = "😾";
  var see_no_evil = "🙈";
  var hear_no_evil = "🙉";
  var speak_no_evil = "🙊";
  var kiss = "💋";
  var love_letter = "💌";
  var cupid = "💘";
  var gift_heart = "💝";
  var sparkling_heart = "💖";
  var heartpulse = "💗";
  var heartbeat = "💓";
  var revolving_hearts = "💞";
  var two_hearts = "💕";
  var heart_decoration = "💟";
  var heavy_heart_exclamation = "❣️";
  var broken_heart = "💔";
  var heart = "❤️";
  var orange_heart = "🧡";
  var yellow_heart = "💛";
  var green_heart = "💚";
  var blue_heart = "💙";
  var purple_heart = "💜";
  var brown_heart = "🤎";
  var black_heart = "🖤";
  var white_heart = "🤍";
  var anger = "💢";
  var boom = "💥";
  var collision = "💥";
  var dizzy = "💫";
  var sweat_drops = "💦";
  var dash$1 = "💨";
  var hole = "🕳️";
  var bomb = "💣";
  var speech_balloon = "💬";
  var eye_speech_bubble = "👁️‍🗨️";
  var left_speech_bubble = "🗨️";
  var right_anger_bubble = "🗯️";
  var thought_balloon = "💭";
  var zzz = "💤";
  var wave = "👋";
  var raised_back_of_hand = "🤚";
  var raised_hand_with_fingers_splayed = "🖐️";
  var hand = "✋";
  var raised_hand = "✋";
  var vulcan_salute = "🖖";
  var ok_hand = "👌";
  var pinched_fingers = "🤌";
  var pinching_hand = "🤏";
  var v = "✌️";
  var crossed_fingers = "🤞";
  var love_you_gesture = "🤟";
  var metal = "🤘";
  var call_me_hand = "🤙";
  var point_left = "👈";
  var point_right = "👉";
  var point_up_2 = "👆";
  var middle_finger = "🖕";
  var fu = "🖕";
  var point_down = "👇";
  var point_up = "☝️";
  var thumbsup = "👍";
  var thumbsdown = "👎";
  var fist_raised = "✊";
  var fist = "✊";
  var fist_oncoming = "👊";
  var facepunch = "👊";
  var punch = "👊";
  var fist_left = "🤛";
  var fist_right = "🤜";
  var clap = "👏";
  var raised_hands = "🙌";
  var open_hands = "👐";
  var palms_up_together = "🤲";
  var handshake = "🤝";
  var pray = "🙏";
  var writing_hand = "✍️";
  var nail_care = "💅";
  var selfie = "🤳";
  var muscle = "💪";
  var mechanical_arm = "🦾";
  var mechanical_leg = "🦿";
  var leg$1 = "🦵";
  var foot = "🦶";
  var ear = "👂";
  var ear_with_hearing_aid = "🦻";
  var nose = "👃";
  var brain = "🧠";
  var anatomical_heart = "🫀";
  var lungs = "🫁";
  var tooth = "🦷";
  var bone = "🦴";
  var eyes = "👀";
  var eye = "👁️";
  var tongue = "👅";
  var lips = "👄";
  var baby = "👶";
  var child = "🧒";
  var boy = "👦";
  var girl = "👧";
  var adult = "🧑";
  var blond_haired_person = "👱";
  var man = "👨";
  var bearded_person = "🧔";
  var red_haired_man = "👨‍🦰";
  var curly_haired_man = "👨‍🦱";
  var white_haired_man = "👨‍🦳";
  var bald_man = "👨‍🦲";
  var woman = "👩";
  var red_haired_woman = "👩‍🦰";
  var person_red_hair = "🧑‍🦰";
  var curly_haired_woman = "👩‍🦱";
  var person_curly_hair = "🧑‍🦱";
  var white_haired_woman = "👩‍🦳";
  var person_white_hair = "🧑‍🦳";
  var bald_woman = "👩‍🦲";
  var person_bald = "🧑‍🦲";
  var blond_haired_woman = "👱‍♀️";
  var blonde_woman = "👱‍♀️";
  var blond_haired_man = "👱‍♂️";
  var older_adult = "🧓";
  var older_man = "👴";
  var older_woman = "👵";
  var frowning_person = "🙍";
  var frowning_man = "🙍‍♂️";
  var frowning_woman = "🙍‍♀️";
  var pouting_face = "🙎";
  var pouting_man = "🙎‍♂️";
  var pouting_woman = "🙎‍♀️";
  var no_good = "🙅";
  var no_good_man = "🙅‍♂️";
  var ng_man = "🙅‍♂️";
  var no_good_woman = "🙅‍♀️";
  var ng_woman = "🙅‍♀️";
  var ok_person = "🙆";
  var ok_man = "🙆‍♂️";
  var ok_woman = "🙆‍♀️";
  var tipping_hand_person = "💁";
  var information_desk_person = "💁";
  var tipping_hand_man = "💁‍♂️";
  var sassy_man = "💁‍♂️";
  var tipping_hand_woman = "💁‍♀️";
  var sassy_woman = "💁‍♀️";
  var raising_hand = "🙋";
  var raising_hand_man = "🙋‍♂️";
  var raising_hand_woman = "🙋‍♀️";
  var deaf_person = "🧏";
  var deaf_man = "🧏‍♂️";
  var deaf_woman = "🧏‍♀️";
  var bow = "🙇";
  var bowing_man = "🙇‍♂️";
  var bowing_woman = "🙇‍♀️";
  var facepalm = "🤦";
  var man_facepalming = "🤦‍♂️";
  var woman_facepalming = "🤦‍♀️";
  var shrug = "🤷";
  var man_shrugging = "🤷‍♂️";
  var woman_shrugging = "🤷‍♀️";
  var health_worker = "🧑‍⚕️";
  var man_health_worker = "👨‍⚕️";
  var woman_health_worker = "👩‍⚕️";
  var student = "🧑‍🎓";
  var man_student = "👨‍🎓";
  var woman_student = "👩‍🎓";
  var teacher = "🧑‍🏫";
  var man_teacher = "👨‍🏫";
  var woman_teacher = "👩‍🏫";
  var judge = "🧑‍⚖️";
  var man_judge = "👨‍⚖️";
  var woman_judge = "👩‍⚖️";
  var farmer = "🧑‍🌾";
  var man_farmer = "👨‍🌾";
  var woman_farmer = "👩‍🌾";
  var cook = "🧑‍🍳";
  var man_cook = "👨‍🍳";
  var woman_cook = "👩‍🍳";
  var mechanic = "🧑‍🔧";
  var man_mechanic = "👨‍🔧";
  var woman_mechanic = "👩‍🔧";
  var factory_worker = "🧑‍🏭";
  var man_factory_worker = "👨‍🏭";
  var woman_factory_worker = "👩‍🏭";
  var office_worker = "🧑‍💼";
  var man_office_worker = "👨‍💼";
  var woman_office_worker = "👩‍💼";
  var scientist = "🧑‍🔬";
  var man_scientist = "👨‍🔬";
  var woman_scientist = "👩‍🔬";
  var technologist = "🧑‍💻";
  var man_technologist = "👨‍💻";
  var woman_technologist = "👩‍💻";
  var singer = "🧑‍🎤";
  var man_singer = "👨‍🎤";
  var woman_singer = "👩‍🎤";
  var artist = "🧑‍🎨";
  var man_artist = "👨‍🎨";
  var woman_artist = "👩‍🎨";
  var pilot = "🧑‍✈️";
  var man_pilot = "👨‍✈️";
  var woman_pilot = "👩‍✈️";
  var astronaut = "🧑‍🚀";
  var man_astronaut = "👨‍🚀";
  var woman_astronaut = "👩‍🚀";
  var firefighter = "🧑‍🚒";
  var man_firefighter = "👨‍🚒";
  var woman_firefighter = "👩‍🚒";
  var police_officer = "👮";
  var cop = "👮";
  var policeman = "👮‍♂️";
  var policewoman = "👮‍♀️";
  var detective = "🕵️";
  var male_detective = "🕵️‍♂️";
  var female_detective = "🕵️‍♀️";
  var guard = "💂";
  var guardsman = "💂‍♂️";
  var guardswoman = "💂‍♀️";
  var ninja = "🥷";
  var construction_worker = "👷";
  var construction_worker_man = "👷‍♂️";
  var construction_worker_woman = "👷‍♀️";
  var prince = "🤴";
  var princess = "👸";
  var person_with_turban = "👳";
  var man_with_turban = "👳‍♂️";
  var woman_with_turban = "👳‍♀️";
  var man_with_gua_pi_mao = "👲";
  var woman_with_headscarf = "🧕";
  var person_in_tuxedo = "🤵";
  var man_in_tuxedo = "🤵‍♂️";
  var woman_in_tuxedo = "🤵‍♀️";
  var person_with_veil = "👰";
  var man_with_veil = "👰‍♂️";
  var woman_with_veil = "👰‍♀️";
  var bride_with_veil = "👰‍♀️";
  var pregnant_woman = "🤰";
  var breast_feeding = "🤱";
  var woman_feeding_baby = "👩‍🍼";
  var man_feeding_baby = "👨‍🍼";
  var person_feeding_baby = "🧑‍🍼";
  var angel = "👼";
  var santa = "🎅";
  var mrs_claus = "🤶";
  var mx_claus = "🧑‍🎄";
  var superhero = "🦸";
  var superhero_man = "🦸‍♂️";
  var superhero_woman = "🦸‍♀️";
  var supervillain = "🦹";
  var supervillain_man = "🦹‍♂️";
  var supervillain_woman = "🦹‍♀️";
  var mage = "🧙";
  var mage_man = "🧙‍♂️";
  var mage_woman = "🧙‍♀️";
  var fairy = "🧚";
  var fairy_man = "🧚‍♂️";
  var fairy_woman = "🧚‍♀️";
  var vampire = "🧛";
  var vampire_man = "🧛‍♂️";
  var vampire_woman = "🧛‍♀️";
  var merperson = "🧜";
  var merman = "🧜‍♂️";
  var mermaid = "🧜‍♀️";
  var elf = "🧝";
  var elf_man = "🧝‍♂️";
  var elf_woman = "🧝‍♀️";
  var genie = "🧞";
  var genie_man = "🧞‍♂️";
  var genie_woman = "🧞‍♀️";
  var zombie = "🧟";
  var zombie_man = "🧟‍♂️";
  var zombie_woman = "🧟‍♀️";
  var massage = "💆";
  var massage_man = "💆‍♂️";
  var massage_woman = "💆‍♀️";
  var haircut = "💇";
  var haircut_man = "💇‍♂️";
  var haircut_woman = "💇‍♀️";
  var walking = "🚶";
  var walking_man = "🚶‍♂️";
  var walking_woman = "🚶‍♀️";
  var standing_person = "🧍";
  var standing_man = "🧍‍♂️";
  var standing_woman = "🧍‍♀️";
  var kneeling_person = "🧎";
  var kneeling_man = "🧎‍♂️";
  var kneeling_woman = "🧎‍♀️";
  var person_with_probing_cane = "🧑‍🦯";
  var man_with_probing_cane = "👨‍🦯";
  var woman_with_probing_cane = "👩‍🦯";
  var person_in_motorized_wheelchair = "🧑‍🦼";
  var man_in_motorized_wheelchair = "👨‍🦼";
  var woman_in_motorized_wheelchair = "👩‍🦼";
  var person_in_manual_wheelchair = "🧑‍🦽";
  var man_in_manual_wheelchair = "👨‍🦽";
  var woman_in_manual_wheelchair = "👩‍🦽";
  var runner = "🏃";
  var running = "🏃";
  var running_man = "🏃‍♂️";
  var running_woman = "🏃‍♀️";
  var woman_dancing = "💃";
  var dancer = "💃";
  var man_dancing = "🕺";
  var business_suit_levitating = "🕴️";
  var dancers = "👯";
  var dancing_men = "👯‍♂️";
  var dancing_women = "👯‍♀️";
  var sauna_person = "🧖";
  var sauna_man = "🧖‍♂️";
  var sauna_woman = "🧖‍♀️";
  var climbing = "🧗";
  var climbing_man = "🧗‍♂️";
  var climbing_woman = "🧗‍♀️";
  var person_fencing = "🤺";
  var horse_racing = "🏇";
  var skier = "⛷️";
  var snowboarder = "🏂";
  var golfing = "🏌️";
  var golfing_man = "🏌️‍♂️";
  var golfing_woman = "🏌️‍♀️";
  var surfer = "🏄";
  var surfing_man = "🏄‍♂️";
  var surfing_woman = "🏄‍♀️";
  var rowboat = "🚣";
  var rowing_man = "🚣‍♂️";
  var rowing_woman = "🚣‍♀️";
  var swimmer = "🏊";
  var swimming_man = "🏊‍♂️";
  var swimming_woman = "🏊‍♀️";
  var bouncing_ball_person = "⛹️";
  var bouncing_ball_man = "⛹️‍♂️";
  var basketball_man = "⛹️‍♂️";
  var bouncing_ball_woman = "⛹️‍♀️";
  var basketball_woman = "⛹️‍♀️";
  var weight_lifting = "🏋️";
  var weight_lifting_man = "🏋️‍♂️";
  var weight_lifting_woman = "🏋️‍♀️";
  var bicyclist = "🚴";
  var biking_man = "🚴‍♂️";
  var biking_woman = "🚴‍♀️";
  var mountain_bicyclist = "🚵";
  var mountain_biking_man = "🚵‍♂️";
  var mountain_biking_woman = "🚵‍♀️";
  var cartwheeling = "🤸";
  var man_cartwheeling = "🤸‍♂️";
  var woman_cartwheeling = "🤸‍♀️";
  var wrestling = "🤼";
  var men_wrestling = "🤼‍♂️";
  var women_wrestling = "🤼‍♀️";
  var water_polo = "🤽";
  var man_playing_water_polo = "🤽‍♂️";
  var woman_playing_water_polo = "🤽‍♀️";
  var handball_person = "🤾";
  var man_playing_handball = "🤾‍♂️";
  var woman_playing_handball = "🤾‍♀️";
  var juggling_person = "🤹";
  var man_juggling = "🤹‍♂️";
  var woman_juggling = "🤹‍♀️";
  var lotus_position = "🧘";
  var lotus_position_man = "🧘‍♂️";
  var lotus_position_woman = "🧘‍♀️";
  var bath = "🛀";
  var sleeping_bed = "🛌";
  var people_holding_hands = "🧑‍🤝‍🧑";
  var two_women_holding_hands = "👭";
  var couple = "👫";
  var two_men_holding_hands = "👬";
  var couplekiss = "💏";
  var couplekiss_man_woman = "👩‍❤️‍💋‍👨";
  var couplekiss_man_man = "👨‍❤️‍💋‍👨";
  var couplekiss_woman_woman = "👩‍❤️‍💋‍👩";
  var couple_with_heart = "💑";
  var couple_with_heart_woman_man = "👩‍❤️‍👨";
  var couple_with_heart_man_man = "👨‍❤️‍👨";
  var couple_with_heart_woman_woman = "👩‍❤️‍👩";
  var family = "👪";
  var family_man_woman_boy = "👨‍👩‍👦";
  var family_man_woman_girl = "👨‍👩‍👧";
  var family_man_woman_girl_boy = "👨‍👩‍👧‍👦";
  var family_man_woman_boy_boy = "👨‍👩‍👦‍👦";
  var family_man_woman_girl_girl = "👨‍👩‍👧‍👧";
  var family_man_man_boy = "👨‍👨‍👦";
  var family_man_man_girl = "👨‍👨‍👧";
  var family_man_man_girl_boy = "👨‍👨‍👧‍👦";
  var family_man_man_boy_boy = "👨‍👨‍👦‍👦";
  var family_man_man_girl_girl = "👨‍👨‍👧‍👧";
  var family_woman_woman_boy = "👩‍👩‍👦";
  var family_woman_woman_girl = "👩‍👩‍👧";
  var family_woman_woman_girl_boy = "👩‍👩‍👧‍👦";
  var family_woman_woman_boy_boy = "👩‍👩‍👦‍👦";
  var family_woman_woman_girl_girl = "👩‍👩‍👧‍👧";
  var family_man_boy = "👨‍👦";
  var family_man_boy_boy = "👨‍👦‍👦";
  var family_man_girl = "👨‍👧";
  var family_man_girl_boy = "👨‍👧‍👦";
  var family_man_girl_girl = "👨‍👧‍👧";
  var family_woman_boy = "👩‍👦";
  var family_woman_boy_boy = "👩‍👦‍👦";
  var family_woman_girl = "👩‍👧";
  var family_woman_girl_boy = "👩‍👧‍👦";
  var family_woman_girl_girl = "👩‍👧‍👧";
  var speaking_head = "🗣️";
  var bust_in_silhouette = "👤";
  var busts_in_silhouette = "👥";
  var people_hugging = "🫂";
  var footprints = "👣";
  var monkey_face = "🐵";
  var monkey = "🐒";
  var gorilla = "🦍";
  var orangutan = "🦧";
  var dog = "🐶";
  var dog2 = "🐕";
  var guide_dog = "🦮";
  var service_dog = "🐕‍🦺";
  var poodle = "🐩";
  var wolf = "🐺";
  var fox_face = "🦊";
  var raccoon = "🦝";
  var cat = "🐱";
  var cat2 = "🐈";
  var black_cat = "🐈‍⬛";
  var lion = "🦁";
  var tiger = "🐯";
  var tiger2 = "🐅";
  var leopard = "🐆";
  var horse = "🐴";
  var racehorse = "🐎";
  var unicorn = "🦄";
  var zebra = "🦓";
  var deer = "🦌";
  var bison = "🦬";
  var cow = "🐮";
  var ox = "🐂";
  var water_buffalo = "🐃";
  var cow2 = "🐄";
  var pig = "🐷";
  var pig2 = "🐖";
  var boar = "🐗";
  var pig_nose = "🐽";
  var ram = "🐏";
  var sheep = "🐑";
  var goat = "🐐";
  var dromedary_camel = "🐪";
  var camel = "🐫";
  var llama = "🦙";
  var giraffe = "🦒";
  var elephant = "🐘";
  var mammoth = "🦣";
  var rhinoceros = "🦏";
  var hippopotamus = "🦛";
  var mouse = "🐭";
  var mouse2 = "🐁";
  var rat = "🐀";
  var hamster = "🐹";
  var rabbit = "🐰";
  var rabbit2 = "🐇";
  var chipmunk = "🐿️";
  var beaver = "🦫";
  var hedgehog = "🦔";
  var bat = "🦇";
  var bear = "🐻";
  var polar_bear = "🐻‍❄️";
  var koala = "🐨";
  var panda_face = "🐼";
  var sloth = "🦥";
  var otter = "🦦";
  var skunk = "🦨";
  var kangaroo = "🦘";
  var badger = "🦡";
  var feet = "🐾";
  var paw_prints = "🐾";
  var turkey = "🦃";
  var chicken = "🐔";
  var rooster = "🐓";
  var hatching_chick = "🐣";
  var baby_chick = "🐤";
  var hatched_chick = "🐥";
  var bird = "🐦";
  var penguin = "🐧";
  var dove = "🕊️";
  var eagle = "🦅";
  var duck = "🦆";
  var swan = "🦢";
  var owl = "🦉";
  var dodo = "🦤";
  var feather = "🪶";
  var flamingo = "🦩";
  var peacock = "🦚";
  var parrot = "🦜";
  var frog = "🐸";
  var crocodile = "🐊";
  var turtle = "🐢";
  var lizard = "🦎";
  var snake = "🐍";
  var dragon_face = "🐲";
  var dragon = "🐉";
  var sauropod = "🦕";
  var whale = "🐳";
  var whale2 = "🐋";
  var dolphin = "🐬";
  var flipper = "🐬";
  var seal = "🦭";
  var fish = "🐟";
  var tropical_fish = "🐠";
  var blowfish = "🐡";
  var shark = "🦈";
  var octopus = "🐙";
  var shell = "🐚";
  var snail = "🐌";
  var butterfly = "🦋";
  var bug = "🐛";
  var ant = "🐜";
  var bee = "🐝";
  var honeybee = "🐝";
  var beetle = "🪲";
  var lady_beetle = "🐞";
  var cricket = "🦗";
  var cockroach = "🪳";
  var spider = "🕷️";
  var spider_web = "🕸️";
  var scorpion = "🦂";
  var mosquito = "🦟";
  var fly = "🪰";
  var worm = "🪱";
  var microbe = "🦠";
  var bouquet = "💐";
  var cherry_blossom = "🌸";
  var white_flower = "💮";
  var rosette = "🏵️";
  var rose = "🌹";
  var wilted_flower = "🥀";
  var hibiscus = "🌺";
  var sunflower = "🌻";
  var blossom = "🌼";
  var tulip = "🌷";
  var seedling = "🌱";
  var potted_plant = "🪴";
  var evergreen_tree = "🌲";
  var deciduous_tree = "🌳";
  var palm_tree = "🌴";
  var cactus = "🌵";
  var ear_of_rice = "🌾";
  var herb = "🌿";
  var shamrock = "☘️";
  var four_leaf_clover = "🍀";
  var maple_leaf = "🍁";
  var fallen_leaf = "🍂";
  var leaves = "🍃";
  var grapes = "🍇";
  var melon = "🍈";
  var watermelon = "🍉";
  var tangerine = "🍊";
  var orange = "🍊";
  var mandarin = "🍊";
  var lemon = "🍋";
  var banana = "🍌";
  var pineapple = "🍍";
  var mango = "🥭";
  var apple = "🍎";
  var green_apple = "🍏";
  var pear = "🍐";
  var peach = "🍑";
  var cherries = "🍒";
  var strawberry = "🍓";
  var blueberries = "🫐";
  var kiwi_fruit = "🥝";
  var tomato = "🍅";
  var olive = "🫒";
  var coconut = "🥥";
  var avocado = "🥑";
  var eggplant = "🍆";
  var potato = "🥔";
  var carrot = "🥕";
  var corn = "🌽";
  var hot_pepper = "🌶️";
  var bell_pepper = "🫑";
  var cucumber = "🥒";
  var leafy_green = "🥬";
  var broccoli = "🥦";
  var garlic = "🧄";
  var onion = "🧅";
  var mushroom = "🍄";
  var peanuts = "🥜";
  var chestnut = "🌰";
  var bread = "🍞";
  var croissant = "🥐";
  var baguette_bread = "🥖";
  var flatbread = "🫓";
  var pretzel = "🥨";
  var bagel = "🥯";
  var pancakes = "🥞";
  var waffle = "🧇";
  var cheese = "🧀";
  var meat_on_bone = "🍖";
  var poultry_leg = "🍗";
  var cut_of_meat = "🥩";
  var bacon = "🥓";
  var hamburger = "🍔";
  var fries = "🍟";
  var pizza = "🍕";
  var hotdog = "🌭";
  var sandwich = "🥪";
  var taco = "🌮";
  var burrito = "🌯";
  var tamale = "🫔";
  var stuffed_flatbread = "🥙";
  var falafel = "🧆";
  var egg = "🥚";
  var fried_egg = "🍳";
  var shallow_pan_of_food = "🥘";
  var stew = "🍲";
  var fondue = "🫕";
  var bowl_with_spoon = "🥣";
  var green_salad = "🥗";
  var popcorn = "🍿";
  var butter = "🧈";
  var salt = "🧂";
  var canned_food = "🥫";
  var bento = "🍱";
  var rice_cracker = "🍘";
  var rice_ball = "🍙";
  var rice = "🍚";
  var curry = "🍛";
  var ramen = "🍜";
  var spaghetti = "🍝";
  var sweet_potato = "🍠";
  var oden = "🍢";
  var sushi = "🍣";
  var fried_shrimp = "🍤";
  var fish_cake = "🍥";
  var moon_cake = "🥮";
  var dango = "🍡";
  var dumpling = "🥟";
  var fortune_cookie = "🥠";
  var takeout_box = "🥡";
  var crab = "🦀";
  var lobster = "🦞";
  var shrimp = "🦐";
  var squid = "🦑";
  var oyster = "🦪";
  var icecream = "🍦";
  var shaved_ice = "🍧";
  var ice_cream = "🍨";
  var doughnut = "🍩";
  var cookie = "🍪";
  var birthday = "🎂";
  var cake = "🍰";
  var cupcake = "🧁";
  var pie = "🥧";
  var chocolate_bar = "🍫";
  var candy = "🍬";
  var lollipop = "🍭";
  var custard = "🍮";
  var honey_pot = "🍯";
  var baby_bottle = "🍼";
  var milk_glass = "🥛";
  var coffee = "☕";
  var teapot = "🫖";
  var tea = "🍵";
  var sake = "🍶";
  var champagne = "🍾";
  var wine_glass = "🍷";
  var cocktail = "🍸";
  var tropical_drink = "🍹";
  var beer = "🍺";
  var beers = "🍻";
  var clinking_glasses = "🥂";
  var tumbler_glass = "🥃";
  var cup_with_straw = "🥤";
  var bubble_tea = "🧋";
  var beverage_box = "🧃";
  var mate = "🧉";
  var ice_cube = "🧊";
  var chopsticks = "🥢";
  var plate_with_cutlery = "🍽️";
  var fork_and_knife = "🍴";
  var spoon = "🥄";
  var hocho = "🔪";
  var knife = "🔪";
  var amphora = "🏺";
  var earth_africa = "🌍";
  var earth_americas = "🌎";
  var earth_asia = "🌏";
  var globe_with_meridians = "🌐";
  var world_map = "🗺️";
  var japan = "🗾";
  var compass = "🧭";
  var mountain_snow = "🏔️";
  var mountain = "⛰️";
  var volcano = "🌋";
  var mount_fuji = "🗻";
  var camping = "🏕️";
  var beach_umbrella = "🏖️";
  var desert = "🏜️";
  var desert_island = "🏝️";
  var national_park = "🏞️";
  var stadium = "🏟️";
  var classical_building = "🏛️";
  var building_construction = "🏗️";
  var bricks = "🧱";
  var rock = "🪨";
  var wood = "🪵";
  var hut = "🛖";
  var houses = "🏘️";
  var derelict_house = "🏚️";
  var house = "🏠";
  var house_with_garden = "🏡";
  var office = "🏢";
  var post_office = "🏣";
  var european_post_office = "🏤";
  var hospital = "🏥";
  var bank = "🏦";
  var hotel = "🏨";
  var love_hotel = "🏩";
  var convenience_store = "🏪";
  var school = "🏫";
  var department_store = "🏬";
  var factory = "🏭";
  var japanese_castle = "🏯";
  var european_castle = "🏰";
  var wedding = "💒";
  var tokyo_tower = "🗼";
  var statue_of_liberty = "🗽";
  var church = "⛪";
  var mosque = "🕌";
  var hindu_temple = "🛕";
  var synagogue = "🕍";
  var shinto_shrine = "⛩️";
  var kaaba = "🕋";
  var fountain = "⛲";
  var tent = "⛺";
  var foggy = "🌁";
  var night_with_stars = "🌃";
  var cityscape = "🏙️";
  var sunrise_over_mountains = "🌄";
  var sunrise = "🌅";
  var city_sunset = "🌆";
  var city_sunrise = "🌇";
  var bridge_at_night = "🌉";
  var hotsprings = "♨️";
  var carousel_horse = "🎠";
  var ferris_wheel = "🎡";
  var roller_coaster = "🎢";
  var barber = "💈";
  var circus_tent = "🎪";
  var steam_locomotive = "🚂";
  var railway_car = "🚃";
  var bullettrain_side = "🚄";
  var bullettrain_front = "🚅";
  var train2 = "🚆";
  var metro = "🚇";
  var light_rail = "🚈";
  var station = "🚉";
  var tram = "🚊";
  var monorail = "🚝";
  var mountain_railway = "🚞";
  var train = "🚋";
  var bus = "🚌";
  var oncoming_bus = "🚍";
  var trolleybus = "🚎";
  var minibus = "🚐";
  var ambulance = "🚑";
  var fire_engine = "🚒";
  var police_car = "🚓";
  var oncoming_police_car = "🚔";
  var taxi = "🚕";
  var oncoming_taxi = "🚖";
  var car = "🚗";
  var red_car = "🚗";
  var oncoming_automobile = "🚘";
  var blue_car = "🚙";
  var pickup_truck = "🛻";
  var truck = "🚚";
  var articulated_lorry = "🚛";
  var tractor = "🚜";
  var racing_car = "🏎️";
  var motorcycle = "🏍️";
  var motor_scooter = "🛵";
  var manual_wheelchair = "🦽";
  var motorized_wheelchair = "🦼";
  var auto_rickshaw = "🛺";
  var bike = "🚲";
  var kick_scooter = "🛴";
  var skateboard = "🛹";
  var roller_skate = "🛼";
  var busstop = "🚏";
  var motorway = "🛣️";
  var railway_track = "🛤️";
  var oil_drum = "🛢️";
  var fuelpump = "⛽";
  var rotating_light = "🚨";
  var traffic_light = "🚥";
  var vertical_traffic_light = "🚦";
  var stop_sign = "🛑";
  var construction = "🚧";
  var anchor = "⚓";
  var boat = "⛵";
  var sailboat = "⛵";
  var canoe = "🛶";
  var speedboat = "🚤";
  var passenger_ship = "🛳️";
  var ferry = "⛴️";
  var motor_boat = "🛥️";
  var ship = "🚢";
  var airplane = "✈️";
  var small_airplane = "🛩️";
  var flight_departure = "🛫";
  var flight_arrival = "🛬";
  var parachute = "🪂";
  var seat = "💺";
  var helicopter = "🚁";
  var suspension_railway = "🚟";
  var mountain_cableway = "🚠";
  var aerial_tramway = "🚡";
  var artificial_satellite = "🛰️";
  var rocket = "🚀";
  var flying_saucer = "🛸";
  var bellhop_bell = "🛎️";
  var luggage = "🧳";
  var hourglass = "⌛";
  var hourglass_flowing_sand = "⏳";
  var watch = "⌚";
  var alarm_clock = "⏰";
  var stopwatch = "⏱️";
  var timer_clock = "⏲️";
  var mantelpiece_clock = "🕰️";
  var clock12 = "🕛";
  var clock1230 = "🕧";
  var clock1 = "🕐";
  var clock130 = "🕜";
  var clock2 = "🕑";
  var clock230 = "🕝";
  var clock3 = "🕒";
  var clock330 = "🕞";
  var clock4 = "🕓";
  var clock430 = "🕟";
  var clock5 = "🕔";
  var clock530 = "🕠";
  var clock6 = "🕕";
  var clock630 = "🕡";
  var clock7 = "🕖";
  var clock730 = "🕢";
  var clock8 = "🕗";
  var clock830 = "🕣";
  var clock9 = "🕘";
  var clock930 = "🕤";
  var clock10 = "🕙";
  var clock1030 = "🕥";
  var clock11 = "🕚";
  var clock1130 = "🕦";
  var new_moon = "🌑";
  var waxing_crescent_moon = "🌒";
  var first_quarter_moon = "🌓";
  var moon = "🌔";
  var waxing_gibbous_moon = "🌔";
  var full_moon = "🌕";
  var waning_gibbous_moon = "🌖";
  var last_quarter_moon = "🌗";
  var waning_crescent_moon = "🌘";
  var crescent_moon = "🌙";
  var new_moon_with_face = "🌚";
  var first_quarter_moon_with_face = "🌛";
  var last_quarter_moon_with_face = "🌜";
  var thermometer = "🌡️";
  var sunny = "☀️";
  var full_moon_with_face = "🌝";
  var sun_with_face = "🌞";
  var ringed_planet = "🪐";
  var star$1 = "⭐";
  var star2 = "🌟";
  var stars = "🌠";
  var milky_way = "🌌";
  var cloud = "☁️";
  var partly_sunny = "⛅";
  var cloud_with_lightning_and_rain = "⛈️";
  var sun_behind_small_cloud = "🌤️";
  var sun_behind_large_cloud = "🌥️";
  var sun_behind_rain_cloud = "🌦️";
  var cloud_with_rain = "🌧️";
  var cloud_with_snow = "🌨️";
  var cloud_with_lightning = "🌩️";
  var tornado = "🌪️";
  var fog = "🌫️";
  var wind_face = "🌬️";
  var cyclone = "🌀";
  var rainbow = "🌈";
  var closed_umbrella = "🌂";
  var open_umbrella = "☂️";
  var umbrella = "☔";
  var parasol_on_ground = "⛱️";
  var zap = "⚡";
  var snowflake = "❄️";
  var snowman_with_snow = "☃️";
  var snowman = "⛄";
  var comet = "☄️";
  var fire = "🔥";
  var droplet = "💧";
  var ocean = "🌊";
  var jack_o_lantern = "🎃";
  var christmas_tree = "🎄";
  var fireworks = "🎆";
  var sparkler = "🎇";
  var firecracker = "🧨";
  var sparkles = "✨";
  var balloon = "🎈";
  var tada = "🎉";
  var confetti_ball = "🎊";
  var tanabata_tree = "🎋";
  var bamboo = "🎍";
  var dolls = "🎎";
  var flags = "🎏";
  var wind_chime = "🎐";
  var rice_scene = "🎑";
  var red_envelope = "🧧";
  var ribbon = "🎀";
  var gift = "🎁";
  var reminder_ribbon = "🎗️";
  var tickets = "🎟️";
  var ticket = "🎫";
  var medal_military = "🎖️";
  var trophy = "🏆";
  var medal_sports = "🏅";
  var soccer = "⚽";
  var baseball = "⚾";
  var softball = "🥎";
  var basketball = "🏀";
  var volleyball = "🏐";
  var football = "🏈";
  var rugby_football = "🏉";
  var tennis = "🎾";
  var flying_disc = "🥏";
  var bowling = "🎳";
  var cricket_game = "🏏";
  var field_hockey = "🏑";
  var ice_hockey = "🏒";
  var lacrosse = "🥍";
  var ping_pong = "🏓";
  var badminton = "🏸";
  var boxing_glove = "🥊";
  var martial_arts_uniform = "🥋";
  var goal_net = "🥅";
  var golf = "⛳";
  var ice_skate = "⛸️";
  var fishing_pole_and_fish = "🎣";
  var diving_mask = "🤿";
  var running_shirt_with_sash = "🎽";
  var ski = "🎿";
  var sled = "🛷";
  var curling_stone = "🥌";
  var dart = "🎯";
  var yo_yo = "🪀";
  var kite = "🪁";
  var crystal_ball = "🔮";
  var magic_wand = "🪄";
  var nazar_amulet = "🧿";
  var video_game = "🎮";
  var joystick = "🕹️";
  var slot_machine = "🎰";
  var game_die = "🎲";
  var jigsaw = "🧩";
  var teddy_bear = "🧸";
  var pinata = "🪅";
  var nesting_dolls = "🪆";
  var spades$1 = "♠️";
  var hearts$1 = "♥️";
  var diamonds = "♦️";
  var clubs$1 = "♣️";
  var chess_pawn = "♟️";
  var black_joker = "🃏";
  var mahjong = "🀄";
  var flower_playing_cards = "🎴";
  var performing_arts = "🎭";
  var framed_picture = "🖼️";
  var art = "🎨";
  var thread = "🧵";
  var sewing_needle = "🪡";
  var yarn = "🧶";
  var knot = "🪢";
  var eyeglasses = "👓";
  var dark_sunglasses = "🕶️";
  var goggles = "🥽";
  var lab_coat = "🥼";
  var safety_vest = "🦺";
  var necktie = "👔";
  var shirt = "👕";
  var tshirt = "👕";
  var jeans = "👖";
  var scarf = "🧣";
  var gloves = "🧤";
  var coat = "🧥";
  var socks = "🧦";
  var dress = "👗";
  var kimono = "👘";
  var sari = "🥻";
  var one_piece_swimsuit = "🩱";
  var swim_brief = "🩲";
  var shorts = "🩳";
  var bikini = "👙";
  var womans_clothes = "👚";
  var purse = "👛";
  var handbag = "👜";
  var pouch = "👝";
  var shopping = "🛍️";
  var school_satchel = "🎒";
  var thong_sandal = "🩴";
  var mans_shoe = "👞";
  var shoe = "👞";
  var athletic_shoe = "👟";
  var hiking_boot = "🥾";
  var flat_shoe = "🥿";
  var high_heel = "👠";
  var sandal = "👡";
  var ballet_shoes = "🩰";
  var boot = "👢";
  var crown = "👑";
  var womans_hat = "👒";
  var tophat = "🎩";
  var mortar_board = "🎓";
  var billed_cap = "🧢";
  var military_helmet = "🪖";
  var rescue_worker_helmet = "⛑️";
  var prayer_beads = "📿";
  var lipstick = "💄";
  var ring$1 = "💍";
  var gem = "💎";
  var mute = "🔇";
  var speaker = "🔈";
  var sound = "🔉";
  var loud_sound = "🔊";
  var loudspeaker = "📢";
  var mega = "📣";
  var postal_horn = "📯";
  var bell = "🔔";
  var no_bell = "🔕";
  var musical_score = "🎼";
  var musical_note = "🎵";
  var notes = "🎶";
  var studio_microphone = "🎙️";
  var level_slider = "🎚️";
  var control_knobs = "🎛️";
  var microphone = "🎤";
  var headphones = "🎧";
  var radio = "📻";
  var saxophone = "🎷";
  var accordion = "🪗";
  var guitar = "🎸";
  var musical_keyboard = "🎹";
  var trumpet = "🎺";
  var violin = "🎻";
  var banjo = "🪕";
  var drum = "🥁";
  var long_drum = "🪘";
  var iphone = "📱";
  var calling = "📲";
  var phone$1 = "☎️";
  var telephone = "☎️";
  var telephone_receiver = "📞";
  var pager = "📟";
  var fax = "📠";
  var battery = "🔋";
  var electric_plug = "🔌";
  var computer = "💻";
  var desktop_computer = "🖥️";
  var printer = "🖨️";
  var keyboard = "⌨️";
  var computer_mouse = "🖱️";
  var trackball = "🖲️";
  var minidisc = "💽";
  var floppy_disk = "💾";
  var cd = "💿";
  var dvd = "📀";
  var abacus = "🧮";
  var movie_camera = "🎥";
  var film_strip = "🎞️";
  var film_projector = "📽️";
  var clapper = "🎬";
  var tv = "📺";
  var camera = "📷";
  var camera_flash = "📸";
  var video_camera = "📹";
  var vhs = "📼";
  var mag = "🔍";
  var mag_right = "🔎";
  var candle = "🕯️";
  var bulb = "💡";
  var flashlight = "🔦";
  var izakaya_lantern = "🏮";
  var lantern = "🏮";
  var diya_lamp = "🪔";
  var notebook_with_decorative_cover = "📔";
  var closed_book = "📕";
  var book = "📖";
  var open_book = "📖";
  var green_book = "📗";
  var blue_book = "📘";
  var orange_book = "📙";
  var books = "📚";
  var notebook = "📓";
  var ledger = "📒";
  var page_with_curl = "📃";
  var scroll = "📜";
  var page_facing_up = "📄";
  var newspaper = "📰";
  var newspaper_roll = "🗞️";
  var bookmark_tabs = "📑";
  var bookmark = "🔖";
  var label = "🏷️";
  var moneybag = "💰";
  var coin = "🪙";
  var yen$1 = "💴";
  var dollar$1 = "💵";
  var euro$1 = "💶";
  var pound$1 = "💷";
  var money_with_wings = "💸";
  var credit_card = "💳";
  var receipt = "🧾";
  var chart = "💹";
  var envelope = "✉️";
  var email = "📧";
  var incoming_envelope = "📨";
  var envelope_with_arrow = "📩";
  var outbox_tray = "📤";
  var inbox_tray = "📥";
  var mailbox = "📫";
  var mailbox_closed = "📪";
  var mailbox_with_mail = "📬";
  var mailbox_with_no_mail = "📭";
  var postbox = "📮";
  var ballot_box = "🗳️";
  var pencil2 = "✏️";
  var black_nib = "✒️";
  var fountain_pen = "🖋️";
  var pen = "🖊️";
  var paintbrush = "🖌️";
  var crayon = "🖍️";
  var memo = "📝";
  var pencil = "📝";
  var briefcase = "💼";
  var file_folder = "📁";
  var open_file_folder = "📂";
  var card_index_dividers = "🗂️";
  var date = "📅";
  var calendar = "📆";
  var spiral_notepad = "🗒️";
  var spiral_calendar = "🗓️";
  var card_index = "📇";
  var chart_with_upwards_trend = "📈";
  var chart_with_downwards_trend = "📉";
  var bar_chart = "📊";
  var clipboard$1 = "📋";
  var pushpin = "📌";
  var round_pushpin = "📍";
  var paperclip = "📎";
  var paperclips = "🖇️";
  var straight_ruler = "📏";
  var triangular_ruler = "📐";
  var scissors = "✂️";
  var card_file_box = "🗃️";
  var file_cabinet = "🗄️";
  var wastebasket = "🗑️";
  var lock = "🔒";
  var unlock = "🔓";
  var lock_with_ink_pen = "🔏";
  var closed_lock_with_key = "🔐";
  var key = "🔑";
  var old_key = "🗝️";
  var hammer = "🔨";
  var axe = "🪓";
  var pick = "⛏️";
  var hammer_and_pick = "⚒️";
  var hammer_and_wrench = "🛠️";
  var dagger$1 = "🗡️";
  var crossed_swords = "⚔️";
  var gun = "🔫";
  var boomerang = "🪃";
  var bow_and_arrow = "🏹";
  var shield = "🛡️";
  var carpentry_saw = "🪚";
  var wrench = "🔧";
  var screwdriver = "🪛";
  var nut_and_bolt = "🔩";
  var gear = "⚙️";
  var clamp = "🗜️";
  var balance_scale = "⚖️";
  var probing_cane = "🦯";
  var link$2 = "🔗";
  var chains = "⛓️";
  var hook = "🪝";
  var toolbox = "🧰";
  var magnet = "🧲";
  var ladder = "🪜";
  var alembic = "⚗️";
  var test_tube = "🧪";
  var petri_dish = "🧫";
  var dna = "🧬";
  var microscope = "🔬";
  var telescope = "🔭";
  var satellite = "📡";
  var syringe = "💉";
  var drop_of_blood = "🩸";
  var pill = "💊";
  var adhesive_bandage = "🩹";
  var stethoscope = "🩺";
  var door = "🚪";
  var elevator = "🛗";
  var mirror = "🪞";
  var window$1 = "🪟";
  var bed = "🛏️";
  var couch_and_lamp = "🛋️";
  var chair = "🪑";
  var toilet = "🚽";
  var plunger = "🪠";
  var shower = "🚿";
  var bathtub = "🛁";
  var mouse_trap = "🪤";
  var razor = "🪒";
  var lotion_bottle = "🧴";
  var safety_pin = "🧷";
  var broom = "🧹";
  var basket = "🧺";
  var roll_of_paper = "🧻";
  var bucket = "🪣";
  var soap = "🧼";
  var toothbrush = "🪥";
  var sponge = "🧽";
  var fire_extinguisher = "🧯";
  var shopping_cart = "🛒";
  var smoking = "🚬";
  var coffin = "⚰️";
  var headstone = "🪦";
  var funeral_urn = "⚱️";
  var moyai = "🗿";
  var placard = "🪧";
  var atm = "🏧";
  var put_litter_in_its_place = "🚮";
  var potable_water = "🚰";
  var wheelchair = "♿";
  var mens = "🚹";
  var womens = "🚺";
  var restroom = "🚻";
  var baby_symbol = "🚼";
  var wc = "🚾";
  var passport_control = "🛂";
  var customs = "🛃";
  var baggage_claim = "🛄";
  var left_luggage = "🛅";
  var warning = "⚠️";
  var children_crossing = "🚸";
  var no_entry = "⛔";
  var no_entry_sign = "🚫";
  var no_bicycles = "🚳";
  var no_smoking = "🚭";
  var do_not_litter = "🚯";
  var no_pedestrians = "🚷";
  var no_mobile_phones = "📵";
  var underage = "🔞";
  var radioactive = "☢️";
  var biohazard = "☣️";
  var arrow_up = "⬆️";
  var arrow_upper_right = "↗️";
  var arrow_right = "➡️";
  var arrow_lower_right = "↘️";
  var arrow_down = "⬇️";
  var arrow_lower_left = "↙️";
  var arrow_left = "⬅️";
  var arrow_upper_left = "↖️";
  var arrow_up_down = "↕️";
  var left_right_arrow = "↔️";
  var leftwards_arrow_with_hook = "↩️";
  var arrow_right_hook = "↪️";
  var arrow_heading_up = "⤴️";
  var arrow_heading_down = "⤵️";
  var arrows_clockwise = "🔃";
  var arrows_counterclockwise = "🔄";
  var back = "🔙";
  var end = "🔚";
  var on = "🔛";
  var soon = "🔜";
  var top$1 = "🔝";
  var place_of_worship = "🛐";
  var atom_symbol = "⚛️";
  var om = "🕉️";
  var star_of_david = "✡️";
  var wheel_of_dharma = "☸️";
  var yin_yang = "☯️";
  var latin_cross = "✝️";
  var orthodox_cross = "☦️";
  var star_and_crescent = "☪️";
  var peace_symbol = "☮️";
  var menorah = "🕎";
  var six_pointed_star = "🔯";
  var aries = "♈";
  var taurus = "♉";
  var gemini = "♊";
  var cancer = "♋";
  var leo = "♌";
  var virgo = "♍";
  var libra = "♎";
  var scorpius = "♏";
  var sagittarius = "♐";
  var capricorn = "♑";
  var aquarius = "♒";
  var pisces = "♓";
  var ophiuchus = "⛎";
  var twisted_rightwards_arrows = "🔀";
  var repeat = "🔁";
  var repeat_one = "🔂";
  var arrow_forward = "▶️";
  var fast_forward = "⏩";
  var next_track_button = "⏭️";
  var play_or_pause_button = "⏯️";
  var arrow_backward = "◀️";
  var rewind = "⏪";
  var previous_track_button = "⏮️";
  var arrow_up_small = "🔼";
  var arrow_double_up = "⏫";
  var arrow_down_small = "🔽";
  var arrow_double_down = "⏬";
  var pause_button = "⏸️";
  var stop_button = "⏹️";
  var record_button = "⏺️";
  var eject_button = "⏏️";
  var cinema = "🎦";
  var low_brightness = "🔅";
  var high_brightness = "🔆";
  var signal_strength = "📶";
  var vibration_mode = "📳";
  var mobile_phone_off = "📴";
  var female_sign = "♀️";
  var male_sign = "♂️";
  var transgender_symbol = "⚧️";
  var heavy_multiplication_x = "✖️";
  var heavy_plus_sign = "➕";
  var heavy_minus_sign = "➖";
  var heavy_division_sign = "➗";
  var infinity = "♾️";
  var bangbang = "‼️";
  var interrobang = "⁉️";
  var question = "❓";
  var grey_question = "❔";
  var grey_exclamation = "❕";
  var exclamation = "❗";
  var heavy_exclamation_mark = "❗";
  var wavy_dash = "〰️";
  var currency_exchange = "💱";
  var heavy_dollar_sign = "💲";
  var medical_symbol = "⚕️";
  var recycle = "♻️";
  var fleur_de_lis = "⚜️";
  var trident = "🔱";
  var name_badge = "📛";
  var beginner = "🔰";
  var o = "⭕";
  var white_check_mark = "✅";
  var ballot_box_with_check = "☑️";
  var heavy_check_mark = "✔️";
  var x = "❌";
  var negative_squared_cross_mark = "❎";
  var curly_loop = "➰";
  var loop = "➿";
  var part_alternation_mark = "〽️";
  var eight_spoked_asterisk = "✳️";
  var eight_pointed_black_star = "✴️";
  var sparkle = "❇️";
  var copyright = "©️";
  var registered = "®️";
  var tm = "™️";
  var hash = "#️⃣";
  var asterisk = "*️⃣";
  var zero$1 = "0️⃣";
  var one = "1️⃣";
  var two = "2️⃣";
  var three = "3️⃣";
  var four = "4️⃣";
  var five = "5️⃣";
  var six = "6️⃣";
  var seven = "7️⃣";
  var eight = "8️⃣";
  var nine = "9️⃣";
  var keycap_ten = "🔟";
  var capital_abcd = "🔠";
  var abcd = "🔡";
  var symbols = "🔣";
  var abc = "🔤";
  var a = "🅰️";
  var ab = "🆎";
  var b = "🅱️";
  var cl = "🆑";
  var cool = "🆒";
  var free = "🆓";
  var information_source = "ℹ️";
  var id = "🆔";
  var m = "Ⓜ️";
  var ng = "🆖";
  var o2 = "🅾️";
  var ok = "🆗";
  var parking = "🅿️";
  var sos = "🆘";
  var up = "🆙";
  var vs = "🆚";
  var koko = "🈁";
  var sa = "🈂️";
  var ideograph_advantage = "🉐";
  var accept = "🉑";
  var congratulations = "㊗️";
  var secret = "㊙️";
  var u6e80 = "🈵";
  var red_circle = "🔴";
  var orange_circle = "🟠";
  var yellow_circle = "🟡";
  var green_circle = "🟢";
  var large_blue_circle = "🔵";
  var purple_circle = "🟣";
  var brown_circle = "🟤";
  var black_circle = "⚫";
  var white_circle = "⚪";
  var red_square = "🟥";
  var orange_square = "🟧";
  var yellow_square = "🟨";
  var green_square = "🟩";
  var blue_square = "🟦";
  var purple_square = "🟪";
  var brown_square = "🟫";
  var black_large_square = "⬛";
  var white_large_square = "⬜";
  var black_medium_square = "◼️";
  var white_medium_square = "◻️";
  var black_medium_small_square = "◾";
  var white_medium_small_square = "◽";
  var black_small_square = "▪️";
  var white_small_square = "▫️";
  var large_orange_diamond = "🔶";
  var large_blue_diamond = "🔷";
  var small_orange_diamond = "🔸";
  var small_blue_diamond = "🔹";
  var small_red_triangle = "🔺";
  var small_red_triangle_down = "🔻";
  var diamond_shape_with_a_dot_inside = "💠";
  var radio_button = "🔘";
  var white_square_button = "🔳";
  var black_square_button = "🔲";
  var checkered_flag = "🏁";
  var triangular_flag_on_post = "🚩";
  var crossed_flags = "🎌";
  var black_flag = "🏴";
  var white_flag = "🏳️";
  var rainbow_flag = "🏳️‍🌈";
  var transgender_flag = "🏳️‍⚧️";
  var pirate_flag = "🏴‍☠️";
  var ascension_island = "🇦🇨";
  var andorra = "🇦🇩";
  var united_arab_emirates = "🇦🇪";
  var afghanistan = "🇦🇫";
  var antigua_barbuda = "🇦🇬";
  var anguilla = "🇦🇮";
  var albania = "🇦🇱";
  var armenia = "🇦🇲";
  var angola = "🇦🇴";
  var antarctica = "🇦🇶";
  var argentina = "🇦🇷";
  var american_samoa = "🇦🇸";
  var austria = "🇦🇹";
  var australia = "🇦🇺";
  var aruba = "🇦🇼";
  var aland_islands = "🇦🇽";
  var azerbaijan = "🇦🇿";
  var bosnia_herzegovina = "🇧🇦";
  var barbados = "🇧🇧";
  var bangladesh = "🇧🇩";
  var belgium = "🇧🇪";
  var burkina_faso = "🇧🇫";
  var bulgaria = "🇧🇬";
  var bahrain = "🇧🇭";
  var burundi = "🇧🇮";
  var benin = "🇧🇯";
  var st_barthelemy = "🇧🇱";
  var bermuda = "🇧🇲";
  var brunei = "🇧🇳";
  var bolivia = "🇧🇴";
  var caribbean_netherlands = "🇧🇶";
  var brazil = "🇧🇷";
  var bahamas = "🇧🇸";
  var bhutan = "🇧🇹";
  var bouvet_island = "🇧🇻";
  var botswana = "🇧🇼";
  var belarus = "🇧🇾";
  var belize = "🇧🇿";
  var canada = "🇨🇦";
  var cocos_islands = "🇨🇨";
  var congo_kinshasa = "🇨🇩";
  var central_african_republic = "🇨🇫";
  var congo_brazzaville = "🇨🇬";
  var switzerland = "🇨🇭";
  var cote_divoire = "🇨🇮";
  var cook_islands = "🇨🇰";
  var chile = "🇨🇱";
  var cameroon = "🇨🇲";
  var cn = "🇨🇳";
  var colombia = "🇨🇴";
  var clipperton_island = "🇨🇵";
  var costa_rica = "🇨🇷";
  var cuba = "🇨🇺";
  var cape_verde = "🇨🇻";
  var curacao = "🇨🇼";
  var christmas_island = "🇨🇽";
  var cyprus = "🇨🇾";
  var czech_republic = "🇨🇿";
  var de = "🇩🇪";
  var diego_garcia = "🇩🇬";
  var djibouti = "🇩🇯";
  var denmark = "🇩🇰";
  var dominica = "🇩🇲";
  var dominican_republic = "🇩🇴";
  var algeria = "🇩🇿";
  var ceuta_melilla = "🇪🇦";
  var ecuador = "🇪🇨";
  var estonia = "🇪🇪";
  var egypt = "🇪🇬";
  var western_sahara = "🇪🇭";
  var eritrea = "🇪🇷";
  var es = "🇪🇸";
  var ethiopia = "🇪🇹";
  var eu = "🇪🇺";
  var european_union = "🇪🇺";
  var finland = "🇫🇮";
  var fiji = "🇫🇯";
  var falkland_islands = "🇫🇰";
  var micronesia = "🇫🇲";
  var faroe_islands = "🇫🇴";
  var fr = "🇫🇷";
  var gabon = "🇬🇦";
  var gb = "🇬🇧";
  var uk = "🇬🇧";
  var grenada = "🇬🇩";
  var georgia = "🇬🇪";
  var french_guiana = "🇬🇫";
  var guernsey = "🇬🇬";
  var ghana = "🇬🇭";
  var gibraltar = "🇬🇮";
  var greenland = "🇬🇱";
  var gambia = "🇬🇲";
  var guinea = "🇬🇳";
  var guadeloupe = "🇬🇵";
  var equatorial_guinea = "🇬🇶";
  var greece = "🇬🇷";
  var south_georgia_south_sandwich_islands = "🇬🇸";
  var guatemala = "🇬🇹";
  var guam = "🇬🇺";
  var guinea_bissau = "🇬🇼";
  var guyana = "🇬🇾";
  var hong_kong = "🇭🇰";
  var heard_mcdonald_islands = "🇭🇲";
  var honduras = "🇭🇳";
  var croatia = "🇭🇷";
  var haiti = "🇭🇹";
  var hungary = "🇭🇺";
  var canary_islands = "🇮🇨";
  var indonesia = "🇮🇩";
  var ireland = "🇮🇪";
  var israel = "🇮🇱";
  var isle_of_man = "🇮🇲";
  var india = "🇮🇳";
  var british_indian_ocean_territory = "🇮🇴";
  var iraq = "🇮🇶";
  var iran = "🇮🇷";
  var iceland = "🇮🇸";
  var it$1 = "🇮🇹";
  var jersey = "🇯🇪";
  var jamaica = "🇯🇲";
  var jordan = "🇯🇴";
  var jp = "🇯🇵";
  var kenya = "🇰🇪";
  var kyrgyzstan = "🇰🇬";
  var cambodia = "🇰🇭";
  var kiribati = "🇰🇮";
  var comoros = "🇰🇲";
  var st_kitts_nevis = "🇰🇳";
  var north_korea = "🇰🇵";
  var kr = "🇰🇷";
  var kuwait = "🇰🇼";
  var cayman_islands = "🇰🇾";
  var kazakhstan = "🇰🇿";
  var laos = "🇱🇦";
  var lebanon = "🇱🇧";
  var st_lucia = "🇱🇨";
  var liechtenstein = "🇱🇮";
  var sri_lanka = "🇱🇰";
  var liberia = "🇱🇷";
  var lesotho = "🇱🇸";
  var lithuania = "🇱🇹";
  var luxembourg = "🇱🇺";
  var latvia = "🇱🇻";
  var libya = "🇱🇾";
  var morocco = "🇲🇦";
  var monaco = "🇲🇨";
  var moldova = "🇲🇩";
  var montenegro = "🇲🇪";
  var st_martin = "🇲🇫";
  var madagascar = "🇲🇬";
  var marshall_islands = "🇲🇭";
  var macedonia = "🇲🇰";
  var mali = "🇲🇱";
  var myanmar = "🇲🇲";
  var mongolia = "🇲🇳";
  var macau = "🇲🇴";
  var northern_mariana_islands = "🇲🇵";
  var martinique = "🇲🇶";
  var mauritania = "🇲🇷";
  var montserrat = "🇲🇸";
  var malta = "🇲🇹";
  var mauritius = "🇲🇺";
  var maldives = "🇲🇻";
  var malawi = "🇲🇼";
  var mexico = "🇲🇽";
  var malaysia = "🇲🇾";
  var mozambique = "🇲🇿";
  var namibia = "🇳🇦";
  var new_caledonia = "🇳🇨";
  var niger = "🇳🇪";
  var norfolk_island = "🇳🇫";
  var nigeria = "🇳🇬";
  var nicaragua = "🇳🇮";
  var netherlands = "🇳🇱";
  var norway = "🇳🇴";
  var nepal = "🇳🇵";
  var nauru = "🇳🇷";
  var niue = "🇳🇺";
  var new_zealand = "🇳🇿";
  var oman = "🇴🇲";
  var panama = "🇵🇦";
  var peru = "🇵🇪";
  var french_polynesia = "🇵🇫";
  var papua_new_guinea = "🇵🇬";
  var philippines = "🇵🇭";
  var pakistan = "🇵🇰";
  var poland = "🇵🇱";
  var st_pierre_miquelon = "🇵🇲";
  var pitcairn_islands = "🇵🇳";
  var puerto_rico = "🇵🇷";
  var palestinian_territories = "🇵🇸";
  var portugal = "🇵🇹";
  var palau = "🇵🇼";
  var paraguay = "🇵🇾";
  var qatar = "🇶🇦";
  var reunion = "🇷🇪";
  var romania = "🇷🇴";
  var serbia = "🇷🇸";
  var ru = "🇷🇺";
  var rwanda = "🇷🇼";
  var saudi_arabia = "🇸🇦";
  var solomon_islands = "🇸🇧";
  var seychelles = "🇸🇨";
  var sudan = "🇸🇩";
  var sweden = "🇸🇪";
  var singapore = "🇸🇬";
  var st_helena = "🇸🇭";
  var slovenia = "🇸🇮";
  var svalbard_jan_mayen = "🇸🇯";
  var slovakia = "🇸🇰";
  var sierra_leone = "🇸🇱";
  var san_marino = "🇸🇲";
  var senegal = "🇸🇳";
  var somalia = "🇸🇴";
  var suriname = "🇸🇷";
  var south_sudan = "🇸🇸";
  var sao_tome_principe = "🇸🇹";
  var el_salvador = "🇸🇻";
  var sint_maarten = "🇸🇽";
  var syria = "🇸🇾";
  var swaziland = "🇸🇿";
  var tristan_da_cunha = "🇹🇦";
  var turks_caicos_islands = "🇹🇨";
  var chad = "🇹🇩";
  var french_southern_territories = "🇹🇫";
  var togo = "🇹🇬";
  var thailand = "🇹🇭";
  var tajikistan = "🇹🇯";
  var tokelau = "🇹🇰";
  var timor_leste = "🇹🇱";
  var turkmenistan = "🇹🇲";
  var tunisia = "🇹🇳";
  var tonga = "🇹🇴";
  var tr = "🇹🇷";
  var trinidad_tobago = "🇹🇹";
  var tuvalu = "🇹🇻";
  var taiwan = "🇹🇼";
  var tanzania = "🇹🇿";
  var ukraine = "🇺🇦";
  var uganda = "🇺🇬";
  var us_outlying_islands = "🇺🇲";
  var united_nations = "🇺🇳";
  var us = "🇺🇸";
  var uruguay = "🇺🇾";
  var uzbekistan = "🇺🇿";
  var vatican_city = "🇻🇦";
  var st_vincent_grenadines = "🇻🇨";
  var venezuela = "🇻🇪";
  var british_virgin_islands = "🇻🇬";
  var us_virgin_islands = "🇻🇮";
  var vietnam = "🇻🇳";
  var vanuatu = "🇻🇺";
  var wallis_futuna = "🇼🇫";
  var samoa = "🇼🇸";
  var kosovo = "🇽🇰";
  var yemen = "🇾🇪";
  var mayotte = "🇾🇹";
  var south_africa = "🇿🇦";
  var zambia = "🇿🇲";
  var zimbabwe = "🇿🇼";
  var england = "🏴󠁧󠁢󠁥󠁮󠁧󠁿";
  var scotland = "🏴󠁧󠁢󠁳󠁣󠁴󠁿";
  var wales = "🏴󠁧󠁢󠁷󠁬󠁳󠁿";
  var full = {
  	"100": "💯",
  	"1234": "🔢",
  	grinning: grinning,
  	smiley: smiley,
  	smile: smile$1,
  	grin: grin,
  	laughing: laughing,
  	satisfied: satisfied,
  	sweat_smile: sweat_smile,
  	rofl: rofl,
  	joy: joy,
  	slightly_smiling_face: slightly_smiling_face,
  	upside_down_face: upside_down_face,
  	wink: wink,
  	blush: blush,
  	innocent: innocent,
  	smiling_face_with_three_hearts: smiling_face_with_three_hearts,
  	heart_eyes: heart_eyes,
  	star_struck: star_struck,
  	kissing_heart: kissing_heart,
  	kissing: kissing,
  	relaxed: relaxed,
  	kissing_closed_eyes: kissing_closed_eyes,
  	kissing_smiling_eyes: kissing_smiling_eyes,
  	smiling_face_with_tear: smiling_face_with_tear,
  	yum: yum,
  	stuck_out_tongue: stuck_out_tongue,
  	stuck_out_tongue_winking_eye: stuck_out_tongue_winking_eye,
  	zany_face: zany_face,
  	stuck_out_tongue_closed_eyes: stuck_out_tongue_closed_eyes,
  	money_mouth_face: money_mouth_face,
  	hugs: hugs,
  	hand_over_mouth: hand_over_mouth,
  	shushing_face: shushing_face,
  	thinking: thinking,
  	zipper_mouth_face: zipper_mouth_face,
  	raised_eyebrow: raised_eyebrow,
  	neutral_face: neutral_face,
  	expressionless: expressionless,
  	no_mouth: no_mouth,
  	smirk: smirk,
  	unamused: unamused,
  	roll_eyes: roll_eyes,
  	grimacing: grimacing,
  	lying_face: lying_face,
  	relieved: relieved,
  	pensive: pensive,
  	sleepy: sleepy,
  	drooling_face: drooling_face,
  	sleeping: sleeping,
  	mask: mask,
  	face_with_thermometer: face_with_thermometer,
  	face_with_head_bandage: face_with_head_bandage,
  	nauseated_face: nauseated_face,
  	vomiting_face: vomiting_face,
  	sneezing_face: sneezing_face,
  	hot_face: hot_face,
  	cold_face: cold_face,
  	woozy_face: woozy_face,
  	dizzy_face: dizzy_face,
  	exploding_head: exploding_head,
  	cowboy_hat_face: cowboy_hat_face,
  	partying_face: partying_face,
  	disguised_face: disguised_face,
  	sunglasses: sunglasses,
  	nerd_face: nerd_face,
  	monocle_face: monocle_face,
  	confused: confused,
  	worried: worried,
  	slightly_frowning_face: slightly_frowning_face,
  	frowning_face: frowning_face,
  	open_mouth: open_mouth,
  	hushed: hushed,
  	astonished: astonished,
  	flushed: flushed,
  	pleading_face: pleading_face,
  	frowning: frowning,
  	anguished: anguished,
  	fearful: fearful,
  	cold_sweat: cold_sweat,
  	disappointed_relieved: disappointed_relieved,
  	cry: cry,
  	sob: sob,
  	scream: scream,
  	confounded: confounded,
  	persevere: persevere,
  	disappointed: disappointed,
  	sweat: sweat,
  	weary: weary,
  	tired_face: tired_face,
  	yawning_face: yawning_face,
  	triumph: triumph,
  	rage: rage,
  	pout: pout,
  	angry: angry,
  	cursing_face: cursing_face,
  	smiling_imp: smiling_imp,
  	imp: imp,
  	skull: skull,
  	skull_and_crossbones: skull_and_crossbones,
  	hankey: hankey,
  	poop: poop,
  	shit: shit,
  	clown_face: clown_face,
  	japanese_ogre: japanese_ogre,
  	japanese_goblin: japanese_goblin,
  	ghost: ghost,
  	alien: alien,
  	space_invader: space_invader,
  	robot: robot,
  	smiley_cat: smiley_cat,
  	smile_cat: smile_cat,
  	joy_cat: joy_cat,
  	heart_eyes_cat: heart_eyes_cat,
  	smirk_cat: smirk_cat,
  	kissing_cat: kissing_cat,
  	scream_cat: scream_cat,
  	crying_cat_face: crying_cat_face,
  	pouting_cat: pouting_cat,
  	see_no_evil: see_no_evil,
  	hear_no_evil: hear_no_evil,
  	speak_no_evil: speak_no_evil,
  	kiss: kiss,
  	love_letter: love_letter,
  	cupid: cupid,
  	gift_heart: gift_heart,
  	sparkling_heart: sparkling_heart,
  	heartpulse: heartpulse,
  	heartbeat: heartbeat,
  	revolving_hearts: revolving_hearts,
  	two_hearts: two_hearts,
  	heart_decoration: heart_decoration,
  	heavy_heart_exclamation: heavy_heart_exclamation,
  	broken_heart: broken_heart,
  	heart: heart,
  	orange_heart: orange_heart,
  	yellow_heart: yellow_heart,
  	green_heart: green_heart,
  	blue_heart: blue_heart,
  	purple_heart: purple_heart,
  	brown_heart: brown_heart,
  	black_heart: black_heart,
  	white_heart: white_heart,
  	anger: anger,
  	boom: boom,
  	collision: collision,
  	dizzy: dizzy,
  	sweat_drops: sweat_drops,
  	dash: dash$1,
  	hole: hole,
  	bomb: bomb,
  	speech_balloon: speech_balloon,
  	eye_speech_bubble: eye_speech_bubble,
  	left_speech_bubble: left_speech_bubble,
  	right_anger_bubble: right_anger_bubble,
  	thought_balloon: thought_balloon,
  	zzz: zzz,
  	wave: wave,
  	raised_back_of_hand: raised_back_of_hand,
  	raised_hand_with_fingers_splayed: raised_hand_with_fingers_splayed,
  	hand: hand,
  	raised_hand: raised_hand,
  	vulcan_salute: vulcan_salute,
  	ok_hand: ok_hand,
  	pinched_fingers: pinched_fingers,
  	pinching_hand: pinching_hand,
  	v: v,
  	crossed_fingers: crossed_fingers,
  	love_you_gesture: love_you_gesture,
  	metal: metal,
  	call_me_hand: call_me_hand,
  	point_left: point_left,
  	point_right: point_right,
  	point_up_2: point_up_2,
  	middle_finger: middle_finger,
  	fu: fu,
  	point_down: point_down,
  	point_up: point_up,
  	"+1": "👍",
  	thumbsup: thumbsup,
  	"-1": "👎",
  	thumbsdown: thumbsdown,
  	fist_raised: fist_raised,
  	fist: fist,
  	fist_oncoming: fist_oncoming,
  	facepunch: facepunch,
  	punch: punch,
  	fist_left: fist_left,
  	fist_right: fist_right,
  	clap: clap,
  	raised_hands: raised_hands,
  	open_hands: open_hands,
  	palms_up_together: palms_up_together,
  	handshake: handshake,
  	pray: pray,
  	writing_hand: writing_hand,
  	nail_care: nail_care,
  	selfie: selfie,
  	muscle: muscle,
  	mechanical_arm: mechanical_arm,
  	mechanical_leg: mechanical_leg,
  	leg: leg$1,
  	foot: foot,
  	ear: ear,
  	ear_with_hearing_aid: ear_with_hearing_aid,
  	nose: nose,
  	brain: brain,
  	anatomical_heart: anatomical_heart,
  	lungs: lungs,
  	tooth: tooth,
  	bone: bone,
  	eyes: eyes,
  	eye: eye,
  	tongue: tongue,
  	lips: lips,
  	baby: baby,
  	child: child,
  	boy: boy,
  	girl: girl,
  	adult: adult,
  	blond_haired_person: blond_haired_person,
  	man: man,
  	bearded_person: bearded_person,
  	red_haired_man: red_haired_man,
  	curly_haired_man: curly_haired_man,
  	white_haired_man: white_haired_man,
  	bald_man: bald_man,
  	woman: woman,
  	red_haired_woman: red_haired_woman,
  	person_red_hair: person_red_hair,
  	curly_haired_woman: curly_haired_woman,
  	person_curly_hair: person_curly_hair,
  	white_haired_woman: white_haired_woman,
  	person_white_hair: person_white_hair,
  	bald_woman: bald_woman,
  	person_bald: person_bald,
  	blond_haired_woman: blond_haired_woman,
  	blonde_woman: blonde_woman,
  	blond_haired_man: blond_haired_man,
  	older_adult: older_adult,
  	older_man: older_man,
  	older_woman: older_woman,
  	frowning_person: frowning_person,
  	frowning_man: frowning_man,
  	frowning_woman: frowning_woman,
  	pouting_face: pouting_face,
  	pouting_man: pouting_man,
  	pouting_woman: pouting_woman,
  	no_good: no_good,
  	no_good_man: no_good_man,
  	ng_man: ng_man,
  	no_good_woman: no_good_woman,
  	ng_woman: ng_woman,
  	ok_person: ok_person,
  	ok_man: ok_man,
  	ok_woman: ok_woman,
  	tipping_hand_person: tipping_hand_person,
  	information_desk_person: information_desk_person,
  	tipping_hand_man: tipping_hand_man,
  	sassy_man: sassy_man,
  	tipping_hand_woman: tipping_hand_woman,
  	sassy_woman: sassy_woman,
  	raising_hand: raising_hand,
  	raising_hand_man: raising_hand_man,
  	raising_hand_woman: raising_hand_woman,
  	deaf_person: deaf_person,
  	deaf_man: deaf_man,
  	deaf_woman: deaf_woman,
  	bow: bow,
  	bowing_man: bowing_man,
  	bowing_woman: bowing_woman,
  	facepalm: facepalm,
  	man_facepalming: man_facepalming,
  	woman_facepalming: woman_facepalming,
  	shrug: shrug,
  	man_shrugging: man_shrugging,
  	woman_shrugging: woman_shrugging,
  	health_worker: health_worker,
  	man_health_worker: man_health_worker,
  	woman_health_worker: woman_health_worker,
  	student: student,
  	man_student: man_student,
  	woman_student: woman_student,
  	teacher: teacher,
  	man_teacher: man_teacher,
  	woman_teacher: woman_teacher,
  	judge: judge,
  	man_judge: man_judge,
  	woman_judge: woman_judge,
  	farmer: farmer,
  	man_farmer: man_farmer,
  	woman_farmer: woman_farmer,
  	cook: cook,
  	man_cook: man_cook,
  	woman_cook: woman_cook,
  	mechanic: mechanic,
  	man_mechanic: man_mechanic,
  	woman_mechanic: woman_mechanic,
  	factory_worker: factory_worker,
  	man_factory_worker: man_factory_worker,
  	woman_factory_worker: woman_factory_worker,
  	office_worker: office_worker,
  	man_office_worker: man_office_worker,
  	woman_office_worker: woman_office_worker,
  	scientist: scientist,
  	man_scientist: man_scientist,
  	woman_scientist: woman_scientist,
  	technologist: technologist,
  	man_technologist: man_technologist,
  	woman_technologist: woman_technologist,
  	singer: singer,
  	man_singer: man_singer,
  	woman_singer: woman_singer,
  	artist: artist,
  	man_artist: man_artist,
  	woman_artist: woman_artist,
  	pilot: pilot,
  	man_pilot: man_pilot,
  	woman_pilot: woman_pilot,
  	astronaut: astronaut,
  	man_astronaut: man_astronaut,
  	woman_astronaut: woman_astronaut,
  	firefighter: firefighter,
  	man_firefighter: man_firefighter,
  	woman_firefighter: woman_firefighter,
  	police_officer: police_officer,
  	cop: cop,
  	policeman: policeman,
  	policewoman: policewoman,
  	detective: detective,
  	male_detective: male_detective,
  	female_detective: female_detective,
  	guard: guard,
  	guardsman: guardsman,
  	guardswoman: guardswoman,
  	ninja: ninja,
  	construction_worker: construction_worker,
  	construction_worker_man: construction_worker_man,
  	construction_worker_woman: construction_worker_woman,
  	prince: prince,
  	princess: princess,
  	person_with_turban: person_with_turban,
  	man_with_turban: man_with_turban,
  	woman_with_turban: woman_with_turban,
  	man_with_gua_pi_mao: man_with_gua_pi_mao,
  	woman_with_headscarf: woman_with_headscarf,
  	person_in_tuxedo: person_in_tuxedo,
  	man_in_tuxedo: man_in_tuxedo,
  	woman_in_tuxedo: woman_in_tuxedo,
  	person_with_veil: person_with_veil,
  	man_with_veil: man_with_veil,
  	woman_with_veil: woman_with_veil,
  	bride_with_veil: bride_with_veil,
  	pregnant_woman: pregnant_woman,
  	breast_feeding: breast_feeding,
  	woman_feeding_baby: woman_feeding_baby,
  	man_feeding_baby: man_feeding_baby,
  	person_feeding_baby: person_feeding_baby,
  	angel: angel,
  	santa: santa,
  	mrs_claus: mrs_claus,
  	mx_claus: mx_claus,
  	superhero: superhero,
  	superhero_man: superhero_man,
  	superhero_woman: superhero_woman,
  	supervillain: supervillain,
  	supervillain_man: supervillain_man,
  	supervillain_woman: supervillain_woman,
  	mage: mage,
  	mage_man: mage_man,
  	mage_woman: mage_woman,
  	fairy: fairy,
  	fairy_man: fairy_man,
  	fairy_woman: fairy_woman,
  	vampire: vampire,
  	vampire_man: vampire_man,
  	vampire_woman: vampire_woman,
  	merperson: merperson,
  	merman: merman,
  	mermaid: mermaid,
  	elf: elf,
  	elf_man: elf_man,
  	elf_woman: elf_woman,
  	genie: genie,
  	genie_man: genie_man,
  	genie_woman: genie_woman,
  	zombie: zombie,
  	zombie_man: zombie_man,
  	zombie_woman: zombie_woman,
  	massage: massage,
  	massage_man: massage_man,
  	massage_woman: massage_woman,
  	haircut: haircut,
  	haircut_man: haircut_man,
  	haircut_woman: haircut_woman,
  	walking: walking,
  	walking_man: walking_man,
  	walking_woman: walking_woman,
  	standing_person: standing_person,
  	standing_man: standing_man,
  	standing_woman: standing_woman,
  	kneeling_person: kneeling_person,
  	kneeling_man: kneeling_man,
  	kneeling_woman: kneeling_woman,
  	person_with_probing_cane: person_with_probing_cane,
  	man_with_probing_cane: man_with_probing_cane,
  	woman_with_probing_cane: woman_with_probing_cane,
  	person_in_motorized_wheelchair: person_in_motorized_wheelchair,
  	man_in_motorized_wheelchair: man_in_motorized_wheelchair,
  	woman_in_motorized_wheelchair: woman_in_motorized_wheelchair,
  	person_in_manual_wheelchair: person_in_manual_wheelchair,
  	man_in_manual_wheelchair: man_in_manual_wheelchair,
  	woman_in_manual_wheelchair: woman_in_manual_wheelchair,
  	runner: runner,
  	running: running,
  	running_man: running_man,
  	running_woman: running_woman,
  	woman_dancing: woman_dancing,
  	dancer: dancer,
  	man_dancing: man_dancing,
  	business_suit_levitating: business_suit_levitating,
  	dancers: dancers,
  	dancing_men: dancing_men,
  	dancing_women: dancing_women,
  	sauna_person: sauna_person,
  	sauna_man: sauna_man,
  	sauna_woman: sauna_woman,
  	climbing: climbing,
  	climbing_man: climbing_man,
  	climbing_woman: climbing_woman,
  	person_fencing: person_fencing,
  	horse_racing: horse_racing,
  	skier: skier,
  	snowboarder: snowboarder,
  	golfing: golfing,
  	golfing_man: golfing_man,
  	golfing_woman: golfing_woman,
  	surfer: surfer,
  	surfing_man: surfing_man,
  	surfing_woman: surfing_woman,
  	rowboat: rowboat,
  	rowing_man: rowing_man,
  	rowing_woman: rowing_woman,
  	swimmer: swimmer,
  	swimming_man: swimming_man,
  	swimming_woman: swimming_woman,
  	bouncing_ball_person: bouncing_ball_person,
  	bouncing_ball_man: bouncing_ball_man,
  	basketball_man: basketball_man,
  	bouncing_ball_woman: bouncing_ball_woman,
  	basketball_woman: basketball_woman,
  	weight_lifting: weight_lifting,
  	weight_lifting_man: weight_lifting_man,
  	weight_lifting_woman: weight_lifting_woman,
  	bicyclist: bicyclist,
  	biking_man: biking_man,
  	biking_woman: biking_woman,
  	mountain_bicyclist: mountain_bicyclist,
  	mountain_biking_man: mountain_biking_man,
  	mountain_biking_woman: mountain_biking_woman,
  	cartwheeling: cartwheeling,
  	man_cartwheeling: man_cartwheeling,
  	woman_cartwheeling: woman_cartwheeling,
  	wrestling: wrestling,
  	men_wrestling: men_wrestling,
  	women_wrestling: women_wrestling,
  	water_polo: water_polo,
  	man_playing_water_polo: man_playing_water_polo,
  	woman_playing_water_polo: woman_playing_water_polo,
  	handball_person: handball_person,
  	man_playing_handball: man_playing_handball,
  	woman_playing_handball: woman_playing_handball,
  	juggling_person: juggling_person,
  	man_juggling: man_juggling,
  	woman_juggling: woman_juggling,
  	lotus_position: lotus_position,
  	lotus_position_man: lotus_position_man,
  	lotus_position_woman: lotus_position_woman,
  	bath: bath,
  	sleeping_bed: sleeping_bed,
  	people_holding_hands: people_holding_hands,
  	two_women_holding_hands: two_women_holding_hands,
  	couple: couple,
  	two_men_holding_hands: two_men_holding_hands,
  	couplekiss: couplekiss,
  	couplekiss_man_woman: couplekiss_man_woman,
  	couplekiss_man_man: couplekiss_man_man,
  	couplekiss_woman_woman: couplekiss_woman_woman,
  	couple_with_heart: couple_with_heart,
  	couple_with_heart_woman_man: couple_with_heart_woman_man,
  	couple_with_heart_man_man: couple_with_heart_man_man,
  	couple_with_heart_woman_woman: couple_with_heart_woman_woman,
  	family: family,
  	family_man_woman_boy: family_man_woman_boy,
  	family_man_woman_girl: family_man_woman_girl,
  	family_man_woman_girl_boy: family_man_woman_girl_boy,
  	family_man_woman_boy_boy: family_man_woman_boy_boy,
  	family_man_woman_girl_girl: family_man_woman_girl_girl,
  	family_man_man_boy: family_man_man_boy,
  	family_man_man_girl: family_man_man_girl,
  	family_man_man_girl_boy: family_man_man_girl_boy,
  	family_man_man_boy_boy: family_man_man_boy_boy,
  	family_man_man_girl_girl: family_man_man_girl_girl,
  	family_woman_woman_boy: family_woman_woman_boy,
  	family_woman_woman_girl: family_woman_woman_girl,
  	family_woman_woman_girl_boy: family_woman_woman_girl_boy,
  	family_woman_woman_boy_boy: family_woman_woman_boy_boy,
  	family_woman_woman_girl_girl: family_woman_woman_girl_girl,
  	family_man_boy: family_man_boy,
  	family_man_boy_boy: family_man_boy_boy,
  	family_man_girl: family_man_girl,
  	family_man_girl_boy: family_man_girl_boy,
  	family_man_girl_girl: family_man_girl_girl,
  	family_woman_boy: family_woman_boy,
  	family_woman_boy_boy: family_woman_boy_boy,
  	family_woman_girl: family_woman_girl,
  	family_woman_girl_boy: family_woman_girl_boy,
  	family_woman_girl_girl: family_woman_girl_girl,
  	speaking_head: speaking_head,
  	bust_in_silhouette: bust_in_silhouette,
  	busts_in_silhouette: busts_in_silhouette,
  	people_hugging: people_hugging,
  	footprints: footprints,
  	monkey_face: monkey_face,
  	monkey: monkey,
  	gorilla: gorilla,
  	orangutan: orangutan,
  	dog: dog,
  	dog2: dog2,
  	guide_dog: guide_dog,
  	service_dog: service_dog,
  	poodle: poodle,
  	wolf: wolf,
  	fox_face: fox_face,
  	raccoon: raccoon,
  	cat: cat,
  	cat2: cat2,
  	black_cat: black_cat,
  	lion: lion,
  	tiger: tiger,
  	tiger2: tiger2,
  	leopard: leopard,
  	horse: horse,
  	racehorse: racehorse,
  	unicorn: unicorn,
  	zebra: zebra,
  	deer: deer,
  	bison: bison,
  	cow: cow,
  	ox: ox,
  	water_buffalo: water_buffalo,
  	cow2: cow2,
  	pig: pig,
  	pig2: pig2,
  	boar: boar,
  	pig_nose: pig_nose,
  	ram: ram,
  	sheep: sheep,
  	goat: goat,
  	dromedary_camel: dromedary_camel,
  	camel: camel,
  	llama: llama,
  	giraffe: giraffe,
  	elephant: elephant,
  	mammoth: mammoth,
  	rhinoceros: rhinoceros,
  	hippopotamus: hippopotamus,
  	mouse: mouse,
  	mouse2: mouse2,
  	rat: rat,
  	hamster: hamster,
  	rabbit: rabbit,
  	rabbit2: rabbit2,
  	chipmunk: chipmunk,
  	beaver: beaver,
  	hedgehog: hedgehog,
  	bat: bat,
  	bear: bear,
  	polar_bear: polar_bear,
  	koala: koala,
  	panda_face: panda_face,
  	sloth: sloth,
  	otter: otter,
  	skunk: skunk,
  	kangaroo: kangaroo,
  	badger: badger,
  	feet: feet,
  	paw_prints: paw_prints,
  	turkey: turkey,
  	chicken: chicken,
  	rooster: rooster,
  	hatching_chick: hatching_chick,
  	baby_chick: baby_chick,
  	hatched_chick: hatched_chick,
  	bird: bird,
  	penguin: penguin,
  	dove: dove,
  	eagle: eagle,
  	duck: duck,
  	swan: swan,
  	owl: owl,
  	dodo: dodo,
  	feather: feather,
  	flamingo: flamingo,
  	peacock: peacock,
  	parrot: parrot,
  	frog: frog,
  	crocodile: crocodile,
  	turtle: turtle,
  	lizard: lizard,
  	snake: snake,
  	dragon_face: dragon_face,
  	dragon: dragon,
  	sauropod: sauropod,
  	"t-rex": "🦖",
  	whale: whale,
  	whale2: whale2,
  	dolphin: dolphin,
  	flipper: flipper,
  	seal: seal,
  	fish: fish,
  	tropical_fish: tropical_fish,
  	blowfish: blowfish,
  	shark: shark,
  	octopus: octopus,
  	shell: shell,
  	snail: snail,
  	butterfly: butterfly,
  	bug: bug,
  	ant: ant,
  	bee: bee,
  	honeybee: honeybee,
  	beetle: beetle,
  	lady_beetle: lady_beetle,
  	cricket: cricket,
  	cockroach: cockroach,
  	spider: spider,
  	spider_web: spider_web,
  	scorpion: scorpion,
  	mosquito: mosquito,
  	fly: fly,
  	worm: worm,
  	microbe: microbe,
  	bouquet: bouquet,
  	cherry_blossom: cherry_blossom,
  	white_flower: white_flower,
  	rosette: rosette,
  	rose: rose,
  	wilted_flower: wilted_flower,
  	hibiscus: hibiscus,
  	sunflower: sunflower,
  	blossom: blossom,
  	tulip: tulip,
  	seedling: seedling,
  	potted_plant: potted_plant,
  	evergreen_tree: evergreen_tree,
  	deciduous_tree: deciduous_tree,
  	palm_tree: palm_tree,
  	cactus: cactus,
  	ear_of_rice: ear_of_rice,
  	herb: herb,
  	shamrock: shamrock,
  	four_leaf_clover: four_leaf_clover,
  	maple_leaf: maple_leaf,
  	fallen_leaf: fallen_leaf,
  	leaves: leaves,
  	grapes: grapes,
  	melon: melon,
  	watermelon: watermelon,
  	tangerine: tangerine,
  	orange: orange,
  	mandarin: mandarin,
  	lemon: lemon,
  	banana: banana,
  	pineapple: pineapple,
  	mango: mango,
  	apple: apple,
  	green_apple: green_apple,
  	pear: pear,
  	peach: peach,
  	cherries: cherries,
  	strawberry: strawberry,
  	blueberries: blueberries,
  	kiwi_fruit: kiwi_fruit,
  	tomato: tomato,
  	olive: olive,
  	coconut: coconut,
  	avocado: avocado,
  	eggplant: eggplant,
  	potato: potato,
  	carrot: carrot,
  	corn: corn,
  	hot_pepper: hot_pepper,
  	bell_pepper: bell_pepper,
  	cucumber: cucumber,
  	leafy_green: leafy_green,
  	broccoli: broccoli,
  	garlic: garlic,
  	onion: onion,
  	mushroom: mushroom,
  	peanuts: peanuts,
  	chestnut: chestnut,
  	bread: bread,
  	croissant: croissant,
  	baguette_bread: baguette_bread,
  	flatbread: flatbread,
  	pretzel: pretzel,
  	bagel: bagel,
  	pancakes: pancakes,
  	waffle: waffle,
  	cheese: cheese,
  	meat_on_bone: meat_on_bone,
  	poultry_leg: poultry_leg,
  	cut_of_meat: cut_of_meat,
  	bacon: bacon,
  	hamburger: hamburger,
  	fries: fries,
  	pizza: pizza,
  	hotdog: hotdog,
  	sandwich: sandwich,
  	taco: taco,
  	burrito: burrito,
  	tamale: tamale,
  	stuffed_flatbread: stuffed_flatbread,
  	falafel: falafel,
  	egg: egg,
  	fried_egg: fried_egg,
  	shallow_pan_of_food: shallow_pan_of_food,
  	stew: stew,
  	fondue: fondue,
  	bowl_with_spoon: bowl_with_spoon,
  	green_salad: green_salad,
  	popcorn: popcorn,
  	butter: butter,
  	salt: salt,
  	canned_food: canned_food,
  	bento: bento,
  	rice_cracker: rice_cracker,
  	rice_ball: rice_ball,
  	rice: rice,
  	curry: curry,
  	ramen: ramen,
  	spaghetti: spaghetti,
  	sweet_potato: sweet_potato,
  	oden: oden,
  	sushi: sushi,
  	fried_shrimp: fried_shrimp,
  	fish_cake: fish_cake,
  	moon_cake: moon_cake,
  	dango: dango,
  	dumpling: dumpling,
  	fortune_cookie: fortune_cookie,
  	takeout_box: takeout_box,
  	crab: crab,
  	lobster: lobster,
  	shrimp: shrimp,
  	squid: squid,
  	oyster: oyster,
  	icecream: icecream,
  	shaved_ice: shaved_ice,
  	ice_cream: ice_cream,
  	doughnut: doughnut,
  	cookie: cookie,
  	birthday: birthday,
  	cake: cake,
  	cupcake: cupcake,
  	pie: pie,
  	chocolate_bar: chocolate_bar,
  	candy: candy,
  	lollipop: lollipop,
  	custard: custard,
  	honey_pot: honey_pot,
  	baby_bottle: baby_bottle,
  	milk_glass: milk_glass,
  	coffee: coffee,
  	teapot: teapot,
  	tea: tea,
  	sake: sake,
  	champagne: champagne,
  	wine_glass: wine_glass,
  	cocktail: cocktail,
  	tropical_drink: tropical_drink,
  	beer: beer,
  	beers: beers,
  	clinking_glasses: clinking_glasses,
  	tumbler_glass: tumbler_glass,
  	cup_with_straw: cup_with_straw,
  	bubble_tea: bubble_tea,
  	beverage_box: beverage_box,
  	mate: mate,
  	ice_cube: ice_cube,
  	chopsticks: chopsticks,
  	plate_with_cutlery: plate_with_cutlery,
  	fork_and_knife: fork_and_knife,
  	spoon: spoon,
  	hocho: hocho,
  	knife: knife,
  	amphora: amphora,
  	earth_africa: earth_africa,
  	earth_americas: earth_americas,
  	earth_asia: earth_asia,
  	globe_with_meridians: globe_with_meridians,
  	world_map: world_map,
  	japan: japan,
  	compass: compass,
  	mountain_snow: mountain_snow,
  	mountain: mountain,
  	volcano: volcano,
  	mount_fuji: mount_fuji,
  	camping: camping,
  	beach_umbrella: beach_umbrella,
  	desert: desert,
  	desert_island: desert_island,
  	national_park: national_park,
  	stadium: stadium,
  	classical_building: classical_building,
  	building_construction: building_construction,
  	bricks: bricks,
  	rock: rock,
  	wood: wood,
  	hut: hut,
  	houses: houses,
  	derelict_house: derelict_house,
  	house: house,
  	house_with_garden: house_with_garden,
  	office: office,
  	post_office: post_office,
  	european_post_office: european_post_office,
  	hospital: hospital,
  	bank: bank,
  	hotel: hotel,
  	love_hotel: love_hotel,
  	convenience_store: convenience_store,
  	school: school,
  	department_store: department_store,
  	factory: factory,
  	japanese_castle: japanese_castle,
  	european_castle: european_castle,
  	wedding: wedding,
  	tokyo_tower: tokyo_tower,
  	statue_of_liberty: statue_of_liberty,
  	church: church,
  	mosque: mosque,
  	hindu_temple: hindu_temple,
  	synagogue: synagogue,
  	shinto_shrine: shinto_shrine,
  	kaaba: kaaba,
  	fountain: fountain,
  	tent: tent,
  	foggy: foggy,
  	night_with_stars: night_with_stars,
  	cityscape: cityscape,
  	sunrise_over_mountains: sunrise_over_mountains,
  	sunrise: sunrise,
  	city_sunset: city_sunset,
  	city_sunrise: city_sunrise,
  	bridge_at_night: bridge_at_night,
  	hotsprings: hotsprings,
  	carousel_horse: carousel_horse,
  	ferris_wheel: ferris_wheel,
  	roller_coaster: roller_coaster,
  	barber: barber,
  	circus_tent: circus_tent,
  	steam_locomotive: steam_locomotive,
  	railway_car: railway_car,
  	bullettrain_side: bullettrain_side,
  	bullettrain_front: bullettrain_front,
  	train2: train2,
  	metro: metro,
  	light_rail: light_rail,
  	station: station,
  	tram: tram,
  	monorail: monorail,
  	mountain_railway: mountain_railway,
  	train: train,
  	bus: bus,
  	oncoming_bus: oncoming_bus,
  	trolleybus: trolleybus,
  	minibus: minibus,
  	ambulance: ambulance,
  	fire_engine: fire_engine,
  	police_car: police_car,
  	oncoming_police_car: oncoming_police_car,
  	taxi: taxi,
  	oncoming_taxi: oncoming_taxi,
  	car: car,
  	red_car: red_car,
  	oncoming_automobile: oncoming_automobile,
  	blue_car: blue_car,
  	pickup_truck: pickup_truck,
  	truck: truck,
  	articulated_lorry: articulated_lorry,
  	tractor: tractor,
  	racing_car: racing_car,
  	motorcycle: motorcycle,
  	motor_scooter: motor_scooter,
  	manual_wheelchair: manual_wheelchair,
  	motorized_wheelchair: motorized_wheelchair,
  	auto_rickshaw: auto_rickshaw,
  	bike: bike,
  	kick_scooter: kick_scooter,
  	skateboard: skateboard,
  	roller_skate: roller_skate,
  	busstop: busstop,
  	motorway: motorway,
  	railway_track: railway_track,
  	oil_drum: oil_drum,
  	fuelpump: fuelpump,
  	rotating_light: rotating_light,
  	traffic_light: traffic_light,
  	vertical_traffic_light: vertical_traffic_light,
  	stop_sign: stop_sign,
  	construction: construction,
  	anchor: anchor,
  	boat: boat,
  	sailboat: sailboat,
  	canoe: canoe,
  	speedboat: speedboat,
  	passenger_ship: passenger_ship,
  	ferry: ferry,
  	motor_boat: motor_boat,
  	ship: ship,
  	airplane: airplane,
  	small_airplane: small_airplane,
  	flight_departure: flight_departure,
  	flight_arrival: flight_arrival,
  	parachute: parachute,
  	seat: seat,
  	helicopter: helicopter,
  	suspension_railway: suspension_railway,
  	mountain_cableway: mountain_cableway,
  	aerial_tramway: aerial_tramway,
  	artificial_satellite: artificial_satellite,
  	rocket: rocket,
  	flying_saucer: flying_saucer,
  	bellhop_bell: bellhop_bell,
  	luggage: luggage,
  	hourglass: hourglass,
  	hourglass_flowing_sand: hourglass_flowing_sand,
  	watch: watch,
  	alarm_clock: alarm_clock,
  	stopwatch: stopwatch,
  	timer_clock: timer_clock,
  	mantelpiece_clock: mantelpiece_clock,
  	clock12: clock12,
  	clock1230: clock1230,
  	clock1: clock1,
  	clock130: clock130,
  	clock2: clock2,
  	clock230: clock230,
  	clock3: clock3,
  	clock330: clock330,
  	clock4: clock4,
  	clock430: clock430,
  	clock5: clock5,
  	clock530: clock530,
  	clock6: clock6,
  	clock630: clock630,
  	clock7: clock7,
  	clock730: clock730,
  	clock8: clock8,
  	clock830: clock830,
  	clock9: clock9,
  	clock930: clock930,
  	clock10: clock10,
  	clock1030: clock1030,
  	clock11: clock11,
  	clock1130: clock1130,
  	new_moon: new_moon,
  	waxing_crescent_moon: waxing_crescent_moon,
  	first_quarter_moon: first_quarter_moon,
  	moon: moon,
  	waxing_gibbous_moon: waxing_gibbous_moon,
  	full_moon: full_moon,
  	waning_gibbous_moon: waning_gibbous_moon,
  	last_quarter_moon: last_quarter_moon,
  	waning_crescent_moon: waning_crescent_moon,
  	crescent_moon: crescent_moon,
  	new_moon_with_face: new_moon_with_face,
  	first_quarter_moon_with_face: first_quarter_moon_with_face,
  	last_quarter_moon_with_face: last_quarter_moon_with_face,
  	thermometer: thermometer,
  	sunny: sunny,
  	full_moon_with_face: full_moon_with_face,
  	sun_with_face: sun_with_face,
  	ringed_planet: ringed_planet,
  	star: star$1,
  	star2: star2,
  	stars: stars,
  	milky_way: milky_way,
  	cloud: cloud,
  	partly_sunny: partly_sunny,
  	cloud_with_lightning_and_rain: cloud_with_lightning_and_rain,
  	sun_behind_small_cloud: sun_behind_small_cloud,
  	sun_behind_large_cloud: sun_behind_large_cloud,
  	sun_behind_rain_cloud: sun_behind_rain_cloud,
  	cloud_with_rain: cloud_with_rain,
  	cloud_with_snow: cloud_with_snow,
  	cloud_with_lightning: cloud_with_lightning,
  	tornado: tornado,
  	fog: fog,
  	wind_face: wind_face,
  	cyclone: cyclone,
  	rainbow: rainbow,
  	closed_umbrella: closed_umbrella,
  	open_umbrella: open_umbrella,
  	umbrella: umbrella,
  	parasol_on_ground: parasol_on_ground,
  	zap: zap,
  	snowflake: snowflake,
  	snowman_with_snow: snowman_with_snow,
  	snowman: snowman,
  	comet: comet,
  	fire: fire,
  	droplet: droplet,
  	ocean: ocean,
  	jack_o_lantern: jack_o_lantern,
  	christmas_tree: christmas_tree,
  	fireworks: fireworks,
  	sparkler: sparkler,
  	firecracker: firecracker,
  	sparkles: sparkles,
  	balloon: balloon,
  	tada: tada,
  	confetti_ball: confetti_ball,
  	tanabata_tree: tanabata_tree,
  	bamboo: bamboo,
  	dolls: dolls,
  	flags: flags,
  	wind_chime: wind_chime,
  	rice_scene: rice_scene,
  	red_envelope: red_envelope,
  	ribbon: ribbon,
  	gift: gift,
  	reminder_ribbon: reminder_ribbon,
  	tickets: tickets,
  	ticket: ticket,
  	medal_military: medal_military,
  	trophy: trophy,
  	medal_sports: medal_sports,
  	"1st_place_medal": "🥇",
  	"2nd_place_medal": "🥈",
  	"3rd_place_medal": "🥉",
  	soccer: soccer,
  	baseball: baseball,
  	softball: softball,
  	basketball: basketball,
  	volleyball: volleyball,
  	football: football,
  	rugby_football: rugby_football,
  	tennis: tennis,
  	flying_disc: flying_disc,
  	bowling: bowling,
  	cricket_game: cricket_game,
  	field_hockey: field_hockey,
  	ice_hockey: ice_hockey,
  	lacrosse: lacrosse,
  	ping_pong: ping_pong,
  	badminton: badminton,
  	boxing_glove: boxing_glove,
  	martial_arts_uniform: martial_arts_uniform,
  	goal_net: goal_net,
  	golf: golf,
  	ice_skate: ice_skate,
  	fishing_pole_and_fish: fishing_pole_and_fish,
  	diving_mask: diving_mask,
  	running_shirt_with_sash: running_shirt_with_sash,
  	ski: ski,
  	sled: sled,
  	curling_stone: curling_stone,
  	dart: dart,
  	yo_yo: yo_yo,
  	kite: kite,
  	"8ball": "🎱",
  	crystal_ball: crystal_ball,
  	magic_wand: magic_wand,
  	nazar_amulet: nazar_amulet,
  	video_game: video_game,
  	joystick: joystick,
  	slot_machine: slot_machine,
  	game_die: game_die,
  	jigsaw: jigsaw,
  	teddy_bear: teddy_bear,
  	pinata: pinata,
  	nesting_dolls: nesting_dolls,
  	spades: spades$1,
  	hearts: hearts$1,
  	diamonds: diamonds,
  	clubs: clubs$1,
  	chess_pawn: chess_pawn,
  	black_joker: black_joker,
  	mahjong: mahjong,
  	flower_playing_cards: flower_playing_cards,
  	performing_arts: performing_arts,
  	framed_picture: framed_picture,
  	art: art,
  	thread: thread,
  	sewing_needle: sewing_needle,
  	yarn: yarn,
  	knot: knot,
  	eyeglasses: eyeglasses,
  	dark_sunglasses: dark_sunglasses,
  	goggles: goggles,
  	lab_coat: lab_coat,
  	safety_vest: safety_vest,
  	necktie: necktie,
  	shirt: shirt,
  	tshirt: tshirt,
  	jeans: jeans,
  	scarf: scarf,
  	gloves: gloves,
  	coat: coat,
  	socks: socks,
  	dress: dress,
  	kimono: kimono,
  	sari: sari,
  	one_piece_swimsuit: one_piece_swimsuit,
  	swim_brief: swim_brief,
  	shorts: shorts,
  	bikini: bikini,
  	womans_clothes: womans_clothes,
  	purse: purse,
  	handbag: handbag,
  	pouch: pouch,
  	shopping: shopping,
  	school_satchel: school_satchel,
  	thong_sandal: thong_sandal,
  	mans_shoe: mans_shoe,
  	shoe: shoe,
  	athletic_shoe: athletic_shoe,
  	hiking_boot: hiking_boot,
  	flat_shoe: flat_shoe,
  	high_heel: high_heel,
  	sandal: sandal,
  	ballet_shoes: ballet_shoes,
  	boot: boot,
  	crown: crown,
  	womans_hat: womans_hat,
  	tophat: tophat,
  	mortar_board: mortar_board,
  	billed_cap: billed_cap,
  	military_helmet: military_helmet,
  	rescue_worker_helmet: rescue_worker_helmet,
  	prayer_beads: prayer_beads,
  	lipstick: lipstick,
  	ring: ring$1,
  	gem: gem,
  	mute: mute,
  	speaker: speaker,
  	sound: sound,
  	loud_sound: loud_sound,
  	loudspeaker: loudspeaker,
  	mega: mega,
  	postal_horn: postal_horn,
  	bell: bell,
  	no_bell: no_bell,
  	musical_score: musical_score,
  	musical_note: musical_note,
  	notes: notes,
  	studio_microphone: studio_microphone,
  	level_slider: level_slider,
  	control_knobs: control_knobs,
  	microphone: microphone,
  	headphones: headphones,
  	radio: radio,
  	saxophone: saxophone,
  	accordion: accordion,
  	guitar: guitar,
  	musical_keyboard: musical_keyboard,
  	trumpet: trumpet,
  	violin: violin,
  	banjo: banjo,
  	drum: drum,
  	long_drum: long_drum,
  	iphone: iphone,
  	calling: calling,
  	phone: phone$1,
  	telephone: telephone,
  	telephone_receiver: telephone_receiver,
  	pager: pager,
  	fax: fax,
  	battery: battery,
  	electric_plug: electric_plug,
  	computer: computer,
  	desktop_computer: desktop_computer,
  	printer: printer,
  	keyboard: keyboard,
  	computer_mouse: computer_mouse,
  	trackball: trackball,
  	minidisc: minidisc,
  	floppy_disk: floppy_disk,
  	cd: cd,
  	dvd: dvd,
  	abacus: abacus,
  	movie_camera: movie_camera,
  	film_strip: film_strip,
  	film_projector: film_projector,
  	clapper: clapper,
  	tv: tv,
  	camera: camera,
  	camera_flash: camera_flash,
  	video_camera: video_camera,
  	vhs: vhs,
  	mag: mag,
  	mag_right: mag_right,
  	candle: candle,
  	bulb: bulb,
  	flashlight: flashlight,
  	izakaya_lantern: izakaya_lantern,
  	lantern: lantern,
  	diya_lamp: diya_lamp,
  	notebook_with_decorative_cover: notebook_with_decorative_cover,
  	closed_book: closed_book,
  	book: book,
  	open_book: open_book,
  	green_book: green_book,
  	blue_book: blue_book,
  	orange_book: orange_book,
  	books: books,
  	notebook: notebook,
  	ledger: ledger,
  	page_with_curl: page_with_curl,
  	scroll: scroll,
  	page_facing_up: page_facing_up,
  	newspaper: newspaper,
  	newspaper_roll: newspaper_roll,
  	bookmark_tabs: bookmark_tabs,
  	bookmark: bookmark,
  	label: label,
  	moneybag: moneybag,
  	coin: coin,
  	yen: yen$1,
  	dollar: dollar$1,
  	euro: euro$1,
  	pound: pound$1,
  	money_with_wings: money_with_wings,
  	credit_card: credit_card,
  	receipt: receipt,
  	chart: chart,
  	envelope: envelope,
  	email: email,
  	"e-mail": "📧",
  	incoming_envelope: incoming_envelope,
  	envelope_with_arrow: envelope_with_arrow,
  	outbox_tray: outbox_tray,
  	inbox_tray: inbox_tray,
  	"package": "📦",
  	mailbox: mailbox,
  	mailbox_closed: mailbox_closed,
  	mailbox_with_mail: mailbox_with_mail,
  	mailbox_with_no_mail: mailbox_with_no_mail,
  	postbox: postbox,
  	ballot_box: ballot_box,
  	pencil2: pencil2,
  	black_nib: black_nib,
  	fountain_pen: fountain_pen,
  	pen: pen,
  	paintbrush: paintbrush,
  	crayon: crayon,
  	memo: memo,
  	pencil: pencil,
  	briefcase: briefcase,
  	file_folder: file_folder,
  	open_file_folder: open_file_folder,
  	card_index_dividers: card_index_dividers,
  	date: date,
  	calendar: calendar,
  	spiral_notepad: spiral_notepad,
  	spiral_calendar: spiral_calendar,
  	card_index: card_index,
  	chart_with_upwards_trend: chart_with_upwards_trend,
  	chart_with_downwards_trend: chart_with_downwards_trend,
  	bar_chart: bar_chart,
  	clipboard: clipboard$1,
  	pushpin: pushpin,
  	round_pushpin: round_pushpin,
  	paperclip: paperclip,
  	paperclips: paperclips,
  	straight_ruler: straight_ruler,
  	triangular_ruler: triangular_ruler,
  	scissors: scissors,
  	card_file_box: card_file_box,
  	file_cabinet: file_cabinet,
  	wastebasket: wastebasket,
  	lock: lock,
  	unlock: unlock,
  	lock_with_ink_pen: lock_with_ink_pen,
  	closed_lock_with_key: closed_lock_with_key,
  	key: key,
  	old_key: old_key,
  	hammer: hammer,
  	axe: axe,
  	pick: pick,
  	hammer_and_pick: hammer_and_pick,
  	hammer_and_wrench: hammer_and_wrench,
  	dagger: dagger$1,
  	crossed_swords: crossed_swords,
  	gun: gun,
  	boomerang: boomerang,
  	bow_and_arrow: bow_and_arrow,
  	shield: shield,
  	carpentry_saw: carpentry_saw,
  	wrench: wrench,
  	screwdriver: screwdriver,
  	nut_and_bolt: nut_and_bolt,
  	gear: gear,
  	clamp: clamp,
  	balance_scale: balance_scale,
  	probing_cane: probing_cane,
  	link: link$2,
  	chains: chains,
  	hook: hook,
  	toolbox: toolbox,
  	magnet: magnet,
  	ladder: ladder,
  	alembic: alembic,
  	test_tube: test_tube,
  	petri_dish: petri_dish,
  	dna: dna,
  	microscope: microscope,
  	telescope: telescope,
  	satellite: satellite,
  	syringe: syringe,
  	drop_of_blood: drop_of_blood,
  	pill: pill,
  	adhesive_bandage: adhesive_bandage,
  	stethoscope: stethoscope,
  	door: door,
  	elevator: elevator,
  	mirror: mirror,
  	window: window$1,
  	bed: bed,
  	couch_and_lamp: couch_and_lamp,
  	chair: chair,
  	toilet: toilet,
  	plunger: plunger,
  	shower: shower,
  	bathtub: bathtub,
  	mouse_trap: mouse_trap,
  	razor: razor,
  	lotion_bottle: lotion_bottle,
  	safety_pin: safety_pin,
  	broom: broom,
  	basket: basket,
  	roll_of_paper: roll_of_paper,
  	bucket: bucket,
  	soap: soap,
  	toothbrush: toothbrush,
  	sponge: sponge,
  	fire_extinguisher: fire_extinguisher,
  	shopping_cart: shopping_cart,
  	smoking: smoking,
  	coffin: coffin,
  	headstone: headstone,
  	funeral_urn: funeral_urn,
  	moyai: moyai,
  	placard: placard,
  	atm: atm,
  	put_litter_in_its_place: put_litter_in_its_place,
  	potable_water: potable_water,
  	wheelchair: wheelchair,
  	mens: mens,
  	womens: womens,
  	restroom: restroom,
  	baby_symbol: baby_symbol,
  	wc: wc,
  	passport_control: passport_control,
  	customs: customs,
  	baggage_claim: baggage_claim,
  	left_luggage: left_luggage,
  	warning: warning,
  	children_crossing: children_crossing,
  	no_entry: no_entry,
  	no_entry_sign: no_entry_sign,
  	no_bicycles: no_bicycles,
  	no_smoking: no_smoking,
  	do_not_litter: do_not_litter,
  	"non-potable_water": "🚱",
  	no_pedestrians: no_pedestrians,
  	no_mobile_phones: no_mobile_phones,
  	underage: underage,
  	radioactive: radioactive,
  	biohazard: biohazard,
  	arrow_up: arrow_up,
  	arrow_upper_right: arrow_upper_right,
  	arrow_right: arrow_right,
  	arrow_lower_right: arrow_lower_right,
  	arrow_down: arrow_down,
  	arrow_lower_left: arrow_lower_left,
  	arrow_left: arrow_left,
  	arrow_upper_left: arrow_upper_left,
  	arrow_up_down: arrow_up_down,
  	left_right_arrow: left_right_arrow,
  	leftwards_arrow_with_hook: leftwards_arrow_with_hook,
  	arrow_right_hook: arrow_right_hook,
  	arrow_heading_up: arrow_heading_up,
  	arrow_heading_down: arrow_heading_down,
  	arrows_clockwise: arrows_clockwise,
  	arrows_counterclockwise: arrows_counterclockwise,
  	back: back,
  	end: end,
  	on: on,
  	soon: soon,
  	top: top$1,
  	place_of_worship: place_of_worship,
  	atom_symbol: atom_symbol,
  	om: om,
  	star_of_david: star_of_david,
  	wheel_of_dharma: wheel_of_dharma,
  	yin_yang: yin_yang,
  	latin_cross: latin_cross,
  	orthodox_cross: orthodox_cross,
  	star_and_crescent: star_and_crescent,
  	peace_symbol: peace_symbol,
  	menorah: menorah,
  	six_pointed_star: six_pointed_star,
  	aries: aries,
  	taurus: taurus,
  	gemini: gemini,
  	cancer: cancer,
  	leo: leo,
  	virgo: virgo,
  	libra: libra,
  	scorpius: scorpius,
  	sagittarius: sagittarius,
  	capricorn: capricorn,
  	aquarius: aquarius,
  	pisces: pisces,
  	ophiuchus: ophiuchus,
  	twisted_rightwards_arrows: twisted_rightwards_arrows,
  	repeat: repeat,
  	repeat_one: repeat_one,
  	arrow_forward: arrow_forward,
  	fast_forward: fast_forward,
  	next_track_button: next_track_button,
  	play_or_pause_button: play_or_pause_button,
  	arrow_backward: arrow_backward,
  	rewind: rewind,
  	previous_track_button: previous_track_button,
  	arrow_up_small: arrow_up_small,
  	arrow_double_up: arrow_double_up,
  	arrow_down_small: arrow_down_small,
  	arrow_double_down: arrow_double_down,
  	pause_button: pause_button,
  	stop_button: stop_button,
  	record_button: record_button,
  	eject_button: eject_button,
  	cinema: cinema,
  	low_brightness: low_brightness,
  	high_brightness: high_brightness,
  	signal_strength: signal_strength,
  	vibration_mode: vibration_mode,
  	mobile_phone_off: mobile_phone_off,
  	female_sign: female_sign,
  	male_sign: male_sign,
  	transgender_symbol: transgender_symbol,
  	heavy_multiplication_x: heavy_multiplication_x,
  	heavy_plus_sign: heavy_plus_sign,
  	heavy_minus_sign: heavy_minus_sign,
  	heavy_division_sign: heavy_division_sign,
  	infinity: infinity,
  	bangbang: bangbang,
  	interrobang: interrobang,
  	question: question,
  	grey_question: grey_question,
  	grey_exclamation: grey_exclamation,
  	exclamation: exclamation,
  	heavy_exclamation_mark: heavy_exclamation_mark,
  	wavy_dash: wavy_dash,
  	currency_exchange: currency_exchange,
  	heavy_dollar_sign: heavy_dollar_sign,
  	medical_symbol: medical_symbol,
  	recycle: recycle,
  	fleur_de_lis: fleur_de_lis,
  	trident: trident,
  	name_badge: name_badge,
  	beginner: beginner,
  	o: o,
  	white_check_mark: white_check_mark,
  	ballot_box_with_check: ballot_box_with_check,
  	heavy_check_mark: heavy_check_mark,
  	x: x,
  	negative_squared_cross_mark: negative_squared_cross_mark,
  	curly_loop: curly_loop,
  	loop: loop,
  	part_alternation_mark: part_alternation_mark,
  	eight_spoked_asterisk: eight_spoked_asterisk,
  	eight_pointed_black_star: eight_pointed_black_star,
  	sparkle: sparkle,
  	copyright: copyright,
  	registered: registered,
  	tm: tm,
  	hash: hash,
  	asterisk: asterisk,
  	zero: zero$1,
  	one: one,
  	two: two,
  	three: three,
  	four: four,
  	five: five,
  	six: six,
  	seven: seven,
  	eight: eight,
  	nine: nine,
  	keycap_ten: keycap_ten,
  	capital_abcd: capital_abcd,
  	abcd: abcd,
  	symbols: symbols,
  	abc: abc,
  	a: a,
  	ab: ab,
  	b: b,
  	cl: cl,
  	cool: cool,
  	free: free,
  	information_source: information_source,
  	id: id,
  	m: m,
  	"new": "🆕",
  	ng: ng,
  	o2: o2,
  	ok: ok,
  	parking: parking,
  	sos: sos,
  	up: up,
  	vs: vs,
  	koko: koko,
  	sa: sa,
  	ideograph_advantage: ideograph_advantage,
  	accept: accept,
  	congratulations: congratulations,
  	secret: secret,
  	u6e80: u6e80,
  	red_circle: red_circle,
  	orange_circle: orange_circle,
  	yellow_circle: yellow_circle,
  	green_circle: green_circle,
  	large_blue_circle: large_blue_circle,
  	purple_circle: purple_circle,
  	brown_circle: brown_circle,
  	black_circle: black_circle,
  	white_circle: white_circle,
  	red_square: red_square,
  	orange_square: orange_square,
  	yellow_square: yellow_square,
  	green_square: green_square,
  	blue_square: blue_square,
  	purple_square: purple_square,
  	brown_square: brown_square,
  	black_large_square: black_large_square,
  	white_large_square: white_large_square,
  	black_medium_square: black_medium_square,
  	white_medium_square: white_medium_square,
  	black_medium_small_square: black_medium_small_square,
  	white_medium_small_square: white_medium_small_square,
  	black_small_square: black_small_square,
  	white_small_square: white_small_square,
  	large_orange_diamond: large_orange_diamond,
  	large_blue_diamond: large_blue_diamond,
  	small_orange_diamond: small_orange_diamond,
  	small_blue_diamond: small_blue_diamond,
  	small_red_triangle: small_red_triangle,
  	small_red_triangle_down: small_red_triangle_down,
  	diamond_shape_with_a_dot_inside: diamond_shape_with_a_dot_inside,
  	radio_button: radio_button,
  	white_square_button: white_square_button,
  	black_square_button: black_square_button,
  	checkered_flag: checkered_flag,
  	triangular_flag_on_post: triangular_flag_on_post,
  	crossed_flags: crossed_flags,
  	black_flag: black_flag,
  	white_flag: white_flag,
  	rainbow_flag: rainbow_flag,
  	transgender_flag: transgender_flag,
  	pirate_flag: pirate_flag,
  	ascension_island: ascension_island,
  	andorra: andorra,
  	united_arab_emirates: united_arab_emirates,
  	afghanistan: afghanistan,
  	antigua_barbuda: antigua_barbuda,
  	anguilla: anguilla,
  	albania: albania,
  	armenia: armenia,
  	angola: angola,
  	antarctica: antarctica,
  	argentina: argentina,
  	american_samoa: american_samoa,
  	austria: austria,
  	australia: australia,
  	aruba: aruba,
  	aland_islands: aland_islands,
  	azerbaijan: azerbaijan,
  	bosnia_herzegovina: bosnia_herzegovina,
  	barbados: barbados,
  	bangladesh: bangladesh,
  	belgium: belgium,
  	burkina_faso: burkina_faso,
  	bulgaria: bulgaria,
  	bahrain: bahrain,
  	burundi: burundi,
  	benin: benin,
  	st_barthelemy: st_barthelemy,
  	bermuda: bermuda,
  	brunei: brunei,
  	bolivia: bolivia,
  	caribbean_netherlands: caribbean_netherlands,
  	brazil: brazil,
  	bahamas: bahamas,
  	bhutan: bhutan,
  	bouvet_island: bouvet_island,
  	botswana: botswana,
  	belarus: belarus,
  	belize: belize,
  	canada: canada,
  	cocos_islands: cocos_islands,
  	congo_kinshasa: congo_kinshasa,
  	central_african_republic: central_african_republic,
  	congo_brazzaville: congo_brazzaville,
  	switzerland: switzerland,
  	cote_divoire: cote_divoire,
  	cook_islands: cook_islands,
  	chile: chile,
  	cameroon: cameroon,
  	cn: cn,
  	colombia: colombia,
  	clipperton_island: clipperton_island,
  	costa_rica: costa_rica,
  	cuba: cuba,
  	cape_verde: cape_verde,
  	curacao: curacao,
  	christmas_island: christmas_island,
  	cyprus: cyprus,
  	czech_republic: czech_republic,
  	de: de,
  	diego_garcia: diego_garcia,
  	djibouti: djibouti,
  	denmark: denmark,
  	dominica: dominica,
  	dominican_republic: dominican_republic,
  	algeria: algeria,
  	ceuta_melilla: ceuta_melilla,
  	ecuador: ecuador,
  	estonia: estonia,
  	egypt: egypt,
  	western_sahara: western_sahara,
  	eritrea: eritrea,
  	es: es,
  	ethiopia: ethiopia,
  	eu: eu,
  	european_union: european_union,
  	finland: finland,
  	fiji: fiji,
  	falkland_islands: falkland_islands,
  	micronesia: micronesia,
  	faroe_islands: faroe_islands,
  	fr: fr,
  	gabon: gabon,
  	gb: gb,
  	uk: uk,
  	grenada: grenada,
  	georgia: georgia,
  	french_guiana: french_guiana,
  	guernsey: guernsey,
  	ghana: ghana,
  	gibraltar: gibraltar,
  	greenland: greenland,
  	gambia: gambia,
  	guinea: guinea,
  	guadeloupe: guadeloupe,
  	equatorial_guinea: equatorial_guinea,
  	greece: greece,
  	south_georgia_south_sandwich_islands: south_georgia_south_sandwich_islands,
  	guatemala: guatemala,
  	guam: guam,
  	guinea_bissau: guinea_bissau,
  	guyana: guyana,
  	hong_kong: hong_kong,
  	heard_mcdonald_islands: heard_mcdonald_islands,
  	honduras: honduras,
  	croatia: croatia,
  	haiti: haiti,
  	hungary: hungary,
  	canary_islands: canary_islands,
  	indonesia: indonesia,
  	ireland: ireland,
  	israel: israel,
  	isle_of_man: isle_of_man,
  	india: india,
  	british_indian_ocean_territory: british_indian_ocean_territory,
  	iraq: iraq,
  	iran: iran,
  	iceland: iceland,
  	it: it$1,
  	jersey: jersey,
  	jamaica: jamaica,
  	jordan: jordan,
  	jp: jp,
  	kenya: kenya,
  	kyrgyzstan: kyrgyzstan,
  	cambodia: cambodia,
  	kiribati: kiribati,
  	comoros: comoros,
  	st_kitts_nevis: st_kitts_nevis,
  	north_korea: north_korea,
  	kr: kr,
  	kuwait: kuwait,
  	cayman_islands: cayman_islands,
  	kazakhstan: kazakhstan,
  	laos: laos,
  	lebanon: lebanon,
  	st_lucia: st_lucia,
  	liechtenstein: liechtenstein,
  	sri_lanka: sri_lanka,
  	liberia: liberia,
  	lesotho: lesotho,
  	lithuania: lithuania,
  	luxembourg: luxembourg,
  	latvia: latvia,
  	libya: libya,
  	morocco: morocco,
  	monaco: monaco,
  	moldova: moldova,
  	montenegro: montenegro,
  	st_martin: st_martin,
  	madagascar: madagascar,
  	marshall_islands: marshall_islands,
  	macedonia: macedonia,
  	mali: mali,
  	myanmar: myanmar,
  	mongolia: mongolia,
  	macau: macau,
  	northern_mariana_islands: northern_mariana_islands,
  	martinique: martinique,
  	mauritania: mauritania,
  	montserrat: montserrat,
  	malta: malta,
  	mauritius: mauritius,
  	maldives: maldives,
  	malawi: malawi,
  	mexico: mexico,
  	malaysia: malaysia,
  	mozambique: mozambique,
  	namibia: namibia,
  	new_caledonia: new_caledonia,
  	niger: niger,
  	norfolk_island: norfolk_island,
  	nigeria: nigeria,
  	nicaragua: nicaragua,
  	netherlands: netherlands,
  	norway: norway,
  	nepal: nepal,
  	nauru: nauru,
  	niue: niue,
  	new_zealand: new_zealand,
  	oman: oman,
  	panama: panama,
  	peru: peru,
  	french_polynesia: french_polynesia,
  	papua_new_guinea: papua_new_guinea,
  	philippines: philippines,
  	pakistan: pakistan,
  	poland: poland,
  	st_pierre_miquelon: st_pierre_miquelon,
  	pitcairn_islands: pitcairn_islands,
  	puerto_rico: puerto_rico,
  	palestinian_territories: palestinian_territories,
  	portugal: portugal,
  	palau: palau,
  	paraguay: paraguay,
  	qatar: qatar,
  	reunion: reunion,
  	romania: romania,
  	serbia: serbia,
  	ru: ru,
  	rwanda: rwanda,
  	saudi_arabia: saudi_arabia,
  	solomon_islands: solomon_islands,
  	seychelles: seychelles,
  	sudan: sudan,
  	sweden: sweden,
  	singapore: singapore,
  	st_helena: st_helena,
  	slovenia: slovenia,
  	svalbard_jan_mayen: svalbard_jan_mayen,
  	slovakia: slovakia,
  	sierra_leone: sierra_leone,
  	san_marino: san_marino,
  	senegal: senegal,
  	somalia: somalia,
  	suriname: suriname,
  	south_sudan: south_sudan,
  	sao_tome_principe: sao_tome_principe,
  	el_salvador: el_salvador,
  	sint_maarten: sint_maarten,
  	syria: syria,
  	swaziland: swaziland,
  	tristan_da_cunha: tristan_da_cunha,
  	turks_caicos_islands: turks_caicos_islands,
  	chad: chad,
  	french_southern_territories: french_southern_territories,
  	togo: togo,
  	thailand: thailand,
  	tajikistan: tajikistan,
  	tokelau: tokelau,
  	timor_leste: timor_leste,
  	turkmenistan: turkmenistan,
  	tunisia: tunisia,
  	tonga: tonga,
  	tr: tr,
  	trinidad_tobago: trinidad_tobago,
  	tuvalu: tuvalu,
  	taiwan: taiwan,
  	tanzania: tanzania,
  	ukraine: ukraine,
  	uganda: uganda,
  	us_outlying_islands: us_outlying_islands,
  	united_nations: united_nations,
  	us: us,
  	uruguay: uruguay,
  	uzbekistan: uzbekistan,
  	vatican_city: vatican_city,
  	st_vincent_grenadines: st_vincent_grenadines,
  	venezuela: venezuela,
  	british_virgin_islands: british_virgin_islands,
  	us_virgin_islands: us_virgin_islands,
  	vietnam: vietnam,
  	vanuatu: vanuatu,
  	wallis_futuna: wallis_futuna,
  	samoa: samoa,
  	kosovo: kosovo,
  	yemen: yemen,
  	mayotte: mayotte,
  	south_africa: south_africa,
  	zambia: zambia,
  	zimbabwe: zimbabwe,
  	england: england,
  	scotland: scotland,
  	wales: wales
  };

  var full$1 = /*#__PURE__*/Object.freeze({
    __proto__: null,
    grinning: grinning,
    smiley: smiley,
    smile: smile$1,
    grin: grin,
    laughing: laughing,
    satisfied: satisfied,
    sweat_smile: sweat_smile,
    rofl: rofl,
    joy: joy,
    slightly_smiling_face: slightly_smiling_face,
    upside_down_face: upside_down_face,
    wink: wink,
    blush: blush,
    innocent: innocent,
    smiling_face_with_three_hearts: smiling_face_with_three_hearts,
    heart_eyes: heart_eyes,
    star_struck: star_struck,
    kissing_heart: kissing_heart,
    kissing: kissing,
    relaxed: relaxed,
    kissing_closed_eyes: kissing_closed_eyes,
    kissing_smiling_eyes: kissing_smiling_eyes,
    smiling_face_with_tear: smiling_face_with_tear,
    yum: yum,
    stuck_out_tongue: stuck_out_tongue,
    stuck_out_tongue_winking_eye: stuck_out_tongue_winking_eye,
    zany_face: zany_face,
    stuck_out_tongue_closed_eyes: stuck_out_tongue_closed_eyes,
    money_mouth_face: money_mouth_face,
    hugs: hugs,
    hand_over_mouth: hand_over_mouth,
    shushing_face: shushing_face,
    thinking: thinking,
    zipper_mouth_face: zipper_mouth_face,
    raised_eyebrow: raised_eyebrow,
    neutral_face: neutral_face,
    expressionless: expressionless,
    no_mouth: no_mouth,
    smirk: smirk,
    unamused: unamused,
    roll_eyes: roll_eyes,
    grimacing: grimacing,
    lying_face: lying_face,
    relieved: relieved,
    pensive: pensive,
    sleepy: sleepy,
    drooling_face: drooling_face,
    sleeping: sleeping,
    mask: mask,
    face_with_thermometer: face_with_thermometer,
    face_with_head_bandage: face_with_head_bandage,
    nauseated_face: nauseated_face,
    vomiting_face: vomiting_face,
    sneezing_face: sneezing_face,
    hot_face: hot_face,
    cold_face: cold_face,
    woozy_face: woozy_face,
    dizzy_face: dizzy_face,
    exploding_head: exploding_head,
    cowboy_hat_face: cowboy_hat_face,
    partying_face: partying_face,
    disguised_face: disguised_face,
    sunglasses: sunglasses,
    nerd_face: nerd_face,
    monocle_face: monocle_face,
    confused: confused,
    worried: worried,
    slightly_frowning_face: slightly_frowning_face,
    frowning_face: frowning_face,
    open_mouth: open_mouth,
    hushed: hushed,
    astonished: astonished,
    flushed: flushed,
    pleading_face: pleading_face,
    frowning: frowning,
    anguished: anguished,
    fearful: fearful,
    cold_sweat: cold_sweat,
    disappointed_relieved: disappointed_relieved,
    cry: cry,
    sob: sob,
    scream: scream,
    confounded: confounded,
    persevere: persevere,
    disappointed: disappointed,
    sweat: sweat,
    weary: weary,
    tired_face: tired_face,
    yawning_face: yawning_face,
    triumph: triumph,
    rage: rage,
    pout: pout,
    angry: angry,
    cursing_face: cursing_face,
    smiling_imp: smiling_imp,
    imp: imp,
    skull: skull,
    skull_and_crossbones: skull_and_crossbones,
    hankey: hankey,
    poop: poop,
    shit: shit,
    clown_face: clown_face,
    japanese_ogre: japanese_ogre,
    japanese_goblin: japanese_goblin,
    ghost: ghost,
    alien: alien,
    space_invader: space_invader,
    robot: robot,
    smiley_cat: smiley_cat,
    smile_cat: smile_cat,
    joy_cat: joy_cat,
    heart_eyes_cat: heart_eyes_cat,
    smirk_cat: smirk_cat,
    kissing_cat: kissing_cat,
    scream_cat: scream_cat,
    crying_cat_face: crying_cat_face,
    pouting_cat: pouting_cat,
    see_no_evil: see_no_evil,
    hear_no_evil: hear_no_evil,
    speak_no_evil: speak_no_evil,
    kiss: kiss,
    love_letter: love_letter,
    cupid: cupid,
    gift_heart: gift_heart,
    sparkling_heart: sparkling_heart,
    heartpulse: heartpulse,
    heartbeat: heartbeat,
    revolving_hearts: revolving_hearts,
    two_hearts: two_hearts,
    heart_decoration: heart_decoration,
    heavy_heart_exclamation: heavy_heart_exclamation,
    broken_heart: broken_heart,
    heart: heart,
    orange_heart: orange_heart,
    yellow_heart: yellow_heart,
    green_heart: green_heart,
    blue_heart: blue_heart,
    purple_heart: purple_heart,
    brown_heart: brown_heart,
    black_heart: black_heart,
    white_heart: white_heart,
    anger: anger,
    boom: boom,
    collision: collision,
    dizzy: dizzy,
    sweat_drops: sweat_drops,
    dash: dash$1,
    hole: hole,
    bomb: bomb,
    speech_balloon: speech_balloon,
    eye_speech_bubble: eye_speech_bubble,
    left_speech_bubble: left_speech_bubble,
    right_anger_bubble: right_anger_bubble,
    thought_balloon: thought_balloon,
    zzz: zzz,
    wave: wave,
    raised_back_of_hand: raised_back_of_hand,
    raised_hand_with_fingers_splayed: raised_hand_with_fingers_splayed,
    hand: hand,
    raised_hand: raised_hand,
    vulcan_salute: vulcan_salute,
    ok_hand: ok_hand,
    pinched_fingers: pinched_fingers,
    pinching_hand: pinching_hand,
    v: v,
    crossed_fingers: crossed_fingers,
    love_you_gesture: love_you_gesture,
    metal: metal,
    call_me_hand: call_me_hand,
    point_left: point_left,
    point_right: point_right,
    point_up_2: point_up_2,
    middle_finger: middle_finger,
    fu: fu,
    point_down: point_down,
    point_up: point_up,
    thumbsup: thumbsup,
    thumbsdown: thumbsdown,
    fist_raised: fist_raised,
    fist: fist,
    fist_oncoming: fist_oncoming,
    facepunch: facepunch,
    punch: punch,
    fist_left: fist_left,
    fist_right: fist_right,
    clap: clap,
    raised_hands: raised_hands,
    open_hands: open_hands,
    palms_up_together: palms_up_together,
    handshake: handshake,
    pray: pray,
    writing_hand: writing_hand,
    nail_care: nail_care,
    selfie: selfie,
    muscle: muscle,
    mechanical_arm: mechanical_arm,
    mechanical_leg: mechanical_leg,
    leg: leg$1,
    foot: foot,
    ear: ear,
    ear_with_hearing_aid: ear_with_hearing_aid,
    nose: nose,
    brain: brain,
    anatomical_heart: anatomical_heart,
    lungs: lungs,
    tooth: tooth,
    bone: bone,
    eyes: eyes,
    eye: eye,
    tongue: tongue,
    lips: lips,
    baby: baby,
    child: child,
    boy: boy,
    girl: girl,
    adult: adult,
    blond_haired_person: blond_haired_person,
    man: man,
    bearded_person: bearded_person,
    red_haired_man: red_haired_man,
    curly_haired_man: curly_haired_man,
    white_haired_man: white_haired_man,
    bald_man: bald_man,
    woman: woman,
    red_haired_woman: red_haired_woman,
    person_red_hair: person_red_hair,
    curly_haired_woman: curly_haired_woman,
    person_curly_hair: person_curly_hair,
    white_haired_woman: white_haired_woman,
    person_white_hair: person_white_hair,
    bald_woman: bald_woman,
    person_bald: person_bald,
    blond_haired_woman: blond_haired_woman,
    blonde_woman: blonde_woman,
    blond_haired_man: blond_haired_man,
    older_adult: older_adult,
    older_man: older_man,
    older_woman: older_woman,
    frowning_person: frowning_person,
    frowning_man: frowning_man,
    frowning_woman: frowning_woman,
    pouting_face: pouting_face,
    pouting_man: pouting_man,
    pouting_woman: pouting_woman,
    no_good: no_good,
    no_good_man: no_good_man,
    ng_man: ng_man,
    no_good_woman: no_good_woman,
    ng_woman: ng_woman,
    ok_person: ok_person,
    ok_man: ok_man,
    ok_woman: ok_woman,
    tipping_hand_person: tipping_hand_person,
    information_desk_person: information_desk_person,
    tipping_hand_man: tipping_hand_man,
    sassy_man: sassy_man,
    tipping_hand_woman: tipping_hand_woman,
    sassy_woman: sassy_woman,
    raising_hand: raising_hand,
    raising_hand_man: raising_hand_man,
    raising_hand_woman: raising_hand_woman,
    deaf_person: deaf_person,
    deaf_man: deaf_man,
    deaf_woman: deaf_woman,
    bow: bow,
    bowing_man: bowing_man,
    bowing_woman: bowing_woman,
    facepalm: facepalm,
    man_facepalming: man_facepalming,
    woman_facepalming: woman_facepalming,
    shrug: shrug,
    man_shrugging: man_shrugging,
    woman_shrugging: woman_shrugging,
    health_worker: health_worker,
    man_health_worker: man_health_worker,
    woman_health_worker: woman_health_worker,
    student: student,
    man_student: man_student,
    woman_student: woman_student,
    teacher: teacher,
    man_teacher: man_teacher,
    woman_teacher: woman_teacher,
    judge: judge,
    man_judge: man_judge,
    woman_judge: woman_judge,
    farmer: farmer,
    man_farmer: man_farmer,
    woman_farmer: woman_farmer,
    cook: cook,
    man_cook: man_cook,
    woman_cook: woman_cook,
    mechanic: mechanic,
    man_mechanic: man_mechanic,
    woman_mechanic: woman_mechanic,
    factory_worker: factory_worker,
    man_factory_worker: man_factory_worker,
    woman_factory_worker: woman_factory_worker,
    office_worker: office_worker,
    man_office_worker: man_office_worker,
    woman_office_worker: woman_office_worker,
    scientist: scientist,
    man_scientist: man_scientist,
    woman_scientist: woman_scientist,
    technologist: technologist,
    man_technologist: man_technologist,
    woman_technologist: woman_technologist,
    singer: singer,
    man_singer: man_singer,
    woman_singer: woman_singer,
    artist: artist,
    man_artist: man_artist,
    woman_artist: woman_artist,
    pilot: pilot,
    man_pilot: man_pilot,
    woman_pilot: woman_pilot,
    astronaut: astronaut,
    man_astronaut: man_astronaut,
    woman_astronaut: woman_astronaut,
    firefighter: firefighter,
    man_firefighter: man_firefighter,
    woman_firefighter: woman_firefighter,
    police_officer: police_officer,
    cop: cop,
    policeman: policeman,
    policewoman: policewoman,
    detective: detective,
    male_detective: male_detective,
    female_detective: female_detective,
    guard: guard,
    guardsman: guardsman,
    guardswoman: guardswoman,
    ninja: ninja,
    construction_worker: construction_worker,
    construction_worker_man: construction_worker_man,
    construction_worker_woman: construction_worker_woman,
    prince: prince,
    princess: princess,
    person_with_turban: person_with_turban,
    man_with_turban: man_with_turban,
    woman_with_turban: woman_with_turban,
    man_with_gua_pi_mao: man_with_gua_pi_mao,
    woman_with_headscarf: woman_with_headscarf,
    person_in_tuxedo: person_in_tuxedo,
    man_in_tuxedo: man_in_tuxedo,
    woman_in_tuxedo: woman_in_tuxedo,
    person_with_veil: person_with_veil,
    man_with_veil: man_with_veil,
    woman_with_veil: woman_with_veil,
    bride_with_veil: bride_with_veil,
    pregnant_woman: pregnant_woman,
    breast_feeding: breast_feeding,
    woman_feeding_baby: woman_feeding_baby,
    man_feeding_baby: man_feeding_baby,
    person_feeding_baby: person_feeding_baby,
    angel: angel,
    santa: santa,
    mrs_claus: mrs_claus,
    mx_claus: mx_claus,
    superhero: superhero,
    superhero_man: superhero_man,
    superhero_woman: superhero_woman,
    supervillain: supervillain,
    supervillain_man: supervillain_man,
    supervillain_woman: supervillain_woman,
    mage: mage,
    mage_man: mage_man,
    mage_woman: mage_woman,
    fairy: fairy,
    fairy_man: fairy_man,
    fairy_woman: fairy_woman,
    vampire: vampire,
    vampire_man: vampire_man,
    vampire_woman: vampire_woman,
    merperson: merperson,
    merman: merman,
    mermaid: mermaid,
    elf: elf,
    elf_man: elf_man,
    elf_woman: elf_woman,
    genie: genie,
    genie_man: genie_man,
    genie_woman: genie_woman,
    zombie: zombie,
    zombie_man: zombie_man,
    zombie_woman: zombie_woman,
    massage: massage,
    massage_man: massage_man,
    massage_woman: massage_woman,
    haircut: haircut,
    haircut_man: haircut_man,
    haircut_woman: haircut_woman,
    walking: walking,
    walking_man: walking_man,
    walking_woman: walking_woman,
    standing_person: standing_person,
    standing_man: standing_man,
    standing_woman: standing_woman,
    kneeling_person: kneeling_person,
    kneeling_man: kneeling_man,
    kneeling_woman: kneeling_woman,
    person_with_probing_cane: person_with_probing_cane,
    man_with_probing_cane: man_with_probing_cane,
    woman_with_probing_cane: woman_with_probing_cane,
    person_in_motorized_wheelchair: person_in_motorized_wheelchair,
    man_in_motorized_wheelchair: man_in_motorized_wheelchair,
    woman_in_motorized_wheelchair: woman_in_motorized_wheelchair,
    person_in_manual_wheelchair: person_in_manual_wheelchair,
    man_in_manual_wheelchair: man_in_manual_wheelchair,
    woman_in_manual_wheelchair: woman_in_manual_wheelchair,
    runner: runner,
    running: running,
    running_man: running_man,
    running_woman: running_woman,
    woman_dancing: woman_dancing,
    dancer: dancer,
    man_dancing: man_dancing,
    business_suit_levitating: business_suit_levitating,
    dancers: dancers,
    dancing_men: dancing_men,
    dancing_women: dancing_women,
    sauna_person: sauna_person,
    sauna_man: sauna_man,
    sauna_woman: sauna_woman,
    climbing: climbing,
    climbing_man: climbing_man,
    climbing_woman: climbing_woman,
    person_fencing: person_fencing,
    horse_racing: horse_racing,
    skier: skier,
    snowboarder: snowboarder,
    golfing: golfing,
    golfing_man: golfing_man,
    golfing_woman: golfing_woman,
    surfer: surfer,
    surfing_man: surfing_man,
    surfing_woman: surfing_woman,
    rowboat: rowboat,
    rowing_man: rowing_man,
    rowing_woman: rowing_woman,
    swimmer: swimmer,
    swimming_man: swimming_man,
    swimming_woman: swimming_woman,
    bouncing_ball_person: bouncing_ball_person,
    bouncing_ball_man: bouncing_ball_man,
    basketball_man: basketball_man,
    bouncing_ball_woman: bouncing_ball_woman,
    basketball_woman: basketball_woman,
    weight_lifting: weight_lifting,
    weight_lifting_man: weight_lifting_man,
    weight_lifting_woman: weight_lifting_woman,
    bicyclist: bicyclist,
    biking_man: biking_man,
    biking_woman: biking_woman,
    mountain_bicyclist: mountain_bicyclist,
    mountain_biking_man: mountain_biking_man,
    mountain_biking_woman: mountain_biking_woman,
    cartwheeling: cartwheeling,
    man_cartwheeling: man_cartwheeling,
    woman_cartwheeling: woman_cartwheeling,
    wrestling: wrestling,
    men_wrestling: men_wrestling,
    women_wrestling: women_wrestling,
    water_polo: water_polo,
    man_playing_water_polo: man_playing_water_polo,
    woman_playing_water_polo: woman_playing_water_polo,
    handball_person: handball_person,
    man_playing_handball: man_playing_handball,
    woman_playing_handball: woman_playing_handball,
    juggling_person: juggling_person,
    man_juggling: man_juggling,
    woman_juggling: woman_juggling,
    lotus_position: lotus_position,
    lotus_position_man: lotus_position_man,
    lotus_position_woman: lotus_position_woman,
    bath: bath,
    sleeping_bed: sleeping_bed,
    people_holding_hands: people_holding_hands,
    two_women_holding_hands: two_women_holding_hands,
    couple: couple,
    two_men_holding_hands: two_men_holding_hands,
    couplekiss: couplekiss,
    couplekiss_man_woman: couplekiss_man_woman,
    couplekiss_man_man: couplekiss_man_man,
    couplekiss_woman_woman: couplekiss_woman_woman,
    couple_with_heart: couple_with_heart,
    couple_with_heart_woman_man: couple_with_heart_woman_man,
    couple_with_heart_man_man: couple_with_heart_man_man,
    couple_with_heart_woman_woman: couple_with_heart_woman_woman,
    family: family,
    family_man_woman_boy: family_man_woman_boy,
    family_man_woman_girl: family_man_woman_girl,
    family_man_woman_girl_boy: family_man_woman_girl_boy,
    family_man_woman_boy_boy: family_man_woman_boy_boy,
    family_man_woman_girl_girl: family_man_woman_girl_girl,
    family_man_man_boy: family_man_man_boy,
    family_man_man_girl: family_man_man_girl,
    family_man_man_girl_boy: family_man_man_girl_boy,
    family_man_man_boy_boy: family_man_man_boy_boy,
    family_man_man_girl_girl: family_man_man_girl_girl,
    family_woman_woman_boy: family_woman_woman_boy,
    family_woman_woman_girl: family_woman_woman_girl,
    family_woman_woman_girl_boy: family_woman_woman_girl_boy,
    family_woman_woman_boy_boy: family_woman_woman_boy_boy,
    family_woman_woman_girl_girl: family_woman_woman_girl_girl,
    family_man_boy: family_man_boy,
    family_man_boy_boy: family_man_boy_boy,
    family_man_girl: family_man_girl,
    family_man_girl_boy: family_man_girl_boy,
    family_man_girl_girl: family_man_girl_girl,
    family_woman_boy: family_woman_boy,
    family_woman_boy_boy: family_woman_boy_boy,
    family_woman_girl: family_woman_girl,
    family_woman_girl_boy: family_woman_girl_boy,
    family_woman_girl_girl: family_woman_girl_girl,
    speaking_head: speaking_head,
    bust_in_silhouette: bust_in_silhouette,
    busts_in_silhouette: busts_in_silhouette,
    people_hugging: people_hugging,
    footprints: footprints,
    monkey_face: monkey_face,
    monkey: monkey,
    gorilla: gorilla,
    orangutan: orangutan,
    dog: dog,
    dog2: dog2,
    guide_dog: guide_dog,
    service_dog: service_dog,
    poodle: poodle,
    wolf: wolf,
    fox_face: fox_face,
    raccoon: raccoon,
    cat: cat,
    cat2: cat2,
    black_cat: black_cat,
    lion: lion,
    tiger: tiger,
    tiger2: tiger2,
    leopard: leopard,
    horse: horse,
    racehorse: racehorse,
    unicorn: unicorn,
    zebra: zebra,
    deer: deer,
    bison: bison,
    cow: cow,
    ox: ox,
    water_buffalo: water_buffalo,
    cow2: cow2,
    pig: pig,
    pig2: pig2,
    boar: boar,
    pig_nose: pig_nose,
    ram: ram,
    sheep: sheep,
    goat: goat,
    dromedary_camel: dromedary_camel,
    camel: camel,
    llama: llama,
    giraffe: giraffe,
    elephant: elephant,
    mammoth: mammoth,
    rhinoceros: rhinoceros,
    hippopotamus: hippopotamus,
    mouse: mouse,
    mouse2: mouse2,
    rat: rat,
    hamster: hamster,
    rabbit: rabbit,
    rabbit2: rabbit2,
    chipmunk: chipmunk,
    beaver: beaver,
    hedgehog: hedgehog,
    bat: bat,
    bear: bear,
    polar_bear: polar_bear,
    koala: koala,
    panda_face: panda_face,
    sloth: sloth,
    otter: otter,
    skunk: skunk,
    kangaroo: kangaroo,
    badger: badger,
    feet: feet,
    paw_prints: paw_prints,
    turkey: turkey,
    chicken: chicken,
    rooster: rooster,
    hatching_chick: hatching_chick,
    baby_chick: baby_chick,
    hatched_chick: hatched_chick,
    bird: bird,
    penguin: penguin,
    dove: dove,
    eagle: eagle,
    duck: duck,
    swan: swan,
    owl: owl,
    dodo: dodo,
    feather: feather,
    flamingo: flamingo,
    peacock: peacock,
    parrot: parrot,
    frog: frog,
    crocodile: crocodile,
    turtle: turtle,
    lizard: lizard,
    snake: snake,
    dragon_face: dragon_face,
    dragon: dragon,
    sauropod: sauropod,
    whale: whale,
    whale2: whale2,
    dolphin: dolphin,
    flipper: flipper,
    seal: seal,
    fish: fish,
    tropical_fish: tropical_fish,
    blowfish: blowfish,
    shark: shark,
    octopus: octopus,
    shell: shell,
    snail: snail,
    butterfly: butterfly,
    bug: bug,
    ant: ant,
    bee: bee,
    honeybee: honeybee,
    beetle: beetle,
    lady_beetle: lady_beetle,
    cricket: cricket,
    cockroach: cockroach,
    spider: spider,
    spider_web: spider_web,
    scorpion: scorpion,
    mosquito: mosquito,
    fly: fly,
    worm: worm,
    microbe: microbe,
    bouquet: bouquet,
    cherry_blossom: cherry_blossom,
    white_flower: white_flower,
    rosette: rosette,
    rose: rose,
    wilted_flower: wilted_flower,
    hibiscus: hibiscus,
    sunflower: sunflower,
    blossom: blossom,
    tulip: tulip,
    seedling: seedling,
    potted_plant: potted_plant,
    evergreen_tree: evergreen_tree,
    deciduous_tree: deciduous_tree,
    palm_tree: palm_tree,
    cactus: cactus,
    ear_of_rice: ear_of_rice,
    herb: herb,
    shamrock: shamrock,
    four_leaf_clover: four_leaf_clover,
    maple_leaf: maple_leaf,
    fallen_leaf: fallen_leaf,
    leaves: leaves,
    grapes: grapes,
    melon: melon,
    watermelon: watermelon,
    tangerine: tangerine,
    orange: orange,
    mandarin: mandarin,
    lemon: lemon,
    banana: banana,
    pineapple: pineapple,
    mango: mango,
    apple: apple,
    green_apple: green_apple,
    pear: pear,
    peach: peach,
    cherries: cherries,
    strawberry: strawberry,
    blueberries: blueberries,
    kiwi_fruit: kiwi_fruit,
    tomato: tomato,
    olive: olive,
    coconut: coconut,
    avocado: avocado,
    eggplant: eggplant,
    potato: potato,
    carrot: carrot,
    corn: corn,
    hot_pepper: hot_pepper,
    bell_pepper: bell_pepper,
    cucumber: cucumber,
    leafy_green: leafy_green,
    broccoli: broccoli,
    garlic: garlic,
    onion: onion,
    mushroom: mushroom,
    peanuts: peanuts,
    chestnut: chestnut,
    bread: bread,
    croissant: croissant,
    baguette_bread: baguette_bread,
    flatbread: flatbread,
    pretzel: pretzel,
    bagel: bagel,
    pancakes: pancakes,
    waffle: waffle,
    cheese: cheese,
    meat_on_bone: meat_on_bone,
    poultry_leg: poultry_leg,
    cut_of_meat: cut_of_meat,
    bacon: bacon,
    hamburger: hamburger,
    fries: fries,
    pizza: pizza,
    hotdog: hotdog,
    sandwich: sandwich,
    taco: taco,
    burrito: burrito,
    tamale: tamale,
    stuffed_flatbread: stuffed_flatbread,
    falafel: falafel,
    egg: egg,
    fried_egg: fried_egg,
    shallow_pan_of_food: shallow_pan_of_food,
    stew: stew,
    fondue: fondue,
    bowl_with_spoon: bowl_with_spoon,
    green_salad: green_salad,
    popcorn: popcorn,
    butter: butter,
    salt: salt,
    canned_food: canned_food,
    bento: bento,
    rice_cracker: rice_cracker,
    rice_ball: rice_ball,
    rice: rice,
    curry: curry,
    ramen: ramen,
    spaghetti: spaghetti,
    sweet_potato: sweet_potato,
    oden: oden,
    sushi: sushi,
    fried_shrimp: fried_shrimp,
    fish_cake: fish_cake,
    moon_cake: moon_cake,
    dango: dango,
    dumpling: dumpling,
    fortune_cookie: fortune_cookie,
    takeout_box: takeout_box,
    crab: crab,
    lobster: lobster,
    shrimp: shrimp,
    squid: squid,
    oyster: oyster,
    icecream: icecream,
    shaved_ice: shaved_ice,
    ice_cream: ice_cream,
    doughnut: doughnut,
    cookie: cookie,
    birthday: birthday,
    cake: cake,
    cupcake: cupcake,
    pie: pie,
    chocolate_bar: chocolate_bar,
    candy: candy,
    lollipop: lollipop,
    custard: custard,
    honey_pot: honey_pot,
    baby_bottle: baby_bottle,
    milk_glass: milk_glass,
    coffee: coffee,
    teapot: teapot,
    tea: tea,
    sake: sake,
    champagne: champagne,
    wine_glass: wine_glass,
    cocktail: cocktail,
    tropical_drink: tropical_drink,
    beer: beer,
    beers: beers,
    clinking_glasses: clinking_glasses,
    tumbler_glass: tumbler_glass,
    cup_with_straw: cup_with_straw,
    bubble_tea: bubble_tea,
    beverage_box: beverage_box,
    mate: mate,
    ice_cube: ice_cube,
    chopsticks: chopsticks,
    plate_with_cutlery: plate_with_cutlery,
    fork_and_knife: fork_and_knife,
    spoon: spoon,
    hocho: hocho,
    knife: knife,
    amphora: amphora,
    earth_africa: earth_africa,
    earth_americas: earth_americas,
    earth_asia: earth_asia,
    globe_with_meridians: globe_with_meridians,
    world_map: world_map,
    japan: japan,
    compass: compass,
    mountain_snow: mountain_snow,
    mountain: mountain,
    volcano: volcano,
    mount_fuji: mount_fuji,
    camping: camping,
    beach_umbrella: beach_umbrella,
    desert: desert,
    desert_island: desert_island,
    national_park: national_park,
    stadium: stadium,
    classical_building: classical_building,
    building_construction: building_construction,
    bricks: bricks,
    rock: rock,
    wood: wood,
    hut: hut,
    houses: houses,
    derelict_house: derelict_house,
    house: house,
    house_with_garden: house_with_garden,
    office: office,
    post_office: post_office,
    european_post_office: european_post_office,
    hospital: hospital,
    bank: bank,
    hotel: hotel,
    love_hotel: love_hotel,
    convenience_store: convenience_store,
    school: school,
    department_store: department_store,
    factory: factory,
    japanese_castle: japanese_castle,
    european_castle: european_castle,
    wedding: wedding,
    tokyo_tower: tokyo_tower,
    statue_of_liberty: statue_of_liberty,
    church: church,
    mosque: mosque,
    hindu_temple: hindu_temple,
    synagogue: synagogue,
    shinto_shrine: shinto_shrine,
    kaaba: kaaba,
    fountain: fountain,
    tent: tent,
    foggy: foggy,
    night_with_stars: night_with_stars,
    cityscape: cityscape,
    sunrise_over_mountains: sunrise_over_mountains,
    sunrise: sunrise,
    city_sunset: city_sunset,
    city_sunrise: city_sunrise,
    bridge_at_night: bridge_at_night,
    hotsprings: hotsprings,
    carousel_horse: carousel_horse,
    ferris_wheel: ferris_wheel,
    roller_coaster: roller_coaster,
    barber: barber,
    circus_tent: circus_tent,
    steam_locomotive: steam_locomotive,
    railway_car: railway_car,
    bullettrain_side: bullettrain_side,
    bullettrain_front: bullettrain_front,
    train2: train2,
    metro: metro,
    light_rail: light_rail,
    station: station,
    tram: tram,
    monorail: monorail,
    mountain_railway: mountain_railway,
    train: train,
    bus: bus,
    oncoming_bus: oncoming_bus,
    trolleybus: trolleybus,
    minibus: minibus,
    ambulance: ambulance,
    fire_engine: fire_engine,
    police_car: police_car,
    oncoming_police_car: oncoming_police_car,
    taxi: taxi,
    oncoming_taxi: oncoming_taxi,
    car: car,
    red_car: red_car,
    oncoming_automobile: oncoming_automobile,
    blue_car: blue_car,
    pickup_truck: pickup_truck,
    truck: truck,
    articulated_lorry: articulated_lorry,
    tractor: tractor,
    racing_car: racing_car,
    motorcycle: motorcycle,
    motor_scooter: motor_scooter,
    manual_wheelchair: manual_wheelchair,
    motorized_wheelchair: motorized_wheelchair,
    auto_rickshaw: auto_rickshaw,
    bike: bike,
    kick_scooter: kick_scooter,
    skateboard: skateboard,
    roller_skate: roller_skate,
    busstop: busstop,
    motorway: motorway,
    railway_track: railway_track,
    oil_drum: oil_drum,
    fuelpump: fuelpump,
    rotating_light: rotating_light,
    traffic_light: traffic_light,
    vertical_traffic_light: vertical_traffic_light,
    stop_sign: stop_sign,
    construction: construction,
    anchor: anchor,
    boat: boat,
    sailboat: sailboat,
    canoe: canoe,
    speedboat: speedboat,
    passenger_ship: passenger_ship,
    ferry: ferry,
    motor_boat: motor_boat,
    ship: ship,
    airplane: airplane,
    small_airplane: small_airplane,
    flight_departure: flight_departure,
    flight_arrival: flight_arrival,
    parachute: parachute,
    seat: seat,
    helicopter: helicopter,
    suspension_railway: suspension_railway,
    mountain_cableway: mountain_cableway,
    aerial_tramway: aerial_tramway,
    artificial_satellite: artificial_satellite,
    rocket: rocket,
    flying_saucer: flying_saucer,
    bellhop_bell: bellhop_bell,
    luggage: luggage,
    hourglass: hourglass,
    hourglass_flowing_sand: hourglass_flowing_sand,
    watch: watch,
    alarm_clock: alarm_clock,
    stopwatch: stopwatch,
    timer_clock: timer_clock,
    mantelpiece_clock: mantelpiece_clock,
    clock12: clock12,
    clock1230: clock1230,
    clock1: clock1,
    clock130: clock130,
    clock2: clock2,
    clock230: clock230,
    clock3: clock3,
    clock330: clock330,
    clock4: clock4,
    clock430: clock430,
    clock5: clock5,
    clock530: clock530,
    clock6: clock6,
    clock630: clock630,
    clock7: clock7,
    clock730: clock730,
    clock8: clock8,
    clock830: clock830,
    clock9: clock9,
    clock930: clock930,
    clock10: clock10,
    clock1030: clock1030,
    clock11: clock11,
    clock1130: clock1130,
    new_moon: new_moon,
    waxing_crescent_moon: waxing_crescent_moon,
    first_quarter_moon: first_quarter_moon,
    moon: moon,
    waxing_gibbous_moon: waxing_gibbous_moon,
    full_moon: full_moon,
    waning_gibbous_moon: waning_gibbous_moon,
    last_quarter_moon: last_quarter_moon,
    waning_crescent_moon: waning_crescent_moon,
    crescent_moon: crescent_moon,
    new_moon_with_face: new_moon_with_face,
    first_quarter_moon_with_face: first_quarter_moon_with_face,
    last_quarter_moon_with_face: last_quarter_moon_with_face,
    thermometer: thermometer,
    sunny: sunny,
    full_moon_with_face: full_moon_with_face,
    sun_with_face: sun_with_face,
    ringed_planet: ringed_planet,
    star: star$1,
    star2: star2,
    stars: stars,
    milky_way: milky_way,
    cloud: cloud,
    partly_sunny: partly_sunny,
    cloud_with_lightning_and_rain: cloud_with_lightning_and_rain,
    sun_behind_small_cloud: sun_behind_small_cloud,
    sun_behind_large_cloud: sun_behind_large_cloud,
    sun_behind_rain_cloud: sun_behind_rain_cloud,
    cloud_with_rain: cloud_with_rain,
    cloud_with_snow: cloud_with_snow,
    cloud_with_lightning: cloud_with_lightning,
    tornado: tornado,
    fog: fog,
    wind_face: wind_face,
    cyclone: cyclone,
    rainbow: rainbow,
    closed_umbrella: closed_umbrella,
    open_umbrella: open_umbrella,
    umbrella: umbrella,
    parasol_on_ground: parasol_on_ground,
    zap: zap,
    snowflake: snowflake,
    snowman_with_snow: snowman_with_snow,
    snowman: snowman,
    comet: comet,
    fire: fire,
    droplet: droplet,
    ocean: ocean,
    jack_o_lantern: jack_o_lantern,
    christmas_tree: christmas_tree,
    fireworks: fireworks,
    sparkler: sparkler,
    firecracker: firecracker,
    sparkles: sparkles,
    balloon: balloon,
    tada: tada,
    confetti_ball: confetti_ball,
    tanabata_tree: tanabata_tree,
    bamboo: bamboo,
    dolls: dolls,
    flags: flags,
    wind_chime: wind_chime,
    rice_scene: rice_scene,
    red_envelope: red_envelope,
    ribbon: ribbon,
    gift: gift,
    reminder_ribbon: reminder_ribbon,
    tickets: tickets,
    ticket: ticket,
    medal_military: medal_military,
    trophy: trophy,
    medal_sports: medal_sports,
    soccer: soccer,
    baseball: baseball,
    softball: softball,
    basketball: basketball,
    volleyball: volleyball,
    football: football,
    rugby_football: rugby_football,
    tennis: tennis,
    flying_disc: flying_disc,
    bowling: bowling,
    cricket_game: cricket_game,
    field_hockey: field_hockey,
    ice_hockey: ice_hockey,
    lacrosse: lacrosse,
    ping_pong: ping_pong,
    badminton: badminton,
    boxing_glove: boxing_glove,
    martial_arts_uniform: martial_arts_uniform,
    goal_net: goal_net,
    golf: golf,
    ice_skate: ice_skate,
    fishing_pole_and_fish: fishing_pole_and_fish,
    diving_mask: diving_mask,
    running_shirt_with_sash: running_shirt_with_sash,
    ski: ski,
    sled: sled,
    curling_stone: curling_stone,
    dart: dart,
    yo_yo: yo_yo,
    kite: kite,
    crystal_ball: crystal_ball,
    magic_wand: magic_wand,
    nazar_amulet: nazar_amulet,
    video_game: video_game,
    joystick: joystick,
    slot_machine: slot_machine,
    game_die: game_die,
    jigsaw: jigsaw,
    teddy_bear: teddy_bear,
    pinata: pinata,
    nesting_dolls: nesting_dolls,
    spades: spades$1,
    hearts: hearts$1,
    diamonds: diamonds,
    clubs: clubs$1,
    chess_pawn: chess_pawn,
    black_joker: black_joker,
    mahjong: mahjong,
    flower_playing_cards: flower_playing_cards,
    performing_arts: performing_arts,
    framed_picture: framed_picture,
    art: art,
    thread: thread,
    sewing_needle: sewing_needle,
    yarn: yarn,
    knot: knot,
    eyeglasses: eyeglasses,
    dark_sunglasses: dark_sunglasses,
    goggles: goggles,
    lab_coat: lab_coat,
    safety_vest: safety_vest,
    necktie: necktie,
    shirt: shirt,
    tshirt: tshirt,
    jeans: jeans,
    scarf: scarf,
    gloves: gloves,
    coat: coat,
    socks: socks,
    dress: dress,
    kimono: kimono,
    sari: sari,
    one_piece_swimsuit: one_piece_swimsuit,
    swim_brief: swim_brief,
    shorts: shorts,
    bikini: bikini,
    womans_clothes: womans_clothes,
    purse: purse,
    handbag: handbag,
    pouch: pouch,
    shopping: shopping,
    school_satchel: school_satchel,
    thong_sandal: thong_sandal,
    mans_shoe: mans_shoe,
    shoe: shoe,
    athletic_shoe: athletic_shoe,
    hiking_boot: hiking_boot,
    flat_shoe: flat_shoe,
    high_heel: high_heel,
    sandal: sandal,
    ballet_shoes: ballet_shoes,
    boot: boot,
    crown: crown,
    womans_hat: womans_hat,
    tophat: tophat,
    mortar_board: mortar_board,
    billed_cap: billed_cap,
    military_helmet: military_helmet,
    rescue_worker_helmet: rescue_worker_helmet,
    prayer_beads: prayer_beads,
    lipstick: lipstick,
    ring: ring$1,
    gem: gem,
    mute: mute,
    speaker: speaker,
    sound: sound,
    loud_sound: loud_sound,
    loudspeaker: loudspeaker,
    mega: mega,
    postal_horn: postal_horn,
    bell: bell,
    no_bell: no_bell,
    musical_score: musical_score,
    musical_note: musical_note,
    notes: notes,
    studio_microphone: studio_microphone,
    level_slider: level_slider,
    control_knobs: control_knobs,
    microphone: microphone,
    headphones: headphones,
    radio: radio,
    saxophone: saxophone,
    accordion: accordion,
    guitar: guitar,
    musical_keyboard: musical_keyboard,
    trumpet: trumpet,
    violin: violin,
    banjo: banjo,
    drum: drum,
    long_drum: long_drum,
    iphone: iphone,
    calling: calling,
    phone: phone$1,
    telephone: telephone,
    telephone_receiver: telephone_receiver,
    pager: pager,
    fax: fax,
    battery: battery,
    electric_plug: electric_plug,
    computer: computer,
    desktop_computer: desktop_computer,
    printer: printer,
    keyboard: keyboard,
    computer_mouse: computer_mouse,
    trackball: trackball,
    minidisc: minidisc,
    floppy_disk: floppy_disk,
    cd: cd,
    dvd: dvd,
    abacus: abacus,
    movie_camera: movie_camera,
    film_strip: film_strip,
    film_projector: film_projector,
    clapper: clapper,
    tv: tv,
    camera: camera,
    camera_flash: camera_flash,
    video_camera: video_camera,
    vhs: vhs,
    mag: mag,
    mag_right: mag_right,
    candle: candle,
    bulb: bulb,
    flashlight: flashlight,
    izakaya_lantern: izakaya_lantern,
    lantern: lantern,
    diya_lamp: diya_lamp,
    notebook_with_decorative_cover: notebook_with_decorative_cover,
    closed_book: closed_book,
    book: book,
    open_book: open_book,
    green_book: green_book,
    blue_book: blue_book,
    orange_book: orange_book,
    books: books,
    notebook: notebook,
    ledger: ledger,
    page_with_curl: page_with_curl,
    scroll: scroll,
    page_facing_up: page_facing_up,
    newspaper: newspaper,
    newspaper_roll: newspaper_roll,
    bookmark_tabs: bookmark_tabs,
    bookmark: bookmark,
    label: label,
    moneybag: moneybag,
    coin: coin,
    yen: yen$1,
    dollar: dollar$1,
    euro: euro$1,
    pound: pound$1,
    money_with_wings: money_with_wings,
    credit_card: credit_card,
    receipt: receipt,
    chart: chart,
    envelope: envelope,
    email: email,
    incoming_envelope: incoming_envelope,
    envelope_with_arrow: envelope_with_arrow,
    outbox_tray: outbox_tray,
    inbox_tray: inbox_tray,
    mailbox: mailbox,
    mailbox_closed: mailbox_closed,
    mailbox_with_mail: mailbox_with_mail,
    mailbox_with_no_mail: mailbox_with_no_mail,
    postbox: postbox,
    ballot_box: ballot_box,
    pencil2: pencil2,
    black_nib: black_nib,
    fountain_pen: fountain_pen,
    pen: pen,
    paintbrush: paintbrush,
    crayon: crayon,
    memo: memo,
    pencil: pencil,
    briefcase: briefcase,
    file_folder: file_folder,
    open_file_folder: open_file_folder,
    card_index_dividers: card_index_dividers,
    date: date,
    calendar: calendar,
    spiral_notepad: spiral_notepad,
    spiral_calendar: spiral_calendar,
    card_index: card_index,
    chart_with_upwards_trend: chart_with_upwards_trend,
    chart_with_downwards_trend: chart_with_downwards_trend,
    bar_chart: bar_chart,
    clipboard: clipboard$1,
    pushpin: pushpin,
    round_pushpin: round_pushpin,
    paperclip: paperclip,
    paperclips: paperclips,
    straight_ruler: straight_ruler,
    triangular_ruler: triangular_ruler,
    scissors: scissors,
    card_file_box: card_file_box,
    file_cabinet: file_cabinet,
    wastebasket: wastebasket,
    lock: lock,
    unlock: unlock,
    lock_with_ink_pen: lock_with_ink_pen,
    closed_lock_with_key: closed_lock_with_key,
    key: key,
    old_key: old_key,
    hammer: hammer,
    axe: axe,
    pick: pick,
    hammer_and_pick: hammer_and_pick,
    hammer_and_wrench: hammer_and_wrench,
    dagger: dagger$1,
    crossed_swords: crossed_swords,
    gun: gun,
    boomerang: boomerang,
    bow_and_arrow: bow_and_arrow,
    shield: shield,
    carpentry_saw: carpentry_saw,
    wrench: wrench,
    screwdriver: screwdriver,
    nut_and_bolt: nut_and_bolt,
    gear: gear,
    clamp: clamp,
    balance_scale: balance_scale,
    probing_cane: probing_cane,
    link: link$2,
    chains: chains,
    hook: hook,
    toolbox: toolbox,
    magnet: magnet,
    ladder: ladder,
    alembic: alembic,
    test_tube: test_tube,
    petri_dish: petri_dish,
    dna: dna,
    microscope: microscope,
    telescope: telescope,
    satellite: satellite,
    syringe: syringe,
    drop_of_blood: drop_of_blood,
    pill: pill,
    adhesive_bandage: adhesive_bandage,
    stethoscope: stethoscope,
    door: door,
    elevator: elevator,
    mirror: mirror,
    window: window$1,
    bed: bed,
    couch_and_lamp: couch_and_lamp,
    chair: chair,
    toilet: toilet,
    plunger: plunger,
    shower: shower,
    bathtub: bathtub,
    mouse_trap: mouse_trap,
    razor: razor,
    lotion_bottle: lotion_bottle,
    safety_pin: safety_pin,
    broom: broom,
    basket: basket,
    roll_of_paper: roll_of_paper,
    bucket: bucket,
    soap: soap,
    toothbrush: toothbrush,
    sponge: sponge,
    fire_extinguisher: fire_extinguisher,
    shopping_cart: shopping_cart,
    smoking: smoking,
    coffin: coffin,
    headstone: headstone,
    funeral_urn: funeral_urn,
    moyai: moyai,
    placard: placard,
    atm: atm,
    put_litter_in_its_place: put_litter_in_its_place,
    potable_water: potable_water,
    wheelchair: wheelchair,
    mens: mens,
    womens: womens,
    restroom: restroom,
    baby_symbol: baby_symbol,
    wc: wc,
    passport_control: passport_control,
    customs: customs,
    baggage_claim: baggage_claim,
    left_luggage: left_luggage,
    warning: warning,
    children_crossing: children_crossing,
    no_entry: no_entry,
    no_entry_sign: no_entry_sign,
    no_bicycles: no_bicycles,
    no_smoking: no_smoking,
    do_not_litter: do_not_litter,
    no_pedestrians: no_pedestrians,
    no_mobile_phones: no_mobile_phones,
    underage: underage,
    radioactive: radioactive,
    biohazard: biohazard,
    arrow_up: arrow_up,
    arrow_upper_right: arrow_upper_right,
    arrow_right: arrow_right,
    arrow_lower_right: arrow_lower_right,
    arrow_down: arrow_down,
    arrow_lower_left: arrow_lower_left,
    arrow_left: arrow_left,
    arrow_upper_left: arrow_upper_left,
    arrow_up_down: arrow_up_down,
    left_right_arrow: left_right_arrow,
    leftwards_arrow_with_hook: leftwards_arrow_with_hook,
    arrow_right_hook: arrow_right_hook,
    arrow_heading_up: arrow_heading_up,
    arrow_heading_down: arrow_heading_down,
    arrows_clockwise: arrows_clockwise,
    arrows_counterclockwise: arrows_counterclockwise,
    back: back,
    end: end,
    on: on,
    soon: soon,
    top: top$1,
    place_of_worship: place_of_worship,
    atom_symbol: atom_symbol,
    om: om,
    star_of_david: star_of_david,
    wheel_of_dharma: wheel_of_dharma,
    yin_yang: yin_yang,
    latin_cross: latin_cross,
    orthodox_cross: orthodox_cross,
    star_and_crescent: star_and_crescent,
    peace_symbol: peace_symbol,
    menorah: menorah,
    six_pointed_star: six_pointed_star,
    aries: aries,
    taurus: taurus,
    gemini: gemini,
    cancer: cancer,
    leo: leo,
    virgo: virgo,
    libra: libra,
    scorpius: scorpius,
    sagittarius: sagittarius,
    capricorn: capricorn,
    aquarius: aquarius,
    pisces: pisces,
    ophiuchus: ophiuchus,
    twisted_rightwards_arrows: twisted_rightwards_arrows,
    repeat: repeat,
    repeat_one: repeat_one,
    arrow_forward: arrow_forward,
    fast_forward: fast_forward,
    next_track_button: next_track_button,
    play_or_pause_button: play_or_pause_button,
    arrow_backward: arrow_backward,
    rewind: rewind,
    previous_track_button: previous_track_button,
    arrow_up_small: arrow_up_small,
    arrow_double_up: arrow_double_up,
    arrow_down_small: arrow_down_small,
    arrow_double_down: arrow_double_down,
    pause_button: pause_button,
    stop_button: stop_button,
    record_button: record_button,
    eject_button: eject_button,
    cinema: cinema,
    low_brightness: low_brightness,
    high_brightness: high_brightness,
    signal_strength: signal_strength,
    vibration_mode: vibration_mode,
    mobile_phone_off: mobile_phone_off,
    female_sign: female_sign,
    male_sign: male_sign,
    transgender_symbol: transgender_symbol,
    heavy_multiplication_x: heavy_multiplication_x,
    heavy_plus_sign: heavy_plus_sign,
    heavy_minus_sign: heavy_minus_sign,
    heavy_division_sign: heavy_division_sign,
    infinity: infinity,
    bangbang: bangbang,
    interrobang: interrobang,
    question: question,
    grey_question: grey_question,
    grey_exclamation: grey_exclamation,
    exclamation: exclamation,
    heavy_exclamation_mark: heavy_exclamation_mark,
    wavy_dash: wavy_dash,
    currency_exchange: currency_exchange,
    heavy_dollar_sign: heavy_dollar_sign,
    medical_symbol: medical_symbol,
    recycle: recycle,
    fleur_de_lis: fleur_de_lis,
    trident: trident,
    name_badge: name_badge,
    beginner: beginner,
    o: o,
    white_check_mark: white_check_mark,
    ballot_box_with_check: ballot_box_with_check,
    heavy_check_mark: heavy_check_mark,
    x: x,
    negative_squared_cross_mark: negative_squared_cross_mark,
    curly_loop: curly_loop,
    loop: loop,
    part_alternation_mark: part_alternation_mark,
    eight_spoked_asterisk: eight_spoked_asterisk,
    eight_pointed_black_star: eight_pointed_black_star,
    sparkle: sparkle,
    copyright: copyright,
    registered: registered,
    tm: tm,
    hash: hash,
    asterisk: asterisk,
    zero: zero$1,
    one: one,
    two: two,
    three: three,
    four: four,
    five: five,
    six: six,
    seven: seven,
    eight: eight,
    nine: nine,
    keycap_ten: keycap_ten,
    capital_abcd: capital_abcd,
    abcd: abcd,
    symbols: symbols,
    abc: abc,
    a: a,
    ab: ab,
    b: b,
    cl: cl,
    cool: cool,
    free: free,
    information_source: information_source,
    id: id,
    m: m,
    ng: ng,
    o2: o2,
    ok: ok,
    parking: parking,
    sos: sos,
    up: up,
    vs: vs,
    koko: koko,
    sa: sa,
    ideograph_advantage: ideograph_advantage,
    accept: accept,
    congratulations: congratulations,
    secret: secret,
    u6e80: u6e80,
    red_circle: red_circle,
    orange_circle: orange_circle,
    yellow_circle: yellow_circle,
    green_circle: green_circle,
    large_blue_circle: large_blue_circle,
    purple_circle: purple_circle,
    brown_circle: brown_circle,
    black_circle: black_circle,
    white_circle: white_circle,
    red_square: red_square,
    orange_square: orange_square,
    yellow_square: yellow_square,
    green_square: green_square,
    blue_square: blue_square,
    purple_square: purple_square,
    brown_square: brown_square,
    black_large_square: black_large_square,
    white_large_square: white_large_square,
    black_medium_square: black_medium_square,
    white_medium_square: white_medium_square,
    black_medium_small_square: black_medium_small_square,
    white_medium_small_square: white_medium_small_square,
    black_small_square: black_small_square,
    white_small_square: white_small_square,
    large_orange_diamond: large_orange_diamond,
    large_blue_diamond: large_blue_diamond,
    small_orange_diamond: small_orange_diamond,
    small_blue_diamond: small_blue_diamond,
    small_red_triangle: small_red_triangle,
    small_red_triangle_down: small_red_triangle_down,
    diamond_shape_with_a_dot_inside: diamond_shape_with_a_dot_inside,
    radio_button: radio_button,
    white_square_button: white_square_button,
    black_square_button: black_square_button,
    checkered_flag: checkered_flag,
    triangular_flag_on_post: triangular_flag_on_post,
    crossed_flags: crossed_flags,
    black_flag: black_flag,
    white_flag: white_flag,
    rainbow_flag: rainbow_flag,
    transgender_flag: transgender_flag,
    pirate_flag: pirate_flag,
    ascension_island: ascension_island,
    andorra: andorra,
    united_arab_emirates: united_arab_emirates,
    afghanistan: afghanistan,
    antigua_barbuda: antigua_barbuda,
    anguilla: anguilla,
    albania: albania,
    armenia: armenia,
    angola: angola,
    antarctica: antarctica,
    argentina: argentina,
    american_samoa: american_samoa,
    austria: austria,
    australia: australia,
    aruba: aruba,
    aland_islands: aland_islands,
    azerbaijan: azerbaijan,
    bosnia_herzegovina: bosnia_herzegovina,
    barbados: barbados,
    bangladesh: bangladesh,
    belgium: belgium,
    burkina_faso: burkina_faso,
    bulgaria: bulgaria,
    bahrain: bahrain,
    burundi: burundi,
    benin: benin,
    st_barthelemy: st_barthelemy,
    bermuda: bermuda,
    brunei: brunei,
    bolivia: bolivia,
    caribbean_netherlands: caribbean_netherlands,
    brazil: brazil,
    bahamas: bahamas,
    bhutan: bhutan,
    bouvet_island: bouvet_island,
    botswana: botswana,
    belarus: belarus,
    belize: belize,
    canada: canada,
    cocos_islands: cocos_islands,
    congo_kinshasa: congo_kinshasa,
    central_african_republic: central_african_republic,
    congo_brazzaville: congo_brazzaville,
    switzerland: switzerland,
    cote_divoire: cote_divoire,
    cook_islands: cook_islands,
    chile: chile,
    cameroon: cameroon,
    cn: cn,
    colombia: colombia,
    clipperton_island: clipperton_island,
    costa_rica: costa_rica,
    cuba: cuba,
    cape_verde: cape_verde,
    curacao: curacao,
    christmas_island: christmas_island,
    cyprus: cyprus,
    czech_republic: czech_republic,
    de: de,
    diego_garcia: diego_garcia,
    djibouti: djibouti,
    denmark: denmark,
    dominica: dominica,
    dominican_republic: dominican_republic,
    algeria: algeria,
    ceuta_melilla: ceuta_melilla,
    ecuador: ecuador,
    estonia: estonia,
    egypt: egypt,
    western_sahara: western_sahara,
    eritrea: eritrea,
    es: es,
    ethiopia: ethiopia,
    eu: eu,
    european_union: european_union,
    finland: finland,
    fiji: fiji,
    falkland_islands: falkland_islands,
    micronesia: micronesia,
    faroe_islands: faroe_islands,
    fr: fr,
    gabon: gabon,
    gb: gb,
    uk: uk,
    grenada: grenada,
    georgia: georgia,
    french_guiana: french_guiana,
    guernsey: guernsey,
    ghana: ghana,
    gibraltar: gibraltar,
    greenland: greenland,
    gambia: gambia,
    guinea: guinea,
    guadeloupe: guadeloupe,
    equatorial_guinea: equatorial_guinea,
    greece: greece,
    south_georgia_south_sandwich_islands: south_georgia_south_sandwich_islands,
    guatemala: guatemala,
    guam: guam,
    guinea_bissau: guinea_bissau,
    guyana: guyana,
    hong_kong: hong_kong,
    heard_mcdonald_islands: heard_mcdonald_islands,
    honduras: honduras,
    croatia: croatia,
    haiti: haiti,
    hungary: hungary,
    canary_islands: canary_islands,
    indonesia: indonesia,
    ireland: ireland,
    israel: israel,
    isle_of_man: isle_of_man,
    india: india,
    british_indian_ocean_territory: british_indian_ocean_territory,
    iraq: iraq,
    iran: iran,
    iceland: iceland,
    it: it$1,
    jersey: jersey,
    jamaica: jamaica,
    jordan: jordan,
    jp: jp,
    kenya: kenya,
    kyrgyzstan: kyrgyzstan,
    cambodia: cambodia,
    kiribati: kiribati,
    comoros: comoros,
    st_kitts_nevis: st_kitts_nevis,
    north_korea: north_korea,
    kr: kr,
    kuwait: kuwait,
    cayman_islands: cayman_islands,
    kazakhstan: kazakhstan,
    laos: laos,
    lebanon: lebanon,
    st_lucia: st_lucia,
    liechtenstein: liechtenstein,
    sri_lanka: sri_lanka,
    liberia: liberia,
    lesotho: lesotho,
    lithuania: lithuania,
    luxembourg: luxembourg,
    latvia: latvia,
    libya: libya,
    morocco: morocco,
    monaco: monaco,
    moldova: moldova,
    montenegro: montenegro,
    st_martin: st_martin,
    madagascar: madagascar,
    marshall_islands: marshall_islands,
    macedonia: macedonia,
    mali: mali,
    myanmar: myanmar,
    mongolia: mongolia,
    macau: macau,
    northern_mariana_islands: northern_mariana_islands,
    martinique: martinique,
    mauritania: mauritania,
    montserrat: montserrat,
    malta: malta,
    mauritius: mauritius,
    maldives: maldives,
    malawi: malawi,
    mexico: mexico,
    malaysia: malaysia,
    mozambique: mozambique,
    namibia: namibia,
    new_caledonia: new_caledonia,
    niger: niger,
    norfolk_island: norfolk_island,
    nigeria: nigeria,
    nicaragua: nicaragua,
    netherlands: netherlands,
    norway: norway,
    nepal: nepal,
    nauru: nauru,
    niue: niue,
    new_zealand: new_zealand,
    oman: oman,
    panama: panama,
    peru: peru,
    french_polynesia: french_polynesia,
    papua_new_guinea: papua_new_guinea,
    philippines: philippines,
    pakistan: pakistan,
    poland: poland,
    st_pierre_miquelon: st_pierre_miquelon,
    pitcairn_islands: pitcairn_islands,
    puerto_rico: puerto_rico,
    palestinian_territories: palestinian_territories,
    portugal: portugal,
    palau: palau,
    paraguay: paraguay,
    qatar: qatar,
    reunion: reunion,
    romania: romania,
    serbia: serbia,
    ru: ru,
    rwanda: rwanda,
    saudi_arabia: saudi_arabia,
    solomon_islands: solomon_islands,
    seychelles: seychelles,
    sudan: sudan,
    sweden: sweden,
    singapore: singapore,
    st_helena: st_helena,
    slovenia: slovenia,
    svalbard_jan_mayen: svalbard_jan_mayen,
    slovakia: slovakia,
    sierra_leone: sierra_leone,
    san_marino: san_marino,
    senegal: senegal,
    somalia: somalia,
    suriname: suriname,
    south_sudan: south_sudan,
    sao_tome_principe: sao_tome_principe,
    el_salvador: el_salvador,
    sint_maarten: sint_maarten,
    syria: syria,
    swaziland: swaziland,
    tristan_da_cunha: tristan_da_cunha,
    turks_caicos_islands: turks_caicos_islands,
    chad: chad,
    french_southern_territories: french_southern_territories,
    togo: togo,
    thailand: thailand,
    tajikistan: tajikistan,
    tokelau: tokelau,
    timor_leste: timor_leste,
    turkmenistan: turkmenistan,
    tunisia: tunisia,
    tonga: tonga,
    tr: tr,
    trinidad_tobago: trinidad_tobago,
    tuvalu: tuvalu,
    taiwan: taiwan,
    tanzania: tanzania,
    ukraine: ukraine,
    uganda: uganda,
    us_outlying_islands: us_outlying_islands,
    united_nations: united_nations,
    us: us,
    uruguay: uruguay,
    uzbekistan: uzbekistan,
    vatican_city: vatican_city,
    st_vincent_grenadines: st_vincent_grenadines,
    venezuela: venezuela,
    british_virgin_islands: british_virgin_islands,
    us_virgin_islands: us_virgin_islands,
    vietnam: vietnam,
    vanuatu: vanuatu,
    wallis_futuna: wallis_futuna,
    samoa: samoa,
    kosovo: kosovo,
    yemen: yemen,
    mayotte: mayotte,
    south_africa: south_africa,
    zambia: zambia,
    zimbabwe: zimbabwe,
    england: england,
    scotland: scotland,
    wales: wales,
    'default': full
  });

  var render = function emoji_html(tokens, idx /*, options, env */) {
    return tokens[idx].content;
  };

  // Emojies & shortcuts replacement logic.


  var replace$1 = function create_rule(md, emojies, shortcuts, scanRE, replaceRE) {
    var arrayReplaceAt = md.utils.arrayReplaceAt,
        ucm = md.utils.lib.ucmicro,
        ZPCc = new RegExp([ ucm.Z.source, ucm.P.source, ucm.Cc.source ].join('|'));

    function splitTextToken(text, level, Token) {
      var token, last_pos = 0, nodes = [];

      text.replace(replaceRE, function (match, offset, src) {
        var emoji_name;
        // Validate emoji name
        if (shortcuts.hasOwnProperty(match)) {
          // replace shortcut with full name
          emoji_name = shortcuts[match];

          // Don't allow letters before any shortcut (as in no ":/" in http://)
          if (offset > 0 && !ZPCc.test(src[offset - 1])) {
            return;
          }

          // Don't allow letters after any shortcut
          if (offset + match.length < src.length && !ZPCc.test(src[offset + match.length])) {
            return;
          }
        } else {
          emoji_name = match.slice(1, -1);
        }

        // Add new tokens to pending list
        if (offset > last_pos) {
          token         = new Token('text', '', 0);
          token.content = text.slice(last_pos, offset);
          nodes.push(token);
        }

        token         = new Token('emoji', '', 0);
        token.markup  = emoji_name;
        token.content = emojies[emoji_name];
        nodes.push(token);

        last_pos = offset + match.length;
      });

      if (last_pos < text.length) {
        token         = new Token('text', '', 0);
        token.content = text.slice(last_pos);
        nodes.push(token);
      }

      return nodes;
    }

    return function emoji_replace(state) {
      var i, j, l, tokens, token,
          blockTokens = state.tokens,
          autolinkLevel = 0;

      for (j = 0, l = blockTokens.length; j < l; j++) {
        if (blockTokens[j].type !== 'inline') { continue; }
        tokens = blockTokens[j].children;

        // We scan from the end, to keep position when new tags added.
        // Use reversed logic in links start/end match
        for (i = tokens.length - 1; i >= 0; i--) {
          token = tokens[i];

          if (token.type === 'link_open' || token.type === 'link_close') {
            if (token.info === 'auto') { autolinkLevel -= token.nesting; }
          }

          if (token.type === 'text' && autolinkLevel === 0 && scanRE.test(token.content)) {
            // replace current node
            blockTokens[j].children = tokens = arrayReplaceAt(
              tokens, i, splitTextToken(token.content, token.level, state.Token)
            );
          }
        }
      }
    };
  };

  // Convert input options to more useable format


  function quoteRE$1(str) {
    return str.replace(/[.?*+^$[\]\\(){}|-]/g, '\\$&');
  }


  var normalize_opts = function normalize_opts(options) {
    var emojies = options.defs,
        shortcuts;

    // Filter emojies by whitelist, if needed
    if (options.enabled.length) {
      emojies = Object.keys(emojies).reduce(function (acc, key) {
        if (options.enabled.indexOf(key) >= 0) {
          acc[key] = emojies[key];
        }
        return acc;
      }, {});
    }

    // Flatten shortcuts to simple object: { alias: emoji_name }
    shortcuts = Object.keys(options.shortcuts).reduce(function (acc, key) {
      // Skip aliases for filtered emojies, to reduce regexp
      if (!emojies[key]) { return acc; }

      if (Array.isArray(options.shortcuts[key])) {
        options.shortcuts[key].forEach(function (alias) {
          acc[alias] = key;
        });
        return acc;
      }

      acc[options.shortcuts[key]] = key;
      return acc;
    }, {});

    var keys = Object.keys(emojies),
        names;

    // If no definitions are given, return empty regex to avoid replacements with 'undefined'.
    if (keys.length === 0) {
      names = '^$';
    } else {
      // Compile regexp
      names = keys
        .map(function (name) { return ':' + name + ':'; })
        .concat(Object.keys(shortcuts))
        .sort()
        .reverse()
        .map(function (name) { return quoteRE$1(name); })
        .join('|');
    }
    var scanRE = RegExp(names);
    var replaceRE = RegExp(names, 'g');

    return {
      defs: emojies,
      shortcuts: shortcuts,
      scanRE: scanRE,
      replaceRE: replaceRE
    };
  };

  var bare = function emoji_plugin(md, options) {
    var defaults = {
      defs: {},
      shortcuts: {},
      enabled: []
    };

    var opts = normalize_opts(md.utils.assign({}, defaults, options || {}));

    md.renderer.rules.emoji = render;

    md.core.ruler.push('emoji', replace$1(md, opts.defs, opts.shortcuts, opts.scanRE, opts.replaceRE));
  };

  var emojies_defs = getCjsExportFromNamespace(full$1);

  var markdownItEmoji = function emoji_plugin(md, options) {
    var defaults = {
      defs: emojies_defs,
      shortcuts: shortcuts$1,
      enabled: []
    };

    var opts = md.utils.assign({}, defaults, options || {});

    bare(md, opts);
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2018 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var EmojiQueryState = function EmojiQueryState(state, options) {
      var this$1$1 = this;

      this.state = state;
      this.provider = options.provider;
      this.provider.event.on('closed', function () {
          if(this$1$1.active) {
              var ref = this$1$1.state.schema.marks;
              var emojiQuery = ref.emojiQuery;
              this$1$1.view.dispatch(this$1$1.state.tr.removeMark(0, this$1$1.state.doc.nodeSize -2, emojiQuery));
          }
      }).on('focus', function () {
          this$1$1.view.focus();
      });
      this.reset();
  };

  EmojiQueryState.prototype.findQueryNode = function findQueryNode () {
      return $(this.view.dom).find('[data-emoji-query]');
  };

  EmojiQueryState.prototype.update = function update (state, view) {
      this.view = view;
      this.state = state;
      var ref = state.schema.marks;
          var emojiQuery = ref.emojiQuery;
      var doc = state.doc;
          var selection = state.selection;
      var $from = selection.$from;
          var from = selection.from;
          var to = selection.to;

      this.active = doc.rangeHasMark(from - 1, to, emojiQuery);

      if (!this.active) {
          return this.reset();
      }

      var $query = this.findQueryNode();
      var $pos = doc.resolve(from - 1);

      this.queryMark = {
          start: $pos.path[$pos.path.length - 1],
          end: to
      };

      var nodeBefore = $from.nodeBefore;

      if(!nodeBefore.text.length || nodeBefore.text.length > 1) {
          this.provider.reset();
          return;
      }

      var query = nodeBefore.text.substr(1);

      if(query != this.query) {
          this.query = query;
          this.provider.query(this, $query[0]);
      }
  };

  EmojiQueryState.prototype.reset = function reset () {
      this.active = false;
      this.query = null;
      if(this.view) {
          this.provider.reset();
      }
  };

  EmojiQueryState.prototype.addEmoji = function addEmoji (item) {
      var ref = this.state.schema.nodes;
          var emoji = ref.emoji;
      var ref$1 = this.state.schema.marks;
          var emojiQuery = ref$1.emojiQuery;

      var nodes = [emoji.create({
          'data-name': String(item.name),
          alt: item.alt,
          src: item.src
      }, null)];


      var tr = this.state.tr
          .removeMark(0, this.state.doc.nodeSize -2, emojiQuery)
          .replaceWith(this.queryMark.start, this.queryMark.end, nodes);

      if(isChromeWithSelectionBug) {
          document.getSelection().empty();
      }

      this.view.dispatch(tr);
      this.view.focus();
  };

  var SimpleEmojiState = function SimpleEmojiState(provider) {
      var this$1$1 = this;

      this.provider = provider;
      this.provider.event.on('focus', function () {
          if(this$1$1.view) {
              this$1$1.view.focus();
          }
      });
      this.reset();
  };

  SimpleEmojiState.prototype.update = function update (state, view, node) {
      this.view = view;
      this.state = state;
      this.provider.query(this, node, true);
  };

  SimpleEmojiState.prototype.reset = function reset () {
      if(this.view) {
          this.provider.reset();
          this.view.focus();
      }
  };

  SimpleEmojiState.prototype.addEmoji = function addEmoji (item) {
      var ref = this.state.schema.nodes;
          var emoji = ref.emoji;

      var node = emoji.create({
          'data-name': String(item.name),
          alt: item.alt,
          src: item.src
      }, null);

      var tr = this.state.tr.replaceSelectionWith(node);

      this.view.dispatch(tr);
      this.view.focus();
      this.reset();
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var userFlag = undefined;
  var findUserFlag = function() {
      if(userFlag) {
          return userFlag;
      }

      var directMapping = {
          'en-us': 'us',
          'en': 'us',
          'en-gb': 'uk',
          'pt-br': 'portugal',
          'fa-ir': 'iran',
          'zh-cn': 'cn',
          'zh-tw': 'cn',
          'ja': 'jp',
          'ko': 'kr',
          'ar': 'united_arab_emirates',
          'uk': 'ukraine',
          'ru' : 'ru',
          'vi': 'vietnam',
          'sv': 'sweden',
          'nb-no': 'norway',
          'it' : 'it',
          'fr': 'fr',
          'es': 'es',
          'de': 'de',
          'da': 'denmark',
          'cs': 'czech_republic',
          'ca': 'es', // sorry for that ;)
          'an': 'es'
      };

      var result = '\uD83C\uDDE9\uD83C\uDDEA';

      try {
          var language = humhub.require('user').getLocale().toLowerCase();

          if(directMapping[language]) {
              return getCharByName(directMapping[language]);
          }

          $.each(getByCategory('flags'), function (index, flag) {
              if(flag && flag.keywords && flag.keywords.indexOf(language) >= 0) {
                  result = flag.char;
                  return false;
              }
          });
      } catch(e) {
          console.error('Error while determining user flag in emoji chooser ');
          console.error(e);
      }

      return userFlag = result;
  };

  var chooser = undefined;

  var EmojiChooser = function EmojiChooser(provider) {
      this.provider = provider;
      this.categoryOrder = ['people', 'animals_and_nature', 'food_and_drink', 'activity', 'travel_and_places', 'objects', 'symbols', 'flags', 'search'];
      this.categories = {
          people: {$icon: getCharToDom('\uD83D\uDE00')},
          animals_and_nature: {$icon: getCharToDom('\uD83D\uDC3B')},
          food_and_drink: {$icon: getCharToDom('\uD83C\uDF82')},
          activity: {$icon: getCharToDom('\u26BD')},
          travel_and_places: {$icon: getCharToDom('\u2708\uFE0F')},
          objects: {$icon: getCharToDom('\uD83D\uDDA5')},
          symbols: {$icon: getCharToDom('\u2764\uFE0F')},
          flags: {$icon: getCharToDom(findUserFlag())},
          search: {$icon: getCharToDom('\uD83D\uDD0D')}
      };
  };

  EmojiChooser.prototype.update = function update (provider, focus) {
      this.provider = provider;
      var position = provider.$node.offset();

      if(!this.$) {
          this.initDom();
          this.initCategory(this.categoryOrder[0]);
      }

      if(!humhub.require('ui.view').isSmall()) {
          this.$.css({
              top: position.top + provider.$node.outerHeight() - 5,
              left: position.left,
          }).show();
      } else {
          this.$.css({
              top: 5,
              position: 'fixed',
              left: 0,
              right: 0,
              margin: 'auto'
          }).show();
      }


      if(focus) {
          this.$.find('.humhub-emoji-chooser-search').focus();
      }
  };

  EmojiChooser.prototype.initDom = function initDom () {
      var that = this;
      this.$ = $('<div class="atwho-view humhub-richtext-provider humhub-emoji-chooser"><div><input type="text" class="form-control humhub-emoji-chooser-search"></div></div>')
          .hide().appendTo($('body'))
          .on('hidden', function () {
              if(that.provider) {
                  that.provider.reset();
              }
          });

      this.$.find('.humhub-emoji-chooser-search').on('keydown', function(e) {
          switch (e.which) {
              case 9:
                  e.preventDefault();
                  that.nextCategory();
                  break;
              case 27:
                  that.provider.reset();
                  break;
              case 13:
                  e.preventDefault();
                  that.provider.select();
                  break;
              case 37:
                  that.prev();
                  break;
              case 38:
                  that.up();
                  break;
              case 39:
                  that.next();
                  break;
              case 40:
                  that.down();
                  break;
          }
      }).on('keyup', function(e) {
          var keyCode = e.keyCode || e.which;
          // This line should prevent processing in case user presses down/up on desktop, android chrome does not send
          // always send keyCode 229 so we can skip this check in this case
          if (keyCode !== 229 && keyCode !== 8 && !/[a-z0-9\d]/i.test(String.fromCharCode(keyCode))) {
              return;
          }

          var val = $(this).val();
          if(!val.length && that.lastActiveCategory) {
              that.openCategory(that.lastActiveCategory);
              return;
          }

          var currentlyActive = that.getActiveCategoryMenuItem().attr('data-emoji-nav-item');
          if(currentlyActive !== 'search') {
              that.lastActiveCategory = currentlyActive;
          }

          that.updateSearch(val);
      });

      this.initNav();
  };

  EmojiChooser.prototype.initNav = function initNav () {
          var this$1$1 = this;

      var $nav = $('<div class="emoji-nav">').appendTo(this.$);

      this.categoryOrder.forEach(function (categoryName, index) {
          var categoryDef = this$1$1.categories[categoryName];
          var $item = $('<span class="emoji-nav-item" title="'+this$1$1.translate(categoryName)+'">').attr('data-emoji-nav-item', categoryName).append(categoryDef.$icon).on('click', function () {
              this$1$1.openCategory(categoryName);
              this$1$1.provider.event.trigger('focus');
          });

          if(index === 0) {
              $item.addClass('cur');
          }

          $nav.append($item);
      });

      $nav.find('[data-emoji-nav-item="search"]').hide();
  };

  EmojiChooser.prototype.clearSearch = function clearSearch () {
      this.$.find('[data-emoji-nav-item="search"]').hide();
      this.$.find('.humhub-emoji-chooser-search').val('');
  };

  EmojiChooser.prototype.updateSearch = function updateSearch (searchStr) {
      this.$.find('[data-emoji-nav-item="search"]').show();
      var result = [];
      var length = searchStr.length;
      this.categoryOrder.forEach(function (categoryName, index) {
          $.each(getByCategory(categoryName), function (index, emoji) {
              if(emoji && emoji.keywords) {
                  $.each(emoji.keywords, function (index, keyword) {
                      if(length < 3) {
                          if(keyword.lastIndexOf(searchStr, 0) === 0) {
                              result.push(emoji);
                              return false;
                          }
                      } else if(keyword.includes(searchStr)) {
                          result.push(emoji);
                          return false;
                      }
                  });
              }
          });
      });

      this.openCategory('search');
      this.setCategoryItems('search', result);
  };

  EmojiChooser.prototype.openCategory = function openCategory (categoryName) {
      this.categories[categoryName];

      if(!this.$.find('[data-emoji-category="'+categoryName+'"]').length) {
          this.initCategory(categoryName);
      }

      if(categoryName !== 'search') {
          this.clearSearch();
      }

      this.$.find('[data-emoji-nav-item]').removeClass('cur');
      this.$.find('[data-emoji-nav-item="'+categoryName+'"]').addClass('cur');
      this.$.find('[data-emoji-category]').hide();
      this.$.find('[data-emoji-category="'+categoryName+'"]').show();
  };

  EmojiChooser.prototype.initCategory = function initCategory (categoryName) {
      var that = this;
      var $category = $('<div>').attr('data-emoji-category', categoryName).on('click', '.atwho-emoji-entry', function()  {
          that.getSelectionNode().removeClass('cur');
          $(this).addClass('cur');
          that.provider.select();
      }).prependTo(this.$);

      $('<ul class="atwo-view-ul humhub-emoji-chooser-item-list">').appendTo($category);
      this.categories[categoryName].$ = $category;
      this.setCategoryItems(categoryName);
  };

  EmojiChooser.prototype.setCategoryItems = function setCategoryItems (categoryName, items) {
      if(!items && categoryName !== 'search') {
          items = getByCategory(categoryName);
      }

      if(!items) {
          items = [];
      }

      var $list = this.categories[categoryName].$.find('.humhub-emoji-chooser-item-list').empty();

      items.forEach(function (emojiDef) {
          var $li = $('<li class="atwho-emoji-entry">').append(getCharToDom(emojiDef.char, emojiDef.name));

          if(categoryName === 'flags' && emojiDef.char === findUserFlag()) {
              $list.prepend($li);
          } else {
              $list.append($li);
          }
      });

      $list.children().first().addClass('cur');
  };

  EmojiChooser.prototype.reset = function reset () {
      this.provder = undefined;
      this.$.remove();
      this.$ = undefined;
  };

  EmojiChooser.prototype.getSelection = function getSelection () {
      var $selection = this.getSelectionNode().find('img');
      return {
          name: $selection.data('name'),
          alt: $selection.attr('alt'),
          src: $selection.attr('src'),
      }
  };

  EmojiChooser.prototype.translate = function translate (key) {
      return this.provider.context.translate(key);
  };

  EmojiChooser.prototype.getSelectionNode = function getSelectionNode () {
      return this.getActiveCategoryTab().find('.cur');
  };

  EmojiChooser.prototype.getActiveCategoryTab = function getActiveCategoryTab () {
      return this.$.find('[data-emoji-category]:visible');
  };

  EmojiChooser.prototype.getActiveCategoryMenuItem = function getActiveCategoryMenuItem () {
      return this.$.find('[data-emoji-nav-item].cur');
  };

  EmojiChooser.prototype.nextCategory = function nextCategory () {
      var $next = this.getActiveCategoryMenuItem().next('[data-emoji-nav-item]:not([data-emoji-nav-item="search"])');
      if(!$next.length) {
          $next = this.$.find('[data-emoji-nav-item]:first');
      }

      this.openCategory($next.attr('data-emoji-nav-item'));
  };

  EmojiChooser.prototype.prev = function prev () {
      var $cur = this.getSelectionNode();
      var $prev = $cur.prev();
      if ($prev.length) {
          $prev.addClass('cur');
          $cur.removeClass('cur');
          this.alignScroll();
      }
  };

  EmojiChooser.prototype.next = function next () {
      var $cur = this.getSelectionNode();
      var $next = $cur.next();
      if ($next.length) {
          $next.addClass('cur');
          $cur.removeClass('cur');
          this.alignScroll();
      }
  };

  EmojiChooser.prototype.up = function up () {
      var $cur = this.getSelectionNode();
      var curPosition = $cur.position();

      for (var $prev = $cur.prev(); $prev.length; $prev = $prev.prev()) {
          var nextPosition = $prev.position();

          if (nextPosition.top < curPosition.top && nextPosition.left === curPosition.left) {
              $prev.addClass('cur');
              $cur.removeClass('cur');
              this.alignScroll();
              return;
          }
      }
  };

  EmojiChooser.prototype.down = function down () {
      var $cur = this.getSelectionNode();
      var curPosition = $cur.position();

      for (var $next = $cur.next(); $next.length; $next = $next.next()) {
          var nextPosition = $next.position();
          if (nextPosition.top > curPosition.top && nextPosition.left === curPosition.left) {
              $next.addClass('cur');
              $cur.removeClass('cur');
              this.alignScroll();
              return;
          }
      }

      // If we did not find a match the line below is probably the last line.
      var $last = this.getActiveCategoryTab().find('.atwho-emoji-entry:last');
      if($last.position().top !== curPosition.top) {
          $last.addClass('cur');
          $cur.removeClass('cur');
          this.alignScroll();
      }
  };

  EmojiChooser.prototype.alignScroll = function alignScroll () {
      var $cur = this.getSelectionNode();
      var $tab = this.getActiveCategoryTab();
      var scrollTop = $tab.scrollTop();
      var scrollBottom = scrollTop + $tab.height();

      var offsetTop = $cur[0].offsetTop;
      var offsetBottom = offsetTop + $cur.height();

      if(offsetTop > scrollBottom || offsetTop < scrollTop || offsetBottom > scrollBottom || offsetBottom < scrollTop) {
          $tab[0].scrollTop = $cur[0].offsetTop;
      }
  };

  var EmojiProvider = function EmojiProvider(context) {
      this.event = $({});
      this.context = context;
  };

  EmojiProvider.prototype.query = function query (state, node, focus) {
      this.state = state;
      this.$node = $(node);
      this.update(focus);
  };
  EmojiProvider.prototype.reset = function reset (query, node) {
      if (this.$node) {
          this.$node = undefined;
          this.getChooser().reset();
          this.event.trigger('closed');
      }
  };
  EmojiProvider.prototype.next = function next () {
      this.getChooser().next();
  };
  EmojiProvider.prototype.prev = function prev () {
      this.getChooser().prev();
  };
  EmojiProvider.prototype.down = function down () {
      this.getChooser().down();
  };
  EmojiProvider.prototype.up = function up () {
      this.getChooser().up();
  };
  EmojiProvider.prototype.select = function select () {
      this.state.addEmoji(this.getChooser().getSelection());
  };
  EmojiProvider.prototype.update = function update (focus) {
      this.getChooser().update(this, focus);
  };
  EmojiProvider.prototype.getChooser = function getChooser () {
      if(!chooser) {
          chooser = new EmojiChooser(this);
      }

      return chooser;
  };


  function getProvider(context) {
      return (context.options.emoji && context.options.emoji.provider)
          ?  context.options.emoji.provider : new EmojiProvider(context);
  }

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2018 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var pluginKey$2 = new PluginKey('emoji');

  var emojiPlugin = function (context) {
      return new Plugin({
          props: {
              transformPastedText: function (text) {
                  text = twemoji.parse(text, context.getPluginOption('emoji', 'twemoji'));

                  return text.replace(/\<img class="emoji"[^\\\>]* alt=\"([^\"]*)\"[^\\\>]*\/>/g, function(match, char) {
                      return ':'+getNameByChar(char)+':';
                  });
              },
          },
          state: {
              init: function init(config, state) {
                  return new EmojiQueryState(state, {
                      provider: getProvider(context)
                  });
              },
              apply: function apply(tr, prevPluginState, oldState, newState) {
                  return prevPluginState;
              }
          },
          key: pluginKey$2,
          view: function (view) {
              var emojiState = pluginKey$2.getState(view.state);

              return {
                  update: function update(view, prevState) {
                      emojiState.update(view.state, view);
                  },
                  destroy: function destroy() {}
              };
          },
      });
  };

  var NodePos = function NodePos(node, pos, parent) {
      if ( pos === void 0 ) pos = 0;

      this.node = node;
      this.pos = pos;
      this.children = [];
      this.content = new NodePosFragment(this);
  };

  NodePos.prototype.push = function push (childNodePos) {
      if(!this.hasChild(childNodePos.pos)) {
          this.children.push(childNodePos);
      }
  };

  NodePos.prototype.hasChild = function hasChild (pos) {
      for(var i = 0; i < this.children.length; i++) {
          if(this.children[i].pos === pos) {
              return true;
          }
      }
  };

  NodePos.prototype.removeMark = function removeMark (mark) {
      var markInstance = this.getMark(mark);
      var index = this.node.marks.indexOf(markInstance);

      if (index > -1) {
          this.node.marks.splice(index, 1);
      }
  };

  NodePos.prototype.hasMark = function hasMark (mark) {
      return this.getMark(mark) != null;
  };
  NodePos.prototype.getMark = function getMark (mark) {
      var result = null;

      if(mark instanceof MarkType) {
          mark = mark.name;
      }

      this.node.marks.forEach(function (activeMark) {
          if(activeMark.type.name === mark) {
              result = activeMark;
          }
      });

      return result;
  };

  NodePos.prototype.isPlain = function isPlain () {
      return !this.node.marks.length;
  };

  NodePos.prototype.addMarks = function addMarks (marks) {
          var this$1$1 = this;

      if(!marks || !marks.length) {
          return;
      }

      marks.forEach(function (mark) {
          this$1$1.node.marks = mark.addToSet(this$1$1.node.marks);
      });
  };

  NodePos.prototype.nodesBetween = function nodesBetween (from, to, f, pos, level) {
          if ( from === void 0 ) from = 0;
          if ( pos === void 0 ) pos = 0;
          if ( level === void 0 ) level = 1;

      this.content.nodesBetween(from, to, function (childNode, childPos, parent, i, level) {
          f(childNode, childPos , parent, i, level);
      }, pos, this.node, level);
  };
  NodePos.prototype.start = function start () {
      return this.pos;
  };

  NodePos.prototype.end = function end () {
      return this.pos + this.node.nodeSize;
  };

  var NodePosFragment = function NodePosFragment(nodePos) {
      this.nodePos = nodePos;
      this.fragment = nodePos.node.content;
      this.size = this.fragment.size;
      this.content = this.fragment.content;
  };

  NodePosFragment.prototype.nodesBetween = function nodesBetween (from, to, f, nodeStart, parent, level) {
          if ( nodeStart === void 0 ) nodeStart = 0;

      for (var i = 0, pos = 0; pos < to; i++) {
          var child = this.content[i], end = pos + child.nodeSize;
          if (end > from && f(child, nodeStart + pos, parent, i, level) !== false && child.content.size) {
              var start = pos + 1;
              var childNodePos = new NodePos(child, start);

              childNodePos.nodesBetween(Math.max(0, from - start),
                                      Math.min(child.content.size, to - start),
                                      f, nodeStart + start, level + 1);
          }
          pos = end;
      }
  };

  var $node = function (node, pos) {
      if ( pos === void 0 ) pos = 0;

      if (!(this instanceof $node)) {
          return new $node(node,pos);
      }

      this.tree = [];
      this.flat = [];
      this.filters = [];
      this.findFlag = false;

      if (node) {
          this.push(new NodePos(node, pos));
      }
  };

  $node.prototype.push = function (nodePos, parentPos) {
      if(this._hasNodePos(nodePos.pos)) {
          return;
      }

      this.flat.push(nodePos);

      if (parentPos) {
          parentPos.push(nodePos);
      } else {
          this.tree.push(nodePos);
      }
  };

  $node.prototype.find = function (selector) {
      this.filters = [];

      if(!selector) {
          this.findFlag = true;
          return this;
      }

      return this.type(selector, false);
  };

  $node.prototype._hasNodePos = function(pos) {
      for(var i = 0; i < this.flat.length; i++) {
          if(this.flat[i].pos === pos) {
              return true;
          }
      }
  };

  $node.prototype.size = function() {
      return this.flat.length;
  };

  $node.prototype.type = function (selector, includeSelf) {
      var typeFilter = function (node, filter) {
          var result = false;
          if (Array.isArray(filter)) {
              filter.forEach(function (type) {
                  if (typeFilter((type))) {
                      result = true;
                  }
              });
          } else if (filter instanceof NodeType$1) {
              result = node.type === filter;
          } else if (typeof filter === 'string') {
              result = node.type.name === filter;
          }
          return result;
      };

      return this.where(function (node) {
          return typeFilter(node, selector);
      }, includeSelf);
  };

  $node.prototype.between = function (from, to) {
      return this.where(function (node, pos) {
          var $pos = node.resolve(pos);
          return from <= $pos.start && to >= $pos.end;
      });
  };

  $node.prototype.from = function (from) {
      return this.where(function (node, pos) {
          return from <= node.resolve(pos).start;
      });
  };

  $node.prototype.to = function (from, to) {
      return this.where(function (node, pos) {
          return to >= node.resolve(pos).end;
      });
  };

  $node.prototype.mark = function (filterMark, attributes) {
      if (!filterMark) {
          this.where(function (node) {
              return !node.marks.length
          });
      }

      var markFilter = function (node, attributes, filter) {
          var result = false;
          if (Array.isArray(filter)) {
              result = true;
              filter.forEach(function (type) {
                  result = result && markFilter(node, attributes, type);
              });
          } else {
              result = hasMark(node, filter);
          }

          return result;
      };

      return this.where(function (node) {
          return markFilter(node, attributes, filterMark);
      });
  };


  $node.prototype.markup = function (type, attrs, marks) {
      return this.where(function (node) {
          return node.hasMarkup(type, attrs, marks);
      });
  };

  $node.prototype.text = function (search) {
      return this.where(function (node) {
          return node.isText && ((search) ? node.text === search : true)
      })
  };

  $node.prototype.contains = function (search) {
      return this.where(function (node) {
          return node.textContent.indexOf(search) >= 0
      })
  };

  $node.prototype.textBlock = function () {
      return this.where(function (node) {
          return node.isTextblock
      })
  };

  $node.prototype.block = function () {
      return this.where(function (node) {
          return node.isBlock
      })
  };

  $node.prototype.inline = function () {
      return this.where(function (node) {
          return node.isInline
      })
  };

  $node.prototype.leaf = function () {
      return this.where(function (node) {
          return node.isLeaf
      })
  };

  $node.prototype.canAppend = function (node) {
      return this.where(function (node) {
          return node.canAppend(node)
      })
  };

  $node.prototype.sameMarkup = function (node) {
      return this.where(function (node) {
          return node.sameMarkup(node)
      })
  };

  $node.prototype.not = function () {
      this.notFlag = true;
      return this;
  };

  $node.prototype.delete = function (view) {
      var tr = view.state.tr;
      this.tree.reverse().forEach(function (nodePos) {
          tr = tr.delete(nodePos.start(), nodePos.end());
      });
      view.dispatch(tr);
  };

  $node.prototype.get = function (index) {
      return this.tree[index];
  };

  $node.prototype.append = function (node, view) {
      var tr = view.state.tr;
      var doc = view.state.doc;

      this.flat.reverse().forEach(function (nodePos) {
          tr = tr.setSelection(new TextSelection(doc.resolve(nodePos.end()))).replaceSelectionWith(node);
      });

      view.dispatch(tr);
  };

  $node.prototype.replaceWith = function (node, view, dispatch) {
      if ( dispatch === void 0 ) dispatch = true;

      var tr = view.state.tr;
      var doc = view.state.doc;

      this.flat.reverse().forEach(function (nodePos) {
          tr = tr.setSelection(new TextSelection(doc.resolve(nodePos.start()), doc.resolve(nodePos.end()))).replaceSelectionWith(node);
      });

      if(dispatch) {
          view.dispatch(tr);
      }
  };

  $node.prototype.removeMark = function (mark, state) {
      var tr = state.tr;
      var doc = state.doc;
      this.flat.forEach(function (nodePos) {
          nodePos.removeMark(mark);
          tr = tr.setSelection(new TextSelection(doc.resolve(nodePos.start())), doc.resolve(nodePos.end())).replaceSelectionWith(nodePos.node, false);
      });
  };

  $node.prototype.getMark = function(mark) {
      if(!this.flat.length) {
          return;
      }

      return this.flat[0].getMark(mark);
  };

  $node.prototype.where = function (filter, includeSelf) {
      var this$1$1 = this;
      if ( includeSelf === void 0 ) includeSelf = true;

      var addFilter = (this.notFlag)
                      ? function (node, pos, parent, searchRoot) {
                          return !filter(node, pos, parent, searchRoot)
                      }
                      : filter;

      this.filters.push(addFilter);

      var $result = new $node();
      $result.filters = this.filters;

      this.tree.forEach(function (rootNodePos) {

          var branchMatch = [];

          if (!this$1$1.findFlag && includeSelf && checkFilter(this$1$1.filters, rootNodePos.node, rootNodePos.pos)) {
              branchMatch[0] = new NodePos(rootNodePos.node, rootNodePos.pos);
              $result.push(branchMatch[0]);
          }

          var lastLevel = 1;
          var startPos = rootNodePos.node.type.name === 'doc' ? 0 : rootNodePos.pos + 1;

          rootNodePos.nodesBetween(0, rootNodePos.content.size, function (childNode, pos, parent, i, level) {
              // We moved one tree level back or switched to another branch
              if(lastLevel >= level) {
                  branchMatch = clearLevelBranch(branchMatch, level);
              }

              if (checkFilter(this$1$1.filters, childNode, pos, parent)) {
                  var nodePos = new NodePos(childNode, pos);
                  $result.push(nodePos, findBranchMatch(branchMatch, level));
                  branchMatch[level] = nodePos;
              }

              lastLevel = level;
          }, startPos);

      });

      this.notFlag = false;
      this.findFlag = false;
      return $result;
  };

  var clearLevelBranch = function(branchMatches, level) {
      var result = [];
      branchMatches.forEach(function (val, index) {
          result[index] = (index >= level) ? null : branchMatches[index];
      });
      return result;
  };

  var findBranchMatch = function(branchMatches, level) {
      for(var i = level - 1; i >= 0; i--) {
          if(branchMatches[i]) {
              return branchMatches[i];
          }
      }
  };

  var checkFilter = function (filters, node, pos, parent, searchRoot) {
      for (var i = 0; i < filters.length; i++) {
          if (!filters[i](node, pos, parent, searchRoot)) {
              return false;
          }
      }
      return true;
  };

  var hasMark = function (node, markType) {
      var result = false;

      if(!node) {
          return false;
      }



      node.marks.forEach(function (mark) {
          if(markType instanceof Mark$1 && mark.eq(markType)) {
              result = true;
          }else if (markType instanceof MarkType && ((mark.type.name === markType.name) || mark.eq(markType))) {
              result = true;
          } else if (typeof markType === 'string' && mark.type.name === markType) {
              result = true;
          }
      });
      return result;
  };

  function quoteRE(str) {
      return str.replace(/[.?*+^$[\]\\(){}|-]/g, '\\$&');
  }

  // all emoji shortcuts in string seperated by |
  var shortcutStr = Object.keys(shortcuts)
      .sort()
      .reverse()
      .map(function (shortcut) {return quoteRE(shortcut); })
      .join('|');

  var scanRE = new RegExp('(?:^|\\ )('+shortcutStr+')$');

  var emojiAutoCompleteRule = function(schema) {

      return new InputRule(scanRE, function (state, match, start, end) {
          // Only handle match if match is at the end of the match input
          if(match.index !== (match.input.length - match[0].length)) {
              return false;
          }

          // Match e.g. :) => smiley
          var emojiDef = getEmojiDefinitionByShortcut(match[1]);
          if(emojiDef.name && emojiDef.emoji && emojiDef.$dom) {
              var node = state.schema.nodes.emoji.create({
                  'data-name': emojiDef.name,
                  alt: emojiDef.$dom.attr('alt'),
                  src: emojiDef.$dom.attr('src')
              });

              start = start + (match[0].length - match[1].length);

              return state.tr.delete(start, end).replaceSelectionWith(node, false);
          }

          return false;
      })
  };

  var emojiChooser = function(schema) {
      return new InputRule(new RegExp('(^|\\ +)(:$)'), function (state, match, start, end) {
          if(humhub
              && humhub.modules
              && humhub.modules.ui
              && humhub.modules.ui.view
              && humhub.modules.ui.view.isSmall()) {
              return;
          }

          var mark = schema.mark('emojiQuery');
          var emojiText = schema.text(':', [mark]);

          // Prevents an error log when using IME
          if(hasMark(state.selection.$anchor.nodeBefore, mark)) {
              return;
          }

          start = start + (match[0].length -1);

          return state.tr
              .removeMark(0, state.doc.nodeSize -2, mark)
              .setSelection(TextSelection.create(state.doc,  start, end))
              .replaceSelectionWith(emojiText, false);
      })
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2018 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var keymap$1 = function () {
      var result = {};
      result['ArrowLeft'] = function (state, dispatch) {
          var emojiState = pluginKey$2.getState(state);

          if(emojiState.active) {
              emojiState.provider.prev();
              return true;
          }

          return false;
      };

      result['ArrowDown'] = function (state, dispatch) {
          var emojiState = pluginKey$2.getState(state);

          if(emojiState.active) {
              emojiState.provider.down();
              return true;
          }

          return false;
      };

      result['ArrowRight'] = function (state, dispatch) {
          var emojiState  = pluginKey$2.getState(state);

          if(emojiState  && emojiState.active) {
              emojiState.provider.next();
              return true;
          }

          return false;
      };

      result['Tab'] = function (state, dispatch) {
          var emojiState  = pluginKey$2.getState(state);

          if(emojiState  && emojiState.active) {
              emojiState.provider.getChooser().nextCategory();
              return true;
          }

          return false;
      };

      result['ArrowUp'] = function (state, dispatch) {
          var emojiState  = pluginKey$2.getState(state);

          if(emojiState  && emojiState.active) {
              emojiState.provider.up();
              return true;
          }

          return false;
      };

      result['Enter'] = function (state, dispatch) {
          var emojiState  = pluginKey$2.getState(state);

          if(emojiState  && emojiState.active) {
              emojiState.provider.select();
              return true;
          }

          return false;
      };

      result['Escape'] = function (state, dispatch) {
          var emojiState  = pluginKey$2.getState(state);

          if(emojiState  && emojiState.active) {
              emojiState.provider.reset();
              return true;
          }

          return false;
      };

      return result;
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  function insertEmoji(context) {
      return new MenuItem$1({
          title: context.translate("Insert Emoji"),
          icon: icons$1.emoji,
          sortOrder: 350,
          enable: function enable(state) {
              return canInsert(state, context.schema.nodes.image) && canInsertLink(state)
          },
          run: function run(state, _, view, e) {
              if (!$('.humhub-richtext-provider:visible').length) {
                  setTimeout(function () {
                      new SimpleEmojiState(getProvider(context)).update(state, view, e.target);
                  }, 50);
              }
          }
      })
  }

  function menu$d(context) {
      return [
          {
              id: 'insertEmoji',
              node: 'emoji',
              item: insertEmoji(context)
          }
      ]
  }

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var emoji = {
      id: 'emoji',
      schema: schema$e,
      menu: function (context) { return menu$d(context); },
      inputRules: function (schema) {
          return [
              emojiAutoCompleteRule(),
              emojiChooser(schema)
          ]
      },
      keymap: function (context) { return keymap$1()},
      plugins: function (context) {
          return [
              emojiPlugin(context)
          ]
      },
      registerMarkdownIt: function (markdownIt) {
          markdownIt.use(markdownItEmoji, getMarkdownItOpts());
          markdownIt.renderer.rules.emoji = function(token, idx) {
              var emojiToken = token[idx];

              // Not that clean but unfortunately we don't have access to the editor context here...
              var config = humhub.config.get('ui.richtext.prosemirror', 'emoji');
              var twemojiConfig = config.twemoji || {};
              twemojiConfig.attributes = function (icon, variant) {
                  return {
                      'data-name': emojiToken.markup
                  }
              };
              return twemoji.parse(emojiToken.content, twemojiConfig);
          };
      }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var schema$d = {
      nodes: {
          hard_break: {
              sortOrder: 1100,
              inline: true,
              group: "inline",
              selectable: false,
              parseDOM: [{tag: "br"}],
              toDOM: function () {
                  return ["br"]
              },
              parseMarkdown: {hardbreak: {node: "hard_break"}},
              toMarkdown: function (state, node, parent, index) {
                  for (var i = index + 1; i < parent.childCount; i++)
                  { if (parent.child(i).type != node.type) {

                      (state.table) ? state.write('<br>') : state.write("\\\n");
                      return
                  } }
              }
          }
      }
  };

  var Aacute = "Á";
  var aacute = "á";
  var Abreve = "Ă";
  var abreve = "ă";
  var ac = "∾";
  var acd = "∿";
  var acE = "∾̳";
  var Acirc = "Â";
  var acirc = "â";
  var acute = "´";
  var Acy = "А";
  var acy = "а";
  var AElig = "Æ";
  var aelig = "æ";
  var af = "⁡";
  var Afr = "𝔄";
  var afr = "𝔞";
  var Agrave = "À";
  var agrave = "à";
  var alefsym = "ℵ";
  var aleph = "ℵ";
  var Alpha = "Α";
  var alpha = "α";
  var Amacr = "Ā";
  var amacr = "ā";
  var amalg = "⨿";
  var amp = "&";
  var AMP = "&";
  var andand = "⩕";
  var And = "⩓";
  var and = "∧";
  var andd = "⩜";
  var andslope = "⩘";
  var andv = "⩚";
  var ang = "∠";
  var ange = "⦤";
  var angle = "∠";
  var angmsdaa = "⦨";
  var angmsdab = "⦩";
  var angmsdac = "⦪";
  var angmsdad = "⦫";
  var angmsdae = "⦬";
  var angmsdaf = "⦭";
  var angmsdag = "⦮";
  var angmsdah = "⦯";
  var angmsd = "∡";
  var angrt = "∟";
  var angrtvb = "⊾";
  var angrtvbd = "⦝";
  var angsph = "∢";
  var angst = "Å";
  var angzarr = "⍼";
  var Aogon = "Ą";
  var aogon = "ą";
  var Aopf = "𝔸";
  var aopf = "𝕒";
  var apacir = "⩯";
  var ap = "≈";
  var apE = "⩰";
  var ape = "≊";
  var apid = "≋";
  var apos = "'";
  var ApplyFunction = "⁡";
  var approx = "≈";
  var approxeq = "≊";
  var Aring = "Å";
  var aring = "å";
  var Ascr = "𝒜";
  var ascr = "𝒶";
  var Assign = "≔";
  var ast = "*";
  var asymp = "≈";
  var asympeq = "≍";
  var Atilde = "Ã";
  var atilde = "ã";
  var Auml = "Ä";
  var auml = "ä";
  var awconint = "∳";
  var awint = "⨑";
  var backcong = "≌";
  var backepsilon = "϶";
  var backprime = "‵";
  var backsim = "∽";
  var backsimeq = "⋍";
  var Backslash = "∖";
  var Barv = "⫧";
  var barvee = "⊽";
  var barwed = "⌅";
  var Barwed = "⌆";
  var barwedge = "⌅";
  var bbrk = "⎵";
  var bbrktbrk = "⎶";
  var bcong = "≌";
  var Bcy = "Б";
  var bcy = "б";
  var bdquo = "„";
  var becaus = "∵";
  var because = "∵";
  var Because = "∵";
  var bemptyv = "⦰";
  var bepsi = "϶";
  var bernou = "ℬ";
  var Bernoullis = "ℬ";
  var Beta = "Β";
  var beta = "β";
  var beth = "ℶ";
  var between = "≬";
  var Bfr = "𝔅";
  var bfr = "𝔟";
  var bigcap = "⋂";
  var bigcirc = "◯";
  var bigcup = "⋃";
  var bigodot = "⨀";
  var bigoplus = "⨁";
  var bigotimes = "⨂";
  var bigsqcup = "⨆";
  var bigstar = "★";
  var bigtriangledown = "▽";
  var bigtriangleup = "△";
  var biguplus = "⨄";
  var bigvee = "⋁";
  var bigwedge = "⋀";
  var bkarow = "⤍";
  var blacklozenge = "⧫";
  var blacksquare = "▪";
  var blacktriangle = "▴";
  var blacktriangledown = "▾";
  var blacktriangleleft = "◂";
  var blacktriangleright = "▸";
  var blank = "␣";
  var blk12 = "▒";
  var blk14 = "░";
  var blk34 = "▓";
  var block$1 = "█";
  var bne = "=⃥";
  var bnequiv = "≡⃥";
  var bNot = "⫭";
  var bnot = "⌐";
  var Bopf = "𝔹";
  var bopf = "𝕓";
  var bot = "⊥";
  var bottom = "⊥";
  var bowtie = "⋈";
  var boxbox = "⧉";
  var boxdl = "┐";
  var boxdL = "╕";
  var boxDl = "╖";
  var boxDL = "╗";
  var boxdr = "┌";
  var boxdR = "╒";
  var boxDr = "╓";
  var boxDR = "╔";
  var boxh = "─";
  var boxH = "═";
  var boxhd = "┬";
  var boxHd = "╤";
  var boxhD = "╥";
  var boxHD = "╦";
  var boxhu = "┴";
  var boxHu = "╧";
  var boxhU = "╨";
  var boxHU = "╩";
  var boxminus = "⊟";
  var boxplus = "⊞";
  var boxtimes = "⊠";
  var boxul = "┘";
  var boxuL = "╛";
  var boxUl = "╜";
  var boxUL = "╝";
  var boxur = "└";
  var boxuR = "╘";
  var boxUr = "╙";
  var boxUR = "╚";
  var boxv = "│";
  var boxV = "║";
  var boxvh = "┼";
  var boxvH = "╪";
  var boxVh = "╫";
  var boxVH = "╬";
  var boxvl = "┤";
  var boxvL = "╡";
  var boxVl = "╢";
  var boxVL = "╣";
  var boxvr = "├";
  var boxvR = "╞";
  var boxVr = "╟";
  var boxVR = "╠";
  var bprime = "‵";
  var breve = "˘";
  var Breve = "˘";
  var brvbar = "¦";
  var bscr = "𝒷";
  var Bscr = "ℬ";
  var bsemi = "⁏";
  var bsim = "∽";
  var bsime = "⋍";
  var bsolb = "⧅";
  var bsol = "\\";
  var bsolhsub = "⟈";
  var bull = "•";
  var bullet = "•";
  var bump = "≎";
  var bumpE = "⪮";
  var bumpe = "≏";
  var Bumpeq = "≎";
  var bumpeq = "≏";
  var Cacute = "Ć";
  var cacute = "ć";
  var capand = "⩄";
  var capbrcup = "⩉";
  var capcap = "⩋";
  var cap = "∩";
  var Cap = "⋒";
  var capcup = "⩇";
  var capdot = "⩀";
  var CapitalDifferentialD = "ⅅ";
  var caps = "∩︀";
  var caret = "⁁";
  var caron = "ˇ";
  var Cayleys = "ℭ";
  var ccaps = "⩍";
  var Ccaron = "Č";
  var ccaron = "č";
  var Ccedil = "Ç";
  var ccedil = "ç";
  var Ccirc = "Ĉ";
  var ccirc = "ĉ";
  var Cconint = "∰";
  var ccups = "⩌";
  var ccupssm = "⩐";
  var Cdot = "Ċ";
  var cdot = "ċ";
  var cedil = "¸";
  var Cedilla = "¸";
  var cemptyv = "⦲";
  var cent = "¢";
  var centerdot = "·";
  var CenterDot = "·";
  var cfr = "𝔠";
  var Cfr = "ℭ";
  var CHcy = "Ч";
  var chcy = "ч";
  var check = "✓";
  var checkmark = "✓";
  var Chi = "Χ";
  var chi = "χ";
  var circ = "ˆ";
  var circeq = "≗";
  var circlearrowleft = "↺";
  var circlearrowright = "↻";
  var circledast = "⊛";
  var circledcirc = "⊚";
  var circleddash = "⊝";
  var CircleDot = "⊙";
  var circledR = "®";
  var circledS = "Ⓢ";
  var CircleMinus = "⊖";
  var CirclePlus = "⊕";
  var CircleTimes = "⊗";
  var cir = "○";
  var cirE = "⧃";
  var cire = "≗";
  var cirfnint = "⨐";
  var cirmid = "⫯";
  var cirscir = "⧂";
  var ClockwiseContourIntegral = "∲";
  var CloseCurlyDoubleQuote = "”";
  var CloseCurlyQuote = "’";
  var clubs = "♣";
  var clubsuit = "♣";
  var colon = ":";
  var Colon = "∷";
  var Colone = "⩴";
  var colone = "≔";
  var coloneq = "≔";
  var comma = ",";
  var commat = "@";
  var comp = "∁";
  var compfn = "∘";
  var complement = "∁";
  var complexes = "ℂ";
  var cong = "≅";
  var congdot = "⩭";
  var Congruent = "≡";
  var conint = "∮";
  var Conint = "∯";
  var ContourIntegral = "∮";
  var copf = "𝕔";
  var Copf = "ℂ";
  var coprod = "∐";
  var Coproduct = "∐";
  var copy = "©";
  var COPY = "©";
  var copysr = "℗";
  var CounterClockwiseContourIntegral = "∳";
  var crarr = "↵";
  var cross = "✗";
  var Cross = "⨯";
  var Cscr = "𝒞";
  var cscr = "𝒸";
  var csub = "⫏";
  var csube = "⫑";
  var csup = "⫐";
  var csupe = "⫒";
  var ctdot = "⋯";
  var cudarrl = "⤸";
  var cudarrr = "⤵";
  var cuepr = "⋞";
  var cuesc = "⋟";
  var cularr = "↶";
  var cularrp = "⤽";
  var cupbrcap = "⩈";
  var cupcap = "⩆";
  var CupCap = "≍";
  var cup = "∪";
  var Cup = "⋓";
  var cupcup = "⩊";
  var cupdot = "⊍";
  var cupor = "⩅";
  var cups = "∪︀";
  var curarr = "↷";
  var curarrm = "⤼";
  var curlyeqprec = "⋞";
  var curlyeqsucc = "⋟";
  var curlyvee = "⋎";
  var curlywedge = "⋏";
  var curren = "¤";
  var curvearrowleft = "↶";
  var curvearrowright = "↷";
  var cuvee = "⋎";
  var cuwed = "⋏";
  var cwconint = "∲";
  var cwint = "∱";
  var cylcty = "⌭";
  var dagger = "†";
  var Dagger = "‡";
  var daleth = "ℸ";
  var darr = "↓";
  var Darr = "↡";
  var dArr = "⇓";
  var dash = "‐";
  var Dashv = "⫤";
  var dashv = "⊣";
  var dbkarow = "⤏";
  var dblac = "˝";
  var Dcaron = "Ď";
  var dcaron = "ď";
  var Dcy = "Д";
  var dcy = "д";
  var ddagger = "‡";
  var ddarr = "⇊";
  var DD = "ⅅ";
  var dd = "ⅆ";
  var DDotrahd = "⤑";
  var ddotseq = "⩷";
  var deg = "°";
  var Del = "∇";
  var Delta = "Δ";
  var delta = "δ";
  var demptyv = "⦱";
  var dfisht = "⥿";
  var Dfr = "𝔇";
  var dfr = "𝔡";
  var dHar = "⥥";
  var dharl = "⇃";
  var dharr = "⇂";
  var DiacriticalAcute = "´";
  var DiacriticalDot = "˙";
  var DiacriticalDoubleAcute = "˝";
  var DiacriticalGrave = "`";
  var DiacriticalTilde = "˜";
  var diam = "⋄";
  var diamond = "⋄";
  var Diamond = "⋄";
  var diamondsuit = "♦";
  var diams = "♦";
  var die = "¨";
  var DifferentialD = "ⅆ";
  var digamma = "ϝ";
  var disin = "⋲";
  var div = "÷";
  var divide = "÷";
  var divideontimes = "⋇";
  var divonx = "⋇";
  var DJcy = "Ђ";
  var djcy = "ђ";
  var dlcorn = "⌞";
  var dlcrop = "⌍";
  var dollar = "$";
  var Dopf = "𝔻";
  var dopf = "𝕕";
  var Dot = "¨";
  var dot = "˙";
  var DotDot = "⃜";
  var doteq = "≐";
  var doteqdot = "≑";
  var DotEqual = "≐";
  var dotminus = "∸";
  var dotplus = "∔";
  var dotsquare = "⊡";
  var doublebarwedge = "⌆";
  var DoubleContourIntegral = "∯";
  var DoubleDot = "¨";
  var DoubleDownArrow = "⇓";
  var DoubleLeftArrow = "⇐";
  var DoubleLeftRightArrow = "⇔";
  var DoubleLeftTee = "⫤";
  var DoubleLongLeftArrow = "⟸";
  var DoubleLongLeftRightArrow = "⟺";
  var DoubleLongRightArrow = "⟹";
  var DoubleRightArrow = "⇒";
  var DoubleRightTee = "⊨";
  var DoubleUpArrow = "⇑";
  var DoubleUpDownArrow = "⇕";
  var DoubleVerticalBar = "∥";
  var DownArrowBar = "⤓";
  var downarrow = "↓";
  var DownArrow = "↓";
  var Downarrow = "⇓";
  var DownArrowUpArrow = "⇵";
  var DownBreve = "̑";
  var downdownarrows = "⇊";
  var downharpoonleft = "⇃";
  var downharpoonright = "⇂";
  var DownLeftRightVector = "⥐";
  var DownLeftTeeVector = "⥞";
  var DownLeftVectorBar = "⥖";
  var DownLeftVector = "↽";
  var DownRightTeeVector = "⥟";
  var DownRightVectorBar = "⥗";
  var DownRightVector = "⇁";
  var DownTeeArrow = "↧";
  var DownTee = "⊤";
  var drbkarow = "⤐";
  var drcorn = "⌟";
  var drcrop = "⌌";
  var Dscr = "𝒟";
  var dscr = "𝒹";
  var DScy = "Ѕ";
  var dscy = "ѕ";
  var dsol = "⧶";
  var Dstrok = "Đ";
  var dstrok = "đ";
  var dtdot = "⋱";
  var dtri = "▿";
  var dtrif = "▾";
  var duarr = "⇵";
  var duhar = "⥯";
  var dwangle = "⦦";
  var DZcy = "Џ";
  var dzcy = "џ";
  var dzigrarr = "⟿";
  var Eacute = "É";
  var eacute = "é";
  var easter = "⩮";
  var Ecaron = "Ě";
  var ecaron = "ě";
  var Ecirc = "Ê";
  var ecirc = "ê";
  var ecir = "≖";
  var ecolon = "≕";
  var Ecy = "Э";
  var ecy = "э";
  var eDDot = "⩷";
  var Edot = "Ė";
  var edot = "ė";
  var eDot = "≑";
  var ee = "ⅇ";
  var efDot = "≒";
  var Efr = "𝔈";
  var efr = "𝔢";
  var eg = "⪚";
  var Egrave = "È";
  var egrave = "è";
  var egs = "⪖";
  var egsdot = "⪘";
  var el = "⪙";
  var Element = "∈";
  var elinters = "⏧";
  var ell = "ℓ";
  var els = "⪕";
  var elsdot = "⪗";
  var Emacr = "Ē";
  var emacr = "ē";
  var empty = "∅";
  var emptyset = "∅";
  var EmptySmallSquare = "◻";
  var emptyv = "∅";
  var EmptyVerySmallSquare = "▫";
  var emsp13 = " ";
  var emsp14 = " ";
  var emsp = " ";
  var ENG = "Ŋ";
  var eng = "ŋ";
  var ensp = " ";
  var Eogon = "Ę";
  var eogon = "ę";
  var Eopf = "𝔼";
  var eopf = "𝕖";
  var epar = "⋕";
  var eparsl = "⧣";
  var eplus = "⩱";
  var epsi = "ε";
  var Epsilon = "Ε";
  var epsilon = "ε";
  var epsiv = "ϵ";
  var eqcirc = "≖";
  var eqcolon = "≕";
  var eqsim = "≂";
  var eqslantgtr = "⪖";
  var eqslantless = "⪕";
  var Equal = "⩵";
  var equals = "=";
  var EqualTilde = "≂";
  var equest = "≟";
  var Equilibrium = "⇌";
  var equiv = "≡";
  var equivDD = "⩸";
  var eqvparsl = "⧥";
  var erarr = "⥱";
  var erDot = "≓";
  var escr = "ℯ";
  var Escr = "ℰ";
  var esdot = "≐";
  var Esim = "⩳";
  var esim = "≂";
  var Eta = "Η";
  var eta = "η";
  var ETH = "Ð";
  var eth = "ð";
  var Euml = "Ë";
  var euml = "ë";
  var euro = "€";
  var excl = "!";
  var exist = "∃";
  var Exists = "∃";
  var expectation = "ℰ";
  var exponentiale = "ⅇ";
  var ExponentialE = "ⅇ";
  var fallingdotseq = "≒";
  var Fcy = "Ф";
  var fcy = "ф";
  var female = "♀";
  var ffilig = "ﬃ";
  var fflig = "ﬀ";
  var ffllig = "ﬄ";
  var Ffr = "𝔉";
  var ffr = "𝔣";
  var filig = "ﬁ";
  var FilledSmallSquare = "◼";
  var FilledVerySmallSquare = "▪";
  var fjlig = "fj";
  var flat = "♭";
  var fllig = "ﬂ";
  var fltns = "▱";
  var fnof = "ƒ";
  var Fopf = "𝔽";
  var fopf = "𝕗";
  var forall = "∀";
  var ForAll = "∀";
  var fork = "⋔";
  var forkv = "⫙";
  var Fouriertrf = "ℱ";
  var fpartint = "⨍";
  var frac12 = "½";
  var frac13 = "⅓";
  var frac14 = "¼";
  var frac15 = "⅕";
  var frac16 = "⅙";
  var frac18 = "⅛";
  var frac23 = "⅔";
  var frac25 = "⅖";
  var frac34 = "¾";
  var frac35 = "⅗";
  var frac38 = "⅜";
  var frac45 = "⅘";
  var frac56 = "⅚";
  var frac58 = "⅝";
  var frac78 = "⅞";
  var frasl = "⁄";
  var frown = "⌢";
  var fscr = "𝒻";
  var Fscr = "ℱ";
  var gacute = "ǵ";
  var Gamma = "Γ";
  var gamma = "γ";
  var Gammad = "Ϝ";
  var gammad = "ϝ";
  var gap = "⪆";
  var Gbreve = "Ğ";
  var gbreve = "ğ";
  var Gcedil = "Ģ";
  var Gcirc = "Ĝ";
  var gcirc = "ĝ";
  var Gcy = "Г";
  var gcy = "г";
  var Gdot = "Ġ";
  var gdot = "ġ";
  var ge = "≥";
  var gE = "≧";
  var gEl = "⪌";
  var gel = "⋛";
  var geq = "≥";
  var geqq = "≧";
  var geqslant = "⩾";
  var gescc = "⪩";
  var ges = "⩾";
  var gesdot = "⪀";
  var gesdoto = "⪂";
  var gesdotol = "⪄";
  var gesl = "⋛︀";
  var gesles = "⪔";
  var Gfr = "𝔊";
  var gfr = "𝔤";
  var gg = "≫";
  var Gg = "⋙";
  var ggg = "⋙";
  var gimel = "ℷ";
  var GJcy = "Ѓ";
  var gjcy = "ѓ";
  var gla = "⪥";
  var gl = "≷";
  var glE = "⪒";
  var glj = "⪤";
  var gnap = "⪊";
  var gnapprox = "⪊";
  var gne = "⪈";
  var gnE = "≩";
  var gneq = "⪈";
  var gneqq = "≩";
  var gnsim = "⋧";
  var Gopf = "𝔾";
  var gopf = "𝕘";
  var grave = "`";
  var GreaterEqual = "≥";
  var GreaterEqualLess = "⋛";
  var GreaterFullEqual = "≧";
  var GreaterGreater = "⪢";
  var GreaterLess = "≷";
  var GreaterSlantEqual = "⩾";
  var GreaterTilde = "≳";
  var Gscr = "𝒢";
  var gscr = "ℊ";
  var gsim = "≳";
  var gsime = "⪎";
  var gsiml = "⪐";
  var gtcc = "⪧";
  var gtcir = "⩺";
  var gt = ">";
  var GT = ">";
  var Gt = "≫";
  var gtdot = "⋗";
  var gtlPar = "⦕";
  var gtquest = "⩼";
  var gtrapprox = "⪆";
  var gtrarr = "⥸";
  var gtrdot = "⋗";
  var gtreqless = "⋛";
  var gtreqqless = "⪌";
  var gtrless = "≷";
  var gtrsim = "≳";
  var gvertneqq = "≩︀";
  var gvnE = "≩︀";
  var Hacek = "ˇ";
  var hairsp = " ";
  var half = "½";
  var hamilt = "ℋ";
  var HARDcy = "Ъ";
  var hardcy = "ъ";
  var harrcir = "⥈";
  var harr = "↔";
  var hArr = "⇔";
  var harrw = "↭";
  var Hat = "^";
  var hbar = "ℏ";
  var Hcirc = "Ĥ";
  var hcirc = "ĥ";
  var hearts = "♥";
  var heartsuit = "♥";
  var hellip = "…";
  var hercon = "⊹";
  var hfr = "𝔥";
  var Hfr = "ℌ";
  var HilbertSpace = "ℋ";
  var hksearow = "⤥";
  var hkswarow = "⤦";
  var hoarr = "⇿";
  var homtht = "∻";
  var hookleftarrow = "↩";
  var hookrightarrow = "↪";
  var hopf = "𝕙";
  var Hopf = "ℍ";
  var horbar = "―";
  var HorizontalLine = "─";
  var hscr = "𝒽";
  var Hscr = "ℋ";
  var hslash = "ℏ";
  var Hstrok = "Ħ";
  var hstrok = "ħ";
  var HumpDownHump = "≎";
  var HumpEqual = "≏";
  var hybull = "⁃";
  var hyphen = "‐";
  var Iacute = "Í";
  var iacute = "í";
  var ic = "⁣";
  var Icirc = "Î";
  var icirc = "î";
  var Icy = "И";
  var icy = "и";
  var Idot = "İ";
  var IEcy = "Е";
  var iecy = "е";
  var iexcl = "¡";
  var iff = "⇔";
  var ifr = "𝔦";
  var Ifr = "ℑ";
  var Igrave = "Ì";
  var igrave = "ì";
  var ii = "ⅈ";
  var iiiint = "⨌";
  var iiint = "∭";
  var iinfin = "⧜";
  var iiota = "℩";
  var IJlig = "Ĳ";
  var ijlig = "ĳ";
  var Imacr = "Ī";
  var imacr = "ī";
  var image$2 = "ℑ";
  var ImaginaryI = "ⅈ";
  var imagline = "ℐ";
  var imagpart = "ℑ";
  var imath = "ı";
  var Im = "ℑ";
  var imof = "⊷";
  var imped = "Ƶ";
  var Implies = "⇒";
  var incare = "℅";
  var infin = "∞";
  var infintie = "⧝";
  var inodot = "ı";
  var intcal = "⊺";
  var int = "∫";
  var Int = "∬";
  var integers = "ℤ";
  var Integral = "∫";
  var intercal = "⊺";
  var Intersection = "⋂";
  var intlarhk = "⨗";
  var intprod = "⨼";
  var InvisibleComma = "⁣";
  var InvisibleTimes = "⁢";
  var IOcy = "Ё";
  var iocy = "ё";
  var Iogon = "Į";
  var iogon = "į";
  var Iopf = "𝕀";
  var iopf = "𝕚";
  var Iota = "Ι";
  var iota = "ι";
  var iprod = "⨼";
  var iquest = "¿";
  var iscr = "𝒾";
  var Iscr = "ℐ";
  var isin = "∈";
  var isindot = "⋵";
  var isinE = "⋹";
  var isins = "⋴";
  var isinsv = "⋳";
  var isinv = "∈";
  var it = "⁢";
  var Itilde = "Ĩ";
  var itilde = "ĩ";
  var Iukcy = "І";
  var iukcy = "і";
  var Iuml = "Ï";
  var iuml = "ï";
  var Jcirc = "Ĵ";
  var jcirc = "ĵ";
  var Jcy = "Й";
  var jcy = "й";
  var Jfr = "𝔍";
  var jfr = "𝔧";
  var jmath = "ȷ";
  var Jopf = "𝕁";
  var jopf = "𝕛";
  var Jscr = "𝒥";
  var jscr = "𝒿";
  var Jsercy = "Ј";
  var jsercy = "ј";
  var Jukcy = "Є";
  var jukcy = "є";
  var Kappa = "Κ";
  var kappa = "κ";
  var kappav = "ϰ";
  var Kcedil = "Ķ";
  var kcedil = "ķ";
  var Kcy = "К";
  var kcy = "к";
  var Kfr = "𝔎";
  var kfr = "𝔨";
  var kgreen = "ĸ";
  var KHcy = "Х";
  var khcy = "х";
  var KJcy = "Ќ";
  var kjcy = "ќ";
  var Kopf = "𝕂";
  var kopf = "𝕜";
  var Kscr = "𝒦";
  var kscr = "𝓀";
  var lAarr = "⇚";
  var Lacute = "Ĺ";
  var lacute = "ĺ";
  var laemptyv = "⦴";
  var lagran = "ℒ";
  var Lambda = "Λ";
  var lambda = "λ";
  var lang = "⟨";
  var Lang = "⟪";
  var langd = "⦑";
  var langle = "⟨";
  var lap = "⪅";
  var Laplacetrf = "ℒ";
  var laquo = "«";
  var larrb = "⇤";
  var larrbfs = "⤟";
  var larr = "←";
  var Larr = "↞";
  var lArr = "⇐";
  var larrfs = "⤝";
  var larrhk = "↩";
  var larrlp = "↫";
  var larrpl = "⤹";
  var larrsim = "⥳";
  var larrtl = "↢";
  var latail = "⤙";
  var lAtail = "⤛";
  var lat = "⪫";
  var late = "⪭";
  var lates = "⪭︀";
  var lbarr = "⤌";
  var lBarr = "⤎";
  var lbbrk = "❲";
  var lbrace = "{";
  var lbrack = "[";
  var lbrke = "⦋";
  var lbrksld = "⦏";
  var lbrkslu = "⦍";
  var Lcaron = "Ľ";
  var lcaron = "ľ";
  var Lcedil = "Ļ";
  var lcedil = "ļ";
  var lceil = "⌈";
  var lcub = "{";
  var Lcy = "Л";
  var lcy = "л";
  var ldca = "⤶";
  var ldquo = "“";
  var ldquor = "„";
  var ldrdhar = "⥧";
  var ldrushar = "⥋";
  var ldsh = "↲";
  var le = "≤";
  var lE = "≦";
  var LeftAngleBracket = "⟨";
  var LeftArrowBar = "⇤";
  var leftarrow = "←";
  var LeftArrow = "←";
  var Leftarrow = "⇐";
  var LeftArrowRightArrow = "⇆";
  var leftarrowtail = "↢";
  var LeftCeiling = "⌈";
  var LeftDoubleBracket = "⟦";
  var LeftDownTeeVector = "⥡";
  var LeftDownVectorBar = "⥙";
  var LeftDownVector = "⇃";
  var LeftFloor = "⌊";
  var leftharpoondown = "↽";
  var leftharpoonup = "↼";
  var leftleftarrows = "⇇";
  var leftrightarrow = "↔";
  var LeftRightArrow = "↔";
  var Leftrightarrow = "⇔";
  var leftrightarrows = "⇆";
  var leftrightharpoons = "⇋";
  var leftrightsquigarrow = "↭";
  var LeftRightVector = "⥎";
  var LeftTeeArrow = "↤";
  var LeftTee = "⊣";
  var LeftTeeVector = "⥚";
  var leftthreetimes = "⋋";
  var LeftTriangleBar = "⧏";
  var LeftTriangle = "⊲";
  var LeftTriangleEqual = "⊴";
  var LeftUpDownVector = "⥑";
  var LeftUpTeeVector = "⥠";
  var LeftUpVectorBar = "⥘";
  var LeftUpVector = "↿";
  var LeftVectorBar = "⥒";
  var LeftVector = "↼";
  var lEg = "⪋";
  var leg = "⋚";
  var leq = "≤";
  var leqq = "≦";
  var leqslant = "⩽";
  var lescc = "⪨";
  var les = "⩽";
  var lesdot = "⩿";
  var lesdoto = "⪁";
  var lesdotor = "⪃";
  var lesg = "⋚︀";
  var lesges = "⪓";
  var lessapprox = "⪅";
  var lessdot = "⋖";
  var lesseqgtr = "⋚";
  var lesseqqgtr = "⪋";
  var LessEqualGreater = "⋚";
  var LessFullEqual = "≦";
  var LessGreater = "≶";
  var lessgtr = "≶";
  var LessLess = "⪡";
  var lesssim = "≲";
  var LessSlantEqual = "⩽";
  var LessTilde = "≲";
  var lfisht = "⥼";
  var lfloor = "⌊";
  var Lfr = "𝔏";
  var lfr = "𝔩";
  var lg = "≶";
  var lgE = "⪑";
  var lHar = "⥢";
  var lhard = "↽";
  var lharu = "↼";
  var lharul = "⥪";
  var lhblk = "▄";
  var LJcy = "Љ";
  var ljcy = "љ";
  var llarr = "⇇";
  var ll = "≪";
  var Ll = "⋘";
  var llcorner = "⌞";
  var Lleftarrow = "⇚";
  var llhard = "⥫";
  var lltri = "◺";
  var Lmidot = "Ŀ";
  var lmidot = "ŀ";
  var lmoustache = "⎰";
  var lmoust = "⎰";
  var lnap = "⪉";
  var lnapprox = "⪉";
  var lne = "⪇";
  var lnE = "≨";
  var lneq = "⪇";
  var lneqq = "≨";
  var lnsim = "⋦";
  var loang = "⟬";
  var loarr = "⇽";
  var lobrk = "⟦";
  var longleftarrow = "⟵";
  var LongLeftArrow = "⟵";
  var Longleftarrow = "⟸";
  var longleftrightarrow = "⟷";
  var LongLeftRightArrow = "⟷";
  var Longleftrightarrow = "⟺";
  var longmapsto = "⟼";
  var longrightarrow = "⟶";
  var LongRightArrow = "⟶";
  var Longrightarrow = "⟹";
  var looparrowleft = "↫";
  var looparrowright = "↬";
  var lopar = "⦅";
  var Lopf = "𝕃";
  var lopf = "𝕝";
  var loplus = "⨭";
  var lotimes = "⨴";
  var lowast = "∗";
  var lowbar = "_";
  var LowerLeftArrow = "↙";
  var LowerRightArrow = "↘";
  var loz = "◊";
  var lozenge = "◊";
  var lozf = "⧫";
  var lpar = "(";
  var lparlt = "⦓";
  var lrarr = "⇆";
  var lrcorner = "⌟";
  var lrhar = "⇋";
  var lrhard = "⥭";
  var lrm = "‎";
  var lrtri = "⊿";
  var lsaquo = "‹";
  var lscr = "𝓁";
  var Lscr = "ℒ";
  var lsh = "↰";
  var Lsh = "↰";
  var lsim = "≲";
  var lsime = "⪍";
  var lsimg = "⪏";
  var lsqb = "[";
  var lsquo = "‘";
  var lsquor = "‚";
  var Lstrok = "Ł";
  var lstrok = "ł";
  var ltcc = "⪦";
  var ltcir = "⩹";
  var lt = "<";
  var LT = "<";
  var Lt = "≪";
  var ltdot = "⋖";
  var lthree = "⋋";
  var ltimes = "⋉";
  var ltlarr = "⥶";
  var ltquest = "⩻";
  var ltri = "◃";
  var ltrie = "⊴";
  var ltrif = "◂";
  var ltrPar = "⦖";
  var lurdshar = "⥊";
  var luruhar = "⥦";
  var lvertneqq = "≨︀";
  var lvnE = "≨︀";
  var macr = "¯";
  var male = "♂";
  var malt = "✠";
  var maltese = "✠";
  var map = "↦";
  var mapsto = "↦";
  var mapstodown = "↧";
  var mapstoleft = "↤";
  var mapstoup = "↥";
  var marker = "▮";
  var mcomma = "⨩";
  var Mcy = "М";
  var mcy = "м";
  var mdash = "—";
  var mDDot = "∺";
  var measuredangle = "∡";
  var MediumSpace = " ";
  var Mellintrf = "ℳ";
  var Mfr = "𝔐";
  var mfr = "𝔪";
  var mho = "℧";
  var micro = "µ";
  var midast = "*";
  var midcir = "⫰";
  var mid = "∣";
  var middot = "·";
  var minusb = "⊟";
  var minus = "−";
  var minusd = "∸";
  var minusdu = "⨪";
  var MinusPlus = "∓";
  var mlcp = "⫛";
  var mldr = "…";
  var mnplus = "∓";
  var models = "⊧";
  var Mopf = "𝕄";
  var mopf = "𝕞";
  var mp = "∓";
  var mscr = "𝓂";
  var Mscr = "ℳ";
  var mstpos = "∾";
  var Mu = "Μ";
  var mu = "μ";
  var multimap = "⊸";
  var mumap = "⊸";
  var nabla = "∇";
  var Nacute = "Ń";
  var nacute = "ń";
  var nang = "∠⃒";
  var nap = "≉";
  var napE = "⩰̸";
  var napid = "≋̸";
  var napos = "ŉ";
  var napprox = "≉";
  var natural = "♮";
  var naturals = "ℕ";
  var natur = "♮";
  var nbsp = " ";
  var nbump = "≎̸";
  var nbumpe = "≏̸";
  var ncap = "⩃";
  var Ncaron = "Ň";
  var ncaron = "ň";
  var Ncedil = "Ņ";
  var ncedil = "ņ";
  var ncong = "≇";
  var ncongdot = "⩭̸";
  var ncup = "⩂";
  var Ncy = "Н";
  var ncy = "н";
  var ndash = "–";
  var nearhk = "⤤";
  var nearr = "↗";
  var neArr = "⇗";
  var nearrow = "↗";
  var ne = "≠";
  var nedot = "≐̸";
  var NegativeMediumSpace = "​";
  var NegativeThickSpace = "​";
  var NegativeThinSpace = "​";
  var NegativeVeryThinSpace = "​";
  var nequiv = "≢";
  var nesear = "⤨";
  var nesim = "≂̸";
  var NestedGreaterGreater = "≫";
  var NestedLessLess = "≪";
  var NewLine = "\n";
  var nexist = "∄";
  var nexists = "∄";
  var Nfr = "𝔑";
  var nfr = "𝔫";
  var ngE = "≧̸";
  var nge = "≱";
  var ngeq = "≱";
  var ngeqq = "≧̸";
  var ngeqslant = "⩾̸";
  var nges = "⩾̸";
  var nGg = "⋙̸";
  var ngsim = "≵";
  var nGt = "≫⃒";
  var ngt = "≯";
  var ngtr = "≯";
  var nGtv = "≫̸";
  var nharr = "↮";
  var nhArr = "⇎";
  var nhpar = "⫲";
  var ni = "∋";
  var nis = "⋼";
  var nisd = "⋺";
  var niv = "∋";
  var NJcy = "Њ";
  var njcy = "њ";
  var nlarr = "↚";
  var nlArr = "⇍";
  var nldr = "‥";
  var nlE = "≦̸";
  var nle = "≰";
  var nleftarrow = "↚";
  var nLeftarrow = "⇍";
  var nleftrightarrow = "↮";
  var nLeftrightarrow = "⇎";
  var nleq = "≰";
  var nleqq = "≦̸";
  var nleqslant = "⩽̸";
  var nles = "⩽̸";
  var nless = "≮";
  var nLl = "⋘̸";
  var nlsim = "≴";
  var nLt = "≪⃒";
  var nlt = "≮";
  var nltri = "⋪";
  var nltrie = "⋬";
  var nLtv = "≪̸";
  var nmid = "∤";
  var NoBreak = "⁠";
  var NonBreakingSpace = " ";
  var nopf = "𝕟";
  var Nopf = "ℕ";
  var Not = "⫬";
  var not = "¬";
  var NotCongruent = "≢";
  var NotCupCap = "≭";
  var NotDoubleVerticalBar = "∦";
  var NotElement = "∉";
  var NotEqual = "≠";
  var NotEqualTilde = "≂̸";
  var NotExists = "∄";
  var NotGreater = "≯";
  var NotGreaterEqual = "≱";
  var NotGreaterFullEqual = "≧̸";
  var NotGreaterGreater = "≫̸";
  var NotGreaterLess = "≹";
  var NotGreaterSlantEqual = "⩾̸";
  var NotGreaterTilde = "≵";
  var NotHumpDownHump = "≎̸";
  var NotHumpEqual = "≏̸";
  var notin = "∉";
  var notindot = "⋵̸";
  var notinE = "⋹̸";
  var notinva = "∉";
  var notinvb = "⋷";
  var notinvc = "⋶";
  var NotLeftTriangleBar = "⧏̸";
  var NotLeftTriangle = "⋪";
  var NotLeftTriangleEqual = "⋬";
  var NotLess = "≮";
  var NotLessEqual = "≰";
  var NotLessGreater = "≸";
  var NotLessLess = "≪̸";
  var NotLessSlantEqual = "⩽̸";
  var NotLessTilde = "≴";
  var NotNestedGreaterGreater = "⪢̸";
  var NotNestedLessLess = "⪡̸";
  var notni = "∌";
  var notniva = "∌";
  var notnivb = "⋾";
  var notnivc = "⋽";
  var NotPrecedes = "⊀";
  var NotPrecedesEqual = "⪯̸";
  var NotPrecedesSlantEqual = "⋠";
  var NotReverseElement = "∌";
  var NotRightTriangleBar = "⧐̸";
  var NotRightTriangle = "⋫";
  var NotRightTriangleEqual = "⋭";
  var NotSquareSubset = "⊏̸";
  var NotSquareSubsetEqual = "⋢";
  var NotSquareSuperset = "⊐̸";
  var NotSquareSupersetEqual = "⋣";
  var NotSubset = "⊂⃒";
  var NotSubsetEqual = "⊈";
  var NotSucceeds = "⊁";
  var NotSucceedsEqual = "⪰̸";
  var NotSucceedsSlantEqual = "⋡";
  var NotSucceedsTilde = "≿̸";
  var NotSuperset = "⊃⃒";
  var NotSupersetEqual = "⊉";
  var NotTilde = "≁";
  var NotTildeEqual = "≄";
  var NotTildeFullEqual = "≇";
  var NotTildeTilde = "≉";
  var NotVerticalBar = "∤";
  var nparallel = "∦";
  var npar = "∦";
  var nparsl = "⫽⃥";
  var npart = "∂̸";
  var npolint = "⨔";
  var npr = "⊀";
  var nprcue = "⋠";
  var nprec = "⊀";
  var npreceq = "⪯̸";
  var npre = "⪯̸";
  var nrarrc = "⤳̸";
  var nrarr = "↛";
  var nrArr = "⇏";
  var nrarrw = "↝̸";
  var nrightarrow = "↛";
  var nRightarrow = "⇏";
  var nrtri = "⋫";
  var nrtrie = "⋭";
  var nsc = "⊁";
  var nsccue = "⋡";
  var nsce = "⪰̸";
  var Nscr = "𝒩";
  var nscr = "𝓃";
  var nshortmid = "∤";
  var nshortparallel = "∦";
  var nsim = "≁";
  var nsime = "≄";
  var nsimeq = "≄";
  var nsmid = "∤";
  var nspar = "∦";
  var nsqsube = "⋢";
  var nsqsupe = "⋣";
  var nsub = "⊄";
  var nsubE = "⫅̸";
  var nsube = "⊈";
  var nsubset = "⊂⃒";
  var nsubseteq = "⊈";
  var nsubseteqq = "⫅̸";
  var nsucc = "⊁";
  var nsucceq = "⪰̸";
  var nsup = "⊅";
  var nsupE = "⫆̸";
  var nsupe = "⊉";
  var nsupset = "⊃⃒";
  var nsupseteq = "⊉";
  var nsupseteqq = "⫆̸";
  var ntgl = "≹";
  var Ntilde = "Ñ";
  var ntilde = "ñ";
  var ntlg = "≸";
  var ntriangleleft = "⋪";
  var ntrianglelefteq = "⋬";
  var ntriangleright = "⋫";
  var ntrianglerighteq = "⋭";
  var Nu = "Ν";
  var nu = "ν";
  var num = "#";
  var numero = "№";
  var numsp = " ";
  var nvap = "≍⃒";
  var nvdash = "⊬";
  var nvDash = "⊭";
  var nVdash = "⊮";
  var nVDash = "⊯";
  var nvge = "≥⃒";
  var nvgt = ">⃒";
  var nvHarr = "⤄";
  var nvinfin = "⧞";
  var nvlArr = "⤂";
  var nvle = "≤⃒";
  var nvlt = "<⃒";
  var nvltrie = "⊴⃒";
  var nvrArr = "⤃";
  var nvrtrie = "⊵⃒";
  var nvsim = "∼⃒";
  var nwarhk = "⤣";
  var nwarr = "↖";
  var nwArr = "⇖";
  var nwarrow = "↖";
  var nwnear = "⤧";
  var Oacute = "Ó";
  var oacute = "ó";
  var oast = "⊛";
  var Ocirc = "Ô";
  var ocirc = "ô";
  var ocir = "⊚";
  var Ocy = "О";
  var ocy = "о";
  var odash = "⊝";
  var Odblac = "Ő";
  var odblac = "ő";
  var odiv = "⨸";
  var odot = "⊙";
  var odsold = "⦼";
  var OElig = "Œ";
  var oelig = "œ";
  var ofcir = "⦿";
  var Ofr = "𝔒";
  var ofr = "𝔬";
  var ogon = "˛";
  var Ograve = "Ò";
  var ograve = "ò";
  var ogt = "⧁";
  var ohbar = "⦵";
  var ohm = "Ω";
  var oint = "∮";
  var olarr = "↺";
  var olcir = "⦾";
  var olcross = "⦻";
  var oline = "‾";
  var olt = "⧀";
  var Omacr = "Ō";
  var omacr = "ō";
  var Omega = "Ω";
  var omega = "ω";
  var Omicron = "Ο";
  var omicron = "ο";
  var omid = "⦶";
  var ominus = "⊖";
  var Oopf = "𝕆";
  var oopf = "𝕠";
  var opar = "⦷";
  var OpenCurlyDoubleQuote = "“";
  var OpenCurlyQuote = "‘";
  var operp = "⦹";
  var oplus = "⊕";
  var orarr = "↻";
  var Or = "⩔";
  var or = "∨";
  var ord = "⩝";
  var order = "ℴ";
  var orderof = "ℴ";
  var ordf = "ª";
  var ordm = "º";
  var origof = "⊶";
  var oror = "⩖";
  var orslope = "⩗";
  var orv = "⩛";
  var oS = "Ⓢ";
  var Oscr = "𝒪";
  var oscr = "ℴ";
  var Oslash = "Ø";
  var oslash = "ø";
  var osol = "⊘";
  var Otilde = "Õ";
  var otilde = "õ";
  var otimesas = "⨶";
  var Otimes = "⨷";
  var otimes = "⊗";
  var Ouml = "Ö";
  var ouml = "ö";
  var ovbar = "⌽";
  var OverBar = "‾";
  var OverBrace = "⏞";
  var OverBracket = "⎴";
  var OverParenthesis = "⏜";
  var para = "¶";
  var parallel = "∥";
  var par = "∥";
  var parsim = "⫳";
  var parsl = "⫽";
  var part = "∂";
  var PartialD = "∂";
  var Pcy = "П";
  var pcy = "п";
  var percnt = "%";
  var period = ".";
  var permil = "‰";
  var perp = "⊥";
  var pertenk = "‱";
  var Pfr = "𝔓";
  var pfr = "𝔭";
  var Phi = "Φ";
  var phi = "φ";
  var phiv = "ϕ";
  var phmmat = "ℳ";
  var phone = "☎";
  var Pi = "Π";
  var pi = "π";
  var pitchfork = "⋔";
  var piv = "ϖ";
  var planck = "ℏ";
  var planckh = "ℎ";
  var plankv = "ℏ";
  var plusacir = "⨣";
  var plusb = "⊞";
  var pluscir = "⨢";
  var plus = "+";
  var plusdo = "∔";
  var plusdu = "⨥";
  var pluse = "⩲";
  var PlusMinus = "±";
  var plusmn = "±";
  var plussim = "⨦";
  var plustwo = "⨧";
  var pm = "±";
  var Poincareplane = "ℌ";
  var pointint = "⨕";
  var popf = "𝕡";
  var Popf = "ℙ";
  var pound = "£";
  var prap = "⪷";
  var Pr = "⪻";
  var pr = "≺";
  var prcue = "≼";
  var precapprox = "⪷";
  var prec = "≺";
  var preccurlyeq = "≼";
  var Precedes = "≺";
  var PrecedesEqual = "⪯";
  var PrecedesSlantEqual = "≼";
  var PrecedesTilde = "≾";
  var preceq = "⪯";
  var precnapprox = "⪹";
  var precneqq = "⪵";
  var precnsim = "⋨";
  var pre = "⪯";
  var prE = "⪳";
  var precsim = "≾";
  var prime = "′";
  var Prime = "″";
  var primes = "ℙ";
  var prnap = "⪹";
  var prnE = "⪵";
  var prnsim = "⋨";
  var prod = "∏";
  var Product = "∏";
  var profalar = "⌮";
  var profline = "⌒";
  var profsurf = "⌓";
  var prop = "∝";
  var Proportional = "∝";
  var Proportion = "∷";
  var propto = "∝";
  var prsim = "≾";
  var prurel = "⊰";
  var Pscr = "𝒫";
  var pscr = "𝓅";
  var Psi = "Ψ";
  var psi = "ψ";
  var puncsp = " ";
  var Qfr = "𝔔";
  var qfr = "𝔮";
  var qint = "⨌";
  var qopf = "𝕢";
  var Qopf = "ℚ";
  var qprime = "⁗";
  var Qscr = "𝒬";
  var qscr = "𝓆";
  var quaternions = "ℍ";
  var quatint = "⨖";
  var quest = "?";
  var questeq = "≟";
  var quot = "\"";
  var QUOT = "\"";
  var rAarr = "⇛";
  var race = "∽̱";
  var Racute = "Ŕ";
  var racute = "ŕ";
  var radic = "√";
  var raemptyv = "⦳";
  var rang = "⟩";
  var Rang = "⟫";
  var rangd = "⦒";
  var range = "⦥";
  var rangle = "⟩";
  var raquo = "»";
  var rarrap = "⥵";
  var rarrb = "⇥";
  var rarrbfs = "⤠";
  var rarrc = "⤳";
  var rarr = "→";
  var Rarr = "↠";
  var rArr = "⇒";
  var rarrfs = "⤞";
  var rarrhk = "↪";
  var rarrlp = "↬";
  var rarrpl = "⥅";
  var rarrsim = "⥴";
  var Rarrtl = "⤖";
  var rarrtl = "↣";
  var rarrw = "↝";
  var ratail = "⤚";
  var rAtail = "⤜";
  var ratio = "∶";
  var rationals = "ℚ";
  var rbarr = "⤍";
  var rBarr = "⤏";
  var RBarr = "⤐";
  var rbbrk = "❳";
  var rbrace = "}";
  var rbrack = "]";
  var rbrke = "⦌";
  var rbrksld = "⦎";
  var rbrkslu = "⦐";
  var Rcaron = "Ř";
  var rcaron = "ř";
  var Rcedil = "Ŗ";
  var rcedil = "ŗ";
  var rceil = "⌉";
  var rcub = "}";
  var Rcy = "Р";
  var rcy = "р";
  var rdca = "⤷";
  var rdldhar = "⥩";
  var rdquo = "”";
  var rdquor = "”";
  var rdsh = "↳";
  var real = "ℜ";
  var realine = "ℛ";
  var realpart = "ℜ";
  var reals = "ℝ";
  var Re = "ℜ";
  var rect = "▭";
  var reg = "®";
  var REG = "®";
  var ReverseElement = "∋";
  var ReverseEquilibrium = "⇋";
  var ReverseUpEquilibrium = "⥯";
  var rfisht = "⥽";
  var rfloor = "⌋";
  var rfr = "𝔯";
  var Rfr = "ℜ";
  var rHar = "⥤";
  var rhard = "⇁";
  var rharu = "⇀";
  var rharul = "⥬";
  var Rho = "Ρ";
  var rho = "ρ";
  var rhov = "ϱ";
  var RightAngleBracket = "⟩";
  var RightArrowBar = "⇥";
  var rightarrow = "→";
  var RightArrow = "→";
  var Rightarrow = "⇒";
  var RightArrowLeftArrow = "⇄";
  var rightarrowtail = "↣";
  var RightCeiling = "⌉";
  var RightDoubleBracket = "⟧";
  var RightDownTeeVector = "⥝";
  var RightDownVectorBar = "⥕";
  var RightDownVector = "⇂";
  var RightFloor = "⌋";
  var rightharpoondown = "⇁";
  var rightharpoonup = "⇀";
  var rightleftarrows = "⇄";
  var rightleftharpoons = "⇌";
  var rightrightarrows = "⇉";
  var rightsquigarrow = "↝";
  var RightTeeArrow = "↦";
  var RightTee = "⊢";
  var RightTeeVector = "⥛";
  var rightthreetimes = "⋌";
  var RightTriangleBar = "⧐";
  var RightTriangle = "⊳";
  var RightTriangleEqual = "⊵";
  var RightUpDownVector = "⥏";
  var RightUpTeeVector = "⥜";
  var RightUpVectorBar = "⥔";
  var RightUpVector = "↾";
  var RightVectorBar = "⥓";
  var RightVector = "⇀";
  var ring = "˚";
  var risingdotseq = "≓";
  var rlarr = "⇄";
  var rlhar = "⇌";
  var rlm = "‏";
  var rmoustache = "⎱";
  var rmoust = "⎱";
  var rnmid = "⫮";
  var roang = "⟭";
  var roarr = "⇾";
  var robrk = "⟧";
  var ropar = "⦆";
  var ropf = "𝕣";
  var Ropf = "ℝ";
  var roplus = "⨮";
  var rotimes = "⨵";
  var RoundImplies = "⥰";
  var rpar = ")";
  var rpargt = "⦔";
  var rppolint = "⨒";
  var rrarr = "⇉";
  var Rrightarrow = "⇛";
  var rsaquo = "›";
  var rscr = "𝓇";
  var Rscr = "ℛ";
  var rsh = "↱";
  var Rsh = "↱";
  var rsqb = "]";
  var rsquo = "’";
  var rsquor = "’";
  var rthree = "⋌";
  var rtimes = "⋊";
  var rtri = "▹";
  var rtrie = "⊵";
  var rtrif = "▸";
  var rtriltri = "⧎";
  var RuleDelayed = "⧴";
  var ruluhar = "⥨";
  var rx = "℞";
  var Sacute = "Ś";
  var sacute = "ś";
  var sbquo = "‚";
  var scap = "⪸";
  var Scaron = "Š";
  var scaron = "š";
  var Sc = "⪼";
  var sc = "≻";
  var sccue = "≽";
  var sce = "⪰";
  var scE = "⪴";
  var Scedil = "Ş";
  var scedil = "ş";
  var Scirc = "Ŝ";
  var scirc = "ŝ";
  var scnap = "⪺";
  var scnE = "⪶";
  var scnsim = "⋩";
  var scpolint = "⨓";
  var scsim = "≿";
  var Scy = "С";
  var scy = "с";
  var sdotb = "⊡";
  var sdot = "⋅";
  var sdote = "⩦";
  var searhk = "⤥";
  var searr = "↘";
  var seArr = "⇘";
  var searrow = "↘";
  var sect = "§";
  var semi = ";";
  var seswar = "⤩";
  var setminus = "∖";
  var setmn = "∖";
  var sext = "✶";
  var Sfr = "𝔖";
  var sfr = "𝔰";
  var sfrown = "⌢";
  var sharp = "♯";
  var SHCHcy = "Щ";
  var shchcy = "щ";
  var SHcy = "Ш";
  var shcy = "ш";
  var ShortDownArrow = "↓";
  var ShortLeftArrow = "←";
  var shortmid = "∣";
  var shortparallel = "∥";
  var ShortRightArrow = "→";
  var ShortUpArrow = "↑";
  var shy = "­";
  var Sigma = "Σ";
  var sigma = "σ";
  var sigmaf = "ς";
  var sigmav = "ς";
  var sim = "∼";
  var simdot = "⩪";
  var sime = "≃";
  var simeq = "≃";
  var simg = "⪞";
  var simgE = "⪠";
  var siml = "⪝";
  var simlE = "⪟";
  var simne = "≆";
  var simplus = "⨤";
  var simrarr = "⥲";
  var slarr = "←";
  var SmallCircle = "∘";
  var smallsetminus = "∖";
  var smashp = "⨳";
  var smeparsl = "⧤";
  var smid = "∣";
  var smile = "⌣";
  var smt = "⪪";
  var smte = "⪬";
  var smtes = "⪬︀";
  var SOFTcy = "Ь";
  var softcy = "ь";
  var solbar = "⌿";
  var solb = "⧄";
  var sol = "/";
  var Sopf = "𝕊";
  var sopf = "𝕤";
  var spades = "♠";
  var spadesuit = "♠";
  var spar = "∥";
  var sqcap = "⊓";
  var sqcaps = "⊓︀";
  var sqcup = "⊔";
  var sqcups = "⊔︀";
  var Sqrt = "√";
  var sqsub = "⊏";
  var sqsube = "⊑";
  var sqsubset = "⊏";
  var sqsubseteq = "⊑";
  var sqsup = "⊐";
  var sqsupe = "⊒";
  var sqsupset = "⊐";
  var sqsupseteq = "⊒";
  var square = "□";
  var Square = "□";
  var SquareIntersection = "⊓";
  var SquareSubset = "⊏";
  var SquareSubsetEqual = "⊑";
  var SquareSuperset = "⊐";
  var SquareSupersetEqual = "⊒";
  var SquareUnion = "⊔";
  var squarf = "▪";
  var squ = "□";
  var squf = "▪";
  var srarr = "→";
  var Sscr = "𝒮";
  var sscr = "𝓈";
  var ssetmn = "∖";
  var ssmile = "⌣";
  var sstarf = "⋆";
  var Star = "⋆";
  var star = "☆";
  var starf = "★";
  var straightepsilon = "ϵ";
  var straightphi = "ϕ";
  var strns = "¯";
  var sub = "⊂";
  var Sub = "⋐";
  var subdot = "⪽";
  var subE = "⫅";
  var sube = "⊆";
  var subedot = "⫃";
  var submult = "⫁";
  var subnE = "⫋";
  var subne = "⊊";
  var subplus = "⪿";
  var subrarr = "⥹";
  var subset = "⊂";
  var Subset = "⋐";
  var subseteq = "⊆";
  var subseteqq = "⫅";
  var SubsetEqual = "⊆";
  var subsetneq = "⊊";
  var subsetneqq = "⫋";
  var subsim = "⫇";
  var subsub = "⫕";
  var subsup = "⫓";
  var succapprox = "⪸";
  var succ = "≻";
  var succcurlyeq = "≽";
  var Succeeds = "≻";
  var SucceedsEqual = "⪰";
  var SucceedsSlantEqual = "≽";
  var SucceedsTilde = "≿";
  var succeq = "⪰";
  var succnapprox = "⪺";
  var succneqq = "⪶";
  var succnsim = "⋩";
  var succsim = "≿";
  var SuchThat = "∋";
  var sum = "∑";
  var Sum = "∑";
  var sung = "♪";
  var sup1 = "¹";
  var sup2 = "²";
  var sup3 = "³";
  var sup = "⊃";
  var Sup = "⋑";
  var supdot = "⪾";
  var supdsub = "⫘";
  var supE = "⫆";
  var supe = "⊇";
  var supedot = "⫄";
  var Superset = "⊃";
  var SupersetEqual = "⊇";
  var suphsol = "⟉";
  var suphsub = "⫗";
  var suplarr = "⥻";
  var supmult = "⫂";
  var supnE = "⫌";
  var supne = "⊋";
  var supplus = "⫀";
  var supset = "⊃";
  var Supset = "⋑";
  var supseteq = "⊇";
  var supseteqq = "⫆";
  var supsetneq = "⊋";
  var supsetneqq = "⫌";
  var supsim = "⫈";
  var supsub = "⫔";
  var supsup = "⫖";
  var swarhk = "⤦";
  var swarr = "↙";
  var swArr = "⇙";
  var swarrow = "↙";
  var swnwar = "⤪";
  var szlig = "ß";
  var Tab = "\t";
  var target = "⌖";
  var Tau = "Τ";
  var tau = "τ";
  var tbrk = "⎴";
  var Tcaron = "Ť";
  var tcaron = "ť";
  var Tcedil = "Ţ";
  var tcedil = "ţ";
  var Tcy = "Т";
  var tcy = "т";
  var tdot = "⃛";
  var telrec = "⌕";
  var Tfr = "𝔗";
  var tfr = "𝔱";
  var there4 = "∴";
  var therefore = "∴";
  var Therefore = "∴";
  var Theta = "Θ";
  var theta = "θ";
  var thetasym = "ϑ";
  var thetav = "ϑ";
  var thickapprox = "≈";
  var thicksim = "∼";
  var ThickSpace = "  ";
  var ThinSpace = " ";
  var thinsp = " ";
  var thkap = "≈";
  var thksim = "∼";
  var THORN = "Þ";
  var thorn = "þ";
  var tilde = "˜";
  var Tilde = "∼";
  var TildeEqual = "≃";
  var TildeFullEqual = "≅";
  var TildeTilde = "≈";
  var timesbar = "⨱";
  var timesb = "⊠";
  var times = "×";
  var timesd = "⨰";
  var tint = "∭";
  var toea = "⤨";
  var topbot = "⌶";
  var topcir = "⫱";
  var top = "⊤";
  var Topf = "𝕋";
  var topf = "𝕥";
  var topfork = "⫚";
  var tosa = "⤩";
  var tprime = "‴";
  var trade = "™";
  var TRADE = "™";
  var triangle = "▵";
  var triangledown = "▿";
  var triangleleft = "◃";
  var trianglelefteq = "⊴";
  var triangleq = "≜";
  var triangleright = "▹";
  var trianglerighteq = "⊵";
  var tridot = "◬";
  var trie = "≜";
  var triminus = "⨺";
  var TripleDot = "⃛";
  var triplus = "⨹";
  var trisb = "⧍";
  var tritime = "⨻";
  var trpezium = "⏢";
  var Tscr = "𝒯";
  var tscr = "𝓉";
  var TScy = "Ц";
  var tscy = "ц";
  var TSHcy = "Ћ";
  var tshcy = "ћ";
  var Tstrok = "Ŧ";
  var tstrok = "ŧ";
  var twixt = "≬";
  var twoheadleftarrow = "↞";
  var twoheadrightarrow = "↠";
  var Uacute = "Ú";
  var uacute = "ú";
  var uarr = "↑";
  var Uarr = "↟";
  var uArr = "⇑";
  var Uarrocir = "⥉";
  var Ubrcy = "Ў";
  var ubrcy = "ў";
  var Ubreve = "Ŭ";
  var ubreve = "ŭ";
  var Ucirc = "Û";
  var ucirc = "û";
  var Ucy = "У";
  var ucy = "у";
  var udarr = "⇅";
  var Udblac = "Ű";
  var udblac = "ű";
  var udhar = "⥮";
  var ufisht = "⥾";
  var Ufr = "𝔘";
  var ufr = "𝔲";
  var Ugrave = "Ù";
  var ugrave = "ù";
  var uHar = "⥣";
  var uharl = "↿";
  var uharr = "↾";
  var uhblk = "▀";
  var ulcorn = "⌜";
  var ulcorner = "⌜";
  var ulcrop = "⌏";
  var ultri = "◸";
  var Umacr = "Ū";
  var umacr = "ū";
  var uml = "¨";
  var UnderBar = "_";
  var UnderBrace = "⏟";
  var UnderBracket = "⎵";
  var UnderParenthesis = "⏝";
  var Union = "⋃";
  var UnionPlus = "⊎";
  var Uogon = "Ų";
  var uogon = "ų";
  var Uopf = "𝕌";
  var uopf = "𝕦";
  var UpArrowBar = "⤒";
  var uparrow = "↑";
  var UpArrow = "↑";
  var Uparrow = "⇑";
  var UpArrowDownArrow = "⇅";
  var updownarrow = "↕";
  var UpDownArrow = "↕";
  var Updownarrow = "⇕";
  var UpEquilibrium = "⥮";
  var upharpoonleft = "↿";
  var upharpoonright = "↾";
  var uplus = "⊎";
  var UpperLeftArrow = "↖";
  var UpperRightArrow = "↗";
  var upsi = "υ";
  var Upsi = "ϒ";
  var upsih = "ϒ";
  var Upsilon = "Υ";
  var upsilon = "υ";
  var UpTeeArrow = "↥";
  var UpTee = "⊥";
  var upuparrows = "⇈";
  var urcorn = "⌝";
  var urcorner = "⌝";
  var urcrop = "⌎";
  var Uring = "Ů";
  var uring = "ů";
  var urtri = "◹";
  var Uscr = "𝒰";
  var uscr = "𝓊";
  var utdot = "⋰";
  var Utilde = "Ũ";
  var utilde = "ũ";
  var utri = "▵";
  var utrif = "▴";
  var uuarr = "⇈";
  var Uuml = "Ü";
  var uuml = "ü";
  var uwangle = "⦧";
  var vangrt = "⦜";
  var varepsilon = "ϵ";
  var varkappa = "ϰ";
  var varnothing = "∅";
  var varphi = "ϕ";
  var varpi = "ϖ";
  var varpropto = "∝";
  var varr = "↕";
  var vArr = "⇕";
  var varrho = "ϱ";
  var varsigma = "ς";
  var varsubsetneq = "⊊︀";
  var varsubsetneqq = "⫋︀";
  var varsupsetneq = "⊋︀";
  var varsupsetneqq = "⫌︀";
  var vartheta = "ϑ";
  var vartriangleleft = "⊲";
  var vartriangleright = "⊳";
  var vBar = "⫨";
  var Vbar = "⫫";
  var vBarv = "⫩";
  var Vcy = "В";
  var vcy = "в";
  var vdash = "⊢";
  var vDash = "⊨";
  var Vdash = "⊩";
  var VDash = "⊫";
  var Vdashl = "⫦";
  var veebar = "⊻";
  var vee = "∨";
  var Vee = "⋁";
  var veeeq = "≚";
  var vellip = "⋮";
  var verbar = "|";
  var Verbar = "‖";
  var vert = "|";
  var Vert = "‖";
  var VerticalBar = "∣";
  var VerticalLine = "|";
  var VerticalSeparator = "❘";
  var VerticalTilde = "≀";
  var VeryThinSpace = " ";
  var Vfr = "𝔙";
  var vfr = "𝔳";
  var vltri = "⊲";
  var vnsub = "⊂⃒";
  var vnsup = "⊃⃒";
  var Vopf = "𝕍";
  var vopf = "𝕧";
  var vprop = "∝";
  var vrtri = "⊳";
  var Vscr = "𝒱";
  var vscr = "𝓋";
  var vsubnE = "⫋︀";
  var vsubne = "⊊︀";
  var vsupnE = "⫌︀";
  var vsupne = "⊋︀";
  var Vvdash = "⊪";
  var vzigzag = "⦚";
  var Wcirc = "Ŵ";
  var wcirc = "ŵ";
  var wedbar = "⩟";
  var wedge = "∧";
  var Wedge = "⋀";
  var wedgeq = "≙";
  var weierp = "℘";
  var Wfr = "𝔚";
  var wfr = "𝔴";
  var Wopf = "𝕎";
  var wopf = "𝕨";
  var wp = "℘";
  var wr = "≀";
  var wreath = "≀";
  var Wscr = "𝒲";
  var wscr = "𝓌";
  var xcap = "⋂";
  var xcirc = "◯";
  var xcup = "⋃";
  var xdtri = "▽";
  var Xfr = "𝔛";
  var xfr = "𝔵";
  var xharr = "⟷";
  var xhArr = "⟺";
  var Xi = "Ξ";
  var xi = "ξ";
  var xlarr = "⟵";
  var xlArr = "⟸";
  var xmap = "⟼";
  var xnis = "⋻";
  var xodot = "⨀";
  var Xopf = "𝕏";
  var xopf = "𝕩";
  var xoplus = "⨁";
  var xotime = "⨂";
  var xrarr = "⟶";
  var xrArr = "⟹";
  var Xscr = "𝒳";
  var xscr = "𝓍";
  var xsqcup = "⨆";
  var xuplus = "⨄";
  var xutri = "△";
  var xvee = "⋁";
  var xwedge = "⋀";
  var Yacute = "Ý";
  var yacute = "ý";
  var YAcy = "Я";
  var yacy = "я";
  var Ycirc = "Ŷ";
  var ycirc = "ŷ";
  var Ycy = "Ы";
  var ycy = "ы";
  var yen = "¥";
  var Yfr = "𝔜";
  var yfr = "𝔶";
  var YIcy = "Ї";
  var yicy = "ї";
  var Yopf = "𝕐";
  var yopf = "𝕪";
  var Yscr = "𝒴";
  var yscr = "𝓎";
  var YUcy = "Ю";
  var yucy = "ю";
  var yuml = "ÿ";
  var Yuml = "Ÿ";
  var Zacute = "Ź";
  var zacute = "ź";
  var Zcaron = "Ž";
  var zcaron = "ž";
  var Zcy = "З";
  var zcy = "з";
  var Zdot = "Ż";
  var zdot = "ż";
  var zeetrf = "ℨ";
  var ZeroWidthSpace = "​";
  var Zeta = "Ζ";
  var zeta = "ζ";
  var zfr = "𝔷";
  var Zfr = "ℨ";
  var ZHcy = "Ж";
  var zhcy = "ж";
  var zigrarr = "⇝";
  var zopf = "𝕫";
  var Zopf = "ℤ";
  var Zscr = "𝒵";
  var zscr = "𝓏";
  var zwj = "‍";
  var zwnj = "‌";
  var entities$1 = {
  	Aacute: Aacute,
  	aacute: aacute,
  	Abreve: Abreve,
  	abreve: abreve,
  	ac: ac,
  	acd: acd,
  	acE: acE,
  	Acirc: Acirc,
  	acirc: acirc,
  	acute: acute,
  	Acy: Acy,
  	acy: acy,
  	AElig: AElig,
  	aelig: aelig,
  	af: af,
  	Afr: Afr,
  	afr: afr,
  	Agrave: Agrave,
  	agrave: agrave,
  	alefsym: alefsym,
  	aleph: aleph,
  	Alpha: Alpha,
  	alpha: alpha,
  	Amacr: Amacr,
  	amacr: amacr,
  	amalg: amalg,
  	amp: amp,
  	AMP: AMP,
  	andand: andand,
  	And: And,
  	and: and,
  	andd: andd,
  	andslope: andslope,
  	andv: andv,
  	ang: ang,
  	ange: ange,
  	angle: angle,
  	angmsdaa: angmsdaa,
  	angmsdab: angmsdab,
  	angmsdac: angmsdac,
  	angmsdad: angmsdad,
  	angmsdae: angmsdae,
  	angmsdaf: angmsdaf,
  	angmsdag: angmsdag,
  	angmsdah: angmsdah,
  	angmsd: angmsd,
  	angrt: angrt,
  	angrtvb: angrtvb,
  	angrtvbd: angrtvbd,
  	angsph: angsph,
  	angst: angst,
  	angzarr: angzarr,
  	Aogon: Aogon,
  	aogon: aogon,
  	Aopf: Aopf,
  	aopf: aopf,
  	apacir: apacir,
  	ap: ap,
  	apE: apE,
  	ape: ape,
  	apid: apid,
  	apos: apos,
  	ApplyFunction: ApplyFunction,
  	approx: approx,
  	approxeq: approxeq,
  	Aring: Aring,
  	aring: aring,
  	Ascr: Ascr,
  	ascr: ascr,
  	Assign: Assign,
  	ast: ast,
  	asymp: asymp,
  	asympeq: asympeq,
  	Atilde: Atilde,
  	atilde: atilde,
  	Auml: Auml,
  	auml: auml,
  	awconint: awconint,
  	awint: awint,
  	backcong: backcong,
  	backepsilon: backepsilon,
  	backprime: backprime,
  	backsim: backsim,
  	backsimeq: backsimeq,
  	Backslash: Backslash,
  	Barv: Barv,
  	barvee: barvee,
  	barwed: barwed,
  	Barwed: Barwed,
  	barwedge: barwedge,
  	bbrk: bbrk,
  	bbrktbrk: bbrktbrk,
  	bcong: bcong,
  	Bcy: Bcy,
  	bcy: bcy,
  	bdquo: bdquo,
  	becaus: becaus,
  	because: because,
  	Because: Because,
  	bemptyv: bemptyv,
  	bepsi: bepsi,
  	bernou: bernou,
  	Bernoullis: Bernoullis,
  	Beta: Beta,
  	beta: beta,
  	beth: beth,
  	between: between,
  	Bfr: Bfr,
  	bfr: bfr,
  	bigcap: bigcap,
  	bigcirc: bigcirc,
  	bigcup: bigcup,
  	bigodot: bigodot,
  	bigoplus: bigoplus,
  	bigotimes: bigotimes,
  	bigsqcup: bigsqcup,
  	bigstar: bigstar,
  	bigtriangledown: bigtriangledown,
  	bigtriangleup: bigtriangleup,
  	biguplus: biguplus,
  	bigvee: bigvee,
  	bigwedge: bigwedge,
  	bkarow: bkarow,
  	blacklozenge: blacklozenge,
  	blacksquare: blacksquare,
  	blacktriangle: blacktriangle,
  	blacktriangledown: blacktriangledown,
  	blacktriangleleft: blacktriangleleft,
  	blacktriangleright: blacktriangleright,
  	blank: blank,
  	blk12: blk12,
  	blk14: blk14,
  	blk34: blk34,
  	block: block$1,
  	bne: bne,
  	bnequiv: bnequiv,
  	bNot: bNot,
  	bnot: bnot,
  	Bopf: Bopf,
  	bopf: bopf,
  	bot: bot,
  	bottom: bottom,
  	bowtie: bowtie,
  	boxbox: boxbox,
  	boxdl: boxdl,
  	boxdL: boxdL,
  	boxDl: boxDl,
  	boxDL: boxDL,
  	boxdr: boxdr,
  	boxdR: boxdR,
  	boxDr: boxDr,
  	boxDR: boxDR,
  	boxh: boxh,
  	boxH: boxH,
  	boxhd: boxhd,
  	boxHd: boxHd,
  	boxhD: boxhD,
  	boxHD: boxHD,
  	boxhu: boxhu,
  	boxHu: boxHu,
  	boxhU: boxhU,
  	boxHU: boxHU,
  	boxminus: boxminus,
  	boxplus: boxplus,
  	boxtimes: boxtimes,
  	boxul: boxul,
  	boxuL: boxuL,
  	boxUl: boxUl,
  	boxUL: boxUL,
  	boxur: boxur,
  	boxuR: boxuR,
  	boxUr: boxUr,
  	boxUR: boxUR,
  	boxv: boxv,
  	boxV: boxV,
  	boxvh: boxvh,
  	boxvH: boxvH,
  	boxVh: boxVh,
  	boxVH: boxVH,
  	boxvl: boxvl,
  	boxvL: boxvL,
  	boxVl: boxVl,
  	boxVL: boxVL,
  	boxvr: boxvr,
  	boxvR: boxvR,
  	boxVr: boxVr,
  	boxVR: boxVR,
  	bprime: bprime,
  	breve: breve,
  	Breve: Breve,
  	brvbar: brvbar,
  	bscr: bscr,
  	Bscr: Bscr,
  	bsemi: bsemi,
  	bsim: bsim,
  	bsime: bsime,
  	bsolb: bsolb,
  	bsol: bsol,
  	bsolhsub: bsolhsub,
  	bull: bull,
  	bullet: bullet,
  	bump: bump,
  	bumpE: bumpE,
  	bumpe: bumpe,
  	Bumpeq: Bumpeq,
  	bumpeq: bumpeq,
  	Cacute: Cacute,
  	cacute: cacute,
  	capand: capand,
  	capbrcup: capbrcup,
  	capcap: capcap,
  	cap: cap,
  	Cap: Cap,
  	capcup: capcup,
  	capdot: capdot,
  	CapitalDifferentialD: CapitalDifferentialD,
  	caps: caps,
  	caret: caret,
  	caron: caron,
  	Cayleys: Cayleys,
  	ccaps: ccaps,
  	Ccaron: Ccaron,
  	ccaron: ccaron,
  	Ccedil: Ccedil,
  	ccedil: ccedil,
  	Ccirc: Ccirc,
  	ccirc: ccirc,
  	Cconint: Cconint,
  	ccups: ccups,
  	ccupssm: ccupssm,
  	Cdot: Cdot,
  	cdot: cdot,
  	cedil: cedil,
  	Cedilla: Cedilla,
  	cemptyv: cemptyv,
  	cent: cent,
  	centerdot: centerdot,
  	CenterDot: CenterDot,
  	cfr: cfr,
  	Cfr: Cfr,
  	CHcy: CHcy,
  	chcy: chcy,
  	check: check,
  	checkmark: checkmark,
  	Chi: Chi,
  	chi: chi,
  	circ: circ,
  	circeq: circeq,
  	circlearrowleft: circlearrowleft,
  	circlearrowright: circlearrowright,
  	circledast: circledast,
  	circledcirc: circledcirc,
  	circleddash: circleddash,
  	CircleDot: CircleDot,
  	circledR: circledR,
  	circledS: circledS,
  	CircleMinus: CircleMinus,
  	CirclePlus: CirclePlus,
  	CircleTimes: CircleTimes,
  	cir: cir,
  	cirE: cirE,
  	cire: cire,
  	cirfnint: cirfnint,
  	cirmid: cirmid,
  	cirscir: cirscir,
  	ClockwiseContourIntegral: ClockwiseContourIntegral,
  	CloseCurlyDoubleQuote: CloseCurlyDoubleQuote,
  	CloseCurlyQuote: CloseCurlyQuote,
  	clubs: clubs,
  	clubsuit: clubsuit,
  	colon: colon,
  	Colon: Colon,
  	Colone: Colone,
  	colone: colone,
  	coloneq: coloneq,
  	comma: comma,
  	commat: commat,
  	comp: comp,
  	compfn: compfn,
  	complement: complement,
  	complexes: complexes,
  	cong: cong,
  	congdot: congdot,
  	Congruent: Congruent,
  	conint: conint,
  	Conint: Conint,
  	ContourIntegral: ContourIntegral,
  	copf: copf,
  	Copf: Copf,
  	coprod: coprod,
  	Coproduct: Coproduct,
  	copy: copy,
  	COPY: COPY,
  	copysr: copysr,
  	CounterClockwiseContourIntegral: CounterClockwiseContourIntegral,
  	crarr: crarr,
  	cross: cross,
  	Cross: Cross,
  	Cscr: Cscr,
  	cscr: cscr,
  	csub: csub,
  	csube: csube,
  	csup: csup,
  	csupe: csupe,
  	ctdot: ctdot,
  	cudarrl: cudarrl,
  	cudarrr: cudarrr,
  	cuepr: cuepr,
  	cuesc: cuesc,
  	cularr: cularr,
  	cularrp: cularrp,
  	cupbrcap: cupbrcap,
  	cupcap: cupcap,
  	CupCap: CupCap,
  	cup: cup,
  	Cup: Cup,
  	cupcup: cupcup,
  	cupdot: cupdot,
  	cupor: cupor,
  	cups: cups,
  	curarr: curarr,
  	curarrm: curarrm,
  	curlyeqprec: curlyeqprec,
  	curlyeqsucc: curlyeqsucc,
  	curlyvee: curlyvee,
  	curlywedge: curlywedge,
  	curren: curren,
  	curvearrowleft: curvearrowleft,
  	curvearrowright: curvearrowright,
  	cuvee: cuvee,
  	cuwed: cuwed,
  	cwconint: cwconint,
  	cwint: cwint,
  	cylcty: cylcty,
  	dagger: dagger,
  	Dagger: Dagger,
  	daleth: daleth,
  	darr: darr,
  	Darr: Darr,
  	dArr: dArr,
  	dash: dash,
  	Dashv: Dashv,
  	dashv: dashv,
  	dbkarow: dbkarow,
  	dblac: dblac,
  	Dcaron: Dcaron,
  	dcaron: dcaron,
  	Dcy: Dcy,
  	dcy: dcy,
  	ddagger: ddagger,
  	ddarr: ddarr,
  	DD: DD,
  	dd: dd,
  	DDotrahd: DDotrahd,
  	ddotseq: ddotseq,
  	deg: deg,
  	Del: Del,
  	Delta: Delta,
  	delta: delta,
  	demptyv: demptyv,
  	dfisht: dfisht,
  	Dfr: Dfr,
  	dfr: dfr,
  	dHar: dHar,
  	dharl: dharl,
  	dharr: dharr,
  	DiacriticalAcute: DiacriticalAcute,
  	DiacriticalDot: DiacriticalDot,
  	DiacriticalDoubleAcute: DiacriticalDoubleAcute,
  	DiacriticalGrave: DiacriticalGrave,
  	DiacriticalTilde: DiacriticalTilde,
  	diam: diam,
  	diamond: diamond,
  	Diamond: Diamond,
  	diamondsuit: diamondsuit,
  	diams: diams,
  	die: die,
  	DifferentialD: DifferentialD,
  	digamma: digamma,
  	disin: disin,
  	div: div,
  	divide: divide,
  	divideontimes: divideontimes,
  	divonx: divonx,
  	DJcy: DJcy,
  	djcy: djcy,
  	dlcorn: dlcorn,
  	dlcrop: dlcrop,
  	dollar: dollar,
  	Dopf: Dopf,
  	dopf: dopf,
  	Dot: Dot,
  	dot: dot,
  	DotDot: DotDot,
  	doteq: doteq,
  	doteqdot: doteqdot,
  	DotEqual: DotEqual,
  	dotminus: dotminus,
  	dotplus: dotplus,
  	dotsquare: dotsquare,
  	doublebarwedge: doublebarwedge,
  	DoubleContourIntegral: DoubleContourIntegral,
  	DoubleDot: DoubleDot,
  	DoubleDownArrow: DoubleDownArrow,
  	DoubleLeftArrow: DoubleLeftArrow,
  	DoubleLeftRightArrow: DoubleLeftRightArrow,
  	DoubleLeftTee: DoubleLeftTee,
  	DoubleLongLeftArrow: DoubleLongLeftArrow,
  	DoubleLongLeftRightArrow: DoubleLongLeftRightArrow,
  	DoubleLongRightArrow: DoubleLongRightArrow,
  	DoubleRightArrow: DoubleRightArrow,
  	DoubleRightTee: DoubleRightTee,
  	DoubleUpArrow: DoubleUpArrow,
  	DoubleUpDownArrow: DoubleUpDownArrow,
  	DoubleVerticalBar: DoubleVerticalBar,
  	DownArrowBar: DownArrowBar,
  	downarrow: downarrow,
  	DownArrow: DownArrow,
  	Downarrow: Downarrow,
  	DownArrowUpArrow: DownArrowUpArrow,
  	DownBreve: DownBreve,
  	downdownarrows: downdownarrows,
  	downharpoonleft: downharpoonleft,
  	downharpoonright: downharpoonright,
  	DownLeftRightVector: DownLeftRightVector,
  	DownLeftTeeVector: DownLeftTeeVector,
  	DownLeftVectorBar: DownLeftVectorBar,
  	DownLeftVector: DownLeftVector,
  	DownRightTeeVector: DownRightTeeVector,
  	DownRightVectorBar: DownRightVectorBar,
  	DownRightVector: DownRightVector,
  	DownTeeArrow: DownTeeArrow,
  	DownTee: DownTee,
  	drbkarow: drbkarow,
  	drcorn: drcorn,
  	drcrop: drcrop,
  	Dscr: Dscr,
  	dscr: dscr,
  	DScy: DScy,
  	dscy: dscy,
  	dsol: dsol,
  	Dstrok: Dstrok,
  	dstrok: dstrok,
  	dtdot: dtdot,
  	dtri: dtri,
  	dtrif: dtrif,
  	duarr: duarr,
  	duhar: duhar,
  	dwangle: dwangle,
  	DZcy: DZcy,
  	dzcy: dzcy,
  	dzigrarr: dzigrarr,
  	Eacute: Eacute,
  	eacute: eacute,
  	easter: easter,
  	Ecaron: Ecaron,
  	ecaron: ecaron,
  	Ecirc: Ecirc,
  	ecirc: ecirc,
  	ecir: ecir,
  	ecolon: ecolon,
  	Ecy: Ecy,
  	ecy: ecy,
  	eDDot: eDDot,
  	Edot: Edot,
  	edot: edot,
  	eDot: eDot,
  	ee: ee,
  	efDot: efDot,
  	Efr: Efr,
  	efr: efr,
  	eg: eg,
  	Egrave: Egrave,
  	egrave: egrave,
  	egs: egs,
  	egsdot: egsdot,
  	el: el,
  	Element: Element,
  	elinters: elinters,
  	ell: ell,
  	els: els,
  	elsdot: elsdot,
  	Emacr: Emacr,
  	emacr: emacr,
  	empty: empty,
  	emptyset: emptyset,
  	EmptySmallSquare: EmptySmallSquare,
  	emptyv: emptyv,
  	EmptyVerySmallSquare: EmptyVerySmallSquare,
  	emsp13: emsp13,
  	emsp14: emsp14,
  	emsp: emsp,
  	ENG: ENG,
  	eng: eng,
  	ensp: ensp,
  	Eogon: Eogon,
  	eogon: eogon,
  	Eopf: Eopf,
  	eopf: eopf,
  	epar: epar,
  	eparsl: eparsl,
  	eplus: eplus,
  	epsi: epsi,
  	Epsilon: Epsilon,
  	epsilon: epsilon,
  	epsiv: epsiv,
  	eqcirc: eqcirc,
  	eqcolon: eqcolon,
  	eqsim: eqsim,
  	eqslantgtr: eqslantgtr,
  	eqslantless: eqslantless,
  	Equal: Equal,
  	equals: equals,
  	EqualTilde: EqualTilde,
  	equest: equest,
  	Equilibrium: Equilibrium,
  	equiv: equiv,
  	equivDD: equivDD,
  	eqvparsl: eqvparsl,
  	erarr: erarr,
  	erDot: erDot,
  	escr: escr,
  	Escr: Escr,
  	esdot: esdot,
  	Esim: Esim,
  	esim: esim,
  	Eta: Eta,
  	eta: eta,
  	ETH: ETH,
  	eth: eth,
  	Euml: Euml,
  	euml: euml,
  	euro: euro,
  	excl: excl,
  	exist: exist,
  	Exists: Exists,
  	expectation: expectation,
  	exponentiale: exponentiale,
  	ExponentialE: ExponentialE,
  	fallingdotseq: fallingdotseq,
  	Fcy: Fcy,
  	fcy: fcy,
  	female: female,
  	ffilig: ffilig,
  	fflig: fflig,
  	ffllig: ffllig,
  	Ffr: Ffr,
  	ffr: ffr,
  	filig: filig,
  	FilledSmallSquare: FilledSmallSquare,
  	FilledVerySmallSquare: FilledVerySmallSquare,
  	fjlig: fjlig,
  	flat: flat,
  	fllig: fllig,
  	fltns: fltns,
  	fnof: fnof,
  	Fopf: Fopf,
  	fopf: fopf,
  	forall: forall,
  	ForAll: ForAll,
  	fork: fork,
  	forkv: forkv,
  	Fouriertrf: Fouriertrf,
  	fpartint: fpartint,
  	frac12: frac12,
  	frac13: frac13,
  	frac14: frac14,
  	frac15: frac15,
  	frac16: frac16,
  	frac18: frac18,
  	frac23: frac23,
  	frac25: frac25,
  	frac34: frac34,
  	frac35: frac35,
  	frac38: frac38,
  	frac45: frac45,
  	frac56: frac56,
  	frac58: frac58,
  	frac78: frac78,
  	frasl: frasl,
  	frown: frown,
  	fscr: fscr,
  	Fscr: Fscr,
  	gacute: gacute,
  	Gamma: Gamma,
  	gamma: gamma,
  	Gammad: Gammad,
  	gammad: gammad,
  	gap: gap,
  	Gbreve: Gbreve,
  	gbreve: gbreve,
  	Gcedil: Gcedil,
  	Gcirc: Gcirc,
  	gcirc: gcirc,
  	Gcy: Gcy,
  	gcy: gcy,
  	Gdot: Gdot,
  	gdot: gdot,
  	ge: ge,
  	gE: gE,
  	gEl: gEl,
  	gel: gel,
  	geq: geq,
  	geqq: geqq,
  	geqslant: geqslant,
  	gescc: gescc,
  	ges: ges,
  	gesdot: gesdot,
  	gesdoto: gesdoto,
  	gesdotol: gesdotol,
  	gesl: gesl,
  	gesles: gesles,
  	Gfr: Gfr,
  	gfr: gfr,
  	gg: gg,
  	Gg: Gg,
  	ggg: ggg,
  	gimel: gimel,
  	GJcy: GJcy,
  	gjcy: gjcy,
  	gla: gla,
  	gl: gl,
  	glE: glE,
  	glj: glj,
  	gnap: gnap,
  	gnapprox: gnapprox,
  	gne: gne,
  	gnE: gnE,
  	gneq: gneq,
  	gneqq: gneqq,
  	gnsim: gnsim,
  	Gopf: Gopf,
  	gopf: gopf,
  	grave: grave,
  	GreaterEqual: GreaterEqual,
  	GreaterEqualLess: GreaterEqualLess,
  	GreaterFullEqual: GreaterFullEqual,
  	GreaterGreater: GreaterGreater,
  	GreaterLess: GreaterLess,
  	GreaterSlantEqual: GreaterSlantEqual,
  	GreaterTilde: GreaterTilde,
  	Gscr: Gscr,
  	gscr: gscr,
  	gsim: gsim,
  	gsime: gsime,
  	gsiml: gsiml,
  	gtcc: gtcc,
  	gtcir: gtcir,
  	gt: gt,
  	GT: GT,
  	Gt: Gt,
  	gtdot: gtdot,
  	gtlPar: gtlPar,
  	gtquest: gtquest,
  	gtrapprox: gtrapprox,
  	gtrarr: gtrarr,
  	gtrdot: gtrdot,
  	gtreqless: gtreqless,
  	gtreqqless: gtreqqless,
  	gtrless: gtrless,
  	gtrsim: gtrsim,
  	gvertneqq: gvertneqq,
  	gvnE: gvnE,
  	Hacek: Hacek,
  	hairsp: hairsp,
  	half: half,
  	hamilt: hamilt,
  	HARDcy: HARDcy,
  	hardcy: hardcy,
  	harrcir: harrcir,
  	harr: harr,
  	hArr: hArr,
  	harrw: harrw,
  	Hat: Hat,
  	hbar: hbar,
  	Hcirc: Hcirc,
  	hcirc: hcirc,
  	hearts: hearts,
  	heartsuit: heartsuit,
  	hellip: hellip,
  	hercon: hercon,
  	hfr: hfr,
  	Hfr: Hfr,
  	HilbertSpace: HilbertSpace,
  	hksearow: hksearow,
  	hkswarow: hkswarow,
  	hoarr: hoarr,
  	homtht: homtht,
  	hookleftarrow: hookleftarrow,
  	hookrightarrow: hookrightarrow,
  	hopf: hopf,
  	Hopf: Hopf,
  	horbar: horbar,
  	HorizontalLine: HorizontalLine,
  	hscr: hscr,
  	Hscr: Hscr,
  	hslash: hslash,
  	Hstrok: Hstrok,
  	hstrok: hstrok,
  	HumpDownHump: HumpDownHump,
  	HumpEqual: HumpEqual,
  	hybull: hybull,
  	hyphen: hyphen,
  	Iacute: Iacute,
  	iacute: iacute,
  	ic: ic,
  	Icirc: Icirc,
  	icirc: icirc,
  	Icy: Icy,
  	icy: icy,
  	Idot: Idot,
  	IEcy: IEcy,
  	iecy: iecy,
  	iexcl: iexcl,
  	iff: iff,
  	ifr: ifr,
  	Ifr: Ifr,
  	Igrave: Igrave,
  	igrave: igrave,
  	ii: ii,
  	iiiint: iiiint,
  	iiint: iiint,
  	iinfin: iinfin,
  	iiota: iiota,
  	IJlig: IJlig,
  	ijlig: ijlig,
  	Imacr: Imacr,
  	imacr: imacr,
  	image: image$2,
  	ImaginaryI: ImaginaryI,
  	imagline: imagline,
  	imagpart: imagpart,
  	imath: imath,
  	Im: Im,
  	imof: imof,
  	imped: imped,
  	Implies: Implies,
  	incare: incare,
  	"in": "∈",
  	infin: infin,
  	infintie: infintie,
  	inodot: inodot,
  	intcal: intcal,
  	int: int,
  	Int: Int,
  	integers: integers,
  	Integral: Integral,
  	intercal: intercal,
  	Intersection: Intersection,
  	intlarhk: intlarhk,
  	intprod: intprod,
  	InvisibleComma: InvisibleComma,
  	InvisibleTimes: InvisibleTimes,
  	IOcy: IOcy,
  	iocy: iocy,
  	Iogon: Iogon,
  	iogon: iogon,
  	Iopf: Iopf,
  	iopf: iopf,
  	Iota: Iota,
  	iota: iota,
  	iprod: iprod,
  	iquest: iquest,
  	iscr: iscr,
  	Iscr: Iscr,
  	isin: isin,
  	isindot: isindot,
  	isinE: isinE,
  	isins: isins,
  	isinsv: isinsv,
  	isinv: isinv,
  	it: it,
  	Itilde: Itilde,
  	itilde: itilde,
  	Iukcy: Iukcy,
  	iukcy: iukcy,
  	Iuml: Iuml,
  	iuml: iuml,
  	Jcirc: Jcirc,
  	jcirc: jcirc,
  	Jcy: Jcy,
  	jcy: jcy,
  	Jfr: Jfr,
  	jfr: jfr,
  	jmath: jmath,
  	Jopf: Jopf,
  	jopf: jopf,
  	Jscr: Jscr,
  	jscr: jscr,
  	Jsercy: Jsercy,
  	jsercy: jsercy,
  	Jukcy: Jukcy,
  	jukcy: jukcy,
  	Kappa: Kappa,
  	kappa: kappa,
  	kappav: kappav,
  	Kcedil: Kcedil,
  	kcedil: kcedil,
  	Kcy: Kcy,
  	kcy: kcy,
  	Kfr: Kfr,
  	kfr: kfr,
  	kgreen: kgreen,
  	KHcy: KHcy,
  	khcy: khcy,
  	KJcy: KJcy,
  	kjcy: kjcy,
  	Kopf: Kopf,
  	kopf: kopf,
  	Kscr: Kscr,
  	kscr: kscr,
  	lAarr: lAarr,
  	Lacute: Lacute,
  	lacute: lacute,
  	laemptyv: laemptyv,
  	lagran: lagran,
  	Lambda: Lambda,
  	lambda: lambda,
  	lang: lang,
  	Lang: Lang,
  	langd: langd,
  	langle: langle,
  	lap: lap,
  	Laplacetrf: Laplacetrf,
  	laquo: laquo,
  	larrb: larrb,
  	larrbfs: larrbfs,
  	larr: larr,
  	Larr: Larr,
  	lArr: lArr,
  	larrfs: larrfs,
  	larrhk: larrhk,
  	larrlp: larrlp,
  	larrpl: larrpl,
  	larrsim: larrsim,
  	larrtl: larrtl,
  	latail: latail,
  	lAtail: lAtail,
  	lat: lat,
  	late: late,
  	lates: lates,
  	lbarr: lbarr,
  	lBarr: lBarr,
  	lbbrk: lbbrk,
  	lbrace: lbrace,
  	lbrack: lbrack,
  	lbrke: lbrke,
  	lbrksld: lbrksld,
  	lbrkslu: lbrkslu,
  	Lcaron: Lcaron,
  	lcaron: lcaron,
  	Lcedil: Lcedil,
  	lcedil: lcedil,
  	lceil: lceil,
  	lcub: lcub,
  	Lcy: Lcy,
  	lcy: lcy,
  	ldca: ldca,
  	ldquo: ldquo,
  	ldquor: ldquor,
  	ldrdhar: ldrdhar,
  	ldrushar: ldrushar,
  	ldsh: ldsh,
  	le: le,
  	lE: lE,
  	LeftAngleBracket: LeftAngleBracket,
  	LeftArrowBar: LeftArrowBar,
  	leftarrow: leftarrow,
  	LeftArrow: LeftArrow,
  	Leftarrow: Leftarrow,
  	LeftArrowRightArrow: LeftArrowRightArrow,
  	leftarrowtail: leftarrowtail,
  	LeftCeiling: LeftCeiling,
  	LeftDoubleBracket: LeftDoubleBracket,
  	LeftDownTeeVector: LeftDownTeeVector,
  	LeftDownVectorBar: LeftDownVectorBar,
  	LeftDownVector: LeftDownVector,
  	LeftFloor: LeftFloor,
  	leftharpoondown: leftharpoondown,
  	leftharpoonup: leftharpoonup,
  	leftleftarrows: leftleftarrows,
  	leftrightarrow: leftrightarrow,
  	LeftRightArrow: LeftRightArrow,
  	Leftrightarrow: Leftrightarrow,
  	leftrightarrows: leftrightarrows,
  	leftrightharpoons: leftrightharpoons,
  	leftrightsquigarrow: leftrightsquigarrow,
  	LeftRightVector: LeftRightVector,
  	LeftTeeArrow: LeftTeeArrow,
  	LeftTee: LeftTee,
  	LeftTeeVector: LeftTeeVector,
  	leftthreetimes: leftthreetimes,
  	LeftTriangleBar: LeftTriangleBar,
  	LeftTriangle: LeftTriangle,
  	LeftTriangleEqual: LeftTriangleEqual,
  	LeftUpDownVector: LeftUpDownVector,
  	LeftUpTeeVector: LeftUpTeeVector,
  	LeftUpVectorBar: LeftUpVectorBar,
  	LeftUpVector: LeftUpVector,
  	LeftVectorBar: LeftVectorBar,
  	LeftVector: LeftVector,
  	lEg: lEg,
  	leg: leg,
  	leq: leq,
  	leqq: leqq,
  	leqslant: leqslant,
  	lescc: lescc,
  	les: les,
  	lesdot: lesdot,
  	lesdoto: lesdoto,
  	lesdotor: lesdotor,
  	lesg: lesg,
  	lesges: lesges,
  	lessapprox: lessapprox,
  	lessdot: lessdot,
  	lesseqgtr: lesseqgtr,
  	lesseqqgtr: lesseqqgtr,
  	LessEqualGreater: LessEqualGreater,
  	LessFullEqual: LessFullEqual,
  	LessGreater: LessGreater,
  	lessgtr: lessgtr,
  	LessLess: LessLess,
  	lesssim: lesssim,
  	LessSlantEqual: LessSlantEqual,
  	LessTilde: LessTilde,
  	lfisht: lfisht,
  	lfloor: lfloor,
  	Lfr: Lfr,
  	lfr: lfr,
  	lg: lg,
  	lgE: lgE,
  	lHar: lHar,
  	lhard: lhard,
  	lharu: lharu,
  	lharul: lharul,
  	lhblk: lhblk,
  	LJcy: LJcy,
  	ljcy: ljcy,
  	llarr: llarr,
  	ll: ll,
  	Ll: Ll,
  	llcorner: llcorner,
  	Lleftarrow: Lleftarrow,
  	llhard: llhard,
  	lltri: lltri,
  	Lmidot: Lmidot,
  	lmidot: lmidot,
  	lmoustache: lmoustache,
  	lmoust: lmoust,
  	lnap: lnap,
  	lnapprox: lnapprox,
  	lne: lne,
  	lnE: lnE,
  	lneq: lneq,
  	lneqq: lneqq,
  	lnsim: lnsim,
  	loang: loang,
  	loarr: loarr,
  	lobrk: lobrk,
  	longleftarrow: longleftarrow,
  	LongLeftArrow: LongLeftArrow,
  	Longleftarrow: Longleftarrow,
  	longleftrightarrow: longleftrightarrow,
  	LongLeftRightArrow: LongLeftRightArrow,
  	Longleftrightarrow: Longleftrightarrow,
  	longmapsto: longmapsto,
  	longrightarrow: longrightarrow,
  	LongRightArrow: LongRightArrow,
  	Longrightarrow: Longrightarrow,
  	looparrowleft: looparrowleft,
  	looparrowright: looparrowright,
  	lopar: lopar,
  	Lopf: Lopf,
  	lopf: lopf,
  	loplus: loplus,
  	lotimes: lotimes,
  	lowast: lowast,
  	lowbar: lowbar,
  	LowerLeftArrow: LowerLeftArrow,
  	LowerRightArrow: LowerRightArrow,
  	loz: loz,
  	lozenge: lozenge,
  	lozf: lozf,
  	lpar: lpar,
  	lparlt: lparlt,
  	lrarr: lrarr,
  	lrcorner: lrcorner,
  	lrhar: lrhar,
  	lrhard: lrhard,
  	lrm: lrm,
  	lrtri: lrtri,
  	lsaquo: lsaquo,
  	lscr: lscr,
  	Lscr: Lscr,
  	lsh: lsh,
  	Lsh: Lsh,
  	lsim: lsim,
  	lsime: lsime,
  	lsimg: lsimg,
  	lsqb: lsqb,
  	lsquo: lsquo,
  	lsquor: lsquor,
  	Lstrok: Lstrok,
  	lstrok: lstrok,
  	ltcc: ltcc,
  	ltcir: ltcir,
  	lt: lt,
  	LT: LT,
  	Lt: Lt,
  	ltdot: ltdot,
  	lthree: lthree,
  	ltimes: ltimes,
  	ltlarr: ltlarr,
  	ltquest: ltquest,
  	ltri: ltri,
  	ltrie: ltrie,
  	ltrif: ltrif,
  	ltrPar: ltrPar,
  	lurdshar: lurdshar,
  	luruhar: luruhar,
  	lvertneqq: lvertneqq,
  	lvnE: lvnE,
  	macr: macr,
  	male: male,
  	malt: malt,
  	maltese: maltese,
  	"Map": "⤅",
  	map: map,
  	mapsto: mapsto,
  	mapstodown: mapstodown,
  	mapstoleft: mapstoleft,
  	mapstoup: mapstoup,
  	marker: marker,
  	mcomma: mcomma,
  	Mcy: Mcy,
  	mcy: mcy,
  	mdash: mdash,
  	mDDot: mDDot,
  	measuredangle: measuredangle,
  	MediumSpace: MediumSpace,
  	Mellintrf: Mellintrf,
  	Mfr: Mfr,
  	mfr: mfr,
  	mho: mho,
  	micro: micro,
  	midast: midast,
  	midcir: midcir,
  	mid: mid,
  	middot: middot,
  	minusb: minusb,
  	minus: minus,
  	minusd: minusd,
  	minusdu: minusdu,
  	MinusPlus: MinusPlus,
  	mlcp: mlcp,
  	mldr: mldr,
  	mnplus: mnplus,
  	models: models,
  	Mopf: Mopf,
  	mopf: mopf,
  	mp: mp,
  	mscr: mscr,
  	Mscr: Mscr,
  	mstpos: mstpos,
  	Mu: Mu,
  	mu: mu,
  	multimap: multimap,
  	mumap: mumap,
  	nabla: nabla,
  	Nacute: Nacute,
  	nacute: nacute,
  	nang: nang,
  	nap: nap,
  	napE: napE,
  	napid: napid,
  	napos: napos,
  	napprox: napprox,
  	natural: natural,
  	naturals: naturals,
  	natur: natur,
  	nbsp: nbsp,
  	nbump: nbump,
  	nbumpe: nbumpe,
  	ncap: ncap,
  	Ncaron: Ncaron,
  	ncaron: ncaron,
  	Ncedil: Ncedil,
  	ncedil: ncedil,
  	ncong: ncong,
  	ncongdot: ncongdot,
  	ncup: ncup,
  	Ncy: Ncy,
  	ncy: ncy,
  	ndash: ndash,
  	nearhk: nearhk,
  	nearr: nearr,
  	neArr: neArr,
  	nearrow: nearrow,
  	ne: ne,
  	nedot: nedot,
  	NegativeMediumSpace: NegativeMediumSpace,
  	NegativeThickSpace: NegativeThickSpace,
  	NegativeThinSpace: NegativeThinSpace,
  	NegativeVeryThinSpace: NegativeVeryThinSpace,
  	nequiv: nequiv,
  	nesear: nesear,
  	nesim: nesim,
  	NestedGreaterGreater: NestedGreaterGreater,
  	NestedLessLess: NestedLessLess,
  	NewLine: NewLine,
  	nexist: nexist,
  	nexists: nexists,
  	Nfr: Nfr,
  	nfr: nfr,
  	ngE: ngE,
  	nge: nge,
  	ngeq: ngeq,
  	ngeqq: ngeqq,
  	ngeqslant: ngeqslant,
  	nges: nges,
  	nGg: nGg,
  	ngsim: ngsim,
  	nGt: nGt,
  	ngt: ngt,
  	ngtr: ngtr,
  	nGtv: nGtv,
  	nharr: nharr,
  	nhArr: nhArr,
  	nhpar: nhpar,
  	ni: ni,
  	nis: nis,
  	nisd: nisd,
  	niv: niv,
  	NJcy: NJcy,
  	njcy: njcy,
  	nlarr: nlarr,
  	nlArr: nlArr,
  	nldr: nldr,
  	nlE: nlE,
  	nle: nle,
  	nleftarrow: nleftarrow,
  	nLeftarrow: nLeftarrow,
  	nleftrightarrow: nleftrightarrow,
  	nLeftrightarrow: nLeftrightarrow,
  	nleq: nleq,
  	nleqq: nleqq,
  	nleqslant: nleqslant,
  	nles: nles,
  	nless: nless,
  	nLl: nLl,
  	nlsim: nlsim,
  	nLt: nLt,
  	nlt: nlt,
  	nltri: nltri,
  	nltrie: nltrie,
  	nLtv: nLtv,
  	nmid: nmid,
  	NoBreak: NoBreak,
  	NonBreakingSpace: NonBreakingSpace,
  	nopf: nopf,
  	Nopf: Nopf,
  	Not: Not,
  	not: not,
  	NotCongruent: NotCongruent,
  	NotCupCap: NotCupCap,
  	NotDoubleVerticalBar: NotDoubleVerticalBar,
  	NotElement: NotElement,
  	NotEqual: NotEqual,
  	NotEqualTilde: NotEqualTilde,
  	NotExists: NotExists,
  	NotGreater: NotGreater,
  	NotGreaterEqual: NotGreaterEqual,
  	NotGreaterFullEqual: NotGreaterFullEqual,
  	NotGreaterGreater: NotGreaterGreater,
  	NotGreaterLess: NotGreaterLess,
  	NotGreaterSlantEqual: NotGreaterSlantEqual,
  	NotGreaterTilde: NotGreaterTilde,
  	NotHumpDownHump: NotHumpDownHump,
  	NotHumpEqual: NotHumpEqual,
  	notin: notin,
  	notindot: notindot,
  	notinE: notinE,
  	notinva: notinva,
  	notinvb: notinvb,
  	notinvc: notinvc,
  	NotLeftTriangleBar: NotLeftTriangleBar,
  	NotLeftTriangle: NotLeftTriangle,
  	NotLeftTriangleEqual: NotLeftTriangleEqual,
  	NotLess: NotLess,
  	NotLessEqual: NotLessEqual,
  	NotLessGreater: NotLessGreater,
  	NotLessLess: NotLessLess,
  	NotLessSlantEqual: NotLessSlantEqual,
  	NotLessTilde: NotLessTilde,
  	NotNestedGreaterGreater: NotNestedGreaterGreater,
  	NotNestedLessLess: NotNestedLessLess,
  	notni: notni,
  	notniva: notniva,
  	notnivb: notnivb,
  	notnivc: notnivc,
  	NotPrecedes: NotPrecedes,
  	NotPrecedesEqual: NotPrecedesEqual,
  	NotPrecedesSlantEqual: NotPrecedesSlantEqual,
  	NotReverseElement: NotReverseElement,
  	NotRightTriangleBar: NotRightTriangleBar,
  	NotRightTriangle: NotRightTriangle,
  	NotRightTriangleEqual: NotRightTriangleEqual,
  	NotSquareSubset: NotSquareSubset,
  	NotSquareSubsetEqual: NotSquareSubsetEqual,
  	NotSquareSuperset: NotSquareSuperset,
  	NotSquareSupersetEqual: NotSquareSupersetEqual,
  	NotSubset: NotSubset,
  	NotSubsetEqual: NotSubsetEqual,
  	NotSucceeds: NotSucceeds,
  	NotSucceedsEqual: NotSucceedsEqual,
  	NotSucceedsSlantEqual: NotSucceedsSlantEqual,
  	NotSucceedsTilde: NotSucceedsTilde,
  	NotSuperset: NotSuperset,
  	NotSupersetEqual: NotSupersetEqual,
  	NotTilde: NotTilde,
  	NotTildeEqual: NotTildeEqual,
  	NotTildeFullEqual: NotTildeFullEqual,
  	NotTildeTilde: NotTildeTilde,
  	NotVerticalBar: NotVerticalBar,
  	nparallel: nparallel,
  	npar: npar,
  	nparsl: nparsl,
  	npart: npart,
  	npolint: npolint,
  	npr: npr,
  	nprcue: nprcue,
  	nprec: nprec,
  	npreceq: npreceq,
  	npre: npre,
  	nrarrc: nrarrc,
  	nrarr: nrarr,
  	nrArr: nrArr,
  	nrarrw: nrarrw,
  	nrightarrow: nrightarrow,
  	nRightarrow: nRightarrow,
  	nrtri: nrtri,
  	nrtrie: nrtrie,
  	nsc: nsc,
  	nsccue: nsccue,
  	nsce: nsce,
  	Nscr: Nscr,
  	nscr: nscr,
  	nshortmid: nshortmid,
  	nshortparallel: nshortparallel,
  	nsim: nsim,
  	nsime: nsime,
  	nsimeq: nsimeq,
  	nsmid: nsmid,
  	nspar: nspar,
  	nsqsube: nsqsube,
  	nsqsupe: nsqsupe,
  	nsub: nsub,
  	nsubE: nsubE,
  	nsube: nsube,
  	nsubset: nsubset,
  	nsubseteq: nsubseteq,
  	nsubseteqq: nsubseteqq,
  	nsucc: nsucc,
  	nsucceq: nsucceq,
  	nsup: nsup,
  	nsupE: nsupE,
  	nsupe: nsupe,
  	nsupset: nsupset,
  	nsupseteq: nsupseteq,
  	nsupseteqq: nsupseteqq,
  	ntgl: ntgl,
  	Ntilde: Ntilde,
  	ntilde: ntilde,
  	ntlg: ntlg,
  	ntriangleleft: ntriangleleft,
  	ntrianglelefteq: ntrianglelefteq,
  	ntriangleright: ntriangleright,
  	ntrianglerighteq: ntrianglerighteq,
  	Nu: Nu,
  	nu: nu,
  	num: num,
  	numero: numero,
  	numsp: numsp,
  	nvap: nvap,
  	nvdash: nvdash,
  	nvDash: nvDash,
  	nVdash: nVdash,
  	nVDash: nVDash,
  	nvge: nvge,
  	nvgt: nvgt,
  	nvHarr: nvHarr,
  	nvinfin: nvinfin,
  	nvlArr: nvlArr,
  	nvle: nvle,
  	nvlt: nvlt,
  	nvltrie: nvltrie,
  	nvrArr: nvrArr,
  	nvrtrie: nvrtrie,
  	nvsim: nvsim,
  	nwarhk: nwarhk,
  	nwarr: nwarr,
  	nwArr: nwArr,
  	nwarrow: nwarrow,
  	nwnear: nwnear,
  	Oacute: Oacute,
  	oacute: oacute,
  	oast: oast,
  	Ocirc: Ocirc,
  	ocirc: ocirc,
  	ocir: ocir,
  	Ocy: Ocy,
  	ocy: ocy,
  	odash: odash,
  	Odblac: Odblac,
  	odblac: odblac,
  	odiv: odiv,
  	odot: odot,
  	odsold: odsold,
  	OElig: OElig,
  	oelig: oelig,
  	ofcir: ofcir,
  	Ofr: Ofr,
  	ofr: ofr,
  	ogon: ogon,
  	Ograve: Ograve,
  	ograve: ograve,
  	ogt: ogt,
  	ohbar: ohbar,
  	ohm: ohm,
  	oint: oint,
  	olarr: olarr,
  	olcir: olcir,
  	olcross: olcross,
  	oline: oline,
  	olt: olt,
  	Omacr: Omacr,
  	omacr: omacr,
  	Omega: Omega,
  	omega: omega,
  	Omicron: Omicron,
  	omicron: omicron,
  	omid: omid,
  	ominus: ominus,
  	Oopf: Oopf,
  	oopf: oopf,
  	opar: opar,
  	OpenCurlyDoubleQuote: OpenCurlyDoubleQuote,
  	OpenCurlyQuote: OpenCurlyQuote,
  	operp: operp,
  	oplus: oplus,
  	orarr: orarr,
  	Or: Or,
  	or: or,
  	ord: ord,
  	order: order,
  	orderof: orderof,
  	ordf: ordf,
  	ordm: ordm,
  	origof: origof,
  	oror: oror,
  	orslope: orslope,
  	orv: orv,
  	oS: oS,
  	Oscr: Oscr,
  	oscr: oscr,
  	Oslash: Oslash,
  	oslash: oslash,
  	osol: osol,
  	Otilde: Otilde,
  	otilde: otilde,
  	otimesas: otimesas,
  	Otimes: Otimes,
  	otimes: otimes,
  	Ouml: Ouml,
  	ouml: ouml,
  	ovbar: ovbar,
  	OverBar: OverBar,
  	OverBrace: OverBrace,
  	OverBracket: OverBracket,
  	OverParenthesis: OverParenthesis,
  	para: para,
  	parallel: parallel,
  	par: par,
  	parsim: parsim,
  	parsl: parsl,
  	part: part,
  	PartialD: PartialD,
  	Pcy: Pcy,
  	pcy: pcy,
  	percnt: percnt,
  	period: period,
  	permil: permil,
  	perp: perp,
  	pertenk: pertenk,
  	Pfr: Pfr,
  	pfr: pfr,
  	Phi: Phi,
  	phi: phi,
  	phiv: phiv,
  	phmmat: phmmat,
  	phone: phone,
  	Pi: Pi,
  	pi: pi,
  	pitchfork: pitchfork,
  	piv: piv,
  	planck: planck,
  	planckh: planckh,
  	plankv: plankv,
  	plusacir: plusacir,
  	plusb: plusb,
  	pluscir: pluscir,
  	plus: plus,
  	plusdo: plusdo,
  	plusdu: plusdu,
  	pluse: pluse,
  	PlusMinus: PlusMinus,
  	plusmn: plusmn,
  	plussim: plussim,
  	plustwo: plustwo,
  	pm: pm,
  	Poincareplane: Poincareplane,
  	pointint: pointint,
  	popf: popf,
  	Popf: Popf,
  	pound: pound,
  	prap: prap,
  	Pr: Pr,
  	pr: pr,
  	prcue: prcue,
  	precapprox: precapprox,
  	prec: prec,
  	preccurlyeq: preccurlyeq,
  	Precedes: Precedes,
  	PrecedesEqual: PrecedesEqual,
  	PrecedesSlantEqual: PrecedesSlantEqual,
  	PrecedesTilde: PrecedesTilde,
  	preceq: preceq,
  	precnapprox: precnapprox,
  	precneqq: precneqq,
  	precnsim: precnsim,
  	pre: pre,
  	prE: prE,
  	precsim: precsim,
  	prime: prime,
  	Prime: Prime,
  	primes: primes,
  	prnap: prnap,
  	prnE: prnE,
  	prnsim: prnsim,
  	prod: prod,
  	Product: Product,
  	profalar: profalar,
  	profline: profline,
  	profsurf: profsurf,
  	prop: prop,
  	Proportional: Proportional,
  	Proportion: Proportion,
  	propto: propto,
  	prsim: prsim,
  	prurel: prurel,
  	Pscr: Pscr,
  	pscr: pscr,
  	Psi: Psi,
  	psi: psi,
  	puncsp: puncsp,
  	Qfr: Qfr,
  	qfr: qfr,
  	qint: qint,
  	qopf: qopf,
  	Qopf: Qopf,
  	qprime: qprime,
  	Qscr: Qscr,
  	qscr: qscr,
  	quaternions: quaternions,
  	quatint: quatint,
  	quest: quest,
  	questeq: questeq,
  	quot: quot,
  	QUOT: QUOT,
  	rAarr: rAarr,
  	race: race,
  	Racute: Racute,
  	racute: racute,
  	radic: radic,
  	raemptyv: raemptyv,
  	rang: rang,
  	Rang: Rang,
  	rangd: rangd,
  	range: range,
  	rangle: rangle,
  	raquo: raquo,
  	rarrap: rarrap,
  	rarrb: rarrb,
  	rarrbfs: rarrbfs,
  	rarrc: rarrc,
  	rarr: rarr,
  	Rarr: Rarr,
  	rArr: rArr,
  	rarrfs: rarrfs,
  	rarrhk: rarrhk,
  	rarrlp: rarrlp,
  	rarrpl: rarrpl,
  	rarrsim: rarrsim,
  	Rarrtl: Rarrtl,
  	rarrtl: rarrtl,
  	rarrw: rarrw,
  	ratail: ratail,
  	rAtail: rAtail,
  	ratio: ratio,
  	rationals: rationals,
  	rbarr: rbarr,
  	rBarr: rBarr,
  	RBarr: RBarr,
  	rbbrk: rbbrk,
  	rbrace: rbrace,
  	rbrack: rbrack,
  	rbrke: rbrke,
  	rbrksld: rbrksld,
  	rbrkslu: rbrkslu,
  	Rcaron: Rcaron,
  	rcaron: rcaron,
  	Rcedil: Rcedil,
  	rcedil: rcedil,
  	rceil: rceil,
  	rcub: rcub,
  	Rcy: Rcy,
  	rcy: rcy,
  	rdca: rdca,
  	rdldhar: rdldhar,
  	rdquo: rdquo,
  	rdquor: rdquor,
  	rdsh: rdsh,
  	real: real,
  	realine: realine,
  	realpart: realpart,
  	reals: reals,
  	Re: Re,
  	rect: rect,
  	reg: reg,
  	REG: REG,
  	ReverseElement: ReverseElement,
  	ReverseEquilibrium: ReverseEquilibrium,
  	ReverseUpEquilibrium: ReverseUpEquilibrium,
  	rfisht: rfisht,
  	rfloor: rfloor,
  	rfr: rfr,
  	Rfr: Rfr,
  	rHar: rHar,
  	rhard: rhard,
  	rharu: rharu,
  	rharul: rharul,
  	Rho: Rho,
  	rho: rho,
  	rhov: rhov,
  	RightAngleBracket: RightAngleBracket,
  	RightArrowBar: RightArrowBar,
  	rightarrow: rightarrow,
  	RightArrow: RightArrow,
  	Rightarrow: Rightarrow,
  	RightArrowLeftArrow: RightArrowLeftArrow,
  	rightarrowtail: rightarrowtail,
  	RightCeiling: RightCeiling,
  	RightDoubleBracket: RightDoubleBracket,
  	RightDownTeeVector: RightDownTeeVector,
  	RightDownVectorBar: RightDownVectorBar,
  	RightDownVector: RightDownVector,
  	RightFloor: RightFloor,
  	rightharpoondown: rightharpoondown,
  	rightharpoonup: rightharpoonup,
  	rightleftarrows: rightleftarrows,
  	rightleftharpoons: rightleftharpoons,
  	rightrightarrows: rightrightarrows,
  	rightsquigarrow: rightsquigarrow,
  	RightTeeArrow: RightTeeArrow,
  	RightTee: RightTee,
  	RightTeeVector: RightTeeVector,
  	rightthreetimes: rightthreetimes,
  	RightTriangleBar: RightTriangleBar,
  	RightTriangle: RightTriangle,
  	RightTriangleEqual: RightTriangleEqual,
  	RightUpDownVector: RightUpDownVector,
  	RightUpTeeVector: RightUpTeeVector,
  	RightUpVectorBar: RightUpVectorBar,
  	RightUpVector: RightUpVector,
  	RightVectorBar: RightVectorBar,
  	RightVector: RightVector,
  	ring: ring,
  	risingdotseq: risingdotseq,
  	rlarr: rlarr,
  	rlhar: rlhar,
  	rlm: rlm,
  	rmoustache: rmoustache,
  	rmoust: rmoust,
  	rnmid: rnmid,
  	roang: roang,
  	roarr: roarr,
  	robrk: robrk,
  	ropar: ropar,
  	ropf: ropf,
  	Ropf: Ropf,
  	roplus: roplus,
  	rotimes: rotimes,
  	RoundImplies: RoundImplies,
  	rpar: rpar,
  	rpargt: rpargt,
  	rppolint: rppolint,
  	rrarr: rrarr,
  	Rrightarrow: Rrightarrow,
  	rsaquo: rsaquo,
  	rscr: rscr,
  	Rscr: Rscr,
  	rsh: rsh,
  	Rsh: Rsh,
  	rsqb: rsqb,
  	rsquo: rsquo,
  	rsquor: rsquor,
  	rthree: rthree,
  	rtimes: rtimes,
  	rtri: rtri,
  	rtrie: rtrie,
  	rtrif: rtrif,
  	rtriltri: rtriltri,
  	RuleDelayed: RuleDelayed,
  	ruluhar: ruluhar,
  	rx: rx,
  	Sacute: Sacute,
  	sacute: sacute,
  	sbquo: sbquo,
  	scap: scap,
  	Scaron: Scaron,
  	scaron: scaron,
  	Sc: Sc,
  	sc: sc,
  	sccue: sccue,
  	sce: sce,
  	scE: scE,
  	Scedil: Scedil,
  	scedil: scedil,
  	Scirc: Scirc,
  	scirc: scirc,
  	scnap: scnap,
  	scnE: scnE,
  	scnsim: scnsim,
  	scpolint: scpolint,
  	scsim: scsim,
  	Scy: Scy,
  	scy: scy,
  	sdotb: sdotb,
  	sdot: sdot,
  	sdote: sdote,
  	searhk: searhk,
  	searr: searr,
  	seArr: seArr,
  	searrow: searrow,
  	sect: sect,
  	semi: semi,
  	seswar: seswar,
  	setminus: setminus,
  	setmn: setmn,
  	sext: sext,
  	Sfr: Sfr,
  	sfr: sfr,
  	sfrown: sfrown,
  	sharp: sharp,
  	SHCHcy: SHCHcy,
  	shchcy: shchcy,
  	SHcy: SHcy,
  	shcy: shcy,
  	ShortDownArrow: ShortDownArrow,
  	ShortLeftArrow: ShortLeftArrow,
  	shortmid: shortmid,
  	shortparallel: shortparallel,
  	ShortRightArrow: ShortRightArrow,
  	ShortUpArrow: ShortUpArrow,
  	shy: shy,
  	Sigma: Sigma,
  	sigma: sigma,
  	sigmaf: sigmaf,
  	sigmav: sigmav,
  	sim: sim,
  	simdot: simdot,
  	sime: sime,
  	simeq: simeq,
  	simg: simg,
  	simgE: simgE,
  	siml: siml,
  	simlE: simlE,
  	simne: simne,
  	simplus: simplus,
  	simrarr: simrarr,
  	slarr: slarr,
  	SmallCircle: SmallCircle,
  	smallsetminus: smallsetminus,
  	smashp: smashp,
  	smeparsl: smeparsl,
  	smid: smid,
  	smile: smile,
  	smt: smt,
  	smte: smte,
  	smtes: smtes,
  	SOFTcy: SOFTcy,
  	softcy: softcy,
  	solbar: solbar,
  	solb: solb,
  	sol: sol,
  	Sopf: Sopf,
  	sopf: sopf,
  	spades: spades,
  	spadesuit: spadesuit,
  	spar: spar,
  	sqcap: sqcap,
  	sqcaps: sqcaps,
  	sqcup: sqcup,
  	sqcups: sqcups,
  	Sqrt: Sqrt,
  	sqsub: sqsub,
  	sqsube: sqsube,
  	sqsubset: sqsubset,
  	sqsubseteq: sqsubseteq,
  	sqsup: sqsup,
  	sqsupe: sqsupe,
  	sqsupset: sqsupset,
  	sqsupseteq: sqsupseteq,
  	square: square,
  	Square: Square,
  	SquareIntersection: SquareIntersection,
  	SquareSubset: SquareSubset,
  	SquareSubsetEqual: SquareSubsetEqual,
  	SquareSuperset: SquareSuperset,
  	SquareSupersetEqual: SquareSupersetEqual,
  	SquareUnion: SquareUnion,
  	squarf: squarf,
  	squ: squ,
  	squf: squf,
  	srarr: srarr,
  	Sscr: Sscr,
  	sscr: sscr,
  	ssetmn: ssetmn,
  	ssmile: ssmile,
  	sstarf: sstarf,
  	Star: Star,
  	star: star,
  	starf: starf,
  	straightepsilon: straightepsilon,
  	straightphi: straightphi,
  	strns: strns,
  	sub: sub,
  	Sub: Sub,
  	subdot: subdot,
  	subE: subE,
  	sube: sube,
  	subedot: subedot,
  	submult: submult,
  	subnE: subnE,
  	subne: subne,
  	subplus: subplus,
  	subrarr: subrarr,
  	subset: subset,
  	Subset: Subset,
  	subseteq: subseteq,
  	subseteqq: subseteqq,
  	SubsetEqual: SubsetEqual,
  	subsetneq: subsetneq,
  	subsetneqq: subsetneqq,
  	subsim: subsim,
  	subsub: subsub,
  	subsup: subsup,
  	succapprox: succapprox,
  	succ: succ,
  	succcurlyeq: succcurlyeq,
  	Succeeds: Succeeds,
  	SucceedsEqual: SucceedsEqual,
  	SucceedsSlantEqual: SucceedsSlantEqual,
  	SucceedsTilde: SucceedsTilde,
  	succeq: succeq,
  	succnapprox: succnapprox,
  	succneqq: succneqq,
  	succnsim: succnsim,
  	succsim: succsim,
  	SuchThat: SuchThat,
  	sum: sum,
  	Sum: Sum,
  	sung: sung,
  	sup1: sup1,
  	sup2: sup2,
  	sup3: sup3,
  	sup: sup,
  	Sup: Sup,
  	supdot: supdot,
  	supdsub: supdsub,
  	supE: supE,
  	supe: supe,
  	supedot: supedot,
  	Superset: Superset,
  	SupersetEqual: SupersetEqual,
  	suphsol: suphsol,
  	suphsub: suphsub,
  	suplarr: suplarr,
  	supmult: supmult,
  	supnE: supnE,
  	supne: supne,
  	supplus: supplus,
  	supset: supset,
  	Supset: Supset,
  	supseteq: supseteq,
  	supseteqq: supseteqq,
  	supsetneq: supsetneq,
  	supsetneqq: supsetneqq,
  	supsim: supsim,
  	supsub: supsub,
  	supsup: supsup,
  	swarhk: swarhk,
  	swarr: swarr,
  	swArr: swArr,
  	swarrow: swarrow,
  	swnwar: swnwar,
  	szlig: szlig,
  	Tab: Tab,
  	target: target,
  	Tau: Tau,
  	tau: tau,
  	tbrk: tbrk,
  	Tcaron: Tcaron,
  	tcaron: tcaron,
  	Tcedil: Tcedil,
  	tcedil: tcedil,
  	Tcy: Tcy,
  	tcy: tcy,
  	tdot: tdot,
  	telrec: telrec,
  	Tfr: Tfr,
  	tfr: tfr,
  	there4: there4,
  	therefore: therefore,
  	Therefore: Therefore,
  	Theta: Theta,
  	theta: theta,
  	thetasym: thetasym,
  	thetav: thetav,
  	thickapprox: thickapprox,
  	thicksim: thicksim,
  	ThickSpace: ThickSpace,
  	ThinSpace: ThinSpace,
  	thinsp: thinsp,
  	thkap: thkap,
  	thksim: thksim,
  	THORN: THORN,
  	thorn: thorn,
  	tilde: tilde,
  	Tilde: Tilde,
  	TildeEqual: TildeEqual,
  	TildeFullEqual: TildeFullEqual,
  	TildeTilde: TildeTilde,
  	timesbar: timesbar,
  	timesb: timesb,
  	times: times,
  	timesd: timesd,
  	tint: tint,
  	toea: toea,
  	topbot: topbot,
  	topcir: topcir,
  	top: top,
  	Topf: Topf,
  	topf: topf,
  	topfork: topfork,
  	tosa: tosa,
  	tprime: tprime,
  	trade: trade,
  	TRADE: TRADE,
  	triangle: triangle,
  	triangledown: triangledown,
  	triangleleft: triangleleft,
  	trianglelefteq: trianglelefteq,
  	triangleq: triangleq,
  	triangleright: triangleright,
  	trianglerighteq: trianglerighteq,
  	tridot: tridot,
  	trie: trie,
  	triminus: triminus,
  	TripleDot: TripleDot,
  	triplus: triplus,
  	trisb: trisb,
  	tritime: tritime,
  	trpezium: trpezium,
  	Tscr: Tscr,
  	tscr: tscr,
  	TScy: TScy,
  	tscy: tscy,
  	TSHcy: TSHcy,
  	tshcy: tshcy,
  	Tstrok: Tstrok,
  	tstrok: tstrok,
  	twixt: twixt,
  	twoheadleftarrow: twoheadleftarrow,
  	twoheadrightarrow: twoheadrightarrow,
  	Uacute: Uacute,
  	uacute: uacute,
  	uarr: uarr,
  	Uarr: Uarr,
  	uArr: uArr,
  	Uarrocir: Uarrocir,
  	Ubrcy: Ubrcy,
  	ubrcy: ubrcy,
  	Ubreve: Ubreve,
  	ubreve: ubreve,
  	Ucirc: Ucirc,
  	ucirc: ucirc,
  	Ucy: Ucy,
  	ucy: ucy,
  	udarr: udarr,
  	Udblac: Udblac,
  	udblac: udblac,
  	udhar: udhar,
  	ufisht: ufisht,
  	Ufr: Ufr,
  	ufr: ufr,
  	Ugrave: Ugrave,
  	ugrave: ugrave,
  	uHar: uHar,
  	uharl: uharl,
  	uharr: uharr,
  	uhblk: uhblk,
  	ulcorn: ulcorn,
  	ulcorner: ulcorner,
  	ulcrop: ulcrop,
  	ultri: ultri,
  	Umacr: Umacr,
  	umacr: umacr,
  	uml: uml,
  	UnderBar: UnderBar,
  	UnderBrace: UnderBrace,
  	UnderBracket: UnderBracket,
  	UnderParenthesis: UnderParenthesis,
  	Union: Union,
  	UnionPlus: UnionPlus,
  	Uogon: Uogon,
  	uogon: uogon,
  	Uopf: Uopf,
  	uopf: uopf,
  	UpArrowBar: UpArrowBar,
  	uparrow: uparrow,
  	UpArrow: UpArrow,
  	Uparrow: Uparrow,
  	UpArrowDownArrow: UpArrowDownArrow,
  	updownarrow: updownarrow,
  	UpDownArrow: UpDownArrow,
  	Updownarrow: Updownarrow,
  	UpEquilibrium: UpEquilibrium,
  	upharpoonleft: upharpoonleft,
  	upharpoonright: upharpoonright,
  	uplus: uplus,
  	UpperLeftArrow: UpperLeftArrow,
  	UpperRightArrow: UpperRightArrow,
  	upsi: upsi,
  	Upsi: Upsi,
  	upsih: upsih,
  	Upsilon: Upsilon,
  	upsilon: upsilon,
  	UpTeeArrow: UpTeeArrow,
  	UpTee: UpTee,
  	upuparrows: upuparrows,
  	urcorn: urcorn,
  	urcorner: urcorner,
  	urcrop: urcrop,
  	Uring: Uring,
  	uring: uring,
  	urtri: urtri,
  	Uscr: Uscr,
  	uscr: uscr,
  	utdot: utdot,
  	Utilde: Utilde,
  	utilde: utilde,
  	utri: utri,
  	utrif: utrif,
  	uuarr: uuarr,
  	Uuml: Uuml,
  	uuml: uuml,
  	uwangle: uwangle,
  	vangrt: vangrt,
  	varepsilon: varepsilon,
  	varkappa: varkappa,
  	varnothing: varnothing,
  	varphi: varphi,
  	varpi: varpi,
  	varpropto: varpropto,
  	varr: varr,
  	vArr: vArr,
  	varrho: varrho,
  	varsigma: varsigma,
  	varsubsetneq: varsubsetneq,
  	varsubsetneqq: varsubsetneqq,
  	varsupsetneq: varsupsetneq,
  	varsupsetneqq: varsupsetneqq,
  	vartheta: vartheta,
  	vartriangleleft: vartriangleleft,
  	vartriangleright: vartriangleright,
  	vBar: vBar,
  	Vbar: Vbar,
  	vBarv: vBarv,
  	Vcy: Vcy,
  	vcy: vcy,
  	vdash: vdash,
  	vDash: vDash,
  	Vdash: Vdash,
  	VDash: VDash,
  	Vdashl: Vdashl,
  	veebar: veebar,
  	vee: vee,
  	Vee: Vee,
  	veeeq: veeeq,
  	vellip: vellip,
  	verbar: verbar,
  	Verbar: Verbar,
  	vert: vert,
  	Vert: Vert,
  	VerticalBar: VerticalBar,
  	VerticalLine: VerticalLine,
  	VerticalSeparator: VerticalSeparator,
  	VerticalTilde: VerticalTilde,
  	VeryThinSpace: VeryThinSpace,
  	Vfr: Vfr,
  	vfr: vfr,
  	vltri: vltri,
  	vnsub: vnsub,
  	vnsup: vnsup,
  	Vopf: Vopf,
  	vopf: vopf,
  	vprop: vprop,
  	vrtri: vrtri,
  	Vscr: Vscr,
  	vscr: vscr,
  	vsubnE: vsubnE,
  	vsubne: vsubne,
  	vsupnE: vsupnE,
  	vsupne: vsupne,
  	Vvdash: Vvdash,
  	vzigzag: vzigzag,
  	Wcirc: Wcirc,
  	wcirc: wcirc,
  	wedbar: wedbar,
  	wedge: wedge,
  	Wedge: Wedge,
  	wedgeq: wedgeq,
  	weierp: weierp,
  	Wfr: Wfr,
  	wfr: wfr,
  	Wopf: Wopf,
  	wopf: wopf,
  	wp: wp,
  	wr: wr,
  	wreath: wreath,
  	Wscr: Wscr,
  	wscr: wscr,
  	xcap: xcap,
  	xcirc: xcirc,
  	xcup: xcup,
  	xdtri: xdtri,
  	Xfr: Xfr,
  	xfr: xfr,
  	xharr: xharr,
  	xhArr: xhArr,
  	Xi: Xi,
  	xi: xi,
  	xlarr: xlarr,
  	xlArr: xlArr,
  	xmap: xmap,
  	xnis: xnis,
  	xodot: xodot,
  	Xopf: Xopf,
  	xopf: xopf,
  	xoplus: xoplus,
  	xotime: xotime,
  	xrarr: xrarr,
  	xrArr: xrArr,
  	Xscr: Xscr,
  	xscr: xscr,
  	xsqcup: xsqcup,
  	xuplus: xuplus,
  	xutri: xutri,
  	xvee: xvee,
  	xwedge: xwedge,
  	Yacute: Yacute,
  	yacute: yacute,
  	YAcy: YAcy,
  	yacy: yacy,
  	Ycirc: Ycirc,
  	ycirc: ycirc,
  	Ycy: Ycy,
  	ycy: ycy,
  	yen: yen,
  	Yfr: Yfr,
  	yfr: yfr,
  	YIcy: YIcy,
  	yicy: yicy,
  	Yopf: Yopf,
  	yopf: yopf,
  	Yscr: Yscr,
  	yscr: yscr,
  	YUcy: YUcy,
  	yucy: yucy,
  	yuml: yuml,
  	Yuml: Yuml,
  	Zacute: Zacute,
  	zacute: zacute,
  	Zcaron: Zcaron,
  	zcaron: zcaron,
  	Zcy: Zcy,
  	zcy: zcy,
  	Zdot: Zdot,
  	zdot: zdot,
  	zeetrf: zeetrf,
  	ZeroWidthSpace: ZeroWidthSpace,
  	Zeta: Zeta,
  	zeta: zeta,
  	zfr: zfr,
  	Zfr: Zfr,
  	ZHcy: ZHcy,
  	zhcy: zhcy,
  	zigrarr: zigrarr,
  	zopf: zopf,
  	Zopf: Zopf,
  	Zscr: Zscr,
  	zscr: zscr,
  	zwj: zwj,
  	zwnj: zwnj
  };

  var entities$2 = /*#__PURE__*/Object.freeze({
    __proto__: null,
    Aacute: Aacute,
    aacute: aacute,
    Abreve: Abreve,
    abreve: abreve,
    ac: ac,
    acd: acd,
    acE: acE,
    Acirc: Acirc,
    acirc: acirc,
    acute: acute,
    Acy: Acy,
    acy: acy,
    AElig: AElig,
    aelig: aelig,
    af: af,
    Afr: Afr,
    afr: afr,
    Agrave: Agrave,
    agrave: agrave,
    alefsym: alefsym,
    aleph: aleph,
    Alpha: Alpha,
    alpha: alpha,
    Amacr: Amacr,
    amacr: amacr,
    amalg: amalg,
    amp: amp,
    AMP: AMP,
    andand: andand,
    And: And,
    and: and,
    andd: andd,
    andslope: andslope,
    andv: andv,
    ang: ang,
    ange: ange,
    angle: angle,
    angmsdaa: angmsdaa,
    angmsdab: angmsdab,
    angmsdac: angmsdac,
    angmsdad: angmsdad,
    angmsdae: angmsdae,
    angmsdaf: angmsdaf,
    angmsdag: angmsdag,
    angmsdah: angmsdah,
    angmsd: angmsd,
    angrt: angrt,
    angrtvb: angrtvb,
    angrtvbd: angrtvbd,
    angsph: angsph,
    angst: angst,
    angzarr: angzarr,
    Aogon: Aogon,
    aogon: aogon,
    Aopf: Aopf,
    aopf: aopf,
    apacir: apacir,
    ap: ap,
    apE: apE,
    ape: ape,
    apid: apid,
    apos: apos,
    ApplyFunction: ApplyFunction,
    approx: approx,
    approxeq: approxeq,
    Aring: Aring,
    aring: aring,
    Ascr: Ascr,
    ascr: ascr,
    Assign: Assign,
    ast: ast,
    asymp: asymp,
    asympeq: asympeq,
    Atilde: Atilde,
    atilde: atilde,
    Auml: Auml,
    auml: auml,
    awconint: awconint,
    awint: awint,
    backcong: backcong,
    backepsilon: backepsilon,
    backprime: backprime,
    backsim: backsim,
    backsimeq: backsimeq,
    Backslash: Backslash,
    Barv: Barv,
    barvee: barvee,
    barwed: barwed,
    Barwed: Barwed,
    barwedge: barwedge,
    bbrk: bbrk,
    bbrktbrk: bbrktbrk,
    bcong: bcong,
    Bcy: Bcy,
    bcy: bcy,
    bdquo: bdquo,
    becaus: becaus,
    because: because,
    Because: Because,
    bemptyv: bemptyv,
    bepsi: bepsi,
    bernou: bernou,
    Bernoullis: Bernoullis,
    Beta: Beta,
    beta: beta,
    beth: beth,
    between: between,
    Bfr: Bfr,
    bfr: bfr,
    bigcap: bigcap,
    bigcirc: bigcirc,
    bigcup: bigcup,
    bigodot: bigodot,
    bigoplus: bigoplus,
    bigotimes: bigotimes,
    bigsqcup: bigsqcup,
    bigstar: bigstar,
    bigtriangledown: bigtriangledown,
    bigtriangleup: bigtriangleup,
    biguplus: biguplus,
    bigvee: bigvee,
    bigwedge: bigwedge,
    bkarow: bkarow,
    blacklozenge: blacklozenge,
    blacksquare: blacksquare,
    blacktriangle: blacktriangle,
    blacktriangledown: blacktriangledown,
    blacktriangleleft: blacktriangleleft,
    blacktriangleright: blacktriangleright,
    blank: blank,
    blk12: blk12,
    blk14: blk14,
    blk34: blk34,
    block: block$1,
    bne: bne,
    bnequiv: bnequiv,
    bNot: bNot,
    bnot: bnot,
    Bopf: Bopf,
    bopf: bopf,
    bot: bot,
    bottom: bottom,
    bowtie: bowtie,
    boxbox: boxbox,
    boxdl: boxdl,
    boxdL: boxdL,
    boxDl: boxDl,
    boxDL: boxDL,
    boxdr: boxdr,
    boxdR: boxdR,
    boxDr: boxDr,
    boxDR: boxDR,
    boxh: boxh,
    boxH: boxH,
    boxhd: boxhd,
    boxHd: boxHd,
    boxhD: boxhD,
    boxHD: boxHD,
    boxhu: boxhu,
    boxHu: boxHu,
    boxhU: boxhU,
    boxHU: boxHU,
    boxminus: boxminus,
    boxplus: boxplus,
    boxtimes: boxtimes,
    boxul: boxul,
    boxuL: boxuL,
    boxUl: boxUl,
    boxUL: boxUL,
    boxur: boxur,
    boxuR: boxuR,
    boxUr: boxUr,
    boxUR: boxUR,
    boxv: boxv,
    boxV: boxV,
    boxvh: boxvh,
    boxvH: boxvH,
    boxVh: boxVh,
    boxVH: boxVH,
    boxvl: boxvl,
    boxvL: boxvL,
    boxVl: boxVl,
    boxVL: boxVL,
    boxvr: boxvr,
    boxvR: boxvR,
    boxVr: boxVr,
    boxVR: boxVR,
    bprime: bprime,
    breve: breve,
    Breve: Breve,
    brvbar: brvbar,
    bscr: bscr,
    Bscr: Bscr,
    bsemi: bsemi,
    bsim: bsim,
    bsime: bsime,
    bsolb: bsolb,
    bsol: bsol,
    bsolhsub: bsolhsub,
    bull: bull,
    bullet: bullet,
    bump: bump,
    bumpE: bumpE,
    bumpe: bumpe,
    Bumpeq: Bumpeq,
    bumpeq: bumpeq,
    Cacute: Cacute,
    cacute: cacute,
    capand: capand,
    capbrcup: capbrcup,
    capcap: capcap,
    cap: cap,
    Cap: Cap,
    capcup: capcup,
    capdot: capdot,
    CapitalDifferentialD: CapitalDifferentialD,
    caps: caps,
    caret: caret,
    caron: caron,
    Cayleys: Cayleys,
    ccaps: ccaps,
    Ccaron: Ccaron,
    ccaron: ccaron,
    Ccedil: Ccedil,
    ccedil: ccedil,
    Ccirc: Ccirc,
    ccirc: ccirc,
    Cconint: Cconint,
    ccups: ccups,
    ccupssm: ccupssm,
    Cdot: Cdot,
    cdot: cdot,
    cedil: cedil,
    Cedilla: Cedilla,
    cemptyv: cemptyv,
    cent: cent,
    centerdot: centerdot,
    CenterDot: CenterDot,
    cfr: cfr,
    Cfr: Cfr,
    CHcy: CHcy,
    chcy: chcy,
    check: check,
    checkmark: checkmark,
    Chi: Chi,
    chi: chi,
    circ: circ,
    circeq: circeq,
    circlearrowleft: circlearrowleft,
    circlearrowright: circlearrowright,
    circledast: circledast,
    circledcirc: circledcirc,
    circleddash: circleddash,
    CircleDot: CircleDot,
    circledR: circledR,
    circledS: circledS,
    CircleMinus: CircleMinus,
    CirclePlus: CirclePlus,
    CircleTimes: CircleTimes,
    cir: cir,
    cirE: cirE,
    cire: cire,
    cirfnint: cirfnint,
    cirmid: cirmid,
    cirscir: cirscir,
    ClockwiseContourIntegral: ClockwiseContourIntegral,
    CloseCurlyDoubleQuote: CloseCurlyDoubleQuote,
    CloseCurlyQuote: CloseCurlyQuote,
    clubs: clubs,
    clubsuit: clubsuit,
    colon: colon,
    Colon: Colon,
    Colone: Colone,
    colone: colone,
    coloneq: coloneq,
    comma: comma,
    commat: commat,
    comp: comp,
    compfn: compfn,
    complement: complement,
    complexes: complexes,
    cong: cong,
    congdot: congdot,
    Congruent: Congruent,
    conint: conint,
    Conint: Conint,
    ContourIntegral: ContourIntegral,
    copf: copf,
    Copf: Copf,
    coprod: coprod,
    Coproduct: Coproduct,
    copy: copy,
    COPY: COPY,
    copysr: copysr,
    CounterClockwiseContourIntegral: CounterClockwiseContourIntegral,
    crarr: crarr,
    cross: cross,
    Cross: Cross,
    Cscr: Cscr,
    cscr: cscr,
    csub: csub,
    csube: csube,
    csup: csup,
    csupe: csupe,
    ctdot: ctdot,
    cudarrl: cudarrl,
    cudarrr: cudarrr,
    cuepr: cuepr,
    cuesc: cuesc,
    cularr: cularr,
    cularrp: cularrp,
    cupbrcap: cupbrcap,
    cupcap: cupcap,
    CupCap: CupCap,
    cup: cup,
    Cup: Cup,
    cupcup: cupcup,
    cupdot: cupdot,
    cupor: cupor,
    cups: cups,
    curarr: curarr,
    curarrm: curarrm,
    curlyeqprec: curlyeqprec,
    curlyeqsucc: curlyeqsucc,
    curlyvee: curlyvee,
    curlywedge: curlywedge,
    curren: curren,
    curvearrowleft: curvearrowleft,
    curvearrowright: curvearrowright,
    cuvee: cuvee,
    cuwed: cuwed,
    cwconint: cwconint,
    cwint: cwint,
    cylcty: cylcty,
    dagger: dagger,
    Dagger: Dagger,
    daleth: daleth,
    darr: darr,
    Darr: Darr,
    dArr: dArr,
    dash: dash,
    Dashv: Dashv,
    dashv: dashv,
    dbkarow: dbkarow,
    dblac: dblac,
    Dcaron: Dcaron,
    dcaron: dcaron,
    Dcy: Dcy,
    dcy: dcy,
    ddagger: ddagger,
    ddarr: ddarr,
    DD: DD,
    dd: dd,
    DDotrahd: DDotrahd,
    ddotseq: ddotseq,
    deg: deg,
    Del: Del,
    Delta: Delta,
    delta: delta,
    demptyv: demptyv,
    dfisht: dfisht,
    Dfr: Dfr,
    dfr: dfr,
    dHar: dHar,
    dharl: dharl,
    dharr: dharr,
    DiacriticalAcute: DiacriticalAcute,
    DiacriticalDot: DiacriticalDot,
    DiacriticalDoubleAcute: DiacriticalDoubleAcute,
    DiacriticalGrave: DiacriticalGrave,
    DiacriticalTilde: DiacriticalTilde,
    diam: diam,
    diamond: diamond,
    Diamond: Diamond,
    diamondsuit: diamondsuit,
    diams: diams,
    die: die,
    DifferentialD: DifferentialD,
    digamma: digamma,
    disin: disin,
    div: div,
    divide: divide,
    divideontimes: divideontimes,
    divonx: divonx,
    DJcy: DJcy,
    djcy: djcy,
    dlcorn: dlcorn,
    dlcrop: dlcrop,
    dollar: dollar,
    Dopf: Dopf,
    dopf: dopf,
    Dot: Dot,
    dot: dot,
    DotDot: DotDot,
    doteq: doteq,
    doteqdot: doteqdot,
    DotEqual: DotEqual,
    dotminus: dotminus,
    dotplus: dotplus,
    dotsquare: dotsquare,
    doublebarwedge: doublebarwedge,
    DoubleContourIntegral: DoubleContourIntegral,
    DoubleDot: DoubleDot,
    DoubleDownArrow: DoubleDownArrow,
    DoubleLeftArrow: DoubleLeftArrow,
    DoubleLeftRightArrow: DoubleLeftRightArrow,
    DoubleLeftTee: DoubleLeftTee,
    DoubleLongLeftArrow: DoubleLongLeftArrow,
    DoubleLongLeftRightArrow: DoubleLongLeftRightArrow,
    DoubleLongRightArrow: DoubleLongRightArrow,
    DoubleRightArrow: DoubleRightArrow,
    DoubleRightTee: DoubleRightTee,
    DoubleUpArrow: DoubleUpArrow,
    DoubleUpDownArrow: DoubleUpDownArrow,
    DoubleVerticalBar: DoubleVerticalBar,
    DownArrowBar: DownArrowBar,
    downarrow: downarrow,
    DownArrow: DownArrow,
    Downarrow: Downarrow,
    DownArrowUpArrow: DownArrowUpArrow,
    DownBreve: DownBreve,
    downdownarrows: downdownarrows,
    downharpoonleft: downharpoonleft,
    downharpoonright: downharpoonright,
    DownLeftRightVector: DownLeftRightVector,
    DownLeftTeeVector: DownLeftTeeVector,
    DownLeftVectorBar: DownLeftVectorBar,
    DownLeftVector: DownLeftVector,
    DownRightTeeVector: DownRightTeeVector,
    DownRightVectorBar: DownRightVectorBar,
    DownRightVector: DownRightVector,
    DownTeeArrow: DownTeeArrow,
    DownTee: DownTee,
    drbkarow: drbkarow,
    drcorn: drcorn,
    drcrop: drcrop,
    Dscr: Dscr,
    dscr: dscr,
    DScy: DScy,
    dscy: dscy,
    dsol: dsol,
    Dstrok: Dstrok,
    dstrok: dstrok,
    dtdot: dtdot,
    dtri: dtri,
    dtrif: dtrif,
    duarr: duarr,
    duhar: duhar,
    dwangle: dwangle,
    DZcy: DZcy,
    dzcy: dzcy,
    dzigrarr: dzigrarr,
    Eacute: Eacute,
    eacute: eacute,
    easter: easter,
    Ecaron: Ecaron,
    ecaron: ecaron,
    Ecirc: Ecirc,
    ecirc: ecirc,
    ecir: ecir,
    ecolon: ecolon,
    Ecy: Ecy,
    ecy: ecy,
    eDDot: eDDot,
    Edot: Edot,
    edot: edot,
    eDot: eDot,
    ee: ee,
    efDot: efDot,
    Efr: Efr,
    efr: efr,
    eg: eg,
    Egrave: Egrave,
    egrave: egrave,
    egs: egs,
    egsdot: egsdot,
    el: el,
    Element: Element,
    elinters: elinters,
    ell: ell,
    els: els,
    elsdot: elsdot,
    Emacr: Emacr,
    emacr: emacr,
    empty: empty,
    emptyset: emptyset,
    EmptySmallSquare: EmptySmallSquare,
    emptyv: emptyv,
    EmptyVerySmallSquare: EmptyVerySmallSquare,
    emsp13: emsp13,
    emsp14: emsp14,
    emsp: emsp,
    ENG: ENG,
    eng: eng,
    ensp: ensp,
    Eogon: Eogon,
    eogon: eogon,
    Eopf: Eopf,
    eopf: eopf,
    epar: epar,
    eparsl: eparsl,
    eplus: eplus,
    epsi: epsi,
    Epsilon: Epsilon,
    epsilon: epsilon,
    epsiv: epsiv,
    eqcirc: eqcirc,
    eqcolon: eqcolon,
    eqsim: eqsim,
    eqslantgtr: eqslantgtr,
    eqslantless: eqslantless,
    Equal: Equal,
    equals: equals,
    EqualTilde: EqualTilde,
    equest: equest,
    Equilibrium: Equilibrium,
    equiv: equiv,
    equivDD: equivDD,
    eqvparsl: eqvparsl,
    erarr: erarr,
    erDot: erDot,
    escr: escr,
    Escr: Escr,
    esdot: esdot,
    Esim: Esim,
    esim: esim,
    Eta: Eta,
    eta: eta,
    ETH: ETH,
    eth: eth,
    Euml: Euml,
    euml: euml,
    euro: euro,
    excl: excl,
    exist: exist,
    Exists: Exists,
    expectation: expectation,
    exponentiale: exponentiale,
    ExponentialE: ExponentialE,
    fallingdotseq: fallingdotseq,
    Fcy: Fcy,
    fcy: fcy,
    female: female,
    ffilig: ffilig,
    fflig: fflig,
    ffllig: ffllig,
    Ffr: Ffr,
    ffr: ffr,
    filig: filig,
    FilledSmallSquare: FilledSmallSquare,
    FilledVerySmallSquare: FilledVerySmallSquare,
    fjlig: fjlig,
    flat: flat,
    fllig: fllig,
    fltns: fltns,
    fnof: fnof,
    Fopf: Fopf,
    fopf: fopf,
    forall: forall,
    ForAll: ForAll,
    fork: fork,
    forkv: forkv,
    Fouriertrf: Fouriertrf,
    fpartint: fpartint,
    frac12: frac12,
    frac13: frac13,
    frac14: frac14,
    frac15: frac15,
    frac16: frac16,
    frac18: frac18,
    frac23: frac23,
    frac25: frac25,
    frac34: frac34,
    frac35: frac35,
    frac38: frac38,
    frac45: frac45,
    frac56: frac56,
    frac58: frac58,
    frac78: frac78,
    frasl: frasl,
    frown: frown,
    fscr: fscr,
    Fscr: Fscr,
    gacute: gacute,
    Gamma: Gamma,
    gamma: gamma,
    Gammad: Gammad,
    gammad: gammad,
    gap: gap,
    Gbreve: Gbreve,
    gbreve: gbreve,
    Gcedil: Gcedil,
    Gcirc: Gcirc,
    gcirc: gcirc,
    Gcy: Gcy,
    gcy: gcy,
    Gdot: Gdot,
    gdot: gdot,
    ge: ge,
    gE: gE,
    gEl: gEl,
    gel: gel,
    geq: geq,
    geqq: geqq,
    geqslant: geqslant,
    gescc: gescc,
    ges: ges,
    gesdot: gesdot,
    gesdoto: gesdoto,
    gesdotol: gesdotol,
    gesl: gesl,
    gesles: gesles,
    Gfr: Gfr,
    gfr: gfr,
    gg: gg,
    Gg: Gg,
    ggg: ggg,
    gimel: gimel,
    GJcy: GJcy,
    gjcy: gjcy,
    gla: gla,
    gl: gl,
    glE: glE,
    glj: glj,
    gnap: gnap,
    gnapprox: gnapprox,
    gne: gne,
    gnE: gnE,
    gneq: gneq,
    gneqq: gneqq,
    gnsim: gnsim,
    Gopf: Gopf,
    gopf: gopf,
    grave: grave,
    GreaterEqual: GreaterEqual,
    GreaterEqualLess: GreaterEqualLess,
    GreaterFullEqual: GreaterFullEqual,
    GreaterGreater: GreaterGreater,
    GreaterLess: GreaterLess,
    GreaterSlantEqual: GreaterSlantEqual,
    GreaterTilde: GreaterTilde,
    Gscr: Gscr,
    gscr: gscr,
    gsim: gsim,
    gsime: gsime,
    gsiml: gsiml,
    gtcc: gtcc,
    gtcir: gtcir,
    gt: gt,
    GT: GT,
    Gt: Gt,
    gtdot: gtdot,
    gtlPar: gtlPar,
    gtquest: gtquest,
    gtrapprox: gtrapprox,
    gtrarr: gtrarr,
    gtrdot: gtrdot,
    gtreqless: gtreqless,
    gtreqqless: gtreqqless,
    gtrless: gtrless,
    gtrsim: gtrsim,
    gvertneqq: gvertneqq,
    gvnE: gvnE,
    Hacek: Hacek,
    hairsp: hairsp,
    half: half,
    hamilt: hamilt,
    HARDcy: HARDcy,
    hardcy: hardcy,
    harrcir: harrcir,
    harr: harr,
    hArr: hArr,
    harrw: harrw,
    Hat: Hat,
    hbar: hbar,
    Hcirc: Hcirc,
    hcirc: hcirc,
    hearts: hearts,
    heartsuit: heartsuit,
    hellip: hellip,
    hercon: hercon,
    hfr: hfr,
    Hfr: Hfr,
    HilbertSpace: HilbertSpace,
    hksearow: hksearow,
    hkswarow: hkswarow,
    hoarr: hoarr,
    homtht: homtht,
    hookleftarrow: hookleftarrow,
    hookrightarrow: hookrightarrow,
    hopf: hopf,
    Hopf: Hopf,
    horbar: horbar,
    HorizontalLine: HorizontalLine,
    hscr: hscr,
    Hscr: Hscr,
    hslash: hslash,
    Hstrok: Hstrok,
    hstrok: hstrok,
    HumpDownHump: HumpDownHump,
    HumpEqual: HumpEqual,
    hybull: hybull,
    hyphen: hyphen,
    Iacute: Iacute,
    iacute: iacute,
    ic: ic,
    Icirc: Icirc,
    icirc: icirc,
    Icy: Icy,
    icy: icy,
    Idot: Idot,
    IEcy: IEcy,
    iecy: iecy,
    iexcl: iexcl,
    iff: iff,
    ifr: ifr,
    Ifr: Ifr,
    Igrave: Igrave,
    igrave: igrave,
    ii: ii,
    iiiint: iiiint,
    iiint: iiint,
    iinfin: iinfin,
    iiota: iiota,
    IJlig: IJlig,
    ijlig: ijlig,
    Imacr: Imacr,
    imacr: imacr,
    image: image$2,
    ImaginaryI: ImaginaryI,
    imagline: imagline,
    imagpart: imagpart,
    imath: imath,
    Im: Im,
    imof: imof,
    imped: imped,
    Implies: Implies,
    incare: incare,
    infin: infin,
    infintie: infintie,
    inodot: inodot,
    intcal: intcal,
    int: int,
    Int: Int,
    integers: integers,
    Integral: Integral,
    intercal: intercal,
    Intersection: Intersection,
    intlarhk: intlarhk,
    intprod: intprod,
    InvisibleComma: InvisibleComma,
    InvisibleTimes: InvisibleTimes,
    IOcy: IOcy,
    iocy: iocy,
    Iogon: Iogon,
    iogon: iogon,
    Iopf: Iopf,
    iopf: iopf,
    Iota: Iota,
    iota: iota,
    iprod: iprod,
    iquest: iquest,
    iscr: iscr,
    Iscr: Iscr,
    isin: isin,
    isindot: isindot,
    isinE: isinE,
    isins: isins,
    isinsv: isinsv,
    isinv: isinv,
    it: it,
    Itilde: Itilde,
    itilde: itilde,
    Iukcy: Iukcy,
    iukcy: iukcy,
    Iuml: Iuml,
    iuml: iuml,
    Jcirc: Jcirc,
    jcirc: jcirc,
    Jcy: Jcy,
    jcy: jcy,
    Jfr: Jfr,
    jfr: jfr,
    jmath: jmath,
    Jopf: Jopf,
    jopf: jopf,
    Jscr: Jscr,
    jscr: jscr,
    Jsercy: Jsercy,
    jsercy: jsercy,
    Jukcy: Jukcy,
    jukcy: jukcy,
    Kappa: Kappa,
    kappa: kappa,
    kappav: kappav,
    Kcedil: Kcedil,
    kcedil: kcedil,
    Kcy: Kcy,
    kcy: kcy,
    Kfr: Kfr,
    kfr: kfr,
    kgreen: kgreen,
    KHcy: KHcy,
    khcy: khcy,
    KJcy: KJcy,
    kjcy: kjcy,
    Kopf: Kopf,
    kopf: kopf,
    Kscr: Kscr,
    kscr: kscr,
    lAarr: lAarr,
    Lacute: Lacute,
    lacute: lacute,
    laemptyv: laemptyv,
    lagran: lagran,
    Lambda: Lambda,
    lambda: lambda,
    lang: lang,
    Lang: Lang,
    langd: langd,
    langle: langle,
    lap: lap,
    Laplacetrf: Laplacetrf,
    laquo: laquo,
    larrb: larrb,
    larrbfs: larrbfs,
    larr: larr,
    Larr: Larr,
    lArr: lArr,
    larrfs: larrfs,
    larrhk: larrhk,
    larrlp: larrlp,
    larrpl: larrpl,
    larrsim: larrsim,
    larrtl: larrtl,
    latail: latail,
    lAtail: lAtail,
    lat: lat,
    late: late,
    lates: lates,
    lbarr: lbarr,
    lBarr: lBarr,
    lbbrk: lbbrk,
    lbrace: lbrace,
    lbrack: lbrack,
    lbrke: lbrke,
    lbrksld: lbrksld,
    lbrkslu: lbrkslu,
    Lcaron: Lcaron,
    lcaron: lcaron,
    Lcedil: Lcedil,
    lcedil: lcedil,
    lceil: lceil,
    lcub: lcub,
    Lcy: Lcy,
    lcy: lcy,
    ldca: ldca,
    ldquo: ldquo,
    ldquor: ldquor,
    ldrdhar: ldrdhar,
    ldrushar: ldrushar,
    ldsh: ldsh,
    le: le,
    lE: lE,
    LeftAngleBracket: LeftAngleBracket,
    LeftArrowBar: LeftArrowBar,
    leftarrow: leftarrow,
    LeftArrow: LeftArrow,
    Leftarrow: Leftarrow,
    LeftArrowRightArrow: LeftArrowRightArrow,
    leftarrowtail: leftarrowtail,
    LeftCeiling: LeftCeiling,
    LeftDoubleBracket: LeftDoubleBracket,
    LeftDownTeeVector: LeftDownTeeVector,
    LeftDownVectorBar: LeftDownVectorBar,
    LeftDownVector: LeftDownVector,
    LeftFloor: LeftFloor,
    leftharpoondown: leftharpoondown,
    leftharpoonup: leftharpoonup,
    leftleftarrows: leftleftarrows,
    leftrightarrow: leftrightarrow,
    LeftRightArrow: LeftRightArrow,
    Leftrightarrow: Leftrightarrow,
    leftrightarrows: leftrightarrows,
    leftrightharpoons: leftrightharpoons,
    leftrightsquigarrow: leftrightsquigarrow,
    LeftRightVector: LeftRightVector,
    LeftTeeArrow: LeftTeeArrow,
    LeftTee: LeftTee,
    LeftTeeVector: LeftTeeVector,
    leftthreetimes: leftthreetimes,
    LeftTriangleBar: LeftTriangleBar,
    LeftTriangle: LeftTriangle,
    LeftTriangleEqual: LeftTriangleEqual,
    LeftUpDownVector: LeftUpDownVector,
    LeftUpTeeVector: LeftUpTeeVector,
    LeftUpVectorBar: LeftUpVectorBar,
    LeftUpVector: LeftUpVector,
    LeftVectorBar: LeftVectorBar,
    LeftVector: LeftVector,
    lEg: lEg,
    leg: leg,
    leq: leq,
    leqq: leqq,
    leqslant: leqslant,
    lescc: lescc,
    les: les,
    lesdot: lesdot,
    lesdoto: lesdoto,
    lesdotor: lesdotor,
    lesg: lesg,
    lesges: lesges,
    lessapprox: lessapprox,
    lessdot: lessdot,
    lesseqgtr: lesseqgtr,
    lesseqqgtr: lesseqqgtr,
    LessEqualGreater: LessEqualGreater,
    LessFullEqual: LessFullEqual,
    LessGreater: LessGreater,
    lessgtr: lessgtr,
    LessLess: LessLess,
    lesssim: lesssim,
    LessSlantEqual: LessSlantEqual,
    LessTilde: LessTilde,
    lfisht: lfisht,
    lfloor: lfloor,
    Lfr: Lfr,
    lfr: lfr,
    lg: lg,
    lgE: lgE,
    lHar: lHar,
    lhard: lhard,
    lharu: lharu,
    lharul: lharul,
    lhblk: lhblk,
    LJcy: LJcy,
    ljcy: ljcy,
    llarr: llarr,
    ll: ll,
    Ll: Ll,
    llcorner: llcorner,
    Lleftarrow: Lleftarrow,
    llhard: llhard,
    lltri: lltri,
    Lmidot: Lmidot,
    lmidot: lmidot,
    lmoustache: lmoustache,
    lmoust: lmoust,
    lnap: lnap,
    lnapprox: lnapprox,
    lne: lne,
    lnE: lnE,
    lneq: lneq,
    lneqq: lneqq,
    lnsim: lnsim,
    loang: loang,
    loarr: loarr,
    lobrk: lobrk,
    longleftarrow: longleftarrow,
    LongLeftArrow: LongLeftArrow,
    Longleftarrow: Longleftarrow,
    longleftrightarrow: longleftrightarrow,
    LongLeftRightArrow: LongLeftRightArrow,
    Longleftrightarrow: Longleftrightarrow,
    longmapsto: longmapsto,
    longrightarrow: longrightarrow,
    LongRightArrow: LongRightArrow,
    Longrightarrow: Longrightarrow,
    looparrowleft: looparrowleft,
    looparrowright: looparrowright,
    lopar: lopar,
    Lopf: Lopf,
    lopf: lopf,
    loplus: loplus,
    lotimes: lotimes,
    lowast: lowast,
    lowbar: lowbar,
    LowerLeftArrow: LowerLeftArrow,
    LowerRightArrow: LowerRightArrow,
    loz: loz,
    lozenge: lozenge,
    lozf: lozf,
    lpar: lpar,
    lparlt: lparlt,
    lrarr: lrarr,
    lrcorner: lrcorner,
    lrhar: lrhar,
    lrhard: lrhard,
    lrm: lrm,
    lrtri: lrtri,
    lsaquo: lsaquo,
    lscr: lscr,
    Lscr: Lscr,
    lsh: lsh,
    Lsh: Lsh,
    lsim: lsim,
    lsime: lsime,
    lsimg: lsimg,
    lsqb: lsqb,
    lsquo: lsquo,
    lsquor: lsquor,
    Lstrok: Lstrok,
    lstrok: lstrok,
    ltcc: ltcc,
    ltcir: ltcir,
    lt: lt,
    LT: LT,
    Lt: Lt,
    ltdot: ltdot,
    lthree: lthree,
    ltimes: ltimes,
    ltlarr: ltlarr,
    ltquest: ltquest,
    ltri: ltri,
    ltrie: ltrie,
    ltrif: ltrif,
    ltrPar: ltrPar,
    lurdshar: lurdshar,
    luruhar: luruhar,
    lvertneqq: lvertneqq,
    lvnE: lvnE,
    macr: macr,
    male: male,
    malt: malt,
    maltese: maltese,
    map: map,
    mapsto: mapsto,
    mapstodown: mapstodown,
    mapstoleft: mapstoleft,
    mapstoup: mapstoup,
    marker: marker,
    mcomma: mcomma,
    Mcy: Mcy,
    mcy: mcy,
    mdash: mdash,
    mDDot: mDDot,
    measuredangle: measuredangle,
    MediumSpace: MediumSpace,
    Mellintrf: Mellintrf,
    Mfr: Mfr,
    mfr: mfr,
    mho: mho,
    micro: micro,
    midast: midast,
    midcir: midcir,
    mid: mid,
    middot: middot,
    minusb: minusb,
    minus: minus,
    minusd: minusd,
    minusdu: minusdu,
    MinusPlus: MinusPlus,
    mlcp: mlcp,
    mldr: mldr,
    mnplus: mnplus,
    models: models,
    Mopf: Mopf,
    mopf: mopf,
    mp: mp,
    mscr: mscr,
    Mscr: Mscr,
    mstpos: mstpos,
    Mu: Mu,
    mu: mu,
    multimap: multimap,
    mumap: mumap,
    nabla: nabla,
    Nacute: Nacute,
    nacute: nacute,
    nang: nang,
    nap: nap,
    napE: napE,
    napid: napid,
    napos: napos,
    napprox: napprox,
    natural: natural,
    naturals: naturals,
    natur: natur,
    nbsp: nbsp,
    nbump: nbump,
    nbumpe: nbumpe,
    ncap: ncap,
    Ncaron: Ncaron,
    ncaron: ncaron,
    Ncedil: Ncedil,
    ncedil: ncedil,
    ncong: ncong,
    ncongdot: ncongdot,
    ncup: ncup,
    Ncy: Ncy,
    ncy: ncy,
    ndash: ndash,
    nearhk: nearhk,
    nearr: nearr,
    neArr: neArr,
    nearrow: nearrow,
    ne: ne,
    nedot: nedot,
    NegativeMediumSpace: NegativeMediumSpace,
    NegativeThickSpace: NegativeThickSpace,
    NegativeThinSpace: NegativeThinSpace,
    NegativeVeryThinSpace: NegativeVeryThinSpace,
    nequiv: nequiv,
    nesear: nesear,
    nesim: nesim,
    NestedGreaterGreater: NestedGreaterGreater,
    NestedLessLess: NestedLessLess,
    NewLine: NewLine,
    nexist: nexist,
    nexists: nexists,
    Nfr: Nfr,
    nfr: nfr,
    ngE: ngE,
    nge: nge,
    ngeq: ngeq,
    ngeqq: ngeqq,
    ngeqslant: ngeqslant,
    nges: nges,
    nGg: nGg,
    ngsim: ngsim,
    nGt: nGt,
    ngt: ngt,
    ngtr: ngtr,
    nGtv: nGtv,
    nharr: nharr,
    nhArr: nhArr,
    nhpar: nhpar,
    ni: ni,
    nis: nis,
    nisd: nisd,
    niv: niv,
    NJcy: NJcy,
    njcy: njcy,
    nlarr: nlarr,
    nlArr: nlArr,
    nldr: nldr,
    nlE: nlE,
    nle: nle,
    nleftarrow: nleftarrow,
    nLeftarrow: nLeftarrow,
    nleftrightarrow: nleftrightarrow,
    nLeftrightarrow: nLeftrightarrow,
    nleq: nleq,
    nleqq: nleqq,
    nleqslant: nleqslant,
    nles: nles,
    nless: nless,
    nLl: nLl,
    nlsim: nlsim,
    nLt: nLt,
    nlt: nlt,
    nltri: nltri,
    nltrie: nltrie,
    nLtv: nLtv,
    nmid: nmid,
    NoBreak: NoBreak,
    NonBreakingSpace: NonBreakingSpace,
    nopf: nopf,
    Nopf: Nopf,
    Not: Not,
    not: not,
    NotCongruent: NotCongruent,
    NotCupCap: NotCupCap,
    NotDoubleVerticalBar: NotDoubleVerticalBar,
    NotElement: NotElement,
    NotEqual: NotEqual,
    NotEqualTilde: NotEqualTilde,
    NotExists: NotExists,
    NotGreater: NotGreater,
    NotGreaterEqual: NotGreaterEqual,
    NotGreaterFullEqual: NotGreaterFullEqual,
    NotGreaterGreater: NotGreaterGreater,
    NotGreaterLess: NotGreaterLess,
    NotGreaterSlantEqual: NotGreaterSlantEqual,
    NotGreaterTilde: NotGreaterTilde,
    NotHumpDownHump: NotHumpDownHump,
    NotHumpEqual: NotHumpEqual,
    notin: notin,
    notindot: notindot,
    notinE: notinE,
    notinva: notinva,
    notinvb: notinvb,
    notinvc: notinvc,
    NotLeftTriangleBar: NotLeftTriangleBar,
    NotLeftTriangle: NotLeftTriangle,
    NotLeftTriangleEqual: NotLeftTriangleEqual,
    NotLess: NotLess,
    NotLessEqual: NotLessEqual,
    NotLessGreater: NotLessGreater,
    NotLessLess: NotLessLess,
    NotLessSlantEqual: NotLessSlantEqual,
    NotLessTilde: NotLessTilde,
    NotNestedGreaterGreater: NotNestedGreaterGreater,
    NotNestedLessLess: NotNestedLessLess,
    notni: notni,
    notniva: notniva,
    notnivb: notnivb,
    notnivc: notnivc,
    NotPrecedes: NotPrecedes,
    NotPrecedesEqual: NotPrecedesEqual,
    NotPrecedesSlantEqual: NotPrecedesSlantEqual,
    NotReverseElement: NotReverseElement,
    NotRightTriangleBar: NotRightTriangleBar,
    NotRightTriangle: NotRightTriangle,
    NotRightTriangleEqual: NotRightTriangleEqual,
    NotSquareSubset: NotSquareSubset,
    NotSquareSubsetEqual: NotSquareSubsetEqual,
    NotSquareSuperset: NotSquareSuperset,
    NotSquareSupersetEqual: NotSquareSupersetEqual,
    NotSubset: NotSubset,
    NotSubsetEqual: NotSubsetEqual,
    NotSucceeds: NotSucceeds,
    NotSucceedsEqual: NotSucceedsEqual,
    NotSucceedsSlantEqual: NotSucceedsSlantEqual,
    NotSucceedsTilde: NotSucceedsTilde,
    NotSuperset: NotSuperset,
    NotSupersetEqual: NotSupersetEqual,
    NotTilde: NotTilde,
    NotTildeEqual: NotTildeEqual,
    NotTildeFullEqual: NotTildeFullEqual,
    NotTildeTilde: NotTildeTilde,
    NotVerticalBar: NotVerticalBar,
    nparallel: nparallel,
    npar: npar,
    nparsl: nparsl,
    npart: npart,
    npolint: npolint,
    npr: npr,
    nprcue: nprcue,
    nprec: nprec,
    npreceq: npreceq,
    npre: npre,
    nrarrc: nrarrc,
    nrarr: nrarr,
    nrArr: nrArr,
    nrarrw: nrarrw,
    nrightarrow: nrightarrow,
    nRightarrow: nRightarrow,
    nrtri: nrtri,
    nrtrie: nrtrie,
    nsc: nsc,
    nsccue: nsccue,
    nsce: nsce,
    Nscr: Nscr,
    nscr: nscr,
    nshortmid: nshortmid,
    nshortparallel: nshortparallel,
    nsim: nsim,
    nsime: nsime,
    nsimeq: nsimeq,
    nsmid: nsmid,
    nspar: nspar,
    nsqsube: nsqsube,
    nsqsupe: nsqsupe,
    nsub: nsub,
    nsubE: nsubE,
    nsube: nsube,
    nsubset: nsubset,
    nsubseteq: nsubseteq,
    nsubseteqq: nsubseteqq,
    nsucc: nsucc,
    nsucceq: nsucceq,
    nsup: nsup,
    nsupE: nsupE,
    nsupe: nsupe,
    nsupset: nsupset,
    nsupseteq: nsupseteq,
    nsupseteqq: nsupseteqq,
    ntgl: ntgl,
    Ntilde: Ntilde,
    ntilde: ntilde,
    ntlg: ntlg,
    ntriangleleft: ntriangleleft,
    ntrianglelefteq: ntrianglelefteq,
    ntriangleright: ntriangleright,
    ntrianglerighteq: ntrianglerighteq,
    Nu: Nu,
    nu: nu,
    num: num,
    numero: numero,
    numsp: numsp,
    nvap: nvap,
    nvdash: nvdash,
    nvDash: nvDash,
    nVdash: nVdash,
    nVDash: nVDash,
    nvge: nvge,
    nvgt: nvgt,
    nvHarr: nvHarr,
    nvinfin: nvinfin,
    nvlArr: nvlArr,
    nvle: nvle,
    nvlt: nvlt,
    nvltrie: nvltrie,
    nvrArr: nvrArr,
    nvrtrie: nvrtrie,
    nvsim: nvsim,
    nwarhk: nwarhk,
    nwarr: nwarr,
    nwArr: nwArr,
    nwarrow: nwarrow,
    nwnear: nwnear,
    Oacute: Oacute,
    oacute: oacute,
    oast: oast,
    Ocirc: Ocirc,
    ocirc: ocirc,
    ocir: ocir,
    Ocy: Ocy,
    ocy: ocy,
    odash: odash,
    Odblac: Odblac,
    odblac: odblac,
    odiv: odiv,
    odot: odot,
    odsold: odsold,
    OElig: OElig,
    oelig: oelig,
    ofcir: ofcir,
    Ofr: Ofr,
    ofr: ofr,
    ogon: ogon,
    Ograve: Ograve,
    ograve: ograve,
    ogt: ogt,
    ohbar: ohbar,
    ohm: ohm,
    oint: oint,
    olarr: olarr,
    olcir: olcir,
    olcross: olcross,
    oline: oline,
    olt: olt,
    Omacr: Omacr,
    omacr: omacr,
    Omega: Omega,
    omega: omega,
    Omicron: Omicron,
    omicron: omicron,
    omid: omid,
    ominus: ominus,
    Oopf: Oopf,
    oopf: oopf,
    opar: opar,
    OpenCurlyDoubleQuote: OpenCurlyDoubleQuote,
    OpenCurlyQuote: OpenCurlyQuote,
    operp: operp,
    oplus: oplus,
    orarr: orarr,
    Or: Or,
    or: or,
    ord: ord,
    order: order,
    orderof: orderof,
    ordf: ordf,
    ordm: ordm,
    origof: origof,
    oror: oror,
    orslope: orslope,
    orv: orv,
    oS: oS,
    Oscr: Oscr,
    oscr: oscr,
    Oslash: Oslash,
    oslash: oslash,
    osol: osol,
    Otilde: Otilde,
    otilde: otilde,
    otimesas: otimesas,
    Otimes: Otimes,
    otimes: otimes,
    Ouml: Ouml,
    ouml: ouml,
    ovbar: ovbar,
    OverBar: OverBar,
    OverBrace: OverBrace,
    OverBracket: OverBracket,
    OverParenthesis: OverParenthesis,
    para: para,
    parallel: parallel,
    par: par,
    parsim: parsim,
    parsl: parsl,
    part: part,
    PartialD: PartialD,
    Pcy: Pcy,
    pcy: pcy,
    percnt: percnt,
    period: period,
    permil: permil,
    perp: perp,
    pertenk: pertenk,
    Pfr: Pfr,
    pfr: pfr,
    Phi: Phi,
    phi: phi,
    phiv: phiv,
    phmmat: phmmat,
    phone: phone,
    Pi: Pi,
    pi: pi,
    pitchfork: pitchfork,
    piv: piv,
    planck: planck,
    planckh: planckh,
    plankv: plankv,
    plusacir: plusacir,
    plusb: plusb,
    pluscir: pluscir,
    plus: plus,
    plusdo: plusdo,
    plusdu: plusdu,
    pluse: pluse,
    PlusMinus: PlusMinus,
    plusmn: plusmn,
    plussim: plussim,
    plustwo: plustwo,
    pm: pm,
    Poincareplane: Poincareplane,
    pointint: pointint,
    popf: popf,
    Popf: Popf,
    pound: pound,
    prap: prap,
    Pr: Pr,
    pr: pr,
    prcue: prcue,
    precapprox: precapprox,
    prec: prec,
    preccurlyeq: preccurlyeq,
    Precedes: Precedes,
    PrecedesEqual: PrecedesEqual,
    PrecedesSlantEqual: PrecedesSlantEqual,
    PrecedesTilde: PrecedesTilde,
    preceq: preceq,
    precnapprox: precnapprox,
    precneqq: precneqq,
    precnsim: precnsim,
    pre: pre,
    prE: prE,
    precsim: precsim,
    prime: prime,
    Prime: Prime,
    primes: primes,
    prnap: prnap,
    prnE: prnE,
    prnsim: prnsim,
    prod: prod,
    Product: Product,
    profalar: profalar,
    profline: profline,
    profsurf: profsurf,
    prop: prop,
    Proportional: Proportional,
    Proportion: Proportion,
    propto: propto,
    prsim: prsim,
    prurel: prurel,
    Pscr: Pscr,
    pscr: pscr,
    Psi: Psi,
    psi: psi,
    puncsp: puncsp,
    Qfr: Qfr,
    qfr: qfr,
    qint: qint,
    qopf: qopf,
    Qopf: Qopf,
    qprime: qprime,
    Qscr: Qscr,
    qscr: qscr,
    quaternions: quaternions,
    quatint: quatint,
    quest: quest,
    questeq: questeq,
    quot: quot,
    QUOT: QUOT,
    rAarr: rAarr,
    race: race,
    Racute: Racute,
    racute: racute,
    radic: radic,
    raemptyv: raemptyv,
    rang: rang,
    Rang: Rang,
    rangd: rangd,
    range: range,
    rangle: rangle,
    raquo: raquo,
    rarrap: rarrap,
    rarrb: rarrb,
    rarrbfs: rarrbfs,
    rarrc: rarrc,
    rarr: rarr,
    Rarr: Rarr,
    rArr: rArr,
    rarrfs: rarrfs,
    rarrhk: rarrhk,
    rarrlp: rarrlp,
    rarrpl: rarrpl,
    rarrsim: rarrsim,
    Rarrtl: Rarrtl,
    rarrtl: rarrtl,
    rarrw: rarrw,
    ratail: ratail,
    rAtail: rAtail,
    ratio: ratio,
    rationals: rationals,
    rbarr: rbarr,
    rBarr: rBarr,
    RBarr: RBarr,
    rbbrk: rbbrk,
    rbrace: rbrace,
    rbrack: rbrack,
    rbrke: rbrke,
    rbrksld: rbrksld,
    rbrkslu: rbrkslu,
    Rcaron: Rcaron,
    rcaron: rcaron,
    Rcedil: Rcedil,
    rcedil: rcedil,
    rceil: rceil,
    rcub: rcub,
    Rcy: Rcy,
    rcy: rcy,
    rdca: rdca,
    rdldhar: rdldhar,
    rdquo: rdquo,
    rdquor: rdquor,
    rdsh: rdsh,
    real: real,
    realine: realine,
    realpart: realpart,
    reals: reals,
    Re: Re,
    rect: rect,
    reg: reg,
    REG: REG,
    ReverseElement: ReverseElement,
    ReverseEquilibrium: ReverseEquilibrium,
    ReverseUpEquilibrium: ReverseUpEquilibrium,
    rfisht: rfisht,
    rfloor: rfloor,
    rfr: rfr,
    Rfr: Rfr,
    rHar: rHar,
    rhard: rhard,
    rharu: rharu,
    rharul: rharul,
    Rho: Rho,
    rho: rho,
    rhov: rhov,
    RightAngleBracket: RightAngleBracket,
    RightArrowBar: RightArrowBar,
    rightarrow: rightarrow,
    RightArrow: RightArrow,
    Rightarrow: Rightarrow,
    RightArrowLeftArrow: RightArrowLeftArrow,
    rightarrowtail: rightarrowtail,
    RightCeiling: RightCeiling,
    RightDoubleBracket: RightDoubleBracket,
    RightDownTeeVector: RightDownTeeVector,
    RightDownVectorBar: RightDownVectorBar,
    RightDownVector: RightDownVector,
    RightFloor: RightFloor,
    rightharpoondown: rightharpoondown,
    rightharpoonup: rightharpoonup,
    rightleftarrows: rightleftarrows,
    rightleftharpoons: rightleftharpoons,
    rightrightarrows: rightrightarrows,
    rightsquigarrow: rightsquigarrow,
    RightTeeArrow: RightTeeArrow,
    RightTee: RightTee,
    RightTeeVector: RightTeeVector,
    rightthreetimes: rightthreetimes,
    RightTriangleBar: RightTriangleBar,
    RightTriangle: RightTriangle,
    RightTriangleEqual: RightTriangleEqual,
    RightUpDownVector: RightUpDownVector,
    RightUpTeeVector: RightUpTeeVector,
    RightUpVectorBar: RightUpVectorBar,
    RightUpVector: RightUpVector,
    RightVectorBar: RightVectorBar,
    RightVector: RightVector,
    ring: ring,
    risingdotseq: risingdotseq,
    rlarr: rlarr,
    rlhar: rlhar,
    rlm: rlm,
    rmoustache: rmoustache,
    rmoust: rmoust,
    rnmid: rnmid,
    roang: roang,
    roarr: roarr,
    robrk: robrk,
    ropar: ropar,
    ropf: ropf,
    Ropf: Ropf,
    roplus: roplus,
    rotimes: rotimes,
    RoundImplies: RoundImplies,
    rpar: rpar,
    rpargt: rpargt,
    rppolint: rppolint,
    rrarr: rrarr,
    Rrightarrow: Rrightarrow,
    rsaquo: rsaquo,
    rscr: rscr,
    Rscr: Rscr,
    rsh: rsh,
    Rsh: Rsh,
    rsqb: rsqb,
    rsquo: rsquo,
    rsquor: rsquor,
    rthree: rthree,
    rtimes: rtimes,
    rtri: rtri,
    rtrie: rtrie,
    rtrif: rtrif,
    rtriltri: rtriltri,
    RuleDelayed: RuleDelayed,
    ruluhar: ruluhar,
    rx: rx,
    Sacute: Sacute,
    sacute: sacute,
    sbquo: sbquo,
    scap: scap,
    Scaron: Scaron,
    scaron: scaron,
    Sc: Sc,
    sc: sc,
    sccue: sccue,
    sce: sce,
    scE: scE,
    Scedil: Scedil,
    scedil: scedil,
    Scirc: Scirc,
    scirc: scirc,
    scnap: scnap,
    scnE: scnE,
    scnsim: scnsim,
    scpolint: scpolint,
    scsim: scsim,
    Scy: Scy,
    scy: scy,
    sdotb: sdotb,
    sdot: sdot,
    sdote: sdote,
    searhk: searhk,
    searr: searr,
    seArr: seArr,
    searrow: searrow,
    sect: sect,
    semi: semi,
    seswar: seswar,
    setminus: setminus,
    setmn: setmn,
    sext: sext,
    Sfr: Sfr,
    sfr: sfr,
    sfrown: sfrown,
    sharp: sharp,
    SHCHcy: SHCHcy,
    shchcy: shchcy,
    SHcy: SHcy,
    shcy: shcy,
    ShortDownArrow: ShortDownArrow,
    ShortLeftArrow: ShortLeftArrow,
    shortmid: shortmid,
    shortparallel: shortparallel,
    ShortRightArrow: ShortRightArrow,
    ShortUpArrow: ShortUpArrow,
    shy: shy,
    Sigma: Sigma,
    sigma: sigma,
    sigmaf: sigmaf,
    sigmav: sigmav,
    sim: sim,
    simdot: simdot,
    sime: sime,
    simeq: simeq,
    simg: simg,
    simgE: simgE,
    siml: siml,
    simlE: simlE,
    simne: simne,
    simplus: simplus,
    simrarr: simrarr,
    slarr: slarr,
    SmallCircle: SmallCircle,
    smallsetminus: smallsetminus,
    smashp: smashp,
    smeparsl: smeparsl,
    smid: smid,
    smile: smile,
    smt: smt,
    smte: smte,
    smtes: smtes,
    SOFTcy: SOFTcy,
    softcy: softcy,
    solbar: solbar,
    solb: solb,
    sol: sol,
    Sopf: Sopf,
    sopf: sopf,
    spades: spades,
    spadesuit: spadesuit,
    spar: spar,
    sqcap: sqcap,
    sqcaps: sqcaps,
    sqcup: sqcup,
    sqcups: sqcups,
    Sqrt: Sqrt,
    sqsub: sqsub,
    sqsube: sqsube,
    sqsubset: sqsubset,
    sqsubseteq: sqsubseteq,
    sqsup: sqsup,
    sqsupe: sqsupe,
    sqsupset: sqsupset,
    sqsupseteq: sqsupseteq,
    square: square,
    Square: Square,
    SquareIntersection: SquareIntersection,
    SquareSubset: SquareSubset,
    SquareSubsetEqual: SquareSubsetEqual,
    SquareSuperset: SquareSuperset,
    SquareSupersetEqual: SquareSupersetEqual,
    SquareUnion: SquareUnion,
    squarf: squarf,
    squ: squ,
    squf: squf,
    srarr: srarr,
    Sscr: Sscr,
    sscr: sscr,
    ssetmn: ssetmn,
    ssmile: ssmile,
    sstarf: sstarf,
    Star: Star,
    star: star,
    starf: starf,
    straightepsilon: straightepsilon,
    straightphi: straightphi,
    strns: strns,
    sub: sub,
    Sub: Sub,
    subdot: subdot,
    subE: subE,
    sube: sube,
    subedot: subedot,
    submult: submult,
    subnE: subnE,
    subne: subne,
    subplus: subplus,
    subrarr: subrarr,
    subset: subset,
    Subset: Subset,
    subseteq: subseteq,
    subseteqq: subseteqq,
    SubsetEqual: SubsetEqual,
    subsetneq: subsetneq,
    subsetneqq: subsetneqq,
    subsim: subsim,
    subsub: subsub,
    subsup: subsup,
    succapprox: succapprox,
    succ: succ,
    succcurlyeq: succcurlyeq,
    Succeeds: Succeeds,
    SucceedsEqual: SucceedsEqual,
    SucceedsSlantEqual: SucceedsSlantEqual,
    SucceedsTilde: SucceedsTilde,
    succeq: succeq,
    succnapprox: succnapprox,
    succneqq: succneqq,
    succnsim: succnsim,
    succsim: succsim,
    SuchThat: SuchThat,
    sum: sum,
    Sum: Sum,
    sung: sung,
    sup1: sup1,
    sup2: sup2,
    sup3: sup3,
    sup: sup,
    Sup: Sup,
    supdot: supdot,
    supdsub: supdsub,
    supE: supE,
    supe: supe,
    supedot: supedot,
    Superset: Superset,
    SupersetEqual: SupersetEqual,
    suphsol: suphsol,
    suphsub: suphsub,
    suplarr: suplarr,
    supmult: supmult,
    supnE: supnE,
    supne: supne,
    supplus: supplus,
    supset: supset,
    Supset: Supset,
    supseteq: supseteq,
    supseteqq: supseteqq,
    supsetneq: supsetneq,
    supsetneqq: supsetneqq,
    supsim: supsim,
    supsub: supsub,
    supsup: supsup,
    swarhk: swarhk,
    swarr: swarr,
    swArr: swArr,
    swarrow: swarrow,
    swnwar: swnwar,
    szlig: szlig,
    Tab: Tab,
    target: target,
    Tau: Tau,
    tau: tau,
    tbrk: tbrk,
    Tcaron: Tcaron,
    tcaron: tcaron,
    Tcedil: Tcedil,
    tcedil: tcedil,
    Tcy: Tcy,
    tcy: tcy,
    tdot: tdot,
    telrec: telrec,
    Tfr: Tfr,
    tfr: tfr,
    there4: there4,
    therefore: therefore,
    Therefore: Therefore,
    Theta: Theta,
    theta: theta,
    thetasym: thetasym,
    thetav: thetav,
    thickapprox: thickapprox,
    thicksim: thicksim,
    ThickSpace: ThickSpace,
    ThinSpace: ThinSpace,
    thinsp: thinsp,
    thkap: thkap,
    thksim: thksim,
    THORN: THORN,
    thorn: thorn,
    tilde: tilde,
    Tilde: Tilde,
    TildeEqual: TildeEqual,
    TildeFullEqual: TildeFullEqual,
    TildeTilde: TildeTilde,
    timesbar: timesbar,
    timesb: timesb,
    times: times,
    timesd: timesd,
    tint: tint,
    toea: toea,
    topbot: topbot,
    topcir: topcir,
    top: top,
    Topf: Topf,
    topf: topf,
    topfork: topfork,
    tosa: tosa,
    tprime: tprime,
    trade: trade,
    TRADE: TRADE,
    triangle: triangle,
    triangledown: triangledown,
    triangleleft: triangleleft,
    trianglelefteq: trianglelefteq,
    triangleq: triangleq,
    triangleright: triangleright,
    trianglerighteq: trianglerighteq,
    tridot: tridot,
    trie: trie,
    triminus: triminus,
    TripleDot: TripleDot,
    triplus: triplus,
    trisb: trisb,
    tritime: tritime,
    trpezium: trpezium,
    Tscr: Tscr,
    tscr: tscr,
    TScy: TScy,
    tscy: tscy,
    TSHcy: TSHcy,
    tshcy: tshcy,
    Tstrok: Tstrok,
    tstrok: tstrok,
    twixt: twixt,
    twoheadleftarrow: twoheadleftarrow,
    twoheadrightarrow: twoheadrightarrow,
    Uacute: Uacute,
    uacute: uacute,
    uarr: uarr,
    Uarr: Uarr,
    uArr: uArr,
    Uarrocir: Uarrocir,
    Ubrcy: Ubrcy,
    ubrcy: ubrcy,
    Ubreve: Ubreve,
    ubreve: ubreve,
    Ucirc: Ucirc,
    ucirc: ucirc,
    Ucy: Ucy,
    ucy: ucy,
    udarr: udarr,
    Udblac: Udblac,
    udblac: udblac,
    udhar: udhar,
    ufisht: ufisht,
    Ufr: Ufr,
    ufr: ufr,
    Ugrave: Ugrave,
    ugrave: ugrave,
    uHar: uHar,
    uharl: uharl,
    uharr: uharr,
    uhblk: uhblk,
    ulcorn: ulcorn,
    ulcorner: ulcorner,
    ulcrop: ulcrop,
    ultri: ultri,
    Umacr: Umacr,
    umacr: umacr,
    uml: uml,
    UnderBar: UnderBar,
    UnderBrace: UnderBrace,
    UnderBracket: UnderBracket,
    UnderParenthesis: UnderParenthesis,
    Union: Union,
    UnionPlus: UnionPlus,
    Uogon: Uogon,
    uogon: uogon,
    Uopf: Uopf,
    uopf: uopf,
    UpArrowBar: UpArrowBar,
    uparrow: uparrow,
    UpArrow: UpArrow,
    Uparrow: Uparrow,
    UpArrowDownArrow: UpArrowDownArrow,
    updownarrow: updownarrow,
    UpDownArrow: UpDownArrow,
    Updownarrow: Updownarrow,
    UpEquilibrium: UpEquilibrium,
    upharpoonleft: upharpoonleft,
    upharpoonright: upharpoonright,
    uplus: uplus,
    UpperLeftArrow: UpperLeftArrow,
    UpperRightArrow: UpperRightArrow,
    upsi: upsi,
    Upsi: Upsi,
    upsih: upsih,
    Upsilon: Upsilon,
    upsilon: upsilon,
    UpTeeArrow: UpTeeArrow,
    UpTee: UpTee,
    upuparrows: upuparrows,
    urcorn: urcorn,
    urcorner: urcorner,
    urcrop: urcrop,
    Uring: Uring,
    uring: uring,
    urtri: urtri,
    Uscr: Uscr,
    uscr: uscr,
    utdot: utdot,
    Utilde: Utilde,
    utilde: utilde,
    utri: utri,
    utrif: utrif,
    uuarr: uuarr,
    Uuml: Uuml,
    uuml: uuml,
    uwangle: uwangle,
    vangrt: vangrt,
    varepsilon: varepsilon,
    varkappa: varkappa,
    varnothing: varnothing,
    varphi: varphi,
    varpi: varpi,
    varpropto: varpropto,
    varr: varr,
    vArr: vArr,
    varrho: varrho,
    varsigma: varsigma,
    varsubsetneq: varsubsetneq,
    varsubsetneqq: varsubsetneqq,
    varsupsetneq: varsupsetneq,
    varsupsetneqq: varsupsetneqq,
    vartheta: vartheta,
    vartriangleleft: vartriangleleft,
    vartriangleright: vartriangleright,
    vBar: vBar,
    Vbar: Vbar,
    vBarv: vBarv,
    Vcy: Vcy,
    vcy: vcy,
    vdash: vdash,
    vDash: vDash,
    Vdash: Vdash,
    VDash: VDash,
    Vdashl: Vdashl,
    veebar: veebar,
    vee: vee,
    Vee: Vee,
    veeeq: veeeq,
    vellip: vellip,
    verbar: verbar,
    Verbar: Verbar,
    vert: vert,
    Vert: Vert,
    VerticalBar: VerticalBar,
    VerticalLine: VerticalLine,
    VerticalSeparator: VerticalSeparator,
    VerticalTilde: VerticalTilde,
    VeryThinSpace: VeryThinSpace,
    Vfr: Vfr,
    vfr: vfr,
    vltri: vltri,
    vnsub: vnsub,
    vnsup: vnsup,
    Vopf: Vopf,
    vopf: vopf,
    vprop: vprop,
    vrtri: vrtri,
    Vscr: Vscr,
    vscr: vscr,
    vsubnE: vsubnE,
    vsubne: vsubne,
    vsupnE: vsupnE,
    vsupne: vsupne,
    Vvdash: Vvdash,
    vzigzag: vzigzag,
    Wcirc: Wcirc,
    wcirc: wcirc,
    wedbar: wedbar,
    wedge: wedge,
    Wedge: Wedge,
    wedgeq: wedgeq,
    weierp: weierp,
    Wfr: Wfr,
    wfr: wfr,
    Wopf: Wopf,
    wopf: wopf,
    wp: wp,
    wr: wr,
    wreath: wreath,
    Wscr: Wscr,
    wscr: wscr,
    xcap: xcap,
    xcirc: xcirc,
    xcup: xcup,
    xdtri: xdtri,
    Xfr: Xfr,
    xfr: xfr,
    xharr: xharr,
    xhArr: xhArr,
    Xi: Xi,
    xi: xi,
    xlarr: xlarr,
    xlArr: xlArr,
    xmap: xmap,
    xnis: xnis,
    xodot: xodot,
    Xopf: Xopf,
    xopf: xopf,
    xoplus: xoplus,
    xotime: xotime,
    xrarr: xrarr,
    xrArr: xrArr,
    Xscr: Xscr,
    xscr: xscr,
    xsqcup: xsqcup,
    xuplus: xuplus,
    xutri: xutri,
    xvee: xvee,
    xwedge: xwedge,
    Yacute: Yacute,
    yacute: yacute,
    YAcy: YAcy,
    yacy: yacy,
    Ycirc: Ycirc,
    ycirc: ycirc,
    Ycy: Ycy,
    ycy: ycy,
    yen: yen,
    Yfr: Yfr,
    yfr: yfr,
    YIcy: YIcy,
    yicy: yicy,
    Yopf: Yopf,
    yopf: yopf,
    Yscr: Yscr,
    yscr: yscr,
    YUcy: YUcy,
    yucy: yucy,
    yuml: yuml,
    Yuml: Yuml,
    Zacute: Zacute,
    zacute: zacute,
    Zcaron: Zcaron,
    zcaron: zcaron,
    Zcy: Zcy,
    zcy: zcy,
    Zdot: Zdot,
    zdot: zdot,
    zeetrf: zeetrf,
    ZeroWidthSpace: ZeroWidthSpace,
    Zeta: Zeta,
    zeta: zeta,
    zfr: zfr,
    Zfr: Zfr,
    ZHcy: ZHcy,
    zhcy: zhcy,
    zigrarr: zigrarr,
    zopf: zopf,
    Zopf: Zopf,
    Zscr: Zscr,
    zscr: zscr,
    zwj: zwj,
    zwnj: zwnj,
    'default': entities$1
  });

  var require$$0 = getCjsExportFromNamespace(entities$2);

  /*eslint quotes:0*/
  var entities = require$$0;

  var utils = createCommonjsModule(function (module, exports) {


  function _class(obj) { return Object.prototype.toString.call(obj); }

  function isString(obj) { return _class(obj) === '[object String]'; }

  var _hasOwnProperty = Object.prototype.hasOwnProperty;

  function has(object, key) {
    return _hasOwnProperty.call(object, key);
  }

  // Merge objects
  //
  function assign(obj /*from1, from2, from3, ...*/) {
    var sources = Array.prototype.slice.call(arguments, 1);

    sources.forEach(function (source) {
      if (!source) { return; }

      if (typeof source !== 'object') {
        throw new TypeError(source + 'must be object');
      }

      Object.keys(source).forEach(function (key) {
        obj[key] = source[key];
      });
    });

    return obj;
  }

  // Remove element from array and put another array at those position.
  // Useful for some operations with tokens
  function arrayReplaceAt(src, pos, newElements) {
    return [].concat(src.slice(0, pos), newElements, src.slice(pos + 1));
  }

  ////////////////////////////////////////////////////////////////////////////////

  function isValidEntityCode(c) {
    /*eslint no-bitwise:0*/
    // broken sequence
    if (c >= 0xD800 && c <= 0xDFFF) { return false; }
    // never used
    if (c >= 0xFDD0 && c <= 0xFDEF) { return false; }
    if ((c & 0xFFFF) === 0xFFFF || (c & 0xFFFF) === 0xFFFE) { return false; }
    // control codes
    if (c >= 0x00 && c <= 0x08) { return false; }
    if (c === 0x0B) { return false; }
    if (c >= 0x0E && c <= 0x1F) { return false; }
    if (c >= 0x7F && c <= 0x9F) { return false; }
    // out of range
    if (c > 0x10FFFF) { return false; }
    return true;
  }

  function fromCodePoint(c) {
    /*eslint no-bitwise:0*/
    if (c > 0xffff) {
      c -= 0x10000;
      var surrogate1 = 0xd800 + (c >> 10),
          surrogate2 = 0xdc00 + (c & 0x3ff);

      return String.fromCharCode(surrogate1, surrogate2);
    }
    return String.fromCharCode(c);
  }


  var UNESCAPE_MD_RE  = /\\([!"#$%&'()*+,\-.\/:;<=>?@[\\\]^_`{|}~])/g;
  var ENTITY_RE       = /&([a-z#][a-z0-9]{1,31});/gi;
  var UNESCAPE_ALL_RE = new RegExp(UNESCAPE_MD_RE.source + '|' + ENTITY_RE.source, 'gi');

  var DIGITAL_ENTITY_TEST_RE = /^#((?:x[a-f0-9]{1,8}|[0-9]{1,8}))/i;



  function replaceEntityPattern(match, name) {
    var code = 0;

    if (has(entities, name)) {
      return entities[name];
    }

    if (name.charCodeAt(0) === 0x23/* # */ && DIGITAL_ENTITY_TEST_RE.test(name)) {
      code = name[1].toLowerCase() === 'x' ?
        parseInt(name.slice(2), 16) : parseInt(name.slice(1), 10);

      if (isValidEntityCode(code)) {
        return fromCodePoint(code);
      }
    }

    return match;
  }

  /*function replaceEntities(str) {
    if (str.indexOf('&') < 0) { return str; }

    return str.replace(ENTITY_RE, replaceEntityPattern);
  }*/

  function unescapeMd(str) {
    if (str.indexOf('\\') < 0) { return str; }
    return str.replace(UNESCAPE_MD_RE, '$1');
  }

  function unescapeAll(str) {
    if (str.indexOf('\\') < 0 && str.indexOf('&') < 0) { return str; }

    return str.replace(UNESCAPE_ALL_RE, function (match, escaped, entity) {
      if (escaped) { return escaped; }
      return replaceEntityPattern(match, entity);
    });
  }

  ////////////////////////////////////////////////////////////////////////////////

  var HTML_ESCAPE_TEST_RE = /[&<>"]/;
  var HTML_ESCAPE_REPLACE_RE = /[&<>"]/g;
  var HTML_REPLACEMENTS = {
    '&': '&amp;',
    '<': '&lt;',
    '>': '&gt;',
    '"': '&quot;'
  };

  function replaceUnsafeChar(ch) {
    return HTML_REPLACEMENTS[ch];
  }

  function escapeHtml(str) {
    if (HTML_ESCAPE_TEST_RE.test(str)) {
      return str.replace(HTML_ESCAPE_REPLACE_RE, replaceUnsafeChar);
    }
    return str;
  }

  ////////////////////////////////////////////////////////////////////////////////

  var REGEXP_ESCAPE_RE = /[.?*+^$[\]\\(){}|-]/g;

  function escapeRE(str) {
    return str.replace(REGEXP_ESCAPE_RE, '\\$&');
  }

  ////////////////////////////////////////////////////////////////////////////////

  function isSpace(code) {
    switch (code) {
      case 0x09:
      case 0x20:
        return true;
    }
    return false;
  }

  // Zs (unicode class) || [\t\f\v\r\n]
  function isWhiteSpace(code) {
    if (code >= 0x2000 && code <= 0x200A) { return true; }
    switch (code) {
      case 0x09: // \t
      case 0x0A: // \n
      case 0x0B: // \v
      case 0x0C: // \f
      case 0x0D: // \r
      case 0x20:
      case 0xA0:
      case 0x1680:
      case 0x202F:
      case 0x205F:
      case 0x3000:
        return true;
    }
    return false;
  }

  ////////////////////////////////////////////////////////////////////////////////

  /*eslint-disable max-len*/


  // Currently without astral characters support.
  function isPunctChar(ch) {
    return regex$4.test(ch);
  }


  // Markdown ASCII punctuation characters.
  //
  // !, ", #, $, %, &, ', (, ), *, +, ,, -, ., /, :, ;, <, =, >, ?, @, [, \, ], ^, _, `, {, |, }, or ~
  // http://spec.commonmark.org/0.15/#ascii-punctuation-character
  //
  // Don't confuse with unicode punctuation !!! It lacks some chars in ascii range.
  //
  function isMdAsciiPunct(ch) {
    switch (ch) {
      case 0x21/* ! */:
      case 0x22/* " */:
      case 0x23/* # */:
      case 0x24/* $ */:
      case 0x25/* % */:
      case 0x26/* & */:
      case 0x27/* ' */:
      case 0x28/* ( */:
      case 0x29/* ) */:
      case 0x2A/* * */:
      case 0x2B/* + */:
      case 0x2C/* , */:
      case 0x2D/* - */:
      case 0x2E/* . */:
      case 0x2F/* / */:
      case 0x3A/* : */:
      case 0x3B/* ; */:
      case 0x3C/* < */:
      case 0x3D/* = */:
      case 0x3E/* > */:
      case 0x3F/* ? */:
      case 0x40/* @ */:
      case 0x5B/* [ */:
      case 0x5C/* \ */:
      case 0x5D/* ] */:
      case 0x5E/* ^ */:
      case 0x5F/* _ */:
      case 0x60/* ` */:
      case 0x7B/* { */:
      case 0x7C/* | */:
      case 0x7D/* } */:
      case 0x7E/* ~ */:
        return true;
      default:
        return false;
    }
  }

  // Hepler to unify [reference labels].
  //
  function normalizeReference(str) {
    // Trim and collapse whitespace
    //
    str = str.trim().replace(/\s+/g, ' ');

    // In node v10 'ẞ'.toLowerCase() === 'Ṿ', which is presumed to be a bug
    // fixed in v12 (couldn't find any details).
    //
    // So treat this one as a special case
    // (remove this when node v10 is no longer supported).
    //
    if ('ẞ'.toLowerCase() === 'Ṿ') {
      str = str.replace(/ẞ/g, 'ß');
    }

    // .toLowerCase().toUpperCase() should get rid of all differences
    // between letter variants.
    //
    // Simple .toLowerCase() doesn't normalize 125 code points correctly,
    // and .toUpperCase doesn't normalize 6 of them (list of exceptions:
    // İ, ϴ, ẞ, Ω, K, Å - those are already uppercased, but have differently
    // uppercased versions).
    //
    // Here's an example showing how it happens. Lets take greek letter omega:
    // uppercase U+0398 (Θ), U+03f4 (ϴ) and lowercase U+03b8 (θ), U+03d1 (ϑ)
    //
    // Unicode entries:
    // 0398;GREEK CAPITAL LETTER THETA;Lu;0;L;;;;;N;;;;03B8;
    // 03B8;GREEK SMALL LETTER THETA;Ll;0;L;;;;;N;;;0398;;0398
    // 03D1;GREEK THETA SYMBOL;Ll;0;L;<compat> 03B8;;;;N;GREEK SMALL LETTER SCRIPT THETA;;0398;;0398
    // 03F4;GREEK CAPITAL THETA SYMBOL;Lu;0;L;<compat> 0398;;;;N;;;;03B8;
    //
    // Case-insensitive comparison should treat all of them as equivalent.
    //
    // But .toLowerCase() doesn't change ϑ (it's already lowercase),
    // and .toUpperCase() doesn't change ϴ (already uppercase).
    //
    // Applying first lower then upper case normalizes any character:
    // '\u0398\u03f4\u03b8\u03d1'.toLowerCase().toUpperCase() === '\u0398\u0398\u0398\u0398'
    //
    // Note: this is equivalent to unicode case folding; unicode normalization
    // is a different step that is not required here.
    //
    // Final result should be uppercased, because it's later stored in an object
    // (this avoid a conflict with Object.prototype members,
    // most notably, `__proto__`)
    //
    return str.toLowerCase().toUpperCase();
  }

  ////////////////////////////////////////////////////////////////////////////////

  // Re-export libraries commonly used in both markdown-it and its plugins,
  // so plugins won't have to depend on them explicitly, which reduces their
  // bundled size (e.g. a browser build).
  //
  exports.lib                 = {};
  exports.lib.mdurl           = mdurl;
  exports.lib.ucmicro         = uc_micro;

  exports.assign              = assign;
  exports.isString            = isString;
  exports.has                 = has;
  exports.unescapeMd          = unescapeMd;
  exports.unescapeAll         = unescapeAll;
  exports.isValidEntityCode   = isValidEntityCode;
  exports.fromCodePoint       = fromCodePoint;
  // exports.replaceEntities     = replaceEntities;
  exports.escapeHtml          = escapeHtml;
  exports.arrayReplaceAt      = arrayReplaceAt;
  exports.isSpace             = isSpace;
  exports.isWhiteSpace        = isWhiteSpace;
  exports.isMdAsciiPunct      = isMdAsciiPunct;
  exports.isPunctChar         = isPunctChar;
  exports.escapeRE            = escapeRE;
  exports.normalizeReference  = normalizeReference;
  });
  utils.lib;
  utils.assign;
  utils.isString;
  utils.has;
  utils.unescapeMd;
  utils.unescapeAll;
  utils.isValidEntityCode;
  utils.fromCodePoint;
  utils.escapeHtml;
  utils.arrayReplaceAt;
  utils.isSpace;
  utils.isWhiteSpace;
  utils.isMdAsciiPunct;
  utils.isPunctChar;
  utils.escapeRE;
  utils.normalizeReference;

  // Parse link label

  var parse_link_label = function parseLinkLabel(state, start, disableNested) {
    var level, found, marker, prevPos,
        labelEnd = -1,
        max = state.posMax,
        oldPos = state.pos;

    state.pos = start + 1;
    level = 1;

    while (state.pos < max) {
      marker = state.src.charCodeAt(state.pos);
      if (marker === 0x5D /* ] */) {
        level--;
        if (level === 0) {
          found = true;
          break;
        }
      }

      prevPos = state.pos;
      state.md.inline.skipToken(state);
      if (marker === 0x5B /* [ */) {
        if (prevPos === state.pos - 1) {
          // increase level if we find text `[`, which is not a part of any token
          level++;
        } else if (disableNested) {
          state.pos = oldPos;
          return -1;
        }
      }
    }

    if (found) {
      labelEnd = state.pos;
    }

    // restore old state
    state.pos = oldPos;

    return labelEnd;
  };

  var unescapeAll$2 = utils.unescapeAll;


  var parse_link_destination = function parseLinkDestination(str, pos, max) {
    var code, level,
        lines = 0,
        start = pos,
        result = {
          ok: false,
          pos: 0,
          lines: 0,
          str: ''
        };

    if (str.charCodeAt(pos) === 0x3C /* < */) {
      pos++;
      while (pos < max) {
        code = str.charCodeAt(pos);
        if (code === 0x0A /* \n */) { return result; }
        if (code === 0x3C /* < */) { return result; }
        if (code === 0x3E /* > */) {
          result.pos = pos + 1;
          result.str = unescapeAll$2(str.slice(start + 1, pos));
          result.ok = true;
          return result;
        }
        if (code === 0x5C /* \ */ && pos + 1 < max) {
          pos += 2;
          continue;
        }

        pos++;
      }

      // no closing '>'
      return result;
    }

    // this should be ... } else { ... branch

    level = 0;
    while (pos < max) {
      code = str.charCodeAt(pos);

      if (code === 0x20) { break; }

      // ascii control characters
      if (code < 0x20 || code === 0x7F) { break; }

      if (code === 0x5C /* \ */ && pos + 1 < max) {
        if (str.charCodeAt(pos + 1) === 0x20) { break; }
        pos += 2;
        continue;
      }

      if (code === 0x28 /* ( */) {
        level++;
        if (level > 32) { return result; }
      }

      if (code === 0x29 /* ) */) {
        if (level === 0) { break; }
        level--;
      }

      pos++;
    }

    if (start === pos) { return result; }
    if (level !== 0) { return result; }

    result.str = unescapeAll$2(str.slice(start, pos));
    result.lines = lines;
    result.pos = pos;
    result.ok = true;
    return result;
  };

  var unescapeAll$1 = utils.unescapeAll;


  var parse_link_title = function parseLinkTitle(str, pos, max) {
    var code,
        marker,
        lines = 0,
        start = pos,
        result = {
          ok: false,
          pos: 0,
          lines: 0,
          str: ''
        };

    if (pos >= max) { return result; }

    marker = str.charCodeAt(pos);

    if (marker !== 0x22 /* " */ && marker !== 0x27 /* ' */ && marker !== 0x28 /* ( */) { return result; }

    pos++;

    // if opening marker is "(", switch it to closing marker ")"
    if (marker === 0x28) { marker = 0x29; }

    while (pos < max) {
      code = str.charCodeAt(pos);
      if (code === marker) {
        result.pos = pos + 1;
        result.lines = lines;
        result.str = unescapeAll$1(str.slice(start + 1, pos));
        result.ok = true;
        return result;
      } else if (code === 0x28 /* ( */ && marker === 0x29 /* ) */) {
        return result;
      } else if (code === 0x0A) {
        lines++;
      } else if (code === 0x5C /* \ */ && pos + 1 < max) {
        pos++;
        if (str.charCodeAt(pos) === 0x0A) {
          lines++;
        }
      }

      pos++;
    }

    return result;
  };

  var parseLinkLabel       = parse_link_label;
  var parseLinkDestination = parse_link_destination;
  var parseLinkTitle       = parse_link_title;

  var helpers = {
  	parseLinkLabel: parseLinkLabel,
  	parseLinkDestination: parseLinkDestination,
  	parseLinkTitle: parseLinkTitle
  };

  var assign$1          = utils.assign;
  var unescapeAll     = utils.unescapeAll;
  var escapeHtml      = utils.escapeHtml;


  ////////////////////////////////////////////////////////////////////////////////

  var default_rules = {};


  default_rules.code_inline = function (tokens, idx, options, env, slf) {
    var token = tokens[idx];

    return  '<code' + slf.renderAttrs(token) + '>' +
            escapeHtml(tokens[idx].content) +
            '</code>';
  };


  default_rules.code_block = function (tokens, idx, options, env, slf) {
    var token = tokens[idx];

    return  '<pre' + slf.renderAttrs(token) + '><code>' +
            escapeHtml(tokens[idx].content) +
            '</code></pre>\n';
  };


  default_rules.fence = function (tokens, idx, options, env, slf) {
    var token = tokens[idx],
        info = token.info ? unescapeAll(token.info).trim() : '',
        langName = '',
        langAttrs = '',
        highlighted, i, arr, tmpAttrs, tmpToken;

    if (info) {
      arr = info.split(/(\s+)/g);
      langName = arr[0];
      langAttrs = arr.slice(2).join('');
    }

    if (options.highlight) {
      highlighted = options.highlight(token.content, langName, langAttrs) || escapeHtml(token.content);
    } else {
      highlighted = escapeHtml(token.content);
    }

    if (highlighted.indexOf('<pre') === 0) {
      return highlighted + '\n';
    }

    // If language exists, inject class gently, without modifying original token.
    // May be, one day we will add .deepClone() for token and simplify this part, but
    // now we prefer to keep things local.
    if (info) {
      i        = token.attrIndex('class');
      tmpAttrs = token.attrs ? token.attrs.slice() : [];

      if (i < 0) {
        tmpAttrs.push([ 'class', options.langPrefix + langName ]);
      } else {
        tmpAttrs[i] = tmpAttrs[i].slice();
        tmpAttrs[i][1] += ' ' + options.langPrefix + langName;
      }

      // Fake token just to render attributes
      tmpToken = {
        attrs: tmpAttrs
      };

      return  '<pre><code' + slf.renderAttrs(tmpToken) + '>'
            + highlighted
            + '</code></pre>\n';
    }


    return  '<pre><code' + slf.renderAttrs(token) + '>'
          + highlighted
          + '</code></pre>\n';
  };


  default_rules.image = function (tokens, idx, options, env, slf) {
    var token = tokens[idx];

    // "alt" attr MUST be set, even if empty. Because it's mandatory and
    // should be placed on proper position for tests.
    //
    // Replace content with actual value

    token.attrs[token.attrIndex('alt')][1] =
      slf.renderInlineAsText(token.children, options, env);

    return slf.renderToken(tokens, idx, options);
  };


  default_rules.hardbreak = function (tokens, idx, options /*, env */) {
    return options.xhtmlOut ? '<br />\n' : '<br>\n';
  };
  default_rules.softbreak = function (tokens, idx, options /*, env */) {
    return options.breaks ? (options.xhtmlOut ? '<br />\n' : '<br>\n') : '\n';
  };


  default_rules.text = function (tokens, idx /*, options, env */) {
    return escapeHtml(tokens[idx].content);
  };


  default_rules.html_block = function (tokens, idx /*, options, env */) {
    return tokens[idx].content;
  };
  default_rules.html_inline = function (tokens, idx /*, options, env */) {
    return tokens[idx].content;
  };


  /**
   * new Renderer()
   *
   * Creates new [[Renderer]] instance and fill [[Renderer#rules]] with defaults.
   **/
  function Renderer() {

    /**
     * Renderer#rules -> Object
     *
     * Contains render rules for tokens. Can be updated and extended.
     *
     * ##### Example
     *
     * ```javascript
     * var md = require('markdown-it')();
     *
     * md.renderer.rules.strong_open  = function () { return '<b>'; };
     * md.renderer.rules.strong_close = function () { return '</b>'; };
     *
     * var result = md.renderInline(...);
     * ```
     *
     * Each rule is called as independent static function with fixed signature:
     *
     * ```javascript
     * function my_token_render(tokens, idx, options, env, renderer) {
     *   // ...
     *   return renderedHTML;
     * }
     * ```
     *
     * See [source code](https://github.com/markdown-it/markdown-it/blob/master/lib/renderer.js)
     * for more details and examples.
     **/
    this.rules = assign$1({}, default_rules);
  }


  /**
   * Renderer.renderAttrs(token) -> String
   *
   * Render token attributes to string.
   **/
  Renderer.prototype.renderAttrs = function renderAttrs(token) {
    var i, l, result;

    if (!token.attrs) { return ''; }

    result = '';

    for (i = 0, l = token.attrs.length; i < l; i++) {
      result += ' ' + escapeHtml(token.attrs[i][0]) + '="' + escapeHtml(token.attrs[i][1]) + '"';
    }

    return result;
  };


  /**
   * Renderer.renderToken(tokens, idx, options) -> String
   * - tokens (Array): list of tokens
   * - idx (Numbed): token index to render
   * - options (Object): params of parser instance
   *
   * Default token renderer. Can be overriden by custom function
   * in [[Renderer#rules]].
   **/
  Renderer.prototype.renderToken = function renderToken(tokens, idx, options) {
    var nextToken,
        result = '',
        needLf = false,
        token = tokens[idx];

    // Tight list paragraphs
    if (token.hidden) {
      return '';
    }

    // Insert a newline between hidden paragraph and subsequent opening
    // block-level tag.
    //
    // For example, here we should insert a newline before blockquote:
    //  - a
    //    >
    //
    if (token.block && token.nesting !== -1 && idx && tokens[idx - 1].hidden) {
      result += '\n';
    }

    // Add token name, e.g. `<img`
    result += (token.nesting === -1 ? '</' : '<') + token.tag;

    // Encode attributes, e.g. `<img src="foo"`
    result += this.renderAttrs(token);

    // Add a slash for self-closing tags, e.g. `<img src="foo" /`
    if (token.nesting === 0 && options.xhtmlOut) {
      result += ' /';
    }

    // Check if we need to add a newline after this tag
    if (token.block) {
      needLf = true;

      if (token.nesting === 1) {
        if (idx + 1 < tokens.length) {
          nextToken = tokens[idx + 1];

          if (nextToken.type === 'inline' || nextToken.hidden) {
            // Block-level tag containing an inline tag.
            //
            needLf = false;

          } else if (nextToken.nesting === -1 && nextToken.tag === token.tag) {
            // Opening tag + closing tag of the same type. E.g. `<li></li>`.
            //
            needLf = false;
          }
        }
      }
    }

    result += needLf ? '>\n' : '>';

    return result;
  };


  /**
   * Renderer.renderInline(tokens, options, env) -> String
   * - tokens (Array): list on block tokens to renter
   * - options (Object): params of parser instance
   * - env (Object): additional data from parsed input (references, for example)
   *
   * The same as [[Renderer.render]], but for single token of `inline` type.
   **/
  Renderer.prototype.renderInline = function (tokens, options, env) {
    var type,
        result = '',
        rules = this.rules;

    for (var i = 0, len = tokens.length; i < len; i++) {
      type = tokens[i].type;

      if (typeof rules[type] !== 'undefined') {
        result += rules[type](tokens, i, options, env, this);
      } else {
        result += this.renderToken(tokens, i, options);
      }
    }

    return result;
  };


  /** internal
   * Renderer.renderInlineAsText(tokens, options, env) -> String
   * - tokens (Array): list on block tokens to renter
   * - options (Object): params of parser instance
   * - env (Object): additional data from parsed input (references, for example)
   *
   * Special kludge for image `alt` attributes to conform CommonMark spec.
   * Don't try to use it! Spec requires to show `alt` content with stripped markup,
   * instead of simple escaping.
   **/
  Renderer.prototype.renderInlineAsText = function (tokens, options, env) {
    var result = '';

    for (var i = 0, len = tokens.length; i < len; i++) {
      if (tokens[i].type === 'text') {
        result += tokens[i].content;
      } else if (tokens[i].type === 'image') {
        result += this.renderInlineAsText(tokens[i].children, options, env);
      } else if (tokens[i].type === 'softbreak') {
        result += '\n';
      }
    }

    return result;
  };


  /**
   * Renderer.render(tokens, options, env) -> String
   * - tokens (Array): list on block tokens to renter
   * - options (Object): params of parser instance
   * - env (Object): additional data from parsed input (references, for example)
   *
   * Takes token stream and generates HTML. Probably, you will never need to call
   * this method directly.
   **/
  Renderer.prototype.render = function (tokens, options, env) {
    var i, len, type,
        result = '',
        rules = this.rules;

    for (i = 0, len = tokens.length; i < len; i++) {
      type = tokens[i].type;

      if (type === 'inline') {
        result += this.renderInline(tokens[i].children, options, env);
      } else if (typeof rules[type] !== 'undefined') {
        result += rules[tokens[i].type](tokens, i, options, env, this);
      } else {
        result += this.renderToken(tokens, i, options, env);
      }
    }

    return result;
  };

  var renderer = Renderer;

  /**
   * class Ruler
   *
   * Helper class, used by [[MarkdownIt#core]], [[MarkdownIt#block]] and
   * [[MarkdownIt#inline]] to manage sequences of functions (rules):
   *
   * - keep rules in defined order
   * - assign the name to each rule
   * - enable/disable rules
   * - add/replace rules
   * - allow assign rules to additional named chains (in the same)
   * - cacheing lists of active rules
   *
   * You will not need use this class directly until write plugins. For simple
   * rules control use [[MarkdownIt.disable]], [[MarkdownIt.enable]] and
   * [[MarkdownIt.use]].
   **/


  /**
   * new Ruler()
   **/
  function Ruler() {
    // List of added rules. Each element is:
    //
    // {
    //   name: XXX,
    //   enabled: Boolean,
    //   fn: Function(),
    //   alt: [ name2, name3 ]
    // }
    //
    this.__rules__ = [];

    // Cached rule chains.
    //
    // First level - chain name, '' for default.
    // Second level - diginal anchor for fast filtering by charcodes.
    //
    this.__cache__ = null;
  }

  ////////////////////////////////////////////////////////////////////////////////
  // Helper methods, should not be used directly


  // Find rule index by name
  //
  Ruler.prototype.__find__ = function (name) {
    for (var i = 0; i < this.__rules__.length; i++) {
      if (this.__rules__[i].name === name) {
        return i;
      }
    }
    return -1;
  };


  // Build rules lookup cache
  //
  Ruler.prototype.__compile__ = function () {
    var self = this;
    var chains = [ '' ];

    // collect unique names
    self.__rules__.forEach(function (rule) {
      if (!rule.enabled) { return; }

      rule.alt.forEach(function (altName) {
        if (chains.indexOf(altName) < 0) {
          chains.push(altName);
        }
      });
    });

    self.__cache__ = {};

    chains.forEach(function (chain) {
      self.__cache__[chain] = [];
      self.__rules__.forEach(function (rule) {
        if (!rule.enabled) { return; }

        if (chain && rule.alt.indexOf(chain) < 0) { return; }

        self.__cache__[chain].push(rule.fn);
      });
    });
  };


  /**
   * Ruler.at(name, fn [, options])
   * - name (String): rule name to replace.
   * - fn (Function): new rule function.
   * - options (Object): new rule options (not mandatory).
   *
   * Replace rule by name with new function & options. Throws error if name not
   * found.
   *
   * ##### Options:
   *
   * - __alt__ - array with names of "alternate" chains.
   *
   * ##### Example
   *
   * Replace existing typographer replacement rule with new one:
   *
   * ```javascript
   * var md = require('markdown-it')();
   *
   * md.core.ruler.at('replacements', function replace(state) {
   *   //...
   * });
   * ```
   **/
  Ruler.prototype.at = function (name, fn, options) {
    var index = this.__find__(name);
    var opt = options || {};

    if (index === -1) { throw new Error('Parser rule not found: ' + name); }

    this.__rules__[index].fn = fn;
    this.__rules__[index].alt = opt.alt || [];
    this.__cache__ = null;
  };


  /**
   * Ruler.before(beforeName, ruleName, fn [, options])
   * - beforeName (String): new rule will be added before this one.
   * - ruleName (String): name of added rule.
   * - fn (Function): rule function.
   * - options (Object): rule options (not mandatory).
   *
   * Add new rule to chain before one with given name. See also
   * [[Ruler.after]], [[Ruler.push]].
   *
   * ##### Options:
   *
   * - __alt__ - array with names of "alternate" chains.
   *
   * ##### Example
   *
   * ```javascript
   * var md = require('markdown-it')();
   *
   * md.block.ruler.before('paragraph', 'my_rule', function replace(state) {
   *   //...
   * });
   * ```
   **/
  Ruler.prototype.before = function (beforeName, ruleName, fn, options) {
    var index = this.__find__(beforeName);
    var opt = options || {};

    if (index === -1) { throw new Error('Parser rule not found: ' + beforeName); }

    this.__rules__.splice(index, 0, {
      name: ruleName,
      enabled: true,
      fn: fn,
      alt: opt.alt || []
    });

    this.__cache__ = null;
  };


  /**
   * Ruler.after(afterName, ruleName, fn [, options])
   * - afterName (String): new rule will be added after this one.
   * - ruleName (String): name of added rule.
   * - fn (Function): rule function.
   * - options (Object): rule options (not mandatory).
   *
   * Add new rule to chain after one with given name. See also
   * [[Ruler.before]], [[Ruler.push]].
   *
   * ##### Options:
   *
   * - __alt__ - array with names of "alternate" chains.
   *
   * ##### Example
   *
   * ```javascript
   * var md = require('markdown-it')();
   *
   * md.inline.ruler.after('text', 'my_rule', function replace(state) {
   *   //...
   * });
   * ```
   **/
  Ruler.prototype.after = function (afterName, ruleName, fn, options) {
    var index = this.__find__(afterName);
    var opt = options || {};

    if (index === -1) { throw new Error('Parser rule not found: ' + afterName); }

    this.__rules__.splice(index + 1, 0, {
      name: ruleName,
      enabled: true,
      fn: fn,
      alt: opt.alt || []
    });

    this.__cache__ = null;
  };

  /**
   * Ruler.push(ruleName, fn [, options])
   * - ruleName (String): name of added rule.
   * - fn (Function): rule function.
   * - options (Object): rule options (not mandatory).
   *
   * Push new rule to the end of chain. See also
   * [[Ruler.before]], [[Ruler.after]].
   *
   * ##### Options:
   *
   * - __alt__ - array with names of "alternate" chains.
   *
   * ##### Example
   *
   * ```javascript
   * var md = require('markdown-it')();
   *
   * md.core.ruler.push('my_rule', function replace(state) {
   *   //...
   * });
   * ```
   **/
  Ruler.prototype.push = function (ruleName, fn, options) {
    var opt = options || {};

    this.__rules__.push({
      name: ruleName,
      enabled: true,
      fn: fn,
      alt: opt.alt || []
    });

    this.__cache__ = null;
  };


  /**
   * Ruler.enable(list [, ignoreInvalid]) -> Array
   * - list (String|Array): list of rule names to enable.
   * - ignoreInvalid (Boolean): set `true` to ignore errors when rule not found.
   *
   * Enable rules with given names. If any rule name not found - throw Error.
   * Errors can be disabled by second param.
   *
   * Returns list of found rule names (if no exception happened).
   *
   * See also [[Ruler.disable]], [[Ruler.enableOnly]].
   **/
  Ruler.prototype.enable = function (list, ignoreInvalid) {
    if (!Array.isArray(list)) { list = [ list ]; }

    var result = [];

    // Search by name and enable
    list.forEach(function (name) {
      var idx = this.__find__(name);

      if (idx < 0) {
        if (ignoreInvalid) { return; }
        throw new Error('Rules manager: invalid rule name ' + name);
      }
      this.__rules__[idx].enabled = true;
      result.push(name);
    }, this);

    this.__cache__ = null;
    return result;
  };


  /**
   * Ruler.enableOnly(list [, ignoreInvalid])
   * - list (String|Array): list of rule names to enable (whitelist).
   * - ignoreInvalid (Boolean): set `true` to ignore errors when rule not found.
   *
   * Enable rules with given names, and disable everything else. If any rule name
   * not found - throw Error. Errors can be disabled by second param.
   *
   * See also [[Ruler.disable]], [[Ruler.enable]].
   **/
  Ruler.prototype.enableOnly = function (list, ignoreInvalid) {
    if (!Array.isArray(list)) { list = [ list ]; }

    this.__rules__.forEach(function (rule) { rule.enabled = false; });

    this.enable(list, ignoreInvalid);
  };


  /**
   * Ruler.disable(list [, ignoreInvalid]) -> Array
   * - list (String|Array): list of rule names to disable.
   * - ignoreInvalid (Boolean): set `true` to ignore errors when rule not found.
   *
   * Disable rules with given names. If any rule name not found - throw Error.
   * Errors can be disabled by second param.
   *
   * Returns list of found rule names (if no exception happened).
   *
   * See also [[Ruler.enable]], [[Ruler.enableOnly]].
   **/
  Ruler.prototype.disable = function (list, ignoreInvalid) {
    if (!Array.isArray(list)) { list = [ list ]; }

    var result = [];

    // Search by name and disable
    list.forEach(function (name) {
      var idx = this.__find__(name);

      if (idx < 0) {
        if (ignoreInvalid) { return; }
        throw new Error('Rules manager: invalid rule name ' + name);
      }
      this.__rules__[idx].enabled = false;
      result.push(name);
    }, this);

    this.__cache__ = null;
    return result;
  };


  /**
   * Ruler.getRules(chainName) -> Array
   *
   * Return array of active functions (rules) for given chain name. It analyzes
   * rules configuration, compiles caches if not exists and returns result.
   *
   * Default chain name is `''` (empty string). It can't be skipped. That's
   * done intentionally, to keep signature monomorphic for high speed.
   **/
  Ruler.prototype.getRules = function (chainName) {
    if (this.__cache__ === null) {
      this.__compile__();
    }

    // Chain can be empty, if rules disabled. But we still have to return Array.
    return this.__cache__[chainName] || [];
  };

  var ruler = Ruler;

  // Normalize input string


  // https://spec.commonmark.org/0.29/#line-ending
  var NEWLINES_RE  = /\r\n?|\n/g;
  var NULL_RE      = /\0/g;


  var normalize = function normalize(state) {
    var str;

    // Normalize newlines
    str = state.src.replace(NEWLINES_RE, '\n');

    // Replace NULL characters
    str = str.replace(NULL_RE, '\uFFFD');

    state.src = str;
  };

  var block = function block(state) {
    var token;

    if (state.inlineMode) {
      token          = new state.Token('inline', '', 0);
      token.content  = state.src;
      token.map      = [ 0, 1 ];
      token.children = [];
      state.tokens.push(token);
    } else {
      state.md.block.parse(state.src, state.md, state.env, state.tokens);
    }
  };

  var inline = function inline(state) {
    var tokens = state.tokens, tok, i, l;

    // Parse inlines
    for (i = 0, l = tokens.length; i < l; i++) {
      tok = tokens[i];
      if (tok.type === 'inline') {
        state.md.inline.parse(tok.content, state.md, state.env, tok.children);
      }
    }
  };

  var arrayReplaceAt = utils.arrayReplaceAt;


  function isLinkOpen(str) {
    return /^<a[>\s]/i.test(str);
  }
  function isLinkClose(str) {
    return /^<\/a\s*>/i.test(str);
  }


  var linkify$1 = function linkify(state) {
    var i, j, l, tokens, token, currentToken, nodes, ln, text, pos, lastPos,
        level, htmlLinkLevel, url, fullUrl, urlText,
        blockTokens = state.tokens,
        links;

    if (!state.md.options.linkify) { return; }

    for (j = 0, l = blockTokens.length; j < l; j++) {
      if (blockTokens[j].type !== 'inline' ||
          !state.md.linkify.pretest(blockTokens[j].content)) {
        continue;
      }

      tokens = blockTokens[j].children;

      htmlLinkLevel = 0;

      // We scan from the end, to keep position when new tags added.
      // Use reversed logic in links start/end match
      for (i = tokens.length - 1; i >= 0; i--) {
        currentToken = tokens[i];

        // Skip content of markdown links
        if (currentToken.type === 'link_close') {
          i--;
          while (tokens[i].level !== currentToken.level && tokens[i].type !== 'link_open') {
            i--;
          }
          continue;
        }

        // Skip content of html tag links
        if (currentToken.type === 'html_inline') {
          if (isLinkOpen(currentToken.content) && htmlLinkLevel > 0) {
            htmlLinkLevel--;
          }
          if (isLinkClose(currentToken.content)) {
            htmlLinkLevel++;
          }
        }
        if (htmlLinkLevel > 0) { continue; }

        if (currentToken.type === 'text' && state.md.linkify.test(currentToken.content)) {

          text = currentToken.content;
          links = state.md.linkify.match(text);

          // Now split string to nodes
          nodes = [];
          level = currentToken.level;
          lastPos = 0;

          for (ln = 0; ln < links.length; ln++) {

            url = links[ln].url;
            fullUrl = state.md.normalizeLink(url);
            if (!state.md.validateLink(fullUrl)) { continue; }

            urlText = links[ln].text;

            // Linkifier might send raw hostnames like "example.com", where url
            // starts with domain name. So we prepend http:// in those cases,
            // and remove it afterwards.
            //
            if (!links[ln].schema) {
              urlText = state.md.normalizeLinkText('http://' + urlText).replace(/^http:\/\//, '');
            } else if (links[ln].schema === 'mailto:' && !/^mailto:/i.test(urlText)) {
              urlText = state.md.normalizeLinkText('mailto:' + urlText).replace(/^mailto:/, '');
            } else {
              urlText = state.md.normalizeLinkText(urlText);
            }

            pos = links[ln].index;

            if (pos > lastPos) {
              token         = new state.Token('text', '', 0);
              token.content = text.slice(lastPos, pos);
              token.level   = level;
              nodes.push(token);
            }

            token         = new state.Token('link_open', 'a', 1);
            token.attrs   = [ [ 'href', fullUrl ] ];
            token.level   = level++;
            token.markup  = 'linkify';
            token.info    = 'auto';
            nodes.push(token);

            token         = new state.Token('text', '', 0);
            token.content = urlText;
            token.level   = level;
            nodes.push(token);

            token         = new state.Token('link_close', 'a', -1);
            token.level   = --level;
            token.markup  = 'linkify';
            token.info    = 'auto';
            nodes.push(token);

            lastPos = links[ln].lastIndex;
          }
          if (lastPos < text.length) {
            token         = new state.Token('text', '', 0);
            token.content = text.slice(lastPos);
            token.level   = level;
            nodes.push(token);
          }

          // replace current node
          blockTokens[j].children = tokens = arrayReplaceAt(tokens, i, nodes);
        }
      }
    }
  };

  // Simple typographic replacements

  // TODO:
  // - fractionals 1/2, 1/4, 3/4 -> ½, ¼, ¾
  // - miltiplication 2 x 4 -> 2 × 4

  var RARE_RE = /\+-|\.\.|\?\?\?\?|!!!!|,,|--/;

  // Workaround for phantomjs - need regex without /g flag,
  // or root check will fail every second time
  var SCOPED_ABBR_TEST_RE = /\((c|tm|r|p)\)/i;

  var SCOPED_ABBR_RE = /\((c|tm|r|p)\)/ig;
  var SCOPED_ABBR = {
    c: '©',
    r: '®',
    p: '§',
    tm: '™'
  };

  function replaceFn(match, name) {
    return SCOPED_ABBR[name.toLowerCase()];
  }

  function replace_scoped(inlineTokens) {
    var i, token, inside_autolink = 0;

    for (i = inlineTokens.length - 1; i >= 0; i--) {
      token = inlineTokens[i];

      if (token.type === 'text' && !inside_autolink) {
        token.content = token.content.replace(SCOPED_ABBR_RE, replaceFn);
      }

      if (token.type === 'link_open' && token.info === 'auto') {
        inside_autolink--;
      }

      if (token.type === 'link_close' && token.info === 'auto') {
        inside_autolink++;
      }
    }
  }

  function replace_rare(inlineTokens) {
    var i, token, inside_autolink = 0;

    for (i = inlineTokens.length - 1; i >= 0; i--) {
      token = inlineTokens[i];

      if (token.type === 'text' && !inside_autolink) {
        if (RARE_RE.test(token.content)) {
          token.content = token.content
            .replace(/\+-/g, '±')
            // .., ..., ....... -> …
            // but ?..... & !..... -> ?.. & !..
            .replace(/\.{2,}/g, '…').replace(/([?!])…/g, '$1..')
            .replace(/([?!]){4,}/g, '$1$1$1').replace(/,{2,}/g, ',')
            // em-dash
            .replace(/(^|[^-])---(?=[^-]|$)/mg, '$1\u2014')
            // en-dash
            .replace(/(^|\s)--(?=\s|$)/mg, '$1\u2013')
            .replace(/(^|[^-\s])--(?=[^-\s]|$)/mg, '$1\u2013');
        }
      }

      if (token.type === 'link_open' && token.info === 'auto') {
        inside_autolink--;
      }

      if (token.type === 'link_close' && token.info === 'auto') {
        inside_autolink++;
      }
    }
  }


  var replacements = function replace(state) {
    var blkIdx;

    if (!state.md.options.typographer) { return; }

    for (blkIdx = state.tokens.length - 1; blkIdx >= 0; blkIdx--) {

      if (state.tokens[blkIdx].type !== 'inline') { continue; }

      if (SCOPED_ABBR_TEST_RE.test(state.tokens[blkIdx].content)) {
        replace_scoped(state.tokens[blkIdx].children);
      }

      if (RARE_RE.test(state.tokens[blkIdx].content)) {
        replace_rare(state.tokens[blkIdx].children);
      }

    }
  };

  var isWhiteSpace$1   = utils.isWhiteSpace;
  var isPunctChar$1    = utils.isPunctChar;
  var isMdAsciiPunct$1 = utils.isMdAsciiPunct;

  var QUOTE_TEST_RE = /['"]/;
  var QUOTE_RE = /['"]/g;
  var APOSTROPHE = '\u2019'; /* ’ */


  function replaceAt(str, index, ch) {
    return str.substr(0, index) + ch + str.substr(index + 1);
  }

  function process_inlines(tokens, state) {
    var i, token, text, t, pos, max, thisLevel, item, lastChar, nextChar,
        isLastPunctChar, isNextPunctChar, isLastWhiteSpace, isNextWhiteSpace,
        canOpen, canClose, j, isSingle, stack, openQuote, closeQuote;

    stack = [];

    for (i = 0; i < tokens.length; i++) {
      token = tokens[i];

      thisLevel = tokens[i].level;

      for (j = stack.length - 1; j >= 0; j--) {
        if (stack[j].level <= thisLevel) { break; }
      }
      stack.length = j + 1;

      if (token.type !== 'text') { continue; }

      text = token.content;
      pos = 0;
      max = text.length;

      /*eslint no-labels:0,block-scoped-var:0*/
      OUTER:
      while (pos < max) {
        QUOTE_RE.lastIndex = pos;
        t = QUOTE_RE.exec(text);
        if (!t) { break; }

        canOpen = canClose = true;
        pos = t.index + 1;
        isSingle = (t[0] === "'");

        // Find previous character,
        // default to space if it's the beginning of the line
        //
        lastChar = 0x20;

        if (t.index - 1 >= 0) {
          lastChar = text.charCodeAt(t.index - 1);
        } else {
          for (j = i - 1; j >= 0; j--) {
            if (tokens[j].type === 'softbreak' || tokens[j].type === 'hardbreak') { break; } // lastChar defaults to 0x20
            if (!tokens[j].content) { continue; } // should skip all tokens except 'text', 'html_inline' or 'code_inline'

            lastChar = tokens[j].content.charCodeAt(tokens[j].content.length - 1);
            break;
          }
        }

        // Find next character,
        // default to space if it's the end of the line
        //
        nextChar = 0x20;

        if (pos < max) {
          nextChar = text.charCodeAt(pos);
        } else {
          for (j = i + 1; j < tokens.length; j++) {
            if (tokens[j].type === 'softbreak' || tokens[j].type === 'hardbreak') { break; } // nextChar defaults to 0x20
            if (!tokens[j].content) { continue; } // should skip all tokens except 'text', 'html_inline' or 'code_inline'

            nextChar = tokens[j].content.charCodeAt(0);
            break;
          }
        }

        isLastPunctChar = isMdAsciiPunct$1(lastChar) || isPunctChar$1(String.fromCharCode(lastChar));
        isNextPunctChar = isMdAsciiPunct$1(nextChar) || isPunctChar$1(String.fromCharCode(nextChar));

        isLastWhiteSpace = isWhiteSpace$1(lastChar);
        isNextWhiteSpace = isWhiteSpace$1(nextChar);

        if (isNextWhiteSpace) {
          canOpen = false;
        } else if (isNextPunctChar) {
          if (!(isLastWhiteSpace || isLastPunctChar)) {
            canOpen = false;
          }
        }

        if (isLastWhiteSpace) {
          canClose = false;
        } else if (isLastPunctChar) {
          if (!(isNextWhiteSpace || isNextPunctChar)) {
            canClose = false;
          }
        }

        if (nextChar === 0x22 /* " */ && t[0] === '"') {
          if (lastChar >= 0x30 /* 0 */ && lastChar <= 0x39 /* 9 */) {
            // special case: 1"" - count first quote as an inch
            canClose = canOpen = false;
          }
        }

        if (canOpen && canClose) {
          // Replace quotes in the middle of punctuation sequence, but not
          // in the middle of the words, i.e.:
          //
          // 1. foo " bar " baz - not replaced
          // 2. foo-"-bar-"-baz - replaced
          // 3. foo"bar"baz     - not replaced
          //
          canOpen = isLastPunctChar;
          canClose = isNextPunctChar;
        }

        if (!canOpen && !canClose) {
          // middle of word
          if (isSingle) {
            token.content = replaceAt(token.content, t.index, APOSTROPHE);
          }
          continue;
        }

        if (canClose) {
          // this could be a closing quote, rewind the stack to get a match
          for (j = stack.length - 1; j >= 0; j--) {
            item = stack[j];
            if (stack[j].level < thisLevel) { break; }
            if (item.single === isSingle && stack[j].level === thisLevel) {
              item = stack[j];

              if (isSingle) {
                openQuote = state.md.options.quotes[2];
                closeQuote = state.md.options.quotes[3];
              } else {
                openQuote = state.md.options.quotes[0];
                closeQuote = state.md.options.quotes[1];
              }

              // replace token.content *before* tokens[item.token].content,
              // because, if they are pointing at the same token, replaceAt
              // could mess up indices when quote length != 1
              token.content = replaceAt(token.content, t.index, closeQuote);
              tokens[item.token].content = replaceAt(
                tokens[item.token].content, item.pos, openQuote);

              pos += closeQuote.length - 1;
              if (item.token === i) { pos += openQuote.length - 1; }

              text = token.content;
              max = text.length;

              stack.length = j;
              continue OUTER;
            }
          }
        }

        if (canOpen) {
          stack.push({
            token: i,
            pos: t.index,
            single: isSingle,
            level: thisLevel
          });
        } else if (canClose && isSingle) {
          token.content = replaceAt(token.content, t.index, APOSTROPHE);
        }
      }
    }
  }


  var smartquotes = function smartquotes(state) {
    /*eslint max-depth:0*/
    var blkIdx;

    if (!state.md.options.typographer) { return; }

    for (blkIdx = state.tokens.length - 1; blkIdx >= 0; blkIdx--) {

      if (state.tokens[blkIdx].type !== 'inline' ||
          !QUOTE_TEST_RE.test(state.tokens[blkIdx].content)) {
        continue;
      }

      process_inlines(state.tokens[blkIdx].children, state);
    }
  };

  // Token class


  /**
   * class Token
   **/

  /**
   * new Token(type, tag, nesting)
   *
   * Create new token and fill passed properties.
   **/
  function Token(type, tag, nesting) {
    /**
     * Token#type -> String
     *
     * Type of the token (string, e.g. "paragraph_open")
     **/
    this.type     = type;

    /**
     * Token#tag -> String
     *
     * html tag name, e.g. "p"
     **/
    this.tag      = tag;

    /**
     * Token#attrs -> Array
     *
     * Html attributes. Format: `[ [ name1, value1 ], [ name2, value2 ] ]`
     **/
    this.attrs    = null;

    /**
     * Token#map -> Array
     *
     * Source map info. Format: `[ line_begin, line_end ]`
     **/
    this.map      = null;

    /**
     * Token#nesting -> Number
     *
     * Level change (number in {-1, 0, 1} set), where:
     *
     * -  `1` means the tag is opening
     * -  `0` means the tag is self-closing
     * - `-1` means the tag is closing
     **/
    this.nesting  = nesting;

    /**
     * Token#level -> Number
     *
     * nesting level, the same as `state.level`
     **/
    this.level    = 0;

    /**
     * Token#children -> Array
     *
     * An array of child nodes (inline and img tokens)
     **/
    this.children = null;

    /**
     * Token#content -> String
     *
     * In a case of self-closing tag (code, html, fence, etc.),
     * it has contents of this tag.
     **/
    this.content  = '';

    /**
     * Token#markup -> String
     *
     * '*' or '_' for emphasis, fence string for fence, etc.
     **/
    this.markup   = '';

    /**
     * Token#info -> String
     *
     * Additional information:
     *
     * - Info string for "fence" tokens
     * - The value "auto" for autolink "link_open" and "link_close" tokens
     * - The string value of the item marker for ordered-list "list_item_open" tokens
     **/
    this.info     = '';

    /**
     * Token#meta -> Object
     *
     * A place for plugins to store an arbitrary data
     **/
    this.meta     = null;

    /**
     * Token#block -> Boolean
     *
     * True for block-level tokens, false for inline tokens.
     * Used in renderer to calculate line breaks
     **/
    this.block    = false;

    /**
     * Token#hidden -> Boolean
     *
     * If it's true, ignore this element when rendering. Used for tight lists
     * to hide paragraphs.
     **/
    this.hidden   = false;
  }


  /**
   * Token.attrIndex(name) -> Number
   *
   * Search attribute index by name.
   **/
  Token.prototype.attrIndex = function attrIndex(name) {
    var attrs, i, len;

    if (!this.attrs) { return -1; }

    attrs = this.attrs;

    for (i = 0, len = attrs.length; i < len; i++) {
      if (attrs[i][0] === name) { return i; }
    }
    return -1;
  };


  /**
   * Token.attrPush(attrData)
   *
   * Add `[ name, value ]` attribute to list. Init attrs if necessary
   **/
  Token.prototype.attrPush = function attrPush(attrData) {
    if (this.attrs) {
      this.attrs.push(attrData);
    } else {
      this.attrs = [ attrData ];
    }
  };


  /**
   * Token.attrSet(name, value)
   *
   * Set `name` attribute to `value`. Override old value if exists.
   **/
  Token.prototype.attrSet = function attrSet(name, value) {
    var idx = this.attrIndex(name),
        attrData = [ name, value ];

    if (idx < 0) {
      this.attrPush(attrData);
    } else {
      this.attrs[idx] = attrData;
    }
  };


  /**
   * Token.attrGet(name)
   *
   * Get the value of attribute `name`, or null if it does not exist.
   **/
  Token.prototype.attrGet = function attrGet(name) {
    var idx = this.attrIndex(name), value = null;
    if (idx >= 0) {
      value = this.attrs[idx][1];
    }
    return value;
  };


  /**
   * Token.attrJoin(name, value)
   *
   * Join value to existing attribute via space. Or create new attribute if not
   * exists. Useful to operate with token classes.
   **/
  Token.prototype.attrJoin = function attrJoin(name, value) {
    var idx = this.attrIndex(name);

    if (idx < 0) {
      this.attrPush([ name, value ]);
    } else {
      this.attrs[idx][1] = this.attrs[idx][1] + ' ' + value;
    }
  };


  var token = Token;

  function StateCore(src, md, env) {
    this.src = src;
    this.env = env;
    this.tokens = [];
    this.inlineMode = false;
    this.md = md; // link to parser instance
  }

  // re-export Token class to use in core rules
  StateCore.prototype.Token = token;


  var state_core = StateCore;

  var _rules$2 = [
    [ 'normalize',      normalize      ],
    [ 'block',          block          ],
    [ 'inline',         inline         ],
    [ 'linkify',        linkify$1        ],
    [ 'replacements',   replacements   ],
    [ 'smartquotes',    smartquotes    ]
  ];


  /**
   * new Core()
   **/
  function Core() {
    /**
     * Core#ruler -> Ruler
     *
     * [[Ruler]] instance. Keep configuration of core rules.
     **/
    this.ruler = new ruler();

    for (var i = 0; i < _rules$2.length; i++) {
      this.ruler.push(_rules$2[i][0], _rules$2[i][1]);
    }
  }


  /**
   * Core.process(state)
   *
   * Executes core chain rules.
   **/
  Core.prototype.process = function (state) {
    var i, l, rules;

    rules = this.ruler.getRules('');

    for (i = 0, l = rules.length; i < l; i++) {
      rules[i](state);
    }
  };

  Core.prototype.State = state_core;


  var parser_core = Core;

  var isSpace$d = utils.isSpace;


  function getLine$1(state, line) {
    var pos = state.bMarks[line] + state.tShift[line],
        max = state.eMarks[line];

    return state.src.substr(pos, max - pos);
  }

  function escapedSplit$1(str) {
    var result = [],
        pos = 0,
        max = str.length,
        ch,
        isEscaped = false,
        lastPos = 0,
        current = '';

    ch  = str.charCodeAt(pos);

    while (pos < max) {
      if (ch === 0x7c/* | */) {
        if (!isEscaped) {
          // pipe separating cells, '|'
          result.push(current + str.substring(lastPos, pos));
          current = '';
          lastPos = pos + 1;
        } else {
          // escaped pipe, '\|'
          current += str.substring(lastPos, pos - 1);
          lastPos = pos;
        }
      }

      isEscaped = (ch === 0x5c/* \ */);
      pos++;

      ch = str.charCodeAt(pos);
    }

    result.push(current + str.substring(lastPos));

    return result;
  }


  var table$1 = function table(state, startLine, endLine, silent) {
    var ch, lineText, pos, i, l, nextLine, columns, columnCount, token,
        aligns, t, tableLines, tbodyLines, oldParentType, terminate,
        terminatorRules, firstCh, secondCh;

    // should have at least two lines
    if (startLine + 2 > endLine) { return false; }

    nextLine = startLine + 1;

    if (state.sCount[nextLine] < state.blkIndent) { return false; }

    // if it's indented more than 3 spaces, it should be a code block
    if (state.sCount[nextLine] - state.blkIndent >= 4) { return false; }

    // first character of the second line should be '|', '-', ':',
    // and no other characters are allowed but spaces;
    // basically, this is the equivalent of /^[-:|][-:|\s]*$/ regexp

    pos = state.bMarks[nextLine] + state.tShift[nextLine];
    if (pos >= state.eMarks[nextLine]) { return false; }

    firstCh = state.src.charCodeAt(pos++);
    if (firstCh !== 0x7C/* | */ && firstCh !== 0x2D/* - */ && firstCh !== 0x3A/* : */) { return false; }

    if (pos >= state.eMarks[nextLine]) { return false; }

    secondCh = state.src.charCodeAt(pos++);
    if (secondCh !== 0x7C/* | */ && secondCh !== 0x2D/* - */ && secondCh !== 0x3A/* : */ && !isSpace$d(secondCh)) {
      return false;
    }

    // if first character is '-', then second character must not be a space
    // (due to parsing ambiguity with list)
    if (firstCh === 0x2D/* - */ && isSpace$d(secondCh)) { return false; }

    while (pos < state.eMarks[nextLine]) {
      ch = state.src.charCodeAt(pos);

      if (ch !== 0x7C/* | */ && ch !== 0x2D/* - */ && ch !== 0x3A/* : */ && !isSpace$d(ch)) { return false; }

      pos++;
    }

    lineText = getLine$1(state, startLine + 1);

    columns = lineText.split('|');
    aligns = [];
    for (i = 0; i < columns.length; i++) {
      t = columns[i].trim();
      if (!t) {
        // allow empty columns before and after table, but not in between columns;
        // e.g. allow ` |---| `, disallow ` ---||--- `
        if (i === 0 || i === columns.length - 1) {
          continue;
        } else {
          return false;
        }
      }

      if (!/^:?-+:?$/.test(t)) { return false; }
      if (t.charCodeAt(t.length - 1) === 0x3A/* : */) {
        aligns.push(t.charCodeAt(0) === 0x3A/* : */ ? 'center' : 'right');
      } else if (t.charCodeAt(0) === 0x3A/* : */) {
        aligns.push('left');
      } else {
        aligns.push('');
      }
    }

    lineText = getLine$1(state, startLine).trim();
    if (lineText.indexOf('|') === -1) { return false; }
    if (state.sCount[startLine] - state.blkIndent >= 4) { return false; }
    columns = escapedSplit$1(lineText);
    if (columns.length && columns[0] === '') { columns.shift(); }
    if (columns.length && columns[columns.length - 1] === '') { columns.pop(); }

    // header row will define an amount of columns in the entire table,
    // and align row should be exactly the same (the rest of the rows can differ)
    columnCount = columns.length;
    if (columnCount === 0 || columnCount !== aligns.length) { return false; }

    if (silent) { return true; }

    oldParentType = state.parentType;
    state.parentType = 'table';

    // use 'blockquote' lists for termination because it's
    // the most similar to tables
    terminatorRules = state.md.block.ruler.getRules('blockquote');

    token     = state.push('table_open', 'table', 1);
    token.map = tableLines = [ startLine, 0 ];

    token     = state.push('thead_open', 'thead', 1);
    token.map = [ startLine, startLine + 1 ];

    token     = state.push('tr_open', 'tr', 1);
    token.map = [ startLine, startLine + 1 ];

    for (i = 0; i < columns.length; i++) {
      token          = state.push('th_open', 'th', 1);
      if (aligns[i]) {
        token.attrs  = [ [ 'style', 'text-align:' + aligns[i] ] ];
      }

      token          = state.push('inline', '', 0);
      token.content  = columns[i].trim();
      token.children = [];

      token          = state.push('th_close', 'th', -1);
    }

    token     = state.push('tr_close', 'tr', -1);
    token     = state.push('thead_close', 'thead', -1);

    for (nextLine = startLine + 2; nextLine < endLine; nextLine++) {
      if (state.sCount[nextLine] < state.blkIndent) { break; }

      terminate = false;
      for (i = 0, l = terminatorRules.length; i < l; i++) {
        if (terminatorRules[i](state, nextLine, endLine, true)) {
          terminate = true;
          break;
        }
      }

      if (terminate) { break; }
      lineText = getLine$1(state, nextLine).trim();
      if (!lineText) { break; }
      if (state.sCount[nextLine] - state.blkIndent >= 4) { break; }
      columns = escapedSplit$1(lineText);
      if (columns.length && columns[0] === '') { columns.shift(); }
      if (columns.length && columns[columns.length - 1] === '') { columns.pop(); }

      if (nextLine === startLine + 2) {
        token     = state.push('tbody_open', 'tbody', 1);
        token.map = tbodyLines = [ startLine + 2, 0 ];
      }

      token     = state.push('tr_open', 'tr', 1);
      token.map = [ nextLine, nextLine + 1 ];

      for (i = 0; i < columnCount; i++) {
        token          = state.push('td_open', 'td', 1);
        if (aligns[i]) {
          token.attrs  = [ [ 'style', 'text-align:' + aligns[i] ] ];
        }

        token          = state.push('inline', '', 0);
        token.content  = columns[i] ? columns[i].trim() : '';
        token.children = [];

        token          = state.push('td_close', 'td', -1);
      }
      token = state.push('tr_close', 'tr', -1);
    }

    if (tbodyLines) {
      token = state.push('tbody_close', 'tbody', -1);
      tbodyLines[1] = nextLine;
    }

    token = state.push('table_close', 'table', -1);
    tableLines[1] = nextLine;

    state.parentType = oldParentType;
    state.line = nextLine;
    return true;
  };

  // Code block (4 spaces padded)


  var code = function code(state, startLine, endLine/*, silent*/) {
    var nextLine, last, token;

    if (state.sCount[startLine] - state.blkIndent < 4) { return false; }

    last = nextLine = startLine + 1;

    while (nextLine < endLine) {
      if (state.isEmpty(nextLine)) {
        nextLine++;
        continue;
      }

      if (state.sCount[nextLine] - state.blkIndent >= 4) {
        nextLine++;
        last = nextLine;
        continue;
      }
      break;
    }

    state.line = last;

    token         = state.push('code_block', 'code', 0);
    token.content = state.getLines(startLine, last, 4 + state.blkIndent, false) + '\n';
    token.map     = [ startLine, state.line ];

    return true;
  };

  // fences (``` lang, ~~~ lang)


  var fence = function fence(state, startLine, endLine, silent) {
    var marker, len, params, nextLine, mem, token, markup,
        haveEndMarker = false,
        pos = state.bMarks[startLine] + state.tShift[startLine],
        max = state.eMarks[startLine];

    // if it's indented more than 3 spaces, it should be a code block
    if (state.sCount[startLine] - state.blkIndent >= 4) { return false; }

    if (pos + 3 > max) { return false; }

    marker = state.src.charCodeAt(pos);

    if (marker !== 0x7E/* ~ */ && marker !== 0x60 /* ` */) {
      return false;
    }

    // scan marker length
    mem = pos;
    pos = state.skipChars(pos, marker);

    len = pos - mem;

    if (len < 3) { return false; }

    markup = state.src.slice(mem, pos);
    params = state.src.slice(pos, max);

    if (marker === 0x60 /* ` */) {
      if (params.indexOf(String.fromCharCode(marker)) >= 0) {
        return false;
      }
    }

    // Since start is found, we can report success here in validation mode
    if (silent) { return true; }

    // search end of block
    nextLine = startLine;

    for (;;) {
      nextLine++;
      if (nextLine >= endLine) {
        // unclosed block should be autoclosed by end of document.
        // also block seems to be autoclosed by end of parent
        break;
      }

      pos = mem = state.bMarks[nextLine] + state.tShift[nextLine];
      max = state.eMarks[nextLine];

      if (pos < max && state.sCount[nextLine] < state.blkIndent) {
        // non-empty line with negative indent should stop the list:
        // - ```
        //  test
        break;
      }

      if (state.src.charCodeAt(pos) !== marker) { continue; }

      if (state.sCount[nextLine] - state.blkIndent >= 4) {
        // closing fence should be indented less than 4 spaces
        continue;
      }

      pos = state.skipChars(pos, marker);

      // closing code fence must be at least as long as the opening one
      if (pos - mem < len) { continue; }

      // make sure tail has spaces only
      pos = state.skipSpaces(pos);

      if (pos < max) { continue; }

      haveEndMarker = true;
      // found!
      break;
    }

    // If a fence has heading spaces, they should be removed from its inner block
    len = state.sCount[startLine];

    state.line = nextLine + (haveEndMarker ? 1 : 0);

    token         = state.push('fence', 'code', 0);
    token.info    = params;
    token.content = state.getLines(startLine + 1, nextLine, len, true);
    token.markup  = markup;
    token.map     = [ startLine, state.line ];

    return true;
  };

  var isSpace$c = utils.isSpace;


  var blockquote = function blockquote(state, startLine, endLine, silent) {
    var adjustTab,
        ch,
        i,
        initial,
        l,
        lastLineEmpty,
        lines,
        nextLine,
        offset,
        oldBMarks,
        oldBSCount,
        oldIndent,
        oldParentType,
        oldSCount,
        oldTShift,
        spaceAfterMarker,
        terminate,
        terminatorRules,
        token,
        isOutdented,
        oldLineMax = state.lineMax,
        pos = state.bMarks[startLine] + state.tShift[startLine],
        max = state.eMarks[startLine];

    // if it's indented more than 3 spaces, it should be a code block
    if (state.sCount[startLine] - state.blkIndent >= 4) { return false; }

    // check the block quote marker
    if (state.src.charCodeAt(pos++) !== 0x3E/* > */) { return false; }

    // we know that it's going to be a valid blockquote,
    // so no point trying to find the end of it in silent mode
    if (silent) { return true; }

    // set offset past spaces and ">"
    initial = offset = state.sCount[startLine] + 1;

    // skip one optional space after '>'
    if (state.src.charCodeAt(pos) === 0x20 /* space */) {
      // ' >   test '
      //     ^ -- position start of line here:
      pos++;
      initial++;
      offset++;
      adjustTab = false;
      spaceAfterMarker = true;
    } else if (state.src.charCodeAt(pos) === 0x09 /* tab */) {
      spaceAfterMarker = true;

      if ((state.bsCount[startLine] + offset) % 4 === 3) {
        // '  >\t  test '
        //       ^ -- position start of line here (tab has width===1)
        pos++;
        initial++;
        offset++;
        adjustTab = false;
      } else {
        // ' >\t  test '
        //    ^ -- position start of line here + shift bsCount slightly
        //         to make extra space appear
        adjustTab = true;
      }
    } else {
      spaceAfterMarker = false;
    }

    oldBMarks = [ state.bMarks[startLine] ];
    state.bMarks[startLine] = pos;

    while (pos < max) {
      ch = state.src.charCodeAt(pos);

      if (isSpace$c(ch)) {
        if (ch === 0x09) {
          offset += 4 - (offset + state.bsCount[startLine] + (adjustTab ? 1 : 0)) % 4;
        } else {
          offset++;
        }
      } else {
        break;
      }

      pos++;
    }

    oldBSCount = [ state.bsCount[startLine] ];
    state.bsCount[startLine] = state.sCount[startLine] + 1 + (spaceAfterMarker ? 1 : 0);

    lastLineEmpty = pos >= max;

    oldSCount = [ state.sCount[startLine] ];
    state.sCount[startLine] = offset - initial;

    oldTShift = [ state.tShift[startLine] ];
    state.tShift[startLine] = pos - state.bMarks[startLine];

    terminatorRules = state.md.block.ruler.getRules('blockquote');

    oldParentType = state.parentType;
    state.parentType = 'blockquote';

    // Search the end of the block
    //
    // Block ends with either:
    //  1. an empty line outside:
    //     ```
    //     > test
    //
    //     ```
    //  2. an empty line inside:
    //     ```
    //     >
    //     test
    //     ```
    //  3. another tag:
    //     ```
    //     > test
    //      - - -
    //     ```
    for (nextLine = startLine + 1; nextLine < endLine; nextLine++) {
      // check if it's outdented, i.e. it's inside list item and indented
      // less than said list item:
      //
      // ```
      // 1. anything
      //    > current blockquote
      // 2. checking this line
      // ```
      isOutdented = state.sCount[nextLine] < state.blkIndent;

      pos = state.bMarks[nextLine] + state.tShift[nextLine];
      max = state.eMarks[nextLine];

      if (pos >= max) {
        // Case 1: line is not inside the blockquote, and this line is empty.
        break;
      }

      if (state.src.charCodeAt(pos++) === 0x3E/* > */ && !isOutdented) {
        // This line is inside the blockquote.

        // set offset past spaces and ">"
        initial = offset = state.sCount[nextLine] + 1;

        // skip one optional space after '>'
        if (state.src.charCodeAt(pos) === 0x20 /* space */) {
          // ' >   test '
          //     ^ -- position start of line here:
          pos++;
          initial++;
          offset++;
          adjustTab = false;
          spaceAfterMarker = true;
        } else if (state.src.charCodeAt(pos) === 0x09 /* tab */) {
          spaceAfterMarker = true;

          if ((state.bsCount[nextLine] + offset) % 4 === 3) {
            // '  >\t  test '
            //       ^ -- position start of line here (tab has width===1)
            pos++;
            initial++;
            offset++;
            adjustTab = false;
          } else {
            // ' >\t  test '
            //    ^ -- position start of line here + shift bsCount slightly
            //         to make extra space appear
            adjustTab = true;
          }
        } else {
          spaceAfterMarker = false;
        }

        oldBMarks.push(state.bMarks[nextLine]);
        state.bMarks[nextLine] = pos;

        while (pos < max) {
          ch = state.src.charCodeAt(pos);

          if (isSpace$c(ch)) {
            if (ch === 0x09) {
              offset += 4 - (offset + state.bsCount[nextLine] + (adjustTab ? 1 : 0)) % 4;
            } else {
              offset++;
            }
          } else {
            break;
          }

          pos++;
        }

        lastLineEmpty = pos >= max;

        oldBSCount.push(state.bsCount[nextLine]);
        state.bsCount[nextLine] = state.sCount[nextLine] + 1 + (spaceAfterMarker ? 1 : 0);

        oldSCount.push(state.sCount[nextLine]);
        state.sCount[nextLine] = offset - initial;

        oldTShift.push(state.tShift[nextLine]);
        state.tShift[nextLine] = pos - state.bMarks[nextLine];
        continue;
      }

      // Case 2: line is not inside the blockquote, and the last line was empty.
      if (lastLineEmpty) { break; }

      // Case 3: another tag found.
      terminate = false;
      for (i = 0, l = terminatorRules.length; i < l; i++) {
        if (terminatorRules[i](state, nextLine, endLine, true)) {
          terminate = true;
          break;
        }
      }

      if (terminate) {
        // Quirk to enforce "hard termination mode" for paragraphs;
        // normally if you call `tokenize(state, startLine, nextLine)`,
        // paragraphs will look below nextLine for paragraph continuation,
        // but if blockquote is terminated by another tag, they shouldn't
        state.lineMax = nextLine;

        if (state.blkIndent !== 0) {
          // state.blkIndent was non-zero, we now set it to zero,
          // so we need to re-calculate all offsets to appear as
          // if indent wasn't changed
          oldBMarks.push(state.bMarks[nextLine]);
          oldBSCount.push(state.bsCount[nextLine]);
          oldTShift.push(state.tShift[nextLine]);
          oldSCount.push(state.sCount[nextLine]);
          state.sCount[nextLine] -= state.blkIndent;
        }

        break;
      }

      oldBMarks.push(state.bMarks[nextLine]);
      oldBSCount.push(state.bsCount[nextLine]);
      oldTShift.push(state.tShift[nextLine]);
      oldSCount.push(state.sCount[nextLine]);

      // A negative indentation means that this is a paragraph continuation
      //
      state.sCount[nextLine] = -1;
    }

    oldIndent = state.blkIndent;
    state.blkIndent = 0;

    token        = state.push('blockquote_open', 'blockquote', 1);
    token.markup = '>';
    token.map    = lines = [ startLine, 0 ];

    state.md.block.tokenize(state, startLine, nextLine);

    token        = state.push('blockquote_close', 'blockquote', -1);
    token.markup = '>';

    state.lineMax = oldLineMax;
    state.parentType = oldParentType;
    lines[1] = state.line;

    // Restore original tShift; this might not be necessary since the parser
    // has already been here, but just to make sure we can do that.
    for (i = 0; i < oldTShift.length; i++) {
      state.bMarks[i + startLine] = oldBMarks[i];
      state.tShift[i + startLine] = oldTShift[i];
      state.sCount[i + startLine] = oldSCount[i];
      state.bsCount[i + startLine] = oldBSCount[i];
    }
    state.blkIndent = oldIndent;

    return true;
  };

  var isSpace$b = utils.isSpace;


  var hr = function hr(state, startLine, endLine, silent) {
    var marker, cnt, ch, token,
        pos = state.bMarks[startLine] + state.tShift[startLine],
        max = state.eMarks[startLine];

    // if it's indented more than 3 spaces, it should be a code block
    if (state.sCount[startLine] - state.blkIndent >= 4) { return false; }

    marker = state.src.charCodeAt(pos++);

    // Check hr marker
    if (marker !== 0x2A/* * */ &&
        marker !== 0x2D/* - */ &&
        marker !== 0x5F/* _ */) {
      return false;
    }

    // markers can be mixed with spaces, but there should be at least 3 of them

    cnt = 1;
    while (pos < max) {
      ch = state.src.charCodeAt(pos++);
      if (ch !== marker && !isSpace$b(ch)) { return false; }
      if (ch === marker) { cnt++; }
    }

    if (cnt < 3) { return false; }

    if (silent) { return true; }

    state.line = startLine + 1;

    token        = state.push('hr', 'hr', 0);
    token.map    = [ startLine, state.line ];
    token.markup = Array(cnt + 1).join(String.fromCharCode(marker));

    return true;
  };

  var isSpace$a = utils.isSpace;


  // Search `[-+*][\n ]`, returns next pos after marker on success
  // or -1 on fail.
  function skipBulletListMarker(state, startLine) {
    var marker, pos, max, ch;

    pos = state.bMarks[startLine] + state.tShift[startLine];
    max = state.eMarks[startLine];

    marker = state.src.charCodeAt(pos++);
    // Check bullet
    if (marker !== 0x2A/* * */ &&
        marker !== 0x2D/* - */ &&
        marker !== 0x2B/* + */) {
      return -1;
    }

    if (pos < max) {
      ch = state.src.charCodeAt(pos);

      if (!isSpace$a(ch)) {
        // " -test " - is not a list item
        return -1;
      }
    }

    return pos;
  }

  // Search `\d+[.)][\n ]`, returns next pos after marker on success
  // or -1 on fail.
  function skipOrderedListMarker(state, startLine) {
    var ch,
        start = state.bMarks[startLine] + state.tShift[startLine],
        pos = start,
        max = state.eMarks[startLine];

    // List marker should have at least 2 chars (digit + dot)
    if (pos + 1 >= max) { return -1; }

    ch = state.src.charCodeAt(pos++);

    if (ch < 0x30/* 0 */ || ch > 0x39/* 9 */) { return -1; }

    for (;;) {
      // EOL -> fail
      if (pos >= max) { return -1; }

      ch = state.src.charCodeAt(pos++);

      if (ch >= 0x30/* 0 */ && ch <= 0x39/* 9 */) {

        // List marker should have no more than 9 digits
        // (prevents integer overflow in browsers)
        if (pos - start >= 10) { return -1; }

        continue;
      }

      // found valid marker
      if (ch === 0x29/* ) */ || ch === 0x2e/* . */) {
        break;
      }

      return -1;
    }


    if (pos < max) {
      ch = state.src.charCodeAt(pos);

      if (!isSpace$a(ch)) {
        // " 1.test " - is not a list item
        return -1;
      }
    }
    return pos;
  }

  function markTightParagraphs(state, idx) {
    var i, l,
        level = state.level + 2;

    for (i = idx + 2, l = state.tokens.length - 2; i < l; i++) {
      if (state.tokens[i].level === level && state.tokens[i].type === 'paragraph_open') {
        state.tokens[i + 2].hidden = true;
        state.tokens[i].hidden = true;
        i += 2;
      }
    }
  }


  var list = function list(state, startLine, endLine, silent) {
    var ch,
        contentStart,
        i,
        indent,
        indentAfterMarker,
        initial,
        isOrdered,
        itemLines,
        l,
        listLines,
        listTokIdx,
        markerCharCode,
        markerValue,
        max,
        nextLine,
        offset,
        oldListIndent,
        oldParentType,
        oldSCount,
        oldTShift,
        oldTight,
        pos,
        posAfterMarker,
        prevEmptyEnd,
        start,
        terminate,
        terminatorRules,
        token,
        isTerminatingParagraph = false,
        tight = true;

    // if it's indented more than 3 spaces, it should be a code block
    if (state.sCount[startLine] - state.blkIndent >= 4) { return false; }

    // Special case:
    //  - item 1
    //   - item 2
    //    - item 3
    //     - item 4
    //      - this one is a paragraph continuation
    if (state.listIndent >= 0 &&
        state.sCount[startLine] - state.listIndent >= 4 &&
        state.sCount[startLine] < state.blkIndent) {
      return false;
    }

    // limit conditions when list can interrupt
    // a paragraph (validation mode only)
    if (silent && state.parentType === 'paragraph') {
      // Next list item should still terminate previous list item;
      //
      // This code can fail if plugins use blkIndent as well as lists,
      // but I hope the spec gets fixed long before that happens.
      //
      if (state.tShift[startLine] >= state.blkIndent) {
        isTerminatingParagraph = true;
      }
    }

    // Detect list type and position after marker
    if ((posAfterMarker = skipOrderedListMarker(state, startLine)) >= 0) {
      isOrdered = true;
      start = state.bMarks[startLine] + state.tShift[startLine];
      markerValue = Number(state.src.slice(start, posAfterMarker - 1));

      // If we're starting a new ordered list right after
      // a paragraph, it should start with 1.
      if (isTerminatingParagraph && markerValue !== 1) { return false; }

    } else if ((posAfterMarker = skipBulletListMarker(state, startLine)) >= 0) {
      isOrdered = false;

    } else {
      return false;
    }

    // If we're starting a new unordered list right after
    // a paragraph, first line should not be empty.
    if (isTerminatingParagraph) {
      if (state.skipSpaces(posAfterMarker) >= state.eMarks[startLine]) { return false; }
    }

    // We should terminate list on style change. Remember first one to compare.
    markerCharCode = state.src.charCodeAt(posAfterMarker - 1);

    // For validation mode we can terminate immediately
    if (silent) { return true; }

    // Start list
    listTokIdx = state.tokens.length;

    if (isOrdered) {
      token       = state.push('ordered_list_open', 'ol', 1);
      if (markerValue !== 1) {
        token.attrs = [ [ 'start', markerValue ] ];
      }

    } else {
      token       = state.push('bullet_list_open', 'ul', 1);
    }

    token.map    = listLines = [ startLine, 0 ];
    token.markup = String.fromCharCode(markerCharCode);

    //
    // Iterate list items
    //

    nextLine = startLine;
    prevEmptyEnd = false;
    terminatorRules = state.md.block.ruler.getRules('list');

    oldParentType = state.parentType;
    state.parentType = 'list';

    while (nextLine < endLine) {
      pos = posAfterMarker;
      max = state.eMarks[nextLine];

      initial = offset = state.sCount[nextLine] + posAfterMarker - (state.bMarks[startLine] + state.tShift[startLine]);

      while (pos < max) {
        ch = state.src.charCodeAt(pos);

        if (ch === 0x09) {
          offset += 4 - (offset + state.bsCount[nextLine]) % 4;
        } else if (ch === 0x20) {
          offset++;
        } else {
          break;
        }

        pos++;
      }

      contentStart = pos;

      if (contentStart >= max) {
        // trimming space in "-    \n  3" case, indent is 1 here
        indentAfterMarker = 1;
      } else {
        indentAfterMarker = offset - initial;
      }

      // If we have more than 4 spaces, the indent is 1
      // (the rest is just indented code block)
      if (indentAfterMarker > 4) { indentAfterMarker = 1; }

      // "  -  test"
      //  ^^^^^ - calculating total length of this thing
      indent = initial + indentAfterMarker;

      // Run subparser & write tokens
      token        = state.push('list_item_open', 'li', 1);
      token.markup = String.fromCharCode(markerCharCode);
      token.map    = itemLines = [ startLine, 0 ];
      if (isOrdered) {
        token.info = state.src.slice(start, posAfterMarker - 1);
      }

      // change current state, then restore it after parser subcall
      oldTight = state.tight;
      oldTShift = state.tShift[startLine];
      oldSCount = state.sCount[startLine];

      //  - example list
      // ^ listIndent position will be here
      //   ^ blkIndent position will be here
      //
      oldListIndent = state.listIndent;
      state.listIndent = state.blkIndent;
      state.blkIndent = indent;

      state.tight = true;
      state.tShift[startLine] = contentStart - state.bMarks[startLine];
      state.sCount[startLine] = offset;

      if (contentStart >= max && state.isEmpty(startLine + 1)) {
        // workaround for this case
        // (list item is empty, list terminates before "foo"):
        // ~~~~~~~~
        //   -
        //
        //     foo
        // ~~~~~~~~
        state.line = Math.min(state.line + 2, endLine);
      } else {
        state.md.block.tokenize(state, startLine, endLine, true);
      }

      // If any of list item is tight, mark list as tight
      if (!state.tight || prevEmptyEnd) {
        tight = false;
      }
      // Item become loose if finish with empty line,
      // but we should filter last element, because it means list finish
      prevEmptyEnd = (state.line - startLine) > 1 && state.isEmpty(state.line - 1);

      state.blkIndent = state.listIndent;
      state.listIndent = oldListIndent;
      state.tShift[startLine] = oldTShift;
      state.sCount[startLine] = oldSCount;
      state.tight = oldTight;

      token        = state.push('list_item_close', 'li', -1);
      token.markup = String.fromCharCode(markerCharCode);

      nextLine = startLine = state.line;
      itemLines[1] = nextLine;
      contentStart = state.bMarks[startLine];

      if (nextLine >= endLine) { break; }

      //
      // Try to check if list is terminated or continued.
      //
      if (state.sCount[nextLine] < state.blkIndent) { break; }

      // if it's indented more than 3 spaces, it should be a code block
      if (state.sCount[startLine] - state.blkIndent >= 4) { break; }

      // fail if terminating block found
      terminate = false;
      for (i = 0, l = terminatorRules.length; i < l; i++) {
        if (terminatorRules[i](state, nextLine, endLine, true)) {
          terminate = true;
          break;
        }
      }
      if (terminate) { break; }

      // fail if list has another type
      if (isOrdered) {
        posAfterMarker = skipOrderedListMarker(state, nextLine);
        if (posAfterMarker < 0) { break; }
        start = state.bMarks[nextLine] + state.tShift[nextLine];
      } else {
        posAfterMarker = skipBulletListMarker(state, nextLine);
        if (posAfterMarker < 0) { break; }
      }

      if (markerCharCode !== state.src.charCodeAt(posAfterMarker - 1)) { break; }
    }

    // Finalize list
    if (isOrdered) {
      token = state.push('ordered_list_close', 'ol', -1);
    } else {
      token = state.push('bullet_list_close', 'ul', -1);
    }
    token.markup = String.fromCharCode(markerCharCode);

    listLines[1] = nextLine;
    state.line = nextLine;

    state.parentType = oldParentType;

    // mark paragraphs tight if needed
    if (tight) {
      markTightParagraphs(state, listTokIdx);
    }

    return true;
  };

  var normalizeReference$2   = utils.normalizeReference;
  var isSpace$9              = utils.isSpace;


  var reference = function reference(state, startLine, _endLine, silent) {
    var ch,
        destEndPos,
        destEndLineNo,
        endLine,
        href,
        i,
        l,
        label,
        labelEnd,
        oldParentType,
        res,
        start,
        str,
        terminate,
        terminatorRules,
        title,
        lines = 0,
        pos = state.bMarks[startLine] + state.tShift[startLine],
        max = state.eMarks[startLine],
        nextLine = startLine + 1;

    // if it's indented more than 3 spaces, it should be a code block
    if (state.sCount[startLine] - state.blkIndent >= 4) { return false; }

    if (state.src.charCodeAt(pos) !== 0x5B/* [ */) { return false; }

    // Simple check to quickly interrupt scan on [link](url) at the start of line.
    // Can be useful on practice: https://github.com/markdown-it/markdown-it/issues/54
    while (++pos < max) {
      if (state.src.charCodeAt(pos) === 0x5D /* ] */ &&
          state.src.charCodeAt(pos - 1) !== 0x5C/* \ */) {
        if (pos + 1 === max) { return false; }
        if (state.src.charCodeAt(pos + 1) !== 0x3A/* : */) { return false; }
        break;
      }
    }

    endLine = state.lineMax;

    // jump line-by-line until empty one or EOF
    terminatorRules = state.md.block.ruler.getRules('reference');

    oldParentType = state.parentType;
    state.parentType = 'reference';

    for (; nextLine < endLine && !state.isEmpty(nextLine); nextLine++) {
      // this would be a code block normally, but after paragraph
      // it's considered a lazy continuation regardless of what's there
      if (state.sCount[nextLine] - state.blkIndent > 3) { continue; }

      // quirk for blockquotes, this line should already be checked by that rule
      if (state.sCount[nextLine] < 0) { continue; }

      // Some tags can terminate paragraph without empty line.
      terminate = false;
      for (i = 0, l = terminatorRules.length; i < l; i++) {
        if (terminatorRules[i](state, nextLine, endLine, true)) {
          terminate = true;
          break;
        }
      }
      if (terminate) { break; }
    }

    str = state.getLines(startLine, nextLine, state.blkIndent, false).trim();
    max = str.length;

    for (pos = 1; pos < max; pos++) {
      ch = str.charCodeAt(pos);
      if (ch === 0x5B /* [ */) {
        return false;
      } else if (ch === 0x5D /* ] */) {
        labelEnd = pos;
        break;
      } else if (ch === 0x0A /* \n */) {
        lines++;
      } else if (ch === 0x5C /* \ */) {
        pos++;
        if (pos < max && str.charCodeAt(pos) === 0x0A) {
          lines++;
        }
      }
    }

    if (labelEnd < 0 || str.charCodeAt(labelEnd + 1) !== 0x3A/* : */) { return false; }

    // [label]:   destination   'title'
    //         ^^^ skip optional whitespace here
    for (pos = labelEnd + 2; pos < max; pos++) {
      ch = str.charCodeAt(pos);
      if (ch === 0x0A) {
        lines++;
      } else if (isSpace$9(ch)) ; else {
        break;
      }
    }

    // [label]:   destination   'title'
    //            ^^^^^^^^^^^ parse this
    res = state.md.helpers.parseLinkDestination(str, pos, max);
    if (!res.ok) { return false; }

    href = state.md.normalizeLink(res.str);
    if (!state.md.validateLink(href)) { return false; }

    pos = res.pos;
    lines += res.lines;

    // save cursor state, we could require to rollback later
    destEndPos = pos;
    destEndLineNo = lines;

    // [label]:   destination   'title'
    //                       ^^^ skipping those spaces
    start = pos;
    for (; pos < max; pos++) {
      ch = str.charCodeAt(pos);
      if (ch === 0x0A) {
        lines++;
      } else if (isSpace$9(ch)) ; else {
        break;
      }
    }

    // [label]:   destination   'title'
    //                          ^^^^^^^ parse this
    res = state.md.helpers.parseLinkTitle(str, pos, max);
    if (pos < max && start !== pos && res.ok) {
      title = res.str;
      pos = res.pos;
      lines += res.lines;
    } else {
      title = '';
      pos = destEndPos;
      lines = destEndLineNo;
    }

    // skip trailing spaces until the rest of the line
    while (pos < max) {
      ch = str.charCodeAt(pos);
      if (!isSpace$9(ch)) { break; }
      pos++;
    }

    if (pos < max && str.charCodeAt(pos) !== 0x0A) {
      if (title) {
        // garbage at the end of the line after title,
        // but it could still be a valid reference if we roll back
        title = '';
        pos = destEndPos;
        lines = destEndLineNo;
        while (pos < max) {
          ch = str.charCodeAt(pos);
          if (!isSpace$9(ch)) { break; }
          pos++;
        }
      }
    }

    if (pos < max && str.charCodeAt(pos) !== 0x0A) {
      // garbage at the end of the line
      return false;
    }

    label = normalizeReference$2(str.slice(1, labelEnd));
    if (!label) {
      // CommonMark 0.20 disallows empty labels
      return false;
    }

    // Reference can not terminate anything. This check is for safety only.
    /*istanbul ignore if*/
    if (silent) { return true; }

    if (typeof state.env.references === 'undefined') {
      state.env.references = {};
    }
    if (typeof state.env.references[label] === 'undefined') {
      state.env.references[label] = { title: title, href: href };
    }

    state.parentType = oldParentType;

    state.line = startLine + lines + 1;
    return true;
  };

  // List of valid html blocks names, accorting to commonmark spec


  var html_blocks = [
    'address',
    'article',
    'aside',
    'base',
    'basefont',
    'blockquote',
    'body',
    'caption',
    'center',
    'col',
    'colgroup',
    'dd',
    'details',
    'dialog',
    'dir',
    'div',
    'dl',
    'dt',
    'fieldset',
    'figcaption',
    'figure',
    'footer',
    'form',
    'frame',
    'frameset',
    'h1',
    'h2',
    'h3',
    'h4',
    'h5',
    'h6',
    'head',
    'header',
    'hr',
    'html',
    'iframe',
    'legend',
    'li',
    'link',
    'main',
    'menu',
    'menuitem',
    'nav',
    'noframes',
    'ol',
    'optgroup',
    'option',
    'p',
    'param',
    'section',
    'source',
    'summary',
    'table',
    'tbody',
    'td',
    'tfoot',
    'th',
    'thead',
    'title',
    'tr',
    'track',
    'ul'
  ];

  // Regexps to match html elements

  var attr_name     = '[a-zA-Z_:][a-zA-Z0-9:._-]*';

  var unquoted      = '[^"\'=<>`\\x00-\\x20]+';
  var single_quoted = "'[^']*'";
  var double_quoted = '"[^"]*"';

  var attr_value  = '(?:' + unquoted + '|' + single_quoted + '|' + double_quoted + ')';

  var attribute   = '(?:\\s+' + attr_name + '(?:\\s*=\\s*' + attr_value + ')?)';

  var open_tag    = '<[A-Za-z][A-Za-z0-9\\-]*' + attribute + '*\\s*\\/?>';

  var close_tag   = '<\\/[A-Za-z][A-Za-z0-9\\-]*\\s*>';
  var comment     = '<!---->|<!--(?:-?[^>-])(?:-?[^-])*-->';
  var processing  = '<[?][\\s\\S]*?[?]>';
  var declaration = '<![A-Z]+\\s+[^>]*>';
  var cdata       = '<!\\[CDATA\\[[\\s\\S]*?\\]\\]>';

  var HTML_TAG_RE$1 = new RegExp('^(?:' + open_tag + '|' + close_tag + '|' + comment +
                          '|' + processing + '|' + declaration + '|' + cdata + ')');
  var HTML_OPEN_CLOSE_TAG_RE$1 = new RegExp('^(?:' + open_tag + '|' + close_tag + ')');

  var HTML_TAG_RE_1 = HTML_TAG_RE$1;
  var HTML_OPEN_CLOSE_TAG_RE_1 = HTML_OPEN_CLOSE_TAG_RE$1;

  var html_re = {
  	HTML_TAG_RE: HTML_TAG_RE_1,
  	HTML_OPEN_CLOSE_TAG_RE: HTML_OPEN_CLOSE_TAG_RE_1
  };

  var HTML_OPEN_CLOSE_TAG_RE = html_re.HTML_OPEN_CLOSE_TAG_RE;

  // An array of opening and corresponding closing sequences for html tags,
  // last argument defines whether it can terminate a paragraph or not
  //
  var HTML_SEQUENCES = [
    [ /^<(script|pre|style|textarea)(?=(\s|>|$))/i, /<\/(script|pre|style|textarea)>/i, true ],
    [ /^<!--/,        /-->/,   true ],
    [ /^<\?/,         /\?>/,   true ],
    [ /^<![A-Z]/,     />/,     true ],
    [ /^<!\[CDATA\[/, /\]\]>/, true ],
    [ new RegExp('^</?(' + html_blocks.join('|') + ')(?=(\\s|/?>|$))', 'i'), /^$/, true ],
    [ new RegExp(HTML_OPEN_CLOSE_TAG_RE.source + '\\s*$'),  /^$/, false ]
  ];


  var html_block = function html_block(state, startLine, endLine, silent) {
    var i, nextLine, token, lineText,
        pos = state.bMarks[startLine] + state.tShift[startLine],
        max = state.eMarks[startLine];

    // if it's indented more than 3 spaces, it should be a code block
    if (state.sCount[startLine] - state.blkIndent >= 4) { return false; }

    if (!state.md.options.html) { return false; }

    if (state.src.charCodeAt(pos) !== 0x3C/* < */) { return false; }

    lineText = state.src.slice(pos, max);

    for (i = 0; i < HTML_SEQUENCES.length; i++) {
      if (HTML_SEQUENCES[i][0].test(lineText)) { break; }
    }

    if (i === HTML_SEQUENCES.length) { return false; }

    if (silent) {
      // true if this sequence can be a terminator, false otherwise
      return HTML_SEQUENCES[i][2];
    }

    nextLine = startLine + 1;

    // If we are here - we detected HTML block.
    // Let's roll down till block end.
    if (!HTML_SEQUENCES[i][1].test(lineText)) {
      for (; nextLine < endLine; nextLine++) {
        if (state.sCount[nextLine] < state.blkIndent) { break; }

        pos = state.bMarks[nextLine] + state.tShift[nextLine];
        max = state.eMarks[nextLine];
        lineText = state.src.slice(pos, max);

        if (HTML_SEQUENCES[i][1].test(lineText)) {
          if (lineText.length !== 0) { nextLine++; }
          break;
        }
      }
    }

    state.line = nextLine;

    token         = state.push('html_block', '', 0);
    token.map     = [ startLine, nextLine ];
    token.content = state.getLines(startLine, nextLine, state.blkIndent, true);

    return true;
  };

  var isSpace$8 = utils.isSpace;


  var heading$1 = function heading(state, startLine, endLine, silent) {
    var ch, level, tmp, token,
        pos = state.bMarks[startLine] + state.tShift[startLine],
        max = state.eMarks[startLine];

    // if it's indented more than 3 spaces, it should be a code block
    if (state.sCount[startLine] - state.blkIndent >= 4) { return false; }

    ch  = state.src.charCodeAt(pos);

    if (ch !== 0x23/* # */ || pos >= max) { return false; }

    // count heading level
    level = 1;
    ch = state.src.charCodeAt(++pos);
    while (ch === 0x23/* # */ && pos < max && level <= 6) {
      level++;
      ch = state.src.charCodeAt(++pos);
    }

    if (level > 6 || (pos < max && !isSpace$8(ch))) { return false; }

    if (silent) { return true; }

    // Let's cut tails like '    ###  ' from the end of string

    max = state.skipSpacesBack(max, pos);
    tmp = state.skipCharsBack(max, 0x23, pos); // #
    if (tmp > pos && isSpace$8(state.src.charCodeAt(tmp - 1))) {
      max = tmp;
    }

    state.line = startLine + 1;

    token        = state.push('heading_open', 'h' + String(level), 1);
    token.markup = '########'.slice(0, level);
    token.map    = [ startLine, state.line ];

    token          = state.push('inline', '', 0);
    token.content  = state.src.slice(pos, max).trim();
    token.map      = [ startLine, state.line ];
    token.children = [];

    token        = state.push('heading_close', 'h' + String(level), -1);
    token.markup = '########'.slice(0, level);

    return true;
  };

  // lheading (---, ===)


  var lheading = function lheading(state, startLine, endLine/*, silent*/) {
    var content, terminate, i, l, token, pos, max, level, marker,
        nextLine = startLine + 1, oldParentType,
        terminatorRules = state.md.block.ruler.getRules('paragraph');

    // if it's indented more than 3 spaces, it should be a code block
    if (state.sCount[startLine] - state.blkIndent >= 4) { return false; }

    oldParentType = state.parentType;
    state.parentType = 'paragraph'; // use paragraph to match terminatorRules

    // jump line-by-line until empty one or EOF
    for (; nextLine < endLine && !state.isEmpty(nextLine); nextLine++) {
      // this would be a code block normally, but after paragraph
      // it's considered a lazy continuation regardless of what's there
      if (state.sCount[nextLine] - state.blkIndent > 3) { continue; }

      //
      // Check for underline in setext header
      //
      if (state.sCount[nextLine] >= state.blkIndent) {
        pos = state.bMarks[nextLine] + state.tShift[nextLine];
        max = state.eMarks[nextLine];

        if (pos < max) {
          marker = state.src.charCodeAt(pos);

          if (marker === 0x2D/* - */ || marker === 0x3D/* = */) {
            pos = state.skipChars(pos, marker);
            pos = state.skipSpaces(pos);

            if (pos >= max) {
              level = (marker === 0x3D/* = */ ? 1 : 2);
              break;
            }
          }
        }
      }

      // quirk for blockquotes, this line should already be checked by that rule
      if (state.sCount[nextLine] < 0) { continue; }

      // Some tags can terminate paragraph without empty line.
      terminate = false;
      for (i = 0, l = terminatorRules.length; i < l; i++) {
        if (terminatorRules[i](state, nextLine, endLine, true)) {
          terminate = true;
          break;
        }
      }
      if (terminate) { break; }
    }

    if (!level) {
      // Didn't find valid underline
      return false;
    }

    content = state.getLines(startLine, nextLine, state.blkIndent, false).trim();

    state.line = nextLine + 1;

    token          = state.push('heading_open', 'h' + String(level), 1);
    token.markup   = String.fromCharCode(marker);
    token.map      = [ startLine, state.line ];

    token          = state.push('inline', '', 0);
    token.content  = content;
    token.map      = [ startLine, state.line - 1 ];
    token.children = [];

    token          = state.push('heading_close', 'h' + String(level), -1);
    token.markup   = String.fromCharCode(marker);

    state.parentType = oldParentType;

    return true;
  };

  // Paragraph


  var paragraph$1 = function paragraph(state, startLine/*, endLine*/) {
    var content, terminate, i, l, token, oldParentType,
        nextLine = startLine + 1,
        terminatorRules = state.md.block.ruler.getRules('paragraph'),
        endLine = state.lineMax;

    oldParentType = state.parentType;
    state.parentType = 'paragraph';

    // jump line-by-line until empty one or EOF
    for (; nextLine < endLine && !state.isEmpty(nextLine); nextLine++) {
      // this would be a code block normally, but after paragraph
      // it's considered a lazy continuation regardless of what's there
      if (state.sCount[nextLine] - state.blkIndent > 3) { continue; }

      // quirk for blockquotes, this line should already be checked by that rule
      if (state.sCount[nextLine] < 0) { continue; }

      // Some tags can terminate paragraph without empty line.
      terminate = false;
      for (i = 0, l = terminatorRules.length; i < l; i++) {
        if (terminatorRules[i](state, nextLine, endLine, true)) {
          terminate = true;
          break;
        }
      }
      if (terminate) { break; }
    }

    content = state.getLines(startLine, nextLine, state.blkIndent, false).trim();

    state.line = nextLine;

    token          = state.push('paragraph_open', 'p', 1);
    token.map      = [ startLine, state.line ];

    token          = state.push('inline', '', 0);
    token.content  = content;
    token.map      = [ startLine, state.line ];
    token.children = [];

    token          = state.push('paragraph_close', 'p', -1);

    state.parentType = oldParentType;

    return true;
  };

  var isSpace$7 = utils.isSpace;


  function StateBlock(src, md, env, tokens) {
    var ch, s, start, pos, len, indent, offset, indent_found;

    this.src = src;

    // link to parser instance
    this.md     = md;

    this.env = env;

    //
    // Internal state vartiables
    //

    this.tokens = tokens;

    this.bMarks = [];  // line begin offsets for fast jumps
    this.eMarks = [];  // line end offsets for fast jumps
    this.tShift = [];  // offsets of the first non-space characters (tabs not expanded)
    this.sCount = [];  // indents for each line (tabs expanded)

    // An amount of virtual spaces (tabs expanded) between beginning
    // of each line (bMarks) and real beginning of that line.
    //
    // It exists only as a hack because blockquotes override bMarks
    // losing information in the process.
    //
    // It's used only when expanding tabs, you can think about it as
    // an initial tab length, e.g. bsCount=21 applied to string `\t123`
    // means first tab should be expanded to 4-21%4 === 3 spaces.
    //
    this.bsCount = [];

    // block parser variables
    this.blkIndent  = 0; // required block content indent (for example, if we are
                         // inside a list, it would be positioned after list marker)
    this.line       = 0; // line index in src
    this.lineMax    = 0; // lines count
    this.tight      = false;  // loose/tight mode for lists
    this.ddIndent   = -1; // indent of the current dd block (-1 if there isn't any)
    this.listIndent = -1; // indent of the current list block (-1 if there isn't any)

    // can be 'blockquote', 'list', 'root', 'paragraph' or 'reference'
    // used in lists to determine if they interrupt a paragraph
    this.parentType = 'root';

    this.level = 0;

    // renderer
    this.result = '';

    // Create caches
    // Generate markers.
    s = this.src;
    indent_found = false;

    for (start = pos = indent = offset = 0, len = s.length; pos < len; pos++) {
      ch = s.charCodeAt(pos);

      if (!indent_found) {
        if (isSpace$7(ch)) {
          indent++;

          if (ch === 0x09) {
            offset += 4 - offset % 4;
          } else {
            offset++;
          }
          continue;
        } else {
          indent_found = true;
        }
      }

      if (ch === 0x0A || pos === len - 1) {
        if (ch !== 0x0A) { pos++; }
        this.bMarks.push(start);
        this.eMarks.push(pos);
        this.tShift.push(indent);
        this.sCount.push(offset);
        this.bsCount.push(0);

        indent_found = false;
        indent = 0;
        offset = 0;
        start = pos + 1;
      }
    }

    // Push fake entry to simplify cache bounds checks
    this.bMarks.push(s.length);
    this.eMarks.push(s.length);
    this.tShift.push(0);
    this.sCount.push(0);
    this.bsCount.push(0);

    this.lineMax = this.bMarks.length - 1; // don't count last fake line
  }

  // Push new token to "stream".
  //
  StateBlock.prototype.push = function (type, tag, nesting) {
    var token$1 = new token(type, tag, nesting);
    token$1.block = true;

    if (nesting < 0) { this.level--; } // closing tag
    token$1.level = this.level;
    if (nesting > 0) { this.level++; } // opening tag

    this.tokens.push(token$1);
    return token$1;
  };

  StateBlock.prototype.isEmpty = function isEmpty(line) {
    return this.bMarks[line] + this.tShift[line] >= this.eMarks[line];
  };

  StateBlock.prototype.skipEmptyLines = function skipEmptyLines(from) {
    for (var max = this.lineMax; from < max; from++) {
      if (this.bMarks[from] + this.tShift[from] < this.eMarks[from]) {
        break;
      }
    }
    return from;
  };

  // Skip spaces from given position.
  StateBlock.prototype.skipSpaces = function skipSpaces(pos) {
    var ch;

    for (var max = this.src.length; pos < max; pos++) {
      ch = this.src.charCodeAt(pos);
      if (!isSpace$7(ch)) { break; }
    }
    return pos;
  };

  // Skip spaces from given position in reverse.
  StateBlock.prototype.skipSpacesBack = function skipSpacesBack(pos, min) {
    if (pos <= min) { return pos; }

    while (pos > min) {
      if (!isSpace$7(this.src.charCodeAt(--pos))) { return pos + 1; }
    }
    return pos;
  };

  // Skip char codes from given position
  StateBlock.prototype.skipChars = function skipChars(pos, code) {
    for (var max = this.src.length; pos < max; pos++) {
      if (this.src.charCodeAt(pos) !== code) { break; }
    }
    return pos;
  };

  // Skip char codes reverse from given position - 1
  StateBlock.prototype.skipCharsBack = function skipCharsBack(pos, code, min) {
    if (pos <= min) { return pos; }

    while (pos > min) {
      if (code !== this.src.charCodeAt(--pos)) { return pos + 1; }
    }
    return pos;
  };

  // cut lines range from source.
  StateBlock.prototype.getLines = function getLines(begin, end, indent, keepLastLF) {
    var i, lineIndent, ch, first, last, queue, lineStart,
        line = begin;

    if (begin >= end) {
      return '';
    }

    queue = new Array(end - begin);

    for (i = 0; line < end; line++, i++) {
      lineIndent = 0;
      lineStart = first = this.bMarks[line];

      if (line + 1 < end || keepLastLF) {
        // No need for bounds check because we have fake entry on tail.
        last = this.eMarks[line] + 1;
      } else {
        last = this.eMarks[line];
      }

      while (first < last && lineIndent < indent) {
        ch = this.src.charCodeAt(first);

        if (isSpace$7(ch)) {
          if (ch === 0x09) {
            lineIndent += 4 - (lineIndent + this.bsCount[line]) % 4;
          } else {
            lineIndent++;
          }
        } else if (first - lineStart < this.tShift[line]) {
          // patched tShift masked characters to look like spaces (blockquotes, list markers)
          lineIndent++;
        } else {
          break;
        }

        first++;
      }

      if (lineIndent > indent) {
        // partially expanding tabs in code blocks, e.g '\t\tfoobar'
        // with indent=2 becomes '  \tfoobar'
        queue[i] = new Array(lineIndent - indent + 1).join(' ') + this.src.slice(first, last);
      } else {
        queue[i] = this.src.slice(first, last);
      }
    }

    return queue.join('');
  };

  // re-export Token class to use in block rules
  StateBlock.prototype.Token = token;


  var state_block = StateBlock;

  var _rules$1 = [
    // First 2 params - rule name & source. Secondary array - list of rules,
    // which can be terminated by this one.
    [ 'table',      table$1,      [ 'paragraph', 'reference' ] ],
    [ 'code',       code ],
    [ 'fence',      fence,      [ 'paragraph', 'reference', 'blockquote', 'list' ] ],
    [ 'blockquote', blockquote, [ 'paragraph', 'reference', 'blockquote', 'list' ] ],
    [ 'hr',         hr,         [ 'paragraph', 'reference', 'blockquote', 'list' ] ],
    [ 'list',       list,       [ 'paragraph', 'reference', 'blockquote' ] ],
    [ 'reference',  reference ],
    [ 'html_block', html_block, [ 'paragraph', 'reference', 'blockquote' ] ],
    [ 'heading',    heading$1,    [ 'paragraph', 'reference', 'blockquote' ] ],
    [ 'lheading',   lheading ],
    [ 'paragraph',  paragraph$1 ]
  ];


  /**
   * new ParserBlock()
   **/
  function ParserBlock() {
    /**
     * ParserBlock#ruler -> Ruler
     *
     * [[Ruler]] instance. Keep configuration of block rules.
     **/
    this.ruler = new ruler();

    for (var i = 0; i < _rules$1.length; i++) {
      this.ruler.push(_rules$1[i][0], _rules$1[i][1], { alt: (_rules$1[i][2] || []).slice() });
    }
  }


  // Generate tokens for input range
  //
  ParserBlock.prototype.tokenize = function (state, startLine, endLine) {
    var ok, i,
        rules = this.ruler.getRules(''),
        len = rules.length,
        line = startLine,
        hasEmptyLines = false,
        maxNesting = state.md.options.maxNesting;

    while (line < endLine) {
      state.line = line = state.skipEmptyLines(line);
      if (line >= endLine) { break; }

      // Termination condition for nested calls.
      // Nested calls currently used for blockquotes & lists
      if (state.sCount[line] < state.blkIndent) { break; }

      // If nesting level exceeded - skip tail to the end. That's not ordinary
      // situation and we should not care about content.
      if (state.level >= maxNesting) {
        state.line = endLine;
        break;
      }

      // Try all possible rules.
      // On success, rule should:
      //
      // - update `state.line`
      // - update `state.tokens`
      // - return true

      for (i = 0; i < len; i++) {
        ok = rules[i](state, line, endLine, false);
        if (ok) { break; }
      }

      // set state.tight if we had an empty line before current tag
      // i.e. latest empty line should not count
      state.tight = !hasEmptyLines;

      // paragraph might "eat" one newline after it in nested lists
      if (state.isEmpty(state.line - 1)) {
        hasEmptyLines = true;
      }

      line = state.line;

      if (line < endLine && state.isEmpty(line)) {
        hasEmptyLines = true;
        line++;
        state.line = line;
      }
    }
  };


  /**
   * ParserBlock.parse(str, md, env, outTokens)
   *
   * Process input string and push block tokens into `outTokens`
   **/
  ParserBlock.prototype.parse = function (src, md, env, outTokens) {
    var state;

    if (!src) { return; }

    state = new this.State(src, md, env, outTokens);

    this.tokenize(state, state.line, state.lineMax);
  };


  ParserBlock.prototype.State = state_block;


  var parser_block = ParserBlock;

  // Skip text characters for text token, place those to pending buffer


  // Rule to skip pure text
  // '{}$%@~+=:' reserved for extentions

  // !, ", #, $, %, &, ', (, ), *, +, ,, -, ., /, :, ;, <, =, >, ?, @, [, \, ], ^, _, `, {, |, }, or ~

  // !!!! Don't confuse with "Markdown ASCII Punctuation" chars
  // http://spec.commonmark.org/0.15/#ascii-punctuation-character
  function isTerminatorChar(ch) {
    switch (ch) {
      case 0x0A/* \n */:
      case 0x21/* ! */:
      case 0x23/* # */:
      case 0x24/* $ */:
      case 0x25/* % */:
      case 0x26/* & */:
      case 0x2A/* * */:
      case 0x2B/* + */:
      case 0x2D/* - */:
      case 0x3A/* : */:
      case 0x3C/* < */:
      case 0x3D/* = */:
      case 0x3E/* > */:
      case 0x40/* @ */:
      case 0x5B/* [ */:
      case 0x5C/* \ */:
      case 0x5D/* ] */:
      case 0x5E/* ^ */:
      case 0x5F/* _ */:
      case 0x60/* ` */:
      case 0x7B/* { */:
      case 0x7D/* } */:
      case 0x7E/* ~ */:
        return true;
      default:
        return false;
    }
  }

  var text$1 = function text(state, silent) {
    var pos = state.pos;

    while (pos < state.posMax && !isTerminatorChar(state.src.charCodeAt(pos))) {
      pos++;
    }

    if (pos === state.pos) { return false; }

    if (!silent) { state.pending += state.src.slice(state.pos, pos); }

    state.pos = pos;

    return true;
  };

  var isSpace$6 = utils.isSpace;


  var newline = function newline(state, silent) {
    var pmax, max, pos = state.pos;

    if (state.src.charCodeAt(pos) !== 0x0A/* \n */) { return false; }

    pmax = state.pending.length - 1;
    max = state.posMax;

    // '  \n' -> hardbreak
    // Lookup in pending chars is bad practice! Don't copy to other rules!
    // Pending string is stored in concat mode, indexed lookups will cause
    // convertion to flat mode.
    if (!silent) {
      if (pmax >= 0 && state.pending.charCodeAt(pmax) === 0x20) {
        if (pmax >= 1 && state.pending.charCodeAt(pmax - 1) === 0x20) {
          state.pending = state.pending.replace(/ +$/, '');
          state.push('hardbreak', 'br', 0);
        } else {
          state.pending = state.pending.slice(0, -1);
          state.push('softbreak', 'br', 0);
        }

      } else {
        state.push('softbreak', 'br', 0);
      }
    }

    pos++;

    // skip heading spaces for next line
    while (pos < max && isSpace$6(state.src.charCodeAt(pos))) { pos++; }

    state.pos = pos;
    return true;
  };

  var isSpace$5 = utils.isSpace;

  var ESCAPED = [];

  for (var i = 0; i < 256; i++) { ESCAPED.push(0); }

  '\\!"#$%&\'()*+,./:;<=>?@[]^_`{|}~-'
    .split('').forEach(function (ch) { ESCAPED[ch.charCodeAt(0)] = 1; });


  var _escape = function escape(state, silent) {
    var ch, pos = state.pos, max = state.posMax;

    if (state.src.charCodeAt(pos) !== 0x5C/* \ */) { return false; }

    pos++;

    if (pos < max) {
      ch = state.src.charCodeAt(pos);

      if (ch < 256 && ESCAPED[ch] !== 0) {
        if (!silent) { state.pending += state.src[pos]; }
        state.pos += 2;
        return true;
      }

      if (ch === 0x0A) {
        if (!silent) {
          state.push('hardbreak', 'br', 0);
        }

        pos++;
        // skip leading whitespaces from next line
        while (pos < max) {
          ch = state.src.charCodeAt(pos);
          if (!isSpace$5(ch)) { break; }
          pos++;
        }

        state.pos = pos;
        return true;
      }
    }

    if (!silent) { state.pending += '\\'; }
    state.pos++;
    return true;
  };

  // Parse backticks


  var backticks = function backtick(state, silent) {
    var start, max, marker, token, matchStart, matchEnd, openerLength, closerLength,
        pos = state.pos,
        ch = state.src.charCodeAt(pos);

    if (ch !== 0x60/* ` */) { return false; }

    start = pos;
    pos++;
    max = state.posMax;

    // scan marker length
    while (pos < max && state.src.charCodeAt(pos) === 0x60/* ` */) { pos++; }

    marker = state.src.slice(start, pos);
    openerLength = marker.length;

    if (state.backticksScanned && (state.backticks[openerLength] || 0) <= start) {
      if (!silent) { state.pending += marker; }
      state.pos += openerLength;
      return true;
    }

    matchStart = matchEnd = pos;

    // Nothing found in the cache, scan until the end of the line (or until marker is found)
    while ((matchStart = state.src.indexOf('`', matchEnd)) !== -1) {
      matchEnd = matchStart + 1;

      // scan marker length
      while (matchEnd < max && state.src.charCodeAt(matchEnd) === 0x60/* ` */) { matchEnd++; }

      closerLength = matchEnd - matchStart;

      if (closerLength === openerLength) {
        // Found matching closer length.
        if (!silent) {
          token     = state.push('code_inline', 'code', 0);
          token.markup  = marker;
          token.content = state.src.slice(pos, matchStart)
            .replace(/\n/g, ' ')
            .replace(/^ (.+) $/, '$1');
        }
        state.pos = matchEnd;
        return true;
      }

      // Some different length found, put it in cache as upper limit of where closer can be found
      state.backticks[closerLength] = matchStart;
    }

    // Scanned through the end, didn't find anything
    state.backticksScanned = true;

    if (!silent) { state.pending += marker; }
    state.pos += openerLength;
    return true;
  };

  // ~~strike through~~


  // Insert each marker as a separate text token, and add it to delimiter list
  //
  var tokenize$1 = function strikethrough(state, silent) {
    var i, scanned, token, len, ch,
        start = state.pos,
        marker = state.src.charCodeAt(start);

    if (silent) { return false; }

    if (marker !== 0x7E/* ~ */) { return false; }

    scanned = state.scanDelims(state.pos, true);
    len = scanned.length;
    ch = String.fromCharCode(marker);

    if (len < 2) { return false; }

    if (len % 2) {
      token         = state.push('text', '', 0);
      token.content = ch;
      len--;
    }

    for (i = 0; i < len; i += 2) {
      token         = state.push('text', '', 0);
      token.content = ch + ch;

      state.delimiters.push({
        marker: marker,
        length: 0,     // disable "rule of 3" length checks meant for emphasis
        jump:   i / 2, // for `~~` 1 marker = 2 characters
        token:  state.tokens.length - 1,
        end:    -1,
        open:   scanned.can_open,
        close:  scanned.can_close
      });
    }

    state.pos += scanned.length;

    return true;
  };


  function postProcess$1(state, delimiters) {
    var i, j,
        startDelim,
        endDelim,
        token,
        loneMarkers = [],
        max = delimiters.length;

    for (i = 0; i < max; i++) {
      startDelim = delimiters[i];

      if (startDelim.marker !== 0x7E/* ~ */) {
        continue;
      }

      if (startDelim.end === -1) {
        continue;
      }

      endDelim = delimiters[startDelim.end];

      token         = state.tokens[startDelim.token];
      token.type    = 's_open';
      token.tag     = 's';
      token.nesting = 1;
      token.markup  = '~~';
      token.content = '';

      token         = state.tokens[endDelim.token];
      token.type    = 's_close';
      token.tag     = 's';
      token.nesting = -1;
      token.markup  = '~~';
      token.content = '';

      if (state.tokens[endDelim.token - 1].type === 'text' &&
          state.tokens[endDelim.token - 1].content === '~') {

        loneMarkers.push(endDelim.token - 1);
      }
    }

    // If a marker sequence has an odd number of characters, it's splitted
    // like this: `~~~~~` -> `~` + `~~` + `~~`, leaving one marker at the
    // start of the sequence.
    //
    // So, we have to move all those markers after subsequent s_close tags.
    //
    while (loneMarkers.length) {
      i = loneMarkers.pop();
      j = i + 1;

      while (j < state.tokens.length && state.tokens[j].type === 's_close') {
        j++;
      }

      j--;

      if (i !== j) {
        token = state.tokens[j];
        state.tokens[j] = state.tokens[i];
        state.tokens[i] = token;
      }
    }
  }


  // Walk through delimiter list and replace text tokens with tags
  //
  var postProcess_1$1 = function strikethrough(state) {
    var curr,
        tokens_meta = state.tokens_meta,
        max = state.tokens_meta.length;

    postProcess$1(state, state.delimiters);

    for (curr = 0; curr < max; curr++) {
      if (tokens_meta[curr] && tokens_meta[curr].delimiters) {
        postProcess$1(state, tokens_meta[curr].delimiters);
      }
    }
  };

  var strikethrough$1 = {
  	tokenize: tokenize$1,
  	postProcess: postProcess_1$1
  };

  // Process *this* and _that_


  // Insert each marker as a separate text token, and add it to delimiter list
  //
  var tokenize = function emphasis(state, silent) {
    var i, scanned, token,
        start = state.pos,
        marker = state.src.charCodeAt(start);

    if (silent) { return false; }

    if (marker !== 0x5F /* _ */ && marker !== 0x2A /* * */) { return false; }

    scanned = state.scanDelims(state.pos, marker === 0x2A);

    for (i = 0; i < scanned.length; i++) {
      token         = state.push('text', '', 0);
      token.content = String.fromCharCode(marker);

      state.delimiters.push({
        // Char code of the starting marker (number).
        //
        marker: marker,

        // Total length of these series of delimiters.
        //
        length: scanned.length,

        // An amount of characters before this one that's equivalent to
        // current one. In plain English: if this delimiter does not open
        // an emphasis, neither do previous `jump` characters.
        //
        // Used to skip sequences like "*****" in one step, for 1st asterisk
        // value will be 0, for 2nd it's 1 and so on.
        //
        jump:   i,

        // A position of the token this delimiter corresponds to.
        //
        token:  state.tokens.length - 1,

        // If this delimiter is matched as a valid opener, `end` will be
        // equal to its position, otherwise it's `-1`.
        //
        end:    -1,

        // Boolean flags that determine if this delimiter could open or close
        // an emphasis.
        //
        open:   scanned.can_open,
        close:  scanned.can_close
      });
    }

    state.pos += scanned.length;

    return true;
  };


  function postProcess(state, delimiters) {
    var i,
        startDelim,
        endDelim,
        token,
        ch,
        isStrong,
        max = delimiters.length;

    for (i = max - 1; i >= 0; i--) {
      startDelim = delimiters[i];

      if (startDelim.marker !== 0x5F/* _ */ && startDelim.marker !== 0x2A/* * */) {
        continue;
      }

      // Process only opening markers
      if (startDelim.end === -1) {
        continue;
      }

      endDelim = delimiters[startDelim.end];

      // If the previous delimiter has the same marker and is adjacent to this one,
      // merge those into one strong delimiter.
      //
      // `<em><em>whatever</em></em>` -> `<strong>whatever</strong>`
      //
      isStrong = i > 0 &&
                 delimiters[i - 1].end === startDelim.end + 1 &&
                 delimiters[i - 1].token === startDelim.token - 1 &&
                 delimiters[startDelim.end + 1].token === endDelim.token + 1 &&
                 delimiters[i - 1].marker === startDelim.marker;

      ch = String.fromCharCode(startDelim.marker);

      token         = state.tokens[startDelim.token];
      token.type    = isStrong ? 'strong_open' : 'em_open';
      token.tag     = isStrong ? 'strong' : 'em';
      token.nesting = 1;
      token.markup  = isStrong ? ch + ch : ch;
      token.content = '';

      token         = state.tokens[endDelim.token];
      token.type    = isStrong ? 'strong_close' : 'em_close';
      token.tag     = isStrong ? 'strong' : 'em';
      token.nesting = -1;
      token.markup  = isStrong ? ch + ch : ch;
      token.content = '';

      if (isStrong) {
        state.tokens[delimiters[i - 1].token].content = '';
        state.tokens[delimiters[startDelim.end + 1].token].content = '';
        i--;
      }
    }
  }


  // Walk through delimiter list and replace text tokens with tags
  //
  var postProcess_1 = function emphasis(state) {
    var curr,
        tokens_meta = state.tokens_meta,
        max = state.tokens_meta.length;

    postProcess(state, state.delimiters);

    for (curr = 0; curr < max; curr++) {
      if (tokens_meta[curr] && tokens_meta[curr].delimiters) {
        postProcess(state, tokens_meta[curr].delimiters);
      }
    }
  };

  var emphasis = {
  	tokenize: tokenize,
  	postProcess: postProcess_1
  };

  var normalizeReference$1   = utils.normalizeReference;
  var isSpace$4              = utils.isSpace;


  var link$1 = function link(state, silent) {
    var attrs,
        code,
        label,
        labelEnd,
        labelStart,
        pos,
        res,
        ref,
        token,
        href = '',
        title = '',
        oldPos = state.pos,
        max = state.posMax,
        start = state.pos,
        parseReference = true;

    if (state.src.charCodeAt(state.pos) !== 0x5B/* [ */) { return false; }

    labelStart = state.pos + 1;
    labelEnd = state.md.helpers.parseLinkLabel(state, state.pos, true);

    // parser failed to find ']', so it's not a valid link
    if (labelEnd < 0) { return false; }

    pos = labelEnd + 1;
    if (pos < max && state.src.charCodeAt(pos) === 0x28/* ( */) {
      //
      // Inline link
      //

      // might have found a valid shortcut link, disable reference parsing
      parseReference = false;

      // [link](  <href>  "title"  )
      //        ^^ skipping these spaces
      pos++;
      for (; pos < max; pos++) {
        code = state.src.charCodeAt(pos);
        if (!isSpace$4(code) && code !== 0x0A) { break; }
      }
      if (pos >= max) { return false; }

      // [link](  <href>  "title"  )
      //          ^^^^^^ parsing link destination
      start = pos;
      res = state.md.helpers.parseLinkDestination(state.src, pos, state.posMax);
      if (res.ok) {
        href = state.md.normalizeLink(res.str);
        if (state.md.validateLink(href)) {
          pos = res.pos;
        } else {
          href = '';
        }

        // [link](  <href>  "title"  )
        //                ^^ skipping these spaces
        start = pos;
        for (; pos < max; pos++) {
          code = state.src.charCodeAt(pos);
          if (!isSpace$4(code) && code !== 0x0A) { break; }
        }

        // [link](  <href>  "title"  )
        //                  ^^^^^^^ parsing link title
        res = state.md.helpers.parseLinkTitle(state.src, pos, state.posMax);
        if (pos < max && start !== pos && res.ok) {
          title = res.str;
          pos = res.pos;

          // [link](  <href>  "title"  )
          //                         ^^ skipping these spaces
          for (; pos < max; pos++) {
            code = state.src.charCodeAt(pos);
            if (!isSpace$4(code) && code !== 0x0A) { break; }
          }
        }
      }

      if (pos >= max || state.src.charCodeAt(pos) !== 0x29/* ) */) {
        // parsing a valid shortcut link failed, fallback to reference
        parseReference = true;
      }
      pos++;
    }

    if (parseReference) {
      //
      // Link reference
      //
      if (typeof state.env.references === 'undefined') { return false; }

      if (pos < max && state.src.charCodeAt(pos) === 0x5B/* [ */) {
        start = pos + 1;
        pos = state.md.helpers.parseLinkLabel(state, pos);
        if (pos >= 0) {
          label = state.src.slice(start, pos++);
        } else {
          pos = labelEnd + 1;
        }
      } else {
        pos = labelEnd + 1;
      }

      // covers label === '' and label === undefined
      // (collapsed reference link and shortcut reference link respectively)
      if (!label) { label = state.src.slice(labelStart, labelEnd); }

      ref = state.env.references[normalizeReference$1(label)];
      if (!ref) {
        state.pos = oldPos;
        return false;
      }
      href = ref.href;
      title = ref.title;
    }

    //
    // We found the end of the link, and know for a fact it's a valid link;
    // so all that's left to do is to call tokenizer.
    //
    if (!silent) {
      state.pos = labelStart;
      state.posMax = labelEnd;

      token        = state.push('link_open', 'a', 1);
      token.attrs  = attrs = [ [ 'href', href ] ];
      if (title) {
        attrs.push([ 'title', title ]);
      }

      state.md.inline.tokenize(state);

      token        = state.push('link_close', 'a', -1);
    }

    state.pos = pos;
    state.posMax = max;
    return true;
  };

  var normalizeReference   = utils.normalizeReference;
  var isSpace$3              = utils.isSpace;


  var image$1 = function image(state, silent) {
    var attrs,
        code,
        content,
        label,
        labelEnd,
        labelStart,
        pos,
        ref,
        res,
        title,
        token,
        tokens,
        start,
        href = '',
        oldPos = state.pos,
        max = state.posMax;

    if (state.src.charCodeAt(state.pos) !== 0x21/* ! */) { return false; }
    if (state.src.charCodeAt(state.pos + 1) !== 0x5B/* [ */) { return false; }

    labelStart = state.pos + 2;
    labelEnd = state.md.helpers.parseLinkLabel(state, state.pos + 1, false);

    // parser failed to find ']', so it's not a valid link
    if (labelEnd < 0) { return false; }

    pos = labelEnd + 1;
    if (pos < max && state.src.charCodeAt(pos) === 0x28/* ( */) {
      //
      // Inline link
      //

      // [link](  <href>  "title"  )
      //        ^^ skipping these spaces
      pos++;
      for (; pos < max; pos++) {
        code = state.src.charCodeAt(pos);
        if (!isSpace$3(code) && code !== 0x0A) { break; }
      }
      if (pos >= max) { return false; }

      // [link](  <href>  "title"  )
      //          ^^^^^^ parsing link destination
      start = pos;
      res = state.md.helpers.parseLinkDestination(state.src, pos, state.posMax);
      if (res.ok) {
        href = state.md.normalizeLink(res.str);
        if (state.md.validateLink(href)) {
          pos = res.pos;
        } else {
          href = '';
        }
      }

      // [link](  <href>  "title"  )
      //                ^^ skipping these spaces
      start = pos;
      for (; pos < max; pos++) {
        code = state.src.charCodeAt(pos);
        if (!isSpace$3(code) && code !== 0x0A) { break; }
      }

      // [link](  <href>  "title"  )
      //                  ^^^^^^^ parsing link title
      res = state.md.helpers.parseLinkTitle(state.src, pos, state.posMax);
      if (pos < max && start !== pos && res.ok) {
        title = res.str;
        pos = res.pos;

        // [link](  <href>  "title"  )
        //                         ^^ skipping these spaces
        for (; pos < max; pos++) {
          code = state.src.charCodeAt(pos);
          if (!isSpace$3(code) && code !== 0x0A) { break; }
        }
      } else {
        title = '';
      }

      if (pos >= max || state.src.charCodeAt(pos) !== 0x29/* ) */) {
        state.pos = oldPos;
        return false;
      }
      pos++;
    } else {
      //
      // Link reference
      //
      if (typeof state.env.references === 'undefined') { return false; }

      if (pos < max && state.src.charCodeAt(pos) === 0x5B/* [ */) {
        start = pos + 1;
        pos = state.md.helpers.parseLinkLabel(state, pos);
        if (pos >= 0) {
          label = state.src.slice(start, pos++);
        } else {
          pos = labelEnd + 1;
        }
      } else {
        pos = labelEnd + 1;
      }

      // covers label === '' and label === undefined
      // (collapsed reference link and shortcut reference link respectively)
      if (!label) { label = state.src.slice(labelStart, labelEnd); }

      ref = state.env.references[normalizeReference(label)];
      if (!ref) {
        state.pos = oldPos;
        return false;
      }
      href = ref.href;
      title = ref.title;
    }

    //
    // We found the end of the link, and know for a fact it's a valid link;
    // so all that's left to do is to call tokenizer.
    //
    if (!silent) {
      content = state.src.slice(labelStart, labelEnd);

      state.md.inline.parse(
        content,
        state.md,
        state.env,
        tokens = []
      );

      token          = state.push('image', 'img', 0);
      token.attrs    = attrs = [ [ 'src', href ], [ 'alt', '' ] ];
      token.children = tokens;
      token.content  = content;

      if (title) {
        attrs.push([ 'title', title ]);
      }
    }

    state.pos = pos;
    state.posMax = max;
    return true;
  };

  // Process autolinks '<protocol:...>'


  /*eslint max-len:0*/
  var EMAIL_RE    = /^([a-zA-Z0-9.!#$%&'*+\/=?^_`{|}~-]+@[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?(?:\.[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)*)$/;
  var AUTOLINK_RE = /^([a-zA-Z][a-zA-Z0-9+.\-]{1,31}):([^<>\x00-\x20]*)$/;


  var autolink = function autolink(state, silent) {
    var url, fullUrl, token, ch, start, max,
        pos = state.pos;

    if (state.src.charCodeAt(pos) !== 0x3C/* < */) { return false; }

    start = state.pos;
    max = state.posMax;

    for (;;) {
      if (++pos >= max) { return false; }

      ch = state.src.charCodeAt(pos);

      if (ch === 0x3C /* < */) { return false; }
      if (ch === 0x3E /* > */) { break; }
    }

    url = state.src.slice(start + 1, pos);

    if (AUTOLINK_RE.test(url)) {
      fullUrl = state.md.normalizeLink(url);
      if (!state.md.validateLink(fullUrl)) { return false; }

      if (!silent) {
        token         = state.push('link_open', 'a', 1);
        token.attrs   = [ [ 'href', fullUrl ] ];
        token.markup  = 'autolink';
        token.info    = 'auto';

        token         = state.push('text', '', 0);
        token.content = state.md.normalizeLinkText(url);

        token         = state.push('link_close', 'a', -1);
        token.markup  = 'autolink';
        token.info    = 'auto';
      }

      state.pos += url.length + 2;
      return true;
    }

    if (EMAIL_RE.test(url)) {
      fullUrl = state.md.normalizeLink('mailto:' + url);
      if (!state.md.validateLink(fullUrl)) { return false; }

      if (!silent) {
        token         = state.push('link_open', 'a', 1);
        token.attrs   = [ [ 'href', fullUrl ] ];
        token.markup  = 'autolink';
        token.info    = 'auto';

        token         = state.push('text', '', 0);
        token.content = state.md.normalizeLinkText(url);

        token         = state.push('link_close', 'a', -1);
        token.markup  = 'autolink';
        token.info    = 'auto';
      }

      state.pos += url.length + 2;
      return true;
    }

    return false;
  };

  var HTML_TAG_RE = html_re.HTML_TAG_RE;


  function isLetter(ch) {
    /*eslint no-bitwise:0*/
    var lc = ch | 0x20; // to lower case
    return (lc >= 0x61/* a */) && (lc <= 0x7a/* z */);
  }


  var html_inline = function html_inline(state, silent) {
    var ch, match, max, token,
        pos = state.pos;

    if (!state.md.options.html) { return false; }

    // Check start
    max = state.posMax;
    if (state.src.charCodeAt(pos) !== 0x3C/* < */ ||
        pos + 2 >= max) {
      return false;
    }

    // Quick fail on second char
    ch = state.src.charCodeAt(pos + 1);
    if (ch !== 0x21/* ! */ &&
        ch !== 0x3F/* ? */ &&
        ch !== 0x2F/* / */ &&
        !isLetter(ch)) {
      return false;
    }

    match = state.src.slice(pos).match(HTML_TAG_RE);
    if (!match) { return false; }

    if (!silent) {
      token         = state.push('html_inline', '', 0);
      token.content = state.src.slice(pos, pos + match[0].length);
    }
    state.pos += match[0].length;
    return true;
  };

  var has               = utils.has;
  var isValidEntityCode = utils.isValidEntityCode;
  var fromCodePoint     = utils.fromCodePoint;


  var DIGITAL_RE = /^&#((?:x[a-f0-9]{1,6}|[0-9]{1,7}));/i;
  var NAMED_RE   = /^&([a-z][a-z0-9]{1,31});/i;


  var entity = function entity(state, silent) {
    var ch, code, match, pos = state.pos, max = state.posMax;

    if (state.src.charCodeAt(pos) !== 0x26/* & */) { return false; }

    if (pos + 1 < max) {
      ch = state.src.charCodeAt(pos + 1);

      if (ch === 0x23 /* # */) {
        match = state.src.slice(pos).match(DIGITAL_RE);
        if (match) {
          if (!silent) {
            code = match[1][0].toLowerCase() === 'x' ? parseInt(match[1].slice(1), 16) : parseInt(match[1], 10);
            state.pending += isValidEntityCode(code) ? fromCodePoint(code) : fromCodePoint(0xFFFD);
          }
          state.pos += match[0].length;
          return true;
        }
      } else {
        match = state.src.slice(pos).match(NAMED_RE);
        if (match) {
          if (has(entities, match[1])) {
            if (!silent) { state.pending += entities[match[1]]; }
            state.pos += match[0].length;
            return true;
          }
        }
      }
    }

    if (!silent) { state.pending += '&'; }
    state.pos++;
    return true;
  };

  // For each opening emphasis-like marker find a matching closing one


  function processDelimiters(state, delimiters) {
    var closerIdx, openerIdx, closer, opener, minOpenerIdx, newMinOpenerIdx,
        isOddMatch, lastJump,
        openersBottom = {},
        max = delimiters.length;

    for (closerIdx = 0; closerIdx < max; closerIdx++) {
      closer = delimiters[closerIdx];

      // Length is only used for emphasis-specific "rule of 3",
      // if it's not defined (in strikethrough or 3rd party plugins),
      // we can default it to 0 to disable those checks.
      //
      closer.length = closer.length || 0;

      if (!closer.close) { continue; }

      // Previously calculated lower bounds (previous fails)
      // for each marker, each delimiter length modulo 3,
      // and for whether this closer can be an opener;
      // https://github.com/commonmark/cmark/commit/34250e12ccebdc6372b8b49c44fab57c72443460
      if (!openersBottom.hasOwnProperty(closer.marker)) {
        openersBottom[closer.marker] = [ -1, -1, -1, -1, -1, -1 ];
      }

      minOpenerIdx = openersBottom[closer.marker][(closer.open ? 3 : 0) + (closer.length % 3)];

      openerIdx = closerIdx - closer.jump - 1;

      // avoid crash if `closer.jump` is pointing outside of the array, see #742
      if (openerIdx < -1) { openerIdx = -1; }

      newMinOpenerIdx = openerIdx;

      for (; openerIdx > minOpenerIdx; openerIdx -= opener.jump + 1) {
        opener = delimiters[openerIdx];

        if (opener.marker !== closer.marker) { continue; }

        if (opener.open && opener.end < 0) {

          isOddMatch = false;

          // from spec:
          //
          // If one of the delimiters can both open and close emphasis, then the
          // sum of the lengths of the delimiter runs containing the opening and
          // closing delimiters must not be a multiple of 3 unless both lengths
          // are multiples of 3.
          //
          if (opener.close || closer.open) {
            if ((opener.length + closer.length) % 3 === 0) {
              if (opener.length % 3 !== 0 || closer.length % 3 !== 0) {
                isOddMatch = true;
              }
            }
          }

          if (!isOddMatch) {
            // If previous delimiter cannot be an opener, we can safely skip
            // the entire sequence in future checks. This is required to make
            // sure algorithm has linear complexity (see *_*_*_*_*_... case).
            //
            lastJump = openerIdx > 0 && !delimiters[openerIdx - 1].open ?
              delimiters[openerIdx - 1].jump + 1 :
              0;

            closer.jump  = closerIdx - openerIdx + lastJump;
            closer.open  = false;
            opener.end   = closerIdx;
            opener.jump  = lastJump;
            opener.close = false;
            newMinOpenerIdx = -1;
            break;
          }
        }
      }

      if (newMinOpenerIdx !== -1) {
        // If match for this delimiter run failed, we want to set lower bound for
        // future lookups. This is required to make sure algorithm has linear
        // complexity.
        //
        // See details here:
        // https://github.com/commonmark/cmark/issues/178#issuecomment-270417442
        //
        openersBottom[closer.marker][(closer.open ? 3 : 0) + ((closer.length || 0) % 3)] = newMinOpenerIdx;
      }
    }
  }


  var balance_pairs = function link_pairs(state) {
    var curr,
        tokens_meta = state.tokens_meta,
        max = state.tokens_meta.length;

    processDelimiters(state, state.delimiters);

    for (curr = 0; curr < max; curr++) {
      if (tokens_meta[curr] && tokens_meta[curr].delimiters) {
        processDelimiters(state, tokens_meta[curr].delimiters);
      }
    }
  };

  // Clean up tokens after emphasis and strikethrough postprocessing:


  var text_collapse = function text_collapse(state) {
    var curr, last,
        level = 0,
        tokens = state.tokens,
        max = state.tokens.length;

    for (curr = last = 0; curr < max; curr++) {
      // re-calculate levels after emphasis/strikethrough turns some text nodes
      // into opening/closing tags
      if (tokens[curr].nesting < 0) { level--; } // closing tag
      tokens[curr].level = level;
      if (tokens[curr].nesting > 0) { level++; } // opening tag

      if (tokens[curr].type === 'text' &&
          curr + 1 < max &&
          tokens[curr + 1].type === 'text') {

        // collapse two adjacent text nodes
        tokens[curr + 1].content = tokens[curr].content + tokens[curr + 1].content;
      } else {
        if (curr !== last) { tokens[last] = tokens[curr]; }

        last++;
      }
    }

    if (curr !== last) {
      tokens.length = last;
    }
  };

  var isWhiteSpace   = utils.isWhiteSpace;
  var isPunctChar    = utils.isPunctChar;
  var isMdAsciiPunct = utils.isMdAsciiPunct;


  function StateInline(src, md, env, outTokens) {
    this.src = src;
    this.env = env;
    this.md = md;
    this.tokens = outTokens;
    this.tokens_meta = Array(outTokens.length);

    this.pos = 0;
    this.posMax = this.src.length;
    this.level = 0;
    this.pending = '';
    this.pendingLevel = 0;

    // Stores { start: end } pairs. Useful for backtrack
    // optimization of pairs parse (emphasis, strikes).
    this.cache = {};

    // List of emphasis-like delimiters for current tag
    this.delimiters = [];

    // Stack of delimiter lists for upper level tags
    this._prev_delimiters = [];

    // backtick length => last seen position
    this.backticks = {};
    this.backticksScanned = false;
  }


  // Flush pending text
  //
  StateInline.prototype.pushPending = function () {
    var token$1 = new token('text', '', 0);
    token$1.content = this.pending;
    token$1.level = this.pendingLevel;
    this.tokens.push(token$1);
    this.pending = '';
    return token$1;
  };


  // Push new token to "stream".
  // If pending text exists - flush it as text token
  //
  StateInline.prototype.push = function (type, tag, nesting) {
    if (this.pending) {
      this.pushPending();
    }

    var token$1 = new token(type, tag, nesting);
    var token_meta = null;

    if (nesting < 0) {
      // closing tag
      this.level--;
      this.delimiters = this._prev_delimiters.pop();
    }

    token$1.level = this.level;

    if (nesting > 0) {
      // opening tag
      this.level++;
      this._prev_delimiters.push(this.delimiters);
      this.delimiters = [];
      token_meta = { delimiters: this.delimiters };
    }

    this.pendingLevel = this.level;
    this.tokens.push(token$1);
    this.tokens_meta.push(token_meta);
    return token$1;
  };


  // Scan a sequence of emphasis-like markers, and determine whether
  // it can start an emphasis sequence or end an emphasis sequence.
  //
  //  - start - position to scan from (it should point at a valid marker);
  //  - canSplitWord - determine if these markers can be found inside a word
  //
  StateInline.prototype.scanDelims = function (start, canSplitWord) {
    var pos = start, lastChar, nextChar, count, can_open, can_close,
        isLastWhiteSpace, isLastPunctChar,
        isNextWhiteSpace, isNextPunctChar,
        left_flanking = true,
        right_flanking = true,
        max = this.posMax,
        marker = this.src.charCodeAt(start);

    // treat beginning of the line as a whitespace
    lastChar = start > 0 ? this.src.charCodeAt(start - 1) : 0x20;

    while (pos < max && this.src.charCodeAt(pos) === marker) { pos++; }

    count = pos - start;

    // treat end of the line as a whitespace
    nextChar = pos < max ? this.src.charCodeAt(pos) : 0x20;

    isLastPunctChar = isMdAsciiPunct(lastChar) || isPunctChar(String.fromCharCode(lastChar));
    isNextPunctChar = isMdAsciiPunct(nextChar) || isPunctChar(String.fromCharCode(nextChar));

    isLastWhiteSpace = isWhiteSpace(lastChar);
    isNextWhiteSpace = isWhiteSpace(nextChar);

    if (isNextWhiteSpace) {
      left_flanking = false;
    } else if (isNextPunctChar) {
      if (!(isLastWhiteSpace || isLastPunctChar)) {
        left_flanking = false;
      }
    }

    if (isLastWhiteSpace) {
      right_flanking = false;
    } else if (isLastPunctChar) {
      if (!(isNextWhiteSpace || isNextPunctChar)) {
        right_flanking = false;
      }
    }

    if (!canSplitWord) {
      can_open  = left_flanking  && (!right_flanking || isLastPunctChar);
      can_close = right_flanking && (!left_flanking  || isNextPunctChar);
    } else {
      can_open  = left_flanking;
      can_close = right_flanking;
    }

    return {
      can_open:  can_open,
      can_close: can_close,
      length:    count
    };
  };


  // re-export Token class to use in block rules
  StateInline.prototype.Token = token;


  var state_inline = StateInline;

  ////////////////////////////////////////////////////////////////////////////////
  // Parser rules

  var _rules = [
    [ 'text',            text$1 ],
    [ 'newline',         newline ],
    [ 'escape',          _escape ],
    [ 'backticks',       backticks ],
    [ 'strikethrough',   strikethrough$1.tokenize ],
    [ 'emphasis',        emphasis.tokenize ],
    [ 'link',            link$1 ],
    [ 'image',           image$1 ],
    [ 'autolink',        autolink ],
    [ 'html_inline',     html_inline ],
    [ 'entity',          entity ]
  ];

  var _rules2 = [
    [ 'balance_pairs',   balance_pairs ],
    [ 'strikethrough',   strikethrough$1.postProcess ],
    [ 'emphasis',        emphasis.postProcess ],
    [ 'text_collapse',   text_collapse ]
  ];


  /**
   * new ParserInline()
   **/
  function ParserInline() {
    var i;

    /**
     * ParserInline#ruler -> Ruler
     *
     * [[Ruler]] instance. Keep configuration of inline rules.
     **/
    this.ruler = new ruler();

    for (i = 0; i < _rules.length; i++) {
      this.ruler.push(_rules[i][0], _rules[i][1]);
    }

    /**
     * ParserInline#ruler2 -> Ruler
     *
     * [[Ruler]] instance. Second ruler used for post-processing
     * (e.g. in emphasis-like rules).
     **/
    this.ruler2 = new ruler();

    for (i = 0; i < _rules2.length; i++) {
      this.ruler2.push(_rules2[i][0], _rules2[i][1]);
    }
  }


  // Skip single token by running all rules in validation mode;
  // returns `true` if any rule reported success
  //
  ParserInline.prototype.skipToken = function (state) {
    var ok, i, pos = state.pos,
        rules = this.ruler.getRules(''),
        len = rules.length,
        maxNesting = state.md.options.maxNesting,
        cache = state.cache;


    if (typeof cache[pos] !== 'undefined') {
      state.pos = cache[pos];
      return;
    }

    if (state.level < maxNesting) {
      for (i = 0; i < len; i++) {
        // Increment state.level and decrement it later to limit recursion.
        // It's harmless to do here, because no tokens are created. But ideally,
        // we'd need a separate private state variable for this purpose.
        //
        state.level++;
        ok = rules[i](state, true);
        state.level--;

        if (ok) { break; }
      }
    } else {
      // Too much nesting, just skip until the end of the paragraph.
      //
      // NOTE: this will cause links to behave incorrectly in the following case,
      //       when an amount of `[` is exactly equal to `maxNesting + 1`:
      //
      //       [[[[[[[[[[[[[[[[[[[[[foo]()
      //
      // TODO: remove this workaround when CM standard will allow nested links
      //       (we can replace it by preventing links from being parsed in
      //       validation mode)
      //
      state.pos = state.posMax;
    }

    if (!ok) { state.pos++; }
    cache[pos] = state.pos;
  };


  // Generate tokens for input range
  //
  ParserInline.prototype.tokenize = function (state) {
    var ok, i,
        rules = this.ruler.getRules(''),
        len = rules.length,
        end = state.posMax,
        maxNesting = state.md.options.maxNesting;

    while (state.pos < end) {
      // Try all possible rules.
      // On success, rule should:
      //
      // - update `state.pos`
      // - update `state.tokens`
      // - return true

      if (state.level < maxNesting) {
        for (i = 0; i < len; i++) {
          ok = rules[i](state, false);
          if (ok) { break; }
        }
      }

      if (ok) {
        if (state.pos >= end) { break; }
        continue;
      }

      state.pending += state.src[state.pos++];
    }

    if (state.pending) {
      state.pushPending();
    }
  };


  /**
   * ParserInline.parse(str, md, env, outTokens)
   *
   * Process input string and push inline tokens into `outTokens`
   **/
  ParserInline.prototype.parse = function (str, md, env, outTokens) {
    var i, rules, len;
    var state = new this.State(str, md, env, outTokens);

    this.tokenize(state);

    rules = this.ruler2.getRules('');
    len = rules.length;

    for (i = 0; i < len; i++) {
      rules[i](state);
    }
  };


  ParserInline.prototype.State = state_inline;


  var parser_inline = ParserInline;

  var re = function (opts) {
    var re = {};

    // Use direct extract instead of `regenerate` to reduse browserified size
    re.src_Any = regex$3.source;
    re.src_Cc  = regex$2.source;
    re.src_Z   = regex.source;
    re.src_P   = regex$4.source;

    // \p{\Z\P\Cc\CF} (white spaces + control + format + punctuation)
    re.src_ZPCc = [ re.src_Z, re.src_P, re.src_Cc ].join('|');

    // \p{\Z\Cc} (white spaces + control)
    re.src_ZCc = [ re.src_Z, re.src_Cc ].join('|');

    // Experimental. List of chars, completely prohibited in links
    // because can separate it from other part of text
    var text_separators = '[><\uff5c]';

    // All possible word characters (everything without punctuation, spaces & controls)
    // Defined via punctuation & spaces to save space
    // Should be something like \p{\L\N\S\M} (\w but without `_`)
    re.src_pseudo_letter       = '(?:(?!' + text_separators + '|' + re.src_ZPCc + ')' + re.src_Any + ')';
    // The same as abothe but without [0-9]
    // var src_pseudo_letter_non_d = '(?:(?![0-9]|' + src_ZPCc + ')' + src_Any + ')';

    ////////////////////////////////////////////////////////////////////////////////

    re.src_ip4 =

      '(?:(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)';

    // Prohibit any of "@/[]()" in user/pass to avoid wrong domain fetch.
    re.src_auth    = '(?:(?:(?!' + re.src_ZCc + '|[@/\\[\\]()]).)+@)?';

    re.src_port =

      '(?::(?:6(?:[0-4]\\d{3}|5(?:[0-4]\\d{2}|5(?:[0-2]\\d|3[0-5])))|[1-5]?\\d{1,4}))?';

    re.src_host_terminator =

      '(?=$|' + text_separators + '|' + re.src_ZPCc + ')(?!-|_|:\\d|\\.-|\\.(?!$|' + re.src_ZPCc + '))';

    re.src_path =

      '(?:' +
        '[/?#]' +
          '(?:' +
            '(?!' + re.src_ZCc + '|' + text_separators + '|[()[\\]{}.,"\'?!\\-]).|' +
            '\\[(?:(?!' + re.src_ZCc + '|\\]).)*\\]|' +
            '\\((?:(?!' + re.src_ZCc + '|[)]).)*\\)|' +
            '\\{(?:(?!' + re.src_ZCc + '|[}]).)*\\}|' +
            '\\"(?:(?!' + re.src_ZCc + '|["]).)+\\"|' +
            "\\'(?:(?!" + re.src_ZCc + "|[']).)+\\'|" +
            "\\'(?=" + re.src_pseudo_letter + '|[-]).|' +  // allow `I'm_king` if no pair found
            '\\.{2,}[a-zA-Z0-9%/&]|' + // google has many dots in "google search" links (#66, #81).
                                       // github has ... in commit range links,
                                       // Restrict to
                                       // - english
                                       // - percent-encoded
                                       // - parts of file path
                                       // - params separator
                                       // until more examples found.
            '\\.(?!' + re.src_ZCc + '|[.]).|' +
            (opts && opts['---'] ?
              '\\-(?!--(?:[^-]|$))(?:-*)|' // `---` => long dash, terminate
              :
              '\\-+|'
            ) +
            '\\,(?!' + re.src_ZCc + ').|' +       // allow `,,,` in paths
            '\\!+(?!' + re.src_ZCc + '|[!]).|' +  // allow `!!!` in paths, but not at the end
            '\\?(?!' + re.src_ZCc + '|[?]).' +
          ')+' +
        '|\\/' +
      ')?';

    // Allow anything in markdown spec, forbid quote (") at the first position
    // because emails enclosed in quotes are far more common
    re.src_email_name =

      '[\\-;:&=\\+\\$,\\.a-zA-Z0-9_][\\-;:&=\\+\\$,\\"\\.a-zA-Z0-9_]*';

    re.src_xn =

      'xn--[a-z0-9\\-]{1,59}';

    // More to read about domain names
    // http://serverfault.com/questions/638260/

    re.src_domain_root =

      // Allow letters & digits (http://test1)
      '(?:' +
        re.src_xn +
        '|' +
        re.src_pseudo_letter + '{1,63}' +
      ')';

    re.src_domain =

      '(?:' +
        re.src_xn +
        '|' +
        '(?:' + re.src_pseudo_letter + ')' +
        '|' +
        '(?:' + re.src_pseudo_letter + '(?:-|' + re.src_pseudo_letter + '){0,61}' + re.src_pseudo_letter + ')' +
      ')';

    re.src_host =

      '(?:' +
      // Don't need IP check, because digits are already allowed in normal domain names
      //   src_ip4 +
      // '|' +
        '(?:(?:(?:' + re.src_domain + ')\\.)*' + re.src_domain/*_root*/ + ')' +
      ')';

    re.tpl_host_fuzzy =

      '(?:' +
        re.src_ip4 +
      '|' +
        '(?:(?:(?:' + re.src_domain + ')\\.)+(?:%TLDS%))' +
      ')';

    re.tpl_host_no_ip_fuzzy =

      '(?:(?:(?:' + re.src_domain + ')\\.)+(?:%TLDS%))';

    re.src_host_strict =

      re.src_host + re.src_host_terminator;

    re.tpl_host_fuzzy_strict =

      re.tpl_host_fuzzy + re.src_host_terminator;

    re.src_host_port_strict =

      re.src_host + re.src_port + re.src_host_terminator;

    re.tpl_host_port_fuzzy_strict =

      re.tpl_host_fuzzy + re.src_port + re.src_host_terminator;

    re.tpl_host_port_no_ip_fuzzy_strict =

      re.tpl_host_no_ip_fuzzy + re.src_port + re.src_host_terminator;


    ////////////////////////////////////////////////////////////////////////////////
    // Main rules

    // Rude test fuzzy links by host, for quick deny
    re.tpl_host_fuzzy_test =

      'localhost|www\\.|\\.\\d{1,3}\\.|(?:\\.(?:%TLDS%)(?:' + re.src_ZPCc + '|>|$))';

    re.tpl_email_fuzzy =

        '(^|' + text_separators + '|"|\\(|' + re.src_ZCc + ')' +
        '(' + re.src_email_name + '@' + re.tpl_host_fuzzy_strict + ')';

    re.tpl_link_fuzzy =
        // Fuzzy link can't be prepended with .:/\- and non punctuation.
        // but can start with > (markdown blockquote)
        '(^|(?![.:/\\-_@])(?:[$+<=>^`|\uff5c]|' + re.src_ZPCc + '))' +
        '((?![$+<=>^`|\uff5c])' + re.tpl_host_port_fuzzy_strict + re.src_path + ')';

    re.tpl_link_no_ip_fuzzy =
        // Fuzzy link can't be prepended with .:/\- and non punctuation.
        // but can start with > (markdown blockquote)
        '(^|(?![.:/\\-_@])(?:[$+<=>^`|\uff5c]|' + re.src_ZPCc + '))' +
        '((?![$+<=>^`|\uff5c])' + re.tpl_host_port_no_ip_fuzzy_strict + re.src_path + ')';

    return re;
  };

  ////////////////////////////////////////////////////////////////////////////////
  // Helpers

  // Merge objects
  //
  function assign(obj /*from1, from2, from3, ...*/) {
    var sources = Array.prototype.slice.call(arguments, 1);

    sources.forEach(function (source) {
      if (!source) { return; }

      Object.keys(source).forEach(function (key) {
        obj[key] = source[key];
      });
    });

    return obj;
  }

  function _class(obj) { return Object.prototype.toString.call(obj); }
  function isString(obj) { return _class(obj) === '[object String]'; }
  function isObject(obj) { return _class(obj) === '[object Object]'; }
  function isRegExp(obj) { return _class(obj) === '[object RegExp]'; }
  function isFunction(obj) { return _class(obj) === '[object Function]'; }


  function escapeRE(str) { return str.replace(/[.?*+^$[\]\\(){}|-]/g, '\\$&'); }

  ////////////////////////////////////////////////////////////////////////////////


  var defaultOptions = {
    fuzzyLink: true,
    fuzzyEmail: true,
    fuzzyIP: false
  };


  function isOptionsObj(obj) {
    return Object.keys(obj || {}).reduce(function (acc, k) {
      return acc || defaultOptions.hasOwnProperty(k);
    }, false);
  }


  var defaultSchemas = {
    'http:': {
      validate: function (text, pos, self) {
        var tail = text.slice(pos);

        if (!self.re.http) {
          // compile lazily, because "host"-containing variables can change on tlds update.
          self.re.http =  new RegExp(
            '^\\/\\/' + self.re.src_auth + self.re.src_host_port_strict + self.re.src_path, 'i'
          );
        }
        if (self.re.http.test(tail)) {
          return tail.match(self.re.http)[0].length;
        }
        return 0;
      }
    },
    'https:':  'http:',
    'ftp:':    'http:',
    '//':      {
      validate: function (text, pos, self) {
        var tail = text.slice(pos);

        if (!self.re.no_http) {
        // compile lazily, because "host"-containing variables can change on tlds update.
          self.re.no_http =  new RegExp(
            '^' +
            self.re.src_auth +
            // Don't allow single-level domains, because of false positives like '//test'
            // with code comments
            '(?:localhost|(?:(?:' + self.re.src_domain + ')\\.)+' + self.re.src_domain_root + ')' +
            self.re.src_port +
            self.re.src_host_terminator +
            self.re.src_path,

            'i'
          );
        }

        if (self.re.no_http.test(tail)) {
          // should not be `://` & `///`, that protects from errors in protocol name
          if (pos >= 3 && text[pos - 3] === ':') { return 0; }
          if (pos >= 3 && text[pos - 3] === '/') { return 0; }
          return tail.match(self.re.no_http)[0].length;
        }
        return 0;
      }
    },
    'mailto:': {
      validate: function (text, pos, self) {
        var tail = text.slice(pos);

        if (!self.re.mailto) {
          self.re.mailto =  new RegExp(
            '^' + self.re.src_email_name + '@' + self.re.src_host_strict, 'i'
          );
        }
        if (self.re.mailto.test(tail)) {
          return tail.match(self.re.mailto)[0].length;
        }
        return 0;
      }
    }
  };

  /*eslint-disable max-len*/

  // RE pattern for 2-character tlds (autogenerated by ./support/tlds_2char_gen.js)
  var tlds_2ch_src_re = 'a[cdefgilmnoqrstuwxz]|b[abdefghijmnorstvwyz]|c[acdfghiklmnoruvwxyz]|d[ejkmoz]|e[cegrstu]|f[ijkmor]|g[abdefghilmnpqrstuwy]|h[kmnrtu]|i[delmnoqrst]|j[emop]|k[eghimnprwyz]|l[abcikrstuvy]|m[acdeghklmnopqrstuvwxyz]|n[acefgilopruz]|om|p[aefghklmnrstwy]|qa|r[eosuw]|s[abcdeghijklmnortuvxyz]|t[cdfghjklmnortvwz]|u[agksyz]|v[aceginu]|w[fs]|y[et]|z[amw]';

  // DON'T try to make PRs with changes. Extend TLDs with LinkifyIt.tlds() instead
  var tlds_default = 'biz|com|edu|gov|net|org|pro|web|xxx|aero|asia|coop|info|museum|name|shop|рф'.split('|');

  /*eslint-enable max-len*/

  ////////////////////////////////////////////////////////////////////////////////

  function resetScanCache(self) {
    self.__index__ = -1;
    self.__text_cache__   = '';
  }

  function createValidator(re) {
    return function (text, pos) {
      var tail = text.slice(pos);

      if (re.test(tail)) {
        return tail.match(re)[0].length;
      }
      return 0;
    };
  }

  function createNormalizer() {
    return function (match, self) {
      self.normalize(match);
    };
  }

  // Schemas compiler. Build regexps.
  //
  function compile(self) {

    // Load & clone RE patterns.
    var re$1 = self.re = re(self.__opts__);

    // Define dynamic patterns
    var tlds = self.__tlds__.slice();

    self.onCompile();

    if (!self.__tlds_replaced__) {
      tlds.push(tlds_2ch_src_re);
    }
    tlds.push(re$1.src_xn);

    re$1.src_tlds = tlds.join('|');

    function untpl(tpl) { return tpl.replace('%TLDS%', re$1.src_tlds); }

    re$1.email_fuzzy      = RegExp(untpl(re$1.tpl_email_fuzzy), 'i');
    re$1.link_fuzzy       = RegExp(untpl(re$1.tpl_link_fuzzy), 'i');
    re$1.link_no_ip_fuzzy = RegExp(untpl(re$1.tpl_link_no_ip_fuzzy), 'i');
    re$1.host_fuzzy_test  = RegExp(untpl(re$1.tpl_host_fuzzy_test), 'i');

    //
    // Compile each schema
    //

    var aliases = [];

    self.__compiled__ = {}; // Reset compiled data

    function schemaError(name, val) {
      throw new Error('(LinkifyIt) Invalid schema "' + name + '": ' + val);
    }

    Object.keys(self.__schemas__).forEach(function (name) {
      var val = self.__schemas__[name];

      // skip disabled methods
      if (val === null) { return; }

      var compiled = { validate: null, link: null };

      self.__compiled__[name] = compiled;

      if (isObject(val)) {
        if (isRegExp(val.validate)) {
          compiled.validate = createValidator(val.validate);
        } else if (isFunction(val.validate)) {
          compiled.validate = val.validate;
        } else {
          schemaError(name, val);
        }

        if (isFunction(val.normalize)) {
          compiled.normalize = val.normalize;
        } else if (!val.normalize) {
          compiled.normalize = createNormalizer();
        } else {
          schemaError(name, val);
        }

        return;
      }

      if (isString(val)) {
        aliases.push(name);
        return;
      }

      schemaError(name, val);
    });

    //
    // Compile postponed aliases
    //

    aliases.forEach(function (alias) {
      if (!self.__compiled__[self.__schemas__[alias]]) {
        // Silently fail on missed schemas to avoid errons on disable.
        // schemaError(alias, self.__schemas__[alias]);
        return;
      }

      self.__compiled__[alias].validate =
        self.__compiled__[self.__schemas__[alias]].validate;
      self.__compiled__[alias].normalize =
        self.__compiled__[self.__schemas__[alias]].normalize;
    });

    //
    // Fake record for guessed links
    //
    self.__compiled__[''] = { validate: null, normalize: createNormalizer() };

    //
    // Build schema condition
    //
    var slist = Object.keys(self.__compiled__)
                        .filter(function (name) {
                          // Filter disabled & fake schemas
                          return name.length > 0 && self.__compiled__[name];
                        })
                        .map(escapeRE)
                        .join('|');
    // (?!_) cause 1.5x slowdown
    self.re.schema_test   = RegExp('(^|(?!_)(?:[><\uff5c]|' + re$1.src_ZPCc + '))(' + slist + ')', 'i');
    self.re.schema_search = RegExp('(^|(?!_)(?:[><\uff5c]|' + re$1.src_ZPCc + '))(' + slist + ')', 'ig');

    self.re.pretest = RegExp(
      '(' + self.re.schema_test.source + ')|(' + self.re.host_fuzzy_test.source + ')|@',
      'i'
    );

    //
    // Cleanup
    //

    resetScanCache(self);
  }

  /**
   * class Match
   *
   * Match result. Single element of array, returned by [[LinkifyIt#match]]
   **/
  function Match(self, shift) {
    var start = self.__index__,
        end   = self.__last_index__,
        text  = self.__text_cache__.slice(start, end);

    /**
     * Match#schema -> String
     *
     * Prefix (protocol) for matched string.
     **/
    this.schema    = self.__schema__.toLowerCase();
    /**
     * Match#index -> Number
     *
     * First position of matched string.
     **/
    this.index     = start + shift;
    /**
     * Match#lastIndex -> Number
     *
     * Next position after matched string.
     **/
    this.lastIndex = end + shift;
    /**
     * Match#raw -> String
     *
     * Matched string.
     **/
    this.raw       = text;
    /**
     * Match#text -> String
     *
     * Notmalized text of matched string.
     **/
    this.text      = text;
    /**
     * Match#url -> String
     *
     * Normalized url of matched string.
     **/
    this.url       = text;
  }

  function createMatch(self, shift) {
    var match = new Match(self, shift);

    self.__compiled__[match.schema].normalize(match, self);

    return match;
  }


  /**
   * class LinkifyIt
   **/

  /**
   * new LinkifyIt(schemas, options)
   * - schemas (Object): Optional. Additional schemas to validate (prefix/validator)
   * - options (Object): { fuzzyLink|fuzzyEmail|fuzzyIP: true|false }
   *
   * Creates new linkifier instance with optional additional schemas.
   * Can be called without `new` keyword for convenience.
   *
   * By default understands:
   *
   * - `http(s)://...` , `ftp://...`, `mailto:...` & `//...` links
   * - "fuzzy" links and emails (example.com, foo@bar.com).
   *
   * `schemas` is an object, where each key/value describes protocol/rule:
   *
   * - __key__ - link prefix (usually, protocol name with `:` at the end, `skype:`
   *   for example). `linkify-it` makes shure that prefix is not preceeded with
   *   alphanumeric char and symbols. Only whitespaces and punctuation allowed.
   * - __value__ - rule to check tail after link prefix
   *   - _String_ - just alias to existing rule
   *   - _Object_
   *     - _validate_ - validator function (should return matched length on success),
   *       or `RegExp`.
   *     - _normalize_ - optional function to normalize text & url of matched result
   *       (for example, for @twitter mentions).
   *
   * `options`:
   *
   * - __fuzzyLink__ - recognige URL-s without `http(s):` prefix. Default `true`.
   * - __fuzzyIP__ - allow IPs in fuzzy links above. Can conflict with some texts
   *   like version numbers. Default `false`.
   * - __fuzzyEmail__ - recognize emails without `mailto:` prefix.
   *
   **/
  function LinkifyIt(schemas, options) {
    if (!(this instanceof LinkifyIt)) {
      return new LinkifyIt(schemas, options);
    }

    if (!options) {
      if (isOptionsObj(schemas)) {
        options = schemas;
        schemas = {};
      }
    }

    this.__opts__           = assign({}, defaultOptions, options);

    // Cache last tested result. Used to skip repeating steps on next `match` call.
    this.__index__          = -1;
    this.__last_index__     = -1; // Next scan position
    this.__schema__         = '';
    this.__text_cache__     = '';

    this.__schemas__        = assign({}, defaultSchemas, schemas);
    this.__compiled__       = {};

    this.__tlds__           = tlds_default;
    this.__tlds_replaced__  = false;

    this.re = {};

    compile(this);
  }


  /** chainable
   * LinkifyIt#add(schema, definition)
   * - schema (String): rule name (fixed pattern prefix)
   * - definition (String|RegExp|Object): schema definition
   *
   * Add new rule definition. See constructor description for details.
   **/
  LinkifyIt.prototype.add = function add(schema, definition) {
    this.__schemas__[schema] = definition;
    compile(this);
    return this;
  };


  /** chainable
   * LinkifyIt#set(options)
   * - options (Object): { fuzzyLink|fuzzyEmail|fuzzyIP: true|false }
   *
   * Set recognition options for links without schema.
   **/
  LinkifyIt.prototype.set = function set(options) {
    this.__opts__ = assign(this.__opts__, options);
    return this;
  };


  /**
   * LinkifyIt#test(text) -> Boolean
   *
   * Searches linkifiable pattern and returns `true` on success or `false` on fail.
   **/
  LinkifyIt.prototype.test = function test(text) {
    // Reset scan cache
    this.__text_cache__ = text;
    this.__index__      = -1;

    if (!text.length) { return false; }

    var m, ml, me, len, shift, next, re, tld_pos, at_pos;

    // try to scan for link with schema - that's the most simple rule
    if (this.re.schema_test.test(text)) {
      re = this.re.schema_search;
      re.lastIndex = 0;
      while ((m = re.exec(text)) !== null) {
        len = this.testSchemaAt(text, m[2], re.lastIndex);
        if (len) {
          this.__schema__     = m[2];
          this.__index__      = m.index + m[1].length;
          this.__last_index__ = m.index + m[0].length + len;
          break;
        }
      }
    }

    if (this.__opts__.fuzzyLink && this.__compiled__['http:']) {
      // guess schemaless links
      tld_pos = text.search(this.re.host_fuzzy_test);
      if (tld_pos >= 0) {
        // if tld is located after found link - no need to check fuzzy pattern
        if (this.__index__ < 0 || tld_pos < this.__index__) {
          if ((ml = text.match(this.__opts__.fuzzyIP ? this.re.link_fuzzy : this.re.link_no_ip_fuzzy)) !== null) {

            shift = ml.index + ml[1].length;

            if (this.__index__ < 0 || shift < this.__index__) {
              this.__schema__     = '';
              this.__index__      = shift;
              this.__last_index__ = ml.index + ml[0].length;
            }
          }
        }
      }
    }

    if (this.__opts__.fuzzyEmail && this.__compiled__['mailto:']) {
      // guess schemaless emails
      at_pos = text.indexOf('@');
      if (at_pos >= 0) {
        // We can't skip this check, because this cases are possible:
        // 192.168.1.1@gmail.com, my.in@example.com
        if ((me = text.match(this.re.email_fuzzy)) !== null) {

          shift = me.index + me[1].length;
          next  = me.index + me[0].length;

          if (this.__index__ < 0 || shift < this.__index__ ||
              (shift === this.__index__ && next > this.__last_index__)) {
            this.__schema__     = 'mailto:';
            this.__index__      = shift;
            this.__last_index__ = next;
          }
        }
      }
    }

    return this.__index__ >= 0;
  };


  /**
   * LinkifyIt#pretest(text) -> Boolean
   *
   * Very quick check, that can give false positives. Returns true if link MAY BE
   * can exists. Can be used for speed optimization, when you need to check that
   * link NOT exists.
   **/
  LinkifyIt.prototype.pretest = function pretest(text) {
    return this.re.pretest.test(text);
  };


  /**
   * LinkifyIt#testSchemaAt(text, name, position) -> Number
   * - text (String): text to scan
   * - name (String): rule (schema) name
   * - position (Number): text offset to check from
   *
   * Similar to [[LinkifyIt#test]] but checks only specific protocol tail exactly
   * at given position. Returns length of found pattern (0 on fail).
   **/
  LinkifyIt.prototype.testSchemaAt = function testSchemaAt(text, schema, pos) {
    // If not supported schema check requested - terminate
    if (!this.__compiled__[schema.toLowerCase()]) {
      return 0;
    }
    return this.__compiled__[schema.toLowerCase()].validate(text, pos, this);
  };


  /**
   * LinkifyIt#match(text) -> Array|null
   *
   * Returns array of found link descriptions or `null` on fail. We strongly
   * recommend to use [[LinkifyIt#test]] first, for best speed.
   *
   * ##### Result match description
   *
   * - __schema__ - link schema, can be empty for fuzzy links, or `//` for
   *   protocol-neutral  links.
   * - __index__ - offset of matched text
   * - __lastIndex__ - index of next char after mathch end
   * - __raw__ - matched text
   * - __text__ - normalized text
   * - __url__ - link, generated from matched text
   **/
  LinkifyIt.prototype.match = function match(text) {
    var shift = 0, result = [];

    // Try to take previous element from cache, if .test() called before
    if (this.__index__ >= 0 && this.__text_cache__ === text) {
      result.push(createMatch(this, shift));
      shift = this.__last_index__;
    }

    // Cut head if cache was used
    var tail = shift ? text.slice(shift) : text;

    // Scan string until end reached
    while (this.test(tail)) {
      result.push(createMatch(this, shift));

      tail = tail.slice(this.__last_index__);
      shift += this.__last_index__;
    }

    if (result.length) {
      return result;
    }

    return null;
  };


  /** chainable
   * LinkifyIt#tlds(list [, keepOld]) -> this
   * - list (Array): list of tlds
   * - keepOld (Boolean): merge with current list if `true` (`false` by default)
   *
   * Load (or merge) new tlds list. Those are user for fuzzy links (without prefix)
   * to avoid false positives. By default this algorythm used:
   *
   * - hostname with any 2-letter root zones are ok.
   * - biz|com|edu|gov|net|org|pro|web|xxx|aero|asia|coop|info|museum|name|shop|рф
   *   are ok.
   * - encoded (`xn--...`) root zones are ok.
   *
   * If list is replaced, then exact match for 2-chars root zones will be checked.
   **/
  LinkifyIt.prototype.tlds = function tlds(list, keepOld) {
    list = Array.isArray(list) ? list : [ list ];

    if (!keepOld) {
      this.__tlds__ = list.slice();
      this.__tlds_replaced__ = true;
      compile(this);
      return this;
    }

    this.__tlds__ = this.__tlds__.concat(list)
                                    .sort()
                                    .filter(function (el, idx, arr) {
                                      return el !== arr[idx - 1];
                                    })
                                    .reverse();

    compile(this);
    return this;
  };

  /**
   * LinkifyIt#normalize(match)
   *
   * Default normalizer (if schema does not define it's own).
   **/
  LinkifyIt.prototype.normalize = function normalize(match) {

    // Do minimal possible changes by default. Need to collect feedback prior
    // to move forward https://github.com/markdown-it/linkify-it/issues/1

    if (!match.schema) { match.url = 'http://' + match.url; }

    if (match.schema === 'mailto:' && !/^mailto:/i.test(match.url)) {
      match.url = 'mailto:' + match.url;
    }
  };


  /**
   * LinkifyIt#onCompile()
   *
   * Override to modify basic RegExp-s.
   **/
  LinkifyIt.prototype.onCompile = function onCompile() {
  };


  var linkifyIt = LinkifyIt;

  // markdown-it default options


  var _default = {
    options: {
      html:         false,        // Enable HTML tags in source
      xhtmlOut:     false,        // Use '/' to close single tags (<br />)
      breaks:       false,        // Convert '\n' in paragraphs into <br>
      langPrefix:   'language-',  // CSS language prefix for fenced blocks
      linkify:      false,        // autoconvert URL-like texts to links

      // Enable some language-neutral replacements + quotes beautification
      typographer:  false,

      // Double + single quotes replacement pairs, when typographer enabled,
      // and smartquotes on. Could be either a String or an Array.
      //
      // For example, you can use '«»„“' for Russian, '„“‚‘' for German,
      // and ['«\xA0', '\xA0»', '‹\xA0', '\xA0›'] for French (including nbsp).
      quotes: '\u201c\u201d\u2018\u2019', /* “”‘’ */

      // Highlighter function. Should return escaped HTML,
      // or '' if the source string is not changed and should be escaped externaly.
      // If result starts with <pre... internal wrapper is skipped.
      //
      // function (/*str, lang*/) { return ''; }
      //
      highlight: null,

      maxNesting:   100            // Internal protection, recursion limit
    },

    components: {

      core: {},
      block: {},
      inline: {}
    }
  };
  _default.options;
  _default.components;

  // "Zero" preset, with nothing enabled. Useful for manual configuring of simple


  var zero = {
    options: {
      html:         false,        // Enable HTML tags in source
      xhtmlOut:     false,        // Use '/' to close single tags (<br />)
      breaks:       false,        // Convert '\n' in paragraphs into <br>
      langPrefix:   'language-',  // CSS language prefix for fenced blocks
      linkify:      false,        // autoconvert URL-like texts to links

      // Enable some language-neutral replacements + quotes beautification
      typographer:  false,

      // Double + single quotes replacement pairs, when typographer enabled,
      // and smartquotes on. Could be either a String or an Array.
      //
      // For example, you can use '«»„“' for Russian, '„“‚‘' for German,
      // and ['«\xA0', '\xA0»', '‹\xA0', '\xA0›'] for French (including nbsp).
      quotes: '\u201c\u201d\u2018\u2019', /* “”‘’ */

      // Highlighter function. Should return escaped HTML,
      // or '' if the source string is not changed and should be escaped externaly.
      // If result starts with <pre... internal wrapper is skipped.
      //
      // function (/*str, lang*/) { return ''; }
      //
      highlight: null,

      maxNesting:   20            // Internal protection, recursion limit
    },

    components: {

      core: {
        rules: [
          'normalize',
          'block',
          'inline'
        ]
      },

      block: {
        rules: [
          'paragraph'
        ]
      },

      inline: {
        rules: [
          'text'
        ],
        rules2: [
          'balance_pairs',
          'text_collapse'
        ]
      }
    }
  };
  zero.options;
  zero.components;

  // Commonmark default options


  var commonmark = {
    options: {
      html:         true,         // Enable HTML tags in source
      xhtmlOut:     true,         // Use '/' to close single tags (<br />)
      breaks:       false,        // Convert '\n' in paragraphs into <br>
      langPrefix:   'language-',  // CSS language prefix for fenced blocks
      linkify:      false,        // autoconvert URL-like texts to links

      // Enable some language-neutral replacements + quotes beautification
      typographer:  false,

      // Double + single quotes replacement pairs, when typographer enabled,
      // and smartquotes on. Could be either a String or an Array.
      //
      // For example, you can use '«»„“' for Russian, '„“‚‘' for German,
      // and ['«\xA0', '\xA0»', '‹\xA0', '\xA0›'] for French (including nbsp).
      quotes: '\u201c\u201d\u2018\u2019', /* “”‘’ */

      // Highlighter function. Should return escaped HTML,
      // or '' if the source string is not changed and should be escaped externaly.
      // If result starts with <pre... internal wrapper is skipped.
      //
      // function (/*str, lang*/) { return ''; }
      //
      highlight: null,

      maxNesting:   20            // Internal protection, recursion limit
    },

    components: {

      core: {
        rules: [
          'normalize',
          'block',
          'inline'
        ]
      },

      block: {
        rules: [
          'blockquote',
          'code',
          'fence',
          'heading',
          'hr',
          'html_block',
          'lheading',
          'list',
          'reference',
          'paragraph'
        ]
      },

      inline: {
        rules: [
          'autolink',
          'backticks',
          'emphasis',
          'entity',
          'escape',
          'html_inline',
          'image',
          'link',
          'newline',
          'text'
        ],
        rules2: [
          'balance_pairs',
          'emphasis',
          'text_collapse'
        ]
      }
    }
  };
  commonmark.options;
  commonmark.components;

  var config = {
    default: _default,
    zero: zero,
    commonmark: commonmark
  };

  ////////////////////////////////////////////////////////////////////////////////
  //
  // This validator can prohibit more than really needed to prevent XSS. It's a
  // tradeoff to keep code simple and to be secure by default.
  //
  // If you need different setup - override validator method as you wish. Or
  // replace it with dummy function and use external sanitizer.
  //

  var BAD_PROTO_RE = /^(vbscript|javascript|file|data):/;
  var GOOD_DATA_RE = /^data:image\/(gif|png|jpeg|webp);/;

  function validateLink(url) {
    // url should be normalized at this point, and existing entities are decoded
    var str = url.trim().toLowerCase();

    return BAD_PROTO_RE.test(str) ? (GOOD_DATA_RE.test(str) ? true : false) : true;
  }

  ////////////////////////////////////////////////////////////////////////////////


  var RECODE_HOSTNAME_FOR = [ 'http:', 'https:', 'mailto:' ];

  function normalizeLink(url) {
    var parsed = mdurl.parse(url, true);

    if (parsed.hostname) {
      // Encode hostnames in urls like:
      // `http://host/`, `https://host/`, `mailto:user@host`, `//host/`
      //
      // We don't encode unknown schemas, because it's likely that we encode
      // something we shouldn't (e.g. `skype:name` treated as `skype:host`)
      //
      if (!parsed.protocol || RECODE_HOSTNAME_FOR.indexOf(parsed.protocol) >= 0) {
        try {
          parsed.hostname = punycode.toASCII(parsed.hostname);
        } catch (er) { /**/ }
      }
    }

    return mdurl.encode(mdurl.format(parsed));
  }

  function normalizeLinkText(url) {
    var parsed = mdurl.parse(url, true);

    if (parsed.hostname) {
      // Encode hostnames in urls like:
      // `http://host/`, `https://host/`, `mailto:user@host`, `//host/`
      //
      // We don't encode unknown schemas, because it's likely that we encode
      // something we shouldn't (e.g. `skype:name` treated as `skype:host`)
      //
      if (!parsed.protocol || RECODE_HOSTNAME_FOR.indexOf(parsed.protocol) >= 0) {
        try {
          parsed.hostname = punycode.toUnicode(parsed.hostname);
        } catch (er) { /**/ }
      }
    }

    // add '%' to exclude list because of https://github.com/markdown-it/markdown-it/issues/720
    return mdurl.decode(mdurl.format(parsed), mdurl.decode.defaultChars + '%');
  }


  /**
   * class MarkdownIt
   *
   * Main parser/renderer class.
   *
   * ##### Usage
   *
   * ```javascript
   * // node.js, "classic" way:
   * var MarkdownIt = require('markdown-it'),
   *     md = new MarkdownIt();
   * var result = md.render('# markdown-it rulezz!');
   *
   * // node.js, the same, but with sugar:
   * var md = require('markdown-it')();
   * var result = md.render('# markdown-it rulezz!');
   *
   * // browser without AMD, added to "window" on script load
   * // Note, there are no dash.
   * var md = window.markdownit();
   * var result = md.render('# markdown-it rulezz!');
   * ```
   *
   * Single line rendering, without paragraph wrap:
   *
   * ```javascript
   * var md = require('markdown-it')();
   * var result = md.renderInline('__markdown-it__ rulezz!');
   * ```
   **/

  /**
   * new MarkdownIt([presetName, options])
   * - presetName (String): optional, `commonmark` / `zero`
   * - options (Object)
   *
   * Creates parser instanse with given config. Can be called without `new`.
   *
   * ##### presetName
   *
   * MarkdownIt provides named presets as a convenience to quickly
   * enable/disable active syntax rules and options for common use cases.
   *
   * - ["commonmark"](https://github.com/markdown-it/markdown-it/blob/master/lib/presets/commonmark.js) -
   *   configures parser to strict [CommonMark](http://commonmark.org/) mode.
   * - [default](https://github.com/markdown-it/markdown-it/blob/master/lib/presets/default.js) -
   *   similar to GFM, used when no preset name given. Enables all available rules,
   *   but still without html, typographer & autolinker.
   * - ["zero"](https://github.com/markdown-it/markdown-it/blob/master/lib/presets/zero.js) -
   *   all rules disabled. Useful to quickly setup your config via `.enable()`.
   *   For example, when you need only `bold` and `italic` markup and nothing else.
   *
   * ##### options:
   *
   * - __html__ - `false`. Set `true` to enable HTML tags in source. Be careful!
   *   That's not safe! You may need external sanitizer to protect output from XSS.
   *   It's better to extend features via plugins, instead of enabling HTML.
   * - __xhtmlOut__ - `false`. Set `true` to add '/' when closing single tags
   *   (`<br />`). This is needed only for full CommonMark compatibility. In real
   *   world you will need HTML output.
   * - __breaks__ - `false`. Set `true` to convert `\n` in paragraphs into `<br>`.
   * - __langPrefix__ - `language-`. CSS language class prefix for fenced blocks.
   *   Can be useful for external highlighters.
   * - __linkify__ - `false`. Set `true` to autoconvert URL-like text to links.
   * - __typographer__  - `false`. Set `true` to enable [some language-neutral
   *   replacement](https://github.com/markdown-it/markdown-it/blob/master/lib/rules_core/replacements.js) +
   *   quotes beautification (smartquotes).
   * - __quotes__ - `“”‘’`, String or Array. Double + single quotes replacement
   *   pairs, when typographer enabled and smartquotes on. For example, you can
   *   use `'«»„“'` for Russian, `'„“‚‘'` for German, and
   *   `['«\xA0', '\xA0»', '‹\xA0', '\xA0›']` for French (including nbsp).
   * - __highlight__ - `null`. Highlighter function for fenced code blocks.
   *   Highlighter `function (str, lang)` should return escaped HTML. It can also
   *   return empty string if the source was not changed and should be escaped
   *   externaly. If result starts with <pre... internal wrapper is skipped.
   *
   * ##### Example
   *
   * ```javascript
   * // commonmark mode
   * var md = require('markdown-it')('commonmark');
   *
   * // default mode
   * var md = require('markdown-it')();
   *
   * // enable everything
   * var md = require('markdown-it')({
   *   html: true,
   *   linkify: true,
   *   typographer: true
   * });
   * ```
   *
   * ##### Syntax highlighting
   *
   * ```js
   * var hljs = require('highlight.js') // https://highlightjs.org/
   *
   * var md = require('markdown-it')({
   *   highlight: function (str, lang) {
   *     if (lang && hljs.getLanguage(lang)) {
   *       try {
   *         return hljs.highlight(str, { language: lang, ignoreIllegals: true }).value;
   *       } catch (__) {}
   *     }
   *
   *     return ''; // use external default escaping
   *   }
   * });
   * ```
   *
   * Or with full wrapper override (if you need assign class to `<pre>`):
   *
   * ```javascript
   * var hljs = require('highlight.js') // https://highlightjs.org/
   *
   * // Actual default values
   * var md = require('markdown-it')({
   *   highlight: function (str, lang) {
   *     if (lang && hljs.getLanguage(lang)) {
   *       try {
   *         return '<pre class="hljs"><code>' +
   *                hljs.highlight(str, { language: lang, ignoreIllegals: true }).value +
   *                '</code></pre>';
   *       } catch (__) {}
   *     }
   *
   *     return '<pre class="hljs"><code>' + md.utils.escapeHtml(str) + '</code></pre>';
   *   }
   * });
   * ```
   *
   **/
  function MarkdownIt(presetName, options) {
    if (!(this instanceof MarkdownIt)) {
      return new MarkdownIt(presetName, options);
    }

    if (!options) {
      if (!utils.isString(presetName)) {
        options = presetName || {};
        presetName = 'default';
      }
    }

    /**
     * MarkdownIt#inline -> ParserInline
     *
     * Instance of [[ParserInline]]. You may need it to add new rules when
     * writing plugins. For simple rules control use [[MarkdownIt.disable]] and
     * [[MarkdownIt.enable]].
     **/
    this.inline = new parser_inline();

    /**
     * MarkdownIt#block -> ParserBlock
     *
     * Instance of [[ParserBlock]]. You may need it to add new rules when
     * writing plugins. For simple rules control use [[MarkdownIt.disable]] and
     * [[MarkdownIt.enable]].
     **/
    this.block = new parser_block();

    /**
     * MarkdownIt#core -> Core
     *
     * Instance of [[Core]] chain executor. You may need it to add new rules when
     * writing plugins. For simple rules control use [[MarkdownIt.disable]] and
     * [[MarkdownIt.enable]].
     **/
    this.core = new parser_core();

    /**
     * MarkdownIt#renderer -> Renderer
     *
     * Instance of [[Renderer]]. Use it to modify output look. Or to add rendering
     * rules for new token types, generated by plugins.
     *
     * ##### Example
     *
     * ```javascript
     * var md = require('markdown-it')();
     *
     * function myToken(tokens, idx, options, env, self) {
     *   //...
     *   return result;
     * };
     *
     * md.renderer.rules['my_token'] = myToken
     * ```
     *
     * See [[Renderer]] docs and [source code](https://github.com/markdown-it/markdown-it/blob/master/lib/renderer.js).
     **/
    this.renderer = new renderer();

    /**
     * MarkdownIt#linkify -> LinkifyIt
     *
     * [linkify-it](https://github.com/markdown-it/linkify-it) instance.
     * Used by [linkify](https://github.com/markdown-it/markdown-it/blob/master/lib/rules_core/linkify.js)
     * rule.
     **/
    this.linkify = new linkifyIt();

    /**
     * MarkdownIt#validateLink(url) -> Boolean
     *
     * Link validation function. CommonMark allows too much in links. By default
     * we disable `javascript:`, `vbscript:`, `file:` schemas, and almost all `data:...` schemas
     * except some embedded image types.
     *
     * You can change this behaviour:
     *
     * ```javascript
     * var md = require('markdown-it')();
     * // enable everything
     * md.validateLink = function () { return true; }
     * ```
     **/
    this.validateLink = validateLink;

    /**
     * MarkdownIt#normalizeLink(url) -> String
     *
     * Function used to encode link url to a machine-readable format,
     * which includes url-encoding, punycode, etc.
     **/
    this.normalizeLink = normalizeLink;

    /**
     * MarkdownIt#normalizeLinkText(url) -> String
     *
     * Function used to decode link url to a human-readable format`
     **/
    this.normalizeLinkText = normalizeLinkText;


    // Expose utils & helpers for easy acces from plugins

    /**
     * MarkdownIt#utils -> utils
     *
     * Assorted utility functions, useful to write plugins. See details
     * [here](https://github.com/markdown-it/markdown-it/blob/master/lib/common/utils.js).
     **/
    this.utils = utils;

    /**
     * MarkdownIt#helpers -> helpers
     *
     * Link components parser functions, useful to write plugins. See details
     * [here](https://github.com/markdown-it/markdown-it/blob/master/lib/helpers).
     **/
    this.helpers = utils.assign({}, helpers);


    this.options = {};
    this.configure(presetName);

    if (options) { this.set(options); }
  }


  /** chainable
   * MarkdownIt.set(options)
   *
   * Set parser options (in the same format as in constructor). Probably, you
   * will never need it, but you can change options after constructor call.
   *
   * ##### Example
   *
   * ```javascript
   * var md = require('markdown-it')()
   *             .set({ html: true, breaks: true })
   *             .set({ typographer, true });
   * ```
   *
   * __Note:__ To achieve the best possible performance, don't modify a
   * `markdown-it` instance options on the fly. If you need multiple configurations
   * it's best to create multiple instances and initialize each with separate
   * config.
   **/
  MarkdownIt.prototype.set = function (options) {
    utils.assign(this.options, options);
    return this;
  };


  /** chainable, internal
   * MarkdownIt.configure(presets)
   *
   * Batch load of all options and compenent settings. This is internal method,
   * and you probably will not need it. But if you will - see available presets
   * and data structure [here](https://github.com/markdown-it/markdown-it/tree/master/lib/presets)
   *
   * We strongly recommend to use presets instead of direct config loads. That
   * will give better compatibility with next versions.
   **/
  MarkdownIt.prototype.configure = function (presets) {
    var self = this, presetName;

    if (utils.isString(presets)) {
      presetName = presets;
      presets = config[presetName];
      if (!presets) { throw new Error('Wrong `markdown-it` preset "' + presetName + '", check name'); }
    }

    if (!presets) { throw new Error('Wrong `markdown-it` preset, can\'t be empty'); }

    if (presets.options) { self.set(presets.options); }

    if (presets.components) {
      Object.keys(presets.components).forEach(function (name) {
        if (presets.components[name].rules) {
          self[name].ruler.enableOnly(presets.components[name].rules);
        }
        if (presets.components[name].rules2) {
          self[name].ruler2.enableOnly(presets.components[name].rules2);
        }
      });
    }
    return this;
  };


  /** chainable
   * MarkdownIt.enable(list, ignoreInvalid)
   * - list (String|Array): rule name or list of rule names to enable
   * - ignoreInvalid (Boolean): set `true` to ignore errors when rule not found.
   *
   * Enable list or rules. It will automatically find appropriate components,
   * containing rules with given names. If rule not found, and `ignoreInvalid`
   * not set - throws exception.
   *
   * ##### Example
   *
   * ```javascript
   * var md = require('markdown-it')()
   *             .enable(['sub', 'sup'])
   *             .disable('smartquotes');
   * ```
   **/
  MarkdownIt.prototype.enable = function (list, ignoreInvalid) {
    var result = [];

    if (!Array.isArray(list)) { list = [ list ]; }

    [ 'core', 'block', 'inline' ].forEach(function (chain) {
      result = result.concat(this[chain].ruler.enable(list, true));
    }, this);

    result = result.concat(this.inline.ruler2.enable(list, true));

    var missed = list.filter(function (name) { return result.indexOf(name) < 0; });

    if (missed.length && !ignoreInvalid) {
      throw new Error('MarkdownIt. Failed to enable unknown rule(s): ' + missed);
    }

    return this;
  };


  /** chainable
   * MarkdownIt.disable(list, ignoreInvalid)
   * - list (String|Array): rule name or list of rule names to disable.
   * - ignoreInvalid (Boolean): set `true` to ignore errors when rule not found.
   *
   * The same as [[MarkdownIt.enable]], but turn specified rules off.
   **/
  MarkdownIt.prototype.disable = function (list, ignoreInvalid) {
    var result = [];

    if (!Array.isArray(list)) { list = [ list ]; }

    [ 'core', 'block', 'inline' ].forEach(function (chain) {
      result = result.concat(this[chain].ruler.disable(list, true));
    }, this);

    result = result.concat(this.inline.ruler2.disable(list, true));

    var missed = list.filter(function (name) { return result.indexOf(name) < 0; });

    if (missed.length && !ignoreInvalid) {
      throw new Error('MarkdownIt. Failed to disable unknown rule(s): ' + missed);
    }
    return this;
  };


  /** chainable
   * MarkdownIt.use(plugin, params)
   *
   * Load specified plugin with given params into current parser instance.
   * It's just a sugar to call `plugin(md, params)` with curring.
   *
   * ##### Example
   *
   * ```javascript
   * var iterator = require('markdown-it-for-inline');
   * var md = require('markdown-it')()
   *             .use(iterator, 'foo_replace', 'text', function (tokens, idx) {
   *               tokens[idx].content = tokens[idx].content.replace(/foo/g, 'bar');
   *             });
   * ```
   **/
  MarkdownIt.prototype.use = function (plugin /*, params, ... */) {
    var args = [ this ].concat(Array.prototype.slice.call(arguments, 1));
    plugin.apply(plugin, args);
    return this;
  };


  /** internal
   * MarkdownIt.parse(src, env) -> Array
   * - src (String): source string
   * - env (Object): environment sandbox
   *
   * Parse input string and return list of block tokens (special token type
   * "inline" will contain list of inline tokens). You should not call this
   * method directly, until you write custom renderer (for example, to produce
   * AST).
   *
   * `env` is used to pass data between "distributed" rules and return additional
   * metadata like reference info, needed for the renderer. It also can be used to
   * inject data in specific cases. Usually, you will be ok to pass `{}`,
   * and then pass updated object to renderer.
   **/
  MarkdownIt.prototype.parse = function (src, env) {
    if (typeof src !== 'string') {
      throw new Error('Input data should be a String');
    }

    var state = new this.core.State(src, this, env);

    this.core.process(state);

    return state.tokens;
  };


  /**
   * MarkdownIt.render(src [, env]) -> String
   * - src (String): source string
   * - env (Object): environment sandbox
   *
   * Render markdown string into html. It does all magic for you :).
   *
   * `env` can be used to inject additional metadata (`{}` by default).
   * But you will not need it with high probability. See also comment
   * in [[MarkdownIt.parse]].
   **/
  MarkdownIt.prototype.render = function (src, env) {
    env = env || {};

    return this.renderer.render(this.parse(src, env), this.options, env);
  };


  /** internal
   * MarkdownIt.parseInline(src, env) -> Array
   * - src (String): source string
   * - env (Object): environment sandbox
   *
   * The same as [[MarkdownIt.parse]] but skip all block rules. It returns the
   * block tokens list with the single `inline` element, containing parsed inline
   * tokens in `children` property. Also updates `env` object.
   **/
  MarkdownIt.prototype.parseInline = function (src, env) {
    var state = new this.core.State(src, this, env);

    state.inlineMode = true;
    this.core.process(state);

    return state.tokens;
  };


  /**
   * MarkdownIt.renderInline(src [, env]) -> String
   * - src (String): source string
   * - env (Object): environment sandbox
   *
   * Similar to [[MarkdownIt.render]] but for single paragraph content. Result
   * will NOT be wrapped into `<p>` tags.
   **/
  MarkdownIt.prototype.renderInline = function (src, env) {
    env = env || {};

    return this.renderer.render(this.parseInline(src, env), this.options, env);
  };


  var lib = MarkdownIt;

  var markdownIt = lib;

  var inst$1 = new markdownIt();
  var isSpace$2 = inst$1.utils.isSpace;

  function htmlBreak(state, silent) {

      var start  = state.pos;
      var pos = state.pos;
      var max = state.posMax;

      if (silent) { return false; } // don't run any pairs in validation mode
      //if (state.src.charCodeAt(start) !== 0x3c/* < */) { return false; }
      if (state.src.substr(start, 4) !== '<br>') { return false; }

      state.push('hardbreak', 'br', 0);


      pos = pos + 4;

      // skip heading spaces for next line
      while (pos < max && isSpace$2(state.src.charCodeAt(pos))) {
          pos++;
      }

      state.pos = pos;
      return true;
  }

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var hard_break = {
      id: 'hard_break',
      schema: schema$d,
      registerMarkdownIt: function (markdownIt) {
          markdownIt.inline.ruler.before('newline','htmlbreak', htmlBreak);
      }
  };

  var schema$c = {
      nodes: {
          heading:  {
              sortOrder: 400,
              attrs: {level: {default: 1}},
              content: "(text | image)*",
              group: "block",
              defining: true,
              parseDOM: [{tag: "h1", attrs: {level: 1}},
                  {tag: "h2", attrs: {level: 2}},
                  {tag: "h3", attrs: {level: 3}},
                  {tag: "h4", attrs: {level: 4}},
                  {tag: "h5", attrs: {level: 5}},
                  {tag: "h6", attrs: {level: 6}}],
              toDOM: function (node) {
                  return ["h" + node.attrs.level, 0]
              },
              parseMarkdown: {
                  block: "heading", getAttrs: function (tok) {
                      return ({level: +tok.tag.slice(1)});
                  }
              },
              toMarkdown: function (state, node) {
                  state.write(state.repeat("#", node.attrs.level) + " ");
                  state.renderInline(node);
                  state.closeBlock(node);
              }
          }
      }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  // : (NodeType) → InputRule
  // Given a blockquote node type, returns an input rule that turns `"> "`
  // at the start of a textblock into a blockquote.
  var headingRule = function (schema) {
      var maxLevel = 6;
      return textblockTypeInputRule(new RegExp("^(#{1," + maxLevel + "})\\s$"),
          schema.nodes.heading, function (match) { return ({level: match[1].length}); })
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  function makeHeading(context) {
      var r = {};
      for (var i = 1; i <= 6; i++) {
          r["makeHead" + i] = blockTypeItem$1(context.schema.nodes.heading, {
              label: "H" + i,
              title: context.translate("Change to heading")+" " + i + ' (' + Array(i + 1).join("#") + ')',
              sortOrder: i,
              attrs: {level: i}
          });
      }

      return new DropdownSubmenu$1([r.makeHead1, r.makeHead2, r.makeHead3, r.makeHead4, r.makeHead5, r.makeHead6], {label: context.translate("Heading")});
  }

  function menu$c(context) {
      return [
          {
              id: 'makeHeading',
              node: 'heading',
              group: 'types',
              item: makeHeading(context)
          }
      ]
  }

  var n={false:"push",true:"unshift"},e=Object.prototype.hasOwnProperty,t=function(n,t,r,i){var u=n,o=i;if(r&&e.call(t,u)){ throw Error("User defined id attribute '"+n+"' is NOT unique. Please fix it in your markdown to continue."); }for(;e.call(t,u);){ u=n+"-"+o++; }return t[u]=!0,u},r=function n(e,r){r=Object.assign({},n.defaults,r),e.core.ruler.push("anchor",function(n){var e,i={},u=n.tokens,o=Array.isArray(r.level)?(e=r.level,function(n){return e.includes(n)}):function(n){return function(e){return e>=n}}(r.level);u.filter(function(n){return "heading_open"===n.type}).filter(function(n){return o(Number(n.tag.substr(1)))}).forEach(function(e){var o=u[u.indexOf(e)+1].children.filter(function(n){return "text"===n.type||"code_inline"===n.type}).reduce(function(n,e){return n+e.content},""),c=e.attrGet("id");c=null==c?t(r.slugify(o),i,!1,r.uniqueSlugStartIndex):t(c,i,!0,r.uniqueSlugStartIndex),e.attrSet("id",c),r.permalink&&r.renderPermalink(c,r,n,u.indexOf(e)),r.callback&&r.callback(e,{slug:c,title:o});});});};r.defaults={level:1,slugify:function(n){return encodeURIComponent(String(n).trim().toLowerCase().replace(/\s+/g,"-"))},uniqueSlugStartIndex:1,permalink:!1,renderPermalink:function(e,t,r,i){var u,o=[Object.assign(new r.Token("link_open","a",1),{attrs:[].concat(t.permalinkClass?[["class",t.permalinkClass]]:[],[["href",t.permalinkHref(e,r)]],Object.entries(t.permalinkAttrs(e,r)))}),Object.assign(new r.Token("html_block","",0),{content:t.permalinkSymbol}),new r.Token("link_close","a",-1)];t.permalinkSpace&&o[n[!t.permalinkBefore]](Object.assign(new r.Token("text","",0),{content:" "})),(u=r.tokens[i+1].children)[n[t.permalinkBefore]].apply(u,o);},permalinkClass:"header-anchor",permalinkSpace:!0,permalinkSymbol:"¶",permalinkBefore:!1,permalinkHref:function(n){return "#"+n},permalinkAttrs:function(n){return {}}};

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */


  var heading = {
      id: 'heading',
      schema: schema$c,
      menu: function (context) { return menu$c(context); },
      inputRules: function (schema) {return [headingRule(schema)]},
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */
  var schema$b = {
      nodes: {
          horizontal_rule: {
              sortOrder: 300,
              group: "block",
              parseDOM: [{tag: "hr"}],
              toDOM: function toDOM() {
                  return ["div", ["hr"]]
              },
              parseMarkdown: {hr: {node: "horizontal_rule"}},
              toMarkdown: function (state, node) {
                  state.write(node.attrs.markup || "---");
                  state.closeBlock(node);
              }
          }
      }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  function insertHorizontalRule(context) {
      var hr = context.schema.nodes.horizontal_rule;
      return new MenuItem$1({
          title: context.translate("Insert horizontal rule"),
          label: context.translate("Horizontal rule"),
          sortOrder: 200,
          enable: function enable(state) {
              return canInsert(state, hr)
          },
          run: function run(state, dispatch) {
              dispatch(state.tr.replaceSelectionWith(hr.create()));
          }
      })
  }

  function menu$b(context) {
      return [
          {
              id: 'insertHorizontalRule',
              node: 'horizontal_rule',
              group: 'insert',
              item: insertHorizontalRule(context)
          }
      ]
  }

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var horizontal_rule = {
      id: 'horizontal_rule',
      schema: schema$b,
      menu: function (context) { return menu$b(context); }
  };

  var FLOAT_NONE = 0;
  var FLOAT_LEFT = 1;
  var FLOAT_CENTER = 2;
  var FLOAT_RIGHT = 3;

  var FLOAT_ALT_EXT_NONE = '';
  var FLOAT_ALT_EXT_LEFT = '<';
  var FLOAT_ALT_EXT_CENTER = '><';
  var FLOAT_ALT_EXT_RIGHT = '>';

  var FLOAT_MAP = [
      FLOAT_ALT_EXT_NONE,
      FLOAT_ALT_EXT_LEFT,
      FLOAT_ALT_EXT_CENTER,
      FLOAT_ALT_EXT_RIGHT
  ];

  function getAltExtensionByFloat(float) {
      return FLOAT_MAP[parseInt(float)] || FLOAT_ALT_EXT_NONE;
  }

  function parseFloatFromAlt(alt) {
      var float = FLOAT_NONE;
      var ext = FLOAT_ALT_EXT_NONE;

      if(!alt) {
          return {
              alt: alt,
              float: float,
              ext: ext
          }
      }

      if(endsWith$2(alt, FLOAT_ALT_EXT_CENTER)) {
          alt = alt.substring(0, alt.length -2);
          ext = FLOAT_ALT_EXT_CENTER;
          float = FLOAT_CENTER;
      } else if(endsWith$2(alt, FLOAT_ALT_EXT_LEFT)) {
          alt = alt.substring(0, alt.length -1);
          ext = FLOAT_ALT_EXT_LEFT;
          float = FLOAT_LEFT;
      } else if(endsWith$2(alt, FLOAT_ALT_EXT_RIGHT)) {
          alt = alt.substring(0, alt.length -1);
          ext = FLOAT_ALT_EXT_RIGHT;
          float = FLOAT_RIGHT;
      }

      return {
          alt: alt,
          ext: ext,
          float: float
      }
  }

  function getClassForFloat(float) {
      float = parseInt(float);
      switch (float) {
          case FLOAT_LEFT:
              return 'pull-left';
          case FLOAT_CENTER:
              return 'center-block';
          case FLOAT_RIGHT:
              return 'pull-right';
          default:
              return '';
      }
  }

  var endsWith$2 = function(string, suffix) {
      return string.indexOf(suffix, string.length - suffix.length) !== -1;
  };

  var imageFloat = /*#__PURE__*/Object.freeze({
    __proto__: null,
    FLOAT_NONE: FLOAT_NONE,
    FLOAT_LEFT: FLOAT_LEFT,
    FLOAT_CENTER: FLOAT_CENTER,
    FLOAT_RIGHT: FLOAT_RIGHT,
    FLOAT_ALT_EXT_NONE: FLOAT_ALT_EXT_NONE,
    FLOAT_ALT_EXT_LEFT: FLOAT_ALT_EXT_LEFT,
    FLOAT_ALT_EXT_CENTER: FLOAT_ALT_EXT_CENTER,
    FLOAT_ALT_EXT_RIGHT: FLOAT_ALT_EXT_RIGHT,
    FLOAT_MAP: FLOAT_MAP,
    getAltExtensionByFloat: getAltExtensionByFloat,
    parseFloatFromAlt: parseFloatFromAlt,
    getClassForFloat: getClassForFloat
  });

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var schema$a = {
      nodes: {
          image: {
              sortOrder: 1000,
              inline: true,
              attrs: {
                  src: {},
                  alt: {default: null},
                  title: {default: null},
                  width: {default: null},
                  height: {default: null},
                  float: {default: FLOAT_NONE},
                  fileGuid: { default: null},
              },
              group: "inline",
              draggable: true,
              parseDOM: [{
                  tag: "img[src]", getAttrs: function getAttrs(dom) {
                      return {
                          src: dom.getAttribute("src"),
                          title: dom.getAttribute("title"),
                          alt: dom.getAttribute("alt"),
                          width: dom.getAttribute("width"),
                          height: dom.getAttribute("height"),
                          fileGuid: dom.getAttribute("data-file-guid")
                      }
                  }
              }],
              parseMarkdown: {
                  node: "image", getAttrs: function (tok) {
                      var src =  (window.humhub) ? humhub.modules.file.filterFileUrl(tok.attrGet("src")).url : tok.attrGet("src");
                      var fileGuid = (window.humhub) ?  humhub.modules.file.filterFileUrl(tok.attrGet("src")).guid : null;

                      return ({
                          src: src,
                          title: tok.attrGet("title") || null,
                          width: tok.attrGet("width") || null,
                          height: tok.attrGet("height") || null,
                          alt: tok.attrGet("alt") || null,
                          float: tok.attrGet("float") || FLOAT_NONE,
                          fileGuid: fileGuid
                      });
                  }
              },
              toMarkdown: function (state, node) {
                  var resizeAddition = "";

                  if(node.attrs.width || node.attrs.height) {
                      resizeAddition += " =";
                      resizeAddition += (node.attrs.width) ? node.attrs.width : '';
                      resizeAddition += 'x';
                      resizeAddition += (node.attrs.height) ? node.attrs.height : '';
                  }

                  var src = (node.attrs.fileGuid) ? 'file-guid:'+node.attrs.fileGuid  : node.attrs.src;

                  var float = getAltExtensionByFloat(node.attrs.float);

                  state.write("![" + state.esc(node.attrs.alt || "") + float + "](" + state.esc(src) +
                      (node.attrs.title ? " " + state.quote(node.attrs.title) : "") + resizeAddition + ")");
              }
          }
      }
  };

  // Parse image size

  function parseNextNumber(str, pos, max) {
    var code,
    start = pos,
    result = {
      ok: false,
      pos: pos,
      value: ''
    };

    code = str.charCodeAt(pos);

    while (pos < max && (code >= 0x30 /* 0 */ && code <= 0x39 /* 9 */) || code === 0x25 /* % */) {
      code = str.charCodeAt(++pos);
    }

    result.ok = true;
    result.pos = pos;
    result.value = str.slice(start, pos);

    return result;
  }

  var parse_image_size = function parseImageSize(str, pos, max) {
    var code,
    result = {
      ok: false,
      pos: 0,
      width: '',
      height: ''
    };

    if (pos >= max) { return result; }

    code = str.charCodeAt(pos);

    if (code !== 0x3d /* = */) { return result; }

    pos++;

    // size must follow = without any white spaces as follows
    // (1) =300x200
    // (2) =300x
    // (3) =x200
    code = str.charCodeAt(pos);
    if (code !== 0x78 /* x */ && (code < 0x30 || code  > 0x39) /* [0-9] */) {
      return result;
    }

    // parse width
    var resultW = parseNextNumber(str, pos, max);
    pos = resultW.pos;

    // next charactor must be 'x'
    code = str.charCodeAt(pos);
    if (code !== 0x78 /* x */) { return result; }

    pos++;

    // parse height
    var resultH = parseNextNumber(str, pos, max);
    pos = resultH.pos;

    result.width = resultW.value;
    result.height = resultH.value;
    result.pos = pos;
    result.ok = true;
    return result;
  };

  function image_with_size(md, options) {
      return function(state, silent) {
          var attrs,
              code,
              label,
              labelEnd,
              labelStart,
              pos,
              ref,
              res,
              title,
              width = '',
              height = '',
              token,
              tokens,
              start,
              href = '',
              oldPos = state.pos,
              max = state.posMax;

          if (state.src.charCodeAt(state.pos) !== 0x21/* ! */) { return false; }
          if (state.src.charCodeAt(state.pos + 1) !== 0x5B/* [ */) { return false; }

          labelStart = state.pos + 2;
          labelEnd = md.helpers.parseLinkLabel(state, state.pos + 1, false);

          // parser failed to find ']', so it's not a valid link
          if (labelEnd < 0) { return false; }

          pos = labelEnd + 1;
          if (pos < max && state.src.charCodeAt(pos) === 0x28/* ( */) {

              //
              // Inline link
              //

              // [link](  <href>  "title"  )
              //        ^^ skipping these spaces
              pos++;
              for (; pos < max; pos++) {
                  code = state.src.charCodeAt(pos);
                  if (code !== 0x20 && code !== 0x0A) { break; }
              }
              if (pos >= max) { return false; }

              // [link](  <href>  "title"  )
              //          ^^^^^^ parsing link destination
              start = pos;
              res = md.helpers.parseLinkDestination(state.src, pos, state.posMax);
              if (res.ok) {
                  href = state.md.normalizeLink(res.str);
                  if (state.md.validateLink(href)) {
                      pos = res.pos;
                  } else {
                      href = '';
                  }
              }

              // [link](  <href>  "title"  )
              //                ^^ skipping these spaces
              start = pos;
              for (; pos < max; pos++) {
                  code = state.src.charCodeAt(pos);
                  if (code !== 0x20 && code !== 0x0A) { break; }
              }

              // [link](  <href>  "title"  )
              //                  ^^^^^^^ parsing link title
              res = md.helpers.parseLinkTitle(state.src, pos, state.posMax);
              if (pos < max && start !== pos && res.ok) {
                  title = res.str;
                  pos = res.pos;

                  // [link](  <href>  "title"  )
                  //                         ^^ skipping these spaces
                  for (; pos < max; pos++) {
                      code = state.src.charCodeAt(pos);
                      if (code !== 0x20 && code !== 0x0A) { break; }
                  }
              } else {
                  title = '';
              }

              // [link](  <href>  "title" =WxH  )
              //                          ^^^^ parsing image size
              if (pos - 1 >= 0) {
                  code = state.src.charCodeAt(pos - 1);

                  // there must be at least one white spaces
                  // between previous field and the size
                  if (code === 0x20) {
                      res = parse_image_size(state.src, pos, state.posMax);
                      if (res.ok) {
                          width = res.width;
                          height = res.height;
                          pos = res.pos;

                          // [link](  <href>  "title" =WxH  )
                          //                              ^^ skipping these spaces
                          for (; pos < max; pos++) {
                              code = state.src.charCodeAt(pos);
                              if (code !== 0x20 && code !== 0x0A) { break; }
                          }
                      }
                  }
              }

              if (pos >= max || state.src.charCodeAt(pos) !== 0x29/* ) */) {
                  state.pos = oldPos;
                  return false;
              }
              pos++;

          } else {
              //
              // Link reference
              //
              if (typeof state.env.references === 'undefined') { return false; }

              // [foo]  [bar]
              //      ^^ optional whitespace (can include newlines)
              for (; pos < max; pos++) {
                  code = state.src.charCodeAt(pos);
                  if (code !== 0x20 && code !== 0x0A) { break; }
              }

              if (pos < max && state.src.charCodeAt(pos) === 0x5B/* [ */) {
                  start = pos + 1;
                  pos = md.helpers.parseLinkLabel(state, pos);
                  if (pos >= 0) {
                      label = state.src.slice(start, pos++);
                  } else {
                      pos = labelEnd + 1;
                  }
              } else {
                  pos = labelEnd + 1;
              }

              // covers label === '' and label === undefined
              // (collapsed reference link and shortcut reference link respectively)
              if (!label) { label = state.src.slice(labelStart, labelEnd); }

              ref = state.env.references[md.utils.normalizeReference(label)];
              if (!ref) {
                  state.pos = oldPos;
                  return false;
              }
              href = ref.href;
              title = ref.title;
          }

          //
          // We found the end of the link, and know for a fact it's a valid link;
          // so all that's left to do is to call tokenizer.
          //
          if (!silent) {
              state.pos = labelStart;
              state.posMax = labelEnd;

              var newState = new state.md.inline.State(
                  state.src.slice(labelStart, labelEnd),
                  state.md,
                  state.env,
                  tokens = []
              );
              newState.md.inline.tokenize(newState);

              // if 'autofill' option is specified
              // and width/height are both blank,
              // they are filled automatically
              if (options) {
                  if (options.autofill && width === '' && height === '') {
                      try {
                          var dimensions = sizeOf(href);
                          width = dimensions.width;
                          height = dimensions.height;
                      } catch (e) { }
                  }
              }

              token          = state.push('image', 'img', 0);
              token.attrs    = attrs = [ [ 'src', href ], [ 'alt', '' ] ];
              token.children = tokens;

              // Parse image float extension
              var altTextToken = tokens.length ? tokens[tokens.length - 1] : null;

              if(altTextToken) {
                  var ref$1 = imageFloat.parseFloatFromAlt(altTextToken['content']);
                  var float = ref$1.float;
                  var alt = ref$1.alt;
                  altTextToken['content'] = alt;
                  token.attrs.push(['float', float]);
              } else {
                  token.attrs.push(['float', imageFloat.FLOAT_NONE]);
              }

              if (title) {
                  attrs.push([ 'title', title ]);
              }

              if (width !== '') {
                  attrs.push([ 'width', width ]);
              }

              if (height !== '') {
                  attrs.push([ 'height', height ]);
              }
          }

          state.pos = pos;
          state.posMax = max;
          return true;
      };
  }

  var markdownit_imsize = function imsize_plugin(md, options) {
      md.inline.ruler.before('emphasis', 'image', image_with_size(md, options));
  };

  var DEFAULT_LINK_REL = 'noopener noreferrer nofollow';

  function validateHref(href, cfg) {
      cfg = cfg || {};

      return /^https?:\/\//i.test(href) //http:/https:
          || /^mailto:/i.test(href) //mailto:
          || /^ftps?:\/\//i.test(href) //ftp:/ftps:
          || (cfg.anchor && validateAnchor(href)) //anchor
          || (cfg.relative && validateRelative(href)); //relative
  }

  function validateRelative(href) {
      return /^\/[^\/].*$/i.test(href);
  }

  function validateAnchor(href) {
      return /^#((?:[!$&()*+,;=._~:@?-]|%[0-9a-fA-F]{2}|[a-zA-Z0-9])+)$/i.test(href);
  }

  function buildLink(href, attrs, text, validate) {
      attrs = attrs || {};

      if(validate !== false) {
          href = validateHref(href, validate) ? href : '#';
      }

      text = text || href;

      var defaultAttrs = {href: href};

      if(href !== '#') {
          defaultAttrs.target = '_blank';
          defaultAttrs.rel = DEFAULT_LINK_REL;
      }

      return $('<div>').append($('<a>').attr($.extend(defaultAttrs, attrs)).text(text)).html();
  }

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  function insertImageItem(context) {
      return new MenuItem$1({
          title: context.translate("Insert image"),
          label: context.translate("Image"),
          sortOrder: 100,
          enable: function enable(state) {
              return canInsert(state, context.schema.nodes.image) && canInsertLink(state);
          },
          run: function run(state, _, view) {
              if (state.selection instanceof NodeSelection && state.selection.node.type === context.schema.nodes.image) {
                  editNode$1(state.selection.node, context, view);
              } else {
                  promt$1(context.translate("Insert image"), context, null, view);
              }
          }
      })
  }

  function editNode$1(node, context, view) {
      promt$1(context.translate("Edit image"), context, node.attrs, view, node);
  }

  var isDefined = function(obj) {
      if(arguments.length > 1) {
          var result = true;
          this.each(arguments, function(index, value) {
              if(!isDefined(value)) {
                  return false;
              }
          });

          return result;
      }
      return typeof obj !== 'undefined';
  };

  var endsWith$1 = function(val, suffix) {
      if(!isDefined(val) || !isDefined(suffix)) {
          return false;
      }
      return val.indexOf(suffix, val.length - suffix.length) !== -1;
  };

  function promt$1(title, context, attrs, view, node) {
      var state = view.state;

      var ref = state.selection;
      var from = ref.from;
      var to = ref.to;

      var cleanDimension = function(val) {
          val = val.trim();
          if(endsWith$1(val, 'px')) {
              val = val.substring(0, val.length - 2);
          }
          return val;
      };

      var validateDimension = function(val) {
          val = cleanDimension(val);

          if(val.length && !/^[0-9]+%?$/.test(val)) {
              return context.translate('Invalid dimension format used.')
          }
      };

      var validateSource = function(val) {
          if(!validateHref(val)) {
              return context.translate('Invalid image source.')
          }
      };

      openPrompt({
          title: title,
          fields: {
              src: new TextField({
                  label: context.translate("Location"),
                  required: true,
                  value: attrs && attrs.src,
                  validate: validateSource
              }),
              title: new TextField({label: context.translate("Title"), value: attrs && attrs.title}),
              alt: new TextField({
                  label: context.translate("Description"),
                  value: attrs ? attrs.alt : state.doc.textBetween(from, to, " ")
              }),
              width: new TextField({
                  label: context.translate("Width"),
                  value: attrs && attrs.width,
                  clean: cleanDimension,
                  validate: validateDimension
              }),
              height: new TextField({
                  label: context.translate("Height"),
                  value: attrs && attrs.height,
                  clean: cleanDimension,
                  validate: validateDimension
              }),
              float: new SelectField({
                  label: context.translate("Position"),
                  value: attrs && attrs.float,
                  options: [
                      {label: context.translate("Normal"), value: 0},
                      {label: context.translate("Left"), value: 1},
                      {label: context.translate("Center"), value: 2},
                      {label: context.translate("Right"), value: 3}
                  ]
              })
          },
          callback: function callback(attrs) {
              if(node && node.attrs.src === attrs.src) {
                  attrs.fileGuid = node.attrs.fileGuid;
              }

              attrs.float = parseInt(attrs.float);

              view.dispatch(view.state.tr.replaceSelectionWith(context.schema.nodes.image.createAndFill(attrs)));
              view.focus();
          }
      });
  }

  function menu$a(context) {
      return [
          {
              id: 'insertImage',
              node: 'image',
              group: 'insert',
              item: insertImageItem(context)
          }
      ]
  }

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2018 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var imagePlugin = function (context) {

      context.editor.$.on('mouseleave', function (e) {
          var target =  e.toElement || e.relatedTarget;
          if(!$(target).closest('.humhub-richtext-inline-menu').length) {
            $('.humhub-richtext-inline-menu').remove();
          }
      });

      return new Plugin({
          props: {
              nodeViews: {
                  image: function image(node) { return new ImageView(node, context) }
              },
          },
          filterTransaction: function (tr, state) {
              if(!(tr.curSelection instanceof NodeSelection)) {
               $('.humhub-richtext-image-edit').remove();
              }

              return true;
          }
      });
  };

  var ImageView = function ImageView(node, context) {
      var this$1$1 = this;

      // The editor will use this as the node's DOM representation
      this.createDom(node);

      context.event.on('clear, serialize', function() {
         $('.humhub-richtext-inline-menu').remove();
      });


      this.dom.addEventListener("mouseenter", function (e) {
          var $img = $(this$1$1.dom);
          var offset = $img.offset();
          var editorOffset = context.editor.$.offset();

          if(offset.top < editorOffset.top) {
              return;
          }

          var $edit = $('<div>').addClass('humhub-richtext-inline-menu').addClass('humhub-richtext-image-edit')
              .html('<button class="btn btn-primary btn-xs"><i class="fa fa-pencil"></i></button>')
              .css({
                  position: 'absolute',
                  left: offset.left + $img.width() - (25),
                  top: offset.top + 5,
                  'z-index': 997
              }).on('mousedown', function (evt) {
                  var view = context.editor.view;
                  var doc = view.state.doc;
                  view.dispatch(view.state.tr.setSelection(NodeSelection.create(doc, view.posAtDOM(this$1$1.dom))).scrollIntoView());
                  editNode$1(node, context, view);
              });

          $('html').append($edit);
      });

      this.dom.addEventListener("mouseleave", function (e) {
          var target =  e.toElement || e.relatedTarget;
          if(!$(target).closest('.humhub-richtext-inline-menu').length) {
              $('.humhub-richtext-inline-menu').remove();
          }
      });
  };

  ImageView.prototype.createDom = function createDom (node) {
      var src = validateHref(node.attrs.src) ? node.attrs.src : '#';

      this.dom = $('<img>').attr({
          src: src,
          title: node.attrs.title || null,
          width: node.attrs.width || null,
          height: node.attrs.height || null,
          alt: node.attrs.alt || null,
          class: getClassForFloat(node.attrs.float),
          'data-file-guid': node.attrs.fileGuid
      })[0];
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var image = {
      id: 'image',
      schema: schema$a,
      menu: function (context) { return menu$a(context); },
      plugins: function (context) {
          return [
              imagePlugin(context)
          ]
      },
      registerMarkdownIt: function (markdownIt) {
          markdownIt.use(markdownit_imsize);

          var defaultRender = markdownIt.renderer.rules.image || function(tokens, idx, options, env, self) {
              return self.renderToken(tokens, idx, options);
          };

          markdownIt.renderer.rules.image = function (tokens, idx, options, env, self) {

              var imageToken = tokens[idx];
              var srcIndex = imageToken.attrIndex('src');


              var srcFilter = (window.humhub) ? humhub.modules.file.filterFileUrl(imageToken.attrs[srcIndex][1]) : {url : imageToken.attrs[srcIndex][1]};
              imageToken.attrs[srcIndex][1] = validateHref(srcFilter.url) ? srcFilter.url : '#';

              if(srcFilter.guid) {
                  imageToken.attrPush(['data-file-guid', srcFilter.guid]); // add new attribute
              }

              if(env && env.context && env.context.uuid) {
                  imageToken.attrPush(['data-ui-gallery', env.context.uuid]);
              }

              var float = imageToken.attrs[ imageToken.attrIndex('float')][1];

              if(float) {
                  imageToken.attrPush(['class', getClassForFloat(float)]);
                  imageToken.attrs.splice(imageToken.attrIndex('float'), 1);
              }

              // pass token to default renderer.
              return defaultRender(tokens, idx, options, env, self);
          };
      }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var schema$9 = {
      marks: {
          sortOrder: 300,
          link: {
              attrs: {
                  href: {},
                  title: {default: null},
                  target: {default: '_blank'},
                  fileGuid: { default: null},
                  rel: {default: DEFAULT_LINK_REL}
              },
              inclusive: false,
              parseDOM:
                  [{
                      tag: "a[href]", getAttrs: function getAttrs(dom) {
                          var href = dom.getAttribute("href");
                          if (!validateHref(href))  {
                              href = '#';
                          }

                          return {
                              href: href,
                              title: dom.getAttribute("title"),
                              target: dom.getAttribute("target"),
                              fileGuid: dom.getAttribute("data-file-guid")
                          }
                      }
                  }],
              toDOM: function toDOM(node) { var ref = node.attrs;
              var href = ref.href;
              var title = ref.title; return ["a", {href: href, title: title}, 0] },
              parseMarkdown: {
                  mark: "link", getAttrs: function (tok) {
                      var href = (window.humhub) ? humhub.modules.file.filterFileUrl(tok.attrGet("href")).url : tok.attrGet("href");
                      var fileGuid = (window.humhub) ? humhub.modules.file.filterFileUrl(tok.attrGet("href")).guid : null;

                      if (!validateHref(href))  {
                          href = '#';
                      }

                      return ({
                          href: href,
                          title: tok.attrGet("title") || null,
                          fileGuid: fileGuid
                      });
                  }
              },
              toMarkdown: {
                  open: "[",
                  close: function close(state, mark) {
                      var href = (mark.attrs.fileGuid) ? 'file-guid:'+mark.attrs.fileGuid  : mark.attrs.href;
                      return "](" + state.esc(href) + (mark.attrs.title ? " " + state.quote(mark.attrs.title) : "") + ")"
                  }
              }
          }
      }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  function linkItem(context) {
      var mark = context.schema.marks.link;
      return new MenuItem$1({
          title: context.translate("Add or remove link"),
          sortOrder: 500,
          icon: icons$1.link,
          active: function active(state) {
              return markActive(state, mark)
          },
          enable: function enable(state) {
              if(state.selection.empty) {
                  return false;
              }

              return canInsertLink(state);
          },
          run: function run(state, dispatch, view) {
              if (markActive(state, mark)) {
                  toggleMark(mark)(state, dispatch);
                  return true
              }

              promt(context.translate("Create a link"), context);
          }
      })
  }

  function editNode(dom, context) {
      var doc = context.editor.view.state.doc;
      var view = context.editor.view;

      var nodePos = view.posAtDOM(dom);
      var node = doc.nodeAt(nodePos);

      if(node.type != context.schema.nodes.text) {
          return;
      }

      var $pos = doc.resolve(nodePos);
      var $end = $pos.node(0).resolve($pos.pos + node.nodeSize);
      view.dispatch(view.state.tr.setSelection(new TextSelection($pos, $end)).scrollIntoView());

      var mark = getLinkMark(node, context);

      promt(context.translate("Edit link"), context, $.extend({}, mark.attrs, {text: node.text}), node, mark);
  }

  function promt(title, context, attrs, node, mark) {
      var view = context.editor.view;

      var fields =  {
          text:  new TextField({label: "Text", value: attrs && attrs.text}),
          href: new TextField({
              label: context.translate("Link target"),
              value: attrs && attrs.href,
              required: true,
              clean: function (val) {
                  if (!validateHref(val))  {
                      return 'https://' + val;
                  }

                  return val;
              }
          }),
          title: new TextField({label: "Title", value: attrs && attrs.title})
      };

      if(!node) {
          delete fields['text'];
      }

      openPrompt({
          title: title,
          fields: fields,
          callback: function callback(attrs) {
              if(node) {
                  if(mark.attrs.href === attrs.href) {
                      attrs.fileGuid = mark.attrs.fileGuid;
                  }

                  var newSet = mark.type.removeFromSet(node.marks);
                  var newNode = node.copy(attrs.text);
                  newNode.marks = newSet;

                  delete attrs.text;

                  mark.attrs = attrs;

                  newNode.marks = mark.addToSet(newNode.marks);
                  view.dispatch(view.state.tr.replaceSelectionWith(newNode, false));
              } else {
                  toggleMark(context.schema.marks.link, attrs)(view.state, view.dispatch);
              }
              view.focus();
          }
      });
  }

  function getLinkMark(node, context) {
      var result = null;
      node.marks.forEach(function (mark) {
          if(mark.type == context.schema.marks.link) {
              result = mark;
          }
      });

      return result;
  }

  function menu$9(context) {
      return [
          {
              id: 'linkItem',
              mark: 'link',
              group: 'marks',
              item: linkItem(context)
          }
      ]
  }

  var linkPlugin = function (context) {
      return new Plugin({
          props: {
              nodeViews: {
                  link: function link(node) { return new LinkView(node, context) }
              },
              transformPasted: function (slice) {
                  return new Slice$1(linkify(slice.content, context), slice.openStart, slice.openEnd);
              }
          }
      });
  };

  var LinkView = function LinkView(mark, context) {
      var this$1$1 = this;

      // The editor will use this as the node's DOM representation
      this.createDom(mark);
      this.dom.addEventListener("click", function (e) {
          editNode(this$1$1.dom, context);
      });
  };

  LinkView.prototype.createDom = function createDom (mark) {
      this.dom = $(buildLink(mark.attrs.href, {
          'data-file-guid': mark.attrs.fileGuid,
          target: mark.attrs.target || '_blank'
      }))[0];
  };

  LinkView.prototype.stopEvent = function stopEvent () { return true };

  var HTTP_LINK_REGEX = /((https?:\/\/(?:www\.|(?!www))[^\s\.]+\.[^\s]{2,})|[a-zA-Z0-9._-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,6})/ig;


  var linkify = function(fragment, context) {
      var linkified = [];
      var urls = [];
      fragment.forEach(function(child){
          if (child.isText) {
              var text = child.text;
              var pos = 0, match;

              while (match = HTTP_LINK_REGEX.exec(text)) {
                  var start = match.index;
                  var end = start + match[0].length;
                  var link = child.type.schema.marks['link'];

                  // simply copy across the text from before the match
                  if (start > 0) {
                      linkified.push(child.cut(pos, start));
                  }

                  var urlText = text.slice(start, end);

                  if(urlText.indexOf('http') !== 0) {
                      urlText = 'mailto:'+urlText;
                  }

                  urls.push(urlText);
                  linkified.push(
                      child.cut(start, end).mark(link.create({href: urlText}).addToSet(child.marks))
                  );
                  pos = end;
              }

              // copy over whatever is left
              if (pos < text.length) {
                  linkified.push(child.cut(pos));
              }
          } else {
              linkified.push(child.copy(linkify(child.content, context)));
          }
      });

      if(urls.length) {
          context.event.trigger('linkified', [urls, linkified]);
      }
      return Fragment$1.fromArray(linkified)
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var link = {
      id: 'link',
      schema: schema$9,
      menu: function (context) { return menu$9(context); },
      registerMarkdownIt: function (md) {

          var defaultRender = md.renderer.rules.link_open || function(tokens, idx, options, env, self) {
              return self.renderToken(tokens, idx, options);
          };

          md.renderer.rules.link_open = function (tokens, idx, options, env, self) {
              var hrefIndex = tokens[idx].attrIndex('href');

              var hrefFilter = (window.humhub) ? humhub.modules.file.filterFileUrl(tokens[idx].attrs[hrefIndex][1]) : {url: tokens[idx].attrs[hrefIndex][1] };
              tokens[idx].attrs[hrefIndex][1] = hrefFilter.url;

              if(hrefFilter.guid) {
                  tokens[idx].attrPush(['data-file-guid', hrefFilter.guid]); // add new attribute
                  tokens[idx].attrPush(['data-file-download', '']); // add new attribute
                  tokens[idx].attrPush(['data-file-url', hrefFilter.url]); // add new attribute
              }

              if (!validateHref(hrefFilter.url, {anchor: tokens[idx].anchor}))  {
                  tokens[idx].attrs[hrefIndex][1] = '#';
              }

              // If you are sure other plugins can't add `target` - drop check below
              var aIndex = tokens[idx].attrIndex('target');

              if (aIndex < 0) {
                  tokens[idx].attrPush(['target', '_blank']); // add new attribute
              } else if(!tokens[idx].attrs[aIndex][1]) {
                  tokens[idx].attrs[aIndex][1] = '_blank';    // replace value of existing attr
              }

              tokens[idx].attrPush(['rel', DEFAULT_LINK_REL]);

              // pass token to default renderer.
              return defaultRender(tokens, idx, options, env, self);
          };
      },
      plugins: function (context) {
          return [linkPlugin(context)];
      }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */
  var schema$8 = {
      nodes: {
          list_item: {
              sortOrder: 800,
              content: "paragraph block*",
              defining: true,
              parseDOM: [{tag: "li"}],
              toDOM: function toDOM() {
                  return ["li", 0]
              },
              parseMarkdown: {block: "list_item"},
              toMarkdown: function (state, node) {
                  state.renderContent(node);
              }
          }
      }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  function indentListItem(context) {
      return cmdItem(sinkListItem(context.schema.nodes.list_item), {
          title: context.translate("Increase indent"),
          icon: icons$1.indent,
          sortOrder: 120
      });
  }

  function outdentListItem(context) {
      return cmdItem(liftListItem(context.schema.nodes.list_item), {
          title: context.translate("Decrease indent"),
          icon: icons$1.outdent,
          sortOrder: 110
      });
  }

  function menu$8(context) {
      return [
          {
              id: 'outdentListItem',
              node: 'list_item',
              group: 'format',
              item: outdentListItem(context)
          },
          {
              id: 'indentListItem',
              node: 'list_item',
              group: 'format',
              item: indentListItem(context)
          }
      ]
  }

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var list_item = {
      id: 'list_item',
      schema: schema$8,
      menu: function (context) {
          return menu$8(context);
      },
  };

  var schema$7 = {
      nodes: {
          mention: {
              inline: true,
              group: 'inline',
              selectable: true,
              draggable: true,
              attrs: {
                  name: { default: '' },
                  guid: { default: '' },
                  href: { default: '#' },
              },
              parseDOM: [{
                  tag: 'span[data-mention]',
                  getAttrs: function (dom) {
                      return {
                          guid: dom.getAttribute('data-mention'),
                          name: dom.textContent,
                      };
                  },
              }],
              toDOM: function toDOM(node) {
                  var attrs = {
                      'data-mention': node.attrs.guid,
                      contentEditable: 'false',
                      style: 'display:inline-block'
                  };


                  return ['span', attrs, ['span', { style: 'display:block'}, '@'+node.attrs.name] ];
              },
              parseMarkdown: {
                  node: "mention", getAttrs: function(tok) {
                      return ({
                          name: tok.attrGet("name"),
                          guid: tok.attrGet("guid"),
                          href: tok.attrGet("href")
                      })
                  }
              },
              toMarkdown: function (state, node) {
                  var linkMark = $node(node).getMark('link');
                  if(linkMark) {
                      state.write(schema$9.marks.link.toMarkdown.close(state, linkMark));
                  }

                  var ref = node.attrs;
                  var guid = ref.guid;
                  var name = ref.name;
                  var href = ref.href;
                  state.write("["+state.esc(name)+"](mention:" + state.esc(guid) +" "+ state.quote(href)+ ")");

                  if(linkMark) {
                      state.write(schema$9.marks.link.toMarkdown.open);
                  }
              },
          }
      },
      marks: {
          mentionQuery: {
              excludes: "_",
              inclusive: true,
              parseDOM: [
                  { tag: 'span[data-mention-query]' }
              ],
              toDOM: function toDOM(node) {
                  return ['span', {
                      'data-mention-query': true,
                      style: "color: #0078D7"
                  }];
              }
          }
      }
  };

  // https://github.com/ProseMirror/prosemirror/issues/262
  var objectReplacementCharacter = '\ufffc';

  var mentionRule = function(schema) {
      return new InputRule(new RegExp('(^|[\\s\(' + objectReplacementCharacter + '])@$'), function (state, match, start, end) {
          var mark = schema.mark('mentionQuery');
          var mentionText = schema.text('@', [mark]);

          // Prevents an error log when using IME
          if(hasMark(state.selection.$anchor.nodeBefore, mark)) {
              return;
          }

          start = start + (match[0].length -1);

          return state.tr
              .removeMark(0, state.doc.nodeSize -2, mark)
              .setSelection(TextSelection.create(state.doc,  start, end))
              .replaceSelectionWith(mentionText, false);
      })
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var MentionState = function MentionState(state, options) {
      this.state = state;
      this.provider = options.provider;
      this.reset();
  };

  MentionState.prototype.findQueryNode = function findQueryNode () {
      return $(this.view.dom).find('[data-mention-query]');
  };

  MentionState.prototype.update = function update (state, view) {
      this.view = view;
      this.state = state;
      var ref = state.schema.marks;
          var mentionQuery = ref.mentionQuery;
      var doc = state.doc;
          var selection = state.selection;
      var $from = selection.$from;
          var from = selection.from;
          var to = selection.to;

      this.active = doc.rangeHasMark(from - 1, to, mentionQuery);

      if (!this.active) {
          return this.reset();
      }

      var $query = this.findQueryNode();
      if(endsWith($query.text(), '  ')) {
          view.dispatch(this.state.tr.removeMark(0, this.state.doc.nodeSize -2, mentionQuery));
          return this.reset();
      }

      var $pos = doc.resolve(from - 1);

      this.queryMark = {
          start: $pos.path[$pos.path.length - 1],
          end: to
      };

      if(!$from.nodeBefore || !$from.nodeBefore.text) {
          return;
      }

      var query = $from.nodeBefore.text.substr(1);

      if(query != this.query) {
          this.query = query;
          this.provider.query(this, $query[0]);
      }
  };

  MentionState.prototype.reset = function reset () {
      var ref = this.state.schema.marks;
          var mentionQuery = ref.mentionQuery;

      if(this.state.storedMarks && this.state.storedMarks.length) {
          this.state.storedMarks = mentionQuery.removeFromSet(this.state.storedMarks);
      }
      //this.state.storedMarks = [];
      this.active = false;
      this.query = null;
      this.provider.reset();
  };

  MentionState.prototype.addMention = function addMention (item) {
      if(!item || !item.name || !item.guid) {
          this.view.dispatch(this.state.tr.removeMark(0, this.state.doc.nodeSize -2, mentionQuery));
          this.reset();
          return;
      }
      var ref = this.state.schema.nodes;
          var mention = ref.mention;
      var ref$1 = this.state.schema.marks;
          var mentionQuery = ref$1.mentionQuery;

      var nodes = [mention.create({
          name: item.name,
          guid: item.guid,
          href: item.link
      }, null), this.state.schema.text(' ')];


      var tr = this.state.tr
          .removeMark(0, this.state.doc.nodeSize -2, mentionQuery)
          .replaceWith(this.queryMark.start, this.queryMark.end, nodes);

      if(isChromeWithSelectionBug) {
          document.getSelection().empty();
      }

      this.view.dispatch(tr);
      this.view.focus();
  };

  var endsWith = function(string, suffix) {
      return string.indexOf(suffix, string.length - suffix.length) !== -1;
  };

  var pluginKey$1 = new PluginKey('mention');

  var mentionPlugin = function (context) {
      return new Plugin({
          state: {
              init: function init(config, state) {
                  return new MentionState(state, context.options.mention);
              },
              apply: function apply(tr, prevPluginState, oldState, newState) {
                  return prevPluginState;
              }
          },
          key: pluginKey$1,
          view: function (view) {
              var mentionState = pluginKey$1.getState(view.state);

              return {
                  update: function update(view, prevState) {
                      mentionState.update(view.state, view);
                  },
                  destroy: function destroy() {}
              };
          },
          appendTransaction: function (transactions, oldState, newState) {
              return $node(newState.doc).find('mention').mark('link').removeMark('link', newState);
          }
      });
  };

  var keymap = function () {
      var result = {};
      result['ArrowUp'] = function (state, dispatch) {
          var mentionState = pluginKey$1.getState(state);

          if(mentionState.active) {
              mentionState.provider.prev();
              return true;
          }

          return false;
      };

      result['ArrowDown'] = function (state, dispatch) {
          var mentionState = pluginKey$1.getState(state);

          if(mentionState && mentionState.active) {
              mentionState.provider.next();
              return true;
          }

          return false;
      };

      result['Enter'] = function (state, dispatch) {
          var mentionState = pluginKey$1.getState(state);

          if(mentionState && mentionState.active) {
              mentionState.provider.select();
              return true;
          }

          return false;
      };

      result['Escape'] = function (state, dispatch) {
          var mentionState  = pluginKey$1.getState(state);

          if(mentionState  && mentionState.active) {
              mentionState.provider.reset();
              return true;
          }

          return false;
      };

      return result;
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var inst = new markdownIt();
  var isSpace$1 = inst.utils.isSpace;

  function createLinkExtension(id, options) {

      options = options || {};

      options.node = options.node || 'a';
      options.hrefAttr = options.hrefAttr || 'href';
      options.titleAttr = options.titleAttr || 'title';
      options.labelAttr = options.labelAttr || 'label';

      return function (state, silent) {
          var attrs,
              code,
              labelEnd,
              labelStart,
              label,
              pos,
              res,
              title,
              token,
              href = '',
              prefix = id+':';
              state.pos;
              var max = state.posMax,
              start = state.pos;

          if (state.src.charCodeAt(state.pos) !== 0x5B/* [ */) {
              return false;
          }

          labelStart = state.pos + 1;
          labelEnd = state.md.helpers.parseLinkLabel(state, state.pos, true);

          // parser failed to find ']', so it's not a valid link
          if (labelEnd < 0) {
              return false;
          }

          pos = labelEnd + 1;
          if (pos < max && state.src.charCodeAt(pos) === 0x28/* ( */) {
              //
              // Inline link
              //

              // [link](  <id>:<href>  "title"  )
              //        ^^ skipping these spaces
              pos++;
              for (; pos < max; pos++) {
                  code = state.src.charCodeAt(pos);
                  if (!isSpace$1(code) && code !== 0x0A) {
                      break;
                  }
              }
              if (pos >= max) {
                  return false;
              }

              // [link](  <id>:<href>  "title"  )
              //          ^^^^ parsing prefix
              for(var i = 0;i < prefix.length; i++) {
                  if(state.src.charAt(pos++) !== prefix.charAt(i)) {
                      return false;
                  }
              }

              // [link](  <id>:<href>  "title"  )
              //               ^^^^ parsing href
              res = state.md.helpers.parseLinkDestination(state.src, pos, state.posMax);

              if (res.ok) {
                  href = state.md.normalizeLink(res.str);
                  if (state.md.validateLink(href)) {
                      pos = res.pos;
                  } else {
                      href = '';
                  }
              }

              // [link](  <id>:<href>  "title"  )
              //                     ^^ skipping these spaces
              start = pos;
              for (; pos < max; pos++) {
                  code = state.src.charCodeAt(pos);
                  if (!isSpace$1(code) && code !== 0x0A) {
                      break;
                  }
              }

              // [link](  <id>:<href>  "title"  )
              //                       ^^^^^^^ parsing link title
              res = state.md.helpers.parseLinkTitle(state.src, pos, state.posMax);
              if (pos < max && start !== pos && res.ok) {
                  title = res.str;
                  pos = res.pos;

                  // [link](  <id>:<href>  "title"  )
                  //                              ^^ skipping these spaces
                  for (; pos < max; pos++) {
                      code = state.src.charCodeAt(pos);
                      if (!isSpace$1(code) && code !== 0x0A) {
                          break;
                      }
                  }
              } else {
                  title = '';
              }

              if (pos >= max || state.src.charCodeAt(pos) !== 0x29/* ) */) {
                  // parsing a valid shortcut link failed
                  return false;
              }
              pos++;
          } else {
              return false;
          }

          //
          // We found the end of the link, and know for a fact it's a valid link;
          // so all that's left to do is to call tokenizer.
          //
          if (!silent) {
              state.pos = labelStart;
              state.posMax = labelEnd;

              label = state.src.substring(labelStart, labelEnd);

              token = state.push(id, options.node, 0);
              token.attrs = attrs = [[options.hrefAttr, href]];

              if(label) {
                  attrs.push([options.labelAttr, label]);
              }

              if (title) {
                  attrs.push([options.titleAttr, title]);
              }

              while (state.src.charCodeAt(state.pos) !== 0x29/* ) */) {
                  state.pos++;
              }

          // NOTE linkExtensions currently do not support inline formatting:
          // TODO: make _open, _close behavior optional in order to support inline label format state.md.inline.tokenize(state);
          }

          state.pos = pos;
          state.posMax = max;
          return true;
      };
  }

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var mention$1 = {
      id: 'mention',
      schema: schema$7,
      plugins: function (context) {
          if(!context.options.mention || !context.options.mention.provider) {
              return [];
          }
          return [
              mentionPlugin(context)
          ]
      },
      inputRules: function (schema) {return [mentionRule(schema)]},
      keymap: function (context) { return keymap()},
      registerMarkdownIt: function (markdownIt) {
          // [name](mention:guid "href")
          markdownIt.inline.ruler.before('link','mention', createLinkExtension('mention', {
              labelAttr: 'name',
              hrefAttr : 'guid',
              titleAttr: 'href'
          }));

          markdownIt.renderer.rules.mention = function(token, idx) {
              var mentioning = token[idx];
              var href = mentioning.attrGet('href');
              var guid = mentioning.attrGet('guid');
              var label = '@'+mentioning.attrGet('name');


              if(!validateHref(href, {relative: true})) {
                  return buildLink('#',{'class':'not-found'}, label);
              }

              return buildLink(href, {'data-contentcontainer-guid': guid}, label, false);
          };


      }
  };

  var oembed$1 = {
      attrs: {
          href: {},
      },
      marks: "",
      atom: true,
      draggable: true,
      inline: true,
      group: "inline",
      parseDOM: [{
          tag: "[data-oembed]", getAttrs: function getAttrs(dom) {

              return {
                  href: dom.getAttribute("data-oembed")
              };
          }
      }],
      toDOM: function (node) {
          var $oembed = humhub.require('oembed').get(node.attrs.href);

          if ($oembed && $oembed.length) {
              return $oembed.show()[0];
          }

          return $(buildLink(node.attrs.href, {'class': 'not-found'}))[0];
      },
      parseMarkdown: {
          node: "oembed", getAttrs: function(tok) {
              return ({
                  href: tok.attrGet("href")
              });
          }
      },
      toMarkdown: function (state, node) {
          state.write('['+node.attrs.href+'](oembed:'+node.attrs.href+')');
      }
  };

  var schema$6 = {
      nodes: {
          oembed: oembed$1
      }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var oembed_plugin = createLinkExtension('oembed', {node : 'div'});

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var oembed = {
      id: 'oembed',
      schema: schema$6,
      init: function (context, isEdit) {
          if(!isEdit) {
              return;
          }

          context.event.on('linkified', function (evt, urls) {
              var doc = context.editor.view.state.doc;
              if($node(doc).find('oembed').size() >= context.getPluginOption('oembed', 'max', 5)) {
                  return;
              }

              humhub.require('oembed').load(urls).then(function (result) {
                  $.each(result, function(url, oembed) {
                      var $links = $node(context.editor.view.state.doc).find().mark('link').where(function (node) {
                          return $node(node).getMark('link').attrs.href === url;
                      });

                      $links.replaceWith(context.schema.nodes.oembed.create({href:url}), context.editor.view);

                      // We only allow a single oembed per copy/paste
                      return false;
                  });
              });

          });
      },
      registerMarkdownIt: function (markdownIt) {
          markdownIt.inline.ruler.before('link','oembed', oembed_plugin);

          markdownIt.renderer.rules.oembed = function(token, idx) {
              var oembed = token[idx];
              var href = oembed.attrGet('href');
              var $oembed = humhub
                  .require('oembed')
                  .get(href);

              if(!$oembed || !$oembed.length) {
                  return buildLink(href);
              }

              return $('<div>').append($oembed).html();
          };
      }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var focusKey = new PluginKey('focus');

  var focusPlugin = function (context) {
      return new Plugin({
          key: focusKey,
          state: {
              init: function init() {
                  return false;
              },
              apply: function apply(transaction, prevFocused) {
                  var focused = transaction.getMeta(focusKey);
                  if (typeof focused === 'boolean') {
                      return focused;
                  }
                  return prevFocused;
              }
          },
          props: {
              handleDOMEvents: {
                  blur: function (view) {
                      view.dispatch(view.state.tr.setMeta(focusKey, false));
                      return false;

                  },
                  focus: function (view, event) {
                      view.dispatch(view.state.tr.setMeta(focusKey, true));
                      return false;

                  }
              }
          }
      });
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var focus = {
      id: 'focus',
      plugins: function (context) {
          return [
              focusPlugin()
          ];
      },
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */
  var schema$5 = {
      nodes: {
          ordered_list: {
              sortOrder: 600,
              content: "list_item+",
              group: "block",
              attrs: {order: {default: 1}, tight: {default: true}},
              parseDOM: [{
                  tag: "ol", getAttrs: function getAttrs(dom) {
                      return {
                          order: dom.hasAttribute("start") ? +dom.getAttribute("start") : 1,
                          tight: dom.hasAttribute("data-tight")
                      }
                  }
              }],
              toDOM: function toDOM(node) {
                  return ["ol", {
                      start: node.attrs.order == 1 ? null : node.attrs.order,
                      "data-tight": node.attrs.tight ? "true" : null
                  }, 0]
              },
              parseMarkdown:  {
                  block: "ordered_list", getAttrs: function (tok) {
                      return ({order: +tok.attrGet("start") || 1});
                  }
              },
              toMarkdown: function (state, node) {
                  if(state.table) {
                      state.text(node.textContent);
                      return;
                  }

                  var start = node.attrs.order || 1;
                  var maxW = String(start + node.childCount - 1).length;
                  var space = state.repeat(" ", maxW + 2);
                  state.renderList(node, space, function (i) {
                      var nStr = String(start + i);
                      return state.repeat(" ", maxW - nStr.length) + nStr + ". "
                  });
              }
          }
      }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  // : (NodeType) → InputRule
  // Given a list node type, returns an input rule that turns a number
  // followed by a dot at the start of a textblock into an ordered list.
  var orderedListRule = function(schema) {
      return wrappingInputRule(/^(\d+)\.\s$/, schema.nodes.ordered_list, function (match) { return ({order: +match[1]}); },
          function (match, node) { return node.childCount + node.attrs.order == +match[1]; })
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  function wrapOrderedList(context) {
      return cmdItem(wrapInList$1(context.schema.nodes.ordered_list), {
          title: context.translate("Wrap in ordered list"),
          icon: icons$1.orderedList,
          sortOrder: 200
      });
  }

  function menu$7(context) {
      return [
          {
              id: 'wrapOrderedList',
              node: 'ordered_list',
              group: 'format',
              item: wrapOrderedList(context)
          }
      ]
  }

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var ordered_list = {
      id: 'ordered_list',
      menu: function (context) { return menu$7(context); },
      schema: schema$5,
      inputRules: function (schema) {return [orderedListRule(schema)]}
  };

  var schema$4 = {
      nodes: {
          paragraph:  {
              sortOrder: 100,
              content: "inline*",
              group: "block",
              parseDOM: [{tag: "p"}],
              toDOM: function toDOM() {
                  return ["p", 0]
              },
              parseMarkdown: {block: "paragraph"},
              toMarkdown: function (state, node, parent) {

                  state.renderInline(node);

                  if(!state.table) {
                      state.closeBlock(node);
                  } else if(node.content && node.content.size && parent.lastChild !== node) {
                      state.write('<br><br>');
                  }
              }
          }
      }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  function makeParagraph(context) {
      return blockTypeItem$1(context.schema.nodes.paragraph, {
          title: context.translate("Change to paragraph"),
          label: context.translate("Paragraph")
      })
  }

  function menu$6(context) {
      return [
          {
              id: 'makeParagraph',
              node: 'paragraph',
              group: 'types',
              item: makeParagraph(context)
          }
      ]
  }

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */
  var paragraph = {
      id: 'paragraph',
      schema: schema$4,
      menu: function (context) { return menu$6(context); }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var schema$3 = {
      marks: {
          strikethrough: {
              parseDOM: [{tag: "s"}],
              toDOM: function () {
                  return ["s"]
              },
              parseMarkdown: {s: {mark: "strikethrough"}},
              toMarkdown: {open: "~~", close: "~~", mixable: true, expelEnclosingWhitespace: true}
          }
      }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */


  function markStrikethrough(context) {
      return markItem(context.schema.marks.strikethrough, {
          title: context.translate("Toggle strikethrough"),
          icon: icons$1.strikethrough,
          sortOrder: 300
      });
  }

  function menu$5(context) {
      return [
          {
              id: 'markStrikethrough',
              mark: 'strikethrough',
              group: 'marks',
              item: markStrikethrough(context)
          }
      ]
  }

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var strikethrough = {
      id: 'strikethrough',
      schema: schema$3,
      menu: function (context) { return menu$5(context); }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var schema$2 = {
      marks: {
          strong: {
              sortOrder: 200,
              parseDOM: [{tag: "b"}, {tag: "strong"},
                  {
                      style: "font-weight", getAttrs: function (value) {
                      return /^(bold(er)?|[5-9]\d{2,})$/.test(value) && null;
                  }
                  }],
              toDOM: function () {
                  return ["strong"]
              },
              parseMarkdown: {mark: "strong"},
              toMarkdown: {open: "**", close: "**", mixable: true, expelEnclosingWhitespace: true}
          }
      }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */


  function markStrong(context) {
      return markItem(context.schema.marks.strong, {
          title: context.translate("Toggle strong style"),
          icon: icons$1.strong,
          sortOrder: 100
      });
  }

  function menu$4(context) {
      return [
          {
              id: 'markStrong',
              mark: 'strong',
              group: 'marks',
              item: markStrong(context)
          }
      ]
  }

  function findDiffStart(a, b, pos) {
    for (var i = 0;; i++) {
      if (i == a.childCount || i == b.childCount)
        { return a.childCount == b.childCount ? null : pos }

      var childA = a.child(i), childB = b.child(i);
      if (childA == childB) { pos += childA.nodeSize; continue }

      if (!childA.sameMarkup(childB)) { return pos }

      if (childA.isText && childA.text != childB.text) {
        for (var j = 0; childA.text[j] == childB.text[j]; j++)
          { pos++; }
        return pos
      }
      if (childA.content.size || childB.content.size) {
        var inner = findDiffStart(childA.content, childB.content, pos + 1);
        if (inner != null) { return inner }
      }
      pos += childA.nodeSize;
    }
  }

  function findDiffEnd(a, b, posA, posB) {
    for (var iA = a.childCount, iB = b.childCount;;) {
      if (iA == 0 || iB == 0)
        { return iA == iB ? null : {a: posA, b: posB} }

      var childA = a.child(--iA), childB = b.child(--iB), size = childA.nodeSize;
      if (childA == childB) {
        posA -= size; posB -= size;
        continue
      }

      if (!childA.sameMarkup(childB)) { return {a: posA, b: posB} }

      if (childA.isText && childA.text != childB.text) {
        var same = 0, minSize = Math.min(childA.text.length, childB.text.length);
        while (same < minSize && childA.text[childA.text.length - same - 1] == childB.text[childB.text.length - same - 1]) {
          same++; posA--; posB--;
        }
        return {a: posA, b: posB}
      }
      if (childA.content.size || childB.content.size) {
        var inner = findDiffEnd(childA.content, childB.content, posA - 1, posB - 1);
        if (inner) { return inner }
      }
      posA -= size; posB -= size;
    }
  }

  // ::- A fragment represents a node's collection of child nodes.
  //
  // Like nodes, fragments are persistent data structures, and you
  // should not mutate them or their content. Rather, you create new
  // instances whenever needed. The API tries to make this easy.
  var Fragment = function Fragment(content, size) {
    this.content = content;
    // :: number
    // The size of the fragment, which is the total of the size of its
    // content nodes.
    this.size = size || 0;
    if (size == null) { for (var i = 0; i < content.length; i++)
      { this.size += content[i].nodeSize; } }
  };

  var prototypeAccessors$3 = { firstChild: { configurable: true },lastChild: { configurable: true },childCount: { configurable: true } };

  // :: (number, number, (node: Node, start: number, parent: Node, index: number) → ?bool, ?number)
  // Invoke a callback for all descendant nodes between the given two
  // positions (relative to start of this fragment). Doesn't descend
  // into a node when the callback returns `false`.
  Fragment.prototype.nodesBetween = function nodesBetween (from, to, f, nodeStart, parent) {
      if ( nodeStart === void 0 ) nodeStart = 0;

    for (var i = 0, pos = 0; pos < to; i++) {
      var child = this.content[i], end = pos + child.nodeSize;
      if (end > from && f(child, nodeStart + pos, parent, i) !== false && child.content.size) {
        var start = pos + 1;
        child.nodesBetween(Math.max(0, from - start),
                           Math.min(child.content.size, to - start),
                           f, nodeStart + start);
      }
      pos = end;
    }
  };

  // :: ((node: Node, pos: number, parent: Node) → ?bool)
  // Call the given callback for every descendant node. The callback
  // may return `false` to prevent traversal of a given node's children.
  Fragment.prototype.descendants = function descendants (f) {
    this.nodesBetween(0, this.size, f);
  };

  // :: (number, number, ?string, ?string) → string
  // Extract the text between `from` and `to`. See the same method on
  // [`Node`](#model.Node.textBetween).
  Fragment.prototype.textBetween = function textBetween (from, to, blockSeparator, leafText) {
    var text = "", separated = true;
    this.nodesBetween(from, to, function (node, pos) {
      if (node.isText) {
        text += node.text.slice(Math.max(from, pos) - pos, to - pos);
        separated = !blockSeparator;
      } else if (node.isLeaf && leafText) {
        text += leafText;
        separated = !blockSeparator;
      } else if (!separated && node.isBlock) {
        text += blockSeparator;
        separated = true;
      }
    }, 0);
    return text
  };

  // :: (Fragment) → Fragment
  // Create a new fragment containing the combined content of this
  // fragment and the other.
  Fragment.prototype.append = function append (other) {
    if (!other.size) { return this }
    if (!this.size) { return other }
    var last = this.lastChild, first = other.firstChild, content = this.content.slice(), i = 0;
    if (last.isText && last.sameMarkup(first)) {
      content[content.length - 1] = last.withText(last.text + first.text);
      i = 1;
    }
    for (; i < other.content.length; i++) { content.push(other.content[i]); }
    return new Fragment(content, this.size + other.size)
  };

  // :: (number, ?number) → Fragment
  // Cut out the sub-fragment between the two given positions.
  Fragment.prototype.cut = function cut (from, to) {
    if (to == null) { to = this.size; }
    if (from == 0 && to == this.size) { return this }
    var result = [], size = 0;
    if (to > from) { for (var i = 0, pos = 0; pos < to; i++) {
      var child = this.content[i], end = pos + child.nodeSize;
      if (end > from) {
        if (pos < from || end > to) {
          if (child.isText)
            { child = child.cut(Math.max(0, from - pos), Math.min(child.text.length, to - pos)); }
          else
            { child = child.cut(Math.max(0, from - pos - 1), Math.min(child.content.size, to - pos - 1)); }
        }
        result.push(child);
        size += child.nodeSize;
      }
      pos = end;
    } }
    return new Fragment(result, size)
  };

  Fragment.prototype.cutByIndex = function cutByIndex (from, to) {
    if (from == to) { return Fragment.empty }
    if (from == 0 && to == this.content.length) { return this }
    return new Fragment(this.content.slice(from, to))
  };

  // :: (number, Node) → Fragment
  // Create a new fragment in which the node at the given index is
  // replaced by the given node.
  Fragment.prototype.replaceChild = function replaceChild (index, node) {
    var current = this.content[index];
    if (current == node) { return this }
    var copy = this.content.slice();
    var size = this.size + node.nodeSize - current.nodeSize;
    copy[index] = node;
    return new Fragment(copy, size)
  };

  // : (Node) → Fragment
  // Create a new fragment by prepending the given node to this
  // fragment.
  Fragment.prototype.addToStart = function addToStart (node) {
    return new Fragment([node].concat(this.content), this.size + node.nodeSize)
  };

  // : (Node) → Fragment
  // Create a new fragment by appending the given node to this
  // fragment.
  Fragment.prototype.addToEnd = function addToEnd (node) {
    return new Fragment(this.content.concat(node), this.size + node.nodeSize)
  };

  // :: (Fragment) → bool
  // Compare this fragment to another one.
  Fragment.prototype.eq = function eq (other) {
    if (this.content.length != other.content.length) { return false }
    for (var i = 0; i < this.content.length; i++)
      { if (!this.content[i].eq(other.content[i])) { return false } }
    return true
  };

  // :: ?Node
  // The first child of the fragment, or `null` if it is empty.
  prototypeAccessors$3.firstChild.get = function () { return this.content.length ? this.content[0] : null };

  // :: ?Node
  // The last child of the fragment, or `null` if it is empty.
  prototypeAccessors$3.lastChild.get = function () { return this.content.length ? this.content[this.content.length - 1] : null };

  // :: number
  // The number of child nodes in this fragment.
  prototypeAccessors$3.childCount.get = function () { return this.content.length };

  // :: (number) → Node
  // Get the child node at the given index. Raise an error when the
  // index is out of range.
  Fragment.prototype.child = function child (index) {
    var found = this.content[index];
    if (!found) { throw new RangeError("Index " + index + " out of range for " + this) }
    return found
  };

  // :: (number) → ?Node
  // Get the child node at the given index, if it exists.
  Fragment.prototype.maybeChild = function maybeChild (index) {
    return this.content[index]
  };

  // :: ((node: Node, offset: number, index: number))
  // Call `f` for every child node, passing the node, its offset
  // into this parent node, and its index.
  Fragment.prototype.forEach = function forEach (f) {
    for (var i = 0, p = 0; i < this.content.length; i++) {
      var child = this.content[i];
      f(child, p, i);
      p += child.nodeSize;
    }
  };

  // :: (Fragment) → ?number
  // Find the first position at which this fragment and another
  // fragment differ, or `null` if they are the same.
  Fragment.prototype.findDiffStart = function findDiffStart$1 (other, pos) {
      if ( pos === void 0 ) pos = 0;

    return findDiffStart(this, other, pos)
  };

  // :: (Fragment) → ?{a: number, b: number}
  // Find the first position, searching from the end, at which this
  // fragment and the given fragment differ, or `null` if they are the
  // same. Since this position will not be the same in both nodes, an
  // object with two separate positions is returned.
  Fragment.prototype.findDiffEnd = function findDiffEnd$1 (other, pos, otherPos) {
      if ( pos === void 0 ) pos = this.size;
      if ( otherPos === void 0 ) otherPos = other.size;

    return findDiffEnd(this, other, pos, otherPos)
  };

  // : (number, ?number) → {index: number, offset: number}
  // Find the index and inner offset corresponding to a given relative
  // position in this fragment. The result object will be reused
  // (overwritten) the next time the function is called. (Not public.)
  Fragment.prototype.findIndex = function findIndex (pos, round) {
      if ( round === void 0 ) round = -1;

    if (pos == 0) { return retIndex(0, pos) }
    if (pos == this.size) { return retIndex(this.content.length, pos) }
    if (pos > this.size || pos < 0) { throw new RangeError(("Position " + pos + " outside of fragment (" + (this) + ")")) }
    for (var i = 0, curPos = 0;; i++) {
      var cur = this.child(i), end = curPos + cur.nodeSize;
      if (end >= pos) {
        if (end == pos || round > 0) { return retIndex(i + 1, end) }
        return retIndex(i, curPos)
      }
      curPos = end;
    }
  };

  // :: () → string
  // Return a debugging string that describes this fragment.
  Fragment.prototype.toString = function toString () { return "<" + this.toStringInner() + ">" };

  Fragment.prototype.toStringInner = function toStringInner () { return this.content.join(", ") };

  // :: () → ?Object
  // Create a JSON-serializeable representation of this fragment.
  Fragment.prototype.toJSON = function toJSON () {
    return this.content.length ? this.content.map(function (n) { return n.toJSON(); }) : null
  };

  // :: (Schema, ?Object) → Fragment
  // Deserialize a fragment from its JSON representation.
  Fragment.fromJSON = function fromJSON (schema, value) {
    if (!value) { return Fragment.empty }
    if (!Array.isArray(value)) { throw new RangeError("Invalid input for Fragment.fromJSON") }
    return new Fragment(value.map(schema.nodeFromJSON))
  };

  // :: ([Node]) → Fragment
  // Build a fragment from an array of nodes. Ensures that adjacent
  // text nodes with the same marks are joined together.
  Fragment.fromArray = function fromArray (array) {
    if (!array.length) { return Fragment.empty }
    var joined, size = 0;
    for (var i = 0; i < array.length; i++) {
      var node = array[i];
      size += node.nodeSize;
      if (i && node.isText && array[i - 1].sameMarkup(node)) {
        if (!joined) { joined = array.slice(0, i); }
        joined[joined.length - 1] = node.withText(joined[joined.length - 1].text + node.text);
      } else if (joined) {
        joined.push(node);
      }
    }
    return new Fragment(joined || array, size)
  };

  // :: (?union<Fragment, Node, [Node]>) → Fragment
  // Create a fragment from something that can be interpreted as a set
  // of nodes. For `null`, it returns the empty fragment. For a
  // fragment, the fragment itself. For a node or array of nodes, a
  // fragment containing those nodes.
  Fragment.from = function from (nodes) {
    if (!nodes) { return Fragment.empty }
    if (nodes instanceof Fragment) { return nodes }
    if (Array.isArray(nodes)) { return this.fromArray(nodes) }
    if (nodes.attrs) { return new Fragment([nodes], nodes.nodeSize) }
    throw new RangeError("Can not convert " + nodes + " to a Fragment" +
                         (nodes.nodesBetween ? " (looks like multiple versions of prosemirror-model were loaded)" : ""))
  };

  Object.defineProperties( Fragment.prototype, prototypeAccessors$3 );

  var found = {index: 0, offset: 0};
  function retIndex(index, offset) {
    found.index = index;
    found.offset = offset;
    return found
  }

  // :: Fragment
  // An empty fragment. Intended to be reused whenever a node doesn't
  // contain anything (rather than allocating a new empty fragment for
  // each leaf node).
  Fragment.empty = new Fragment([], 0);

  function compareDeep(a, b) {
    if (a === b) { return true }
    if (!(a && typeof a == "object") ||
        !(b && typeof b == "object")) { return false }
    var array = Array.isArray(a);
    if (Array.isArray(b) != array) { return false }
    if (array) {
      if (a.length != b.length) { return false }
      for (var i = 0; i < a.length; i++) { if (!compareDeep(a[i], b[i])) { return false } }
    } else {
      for (var p in a) { if (!(p in b) || !compareDeep(a[p], b[p])) { return false } }
      for (var p$1 in b) { if (!(p$1 in a)) { return false } }
    }
    return true
  }

  // ::- A mark is a piece of information that can be attached to a node,
  // such as it being emphasized, in code font, or a link. It has a type
  // and optionally a set of attributes that provide further information
  // (such as the target of the link). Marks are created through a
  // `Schema`, which controls which types exist and which
  // attributes they have.
  var Mark = function Mark(type, attrs) {
    // :: MarkType
    // The type of this mark.
    this.type = type;
    // :: Object
    // The attributes associated with this mark.
    this.attrs = attrs;
  };

  // :: ([Mark]) → [Mark]
  // Given a set of marks, create a new set which contains this one as
  // well, in the right position. If this mark is already in the set,
  // the set itself is returned. If any marks that are set to be
  // [exclusive](#model.MarkSpec.excludes) with this mark are present,
  // those are replaced by this one.
  Mark.prototype.addToSet = function addToSet (set) {
    var copy, placed = false;
    for (var i = 0; i < set.length; i++) {
      var other = set[i];
      if (this.eq(other)) { return set }
      if (this.type.excludes(other.type)) {
        if (!copy) { copy = set.slice(0, i); }
      } else if (other.type.excludes(this.type)) {
        return set
      } else {
        if (!placed && other.type.rank > this.type.rank) {
          if (!copy) { copy = set.slice(0, i); }
          copy.push(this);
          placed = true;
        }
        if (copy) { copy.push(other); }
      }
    }
    if (!copy) { copy = set.slice(); }
    if (!placed) { copy.push(this); }
    return copy
  };

  // :: ([Mark]) → [Mark]
  // Remove this mark from the given set, returning a new set. If this
  // mark is not in the set, the set itself is returned.
  Mark.prototype.removeFromSet = function removeFromSet (set) {
    for (var i = 0; i < set.length; i++)
      { if (this.eq(set[i]))
        { return set.slice(0, i).concat(set.slice(i + 1)) } }
    return set
  };

  // :: ([Mark]) → bool
  // Test whether this mark is in the given set of marks.
  Mark.prototype.isInSet = function isInSet (set) {
    for (var i = 0; i < set.length; i++)
      { if (this.eq(set[i])) { return true } }
    return false
  };

  // :: (Mark) → bool
  // Test whether this mark has the same type and attributes as
  // another mark.
  Mark.prototype.eq = function eq (other) {
    return this == other ||
      (this.type == other.type && compareDeep(this.attrs, other.attrs))
  };

  // :: () → Object
  // Convert this mark to a JSON-serializeable representation.
  Mark.prototype.toJSON = function toJSON () {
    var obj = {type: this.type.name};
    for (var _ in this.attrs) {
      obj.attrs = this.attrs;
      break
    }
    return obj
  };

  // :: (Schema, Object) → Mark
  Mark.fromJSON = function fromJSON (schema, json) {
    if (!json) { throw new RangeError("Invalid input for Mark.fromJSON") }
    var type = schema.marks[json.type];
    if (!type) { throw new RangeError(("There is no mark type " + (json.type) + " in this schema")) }
    return type.create(json.attrs)
  };

  // :: ([Mark], [Mark]) → bool
  // Test whether two sets of marks are identical.
  Mark.sameSet = function sameSet (a, b) {
    if (a == b) { return true }
    if (a.length != b.length) { return false }
    for (var i = 0; i < a.length; i++)
      { if (!a[i].eq(b[i])) { return false } }
    return true
  };

  // :: (?union<Mark, [Mark]>) → [Mark]
  // Create a properly sorted mark set from null, a single mark, or an
  // unsorted array of marks.
  Mark.setFrom = function setFrom (marks) {
    if (!marks || marks.length == 0) { return Mark.none }
    if (marks instanceof Mark) { return [marks] }
    var copy = marks.slice();
    copy.sort(function (a, b) { return a.type.rank - b.type.rank; });
    return copy
  };

  // :: [Mark] The empty set of marks.
  Mark.none = [];

  // ReplaceError:: class extends Error
  // Error type raised by [`Node.replace`](#model.Node.replace) when
  // given an invalid replacement.

  function ReplaceError(message) {
    var err = Error.call(this, message);
    err.__proto__ = ReplaceError.prototype;
    return err
  }

  ReplaceError.prototype = Object.create(Error.prototype);
  ReplaceError.prototype.constructor = ReplaceError;
  ReplaceError.prototype.name = "ReplaceError";

  // ::- A slice represents a piece cut out of a larger document. It
  // stores not only a fragment, but also the depth up to which nodes on
  // both side are ‘open’ (cut through).
  var Slice = function Slice(content, openStart, openEnd) {
    // :: Fragment The slice's content.
    this.content = content;
    // :: number The open depth at the start.
    this.openStart = openStart;
    // :: number The open depth at the end.
    this.openEnd = openEnd;
  };

  var prototypeAccessors$2 = { size: { configurable: true } };

  // :: number
  // The size this slice would add when inserted into a document.
  prototypeAccessors$2.size.get = function () {
    return this.content.size - this.openStart - this.openEnd
  };

  Slice.prototype.insertAt = function insertAt (pos, fragment) {
    var content = insertInto(this.content, pos + this.openStart, fragment, null);
    return content && new Slice(content, this.openStart, this.openEnd)
  };

  Slice.prototype.removeBetween = function removeBetween (from, to) {
    return new Slice(removeRange(this.content, from + this.openStart, to + this.openStart), this.openStart, this.openEnd)
  };

  // :: (Slice) → bool
  // Tests whether this slice is equal to another slice.
  Slice.prototype.eq = function eq (other) {
    return this.content.eq(other.content) && this.openStart == other.openStart && this.openEnd == other.openEnd
  };

  Slice.prototype.toString = function toString () {
    return this.content + "(" + this.openStart + "," + this.openEnd + ")"
  };

  // :: () → ?Object
  // Convert a slice to a JSON-serializable representation.
  Slice.prototype.toJSON = function toJSON () {
    if (!this.content.size) { return null }
    var json = {content: this.content.toJSON()};
    if (this.openStart > 0) { json.openStart = this.openStart; }
    if (this.openEnd > 0) { json.openEnd = this.openEnd; }
    return json
  };

  // :: (Schema, ?Object) → Slice
  // Deserialize a slice from its JSON representation.
  Slice.fromJSON = function fromJSON (schema, json) {
    if (!json) { return Slice.empty }
    var openStart = json.openStart || 0, openEnd = json.openEnd || 0;
    if (typeof openStart != "number" || typeof openEnd != "number")
      { throw new RangeError("Invalid input for Slice.fromJSON") }
    return new Slice(Fragment.fromJSON(schema, json.content), openStart, openEnd)
  };

  // :: (Fragment, ?bool) → Slice
  // Create a slice from a fragment by taking the maximum possible
  // open value on both side of the fragment.
  Slice.maxOpen = function maxOpen (fragment, openIsolating) {
      if ( openIsolating === void 0 ) openIsolating=true;

    var openStart = 0, openEnd = 0;
    for (var n = fragment.firstChild; n && !n.isLeaf && (openIsolating || !n.type.spec.isolating); n = n.firstChild) { openStart++; }
    for (var n$1 = fragment.lastChild; n$1 && !n$1.isLeaf && (openIsolating || !n$1.type.spec.isolating); n$1 = n$1.lastChild) { openEnd++; }
    return new Slice(fragment, openStart, openEnd)
  };

  Object.defineProperties( Slice.prototype, prototypeAccessors$2 );

  function removeRange(content, from, to) {
    var ref = content.findIndex(from);
    var index = ref.index;
    var offset = ref.offset;
    var child = content.maybeChild(index);
    var ref$1 = content.findIndex(to);
    var indexTo = ref$1.index;
    var offsetTo = ref$1.offset;
    if (offset == from || child.isText) {
      if (offsetTo != to && !content.child(indexTo).isText) { throw new RangeError("Removing non-flat range") }
      return content.cut(0, from).append(content.cut(to))
    }
    if (index != indexTo) { throw new RangeError("Removing non-flat range") }
    return content.replaceChild(index, child.copy(removeRange(child.content, from - offset - 1, to - offset - 1)))
  }

  function insertInto(content, dist, insert, parent) {
    var ref = content.findIndex(dist);
    var index = ref.index;
    var offset = ref.offset;
    var child = content.maybeChild(index);
    if (offset == dist || child.isText) {
      if (parent && !parent.canReplace(index, index, insert)) { return null }
      return content.cut(0, dist).append(insert).append(content.cut(dist))
    }
    var inner = insertInto(child.content, dist - offset - 1, insert);
    return inner && content.replaceChild(index, child.copy(inner))
  }

  // :: Slice
  // The empty slice.
  Slice.empty = new Slice(Fragment.empty, 0, 0);

  function replace($from, $to, slice) {
    if (slice.openStart > $from.depth)
      { throw new ReplaceError("Inserted content deeper than insertion position") }
    if ($from.depth - slice.openStart != $to.depth - slice.openEnd)
      { throw new ReplaceError("Inconsistent open depths") }
    return replaceOuter($from, $to, slice, 0)
  }

  function replaceOuter($from, $to, slice, depth) {
    var index = $from.index(depth), node = $from.node(depth);
    if (index == $to.index(depth) && depth < $from.depth - slice.openStart) {
      var inner = replaceOuter($from, $to, slice, depth + 1);
      return node.copy(node.content.replaceChild(index, inner))
    } else if (!slice.content.size) {
      return close(node, replaceTwoWay($from, $to, depth))
    } else if (!slice.openStart && !slice.openEnd && $from.depth == depth && $to.depth == depth) { // Simple, flat case
      var parent = $from.parent, content = parent.content;
      return close(parent, content.cut(0, $from.parentOffset).append(slice.content).append(content.cut($to.parentOffset)))
    } else {
      var ref = prepareSliceForReplace(slice, $from);
      var start = ref.start;
      var end = ref.end;
      return close(node, replaceThreeWay($from, start, end, $to, depth))
    }
  }

  function checkJoin(main, sub) {
    if (!sub.type.compatibleContent(main.type))
      { throw new ReplaceError("Cannot join " + sub.type.name + " onto " + main.type.name) }
  }

  function joinable($before, $after, depth) {
    var node = $before.node(depth);
    checkJoin(node, $after.node(depth));
    return node
  }

  function addNode(child, target) {
    var last = target.length - 1;
    if (last >= 0 && child.isText && child.sameMarkup(target[last]))
      { target[last] = child.withText(target[last].text + child.text); }
    else
      { target.push(child); }
  }

  function addRange($start, $end, depth, target) {
    var node = ($end || $start).node(depth);
    var startIndex = 0, endIndex = $end ? $end.index(depth) : node.childCount;
    if ($start) {
      startIndex = $start.index(depth);
      if ($start.depth > depth) {
        startIndex++;
      } else if ($start.textOffset) {
        addNode($start.nodeAfter, target);
        startIndex++;
      }
    }
    for (var i = startIndex; i < endIndex; i++) { addNode(node.child(i), target); }
    if ($end && $end.depth == depth && $end.textOffset)
      { addNode($end.nodeBefore, target); }
  }

  function close(node, content) {
    if (!node.type.validContent(content))
      { throw new ReplaceError("Invalid content for node " + node.type.name) }
    return node.copy(content)
  }

  function replaceThreeWay($from, $start, $end, $to, depth) {
    var openStart = $from.depth > depth && joinable($from, $start, depth + 1);
    var openEnd = $to.depth > depth && joinable($end, $to, depth + 1);

    var content = [];
    addRange(null, $from, depth, content);
    if (openStart && openEnd && $start.index(depth) == $end.index(depth)) {
      checkJoin(openStart, openEnd);
      addNode(close(openStart, replaceThreeWay($from, $start, $end, $to, depth + 1)), content);
    } else {
      if (openStart)
        { addNode(close(openStart, replaceTwoWay($from, $start, depth + 1)), content); }
      addRange($start, $end, depth, content);
      if (openEnd)
        { addNode(close(openEnd, replaceTwoWay($end, $to, depth + 1)), content); }
    }
    addRange($to, null, depth, content);
    return new Fragment(content)
  }

  function replaceTwoWay($from, $to, depth) {
    var content = [];
    addRange(null, $from, depth, content);
    if ($from.depth > depth) {
      var type = joinable($from, $to, depth + 1);
      addNode(close(type, replaceTwoWay($from, $to, depth + 1)), content);
    }
    addRange($to, null, depth, content);
    return new Fragment(content)
  }

  function prepareSliceForReplace(slice, $along) {
    var extra = $along.depth - slice.openStart, parent = $along.node(extra);
    var node = parent.copy(slice.content);
    for (var i = extra - 1; i >= 0; i--)
      { node = $along.node(i).copy(Fragment.from(node)); }
    return {start: node.resolveNoCache(slice.openStart + extra),
            end: node.resolveNoCache(node.content.size - slice.openEnd - extra)}
  }

  // ::- You can [_resolve_](#model.Node.resolve) a position to get more
  // information about it. Objects of this class represent such a
  // resolved position, providing various pieces of context information,
  // and some helper methods.
  //
  // Throughout this interface, methods that take an optional `depth`
  // parameter will interpret undefined as `this.depth` and negative
  // numbers as `this.depth + value`.
  var ResolvedPos = function ResolvedPos(pos, path, parentOffset) {
    // :: number The position that was resolved.
    this.pos = pos;
    this.path = path;
    // :: number
    // The number of levels the parent node is from the root. If this
    // position points directly into the root node, it is 0. If it
    // points into a top-level paragraph, 1, and so on.
    this.depth = path.length / 3 - 1;
    // :: number The offset this position has into its parent node.
    this.parentOffset = parentOffset;
  };

  var prototypeAccessors$1 = { parent: { configurable: true },doc: { configurable: true },textOffset: { configurable: true },nodeAfter: { configurable: true },nodeBefore: { configurable: true } };

  ResolvedPos.prototype.resolveDepth = function resolveDepth (val) {
    if (val == null) { return this.depth }
    if (val < 0) { return this.depth + val }
    return val
  };

  // :: Node
  // The parent node that the position points into. Note that even if
  // a position points into a text node, that node is not considered
  // the parent—text nodes are ‘flat’ in this model, and have no content.
  prototypeAccessors$1.parent.get = function () { return this.node(this.depth) };

  // :: Node
  // The root node in which the position was resolved.
  prototypeAccessors$1.doc.get = function () { return this.node(0) };

  // :: (?number) → Node
  // The ancestor node at the given level. `p.node(p.depth)` is the
  // same as `p.parent`.
  ResolvedPos.prototype.node = function node (depth) { return this.path[this.resolveDepth(depth) * 3] };

  // :: (?number) → number
  // The index into the ancestor at the given level. If this points at
  // the 3rd node in the 2nd paragraph on the top level, for example,
  // `p.index(0)` is 1 and `p.index(1)` is 2.
  ResolvedPos.prototype.index = function index (depth) { return this.path[this.resolveDepth(depth) * 3 + 1] };

  // :: (?number) → number
  // The index pointing after this position into the ancestor at the
  // given level.
  ResolvedPos.prototype.indexAfter = function indexAfter (depth) {
    depth = this.resolveDepth(depth);
    return this.index(depth) + (depth == this.depth && !this.textOffset ? 0 : 1)
  };

  // :: (?number) → number
  // The (absolute) position at the start of the node at the given
  // level.
  ResolvedPos.prototype.start = function start (depth) {
    depth = this.resolveDepth(depth);
    return depth == 0 ? 0 : this.path[depth * 3 - 1] + 1
  };

  // :: (?number) → number
  // The (absolute) position at the end of the node at the given
  // level.
  ResolvedPos.prototype.end = function end (depth) {
    depth = this.resolveDepth(depth);
    return this.start(depth) + this.node(depth).content.size
  };

  // :: (?number) → number
  // The (absolute) position directly before the wrapping node at the
  // given level, or, when `depth` is `this.depth + 1`, the original
  // position.
  ResolvedPos.prototype.before = function before (depth) {
    depth = this.resolveDepth(depth);
    if (!depth) { throw new RangeError("There is no position before the top-level node") }
    return depth == this.depth + 1 ? this.pos : this.path[depth * 3 - 1]
  };

  // :: (?number) → number
  // The (absolute) position directly after the wrapping node at the
  // given level, or the original position when `depth` is `this.depth + 1`.
  ResolvedPos.prototype.after = function after (depth) {
    depth = this.resolveDepth(depth);
    if (!depth) { throw new RangeError("There is no position after the top-level node") }
    return depth == this.depth + 1 ? this.pos : this.path[depth * 3 - 1] + this.path[depth * 3].nodeSize
  };

  // :: number
  // When this position points into a text node, this returns the
  // distance between the position and the start of the text node.
  // Will be zero for positions that point between nodes.
  prototypeAccessors$1.textOffset.get = function () { return this.pos - this.path[this.path.length - 1] };

  // :: ?Node
  // Get the node directly after the position, if any. If the position
  // points into a text node, only the part of that node after the
  // position is returned.
  prototypeAccessors$1.nodeAfter.get = function () {
    var parent = this.parent, index = this.index(this.depth);
    if (index == parent.childCount) { return null }
    var dOff = this.pos - this.path[this.path.length - 1], child = parent.child(index);
    return dOff ? parent.child(index).cut(dOff) : child
  };

  // :: ?Node
  // Get the node directly before the position, if any. If the
  // position points into a text node, only the part of that node
  // before the position is returned.
  prototypeAccessors$1.nodeBefore.get = function () {
    var index = this.index(this.depth);
    var dOff = this.pos - this.path[this.path.length - 1];
    if (dOff) { return this.parent.child(index).cut(0, dOff) }
    return index == 0 ? null : this.parent.child(index - 1)
  };

  // :: (number, ?number) → number
  // Get the position at the given index in the parent node at the
  // given depth (which defaults to `this.depth`).
  ResolvedPos.prototype.posAtIndex = function posAtIndex (index, depth) {
    depth = this.resolveDepth(depth);
    var node = this.path[depth * 3], pos = depth == 0 ? 0 : this.path[depth * 3 - 1] + 1;
    for (var i = 0; i < index; i++) { pos += node.child(i).nodeSize; }
    return pos
  };

  // :: () → [Mark]
  // Get the marks at this position, factoring in the surrounding
  // marks' [`inclusive`](#model.MarkSpec.inclusive) property. If the
  // position is at the start of a non-empty node, the marks of the
  // node after it (if any) are returned.
  ResolvedPos.prototype.marks = function marks () {
    var parent = this.parent, index = this.index();

    // In an empty parent, return the empty array
    if (parent.content.size == 0) { return Mark.none }

    // When inside a text node, just return the text node's marks
    if (this.textOffset) { return parent.child(index).marks }

    var main = parent.maybeChild(index - 1), other = parent.maybeChild(index);
    // If the `after` flag is true of there is no node before, make
    // the node after this position the main reference.
    if (!main) { var tmp = main; main = other; other = tmp; }

    // Use all marks in the main node, except those that have
    // `inclusive` set to false and are not present in the other node.
    var marks = main.marks;
    for (var i = 0; i < marks.length; i++)
      { if (marks[i].type.spec.inclusive === false && (!other || !marks[i].isInSet(other.marks)))
        { marks = marks[i--].removeFromSet(marks); } }

    return marks
  };

  // :: (ResolvedPos) → ?[Mark]
  // Get the marks after the current position, if any, except those
  // that are non-inclusive and not present at position `$end`. This
  // is mostly useful for getting the set of marks to preserve after a
  // deletion. Will return `null` if this position is at the end of
  // its parent node or its parent node isn't a textblock (in which
  // case no marks should be preserved).
  ResolvedPos.prototype.marksAcross = function marksAcross ($end) {
    var after = this.parent.maybeChild(this.index());
    if (!after || !after.isInline) { return null }

    var marks = after.marks, next = $end.parent.maybeChild($end.index());
    for (var i = 0; i < marks.length; i++)
      { if (marks[i].type.spec.inclusive === false && (!next || !marks[i].isInSet(next.marks)))
        { marks = marks[i--].removeFromSet(marks); } }
    return marks
  };

  // :: (number) → number
  // The depth up to which this position and the given (non-resolved)
  // position share the same parent nodes.
  ResolvedPos.prototype.sharedDepth = function sharedDepth (pos) {
    for (var depth = this.depth; depth > 0; depth--)
      { if (this.start(depth) <= pos && this.end(depth) >= pos) { return depth } }
    return 0
  };

  // :: (?ResolvedPos, ?(Node) → bool) → ?NodeRange
  // Returns a range based on the place where this position and the
  // given position diverge around block content. If both point into
  // the same textblock, for example, a range around that textblock
  // will be returned. If they point into different blocks, the range
  // around those blocks in their shared ancestor is returned. You can
  // pass in an optional predicate that will be called with a parent
  // node to see if a range into that parent is acceptable.
  ResolvedPos.prototype.blockRange = function blockRange (other, pred) {
      if ( other === void 0 ) other = this;

    if (other.pos < this.pos) { return other.blockRange(this) }
    for (var d = this.depth - (this.parent.inlineContent || this.pos == other.pos ? 1 : 0); d >= 0; d--)
      { if (other.pos <= this.end(d) && (!pred || pred(this.node(d))))
        { return new NodeRange(this, other, d) } }
  };

  // :: (ResolvedPos) → bool
  // Query whether the given position shares the same parent node.
  ResolvedPos.prototype.sameParent = function sameParent (other) {
    return this.pos - this.parentOffset == other.pos - other.parentOffset
  };

  // :: (ResolvedPos) → ResolvedPos
  // Return the greater of this and the given position.
  ResolvedPos.prototype.max = function max (other) {
    return other.pos > this.pos ? other : this
  };

  // :: (ResolvedPos) → ResolvedPos
  // Return the smaller of this and the given position.
  ResolvedPos.prototype.min = function min (other) {
    return other.pos < this.pos ? other : this
  };

  ResolvedPos.prototype.toString = function toString () {
    var str = "";
    for (var i = 1; i <= this.depth; i++)
      { str += (str ? "/" : "") + this.node(i).type.name + "_" + this.index(i - 1); }
    return str + ":" + this.parentOffset
  };

  ResolvedPos.resolve = function resolve (doc, pos) {
    if (!(pos >= 0 && pos <= doc.content.size)) { throw new RangeError("Position " + pos + " out of range") }
    var path = [];
    var start = 0, parentOffset = pos;
    for (var node = doc;;) {
      var ref = node.content.findIndex(parentOffset);
        var index = ref.index;
        var offset = ref.offset;
      var rem = parentOffset - offset;
      path.push(node, index, start + offset);
      if (!rem) { break }
      node = node.child(index);
      if (node.isText) { break }
      parentOffset = rem - 1;
      start += offset + 1;
    }
    return new ResolvedPos(pos, path, parentOffset)
  };

  ResolvedPos.resolveCached = function resolveCached (doc, pos) {
    for (var i = 0; i < resolveCache.length; i++) {
      var cached = resolveCache[i];
      if (cached.pos == pos && cached.doc == doc) { return cached }
    }
    var result = resolveCache[resolveCachePos] = ResolvedPos.resolve(doc, pos);
    resolveCachePos = (resolveCachePos + 1) % resolveCacheSize;
    return result
  };

  Object.defineProperties( ResolvedPos.prototype, prototypeAccessors$1 );

  var resolveCache = [], resolveCachePos = 0, resolveCacheSize = 12;

  // ::- Represents a flat range of content, i.e. one that starts and
  // ends in the same node.
  var NodeRange = function NodeRange($from, $to, depth) {
    // :: ResolvedPos A resolved position along the start of the
    // content. May have a `depth` greater than this object's `depth`
    // property, since these are the positions that were used to
    // compute the range, not re-resolved positions directly at its
    // boundaries.
    this.$from = $from;
    // :: ResolvedPos A position along the end of the content. See
    // caveat for [`$from`](#model.NodeRange.$from).
    this.$to = $to;
    // :: number The depth of the node that this range points into.
    this.depth = depth;
  };

  var prototypeAccessors$1$1 = { start: { configurable: true },end: { configurable: true },parent: { configurable: true },startIndex: { configurable: true },endIndex: { configurable: true } };

  // :: number The position at the start of the range.
  prototypeAccessors$1$1.start.get = function () { return this.$from.before(this.depth + 1) };
  // :: number The position at the end of the range.
  prototypeAccessors$1$1.end.get = function () { return this.$to.after(this.depth + 1) };

  // :: Node The parent node that the range points into.
  prototypeAccessors$1$1.parent.get = function () { return this.$from.node(this.depth) };
  // :: number The start index of the range in the parent node.
  prototypeAccessors$1$1.startIndex.get = function () { return this.$from.index(this.depth) };
  // :: number The end index of the range in the parent node.
  prototypeAccessors$1$1.endIndex.get = function () { return this.$to.indexAfter(this.depth) };

  Object.defineProperties( NodeRange.prototype, prototypeAccessors$1$1 );

  var emptyAttrs = Object.create(null);

  // ::- This class represents a node in the tree that makes up a
  // ProseMirror document. So a document is an instance of `Node`, with
  // children that are also instances of `Node`.
  //
  // Nodes are persistent data structures. Instead of changing them, you
  // create new ones with the content you want. Old ones keep pointing
  // at the old document shape. This is made cheaper by sharing
  // structure between the old and new data as much as possible, which a
  // tree shape like this (without back pointers) makes easy.
  //
  // **Do not** directly mutate the properties of a `Node` object. See
  // [the guide](/docs/guide/#doc) for more information.
  var Node$1 = function Node(type, attrs, content, marks) {
    // :: NodeType
    // The type of node that this is.
    this.type = type;

    // :: Object
    // An object mapping attribute names to values. The kind of
    // attributes allowed and required are
    // [determined](#model.NodeSpec.attrs) by the node type.
    this.attrs = attrs;

    // :: Fragment
    // A container holding the node's children.
    this.content = content || Fragment.empty;

    // :: [Mark]
    // The marks (things like whether it is emphasized or part of a
    // link) applied to this node.
    this.marks = marks || Mark.none;
  };

  var prototypeAccessors = { nodeSize: { configurable: true },childCount: { configurable: true },textContent: { configurable: true },firstChild: { configurable: true },lastChild: { configurable: true },isBlock: { configurable: true },isTextblock: { configurable: true },inlineContent: { configurable: true },isInline: { configurable: true },isText: { configurable: true },isLeaf: { configurable: true },isAtom: { configurable: true } };

  // text:: ?string
  // For text nodes, this contains the node's text content.

  // :: number
  // The size of this node, as defined by the integer-based [indexing
  // scheme](/docs/guide/#doc.indexing). For text nodes, this is the
  // amount of characters. For other leaf nodes, it is one. For
  // non-leaf nodes, it is the size of the content plus two (the start
  // and end token).
  prototypeAccessors.nodeSize.get = function () { return this.isLeaf ? 1 : 2 + this.content.size };

  // :: number
  // The number of children that the node has.
  prototypeAccessors.childCount.get = function () { return this.content.childCount };

  // :: (number) → Node
  // Get the child node at the given index. Raises an error when the
  // index is out of range.
  Node$1.prototype.child = function child (index) { return this.content.child(index) };

  // :: (number) → ?Node
  // Get the child node at the given index, if it exists.
  Node$1.prototype.maybeChild = function maybeChild (index) { return this.content.maybeChild(index) };

  // :: ((node: Node, offset: number, index: number))
  // Call `f` for every child node, passing the node, its offset
  // into this parent node, and its index.
  Node$1.prototype.forEach = function forEach (f) { this.content.forEach(f); };

  // :: (number, number, (node: Node, pos: number, parent: Node, index: number) → ?bool, ?number)
  // Invoke a callback for all descendant nodes recursively between
  // the given two positions that are relative to start of this node's
  // content. The callback is invoked with the node, its
  // parent-relative position, its parent node, and its child index.
  // When the callback returns false for a given node, that node's
  // children will not be recursed over. The last parameter can be
  // used to specify a starting position to count from.
  Node$1.prototype.nodesBetween = function nodesBetween (from, to, f, startPos) {
      if ( startPos === void 0 ) startPos = 0;

    this.content.nodesBetween(from, to, f, startPos, this);
  };

  // :: ((node: Node, pos: number, parent: Node) → ?bool)
  // Call the given callback for every descendant node. Doesn't
  // descend into a node when the callback returns `false`.
  Node$1.prototype.descendants = function descendants (f) {
    this.nodesBetween(0, this.content.size, f);
  };

  // :: string
  // Concatenates all the text nodes found in this fragment and its
  // children.
  prototypeAccessors.textContent.get = function () { return this.textBetween(0, this.content.size, "") };

  // :: (number, number, ?string, ?string) → string
  // Get all text between positions `from` and `to`. When
  // `blockSeparator` is given, it will be inserted whenever a new
  // block node is started. When `leafText` is given, it'll be
  // inserted for every non-text leaf node encountered.
  Node$1.prototype.textBetween = function textBetween (from, to, blockSeparator, leafText) {
    return this.content.textBetween(from, to, blockSeparator, leafText)
  };

  // :: ?Node
  // Returns this node's first child, or `null` if there are no
  // children.
  prototypeAccessors.firstChild.get = function () { return this.content.firstChild };

  // :: ?Node
  // Returns this node's last child, or `null` if there are no
  // children.
  prototypeAccessors.lastChild.get = function () { return this.content.lastChild };

  // :: (Node) → bool
  // Test whether two nodes represent the same piece of document.
  Node$1.prototype.eq = function eq (other) {
    return this == other || (this.sameMarkup(other) && this.content.eq(other.content))
  };

  // :: (Node) → bool
  // Compare the markup (type, attributes, and marks) of this node to
  // those of another. Returns `true` if both have the same markup.
  Node$1.prototype.sameMarkup = function sameMarkup (other) {
    return this.hasMarkup(other.type, other.attrs, other.marks)
  };

  // :: (NodeType, ?Object, ?[Mark]) → bool
  // Check whether this node's markup correspond to the given type,
  // attributes, and marks.
  Node$1.prototype.hasMarkup = function hasMarkup (type, attrs, marks) {
    return this.type == type &&
      compareDeep(this.attrs, attrs || type.defaultAttrs || emptyAttrs) &&
      Mark.sameSet(this.marks, marks || Mark.none)
  };

  // :: (?Fragment) → Node
  // Create a new node with the same markup as this node, containing
  // the given content (or empty, if no content is given).
  Node$1.prototype.copy = function copy (content) {
      if ( content === void 0 ) content = null;

    if (content == this.content) { return this }
    return new this.constructor(this.type, this.attrs, content, this.marks)
  };

  // :: ([Mark]) → Node
  // Create a copy of this node, with the given set of marks instead
  // of the node's own marks.
  Node$1.prototype.mark = function mark (marks) {
    return marks == this.marks ? this : new this.constructor(this.type, this.attrs, this.content, marks)
  };

  // :: (number, ?number) → Node
  // Create a copy of this node with only the content between the
  // given positions. If `to` is not given, it defaults to the end of
  // the node.
  Node$1.prototype.cut = function cut (from, to) {
    if (from == 0 && to == this.content.size) { return this }
    return this.copy(this.content.cut(from, to))
  };

  // :: (number, ?number) → Slice
  // Cut out the part of the document between the given positions, and
  // return it as a `Slice` object.
  Node$1.prototype.slice = function slice (from, to, includeParents) {
      if ( to === void 0 ) to = this.content.size;
      if ( includeParents === void 0 ) includeParents = false;

    if (from == to) { return Slice.empty }

    var $from = this.resolve(from), $to = this.resolve(to);
    var depth = includeParents ? 0 : $from.sharedDepth(to);
    var start = $from.start(depth), node = $from.node(depth);
    var content = node.content.cut($from.pos - start, $to.pos - start);
    return new Slice(content, $from.depth - depth, $to.depth - depth)
  };

  // :: (number, number, Slice) → Node
  // Replace the part of the document between the given positions with
  // the given slice. The slice must 'fit', meaning its open sides
  // must be able to connect to the surrounding content, and its
  // content nodes must be valid children for the node they are placed
  // into. If any of this is violated, an error of type
  // [`ReplaceError`](#model.ReplaceError) is thrown.
  Node$1.prototype.replace = function replace$1 (from, to, slice) {
    return replace(this.resolve(from), this.resolve(to), slice)
  };

  // :: (number) → ?Node
  // Find the node directly after the given position.
  Node$1.prototype.nodeAt = function nodeAt (pos) {
    for (var node = this;;) {
      var ref = node.content.findIndex(pos);
        var index = ref.index;
        var offset = ref.offset;
      node = node.maybeChild(index);
      if (!node) { return null }
      if (offset == pos || node.isText) { return node }
      pos -= offset + 1;
    }
  };

  // :: (number) → {node: ?Node, index: number, offset: number}
  // Find the (direct) child node after the given offset, if any,
  // and return it along with its index and offset relative to this
  // node.
  Node$1.prototype.childAfter = function childAfter (pos) {
    var ref = this.content.findIndex(pos);
      var index = ref.index;
      var offset = ref.offset;
    return {node: this.content.maybeChild(index), index: index, offset: offset}
  };

  // :: (number) → {node: ?Node, index: number, offset: number}
  // Find the (direct) child node before the given offset, if any,
  // and return it along with its index and offset relative to this
  // node.
  Node$1.prototype.childBefore = function childBefore (pos) {
    if (pos == 0) { return {node: null, index: 0, offset: 0} }
    var ref = this.content.findIndex(pos);
      var index = ref.index;
      var offset = ref.offset;
    if (offset < pos) { return {node: this.content.child(index), index: index, offset: offset} }
    var node = this.content.child(index - 1);
    return {node: node, index: index - 1, offset: offset - node.nodeSize}
  };

  // :: (number) → ResolvedPos
  // Resolve the given position in the document, returning an
  // [object](#model.ResolvedPos) with information about its context.
  Node$1.prototype.resolve = function resolve (pos) { return ResolvedPos.resolveCached(this, pos) };

  Node$1.prototype.resolveNoCache = function resolveNoCache (pos) { return ResolvedPos.resolve(this, pos) };

  // :: (number, number, union<Mark, MarkType>) → bool
  // Test whether a given mark or mark type occurs in this document
  // between the two given positions.
  Node$1.prototype.rangeHasMark = function rangeHasMark (from, to, type) {
    var found = false;
    if (to > from) { this.nodesBetween(from, to, function (node) {
      if (type.isInSet(node.marks)) { found = true; }
      return !found
    }); }
    return found
  };

  // :: bool
  // True when this is a block (non-inline node)
  prototypeAccessors.isBlock.get = function () { return this.type.isBlock };

  // :: bool
  // True when this is a textblock node, a block node with inline
  // content.
  prototypeAccessors.isTextblock.get = function () { return this.type.isTextblock };

  // :: bool
  // True when this node allows inline content.
  prototypeAccessors.inlineContent.get = function () { return this.type.inlineContent };

  // :: bool
  // True when this is an inline node (a text node or a node that can
  // appear among text).
  prototypeAccessors.isInline.get = function () { return this.type.isInline };

  // :: bool
  // True when this is a text node.
  prototypeAccessors.isText.get = function () { return this.type.isText };

  // :: bool
  // True when this is a leaf node.
  prototypeAccessors.isLeaf.get = function () { return this.type.isLeaf };

  // :: bool
  // True when this is an atom, i.e. when it does not have directly
  // editable content. This is usually the same as `isLeaf`, but can
  // be configured with the [`atom` property](#model.NodeSpec.atom) on
  // a node's spec (typically used when the node is displayed as an
  // uneditable [node view](#view.NodeView)).
  prototypeAccessors.isAtom.get = function () { return this.type.isAtom };

  // :: () → string
  // Return a string representation of this node for debugging
  // purposes.
  Node$1.prototype.toString = function toString () {
    if (this.type.spec.toDebugString) { return this.type.spec.toDebugString(this) }
    var name = this.type.name;
    if (this.content.size)
      { name += "(" + this.content.toStringInner() + ")"; }
    return wrapMarks(this.marks, name)
  };

  // :: (number) → ContentMatch
  // Get the content match in this node at the given index.
  Node$1.prototype.contentMatchAt = function contentMatchAt (index) {
    var match = this.type.contentMatch.matchFragment(this.content, 0, index);
    if (!match) { throw new Error("Called contentMatchAt on a node with invalid content") }
    return match
  };

  // :: (number, number, ?Fragment, ?number, ?number) → bool
  // Test whether replacing the range between `from` and `to` (by
  // child index) with the given replacement fragment (which defaults
  // to the empty fragment) would leave the node's content valid. You
  // can optionally pass `start` and `end` indices into the
  // replacement fragment.
  Node$1.prototype.canReplace = function canReplace (from, to, replacement, start, end) {
      if ( replacement === void 0 ) replacement = Fragment.empty;
      if ( start === void 0 ) start = 0;
      if ( end === void 0 ) end = replacement.childCount;

    var one = this.contentMatchAt(from).matchFragment(replacement, start, end);
    var two = one && one.matchFragment(this.content, to);
    if (!two || !two.validEnd) { return false }
    for (var i = start; i < end; i++) { if (!this.type.allowsMarks(replacement.child(i).marks)) { return false } }
    return true
  };

  // :: (number, number, NodeType, ?[Mark]) → bool
  // Test whether replacing the range `from` to `to` (by index) with a
  // node of the given type would leave the node's content valid.
  Node$1.prototype.canReplaceWith = function canReplaceWith (from, to, type, marks) {
    if (marks && !this.type.allowsMarks(marks)) { return false }
    var start = this.contentMatchAt(from).matchType(type);
    var end = start && start.matchFragment(this.content, to);
    return end ? end.validEnd : false
  };

  // :: (Node) → bool
  // Test whether the given node's content could be appended to this
  // node. If that node is empty, this will only return true if there
  // is at least one node type that can appear in both nodes (to avoid
  // merging completely incompatible nodes).
  Node$1.prototype.canAppend = function canAppend (other) {
    if (other.content.size) { return this.canReplace(this.childCount, this.childCount, other.content) }
    else { return this.type.compatibleContent(other.type) }
  };

  // :: ()
  // Check whether this node and its descendants conform to the
  // schema, and raise error when they do not.
  Node$1.prototype.check = function check () {
    if (!this.type.validContent(this.content))
      { throw new RangeError(("Invalid content for node " + (this.type.name) + ": " + (this.content.toString().slice(0, 50)))) }
    var copy = Mark.none;
    for (var i = 0; i < this.marks.length; i++) { copy = this.marks[i].addToSet(copy); }
    if (!Mark.sameSet(copy, this.marks))
      { throw new RangeError(("Invalid collection of marks for node " + (this.type.name) + ": " + (this.marks.map(function (m) { return m.type.name; })))) }
    this.content.forEach(function (node) { return node.check(); });
  };

  // :: () → Object
  // Return a JSON-serializeable representation of this node.
  Node$1.prototype.toJSON = function toJSON () {
    var obj = {type: this.type.name};
    for (var _ in this.attrs) {
      obj.attrs = this.attrs;
      break
    }
    if (this.content.size)
      { obj.content = this.content.toJSON(); }
    if (this.marks.length)
      { obj.marks = this.marks.map(function (n) { return n.toJSON(); }); }
    return obj
  };

  // :: (Schema, Object) → Node
  // Deserialize a node from its JSON representation.
  Node$1.fromJSON = function fromJSON (schema, json) {
    if (!json) { throw new RangeError("Invalid input for Node.fromJSON") }
    var marks = null;
    if (json.marks) {
      if (!Array.isArray(json.marks)) { throw new RangeError("Invalid mark data for Node.fromJSON") }
      marks = json.marks.map(schema.markFromJSON);
    }
    if (json.type == "text") {
      if (typeof json.text != "string") { throw new RangeError("Invalid text node in JSON") }
      return schema.text(json.text, marks)
    }
    var content = Fragment.fromJSON(schema, json.content);
    return schema.nodeType(json.type).create(json.attrs, content, marks)
  };

  Object.defineProperties( Node$1.prototype, prototypeAccessors );

  function wrapMarks(marks, str) {
    for (var i = marks.length - 1; i >= 0; i--)
      { str = marks[i].type.name + "(" + str + ")"; }
    return str
  }

  // : (NodeType) → InputRule
  // Given a blockquote node type, returns an input rule that turns `"> "`
  // at the start of a textblock into a blockquote.
  var strongRule = function (schema) {
      return markInputRule(/(?:\*\*|__)([^\*_]+)(?:\*\*|__)$/, schema.marks.strong);
  };

  function hasCodeMark(node)
  {
      if(!node) {
          return false;
      }

      var result = false;
      node.marks.forEach(function (mark) {
          if(mark.type.spec.isCode) {
              result = true;
          }
      });

      return result;
  }

  function markInputRule(regexp, markType, getAttrs) {
      return new InputRule(regexp, function (state, match, start, end) {
          var attrs = getAttrs instanceof Function ? getAttrs(match) : getAttrs;


          var nodeBeforeEnd = state.selection.$to.nodeBefore;

          if(!nodeBeforeEnd || !nodeBeforeEnd.isText
              || nodeBeforeEnd.text.length < match[0].length - 1 // check that the match does not span multiple nodes
              || hasCodeMark(nodeBeforeEnd)
              || markType.isInSet(nodeBeforeEnd.marks)) {
              return null;
          }

          if (match[1]) {
              var tr = state.tr;
              var textStart = start + match[0].indexOf(match[1]);
              var textEnd = textStart + match[1].length;
              if (textEnd < end) { tr.delete(textEnd, end); }
              if (textStart > start) { tr.delete(start, textStart); }
              end = start + match[1].length;
              tr.addMark(start, end, markType.create(attrs));
              tr.removeStoredMark(markType); // Do not continue with mark.
              return tr
          }

          return null;

      })
  }

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var strong = {
      id: 'strong',
      schema: schema$2,
      menu: function (context) { return menu$4(context); },
      inputRules: function (schema) { return [strongRule(schema)]; },
  };

  var nodes = tableNodes({
      tableGroup: "block",
      cellContent: "paragraph+",
      cellAttributes: {
          style: {
              default: null,
              getFromDOM: function getFromDOM(dom) {
                  return dom.style;
              },
              setDOMAttr: function setDOMAttr(value, attrs) {
                  if (value) {
                      attrs.style = value;
                  }
              }
          }
      }
  });

  nodes.table_row.parseMarkdown = {tr: {block: "table_row"}};
  nodes.table_header.parseMarkdown = {th: {block: "table_header"}};
  nodes.table_cell.parseMarkdown = {td: {block: "table_cell"}};

  nodes = Object.assign(nodes, {
      table: {
          content: "(table_row+ | table_head | table_body | table_foot)",
          tableRole: "table",
          isolating: false,
          group: "block",
          parseDOM: [{tag: "table"}],
          toDOM: function toDOM() {
              return ["table", ["tbody", 0]]
          },
          toMarkdown: function (state, node) {
              renderTable(state,node);
          },
          parseMarkdown: {block: "table"}
      },
      table_head: {
          content: "table_row*",
          tableRole: "head",
          parseDOM: [{tag: "thead"}],
          toDOM: function toDOM() {
              return ["thead", 0]
          },
          parseMarkdown: {thead: {block: "table_head"}}
      },
      table_body: {
          content: "table_row*",
          tableRole: "body",
          parseDOM: [{tag: "tbody"}],
          toDOM: function toDOM() {
              return ["tbody", 0]
          },
          parseMarkdown: {tbody: {block: "table_body"}}
      },
      table_foot: {
          content: "table_row*",
          tableRole: "foot",
          parseDOM: [{tag: "tfoot"}],
          toDOM: function toDOM() {
              return ["tfoot", 0]
          },
          parseMarkdown: {tfoot: {block: "table_foot"}}
      }
  });

  var renderTable = function(state, node, withHead) {
      state.table = true;

      if(typeof withHead === 'undefined') {
          withHead = true;
      }

      node.forEach(function (child, _, i) {
          if(child.type.name === 'table_body' || child.type.name === 'table_head') {
              renderTable(state, child, i === 0);
          } else if(withHead && i === 0) {
              renderHeadRow(state,child);
          } else {
              renderRow(state, child);
          }

          if(i !== (node.childCount -1)) {
              state.write("\n");
          }
      });

      state.table = false;
      state.closeBlock(node);
  };

  var renderHeadRow = function(state, node) {
      renderRow(state,node);
      state.write("\n");
      renderRow(state,node, true);
  };

  var renderRow = function(state, node, headMarker) {
      state.write('|');
      node.forEach(function (child, _, i) {
          renderCell(state, child, headMarker);
      });
  };

  var renderCell = function(state, node, headMarker) {
      state.write(' ');
      if(headMarker) {
          (node.textContent.length) ? state.write(state.repeat('-', node.textContent.length)) : state.write('---');
         /* if(node.attrs.style && node.attrs.style.indexOf("text-align:right") >= 0) {
              state.write(':');
          } else {
              state.write(' ');
          }*/
          state.write(' ');
      } else {
          state.renderContent(node);

          state.write(' ');
      }
      state.write('|');
  };


  var schema$1 = {
      nodes: nodes
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  // GFM table, non-standard
  function isSpace(code) {
      switch (code) {
          case 0x09:
          case 0x20:
              return true;
      }
      return false;
  }


  function getLine(state, line) {
    var pos = state.bMarks[line] + state.blkIndent,
        max = state.eMarks[line];

    return state.src.substr(pos, max - pos);
  }

  function escapedSplit(str) {
    var result = [],
        pos = 0,
        max = str.length,
        ch,
        escapes = 0,
        lastPos = 0,
        backTicked = false,
        lastBackTick = 0;

    ch  = str.charCodeAt(pos);

    while (pos < max) {
      if (ch === 0x60/* ` */) {
        if (backTicked) {
          // make \` close code sequence, but not open it;
          // the reason is: `\` is correct code block
          backTicked = false;
          lastBackTick = pos;
        } else if (escapes % 2 === 0) {
          backTicked = true;
          lastBackTick = pos;
        }
      } else if (ch === 0x7c/* | */ && (escapes % 2 === 0) && !backTicked) {
        result.push(str.substring(lastPos, pos));
        lastPos = pos + 1;
      }

      if (ch === 0x5c/* \ */) {
        escapes++;
      } else {
        escapes = 0;
      }

      pos++;

      // If there was an un-closed backtick, go back to just after
      // the last backtick, but as if it was a normal character
      if (pos === max && backTicked) {
        backTicked = false;
        pos = lastBackTick + 1;
      }

      ch = str.charCodeAt(pos);
    }

    result.push(str.substring(lastPos));

    return result;
  }


  var markdownit_table = function table_plugin(state, startLine, endLine, silent) {
    var ch, lineText, pos, i, nextLine, columns, columnCount, token,
        aligns, t, tableLines, tbodyLines;

    // should have at least two lines
    if (startLine + 2 > endLine) { return false; }

    nextLine = startLine + 1;

    if (state.sCount[nextLine] < state.blkIndent) { return false; }

    // if it's indented more than 3 spaces, it should be a code block
    if (state.sCount[nextLine] - state.blkIndent >= 4) { return false; }

    // first character of the second line should be '|', '-', ':',
    // and no other characters are allowed but spaces;
    // basically, this is the equivalent of /^[-:|][-:|\s]*$/ regexp

    pos = state.bMarks[nextLine] + state.tShift[nextLine];
    if (pos >= state.eMarks[nextLine]) { return false; }

    ch = state.src.charCodeAt(pos++);
    if (ch !== 0x7C/* | */ && ch !== 0x2D/* - */ && ch !== 0x3A/* : */) { return false; }

    while (pos < state.eMarks[nextLine]) {
      ch = state.src.charCodeAt(pos);

      if (ch !== 0x7C/* | */ && ch !== 0x2D/* - */ && ch !== 0x3A/* : */ && !isSpace(ch)) { return false; }

      pos++;
    }

    lineText = getLine(state, startLine + 1);

    columns = lineText.split('|');
    aligns = [];
    for (i = 0; i < columns.length; i++) {
      t = columns[i].trim();
      if (!t) {
        // allow empty columns before and after table, but not in between columns;
        // e.g. allow ` |---| `, disallow ` ---||--- `
        if (i === 0 || i === columns.length - 1) {
          continue;
        } else {
          return false;
        }
      }

      if (!/^:?-+:?$/.test(t)) { return false; }
      if (t.charCodeAt(t.length - 1) === 0x3A/* : */) {
        aligns.push(t.charCodeAt(0) === 0x3A/* : */ ? 'center' : 'right');
      } else if (t.charCodeAt(0) === 0x3A/* : */) {
        aligns.push('left');
      } else {
        aligns.push('');
      }
    }

    lineText = getLine(state, startLine).trim();
    if (lineText.indexOf('|') === -1) { return false; }
    if (state.sCount[startLine] - state.blkIndent >= 4) { return false; }
    columns = escapedSplit(lineText.replace(/^\||\|$/g, ''));

    // header row will define an amount of columns in the entire table,
    // and align row shouldn't be smaller than that (the rest of the rows can)
    columnCount = columns.length;
    if (columnCount > aligns.length) { return false; }

    if (silent) { return true; }


    token     = state.push('table_open', 'table', 1);
    token.map = tableLines = [ startLine, 0 ];

    //token     = state.push('thead_open', 'thead', 1);
    //token.map = [ startLine, startLine + 1 ];

    token     = state.push('tr_open', 'tr', 1);
    token.map = [ startLine, startLine + 1 ];

    for (i = 0; i < columns.length; i++) {
      token          = state.push('th_open', 'th', 1);
      token.map      = [ startLine, startLine + 1 ];
      if (aligns[i]) {
        token.attrs  = [ [ 'style', 'text-align:' + aligns[i] ] ];
      }

      token          = state.push('paragraph_open', 'p', 1);
      token.map      = [ startLine, state.line ];

      token          = state.push('inline', '', 0);
      token.content  = columns[i].trim();
      token.map      = [ startLine, startLine + 1 ];
      token.children = [];

        token          = state.push('paragraph_close', 'p', -1);
      token          = state.push('th_close', 'th', -1);
    }

    token     = state.push('tr_close', 'tr', -1);
    //token     = state.push('thead_close', 'thead', -1);

    //token     = state.push('tbody_open', 'tbody', 1);
    token.map = tbodyLines = [ startLine + 2, 0 ];

    for (nextLine = startLine + 2; nextLine < endLine; nextLine++) {
      if (state.sCount[nextLine] < state.blkIndent) { break; }

      lineText = getLine(state, nextLine).trim();
      if (lineText.indexOf('|') === -1) { break; }
      if (state.sCount[nextLine] - state.blkIndent >= 4) { break; }
      columns = escapedSplit(lineText.replace(/^\||\|$/g, ''));

      token = state.push('tr_open', 'tr', 1);
      for (i = 0; i < columnCount; i++) {
        token          = state.push('td_open', 'td', 1);
        if (aligns[i]) {
          token.attrs  = [ [ 'style', 'text-align:' + aligns[i] ] ];
        }

          token          = state.push('paragraph_open', 'p', 1);
          token.map      = [ startLine, state.line ];

        token          = state.push('inline', '', 0);
        token.content  = columns[i] ? columns[i].trim() : '';
        token.children = [];

        token          = state.push('paragraph_close', 'p', -1);

        token          = state.push('td_close', 'td', -1);
      }
      token = state.push('tr_close', 'tr', -1);
    }
    //token = state.push('tbody_close', 'tbody', -1);
    token = state.push('table_close', 'table', -1);

    tableLines[1] = tbodyLines[1] = nextLine;
    state.line = nextLine;
    return true;
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  function wrapTableItem(context) {
      var schema = context.schema;
      var command = wrapIn(schema.nodes.table_header);
      var itemOptions = {
          title: context.translate("Create table"),
          icon: icons$1.table,
          sortOrder: 300,
          run: function run(state, dispatch, view) {
              openPrompt({
                  title: context.translate("Insert table"),
                  fields: {
                      rowCount: new TextField({label: context.translate("Rows"), required: true, value: 1}),
                      columnCount: new TextField({label: context.translate("Columns"), value: 1})
                  },
                  callback: function callback(attrs) {
                      wrapIn(schema.nodes.table_header)(view.state, dispatch);

                      for (var i = 1; i < attrs.columnCount; i++) {
                          addColumnAfter(view.state, dispatch);
                      }

                      toggleHeaderRow(view.state, dispatch);
                      toggleHeaderRow(view.state, dispatch);

                      for (var i$1 = 1; i$1 < attrs.rowCount; i$1++) {
                          addRowAfter(view.state, dispatch);
                          //toggleHeaderRow();
                      }

                      view.focus();
                  }
              });
          },
          enable: function enable(state) {
              return command(state)
          },
          select: function select(state) {
              return command(state)
          }
      };

      return new MenuItem$1(itemOptions);
  }

  function menu$3(context) {
      return [
          {
              id: 'insertTable',
              node: 'table',
              item: wrapTableItem(context)
          },
          {
              id: 'tableOptions',
              node: 'table',
              item: new Dropdown$1(buildTableMenu(context), {
                  icon: icons$1.table,
                  sortOrder: 301
              })
          }
      ]
  }

  var buildTableMenu = function (context) {
      function item(label, cmd, sortOrder) {
          return new MenuItem$1({label: label, select: cmd, run: cmd, sortOrder: sortOrder, title: ''})
      }

      return [
          item(context.translate("Insert column before"), addColumnBefore, 0),
          item(context.translate("Insert column after"), addColumnAfter, 1),
          item(context.translate("Delete column"), deleteColumn, 2),
          item(context.translate("Insert row before"), addRowBefore, 3),
          item(context.translate("Insert row after"), addRowAfter, 4),
          item(context.translate("Delete row"), deleteRow, 5),
          item(context.translate("Delete table"), deleteTable, 6)
      ];
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var table = {
      id: 'table',
      schema: schema$1,
      menu: function (context) { return menu$3(context); },
      registerMarkdownIt: function (markdownIt) {
          markdownIt.block.ruler.at('table', markdownit_table);
      }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */
  var schema = {
      nodes: {
          text: {
              sortOrder: 900,
              group: "inline",
              toDOM: function toDOM(node) {
                  return node.text
              },
              toMarkdown: function (state, node) {
                  var isCodeMark = false;
                  node.marks.forEach(function(mark) {
                      if(mark.type.spec.isCode) {
                          isCodeMark = true;
                      }
                  });

                  var text = node.text;

                  if(isCodeMark) {
                      text = text.replace('`', '');
                  }

                  state.text(text, !isCodeMark);
              }
          }
      }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var text = {
      id: 'text',
      schema: schema
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var attributesPlugin = function (context) {
      return new Plugin({
          props: {
              attributes: context.options.attributes
          }
      });
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var attributes = {
      id: 'attributes',
      plugins: function (context) {
          return [
              attributesPlugin(context)
          ]
      },
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var placeholderPlugin = function (context) {
      return new Plugin({
          state: {
              init: function init(config, state) {
                  if(!isEmpty(state.doc, context)) {
                      return DecorationSet.empty;
                  } else {
                      return DecorationSet.create(state.doc, [createDecoration(state.doc, context)]);
                  }
              },
              apply: function apply(tr, set, state, newState) {
                  // TODO: Currently if we leave the node with an empty e.g heading there is no placeholder
                  // We should check when focusout, if the node is empty and change the first child to a paragraph

                  if(focusKey.getState(newState)) {
                      return DecorationSet.empty;
                  }

                  if (!isEmpty(tr.doc, context)) {
                      return DecorationSet.empty;
                  }

                  return set.add(tr.doc, [createDecoration(tr.doc, context)]);
              }
          },
          props: {
              decorations: function decorations(state) {
                  return this.getState(state);
              }
          }
      });
  };

  var isEmpty = function (doc, context) {
      return doc.childCount === 1
          && doc.firstChild.type.name === 'paragraph'
          && doc.firstChild.content.size === 0 &&
          !context.hasContentDecorations()
  };

  var createDecoration = function(doc, context) {
      var node = document.createElement('div');
      node.textContent = context.options.placeholder.text;
      node.className = context.options.placeholder['class'] || 'placeholder';
      return Decoration.widget(1, node);
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var placeholder = {
      id: 'placeholder',
      plugins: function (context) {
          if(!context.options.placeholder || !context.options.placeholder.text) {
              return [];
          }

          return [
              placeholderPlugin(context)
          ];
      },
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var loader = {
      id: 'loader',
      plugins: function (context) {
          return [loaderPlugin(context)]
      },
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  function triggerUpload(state, view, context, files) {
      // A fresh object to act as the ID for this upload
      var id = {};

      var uploadWidget = humhub.require('ui.widget.Widget').instance($('#'+context.id+'-file-upload'));

      uploadWidget.off('uploadStart.richtext').on('uploadStart.richtext', function (evt, response) {
          // Replace the selection with a placeholder
          loaderStart(context, id, true);
      }).off('uploadEnd.richtext').on('uploadEnd.richtext', function (evt, response) {
          replaceLoader(context, id, createNodesFromResponse(context, response), true);
      }).off('uploadFinish.richtext').on('uploadFinish.richtext', function () {
          // Make sure our loader is removed after upload
          removeLoader(context, id, true);
      });

      if(files) {
          uploadWidget.$.fileupload('add', {files: files});
      } else {
          uploadWidget.run();
      }

  }

  var createNodesFromResponse = function(context, response) {
      var schema = context.schema;
      var nodes = [];

      // Otherwise, insert it at the placeholder's position, and remove the placeholder
          response.result.files.forEach(function (file) {

              var node;

              if(file.error) {
                  return;
              }

              var url = file.url;

              if(file.mimeIcon === 'mime-image') {
                  node = schema.nodes.image.create({src : url, title: file.name, alt: file.name, fileGuid: file.guid});
              } else {
                  var linkMark = schema.marks.link.create({href: url, fileGuid: file.guid});
                  node = schema.text(file.name, [linkMark]);
              }

              nodes.push(node);
          });

      return nodes;
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var uploadFile = function (context) {
      return new MenuItem$1({
          title: context.translate("Upload and include a File"),
          label: context.translate("Upload File"),
          sortOrder: 0,
          enable: function enable(state) {
              return canInsertLink(state);
          },
          run: function run(state, dispatch, view) {
              if (view.state.selection.$from.parent.inlineContent) {
                  triggerUpload(state, view, context);
              }
          }
      });
  };

  function menu$2(context) {
      return [
          {
              id: 'uploadFile',
              mark: 'link',
              group: 'insert',
              item: uploadFile(context)
          }
      ]
  }

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var upload = {
      id: 'upload',
      menu:  function (context) { return menu$2(context); }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var clipboardPlugin = function (context) {

      var parser = getParser(context);
      return new Plugin({
          props: {
              clipboardTextParser: $.proxy(parser.parse, parser),
              transformPasted: function (slice) {
                  if(slice && slice instanceof Node$2 && slice.type == context.schema.nodes.doc) {
                      return new Slice$1(slice.content, 0, 0)
                  } else {
                      try {
                          var selectionMarks = getSelectionMarks(context);

                          if(selectionMarks && selectionMarks.length) {
                              applyMarksToRawText(slice, selectionMarks);
                          }
                      } catch(e) {
                          console.warn(e);
                      }
                  }

                  return slice;
              },
              handleDOMEvents: {
                  paste: function (view, e) {
                      if(e.clipboardData.files && e.clipboardData.files.length) {
                          triggerUpload(view.state, view, context, e.clipboardData.files);
                          e.preventDefault();
                      }
                  }
              }
          },
      });
  };

  function getSelectionMarks(context) {
      if(context.editor.view.state.storedMarks && context.editor.view.state.storedMarks.length) {
          return context.editor.view.state.storedMarks
      }

      var selection = context.editor.view.state.selection;
      var nodeBefore = selection.$from.nodeBefore;
      return nodeBefore ? nodeBefore.marks : null;
  }

  function applyMarksToRawText(slice, marks)
  {
      var fragment = slice.content;
      var firstChild = fragment.firstChild;

      if(!firstChild) {
          return;

      }

      var texts = $node(firstChild).find('text');
      if(texts.flat.length) {
          var firstText = texts.flat[0];
          if(firstText.isPlain()) {
              firstText.addMarks(marks);
          }
      }
  }

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2018 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var clipboard = {
      id: 'clipboard',
      plugins: function (context) {
          return [
              clipboardPlugin(context)
          ]
      },
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2018 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var position = {
      false: 'push',
      true: 'unshift'
  };

  var anchors = {
      id: 'anchor',
      renderOnly: true,
      init: function (context, isEdit) {
          if(!isEdit) {
              context.editor.$.on('mouseenter', ':header', function() {
                  $(this).find('.header-anchor').show();
              }).on('mouseleave', ':header', function() {
                  $(this).find('.header-anchor').hide();
              });
          }
      },
      registerMarkdownIt: function (markdownIt) {
          var anchorOptions =  {permalink: true};
          anchorOptions.renderPermalink =  function (slug, opts, state, idx) {
              var ref;

              var space = function () { return Object.assign(new state.Token('text', '', 0), { content: ' ' }); };

              var linkTokens = [
                  Object.assign(new state.Token('link_open', 'a', 1), {
                      anchor: true,
                      attrs: [
                          ['class', opts.permalinkClass],
                          ['href', opts.permalinkHref(slug, state)],
                          ['style', 'display:none'],
                          ['target', '_self'],
                          ['aria-hidden', 'true']
                      ]
                  }),
                  Object.assign(new state.Token('text', '', 0), {content: opts.permalinkSymbol }),
                  new state.Token('link_close', 'a', -1)
              ];

              // `push` or `unshift` according to position option.
              // Space is at the opposite side.
              linkTokens[position[!opts.permalinkBefore]](space());
              (ref = state.tokens[idx + 1].children)[position[opts.permalinkBefore]].apply(ref, linkTokens);
          };

          if(anchorOptions) {
              markdownIt.use(r, anchorOptions);
          }
      }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2018 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  function fullScreen(context) {
      return new MenuItem$1({
          id: 'fullscreen',
          title: "Fullscreen",
          sortOrder: 300,
          run: function() {
              var $editor = context.editor.$;
              if($editor.is('.fullscreen')) {
                 minimize(context);
              } else {
                  maximize(context);
              }
          },
          icon: icons$1.enlarge
      });
  }

  function minimize(context, menuItem) {
      var $editor = context.editor.$;
      if($editor.is('.fullscreen')) {

          $('body').removeClass('modal-open');
          $editor.removeClass('fullscreen');
          $editor.find('.Prosemirror').blur();

          context.fullScreenMenuItem.switchIcon(icons$1.enlarge);
      }
  }

  function maximize(context, menuItem) {
      var $editor = context.editor.$;
      if(!$editor.is('.fullscreen')) {

          // Fixes a bug in ios safari when displaying a position:fixed element with input focus...
          document.activeElement.blur();
          setTimeout(function () {context.editor.view.focus();}, 200);

          $('body').addClass('modal-open');
          $editor.addClass('fullscreen');

          context.fullScreenMenuItem.switchIcon(icons$1.shrink);
      }
  }

  function menu$1(context) {
      return [
          {
              id: 'fullScreen',
              group: 'resize',
              item: fullScreen(context)
          } ]
  }

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2018 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var fullscreen = {
      id: 'fullscreen',
      init: function init(context) {
          if(context.getPluginOption('fullscreen', 'autoFullScreen') === true) {
              context.editor.$.on('click', '.ProseMirror', function(e) {
                  if(humhub.require('ui.view').isSmall() && !context.editor.$.is('.fullscreen')) {
                      maximize(context);
                  }
              });
          }

          context.editor.$.on('clear', function() {
              minimize(context);
          });
      },
      menu: function (context) {
          var fullScreenMenu = menu$1(context);
          context.fullScreenMenuItem = fullScreenMenu[0].item;
          return fullScreenMenu;
      }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2018 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var SELECTOR_DEFAULT = '.helper-group, .format-group, .insert-dropdown, .ProseMirror-menu-insertTable:not(.hidden), .ProseMirror-menu-fullScreen:not(.hidden)';

  function resizeNav$1(context) {

      return new MenuItem$1({
          id: 'resizeNav',
          title: "More",
          sortOrder: 400,
          run: function run() {
              var $nodes = getNodes(context);
              if(!context.editor.$.find('.helper-group').is(':visible')) {
                  $nodes.fadeIn();
                  this.switchIcon(icons$1.angleDoubleLeft);
                  $(this.dom).data('state', true);
              } else {
                  $nodes.hide();
                  this.switchIcon(icons$1.angleDoubleRight);
                  $(this.dom).data('state', false);
              }
          },
          icon: icons$1.angleDoubleRight
      });
  }

  function getNodes(context) {
      return context.editor.$.find(getSelector(context));
  }

  function getSelector(context) {
      return context.getPluginOption('resizeNav', 'selector', SELECTOR_DEFAULT);
  }

  function menu(context) {
      return [
          {
              id: 'resizeNav',
              group: 'resize',
              item: resizeNav$1(context)
          } ]
  }

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2018 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var resizeNav = {
      id: 'resizeNav',
      init: function init(context) {
          context.event.on('afterMenuBarInit', function (evt, instance) {
              getNodes(context).hide();
          }).on('afterMenuBarUpdate', function (evt, instance) {
              if(!$(instance.menu).find('.ProseMirror-menu-resizeNav').data('state')) {
                  getNodes(context).hide();
              }
          });
      },
      menu: function (context) { return menu(context); }
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2018 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var MaxHeightState = function MaxHeightState(state, options) {
      this.state = state;
      this.context = options.context;
      this.oldStageHeight = 0;
      this.scrollActive = false;
      this.niceScrollInit = false;
      this.initialized = false;
  };

  MaxHeightState.prototype.update = function update () {
      var stageHeight = this.context.editor.getStage()[0].offsetHeight;

      if(stageHeight === this.oldStageHeight) {
          return;
      }

      this.oldStageHeight = stageHeight;

      if(!this.scrollActive && this.context.editor.getStage()[0].scrollHeight > stageHeight) {
          if(!this.niceScrollInit && !humhub.require('ui.view').isSmall() && this.context.editor.getStage().niceScroll) {
              this.context.editor.getStage().niceScroll({
                  cursorwidth: "7",
                  cursorborder: "",
                  cursorcolor: "#606572",
                  cursoropacitymax: "0.3",
                  nativeparentscrolling: false,
                  autohidemode: false,
                  railpadding: {top: 2, right: 3, left: 0, bottom: 2}
              });
          }

          this.niceScrollInit = true;
          this.scrollActive = true;
          this.context.editor.trigger('scrollActive');
      } else if(!this.initialized || this.scrollActive) {
          this.scrollActive = false;
          this.context.editor.trigger('scrollInactive');
      }

      this.initialized = true;
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2018 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var pluginKey = new PluginKey('max_height');

  var maxHeight = {
      id: 'max-height',
      init: function (context, isEdit) {
          if(!isEdit) {
              return;
          }

          context.editor.on('init', function () {
              if(context.options.maxHeight) {
                  context.editor.getStage().css({'max-height': context.options.maxHeight, 'overflow': 'auto'});
              }

              if(!context.editor.isEmpty()) {
                  var maxHeightState = pluginKey.getState(context.editor.view.state);
                  maxHeightState.update();
              }
          });
      },
      plugins: function (context) {
          return [new Plugin({
              state: {
                  init: function init(config, state) {
                      return new MaxHeightState(state, {context: context});
                  },
                  apply: function apply(tr, prevPluginState, newState) {
                      return prevPluginState;
                  }
              },
              key: pluginKey,
              view: function (view) {
                  return {
                      update: function (view, prevState) {
                          var maxHeightState = pluginKey.getState(view.state);
                          maxHeightState.update();
                      },
                      destroy: function () {}
                  }
              }
          })];
      }
  };

  var savePlugin = function (context) {
      return new Plugin({
          props: {
              handleKeyDown: function handleKeyDown(view, event) {
                  if(context.options.keySubmit === false) {
                      return;
                  }

                  if(event.ctrlKey && event.key === 's') {
                      event.preventDefault();
                      var $form = context.editor.$.closest('form');

                      if(!$form.length) {
                          return;
                      }

                      var $submit = $form.find('[type="submit"]');
                      if($submit.length) {
                          context.editor.$.trigger('focusout');
                          $submit.trigger('click');
                      }
                  }
              },
          },
      });
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var mention = {
      id: 'save',
      plugins: function (context) {
          if(!context.options.mention || !context.options.mention.provider) {
              return [];
          }
          return [
              savePlugin(context)
          ]
      },
  };

  var PluginRegistry = function PluginRegistry() {
      this.plugins = [];
      this.pluginMap = {};
      this.presets = new PresetRegistry(this);
      this.editorPresets = new PresetRegistry(this);
  };

  PluginRegistry.prototype.getPresetRegistry = function getPresetRegistry (context) {
      return context.editor.isEdit() ? this.editorPresets : this.presets;
  };

  PluginRegistry.prototype.register = function register (plugin, options) {
      options = options || {};

      this.plugins.push(plugin);
      this.pluginMap[plugin.id] = plugin;

      options = (typeof options === 'string') ? { preset: options } : options;

      if(options.preset) {
          this.addToPreset(plugin, options.preset, options);
      }
  };

  PluginRegistry.prototype.registerPreset = function registerPreset (id, plugins) {
      this.presets.register(id, plugins);
      this.editorPresets.register(id, plugins);

      if(plugins.callback) {
          plugins.callback($.proxy(this.addToPreset, this));
      }
  };

  PluginRegistry.prototype.addToPreset = function addToPreset (plugin, presetId,  options) {
      options = options || {};

      if(typeof plugin === 'string') {
          plugin = this.pluginMap[plugin];
      }

      if(!plugin) {
          console.warn('Could not add plugin to preset '+presetId);
          return;
      }

      if(!plugin.renderOnly) {
          this.editorPresets.add(presetId, plugin, options);
      }

      if(!plugin.editorOnly) {
          this.presets.add(presetId, plugin, options);
      }
  };

  var PresetRegistry = function PresetRegistry(pluginRegistry) {
      this.pluginRegistry = pluginRegistry;
      this.map = {};
  };

  PresetRegistry.prototype.get = function get (presetId) {
      return this.map[presetId];
  };
  PresetRegistry.prototype.register = function register (id, plugins) {
          var this$1$1 = this;


      var result = [];

      if(Array.isArray(plugins)) {
          plugins.forEach(function (pluginId) {
              var plugin = this$1$1.pluginRegistry.pluginMap[pluginId];
              if(plugin) {
                  result.push(plugin);
              }
          });
      } else if(plugins.extend) {
          var toExtend =  this.map[plugins.extend];

          if(!toExtend) {
              console.error('Could not extend richtext preset '+plugins.extend+' preset not registered!');
              return;
          }

          if(plugins.exclude && Array.isArray(plugins.exclude)) {
              toExtend.forEach(function (plugin) {
                  if(plugin && !plugins.exclude.includes(plugin.id)) {
                      result.push(plugin);
                  }
              });
          } else {
              result = toExtend.slice(0);
          }

          if(plugins.include && Array.isArray(plugins.include)) {
              plugins.include.forEach(function (plugin) {
                  if(!this$1$1.pluginRegistry.pluginMap[plugin]) {
                      console.error('Could not include plugin '+plugin+' to preset '+id+' plugin not found!');
                  } else {
                      result.push(this$1$1.pluginRegistry.pluginMap[plugin]);
                  }
              });
          }
      }

      this.map[id] = result;
  };
  PresetRegistry.prototype.add = function add (presetId, plugin, options) {
      options = options || {};
      var preset = this.map[presetId] ? this.map[presetId].slice() : [];

      if(options['before'] && this.pluginRegistry.pluginMap[options['before']]) {
          var index = preset.indexOf(this.pluginRegistry.pluginMap[options['before']]);
          if (index >= 0) {
              preset.splice(index, 0, plugin);
          } else {
              console.warn('Tried appending plugin before non existing preset plugin: '+presetId+' before:'+options['before']);
              preset.push(plugin);
          }
      } else if(options['after'] && this.pluginRegistry.pluginMap[options['after']]) {
          var index$1 = preset.indexOf(this.pluginRegistry.pluginMap[options['after']]);
          if (index$1 >= 0) {
              preset.splice(index$1+1, 0, plugin);
          } else {
              console.warn('Tried appending plugin after non existing preset plugin: '+presetId+' after:'+options['after']);
              preset.push(plugin);
          }
      } else {
          preset.push(plugin);
      }

      this.map[presetId] = preset;
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var registry = new PluginRegistry();

  var registerPlugin = function(plugin, options) {
      registry.register(plugin, options);
  };

  var registerPreset = function(id, plugins) {
      registry.registerPreset(id, plugins);
  };

  registerPlugin(doc, 'markdown');
  registerPlugin(focus, 'markdown');
  registerPlugin(clipboard, 'markdown');
  registerPlugin(loader, 'markdown');
  registerPlugin(paragraph, 'markdown');
  registerPlugin(blockquote$1, 'markdown');
  registerPlugin(bullet_list, 'markdown');
  registerPlugin(strikethrough, 'markdown');
  registerPlugin(em, 'markdown');
  registerPlugin(strong, 'markdown');
  registerPlugin(code$1, 'markdown');
  registerPlugin(link, 'markdown');
  registerPlugin(code_block, 'markdown');
  registerPlugin(emoji);
  registerPlugin(hard_break, 'markdown');

  registerPlugin(horizontal_rule, 'markdown');
  registerPlugin(image, 'markdown');
  registerPlugin(list_item, 'markdown');
  registerPlugin(mention$1);
  registerPlugin(oembed);
  registerPlugin(ordered_list, 'markdown');
  registerPlugin(heading, 'markdown');

  registerPlugin(table, 'markdown');
  registerPlugin(text, 'markdown');
  registerPlugin(attributes, 'markdown');
  registerPlugin(upload, 'markdown');
  registerPlugin(placeholder, 'markdown');
  registerPlugin(anchors);
  registerPlugin(fullscreen, 'markdown');
  registerPlugin(resizeNav, 'markdown');
  registerPlugin(maxHeight, 'markdown');
  registerPlugin(mention, 'markdown');

  registerPreset('normal', {
      extend: 'markdown',
      callback: function(addToPreset) {

          addToPreset('emoji', 'normal', {
              'before': 'hard_break'
          });

          addToPreset('mention', 'normal', {
              'before': 'ordered_list'
          });

          addToPreset('oembed', 'normal', {
              'before': 'ordered_list'
          });
      }
  });

  registerPreset('document', {
      extend: 'normal',
      callback: function(addToPreset) {
          addToPreset('anchor', 'document', {
              'before': 'fullscreen'
          });
      }
  });

  registerPreset('full', {
      extend: 'normal'
  });

  var PresetManager = function PresetManager(options) {
      this.map = {};
      this.options = options;
  };

  PresetManager.prototype.add = function add (options, value) {
      this.map[options.preset] = value;
  };

  PresetManager.prototype.create = function create (context) {
      return this.options.create.apply(null, [context]);
  };

  PresetManager.isCustomPluginSet = function isCustomPluginSet (options) {
      return !!options.exclude || !!options.include;
  };

  PresetManager.prototype.check = function check (context) {
      var options = context.options;

      if(this.options.name && context[this.options.name]) {
          return context[this.options.name];
      }

      var result = [];

      if(!PresetManager.isCustomPluginSet(options) && this.map[options.preset]) {
          result = this.map[options.preset];
      }

      if(!result || (Array.isArray(result) && !result.length)) {
          result = this.create(context);

          if(!PresetManager.isCustomPluginSet(options)) {
              this.add(options, result);
          }
      }


      if(this.options.name) {
          context[this.options.name] = result;
      }

      return result;
  };

  var getPlugins = function(context) {
      var options = context.options;

      if(context.plugins) {
          return context.plugins;
      }

      var presetMap = registry.getPresetRegistry(context);

      var toExtend = presetMap.get(options.preset) ?  presetMap.get(options.preset) : registry.plugins;

      if(!PresetManager.isCustomPluginSet(options)) {
          return context.plugins = toExtend.slice();
      }

      var result = [];
      if(options.exclude) {
          toExtend.forEach(function (plugin) {
              if(plugin && !options.exclude.includes(plugin.id)) {
                  result.push(plugin);
              }
          });
      }

      if(options.include) {
          options.include.forEach(function (pluginId) {
              if(registry.plugins[pluginId]) {
                  result.push(plugins[pluginId]);
              } else {
                  console.error('Could not include plugin '+pluginId+' plugin not registered!');
              }
          });
      }
      return context.plugins = result;
  };

  var buildInputRules = function(context) {
      var plugins = context.plugins;
      var schema = context.schema;

      var rules = smartQuotes.concat([ellipsis, emDash]);
      plugins.forEach(function (plugin) {
          if(plugin.inputRules) {
              rules = rules.concat(plugin.inputRules(schema));
          }
      });

      return inputRules({rules: rules})
  };

  var buildPlugins = function(context) {
      var plugins = context.plugins;

      var result = [];
      plugins.forEach(function (plugin) {

          if(plugin.init) {
              plugin.init(context, context.editor.isEdit());
          }

          if(context.editor.isEdit() && plugin.plugins) {
              var pl = plugin.plugins(context);
              if(pl && pl.length) {
                  result = result.concat(pl);
                  context.prosemirrorPlugins[plugin.id] = pl;
              }
          }
      });

      return result;
  };

  var buildPluginKeymap = function(context) {
      var plugins = context.plugins;

      var result = [];
      plugins.forEach(function (plugin) {
          if(plugin.keymap) {
              result.push(keymap$2(plugin.keymap(context)));
          }
      });

      return result;
  };


  // https://github.com/ProseMirror/prosemirror/issues/710
  var isChromeWithSelectionBug = !!navigator.userAgent.match(/Chrome\/(5[89]|6[012])/);

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var presets$3 = new PresetManager({
      name: 'renderer',
      create: function (options) {
          return createRenderer(options);
      }
  });

  var getRenderer = function (context) {
      return presets$3.check(context);
  };

  var createRenderer = function(context) {
      var markdownItOptions = context && context.options.markdownIt || {html: false, breaks: true, linkify: true};
      var renderer = markdownIt(markdownItOptions);

      var plugins = getPlugins(context);
      plugins.forEach(function (plugin) {
          if(plugin.registerMarkdownIt) {
              plugin.registerMarkdownIt(renderer);
          }
      });

      return renderer;
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var mergeSchema = function(schema, plugin, context) {
      if(Array.isArray(plugin)) {
          plugin.forEach(function (newPlugin) {
              schema = mergeSchema(schema, newPlugin, context);
          });
      } else {
          if($.isFunction(schema)) {
              schema = schema(context);
          }
          schema.nodes = Object.assign(schema.nodes || {}, plugin.schema && plugin.schema.nodes || {});
          schema.marks = Object.assign(schema.marks || {}, plugin.schema && plugin.schema.marks || {});
      }

      return schema;
  };

  var presets$2 = new PresetManager({
      name: 'schema',
      create: function (context) {
          return new Schema(mergeSchema({}, context.plugins, context));
      }
  });

  var getSchema = function(context) {
      return presets$2.check(context);
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var presets$1 = new PresetManager({
      name: 'parser',
      create: function (context) {
          return createParser(context);
      }
  });

  var getParser = function (context) {
      return presets$1.check(context);
  };

  var createParser = function (context) {
      var plugins = getPlugins(context);

      var tokens = {};
      plugins.forEach(function (plugin) {
          if (!plugin.schema) {
              return;
          }

          var schemaSpecs = Object.assign({}, plugin.schema.nodes || {}, plugin.schema.marks || {});

          for (var key in schemaSpecs) {
              var spec = schemaSpecs[key];
              if (spec.parseMarkdown) {
                  if(spec.parseMarkdown.block || spec.parseMarkdown.node || spec.parseMarkdown.mark || spec.parseMarkdown.ignore) {
                      tokens[key] = spec.parseMarkdown;
                  } else {
                      var tokenKey = Object.keys(spec.parseMarkdown)[0];
                      tokens[tokenKey] = spec.parseMarkdown[tokenKey];
                  }
              }
          }
      });

      return new MarkdownParser(context.schema || getSchema(context), getRenderer(context), tokens);
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var presets = new PresetManager({
      name: 'serializer',
      create: function (context) {
          return createSerializer(context);
      }
  });

  var getSerializer = function (context) {
      return presets.check(context);
  };

  var createSerializer = function (context) {
      var plugins = getPlugins(context);
      var nodeSpec = {};
      var markSpec = {};
      plugins.forEach(function (plugin) {
          if (!plugin.schema) {
              return;
          }

          var nodes = plugin.schema.nodes || {};

          for (var key in nodes) {
              var node = nodes[key];
              if(node.toMarkdown) {
                  nodeSpec[key] = node.toMarkdown;
              }
          }

          var marks = plugin.schema.marks || {};

          for (var key$1 in marks) {
              var mark = marks[key$1];
              if(mark.toMarkdown) {
                  markSpec[key$1] = mark.toMarkdown;
              } else {
                  markSpec[key$1] = {open: '', close: ''};
              }
          }
      });

      return new HumHubMarkdownSerializer(nodeSpec, markSpec);
  };

  var HumHubMarkdownSerializer = /*@__PURE__*/(function (MarkdownSerializer) {
      function HumHubMarkdownSerializer () {
          MarkdownSerializer.apply(this, arguments);
      }

      if ( MarkdownSerializer ) HumHubMarkdownSerializer.__proto__ = MarkdownSerializer;
      HumHubMarkdownSerializer.prototype = Object.create( MarkdownSerializer && MarkdownSerializer.prototype );
      HumHubMarkdownSerializer.prototype.constructor = HumHubMarkdownSerializer;

      HumHubMarkdownSerializer.prototype.serialize = function serialize (content, options) {
          var state = new HumHubMarkdownSerializerState(this.nodes, this.marks, options);
          state.renderContent(content);
          return state.out
      };

      return HumHubMarkdownSerializer;
  }(MarkdownSerializer));

  var HumHubMarkdownSerializerState = /*@__PURE__*/(function (MarkdownSerializerState) {
      function HumHubMarkdownSerializerState () {
          MarkdownSerializerState.apply(this, arguments);
      }

      if ( MarkdownSerializerState ) HumHubMarkdownSerializerState.__proto__ = MarkdownSerializerState;
      HumHubMarkdownSerializerState.prototype = Object.create( MarkdownSerializerState && MarkdownSerializerState.prototype );
      HumHubMarkdownSerializerState.prototype.constructor = HumHubMarkdownSerializerState;

      HumHubMarkdownSerializerState.prototype.esc = function esc (str, startOfLine) {
          str = str.replace(/[|`*\\~\[\]]/g, "\\$&");
          if (startOfLine) { str = str.replace(/^[:#\-*+]/, "\\$&").replace(/^(\d+)\./, "$1\\."); }
          return str
      };

      return HumHubMarkdownSerializerState;
  }(MarkdownSerializerState));

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var markdown = /*#__PURE__*/Object.freeze({
    __proto__: null,
    getParser: getParser,
    getSerializer: getSerializer,
    getRenderer: getRenderer,
    createLinkExtension: createLinkExtension,
    MarkdownParser: MarkdownParser,
    MarkdownSerializer: MarkdownSerializer,
    MarkdownSerializerState: MarkdownSerializerState,
    defaultMarkdownParser: defaultMarkdownParser,
    defaultMarkdownSerializer: defaultMarkdownSerializer,
    schema: schema$l
  });

  // :: (options: ?Object) → Plugin
  // Create a plugin that, when added to a ProseMirror instance,
  // causes a decoration to show up at the drop position when something
  // is dragged over the editor.
  //
  //   options::- These options are supported:
  //
  //     color:: ?string
  //     The color of the cursor. Defaults to `black`.
  //
  //     width:: ?number
  //     The precise width of the cursor in pixels. Defaults to 1.
  //
  //     class:: ?string
  //     A CSS class name to add to the cursor element.
  function dropCursor(options) {
    if ( options === void 0 ) { options = {}; }

    return new Plugin({
      view: function view(editorView) { return new DropCursorView(editorView, options) }
    })
  }

  var DropCursorView = function DropCursorView(editorView, options) {
    var this$1$1 = this;

    this.editorView = editorView;
    this.width = options.width || 1;
    this.color = options.color || "black";
    this.class = options.class;
    this.cursorPos = null;
    this.element = null;
    this.timeout = null;

    this.handlers = ["dragover", "dragend", "drop", "dragleave"].map(function (name) {
      var handler = function (e) { return this$1$1[name](e); };
      editorView.dom.addEventListener(name, handler);
      return {name: name, handler: handler}
    });
  };

  DropCursorView.prototype.destroy = function destroy () {
      var this$1$1 = this;

    this.handlers.forEach(function (ref) {
        var name = ref.name;
        var handler = ref.handler;

        return this$1$1.editorView.dom.removeEventListener(name, handler);
      });
  };

  DropCursorView.prototype.update = function update (editorView, prevState) {
    if (this.cursorPos != null && prevState.doc != editorView.state.doc) {
      if (this.cursorPos > editorView.state.doc.content.size) { this.setCursor(null); }
      else { this.updateOverlay(); }
    }
  };

  DropCursorView.prototype.setCursor = function setCursor (pos) {
    if (pos == this.cursorPos) { return }
    this.cursorPos = pos;
    if (pos == null) {
      this.element.parentNode.removeChild(this.element);
      this.element = null;
    } else {
      this.updateOverlay();
    }
  };

  DropCursorView.prototype.updateOverlay = function updateOverlay () {
    var $pos = this.editorView.state.doc.resolve(this.cursorPos), rect;
    if (!$pos.parent.inlineContent) {
      var before = $pos.nodeBefore, after = $pos.nodeAfter;
      if (before || after) {
        var nodeRect = this.editorView.nodeDOM(this.cursorPos - (before ?before.nodeSize : 0)).getBoundingClientRect();
        var top = before ? nodeRect.bottom : nodeRect.top;
        if (before && after)
          { top = (top + this.editorView.nodeDOM(this.cursorPos).getBoundingClientRect().top) / 2; }
        rect = {left: nodeRect.left, right: nodeRect.right, top: top - this.width / 2, bottom: top + this.width / 2};
      }
    }
    if (!rect) {
      var coords = this.editorView.coordsAtPos(this.cursorPos);
      rect = {left: coords.left - this.width / 2, right: coords.left + this.width / 2, top: coords.top, bottom: coords.bottom};
    }

    var parent = this.editorView.dom.offsetParent;
    if (!this.element) {
      this.element = parent.appendChild(document.createElement("div"));
      if (this.class) { this.element.className = this.class; }
      this.element.style.cssText = "position: absolute; z-index: 50; pointer-events: none; background-color: " + this.color;
    }
    var parentLeft, parentTop;
    if (!parent || parent == document.body && getComputedStyle(parent).position == "static") {
      parentLeft = -pageXOffset;
      parentTop = -pageYOffset;
    } else {
      var rect$1 = parent.getBoundingClientRect();
      parentLeft = rect$1.left - parent.scrollLeft;
      parentTop = rect$1.top - parent.scrollTop;
    }
    this.element.style.left = (rect.left - parentLeft) + "px";
    this.element.style.top = (rect.top - parentTop) + "px";
    this.element.style.width = (rect.right - rect.left) + "px";
    this.element.style.height = (rect.bottom - rect.top) + "px";
  };

  DropCursorView.prototype.scheduleRemoval = function scheduleRemoval (timeout) {
      var this$1$1 = this;

    clearTimeout(this.timeout);
    this.timeout = setTimeout(function () { return this$1$1.setCursor(null); }, timeout);
  };

  DropCursorView.prototype.dragover = function dragover (event) {
    if (!this.editorView.editable) { return }
    var pos = this.editorView.posAtCoords({left: event.clientX, top: event.clientY});
    if (pos) {
      var target = pos.pos;
      if (this.editorView.dragging && this.editorView.dragging.slice) {
        target = dropPoint(this.editorView.state.doc, target, this.editorView.dragging.slice);
        if (target == null) { return this.setCursor(null) }
      }
      this.setCursor(target);
      this.scheduleRemoval(5000);
    }
  };

  DropCursorView.prototype.dragend = function dragend () {
    this.scheduleRemoval(20);
  };

  DropCursorView.prototype.drop = function drop () {
    this.scheduleRemoval(20);
  };

  DropCursorView.prototype.dragleave = function dragleave (event) {
    if (event.target == this.editorView.dom || !this.editorView.dom.contains(event.relatedTarget))
      { this.setCursor(null); }
  };

  // ::- Gap cursor selections are represented using this class. Its
  // `$anchor` and `$head` properties both point at the cursor position.
  var GapCursor = /*@__PURE__*/(function (Selection) {
    function GapCursor($pos) {
      Selection.call(this, $pos, $pos);
    }

    if ( Selection ) { GapCursor.__proto__ = Selection; }
    GapCursor.prototype = Object.create( Selection && Selection.prototype );
    GapCursor.prototype.constructor = GapCursor;

    GapCursor.prototype.map = function map (doc, mapping) {
      var $pos = doc.resolve(mapping.map(this.head));
      return GapCursor.valid($pos) ? new GapCursor($pos) : Selection.near($pos)
    };

    GapCursor.prototype.content = function content () { return Slice$1.empty };

    GapCursor.prototype.eq = function eq (other) {
      return other instanceof GapCursor && other.head == this.head
    };

    GapCursor.prototype.toJSON = function toJSON () {
      return {type: "gapcursor", pos: this.head}
    };

    GapCursor.fromJSON = function fromJSON (doc, json) {
      if (typeof json.pos != "number") { throw new RangeError("Invalid input for GapCursor.fromJSON") }
      return new GapCursor(doc.resolve(json.pos))
    };

    GapCursor.prototype.getBookmark = function getBookmark () { return new GapBookmark(this.anchor) };

    GapCursor.valid = function valid ($pos) {
      var parent = $pos.parent;
      if (parent.isTextblock || !closedBefore($pos) || !closedAfter($pos)) { return false }
      var override = parent.type.spec.allowGapCursor;
      if (override != null) { return override }
      var deflt = parent.contentMatchAt($pos.index()).defaultType;
      return deflt && deflt.isTextblock
    };

    GapCursor.findFrom = function findFrom ($pos, dir, mustMove) {
      search: for (;;) {
        if (!mustMove && GapCursor.valid($pos)) { return $pos }
        var pos = $pos.pos, next = null;
        // Scan up from this position
        for (var d = $pos.depth;; d--) {
          var parent = $pos.node(d);
          if (dir > 0 ? $pos.indexAfter(d) < parent.childCount : $pos.index(d) > 0) {
            next = parent.child(dir > 0 ? $pos.indexAfter(d) : $pos.index(d) - 1);
            break
          } else if (d == 0) {
            return null
          }
          pos += dir;
          var $cur = $pos.doc.resolve(pos);
          if (GapCursor.valid($cur)) { return $cur }
        }

        // And then down into the next node
        for (;;) {
          var inside = dir > 0 ? next.firstChild : next.lastChild;
          if (!inside) {
            if (next.isAtom && !next.isText && !NodeSelection.isSelectable(next)) {
              $pos = $pos.doc.resolve(pos + next.nodeSize * dir);
              mustMove = false;
              continue search
            }
            break
          }
          next = inside;
          pos += dir;
          var $cur$1 = $pos.doc.resolve(pos);
          if (GapCursor.valid($cur$1)) { return $cur$1 }
        }

        return null
      }
    };

    return GapCursor;
  }(Selection));

  GapCursor.prototype.visible = false;

  Selection.jsonID("gapcursor", GapCursor);

  var GapBookmark = function GapBookmark(pos) {
    this.pos = pos;
  };
  GapBookmark.prototype.map = function map (mapping) {
    return new GapBookmark(mapping.map(this.pos))
  };
  GapBookmark.prototype.resolve = function resolve (doc) {
    var $pos = doc.resolve(this.pos);
    return GapCursor.valid($pos) ? new GapCursor($pos) : Selection.near($pos)
  };

  function closedBefore($pos) {
    for (var d = $pos.depth; d >= 0; d--) {
      var index = $pos.index(d);
      // At the start of this parent, look at next one
      if (index == 0) { continue }
      // See if the node before (or its first ancestor) is closed
      for (var before = $pos.node(d).child(index - 1);; before = before.lastChild) {
        if ((before.childCount == 0 && !before.inlineContent) || before.isAtom || before.type.spec.isolating) { return true }
        if (before.inlineContent) { return false }
      }
    }
    // Hit start of document
    return true
  }

  function closedAfter($pos) {
    for (var d = $pos.depth; d >= 0; d--) {
      var index = $pos.indexAfter(d), parent = $pos.node(d);
      if (index == parent.childCount) { continue }
      for (var after = parent.child(index);; after = after.firstChild) {
        if ((after.childCount == 0 && !after.inlineContent) || after.isAtom || after.type.spec.isolating) { return true }
        if (after.inlineContent) { return false }
      }
    }
    return true
  }

  // :: () → Plugin
  // Create a gap cursor plugin. When enabled, this will capture clicks
  // near and arrow-key-motion past places that don't have a normally
  // selectable position nearby, and create a gap cursor selection for
  // them. The cursor is drawn as an element with class
  // `ProseMirror-gapcursor`. You can either include
  // `style/gapcursor.css` from the package's directory or add your own
  // styles to make it visible.
  var gapCursor = function() {
    return new Plugin({
      props: {
        decorations: drawGapCursor,

        createSelectionBetween: function createSelectionBetween(_view, $anchor, $head) {
          if ($anchor.pos == $head.pos && GapCursor.valid($head)) { return new GapCursor($head) }
        },

        handleClick: handleClick,
        handleKeyDown: handleKeyDown
      }
    })
  };

  var handleKeyDown = keydownHandler({
    "ArrowLeft": arrow("horiz", -1),
    "ArrowRight": arrow("horiz", 1),
    "ArrowUp": arrow("vert", -1),
    "ArrowDown": arrow("vert", 1)
  });

  function arrow(axis, dir) {
    var dirStr = axis == "vert" ? (dir > 0 ? "down" : "up") : (dir > 0 ? "right" : "left");
    return function(state, dispatch, view) {
      var sel = state.selection;
      var $start = dir > 0 ? sel.$to : sel.$from, mustMove = sel.empty;
      if (sel instanceof TextSelection) {
        if (!view.endOfTextblock(dirStr) || $start.depth == 0) { return false }
        mustMove = false;
        $start = state.doc.resolve(dir > 0 ? $start.after() : $start.before());
      }
      var $found = GapCursor.findFrom($start, dir, mustMove);
      if (!$found) { return false }
      if (dispatch) { dispatch(state.tr.setSelection(new GapCursor($found))); }
      return true
    }
  }

  function handleClick(view, pos, event) {
    if (!view.editable) { return false }
    var $pos = view.state.doc.resolve(pos);
    if (!GapCursor.valid($pos)) { return false }
    var ref = view.posAtCoords({left: event.clientX, top: event.clientY});
    var inside = ref.inside;
    if (inside > -1 && NodeSelection.isSelectable(view.state.doc.nodeAt(inside))) { return false }
    view.dispatch(view.state.tr.setSelection(new GapCursor($pos)));
    return true
  }

  function drawGapCursor(state) {
    if (!(state.selection instanceof GapCursor)) { return null }
    var node = document.createElement("div");
    node.className = "ProseMirror-gapcursor";
    return DecorationSet.create(state.doc, [Decoration.widget(state.selection.head, node, {key: "gapcursor"})])
  }

  var mac = typeof navigator != "undefined" ? /Mac/.test(navigator.platform) : false;

  function exitCodeAtLast(state, dispatch) {
      var ref = state.selection;
      var $head = ref.$head;
      var $anchor = ref.$anchor;
      var parent = $head.parent;

      var isBlockQuote = false;
      $anchor.path.forEach(function (item, index) {
          if (!(index % 3) && item.type && item.type.name === 'blockquote') {
              isBlockQuote = true;
          }
      });

      if (!(parent.type.spec.code || isBlockQuote)
          || $anchor.parentOffset != $head.parentOffset
          || !$head.sameParent($anchor)
          || $head.parent.content.size != $head.parentOffset) {

          return false;
      }

      var nodeAfter = state.doc.resolve($head.pos - $head.parentOffset + parent.nodeSize - 1).nodeAfter;
      if (nodeAfter) {
          return false;
      }

      var above = $head.node(-1);
      var after = $head.indexAfter(-1);
      var type = above.contentMatchAt(after).defaultType;

      if (!above.canReplaceWith(after, after, type)) {
          return false;
      }

      if (dispatch) {
          var pos = (!parent.type.spec.code && isBlockQuote) ? $head.after() + 1 : $head.after();
          var tr = state.tr.replaceWith(pos, pos, type.createAndFill());
          tr.setSelection(Selection.near(tr.doc.resolve(pos), 1));
          dispatch(tr.scrollIntoView());
      }

      return true;
  }

  function exitMarkAtLast(state, dispatch) {
      var selection = state.selection;
      if (selection instanceof TextSelection
          && !selection.$head.nodeAfter
          && selection.$head.nodeBefore
          && selection.$head.nodeBefore.isText
          && selection.$head.nodeBefore.marks.length) {

          if (dispatch) {
              selection.$head.nodeBefore.marks.forEach(function (mark) {
                  removeMark(mark.type, state, dispatch);
              });
          }
          return true;
      }

      return false;
  }

  function removeMark(markType, state, dispatch) {
      var ref = state.selection;
      var empty = ref.empty;
      var $cursor = ref.$cursor;
      var ranges = ref.ranges;
      if ((empty && !$cursor) || !markApplies(state.doc, ranges, markType)) { return false; }
      if (dispatch) {
          if ($cursor) {
              if (markType.isInSet(state.storedMarks || $cursor.marks()))
                  { dispatch(state.tr.removeStoredMark(markType)); }
          } else {
              var has = false, tr = state.tr;
              for (var i = 0; !has && i < ranges.length; i++) {
                  var ref$1 = ranges[i];
                  var $from = ref$1.$from;
                  var $to = ref$1.$to;
                  has = state.doc.rangeHasMark($from.pos, $to.pos, markType);
              }
              for (var i$1 = 0; i$1 < ranges.length; i$1++) {
                  var ref$2 = ranges[i$1];
                  var $from$1 = ref$2.$from;
                  var $to$1 = ref$2.$to;
                  if (has) { tr.removeMark($from$1.pos, $to$1.pos, markType); }
              }
              dispatch(tr.scrollIntoView());
          }
      }
      return true
  }

  function markApplies(doc, ranges, type) {
      var loop = function ( i ) {
          var ref = ranges[i];
          var $from = ref.$from;
          var $to = ref.$to;
          var can = $from.depth == 0 ? doc.type.allowsMarkType(type) : false;
          doc.nodesBetween($from.pos, $to.pos, function (node) {
              if (can) { return false }
              can = node.inlineContent && node.type.allowsMarkType(type);
          });
          if (can) { return { v: true } }
      };

      for (var i = 0; i < ranges.length; i++) {
          var returned = loop( i );

          if ( returned ) return returned.v;
      }
      return false
  }

  // :: (Schema, ?Object) → Object
  // Inspect the given schema looking for marks and nodes from the
  // basic schema, and if found, add key bindings related to them.
  // This will add:
  //
  // * **Mod-b** for toggling [strong](#schema-basic.StrongMark)
  // * **Mod-i** for toggling [emphasis](#schema-basic.EmMark)
  // * **Mod-`** for toggling [code font](#schema-basic.CodeMark)
  // * **Ctrl-Shift-0** for making the current textblock a paragraph
  // * **Ctrl-Shift-1** to **Ctrl-Shift-Digit6** for making the current
  //   textblock a heading of the corresponding level
  // * **Ctrl-Shift-Backslash** to make the current textblock a code block
  // * **Ctrl-Shift-8** to wrap the selection in an ordered list
  // * **Ctrl-Shift-9** to wrap the selection in a bullet list
  // * **Ctrl->** to wrap the selection in a block quote
  // * **Enter** to split a non-empty textblock in a list item while at
  //   the same time splitting the list item
  // * **Mod-Enter** to insert a hard break
  // * **Mod-_** to insert a horizontal rule
  // * **Backspace** to undo an input rule
  // * **Alt-ArrowUp** to `joinUp`
  // * **Alt-ArrowDown** to `joinDown`
  // * **Mod-BracketLeft** to `lift`
  // * **Escape** to `selectParentNode`
  //
  // You can suppress or map these bindings by passing a `mapKeys`
  // argument, which maps key names (say `"Mod-B"` to either `false`, to
  // remove the binding, or a new key name string.
  function buildKeymap(context) {
      var keys = {}, type;

      var schema = context.schema;
      var mapKeys = context.options.mapKeys;

      function bind(key, cmd) {
          if (mapKeys) {
              var mapped = mapKeys[key];
              if (mapped === false) { return }
              if (mapped) { key = mapped; }
          }
          keys[key] = cmd;
      }

      bind('ArrowDown', exitCodeAtLast);
      bind('ArrowRight', exitMarkAtLast);

      bind("Mod-z", undo);
      bind("Shift-Mod-z", redo);

      if (!mac) { bind("Mod-y", redo); }

      bind("Alt-ArrowUp", joinUp);
      bind("Alt-ArrowDown", joinDown);
      bind("Mod-BracketLeft", lift$1);
      bind("Escape", selectParentNode);

      if (type = schema.marks.strong)
          { bind("Mod-b", toggleMark(type)); }
      if (type = schema.marks.em)
          { bind("Mod-i", toggleMark(type)); }
      if (type = schema.marks.code)
          { bind("Mod-`", toggleMark(type)); }

      if (type = schema.nodes.bullet_list)
          { bind("Shift-Ctrl-8", wrapInList$1(type)); }
      if (type = schema.nodes.ordered_list)
          { bind("Shift-Ctrl-9", wrapInList$1(type)); }
      if (type = schema.nodes.blockquote)
          { bind("Ctrl->", wrapIn(type)); }
      if (type = schema.nodes.hard_break) {
          var br = type, cmd = chainCommands(exitCode, function (state, dispatch) {
              if(state.selection
                  && state.selection.$anchor.parent
                  && state.selection.$anchor.parent.type === schema.nodes.heading) {
                  splitBlock(state, dispatch);
                  return true;
              }
              dispatch(state.tr.replaceSelectionWith(br.create()).scrollIntoView());
              return true;
          });
          bind("Mod-Enter", cmd);
          bind("Shift-Enter", cmd);
          if (mac) { bind("Ctrl-Enter", cmd); }
      }

      var splitList;

      if (type = schema.nodes.list_item) {
          splitList = splitListItem(type);
          bind("Mod-[", liftListItem(type));
          bind("Mod-]", sinkListItem(type));
      }
      if (type = schema.nodes.paragraph)
          { bind("Shift-Ctrl-0", setBlockType(type)); }
      if (type = schema.nodes.code_block)
          { bind("Shift-Ctrl-\\", setBlockType(type)); }
      if (type = schema.nodes.heading)
          { for (var i = 1; i <= 6; i++) { bind("Shift-Ctrl-" + i, setBlockType(type, {level: i})); } }
      if (type = schema.nodes.horizontal_rule) {
          var hr = type;
          bind("Mod-_", function (state, dispatch) {
              dispatch(state.tr.replaceSelectionWith(hr.create()).scrollIntoView());
              return true
          });
      }

      baseKeymap['Backspace'] = chainCommands(undoInputRule, deleteSelection, joinBackward, selectNodeBackward);

      if(splitList) {
          baseKeymap['Enter'] = chainCommands(splitList, newlineInCode, createParagraphNear, liftEmptyBlock, splitBlock);
      } else {
          baseKeymap['Enter'] = chainCommands(newlineInCode, createParagraphNear, liftEmptyBlock, splitBlock);
      }


      for (var key in baseKeymap) {
          bind(key, baseKeymap[key]);
      }

      return keys
  }

  // !! This module exports helper functions for deriving a set of basic
  // menu items, input rules, or key bindings from a schema. These
  // values need to know about the schema for two reasons—they need
  // access to specific instances of node and mark types, and they need
  // to know which of the node and mark types that they know about are
  // actually present in the schema.
  //
  // The `exampleSetup` plugin ties these together into a plugin that
  // will automatically enable this basic functionality in an editor.

  // :: (Object) → [Plugin]
  // A convenience plugin that bundles together a simple menu with basic
  // key bindings, input rules, and styling for the example schema.
  // Probably only useful for quickly setting up a passable
  // editor—you'll need more control over your settings in most
  // real-world situations.
  //
  //   options::- The following options are recognized:
  //
  //     schema:: Schema
  //     The schema to generate key bindings and menu items for.
  //
  //     mapKeys:: ?Object
  //     Can be used to [adjust](#example-setup.buildKeymap) the key bindings created.
  //
  //     menuBar:: ?bool
  //     Set to false to disable the menu bar.
  //
  //     history:: ?bool
  //     Set to false to disable the history plugin.
  //
  //     floatingMenu:: ?bool
  //     Set to false to make the menu bar non-floating.
  //
  //     menuContent:: [[MenuItem]]
  //     Can be used to override the menu content.
  function setupPlugins(context) {
      var result = buildPluginKeymap(context);

      result = result.concat([
          buildInputRules(context),
          keymap$2(buildKeymap(context)),
          keymap$2(baseKeymap),
          dropCursor(),
          gapCursor(),
          tableEditing(),
          buildMenuBar(context),
          keymap$2({
              "Tab": goToNextCell(1),
              "Shift-Tab": goToNextCell(-1)
          })
      ]);

      if (context.options.history !== false) {
          result.push(history());
      }

      return result.concat(buildPlugins(context));
  }

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  var MentionProvider = function(options) {
      this.event = $({});
      this.options = options;
      if (typeof this.options.minInput === 'undefined') {
          this.options.minInput = 2;
      }
      this.options.minInputText = this.options.minInputText || 'Please type at least '+this.options.minInput+' characters';
  };

  MentionProvider.prototype.query = function(state, node) {
      var this$1$1 = this;

      this.state = state;
      this.$node = $(node);

      if(this.options.minInput > 0 && this.state.query.length < this.options.minInput) {
          this.result = {text: this.options.minInputText};
          this.update();
          return;
      }

      this.loading();
      var queryResult = this.find(this.state.query, node);

      if(queryResult.then) {
          queryResult.then(function (result) {
              this$1$1.updateResult(result);
          });
      } else {
          this.updateResult(queryResult);
      }
  };

  MentionProvider.prototype.loading = function() {
      this.result = {loader: true};
      this.update();
  };

  MentionProvider.prototype.updateResult = function(result) {
      this.result = result;
      this.update();
  };

  MentionProvider.prototype.find = function(query, node) {
      // Abstract method has to be implemented by subclasses
  };

  MentionProvider.prototype.reset = function(query, node) {
      if(this.$container) {
          this.$container.remove();
          this.$container = null;
          this.event.trigger('closed');
      }
  };

  MentionProvider.prototype.prev = function() {
      var $cur = this.$container.find('.cur');
      var $prev = $cur.prev();
      if($prev.length) {
          $prev.addClass('cur');
          $cur.removeClass('cur');
      }
  };

  MentionProvider.prototype.next = function() {
      var $cur = this.$container.find('.cur');
      var $next = $cur.next();
      if($next.length) {
          $next.addClass('cur');
          $cur.removeClass('cur');
      }
  };

  MentionProvider.prototype.select = function() {
      var $cur = this.$container.find('.cur');
      this.state.addMention($cur.data('item'));
      this.reset();
  };

  MentionProvider.prototype.update = function(loading) {
      if(!this.$container) {
          this.$container = $('<div class="atwho-view humhub-richtext-provider">').css({'margin-top': '5px'});
      } else {
          this.$container.empty();
      }

      var position = this.$node.offset();
      this.$container.css({
          top: position.top + this.$node.outerHeight() + 2,
          left: position.left,
      });


      var that = this;
      if(this.result && this.result.length) {
          var $list = $('<ul style="list-style-type: none;padding:0px;margin:0px;">');
          this.result.forEach(function (item) {
              var name =  humhub.modules.util.string.encode(item.name);
              var $li = (item.image) ? $('<li>' + item.image + ' ' + name + '</li>') : $('<li>' + name + '</li>');

              $li.data('item', item).on('click', function () {
                  that.$container.find('.cur').removeClass('cur');
                  $li.addClass('cur');
                  that.select();
              });

              $list.append($li);
          });

          $list.find('li').first().addClass('cur');

          this.$container.append($list);
      } else if(this.result.text) {
          var name =  humhub.modules.util.string.encode(this.result.text);
          this.$container.append($('<span>'+name+'</span>'));
      } else if(this.result.loader) {
          var $loader = humhub.require('ui.loader').set($('<span>'), {
              span: true,
              size: '8px',
              css: {
                  padding: '0px',
                  width: '60px'
              }
          });

          this.$container.append($('<div style="text-align:center;">').append($loader));
      } else {
          this.$container.append($('<span>No Result</span>'));
      }

      $('body').append(this.$container.show());
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2018 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  function uuidv4() {
      return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function(c) {
          var r = Math.random() * 16 | 0, v = c == 'x' ? r : (r & 0x3 | 0x8);
          return v.toString(16);
      });
  }


  var Context = function Context(editor, options) {
      this.event = $({});
      this.uuid = uuidv4();
      this.editor = editor;
      this.editor.context = this;
      this.id = this.editor.$.attr('id');
      this.init(options);

      // This is used to indicate active decorations relevant for some content related assumptions (e.g placeholder plugin)
      this.contentDecorations = [];

      // Map of related prosemirror plugin array by plugin id
      this.prosemirrorPlugins = {};
  };

  Context.prototype.init = function init (options) {
      if(options.pluginOptions) {
          $.extend(options, options.pluginOptions);
      }

      this.options = options;
      this.options.preset = options.preset || 'full';

      if(Array.isArray(options.exclude) && !options.exclude.length) {
          this.options.exclude = undefined;
      }

      if(Array.isArray(options.include) && !options.include.length) {
          this.options.include = undefined;
      }

      getPlugins(this);
      getSchema(this);
  };

  Context.prototype.clear = function clear () {
      this.event.trigger('clear');
  };

  Context.prototype.getPluginOption = function getPluginOption (id, option, defaultValue) {
      var pluginOptions =  this.options[id];

      if(!option) {
          return pluginOptions;
      } else if(pluginOptions) {
          return !(typeof pluginOptions[option] === 'undefined') ? pluginOptions[option] : defaultValue;
      }

      return defaultValue;
  };

  Context.prototype.translate = function translate (key) {
      if(!this.options.translate) {
          return key;
      }

      return this.options.translate(key) || key;
  };

  Context.prototype.getProsemirrorPlugins = function getProsemirrorPlugins (id, prosemirror) {
      return this.prosemirrorPlugins[id];
  };

  Context.prototype.getPlugin = function getPlugin (id, prosemirror) {
      for(var i = 0; i < this.plugins.length; i++) {
          var plugin = this.plugins[i];
          if(plugin.id === id) {
              return plugin;
          }
      }
  };

  Context.prototype.addContentDecoration = function addContentDecoration (id) {
      if(this.contentDecorations.indexOf(id) < 0) {
          this.contentDecorations.push(id);
      }
  };

  Context.prototype.removeContentDecoration = function removeContentDecoration (id) {
      var index = this.contentDecorations.indexOf();
      if(index >= 0) {
          this.contentDecorations.splice(index, 1);
      }
  };

  Context.prototype.hasContentDecorations = function hasContentDecorations () {
      return !!this.contentDecorations.length;
  };

  /*
   * @link https://www.humhub.org/
   * @copyright Copyright (c) 2017 HumHub GmbH & Co. KG
   * @license https://www.humhub.com/licences
   *
   */

  $(document).on('mousedown.richtextProvider', function(evt) {
      if(!$(evt.target).closest('.humhub-richtext-provider:visible').length) {
           $('.humhub-richtext-provider:visible').each(function() {
              var $provider = $(this);

              var provider = $provider.data('provider');
              if(provider && provider.reset) {
                  provider.reset();
              } else {
                  $provider.hide().trigger('hidden');
              }
          });
      }
  });

  var MarkdownEditor = function MarkdownEditor(selector, options) {
      if ( options === void 0 ) options = {};

      this.$ = $(selector);
      this.context = new Context(this, options);
      this.parser = getParser(this.context);
      this.serializer = getSerializer(this.context);
      this.renderer = getRenderer(this.context);

      if(!this.isEdit()) {
          buildPlugins(this.context);
      }
  };

  MarkdownEditor.prototype.isEdit = function isEdit () {
      return this.context.options.edit || this.$.is('.ProsemirrorEditor');
  };

  MarkdownEditor.prototype.clear = function clear () {
      this.view.destroy();
      this.context.clear();
      this.$stage = null;
      this.init();
  };

  MarkdownEditor.prototype.getStage = function getStage () {
      if(!this.$stage) {
          this.$stage = this.$.find('.humhub-ui-richtext');
      }
      return this.$stage;
  };

  MarkdownEditor.prototype.isEmpty = function isEmpty () {
      var doc = this.view.state.doc;
      return doc.childCount === 1 &&
          doc.firstChild.type.name === 'paragraph' &&
          doc.firstChild.content.size === 0 &&
          !this.context.hasContentDecorations()
  };
  MarkdownEditor.prototype.init = function init (md) {
          var this$1$1 = this;
          if ( md === void 0 ) md = "";

      if(this.view) {
          this.view.destroy();
      }

      var editorState = EditorState.create({
          doc: this.parser.parse(md),
          plugins: setupPlugins(this.context)
      });

      var fix = fixTables(editorState);
      editorState = (fix) ? editorState.apply(fix.setMeta("addToHistory", false)) : editorState;

      this.view =  new EditorView(this.$[0], {
          state: editorState
      });

      // TODO: put into menu class...
      if(this.$.is('.focusMenu')) {
          this.$menuBar = this.$.find('.ProseMirror-menubar').hide();

          this.$editor = $(this.view.dom).on('focus', function () {
              this$1$1.$menuBar.show();
          }).on('blur', function (e) {
              if(!this$1$1.$.is('.fullscreen')) {
                  this$1$1.$menuBar.hide();
              }
          });
      }

      this.$editor = $(this.view.dom);

      // Dirty workaround, force inline menus to be removed, this is required e.g. if the editor is removed from dom
      $('.humhub-richtext-inline-menu').remove();
      this.trigger('init');
  };
      
  MarkdownEditor.prototype.serialize = function serialize () {
      this.trigger('serialize');
      return this.serializer.serialize(this.view.state.doc);
  };

  MarkdownEditor.prototype.trigger = function trigger (trigger$1, args) {
      this.context.event.trigger(trigger$1, args);
      this.$.trigger(trigger$1, args);
  };

  MarkdownEditor.prototype.on = function on (event, handler) {
      this.$.on(event, handler);
  };

  MarkdownEditor.prototype.render = function render () {
      return this.renderer.render(this.$.text(), this);
  };

  window.prosemirror = {
      MarkdownEditor: MarkdownEditor,
      state: state,
      view: view,
      transform: transform,
      inputRules: inputRules$1,
      model: model,
      commands: commands,
      history: history$1,
      keymap: keymap$3,
      menu: menu$j,
      loader: loader$1,
      pmmenu: pmmenu,
      prompt: prompt,
      getRenderer: getRenderer,
      plugin: {
          registerPreset: registerPreset,
          registerPlugin: registerPlugin,
          markdown: markdown
      },
      $node: $node,
      MentionProvider: MentionProvider
  };

}());
